/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
 /**
 * \~english
 * @file sphSphere.h
 * Sphere Class header File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphSphere.h
 * Sphereクラスヘッダーファイル
 * @author     hira
 * @date       2009/10/01
 */
#ifndef _SPHSPHERE_H_
#define _SPHSPHERE_H_

#include "base/sphBase.h"
#include "skl_config/SklSolverConfig.h"
/*
#include "solverbase/sphSolverManager.h"
#include "solverbase/sphCouplerManager.h"
#include "parallel/sphParaManager.h"
#include "config/sphElem.h"
#include "solverbase/sphFlowBase.h"
*/

class sphSolverManager;
class sphCouplerManager;
class sphParaManager;
class sphElem;
class sphFlowBase;
class sphSolverBase;
class sphCouplerBase;
//class SklCfg::SklSolverConfig;

/**
 * Sphereクラス.
 * すべてのソルバ、カプラ、プロセスの管理クラス
 */
class sphSphere: public sphBase
{
private:

  /**
   * プロセスグループ情報クラス.
   * プロセス番号リスト、コミュニケータ、グループ、ソルバID/カプラIDの情報を持つ。
   */
  class ProcessGroup {
  public:
    int m_id;                   ///< solverID, couplerID
    int m_nodesize;        /// nodeId size
    int* m_nodelist;        /// nodeId list
    SPL_Comm m_comm;            /// communicator
    SPL_Group m_group;          /// group

    /// コンストラクタ
    ProcessGroup() {
      m_id = 0;
      m_nodesize = 0;
      m_nodelist = NULL;
      m_comm = SPL_COMM_NULL;
      m_group = SPL_GROUP_NULL;
    }

    /// コンストラクタ
    ProcessGroup(int id, int size, int* nodelist) {
      m_id = id;
      m_nodesize = size;
      m_nodelist = nodelist;
      m_comm = SPL_COMM_NULL;
      m_group = SPL_GROUP_NULL;
    }

    /// デストラクタ
    virtual ~ProcessGroup() {
      if (m_nodelist != NULL) delete [] m_nodelist;
    }

    /// コミュニケータ、グループを設定する。
    void setCommGroup(SPL_Comm comm, SPL_Group group) {
      m_comm = comm; m_group = group;
    }

    /// プロセスグループに属するプロセス番号であるかチェックする。
    bool isProcessGroupId(int global_procId) const {
      if (m_nodesize == 0 || m_nodelist == NULL ) {
        if (m_group == SPL_GROUP_NULL) return false;
        /// MPI_GROUP_RANKによりチェックする。
        int rank = SPL_UNDEFINED;
        SPL_Group_rank(m_group, &rank);
        if (rank == SPL_UNDEFINED) return false;
        return true;
      }
      /// プロセスID検索
      register int i;
      for( i=0; i<m_nodesize; i++ ) {
        if (m_nodelist[i] == global_procId) return true;
      }
      return false;
    }

  };    /* ProcessGroup */

  static sphSphere* m_instance;
  sphSphere();
  virtual ~sphSphere();
  const sphSphere::ProcessGroup* getProcessGroupOfSolver(int solverID) const;
  const sphSphere::ProcessGroup* getProcessGroupOfCoupler(int couplerID) const;
  int* mergeIdList(int fst_idcnt, const int* fst_idlist,
                   int snd_idcnt, const int* snd_idlist,
                   int& mrg_idcnt);

protected:

  // プロセスグループ情報
  /**
   * ソルバのプロセスグループリスト<ソルバID, プロセスグループ情報>
   */
  std::map<int, ProcessGroup* > m_solverGroup;

  /**
   * カプラのプロセスグループリスト<カプラID, プロセスグループ情報>
   */
  std::map<int, ProcessGroup* > m_couplerGroup;

  // 管理クラス
  sphSolverManager* m_solverMngr;       ///< ソルバ管理クラス
  sphCouplerManager* m_couplerMngr;     ///< カプラ管理クラス
  sphParaManager* m_paraMngr;           ///< パラレルコントローラ管理クラス
  const sphElem* m_configElem;          ///< コンフィグレーションルート要素
  sphFlowBase* m_flow;                  ///< フローコントローラー
  const SklCfg::SklSolverConfig* m_sklconfig;          ///< Vsphereコンフィグレーション

  bool createProcessSolverGroup();
  bool createProcessCouplerGroup();
  bool createSolver();
  bool createSklSolver();
  bool createCoupler();
  bool createFlow();
  bool createSklFlow();

public:
  static sphSphere* getInstance();
  static void destruct();

  const sphSolverManager* getSolverManager() const;
  sphSolverManager* getSolverManager();
  const sphCouplerManager* getCouplerManager() const;
  sphCouplerManager* getCouplerManager();
  const sphParaManager* getParaManager() const;
  sphParaManager* getParaManager();
  const sphFlowBase* getFlow() const;
  sphFlowBase* getFlow();
  const sphElem* getConfigElem() const;
  const sphElem* getConfigElem();
  const SklCfg::SklSolverConfig* getSklSolverConfig();

  bool addSolver( sphSolverBase* solver,
                  sphSphere* sphObj,
                  const sphElem* config_elem,
                  SPL_Comm comm, SPL_Group group);
  bool addSolver(sphSolverBase* solver,
                  sphSphere* sphObj,
                  const SklCfg::SklSolverConfig* sklconfig,
                  SPL_Comm comm, SPL_Group group);

  bool addCoupler(sphCouplerBase* coupler,
                  sphSphere* sphObj,
                  const sphElem* config_elem,
                  SPL_Comm comm, SPL_Group group);
  void setFlow(sphFlowBase* flow);
  void setConfigElem(const sphElem* root_elem);
  void setSklConfig(const SklCfg::SklSolverConfig* skl_config);

  bool isClassicConfig();
  bool isSphereConfig();

  int executeSolver();

};

#endif /* _SPHSPHERE_H_ */
