/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphErrorHandler.h
 * Error Handler Class Header File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphErrorHandler.h
 * エラーハンドラークラスヘッダーファイル
 * @author     hira
 * @date       2009/10/01
 */

#ifndef _SPHERRORHANDLER_H_
#define _SPHERRORHANDLER_H_

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <deque>
#include <iostream>
#include <map>
#include <sstream>

#include "base/sphMessageDefine.h"
#include "spl/spl.h"
#include "spl/spl_errno.h"
#include "spl/splmpi.h"

/**
 * \~english
 * message type
 *
 * \~japanese
 * メッセージタイプ.
 * MSGTYPE_UNKNOWN:不明
 * MSGTYPE_WARNING:警告メッセージ
 * MSGTYPE_ERROR:エラーメッセージ
 */
enum enum_msgtype {MSGTYPE_UNKNOWN, MSGTYPE_WARNING, MSGTYPE_ERROR};

//#define SPH_ERROR_INFO setFileInfo(__FILE__, __LINE__, __func__),sphErrorHandler::getInstance()->error_fmt;

/**
 * \~english
 * Error Message Class
 *
 * \~japanese
 * エラーメッセージクラス
 *
 * エラー番号、エラーメッセージ、詳細情報、エラー発生ファイル名、行番号、関数名の情報を持つ。
 */
class sphMsg
{
private:

public:
  sphMsg() : m_type(MSGTYPE_UNKNOWN), m_msgNo(0), m_lineno(0) { return; };
  virtual ~sphMsg() {};

  enum_msgtype m_type;          ///< \~english message type  \~japanese メッセージタイプ
  int m_msgNo;                  ///< \~english message number  \~japanese メッセージ番号
  std::string m_message;        ///< \~english message string  \~japanese メッセージ
  std::string m_info;            ///< \~english detail information  \~japanese 詳細情報
  std::string m_filename;        ///< \~english occurred file name   \~japanese エラー発生ファイル名
  int m_lineno;                  ///< \~english occurred file line no   \~japanese エラー発生ファイル行番号
  std::string m_funcname;        ///< \~english occurred function name   \~japanese エラー発生関数名

  /**
   * \~english
   * set error message information
   * @param      msg_no          message number
   * @param      filename        file name
   * @param      lineno          file line number
   * @param      funcname        function name
   * @param      info            detail information
   * @return    void
   *
   * \~japanese
   * エラーメッセージ情報を設定する。
   * @param      msg_no          メッセージ番号
   * @param      filename        ファイル名
   * @param      lineno          ファイル行番号
   * @param      funcname        関数名
   * @param      info            詳細情報
   * @return    void
   */
  void setError(int msg_no, const char* filename = NULL, int lineno = 0, const char* funcname = NULL, const char* info = NULL)
  {
    m_type = MSGTYPE_ERROR;
    m_msgNo = msg_no;
    if (filename) m_filename = filename;
    m_lineno = lineno;
    if (funcname) m_funcname = funcname;
    if (info) {
      m_info = info;
    }
  };


  /**
   * \~english
   * set warning message information
   * @param      msg_no          message number
   * @param      filename        file name
   * @param      lineno          file line number
   * @param      funcname        function name
   * @param      info            detail information
   * @return    void
   *
   * \~japanese
   * 警告メッセージ情報を設定する。
   * @param      msg_no          メッセージ番号
   * @param      filename        ファイル名
   * @param      lineno          ファイル行番号
   * @param      funcname        関数名
   * @param      info            詳細情報
   * @return    void
   */
  void setWarning(int msg_no, const char* filename = NULL, int lineno = 0, const char* funcname = NULL, const char* info = NULL)
  {
    m_type = MSGTYPE_WARNING;
    m_msgNo = msg_no;
    if (filename) m_filename = filename;
    m_lineno = lineno;
    if (funcname) m_funcname = funcname;
    if (info) m_info = info;
  };

  /**
   * \~english
   * set message
   * @param      msg          message
   * @return    void
   *
   * \~japanese
   * 警告メッセージを設定する。
   * @param      msg          メッセージ
   * @return    void
   */
  void setMessage(std::string msg)
  {
    m_message = msg;
  }


  /**
   * \~english
   * set detail information
   * @param      format          detail information format
   * @param      ...            detail information parameter
   * @return    void
   *
   * \~japanese
   * 詳細情報を設定する。
   * @param      format          詳細情報フォーマット
   * @param      ...            詳細情報パラメータ
   * @return    void
   */
  void setInfo(const char* format, ...)
  {
    va_list ap;
    va_start(ap, format);
    char buf[1024];
    vsprintf(buf, format, ap);                  // vasprintf to vsprintf at 2010/05/10 by @hira
    if (buf) m_info = std::string(buf);
    va_end(ap);
  };


  /**
   * \~english
   * get error information string
   * @param      void
   * @return    void
   *
   * \~japanese
   * エラー情報文字列を組み立てる.
   * 書式
   * - Sph_Error[エラー番号]:エラーメッセージ detail=詳細情報 location=ファイル名[行番号]::関数名
   * @param      void
   * @return    void
   */
  std::string getPrintMessage() const
  {
    std::string msg;

    int myID = 0;
    SPL_Comm_rank(SPL_COMM_WORLD, &myID);
    if (myID >= 0) {
      std::ostringstream os;
      os << myID;
      msg += "[rankid=" + os.str() + "]:";
    }

    if (m_type == MSGTYPE_ERROR) {
      msg += "Sph_Error";
    }
    else if (m_type == MSGTYPE_WARNING) {
      msg += "Sph_Warning";
    }
    else {
      msg += "Unknown";
    }
    if (m_msgNo != 0) {
      std::ostringstream os;
      os << m_msgNo;
      msg += "[" + os.str() + "]:";
    }
    if (!m_message.empty()) {
      msg += m_message;
    }
    if (!m_info.empty()) {
      msg += " detail=" + m_info;
    }

    if (!m_filename.empty()) {
      msg += " location=" + m_filename;
    }
    if (m_lineno != 0) {
      std::ostringstream os;
      os << m_lineno;
      msg += "[" + os.str() + "]";
    }
    if (!m_funcname.empty()) {
      msg += "::" + m_funcname;
    }
    return msg;
  }
};



/**
 * \~english
 * Error Handler Class
 *
 * \~japanese
 * エラーハンドラークラス
 *
 * エラーメッセージのスタックを行う。
 */
class sphErrorHandler
{
private:
  /// \~english last message  \~japanese 最終メッセージ
  sphMsg* m_lastErr;
  /// \~english error message stack  \~japanese エラーメッセージスタック
  std::deque<sphMsg*> m_errStack;
  /// \~english message number list  \~japanese メッセージ番号リスト
  std::map<int, std::string> m_msgNumList;
  /// sphErrorHandler instance
  static sphErrorHandler* m_instance;
  /// 警告メッセージのブロックフラグ（true:警告メッセージは無視する。/false:警告メッセージを受け取る。）:SPH_ERRORLEVEL_ERRORの設定
  bool m_sphBlockWarning;
  /// 警告メッセージのブロックフラグ（true:警告メッセージは無視する。/false:警告メッセージを受け取る。）:一時警告メッセージブロック設定
  bool m_tmpBlockWarning;
  /// 警告メッセージのブロックカウント：blockWarning()の入れ子対策
  int m_blockCount;

  const char* _error_filename;
  int _error_lineno;
  const char* _error_funcname;

  sphErrorHandler();
  sphErrorHandler(sphErrorHandler& err);
  void operator = (sphErrorHandler&);
  virtual ~sphErrorHandler();

protected:
  void removeStackEnd();
  std::string getMassage(int err_no) const;

public:
  static sphErrorHandler* getInstance();
  static void destruct();
  void error(sphMsg* err_msg);
  void error(int msg_no, const char* filename = NULL, int lineno = 0, const char* funcname = NULL);
  void error(int msg_no, const char* filename, int lineno, const char* funcname, const char* format, ...);
  void error_info(int msg_no, const char* info);
  void warning(sphMsg* warning_msg);
  void warning(int msg_no, const char* filename = NULL, int lineno = 0, const char* funcname = NULL);
  void warning(int msg_no, const char* filename, int lineno, const char* funcname, const char* format, ...);
  int getLastErrorNo() const;
  bool exists() const;
  bool existsError() const;
  bool existsWarning() const;
  void printStackTrace() const;
  void clear();
  void setInfo(const char* format, ...);

  // 警告メッセージのブロックフラグの設定・解除メソッド
  void blockWarning();
  void releaseWarning();
  bool getDefaultBlockWarning();

  // add 2010/05/10 by @hira for RSCC
  void setFileInfo(const char* filename, int lineno, const char* funcname);
  void error_fmt(int msg_no, const char* format, ...);
  void warning_fmt(int msg_no, const char* format, ...);


  /**
   * \~english
   * Macro:set error message number
   *
   * \~japanese
   * エラーメッセージ番号設定マクロ.
   * ファイル名、行番号、関数名を自動付加する。
   */
  #define SPH_ERROR(msg_no) error(msg_no, __FILE__, __LINE__, __func__)

  /**
   * \~english
   * Macro:set error message number and detail information
   *
   * \~japanese
   * エラーメッセージ番号と詳細情報設定マクロ.
   * ファイル名、行番号、関数名を自動付加する。
   */

//#ifdef __VA_ARGS__
//  #define SPH_ERROR_INFO(msg_no, ...) \
//            error(msg_no, __FILE__, __LINE__, __func__, __VA_ARGS__)
//#else
  #define SPH_ERROR_INFO setFileInfo(__FILE__, __LINE__, __func__);(sphErrorHandler::getInstance()->*(&sphErrorHandler::error_fmt))
//#endif

  /**
   * \~english
   * Macro:set warning message number
   *
   * \~japanese
   * 警告メッセージ番号設定マクロ.
   * ファイル名、行番号、関数名を自動付加する。
   */
  #define SPH_WARNING(msg_no) warning(msg_no, __FILE__, __LINE__, __func__)

  /**
   * \~english
   * Macro:set warning message number and detail information
   *
   * \~japanese
   * 警告メッセージ番号と詳細情報設定マクロ.
   * ファイル名、行番号、関数名を自動付加する。
   */

//#ifdef __VA_ARGS__
//  #define SPH_WARNING_INFO(msg_no, ...) \
//            warning(msg_no, __FILE__, __LINE__, __func__, __VA_ARGS__)
//#else
  #define SPH_WARNING_INFO setFileInfo(__FILE__, __LINE__, __func__);(sphErrorHandler::getInstance()->*(&sphErrorHandler::warning_fmt))
//#endif

};

#endif /* _SPHERRORHANDLER_H_ */
