/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphStepTime.h
 * Sphere Step/Time Class Header File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphStepTime.h
 * Sphereステップ・時刻クラスヘッダーファイル
 * @author     hira
 * @date       2009/10/01
 */

#ifndef _SPHSTEPTIME_H_
#define _SPHSTEPTIME_H_

#define SPH_TIME_REAL           float
//#define SPH_TIME_REAL           double

#include "base/sphBase.h"

class sphStepTime: public sphBase
{

protected:
  size_t m_maxStep;                     ///< ソルバループステップ数
  size_t m_baseStep;                    ///< ソルバ基準ステップ
  size_t m_currentStep;                    ///< ソルバ実行ステップ
  SPH_TIME_REAL m_currentTime;                 ///< ソルバ実行現在時刻
  SPH_TIME_REAL m_baseTime;                 ///< ソルバ基準時刻
  SPH_TIME_REAL m_deltaTime;                ///< ソルバ１ステップ実行時刻

public:
  /**
   * コンストラクタ
   */
  sphStepTime()
  {
    m_maxStep = m_currentStep = m_baseStep = 0;
    m_currentTime = m_baseTime = m_deltaTime = 0.0;
  };

  /**
   * デストラクタ.
   */
  virtual ~sphStepTime()
  {
  };

  /**
   * 実行中のループステップ数を取得する。
   * @param      void
   * @return    ステップ数
   */
  size_t getCurrentStep() const
  {
    return m_currentStep;
  }

  /**
   * 実行中のループステップ数を設定する。
   * @param      void
   * @return    ステップ数
   */
  void setCurrentStep(size_t currentStep)
  {
    m_currentStep = currentStep;
  }

  /**
   * 基準ステップ数を取得する。
   * @param      void
   * @return    基準ステップ数
   */
  size_t getBaseStep() const
  {
    return m_baseStep;
  }

  /**
   * 基準ステップからの実行ステップ数を取得する。
   * @param      void
   * @return    実行ステップ数
   */
  size_t getTotalStep() const
  {
    return ( m_currentStep + m_baseStep );
  }

  /**
   * 基準ステップ数を設定する。
   * @param      baseStep                 baseStep
   * @return    success(=true)/fault(=false)
   */
  bool setBaseStep(size_t baseStep)
  {
    m_baseStep = baseStep;
    return true;
  }

  /**
   * 現在時刻を設定する。
   * @param      currentTime      現在時刻
   * @return    success(=true)/fault(=false)
   */
  bool setCurrentTime(SPH_TIME_REAL currentTime)
  {
    m_currentTime = currentTime;
    return true;
  }

  /**
   * 基準時刻を設定する。
   * @param      baseTime      基準時刻
   * @return    success(=true)/fault(=false)
   */
  bool setBaseTime(SPH_TIME_REAL baseTime)
  {
    m_baseTime = baseTime;
    return true;
  }

  /**
   * 1ステップ時刻を設定する。
   * @param      deltaTime      1ステップ時刻
   * @return    success(=true)/fault(=false)
   */
  bool setDeltaTime(SPH_TIME_REAL deltaTime)
  {
    m_deltaTime = deltaTime;
    return true;
  }

  /**
   * 現在時刻を設定する。
   * @param      void
   * @return    現在時刻
   */
  SPH_TIME_REAL getCurrentTime() const
  {
    return m_currentTime;
  }

  /**
   * 基準時刻を取得する。
   * @param      void
   * @return    基準時刻
   */
  SPH_TIME_REAL getBaseTime() const
  {
    return m_baseTime;
  }

  /**
   * 1ステップ時刻を取得する。
   * @param      void
   * @return    1ステップ時刻
   */
  SPH_TIME_REAL getDeltaTime() const
  {
    return m_deltaTime;
  }

  /**
   * 基準時刻からの現在時刻を取得する。
   * @param      void
   * @return    現在時刻
   */
  SPH_TIME_REAL getTotalTime() const
  {
    return ( m_currentTime + m_baseTime );
  }

  /**
   * 現在時刻に1ステップ時刻を加算する。
   * @param      void
   * @return    success(=true)/fault(=false)
   */
  bool incrementTime()
  {
    if( m_deltaTime == 0.0 ) return false;
    m_currentTime += m_deltaTime;
    return true;
  }

  /**
   * 実行最大ステップ数を取得する.
   * @param               void
   * @return      最大ステップ数
   */
  size_t getMaxStep() const
  {
    return m_maxStep;
  }

  /**
   * 実行最大ステップ数を設定する.
   * @param        max_step       最大ステップ数
   * @return    success(=true)/fault(=false)
   */
  bool setMaxStep(size_t max_step)
  {
    m_maxStep = max_step;
    return true;
  }


  /**
   * メンバの現在ステップ数からインターバルチェックを行う.
   * @param        interval       インターバルステップ数
   * @return       true:インターバルステップである/false:インターバルステップではない
   */
  bool chkInterval(size_t interval) const
  {
    if (interval == 0) return true;
    return ( m_currentStep == (m_currentStep / interval * interval) ) ? true : false;
  }

  /**
   * 引数の現在ステップ数からインターバルチェックを行う.
   * @param        currentStep       現在ステップ数
   * @param        interval       インターバルステップ数
   * @return       true:インターバルステップである/false:インターバルステップではない
   */
  static bool chkInterval(size_t currentStep, size_t interval)
  {
    if (interval == 0) return true;
    return ( currentStep == (currentStep / interval * interval) ) ? true : false;
  }

};

#endif /* _SPHSTEPTIME_H_ */
