/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphElem.h
 * Configuration XML Element Class Header File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphElem.h
 * コンフィギュレーションXML要素クラスヘッダーファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPH_ELEM_H_
#define _SPH_ELEM_H_

#include "config/sphCfgBase.h"
#include "data/sphDataDefs.h"

#define sphString std::string

struct __lesscasecmp
{
   bool operator() (const std::string& a, const std::string& b) const
   {
      return (strcasecmp (a.c_str(), b.c_str()) < 0);
   }
};


/**
 * \~english
 * Configuration XML Element Class
 *
 * \~japanese
 * コンフィギュレーションXML要素クラス
 */
class sphElem : public sphCfgBase
{
  friend class sphConfigController;
  friend class sphElemUtils;
  friend class sphSolverSv;

private:
  void removeSubTreeElem();
  char * normalizePath(const char *path, char *normalized_path);
  std::string getCanonicalizePath(xmlNodePtr node);
  bool isSearchMatch(const sphElem* searchElem) const;
  bool isSphCfgFileElem() const;
  const sphElem* createSearchElem(const std::vector<std::string>* arg_list) const;
  static std::string toLower(std::string value);

protected:
  // member variables
  std::string m_elemName;                         /// element name
  std::map<std::string, std::string, __lesscasecmp>* m_attrList;  /// attribute list
//  std::deque<sphElem*> m_parentElemList;          /// list of pointer to parent element
  std::deque<sphElem*>* m_childElemList;           /// list of pointer to child element
  std::vector<sphElem*>* m_subTreeList;         /// list of xml-sub-tree, <外部XML要素=<sph_xml_config_file>要素の子要素>
  std::string m_fileName;                         /// file name
  const sphElem* m_parentElem;                  /// parent element

  // XML parse methods
  bool parseElem(xmlNodePtr node);
  bool createElem(xmlDocPtr xmlDoc, bool one_file = false) const;
  bool createElem(xmlNodePtr node, bool one_file = false) const;

  // setting methods
  bool setElemName(std::string name);
  bool addAttribute(std::string name, int value);
  bool addAttribute(std::string name, long value);
  bool addAttribute(std::string name, long long value);
  bool addSubTreeList(sphElem* elem);

  // search methods
  const sphElem* getSerialElemBySphElem(const sphElem* searchElem) const;
  int countNumOfSerialElemBySphElem(const sphElem* searchElem) const;
  const sphElem* getWildElemFirst(const char* elem_name) const;
  const sphElem* getWildElemNext(sphElem* elem, const char* elem_name) const;
  const sphElem* getExceptElemBySphElem(const sphElem* searchElem, const sphElem* prevElem = NULL) const;
  int countNumOfExceptElemBySphElem(const sphElem* searchElem) const;

  // write methods
  bool writeXmlFileInCfgFile(const char* base_fname) const;

  // setting methods
  void setSubTreeList(std::vector<sphElem*>* subTreeList);

  // getting methods
  const sphElem* getElement(std::string elem_name = "") const;
  std::string getAttribute(std::string name) const;
  const char* getAttrChar(std::string name) const;
  long long   getAttrInteger(std::string name) const;
  double      getAttrReal (std::string name) const;

  // search methods
  const sphElem* getElemBySerial(const char* elem_name) const;
  const sphElem* getElemBySerial(unsigned int num, const char* elem_name00, ...) const;
  const sphElem* getChildElemFirstBySerial(const char* elem_name = NULL) const;
  const sphElem* getChildElemNextBySerial(const sphElem* elem, const char* elem_name = NULL) const;
  int countNumOfElemBySerial(const char* elem_name = NULL) const;
  int countNumOfElemBySerial(unsigned int num, const char* elem_name00, ...) const;
  bool initialize();

public:
  sphElem();
  sphElem(const char* elem_name);
  sphElem(const sphElem* rootElem, std::vector<sphElem*>* subTreeList);
  sphElem(const sphElem* elem);
  virtual ~sphElem();

  // getting methods
  const char*    getElemName() const;
  const char*    getAttrString(const char* name) const;
  short          getAttrShort (const char* name) const;
  int            getAttrInt   (const char* name) const;
  long           getAttrLong  (const char* name) const;
  long long      getAttrLLong (const char* name) const;
  float          getAttrReal4 (const char* name) const;
  double         getAttrReal8 (const char* name) const;
  SphDCType      getAttrSphDcType (const char* name) const;
  SPL_Datatype   getAttrSplDataType (const char* name) const;
  SphCrdDef      getAttrSphCrdDef(const char* name) const;
  bool           isAttrValue  (const char* name, bool default_value = true) const;
  bool           getAttrFilePath(char* file_path, const char* name = NULL) const;
  bool        existAttribute(const char* name) const;
  bool equalsElemName(std::string elem_name) const;
  bool existChileElement(const char* name) const;

  // search methods
  const sphElem* getElem(const char* elem_name) const;
  const sphElem* getElem(unsigned int num, const char* elem_name00, ...) const;
  const sphElem* getChildElemFirst(const char* elem_name = NULL) const;
  const sphElem* getChildElemFirst(unsigned int num, const char* elem_name00, ...) const;
  const sphElem* getChildElemNext(const sphElem* elem, const char* elem_name = NULL) const;
  const sphElem* getChildElemNext(const sphElem* elem, unsigned int num, const char* elem_name00, ...) const;
  int countNumOfElem(const char* elem_name = NULL) const;
  int countNumOfElem(unsigned int num, const char* elem_name00, ...) const;

  const sphElem* getElemListById(
                        const char* list_name,
                        const char* sub_name,
                        int find_id) const;

  const sphElem* getElemListByLabel(
                        const char* list_name,
                        const char* sub_name,
                        const char* find_label) const;

  const sphElem* getChildElemByAttr(
                        const char* elem_name,
                        const char* attr_name,
                        const char* attr_value) const;

  // write methods
  bool writeXmlFile(const char* fname, bool one_file = false) const;

  // get root element
  const sphElem* getRootElem() const;
  const sphElem* getParentElem(const char* elem_name = NULL) const;
  std::string getUpstreamFileName() const;

  // for create
  bool addAttribute(std::string name, std::string value);
  bool addChildElem(sphElem* child);
  void setAttribute(std::string name, std::string value);

  // for debug
  std::string debugPrintElem() const;
  void debugPrint() const;
};

#endif // _SPH_ELEM_H_

