/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphElemUtils.h
 * Sphere Configuration Element Utilities Class Header File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphElemUtils.h
 * Sphereコンフィギュレーション要素ユーティリティクラスヘッダーファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPHELEMUTILS_H_
#define _SPHELEMUTILS_H_

#include "base/sphBase.h"
#include "config/sphElem.h"


/**
 * \~english
 * Sphere Configuration Element Utilities Class
 *
 * \~japanese
 * Sphereコンフィギュレーション要素Utilitiesクラス
 */
class sphElemUtils : public sphBase
{
protected:
  const sphElem* m_referElem;

  sphElemUtils();


public:
  sphElemUtils(const sphElem* elem);
  virtual ~sphElemUtils();

  int getSolverCount() const;
  int getCouplerCount() const;
  const sphElem* getSolverElem(int index) const;
  const sphElem* getCouplerElem(int index) const;
  const sphElem* getElemBySolverId(int solverId) const;
  const sphElem* getElemByCouplerId(int couplerId) const;
  const sphElem* getElemBySolverLabel(const char* solverLabel) const;
  const sphElem* getElemByCouplerLabel(const char* couplerLabel) const;
  int getNodeCountBySolverId(int solverId) const;
  int getNodeCountByCouplerId(int couplerId) const;
  int getNodeCountFromSteer(const sphElem* solv_elem) const;
  const sphElem* getElemByFileLabel(const char* fileLabel) const;
  const sphElem* getFileElemByFileLabel(const char* fileLabel) const;
  const sphElem* getFileElemByFileId(int fileId) const;
  const sphElem* getFileElem() const;
  bool getDomainInfo(     int dims,
                          size_t size[3],
                          double origin[3],
                          double pitch[3],
                          size_t start[3] = NULL) const;
  const char* getInputBaseDir() const;
  const char* getOutputBaseDir() const;
  bool getFileFormatPath( char* file_path,
                          const char* basedir,
                          int steps, int rankId,
                          bool sub_flag = false);
  bool getFilePath(       char* file_path,
                          const char* basedir) const;
  SphCfgFileIOMode getFileIoMode() const;
  bool isFileInterval(size_t current_step) const;
  SphCfgAssignType getStartIndexAssignType() const;


  /**
   * この要素、又は子要素から属性を取得する(long long)
   * @param[in]     dims      次元数
   * @param[in]     elem_name      検索要素名
   * @param[in]     attr_names      検索属性名リスト（*attr_names[dims])
   * @param[out]     values         属性値リスト（values[dims])
   * @return    success(=true)/fault(=false)
   */
  template <class X> bool getDimsAttributeValues(
                              int dims,
                              const char* elem_name,
                              const char* attr_names[],
                              X* values) const
  {
    if (dims == 0 || elem_name == NULL || attr_names == NULL || values == NULL || m_referElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getDimsAttributeValues arguments");
      return false;
    }

    const sphElem* elem = NULL;
    if (m_referElem->equalsElemName(elem_name)) {
      elem = m_referElem;
    }
    else {
      elem = m_referElem->getElem(elem_name);
    }
    if (elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found %s", elem_name);
      return false;
    }

    register int i;
    for (i=0; i<dims; i++) {
      if (elem->existAttribute(attr_names[i])) {
        values[i] = elem->getAttrLLong(attr_names[i]);
      }
      else {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ATTRIBUTE, "not found %s[@%s], %s", elem_name, attr_names[i], elem->debugPrintElem().c_str());
        return false;
      }
    }
    return true;
  }

  /**
   * この要素、又は子要素から属性を取得する(double)
   * @param[in]     dims      次元数
   * @param[in]     elem_name      検索要素名
   * @param[in]     attr_names      検索属性名リスト（*attr_names[dims])
   * @param[out]     values         属性値リスト（values[dims])
   * @return    success(=true)/fault(=false)
   */
  bool getDimsAttributeValues(
                              int dims,
                              const char* elem_name,
                              const char* attr_names[],
                              double* values) const
  {
    if (dims == 0 || elem_name == NULL || attr_names == NULL || values == NULL || m_referElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getDimsAttributeValues arguments");
      return false;
    }

    const sphElem* elem = NULL;
    if (m_referElem->equalsElemName(elem_name)) {
      elem = m_referElem;
    }
    else {
      elem = m_referElem->getElem(elem_name);
    }
    if (elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found %s", elem_name);
      return false;
    }

    register int i;
    for (i=0; i<dims; i++) {
      if (elem->existAttribute(attr_names[i])) {
        values[i] = elem->getAttrReal8(attr_names[i]);
      }
      else {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ATTRIBUTE, "not found %s[@%s], %s", elem_name, attr_names[i], elem->debugPrintElem().c_str());
        return false;
      }
    }
    return true;
  }


  /**
   * この要素、又は子要素から属性を取得する(float)
   * @param[in]     dims      次元数
   * @param[in]     elem_name      検索要素名
   * @param[in]     attr_names      検索属性名リスト（*attr_names[dims])
   * @param[out]     values         属性値リスト（values[dims])
   * @return    success(=true)/fault(=false)
   */
  bool getDimsAttributeValues(
                            int dims,
                            const char* elem_name,
                            const char* attr_names[],
                            float* values) const
  {
    if (dims == 0 || elem_name == NULL || attr_names == NULL || values == NULL || m_referElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getDimsAttributeValues arguments");
      return false;
    }

    double *dbl_val = new double[dims];
    register int i;
    for (i=0; i<dims; i++) dbl_val[i] = values[i];
    if (!getDimsAttributeValues(dims, elem_name, attr_names, dbl_val)) {
      delete[] dbl_val;
      return false;
    }
    for (i=0; i<dims; i++) values[i] = (float)dbl_val[i];
    delete[] dbl_val;

    return true;
  }

};

#endif /* _SPHELEMUTILS_H_ */
