/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphArray_inline.h
 * Data Array Class Inline Source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphArray_inline.h
 * データ配列クラスインラインソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPH_ARRAY_INLINE_H_
#define _SPH_ARRAY_INLINE_H_

#define SPH_MEMFUN(rettype) \
        SPH_INLINE rettype


/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
template <class DATATYPE>
SPH_MEMFUN()
sphArray<DATATYPE>::sphArray() {
  m_pDims = 0;
  m_dims = 0;
  m_gc = 0;
  m_dtElemNum = 0;
  m_dt = NULL;
  register int i;
  for (i=0; i<SPH_MAX_DIM_SIZE_OF_PRIMITIVE_ARRAY; i++) {
    m_pSz[i] = 1;
    m_sz[i] = 1;
    m_gcl[i] = 0;
  }
}

/**
 * \~english
 * Constructor
 * @param    label      data label
 * @param    dctype      data class type
 * @param    dttype      data type
 * @param    paraCtrlr    Parallel Controller
 *
 * \~japanese
 * コンストラクタ
 * @param    label      データ識別ラベル
 * @param    dctype    データクラスタイプ
 * @param    dttype      データ型タイプ
 * @param    paraCtrlr    パラレルコントローラー
 */
template <class DATATYPE>
SPH_MEMFUN()
sphArray<DATATYPE>::sphArray(
                        const std::string label,
                        SphDCType dctype,
                        SPL_Datatype dttype,
                        const sphParaController *paraCtrlr)
                : sphData(label, dctype, dttype, paraCtrlr)
{
  m_pDims = 0;
  m_dims = 0;
  m_gc = 0;
  m_dtElemNum = 0;
  m_dt = NULL;
  register int i;
  for (i=0; i<SPH_MAX_DIM_SIZE_OF_PRIMITIVE_ARRAY; i++) {
    m_pSz[i] = 1;
    m_sz[i] = 1;
    m_gcl[i] = 0;
  }
}

/**
 * \~english
 * Destoructor
 *
 * \~japanese
 * デストラクタ
 */
template <class DATATYPE>
SPH_MEMFUN()
sphArray<DATATYPE>::~sphArray() {
  // データの破棄
  if( m_dt ) delete [] m_dt;
}



/**
 * データ配列をアロケート、初期化する。
 * @param    array_size  ノードサイズ
 * @param    gc        ガイドセルサイズ
 * @param    data_num  セルのデータ数
 * @param    allocated  データアロケートフラグ（true=アロケートを行う:default）
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::initializeArray(
                        const size_t* array_size,
                        const size_t gc,
                        const size_t data_num,
                        bool allocated)
{
  if (!array_size) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "array_size is null.");
    return false;
  }

  m_gc = gc;
  switch(sphData::m_dcType) {
  case SPH_DC_ARRAY_1D:    /* (i) */
    m_pDims = 1;
    m_pSz[0] = array_size[0] + gc*2;
    m_dims = 1;
    m_sz[0] = array_size[0];
    m_gcl[0] = m_gc;
    m_dtElemNum = 1;
    break;
  case SPH_DC_ARRAY_1DN:  /* (i, n) */
    m_pDims = 2;
    m_pSz[0] = array_size[0] + gc*2;
    m_pSz[1] = data_num;
    m_dims = 1;
    m_sz[0] = array_size[0];
    m_gcl[0] = m_gc;
    m_dtElemNum = data_num;
    break;
  case SPH_DC_ARRAY_1DNRV:  /* (n, i) */
    m_pDims = 2;
    m_pSz[0] = data_num;
    m_pSz[1] = array_size[0] + gc*2;
    m_dims = 1;
    m_sz[0] = array_size[0];
    m_gcl[1] = m_gc;
    m_dtElemNum = data_num;
    break;
  case SPH_DC_ARRAY_2D:    /* (i, j) */
    m_pDims = 2;
    m_pSz[0] = array_size[0] + gc*2;
    m_pSz[1] = array_size[1] + gc*2;
    m_dims = 2;
    m_sz[0] = array_size[0];
    m_sz[1] = array_size[1];
    m_gcl[0] = m_gcl[1] = m_gc;
    m_dtElemNum = 1;
    break;
  case SPH_DC_ARRAY_2DN:  /* (i, j, n) */
    m_pDims = 3;
    m_pSz[0] = array_size[0] + gc*2;
    m_pSz[1] = array_size[1] + gc*2;
    m_pSz[2] = data_num;
    m_dims = 2;
    m_sz[0] = array_size[0];
    m_sz[1] = array_size[1];
    m_gcl[0] = m_gcl[1] = m_gc;
    m_dtElemNum = data_num;
    break;
  case SPH_DC_ARRAY_2DNRV:  /* (n, i, j) */
    m_pDims = 3;
    m_pSz[0] = data_num;
    m_pSz[1] = array_size[0] + gc*2;
    m_pSz[2] = array_size[1] + gc*2;
    m_dims = 2;
    m_sz[0] = array_size[0];
    m_sz[1] = array_size[1];
    m_gcl[1] = m_gcl[2] = m_gc;
    m_dtElemNum = data_num;
    break;
  case SPH_DC_ARRAY_3D:    /* (i, j, k) */
    m_pDims = 3;
    m_pSz[0] = array_size[0] + gc*2;
    m_pSz[1] = array_size[1] + gc*2;
    m_pSz[2] = array_size[2] + gc*2;
    m_dims = 3;
    m_sz[0] = array_size[0];
    m_sz[1] = array_size[1];
    m_sz[2] = array_size[2];
    m_gcl[0] = m_gcl[1] = m_gcl[2] = m_gc;
    m_dtElemNum = 1;
    break;
  case SPH_DC_ARRAY_3DN:  /* (i, j, k, n) */
    m_pDims = 4;
    m_pSz[0] = array_size[0] + gc*2;
    m_pSz[1] = array_size[1] + gc*2;
    m_pSz[2] = array_size[2] + gc*2;
    m_pSz[3] = data_num;
    m_dims = 3;
    m_sz[0] = array_size[0];
    m_sz[1] = array_size[1];
    m_sz[2] = array_size[2];
    m_gcl[0] = m_gcl[1] = m_gcl[2] = m_gc;
    m_dtElemNum = data_num;
    break;
  case SPH_DC_ARRAY_3DNRV:  /* (n, i, j, k) */
    m_pDims = 4;
    m_pSz[0] = data_num;
    m_pSz[1] = array_size[0] + gc*2;
    m_pSz[2] = array_size[1] + gc*2;
    m_pSz[3] = array_size[2] + gc*2;
    m_dims = 3;
    m_sz[0] = array_size[0];
    m_sz[1] = array_size[1];
    m_sz[2] = array_size[2];
    m_gcl[1] = m_gcl[2] = m_gcl[3] = m_gc;
    m_dtElemNum = data_num;
    break;
  default:
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support dctype[=%d].", sphData::m_dcType);
    return false;
  }

  size_t array_len = 1;
  register unsigned n;
  for(n=0; n<m_pDims; n++) array_len *= m_pSz[n];
  if (array_len == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "allocate size is zero");
    return false;
  }

  if (allocated) {
    if( allocateData(array_len) == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not allocate data");
      return false;
    }
    memset(m_dt, 0, sizeof(DATATYPE)*array_len);
  }
  else {
    m_dt = NULL;
  }

  return true;
}


/**
 * \~english
 * get dimension of node (without data lenght)
 * @return    unsigned      dimension of node. (=1, 2 or 3)
 *
 * \~japanese
 * ノードの次元数を取得する。（データ長lは含まない）
 * @return    unsigned      ノードの次元数 (1〜3)
 */
template <class DATATYPE>
SPH_MEMFUN(unsigned)
sphArray<DATATYPE>::getDims() const { return m_dims; }


/**
 * \~english
 * get size of node (without guide cell)
 * @return    const size_t*      node size
 *
 * \~japanese
 * ノードの次元サイズ（ガイドセルを含まない）を取得する。.
 * １セル当りのデータ数は返さない。
 * @return    const size_t*      ノードの次元サイズ（ガイドセルを含まない）{i, j, k}
 */
template <class DATATYPE>
SPH_MEMFUN(const size_t*)
sphArray<DATATYPE>::getArraySize() const { return m_sz; }


/**
 * ノードの次元サイズ（ガイドセルを含む）を取得する。.
 * １セル当りのデータ数は返さない。
 * @return    const size_t*      ノードの次元サイズ（ガイドセルを含む）{i, j, k}
 */
template <class DATATYPE>
SPH_MEMFUN(const size_t*)
sphArray<DATATYPE>::_getArraySize() const
{

  switch(sphData::m_dcType) {
  case SPH_DC_ARRAY_1D:    /* (i) */
  case SPH_DC_ARRAY_1DN:  /* (i, n) */
  case SPH_DC_ARRAY_2D:    /* (i, j) */
  case SPH_DC_ARRAY_2DN:  /* (i, j, n) */
  case SPH_DC_ARRAY_3D:    /* (i, j, k) */
  case SPH_DC_ARRAY_3DN:  /* (i, j, k, n) */
    return m_pSz;
    break;
  case SPH_DC_ARRAY_1DNRV:  /* (n, i) */
  case SPH_DC_ARRAY_2DNRV:  /* (n, i, j) */
  case SPH_DC_ARRAY_3DNRV:  /* (n, i, j, k) */
    return m_pSz+1;
    break;
  }

  return NULL;
}


/**
 * \~english
 * get length of data array
 * @return    size_t*      data length
 *
 * \~japanese
 * データ配列の長さを取得する。
 * @return    size_t*      データ配列長
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphArray<DATATYPE>::getArrayLength() const
{
  size_t array_len = 1;
  register unsigned n;
  for(n=0; n<m_pDims; n++) array_len *= m_pSz[n];
  return array_len;
}


/**
 * \~english
 * get guide cell length
 * @return    size_t*      guide cell length
 *
 * \~japanese
 * ガイドセルサイズを取得する。
 * @return    size_t*      ガイドセルサイズ
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphArray<DATATYPE>::getGc() const { return m_gc; }


/**
 * \~english
 * get data length in cell
 * @return    size_t*      data length in cell
 *
 * \~japanese
 * セルのデータ数を取得する。(=l)
 * @return    size_t      セルのデータ数
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphArray<DATATYPE>::getNumOfDtElem() const { return m_dtElemNum; }


/**
 * \~english
 * get data array pointer(const)
 * @return    const DATATYPE*      data array pointer
 *
 * \~japanese
 * データ配列のポインタを取得する。(const)
 * @return    const DATATYPE*      データ配列のポインタ
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE*)
sphArray<DATATYPE>::getData() const { return m_dt; }


/**
 * \~english
 * get data array pointer
 * @return    DATATYPE*      data array pointer
 *
 * \~japanese
 * データ配列のポインタを取得する。
 * @return    DATATYPE*      データ配列のポインタ
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE*)
sphArray<DATATYPE>::getData() { return m_dt; }


/**
 * \~english
 * get data value in node area.(const)
 * @param    i    I index
 * @return    const DATATYPE&      data value
 *
 * \~japanese
 * ノード領域のデータを取得する。(const).
 * ガイドセルを除いた領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @return  const DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE&)
sphArray<DATATYPE>::val(size_t i) const {
  return m_dt[ (i + m_gcl[0]) ];
}


/**
 * \~english
 * get data value in node area.
 * @param    i    I index
 * @return     DATATYPE&      data value
 *
 * \~japanese
 * ノード領域のデータを取得する。
 * ガイドセルを除いた領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @return   DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE&)
sphArray<DATATYPE>::val(size_t i) {
  return m_dt[ (i + m_gcl[0]) ];
}


/**
 * \~english
 * get data value in node area.(const)
 * @param    i    I index
 * @param    j    J index
 * @return     const DATATYPE&      data value
 *
 * \~japanese
 * ノード領域のデータを取得する。(const)
 * ガイドセルを除いた領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含まない）
 * @return   const DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE&)
sphArray<DATATYPE>::val(size_t i, size_t j) const {
  return m_dt[ m_pSz[0]*(j + m_gcl[1]) + (i + m_gcl[0]) ];
}


/**
 * \~english
 * get data value in node area.
 * @param    i    I index
 * @param    j    J index
 * @return   DATATYPE&      data value
 *
 * \~japanese
 * ノード領域のデータを取得する。
 * ガイドセルを除いた領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含まない）
 * @return   DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE&)
sphArray<DATATYPE>::val(size_t i, size_t j) {
  return m_dt[ m_pSz[0]*(j + m_gcl[1]) + (i + m_gcl[0]) ];
}


/**
 * \~english
 * get data value in node area.(const)
 * @param    i    I index
 * @param    j    J index
 * @param    k    K index
 * @return     const DATATYPE&      data value
 *
 * \~japanese
 * ノード領域のデータを取得する。(const)
 * ガイドセルを除いた領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含まない）
 * @param    k    K方向ノード領域インデックス（ガイドセルは含まない）
 * @return   const DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE&)
sphArray<DATATYPE>::val(size_t i, size_t j, size_t k) const {
  return m_dt[   m_pSz[0]*m_pSz[1]*(k + m_gcl[2])
               + m_pSz[0]*(j + m_gcl[1]) + (i + m_gcl[0]) ];
}


/**
 * \~english
 * get data value in node area.
 * @param    i    I index
 * @param    j    J index
 * @param    k    K index
 * @return    DATATYPE&      data value
 *
 * \~japanese
 * ノード領域のデータを取得する。
 * ガイドセルを除いた領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含まない）
 * @param    k    K方向ノード領域インデックス（ガイドセルは含まない）
 * @return   DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE&)
sphArray<DATATYPE>::val(size_t i, size_t j, size_t k) {
  return m_dt[   m_pSz[0]*m_pSz[1]*(k + m_gcl[2])
               + m_pSz[0]*(j + m_gcl[1]) + (i + m_gcl[0]) ];
}


/**
 * \~english
 * get data value in node area.(const)
 * @param    i    I index
 * @param    j    J index
 * @param    k    K index
 * @param    l    data length
 * @return     const DATATYPE&      data value
 *
 * \~japanese
 * ノード領域のデータを取得する。(const)
 * ガイドセルを除いた領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含まない）
 * @param    k    K方向ノード領域インデックス（ガイドセルは含まない）
 * @param    l    セルのデータ数
 * @return   const DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE&)
sphArray<DATATYPE>::val(size_t i, size_t j, size_t k, size_t l) const {
  return m_dt[   m_pSz[0]*m_pSz[1]*m_pSz[2]*(l + m_gcl[3])
               + m_pSz[0]*m_pSz[1]*(k + m_gcl[2])
               + m_pSz[0]*(j + m_gcl[1])
               + (i + m_gcl[0]) ];
}


/**
 * \~english
 * get data value in node area.
 * @param    i    I index
 * @param    j    J index
 * @param    k    K index
 * @param    l    data length
 * @return   DATATYPE&      data value
 *
 * \~japanese
 * ノード領域のデータを取得する。
 * ガイドセルを除いた領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含まない）
 * @param    k    K方向ノード領域インデックス（ガイドセルは含まない）
 * @param    l    セルのデータ数
 * @return   DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE&)
sphArray<DATATYPE>::val(size_t i, size_t j, size_t k, size_t l) {
  return m_dt[   m_pSz[0]*m_pSz[1]*m_pSz[2]*(l + m_gcl[3])
               + m_pSz[0]*m_pSz[1]*(k + m_gcl[2])
               + m_pSz[0]*(j + m_gcl[1])
               + (i + m_gcl[0]) ];
}


/**
 * \~english
 * get data value in data array.(const)
 * @param    i    I index
 * @return   const DATATYPE&      data value
 *
 * \~japanese
 * データ領域のデータを取得する。(const)
 * ガイドセルを含んだ領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルを含む）
 * @return   const DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE&)
sphArray<DATATYPE>::_val(size_t i) const {
  return m_dt[ i ];
}


/**
 * \~english
 * get data value in data array.
 * @param    i    I index
 * @return   DATATYPE&      data value
 *
 * \~japanese
 * データ領域のデータを取得する。
 * ガイドセルを含んだ領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルを含む）
 * @return   DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE&)
sphArray<DATATYPE>::_val(size_t i) {
  return m_dt[ i ];
}


/**
 * \~english
 * get data value in data array.(const)
 * @param    i    I index
 * @param    j    J index
 * @return   const DATATYPE&      data value
 *
 * \~japanese
 * データ領域のデータを取得する。(const)
 * ガイドセルを含んだ領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルを含む）
 * @param    j    J方向ノード領域インデックス（ガイドセルを含む）
 * @return   const DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE&)
sphArray<DATATYPE>::_val(size_t i, size_t j) const {
  return m_dt[ m_pSz[0]*j + i ];
}


/**
 * \~english
 * get data value in data array.
 * @param    i    I index
 * @param    j    J index
 * @return   DATATYPE&      data value
 *
 * \~japanese
 * データ領域のデータを取得する。
 * ガイドセルを含んだ領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルを含む）
 * @param    j    J方向ノード領域インデックス（ガイドセルを含む）
 * @return   DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE&)
sphArray<DATATYPE>::_val(size_t i, size_t j) {
  return m_dt[ m_pSz[0]*j + i ];
}


/**
 * \~english
 * get data value in data array.(const)
 * @param    i    I index
 * @param    j    J index
 * @param    k    K index
 * @return   const DATATYPE&      data value
 *
 * \~japanese
 * データ領域のデータを取得する。(const)
 * ガイドセルを含んだ領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルを含む）
 * @param    j    J方向ノード領域インデックス（ガイドセルを含む）
 * @param    k    K方向ノード領域インデックス（ガイドセルを含む）
 * @return   const DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE&)
sphArray<DATATYPE>::_val(size_t i, size_t j, size_t k) const {
  return m_dt[ m_pSz[0]*m_pSz[1]*k + m_pSz[0]*j + i ];
}


/**
 * \~english
 * get data value in data array.
 * @param    i    I index
 * @param    j    J index
 * @param    k    K index
 * @return   DATATYPE&      data value
 *
 * \~japanese
 * データ領域のデータを取得する。
 * ガイドセルを含んだ領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルを含む）
 * @param    j    J方向ノード領域インデックス（ガイドセルを含む）
 * @param    k    K方向ノード領域インデックス（ガイドセルを含む）
 * @return   DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE&)
sphArray<DATATYPE>::_val(size_t i, size_t j, size_t k) {
  return m_dt[ m_pSz[0]*m_pSz[1]*k + m_pSz[0]*j + i ];
}


/**
 * \~english
 * get data value in data array.(const)
 * @param    i    I index
 * @param    j    J index
 * @param    k    K index
 * @param    l    data length
 * @return   const DATATYPE&      data value
 *
 * \~japanese
 * データ領域のデータを取得する。(const)
 * ガイドセルを含んだ領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルを含む）
 * @param    j    J方向ノード領域インデックス（ガイドセルを含む）
 * @param    k    K方向ノード領域インデックス（ガイドセルを含む）
 * @param    l    セルのデータ数
 * @return   const DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE&)
sphArray<DATATYPE>::_val(size_t i, size_t j, size_t k, size_t l) const {
  return m_dt[  m_pSz[0]*m_pSz[1]*m_pSz[2]*l
              + m_pSz[0]*m_pSz[1]*k + m_pSz[0]*j + i ];
}


/**
 * \~english
 * get data value in data array.
 * @param    i    I index
 * @param    j    J index
 * @param    k    K index
 * @param    l    data length
 * @return   DATATYPE&      data value
 *
 * \~japanese
 * データ領域のデータを取得する。
 * ガイドセルを含んだ領域を取得する。
 * @param    i    I方向ノード領域インデックス（ガイドセルを含む）
 * @param    j    J方向ノード領域インデックス（ガイドセルを含む）
 * @param    k    K方向ノード領域インデックス（ガイドセルを含む）
 * @param    l    セルのデータ数
 * @return   DATATYPE&     データ値
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE&)
sphArray<DATATYPE>::_val(size_t i, size_t j, size_t k, size_t l) {
  return m_dt[  m_pSz[0]*m_pSz[1]*m_pSz[2]*l
              + m_pSz[0]*m_pSz[1]*k + m_pSz[0]*j + i ];
}


/**
 * データ領域のデータを取得する。（ガイドセルは含まない領域）
 *
 * すべてのクラスタイプにおいてもパラメータは、i,j,k,l固定である。
 * i,j,kはガイドセルを含まない領域のインデックスである。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含まない）
 * @param    k    K方向ノード領域インデックス（ガイドセルは含まない）
 * @param    l    セルのデータ数
 * @return   true=取得成功, false=取得失敗(インデックスが領域外)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::getValue(DATATYPE& val, size_t i, size_t j, size_t k, size_t l) const
{
  if (i>=m_sz[0] || j>=m_sz[1] || k>=m_sz[2] || l>=m_dtElemNum) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "out of range. i,j,k,l:%ld,%ld,%ld,%ld.", i, j, k, l);
    return false;
  }
  size_t array_len = getArrayLength();
  size_t index = getIndex(i, j, k, l);
  if (index >= array_len) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "out of range. i,j,k,l:%ld,%ld,%ld,%ld (index=%ld).", i, j, k, l, index);
    return false;
  }
  val = m_dt[index];
  return true;
}


/**
 * データ領域のデータを取得する。（ガイドセルは含む領域）
 *
 * すべてのクラスタイプにおいてもパラメータは、i,j,k,l固定である。
 * i,j,kはガイドセルを含む領域のインデックスである。
 * @param[out]   val    データ値
 * @param[in]    i    I方向ノード領域インデックス（ガイドセルは含む）
 * @param[in]    j    J方向ノード領域インデックス（ガイドセルは含む）
 * @param[in]    k    K方向ノード領域インデックス（ガイドセルは含む）
 * @param[in]    l    セルのデータ数
 * @return   true=取得成功, false=取得失敗(インデックスが領域外)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::_getValue(DATATYPE& val, size_t i, size_t j, size_t k, size_t l) const
{
  if (i>=m_sz[0]+2*m_gc || j>=m_sz[1]+2*m_gc || k>=m_sz[2]+2*m_gc || l>=m_dtElemNum) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "out of range. i,j,k,l:%ld,%ld,%ld,%ld.", i, j, k, l);
    return false;
  }
  size_t array_len = getArrayLength();
  size_t index = _getIndex(i, j, k, l);
  if ( index >= array_len) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "i,j,k,l:%ld,%ld,%ld,%ld (index=%ld).", i, j, k, l, index);
    return false;
  }
  val = m_dt[index];

  return true;
}

/**
 * データ領域のデータを設定する。（ガイドセルは含まない領域）
 *
 * すべてのクラスタイプにおいてもパラメータは、i,j,k,l固定である。
 * i,j,kはガイドセルを含む領域のインデックスである。
 * @param    val    データ値
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含まない）
 * @param    k    K方向ノード領域インデックス（ガイドセルは含まない）
 * @param    l    セルのデータ数
 * @return   true=設定成功, false=設定失敗(インデックスが領域外)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::setValue(DATATYPE val, size_t i, size_t j, size_t k, size_t l)
{
  if (i>=m_sz[0] || j>=m_sz[1] || k>=m_sz[2] || l>=m_dtElemNum) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "out of range. i,j,k,l:%ld,%ld,%ld,%ld.", i, j, k, l);
    return false;
  }
  size_t array_len = getArrayLength();
  size_t index = getIndex(i, j, k, l);
  if (index >= array_len) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "i,j,k,l:%ld,%ld,%ld,%ld (index=%ld).", i, j, k, l, index);
    return false;
  }
  m_dt[index] = val;
  return true;
}


/**
 * データ領域のデータを設定する。（ガイドセルは含む領域）
 *
 * すべてのクラスタイプにおいてもパラメータは、i,j,k,l固定である。
 * i,j,kはガイドセルを含む領域のインデックスである。
 * @param    val    データ値
 * @param    i    I方向ノード領域インデックス（ガイドセルは含む）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含む）
 * @param    k    K方向ノード領域インデックス（ガイドセルは含む）
 * @param    l    セルのデータ数
 * @return   true=設定成功, false=設定失敗(インデックスが領域外)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::_setValue(DATATYPE val, size_t i, size_t j, size_t k, size_t l)
{
  if (i>=m_sz[0]+2*m_gc || j>=m_sz[1]+2*m_gc || k>=m_sz[2]+2*m_gc || l>=m_dtElemNum) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "out of range. i,j,k,l:%ld,%ld,%ld,%ld.", i, j, k, l);
    return false;
  }
  size_t array_len = getArrayLength();
  size_t index = _getIndex(i, j, k, l);
  if (index >= array_len) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "i,j,k,l:%ld,%ld,%ld,%ld (index=%ld).", i, j, k, l, index);
    return NULL;
  }
  m_dt[index] = val;
  return true;
}


/**
 * \~english
 * get index in data array.
 *
 * Parameter is i,j,k,l of all data class type.
 * @param    i    I index
 * @param    j    J index
 * @param    k    K index
 * @param    l    data length
 * @return   DATATYPE&      data value
 *
 * \~japanese
 * データ領域のインデックスを取得する。（ガイドセルは含まない領域）
 *
 * i,j,kはガイドセルを含まない領域のインデックスである。
 * 戻り値はデータ配列上のインデックスである。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含まない）：1Dの場合はセルのデータ数
 * @param    k    K方向ノード領域インデックス（ガイドセルは含まない）：2Dの場合はセルのデータ数
 * @param    l    セルのデータ数 (3D)
 * @return   １次元データ配列インデックス
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphArray<DATATYPE>::getIndex(size_t i, size_t j, size_t k, size_t l) const
{
  size_t size[4] = {1, 1, 1, 1};
  size_t gc[4] = {0, 0, 0, 0};
  size_t idx[4] = {0, 0, 0, 0};
  size_t dlen = 0;
  bool rev = false;
  register unsigned n;
  switch (m_dcType) {
    case (SPH_DC_ARRAY_1D):
      idx[0]=i;
      break;
    case (SPH_DC_ARRAY_2D):
      idx[0]=i;
      idx[1]=j;
      break;
    case (SPH_DC_ARRAY_3D):
      idx[0]=i;
      idx[1]=j;
      idx[2]=k;
      break;
    case (SPH_DC_ARRAY_1DN):
      idx[0]=i;
      idx[1]=j;
      break;
    case (SPH_DC_ARRAY_2DN):
      idx[0]=i;
      idx[1]=j;
      idx[2]=k;
      break;
    case (SPH_DC_ARRAY_3DN):
      idx[0]=i;
      idx[1]=j;
      idx[2]=k;
      idx[3]=l;
      break;
    case (SPH_DC_ARRAY_1DNRV):
      idx[0]=j;
      idx[1]=i;
      rev = true;
      break;
    case (SPH_DC_ARRAY_2DNRV):
      idx[0]=k;
      idx[1]=i;
      idx[2]=j;
      rev = true;
      break;
    case (SPH_DC_ARRAY_3DNRV):
      idx[0]=l;
      idx[1]=i;
      idx[2]=j;
      idx[3]=k;
      rev = true;
      break;
  }
  for(n=0; n<m_pDims; n++) {
    size[n] = m_pSz[n];
    gc[n] = m_gcl[n];
  }
  size_t index = size[0]*size[1]*size[2]*(idx[3] + gc[3])
                + size[0]*size[1]*(idx[2] + gc[2])
                + size[0]*(idx[1] + gc[1])
                + (idx[0] + gc[0]);
  return index;
}


/**
 * データ領域のインデックスを取得する。（ガイドセルは含む領域）.
 *
 * i,j,kはガイドセルを含む領域のインデックスである。
 * 戻り値はデータ配列上のインデックスである。
 * @param    i    I方向ノード領域インデックス（ガイドセルは含む）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含む）：1Dの場合はセルのデータ数
 * @param    k    K方向ノード領域インデックス（ガイドセルは含む）：2Dの場合はセルのデータ数
 * @param    l    セルのデータ数 (3D)
 * @return   １次元データ配列インデックス
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphArray<DATATYPE>::_getIndex(size_t i, size_t j, size_t k, size_t l) const
{
  size_t size[4] = {1, 1, 1, 1};
  size_t idx[4] = {0, 0, 0, 0};
  size_t dlen = 0;
  bool rev = false;
  register unsigned n;
  switch (m_dcType) {
    case (SPH_DC_ARRAY_1D):
      idx[0]=i;
      break;
    case (SPH_DC_ARRAY_2D):
      idx[0]=i;
      idx[1]=j;
      break;
    case (SPH_DC_ARRAY_3D):
      idx[0]=i;
      idx[1]=j;
      idx[2]=k;
      break;
    case (SPH_DC_ARRAY_1DN):
      idx[0]=i;
      idx[1]=j;
      break;
    case (SPH_DC_ARRAY_2DN):
      idx[0]=i;
      idx[1]=j;
      idx[2]=k;
      break;
    case (SPH_DC_ARRAY_3DN):
      idx[0]=i;
      idx[1]=j;
      idx[2]=k;
      idx[3]=l;
      break;
    case (SPH_DC_ARRAY_1DNRV):
      idx[0]=j;
      idx[1]=i;
      rev = true;
      break;
    case (SPH_DC_ARRAY_2DNRV):
      idx[0]=k;
      idx[1]=i;
      idx[2]=j;
      rev = true;
      break;
    case (SPH_DC_ARRAY_3DNRV):
      idx[0]=l;
      idx[1]=i;
      idx[2]=j;
      idx[3]=k;
      rev = true;
      break;
  }
  for(n=0; n<m_pDims; n++) {
    size[n] = m_pSz[n];
  }
  size_t index = size[0]*size[1]*size[2]*(idx[3])
                + size[0]*size[1]*(idx[2])
                + size[0]*(idx[1])
                + (idx[0]);
  return index;
}


/**
 * \~english
 * communication of guide cells.
 * @param    gc    communication guide cell
 * @param    paraKey    parallel controller key
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ガイドセルの通信を行う。
 * @param    gc          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::comm(size_t gc, int paraKey) { return true; }


/**
 * EV型のガイドセルの通信を行う。
 * @param    gc          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::commPeriodic(unsigned int face_dir, int axis, size_t face, int paraKey) { return true; }


/**
 * \~english
 * create communication type.
 * @param    paraKey    parallel controller key
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ノード間通信タイプの作成を行う。
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::createCommType(int paraKey) { return true; }



/**
 * データ配列のポインタを取得する。
 * @param     void
 * @return    データ配列ポインタ
 */
template <class DATATYPE>
SPH_MEMFUN(const void*)
sphArray<DATATYPE>::getDataPointer() const
{
  return m_dt;
}

/**
 * データ配列のポインタを取得する。
 * @param     void
 * @return    データ配列ポインタ
 */
template <class DATATYPE>
SPH_MEMFUN(void*)
sphArray<DATATYPE>::getDataPointer()
{
  return m_dt;
}


/**
 * データをコピーする.
 *
 * データ領域サイズ、データ型、データクラス、格子当りデータ数は同じでない場合はエラーとする。
 * ガイドセル領域もコピーを行う。（ガイドセルが異なる場合もコピーする）
 * コピー元、コピー先のグローバル始点インデックスは同じインデックスとしてコピーを行う。
 * @param     src_dc           コピー元データ
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::copyArray(const sphData* src_dc)
{
  if (src_dc == NULL) return false;
  if (getDcType() != src_dc->getDcType()) return false;
  if (getDataType() != src_dc->getDataType()) return false;
  if (getDims() != src_dc->getDims()) return false;
  if (getNumOfDtElem() != src_dc->getNumOfDtElem()) return false;

  const sphArray<DATATYPE> *src_vox = dynamic_cast<const sphArray<DATATYPE>*>(src_dc);
  if (src_vox == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphArray(src_dc) dynamic_cast error.");
    return false;
  }

  register size_t i, j, k, n;
  const size_t* dst_sz_excgc = getArraySize();  // dimension size /wo guide cell
  const size_t* src_sz_excgc = src_vox->getArraySize();
  for (i=0; i<m_dims; i++) {
    if (src_sz_excgc[i] != dst_sz_excgc[i]) return false;
  }

  DATATYPE* dst_data = getData();
  const DATATYPE* src_data = src_vox->getData();
  if( !dst_data || !src_data ) return false;

  size_t dst_gc = getGc();
  size_t src_gc = src_vox->getGc();
  size_t dst_size[3] = {1, 1, 1};
  size_t src_size[3] = {1, 1, 1};
  for (i=0; i<m_dims; i++) {
    dst_size[i] = dst_sz_excgc[i] + dst_gc*2;
    src_size[i] = src_sz_excgc[i] + src_gc*2;
  }
  size_t data_num = getNumOfDtElem();
  SphDCType   dctype = getDcType();
  size_t sta[3] = {0, 0, 0};
  int diff;

  // compare whole size
  bool isWhole = true;
  for (i=0; i<m_dims; i++) {
    if (dst_size[i] != src_size[i]) {
      isWhole = false;
      break;
    }
  }

  if( isWhole ) {
    size_t array_len = getArrayLength();
    memcpy(dst_data, src_data, sizeof(DATATYPE)*array_len);
  }
  else {
    diff = src_gc - dst_gc;
    size_t dst_end[3] = {1, 1, 1};
    if( src_gc >= dst_gc ){
      /// コピー先ガイドセルが小さいのですべてコピーする。
      for (i=0; i<m_dims; i++) {
        dst_end[i] = dst_size[i];
        sta[i] = 0;
      }
    }
    else {
      /// コピー先ガイドセルが大きいのでガイドセルの差分を引く。
      for (i=0; i<m_dims; i++) {
        dst_end[i] = dst_size[i]+diff;
        sta[i] = abs(diff);
      }
    }
    size_t src_voxlen = 1;
    size_t dst_voxlen = 1;
    for (i=0; i<m_dims; i++) {
      src_voxlen *= src_size[i];
      dst_voxlen *= dst_size[i];
    }

    for (k=sta[2]; k<dst_end[2]; k++) {
      size_t kk = k+diff;
      if (m_dims <= 2 || (long long)k+diff < 0) kk = 0;
      for(j=sta[1]; j<dst_end[1]; j++){
        size_t jj = j+diff;
        if (m_dims <= 1 || (long long)j+diff < 0) jj = 0;
        for(i=sta[0]; i<dst_end[0]; i++){
          size_t src_idx = src_size[0]*src_size[1]*kk
                          + src_size[0]*jj
                          + i+diff;
          size_t dst_idx = dst_size[0]*dst_size[1]*k
                          + dst_size[0]*j
                          + i;
          if (dctype == SPH_DC_ARRAY_1DN
              || dctype == SPH_DC_ARRAY_2DN
              || dctype == SPH_DC_ARRAY_3DN) {
            for (n=0; n<data_num; n++) {
              dst_data[dst_idx+dst_voxlen*n] = src_data[src_idx+src_voxlen*n];
            }
          }
          else {
            src_idx *= data_num;
            dst_idx *= data_num;
            memcpy( dst_data+dst_idx,
                    src_data+src_idx,
                    sizeof(DATATYPE)*data_num);
          }
        }
      }
    }
  }

  return true;
}

/**
 * データを中央部分にコピーする.
 * データ型、データクラス、格子当りデータ数は同じでない場合はエラーとする。
 * ガイドセル領域もコピーを行う。
 * @param     src_dc           コピー元データ
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::copyCenter(const sphArray<DATATYPE>* src_dc)
{
  if (src_dc == NULL) return false;
  if (getDcType() != src_dc->getDcType()) return false;
  if (getDataType() != src_dc->getDataType()) return false;
  if (getDims() != src_dc->getDims()) return false;
  if (getNumOfDtElem() != src_dc->getNumOfDtElem()) return false;

  register size_t i, j, k, n;
  const size_t* dst_sz_excgc = getArraySize();  // dimension size /wo guide cell
  const size_t* src_sz_excgc = src_dc->getArraySize();

  DATATYPE* dst_data = getData();
  const DATATYPE* src_data = src_dc->getData();
  if( !dst_data || !src_data ) return false;

  size_t dst_gc = getGc();
  size_t src_gc = src_dc->getGc();
  size_t dst_size[3] = {1, 1, 1};
  size_t src_size[3] = {1, 1, 1};
  for (i=0; i<m_dims; i++) {
    dst_size[i] = dst_sz_excgc[i] + dst_gc*2;
    src_size[i] = src_sz_excgc[i] + src_gc*2;
  }
  size_t data_num = getNumOfDtElem();
  SphDCType   dctype = getDcType();

  // compare whole size
  bool isWhole = true;
  for (i=0; i<m_dims; i++) {
    if (dst_size[i] != src_size[i]) {
      isWhole = false;
      break;
    }
  }

  if( isWhole ) {
    size_t array_len = getArrayLength();
    memcpy(dst_data, src_data, sizeof(DATATYPE)*array_len);
    return true;
  }

  size_t max[3] = {1, 1, 1};
  size_t src_dif[3] = {0, 0, 0};
  size_t dst_dif[3] = {0, 0, 0};
  for (i=0; i<m_dims; i++) {
    if( dst_size[i] < src_size[i] ) {
      max[i] = dst_size[i];
      src_dif[i] = (src_size[i] - dst_size[i])/2;
      dst_dif[i] = 0;
    }
    else {
      max[i] = src_size[i];
      src_dif[i] = 0;
      dst_dif[i] = (dst_size[i] - src_size[i])/2;
    }
  }

  size_t src_voxlen = 1;
  size_t dst_voxlen = 1;
  for (i=0; i<m_dims; i++) {
    src_voxlen *= src_size[i];
    dst_voxlen *= dst_size[i];
  }

  for(k=0; k<max[2]; k++) {
    size_t sk = k + src_dif[2];
    size_t dk = k + dst_dif[2];
    for(j=0; j<max[1]; j++) {
      size_t sj = j + src_dif[1];
      size_t dj = j + dst_dif[1];
      for(i=0; i<max[0]; i++){
        size_t src_idx = src_size[0]*src_size[1]*sk
                         + src_size[0]*sj
                         + i+src_dif[0];
        size_t dst_idx = dst_size[0]*dst_size[1]*dk
                         + dst_size[0]*dj
                         + i+dst_dif[0];
        if (dctype == SPH_DC_ARRAY_1DN
            || dctype == SPH_DC_ARRAY_2DN
            || dctype == SPH_DC_ARRAY_3DN) {
          for (n=0; n<data_num; n++) {
            dst_data[dst_idx+dst_voxlen*n] = src_data[src_idx+src_voxlen*n];
          }
        }
        else {
          src_idx *= data_num;
          dst_idx *= data_num;
          memcpy( dst_data+dst_idx,
                  src_data+src_idx,
                  sizeof(DATATYPE)*data_num);
        }
      }
    }
  }

  return true;
}


/**
 * 成分データを中央部分にコピーする.
 * ガイドセル領域もコピーを行う。
 * @param     src_dc           コピー元データ
 * @param     dst_pos          コピー先成分インデックス
 * @param     src_pos          コピー元成分インデックス
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::copyComponent(
                    const sphArray<DATATYPE>* src_dc,
                    size_t dst_pos,
                    size_t src_pos)
{
  if (src_dc == NULL) return false;
  if (getDataType() != src_dc->getDataType()) return false;
  if (getDims() != src_dc->getDims()) return false;

  register size_t i, j, k, n;
  const size_t* dst_sz_excgc = getArraySize();  // dimension size /wo guide cell
  const size_t* src_sz_excgc = src_dc->getArraySize();

  DATATYPE* dst_data = getData();
  const DATATYPE* src_data = src_dc->getData();
  if( !dst_data || !src_data ) return false;

  size_t dst_gc = getGc();
  size_t src_gc = src_dc->getGc();
  size_t dst_size[3] = {1, 1, 1};
  size_t src_size[3] = {1, 1, 1};
  for (i=0; i<m_dims; i++) {
    dst_size[i] = dst_sz_excgc[i] + dst_gc*2;
    src_size[i] = src_sz_excgc[i] + src_gc*2;
  }
  size_t dst_data_num = getNumOfDtElem();
  size_t src_data_num = src_dc->getNumOfDtElem();
  if( (dst_data_num <= dst_pos) || (src_data_num <= src_pos) ) return false;
  SphDCType   src_dctype = src_dc->getDcType();
  SphDCType   dst_dctype = getDcType();

  // compare whole size
  bool isWhole = true;
  for (i=0; i<m_dims; i++) {
    if (dst_size[i] != src_size[i]) {
      isWhole = false;
      break;
    }
  }

  if( isWhole ) {
    size_t array_len = 1;
    for (i=0; i<m_dims; i++) {
      array_len *= src_size[i];
    }
    if ( (src_dctype == dst_dctype)
        && (src_dctype == SPH_DC_ARRAY_1DN
        || src_dctype == SPH_DC_ARRAY_2DN
        || src_dctype == SPH_DC_ARRAY_3DN) ) {
      memcpy( dst_data+array_len*dst_pos,
              src_data+array_len*src_pos,
              sizeof(DATATYPE)*array_len);
      return true;
    }
  }

  size_t max[3] = {1, 1, 1};
  size_t src_dif[3] = {0, 0, 0};
  size_t dst_dif[3] = {0, 0, 0};
  for (i=0; i<m_dims; i++) {
    if( dst_size[i] < src_size[i] ) {
      max[i] = dst_size[i];
      src_dif[i] = (src_size[i] - dst_size[i])/2;
      dst_dif[i] = 0;
    }
    else {
      max[i] = src_size[i];
      src_dif[i] = 0;
      dst_dif[i] = (dst_size[i] - src_size[i])/2;
    }
  }

  size_t src_voxlen = 1;
  size_t dst_voxlen = 1;
  for (i=0; i<m_dims; i++) {
    src_voxlen *= src_size[i];
    dst_voxlen *= dst_size[i];
  }

  for(k=0; k<max[2]; k++) {
    size_t sk = k + src_dif[2];
    size_t dk = k + dst_dif[2];
    for(j=0; j<max[1]; j++) {
      size_t sj = j + src_dif[1];
      size_t dj = j + dst_dif[1];
      for(i=0; i<max[0]; i++){
        size_t src_idx =   src_size[0]*src_size[1]*sk
                         + src_size[0]*sj
                         + i+src_dif[0];
        size_t dst_idx =   dst_size[0]*dst_size[1]*dk
                         + dst_size[0]*dj
                         + i+dst_dif[0];
        if (src_dctype == SPH_DC_ARRAY_1DN
            || src_dctype == SPH_DC_ARRAY_2DN
            || src_dctype == SPH_DC_ARRAY_3DN) {
          src_idx += src_size[0]*src_size[1]*src_size[2]*src_pos;
        }
        else {
          src_idx = src_idx*src_data_num+src_pos;
        }

        if (dst_dctype == SPH_DC_ARRAY_1DN
            || dst_dctype == SPH_DC_ARRAY_2DN
            || dst_dctype == SPH_DC_ARRAY_3DN) {
          dst_idx += dst_size[0]*dst_size[1]*dst_size[2]*dst_pos;
        }
        else {
          dst_idx = dst_idx*dst_data_num+dst_pos;
        }
        dst_data[dst_idx] = src_data[src_idx];
      }
    }
  }

  return true;
}


/**
 * データクラスSPH_DC_ARRAY_3DNからSPH_DC_ARRAY_3DNRV又は反対にコピーする.
 * ガイドセル領域もコピーを行う。
 * @param     src_dc           コピー元データ
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::copyReverse(const sphArray<DATATYPE>* src_dc)
{
  if (src_dc == NULL) return false;
  if (getDataType() != src_dc->getDataType()) return false;
  if (getDims() != src_dc->getDims()) return false;
  if (getNumOfDtElem() != src_dc->getNumOfDtElem()) return false;

  register size_t i, j, k, n;
  const size_t* dst_sz_excgc = getArraySize();  // dimension size /wo guide cell
  const size_t* src_sz_excgc = src_dc->getArraySize();

  DATATYPE* dst_data = getData();
  const DATATYPE* src_data = src_dc->getData();
  if( !dst_data || !src_data ) return false;

  size_t dst_gc = getGc();
  size_t src_gc = src_dc->getGc();
  size_t dst_size[3] = {1, 1, 1};
  size_t src_size[3] = {1, 1, 1};
  for (i=0; i<m_dims; i++) {
    dst_size[i] = dst_sz_excgc[i] + dst_gc*2;
    src_size[i] = src_sz_excgc[i] + src_gc*2;
  }
  size_t data_num = getNumOfDtElem();
  SphDCType   src_dctype = src_dc->getDcType();
  SphDCType   dst_dctype = getDcType();

  size_t max[3] = {1, 1, 1};
  size_t src_dif[3] = {0, 0, 0};
  size_t dst_dif[3] = {0, 0, 0};
  for (i=0; i<m_dims; i++) {
    if( dst_size[i] < src_size[i] ) {
      max[i] = dst_size[i];
      src_dif[i] = (src_size[i] - dst_size[i])/2;
      dst_dif[i] = 0;
    }
    else {
      max[i] = src_size[i];
      src_dif[i] = 0;
      dst_dif[i] = (dst_size[i] - src_size[i])/2;
    }
  }

  size_t src_voxlen = 1;
  size_t dst_voxlen = 1;
  for (i=0; i<m_dims; i++) {
    src_voxlen *= src_size[i];
    dst_voxlen *= dst_size[i];
  }

  for(k=0; k<max[2]; k++) {
    size_t sk = k + src_dif[2];
    size_t dk = k + dst_dif[2];
    for(j=0; j<max[1]; j++) {
      size_t sj = j + src_dif[1];
      size_t dj = j + dst_dif[1];
      for(i=0; i<max[0]; i++){
        size_t src_idx = src_size[0]*src_size[1]*sk
                         + src_size[0]*sj
                         + i+src_dif[0];
        size_t dst_idx = dst_size[0]*dst_size[1]*dk
                         + dst_size[0]*dj
                         + i+dst_dif[0];
        size_t src_rv_idx = 0;
        size_t dst_rv_idx = 0;
        for (n=0; n<data_num; n++) {
          if (src_dctype == SPH_DC_ARRAY_1DNRV
            || src_dctype == SPH_DC_ARRAY_2DNRV
            || src_dctype == SPH_DC_ARRAY_3DNRV) {
            src_rv_idx = src_idx * data_num + n;
          }
          else {
            src_rv_idx = src_idx + src_voxlen*n;
          }
          if (dst_dctype == SPH_DC_ARRAY_1DNRV
            || dst_dctype == SPH_DC_ARRAY_2DNRV
            || dst_dctype == SPH_DC_ARRAY_3DNRV) {
            dst_rv_idx = dst_idx * data_num + n;
          }
          else {
            dst_rv_idx = dst_idx + dst_voxlen*n;
          }
          dst_data[dst_rv_idx] = src_data[src_rv_idx];
        }
      }
    }
  }

  return true;
}


/**
 * データに対して計算を行う.
 * ガイドセル領域も計算を行う。
 * @param     mode             計算モード
 *                1:dst_data[dst_idx] += src_data[src_idx];
 *                2:dst_data[dst_idx] *= stepAvr;
 *                3:dst_data[dst_idx] /= stepAvr;
 *                4:dst_data[dst_idx] = src_data[src_idx]*stepAvr+v00[0];
 * @param     src_dc             計算元データ
 * @param     stepAvr            乗算値
 * @param     v00                加算値（v00[data_num])
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::scaleAvr(int mode,
                          const sphArray<DATATYPE>* src_dc,
                          unsigned stepAvr,
                          double* va)
{
  if (mode == 1) {
    if (src_dc == NULL) return false;
    if (getDataType() != src_dc->getDataType()) return false;
    if (getDims() != src_dc->getDims()) return false;
    if (getNumOfDtElem() != src_dc->getNumOfDtElem()) return false;
  }
  else if (mode == 2) {
  }
  else if (mode == 3) {
    if (stepAvr == 0) return false;
  }
  else if (mode == 4) {
    if (src_dc == NULL) return false;
    if (getDataType() != src_dc->getDataType()) return false;
    if (getDims() != src_dc->getDims()) return false;
    if (getNumOfDtElem() != src_dc->getNumOfDtElem()) return false;
    if (va == NULL) return false;
  }

  register size_t i, j, k, n;
  const size_t* dst_sz_excgc = getArraySize();  // dimension size /wo guide cell
  const size_t* src_sz_excgc = NULL;
  if (src_dc != NULL) {
    src_sz_excgc = src_dc->getArraySize();
    for (i=0; i<m_dims; i++) {
      if (src_sz_excgc[i] != dst_sz_excgc[i]) return false;
    }
  }

  DATATYPE* dst_data = getData();
  const DATATYPE* src_data = NULL;
  if (src_dc != NULL) {
    src_data = src_dc->getData();
  }
  if( !dst_data ) return false;
  if (mode == 1 || mode == 4) {
    if( !src_data ) return false;
  }

  size_t dst_gc = getGc();
  size_t src_gc = 0;
  if (src_dc != NULL) {
    src_gc = src_dc->getGc();
  }
  size_t dst_size[3] = {1, 1, 1};
  size_t src_size[3] = {1, 1, 1};
  size_t src_voxlen = 1;
  size_t dst_voxlen = 1;
  for (i=0; i<m_dims; i++) {
    dst_size[i] = dst_sz_excgc[i] + dst_gc*2;
    dst_voxlen *= dst_size[i];
    if (src_dc != NULL) {
      src_size[i] = src_sz_excgc[i] + src_gc*2;
      src_voxlen *= src_size[i];
    }
  }

  SphDCType   dctype = getDcType();
  size_t data_num = getNumOfDtElem();
  size_t sta[3] = {0, 0, 0};
  int diff = 0;
  size_t dst_end[3] = {1, 1, 1};

  if (src_dc != NULL) {
    diff = src_gc - dst_gc;
    if( src_gc >= dst_gc ){
      /// コピー先ガイドセルが小さいのですべてコピーする。
      for (i=0; i<m_dims; i++) {
        dst_end[i] = dst_size[i];
        sta[i] = 0;
      }
    }
    else {
      /// コピー先ガイドセルが大きいのでガイドセルの差分を引く。
      for (i=0; i<m_dims; i++) {
        dst_end[i] = dst_size[i]+diff;
        sta[i] = abs(diff);
      }
    }
  }
  else {
    for (i=0; i<m_dims; i++) {
      if (dst_size[i] > 0) {
        dst_end[i] = dst_size[i];
      }
    }
  }

  double  ra = 1.0;
  if ( mode==2) ra=(double)stepAvr;
  else if ( mode==3 ) ra=1.0/(double)stepAvr;
  else if ( mode==4) ra=(double)stepAvr;

  for (k=sta[2]; k<dst_end[2]; k++) {
    size_t kk = k+diff;
    if (m_dims <= 2 || (long long)k+diff < 0) kk = 0;
    for(j=sta[1]; j<dst_end[1]; j++){
      size_t jj = j+diff;
      if (m_dims <= 1 || (long long)j+diff < 0) jj = 0;
      for(i=sta[0]; i<dst_end[0]; i++){
        for (n=0; n<data_num; n++) {
          size_t src_idx = 0;
          if (src_data != NULL) {
            src_idx = src_size[0]*src_size[1]*kk
                             + src_size[0]*jj
                             + i+diff;
          }
          size_t dst_idx = dst_size[0]*dst_size[1]*k
                           + dst_size[0]*j
                           + i;
          size_t src_cal_idx = 0;
          size_t dst_cal_idx = 0;
          for (n=0; n<data_num; n++) {
            if (dctype == SPH_DC_ARRAY_1DNRV
              || dctype == SPH_DC_ARRAY_2DNRV
              || dctype == SPH_DC_ARRAY_3DNRV) {
              if (src_data != NULL) src_cal_idx = src_idx * data_num + n;
              dst_cal_idx = dst_idx * data_num + n;
            }
            else {
              if (src_data != NULL) src_cal_idx = src_idx + src_voxlen*n;
              dst_cal_idx = dst_idx + dst_voxlen*n;
            }
            if (mode == 1) {
              dst_data[dst_cal_idx] += src_data[src_cal_idx];
            }
            else if (mode == 2 || mode == 3) {
              dst_data[dst_cal_idx] *= ra;
            }
            else if (mode == 4) {
              dst_data[dst_cal_idx] = src_data[src_cal_idx]*ra+va[n];
            }
          }
        }
      }
    }
  }

  return true;
}


/**
 * データすべてに設定値を設定する。
 * @param     val             設定値
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphArray<DATATYPE>::initValues(DATATYPE val)
{
  register size_t i;
  size_t array_len = getArrayLength();

  // valが0.0の場合は、memsetにより初期化する。
  if (val == (DATATYPE)0.0) {
    memset(m_dt, 0x00, sizeof(DATATYPE)*array_len);
    return true;
  }

  for (i=0; i<array_len; i++) {
    m_dt[i] = val;
  }
  return true;
}

/**
 * プロセスデータ領域のアロケートを行う。
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE*)
sphArray<DATATYPE>::allocateData(size_t array_len)
{
  if (array_len == 0) {
    array_len = getArrayLength();
  }
  if (m_dt != NULL) {
    delete [] m_dt;
  }

  m_dt = new DATATYPE[array_len];
  if( !m_dt ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_dt is null");
    return NULL;
  }
  memset(m_dt, 0, sizeof(DATATYPE)*array_len);

  return m_dt;
}


/**
 * アロケートを行っているメモリサイズを取得する。
 *
 * ボクセルサイズ x データ型サイズをバイト単位で返す。
 * @return    メモリサイズ（バイト）
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphArray<DATATYPE>::getMemoryBytes() const
{
  size_t array_len = getArrayLength();
  return array_len*sizeof(DATATYPE);
}

/**
 * データオブジェクトの添え字演算子
 * @param   index     配列インデックス
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE&)
sphArray<DATATYPE>::operator[](size_t index)
{
  if (m_dt == NULL) assert(0);
  size_t array_len = getArrayLength();
  if (index >= array_len) {
    assert(0);
  }
  return m_dt[index];
}

/**
 * データオブジェクトの添え字演算子(const)
 * @param   index     配列インデックス
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE&)
sphArray<DATATYPE>::operator[](size_t index) const
{
  if (m_dt == NULL) assert(0);
  size_t array_len = getArrayLength();
  if (index >= array_len) {
    assert(0);
  }
  return m_dt[index];
}


/**
 * データ領域のインデックスを取得する。（ガイドセルは含む領域）.
 *
 * i,j,kはガイドセルを含む領域のインデックスである。
 * 戻り値はデータ配列上のインデックスである。
 * @param[in]     index    １次配列インデックス（ガイドセルは含む）
 * @param[out]    i    I方向ノード領域インデックス（ガイドセルは含む）
 * @param[out]    j    J方向ノード領域インデックス（ガイドセルは含む）：1Dの場合はセルのデータ数
 * @param[out]    k    K方向ノード領域インデックス（ガイドセルは含む）：2Dの場合はセルのデータ数
 * @param[out]    l    セルのデータ数 (3D)
 * @return   次元数(=1 or 2 or 3 or 4)
 */
template <class DATATYPE>
SPH_MEMFUN(unsigned int)
sphArray<DATATYPE>::_getIndexDims(size_t index, size_t &i, size_t &j, size_t &k, size_t &l) const
{

  // indexからi,j,kを求める
  register unsigned n;
  size_t len = 1;
  size_t mod = index;
  if (this->m_dcType == SPH_DC_ARRAY_1DNRV
    || this->m_dcType == SPH_DC_ARRAY_2DNRV
    || this->m_dcType == SPH_DC_ARRAY_3DNRV) {

    if (this->m_pDims == 4) {
      len = this->m_pSz[0]*this->m_pSz[1]*this->m_pSz[2];
      k = (size_t)(mod/len);
      mod = mod%len;
    }
    if (this->m_pDims >= 3) {
      len = this->m_pSz[0]*this->m_pSz[1];
      j = (size_t)(mod/len);
      mod = mod%len;
    }
    if (this->m_pDims >= 2) {
      len = this->m_pSz[0];
      i = (size_t)(mod/len);
      mod = mod%len;
    }
    l = mod;
  }
  else {
    if (this->m_pDims == 4) {
      len = this->m_pSz[0]*this->m_pSz[1]*this->m_pSz[2];
      l = (size_t)(mod/len);
      mod = mod%len;
    }
    if (this->m_pDims >= 3) {
      len = this->m_pSz[0]*this->m_pSz[1];
      k = (size_t)(mod/len);
      mod = mod%len;
    }
    if (this->m_pDims >= 2) {
      len = this->m_pSz[0];
      j = (size_t)(mod/len);
      mod = mod%len;
    }
    i = mod;
  }

  return this->m_pDims;
}


#endif // _SPH_ARRAY_INLINE_H_

