/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphData_inline.h
 * Data Class Inline Source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphData_inline.h
 * データクラスインラインソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPH_DATA_INLINE_H_
#define _SPH_DATA_INLINE_H_

#define SPH_MEMFUN(rettype) \
        SPH_INLINE rettype

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
SPH_MEMFUN()
sphData::sphData() {
  m_label = std::string("");
  m_dcType = SPH_DC_INVALID;
  m_dtType = SPL_DATATYPE_NULL;
  m_paraCtrlrMap.clear();
}


/**
 * コンストラクタ
 * @param     label      データ識別ラベル
 * @param     dctype    データクラスタイプ
 * @param     dttype    データ型タイプ
 * @param     paraCtrlr   パラレルコントローラー
 */
SPH_MEMFUN()
sphData::sphData(
            const std::string label,
            SphDCType dctype,
            SPL_Datatype dttype,
            const sphParaController *paraCtrlr) {

  // データクラスの初期化を行う。
  initializeData(label, dctype, dttype, paraCtrlr);

  return;
}


/**
 * \~english
 * Destoructor
 *
 * \~japanese
 * デストラクタ
 */
SPH_MEMFUN()
sphData::~sphData()
{
}


/**
 * \~english
 * get data label
 * @return        data label
 *
 * \~japanese
 * データ識別ラベルを取得する。
 * @return        データ識別ラベル
 */
SPH_MEMFUN(std::string)
sphData::getLabel() const { return m_label; }


/**
 * \~english
 * get data label
 * @return        data label
 *
 * \~japanese
 * データ識別ラベルを取得する。
 * @return        データ識別ラベル
 */
SPH_MEMFUN(void)
sphData::setLabel(const char* label) { m_label = std::string(label); }


/**
 * \~english
 * get data class type
 * @return        data class type
 *
 * \~japanese
 * データクラスタイプを取得する。
 * @return        データクラスタイプ
 */
SPH_MEMFUN(SphDCType)
sphData::getDcType() const { return m_dcType; }


/**
 * \~english
 * get DATATYPE type
 * @return        DATATYPE type
 *
 * \~japanese
 * データ型タイプを取得する。
 * @return        データ型タイプ
 */
SPH_MEMFUN(SPL_Datatype)
sphData::getDataType() const { return m_dtType; }


/**
 * \~english
 * get parallel controller
 *
 * if not paraKey, get default parallel controller
 * @param          paraKey      parallel controller key
 * @return        parallel controller
 *
 * \~japanese
 * パラレルコントローラーを取得する。
 *
 * paraKey省略時は初期設定パラレルコントローラーを取得する。
 * @param      paraKey      パラレルコントローラー識別キー
 * @return    パラレルコントローラー
 */
SPH_MEMFUN(const sphParaController*)
sphData::getParaController(int paraKey) const
{
  unsigned int findKey = paraKey;
  if (paraKey < 0) {
    findKey = m_defualParaKey;
  }

  std::map<unsigned int, const sphParaController*>::const_iterator itr = m_paraCtrlrMap.find(findKey);
  std::map<unsigned int, const sphParaController*>::const_iterator itrEnd = m_paraCtrlrMap.end();
  if( itr == itrEnd ) return NULL;
  return itr->second;

}


/**
 * \~english
 * add parallel controller
 * @param          paraCtrlr      add parallel controller
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * パラレルマネージャの追加を行う。
 * @param   paraCtrlr    追加パラレルコントローラー
 * @return    success(=true)/fault(=false)
 */
SPH_MEMFUN(bool)
sphData::addParaCtrlr(const sphParaController* paraCtrlr)
{
  if (paraCtrlr == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "const sphParaCtrlr* paraCtrlr is null.");
    return false;
  }

  unsigned int paraKey = paraCtrlr->getParaKey();
  // check added paraKey
  const sphParaController* added_para = getParaController(paraKey);
  if (added_para != NULL) {
    // remove parallel controller
    eraseParaCtrlr(paraKey);
  }

  // add parallel controller
  int size = m_paraCtrlrMap.size();
  if (m_paraCtrlrMap.size() <= 0) {
    m_defualParaKey = paraKey;
  }
  if( !((m_paraCtrlrMap.insert(std::make_pair(paraKey, paraCtrlr))).second) ){
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not add parallel controller in data class. [paraKey=%d].", paraKey);
    return false;
  }

  return true;
}


/**
 * \~english
 * erase parallel controller
 * @param      paraKey     removed parallel controller key
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * パラレルコントローラーを削除する。
 * @param      paraKey     削除パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
SPH_MEMFUN(bool)
sphData::eraseParaCtrlr(int paraKey)
{
  unsigned int removeKey = paraKey;
  if (paraKey < 0) {
    removeKey = m_defualParaKey;
  }

  m_paraCtrlrMap.erase(removeKey);
  return true;
}


/**
 * データオブジェクトの原点座標を取得する。
 * @param[out]      org     原点座標
 * @return    success(=true)/fault(=false)
 */
SPH_MEMFUN(bool)
sphData::getDataOrigin(double org[3]) const
{
  const sphParaController* paraCtrl = getParaController();
  if (paraCtrl == NULL) {
    return false;
  }
  return paraCtrl->getNodeOrigin(org);
}


/**
 * データオブジェクトのピッチを取得する。
 * @param[out]      pitch     ピッチ
 * @return    success(=true)/fault(=false)
 */
SPH_MEMFUN(bool)
sphData::getDataPitch(double pitch[3]) const
{
  const sphParaController* paraCtrl = getParaController();
  if (paraCtrl == NULL) {
    return false;
  }
  return paraCtrl->getVoxelPitch(pitch);
}



/**
 * データオブジェクトの始点インデックスを取得する。
 * @param[out]      start_idx     始点インデックス
 * @return    success(=true)/fault(=false)
 */
SPH_MEMFUN(bool)
sphData::getDataStartIdx(size_t start_idx[3]) const
{
  const sphSvParaController* paraCtrl = dynamic_cast<const sphSvParaController*>(getParaController());
  if (paraCtrl == NULL) {
    return false;
  }
  register int i;
  int myID = paraCtrl->getMyID();
  for (i=0; i<getDims(); i++) {
    start_idx[i] = paraCtrl->getVoxelHeadIndex(myID, i);
  };
  return true;
}


/**
 * データクラス初期化を行う。
 * @param     label      データ識別ラベル
 * @param     dctype    データクラスタイプ
 * @param     dttype    データ型タイプ
 * @param     paraCtrlr   パラレルコントローラー
 */
SPH_MEMFUN(bool)
sphData::initializeData(
            const std::string label,
            SphDCType dctype,
            SPL_Datatype dttype,
            const sphParaController *paraCtrlr) {

  m_dcType = dctype;
  // check support data type
  int dt_size = spl_get_datasize(dttype);
  if (dt_size <= 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support SPL_Datatype[=%d]", dttype);
    return false;
  }
  m_dtType = dttype;
  m_label = label;
  m_paraCtrlrMap.clear();
  if (paraCtrlr) {
    if (!addParaCtrlr(paraCtrlr)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONSTRACTOR, "addParaCtrlr error.");
      return false;
    }
  }

  return true;
}


/**
 * 登録されている部分領域のボクセルデータ数を取得する。
 * @return    登録ボクセルデータ数
 */
SPH_MEMFUN(int)
sphData::getNumberOfVoxArray() const
{
  return 0;
}


/**
 * 部分領域管理データクラスであるかチェックする。
 * @return       true=部分領域管理データクラス
 */
SPH_MEMFUN(bool)
sphData::isVoxContainer() const
{
  return false;
}


/**
 * 自ボクセルデータクラスを取得する。（ローカルID検索）
 * @param    localid          ローカルID
 * @return    ボクセルデータクラス
 */
SPH_MEMFUN(sphData*)
sphData::getVoxArray(int localid)
{
  return NULL;
}


/**
 * 自ボクセルデータクラスを取得する。（ローカルID検索）
 * @param    localid          ローカルID
 * @return    ボクセルデータクラス
 */
SPH_MEMFUN(const sphData*)
sphData::getVoxArray(int localid) const
{
  return NULL;
}
#endif // _SPH_DATA_INLINE_H_

