/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphUnstData_inline.h
 * Unstruct Data Class Inline Source File
 * @author     tamaki
 * @date       2009/04/22
 *
 * \~japanese
 * @file sphUnstData_inline.h
 * 非構造格子データクラスインラインソースファイル
 * @author     tamaki
 * @date       2009/04/22
 */

#ifndef _SPH_UNST_DATA_INLINE_H_
#define _SPH_UNST_DATA_INLINE_H_

#define SPH_MEMFUN(rettype) \
        SPH_INLINE rettype


#include "spl/splmpi.h"
#include "parallel/sphUnstParaController.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
template <class DATATYPE>
SPH_MEMFUN()
sphUnstData<DATATYPE>::sphUnstData() {
  m_mesh = NULL;
  m_dtNum = 0;
  m_nPE = 0;
  m_myRank = -1;
  m_comm = SPL_COMM_NULL;
}

/**
 * \~english
 * Constructor
 * @param    label      data label
 * @param    dctype      data class type
 * @param    dttype      data type
 * @param    paraCtrlr    Parallel Controller
 * @param    vlen         data length
 *
 * \~japanese
 * コンストラクタ
 * @param    label      データ識別ラベル
 * @param    dctype      データクラスタイプ
 * @param    dttype      データ型タイプ
 * @param    paraCtrlr    パラレルコントローラー
 * @param    vlen         データ長
 */
template <class DATATYPE>
SPH_MEMFUN()
sphUnstData<DATATYPE>::sphUnstData(const std::string label,
                                   SphDCType dctype,
                                   SPL_Datatype dttype,
                                   const sphParaController *paraCtrlr,
                                   size_t vlen )
             : sphArray<DATATYPE>(label, dctype, dttype, paraCtrlr) {
  m_mesh = NULL;
  m_dtNum = 0;
  m_nPE = 0;
  m_myRank = -1;
  m_comm = SPL_COMM_NULL;
  initializeArray(dctype, dttype, paraCtrlr,vlen);
}


/**
 * \~english
 * Constructor
 * @param    dctype      data class type
 * @param    dttype      data type
 * @param    paraCtrlr    Parallel Controller
 * @param    vlen         data length
 *
 * \~japanese
 * コンストラクタ
 * @param    dctype      データクラスタイプ
 * @param    dttype      データ型タイプ
 * @param    paraCtrlr    パラレルコントローラー
 * @param    vlen         データ長
 */
template <class DATATYPE>
SPH_MEMFUN()
sphUnstData<DATATYPE>::sphUnstData(
                                   SphDCType dctype,
                                   SPL_Datatype dttype,
                                   const sphParaController *paraCtrlr,
                                   size_t vlen )
             : sphArray<DATATYPE>("", dctype, dttype, paraCtrlr) {
  m_mesh = NULL;
  m_dtNum = 0;
  m_nPE = 0;
  m_myRank = -1;
  m_comm = SPL_COMM_NULL;
  initializeArray(dctype, dttype, paraCtrlr,vlen);
}


/**
 * \~english
 * Destoructor
 *
 * \~japanese
 * デストラクタ
 */
template <class DATATYPE>
SPH_MEMFUN()
sphUnstData<DATATYPE>::~sphUnstData() {
}


template <class DATATYPE>
SPH_MEMFUN(bool)
sphUnstData<DATATYPE>::initializeArray(SphDCType dctype,
                                       SPL_Datatype dttype,
                                       const sphParaController *paraCtrlr,
                                       size_t vlen ) {
  const sphUnstParaController* unstParaCntl
                    = dynamic_cast<const sphUnstParaController*>(paraCtrlr);
  if( !unstParaCntl ) {
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "sphParaController is NULL.");
    return false;
  }

  if( !( m_mesh = const_cast<sphUnstMesh*>(unstParaCntl->getMesh()) ) ) {
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "invalid const_cast <sphUnstMesh*>");
    return false;
  }

  m_nPE = m_mesh->getNPE();
  m_myRank = m_mesh->getMyRank();
  m_comm = m_mesh->getComm();

  sphArray<DATATYPE>::m_pDims = 2;
  sphArray<DATATYPE>::m_dims = 2;
  sphArray<DATATYPE>::m_gcl[1] = 0;
  sphArray<DATATYPE>::m_gc = 0;
  switch(dctype){
    case SPH_DC_UNST_NODE:
      m_dtNum = m_mesh->nnNode;
      sphArray<DATATYPE>::m_pSz[0] = m_dtNum;
      sphArray<DATATYPE>::m_pSz[1] = 1;
      sphArray<DATATYPE>::m_dtElemNum = 1;
      break;
    case SPH_DC_UNST_NODE_N:
      m_dtNum = m_mesh->nnNode;
      sphArray<DATATYPE>::m_pSz[0] = m_dtNum;
      sphArray<DATATYPE>::m_pSz[1] = vlen;
      sphArray<DATATYPE>::m_dtElemNum = vlen;
      break;
    case SPH_DC_UNST_NODE_NRV:
      m_dtNum = m_mesh->nnNode;
      sphArray<DATATYPE>::m_pSz[0] = vlen;
      sphArray<DATATYPE>::m_pSz[1] = m_dtNum;
      sphArray<DATATYPE>::m_dtElemNum = vlen;
      break;
    case SPH_DC_UNST_ELEM:
      m_dtNum = m_mesh->nElem;
      sphArray<DATATYPE>::m_pSz[0] = m_dtNum;
      sphArray<DATATYPE>::m_pSz[1] = 1;
      sphArray<DATATYPE>::m_dtElemNum = 1;
      break;
    case SPH_DC_UNST_ELEM_N:
      m_dtNum = m_mesh->nElem;
      sphArray<DATATYPE>::m_pSz[0] = m_dtNum;
      sphArray<DATATYPE>::m_pSz[1] = vlen;
      sphArray<DATATYPE>::m_dtElemNum = vlen;
      break;
    case SPH_DC_UNST_ELEM_NRV:
      m_dtNum = m_mesh->nElem;
      sphArray<DATATYPE>::m_pSz[0] = vlen;
      sphArray<DATATYPE>::m_pSz[1] = m_dtNum;
      sphArray<DATATYPE>::m_dtElemNum = vlen;
      break;
    default:
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support dctype[=%d].", sphData::m_dcType);
      return false;
  }

  size_t array_len = 1;
  register unsigned n;
  for(n=0; n<sphArray<DATATYPE>::m_pDims; n++) array_len *= sphArray<DATATYPE>::m_pSz[n];
  if (array_len == 0) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "allocate size is zero");
    return false;
  }
  sphArray<DATATYPE>::m_dt = new DATATYPE[array_len];
  if( !sphArray<DATATYPE>::m_dt ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_dt is null");
    return false;
  }
  memset(sphArray<DATATYPE>::m_dt, 0, sizeof(DATATYPE)*array_len);

  return true;
}


/**
 * \~english
 * communication of cells.
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ガイドセルの通信を行う。
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphUnstData<DATATYPE>::comm() {
  return true;
}

template <class DATATYPE>
SPH_MEMFUN(bool)
sphUnstData<DATATYPE>::wait() {
  return true;
}

#define NAME_LEN 127

template <class DATATYPE>
SPH_MEMFUN(bool)
sphUnstData<DATATYPE>::writeFile(int istep, const char* file_prefix) {
  if( !file_prefix ) return false;
  int    i, j;
  char   filename[128];
  FILE  *fp;
  if ( m_nPE == 0 ){
    /* ------- Configure output file name ------- */
    if( !file_prefix ) return false;
    sprintf(filename, "%s.%d.dat", file_prefix, istep);
    /* ******** Open result file ******** */
    if( (fp = fopen(filename, "w")) == NULL ) return false;
    /* ******** Write variables ******** */
    fprintf(fp, "%d\n",m_dtNum);
    SphDCType dctype = sphData::getDcType();
    for ( i = 0; i < m_dtNum; i++ ){
      fprintf(fp, "%d", i);
      for( j = 0; j < sphArray<DATATYPE>::m_dtElemNum; j++) {
        size_t index;
        switch(dctype) {
        case SPH_DC_UNST_NODE:
        case SPH_DC_UNST_NODE_N:
        case SPH_DC_UNST_ELEM:
        case SPH_DC_UNST_ELEM_N:
          index = m_dtNum*j + i;
          break;
        case SPH_DC_UNST_NODE_NRV:
        case SPH_DC_UNST_ELEM_NRV:
          index = sphArray<DATATYPE>::m_dtElemNum*i + j;
          break;
        default:
          return false;
        }
        fprintf(fp, " %e", sphArray<DATATYPE>::m_dt[index]);
      }
      fprintf(fp, "\n");
    }
    fclose(fp);
  }
  else if ( m_nPE > 0 ){
    /* ------- Configure output file name ------- */
    if( !file_prefix ) return false;
    sprintf(filename, "%s.%d.%d.dat", file_prefix, istep, m_myRank);
    /* ******** Open result file ******** */
    if( (fp = fopen(filename, "w")) == NULL ) return false;
    /* ******** Write variables ******** */
    fprintf(fp, "%d\n",m_mesh->nNode);
    SphDCType dctype = sphData::getDcType();
    for ( i = 0; i < m_mesh->nNode; i++ ){
      fprintf(fp, "%d", m_mesh->node[i]);
      for( j = 0; j < sphArray<DATATYPE>::m_dtElemNum; j++) {
        size_t index;
        switch(dctype) {
        case SPH_DC_UNST_NODE:
        case SPH_DC_UNST_NODE_N:
        case SPH_DC_UNST_ELEM:
        case SPH_DC_UNST_ELEM_N:
          index = m_mesh->nNode*j + i;
          break; 
        case SPH_DC_UNST_NODE_NRV:
        case SPH_DC_UNST_ELEM_NRV:
          index = sphArray<DATATYPE>::m_dtElemNum*i + j;
          break;
        default:
          return false;
        }
        fprintf(fp, " %e", sphArray<DATATYPE>::m_dt[index]);
      }
      fprintf(fp, "\n");
    }
    fclose(fp);
  }
  return true;
}

#undef NAME_LEN

#endif // _SPH_UNST_DATA_INLINE_H_

