/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphVoxArray_inline.h
 * Voxel Data Class Inline Source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphVoxArray_inline.h
 * ボクセルデータクラスインラインソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPH_VOX_ARRAY_INLINE_H_
#define _SPH_VOX_ARRAY_INLINE_H_

#define SPH_MEMFUN(rettype) \
        SPH_INLINE rettype

#include "data/sphDataUtils.h"
#include "data/sphVoxContainer.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
template <class DATATYPE>
SPH_MEMFUN()
sphVoxArray<DATATYPE>::sphVoxArray()
{
  m_mapParaComm.clear();
  m_mapParaCommEv.clear();
  m_crddef = CRDDEF_UNKNOWN;
  m_parentContainer = NULL;
}


/**
 * コンストラクタ
 * @param    dctype    データクラスタイプ
 * @param    dttype    データ型タイプ
 * @param    label      データ識別ラベル
 * @param    paraCtrlr  パラレルコントローラー
 * @param    gc        ガイドセルサイズ
 * @param    data_num  セルのデータ数
 * @param    crddef    データ定義位置
 */
template <class DATATYPE>
SPH_MEMFUN()
sphVoxArray<DATATYPE>::sphVoxArray(
                          const std::string label,
                          SphDCType dctype,
                          SPL_Datatype dttype,
                          const sphParaController *paraCtrlr,
                          const size_t gc,
                          const size_t data_num,
                          SphCrdDef crddef)
                : sphArray<DATATYPE>(label, dctype, dttype, paraCtrlr)
{
  // ボクセルデータの初期化を行う。
  // データ配列のアロケートを行う。
  initializeVoxArray(label, dctype, dttype, paraCtrlr, gc, data_num, crddef, true);

  return;
}


/**
 * \~english
 * Constructor
 * @param    dctype      data class type
 * @param    dttype      data type
 * @param    array_size  node size
 * @param    gc        guide cell count
 * @param    data_num  data count in cell
 *
 * \~japanese
 * コンストラクタ
 * @param    dctype    データクラスタイプ
 * @param    dttype    データ型タイプ
 * @param    array_size  ノードサイズ
 * @param    gc        ガイドセルサイズ
 * @param    data_num  セルのデータ数
 */
template <class DATATYPE>
SPH_MEMFUN()
sphVoxArray<DATATYPE>::sphVoxArray(SphDCType dctype,
                                    SPL_Datatype dttype,
                                    const size_t* array_size,
                                    const size_t gc,
                                    const size_t data_num)
                : sphArray<DATATYPE>("", dctype, dttype, NULL)
{
  m_mapParaComm.clear();
  m_mapParaCommEv.clear();
  m_crddef = CRDDEF_UNKNOWN;
  m_parentContainer = NULL;

  // check support data type
  if (this->m_dtType == SPL_DATATYPE_NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "SPL_Datatype invalid.");
    return;
  }
  int dt_size = spl_get_datasize(dttype);
  if (dt_size != sizeof(DATATYPE)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not equal SPL_Datatype and DATATYPE of size [%d!=%d]. (%s[%d]:%s)", dt_size, sizeof(DATATYPE));
    return;
  }

  // Initialize Arrary
  if (!this->initializeArray(array_size, gc, data_num)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "initializeArray error");
    return;
  }

  return;
}


/**
 * \~english
 * Destoructor
 *
 * \~japanese
 * デストラクタ
 */
template <class DATATYPE>
SPH_MEMFUN()
sphVoxArray<DATATYPE>::~sphVoxArray()
{
}



/**
 * 設定されているガイドセルの通信を行う。
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::comm()
{

  const sphParaController* paraCtrlr = this->getParaController();
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    return false;
  }

  if( paraCtrlr->isEv() ) {
    // EV分割
    return commEv();
  }
  else {
    // MB分割
    return commMb();
  }

}

/**
 * 通信にてガイドセルの通信を行う。
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::comm(size_t face, int paraKey)
{
  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    return false;
  }

  if( paraCtrlr->isEv() ) {
    // EV分割
    return commEv(face, paraKey);
  }
  else {
    // MB分割
    return commMb(face, paraKey);
  }

}


/**
 * \~english
 * communication of default guide cells.
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * MB型通信にて設定されているガイドセルの通信を行う。
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commMb()
{
  return commMb(this->m_gc);
}

/**
 * MB型通信にてガイドセルの通信を行う。
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commMb(size_t face, int paraKey)
{
  // MB通信テーブルにて非同期送受信を行う。
  if( !commISendRecv(face, DIR_UNKNOWN, paraKey) ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
    return false;
  }

  // MPI通信待ちを行う。
  if( !commWait(face) ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
    return false;
  }

  return true;
}



/**
 * EV型通信にて設定されているガイドセルの通信を行う。
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commEv()
{
  return commEv(this->m_gc);
}

/**
 * EV型通信にてガイドセルの通信を行う。
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commEv(size_t face, int paraKey)
{
  // EV通信テーブルにて非同期送受信を行う。
  // 通信方向
  DirFaceType dir[3] = {X_DIR, Y_DIR, Z_DIR};   // I, J, K方向
  register int i;

  // I, J, K方向のEV型通信を順次行う。
  for (i=0; i<3; i++) {
    if( !commISendRecv(face, dir[i], paraKey) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
      return false;
    }

    // MPI通信待ちを行う。
    if( !commWait(face) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
      return false;
    }
  }

  return true;
}


/**
 * EV型の設定されているガイドセルのPeriodic通信を行う。
 * @param     face_dir       0:X面, 1:Y面, 2:Z面
 * @param     axis           -1:マイナス方向, +1:プラス方向
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commPeriodic(unsigned int face_dir, int axis)
{
  return commPeriodic(face_dir, axis, this->m_gc);
}

/**
 * EV型のガイドセルのPeriodic通信を行う。
 * @param     face_dir       0:X面, 1:Y面, 2:Z面
 * @param     axis           -1:マイナス方向, +1:プラス方向
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commPeriodic(unsigned int face_dir, int axis, size_t face, int paraKey)
{
  if (face == 0) {
    return true;
  }

  // Periodic非同期通信を行う。
  if( !commPeriodicISendRecv(face_dir, axis, face, paraKey) ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
    return false;
  }

  // MPI通信待ちを行う。
  if( !commWait(face) ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
    return false;
  }

  return true;
}


/**
 * EV型のガイドセルのPeriodicMPI通信を行う。
 *
 * @param     face_dir       0:X面, 1:Y面, 2:Z面
 * @param     axis           -1:マイナス方向, +1:プラス方向
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commPeriodicISendRecv(unsigned int face_dir, int axis, size_t face, int paraKey)
{
  if (face == 0) {
    return true;
  }

  const sphSvParaController* paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(paraKey));
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  paraKey = paraCtrlr->getParaKey();
  SPL_Comm splComm = paraCtrlr->getSplComm();
  int myId = paraCtrlr->getMyID();

  // 部分領域サブブロックの親パラレルコントローラを取得する。
  const sphMultiParaController* parent_ctrl = NULL;
  if (paraCtrlr->getParentController() != NULL) {
    parent_ctrl = dynamic_cast<const sphMultiParaController*>(paraCtrlr->getParentController());
    if( !parent_ctrl ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphMultiParaController dynamic_cast error.");
      return false;
    }
  }

  // 通信テーブルの取得
  std::vector<sphCommType*> vectCommType;
  int comm_cnt = this->getCommTypeList(&vectCommType, face, paraKey);
  if (comm_cnt <= 0) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found sphCommType (face=%d, paraKey=%d).", face, paraKey);
    return false;
  }

  void* data = (void*)this->getData();
  if (data == NULL) {
    // warning
    sphArray<DATATYPE>::m_errHandler->SPH_WARNING_INFO(SPH_WARNING_DO_NOTHING, "not null pointer in data.");
    return true;
  }

  register int i, dm;
  int dims = this->m_dims;

  // ----- Communicates COMTYPE1 begin -----
  SPL_Request req[2] = {SPL_REQUEST_NULL, SPL_REQUEST_NULL};
  SPL_Status stat[2];

  bool err_flag = false;

  // 指定方向の通信を行う。
  FaceType comm_face_send;
  FaceType comm_face_recv;
  int idir=0, jdir=0, kdir=0;
  switch( face_dir ) {
     case X_DIR:
       if (axis == PLUS_DIR) {
         comm_face_send = XFACE_PLUS;
         comm_face_recv = XFACE_MINUS;
         idir = 1;
       }
       if (axis == (int)MINUS_DIR) {
         comm_face_send = XFACE_MINUS;
         comm_face_recv = XFACE_PLUS;
         idir = -1;
       }
       break;
     case Y_DIR:
       if (axis == (int)PLUS_DIR) {
         comm_face_send = YFACE_PLUS;
         comm_face_recv = YFACE_MINUS;
         jdir = 1;
       }
       if (axis == (int)MINUS_DIR) {
         comm_face_send = YFACE_MINUS;
         comm_face_recv = YFACE_PLUS;
         idir = -1;
       }
       break;
     case Z_DIR:
       if (axis == (int)PLUS_DIR) {
         comm_face_send = ZFACE_PLUS;
         comm_face_recv = ZFACE_MINUS;
         kdir = 1;
       }
       if (axis == (int)MINUS_DIR) {
         comm_face_send = ZFACE_MINUS;
         comm_face_recv = ZFACE_PLUS;
         kdir = -1;
       }
       break;
  }

  std::vector<sphCommType*>::iterator itComm = vectCommType.begin();
  for (; itComm != vectCommType.end(); itComm++) {
    sphCommType* commType = (*itComm);

    // 最外郭で通信相手が存在する場合は、最外郭でないので除外する。
    if (commType->m_comID >= 0) {
      continue;
    }

    // 通信面
    FaceType face_type = commType->m_commDir;
    int commID = -1;
    if (face_type == comm_face_send) {
      // Periodic通信相手の取得
      commID = paraCtrlr->getCommIDPeriodic(face_dir, axis);
    }
    else if (face_type == comm_face_recv) {
      // Periodic通信相手の取得
      commID = paraCtrlr->getCommIDPeriodic(face_dir, (-1)*axis);
    }
    else {
      continue;
    }

    if (commID < 0) {
      // Periodic通信相手がいない。
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INTERNAL_IMPOSSIBLE, "can not found commid.");
      return false;
    }

    // 部分領域データ管理の場合、通信相手ノードIDからプロセス番号を取得する。
    int procID = commID;
    if (parent_ctrl != NULL) {
      if (parent_ctrl->isContainNodeId(commID)) {
        // 同一プロセス内であるので、MPI通信は行わない。
        continue;
      }
      // 別プロセスの場合、通信相手ノードIDをプロセス番号とする。
      procID = parent_ctrl->getProcessId(commID);
    }

    int tag = 1;

    if (face_type == comm_face_send) {
      // TAG(メッセージタグ)には送信ノードID+受信ノードIDを設定する。(=0xSSSSRRRR)
      tag = (myId<<16) + commID;
      // 非同期送信を行う。
      if (SPL_Isend( data,
                     1,
                     commType->m_comTypeSend,
                     procID,
                     tag,
                     splComm,
                     &(commType->m_request[0])) != SPL_SUCCESS) {
        // error
        sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "Isend error.");
        sphArray<DATATYPE>::m_errHandler->printStackTrace();
        paraCtrlr->abort((int)-1);
        return false;
      }
    }

    if (face_type == comm_face_recv) {
      // 非同期受信を行う。
      // TAG(メッセージタグ)には送信ノードID+受信ノードIDを設定する。(=0xSSSSRRRR)
      tag = (commID<<16) + myId;
      if (SPL_Irecv( data,
                     1,
                     commType->m_comTypeRecv,
                     procID,
                     tag,
                     splComm,
                     &(commType->m_request[0])) != SPL_SUCCESS) {
        // error
        sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Irecv error.");
        sphArray<DATATYPE>::m_errHandler->printStackTrace();
        paraCtrlr->abort((int)-1);
        return false;
      }
    }
  }     // for (; itComm != vectCommType.end(); itComm++)

  return true;
}



/**
 * ガイドセルのデータをPediodicコピーを行う。
 *
 * 同一プロセス内のPediodic通信対象サブブロックデータクラスに対してガイドセル領域のデータコピーを行う。
 * @param    face          通信ガイドセル数
 * @param    paraKey       パラレルコントローラー識別キー
 * @param    send_dir      送受信方向 (X_DIR = 0, Y_DIR = 1, Z_DIR = 2, DIR_UNKNOWN = -1 )
 *                         DIR_UNKNOWNの場合は、全方向
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commPediodicDataCopy(unsigned int face_dir, int axis, size_t face, int paraKey)
{
  if (face == 0) {
    return true;
  }

  // パラレルコントローラの取得を行う。
  const sphSvParaController* paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(paraKey));
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    return false;
  }
  paraKey = paraCtrlr->getParaKey();
  int myId = paraCtrlr->getMyID();

  // 部分領域サブブロックの親パラレルコントローラを取得する。
  const sphMultiParaController* parent_ctrl = NULL;
  if (paraCtrlr->getParentController() != NULL) {
    parent_ctrl = dynamic_cast<const sphMultiParaController*>(paraCtrlr->getParentController());
    if( !parent_ctrl ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphMultiParaController dynamic_cast error.");
      return false;
    }
  }

  const sphVoxContainer<DATATYPE> *parentContainer = dynamic_cast<const sphVoxContainer<DATATYPE>*>(m_parentContainer);
  if( parentContainer == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "m_parentContainer cast error.");
    return false;
  }

  // 通信テーブルの取得
  std::vector<sphCommType*> vectCommType;
  int comm_cnt = this->getCommTypeList(&vectCommType, face, paraKey);
  if (comm_cnt <= 0) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found sphCommType (face=%d, paraKey=%d).", face, paraKey);
    return NULL;
  }
  std::vector<sphCommType*>::iterator itComm = vectCommType.begin();

  // 指定方向の通信を行う。
  FaceType comm_face_send;
  FaceType comm_face_recv;
  switch( face_dir ) {
     case X_DIR:
       if (axis == PLUS_DIR) {
         comm_face_send = XFACE_PLUS;
         comm_face_recv = XFACE_MINUS;
       }
       if (axis == (int)MINUS_DIR) {
         comm_face_send = XFACE_MINUS;
         comm_face_recv = XFACE_PLUS;
       }
       break;
     case Y_DIR:
       if (axis == (int)PLUS_DIR) {
         comm_face_send = YFACE_PLUS;
         comm_face_recv = YFACE_MINUS;
       }
       if (axis == (int)MINUS_DIR) {
         comm_face_send = YFACE_MINUS;
         comm_face_recv = YFACE_PLUS;
       }
       break;
     case Z_DIR:
       if (axis == (int)PLUS_DIR) {
         comm_face_send = ZFACE_PLUS;
         comm_face_recv = ZFACE_MINUS;
       }
       if (axis == (int)MINUS_DIR) {
         comm_face_send = ZFACE_MINUS;
         comm_face_recv = ZFACE_PLUS;
       }
       break;
  }

  // ----- Communicates COMTYPE1 begin -----
  register int nnd = 0;
  for (; itComm != vectCommType.end(); itComm++) {
    sphCommType* commType = (*itComm);

    // 最外郭で通信相手が存在する場合は、最外郭でないので除外する。
    if (commType->m_comID >= 0) continue;

    // 通信面
    FaceType face_type = commType->m_commDir;
    int commID = -1;

    // 自身が受信側の場合、送信元サブブロックの取得
    if (face_type == comm_face_recv) {
      // Periodic送信元相手の取得
      commID = paraCtrlr->getCommIDPeriodic(face_dir, (-1)*axis);
    }
    else {
      continue;
    }
    if (commID < 0) continue;

    // 部分領域データ管理の場合、別プロセスの場合は、データコピーしない。
    if (parent_ctrl != NULL) {
      if (!parent_ctrl->isContainNodeId(commID)) {
        // 別プロセス内であるので、データコピーは行わない。
        continue;
      }
    }

    // データコピー相手のデータクラスの取得を行う。
    const sphVoxArray<DATATYPE> *src_data = parentContainer->getVoxArrayByNodeId(commID);
    if( src_data == NULL ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "src of sphVoxArray is null.");
      return false;
    }
    // コピー相手（=コピー元）から自データにデータコピーを行う。（受信のみ行う）
    // コピー相手（=コピー元）の送信領域
    const sphCommType* src_commType = src_data->getCommTypeByFaceType(face, paraKey, comm_face_send);
    if (src_commType == NULL) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "src of sphCommType is null.");
      return false;
    }

    // コピー相手（=コピー元）の送信領域を自データの受信領域にコピーを行う。
    if (!transferData(src_data, src_commType->m_sendStartIdx, src_commType->m_sendSize, commType->m_recvStartIdx)) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "src of sphCommType is null.");
      return false;
    }

  }

  return true;
}


/**
 * \~english
 * create communication type
 * @param    paraKey    parallel controller key
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 通信タイプの作成を行う。
 *
 * ガイドセル分通信タイプを作成する。
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::createCommType(int paraKey )
{
  m_mapParaComm.clear();

  const sphSvParaController* paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(paraKey));
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }
  paraKey = paraCtrlr->getParaKey();

  // 部分領域サブブロックの親パラレルコントローラを取得する。
  const sphMultiParaController* parent_ctrl = NULL;
  if (paraCtrlr->getParentController() != NULL) {
    parent_ctrl = dynamic_cast<const sphMultiParaController*>(paraCtrlr->getParentController());
    if( !parent_ctrl ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphMultiParaController dynamic_cast error.");
      return false;
    }
  }

  int myID = paraCtrlr->getMyID();
  if( myID < 0 ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not get myID.");
    return false;
  }

  const sphSvNodeInfo* nodeinfo = paraCtrlr->getNodeInfo();
  if (!nodeinfo) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null.");
    return false;
  }

  // 通信相手のノード数を取得する。
  int nodeNum = nodeinfo->getOffsetDivNum();
  if (nodeNum <= 0) {
    return true;
  }

  /// 通信相手のノード番号リストを作成する。
  int* comIDs = new int[nodeNum];
  nodeinfo->getOffsetDivIds(comIDs);
  if (comIDs == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "comIDs is null.");
    return false;
  }

  /// 自ノードの始点、終点インデックスを取得する。
  register int i;
  int dims = this->m_dims;
  const size_t* voxel_min = nodeinfo->getVoxelMin();
  const size_t* voxel_max_src = nodeinfo->getVoxelMax();
  size_t voxel_max[3] = {voxel_max_src[0], voxel_max_src[1], voxel_max_src[2]};

  // CRDDEF_COLLOCATEの場合、サイズが-1小さい為に、自ノードの終点インデックスとアロケートサイズの終点インデックスは異なる。
  const size_t* array_size = this->getArraySize();
  int colocate_flag[3] = {0, 0, 0};
  for (i=0; i<dims; i++) {
    size_t max_idx = voxel_min[i] + array_size[i] - 1;
    if (max_idx <= voxel_max_src[i]) {
      // CRDDEF_COLLOCATEである。
      voxel_max[i] = max_idx;
      colocate_flag[i] = voxel_max_src[i] - max_idx;
    }
  }

  size_t vcell = this->m_gc;

  // アロケートサイズ（
  size_t max[4]={0, 0, 0, 0};
  for (i=0; i<this->m_pDims; i++) max[i] = this->m_pSz[i]-1;

  CommTypeFaceMap mapCommFace;

  /// 通信面毎に派生データ型を作成する。
  register unsigned long nvc;
  register int dm;
  for (nvc=0; nvc<vcell; nvc++) {
    std::vector<sphCommType*> lstCommType;
    lstCommType.clear();

    /// 通信相手ノード毎に派生データ型を作成する。
    for (int nnd=0; nnd<nodeNum; nnd++) {
      size_t sta_comm[3]={0, 0, 0}, end_comm[3]={0, 0, 0};

      // 通信テーブル
      sphCommType* commType = new sphCommType();

      // 部分管理機能:同一プロセス内のノードIDであるかチェックする。
      bool inProcessNode = false;
      int nodeID = comIDs[nnd];     // 通信相手ノードID
      commType->m_comID = nodeID;
      commType->m_tag = nodeID;
      if (parent_ctrl != NULL) {
        inProcessNode = parent_ctrl->isContainNodeId(nodeID);
        if (!inProcessNode) {
          // 別プロセスの場合、通信相手ノードIDはプロセス番号とする。
          commType->m_comID = parent_ctrl->getProcessId(nodeID);
        }
      }
      if (commType->m_comID < 0) commType->m_comID = SPL_PROC_NULL;

      /// 通信相手ノードのボクセル情報の取得を行う。
      const sphSvNodeInfo* otherinfo = paraCtrlr->getNodeInfo(nodeID);
      const size_t* minOther = otherinfo->getVoxelMin();
      const size_t* maxOther_src = otherinfo->getVoxelMax();

      // アロケートサイズがボクセルサイズより少ない場合、自ノード、相手ノードはColocateである。
      size_t other_max[3] = {maxOther_src[0], maxOther_src[1], maxOther_src[2]};
      for (dm=0; dm<dims; dm++) {
        // 自ノードがColocateで最外郭の場合、相手ノードは自ノードサイズより大きくなることはない。
        if (voxel_max[dm] < maxOther_src[dm] && colocate_flag[dm] > 0) {
          other_max[dm] = voxel_max[dm];
        }
      }

      /// 通信面情報を取得する。
      const sphOffsetDiv* divinfo = nodeinfo->getOffsetDivInfo(nodeID);
      if (!divinfo) {
        if (comIDs) { delete[] comIDs; comIDs = NULL; }
        // error
        sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get node info. [nodeid=%d].", nodeID);
        return false;
      }
      const size_t* div_min = divinfo->getOffsetMin();
      const size_t* div_max = divinfo->getOffsetMax();
      const size_t* div_size = divinfo->getOffsetSize();

      /// 受信始点、終点インデックスを算出する。
      getReceiveCommBndIndexes(dims, vcell, nvc,
                        voxel_min, voxel_max,
                        minOther, other_max,
                        div_min, div_max,
                        sta_comm, end_comm);

      // 受信ローカルインデックス
      for (dm=0; dm<dims; dm++) {
        commType->m_recvStartIdx[dm] = sta_comm[dm];
        commType->m_recvSize[dm] = end_comm[dm]-sta_comm[dm]+1;
      }

      // プロセス内のサブブロック間通信の場合は、MPI通信データ型は作成しない。
      if (!inProcessNode) {
        /// 定義座標タイプ別に受信通信データ型を作成する。
        commType->m_comTypeRecv = createCommType(max, sta_comm, end_comm);
      }
      else {
        commType->m_comTypeRecv = SPL_DATATYPE_NULL;
      }

      /// 送信始点、終点インデックスを算出する。
      getSendCommBndIndexes(dims, vcell, nvc,
                        voxel_min, voxel_max,
                        minOther, other_max,
                        div_min, div_max,
                        sta_comm, end_comm);

      // 送信ローカルインデックス
      for (dm=0; dm<dims; dm++) {
        commType->m_sendStartIdx[dm] = sta_comm[dm];
        commType->m_sendSize[dm] = end_comm[dm]-sta_comm[dm]+1;
      }

      // プロセス内のサブブロック間通信の場合は、MPI通信データ型は作成しない。
      if (!inProcessNode) {
        /// 定義座標タイプ別に送信通信データ型を作成する。
        commType->m_comTypeSend = createCommType(max, sta_comm, end_comm);
      }
      else {
        commType->m_comTypeSend = SPL_DATATYPE_NULL;
      }
      /// 通信相手ノードの派生データ型リストに追加する。
      lstCommType.push_back(commType);

    }

    /// 通信面の派生データ型リストのテーブルに追加する。
    if( !((mapCommFace.insert(make_pair(nvc+1, lstCommType))).second) ) {
      if (comIDs) {
        delete[] comIDs;
        comIDs = NULL;
      }
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert mapCommFace. [key=%ld].", nvc+1);
      return false;
    }
  }

  if (comIDs) {
    delete[] comIDs;
    comIDs = NULL;
  }

  // register by pakaKey in map
  if( !((m_mapParaComm.insert(make_pair(paraKey, mapCommFace))).second) ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_mapParaComm. [key=%d].", paraKey);
    return false;
  }

  return true;
}


/**
 * Ev通信用の通信タイプの作成を行う。
 *
 * ガイドセル分通信タイプを作成する。
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::createCommTypeEv(int paraKey)
{
  register int i;

  const sphSvParaController* paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(paraKey));
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }
  paraKey = paraCtrlr->getParaKey();

  if( !paraCtrlr->isEv() ) {
    // 均等分割ではない。
    return true;
  }

  // 部分領域サブブロックの親パラレルコントローラを取得する。
  const sphMultiParaController* parent_ctrl = NULL;
  if (paraCtrlr->getParentController() != NULL) {
    parent_ctrl = dynamic_cast<const sphMultiParaController*>(paraCtrlr->getParentController());
    if( !parent_ctrl ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphMultiParaController dynamic_cast error.");
      return false;
    }
  }

  int myID = paraCtrlr->getMyID();
  if( myID < 0 ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not get myID.");
    return false;
  }

  const sphSvNodeInfo* nodeinfo = paraCtrlr->getNodeInfo();
  if (!nodeinfo) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null.");
    return false;
  }

  /// 自ノードの始点、終点インデックスを取得する。
  int dims = this->m_dims;
  const size_t* voxel_min = nodeinfo->getVoxelMin();
  const size_t* voxel_max_src = nodeinfo->getVoxelMax();
  size_t voxel_max[3] = {voxel_max_src[0], voxel_max_src[1], voxel_max_src[2]};

  // 自ボクセルサイズとアロケートサイズが異なる場合をチェックする。
  const size_t* array_size = this->getArraySize();
  int crddef_flag[3] = {0, 0, 0};
  for (i=0; i<dims; i++) {
    size_t max_idx = voxel_min[i] + array_size[i] - 1;
    voxel_max[i] = max_idx;
    crddef_flag[i] = voxel_max_src[i] - max_idx;
  }

  size_t vcell = this->m_gc;

  // アロケートサイズ（
  size_t max[4]={0, 0, 0, 0};
  for (i=0; i<this->m_pDims; i++) max[i] = this->m_pSz[i]-1;

  CommTypeFaceMap mapCommFace;

  /// 通信面毎に派生データ型を作成する。
  int dirMtrx[6][3] = { {-1, 0, 0},    // XFACE_MINUS
                        {+1, 0, 0},    // XFACE_PLUS
                        {0, -1, 0},    // YFACE_MINUS
                        {0, +1, 0},    // YFACE_PLUS
                        {0, 0, -1},    // ZFACE_MINUS
                        {0, 0, +1}     // ZYFACE_PLUS
                      };
  FaceType face_types[6] = {XFACE_MINUS, XFACE_PLUS, YFACE_MINUS, YFACE_PLUS, ZFACE_MINUS, ZFACE_PLUS};

  register int dir;
  register unsigned long nvc;
  register int dm;
  int commId = -1;
  for (nvc=0; nvc<vcell; nvc++) {
    std::vector<sphCommType*> lstCommType;
    lstCommType.clear();

    for (dir = 0; dir < 6; dir++) {
      size_t sta_comm[3]={0, 0, 0}, end_comm[3]={0, 0, 0};

      int idir = dirMtrx[dir][0];
      int jdir = dirMtrx[dir][1];
      int kdir = dirMtrx[dir][2];

      // 通信テーブル
      sphCommType* commType = new sphCommType();

      // 部分管理機能:同一プロセス内のノードIDであるかチェックする。
      bool inProcessNode = false;
      int nodeID = paraCtrlr->getCommIDEv(idir, jdir, kdir);     // 通信相手ノードID
      commType->m_comID = nodeID;
      commType->m_tag = nodeID;
      if (parent_ctrl != NULL) {
        inProcessNode = parent_ctrl->isContainNodeId(nodeID);
        if (!inProcessNode) {
          // 別プロセスの場合、通信相手ノードIDはプロセス番号とする。
          commType->m_comID = parent_ctrl->getProcessId(nodeID);
        }
      }

      // 通信相手ノード番号
      if (commType->m_comID < 0) commType->m_comID = SPL_PROC_NULL;

      // 通信方向を設定する。
      commType->m_commDir = face_types[dir];

      /// 受信始点、終点インデックスを算出する。
      getReceiveCommBndIndexesEv(dims, vcell, nvc, face_types[dir],
                                 voxel_min, voxel_max,
                                 sta_comm, end_comm);

      // 受信ローカルインデックス
      for (dm=0; dm<dims; dm++) {
        commType->m_recvStartIdx[dm] = sta_comm[dm];
        commType->m_recvSize[dm] = end_comm[dm]-sta_comm[dm]+1;
      }

      // プロセス内のサブブロック間通信の場合は、MPI通信データ型は作成しない。
      if (!inProcessNode) {
        /// 定義座標タイプ別に受信通信データ型を作成する。
        commType->m_comTypeRecv = createCommType(max, sta_comm, end_comm);
      }
      else {
        commType->m_comTypeRecv = SPL_DATATYPE_NULL;
      }

      /// 送信始点、終点インデックスを算出する。
      getSendCommBndIndexesEv(dims, vcell, nvc, face_types[dir],
                              voxel_min, voxel_max,
                              sta_comm, end_comm);

      // 送信ローカルインデックス
      for (dm=0; dm<dims; dm++) {
        commType->m_sendStartIdx[dm] = sta_comm[dm];
        commType->m_sendSize[dm] = end_comm[dm]-sta_comm[dm]+1;
      }

      // プロセス内のサブブロック間通信の場合は、MPI通信データ型は作成しない。
      if (!inProcessNode) {
        /// 定義座標タイプ別に送信通信データ型を作成する。
        commType->m_comTypeSend = createCommType(max, sta_comm, end_comm);
      }
      else {
        commType->m_comTypeSend = SPL_DATATYPE_NULL;
      }

      /// 通信相手ノードの派生データ型リストに追加する。
      lstCommType.push_back(commType);
    }

    if (lstCommType.size() > 0) {
      /// 通信面の派生データ型リストのテーブルに追加する。
      if( !((mapCommFace.insert(make_pair(nvc+1, lstCommType))).second) ) {
        // error
        sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert mapCommFace. [key=%ld].", nvc+1);
        return false;
      }
    }
    else {
      // EV型通信で通信データ型が存在しないことはない。
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not create comm type of EV.");
      return false;
    }
  }

  if (mapCommFace.size() > 0) {
    // register by pakaKey in map
    if( !((m_mapParaCommEv.insert(make_pair(paraKey, mapCommFace))).second) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_mapParaCommEv. [key=%d].", paraKey);
      return false;
    }
  }

  return true;
}


/**
 * \~english
 * create MPI communication type
 * @param    max    maximum communication indexes
 * @param    sta    start communication indexes
 * @param    end    end communication indexes
 * @return    communication type
 *
 * \~japanese
 * MPI通信タイプの作成を行う。
 *
 * MPIの通信タイプの作成関数を呼び出す。
 * @param    max    最大通信領域インデックス
 * @param    sta    開始通信領域インデックス
 * @param    end    終了通信領域インデックス
 * @return    通信タイプ
 */
template <class DATATYPE>
SPH_MEMFUN(SPL_Datatype)
sphVoxArray<DATATYPE>::createCommType(const size_t* max, const size_t* sta, const size_t* end)
{
  SphDCType dcType = this->m_dcType;
  SPL_Datatype dtType = this->m_dtType;
  SPL_Datatype newType = SPL_DATATYPE_NULL;

  switch (dcType) {
    case (SPH_DC_ARRAY_1D):
      splCreateCommType1D(0, max[0], sta[0], end[0], dtType, &newType);
      break;
    case (SPH_DC_ARRAY_1DN):
      splCreateCommType2D(0, max[0], 0,
                          sta[0], end[0], 0, max[1], dtType, &newType);
      break;
    case (SPH_DC_ARRAY_1DNRV):
      splCreateCommType2D(0, max[0], 0,
                          0, max[0], sta[0], end[0],
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_2D):
      splCreateCommType2D(0, max[0], 0,
                          sta[0], end[0], sta[1], end[1],
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_2DN):
      splCreateCommType3D(0, max[0], 0, max[1], 0,
                          sta[0], end[0], sta[1], end[1], 0, max[2],
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_2DNRV):
      splCreateCommType3D(0, max[0], 0, max[1], 0,
                          0, max[0], sta[0], end[0], sta[1], end[1],
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_3D):
      splCreateCommType3D(0, max[0], 0, max[1], 0,
                          sta[0], end[0], sta[1], end[1], sta[2], end[2],
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_3DN):
      splCreateCommType4D(0, max[0], 0, max[1], 0, max[2], 0,
                          sta[0], end[0], sta[1], end[1], sta[2], end[2], 0, max[3],
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_3DNRV):
      splCreateCommType4D(0, max[0], 0, max[1], 0, max[2], 0,
                          0, max[0], sta[0], end[0], sta[1], end[1], sta[2], end[2],
                          dtType, &newType);
      break;
    default:
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support dctype=%d.", (int)dcType);
      break;
  }

  return newType;
}



/**
 * MPI通信タイプの作成を行う。(Staggerdタイプ）（未使用）
 *
 * MPIの通信タイプの作成関数を呼び出す。<br>
 * u,v,wのインデックス要素別に始点、終点インデックスが異なる。
 * @param    max        最大通信領域インデックス
 * @param    sta_stg    開始通信領域インデックス(u[0], v[1], w[3]データ)
 * @param    end_stg    終了通信領域インデックス(u[0], v[1], w[3]データ）
 * @param    sta_col    開始通信領域インデックス(u[1,2], v[0,2], w[0,1]データ)
 * @param    end_col    終了通信領域インデックス(u[1,2], v[0,2], w[0,1]データ)
 * @return    通信タイプ
 */
template <class DATATYPE>
SPH_MEMFUN(SPL_Datatype)
sphVoxArray<DATATYPE>::createCommTypeStaggerd(
                            const size_t* max,
                            const size_t* sta_stg,
                            const size_t* end_stg,
                            const size_t* sta_col,
                            const size_t* end_col)
{
  SphDCType dcType = this->m_dcType;
  SPL_Datatype dtType = this->m_dtType;
  SPL_Datatype newType = SPL_DATATYPE_NULL;

  register size_t i, uvw;
  int dims = 0;

  /// ベクトルデータでない場合は対象外
  switch (dcType) {
  case (SPH_DC_ARRAY_1D):
  case (SPH_DC_ARRAY_1DN):
  case (SPH_DC_ARRAY_1DNRV):
  case (SPH_DC_ARRAY_2D):
  case (SPH_DC_ARRAY_3D):
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support dctype=%d of Staggerd crddef.", (int)dcType);
    break;
  case (SPH_DC_ARRAY_2DN):
  case (SPH_DC_ARRAY_2DNRV):
    /// 2Dの場合は、データ数は2であること。
    if (this->m_dtElemNum != 2) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support dctype=%d and dimension=%d of Staggerd crddef.", (int)dcType, this->m_dtElemNum);
      break;
    }
    dims = 2;
    break;
  case (SPH_DC_ARRAY_3DN):
  case (SPH_DC_ARRAY_3DNRV):
    /// 3Dの場合は、データ数は3であること。
    if (this->m_dtElemNum != 3) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support dctype=%d and dimension=%d of Staggerd crddef.", (int)dcType, this->m_dtElemNum);
      break;
    }
    dims = 3;
    break;
  default:
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support dctype=%d.", (int)dcType);
    break;
  }

  if (dims == 0) {
    return SPL_DATATYPE_NULL;
  }

  /// uvwの始点、終点インデックスのテーブルを作成する。
  /// u = {sta_stg, sta_col, sta_col}
  /// v = {sta_col, sta_stg, sta_col}
  /// v = {sta_col, sta_col, sta_stg}
  size_t vec_sta[3][3], vec_end[3][3];
  memset(vec_sta, 0x00, sizeof(size_t)*3*3);
  memset(vec_end, 0x00, sizeof(size_t)*3*3);
  for (uvw=0; uvw<dims; uvw++) {
    for (i=0; i<dims; i++) {
      if (uvw == i) {
        vec_sta[uvw][i] = sta_stg[i];
        vec_end[uvw][i] = end_stg[i];
      }
      else {
        vec_sta[uvw][i] = sta_col[i];
        vec_end[uvw][i] = end_col[i];
      }
    }
  }

  /// uvwの通信領域インデックスのテーブルを作成する。
  size_t min[3] = {0, 0, 0};
  size_t vec_max[3] = {0, 0, 0};
  for (i=0; i<dims; i++) vec_max[i] = max[i];

  /// MPI通信タイプの作成を行う。
  if (dcType == SPH_DC_ARRAY_2DN || dcType == SPH_DC_ARRAY_3DN) {
    splCreateCommTypeVector(dims, min, vec_max, vec_sta, vec_end, dtType, &newType);
  }
  else {
    splCreateCommTypeVectorRv(dims, min, vec_max, vec_sta, vec_end, dtType, &newType);
  }

  return newType;
}


/**
 * \~english
 * sum data count in my node
 * @param    id    value for count
 * @return    count
 *
 * \~japanese
 * 自ノード内のデータ数を取得する。（ガイドセルを含まない）
 * @param    id      カウントするデータ値
 * @return    データ数
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphVoxArray<DATATYPE>::getNumberOfID(DATATYPE id) const
{
  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return 0;
  }

  size_t num = 0;
  register size_t i, j, k;
  size_t size[3] = {1,1,1};
  size_t max[3] = {1,1,1};
  for (i=0; i<this->m_dims; i++) {
    size[i] = this->m_pSz[i];
    max[i] = this->m_pSz[i]-this->m_gcl[i];
  }
  for(k=this->m_gcl[2]; k<max[2]; k++) {
    for(j=this->m_gcl[1]; j<max[1]; j++) {
      for(i=this->m_gcl[0]; i<max[0]; i++) {
        size_t idx = size[0]*size[1]*(k)
                   + size[0]*(j)
                   + (i);
        if( this->m_dt[idx] == id ) {
          num++;
        }
      }
    }
  }
  return num;
};


/**
 * 自ノード内のデータ数を取得する。（ガイドセルを含む）
 * @param    id      カウントするデータ値
 * @return    データ数
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphVoxArray<DATATYPE>::getNumberOfIDExtGc(DATATYPE id) const
{
  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return 0;
  }

  size_t num = 0;
  size_t array_len = this->getArrayLength();
  register size_t i, j, k;
  for(i=0; i<array_len; i++) {
    if( this->m_dt[i] == id ) {
      num++;
    }
  }
  return num;
};


/**
 * \~english
 * sum data count in all node
 * @param    id    value for count
 * @param    paraKey    parallel conroller key
 * @return    count
 *
 * \~japanese
 * すべてのノード内のデータ数を取得する。（ガイドセルを含まない）
 * @param    id      カウントするデータ値
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    データ数
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphVoxArray<DATATYPE>::getWNumberOfID(DATATYPE id, int paraKey) const
{
  // 自ノードのデータ数をカウントする。
  size_t num = getNumberOfID(id);
  unsigned long tmp_num1, tmp_num2;
  tmp_num1 = (unsigned long)num;

  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if (splIsParallel() && paraCtrlr != NULL) {
    // すべてのノード内のデータ数を合計する。
    if( !paraCtrlr->allreduce(&tmp_num1, &tmp_num2, 1, SPL_UNSIGNED_LONG, SPL_SUM) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_ALLREDUCE_SUM, "paraCtrlr->allreduce error.");
      paraCtrlr->abort(-1);
      return 0;
    }
    num = (size_t)tmp_num2;
  }

  return num;
};


/**
 * \~english
 * sum data count in all node
 * @param    id    value for count
 * @param    paraKey    parallel conroller key
 * @return    count
 *
 * \~japanese
 * すべてのノード内のデータ数を取得する。（ガイドセル領域を含む）
 * @param    id      カウントするデータ値
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    データ数
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphVoxArray<DATATYPE>::getWNumberOfIDExtGc(DATATYPE id, int paraKey) const
{
  size_t num = getNumberOfIDExtGc(id);
  unsigned long tmp_num1, tmp_num2;
  tmp_num1 = (unsigned long)num;

  if (splIsParallel()) {
    const sphParaController* paraCtrlr = this->getParaController(paraKey);
    if( !paraCtrlr ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    }

    if( !paraCtrlr->allreduce(&tmp_num1, &tmp_num2, 1, SPL_UNSIGNED_LONG, SPL_SUM) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_ALLREDUCE_SUM, "paraCtrlr->allreduce error.");
      paraCtrlr->abort(-1);
      return 0;
    }
    num = (size_t)tmp_num2;
  }

  return num;
};


/**
 * コピー元データから重複領域のデータコピーを行う。
 *
 * コピー元、コピー先にはパラレルコントローラが設定されている必要がある。
 * パラレルコントローラから取得したグローバル始点インデックスから重複領域のデータをコピーする。
 * @param    src_dc      コピー元データ
 * @param    src_vc      コピー元のガイドセル数
 * @param    dest_vc    コピー先のガイドセル数
 * @param    src_paraKey    コピー元のパラレルコントローラー識別キー
 * @param    dest_paraKey   コピー先のパラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::copyVoxelArray(
                    const sphData* src_dc,
                    size_t src_vc, size_t dest_vc,
                    int src_paraKey, int dest_paraKey)
{
  register long i, j, k, l;
  if (src_dc == NULL) return false;

  // 部分領域管理データクラスであるかチェックする。
  if (src_dc->isVoxContainer()) {
    const sphVoxContainer<DATATYPE>* src_container = NULL;
    src_container = dynamic_cast<const sphVoxContainer<DATATYPE>*>(src_dc);
    // 部分領域管理データクラスからコピーを行う。
    return copyVoxelContainer(src_container, src_vc, dest_vc, src_paraKey, dest_paraKey);
  }

  const sphVoxArray<DATATYPE> *src_vox = dynamic_cast<const sphVoxArray<DATATYPE>*>(src_dc);
  if (src_vox == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphVoxArray(src_dc) dynamic_cast error.");
    return false;
  }

  const sphSvParaController* src_paraCtrlr = dynamic_cast<const sphSvParaController*>(src_vox->getParaController(src_paraKey));
  if( !src_paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController(src_paraCtrlr) dynamic_cast error.");
    return false;
  }

  const sphSvParaController* dest_paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(dest_paraKey));
  if( !dest_paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController(dest_paraCtrlr) dynamic_cast error.");
    return false;
  }

  if (src_vc > src_vox->m_gc) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "src_vc(=%d) more than src_dc::vcell(=%d).", src_vc, src_vox->m_gc);
    return false;
  }
  if (dest_vc > this->m_gc) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "dest_vc(=%d) more than this::vcell(=%d).", dest_vc, this->m_gc);
    return false;
  }
  if (this->m_dtElemNum != src_vox->m_dtElemNum) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not match data length this len(=%d) and src len(=%d).", this->m_dtElemNum, src_vox->m_dtElemNum);
    return false;
  }
  if (this->m_pDims != src_vox->m_pDims) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not match dimension this dims(=%d) and src dims(=%d).", this->m_pDims, src_vox->m_pDims);
    return false;
  }

  int src_myID = src_paraCtrlr->getMyID();
  int dest_myID = dest_paraCtrlr->getMyID();

  long src_head[3] = {0, 0, 0}, dest_head[3] = {0, 0, 0}, copy_head[3] = {0, 0, 0};
  long src_tail[3] = {0, 0, 0}, dest_tail[3] = {0, 0, 0}, copy_tail[3] = {0, 0, 0};
//  const size_t* src_min = src_paraCtrlr->getReferIndex();
//  const size_t* dest_min = dest_paraCtrlr->getReferIndex();
  for (i=0; i<this->m_dims; i++) {

    src_head[i] = src_paraCtrlr->getVoxelHeadIndex(src_myID, i);
    src_tail[i] = src_paraCtrlr->getVoxelTailIndex(src_myID, i);
    dest_head[i] = dest_paraCtrlr->getVoxelHeadIndex(dest_myID, i);
    dest_tail[i] = dest_paraCtrlr->getVoxelTailIndex(dest_myID, i);

    copy_head[i] = (src_head[i]-(long)src_vc > dest_head[i]-(long)dest_vc) ? src_head[i]-(long)src_vc : dest_head[i]-(long)dest_vc;
    copy_tail[i] = (src_tail[i]+(long)src_vc < dest_tail[i]+(long)dest_vc) ? src_tail[i]+(long)src_vc : dest_tail[i]+(long)dest_vc;

    // 重なりをチェックする
    if (src_head[i]-(long)src_vc > dest_head[i]-(long)dest_vc
        && src_head[i]-(long)src_vc > dest_tail[i]+(long)dest_vc) {
      // 重なりがない。
      return true;
    }
    if (src_tail[i]+(long)src_vc < dest_head[i]-(long)dest_vc
        && src_tail[i]+(long)src_vc < dest_tail[i]+(long)dest_vc) {
      // 重なりがない。
      return true;
    }
  }

  size_t dest_length = this->getArrayLength();
  size_t src_length = src_vox->getArrayLength();
  const size_t *dest_size = this->_getArraySize();
  const size_t *src_size = src_vox->_getArraySize();
  size_t dest_gc = this->m_gc;
  size_t src_gc = src_vox->m_gc;

  // copy data
  for (l=0;l<this->m_dtElemNum; l++) {
    for(k=copy_head[2]; k<=copy_tail[2]; k++) {
      for(j=copy_head[1]; j<=copy_tail[1]; j++) {
        for(i=copy_head[0]; i<=copy_tail[0]; i++) {
          long dest_idx = 0;
          long src_idx = 0;
          if (this->m_dims == 1) {
            dest_idx = this->_getIndex(i-dest_head[0]+dest_gc, l);
            src_idx = src_vox->_getIndex(i-src_head[0]+src_gc, l);
          }
          else if  (this->m_dims == 2) {
            dest_idx = this->_getIndex(i-dest_head[0]+dest_gc, j-dest_head[1]+dest_gc, l);
            src_idx = src_vox->_getIndex(i-src_head[0]+src_gc, j-src_head[1]+src_gc, l);
          }
          else if  (this->m_dims == 3) {
            dest_idx = this->_getIndex(i-dest_head[0]+dest_gc, j-dest_head[1]+dest_gc, k-dest_head[2]+dest_gc, l);
            src_idx = src_vox->_getIndex(i-src_head[0]+src_gc, j-src_head[1]+src_gc, k-src_head[2]+src_gc, l);
          }
          if (dest_idx < 0) continue;
          if (src_idx < 0) continue;
          if (dest_idx >= dest_length) continue;
          if (src_idx >= src_length) continue;

          this->m_dt[dest_idx] = (DATATYPE)src_vox->m_dt[src_idx];
        }
      }
    }
  }

  return true;
}



/**
 * コピー元部分領域管理データクラスから重複領域のデータコピーを行う。
 *
 * コピー元、コピー先にはパラレルコントローラが設定されている必要がある。
 * パラレルコントローラから取得したグローバル始点インデックスから重複領域のデータをコピーする。
 * @param    src_dc      コピー元部分領域管理データクラス
 * @param    src_vc      コピー元のガイドセル数
 * @param    dest_vc    コピー先のガイドセル数
 * @param    src_paraKey    コピー元のパラレルコントローラー識別キー
 * @param    dest_paraKey   コピー先のパラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::copyVoxelContainer(
                    const sphVoxArray<DATATYPE>* src_dc,
                    size_t src_vc, size_t dest_vc,
                    int src_paraKey, int dest_paraKey)
{
  register long i, j, k, l, m, n;

  if (src_dc == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "src_dc is null.");
    return false;
  }

  const sphSvParaController* dest_paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(dest_paraKey));
  if( !dest_paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController(dest_paraCtrlr) dynamic_cast error.");
    return false;
  }

  if (dest_vc > this->m_gc) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "dest_vc(=%d) more than this::vcell(=%d).", dest_vc, this->m_gc);
    return false;
  }

  int dest_myID = dest_paraCtrlr->getMyID();
  long dest_head[3] = {0, 0, 0}, dest_tail[3] = {0, 0, 0};
  for (i=0; i<this->m_dims; i++) {
    dest_head[i] = dest_paraCtrlr->getVoxelHeadIndex(dest_myID, i);
    dest_tail[i] = dest_paraCtrlr->getVoxelTailIndex(dest_myID, i);
  }

  int src_child_count = src_dc->getNumberOfVoxArray();
  // サブブロックデータをコピーする。
  for( n=0; n<src_child_count; n++) {
    const sphVoxArray<DATATYPE>* child_dc = src_dc->getVoxArray(n);

    const sphSvParaController* child_paraCtrlr = dynamic_cast<const sphSvParaController*>(child_dc->getParaController(src_paraKey));
    if( !child_paraCtrlr ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController(src_paraCtrlr) dynamic_cast error.");
      return false;
    }

    if (src_vc > child_dc->m_gc) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "src_vc(=%d) more than src_dc::vcell(=%d).", src_vc, child_dc->m_gc);
      return false;
    }

    if (this->m_dtElemNum != child_dc->m_dtElemNum) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not match data length this len(=%d) and src len(=%d).", this->m_dtElemNum, child_dc->m_dtElemNum);
      return false;
    }

    if (this->m_pDims != child_dc->m_pDims) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not match dimension this dims(=%d) and src dims(=%d).", this->m_pDims, child_dc->m_pDims);
      return false;
    }

    int src_myID = child_paraCtrlr->getMyID();

    bool isCopy = true;
    long src_head[3] = {0, 0, 0}, copy_head[3] = {0, 0, 0};
    long src_tail[3] = {0, 0, 0}, copy_tail[3] = {0, 0, 0};
    for (i=0; i<this->m_dims; i++) {

      src_head[i] = child_paraCtrlr->getVoxelHeadIndex(src_myID, i);
      src_tail[i] = child_paraCtrlr->getVoxelTailIndex(src_myID, i);

      copy_head[i] = (src_head[i]-(long)src_vc > dest_head[i]-(long)dest_vc) ? src_head[i]-(long)src_vc : dest_head[i]-(long)dest_vc;
      copy_tail[i] = (src_tail[i]+(long)src_vc < dest_tail[i]+(long)dest_vc) ? src_tail[i]+(long)src_vc : dest_tail[i]+(long)dest_vc;

      // 重なりをチェックする
      if (src_head[i]-(long)src_vc > dest_head[i]-(long)dest_vc
          && src_head[i]-(long)src_vc > dest_tail[i]+(long)dest_vc) {
        // 重なりがない。
        isCopy = false;
      }
      if (src_tail[i]+(long)src_vc < dest_head[i]-(long)dest_vc
          && src_tail[i]+(long)src_vc < dest_tail[i]+(long)dest_vc) {
        // 重なりがない。
        isCopy = false;
      }
    }
    if (!isCopy) continue;      // 重なりがない。


    size_t dest_length = this->getArrayLength();
    size_t src_length = child_dc->getArrayLength();
    const size_t *dest_size = this->_getArraySize();
    const size_t *src_size = child_dc->_getArraySize();
    size_t dest_gc = this->m_gc;
    size_t src_gc = child_dc->m_gc;

    // 通信領域
    std::vector<sphCommType*> commTypes;
    int comm_cnt= child_dc->getCommTypeList(&commTypes, src_gc, child_paraCtrlr->getParaKey());

    // copy data
    for (l=0;l<this->m_dtElemNum; l++) {
      for(k=copy_head[2]; k<=copy_tail[2]; k++) {
        for(j=copy_head[1]; j<=copy_tail[1]; j++) {
          for(i=copy_head[0]; i<=copy_tail[0]; i++) {
            long dest_idx = 0;
            long src_idx = 0;
            size_t i_idx = 0, j_idx = 0, k_idx = 0;

            // コピー元インデックス
            i_idx = i-src_head[0]+src_gc;
            j_idx = j-src_head[1]+src_gc;
            k_idx = k-src_head[2]+src_gc;
            if (this->m_dims == 1) src_idx = child_dc->_getIndex(i_idx, l);
            else if (this->m_dims == 2) src_idx = child_dc->_getIndex(i_idx, j_idx, l);
            else if  (this->m_dims == 3) src_idx = child_dc->_getIndex(i_idx, j_idx, k_idx, l);

            // 他のサブブロックと重なり領域（受信領域）はコピーしない。
            bool recv = false;
            for (m=0; m<commTypes.size(); m++) {
              sphCommType* comm = commTypes.at(m);
              if (comm == NULL) continue;
              if (comm->m_comID < 0) continue;
              if (comm->m_comTypeSend != SPL_DATATYPE_NULL) continue;
              if (comm->m_comTypeRecv != SPL_DATATYPE_NULL) continue;
              recv = true;
              // 受信データの始点インデックス, サイズ
              recv &= (   i_idx >= comm->m_recvStartIdx[0]
                       && i_idx <= comm->m_recvStartIdx[0] + comm->m_recvSize[0] - 1 );
              if (this->m_dims >= 2) {
                recv &= ( j_idx >= comm->m_recvStartIdx[1]
                       && j_idx <= comm->m_recvStartIdx[1] + comm->m_recvSize[1] - 1 );
              }
              if (this->m_dims == 3) {
                recv &= ( k_idx >= comm->m_recvStartIdx[2]
                       && k_idx <= comm->m_recvStartIdx[2] + comm->m_recvSize[2] - 1 );
              }
              if (recv) {
                break;
              }
            }
            if (recv) {
              continue;
            }

            // コピー先インデックス
            i_idx = i-dest_head[0]+dest_gc;
            j_idx = j-dest_head[1]+dest_gc;
            k_idx = k-dest_head[2]+dest_gc;
            if (this->m_dims == 1) dest_idx = this->_getIndex(i_idx, l);
            else if (this->m_dims == 2) dest_idx = this->_getIndex(i_idx, j_idx, l);
            else if  (this->m_dims == 3) dest_idx = this->_getIndex(i_idx, j_idx, k_idx, l);

            if (dest_idx < 0) continue;
            if (src_idx < 0) continue;
            if (dest_idx >= dest_length) continue;
            if (src_idx >= src_length) continue;

            this->m_dt[dest_idx] = (DATATYPE)child_dc->m_dt[src_idx];
          }
        }
      }
    }
  }

  return true;

}



/**
 * \~english
 * get boundary local index in my node
 * @param[in]  id              boundary data
 * @param[out]    bnd_sta    start boundary local index
 * @param[out]    bnd_len    size boundary area
 * @param[in]  ext_gc          extension guide cell
 * @param[in]  paraKey        parallel controller key
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 境界領域の自ノード内のローカルインデックス、サイズを取得する。（ガイドセルを含まない）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得境界領域始点ローカルインデックス（ガイドセルを含まない）
 * @param[out]    bnd_len    取得境界領域サイズ（ガイドセルを含まない）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::getBndIndex(
              DATATYPE id,
              size_t* bnd_sta,
              size_t* bnd_len,
              const size_t ext_gc,
              int paraKey) const
{
  register size_t i, j, k;

  // 部分領域データクラスの場合は、使用禁止とする。
  if (m_parentContainer != NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "can not use method from sphVoxContainer Object.");
    return false;
  }

  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return false;
  }

  if( !getBndWIndex(id, bnd_sta, bnd_len, ext_gc, paraKey) ) {
    paraCtrlr->abort(-1);
    return false;
  }

  // check
  long check = 1;
  for (int i=0; i<this->m_dims; i++) {
    check *= bnd_len[i];
  }
  if (check <= 0) {
    for (i=0; i<this->m_dims; i++) {
      bnd_sta[i] = bnd_len[i] = 0;
    }
    return true;
  }

  // get local indexes from global indexes
  if (!getGlobal2LocalIndexes(bnd_sta, bnd_len, 0, paraKey)) {
    return false;
  }

  return true;
}


/**
 * 全ノードから境界領域のインデックス、サイズを取得する。（ガイドセルを含まない）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得境界領域始点グローバルインデックス（ガイドセルを含まない）
 * @param[out]    bnd_len    取得境界領域サイズ（ガイドセルを含まない）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::getBndWIndex(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{
  register size_t i, j, k;

  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return false;
  }

  for (i=0; i<this->m_dims; i++) {
    bnd_sta[i] = bnd_len[i] = 0;
  }

  // set initialize index
  size_t my_sta[3] = {0, 0, 0};
  size_t my_end[3] = {0, 0, 0};
  size_t my_len[3] = {1, 1, 1};
  // 自領域の境界領域のローカルインデックス、サイズを取得する。（ガイドセルを含まない）
  if (!getBndMyIndex(id, my_sta, my_len, paraKey, false)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid getBndMyIndex.");
    return false;
  }

  if (!paraCtrlr->isParallel()) {
    for (i=0; i<this->m_dims; i++) {
      bnd_sta[i] = my_sta[i];
      bnd_len[i] = my_len[i];
    }
    return true;
  }

  bool bcheck = true;
  size_t my_size[3] = {0, 0, 0};

  for (i=0; i<this->m_dims; i++) {
    unsigned long mywsta = 0;
    unsigned long mywend = 0;
    if (my_len[i] > 0) {
      my_size[i] = this->m_sz[i];
      my_end[i] = my_sta[i] + my_len[i] - 1;
      int myID = paraCtrlr->getMyID();
      long headIdx = paraCtrlr->getVoxelHeadIndex(myID, i);
      if (headIdx >= 0) {
        my_sta[i] = my_sta[i] + headIdx;
        my_end[i] = my_end[i] + headIdx;

        my_sta[i] = ((long)my_sta[i] - (long)ext_gc < (long)headIdx) ? headIdx : my_sta[i] - ext_gc;
        my_end[i] = (my_end[i] + ext_gc >=my_size[i]+headIdx) ? my_size[i]+headIdx-1 : my_end[i] + ext_gc;
      }
      mywsta = my_sta[i];
      mywend = my_end[i];
    }
    else {
      mywsta = (unsigned long)(-1);
      mywend = 0;
    }
    unsigned long tmp_wsta = 0, tmp_wend = 0;
    if( !paraCtrlr->allreduce(&mywsta, &tmp_wsta, 1,
                              SPL_UNSIGNED_LONG, SPL_MIN) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "paraCtrlr->allreduce error.");
      paraCtrlr->abort(-1);
      return false;
    }
    if( !paraCtrlr->allreduce(&mywend, &tmp_wend, 1,
                              SPL_UNSIGNED_LONG, SPL_MAX) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "paraCtrlr->allreduce error.");
      paraCtrlr->abort(-1);
      return false;
    }

    my_sta[i] = tmp_wsta;
    my_end[i] = tmp_wend;

    // check index
    if (my_sta[i] > my_end[i]) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "my_sta[%d](=%ld) > my_end[%d](=%ld).", i, my_sta[i], i, my_end[i]);
      bcheck &= false;
    }
  }

  if (!bcheck) return false;

  for (i=0; i<this->m_dims; i++) {
    bnd_sta[i] = my_sta[i];
    bnd_len[i] = my_end[i] - my_sta[i] + 1;
  }
  return true;
};


/**
 * \~english
 * get multibox boundary global index in all node
 * @param[in]  id              boundary data
 * @param[out]    bnd_sta    start multibox　boundary global index
 * @param[out]    bnd_len    size multibox　boundary area
 * @param[in]  ext_gc          extension guide cell
 * @param[in]  paraKey        parallel controller key
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 全ノードからMultibox境界領域のインデックス、サイズを取得する。（ガイドセル領域を含まない）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得Multibox境界領域始点グローバルインデックス（ガイドセル領域を含まない）
 * @param[out]    bnd_len    取得Multibox境界領域サイズ（ガイドセル領域を含まない）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::getBndWIndexMb(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{

  register size_t i, j, k;
  const sphSvParaController* paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(paraKey));
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return false;
  }

  for (i=0; i<this->m_dims; i++) {
    bnd_sta[i] = bnd_len[i] = 0;
  }

  if (!paraCtrlr->isParallel()) {
    return getBndWIndex(id, bnd_sta, bnd_len, ext_gc, paraKey);
  }

  // create copy data
  sphVoxArray<DATATYPE>* obj = new sphVoxArray<DATATYPE>(
                                          "",
                                          this->getDcType(),
                                          this->getDataType(),
                                          paraCtrlr,
                                          ext_gc);
  if( !obj ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "create copy data object is null.");
    paraCtrlr->abort(-1);
    return false;
  }

  // copy data
  obj->copyVoxelArray(this, this->getGc(), obj->getGc());

  // Copy guide cell from other node
  if (obj->getGc() > 0) {
    obj->comm();
  }

  // set initialize index
  size_t my_sta[3] = {0, 0, 0};
  size_t my_end[3] = {0, 0, 0};
  size_t my_gc[3] = {0, 0, 0};
  size_t aray_size[3] = {1, 1, 1};
  size_t aray_vcsize[3] = {1, 1, 1};
  const size_t* wsz = paraCtrlr->getWholeVoxelSize();
  for (i=0; i<this->m_dims; i++) {
    my_sta[i] = wsz[i];
    aray_size[i] = obj->m_sz[i];
    aray_vcsize[i] = obj->m_sz[i] + 2 * obj->getGc();
    my_gc[i] = obj->getGc();
  }

  // find index of id
  for(k=0; k<aray_vcsize[2]; k++) {
    for(j=0; j<aray_vcsize[1]; j++) {
      for(i=0; i<aray_vcsize[0]; i++) {
        // calculate data index
        size_t idx = aray_vcsize[0]*aray_vcsize[1]*(k)
                      + aray_vcsize[0]*(j)
                      + i;

        // compare data and id
        if( obj->m_dt[idx] == (DATATYPE)id) {
          if( i < my_sta[0] ) { my_sta[0] = i; }
          if( i > my_end[0] ) { my_end[0] = i; }
          if( j < my_sta[1] ) { my_sta[1] = j; }
          if( j > my_end[1] ) { my_end[1] = j; }
          if( k < my_sta[2] ) { my_sta[2] = k; }
          if( k > my_end[2] ) { my_end[2] = k; }
        }
      }
    }
  }

  bool bcheck = true;
  int myID = paraCtrlr->getMyID();
  for (i=0; i<this->m_dims; i++) {
    if (my_sta[i] > my_end[i]) continue;

    my_sta[i] = (my_sta[i] > 2*my_gc[i]) ? (my_sta[i] - my_gc[i]) - my_gc[i] : 0;
    my_end[i] = (my_end[i] < aray_size[i]) ? (my_end[i] - my_gc[i]) + my_gc[i] : aray_size[i]-1;

    // convert gloval index
    size_t headIdx = paraCtrlr->getVoxelHeadIndex(myID, i);
    my_sta[i] += headIdx;
    my_end[i] += headIdx;

    // check index
    if (my_sta[i] > my_end[i]) bcheck &= false;
  }

  if (bcheck) {
    for (i=0; i<this->m_dims; i++) {
      bnd_sta[i] = my_sta[i];
      bnd_len[i] = my_end[i] - my_sta[i] + 1;
    }
  }

  delete obj;

  return true;
}


/**
 * \~english
 * get multibox boundary global index in node
 * @param[in]  id              boundary data
 * @param[out]    bnd_sta    start multibox　boundary local index
 * @param[out]    bnd_len    size multibox　boundary area
 * @param[in]  ext_gc          extension guide cell
 * @param[in]  paraKey        parallel controller key
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 自ノードからMultibox境界領域のローカルインデックス、サイズを取得する。（ガイドセルは含まない）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得Multibox境界領域始点ローカルインデックス（ガイドセルは含まない）
 * @param[out]    bnd_len    取得Multibox境界領域サイズ（ガイドセルは含まない）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::getBndIndexMb(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{

  register size_t i, j, k;

  // 部分領域データクラスの場合は、使用禁止とする。
  if (m_parentContainer != NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "can not use method from sphVoxContainer Object.");
    return false;
  }

  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return false;
  }

  for (i=0; i<this->m_dims; i++) {
    bnd_sta[i] = bnd_len[i] = 0;
  }

  if( !getBndWIndexMb(id, bnd_sta, bnd_len, ext_gc, paraKey) ) {
    paraCtrlr->abort(-1);
    return false;
  }

  // check
  long check = 1;
  for (int i=0; i<this->m_dims; i++) {
    check *= bnd_len[i];
  }
  if (check <= 0) {
    for (i=0; i<this->m_dims; i++) {
      bnd_sta[i] = bnd_len[i] = 0;
    }
    return true;
  }

  // get local indexes from global indexes
  if (!getGlobal2LocalIndexes(bnd_sta, bnd_len, 0, paraKey)) {
    return false;
  }

  return true;
}



/**
 * 全ノードから境界領域のインデックス、サイズを取得する。（ガイドセルを含む）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得境界領域始点グローバルインデックス（ガイドセルを含む）
 * @param[out]    bnd_len    取得境界領域サイズ（ガイドセルを含む）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::getBndWIndexExtGc(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{
  register size_t i, j, k;

  const sphSvParaController* paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(paraKey));
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return false;
  }

  for (i=0; i<this->m_dims; i++) {
    bnd_sta[i] = bnd_len[i] = 0;
  }

  // set initialize index
  size_t my_sta[3] = {0, 0, 0};
  size_t my_end[3] = {0, 0, 0};
  size_t my_len[3] = {1, 1, 1};
  // 自領域の境界領域のローカルインデックス、サイズを取得する。（ガイドセルを含む）
  if (!getBndMyIndex(id, my_sta, my_len, paraKey, true)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid getBndMyIndex.");
    return false;
  }
  if (!paraCtrlr->isParallel()) {
    for (i=0; i<this->m_dims; i++) {
      bnd_sta[i] = my_sta[i];
      bnd_len[i] = my_len[i];
    }
    return true;
  }

  bool bcheck = true;
  size_t my_size[3] = {0, 0, 0};

  for (i=0; i<this->m_dims; i++) {
    unsigned long mywsta = 0;
    unsigned long mywend = 0;

    if (my_len[i] > 0) {
      my_size[i] = this->m_sz[i]+this->m_gcl[i]*2;
      my_end[i] = my_sta[i] + my_len[i] - 1;
      int myID = paraCtrlr->getMyID();
      long headIdx = paraCtrlr->getVoxelHeadIndex(myID, i);
      if (headIdx >= 0) {
        my_sta[i] = my_sta[i] + headIdx;
        my_end[i] = my_end[i] + headIdx;

        my_sta[i] = ((long)my_sta[i] - (long)ext_gc < (long)headIdx) ? headIdx : my_sta[i] - ext_gc;
        my_end[i] = (my_end[i] + ext_gc >=my_size[i]+headIdx) ? my_size[i]+headIdx-1 : my_end[i] + ext_gc;
      }

      mywsta = my_sta[i];
      mywend = my_end[i];
    }
    else {
      // not found
      mywsta = (unsigned long)(-1);
      mywend = 0;
    }

    unsigned long tmp_wsta = 0, tmp_wend = 0;
    if( !paraCtrlr->allreduce(&mywsta, &tmp_wsta, 1,
                              SPL_UNSIGNED_LONG, SPL_MIN) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "paraCtrlr->allreduce error.");
      paraCtrlr->abort(-1);
      return false;
    }
    if( !paraCtrlr->allreduce(&mywend, &tmp_wend, 1,
                              SPL_UNSIGNED_LONG, SPL_MAX) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "paraCtrlr->allreduce error.");
      paraCtrlr->abort(-1);
      return false;
    }

    my_sta[i] = tmp_wsta;
    my_end[i] = tmp_wend;

    // check index
    if (my_sta[i] > my_end[i]) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "my_sta[%d](=%ld) > my_end[%d](=%ld).", i, my_sta[i], i, my_end[i]);
      bcheck &= false;
    }
  }

  if (!bcheck) return false;

  for (i=0; i<this->m_dims; i++) {
    bnd_sta[i] = my_sta[i];
    bnd_len[i] = my_end[i] - my_sta[i] + 1;
  }
  return true;
};


/**
 * 境界領域の自ノード内のローカルインデックス、サイズを取得する。（ガイドセルを含む）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得境界領域始点ローカルインデックス（ガイドセルを含む）
 * @param[out]    bnd_len    取得境界領域サイズ（ガイドセルを含む）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::getBndIndexExtGc(
              DATATYPE id,
              size_t* bnd_sta,
              size_t* bnd_len,
              const size_t ext_gc,
              int paraKey) const
{
  register size_t i, j, k;

  // 部分領域データクラスの場合は、使用禁止とする。
  if (m_parentContainer != NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "can not use method from sphVoxContainer Object.");
    return false;
  }

  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return false;
  }

  if( !getBndWIndexExtGc(id, bnd_sta, bnd_len, ext_gc, paraKey) ) {
    paraCtrlr->abort(-1);
    return false;
  }

  // check
  long check = 1;
  for (int i=0; i<this->m_dims; i++) {
    check *= bnd_len[i];
  }
  if (check <= 0) {
    for (i=0; i<this->m_dims; i++) {
      bnd_sta[i] = bnd_len[i] = 0;
    }
    return true;
  }

  // get local indexes from global indexes
  if (!getGlobal2LocalIndexes(bnd_sta, bnd_len, this->m_gc, paraKey)) {
    return false;
  }

  return true;
}


/**
 * 全ノードからMultibox境界領域のインデックス、サイズを取得する。（ガイドセル領域を含む）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得Multibox境界領域始点グローバルインデックス（ガイドセル領域を含む）
 * @param[out]    bnd_len    取得Multibox境界領域サイズ（ガイドセル領域を含む）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::getBndWIndexMbExtGc(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{

  register size_t i, j, k;
  const sphSvParaController* paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(paraKey));
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return false;
  }

  for (i=0; i<this->m_dims; i++) {
    bnd_sta[i] = bnd_len[i] = 0;
  }

  if (!paraCtrlr->isParallel()) {
    return getBndWIndex(id, bnd_sta, bnd_len, ext_gc, paraKey);
  }

  size_t gc = this->getGc();
  size_t obj_gc = gc;
  if (ext_gc > gc) {
    obj_gc = ext_gc;
  }
  // create copy data
  sphVoxArray<DATATYPE>* obj = new sphVoxArray<DATATYPE>(
                                          "",
                                          this->getDcType(),
                                          this->getDataType(),
                                          paraCtrlr,
                                          obj_gc);
  if( !obj ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "create copy data object is null.");
    paraCtrlr->abort(-1);
    return false;
  }

  // copy data
  obj->copyVoxelArray(this, this->getGc(), obj->getGc());

  // Copy guide cell from other node
  if (obj->getGc() > 0) {
    obj->comm(ext_gc);
  }

  // set initialize index
  size_t my_sta[3] = {0, 0, 0};
  size_t my_end[3] = {0, 0, 0};
  size_t my_gc[3] = {0, 0, 0};
  size_t aray_size[3] = {1, 1, 1};
  size_t aray_vcsize[3] = {1, 1, 1};
  const size_t* wsz = paraCtrlr->getWholeVoxelSize();
  for (i=0; i<this->m_dims; i++) {
    my_sta[i] = wsz[i];
    aray_size[i] = obj->m_sz[i];
    aray_vcsize[i] = obj->m_sz[i] + 2 * obj->getGc();
    my_gc[i] = obj->getGc();
  }

  // find index of id
  for(k=0; k<aray_vcsize[2]; k++) {
    for(j=0; j<aray_vcsize[1]; j++) {
      for(i=0; i<aray_vcsize[0]; i++) {
        // calculate data index
        size_t idx = aray_vcsize[0]*aray_vcsize[1]*(k)
                      + aray_vcsize[0]*(j)
                      + i;

        // compare data and id
        if( obj->m_dt[idx] == (DATATYPE)id) {
          if( i < my_sta[0] ) { my_sta[0] = i; }
          if( i > my_end[0] ) { my_end[0] = i; }
          if( j < my_sta[1] ) { my_sta[1] = j; }
          if( j > my_end[1] ) { my_end[1] = j; }
          if( k < my_sta[2] ) { my_sta[2] = k; }
          if( k > my_end[2] ) { my_end[2] = k; }
        }
      }
    }
  }

  bool bcheck = true;
  int myID = paraCtrlr->getMyID();
  for (i=0; i<this->m_dims; i++) {
    if (my_sta[i] > my_end[i]) continue;

    my_sta[i] = (my_sta[i]+gc > my_gc[i]+ext_gc) ? (my_sta[i]- my_gc[i]) - ext_gc + gc : 0;
    my_end[i] = ((my_end[i] - my_gc[i]) + ext_gc + gc >= aray_size[i]+gc*2)
                ? aray_size[i]+gc*2-1
                : (my_end[i] - my_gc[i]) + ext_gc + gc;

    // convert gloval index
    size_t headIdx = paraCtrlr->getVoxelHeadIndex(myID, i);
    my_sta[i] += headIdx;
    my_end[i] += headIdx;

    // check index
    if (my_sta[i] > my_end[i]) bcheck &= false;
  }

  if (bcheck) {
    for (i=0; i<this->m_dims; i++) {
      bnd_sta[i] = my_sta[i];
      bnd_len[i] = my_end[i] - my_sta[i] + 1;
    }
  }

  delete obj;

  return true;
}

/**
 * 自ノードからMultibox境界領域のローカルインデックス、サイズを取得する。（ガイドセルを含む）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得Multibox境界領域始点ローカルインデックス（ガイドセルを含む）
 * @param[out]    bnd_len    取得Multibox境界領域サイズ（ガイドセルを含む）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::getBndIndexMbExtGc(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{

  // 部分領域データクラスの場合は、使用禁止とする。
  if (m_parentContainer != NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "can not use method from sphVoxContainer Object.");
    return false;
  }

  register size_t i, j, k;
  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    return false;
  }

  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return false;
  }

  for (i=0; i<this->m_dims; i++) {
    bnd_sta[i] = bnd_len[i] = 0;
  }

  if( !getBndWIndexMbExtGc(id, bnd_sta, bnd_len, ext_gc, paraKey) ) {
    paraCtrlr->abort(-1);
    return false;
  }

  // check
  long check = 1;
  for (int i=0; i<this->m_dims; i++) {
    check *= bnd_len[i];
  }
  if (check <= 0) {
    for (i=0; i<this->m_dims; i++) {
      bnd_sta[i] = bnd_len[i] = 0;
    }
    return true;
  }

  // get local indexes from global indexes
  if (!getGlobal2LocalIndexes(bnd_sta, bnd_len, this->m_gc, paraKey)) {
    return false;
  }

  return true;
}


/**
 * グローバルインデックスから自ノードのローカルインデックス、サイズを取得する。.
 * 指定ガイドセルを含む領域のローカルインデックス、サイズを取得する。
 * つまり、指定ガイドセル＝０の時は、ガイドセルを含まない領域、
 * 指定ガイドセル＝m_gcであれば、ガイドセルを含む領域のインデックス、サイズを取得する。
 * @param[in,out]    min    始点グローバル／ローカルインデックス
 * @param[out]    len        グローバル／ローカルサイズ
 * @param[out]    gc        ガイドセル
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::getGlobal2LocalIndexes(
                              size_t* min,
                              size_t* len,
                              size_t gc,
                              int paraKey) const
{
  if (!min || !len) return false;
  const sphSvParaController* paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(paraKey));
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  int myID = paraCtrlr->getMyID();

  register int i;
  for (i=0; i<this->m_dims; i++) {
    size_t mywsta = paraCtrlr->getVoxelHeadIndex(myID, i);
    size_t mywend = paraCtrlr->getVoxelTailIndex(myID, i);
    mywend = mywend + gc*2;

    if (( min[i]+len[i]-1 < mywsta ) || (min[i] > mywend)) {
      min[i] = len[i] = 0;
    }
    else if( min[i] <= mywsta ) {
      if (min[i]+len[i]-1 > mywend) {
        len[i] = mywend - mywsta + 1;
      }
      else if (min[i]+len[i]-1 <= mywend) {
        len[i] = len[i] - (mywsta - min[i]);
      }
      else {
        len[i] = 0;
      }
      min[i] = 0;
    }
    else if ( min[i] > mywsta ) {
      if (min[i]+len[i]-1 > mywend) {
        len[i] = (mywend - mywsta + 1) - (min[i] - mywsta);
      }
      else if (min[i]+len[i]-1 <= mywend) {
        len[i] = len[i];
      }
      else {
        len[i] = 0;
      }
      min[i] = min[i] - mywsta;
    }
    else {
      min[i] = len[i] = 0;
    }
  }

  // check
  unsigned int check = 1;
  for (int i=0; i<this->m_dims; i++) {
    check *= len[i];
  }
  if (check <= 0) {
    for (int i=0; i<this->m_dims; i++) {
      min[i] = 0;
      len[i] = 0;
    }
  }

  return true;
}


/**
 * \~english
 * add parallel controller
 * @param   paraCtrlr        add parallel controller pointer
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * パラレルマネージャの追加を行う。
 * @param   paraCtrlr    追加パラレルコントローラーポインタ
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::addParaCtrlr(const sphParaController* paraCtrlr)
{
  if (paraCtrlr == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "sphParaCtrlr is null.");
    return false;
  }

  const sphSvParaController* svCtrlr = dynamic_cast<const sphSvParaController*>(paraCtrlr);
  if( !svCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  // Voxel size check
  const size_t* para_size = svCtrlr->getVoxelSize();
  const size_t* data_size = this->getArraySize();
  if (!para_size || !data_size) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "getVoxelSize[=%X] or getArraySize[=%X] is null.", para_size, data_size);
    return false;
  }
  int dim = this->getDims();
  for (int i=0; i<dim; i++) {
    if (para_size[i] != data_size[i]) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "Voxel size invalid in sphParaCtrlr.");
      return false;
    }
  }
  // Parallel Controller Key
  unsigned int paraKey = paraCtrlr->getParaKey();

  // Insert Parallel Controller in m_paraCtrlrMap
  if (!sphData::addParaCtrlr(paraCtrlr)) {
    return false;
  }

  // create comm type
  if (!this->createCommType(paraKey)) {
    return false;
  }

  return true;
}


/**
 * データ定義位置を取得する。
 * @param     void
 * @return    データ定義位置
 */
template <class DATATYPE>
SPH_MEMFUN(SphCrdDef)
sphVoxArray<DATATYPE>::getCrdDef() const
{
  return m_crddef;
}


/**
 * データ定義位置を設定する。
 * @param     crddef            データ定義位置
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::setCrdDef(SphCrdDef crddef)
{
  if (m_crddef != CRDDEF_UNKNOWN) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not set crddef, because of already set m_crddef[=%d].", (int)m_crddef);
    return false;
  }

  m_crddef = crddef;
  return true;
}


/**
 * 受信通信始点、終点インデックスを算出する。
 * @param[in]     dims              次元数
 * @param[in]     vcell             ガイドセル数
 * @param[in]     comm_face         通信面数
 * @param[in]     array_min         自ボクセル始点インデックス
 * @param[in]     array_max         自ボクセル終点インデックス
 * @param[in]     other_min         相手ボクセル始点インデックス
 * @param[in]     other_max         相手ボクセル終点インデックス
 * @param[in]     div_min           接合面始点インデックス
 * @param[in]     div_max           接合面ボクセル終点インデックス
 * @param[out]     bnd_sta           受信通信ボクセル始点インデックス
 * @param[out]     bnd_end           受信通信ボクセル終点インデックス
 * @return    void
 */
template <class DATATYPE>
SPH_MEMFUN(void)
sphVoxArray<DATATYPE>::getReceiveCommBndIndexes(
              int dims,
              size_t vcell,
              size_t comm_face,
              const size_t* array_min,
              const size_t* array_max,
              const size_t* other_min,
              const size_t* other_max,
              const size_t* div_min,
              const size_t* div_max,
              size_t* bnd_sta,
              size_t* bnd_end) const
{
  register int dm, i;
  size_t div_max_crt[3] = {0, 0, 0};
  for (dm=0; dm<dims; dm++) {
    if ( div_max[dm] > array_max[dm]) {
      div_max_crt[dm] = array_max[dm];
    }
    else {
      div_max_crt[dm] = div_max[dm];
    }
    bnd_sta[dm] = 0;
    bnd_end[dm] = 0;
  }

  // create comm contact cell
  for (dm=0; dm<dims; dm++) {
    bnd_sta[dm] = ((long long)(div_min[dm]-comm_face-1) >= (long long)other_min[dm]) ? (long long)div_min[dm]-comm_face-1 : other_min[dm];
    bnd_end[dm] = div_max_crt[dm]+comm_face+1 <= other_max[dm] ? div_max_crt[dm]+comm_face+1 : other_max[dm];

    bnd_sta[dm] = (bnd_sta[dm]-array_min[dm])+vcell;
    bnd_end[dm] = (bnd_end[dm]-array_min[dm])+vcell;

    if ((long long)(div_max_crt[dm] - (long long)other_min[dm]) == -1) {
      bnd_sta[dm] = (div_min[dm]-array_min[dm])+vcell+1;
      bnd_end[dm] = (div_max_crt[dm]-array_min[dm])+vcell+comm_face+1;
    }
    else if ((long long)(div_min[dm] - (long long)other_max) == 1) {
      bnd_sta[dm] = (div_min[dm]-array_min[dm])+vcell-comm_face-1;
      bnd_end[dm] = (div_max_crt[dm]-array_min[dm])+vcell-1;
    }
    if (bnd_end[dm] > array_max[dm] + vcell*2) {
      bnd_end[dm] = array_max[dm] + vcell*2;
    }
  }

  return;
}


/**
 * 送信通信始点、終点インデックスを算出する。
 * @param[in]     dims              次元数
 * @param[in]     vcell             ガイドセル数
 * @param[in]     comm_face         通信面数
 * @param[in]     array_min         自ボクセル始点インデックス
 * @param[in]     array_max         自ボクセル終点インデックス
 * @param[in]     other_min         相手ボクセル始点インデックス
 * @param[in]     array_max         相手ボクセル終点インデックス
 * @param[in]     div_min           接合面始点インデックス
 * @param[in]     div_max           接合面ボクセル終点インデックス
 * @param[out]     bnd_sta           送信通信ボクセル始点インデックス
 * @param[out]     bnd_end           送信通信ボクセル終点インデックス
 * @return    void
 */
template <class DATATYPE>
SPH_MEMFUN(void)
sphVoxArray<DATATYPE>::getSendCommBndIndexes(
              int dims,
              size_t vcell,
              size_t comm_face,
              const size_t* array_min,
              const size_t* array_max,
              const size_t* other_min,
              const size_t* other_max,
              const size_t* div_min,
              const size_t* div_max,
              size_t* bnd_sta,
              size_t* bnd_end) const
{
  register int dm, i;
  size_t div_max_crt[3] = {0, 0, 0};
  for (dm=0; dm<dims; dm++) {
    if ( div_max[dm] > array_max[dm]) {
      div_max_crt[dm] = array_max[dm];
    }
    else {
      div_max_crt[dm] = div_max[dm];
    }
    bnd_sta[dm] = 0;
    bnd_end[dm] = 0;
  }

  for (dm=0; dm<dims; dm++) {
    bnd_sta[dm] = ((long long)(div_min[dm]-comm_face-1) >= (long long)array_min[dm]) ? (long long)div_min[dm]-comm_face-1 : array_min[dm];
    bnd_end[dm] = ((div_max_crt[dm]+comm_face+1) <= array_max[dm]) ? div_max_crt[dm]+comm_face+1 : array_max[dm];

    bnd_sta[dm] = (bnd_sta[dm]-array_min[dm])+vcell;
    bnd_end[dm] = (bnd_end[dm]-array_min[dm])+vcell;

    if ((long long)div_max_crt[dm] - (long long)other_min[dm] == -1) {
      bnd_sta[dm] = (div_min[dm]-array_min[dm])+vcell-comm_face;
      bnd_end[dm] = (div_max_crt[dm]-array_min[dm])+vcell;
    }
    else if ((long long)div_min[dm] - (long long)other_max[dm] == 1) {
      bnd_sta[dm] = (div_min[dm]-array_min[dm])+vcell;
      bnd_end[dm] = (div_max_crt[dm]-array_min[dm])+vcell+comm_face;
    }
    if (bnd_end[dm] > array_max[dm]+vcell*2) {
      bnd_end[dm] = array_max[dm]+vcell*2;
    }
  }

  return;
}



/**
 * 受信通信始点、終点インデックスを算出する。
 * @param[in]     dims              次元数
 * @param[in]     vcell             ガイドセル数
 * @param[in]     comm_face         通信面数
 * @param[in]     face_dir          通信方向
 * @param[in]     array_min         自ボクセル始点インデックス
 * @param[in]     array_max         自ボクセル終点インデックス
 * @param[out]     bnd_sta           受信通信ボクセル始点インデックス
 * @param[out]     bnd_end           受信通信ボクセル終点インデックス
 * @return    void
 */
template <class DATATYPE>
SPH_MEMFUN(void)
sphVoxArray<DATATYPE>::getReceiveCommBndIndexesEv(
              int dims,
              size_t vcell,
              size_t comm_face,
              FaceType face_dir,
              const size_t* array_min,
              const size_t* array_max,
              size_t* bnd_sta,
              size_t* bnd_end) const
{
  register int dm, i;
  int face = 0;
  int dir = 0;
  switch (face_dir) {
    case XFACE_PLUS:
      face = 0;
      dir = 1;
      break;
    case XFACE_MINUS:
      face = 0;
      dir = -1;
      break;
    case YFACE_PLUS:
      face = 1;
      dir = 1;
      break;
    case YFACE_MINUS:
      face = 1;
      dir = -1;
      break;
    case ZFACE_PLUS:
      face = 2;
      dir = 1;
      break;
    case ZFACE_MINUS:
      face = 2;
      dir = -1;
      break;
  }

  // create comm contact cell
  for (dm=0; dm<dims; dm++) {
    if (dm == face && dir == 1) {
      bnd_sta[dm] = array_max[dm]+vcell+1;
      bnd_end[dm] = array_max[dm]+vcell+(comm_face+1);
    }
    else if (dm == face && dir == -1) {
      bnd_sta[dm] = array_min[dm]+vcell-(comm_face+1);
      bnd_end[dm] = array_min[dm]+vcell-1;
    }
    else if (dm != face) {
      bnd_sta[dm] = array_min[dm]+vcell-(comm_face+1);
      bnd_end[dm] = array_max[dm]+vcell+(comm_face+1);
    }
    // ボクセル領域内のローカルインデックスにする。
    bnd_sta[dm] -= array_min[dm];
    bnd_end[dm] -= array_min[dm];
  }

  return;
}



/**
 * 受信通信始点、終点インデックスを算出する。
 * @param[in]     dims              次元数
 * @param[in]     vcell             ガイドセル数
 * @param[in]     comm_face         通信面数
 * @param[in]     face_dir          通信方向
 * @param[in]     array_min         自ボクセル始点インデックス
 * @param[in]     array_max         自ボクセル終点インデックス
 * @param[out]     bnd_sta           受信通信ボクセル始点インデックス
 * @param[out]     bnd_end           受信通信ボクセル終点インデックス
 * @return    void
 */
template <class DATATYPE>
SPH_MEMFUN(void)
sphVoxArray<DATATYPE>::getSendCommBndIndexesEv(
              int dims,
              size_t vcell,
              size_t comm_face,
              FaceType face_dir,
              const size_t* array_min,
              const size_t* array_max,
              size_t* bnd_sta,
              size_t* bnd_end) const
{
  register int dm, i;
  int face = 0;
  int dir = 0;
  switch (face_dir) {
    case XFACE_PLUS:
      face = 0;
      dir = 1;
      break;
    case XFACE_MINUS:
      face = 0;
      dir = -1;
      break;
    case YFACE_PLUS:
      face = 1;
      dir = 1;
      break;
    case YFACE_MINUS:
      face = 1;
      dir = -1;
      break;
    case ZFACE_PLUS:
      face = 2;
      dir = 1;
      break;
    case ZFACE_MINUS:
      face = 2;
      dir = -1;
      break;
  }

  // create comm contact cell
  for (dm=0; dm<dims; dm++) {
    if (dm == face && dir == 1) {
      bnd_sta[dm] = array_max[dm]+vcell-comm_face;
      bnd_end[dm] = array_max[dm]+vcell;
    }
    else if (dm == face && dir == -1) {
      bnd_sta[dm] = array_min[dm]+vcell;
      bnd_end[dm] = array_min[dm]+vcell+comm_face;
    }
    else if (dm != face) {
      bnd_sta[dm] = array_min[dm]+vcell-(comm_face+1);
      bnd_end[dm] = array_max[dm]+vcell+(comm_face+1);
    }
    // ボクセル領域内のローカルインデックスにする。
    bnd_sta[dm] -= array_min[dm];
    bnd_end[dm] -= array_min[dm];
  }

  return;
}



/**
 * データ定義位置がサポート対象であるかチェックする。
 * @param       crddef          データ定義位置
 * @param       dctype          データクラスタイプ
 * @param       data_num        セルのデータ数
 * @return      true:サポート対象/false:サポート対象外
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::isSupportedCrdDef(
                          SphCrdDef crddef,
                          SphDCType dcType,
                          const size_t data_num)
{

  /// 定義座標タイプ別に送信通信データ型を作成する。
  if (crddef == CRDDEF_REGULAR) {
    return true;
  }
  else if (crddef == CRDDEF_COLLOCATE) {
    return true;
  }
  else if (crddef == CRDDEF_STAGGERED1 || crddef == CRDDEF_STAGGERED2) {
    /// ベクトルデータでない場合は対象外
    switch (dcType) {
    case (SPH_DC_ARRAY_1D):
    case (SPH_DC_ARRAY_1DN):
    case (SPH_DC_ARRAY_1DNRV):
    case (SPH_DC_ARRAY_2D):
    case (SPH_DC_ARRAY_3D):
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support Staggerd crddef of dctype=%d.", (int)dcType);
      return false;
    case (SPH_DC_ARRAY_2DN):
    case (SPH_DC_ARRAY_2DNRV):
      /// 2Dの場合は、データ数は2であること。
      if (data_num != 2) {
        // error
        sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support Staggerd crddef of dctype=%d, data_num=%d.", (int)dcType, data_num);
        return false;
      }
      break;
    case (SPH_DC_ARRAY_3DN):
    case (SPH_DC_ARRAY_3DNRV):
      /// 3Dの場合は、データ数は3であること。
      if (data_num != 3) {
        // error
        sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support Staggerd crddef of dctype=%d, data_num=%d.", (int)dcType, data_num);
        return false;
      }
      break;
    default:
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support Staggerd crddef of dctype=%d, data_num=%d.", (int)dcType, data_num);
      return false;
    }
  }
  else {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support crddef=%d.", (int)crddef);
    return false;

  }
  return true;
}


/**
 * ボクセルデータクラスを初期化する。
 * @param    dctype    データクラスタイプ
 * @param    dttype    データ型タイプ
 * @param    label      データ識別ラベル
 * @param    paraCtrlr  パラレルコントローラー
 * @param    gc        ガイドセルサイズ
 * @param    data_num  セルのデータ数
 * @param    crddef    データ定義位置
 * @param    allocated  データアロケートフラグ（true=アロケートを行う:default）
 * @return  true=success
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::initializeVoxArray(
                          const std::string label,
                          SphDCType dctype,
                          SPL_Datatype dttype,
                          const sphParaController *paraCtrlr,
                          const size_t gc,
                          const size_t data_num,
                          SphCrdDef crddef,
                          bool allocated)
{
  m_mapParaComm.clear();
  m_mapParaCommEv.clear();
  m_crddef = crddef;
  m_parentContainer = NULL;

  if (!paraCtrlr) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "sphParaController is null.");
    return false;
  }

  // Voxel Initlalize
  const sphSvParaController* svCtrlr = dynamic_cast<const sphSvParaController*>(paraCtrlr);
  if( !svCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
    return false;
  }

  // check support data type
  if (this->m_dtType == SPL_DATATYPE_NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "SPL_Datatype invalid.");
    return false;
  }
  int dt_size = spl_get_datasize(dttype);
  if (dt_size != sizeof(DATATYPE)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not equal SPL_Datatype and DATATYPE of size [%d!=%d].", dt_size, sizeof(DATATYPE));
    return false;
  }

  // サポート対象のデータ定義位置であるかチェックする。
  if (!isSupportedCrdDef(crddef, dctype, data_num)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support crddef[=%d].", (int)crddef);
    return false;
  }

  // Initialize Arrary
  const size_t* voxel_size = svCtrlr->getVoxelSize();
  register int i;
  int dims = sphDataUtils::getDimsOfSphDCType(dctype);
  size_t data_size[3] = {1, 1, 1};
  for (i=0; i<dims; i++) data_size[i] = voxel_size[i];

  /// データ定義位置によるデータサイズの変更を行う。
  if (crddef == CRDDEF_COLLOCATE) {
    if (svCtrlr->getNumberOfNode() == 1) {
      for (i=0; i<dims; i++) data_size[i] = data_size[i] - 1;
    }
    else {
      /// +I,J,K方向の最外郭ノードであるかチェックする。
      // collocate
      if (svCtrlr->isCommID(1, 0, 0) < 0) {
        data_size[0] = data_size[0] - 1;
      }
      if (svCtrlr->isCommID(0, 1, 0) < 0) {
        data_size[1] = data_size[1] - 1;
      }
      if (svCtrlr->isCommID(0, 0, 1) < 0) {
        data_size[2] = data_size[2] - 1;
      }
    }
  }

  if (!this->initializeArray(data_size, gc, data_num, allocated)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "initializeArray error");
    return false;
  }

  // Parallel Controller Key
  unsigned int paraKey = paraCtrlr->getParaKey();
  // create comm type
  if (!this->createCommType(paraKey)) {
    // データの破棄
    if( sphArray<DATATYPE>::m_dt ) delete [] sphArray<DATATYPE>::m_dt;
    sphArray<DATATYPE>::m_dt = NULL;
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "createCommType error");
    return false;
  }

  // create comm type for EV
  if (!this->createCommTypeEv(paraKey)) {
    // データの破棄
    if( sphArray<DATATYPE>::m_dt ) delete [] sphArray<DATATYPE>::m_dt;
    sphArray<DATATYPE>::m_dt = NULL;
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "createCommType error");
    return false;
  }

  return true;
}


/**
 * 自データから境界領域のインデックス、サイズを取得する。
 *
 * extgcFlag=trueの場合、bnd_sta、bnd_lenはガイドセルを含むインデックス、サイズとなる。
 * extgcFlag=falseの場合、bnd_sta、bnd_lenはガイドセルを含まないインデックス、サイズとなる。
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得境界領域始点ローカルインデックス
 * @param[out]    bnd_len    取得境界領域サイズ
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @param[out]    extgcFlag    ガイドセル領域をチェックするフラグ(true=ガイドセル領域を含む)
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::getBndMyIndex(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      int paraKey,
                      bool extgcFlag) const
{
  register size_t i, j, k;

  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    return false;
  }

  if (this->m_dtElemNum != 1) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "data length in cell=%d.", this->m_dtElemNum);
    return false;
  }

  for (i=0; i<this->m_dims; i++) {
    bnd_sta[i] = bnd_len[i] = 0;
  }

  // set initialize index
  size_t my_sta[3] = {0, 0, 0};
  size_t my_end[3] = {0, 0, 0};
  size_t my_size[3] = {1, 1, 1};
  size_t my_dtsize[3] = {1, 1, 1};
  size_t wsz[3] = {0, 0, 0};
  const size_t* whole_size = paraCtrlr->getWholeVoxelSize();
  for (i=0; i<this->m_dims; i++) {
    my_sta[i] = wsz[i] = whole_size[i];
    my_dtsize[i] = this->m_sz[i]+this->m_gcl[i]*2;

    // ガイドセル領域のチェックフラグ
    if (extgcFlag) {
      // ガイドセル領域もIDを走査する
      my_size[i] = this->m_sz[i]+this->m_gcl[i]*2;
    }
    else {
      // ガイドセル領域はIDを走査しない
      my_size[i] = this->m_sz[i];
    }
  }

  // find index of id
  bool found_flag = false;
  for(k=0; k<my_size[2]; k++) {
    for(j=0; j<my_size[1]; j++) {
      for(i=0; i<my_size[0]; i++) {
        // calculate data index
        size_t idx = 0;
        if (extgcFlag) {
          idx = (my_dtsize[0])*my_dtsize[1]*(k)
                   + my_dtsize[0]*(j)
                   + (i);
        }
        else {
          idx = (my_dtsize[0])*my_dtsize[1]*(k+this->m_gcl[2])
                   + my_dtsize[0]*(j+this->m_gcl[1])
                   + (i+this->m_gcl[0]);
        }

        // compare data and id
        if( this->m_dt[idx] == (DATATYPE)id) {
          if( i < my_sta[0] ) { my_sta[0] = i; }
          if( i > my_end[0] ) { my_end[0] = i; }
          if( j < my_sta[1] ) { my_sta[1] = j; }
          if( j > my_end[1] ) { my_end[1] = j; }
          if( k < my_sta[2] ) { my_sta[2] = k; }
          if( k > my_end[2] ) { my_end[2] = k; }
          found_flag = true;
        }
      }
    }
  }

  //
  if (!found_flag) {
    for (j=0; j<this->m_dims; j++) {
      bnd_sta[j] = bnd_len[j] = 0;
    }
    return true;				// idが対象範囲にない
  }

  for (i=0; i<this->m_dims; i++) {
    // check index
    if (my_sta[i] > my_end[i]) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "my_sta[%d](=%ld) > my_end[%d](=%ld).", i, my_sta[i], i, my_end[i]);

      for (j=0; j<this->m_dims; j++) {
        bnd_sta[j] = bnd_len[j] = 0;
      }
      return false;
    }
    bnd_sta[i] = my_sta[i];
    bnd_len[i] = my_end[i] - my_sta[i] + 1;
  }

  return true;
};


/**
 * ガイドセルの非同期送受信を行う。
 *
 * 送受信待ちは行わない。
 * @param    face          通信ガイドセル数
 * @param    send_dir      送受信方向 (X_DIR = 0, Y_DIR = 1, Z_DIR = 2, DIR_UNKNOWN = -1 )
 *                         DIR_UNKNOWNの場合は、全方向
 * @param    paraKey       パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commISendRecv(size_t face, DirFaceType send_dir, int paraKey)
{
  if (face == 0) {
//    spl_dbg_msg("[WARNING] face in parameter is zero. (%s[%d]:%s)", __FILE__, __LINE__, __func__ );
    return true;
  }

  // パラレルコントローラの取得を行う。
  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    return false;
  }
  // パラレルコントローラ識別IDの取得
  paraKey = paraCtrlr->getParaKey();

  SPL_Comm splComm = paraCtrlr->getSplComm();
  int myId = paraCtrlr->getMyID();

  // 通信テーブルの取得を行う。
  ParaCommMap *mapComm = NULL;
  if( paraCtrlr->isEv() ) mapComm = &m_mapParaCommEv;
  else mapComm = &m_mapParaComm;

  // 通信テーブル数のチェック
  if (mapComm->size() <= 0) {
    return true;
  }

  void* data = (void*)this->getData();
  if (data == NULL) {
    // warning
    sphArray<DATATYPE>::m_errHandler->SPH_WARNING_INFO(SPH_WARNING_DO_NOTHING, "not null pointer in data.");
    return true;
  }

  // パラレルコントローラー識別キーの通信データ型テーブルの取得
  ParaCommMap::const_iterator itrPara = mapComm->find(paraKey);
  if( itrPara == mapComm->end() ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found parallel controller (key=%d).", paraKey);
    return false;
  }
  CommTypeFaceMap mapCommType = itrPara->second;

  // 通信データ型テーブルから通信ガイドセル数の通信データ型クラスの取得
  CommTypeFaceMap::const_iterator itr = mapCommType.find(face);
  if( itr == mapCommType.end() ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found face of parameter(=%d).", face);
    return false;
  }
  std::vector<sphCommType*> vectCommType = itr->second;

  std::vector<sphCommType*>::iterator itComm = vectCommType.begin();

  // ----- Communicates COMTYPE1 begin -----
  register int nnd = 0;
  int tag = 1;
  for (; itComm != vectCommType.end(); itComm++) {
    sphCommType* commType = (*itComm);

    // MPI通信データ型が設定されていないものは除外
    if ( commType->m_comTypeSend == SPL_DATATYPE_NULL
        || commType->m_comTypeRecv == SPL_DATATYPE_NULL) {
      continue;
    }

    // 通信相手のノードIDが設定されていない。
    if (commType->m_comID < 0) continue;

    // 指定通信面の通信を行う
    if (send_dir != DIR_UNKNOWN) {
      // 通信面
      FaceType face_type = commType->m_commDir;
      if      ( (send_dir == X_DIR) && (face_type != XFACE_PLUS && face_type != XFACE_MINUS) ) continue;
      else if ( (send_dir == Y_DIR) && (face_type != YFACE_PLUS && face_type != YFACE_MINUS) ) continue;
      else if ( (send_dir == Z_DIR) && (face_type != ZFACE_PLUS && face_type != ZFACE_MINUS) ) continue;
    }

    // TAG(メッセージタグ)には送信ノードID+受信ノードIDを設定する。(=0xSSSSRRRR)
    tag = (myId<<16) + commType->m_tag;
    // 非同期送信を行う。
    if (SPL_Isend( data,
                   1,
                   commType->m_comTypeSend,
                   commType->m_comID,
                   tag,
                   splComm,
                   &(commType->m_request[0])) != SPL_SUCCESS) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "Isend error.");
      sphArray<DATATYPE>::m_errHandler->printStackTrace();
      paraCtrlr->abort(-1);
      return false;
    }

    // 非同期受信を行う。
    // TAG(メッセージタグ)には送信ノードID+受信ノードIDを設定する。(=0xSSSSRRRR)
    tag = (commType->m_tag<<16) + myId;
    if (SPL_Irecv( data,
                   1,
                   commType->m_comTypeRecv,
                   commType->m_comID,
                   tag,
                   splComm,
                   &(commType->m_request[1]) ) != SPL_SUCCESS) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Irecv error.");
      sphArray<DATATYPE>::m_errHandler->printStackTrace();
      paraCtrlr->abort(-1);
      return false;
    }
  }

  return true;
}


/**
 * ガイドセルの非同期送受信の完了待ちを行う。
 *
 * @param    face          通信ガイドセル数
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commWait(size_t face)
{
  register int i;

  // リクエスト番号を取得する。
  int req_count = getCommRequest(face, NULL);
  SPL_Request *req_list = new SPL_Request[req_count];
  req_count = getCommRequest(face, req_list);

  if (req_count <= 0) return true;

  SPL_Status *stat = new SPL_Status[req_count];
  // 送受信待ちを行う。
  if (SPL_Waitall(req_count, req_list, stat) != SPL_MPI_SUCCESS) {
    delete [] req_list; delete [] stat;
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "mpi_waitall error.");
    return false;
  }

  delete [] req_list; delete [] stat;

  // リクエスト番号のクリアを行う。
  clearCommRequest(face);

  return true;
}



/**
 * ガイドセルの非同期送受信の完了待ちを行う。
 *
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commWaitAll()
{
  register int face;
  size_t gc = this->getGc();

  // すべての通信面で通信の完了待ちを行う。
  // 本来、すべての通信面のリクエストを１回のSPL_Waitallにて行う必要があるが、
  // 通常、いずれかの通信面のみの通信であるので、commWaitを通信面でループする。
  for (face=1; face<=gc; face++) {
    if (!commWait(face)) {
      return false;
    }
  }

  return true;
}


/**
 * ガイドセルのデータコピーを行う。
 *
 * 同一プロセス内のサブブロックデータクラスに対してガイドセル領域のデータコピーを行う。
 * @param    face          通信ガイドセル数
 * @param    paraKey       パラレルコントローラー識別キー
 * @param    send_dir      送受信方向 (X_DIR = 0, Y_DIR = 1, Z_DIR = 2, DIR_UNKNOWN = -1 )
 *                         DIR_UNKNOWNの場合は、全方向
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::commDataCopy(size_t face, DirFaceType send_dir, int paraKey)
{
  if (face == 0) {
//    spl_dbg_msg("[WARNING] face in parameter is zero. (%s[%d]:%s)", __FILE__, __LINE__, __func__ );
    return true;
  }

  // パラレルコントローラの取得を行う。
  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    return false;
  }
  paraKey = paraCtrlr->getParaKey();

  int myId = paraCtrlr->getMyID();

  // 通信テーブルの取得を行う。
  ParaCommMap *mapComm = NULL;
  if( paraCtrlr->isEv() ) mapComm = &m_mapParaCommEv;
  else mapComm = &m_mapParaComm;

  // 通信テーブル数のチェック
  if (mapComm->size() <= 0) {
    return true;
  }

  void* data = (void*)this->getData();
  if (data == NULL) {
    // warning
    sphArray<DATATYPE>::m_errHandler->SPH_WARNING_INFO(SPH_WARNING_DO_NOTHING, "not null pointer in data.");
    return true;
  }

  const sphVoxContainer<DATATYPE> *parentContainer = dynamic_cast<const sphVoxContainer<DATATYPE>*>(m_parentContainer);
  if( parentContainer == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "m_parentContainer cast error.");
    return false;
  }

  // パラレルコントローラー識別キーの通信データ型テーブルの取得
  ParaCommMap::const_iterator itrPara = mapComm->find(paraKey);
  if( itrPara == mapComm->end() ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found parallel controller (key=%d).", paraKey);
    return false;
  }
  CommTypeFaceMap mapCommType = itrPara->second;

  // 通信データ型テーブルから通信ガイドセル数の通信データ型クラスの取得
  CommTypeFaceMap::const_iterator itr = mapCommType.find(face);
  if( itr == mapCommType.end() ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found face of parameter(=%d).", face);
    return false;
  }
  std::vector<sphCommType*> vectCommType = itr->second;

  std::vector<sphCommType*>::iterator itComm = vectCommType.begin();

  // ----- Communicates COMTYPE1 begin -----
  register int nnd = 0;
  for (; itComm != vectCommType.end(); itComm++) {
    sphCommType* commType = (*itComm);

    // MPI通信データ型が設定されているものは除外
    if ( commType->m_comTypeSend != SPL_DATATYPE_NULL
        || commType->m_comTypeRecv != SPL_DATATYPE_NULL) {
      continue;
    }

    // 通信相手のノードIDが設定されていない。
    if (commType->m_comID < 0) continue;

    // 指定通信面の通信を行う
    if (send_dir != DIR_UNKNOWN) {
      // 通信面
      FaceType face_type = commType->m_commDir;
      if      ( (send_dir == X_DIR) && (face_type != XFACE_PLUS && face_type != XFACE_MINUS) ) continue;
      else if ( (send_dir == Y_DIR) && (face_type != YFACE_PLUS && face_type != YFACE_MINUS) ) continue;
      else if ( (send_dir == Z_DIR) && (face_type != ZFACE_PLUS && face_type != ZFACE_MINUS) ) continue;
    }

    // データコピー相手のデータクラスの取得を行う。
    const sphVoxArray<DATATYPE> *src_data = parentContainer->getVoxArrayByNodeId(commType->m_comID);
    if( src_data == NULL ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "src of sphVoxArray is null.");
      return false;
    }
    // コピー相手（=コピー元）から自データにデータコピーを行う。（受信のみ行う）
    // コピー相手（=コピー元）の送信領域
    const sphCommType* src_commType = src_data->getCommType(face, paraKey, myId);
    if (src_commType == NULL) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "src of sphCommType is null.");
      return false;
    }

    // コピー相手（=コピー元）の送信領域を自データの受信領域にコピーを行う。
    if (!transferData(src_data, src_commType->m_sendStartIdx, src_commType->m_sendSize, commType->m_recvStartIdx)) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "src of sphCommType is null.");
      return false;
    }

  }

  return true;
}



/**
 * 通信テーブルから送受信情報を取得する。
 *
 * @param    face          通信ガイドセル数
 * @param    paraKey       パラレルコントローラー識別キー
 * @param    nodeid        送受信ノードID
 * @return    送受信情報(=sphCommType)
 */
template <class DATATYPE>
SPH_MEMFUN(const sphCommType*)
sphVoxArray<DATATYPE>::getCommType(size_t face, int paraKey, int nodeid) const
{
  if (face == 0) {
    return NULL;
  }

  // パラレルコントローラの取得を行う。
  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    return NULL;
  }
  paraKey = paraCtrlr->getParaKey();

  int myId = paraCtrlr->getMyID();

  // 通信テーブルの取得を行う。
  const ParaCommMap *mapComm = NULL;
  if( paraCtrlr->isEv() ) mapComm = &m_mapParaCommEv;
  else mapComm = &m_mapParaComm;

  // 通信テーブル数のチェック
  if (mapComm->size() <= 0) {
    return NULL;
  }

  // パラレルコントローラー識別キーの通信データ型テーブルの取得
  ParaCommMap::const_iterator itrPara = mapComm->find(paraKey);
  if( itrPara == mapComm->end() ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found parallel controller (key=%d).", paraKey);
    return NULL;
  }
  CommTypeFaceMap mapCommType = itrPara->second;

  // 通信データ型テーブルから通信ガイドセル数の通信データ型クラスの取得
  CommTypeFaceMap::const_iterator itr = mapCommType.find(face);
  if( itr == mapCommType.end() ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found face of parameter(=%d).", face);
    return NULL;
  }
  std::vector<sphCommType*> vectCommType = itr->second;

  std::vector<sphCommType*>::iterator itComm = vectCommType.begin();

  // ----- Communicates COMTYPE1 begin -----
  register int nnd = 0;
  for (; itComm != vectCommType.end(); itComm++) {
    sphCommType* commType = (*itComm);
    if (commType->m_comID == nodeid) {
      return commType;
    }
  }

  return NULL;
}


/**
 * 通信テーブルから送受信情報を取得する。
 *
 * @param    face          通信ガイドセル数
 * @param    paraKey       パラレルコントローラー識別キー
 * @param    nodeid        送受信ノードID
 * @return    送受信情報(=sphCommType)
 */
template <class DATATYPE>
SPH_MEMFUN(const sphCommType*)
sphVoxArray<DATATYPE>::getCommTypeByFaceType(size_t face, int paraKey, FaceType facetype) const
{
  if (face == 0) {
    return NULL;
  }

  // 通信テーブルの取得
  std::vector<sphCommType*> commTypes;
  int comm_cnt = this->getCommTypeList(&commTypes, face, paraKey);
  if (comm_cnt <= 0) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found sphCommType (face=%d, paraKey=%d).", face, paraKey);
    return NULL;
  }

  std::vector<sphCommType*>::iterator itComm = commTypes.begin();
  register int nnd = 0;
  for (; itComm != commTypes.end(); itComm++) {
    sphCommType* commType = (*itComm);
    if (commType->m_commDir == facetype) {
      return commType;
    }
  }

  return NULL;
}



/**
 * 通信テーブルから送受信情報テーブルを取得する。
 *
 * @param    face          通信ガイドセル数
 * @param    paraKey       パラレルコントローラー識別キー
 * @return    送受信情報(=sphCommType)
 */
template <class DATATYPE>
SPH_MEMFUN(int)
sphVoxArray<DATATYPE>::getCommTypeList(std::vector<sphCommType*>* commTypes, size_t face, int paraKey) const
{
  if (face == 0) {
    return 0;
  }

  // パラレルコントローラの取得を行う。
  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    return 0;
  }
  paraKey = paraCtrlr->getParaKey();

  int myId = paraCtrlr->getMyID();

  // 通信テーブルの取得を行う。
  const ParaCommMap *mapComm = NULL;
  if( paraCtrlr->isEv() ) mapComm = &m_mapParaCommEv;
  else mapComm = &m_mapParaComm;

  // 通信テーブル数のチェック
  if (mapComm->size() <= 0) {
    return 0;
  }

  // パラレルコントローラー識別キーの通信データ型テーブルの取得
  ParaCommMap::const_iterator itrPara = mapComm->find(paraKey);
  if( itrPara == mapComm->end() ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found parallel controller (key=%d).", paraKey);
    return 0;
  }
  CommTypeFaceMap mapCommType = itrPara->second;

  // 通信データ型テーブルから通信ガイドセル数の通信データ型クラスの取得
  CommTypeFaceMap::const_iterator itr = mapCommType.find(face);
  if( itr == mapCommType.end() ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_FOUND_IN_LIST, "not found face of parameter(=%d).", face);
    return 0;
  }

  std::vector<sphCommType*> vectCommType = itr->second;

  *commTypes = vectCommType;

  return commTypes->size();
}

/**
 * 親データクラス(=部分領域管理データクラス)を設定する。
 *
 * @param   parentContainer     親データクラス(=sphVoxContainer)
 */
template <class DATATYPE>
SPH_MEMFUN(void)
sphVoxArray<DATATYPE>::setParentContainer(const sphVoxArray<DATATYPE> *parentContainer)
{
  m_parentContainer = parentContainer;
}


/**
 * コピー元データの指定領域を自データにコピーする。
 *
 * 始点インデックスはガイドセルを含むローカルインデックスである。
 * コピーサイズはコピー元サイズであるが、自データ領域（ガイドセルを含む）外の領域はコピーしない。
 * @param   src_dc    コピー元データ
 * @param   src_startIdx    コピー元ローカル始点インデックス（ガイドセルを含む）
 * @param   src_startIdx    コピー元サイズ（ガイドセルを含む）
 * @param   dest_startIdx    コピー先ローカル始点インデックス（ガイドセルを含む）
 * @return    true=success
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxArray<DATATYPE>::transferData(
                            const sphVoxArray<DATATYPE> *src_dc,
                            const size_t  src_startIdx[3],
                            const size_t  src_len[3],
                            const size_t  dest_startIdx[3])
{
  register size_t i;
  int dims = this->getDims();

  // データ数のチェックを行う。
  size_t src_elem = src_dc->getNumOfDtElem();
  size_t dest_elem = this->getNumOfDtElem();
  if (src_elem != dest_elem) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "not match data elem[src=%d, dest=%d].", src_elem, dest_elem);
    return false;
  }

  // データサイズ
  const size_t *src_size = src_dc->_getArraySize();
  const size_t *dest_size = this->_getArraySize();

  // 3次元データとする（２次元データの場合、３次元目は不定の為０とする）
  size_t cp_src_startIdx[3] = {0, 0, 0};
  size_t cp_src_endIdx[3] = {0, 0, 0};
  size_t cp_dest_startIdx[3] = {0, 0, 0};
  size_t cp_dest_endIdx[3] = {0, 0, 0};
  for (i=0; i<dims; i++) {
    cp_src_startIdx[i] = src_startIdx[i];
    cp_src_endIdx[i] = cp_src_startIdx[i] + src_len[i] - 1;
    cp_dest_startIdx[i] = dest_startIdx[i];
    cp_dest_endIdx[i] = cp_dest_startIdx[i] + src_len[i] - 1;

    // インデックスチェック
    // 始点インデックスが領域外の場合終了する（エラーとはしない）
    if (cp_src_startIdx[i] >= src_size[i]) return true;
    if (cp_src_endIdx[i] >= src_size[i]) {
      cp_src_endIdx[i] = src_size[i]-1;
    }
    if (cp_dest_startIdx[i] >= dest_size[i])  return true;
    if (cp_dest_endIdx[i] >= dest_size[i]) {
      cp_dest_endIdx[i] = dest_size[i]-1;
    }
  }

  // データ配列長
  size_t src_arraylen = src_dc->getArrayLength();
  size_t dest_arraylen = this->getArrayLength();

  register size_t si, sj, sk, di, dj, dk, l;
  for (l=0; l<src_elem; l++) {
    for (sk=cp_src_startIdx[2], dk=cp_dest_startIdx[2];
            sk<=cp_src_endIdx[2] && dk<=cp_dest_endIdx[2];
            sk++, dk++) {

      for (sj=cp_src_startIdx[1], dj=cp_dest_startIdx[1];
              sj<=cp_src_endIdx[1] && dj<=cp_dest_endIdx[1];
              sj++, dj++) {

        for (si=cp_src_startIdx[0], di=cp_dest_startIdx[0];
                si<=cp_src_endIdx[0] && di<=cp_dest_endIdx[0];
                si++, di++) {
          // コピー元インデックス
          size_t src_idx = 0;
          if (dims == 1) src_idx = src_dc->_getIndex(si, l, 0, 0);
          else if (dims == 2) src_idx = src_dc->_getIndex(si, sj, l, 0);
          else if (dims == 3) src_idx = src_dc->_getIndex(si, sj, sk, l);

          // コピー先インデックス
          size_t dest_idx = 0;
          if (dims == 1) dest_idx = this->_getIndex(di, l, 0, 0);
          else if (dims == 2) dest_idx = this->_getIndex(di, dj, l, 0);
          else if (dims == 3) dest_idx = this->_getIndex(di, dj, dk, l);

          if (src_idx >= src_arraylen) continue;
          if (dest_idx >= dest_arraylen) continue;

          // データコピー
          this->m_dt[dest_idx] = src_dc->m_dt[src_idx];
        }
      }
    }
  }

  return true;
}


/**
 * ボクセルデータクラスの複製を生成する。
 *
 * 生成データクラスの破棄は呼出し側にて行う必要がある。
 * データ識別ラベルは設定しないので、データ識別ラベルの設定、データマネージャへの登録は呼出し側にて行う必要がある。
 * @param   new_gc      生成ガイドセル ( <0 = 元ガイドセルにて生成する。)
 * @param   copy_data   データコピーフラグ (true = 元ボクセルデータをコピーする)
 * @return ボクセルデータオブジェクト
 */
template <class DATATYPE>
SPH_MEMFUN(sphData*)
sphVoxArray<DATATYPE>::createVoxelArray(int new_gc, bool copy_data)
{
  // データクラス作成
  sphVoxArray<DATATYPE>* voxel = NULL;

  // 生成するガイドセルを決定する。
  size_t gc = this->m_gc;
  if (new_gc >= 0) gc = new_gc;

  // データ識別ラベルは設定しない。
  voxel = new sphVoxArray<DATATYPE>(
                              "", this->m_dcType, this->m_dtType,
                              this->getParaController(),
                              gc,
                              this->m_dtElemNum,
                              m_crddef);

  if (voxel == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  else {
    if (voxel->getDataPointer() == NULL) {
      delete voxel;
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data array.");
      return NULL;
    }
  }

  // データコピーを行う。
  if (copy_data && this->m_dt != NULL) {
    voxel->copyArray(this);
  }

  return voxel;
}


/**
 * ガイドセルの非同期送受信を行ったリクエストを取得する。
 *
 * MB型格子の通信テーブル、EV型格子の通信テーブルの両方を検索してMPIリクエストを取得する。
 * req_listは、このメソッド内でnewしているので、呼出側でdeleteを行わなければならない。
 * @param[in]     face          通信ガイドセル数
 * @param[out]    req_list       通信リクエスト番号リスト
 * @param[in]     paraKey       パラレルコントローラー識別キー
 * @return    リクエスト数
 */
template <class DATATYPE>
SPH_MEMFUN(int)
sphVoxArray<DATATYPE>::getCommRequest(size_t face, SPL_Request *req_list)
{
  if (face == 0) {
    return 0;
  }

  // 通信テーブル数のチェック
  if (m_mapParaComm.size() + m_mapParaCommEv.size() <= 0) {
    return 0;
  }

  // 通信テーブル {MB型格子の通信テーブル、EV型格子の通信テーブル}
  ParaCommMap *mapComm[2] = {&m_mapParaCommEv, &m_mapParaComm};

  register int i;
  std::vector<SPL_Request*> req_vec;
  // MB型通信の
  for (i=0; i<2; i++) {
    // パラレルコントローラー識別キーの通信データ型テーブルの取得
    ParaCommMap::const_iterator itrPara = mapComm[i]->begin();
    for (; itrPara != mapComm[i]->end(); itrPara++) {

      // 通信ガイドセル数テーブル
      CommTypeFaceMap mapCommType = itrPara->second;

      // 通信データ型テーブルから通信ガイドセル数の通信データ型クラスの取得
      CommTypeFaceMap::const_iterator itr = mapCommType.find(face);
      if( itr == mapCommType.end() ) continue;

      // 通信情報テーブル
      std::vector<sphCommType*> vectCommType = itr->second;
      std::vector<sphCommType*>::iterator itComm = vectCommType.begin();
      for (; itComm != vectCommType.end(); itComm++) {
        sphCommType* commType = (*itComm);

        // MPIリクエスト番号が設定されているものをカウントする。
        if ( commType->m_request[0] != SPL_REQUEST_NULL ) req_vec.push_back(&(commType->m_request[0]));
        if ( commType->m_request[1] != SPL_REQUEST_NULL ) req_vec.push_back(&(commType->m_request[1]));
      }
    }
  }

  int count = req_vec.size();
  if (req_list) {
    if (count > 0) {
      for (i=0; i<count; i++) {
        memcpy(&req_list[i], req_vec.at(i), sizeof(SPL_Request));
      }
    }
  }

  return count;
}


/**
 * ガイドセルの非同期送受信を行ったリクエスト番号をクリアする。
 *
 * MB型格子の通信テーブル、EV型格子の通信テーブルのすべてのリクエスト番号をクリアする。
 * @param[in]     face          通信ガイドセル数
 * @param[in]     paraKey       パラレルコントローラー識別キー
 * @return     void
 */
template <class DATATYPE>
SPH_MEMFUN(void)
sphVoxArray<DATATYPE>::clearCommRequest(size_t face)
{
  if (face == 0) {
    return;
  }

  // 通信テーブル数のチェック
  if (m_mapParaComm.size() + m_mapParaCommEv.size() <= 0) {
    return;
  }

  // 通信テーブル {MB型格子の通信テーブル、EV型格子の通信テーブル}
  ParaCommMap *mapComm[2] = {&m_mapParaCommEv, &m_mapParaComm};

  register int i;
  // MB型通信の
  for (i=0; i<2; i++) {
    // パラレルコントローラー識別キーの通信データ型テーブルの取得
    ParaCommMap::const_iterator itrPara = mapComm[i]->begin();
    for (; itrPara != mapComm[i]->end(); itrPara++) {

      // 通信ガイドセル数テーブル
      CommTypeFaceMap mapCommType = itrPara->second;

      // 通信データ型テーブルから通信ガイドセル数の通信データ型クラスの取得
      CommTypeFaceMap::const_iterator itr = mapCommType.find(face);
      if( itr == mapCommType.end() ) continue;

      // 通信情報テーブル
      std::vector<sphCommType*> vectCommType = itr->second;
      std::vector<sphCommType*>::iterator itComm = vectCommType.begin();
      for (; itComm != vectCommType.end(); itComm++) {
        sphCommType* commType = (*itComm);

        // MPIリクエスト番号をクリアする。
        commType->m_request[0] = SPL_REQUEST_NULL;
        commType->m_request[1] = SPL_REQUEST_NULL;
      }
    }
  }

  return;
}

/**
 * 自ボクセルデータクラスを取得する。（ローカルID検索）
 * @param    localid          ローカルID
 * @return    ボクセルデータクラス
 */
template <class DATATYPE>
SPH_MEMFUN(sphVoxArray<DATATYPE>*)
sphVoxArray<DATATYPE>::getVoxArray(int localid)
{
  return this;
}


/**
 * 自ボクセルデータクラスを取得する。（ローカルID検索）
 * @param    localid          ローカルID
 * @return    ボクセルデータクラス
 */
template <class DATATYPE>
SPH_MEMFUN(const sphVoxArray<DATATYPE>*)
sphVoxArray<DATATYPE>::getVoxArray(int localid) const
{
  return this;
}


/**
 * 登録されている部分領域のボクセルデータ数を取得する。
 * @return    登録ボクセルデータ数
 */
template <class DATATYPE>
SPH_MEMFUN(int)
sphVoxArray<DATATYPE>::getNumberOfVoxArray() const
{
  return 1;
}


/**
 * データ領域のインデックスをプロセス領域のインデックスから取得する。（ガイドセルは含む領域）.
 *
 * i,j,kはガイドセルを含む領域のインデックスである。
 * 戻り値はデータ配列上のインデックスである。
 * @param    index   １次元データ配列インデックス
 * @param    i    I方向ノード領域プロセスインデックス（ガイドセルは含む）
 * @param    j    J方向ノード領域プロセスインデックス（ガイドセルは含む）：1Dの場合はセルのデータ数
 * @param    k    K方向ノード領域プロセスインデックス（ガイドセルは含む）：2Dの場合はセルのデータ数
 * @param    l    セルのデータ数 (3D)
 * @return   0=データ領域範囲外又はエラー, 1=ガイドセル領域, 2=ガイドセル領域以外の内部領域
 */
template <class DATATYPE>
SPH_MEMFUN(int)
sphVoxArray<DATATYPE>::_getLocalIndex(size_t &index, size_t i, size_t j, size_t k, size_t l) const
{
  unsigned int dims = this->getDims();
  size_t this_gc = this->getGc();
  const size_t *this_size = this->getArraySize();
  size_t this_len = this->getArrayLength();

  if (m_parentContainer != NULL) {
    // 親データクラスからパラレルコントローラの取得を行う
    const sphParaController *parect_ctrl = m_parentContainer->getParaController();
    if (parect_ctrl == NULL) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "Parent sphParaController is null.");
      return 0;
    }

    // プロセス領域の始点インデックス
    size_t proc_start[3] = {0, 0, 0};
    parect_ctrl->getVoxelStartIndex(proc_start);

    // データクラスからパラレルコントローラの取得を行う
    const sphParaController *para_ctrl = this->getParaController();
    if (parect_ctrl == NULL) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "This sphParaController is null.");
      return 0;
    }
    // 自データクラスの始点インデックス
    size_t this_start[3] = {0, 0, 0};
    para_ctrl->getVoxelStartIndex(this_start);

    // 指定インデックスが部分領域内に存在するか？
    int diff[3] = {this_start[0]-proc_start[0], this_start[1]-proc_start[1], this_start[2]-proc_start[2]};
    if (i < diff[0] || i >= diff[0] + this_size[0] + this_gc*2) return 0;
    if (j < diff[1] || j >= diff[1] + this_size[1] + this_gc*2) return 0;
    if (k < diff[2] || k >= diff[2] + this_size[2] + this_gc*2) return 0;

    // ローカルインデックスに変換する
    i = i-diff[0];
    j = j-diff[1];
    k = k-diff[2];
  }

  // インデックスチェック
  if (dims >= 1) {
    if ( i>=this_size[0]+this_gc*2) return 0;    // 領域外
  }
  if (dims >= 2) {
    if ( j>=this_size[1]+this_gc*2) return 0;    // 領域外
  }
  if (dims >= 3) {
    if ( k>=this_size[2]+this_gc*2) return 0;    // 領域外
  }

  // 自データ領域のインデックスを返す
  index = this->_getIndex(i, j, k, l);
  if (index >= this_len) return 0;

  // 取得インデックスがガイドセル領域であるか？
  if (dims >= 1) {
    if (i<this_gc || i>=this_size[0]+this_gc) return 1;    // ガイドセル領域
  }
  if (dims >= 2) {
    if (j<this_gc || j>=this_size[1]+this_gc) return 1;    // ガイドセル領域
  }
  if (dims >= 3) {
    if (k<this_gc || k>=this_size[2]+this_gc) return 1;    // ガイドセル領域
  }
  return 2;     // 内部領域
}


/**
 * データ領域のインデックスをプロセス領域のインデックスから取得する。（ガイドセルは含まない領域）.
 *
 * i,j,kはガイドセルを含まない領域のインデックスである。
 * 戻り値はデータ配列上のインデックスである。
 * @param    index   １次元データ配列インデックス
 * @param    i    I方向ノード領域プロセスインデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域プロセスインデックス（ガイドセルは含まない）：1Dの場合はセルのデータ数
 * @param    k    K方向ノード領域プロセスインデックス（ガイドセルは含まない）：2Dの場合はセルのデータ数
 * @param    l    セルのデータ数 (3D)
 * @return   0=データ領域範囲外又はエラー, 1=ガイドセル領域, 2=ガイドセル領域以外の内部領域
 */
template <class DATATYPE>
SPH_MEMFUN(int)
sphVoxArray<DATATYPE>::getLocalIndex(size_t &index, size_t i, size_t j, size_t k, size_t l) const
{
  unsigned int dims = this->getDims();
  size_t this_gc = this->getGc();
  const size_t *this_size = this->getArraySize();
  size_t this_len = this->getArrayLength();

  if (m_parentContainer != NULL) {
    // 親データクラスからパラレルコントローラの取得を行う
    const sphParaController *parect_ctrl = m_parentContainer->getParaController();
    if (parect_ctrl == NULL) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "Parent sphParaController is null.");
      return 0;
    }

    // プロセス領域の始点インデックス
    size_t proc_start[3] = {0, 0, 0};
    parect_ctrl->getVoxelStartIndex(proc_start);

    // データクラスからパラレルコントローラの取得を行う
    const sphParaController *para_ctrl = this->getParaController();
    if (parect_ctrl == NULL) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "This sphParaController is null.");
      return 0;
    }
    // 自データクラスの始点インデックス
    size_t this_start[3] = {0, 0, 0};
    para_ctrl->getVoxelStartIndex(this_start);

    // 指定インデックスが部分領域内に存在するか？
    int diff[3] = {this_start[0]-proc_start[0], this_start[1]-proc_start[1], this_start[2]-proc_start[2]};
    if (i < diff[0] || i >= diff[0] + this_size[0]) return 0;
    if (j < diff[1] || j >= diff[1] + this_size[1]) return 0;
    if (k < diff[2] || k >= diff[2] + this_size[2]) return 0;

    // ローカルインデックスに変換する
    i = i-diff[0];
    j = j-diff[1];
    k = k-diff[2];
  }

  // インデックスチェック
  if (dims >= 1) {
    if ( i>=this_size[0]) return 0;    // 領域外
  }
  if (dims >= 2) {
    if ( j>=this_size[1]) return 0;    // 領域外
  }
  if (dims >= 3) {
    if ( k>=this_size[2]) return 0;    // 領域外
  }

  // 自データ領域のインデックスを返す
  index = this->getIndex(i, j, k, l);
  if (index >= this_len) return 0;
  return 2;     // 内部領域

}


#ifdef _DEBUGLOG_

template <class DATATYPE>
SPH_MEMFUN(void)
sphVoxArray<DATATYPE>::debugPrint() const
{
  char msg[1024];
  std::string type;

  sprintf(msg, "label:%s", sphData::getLabel().c_str());
  debugLog(msg);

  SPL_Datatype dtype = sphData::getDataType();
  if (dtype == SPL_CHAR) type = "SPL_CHAR";
  if (dtype == SPL_UNSIGNED_CHAR) type = "SPL_UNSIGNED_CHAR";
  if (dtype == SPL_SHORT) type = "SPL_SHORT";
  if (dtype == SPL_UNSIGNED_SHORT) type = "SPL_UNSIGNED_SHORT";
  if (dtype == SPL_INT) type = "SPL_INT";
  if (dtype == SPL_UNSIGNED) type = "SPL_UNSIGNED";
  if (dtype == SPL_LONG) type = "SPL_LONG";
  if (dtype == SPL_UNSIGNED_LONG) type = "SPL_UNSIGNED_LONG";
  if (dtype == SPL_FLOAT) type = "SPL_FLOAT";
  if (dtype == SPL_DOUBLE) type = "SPL_DOUBLE";
  if (dtype == SPL_LONG_LONG_INT) type = "SPL_LONG_LONG_INT";
  if (dtype == SPL_LONG_LONG) type = "SPL_LONG_LONG";
  sprintf(msg, "data type:%s", type.c_str());
  debugLog(msg);

  SphDCType dctype = sphData::getDcType();
  if (dctype == SPH_DC_ARRAY_1D) type = "SPH_DC_ARRAY_1D";
  if (dctype == SPH_DC_ARRAY_1DN) type = "SPH_DC_ARRAY_1DN";
  if (dctype == SPH_DC_ARRAY_1DNRV) type = "SPH_DC_ARRAY_1DNRV";
  if (dctype == SPH_DC_ARRAY_2D) type = "SPH_DC_ARRAY_2D";
  if (dctype == SPH_DC_ARRAY_2DN) type = "SPH_DC_ARRAY_2DN";
  if (dctype == SPH_DC_ARRAY_2DNRV) type = "SPH_DC_ARRAY_2DNRV";
  if (dctype == SPH_DC_ARRAY_3D) type = "SPH_DC_ARRAY_3D";
  if (dctype == SPH_DC_ARRAY_3DN) type = "SPH_DC_ARRAY_3DN";
  if (dctype == SPH_DC_ARRAY_3DNRV) type = "SPH_DC_ARRAY_3DNRV";
  sprintf(msg, "data class:%s", type.c_str());
  debugLog(msg);

  sprintf(msg, "dims:%d", sphArray<DATATYPE>::getDims());
  debugLog(msg);

  const size_t* sz = sphArray<DATATYPE>::getArraySize();
  sprintf(msg, "size:%ld, %ld, %ld", sz[0], sz[1], sz[2]);
  debugLog(msg);

  sprintf(msg, "gc:%ld", sphArray<DATATYPE>::getGc());
  debugLog(msg);

  sprintf(msg, "data_num:%ld", sphArray<DATATYPE>::getNumOfDtElem());
  debugLog(msg);

  SphCrdDef crddef = getCrdDef();
  type = "CRDDEF_UNKNOWN";
  if (crddef == CRDDEF_REGULAR) type = "CRDDEF_REGULAR";
  if (crddef == CRDDEF_COLLOCATE) type = "CRDDEF_COLLOCATE";
  if (crddef == CRDDEF_STAGGERED1) type = "CRDDEF_STAGGERED1";
  if (crddef == CRDDEF_STAGGERED2) type = "CRDDEF_STAGGERED2";
  sprintf(msg, "crddef:%s", type.c_str());
  debugLog(msg);

}
#endif  // _DEBUGLOG_


#endif // _SPH_VOX_ARRAY_INLINE_H_

