/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * @file sphVoxContainer_inline.h
 * 部分領域管理データクラスインラインソースファイル
 * @author     hira
 * @date       2010/10/01
 */

#ifndef _SPHVOXCONTAINER_INLINE_H_
#define _SPHVOXCONTAINER_INLINE_H_

#define SPH_MEMFUN(rettype) \
        SPH_INLINE rettype


#include "data/sphDataUtils.h"

/**
 * コンストラクタ
 */
template <class DATATYPE>
SPH_MEMFUN()
sphVoxContainer<DATATYPE>::sphVoxContainer()
{
  m_voxArrayMap.clear();
  m_evCommBulk = false;
}

/**
 * コンストラクタ
 *
 * 部分領域のボクセルデータクラスを生成する。
 * プロセス領域のデータはアロケートしない。
 * @param    dctype    データクラスタイプ
 * @param    dttype    データ型タイプ
 * @param    label      データ識別ラベル
 * @param    paraCtrlr  パラレルコントローラー
 * @param    gc        ガイドセルサイズ
 * @param    data_num  セルのデータ数
 * @param    crddef    データ定義位置
 */
template <class DATATYPE>
SPH_MEMFUN()
sphVoxContainer<DATATYPE>::sphVoxContainer(
                                  const std::string label,
                                  SphDCType dctype,
                                  SPL_Datatype dttype,
                                  const sphParaController *paraCtrlr,
                                  const size_t gc,
                                  const size_t data_num,
                                  SphCrdDef crddef)
{

  m_voxArrayMap.clear();
  m_evCommBulk = false;

  /// パラレルコントローラを部分領域管理パラレルコントローラクラスにキャストする。
  if (!paraCtrlr) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "sphParaController is null.");
    return;
  }
  const sphMultiParaController* multi_ctrl = dynamic_cast<const sphMultiParaController*>(paraCtrlr);
  if( !multi_ctrl ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphMultiParaController dynamic_cast error.");
    return;
  }

  // プロセス領域のデータクラスの初期化を行う。
  if (!sphData::initializeData(label, dctype, dttype, paraCtrlr)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid sphData::initializeData(...).");
    return;
  }

  // データ配列のアロケートは行わずに、データクラスの初期化を行う。
  if (!sphVoxArray<DATATYPE>::initializeVoxArray(label, dctype, dttype, paraCtrlr, gc, data_num, crddef, false)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid sphVoxArray::initializeVoxArray(...).");
    return;
  }

  /// 部分領域管理パラレルコントローラから部分領域のボクセルデータクラスを生成する。
  register int i;
  int num = multi_ctrl->getNumberOfChildCtrl();
  for (i=0; i<num; i++) {
    // サブブロックの子パラレルコントローラからサブブロック領域のボクセルデータクラスを生成する。
    const sphSvParaController* child_ctrl =  multi_ctrl->getChildParaController(i);
    sphVoxArray<DATATYPE>* child_data = createVoxelArray(
                                      dctype,
                                      dttype,
                                      label,
                                      child_ctrl,
                                      gc,
                                      data_num,
                                      crddef);
    if (child_data == NULL) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid createVoxelArray.");
      return;
    }

    // 親データクラスを設定する。
    child_data->setParentContainer(this);

    // ボクセルデータクラスリストに追加
    if( !addChildVoxelArray(i, child_data) ){
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_voxArrayMap. [localid=%d].", i);
      return;
    }

  }

  return;
}


/**
 * デストラクタ
 */
template <class DATATYPE>
SPH_MEMFUN()
sphVoxContainer<DATATYPE>::~sphVoxContainer()
{
}


/**
 * 登録されている部分領域のボクセルデータ数を取得する。
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(int)
sphVoxContainer<DATATYPE>::getNumberOfVoxArray() const
{
  return m_voxArrayMap.size();
}


/**
 * 登録されている部分領域のボクセルデータクラスを取得する。（ローカルID検索）
 * @param    localid          ローカルID
 * @return    ボクセルデータクラス
 */
template <class DATATYPE>
SPH_MEMFUN(sphVoxArray<DATATYPE>*)
sphVoxContainer<DATATYPE>::getVoxArray(int localid)
{
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.find(localid);
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itrEnd = m_voxArrayMap.end();
  if( itr == itrEnd ) return NULL;
  return itr->second;
}


/**
 * 登録されている部分領域のボクセルデータクラスを取得する。（ローカルID検索）
 * @param    localid          ローカルID
 * @return    ボクセルデータクラス
 */
template <class DATATYPE>
SPH_MEMFUN(const sphVoxArray<DATATYPE>*)
sphVoxContainer<DATATYPE>::getVoxArray(int localid) const
{
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.find(localid);
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itrEnd = m_voxArrayMap.end();
  if( itr == itrEnd ) return NULL;
  return itr->second;
}



/**
 * 登録されている部分領域のボクセルデータクラスを取得する。（ノードID検索）
 * @param    nodeid          ノードID
 * @return    ボクセルデータクラス
 */
template <class DATATYPE>
SPH_MEMFUN(sphVoxArray<DATATYPE>*)
sphVoxContainer<DATATYPE>::getVoxArrayByNodeId(int nodeid)
{
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    // パラレルコントローラの取得
    const sphParaController* child_para = child_data->getParaController();
    if( !child_para ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphParaController is null.");
      return NULL;
    }
    int child_nodeid = paraCtrlr->getMyID();
    if (nodeid == child_nodeid) {
      return child_data;
    }
  }
  return NULL;
}



/**
 * 登録されている部分領域のボクセルデータクラスを取得する。（ノードID検索）
 * @param    nodeid          ローカルID
 * @return    ボクセルデータクラス
 */
template <class DATATYPE>
SPH_MEMFUN(const sphVoxArray<DATATYPE>*)
sphVoxContainer<DATATYPE>::getVoxArrayByNodeId(int nodeid) const
{
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    // パラレルコントローラの取得
    const sphParaController* child_para = child_data->getParaController();
    if( !child_para ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphParaController is null.");
      return NULL;
    }
    int child_nodeid = child_para->getMyID();
    if (nodeid == child_nodeid) {
      return child_data;
    }
  }
  return NULL;
}



/**
 * 設定されているガイドセルの通信を行う。
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::comm()
{
  return comm(this->getGc());
}

/**
 * 通信にてガイドセルの通信を行う。
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::comm(size_t face, int paraKey)
{

  const sphSvParaController* paraCtrlr = dynamic_cast<const sphSvParaController*>(this->getParaController(paraKey));
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaCtrlr dynamic_cast error.");
    return false;
  }

  if( paraCtrlr->isEv() ) {
    if (m_evCommBulk) {
      // EV分割:I,J,K方向一括通信
      return commEvBulk(face, paraKey);
    }
    else {
      // EV分割:I,J,K方向順次通信
      return commEvOrder(face, paraKey);
    }
  }
  else {
    // MB分割
    return commMb(face, paraKey);
  }


  return true;
}


/**
 * EV型の設定されているガイドセルのPeriodic通信を行う。
 * @param     face_dir       0:X面, 1:Y面, 2:Z面
 * @param     axis           -1:マイナス方向, +1:プラス方向
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::commPeriodic(unsigned int face_dir, int axis)
{
  return commPeriodic(face_dir, axis, this->m_gc);
}

/**
 * EV型のガイドセルのPeriodic通信を行う。
 * @param     face_dir       0:X面, 1:Y面, 2:Z面
 * @param     axis           -1:マイナス方向, +1:プラス方向
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::commPeriodic(unsigned int face_dir, int axis, size_t face, int paraKey)
{

  register int i;

  const sphMultiParaController* parent_ctrl = dynamic_cast<const sphMultiParaController*>(this->getParaController(paraKey));
  if( !parent_ctrl ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphMultiParaController dynamic_cast error.");
    return false;
  }

  // I, J, K方向のPeriodic通信を行う。
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr =  m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;

    const sphSvParaController* para_ctrl = dynamic_cast<const sphSvParaController*>(child_data->getParaController(paraKey));
    if( !para_ctrl ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
      return false;
    }

    // Periodic通信相手の取得
    int send_commID = para_ctrl->getCommIDPeriodic(face_dir, axis);
    int recv_commID = para_ctrl->getCommIDPeriodic(face_dir, (-1)*axis);
    if (send_commID < 0 && recv_commID < 0) {
      continue;
    }

    // 同一プロセス内のサブブロックであるか？
    if ( (send_commID < 0 || parent_ctrl->isContainNodeId(send_commID) )
        && (recv_commID < 0 || parent_ctrl->isContainNodeId(recv_commID) ) ) {
      // Periodic通信相手は同一プロセスであるので、MPI通信は行わない。
      continue;
    }

    // EV通信テーブルにて非同期送受信を行う。
    if( !child_data->commPeriodicISendRecv(face_dir, axis, face, paraKey) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
      return false;
    }
  }

  itr =  m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    // 同一プロセス内のサブブロックに対してデータコピーを行う。
    if( !child_data->commPediodicDataCopy(face_dir, axis, face, paraKey) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commDataCopy (face=%d, key=%d).", face, paraKey);
      return false;
    }
  }

  itr =  m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    // MPI通信待ちを行う。
    if( !child_data->commWait(face) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
      return false;
    }
  }

  return true;
}



/**
 * 自ノード内のデータ数を取得する。（ガイドセルを含まない）
 *
 * 部分領域データのデータ数を合計する。
 * @param    id      カウントするデータ値
 * @return    データ数
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphVoxContainer<DATATYPE>::getNumberOfID(DATATYPE id) const
{
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr =  m_voxArrayMap.begin();

  size_t count = 0;
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    count += child_data->getNumberOfID(id);
  }

  return count;
}


/**
 * 自ノード内のデータ数を取得する。（ガイドセルを含む）
 *
 * 部分領域データのデータ数を合計する。
 * @param    id      カウントするデータ値
 * @return    データ数
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphVoxContainer<DATATYPE>::getNumberOfIDExtGc(DATATYPE id) const
{
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr =  m_voxArrayMap.begin();
  size_t count = 0;
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    count += child_data->getNumberOfIDExtGc(id);
  }
  return count;
};


/**
 * データをコピーする.
 *
 * データ領域サイズ、データ型、データクラス、格子当りデータ数は同じでない場合はエラーとする。
 * ガイドセル領域もコピーを行う。（ガイドセルが異なる場合もコピーする）
 * コピー元、コピー先のグローバル始点インデックスは同じインデックスとしてコピーを行う。
 * @param     src_dc           コピー元データ
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::copyArray(const sphData* src_dc)
{
  if (src_dc == NULL) return false;
  if (this->getDcType() != src_dc->getDcType()) return false;
  if (this->getDataType() != src_dc->getDataType()) return false;
  if (this->getDims() != src_dc->getDims()) return false;
  if (this->getNumOfDtElem() != src_dc->getNumOfDtElem()) return false;
  if (!src_dc->isVoxContainer()) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "src_dc is not sphVoxContainer Class.");
    return false;
  }
  if (this->getNumberOfVoxArray() != src_dc->getNumberOfVoxArray()) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not equals Container sphVoxArray count[this=%d, src=%d].", this->getNumberOfVoxArray(), src_dc->getNumberOfVoxArray());
    return false;
  }

  const sphVoxContainer<DATATYPE>* cont_src = dynamic_cast<const sphVoxContainer<DATATYPE>*>(src_dc);
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr =  m_voxArrayMap.begin();
  // 自クラスの部分領域データを取得しコピーを行う。
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;

    const sphVoxArray<DATATYPE>* sub_src = cont_src->getVoxArray(localid);
    if (!child_data->copyArray(sub_src)) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid copyArray.");
      return false;
    }
  }

  return true;
}


/**
 * コピー元データからデータコピーを行う。
 *
 * コピー元データは部分領域データクラス、ボクセルデータクラス指定可能とする。
 * @param    src_dc      コピー元データ
 * @param    src_vc      コピー元のガイドセル数
 * @param    dest_vc    コピー先のガイドセル数
 * @param    src_paraKey    コピー元のパラレルコントローラー識別キー
 * @param    dest_paraKey   コピー先のパラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::copyVoxelArray(
                    const sphData* src_dc,
                    size_t src_vc,
                    size_t dest_vc,
                    int src_paraKey,
                    int dest_paraKey)
{
  register int i;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr =  m_voxArrayMap.begin();

  // 自クラスの部分領域データを取得しコピーを行う。
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    // コピー元データから自クラスの部分領域データにコピーを行う。
    if (!child_data->copyVoxelArray(src_dc, src_vc, dest_vc, src_paraKey, dest_paraKey)) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid copyVoxelArray");
      return false;
    }
  }
  return true;
}


/**
 * 全ノードから境界領域のインデックス、サイズを取得する。（ガイドセルを含まない）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得境界領域始点グローバルインデックス（ガイドセルを含まない）
 * @param[out]    bnd_len    取得境界領域サイズ（ガイドセルを含まない）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::getBndWIndex(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{
  // ガイドセルを含まない境界領域のインデックス、サイズを取得する。
  if (!getBndWIndexByReduce(id, bnd_sta, bnd_len, ext_gc, paraKey, false)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid getBndWIndexByReduce.");
    return false;
  }

  return true;
};


/**
 * 全ノードからMultibox境界領域のインデックス、サイズを取得する。（ガイドセル領域を含まない）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得Multibox境界領域始点グローバルインデックス（ガイドセル領域を含まない）
 * @param[out]    bnd_len    取得Multibox境界領域サイズ（ガイドセル領域を含まない）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::getBndWIndexMb(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{

  return true;
}


/**
 * 自ノードからMultibox境界領域のローカルインデックス、サイズを取得する。（ガイドセルは含まない）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得Multibox境界領域始点ローカルインデックス（ガイドセルは含まない）
 * @param[out]    bnd_len    取得Multibox境界領域サイズ（ガイドセルは含まない）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::getBndIndexMb(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{

  return true;
}



/**
 * 全ノードから境界領域のインデックス、サイズを取得する。（ガイドセルを含む）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得境界領域始点グローバルインデックス（ガイドセルを含む）
 * @param[out]    bnd_len    取得境界領域サイズ（ガイドセルを含む）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::getBndWIndexExtGc(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{
  // ガイドセルを含む境界領域のインデックス、サイズを取得する。
  if (!getBndWIndexByReduce(id, bnd_sta, bnd_len, ext_gc, paraKey, true)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid getBndWIndexByReduce.");
    return false;
  }

  return true;
};


/**
 * 全ノードからMultibox境界領域のインデックス、サイズを取得する。（ガイドセル領域を含む）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得Multibox境界領域始点グローバルインデックス（ガイドセル領域を含む）
 * @param[out]    bnd_len    取得Multibox境界領域サイズ（ガイドセル領域を含む）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::getBndWIndexMbExtGc(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{
  return true;
}

/**
 * 自ノードからMultibox境界領域のローカルインデックス、サイズを取得する。（ガイドセルを含む）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得Multibox境界領域始点ローカルインデックス（ガイドセルを含む）
 * @param[out]    bnd_len    取得Multibox境界領域サイズ（ガイドセルを含む）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::getBndIndexMbExtGc(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey) const
{

  return true;
}



/**
 * ボクセルデータを作成、登録する。
 *
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      label          データ識別ラベル
 * @param      paraCtrlr      パラレルコントローラー
 * @param      gc            ガイドセルサイズ
 * @param      data_num      セルのデータ数
 * @param      crddef      データ定義位置
 * @return ボクセルデータオブジェクト
 */
template <class DATATYPE>
SPH_MEMFUN(sphVoxArray<DATATYPE>*)
sphVoxContainer<DATATYPE>::createVoxelArray(
                      SphDCType dcType,
                      SPL_Datatype dataType,
                      std::string label,
                      const sphParaController *paraCtrlr,
                      const size_t gc,
                      const size_t data_num,
                      SphCrdDef crddef)
{

  // データクラス作成
  sphVoxArray<DATATYPE>* voxel = NULL;
  voxel = new sphVoxArray<DATATYPE>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);

  if (voxel == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  else {
    if (voxel->getDataPointer() == NULL) {
      delete voxel;
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data array.");
      return NULL;
    }
  }

  return voxel;
}


/**
 * ボクセルデータクラスリストにサブブロック領域のデータオブジェクトを追加する。
 * @param localid     サブブロックID
 * @param data        サブブロック領域のデータオブジェクト
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::addChildVoxelArray(int localid, sphVoxArray<DATATYPE>* data)
{
    /// m_voxArrayMapにサブブロック領域のデータオブジェクトを追加する。
    if( !((m_voxArrayMap.insert(std::make_pair(localid, data))).second) ){
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_voxArrayMap [localid=%d].", localid);
      return false;
    }
    return true;
}


/**
 * 全ノードから境界領域のインデックス、サイズを取得する。（ガイドセルを含まない）
 *
 * extgcFlag=trueの場合、bnd_sta、bnd_lenはガイドセルを含むインデックス、サイズとなる。
 * extgcFlag=falseの場合、bnd_sta、bnd_lenはガイドセルを含まないインデックス、サイズとなる。
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta    取得境界領域始点グローバルインデックス（ガイドセルを含まない）
 * @param[out]    bnd_len    取得境界領域サイズ（ガイドセルを含まない）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @param[out]    extgcFlag    ガイドセル領域をチェックするフラグ(true=ガイドセル領域を含む)
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::getBndWIndexByReduce(
                      DATATYPE id,
                      size_t* bnd_sta,
                      size_t* bnd_len,
                      const size_t ext_gc,
                      int paraKey,
                      bool extgcFlag ) const
{

  const sphParaController* paraCtrlr = this->getParaController(paraKey);
  if( !paraCtrlr ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
    return false;
  }

  register int i;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr =  m_voxArrayMap.begin();

  /// 部分領域の境界領域を取得する。
  size_t my_sta[3] = {(size_t)-1, (size_t)-1, (size_t)-1};
  size_t my_end[3] = {0, 0, 0};
  size_t my_len[3] = {0, 0, 0};
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    // 自領域の境界領域のローカルインデックス、サイズを取得する。（ガイドセル領域の走査はextgcFlagによる）
    size_t child_sta[3] = {0, 0, 0};
    size_t child_end[3] = {0, 0, 0};
    size_t child_len[3] = {1, 1, 1};
    size_t child_size[3] = {0, 0, 0};
    if (!child_data->getBndMyIndex(id, child_sta, child_len, paraKey, extgcFlag)) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid getBndMyIndex.");
      return false;
    }

    /// 取得境界領域の最小インデックス、最大インデックスを取得する。
    for (i=0; i<this->getDims(); i++) {
      const sphParaController* child_para = child_data->getParaController(paraKey);
      if( !child_para ) {
        // error
        sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get sphParaController.");
        return false;
      }
      // ガイドセルを含まないボクセルサイズ
      if (extgcFlag) child_size[i] = this->m_sz[i]+this->m_gcl[i]*2;
      else this->m_sz[i];

      // グローバルインデックスに変換
      int myID = child_para->getMyID();
      long headIdx = paraCtrlr->getVoxelHeadIndex(myID, i);
      child_sta[i] = child_sta[i] + headIdx;
      child_end[i] = child_sta[i] + child_len[i] - 1 + headIdx;

      // 拡張ガイドセル数を加算する
      child_sta[i] = ((long)child_sta[i] - (long)ext_gc < (long)headIdx) ? headIdx : child_sta[i] - ext_gc;
      child_end[i] = (child_end[i] + ext_gc >=child_size[i]+headIdx) ? child_size[i]+headIdx-1 : child_end[i] + ext_gc;

      // プロセス領域の最小始点インデックスと最大終点インデックスを取得する。
      if (my_sta[i] > child_sta[i]) my_sta[i] = child_sta[i];
      if (my_end[i] < child_end[i]) my_end[i] = child_end[i];
    }
  }

  size_t tmp_wsta[3] = {0};
  size_t tmp_wend[3] = {0};
  SPL_Datatype dtype = sizeof(size_t)==4?SPL_UNSIGNED:SPL_LONG_LONG;

  // 始点インデックスを取得する。
  if( !paraCtrlr->allreduce(&my_sta, &tmp_wsta, 3, dtype, SPL_MIN) ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "paraCtrlr->allreduce error.");
    paraCtrlr->abort(-1);
    return false;
  }

  // 終点インデックスを取得する。
  if( !paraCtrlr->allreduce(&my_end, &tmp_wend, 3, dtype, SPL_MAX) ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "paraCtrlr->allreduce error.");
    paraCtrlr->abort(-1);
    return false;
  }

  for (i=0; i<this->getDims(); i++) {
    // check index
    if (tmp_wsta[i] > tmp_wend[i]) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "my_sta[%d](=%ld) > my_end[%d](=%ld).", i, my_sta[i], i, my_end[i]);
      return false;
    }
    bnd_sta[i] = tmp_wsta[i];
    bnd_len[i] = tmp_wend[i] - tmp_wsta[i] + 1;
  }

  return true;
};



/**
 * MB型通信にてガイドセルの通信を行う。
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::commMb(size_t face, int paraKey)
{
  if (face == 0) {
    return true;
  }

  // MB型のMPI通信を行う(ISend, IRecvのみの非同期通信)
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr =  m_voxArrayMap.begin();
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;

    // MB通信テーブルにて非同期送受信を行う。
    if( !child_data->commISendRecv(face, DIR_UNKNOWN, paraKey) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
      return false;
    }
  }

  // データコピーを行う。
  itr =  m_voxArrayMap.begin();
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;

    // 同一プロセス内のサブブロックに対してデータコピーを行う。
    if( !child_data->commDataCopy(face, DIR_UNKNOWN, paraKey) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commDataCopy (face=%d, key=%d).", face, paraKey);
      return false;
    }
  }

  // MPI通信のWaitを行う
  this->commWaitAll(face);

  return true;
}


/**
 * MB型通信にてガイドセルの通信を行う。
 *
 * ガイドセル領域のすべて通信を行う。
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::commMb()
{
  return commMb(this->m_gc);
}


/**
 * EV型通信:I,J,K方向順番にてガイドセルの通信を行う。
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::commEvOrder(size_t face, int paraKey)
{
  // 通信方向
  DirFaceType dir[3] = {X_DIR, Y_DIR, Z_DIR};   // I, J, K方向
  register int i;

  // I, J, K方向のEV型通信を順次行う。
  for (i=0; i<3; i++) {

    typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr =  m_voxArrayMap.begin();
    for ( ; itr != m_voxArrayMap.end(); itr++) {
      int localid = itr->first;
      sphVoxArray<DATATYPE>* child_data = itr->second;

      // EV通信テーブルにて非同期送受信を行う。
      if( !child_data->commISendRecv(face, dir[i], paraKey ) ) {
        // error
        sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
        return false;
      }

      // 同一プロセス内のサブブロックに対してデータコピーを行う。
      if( !child_data->commDataCopy(face, dir[i], paraKey) ) {
        // error
        sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commDataCopy (face=%d, key=%d).", face, paraKey);
        return false;
      }

      // MPI通信待ちを行う。
      if( !child_data->commWait(face) ) {
        // error
        sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
        return false;
      }
    }  // for ( ; itr != m_voxArrayMap.end(); itr++)
  }  // for (i=0; i<3; i++)

  return true;
}


/**
 * EV型通信:I,J,K方向順番にてガイドセルの通信を行う。
 *
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::commEvOrder()
{
  return commEvOrder(this->m_gc);
}


/**
 * EV型通信:I,J,K方向一括にてガイドセルの通信を行う。
 *
 * ガイドセル領域のすべて通信を行う。
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::commEvBulk(size_t face, int paraKey)
{
  register int i;

  // I, J, K方向のEV型通信を一括に行う。
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr =  m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;

    // EV通信テーブルにて非同期送受信を行う。
    if( !child_data->commISendRecv(face, DIR_UNKNOWN, paraKey) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
      return false;
    }
  }

  itr =  m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    // 同一プロセス内のサブブロックに対してデータコピーを行う。
    if( !child_data->commDataCopy(face, DIR_UNKNOWN, paraKey) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commDataCopy (face=%d, key=%d).", face, paraKey);
      return false;
    }
  }

  itr =  m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    // MPI通信待ちを行う。
    if( !child_data->commWait(face) ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid commISendRecv (face=%d, key=%d).", face, paraKey);
      return false;
    }
  }

  return true;
}


/**
 * EV型通信:I,J,K方向一括にてガイドセルの通信を行う。
 *
 * ガイドセル領域のすべて通信を行う。
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::commEvBulk()
{
  return commEvBulk(this->m_gc);
}


/**
 * プロセスのボクセルデータクラスを生成する。
 *
 * 子データのサブブロック領域を包括したプロセスのボクセル領域のデータクラスを生成する。
 * 生成データクラスの破棄は呼出し側にて行う必要がある。
 * データ識別ラベルは設定しないので、データ識別ラベルの設定、データマネージャへの登録は呼出し側にて行う必要がある。
 * @param   new_gc      生成ガイドセル ( <0 = 元ガイドセルにて生成する。)
 * @param copy_data     データコピーフラグ（true=部分領域のデータをコピーする）
 * @return ボクセルデータオブジェクト
 */
template <class DATATYPE>
SPH_MEMFUN(sphData*)
sphVoxContainer<DATATYPE>::createVoxelArray(int new_gc, bool copy_data)
{
  // データクラス作成
  sphVoxArray<DATATYPE>* voxel = NULL;

  // 生成するガイドセルを決定する。
  size_t gc = this->m_gc;
  if (new_gc >= 0) gc = new_gc;

  // データ識別ラベルは設定しない。
  voxel = new sphVoxArray<DATATYPE>(
                              "", this->m_dcType, this->m_dtType,
                              this->getParaController(),
                              gc,
                              this->m_dtElemNum,
                              this->m_crddef);

  if (voxel == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  else {
    if (voxel->getDataPointer() == NULL) {
      delete voxel;
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data array.");
      return NULL;
    }
  }

  // データコピーを行う。
  if (copy_data) {
    if ( !voxel->copyVoxelArray(this, gc, gc) ) {
      delete voxel;
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid copyVoxelArray");
      return NULL;
    }
  }

  return voxel;
}



/**
 * 非同期MPI通信の完了待ちを行う。
 *
 * @param    face          通信ガイドセル数
 * @param    paraKey    パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::commWaitAll(size_t face)
{
  register int i;

  // すべてのリクエスト番号の取得を行う。
  std::vector<SPL_Request> req_array;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    int count = child_data->getCommRequest(face, NULL);
    SPL_Request *list = new SPL_Request[count];
    count = child_data->getCommRequest(face, list);

    if (count > 0) {
      for (i=0; i<count; i++) req_array.push_back(list[i]);
      delete [] list;
    }
  }

  if (req_array.size() > 0) {
    // MPI_WAITALLを行う。
    SPL_Status *stat = new SPL_Status[req_array.size()];
    int ret = SPL_Waitall(req_array.size(), &req_array[0], stat);
    delete [] stat;
    if ( ret != SPL_MPI_SUCCESS) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "mpi_waitall error.");
      return false;
    }
  }

  // リクエスト番号のクリアを行う。
  itr =  m_voxArrayMap.begin();
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    child_data->clearCommRequest(face);
  }
  return true;
}


/**
 * EV実行形態のガイドセルの通信方法フラグを設定する。
 *
 * デフォルトのEV実行形態のガイドセルの通信方法は、順次(false)である。
 * @param commflag      EV通信の一括(true)/順次(false)フラグ
 */
template <class DATATYPE>
SPH_MEMFUN(void)
sphVoxContainer<DATATYPE>::setEvCommBulk(bool commflag)
{
  m_evCommBulk = commflag;
}


/**
 * 部分領域管理データクラスであるかチェックする。
 * @return       true=部分領域管理データクラス
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::isVoxContainer() const
{
  return true;
}

/**
 * データすべてに設定値を設定する。
 * @param     val             設定値
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::initValues(DATATYPE val)
{
  // サブブロックデータに対してデータをセットする。
  std::vector<SPL_Request> req_array;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    if (!child_data->initValues(val)) {
      return false;
    }
  }

  return true;
}


/**
 * アロケートを行っているサブブロックのメモリサイズの合計を取得する。
 *
 * ボクセルサイズ x データ型サイズをバイト単位で返す。
 * @return    メモリサイズ（バイト）
 */
template <class DATATYPE>
SPH_MEMFUN(size_t)
sphVoxContainer<DATATYPE>::getMemoryBytes() const
{
  size_t sum_mem = 0;
  // サブブロックデータに対してメモリサイズを合計する。
  std::vector<SPL_Request> req_array;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for( ; itr != m_voxArrayMap.end(); itr++){
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    sum_mem += child_data->getMemoryBytes();
  }

  return sum_mem;
}


/**
 * 全ノードから境界領域のインデックス、サイズを取得し、サブブロック毎に始点インデックス、サイズを取得する。（ガイドセルを含む）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta_list    サブブロックの境界領域始点グローバルインデックス（ガイドセルを含む）
 * @param[out]    bnd_len_list    サブブロックの境界領域サイズ（ガイドセルを含む）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::getBndIndexExtGc_InVoxArray(
                      DATATYPE id,
                      size_t bnd_sta_list[][3],
                      size_t bnd_len_list[][3],
                      const size_t ext_gc,
                      int paraKey) const
{

  size_t bnd_sta[3] = {0};
  size_t bnd_len[3] = {0};
  // ガイドセルを含む境界領域のインデックス、サイズを取得する。
  if (!getBndWIndexByReduce(id, bnd_sta, bnd_len, ext_gc, paraKey, true)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid getBndWIndexByReduce.");
    return false;
  }

  // サブブロックの始点インデックス、サイズを取得する。
  register int i, n;
  int sub_count = this->getNumberOfVoxArray();
  for (n=0; n<sub_count; n++) {
    const sphVoxArray<DATATYPE>* sub_dc = dynamic_cast<const sphVoxArray<DATATYPE>*>(this->getVoxArray(n));

    // サブブロックのローカルインデックスに変換する。
    size_t my_sta[3] = {bnd_sta[0], bnd_sta[1], bnd_sta[2]};
    size_t my_len[3] = {bnd_len[0], bnd_len[1], bnd_len[2]};

    // get local indexes from global indexes
    if (!getGlobal2LocalIndexes(my_sta, my_len, this->m_gc, paraKey)) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid getGlobal2LocalIndexes.");
      return false;
    }

    // 返値にセットする。
    for (i=0; i<3; i++) {
      bnd_sta_list[n][i] = my_sta[i];
      bnd_len_list[n][i] = my_len[i];
    }
  }

  return true;
}

/**
 * 全ノードから境界領域のインデックス、サイズを取得し、サブブロック毎に始点インデックス、サイズを取得する。（ガイドセルを含まない）
 * @param[in]  id              境界領域の検索対象データ値
 * @param[out]    bnd_sta_list    サブブロックの境界領域始点グローバルインデックス（ガイドセルを含む）
 * @param[out]    bnd_len_list    サブブロックの境界領域サイズ（ガイドセルを含む）
 * @param[in]  ext_gc          拡張ガイドセル数
 * @param[in]  paraKey        パラレルコントローラー識別キー
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::getBndIndex_InVoxArray(
                      DATATYPE id,
                      size_t bnd_sta_list[][3],
                      size_t bnd_len_list[][3],
                      const size_t ext_gc,
                      int paraKey) const
{

  size_t bnd_sta[3] = {0};
  size_t bnd_len[3] = {0};
  // ガイドセルを含む境界領域のインデックス、サイズを取得する。
  if (!getBndWIndexByReduce(id, bnd_sta, bnd_len, ext_gc, paraKey, false)) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid getBndWIndexByReduce.");
    return false;
  }

  // サブブロックの始点インデックス、サイズを取得する。
  register int i, n;
  int sub_count = this->getNumberOfVoxArray();
  for (n=0; n<sub_count; n++) {
    const sphVoxArray<DATATYPE>* sub_dc = dynamic_cast<const sphVoxArray<DATATYPE>*>(this->getVoxArray(n));

    // サブブロックのローカルインデックスに変換する。
    size_t my_sta[3] = {bnd_sta[0], bnd_sta[1], bnd_sta[2]};
    size_t my_len[3] = {bnd_len[0], bnd_len[1], bnd_len[2]};

    // get local indexes from global indexes
    if (!getGlobal2LocalIndexes(my_sta, my_len, 0, paraKey)) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid getGlobal2LocalIndexes.");
      return false;
    }

    // 返値にセットする。
    for (i=0; i<3; i++) {
      bnd_sta_list[n][i] = my_sta[i];
      bnd_len_list[n][i] = my_len[i];
    }
  }

  return true;
}


/**
 * データを設定する。（ガイドセルは含む領域）
 *
 * すべてのクラスタイプにおいてもパラメータは、i,j,k,l固定である。
 * i,j,kはガイドセルを含む領域のインデックスである。
 * 重なる部分領域のセルは、すべての部分領域データに対してデータをセットする。
 * @param    val  データ値
 * @param    i    I方向ノード領域インデックス（ガイドセルは含む）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含む）：1Dの場合はセルのデータ数
 * @param    k    K方向ノード領域インデックス（ガイドセルは含む）：2Dの場合はセルのデータ数
 * @param    l    セルのデータ数 : 3Dの場合
 * @return   true=インデックス位置にセルが存在し、データをセットした。
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::_setValue(DATATYPE val, size_t i, size_t j, size_t k, size_t l)
{
  bool exists = false;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;

    size_t child_idx = 0;
    // サブブロックのローカルインデックスを取得する
    // 0=領域外, 1=ガイドセル領域, 2=内部領域
    int idx_loc = child_data->_getLocalIndex(child_idx, i, j, k, l);
    if (idx_loc <= 0) continue;       // サブブロック領域外

    // データを取得する
    DATATYPE* dt = child_data->getData();
    dt[child_idx] = val;
    exists = true;
  }
  return exists;
}


/**
 * データを設定する。（ガイドセルは含まない領域）
 *
 * すべてのクラスタイプにおいてもパラメータは、i,j,k,l固定である。
 * i,j,kはガイドセルを含む領域のインデックスである。
 * 重なる部分領域のセルは、すべての部分領域データに対してデータをセットするが、
 * ガイドセル以外のセルに対してセットを行うので、重なるセルはないはずである。
 * @param    val  データ値
 * @param    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param    j    J方向ノード領域インデックス（ガイドセルは含まない）：1Dの場合はセルのデータ数
 * @param    k    K方向ノード領域インデックス（ガイドセルは含まない）：2Dの場合はセルのデータ数
 * @param    l    セルのデータ数 : 3Dの場合
 * @return   true=インデックス位置にセルが存在し、データをセットした。
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::setValue(DATATYPE val, size_t i, size_t j, size_t k, size_t l)
{
  // プロセス領域の始点インデックス、サイズ
  const sphParaController* proc_ctrl = this->getParaController();
  if( !proc_ctrl ) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphParaController is null.");
    return false;
  }
  const size_t *proc_sz = this->_getArraySize();
  size_t proc_gc = this->getGc();
  size_t proc_start[3] = {0, 0, 0};
  proc_ctrl->getVoxelStartIndex(proc_start);

  bool exists = false;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;
    // パラレルコントローラの取得
    const sphParaController* child_ctrl = child_data->getParaController();
    if( !child_ctrl ) {
      // error
      sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphParaController is null.");
      return false;
    }

    // 部分領域の始点インデックス、サイズ
    const size_t *child_sz = child_data->getArraySize();    // ガイドセルを含まないサイズ
    size_t child_gc = child_data->getGc();
    size_t child_start[3] = {0, 0, 0};
    child_ctrl->getVoxelStartIndex(child_start);
    size_t child_len = child_data->getArrayLength();

    // 指定インデックスが部分領域内に存在するか？
    int diff[3] = {child_start[0]-proc_start[0], child_start[1]-proc_start[1], child_start[2]-proc_start[2]};
    if (i < diff[0] || i >= diff[0] + child_sz[0]) continue;
    if (j < diff[1] || j >= diff[1] + child_sz[1]) continue;
    if (k < diff[2] || k >= diff[2] + child_sz[2]) continue;

    // 部分領域内に存在するのでサブブロックインデックスを取得する。（ガイドセルを含まないインデックス）
    size_t child_idx = child_data->getIndex(i-diff[0], j-diff[1], k-diff[2]);
    if (child_idx >= child_len) continue;

    // データを取得する
    DATATYPE* dt = child_data->getData();
    dt[child_idx] = val;
    exists = true;
  }
  return exists;
}


/**
 * データを取得する。（ガイドセルは含む領域）
 *
 * すべてのクラスタイプにおいてもパラメータは、i,j,k,l固定である。
 * i,j,kはガイドセルを含む領域のインデックスである。
 * 重なる部分領域のセルは、ガイドセル以外のセルを優先に取得する。
 * ガイドセル以外のセルが存在しなければ、ガイドセル領域のセルのデータを取得する。
 * @param[out]    val  データ値
 * @param[in]    i    I方向ノード領域インデックス（ガイドセルは含む）
 * @param[in]    j    J方向ノード領域インデックス（ガイドセルは含む）：1Dの場合はセルのデータ数
 * @param[in]    k    K方向ノード領域インデックス（ガイドセルは含む）：2Dの場合はセルのデータ数
 * @param[in]    l    セルのデータ数 : 3Dの場合
 * @return      true=インデックス位置にセルが存在する。
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::_getValue(DATATYPE& val, size_t i, size_t j, size_t k, size_t l) const
{

  bool exists = false;
  DATATYPE tmp_val = (DATATYPE)0.0;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;

    size_t child_idx = 0;
    // サブブロックのローカルインデックスを取得する
    // 0=領域外, 1=ガイドセル領域, 2=内部領域
    int idx_loc = child_data->_getLocalIndex(child_idx, i, j, k, l);
    if (idx_loc <= 0) continue;       // サブブロック領域外

    // データを取得する
    DATATYPE* dt = child_data->getData();
    if (idx_loc == 2) {
      // ガイドセル領域でないので、データを取得して終了する。
      val = dt[child_idx];
      return true;
    }
    else {
      // ガイドセル領域であるので、一時取得を行う。
      tmp_val = dt[child_idx];
    }
    exists = true;
  }
  val = tmp_val;

  return exists;
}


/**
 * データを取得する。（ガイドセルは含まない領域）
 *
 * すべてのクラスタイプにおいてもパラメータは、i,j,k,l固定である。
 * i,j,kはガイドセルを含む領域のインデックスである。
 * 重なる部分領域のセルは、ガイドセル以外のセルを取得する。
 * ガイドセル領域のセルのデータは取得しない。
 * @param[out]    val  データ値
 * @param[in]    i    I方向ノード領域インデックス（ガイドセルは含まない）
 * @param[in]    j    J方向ノード領域インデックス（ガイドセルは含まない）：1Dの場合はセルのデータ数
 * @param[in]    k    K方向ノード領域インデックス（ガイドセルは含まない）：2Dの場合はセルのデータ数
 * @param[in]    l    セルのデータ数 : 3Dの場合
 * @return      true=インデックス位置にセルが存在する。
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::getValue(DATATYPE& val, size_t i, size_t j, size_t k, size_t l) const
{
  bool exists = false;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;

    size_t child_idx = 0;
    // サブブロックのローカルインデックスを取得する
    // 0=領域外, 2=内部領域
    int idx_loc = child_data->getLocalIndex(child_idx, i, j, k, l);
    if (idx_loc <= 0) continue;       // サブブロック領域外

    // データを取得する
    DATATYPE* dt = child_data->getData();
    val = dt[child_idx];
    return true;
  }
  return exists;
}


/**
 * データオブジェクトの添え字演算子
 * @param   index     配列インデックス
 */
template <class DATATYPE>
SPH_MEMFUN(DATATYPE&)
sphVoxContainer<DATATYPE>::operator[](size_t index)
{
  size_t array_len = this->getArrayLength();
  if (index >= array_len) {
    assert(0);
  }

  // indexからi,j,kを求める
  size_t i=0, j=0, k=0, l=0;
  unsigned int dims = this->_getIndexDims(index, i, j, k, l);

  DATATYPE *tmp_val = NULL;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;

    size_t child_idx = 0;
    // サブブロックのローカルインデックスを取得する
    // 0=領域外, 1=ガイドセル領域, 2=内部領域
    int idx_loc = child_data->_getLocalIndex(child_idx, i, j, k, l);
    if (idx_loc <= 0) continue;       // サブブロック領域外

    // データを取得する
    DATATYPE* dt = child_data->getData();
    if (idx_loc == 2) {
      // ガイドセル領域でないので、データを取得して終了する。
      return dt[child_idx];
    }
    else {
      // ガイドセル領域であるので、一時取得を行う。
      tmp_val = &dt[child_idx];
    }
  }

  if (tmp_val == NULL) {
    // プロセス領域であるが部分データ領域が存在しないので、ダミーデータを返す。
    static DATATYPE dummy = 0;
    return dummy;
  }

  return *tmp_val;
}

/**
 * データオブジェクトの添え字演算子(const)
 * @param   index     配列インデックス
 */
template <class DATATYPE>
SPH_MEMFUN(const DATATYPE&)
sphVoxContainer<DATATYPE>::operator[](size_t index) const
{

  size_t array_len = this->getArrayLength();
  if (index >= array_len) {
    assert(0);
  }

  // indexからi,j,kを求める
  size_t i=0, j=0, k=0, l=0;
  unsigned int dims = this->_getIndexDims(index, i, j, k, l);

  DATATYPE *tmp_val = NULL;
  typename std::map<int, sphVoxArray<DATATYPE>*>::const_iterator itr = m_voxArrayMap.begin();
  for ( ; itr != m_voxArrayMap.end(); itr++) {
    int localid = itr->first;
    sphVoxArray<DATATYPE>* child_data = itr->second;

    size_t child_idx = 0;
    // サブブロックのローカルインデックスを取得する
    // 0=領域外, 1=ガイドセル領域, 2=内部領域
    int idx_loc = child_data->_getLocalIndex(child_idx, i, j, k, l);
    if (idx_loc <= 0) continue;       // サブブロック領域外

    // データを取得する
    DATATYPE* dt = child_data->getData();
    if (idx_loc == 2) {
      // ガイドセル領域でないので、データを取得して終了する。
      return dt[child_idx];
    }
    else {
      // ガイドセル領域であるので、一時取得を行う。
      tmp_val = &dt[child_idx];
    }
  }

  if (tmp_val == NULL) {
    // プロセス領域であるが部分データ領域が存在しないので、ダミーデータを返す。
    static DATATYPE dummy = 0;
    return dummy;
  }

  return *tmp_val;
}


/**
 * データを中央部分にコピーする.
 * データ型、データクラス、格子当りデータ数は同じでない場合はエラーとする。
 * ガイドセル領域もコピーを行う。
 * @param     src_dc           コピー元データ
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::copyCenter(const sphArray<DATATYPE>* src_dc)
{
  const sphVoxContainer<DATATYPE>* src_cont = dynamic_cast<const sphVoxContainer<DATATYPE>*>(src_dc);
  if (src_cont == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "src_dc is not sphVoxContainer.");
    return false;
  }

  // 部分領域データ管理の場合、サブブロック毎に行う。
  int sub_count = this->getNumberOfVoxArray();
  for (unsigned n=0; n<sub_count; n++) {
    // 部分領域データクラスの取得
    sphVoxArray<DATATYPE> *sub_dst = this->getVoxArray(n);
    const sphVoxArray<DATATYPE> *sub_src = src_cont->getVoxArray(n);
    if (!sub_dst->copyCenter(sub_src)) {
      return false;
    }
  }
  return true;

}



/**
 * 成分データを中央部分にコピーする.
 * ガイドセル領域もコピーを行う。
 * @param     src_dc           コピー元データ
 * @param     dst_pos          コピー先成分インデックス
 * @param     src_pos          コピー元成分インデックス
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::copyComponent(
                    const sphArray<DATATYPE>* src_dc,
                    size_t dst_pos,
                    size_t src_pos)
{
  const sphVoxContainer<DATATYPE>* src_cont = dynamic_cast<const sphVoxContainer<DATATYPE>*>(src_dc);
  if (src_cont == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "src_dc is not sphVoxContainer.");
    return false;
  }

  // 部分領域データ管理の場合、サブブロック毎に行う。
  int sub_count = this->getNumberOfVoxArray();
  for (unsigned n=0; n<sub_count; n++) {
    // 部分領域データクラスの取得
    sphVoxArray<DATATYPE> *sub_dst = this->getVoxArray(n);
    const sphVoxArray<DATATYPE> *sub_src = src_cont->getVoxArray(n);
    if (!sub_dst->copyComponent(sub_src, dst_pos, src_pos)) {
      return false;
    }
  }
  return true;
}

/**
 * データクラスSPH_DC_ARRAY_3DNからSPH_DC_ARRAY_3DNRV又は反対にコピーする.
 * ガイドセル領域もコピーを行う。
 * @param     src_dc           コピー元データ
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::copyReverse(const sphArray<DATATYPE>* src_dc)
{
  const sphVoxContainer<DATATYPE>* src_cont = dynamic_cast<const sphVoxContainer<DATATYPE>*>(src_dc);
  if (src_cont == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "src_dc is not sphVoxContainer.");
    return false;
  }

  // 部分領域データ管理の場合、サブブロック毎に行う。
  int sub_count = this->getNumberOfVoxArray();
  for (unsigned n=0; n<sub_count; n++) {
    // 部分領域データクラスの取得
    sphVoxArray<DATATYPE> *sub_dst = this->getVoxArray(n);
    const sphVoxArray<DATATYPE> *sub_src = src_cont->getVoxArray(n);
    if (!sub_dst->copyReverse(sub_src)) {
      return false;
    }
  }
  return true;
}


/**
 * データに対して計算を行う.
 * ガイドセル領域も計算を行う。
 * @param     mode             計算モード
 *                1:dst_data[dst_idx] += src_data[src_idx];
 *                2:dst_data[dst_idx] *= stepAvr;
 *                3:dst_data[dst_idx] /= stepAvr;
 *                4:dst_data[dst_idx] = src_data[src_idx]*stepAvr+v00[0];
 * @param     src_dc             計算元データ
 * @param     stepAvr            乗算値
 * @param     v00                加算値（v00[data_num])
 * @return    success(=true)/fault(=false)
 */
template <class DATATYPE>
SPH_MEMFUN(bool)
sphVoxContainer<DATATYPE>::scaleAvr(int mode,
                          const sphArray<DATATYPE>* src_dc,
                          unsigned stepAvr,
                          double* va)
{
  const sphVoxContainer<DATATYPE>* src_cont = dynamic_cast<const sphVoxContainer<DATATYPE>*>(src_dc);
  if (src_cont == NULL) {
    // error
    sphArray<DATATYPE>::m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "src_dc is not sphVoxContainer.");
    return false;
  }

  // 部分領域データ管理の場合、サブブロック毎に行う。
  int sub_count = this->getNumberOfVoxArray();
  for (unsigned n=0; n<sub_count; n++) {
    // 部分領域データクラスの取得
    sphVoxArray<DATATYPE> *sub_dst = this->getVoxArray(n);
    const sphVoxArray<DATATYPE> *sub_src = src_cont->getVoxArray(n);
    if (!sub_dst->scaleAvr(mode, sub_src, stepAvr, va)) {
      return false;
    }
  }
  return true;
}
#endif /* _SPHVOXCONTAINER_INLINE_H_ */
