/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphArray.h
 * Data Array Class Header File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphArray.h
 * データ配列クラスヘッダーファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPH_ARRAY_H_
#define _SPH_ARRAY_H_

#include <iostream>
#include <new>
#include "sphData.h"

#define SPH_MAX_DIM_SIZE_OF_PRIMITIVE_ARRAY 4

/**
 * \~english
 * Data Array Class
 *
 * \~japanese
 * データ配列クラス
 *
 * データ配列を１次元の配列にて管理を行う。<br>
 * ノードサイズ、次元数、ガイドセルサイズの情報を持つ。
 */
template<class DATATYPE>
class sphArray : public sphData
{
protected:
  /// データ配列の次元数（1-4:lを含む）
  unsigned m_pDims;
  /// ノードの次元数（1-3:lを含まない）
  unsigned m_dims;
  /// データ配列の次元サイズ（ガイドセルを含む）
  size_t m_pSz[SPH_MAX_DIM_SIZE_OF_PRIMITIVE_ARRAY];
  /// ノードの次元サイズ（ガイドセルを含まない）
  size_t m_sz [SPH_MAX_DIM_SIZE_OF_PRIMITIVE_ARRAY];
  /// ガイドセル次元サイズ
  size_t m_gcl[SPH_MAX_DIM_SIZE_OF_PRIMITIVE_ARRAY];
  /// ガイドセルサイズ
  size_t m_gc;
  /// セルのデータ数（=l)
  size_t m_dtElemNum;
  /// データ配列
  DATATYPE* m_dt;

  sphArray();

  sphArray(
           const std::string label,
           SphDCType dctype,
           SPL_Datatype dttype,
           const sphParaController *paraCtrlr);

  virtual bool createCommType(int paraKey = -1);

  bool initializeArray(const size_t* array_size,
                  const size_t gc,
                  const size_t data_num,
                  bool allocated = true);

  DATATYPE* allocateData(size_t array_len = 0);
  unsigned int _getIndexDims(size_t index, size_t &i, size_t &j, size_t &k, size_t &l) const;

public:

  virtual ~sphArray();

  virtual unsigned getDims() const;
  virtual const size_t* getArraySize() const;
  virtual const size_t* _getArraySize() const;
  virtual size_t getArrayLength() const;
  virtual size_t getGc() const;
  virtual size_t getNumOfDtElem() const;
  virtual const void* getDataPointer() const;
  virtual       void* getDataPointer();
  virtual size_t getMemoryBytes() const;

  const DATATYPE* getData() const;
        DATATYPE* getData();

  const DATATYPE& val(size_t i) const;
        DATATYPE& val(size_t i);
  const DATATYPE& val(size_t i, size_t j) const;
        DATATYPE& val(size_t i, size_t j);
  const DATATYPE& val(size_t i, size_t j, size_t k) const;
        DATATYPE& val(size_t i, size_t j, size_t k);
  const DATATYPE& val(size_t i, size_t j, size_t k, size_t l) const;
        DATATYPE& val(size_t i, size_t j, size_t k, size_t l);
  const DATATYPE& _val(size_t i) const;
        DATATYPE& _val(size_t i);
  const DATATYPE& _val(size_t i, size_t j) const;
        DATATYPE& _val(size_t i, size_t j);
  const DATATYPE& _val(size_t i, size_t j, size_t k) const;
        DATATYPE& _val(size_t i, size_t j, size_t k);
  const DATATYPE& _val(size_t i, size_t j, size_t k, size_t l) const;
        DATATYPE& _val(size_t i, size_t j, size_t k, size_t l);

  virtual bool _setValue (DATATYPE  val, size_t i, size_t j=0, size_t k=0, size_t l=0);
  virtual bool setValue  (DATATYPE  val, size_t i, size_t j=0, size_t k=0, size_t l=0);
  virtual bool _getValue (DATATYPE& val, size_t i, size_t j=0, size_t k=0, size_t l=0) const;
  virtual bool getValue  (DATATYPE& val, size_t i, size_t j=0, size_t k=0, size_t l=0) const;
  virtual bool initValues(DATATYPE val);

  size_t getIndex(size_t i, size_t j=0, size_t k=0, size_t l=0) const;
  size_t _getIndex(size_t i, size_t j=0, size_t k=0, size_t l=0) const;

  virtual bool comm(size_t gc = 0, int paraKey = -1);
  virtual bool comm() = 0;
  virtual bool commPeriodic(unsigned int face_dir, int axis) = 0;
  virtual bool commPeriodic(unsigned int face_dir, int axis, size_t face, int paraKey = -1);

  virtual bool copyArray(const sphData* src_dc);
  virtual bool copyCenter(const sphArray<DATATYPE>* src_dc);
  virtual bool copyComponent( const sphArray<DATATYPE>* src_dc,
                      size_t dst_pos,
                      size_t src_pos);
  virtual bool copyReverse(const sphArray<DATATYPE>* src_dc);

  virtual bool scaleAvr(      int mode,
                      const sphArray<DATATYPE>* src_dc,
                      unsigned stepAvr,
                      double* v00);
  virtual DATATYPE& operator[](size_t index);          // 添え字演算子
  virtual const DATATYPE& operator[](size_t index) const;// 添え字演算子(const)

};

#ifdef SPH_INLINE
 #undef SPH_INLINE
#endif /* SPH_INLINE */

#ifndef NO_INLINE
 #define SPH_INLINE inline
#else
 #define SPH_INLINE
#endif /* NO_INLINE */

#include "inline/sphArray_inline.h"

#undef SPH_MAX_DIM_SIZE_OF_PRIMITIVE_ARRAY

#endif // _SPH_ARRAY_H_

