/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphData.h
 * Data Class Header File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphData.h
 * データクラスヘッダーファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPH_DATA_H_
#define _SPH_DATA_H_

#include <iostream>
#include <map>
#include <vector>

#include "base/sphBase.h"
#include "data/sphDataDefs.h"
#include "parallel/sphParaController.h"
#include "parallel/sphSvParaController.h"
#include "spl/spl.h"


/**
 * \~english
 * Data Class
 *
 * \~japanese
 * データ基底クラス
 *
 * データの基本情報を持つ。
 *
 */
class sphData : public virtual sphBase {
private:
  /// パラレルマネージャリスト
  std::map<unsigned int, const sphParaController*>   m_paraCtrlrMap;
  /**
   * デフォルトパラレルマネージャ識別キー
   *
   * 最初に登録されたパラレルマネージャの識別キーであり、既定のパラレルマネージャとする。
   */
  unsigned int m_defualParaKey;

protected:
  /// データ識別ラベル
  std::string m_label;  // data object identifier
  /// データクラスタイプ
  SphDCType   m_dcType; // data-class type
  /// データ型タイプ
  SPL_Datatype m_dtType;

  sphData();
  /// 通信タイプの作成を行う。
  virtual bool createCommType(int paraKey = -1) = 0;

  bool initializeData(
              const std::string label,
              SphDCType dctype,
              SPL_Datatype dttype,
              const sphParaController *paraCtrlr);

public:
  sphData(const std::string label,
          SphDCType dctype,
          SPL_Datatype dttype,
          const sphParaController *paraCtrlr);
  virtual ~sphData();
  std::string getLabel() const;
  void setLabel(const char* label);
  SphDCType getDcType() const;
  SPL_Datatype getDataType() const;

  const sphParaController* getParaController(int paraKey = -1) const;
  bool getDataOrigin(double org[3]) const;
  bool getDataPitch(double org[3]) const;
  bool getDataStartIdx(size_t start_idx[3]) const;

  /// ノード間通信を行う。
  virtual bool comm(size_t gc, int paraKey = -1) = 0;
  virtual bool comm() = 0;
  virtual bool commPeriodic(unsigned int face_dir, int axis) = 0;
  virtual bool commPeriodic(unsigned int face_dir, int axis, size_t face, int paraKey = -1) = 0;

  virtual bool addParaCtrlr(const sphParaController* paraCtrlr);
  virtual bool eraseParaCtrlr(int paraKey = -1);


  // サイズ情報の取得を行う。
  virtual unsigned getDims() const = 0;
  virtual const size_t* getArraySize() const = 0;
  virtual const size_t* _getArraySize() const = 0;
  virtual size_t getArrayLength() const = 0;
  virtual size_t getGc() const = 0;
  virtual size_t getNumOfDtElem() const = 0;
  virtual const void* getDataPointer() const = 0;
  virtual       void* getDataPointer() = 0;

  virtual SphCrdDef getCrdDef() const = 0;
  virtual bool setCrdDef(SphCrdDef crddef) = 0;

  // 部分領域管理データクラス用
  virtual int getNumberOfVoxArray() const;
  virtual sphData* createVoxelArray(int new_gc = -1, bool copy_data = false) {return NULL;};
  virtual bool copyArray(const sphData* src_dc) = 0;
  virtual bool copyVoxelArray(
                const sphData* src_dc,
                size_t src_vc = 0,
                size_t dest_vc = 0,
                int src_paraKey = -1,
                int dest_paraKey = -1) {return NULL;};
  virtual bool isVoxContainer() const;
  virtual sphData* getVoxArray(int localid);
  virtual const sphData* getVoxArray(int localid) const;

};

#ifdef SPH_INLINE
 #undef SPH_INLINE
#endif /* SPH_INLINE */

#ifndef NO_INLINE
 #define SPH_INLINE inline
#else
 #define SPH_INLINE
#endif /* NO_INLINE */

#include "inline/sphData_inline.h"

#endif // _SPH_DATA_H_

