/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphUnstData.h
 * Unstruct Data Class Header File
 * @author     tamaki
 * @date       2009/04/22
 *
 * \~japanese
 * @file sphUnstData.h
 * 非構造格子物理量データクラスヘッダーファイル
 * @author     tamaki
 * @date       2009/04/22
 */

#ifndef _SPH_UNST_DATA_H_
#define _SPH_UNST_DATA_H_

#include <iostream>
#include "data/sphArray.h"
#include "parallel/sphUnstMesh.h"

/**
 * \~english
 * Unstruct Data Data Class
 *
 * \~japanese
 * 非構造格子物理量データクラス
 *
 * - 非構造格子物理量ノードデータの作成、管理。<br>
 * - ノード間通信
 */
template<class DATATYPE>
class sphUnstData : public sphArray<DATATYPE> {
protected:
  sphUnstMesh* m_mesh;
  int m_dtNum;

  // SPL_MPI Infomations
  int          m_nPE;
  int          m_myRank;
  SPL_Comm     m_comm;

protected:
  sphUnstData();

  bool initializeArray(SphDCType dctype,
                       SPL_Datatype dttype,
                       const sphParaController *paraCtrlr,
                       size_t vlen);
  // sphDataクラスの完全仮想関数。sphUnstDataクラスでは使用しない。
  virtual bool createCommType(int paraKey = -1) { return true; };;

public:
  sphUnstData( const std::string label,
               SphDCType dctype,
               SPL_Datatype dttype,
               const sphParaController *paraCtrlr,
               size_t vlen = 1 );

  sphUnstData(
               SphDCType dctype,
               SPL_Datatype dttype,
               const sphParaController *paraCtrlr,
               size_t vlen = 1 );


  virtual ~sphUnstData();
  virtual bool comm();
  virtual bool wait();
  bool writeFile(int istep, const char* file_prefix);
  template <class X>
  bool writeFile(int istep, const char* file_prefix, X coef) {
    if( !file_prefix ) return false;
    int    i, j;
    char   filename[128];
    FILE  *fp;
    if ( m_nPE == 0 ){
      /* ------- Configure output file name ------- */
      if( !file_prefix ) return false;
      sprintf(filename, "%s.%d.dat", file_prefix, istep);
      /* ******** Open result file ******** */
      if( (fp = fopen(filename, "w")) == NULL ) return false;
      /* ******** Write variables ******** */
      fprintf(fp, "%d\n",m_dtNum);
      SphDCType dctype = sphData::getDcType();
      for ( i = 0; i < m_dtNum; i++ ){
        fprintf(fp, "%d", i);
        for( j = 0; j < sphArray<DATATYPE>::m_dtElemNum; j++) {
          size_t index;
          switch(dctype) {
          case SPH_DC_UNST_NODE:
          case SPH_DC_UNST_NODE_N:
          case SPH_DC_UNST_ELEM:
          case SPH_DC_UNST_ELEM_N:
            index = m_dtNum*j + i;
            break;
          case SPH_DC_UNST_NODE_NRV:
          case SPH_DC_UNST_ELEM_NRV:
            index = sphArray<DATATYPE>::m_dtElemNum*i + j;
            break;
          default:
            return false;
          }
          fprintf(fp, " %e", sphArray<DATATYPE>::m_dt[index]*coef);
        }
        fprintf(fp, "\n");
      }
      fclose(fp);
    }
    else if ( m_nPE > 0 ){
      /* ------- Configure output file name ------- */
      if( !file_prefix ) return false;
      sprintf(filename, "%s.%d.%d.dat", file_prefix, istep, m_myRank);
      /* ******** Open result file ******** */
      if( (fp = fopen(filename, "w")) == NULL ) return false;
      /* ******** Write variables ******** */
      fprintf(fp, "%d\n",m_mesh->nNode);
      SphDCType dctype = sphData::getDcType();
      for ( i = 0; i < m_mesh->nNode; i++ ){
        fprintf(fp, "%d", m_mesh->node[i]);
        for( j = 0; j < sphArray<DATATYPE>::m_dtElemNum; j++) {
          size_t index;
          switch(dctype) {
          case SPH_DC_UNST_NODE:
          case SPH_DC_UNST_NODE_N:
          case SPH_DC_UNST_ELEM:
          case SPH_DC_UNST_ELEM_N:
            index = m_mesh->nNode*j + i;
            break;
          case SPH_DC_UNST_NODE_NRV:
          case SPH_DC_UNST_ELEM_NRV:
            index = sphArray<DATATYPE>::m_dtElemNum*i + j;
            break;
          default:
            return false;
          }
          fprintf(fp, " %e", sphArray<DATATYPE>::m_dt[index]*coef);
        }
        fprintf(fp, "\n");
      }
      fclose(fp);
    }
    return true;
  };

public:
  // sphDataクラスの完全仮想関数。sphUnstDataクラスでは使用しない。
  virtual bool comm(size_t gc = 0, int paraKey = -1) { return false; };
  virtual bool commPeriodic(unsigned int face_dir, int axis) { return false; };
  virtual bool commPeriodic(unsigned int face_dir, int axis, size_t face, int paraKey = -1) { return false; };
  virtual unsigned getDims() const { return 0; };
  virtual const size_t* getArraySize() const { return NULL; };
  virtual const size_t* _getArraySize() const { return NULL; };
  virtual size_t getArrayLength() const { return 0; };
  virtual size_t getGc() const { return 0; };
  virtual size_t getNumOfDtElem() const { return 0; };
  virtual SphCrdDef getCrdDef() const { return CRDDEF_UNKNOWN; };
  virtual bool setCrdDef(SphCrdDef crddef) { return false; };

};

#ifdef SPH_INLINE
 #undef SPH_INLINE
#endif /* SPH_INLINE */

#ifndef NO_INLINE
 #define SPH_INLINE inline
#else
 #define SPH_INLINE
#endif /* NO_INLINE */

#include "inline/sphUnstData_inline.h"

#endif // _SPH_UNST_DATA_H_

