/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphVoxArray.h
 * Voxel Data Class Header File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphVoxArray.h
 * ボクセルデータクラスヘッダーファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPH_VOX_ARRAY_H_
#define _SPH_VOX_ARRAY_H_

#include <iostream>
#include <map>
#include <vector>
#include "data/sphArray.h"
#include "parallel/sphParaController.h"

/**
 * ノード間通信タイプクラス
 */
class sphCommType {
  public:
    ///   送信通信タイプ
    SPL_Datatype m_comTypeSend;
    ///   受信通信タイプ
    SPL_Datatype m_comTypeRecv;
    ///  通信ノードID
    int m_comID;
    ///  通信TAG
    int m_tag;
    /// 通信面 (+I, -I, +J, -J, +K, -K)
    FaceType m_commDir;
    /// リクエスト番号
    SPL_Request m_request[2];

    /****   メモリコピー用の領域情報 ****/
    size_t  m_sendStartIdx[3];              ///< 送信データの始点インデックス
    size_t  m_sendSize[3];              ///< 送信データのサイズ
    size_t  m_recvStartIdx[3];              ///< 受信データの始点インデックス
    size_t  m_recvSize[3];              ///< 受信データのサイズ

    /**
     * コンストラクタ
     */
    sphCommType()
    {
      m_comTypeSend = SPL_DATATYPE_NULL;
      m_comTypeRecv = SPL_DATATYPE_NULL;
      m_comID = SPL_PROC_NULL;
      m_tag = SPL_PROC_NULL;
      m_commDir = FACE_UNKNOWN;
      m_request[0] = m_request[1] = SPL_REQUEST_NULL;
      m_sendStartIdx[0] = m_sendStartIdx[1] = m_sendStartIdx[2] = 0;
      m_sendSize[0] = m_sendSize[1] = m_sendSize[2] = 0;
      m_recvStartIdx[0] = m_recvStartIdx[1] = m_recvStartIdx[2] = 0;
      m_recvSize[0] = m_recvSize[1] = m_recvSize[2] = 0;
    }
};


/// \~english  sphCommType Map <Comm face, CommTypeList>  \~japanese  sphCommType Map <送受信面, sphCommTypeリスト>
typedef std::map<size_t, std::vector<sphCommType*> > CommTypeFaceMap;
/// Paralle Controller Map <ParaKey, CommTypeFaceMap>
typedef std::map<int, CommTypeFaceMap > ParaCommMap;

/**
 * \~english
 * Voxel Data Class
 *
 * \~japanese
 * ボクセルデータクラス
 *
 * - ボクセルデータの作成、管理。<br>
 * - ノード間通信
 */
template<class DATATYPE>
class sphVoxArray : public sphArray<DATATYPE> {

template<class DATATYPE> friend class sphVoxContainer;

private:

  /**
  * \~english
  * Map Communication Type for Parallel COntroller
  *
  * \~japanese
  * パラレルコントローラー別の通信タイプマップ
  *
  * パラレルコントローラーの識別Key、通信面、通信タイプのマップとして管理を行う。
  */
  ParaCommMap m_mapParaComm;

  /**
  * EV型通信用パラレルコントローラー別の通信タイプマップ
  *
  * EV型通信のパラレルコントローラーの識別Key、通信面、通信タイプのマップとして管理を行う。
  */
  ParaCommMap m_mapParaCommEv;

  /**
   * データ定義位置
   */
  SphCrdDef m_crddef;

  /**
   * 親ボクセルデータクラス(=sphVoxContainer<DATATYPE>)
   */
  const sphVoxArray<DATATYPE> *m_parentContainer;

  void getReceiveCommBndIndexes(
                int dims, size_t vcell, size_t comm_face,
                const size_t* array_min, const size_t* array_max,
                const size_t* other_min, const size_t* other_max,
                const size_t* div_min, const size_t* div_max,
                size_t* bnd_sta, size_t* bnd_end) const;

  void getSendCommBndIndexes(
                int dims, size_t vcell, size_t comm_face,
                const size_t* array_min, const size_t* array_max,
                const size_t* other_min, const size_t* other_max,
                const size_t* div_min, const size_t* div_max,
                size_t* bnd_sta, size_t* bnd_end) const;

  void getReceiveCommBndIndexesEv(
                int dims, size_t vcell, size_t comm_face, FaceType face_dir,
                const size_t* array_min, const size_t* array_max,
                size_t* bnd_sta, size_t* bnd_end) const;

  void getSendCommBndIndexesEv(
                  int dims, size_t vcell, size_t comm_face, FaceType face_dir,
                  const size_t* array_min, const size_t* array_max,
                  size_t* bnd_sta, size_t* bnd_end) const;

  bool isSupportedCrdDef(
                SphCrdDef crddef,
                SphDCType dctype,
                const size_t data_num);

  virtual bool commWait(size_t face);
  bool commDataCopy(size_t face, DirFaceType send_dir, int paraKey = -1);
  bool commPeriodicISendRecv(unsigned int face_dir, int axis, size_t face, int paraKey);
  bool commPediodicDataCopy(unsigned int face_dir, int axis, size_t face, int paraKey);

  bool getBndMyIndex(
                        DATATYPE id,
                        size_t* bnd_sta,
                        size_t* bnd_len,
                        int paraKey,
                        bool extgcFlag) const;

  bool initializeVoxArray(
              const std::string label,
              SphDCType dctype,
              SPL_Datatype dttype,
              const sphParaController *paraCtrlr,
              const size_t gc = 0,
              const size_t data_num = 1,
              SphCrdDef crddef = CRDDEF_REGULAR,
              bool allocated = true);

  void setParentContainer(const sphVoxArray<DATATYPE> *parentContainer);
  const sphCommType* getCommType(size_t face, int paraKey, int nodeid) const;
  const sphCommType* getCommTypeByFaceType(size_t face, int paraKey, FaceType facetype) const;
  int getCommTypeList(std::vector<sphCommType*>* commTypes, size_t face, int paraKey) const;

protected:
  sphVoxArray();

  virtual bool createCommType(int paraKey = -1);
  virtual bool createCommTypeEv(int paraKey = -1);

  SPL_Datatype createCommType(const size_t* max, const size_t* sta, const size_t* end);

  SPL_Datatype createCommTypeStaggerd(
                    const size_t* max,
                    const size_t* sta_stg,
                    const size_t* end_stg,
                    const size_t* sta_col,
                    const size_t* end_col);

  virtual bool commMb(size_t face, int paraKey = -1);
  virtual bool commMb();
  virtual bool commEv(size_t face, int paraKey = -1);
  virtual bool commEv();
  int getCommRequest(size_t face, SPL_Request *req_list);
  void clearCommRequest(size_t face);
  bool copyVoxelContainer(
                      const sphVoxArray<DATATYPE>* src_dc,
                      size_t src_vc, size_t dest_vc,
                      int src_paraKey = -1, int dest_paraKey = -1);
  int _getLocalIndex(size_t &index, size_t i, size_t j, size_t k, size_t l) const;
  int getLocalIndex(size_t &index, size_t i, size_t j, size_t k, size_t l) const;

public:
  sphVoxArray(
              const std::string label,
              SphDCType dctype,
              SPL_Datatype dttype,
              const sphParaController *paraCtrlr,
              const size_t gc = 0,
              const size_t data_num = 1,
              SphCrdDef crddef = CRDDEF_REGULAR);

  sphVoxArray(SphDCType dctype,
              SPL_Datatype dttype,
              const size_t* array_size,
              const size_t gc = 0,
              const size_t data_num = 1);

  virtual ~sphVoxArray();

  // communicator
  virtual bool comm(size_t face, int paraKey = -1);
  virtual bool comm();
  virtual bool commPeriodic(unsigned int face_dir, int axis);
  virtual bool commPeriodic(unsigned int face_dir, int axis, size_t face, int paraKey = -1);
  bool commISendRecv(size_t face, DirFaceType send_dir = DIR_UNKNOWN, int paraKey = -1);
  bool commWaitAll();

  virtual bool addParaCtrlr(const sphParaController* paraCtrlr);
  virtual SphCrdDef getCrdDef() const;
  virtual bool setCrdDef(SphCrdDef crddef);

  // データ数を取得する。
  virtual size_t getWNumberOfID(DATATYPE id, int paraKey = -1) const;
  virtual size_t getNumberOfID(DATATYPE id) const;
  virtual size_t getWNumberOfIDExtGc(DATATYPE id, int paraKey = -1) const;
  virtual size_t getNumberOfIDExtGc(DATATYPE id) const;

  // データ領域を取得する。（ガイドセルを含まない）
  virtual bool getBndWIndex(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool getBndIndex(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool getBndWIndexMb(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool getBndIndexMb(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  // データ領域を取得する。（ガイドセルを含む）
  virtual bool getBndWIndexExtGc(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool getBndIndexExtGc(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool getBndWIndexMbExtGc(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool getBndIndexMbExtGc(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool copyVoxelArray(
                const sphData* src_dc,
                size_t src_vc = 0,
                size_t dest_vc = 0,
                int src_paraKey = -1,
                int dest_paraKey = -1);

  virtual bool getGlobal2LocalIndexes(
                      size_t* min,
                      size_t* len,
                      size_t gc = 0,
                      int paraKey = -1) const;

  bool transferData( const sphVoxArray<DATATYPE> *src_dc,
                    const size_t  src_startIdx[3],
                    const size_t  src_len[3],
                    const size_t  dest_startIdx[3]);

  virtual sphData* createVoxelArray(int new_gc = -1, bool copy_data = false);

  virtual int getNumberOfVoxArray() const;
  virtual sphVoxArray<DATATYPE>* getVoxArray(int localid);
  virtual const sphVoxArray<DATATYPE>* getVoxArray(int localid) const;

#ifdef _DEBUGLOG_
  void debugPrint() const;
#endif

};

#ifdef SPH_INLINE
 #undef SPH_INLINE
#endif /* SPH_INLINE */

#ifndef NO_INLINE
 #define SPH_INLINE inline
#else
 #define SPH_INLINE
#endif /* NO_INLINE */

#include "inline/sphVoxArray_inline.h"

#endif // _SPH_VOX_ARRAY_H_

