/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * @file sphVoxContainer.h
 * 部分領域管理データクラスヘッダーファイル
 * @author     hira
 * @date       2009/02/20
 */


#ifndef _SPHVOXCONTAINER_H_
#define _SPHVOXCONTAINER_H_

#include "data/sphVoxArray.h"
#include "parallel/sphMultiParaController.h"


/**
 * 部分領域管理データクラス
 */
template<class DATATYPE>
class sphVoxContainer : public sphVoxArray<DATATYPE> {
private:

  /// ボクセルデータクラスリスト (=サブブロック領域)
  std::map<int, sphVoxArray<DATATYPE>*>   m_voxArrayMap;

  /**
   * EV通信の一括/順次フラグ
   *
   * true=I,J,K方向の通信を一括で行う。false= I,J,K方向の通信を順次行う。デフォルト=順次実行(=false)
   */
  bool m_evCommBulk;

  sphVoxArray<DATATYPE>* createVoxelArray(
                        SphDCType dcType,
                        SPL_Datatype dataType,
                        std::string label,
                        const sphParaController *paraCtrlr,
                        const size_t gc,
                        const size_t data_num,
                        SphCrdDef crddef);

  bool addChildVoxelArray(int localid, sphVoxArray<DATATYPE>* data);

  bool getBndWIndexByReduce(
                        DATATYPE id,
                        size_t* bnd_sta,
                        size_t* bnd_len,
                        const size_t ext_gc,
                        int paraKey,
                        bool extgcFlag ) const;
  virtual bool commWaitAll(size_t face);

protected:
  sphVoxContainer();

  virtual bool commMb(size_t face, int paraKey = -1);
  virtual bool commMb();
  bool commEvOrder(size_t face, int paraKey = -1);
  bool commEvOrder();
  bool commEvBulk(size_t face, int paraKey = -1);
  bool commEvBulk();

public:
  sphVoxContainer(
              const std::string label,
              SphDCType dctype,
              SPL_Datatype dttype,
              const sphParaController *paraCtrlr,
              const size_t gc = 0,
              const size_t data_num = 1,
              SphCrdDef crddef = CRDDEF_REGULAR);
  virtual ~sphVoxContainer();

  virtual int getNumberOfVoxArray() const;
  virtual sphVoxArray<DATATYPE>* getVoxArray(int localid);
  virtual const sphVoxArray<DATATYPE>* getVoxArray(int localid) const;
  sphVoxArray<DATATYPE>* getVoxArrayByNodeId(int nodeid);
  const sphVoxArray<DATATYPE>* getVoxArrayByNodeId(int nodeid) const;

  // communicator
  virtual bool comm(size_t face, int paraKey = -1);
  virtual bool comm();
  virtual bool commPeriodic(unsigned int face_dir, int axis);
  virtual bool commPeriodic(unsigned int face_dir, int axis, size_t face, int paraKey = -1);

  // データ数を取得する。
  virtual size_t getNumberOfID(DATATYPE id) const;
  virtual size_t getNumberOfIDExtGc(DATATYPE id) const;

  // データ領域を取得する。（ガイドセルを含まない）
  virtual bool getBndWIndex(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool getBndWIndexMb(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool getBndIndexMb(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  // データ領域を取得する。（ガイドセルを含む）
  virtual bool getBndWIndexExtGc(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool getBndWIndexMbExtGc(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool getBndIndexMbExtGc(
                DATATYPE id,
                size_t* bnd_sta,
                size_t* bnd_len,
                const size_t ext_gc = 0,
                int paraKey = -1) const;

  virtual bool copyVoxelArray(
                const sphData* src_dc,
                size_t src_vc = 0,
                size_t dest_vc = 0,
                int src_paraKey = -1,
                int dest_paraKey = -1);
  virtual bool copyArray(const sphData* src_dc);
  virtual bool copyCenter(const sphArray<DATATYPE>* src_dc);
  virtual bool copyComponent( const sphArray<DATATYPE>* src_dc,
                      size_t dst_pos,
                      size_t src_pos);
  virtual bool copyReverse(const sphArray<DATATYPE>* src_dc);

  virtual bool scaleAvr(      int mode,
                      const sphArray<DATATYPE>* src_dc,
                      unsigned stepAvr,
                      double* v00);

  void setEvCommBulk(bool commflag);

  virtual sphData* createVoxelArray(int new_gc = -1, bool copy_data = false);
  virtual bool isVoxContainer() const;
  virtual bool initValues(DATATYPE val);
  virtual size_t getMemoryBytes() const;

  bool getBndIndexExtGc_InVoxArray(
                        DATATYPE id,
                        size_t bnd_sta_list[][3],
                        size_t bnd_len_list[][3],
                        const size_t ext_gc = 0,
                        int paraKey = -1) const;
  bool getBndIndex_InVoxArray(
                        DATATYPE id,
                        size_t bnd_sta_list[][3],
                        size_t bnd_len_list[][3],
                        const size_t ext_gc = 0,
                        int paraKey = -1) const;

  virtual bool _setValue (DATATYPE  val, size_t i, size_t j=0, size_t k=0, size_t l=0);
  virtual bool setValue  (DATATYPE  val, size_t i, size_t j=0, size_t k=0, size_t l=0);
  virtual bool _getValue (DATATYPE& val, size_t i, size_t j=0, size_t k=0, size_t l=0) const;
  virtual bool getValue  (DATATYPE& val, size_t i, size_t j=0, size_t k=0, size_t l=0) const;
  virtual DATATYPE& operator[](size_t index);          // 添え字演算子
  virtual const DATATYPE& operator[](size_t index) const;  // 添え字演算子(const)

};

#ifdef SPH_INLINE
 #undef SPH_INLINE
#endif /* SPH_INLINE */

#ifndef NO_INLINE
 #define SPH_INLINE inline
#else
 #define SPH_INLINE
#endif /* NO_INLINE */

#include "inline/sphVoxContainer_inline.h"

#endif /* _SPHVOXCONTAINER_H_ */
