/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphFileController.h
 * Sphere File Controller Class header file
 * @author     hira
 * @date       2009/09/07
 *
 * \~japanese
 * @file sphFileController.h
 * Sphere ファイルコントローラークラスヘッダーファイル
 * @author     hira
 * @date       2009/09/07
 */

#ifndef SPHFILECONTROLLER_H_
#define SPHFILECONTROLLER_H_

#include "base/sphBase.h"
#include "spl/splmpi.h"
#include "data/sphData.h"
#include "fileio/sphFile.h"
#include "fileio/sphSvFile.h"
#include "fileio/sphFileHeader.h"
#include "config/sphElem.h"

/**
 * Sphereサポートのファイルの入出力を行う。
 *
 */
class sphFileController: public sphBase
{
protected:
  const char* m_fileType;               ///< ファイルタイプ("sbx","spx")
  const sphElem* m_fileElem;            ///< ファイル要素
  sphSvFile* m_svFileObj;               ///< ファイルIOオブジェクト
  sphSvHeader* m_fileHeader;            ///< ファイルヘッダーオブジェクト
  char m_inputDir[MAXPATHLEN];                ///< 入力フォルダ
  char m_outputDir[MAXPATHLEN];                ///< 入力フォルダ

  // ファイル名書式パラメータ
  int m_formatRankId;                   ///< ファイル名書式：ランク番号
  int m_formatSteps;                 ///< ファイル名書式：ステップ番号

  // ファイル始点インデックス
  size_t m_assignedStartIdx[3];

  // 入出力グループ
  SPL_Comm m_ioComm;

  sphSvFile* createFileObj(const char* file_type);
  sphFileHeader* createFileHeader(const char* file_type);
  const char* addFileList(const sphElem* file_elem);
  sphData* readFile(const sphElem* file_elem);
  bool writeFile(const sphElem* file_elem);
  bool getFileFormatName(const sphElem* file_elem, char* file_name);
  void setGcInSph(int gc_insph);

public:
  sphFileController();
  virtual ~sphFileController();
  bool initFileCtrl(const sphElem* file_elem);
  bool initFileCtrl(const char* format);
  bool isSupprotedFormat();

  sphData* readFile();
  bool writeFile();
  bool writeFile( const char* file_name,
                  int mode,
                  int write_rankid,
                  bool compressFlag,
                  bool extRecodeFlag,
                  unsigned int aux);

  bool getFileInfo(sphFileHeader* header);
  const sphFileHeader* getFileInfo() const;
  sphFileHeader* getFileInfo();
  bool setFileInfo(const sphFileHeader* header);

  sphData* getData();
  void setData(sphData* data);

  /**
   * 入出力制御関数
   */
  /// 読み込みモードを設定する。
  void setReadMode(int mode);
  /// 書き込みモードを設定する。
  void setWriteMode(int mode);
  /// 圧縮書込みを行うかのフラグを設定する。(true:圧縮書込みを行う。, false:圧縮書込みを行わない。)
  void setCompressionFlag(bool flag);
  /// 拡張レコードを書込みかのフラグを設定する。(true:拡張レコードを書き込む。, false:拡張レコードを書き込まない。)
  void setExtRecodeFlag(bool flag);

  /**
   * 読込/書込ファイル関数
   */
  /// 読み込みファイルリストにファイル、始点インデックスを追加する。
  bool addSvFileInfo(const char* filename, const size_t sidx[3]);
  /// 読み込みファイルリストにファイルを追加する。
  bool addSvFileName(const char* filename);
  /// 読み込みファイルリストをクリアする。
  void clearSvFileList();
  /// 読み込みファイルリストを設定する。
  void setSvFileList(splSvFileList* filelist);
  /// 入出力フォルダを設定する
  bool setInputBaseDir(const char* basedir);
  bool setOutputBaseDir(const char* basedir);


  /**
   * ビット操作関数
   */
  sphData* maskUpperBits();
  sphData* shiftLowBits();

  /**
   * ランク・グループ関数
   */
  void setProcGrp(SPL_Comm comm);
  void setRankId(int rankId);

  /**
   * 読込・書込設定
   */
  void setVoxelInfo(
                    int dims,
                    const size_t size[3],
                    const size_t sta_idx[3],
                    size_t gc,
                    const size_t wsize[3] = NULL);
  void setFileFormatRankId(int rankid);
  void setFileFormatSteps(size_t step);
  bool setDataType(SPL_Datatype dtype);
  bool setVectorType(SphDCType dctype, size_t data_num);
  bool setRlen(unsigned int rlen);
  void setBlockSize(size_t block_size);
  bool setStepTime(size_t step, double time);
  bool setCoordsInfo(double org[3], double pitch[3], SphCrdDef crddef);
  bool setAux(unsigned int aux);
  void setAssignedStartIndex(const size_t sta_idx[3]);

};

#endif /* SPHFILECONTROLLER_H_ */
