/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSpxFile.h
 * SPX File Input/Output Class Header File
 * @author     hira
 * @date       2009/09/07
 *
 * \~japanese
 * @file sphSpxFile.h
 * SPXファイル入出力クラスヘッダーファイル
 * @author     hira
 * @date       2009/09/07
 */

#ifndef _SPHSPXFILE_H_
#define _SPHSPXFILE_H_

#include "fileio/sphSvFile.h"
#include "fileio/sphSpxHeader.h"

/**
 * \~english
 * SPX File Input/Output Class
 *
 * \~japanese
 * SPXファイル入出力クラスヘッダーファイル
 *
 * SPXファイルの入出力を行う。<br>
 * libspl.aの低レベル関数群splfio_spxのラッパークラスである。
 */
class sphSpxFile : public sphSvFile
{
private:


protected:
  sphSpxHeader*  m_header;      ///< SPXデータセット
  sphData *m_sphArray;      ///< 生成データクラス

  /**
   * \~english
   * read mode
   * - 0:all nodes read
   * - 1:only readRank read data
   * - 2:only readRank read data, and readRank not expand data.
   *
   * \~japanese
   * 読み込みモード
   * - ０：全てのノードがファイルをオープンする
   * - １：識別子readRankのノードがファイルをオープンし、読み込んだ値を他のノードにブロードキャストする。
   *      このとき、識別子readRankのノードも情報を保持する。
   * - ２：識別子readRankのノードがファイルをオープンし、読み込んだ値を他のノードにブロードキャストする。
   *      このとき、識別子readRankのノードは情報を保持しない。
   */
  int m_readMode;

  /**
   * \~english
   * write mode
   * - 0:all nodes write
   * - 1:only writeRank write data. Output gather data.
   * - 2:only writeRank write data. Output gather data.
   *     and writeRank has not my file object.
   *
   * \~japanese
   * 書込みモード
   * - ０：全てのノードがファイルをオープンする
   * - １：識別子readRankのノードがファイルをオープンし、他のノードから情報を集めてデータを書き出す。
   *      このとき、識別子readRankのノードも出力情報を持っている。
   * - ２：識別子readRankのノードがファイルをオープンし、他のノードから情報を集めてデータを書き出す。
   *      このとき、識別子readRankのノードは出力情報を持たない。
   */
  int m_writeMode;
  bool m_compression;      ///< 圧縮書込みを行うかのフラグ
  bool m_extRecode;      ///< 拡張レコードを書込みかのフラグ

  // 初期化を行う
  bool initialize();

  /**
   * データクラス関数
   */
  /// sphDataクラスを生成する。
  virtual sphData* createSphArray(splSpx* obj);
  /// 圧縮書込みを行うかのフラグを取得する。(true:圧縮書込みを行う。, false:圧縮書込みを行わない。)
  virtual bool isCompressionFlag();
  /// 読み込みモードを取得する。
  virtual int getReadMode() const;
  /// 書き込みモードを取得する。
  virtual int getWriteMode() const;

public:
  sphSpxFile();
  virtual ~sphSpxFile();

  /// 破棄処理を行う。
  virtual bool finalize();

  /**
   * 入出力関数
   */
  /// SBXファイルの読み込みを行う。
  virtual sphData* readFile();
  /// SBXファイルに書込みを行う。
  virtual bool writeFile();
  /// SBXファイルのヘッダー情報のみの取得を行う。
  virtual bool getFileInfo(const char* filename, sphFileHeader* header);

  /**
   * データセット関数
   */
  virtual const sphFileHeader* getFileInfo() const;
  virtual sphFileHeader* getFileInfo();
  virtual bool setFileInfo(const sphFileHeader* header);

  /**
   * 入出力制御関数
   */
  /// 読み込みモードを設定する。
  virtual void setReadMode(int mode);
  /// 書き込みモードを設定する。
  virtual void setWriteMode(int mode);
  /// 圧縮書込みを行うかのフラグを設定する。(true:圧縮書込みを行う。, false:圧縮書込みを行わない。)
  virtual void setCompressionFlag(bool flag);
  /// 拡張レコードを書込みかのフラグを設定する。(true:拡張レコードを書き込む。, false:拡張レコードを書き込まない。)
  virtual void setExtRecodeFlag(bool flag);

  virtual sphData* getData();
  virtual void setData(sphData* data);
};

#endif /* _SPHSPXFILE_H_ */
