/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * @file sphMultiParaController.h
 * 部分領域管理パラレルコントローラーヘッダーファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPHMULTIPARACONTROLLER_H_
#define _SPHMULTIPARACONTROLLER_H_

#include "parallel/sphSvParaController.h"


/**
 * MPI_Allreduceをローカルで行う通信バッファ
 */
typedef struct {
  void* sendbuf;      /// 送信バッファ（ローカル演算結果）
  void* recvbuf;      /// 送信バッファ（プロセス間演算結果）
  unsigned int c;     /// データ数
  SPL_Datatype datatype;    /// データ型
  SPL_Op operation;       /// 演算の種類
} MPI_Allreduce_Buffer;


/**
 * 部分領域管理パラレルコントローラ
 */
class sphMultiParaController : public sphSvParaController
{
private:

  /**
   * MPI_Allreduceをローカルで行う通信バッファテーブル
   * std::map<MPI識別文字列, MPI_Allreduce_Buffer>
   */
  std::map<std::string, MPI_Allreduce_Buffer*> m_mapAllReduceBuffer;


  /**
   * サブブロックパラレルコントローラテーブル
   *
   * サブブロックローカルID（0〜)とサブブロックのパラレルコントローラクラスのテーブル
   */
  std::map<int, sphSvParaController*> m_mapChildPara;

  /**
   * グループID(=プロセス番号)テーブル
   *
   * プロセス番号とサブブロックIDのテーブル
   * プロセス番号 = m_processGroupId[サブブロックID]
   */
  int *m_processGroupId;
  int m_nodeCount;      // サブブロック総数

  bool initializeLocalBlock(
                unsigned int node_count,
                const size_t psize[3],
                const size_t pstart[3],
                const size_t (*node_size)[3],
                const size_t (*node_startidx)[3],
                const unsigned int *groupid_table);

  bool addChildParaController(
                          int localid,
                          sphSvParaController* subctrl);

protected:
  virtual bool initializeVoxel(
                        const size_t wsize[3],
                        const size_t wsta[3],
                        unsigned int div[3],
                        unsigned int localblock_count = 0);
  virtual bool initializeVoxel(
                        const size_t wsize[3],
                        const size_t wsta[3],
                        unsigned int proc_num = 0,
                        unsigned int localblock_count = 0);
  virtual bool initializeVoxel(
                        unsigned int node_count,
                        const size_t wsize[3],
                        const size_t wstart[3],
                        const size_t (*node_size)[3],
                        const size_t (*node_sta)[3],
                        const unsigned int *groupid_table = NULL);

public:
  sphMultiParaController(unsigned int paraKey = 0);
  virtual ~sphMultiParaController();

  const sphSvParaController* getChildParaController(int localid) const;
  const sphSvParaController* getChildParaControllerByNodeId(int nodeid) const;
  virtual int getNumberOfChildCtrl() const;
  bool isContainNodeId(int nodeid) const;
  int getProcessId(int nodeid) const;
  virtual int getNumberOfSubBlock() const;
  virtual bool isMultiParaController() const;
  virtual bool isParallel() const;

#if 0     // not use delete at 2011/01/09 by @hira
  // for MPI Method
  virtual bool allreduce( void* sendbuf, void* recvbuf,
                          unsigned int c, SPL_Datatype datatype,
                          SPL_Op operation,
                          const char*mpi_label = NULL);
  bool execAllreduce(const char*mpi_label);
  bool clearAllreduceBuffer(const char*mpi_label);
  MPI_Allreduce_Buffer* getAllreduceBuffer(const char*mpi_label);
#endif

  // for debug
  virtual std::string toString() const;

  /**
   * 最大値を求める
   * @param sendbuf     元データ
   * @param recvbuf     更新先データ
   * @param len         データ数
   */
  template<class T> void max( T* sendbuf,  T* recvbuf, int len) {
    register int i;
    for (i=0; i<len; i++) {
      if (recvbuf[i] < sendbuf[i] ) {
        recvbuf[i] = sendbuf[i];
      }
    }
  }

  /**
   * 最小値を求める
   * @param sendbuf     元データ
   * @param recvbuf     更新先データ
   * @param len         データ数
   */
  template<class T> void min( T* sendbuf,  T* recvbuf, int len) {
    register int i;
    for (i=0; i<len; i++) {
      if (recvbuf[i] > sendbuf[i] ) {
        recvbuf[i] = sendbuf[i];
      }
    }
  }

  /**
   * 合計値を求める
   * @param sendbuf     元データ
   * @param recvbuf     更新先データ
   * @param len         データ数
   */
  template<class T> void sum( T* sendbuf,  T* recvbuf, int len) {
    register int i;
    for (i=0; i<len; i++) {
      recvbuf[i] += sendbuf[i];
    }
  }

  /**
   * 積算を行う。
   * @param sendbuf     元データ
   * @param recvbuf     更新先データ
   * @param len         データ数
   */
  template<class T> void prod( T* sendbuf,  T* recvbuf, int len) {
    register int i;
    for (i=0; i<len; i++) {
      recvbuf[i] *= sendbuf[i];
    }
  }

  /**
   * 論理積を行う。
   * @param sendbuf     元データ
   * @param recvbuf     更新先データ
   * @param len         データ数
   */
  template<class T> void land( T* sendbuf,  T* recvbuf, int len) {
    register int i;
    for (i=0; i<len; i++) {
      recvbuf[i] = sendbuf[i] && recvbuf[i];
    }
  }


  /**
   * ビット演算の積を行う。
   * @param sendbuf     元データ
   * @param recvbuf     更新先データ
   * @param len         データ数
   */
  template<class T> void band( T* sendbuf,  T* recvbuf, int len) {
    register int i;
    for (i=0; i<len; i++) {
      recvbuf[i] = sendbuf[i] & recvbuf[i];
    }
  }

  /**
   * 論理和を行う。
   * @param sendbuf     元データ
   * @param recvbuf     更新先データ
   * @param len         データ数
   */
  template<class T> void lor( T* sendbuf,  T* recvbuf, int len) {
    register int i;
    for (i=0; i<len; i++) {
      recvbuf[i] = sendbuf[i] || recvbuf[i];
    }
  }

  /**
   * ビット演算の和を行う。
   * @param sendbuf     元データ
   * @param recvbuf     更新先データ
   * @param len         データ数
   */
  template<class T> void bor( T* sendbuf,  T* recvbuf, int len) {
    register int i;
    for (i=0; i<len; i++) {
      recvbuf[i] = sendbuf[i] | recvbuf[i];
    }
  }

  /**
   * 排他的論理和を行う。
   * @param sendbuf     元データ
   * @param recvbuf     更新先データ
   * @param len         データ数
   */
  template<class T> void lxor( T* sendbuf,  T* recvbuf, int len) {
    register int i;
    for (i=0; i<len; i++) {
      recvbuf[i] = sendbuf[i] xor recvbuf[i];
    }
  }

  /**
   * ビット演算の排他的論理和を行う。
   * @param sendbuf     元データ
   * @param recvbuf     更新先データ
   * @param len         データ数
   */
  template<class T> void bxor( T* sendbuf,  T* recvbuf, int len) {
    register int i;
    for (i=0; i<len; i++) {
      recvbuf[i] = sendbuf[i] ^ recvbuf[i];
    }
  }

};

#endif /* _SPHMULTIPARACONTROLLER_H_ */
