/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphParaController.h
 * Parallel Controller Class Header File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphParaController.h
 * パラレルコントローラークラスヘッダーファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPHPARACONTROLLER_H_
#define _SPHPARACONTROLLER_H_

#include "base/sphBase.h"
#include "spl/splmpi.h"
#include "spl/spl.h"


/**
 * \~english
 * MPI Function Return Information
 *
 * \~japanese
 * MPI実行結果構造体
 *
 * MPI関数実行におけるエラーコード、ステータスを設定、参照する構造体である。
 */
typedef struct {
  /// MPI Routine status
  SPL_Status m_mpiStatus;
  /// MPI Routine return code
  int m_mpiErr;
} Spl_Mpi_ReturnInfo;


/**
 * オブジェクトキー管理クラス.
 * オブジェクトの登録、一意のキーの管理を行う。＜br/>
 * 非同期通信時のMPI_Requstの管理を行う。
 */
template <class T> class KeyMapping {
private:
  std::map<int, void*> m_objList;       ///< 登録オブジェクトリスト
  std::list<int> m_delList;             ///< 削除済みキーリスト
  int m_lastKey;                        ///< 最終登録キー

public:
  /** コンストラクタ */
  KeyMapping() { clear(); };

  /** デストラクタ  */
  virtual ~KeyMapping() {
    std::map<int, void*>::iterator itr = m_objList.begin();
    for( ; itr != m_objList.end(); itr++) if( itr->second ) delete (T*)(itr->second);
    m_objList.clear();
    m_delList.clear();
    m_lastKey = 1;
  };

  /**
   * オブジェクトを登録する。
   * @param      obj    登録オブジェクト
   * @return     登録キー( >0:キー番号, <=0:登録エラー)
   */
  int add(T* obj) {
    int key;
    if( m_delList.size() != 0 ){
      /// 削除済みキーリストが存在する場合は、削除済みキーリストからキー番号を取得して登録キーに再使用する。
      std::list<int>::iterator itr = m_delList.begin();
      key = *itr;
      if( !((m_objList.insert(std::make_pair(key, (void*)obj))).second) ) return -1;

      /// 削除済みキーリストから再使用キー番号を削除する。
      m_delList.erase(itr);
    }
    else {
      /// 削除済みキーリストが存在しない場合は、最終登録キーを登録キーとして使用する。
      if( !((m_objList.insert(std::make_pair(m_lastKey, (void*)obj))).second) ) return -1;
      key = m_lastKey;
      m_lastKey++;
    }
    return key;
  }

  /**
   * 登録キーのオブジェクトを削除する。
   * @param      key       キー
   * @return    success(=true)/fault(=false)
   */
  bool del(int key){
    /// 登録オブジェクトリストからキーのオブジェクトを削除する。
    std::map<int, void*>::iterator itr = m_objList.find(key);
    if( itr == m_objList.end() ) return false;
    if( itr->second ) delete (T*)(itr->second);
    m_objList.erase(key);
    if (m_objList.size() == 0) {
      clear();
      return true;
    }

    /// 削除済みキーリストに削除キーを追加する。
    m_delList.push_back(key);
    return true;
  }

  /**
   * オブジェクトを新規生成する。
   * @param      void
   * @return     オブジェクト
   */
  T* create() {
    T* obj = new T;
    return obj;
  }

  /**
   * 登録キーのオブジェクトを取得する。
   * @param      key       キー
   * @return     登録オブジェクト
   */
  T* get(int key) {
    std::map<int, void*>::iterator itr = m_objList.find(key);
    if( itr == m_objList.end() ) return NULL;
    return (T*)(itr->second);
  }

  /**
   * 登録オブジェクトを登録数を取得する。
   * @param      void
   * @return     登録数
   */
  int count() {
    return m_objList.size();
  }

  /**
   * 登録キーリストを取得する。
   * @param      keylist       キーリスト
   * @return     キーリスト数
   */
  int getKeyList(int* keylist) {
    int cnt = 0;
    std::map<int, void*>::iterator itr = m_objList.begin();
    for( ; itr != m_objList.end(); itr++) {
      keylist[cnt] = itr->first;
      cnt++;
    }
    return cnt;
  }

  /**
   * 登録オブジェクトをクリアする。
   * @param      void
   * @return    success(=true)/fault(=false)
   */
  bool clear() {
    m_objList.clear();
    m_delList.clear();
    m_lastKey = 1;
    return true;
  }


  /**
   * 登録オブジェクトを表示する。
   * @param      void
   * @return     void
   */
  void print() {
// ----- modify 20100414 @tamaki begin -----
//    m_objList::iterator itr = m_objList.begin();
// -----------------------------------------
    std::map<int, void*>::iterator itr = m_objList.begin();
// ----- modify 20100414 @tamaki end   -----
    for( ; itr != m_objList.end(); itr++){
      std::cout << "label:" << itr->first << " " << std::endl;
    }
    std::cout << std::endl;
  }
};


/**
 * \~english
 * Parallel Controller Base Class
 *
 * Control Vocel Information.
 * @author       hira
 * @version     1.0
 *
 * \~japanese
 * パラレルコントローラー基底クラス
 *
 * ノード間通信、ノードのボクセル情報の管理を行う。<br>
 * パラレルコントローラーは、固有の識別キー（m_paraKey）を持つ。
 * @author       hira
 * @version     1.0
 */
class sphParaController : public virtual sphBase
{
protected:
  //    const sphSolverConfig* m_solvCfg;
  /// Host Name Table
  std::vector<std::string> m_hostNameList;
  /// MPI Rank Table
  int* m_idTable;
  /// MPI communicator
  SPL_Comm m_splComm;
  /// MPI Group
  SPL_Group m_splGroup;
  /// 識別キー（SPL_COMM_WORLDの中で固有のキー＝sphParaManager::map m_mapParaCtrlのキーと同じ）
  unsigned int m_paraKey;
  /// サブ識別キー（同一のm_splComm中で固有のキー）
  unsigned int m_subKey;
  /// my rank id in process group
  int   m_myID;
  /// 非同期通信のリクエスト登録リスト
  KeyMapping<SPL_Request>* m_immediateRequest;

  bool createHostNameTable();
  bool setIDTable();
  bool createGrpIDTable();
  const int* getGrpIDTable() const;
  void setCommGroup(SPL_Comm comm, SPL_Group group = SPL_GROUP_NULL);

  // MPICH FUNCTIONS //
  bool initMPI(int* argc, char*** argv);
  void finalizeMPI();

public:
  sphParaController(unsigned int paraKey = 0);
  virtual ~sphParaController();

  bool initialize(SPL_Comm comm, SPL_Group group = SPL_GROUP_NULL);
  void finalize();
  const char* getHostName(int id);
  SPL_Comm getSplComm() const;
  void setParaKey(unsigned int paraKey);
  unsigned int getParaKey() const;
  void setSubKey(unsigned int subKey);
  unsigned int getSubKey() const;
  int getMyID() const;
  void setMyID(int myid);
  SPL_Comm createProcessGroup(int idNum, int* idList, SPL_Comm world_comm);
  virtual bool isEv() const = 0;
  virtual bool isMb() const = 0;
  virtual bool isUnst() const = 0;

  // get my Voxel size index
  virtual const size_t* getWholeVoxelSize() const;
  virtual size_t getVoxelHeadIndex(int nodeId, unsigned int dirId) const;
  virtual size_t getVoxelTailIndex(int nodeId, unsigned int dirId) const;
  virtual const size_t* getVoxelSize(int nodeId = -1) const;
  virtual void getVoxelStartIndex(size_t sta_idx[3]) const;
  virtual bool getVoxelOrigin(double org[3]) const = 0;
  virtual bool getVoxelPitch(double pitch[3]) const;
  virtual bool getNodeOrigin(double org[3]) const = 0;
  virtual int getNumberOfNode() const;

  // MPICH FUNCTIONS //
//  SPL_Status getMPIStatus() const;
//  int getMPIError() const;
  unsigned int getNumberOfGroup() const;
  unsigned int getNumberOfWorld() const;
  int getMyWorldRank() const;
  int getMyRank() const;
  virtual bool allreduce( void* sendbuf, void* recvbuf,
                          unsigned int c, SPL_Datatype datatype,
                          SPL_Op operation,
                          const char*mpi_label = NULL) const;
  bool broadcast(void* data, int c, SPL_Datatype datatype, int id) const;
  bool barrier() const;
  bool send(void* data, int c, SPL_Datatype datatype, int dst) const;
  bool recv(void* data, int c, SPL_Datatype datatype, int src) const;
  int iSend(void* data, int c, SPL_Datatype datatype, int dst);
  int iRecv(void* data, int c, SPL_Datatype datatype, int src);
  bool wait(int key);
  bool waitAll(unsigned key_num, int* key_list);
  bool waitRequestAll();
  void abort(int errcode) const;
  bool gather(void* send_data, int send_cnt, SPL_Datatype send_datatype,
      void* recv_data, int recv_cnt, SPL_Datatype recv_datatype,
      int root) const;
  bool allgather(void* send_data, int send_cnt, SPL_Datatype send_datatype,
      void* recv_data, int recv_cnt, SPL_Datatype recv_datatype) const;
  bool createCommGroupByWorld(int idNum, int* idList, SPL_Comm *newComm, SPL_Group *newGrp ) const;
  bool createCommGroupByComm(int idNum, int* idList, SPL_Comm oldComm, SPL_Comm *newComm, SPL_Group *newGrp ) const;
  int getMPIDatatypeSize(SPL_Datatype datatype) const;
  virtual bool isParallel() const = 0;

  // for sphMultiParaController
  virtual int getNumberOfChildCtrl() const;
  virtual bool isMultiParaController() const;
  virtual int getNumberOfSubBlock() const;

  // for debug
  virtual std::string toString() const;

};



#endif /* _SPHPARACONTROLLER_H_ */

