/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSvNodeInfo.h
 * Node Information Classes Header File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphSvNodeInfo.h
 * ノード情報クラス郡ヘッダーファイル
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPHSVNODEINFO_H_
#define _SPHSVNODEINFO_H_


#include <iostream>
#include <vector>
#include <map>
#include <string>

#include "base/sphBase.h"
#include "spl/splmpi.h"


/**
 * \~english
 * Node Offset Information Class
 *
 * \~japanese
 * ノード接合面情報クラス
 */
class sphOffsetDiv {
public:

private:
  int m_rankid;            ///< Contacted node id
  size_t m_offsetMin[3];    //< Contacted min index
  size_t m_offsetMax[3];    //< Contacted max index
  size_t m_offsetSize[3];    //< Contacted size
  FaceType m_face;        //< Contacted direction

public:

  /**
  * \~english
  * Constructor
  *
  * \~japanese
  * コンストラクタ
  */
  sphOffsetDiv() {
    m_rankid = SPL_PROC_NULL;
    m_offsetMin[0] = m_offsetMin[1] = m_offsetMin[2] = 0;
    m_offsetMax[0] = m_offsetMax[1] = m_offsetMax[2] = 0;
    m_offsetSize[0] = m_offsetSize[1] = m_offsetSize[2] = 0;
  };

  /**
  * \~english
  * Set Node Id
  * @param        rankId      Contacted Node ID
  * \~japanese
  * 接合面の相手ノード番号を設定する。
  * @param        rankId      相手ノード番号
  */
  void setRankId(int rankId)
  {
    m_rankid = rankId;
  }

  /**
  * \~english
  * Set Offset Face Type
  * @param        face      Face Tyep
  *
  * \~japanese
  * 接合面タイプを取得する。
  * @param        face      接合面タイプ
  */
  void setFaceType(FaceType face)
  {
    m_face = face;
  }

  /**
  * \~english
  * Set Offset Size
  * @param      min          Contacted start index[3]
  * @param      max          Contacted end index[3]
  *
  * \~japanese
  * 接合面サイズを設定する。
  * @param      min          接合面始点インデックス[3]
  * @param      max          接合面終点インデックス[3]
  */
  void setOffsetSize(size_t* min, size_t* max)
  {
    for (register int i=0; i<3; i++) m_offsetMin[i] = min[i];
    for (register int i=0; i<3; i++) m_offsetMax[i] = max[i];
    for (register int i=0; i<3; i++) m_offsetSize[i] = max[i]-min[i]+1;
  }

  /**
  * 接合面の相手ノード番号を取得する。
  */
  int getRankId() const
  {
    return m_rankid;
  }

  /**
  * \~english
  * Get Offset Face Type
  * @return        FaceType      Face Tyep
  *
  * \~japanese
  * 接合面タイプを設定する。
  * @return        FaceType      接合面タイプ
  */
  FaceType getFaceType() const
  {
    return m_face;
  }


  /**
  * \~english
  * Get Offset start index
  * @return        const size_t*      Contacted start index[3]
  *
  * \~japanese
  * 接合始点インデックスを取得する。
  * @return        const size_t*      接合面始点インデックス[3]
  */
  const size_t* getOffsetMin() const
  {
    return m_offsetMin;
  }

  /**
  * \~english
  * Get Offset end index
  * @return        const size_t*      Contacted end index[3]
  *
  * \~japanese
  * 接合終点インデックスを取得する。
  * @return        const size_t*      接合面終点インデックス[3]
  */
  const size_t* getOffsetMax() const
  {
    return m_offsetMax;
  }


  /**
  * \~english
  * Get Offset size index
  * @return        const size_t*      Contacted size[3]
  *
  * \~japanese
  * 接合サイズを取得する。
  * @return        const size_t*      接合面サイズ[3]
  */
  const size_t* getOffsetSize() const
  {
    return m_offsetSize;
  }

};


/**
 * \~english
 * Node Offset Information List
 *
 * \~japanese
 * ノード接合面リスト
 */
typedef std::vector <sphOffsetDiv*> OffsetDiv;


/**
 * \~english
 * Node Information Class
 *
 * \~japanese
 * ノード情報クラス
 */
class sphSvNodeInfo : public virtual sphBase {
friend class sphSvParaController;

protected:
  int m_nodeID;    ///< ノードID
  int m_groupID;  ///< グループID(=ランクID) for 部分領域データ管理機能
  size_t m_voxSize[3];  ///< voxel size
  size_t m_voxMin[3];  ///< voxel min index (in multibox.xml of Param) I, J, K
  size_t m_voxMax[3];  ///< voxel max index (in multibox.xml of Param) I, J, K

  std::map<std::string, std::string> m_sbxNames;    ///< sbx file names
  OffsetDiv  m_offsetInfo;  ///< offset info vector<SklOffsetDivMb>

  /**
   * \~english
   * Contacted node count List
   *
   * \~japanese
   * 接合ノード数リスト
   *
   * +I,-I,+J,-J,+K,-K方向に接しているノード数
   */
  int m_idMtrxNum[3][3][3];

  /**
   * \~english
   * Contacted node ID List
   *
   * \~japanese
   * 接合ノード番号リスト
   *
   * +I,-I,+J,-J,+K,-K方向に接しているノード番号のリスト
   */
  int* m_idMtrx[3][3][3];

public:
  sphSvNodeInfo();
  virtual ~sphSvNodeInfo();
  int getNodeID() const;
  const size_t* getVoxelSize() const;
  const size_t* getVoxelMin() const;
  const size_t* getVoxelMax() const;
  int  getOffsetDivNum() const;      // get offset node count (=m_offsetInfo.size())
  int  getOffsetDivIds(int* nodeids) const;      // get offset node array  (= m_offsetInfo->m_nodeID array)
  const sphOffsetDiv* getOffsetDivInfo(const int nodeid) const;
  bool setNodeInfo(int nodeid, size_t min_i, size_t min_j, size_t min_k, size_t max_i, size_t max_j, size_t max_k);
  int addOffsetDiv(sphSvNodeInfo* node);
  const char* getSbxFilename(const char* sbx_attr = NULL) const;        // get sbx file name
  bool setSbxFilename(const char* sbx_attr, const char* sbx_file);      // set sbx file name
  int  getCommIDNum (int i, int j, int k) const;
  const int* getCommIDs (int i, int j, int k) const;
  bool createIdMatrix();
  int getGroupID() const;
  void setGroupID(int groupid);

  // for debug
  virtual std::string toString() const;

private:
  bool getNodeFace(int idir, int jdir, int kdir, size_t* face);
  bool getOffsetFace(const size_t* face_src, const size_t* face_dest, size_t* face_offset);
  bool isInsideCoord(const size_t* box, const size_t* coord);
  int getCrossCoords(size_t (*cross_coords)[3],
                      size_t* coord_src_min, size_t* coord_src_max,
                      size_t* coord_dest_min, size_t* coord_dest_max ) const;
  int getContactCoords(size_t (*cross_coords)[3],
                        size_t* coord_src_min, size_t* coord_src_max,
                        size_t* coord_dest_min, size_t* coord_dest_max) const;
};

#endif /* _SPHSVNODEINFO_H_ */
