/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSvParaController.h
 * Simple Voxel Parallel Controller Header File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphSvParaController.h
 * 構造格子パラレルコントローラーヘッダーファイル
 * @author     hira
 * @date       2009/02/20
 */


#ifndef _SPHSVPARACONTROLLER_H_
#define _SPHSVPARACONTROLLER_H_


#include "parallel/sphParaController.h"
#include "parallel/sphSvNodeInfo.h"

// using namespace std;

/**
 * \~english
 * Node Information Map Type
 *
 * \~japanese
 * ノード情報マップ
 *
 * map<キー：ノードID, sphSvNodeInfo*>
 */
typedef std::map<int, sphSvNodeInfo*> MapNodeInfo;



/**
 * \~english
 * Simple Voxel Type Parallel Controller
 *
 * \~japanese
 * 構造格子パラレルコントローラー
 */
class sphSvParaController : public sphParaController
{
friend class sphParaManager;
friend class sphMultiParaController;

protected:
  // voxel coordinates
  double m_voxelOrigin[3];      ///< 原点座標
  double m_voxelPitch[3];       ///< ピッチ

  // node info
  size_t   m_voxSize[3];  ///< 計算領域全体ボクセルサイズ（ガイドセルを含まない）
  MapNodeInfo  *m_nodeInfo;    ///< ノード情報マップ <rankId, sphSvNodeInfo*>
  size_t m_voxMinIdx[3];  ///< 始点インデックス
  unsigned int m_divInfo[3];             ///< 均等分割情報
  unsigned int m_divLocalInfo[3];             ///< プロセス内(サブブロック)の均等分割情報（部分領域データ管理用)
  const sphParaController* m_parentCtrl;    ///< 親パラレルコントローラ(=sphMultiParaController)
  int ***m_nodeEvMap;             ///< ノード番号テーブル（３次元テーブル=[0:I方向分割数-1][0:J方向分割数-1][0:K方向分割数-1]）

  // node info method
  bool addNodeInfo(int nodeid,
              size_t min_i, size_t min_j, size_t min_k,
              size_t max_i, size_t max_j, size_t max_k,
              unsigned int groupid);
  bool addNodeInfo(int nodeid, sphSvNodeInfo* nodeinfo);
  bool makeNodeOffset();
  bool decideDivPattern(
                unsigned int divNum,
                const size_t voxSize[3],
                unsigned int divPttn[3]) const;
  long long calcCommSize(const size_t div[3],
                const size_t voxSize[3]) const;
  bool createVoxelDistribution(
                const size_t voxSize[3],
                const unsigned int divNode[3],
                size_t (*distSize)[3],
                size_t (*distStart)[3] ) const;
  void setVoxelDivInfo(unsigned int div[3]);
  void setVoxelDivLocalInfo(unsigned int div[3]);
  MapNodeInfo* getMapNodeInfo();
  void setMapNodeInfo(MapNodeInfo* info);

  virtual bool initializeVoxel(
                        const size_t wsize[3],
                        const size_t wsta[3],
                        unsigned int div[3],
                        unsigned int localblock_count = 0);

  virtual bool initializeVoxel(
                        const size_t wsize[3],
                        const size_t wsta[3],
                        unsigned int proc_num = 0,
                        unsigned int localblock_count = 0);

  bool initializeVoxel(
                        const size_t wsize[3],
                        const size_t wstart[3],
                        const size_t mysize[3],
                        const size_t mysta[3]);

  virtual bool initializeVoxel(
                        unsigned int nodeCnt,
                        const size_t wsize[3],
                        const size_t wstart[3],
                        const size_t (*node_size)[3],
                        const size_t (*node_sta)[3],
                        const unsigned int *groupid_table = NULL);
  bool createNodeMap();

public:
  sphSvParaController(unsigned int paraKey = 0);
  sphSvParaController(const sphParaController* parent, int myId);
  virtual ~sphSvParaController();

  // VoxelOrigin, VoxelPitch Method
  bool getVoxelOrigin(double org[3]) const;
  bool getVoxelPitch(double pitch[3]) const;
  bool getNodeOrigin(double org[3]) const;
  void setVoxelOrigin(const double org[3]);
  void setVoxelPitch(const double pitch[3]);

  // node info method
  virtual const size_t* getWholeVoxelSize() const;
  virtual const size_t* getVoxelSize(int nodeId = -1) const;
  virtual void getVoxelStartIndex(size_t sta_idx[3]) const;
  const unsigned int* getVoxelDivInfo() const;
  virtual bool isEv() const;
  virtual bool isMb() const;
  virtual bool isUnst() const;
  const sphSvNodeInfo* getNodeInfo(int nodeId = -1) const;  // get node info
  int  getOffsetNodeNum();                  // get offset node count
  int getNumberOfNode() const;
  size_t getVoxelHeadIndex(int nodeId, unsigned int dirId) const;
  size_t getVoxelTailIndex(int nodeId, unsigned int dirId) const;
  int getCommIDNum(int i, int j, int k) const;
  const int* getCommIDs (int i, int j, int k) const;
  int  isCommID(int i, int j, int k) const;
  const size_t* getReferIndex() const;
  int getCommIDEv (int i, int j, int k) const;
  int getCommIDPeriodic (int face, int dir) const;
  void setWholeVoxelSize(const size_t size[3]);
  void setReferIndex(const size_t sidx[3]);
  const sphParaController* getParentController() const;
  bool getCommIDTable(int nIDTable[6]) const;
  virtual bool isParallel() const;
  const int*** getNodeMap(unsigned int& i, unsigned int& j, unsigned int& k) const;
  bool getNodeDivPosition(unsigned int& i_idx, unsigned int& j_idx, unsigned int& k_idx) const;

  // for debug
  virtual std::string toString() const;

};
#endif /* _SPHSVPARACONTROLLER_H_ */
