/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */
/**
 * @file SklCfgUsrDefNode.h
 * コンフィグレーション要素クラスヘッダーファイル
 * @author     riken
 * @date       2010/10/01
 */

#ifndef _SKL_CFG_USER_DEFINE_NODE_
#define _SKL_CFG_USER_DEFINE_NODE_

#include <iostream>
#include <list>

#include "SklCfgDatas.h"
#include "SklCfgBase.h"

using namespace std;

namespace SklCfg {

/**
 * 要素識別子クラス
 */
class CfgIdt : public SklCfgBase {
protected:
  std::string m_idtStr;     ///< 要素識別子文字列
  char* m_name;     ///< name属性
  bool m_idFlg;     ///< id属性設定フラグ
  int m_id;         ///< id属性
  char* m_comment;    ///< comment属性

  void Init();
  bool UpdateIdtStr();
  bool SetName(const char* name);
  bool SetID(int id);
  bool UnsetID();
  bool SetComment(const char* comment);
  virtual bool Parse(xmlNodePtr node);

public:
  CfgIdt();
  CfgIdt(const char* name);
  CfgIdt(const char* name, int id);
  CfgIdt(const char* name,         const char* comment);
  CfgIdt(const char* name, int id, const char* comment);
  ~CfgIdt();
  const char* GetName() const { return m_name; };
  bool IsSetID() const { return m_idFlg; };
  int GetID() const { return m_id; };
  const char* GetComment() const { return m_comment; };
  bool SetIdent(const char* name);
  bool SetIdent(const char* name, int id);
  bool SetIdent(const char* name,         const char* comment);
  bool SetIdent(const char* name, int id, const char* comment);
  bool SetIdtFromStr(const char* idt_str);
  char* GetIdtStr(char* str, size_t slen) const;
  const char* GetIdtStr() const;
  CfgIdt& operator=(const CfgIdt& src);
  static char* genIdt(char* str, size_t slen, const char* name);
  static char* genIdt(char* str, size_t slen, const char* name, int id);
  static char* genIdt(char* str, size_t slen, const char* name,         const char* comment);
  static char* genIdt(char* str, size_t slen, const char* name, int id, const char* comment);
};

/**
 * コンフィグレーション要素基底クラス
 */
class CfgBase : public SklCfgBase {
protected:
  char* m_name;    ///< name属性
  int m_id;        ///< id属性
  bool m_idFlg;     ///< id属性設定フラグ
  char* m_comment; ///< comment属性
  virtual bool Parse(xmlNodePtr node) = 0;
  bool SetName(const char* name);
  bool SetID(const int id);
  bool SetComment(const char* cmt);

public:
  CfgBase();
  virtual ~CfgBase();
  virtual bool SetAttr(xmlNodePtr node) = 0;

  // Getting Methods
  const char* GetName() const;
  bool isSetID() const;
  int GetID() const;
  const char* GetComment() const;

  virtual void Print() const; // for debug
};

/**
 * コンフィグレーションParam要素クラス
 */
class CfgParam : public CfgBase {
protected:

  bool m_valFlag;     ///< value属性設定フラグ
  /// dtype属性
  DataType m_dType; // data type
  /// value属性
  PrmVal m_value;   // value
  bool ClearStr();
  virtual bool Parse(xmlNodePtr node);
  bool SetData(double val);
  bool SetData(int val);
  bool SetData(const char* val);
  bool SetData(DataType type, char* val);

public:
  CfgParam();
  virtual ~CfgParam();
  virtual bool SetAttr(xmlNodePtr node);

  // Getting Methods
  DataType GetDataType() const;
  DataType GetData(PrmVal& val) const;
  bool GetData(int* val) const;
  bool GetData(float* val) const;
  bool GetData(double* val) const;
  bool GetData(const char** val) const;
  bool IsSetValue() const;

  virtual void Print() const; // for debug
};

/**
 * コンフィグレーションElem要素クラス
 */
class CfgElem : public CfgBase {
protected:
  /// 子Param要素リスト
  list<CfgParam*> m_prmList; // CfgParam list
  /// 子Elem要素リスト
  list<CfgElem*>  m_elmList; // CfgElem list
  virtual bool Parse(xmlNodePtr node);

public:
  CfgElem();
  virtual ~CfgElem();
  bool ParseNode(xmlNodePtr node);
  bool AddParam(CfgParam* param);
  bool AddElem(CfgElem* elem);
  virtual bool SetAttr(xmlNodePtr node);
  void ClearList();

  // Getting Methods
  unsigned int GetSize() const;
  unsigned int GetElemSize() const;
  unsigned int GetParamSize() const;
  const CfgElem* GetElemFirst(const char* name = NULL) const;
  const CfgElem* GetElemFirst(const char* name, int id, const char* cmt = NULL) const;
  const CfgElem* GetElemFirst(const char* name,         const char* cmt) const;
  const CfgElem* GetElemNext(const CfgElem* elem,
                             const char* name = NULL) const;
  const CfgElem* GetElemNext(const CfgElem* elem,
                             const char* name, int id, const char* cmt = NULL) const;
  const CfgElem* GetElemNext(const CfgElem* elem,
                             const char* name,         const char* cmt) const;
  const CfgParam* GetParamFirst(const char* name = NULL) const;
  const CfgParam* GetParamFirst(const char* name, int id, const char* cmt = NULL) const;
  const CfgParam* GetParamFirst(const char* name,         const char* cmt) const;
  const CfgParam* GetParamNext(const CfgParam* param,
                               const char* name = NULL) const;
  const CfgParam* GetParamNext(const CfgParam* param,
                               const char* name, int id, const char* cmt = NULL) const;
  const CfgParam* GetParamNext(const CfgParam* param,
                               const char* name,         const char* cmt) const;

/* Add XML Parser Interface 2007.09 begin */
  const CfgElem* GetElemFirst(CfgIdt& idt) const;
  const CfgElem* GetElemNext(const CfgElem* elem, CfgIdt& idt) const;
  const CfgParam* GetParamFirst(CfgIdt& idt) const;
  const CfgParam* GetParamNext(const CfgParam* param, CfgIdt& idt) const;


  bool GetValues(const char* format, ...) const;

  bool GetIntValues   (unsigned int num, ...) const;
  bool GetFloatValues (unsigned int num, ...) const;
  bool GetDoubleValues(unsigned int num, ...) const;
  bool GetStrValues   (unsigned int num, ...) const;
  bool GetElemValues  (unsigned int num, ...) const;

  bool GetValue(CfgIdt ident, int* value) const;
  bool GetValue(CfgIdt ident, float* value) const;
  bool GetValue(CfgIdt ident, double* value) const;
  bool GetValue(CfgIdt ident, const char** value) const;
  bool GetValue(CfgIdt ident, const CfgElem** value) const;

  bool GetValue(const char* ident_str, int* value) const;
  bool GetValue(const char* ident_str, float* value) const;
  bool GetValue(const char* ident_str, double* value) const;
  bool GetValue(const char* ident_str, const char** value) const;
  bool GetValue(const char* ident_str, const CfgElem** value) const;

  int             GetIntValue   (CfgIdt ident) const;
  float           GetFloatValue (CfgIdt ident) const;
  double          GetDoubleValue(CfgIdt ident) const;
  const char*     GetStrValue   (CfgIdt ident) const;
  const CfgElem*  GetElemValue  (CfgIdt ident) const;

  int             GetIntValue   (const char* ident_str) const;
  float           GetFloatValue (const char* ident_str) const;
  double          GetDoubleValue(const char* ident_str) const;
  const char*     GetStrValue   (const char* ident_str) const;
  const CfgElem*  GetElemValue  (const char* ident_str) const;

  bool GetVctValue(CfgIdt u, CfgIdt v, CfgIdt w,
                   int* uv, int* vv, int* wv) const;
  bool GetVctValue(CfgIdt u, CfgIdt v, CfgIdt w,
                   float* uv, float* vv, float* wv) const;
  bool GetVctValue(CfgIdt u, CfgIdt v, CfgIdt w,
                   double* uv, double* vv, double* wv) const;
  bool GetVctValue(CfgIdt u, CfgIdt v, CfgIdt w,
                   const char** uv, const char** vv, const char** wv) const;
  bool GetVctValue(CfgIdt u, CfgIdt v, CfgIdt w,
                   const CfgElem** uv, const CfgElem** vv, const CfgElem** wv) const;

  bool GetVctValue(const char* u, const char* v, const char* w,
                   int* uv, int* vv, int* wv) const;
  bool GetVctValue(const char* u, const char* v, const char* w,
                   float* uv, float* vv, float* wv) const;
  bool GetVctValue(const char* u, const char* v, const char* w,
                   double* uv, double* vv, double* wv) const;
  bool GetVctValue(const char* u, const char* v, const char* w,
                   const char** uv, const char** vv, const char** wv) const;
  bool GetVctValue(const char* u, const char* v, const char* w,
                   const CfgElem** uv, const CfgElem** vv, const CfgElem** wv) const;

  bool GetValue(int* val,            unsigned int num, ...) const;
  bool GetValue(float* val,          unsigned int num, ...) const;
  bool GetValue(double* val,         unsigned int num, ...) const;
  bool GetValue(const char** val,    unsigned int num, ...) const;
  bool GetValue(const CfgElem** val, unsigned int num, ...) const;

/* Add XML Parser Interface 2007.09 end */

  virtual void Print() const; // for debug

};

} // end of namespace SklCfg

#endif // _SKL_CFG_USER_DEFINE_NODE_

