/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */

//
// vfvPathUtil
//
#ifndef _VFV_PATH_UTIL_H_
#define _VFV_PATH_UTIL_H_
namespace CES {

  inline bool SklGetFullPathName(const char* path,
                                 char* resolved_path,
                                 size_t str_len) {
    if( !path || !resolved_path || (str_len <= 0) ) return false;

#if defined(_WIN32)
    if( !_fullpath(resolved_path, path, str_len) ) return false;
#else
    char rpath[MAXPATHLEN];
    if( !realpath(path, rpath) ) return false;
    if( strlen(rpath) > str_len ) return false;
    strncpy(resolved_path, rpath, str_len);
#endif

    return true;
  }

  inline std::string DirName(const std::string& path,
                             const char dc = '/') {
    char* name = strdup( path.c_str() );
    char* p = name;

    for ( ; ; ++p ) {
      if ( ! *p ) {
        if ( p > name ) {
          char rs[2] = {dc, '\0'};
          return rs;
        } else
          return(".");
      }
      if ( *p != dc ) break;
    }

    for ( ; *p; ++p );
    while ( *--p == dc ) continue;
    *++p = '\0';

    while ( --p >= name )
      if ( *p == dc ) break;
    ++p;
    if ( p == name ) return(".");

    while ( --p >= name )
      if ( *p != dc ) break;
    ++p;

    *p = '\0';
    if( p == name ) {
      char rs[2] = {dc, '\0'};
      return rs;
    } else {
      std::string s( name );
      free( name );
      return s;
    }
  }

  inline std::string BaseName(const std::string& path,
                              const std::string& suffix = std::string(""),
                              const char dc = '/') {
    char* name = strdup( path.c_str() );
    char* p = name;

    for ( ; ; ++p ) {
      if ( ! *p ) {
        if ( p > name ) {
          char rs[2] = {dc, '\0'};
          return rs;
        } else
          return "";
      }
      if ( *p != dc ) break;
    }

    for ( ; *p; ++p ) continue;
    while ( *--p == dc ) continue;
    *++p = '\0';

    while ( --p >= name )
      if ( *p == dc ) break;
    ++p;

    if ( suffix.length() > 0 ) {
      const int suffixlen = suffix.length();
      const int stringlen = strlen( p );
      if ( suffixlen < stringlen ) {
        const int off = stringlen - suffixlen;
        if ( !strcmp( p + off, suffix.c_str() ) )
          p[off] = '\0';
      }
    }

    std::string s( p );
    free( name );
    return s;
  }

  inline std::string OmitDots(const std::string& path,
                              const char dc = '/') {
    using namespace std;
    if ( path.empty() ) return path;

    deque<string> elemLst;
    istringstream ss(path);
    std::string selStr; char c;

    // decomposition
//    while ( 1 ) {
    unsigned int FLAG = 1;
    while ( FLAG ) {
      if ( ! ss.get(c) ) {
        elemLst.push_back(selStr);
        break;
      }
      if ( c == dc ) {
        elemLst.push_back(selStr);
        selStr = "";
      } else {
        selStr += c;
      }
    } // end of while(1)

    // remove null or '.' elem
    deque<string>::iterator it;
    for ( it = elemLst.begin(); it != elemLst.end(); ) {
      if ( it->empty() || (*it) == "." )
        it = elemLst.erase(it);
      else
        it ++;
    } // end of for(it)

    // remove '..' elem
    if ( ! elemLst.empty() )
      for ( it = elemLst.begin() + 1; it != elemLst.end(); ) {
        if ( (*it) == ".." ) {
          it = elemLst.erase(--it);
          it = elemLst.erase(it);
        } else
          it ++;
      } // end of for(it)

    // result
    string retPath;
    if ( path[0] == dc ) retPath = dc;
    for ( it = elemLst.begin(); it != elemLst.end(); it++ ) {
      retPath += (*it);
      if ( it != elemLst.end() -1 )
        retPath += dc;
    } // end of for(it)
    return retPath;
  }

} // end of namespace CES

namespace VFV {
  inline char vfvPath_getDelimChar() {
#ifdef WIN32
    return '\\';
#else
    return '/';
#endif
  }

  inline void vfvPath_adjustDelim(std::string& path) {
#ifdef WIN32
    const char newDelim = '\\';
    const char oldDelim = '/';
#else
    const char oldDelim = '\\';
    const char newDelim = '/';
#endif
    size_t pathLen = path.size();
    for ( register size_t i = 0; i < pathLen; i++ ) {
      if ( path[i] == oldDelim )
	path[i] = newDelim;
    } // end of for(i)
  }

  inline bool vfvPath_hasDrive(const std::string& path) {
    if ( path.size() < 2 ) return false;
    char x = path[0];
    if ( ((x >= 'A' && x <= 'Z' ) || (x >= 'a' && x <= 'z')) &&
	 path[1] == ':' )
      return true;
    return false;
  }

  inline std::string vfvPath_emitDrive(std::string& path) {
    // returns drive (ex. 'C:')
    if ( ! vfvPath_hasDrive(path) ) return std::string();
    std::string driveStr = path.substr(0, 2);
    path = path.substr(2);
    return driveStr;
  }

  inline bool vfvPath_isAbsolute(const std::string& path) {
    std::string xpath(path);
    vfvPath_emitDrive(xpath);
    char c1, c2;
    c1 = xpath[0];
    c2 = vfvPath_getDelimChar();
    return (c1 == c2);
//    return (xpath[0] == vfvPath_getDelimChar());
  }

  inline std::string vfvPath_concat(const std::string& path1,
				    const std::string& path2) {
    std::string xpath;
    if ( ! path1.empty() ) {
      xpath = path1;
      xpath += vfvPath_getDelimChar();
    }
    xpath += path2;
    return xpath;
  }

  inline std::string vfvPath_normalize(const std::string& path) {
    std::string xpath(path);
    vfvPath_adjustDelim(xpath);
    std::string driveStr = vfvPath_emitDrive(xpath);
    xpath = CES::OmitDots(xpath, vfvPath_getDelimChar());
    if ( ! driveStr.empty() )
      xpath = driveStr + xpath;
    return xpath;
  }
}

#endif // _VFV_PATH_UTIL_H_


