/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphCouplerBase.h
 * Sphere Coupler super Class Header File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphCouplerBase.h
 * Sphereカプラー基底クラスヘッダーファイル
 * @author     hira
 * @date       2009/10/01
 */
#ifndef _SPHCOUPLERBASE_H_
#define _SPHCOUPLERBASE_H_

#include "app/sphSphereObject.h"
#include "data/sphData.h"
#include "data/sphVoxArray.h"


/**
 * Sphereカプラー基底クラス
 */
class sphCouplerBase: public sphSphereObject
{
private:
  sphData* getDataObjectInDataManager(const char* elem_name );

protected:

  /**
   * 参照・更新ノード情報クラス
   */
  class NodeBoundaryInfo {
  public:
    int m_nodeID;                    ///< 送受信ノードID
    double m_originNode[3];          ///< ノード原点座標
    double m_pitchNode[3];           ///< ノードピッチ
    size_t m_sta_idx[3];             ///< 送受信始点インデックス
    size_t m_size[3];                ///< 送受信サイズ
    SPL_Datatype m_comType;          ///< 送受信通信データ型
    sphData* m_recvBuf;              ///< 受信データバッファ
    size_t m_end_idx[3];             ///< 更新終点インデックス（更新データの終点インデックス）
    int m_crddef;                    ///< 受信データ定義位置
    int m_faceFlag[3][2];            ///< 最外郭面の送信フラグ(-1=最外郭面の送信を行う。)
                                     //   定義位置による無効データが存在しているかチェックを行う為に最外郭情報を送信する。
                                     //   m_invalidFlag[i,j,k方向][マイナス方向, プラス方向]
    int m_dims;                        ///< 次元数
    SPL_Datatype m_dtType;           ///< 送受信データのデータ型

    /** コンストラクタ */
    NodeBoundaryInfo()
    {
      register int i;
      m_nodeID = 0;
      for (i=0; i<3; i++) {
        m_originNode[0] = 0.0;
        m_pitchNode[0] = 0.0;
        m_sta_idx[0] = 0;
        m_end_idx[0] = 0;
        m_size[0] = 0;
      }
      m_comType = SPL_DATATYPE_NULL;
      m_recvBuf = NULL;
      memset(m_faceFlag, 0x00, sizeof(m_faceFlag));
    };

    /** デストラクタ. */
    virtual ~NodeBoundaryInfo()
    {
      if (m_comType != SPL_DATATYPE_NULL) SPL_Type_free(&m_comType);
      if (m_recvBuf != NULL) delete m_recvBuf;
    };
  };

  std::map<int, NodeBoundaryInfo*> m_sendNodeTable;   ///< 送信ノード情報テーブル<送信ノード番号,送信ノード情報>
  std::map<int, NodeBoundaryInfo*> m_recvNodeTable;   ///< 受信ノード情報テーブル<受信ノード番号,受信ノード情報>

  int m_refSolverID;       ///< 参照SolverID
  int m_udtSolverID;       ///< 更新SolverID
  sphData* m_refDataObj;          ///< 参照データオブジェクトポインタ
  sphData* m_uptDataObj;          ///< 更新データオブジェクトポインタ
  SPL_Comm m_couplerComm;               ///< カプラプロセスのコミュニケータ
  SPL_Group m_couplerGroup;             ///< カプラプロセスのグループ

  bool initialize();
  sphData* getReferDataInDataManager(const char* ref_reg_label = NULL);
  sphData* getUpdateDataInDataManager(const char* upd_reg_label = NULL);

  /**
   * 参照データの受信を行う.
   * 参照データのノードからデータを受信する。
   * @param               void
   * @return    success(=true)/fault(=false)
   */
  virtual bool communicateData() = 0;

  /**
   * 更新データオブジェクトにデータの更新を行う.
   * @param               void
   * @return    success(=true)/fault(=false)
   */
  virtual bool interpolate() = 0;

public:
  sphCouplerBase();
  sphCouplerBase(const char* label);
  sphCouplerBase(const char* label, int id);
  virtual ~sphCouplerBase();

  // process communicator
  void setCommGroup(SPL_Comm comm, SPL_Group group);
// ----- modify 20100414 @tamaki begin -----
//  const SPL_Comm getSplComm() const;
//  const SPL_Group getSplGroup() const;
// -----------------------------------------
  SPL_Comm getSplComm() const;
  SPL_Group getSplGroup() const;
// ----- modify 20100414 @tamaki ens   -----
  int getReferSolverId() const;
  int getUpdateSolverId() const;
  sphData* getReferData();
  sphData* getUpdateData();

  // -1 : error
  //  0 : forced terminated
  //  1 : Normality
  /**
   * カプラの初期化を行う（データラベル指定）.
   * 通信ノード、ボクセル情報のテーブルを作成する。
   * @param               ref_reg_label         参照データ登録ラベル
   * @param               upd_reg_label         更新データ登録ラベル
   * @return              -1:error, 0:forced terminated, 1:normality
   */
  virtual int initializeCoupler(const char* ref_reg_label = NULL, const char* upd_reg_label = NULL) = 0;

  /**
   * カプラの実行を行う.
   * @param               void
   * @return              -1:error, 0:forced terminated, 1:normality
   */
  virtual int executeCoupler() = 0;

  /**
   * カプラの破棄処理を行う.
   * @param               void
   * @return              -1:error, 0:forced terminated, 1:normality
   */
  virtual int postCoupler() = 0;


};

#endif /* _SPHCOUPLERBASE_H_ */
