/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSolverBase.h
 * Sphere Solver super Class Header File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphSolverBase.h
 * Sphereソルバー基底クラスヘッダーファイル
 * @author     hira
 * @date       2009/10/01
 */

#ifndef _SPHSOLVERBASE_H_
#define _SPHSOLVERBASE_H_

#include "app/sphSphereObject.h"
#include "parallel/sphParaController.h"
#include "data/sphDataManager.h"
#include "fileio/sphFileHeader.h"
#include "fileio/sphSbxHeader.h"
#include "fileio/sphSpxHeader.h"
#include "fileio/sphSphHeader.h"
#include "fileio/sphTextFile.h"
#include "fileio/sphFileController.h"
#include "base/sphStepTime.h"
#include "skl_config/SklSolverConfig.h"

/**
 * Sphereソルバー基底クラス
 */
class sphSolverBase: public sphSphereObject
{
private:
  bool getGcOutputFlag(const sphElem* file_elem);
  const sphElem* getChildFileElem(const sphElem* data_elem, const char* type);
  const sphElem* createSklInFileElem(const char* file_label);
  const sphElem* createSklOutFileElem(const char* file_label);
  const sphElem* createSklTextFileElem(const char* file_label);

protected:
  int m_solverDims;                     ///< ソルバ次元数
  std::string m_solverMode;           ///< ソルバ実行モード (sph_check:メインループ実行無しモード)
  SPL_Comm m_solverComm;              ///< ソルバプロセスのコミュニケータ
  SPL_Group m_solverGroup;            ///< ソルバプロセスのグループ
  sphDataManager* m_dataMngr;         ///< データマネージャー
  sphParaController* m_paraCtrl;      ///< パラレルコントローラ
  sphFileController* m_fileCtrl;      ///< ファイルコントローラ
  const SklCfg::SklSolverConfig* m_sklconfig;          ///< Vsphereコンフィグレーション

  sphSolverBase();
  virtual bool setParaController(sphParaController* paraCtrl);

  sphData* allocateDataObj(const sphElem* data_elem);
  bool writeFile(const sphElem* file_elem, sphData* data, size_t step, double time, bool gc_flag = false, bool force = false);
  bool writeFileVoxArray(const sphElem* file_elem, sphData* data, size_t step, double time, bool gc_flag = false, bool force = false);
  bool writeFileVoxContainer(const sphElem* file_elem, sphData* data, size_t step, double time, bool gc_flag = false, bool force = false);
  bool getFileHeader(const sphElem* file_elem, sphFileHeader* header);
  bool loadFile(const sphElem* file_elem, sphData* dataObj);
  bool writeFileExcept(const sphElem* file_elem,
                              sphData* dataObj,
                              double vox_org[3],
                              double vox_pit[3],
                              size_t step,
                              double time,
                              size_t vox_sidx[3],
                              int blockid = -1);

public:
  sphSolverBase(const char* solv_label);
  sphSolverBase(const char* solv_label, int solv_id);
  virtual ~sphSolverBase();
  bool initialize();
  bool setupFileController();

  // getter, setter
  const std::string getSolverMode() const;
  const sphDataManager* getDataManager() const;
        sphDataManager* getDataManager();
  const sphParaController* getParaController() const;
        sphParaController* getParaController();
  void setCommGroup(SPL_Comm comm, SPL_Group group);
// ----- modify 20100414 @tamaki begin -----
//  const SPL_Comm getSplComm() const;
//  const SPL_Group getSplGroup() const;
// -----------------------------------------
  SPL_Comm getSplComm() const;
  SPL_Group getSplGroup() const;
// ----- modify 20100414 @tamaki end   -----
  const sphStepTime* getStepTime() const;
        sphStepTime* getStepTime();

  sphData* getDataObj(const char* label);
  bool registerDataObj(const char* label, sphData* dataObj);
  bool deleteDataObj(const char* label);
  bool deleteDataObj(sphData* dataObj);
  int getDims() const;
  void setDims(int dims);
  bool isCheckMode() const;
  bool getVoxelSize(size_t size[3]);
  bool getVoxelStartIndex(size_t sidx[3]);
  bool getWholeVoxelSize(size_t wsize[3]);

  /**
   * ソルバの初期化を行う.
   * @param               steptime          実行ステップ・時刻クラス
   * @return              -1:error, 0:forced terminated, 1:normality
   */
  virtual int initializeSolver(sphStepTime* steptime) = 0;

  /**
   * ソルバのステップ実行を行う.
   * @param               steptime          実行ステップ・時刻クラス
   * @return              -1:error, 0:forced terminated, 1:normality
   */
  virtual int loopSolver(sphStepTime* steptime) = 0;

  /**
   * ソルバの破棄処理を行う.
   * @param               steptime          実行ステップ・時刻クラス
   * @return              -1:error, 0:forced terminated, 1:normality
   */
  virtual int postSolver(sphStepTime* steptime) = 0;

  /**
   * ソルバの使用方法の表示を行う.
   * @param               cmd           コマンド
   * @return              -1:error, 0:forced terminated, 1:normality
   */
  virtual int printSolverUsage(const char* cmd) = 0;


  // データオブジェクトの生成
  virtual sphData* allocateDataObj(const char* cfg_label);
  virtual sphData* allocateDataObj(
                        const char* reg_label,
                        SphDCType dcType,
                        SPL_Datatype dataType,
                        const size_t gc = 0,
                        const size_t data_num = 1,
                        SphCrdDef crddef = CRDDEF_REGULAR);
  virtual sphData* allocateDataObj(
                        SphDCType dcType,
                        SPL_Datatype dataType,
                        const size_t gc = 0,
                        const size_t data_num = 1,
                        SphCrdDef crddef = CRDDEF_REGULAR);
  virtual sphData* allocateDataObjExcept(
                        const char* reg_label,
                        SphDCType dcType,
                        SPL_Datatype dataType,
                        const size_t size[3],
                        const size_t gc = 0,
                        const size_t data_num = 1);
  virtual sphData* allocateDataObjExcept(
                        SphDCType dcType,
                        SPL_Datatype dataType,
                        const size_t size[3],
                        const size_t gc = 0,
                        const size_t data_num = 1);

  // ファイル入出力
  virtual bool getFileHeader(const char* file_label, sphFileHeader* header);
  virtual sphData* loadFile(
                        const char* file_label,
                        const size_t size[3],
                        const size_t start_idx[3],
                        size_t gc);
  virtual bool loadFile(const char* file_label, sphData* data);
  virtual bool writeFile(const char* file_label, const char* data_reg_label);
  virtual bool writeFile(const char* data_cfg_label);
  virtual bool writeFile(const char* file_label, sphData* data);
  virtual bool writeFile(const char* data_cfg_label, size_t step, double time, bool gc_flag = false, bool force = false);
  virtual bool writeFile(const char* file_label, sphData* dataObj, size_t step, double time, bool gc_flag = false, bool force = false);
  virtual bool getTextFile(const char* file_label, sphTextFile* text_file, bool initopen = true);

  // Vsphere Configuration
  void setSklConfig(const SklCfg::SklSolverConfig* skl_config);
  const SklCfg::SklSolverConfig* getSklConfig() const;
  bool isClassicConfig();

};

#endif /* _SPHSOLVERBASE_H_ */
