/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSolverUnst.h
 * Sphere Unstructured Mesh Solver super Class Header File
 * @author     hira
 * @date       2010/04/01
 *
 * \~japanese
 * @file sphSolverUnst.h
 * Sphere非構造格子ソルバー基底クラスヘッダーファイル
 * @author     hira
 * @date       2010/04/01
 */

#ifndef _SPHSOLVERUNST_H_
#define _SPHSOLVERUNST_H_

#include "solverbase/sphSolverBase.h"
#include "parallel/sphUnstParaController.h"
#include "data/sphData.h"
#include "fileio/sphFileController.h"

/****************************************************
 *             Matrix data structure                *
****************************************************/
class Matrix {
public:
  int        nNode;
  int        nElem;
  double    *mat;
  const int *nop;
  Matrix() {
    nNode = 0;
    nElem = 0;
    mat = NULL;
    nop = NULL;
  };
  ~Matrix() {
    if( mat ) delete [] mat;
  };
};

class MatrixF {
public:
  int        nNode;
  int        nElem;
  float     *mat;
  const int *nop;
  MatrixF() {
    nNode = 0;
    nElem = 0;
    mat = NULL;
    nop = NULL;
  };
  ~MatrixF() {
    if( mat ) delete [] mat;
  };
};

/****************************************************
 *             Tensor data structure                *
****************************************************/
class Matrix3 {
public:
  int     nNode;
  int     nElem;
  Matrix *x;
  Matrix *y;
  Matrix *z;
  Matrix3() {
    nNode = 0;
    nElem = 0;
    x = y = z = NULL;
  };
  ~Matrix3() {
     if( x ) delete [] x;
     if( y ) delete [] y;
     if( z ) delete [] z;
  };
};

class MatrixF3 {
public:
  int     nNode;
  int     nElem;
  MatrixF *x;
  MatrixF *y;
  MatrixF *z;
  MatrixF3() {
    nNode = 0;
    nElem = 0;
    x = y = z = NULL;
  };
  ~MatrixF3() {
     if( x ) delete [] x;
     if( y ) delete [] y;
     if( z ) delete [] z;
  };
};

class Matrix9 {
public:
  int     nNode;
  int     nElem;
  Matrix *xx;
  Matrix *xy;
  Matrix *xz;
  Matrix *yx;
  Matrix *yy;
  Matrix *yz;
  Matrix *zx;
  Matrix *zy;
  Matrix *zz;
  Matrix9() {
    nNode = 0;
    nElem = 0;
    xx = xy = xz = yx = yy = yz = zx = zy = zz = NULL;
  };
  ~Matrix9() {
     if( xx ) delete [] xx;
     if( xy ) delete [] xy;
     if( xz ) delete [] xz;
     if( yx ) delete [] yx;
     if( yy ) delete [] yy;
     if( yz ) delete [] yz;
     if( zx ) delete [] zx;
     if( zy ) delete [] zy;
     if( zz ) delete [] zz;
   };
};

class MatrixF9 {
public:
  int     nNode;
  int     nElem;
  MatrixF *xx;
  MatrixF *xy;
  MatrixF *xz;
  MatrixF *yx;
  MatrixF *yy;
  MatrixF *yz;
  MatrixF *zx;
  MatrixF *zy;
  MatrixF *zz;
  MatrixF9() {
    nNode = 0;
    nElem = 0;
    xx = xy = xz = yx = yy = yz = zx = zy = zz = NULL;
  };
  ~MatrixF9() {
     if( xx ) delete [] xx;
     if( xy ) delete [] xy;
     if( xz ) delete [] xz;
     if( yx ) delete [] yx;
     if( yy ) delete [] yy;
     if( yz ) delete [] yz;
     if( zx ) delete [] zx;
     if( zy ) delete [] zy;
     if( zz ) delete [] zz;
   };
};

/**
 * Sphere非構造格子ソルバー基底クラス
 */
class sphSolverUnst : public sphSolverBase
{

private:
  bool initialize();

protected:
  sphUnstParaController* m_unstParaCtrl;    ///< UNSTパラレルコントローラ（sphSolverBase::m_paraCtrlと同一ポインタ）

  sphSolverUnst();

  virtual int setupSolver();
  virtual int setupDataList();

  virtual int initializeUnstMesh();
  virtual int initializeUnstMesh(const char* fn_node,
                                 const char* fn_elem,
                                 const char* fn_grp = NULL,
                                 const char* fn_bnd = NULL,
                                 sphUnstMesh::MeshDataType dtype = sphUnstMesh::MESH_DOUBLE);
  virtual int initializeUnstMesh(const char* dist_fname,
                                 sphUnstMesh::MeshDataType dtype = sphUnstMesh::MESH_DOUBLE);

  virtual bool setParaController(sphParaController* paraCtrl);

  bool ShapeFunction (int Enum, double *detJ, double N[8][8],
                      double NX[8][8], double NY[8][8], double NZ[8][8]);
  bool ShapeFunctionF(int Enum, float *detJ, float N[8][8],
                      float NX[8][8], float NY[8][8], float NZ[8][8]);
  void ZeroClear (int n, double *v);
  void ZeroClearF(int n, float *v);

public:

  sphSolverUnst(const char* solv_label);
  sphSolverUnst(const char* solv_label, int solv_id);
  virtual ~sphSolverUnst();

  // -1 : error
  //  0 : forced terminated
  //  1 : Normality
  virtual int initializeSolver(sphStepTime* steptime);
  virtual int loopSolver(sphStepTime* steptime) = 0;
  virtual int postSolver(sphStepTime* steptime) = 0;
  virtual int printSolverUsage(const char* cmd) = 0;

  const sphUnstParaController* getUnstParaController() const { return m_unstParaCtrl; };
        sphUnstParaController* getUnstParaController() { return m_unstParaCtrl; };

  /* Make Matrix (double) */
  bool MakeAdvMatrix(Matrix *advMAT, double* dp_u, double* dp_v, double* dp_w);
  bool MakeDiffMatrix(double Reinv, Matrix9 *diffMAT);
  bool MakeDivMatrix(Matrix3 *divMAT);
  bool MakeGradMatrix(Matrix3 *gradMAT);
  bool MakeMassMatrix(Matrix *massMAT);
  bool MakePoisson(Matrix *Poisson);

  /* Make Matrix (float) */
  bool MakeAdvMatrix(MatrixF *advMAT, float* dp_u, float* dp_v, float* dp_w);
  bool MakeDiffMatrix(float Reinv, MatrixF9 *diffMAT);
  bool MakeDivMatrix(MatrixF3 *divMAT);
  bool MakeGradMatrix(MatrixF3 *gradMAT);
  bool MakeMassMatrix(MatrixF *massMAT);
  bool MakePoisson(MatrixF *Poisson);

  /* Matrix - vector product */
  bool MatVecProduct(int type, Matrix *A, double *x, double *b );
  bool MatVecProduct(int type, MatrixF *A, float *x, float *b );

};

#endif /* _SPHSOLVERUNST_H_ */
