/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sph_win32_util.h
 * utility function for Windows
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sph_win32_util.h
 * Windows用ユーティリティファイル
 * @author     hira
 * @date       2009/02/20
 */


#ifndef _SPH_WIN32_UTIL_H
#define _SPH_WIN32_UTIL_H
#if defined(_WIN32) || defined(WIN32)

#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <io.h>
#include <math.h>
#include    <time.h>

#define MAXPATHLEN _MAX_PATH

#define strcasecmp(s1, s2) stricmp(s1, s2)
#define strncasecmp(s1, s2, n) strnicmp(s1, s2, n)
#define pow(x, y) pow((double)x, (double)y)
#define sleep(seconds) (Sleep(seconds*1000))
#define usleep(useconds) (Sleep(useconds))
#define chdir(path) SetCurrentDirectory(path)
// #define open(fd, O_RDONLY) open(fd, O_RDONLY|O_BINARY)
#define O_RDONLY        _O_RDONLY|O_BINARY
#define O_WRONLY        _O_WRONLY|O_BINARY    // not use
#define O_RDWR          _O_RDWR|O_BINARY    // not use
#define O_APPEND        _O_APPEND|O_BINARY    // not use
#define O_CREAT         _O_CREAT|O_BINARY    // not use

/* Encoding of the file mode.  */

#define __S_IFMT        0170000 /* These bits determine file type.  */

/* File types.  */
#define __S_IFDIR       0040000 /* Directory.  */
#define __S_IFCHR       0020000 /* Character device.  */
#define __S_IFBLK       0060000 /* Block device.  */
#define __S_IFREG       0100000 /* Regular file.  */
#define __S_IFIFO       0010000 /* FIFO.  */
#define __S_IFLNK       0120000 /* Symbolic link.  */

/* Test macros for file types.  */
#define __S_ISTYPE(mode, mask)  (((mode) & __S_IFMT) == (mask))
#define S_ISDIR(mode)    __S_ISTYPE((mode), __S_IFDIR)
#define S_ISCHR(mode)    __S_ISTYPE((mode), __S_IFCHR)
#define S_ISBLK(mode)    __S_ISTYPE((mode), __S_IFBLK)
#define S_ISREG(mode)    __S_ISTYPE((mode), __S_IFREG)

struct timezone {
    int tz_minuteswest;
    int tz_dsttime;
};

#define EPOCHFILETIME (116444736000000000i64)

static char* dirname(char *path)
{

  static char buffer[_MAX_PATH];
  const char *p;
  const char *dir_p = NULL;
  memset(buffer, NULL, sizeof(buffer));

  // null check
  if (path==NULL)
  {
    strcpy(buffer, ".");
    return buffer;
  }

  // length check
  size_t len = strlen(path);
  if (len <= 0)
  {
    strcpy(buffer, ".");
    return buffer;
  }
  if (len > sizeof(buffer))
  {
    return buffer;
  }

  for (p = path; *p; p++)
  {
    // check shift-jis 2bytes character
    if((0x81 <= (unsigned char)*p && (unsigned char)*p <= 0x9F)
      || (0xE0 <= (unsigned char)*p && (unsigned char)*p <= 0xFC))
    {
      p++;
      continue;
    }
    else if (*p == '\\' || *p == '/')
    {
      dir_p = p;
    }
  }

  // not found FILE_SEPALATER
  if (dir_p == NULL)
  {
    strcpy(buffer, ".");
  }
  else if (dir_p == path)
  {
    memcpy(buffer, path, 1);
  }
  else if (dir_p == path+len-1)
  {
    memcpy(buffer, dir_p, 1);
  }
  else
  {
    memcpy(buffer, path, dir_p-path);
  }
  return buffer;
}

static char* basename(char *path)
{
  static char buffer[_MAX_PATH];
  const char *p;
  char *base_p = path;
  memset(buffer, NULL, sizeof(buffer));

  // null check
  if (path==NULL)
  {
    strcpy(buffer, ".");
    return buffer;
  }

  // length check
  size_t len = strlen(path);
  if (len <= 0)
  {
    strcpy(buffer, ".");
    return buffer;
  }
  if (len > sizeof(buffer))
  {
    return buffer;
  }

  for (p = base_p; *p; p++)
  {
    // check shift-jis 2bytes character
    if((0x81 <= (unsigned char)*p && (unsigned char)*p <= 0x9F)
      || (0xE0 <= (unsigned char)*p && (unsigned char)*p <= 0xFC))
    {
      p++;
      continue;
    }
    else if (*p == '\\' || *p == '/')
    {
      if (! *(p+1))
      {
        len--;
        break;
      }
      base_p = (char*)p + 1;
    }
  }
  if (len - (base_p - path) > 0)
  {
    strncpy(buffer, base_p, len - (base_p - path));
  }
  else
  {
    strcpy(buffer, base_p);
  }
  return buffer;
}



static int gettimeofday(struct timeval *tv, struct timezone *tz)
{
    FILETIME        tagFileTime;
    LARGE_INTEGER   largeInt;
    __int64         val64;
    static int      tzflag;

    if (tv)
    {
        GetSystemTimeAsFileTime(&tagFileTime);

        largeInt.LowPart  = tagFileTime.dwLowDateTime;
        largeInt.HighPart = tagFileTime.dwHighDateTime;
        val64 = largeInt.QuadPart;
        val64 = val64 - EPOCHFILETIME;
        val64 = val64 / 10;
        tv->tv_sec  = (long)(val64 / 1000000);
        tv->tv_usec = (long)(val64 % 1000000);
    }

    if (tz)
    {
        if (!tzflag)
        {
            _tzset();
            tzflag++;
        }

        long _Timezone = 0;
         _get_timezone(&_Timezone);
         tz->tz_minuteswest = _Timezone / 60;

         int _Daylight = 0;
         _get_daylight(&_Daylight);
         tz->tz_dsttime = _Daylight;
    }

    return 0;
}


#define MPICH_REG_SUB_KEY   "SOFTWARE\\MPICH2"
#define MPICH_REG_PATH_KEY   "Path"

static bool GetMpichPath(char *path)
{
  HKEY hKey;
  LONG    lResult;
  DWORD   dwType;
  DWORD   dwSize;
  char    szBuf[MAX_PATH];
  bool ret = false;

  // open Registry
  if(  RegOpenKeyEx(HKEY_LOCAL_MACHINE, MPICH_REG_SUB_KEY, 0, KEY_READ, &hKey) != ERROR_SUCCESS)
    return false;

  // get Value size
  lResult = RegQueryValueEx(hKey, MPICH_REG_PATH_KEY, NULL, &dwType, NULL, &dwSize);
  lResult = RegQueryValueEx(hKey, MPICH_REG_PATH_KEY, NULL, &dwType, (LPBYTE)szBuf,&dwSize);
  if( lResult == ERROR_SUCCESS ) {
    strcpy(path, szBuf);
    ret = true;
  }

  lResult = RegCloseKey(hKey);
  return ret;

}


#endif /* _WIN32 or WIN32 */
#endif /* _SPH_WIN32_UTIL_H */
