/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file splfio_sbx.h
 * Declaration of struct splSbx for Sphere Low Level Library
 *
 * DO NOT INCLUDE THIS FILE DIRECTLY
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file splfio_sbx.h
 * SBXファイルのヘッダーレコードの構造体定義、SBXファイル入出力関数ヘッダーファイル
 *
 * このヘッダーファイルの直接インクルード禁止
 * @author     hira
 * @date       2009/02/20
 */


#ifndef _SPL_FIO_SBX_H_
#define _SPL_FIO_SBX_H_

#include "spl.h"
#include "splfio_sv_data.h"
#include "splfio_sv_file.h"
#include "splmpi.h"

#if defined(__cplusplus)
extern "C" {
#endif


/**
 * \~english
 * splSbx is structure to store SBX data.
 *
 * \~japanese
 * SBXデータの格納用構造体
 */
typedef struct _spl_sbx_ {
  /** whole voxel size for extended record*/
  size_t wsize[3];

  /**
   * global start index for node
   *
   * start_idx[0] x-dir start
   * start_idx[1] y-dir start
   * start_idx[2] z-dir start
   */
  size_t start_idx[3];

  /** dimension ( 1:1D, 2:2D, 3:3D ) */
  unsigned int dims;

  /** vector length */
  unsigned int vlen;

  /** data type ( 1:unsigned char, 2:unsigned short, 4:unsigned int ) */
  unsigned int dtype;

  /** guide cell size */
  size_t gc;

  /** data type ( 4:float, 8:double ) */
  unsigned int rlen;

  /**
   * origin definision
   *
   *  1:Regular
   *  2:Collocate
   *  3:Staggered(1)
   *  4:Staggered(2)
   */
  unsigned int crddef;

  /** extended data */
  unsigned int aux;

  /** voxel size */
  size_t size[3];

  /** origin */
  double origin[3];

  /** pitch */
  double pitch[3];

  /** block size */
  size_t block_size;

#if 0 /* SBX has no time-record */
  /** step */
  size_t step;

  /** time */
  double time;
#endif

  /** data */
  splData data;

} splSbx;

extern SplBool splInitSbx(splSbx* obj);
extern SplBool splFinalizeSbx(splSbx* obj, SplBool mem_free_flag);
extern SplBool splSbxSetValues(splSbx* obj, unsigned int dims, unsigned int vlen,
                            unsigned int dtype, size_t gc,
                            unsigned int rlen, unsigned int crddef,
                            size_t* sta_idx, size_t* size,
                            unsigned char* dt);
extern SplBool splSbxSetCoords(splSbx* obj,unsigned int dims, double* org, double* pit);
extern SplBool splSbxSetWholeSize(splSbx* obj, unsigned int dims, size_t* wsize);

/*
 * Getting SBX header infomation(without data-record).
 *
 * mode : SBX read mode
 *  0 : All nodes open the file and it reads data.
 *  1 : Only the node of rank 0 opens the file and data is broadcast
 *      to other nodes. Rank 0 also sets the data read from the file
 *      in the splSbx object.
 *  2 : Only the node of rank 0 opens the file and data is broadcast
 *      to other nodes. However, rank 0 only reads data from the file,
 *      and doesn't set data in the splSbx object.
 */
extern SplBool splGetSbxInfo(const char* fname, splSbx* obj, int mode,
                             int readRank, SPL_Comm grp);

/*
 * Before the call of splReadSbx,
 * splSbx object must be set the following values.
 *  wsize, start_idx, dims, size, dtype,
 *  rlen, data, and vlen(=1:default value).
 *
 * mode : SBX read mode
 *  0 : All nodes open the file and it reads data.
 *  1 : Only the node of rank 0 opens the file and data is broadcast
 *      to other nodes. Rank 0 also sets the data read from the file
 *      in the splSbx object.
 *  2 : Only the node of rank 0 opens the file and data is broadcast
 *      to other nodes. However, rank 0 only reads data from the file,
 *      and doesn't set data in the splSbx object.
 */
extern SplBool splReadSbx(splSvFileList* file_list, int mode,
                          splSbx* obj, int readRank, SPL_Comm grp);

/*
 * mode : SBX write mode
 *  0 : All nodes output data to each file.
 *  1 : The data of each node is gathered in rank 0, and rank 0 writes
 *      the data to one file.
 */
extern SplBool splWriteSbx(const char* fname, int mode,
                           SplBool cmp_flag, SplBool ext_rec_flag,
                           size_t block_size, const splSbx* obj,
                           int writeRank, SPL_Comm grp);

#if 1
extern void splPrintSbx(const splSbx* obj, SPL_Comm grp);
#endif

#if defined(__cplusplus)
}
#endif /* extern "C" */

#endif /* _SPL_FIO_SBX_H_ */

