/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file splfio_spx.h
 * Declaration of struct splSpx for Sphere Low Level Library
 *
 * DO NOT INCLUDE THIS FILE DIRECTLY
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file splfio_spx.h
 * SPXファイルのヘッダーレコードの構造体定義、SPXファイル入出力関数ヘッダーファイル
 *
 * このヘッダーファイルの直接インクルード禁止
 * @author     hira
 * @date       2009/02/20
 */


#ifndef _SPL_FIO_SPX_H_
#define _SPL_FIO_SPX_H_

#include "spl.h"
#include "splfio_sv_data.h"
#include "splfio_sv_file.h"
#include "splmpi.h"

#if defined(__cplusplus)
extern "C" {
#endif


/**
 * \~english
 * splSpx is structure to store SPX data.
 *
 * \~japanese
 * SPXデータの格納用構造体
 */
typedef struct _spl_spx_ {
  /** whole voxel size for extended record*/
  size_t wsize[3];

  /**
   * groval index for node
   *
   * start_idx[0] x-dir start
   * start_idx[1] y-dir start
   * start_idx[2] z-dir start
   */
  size_t start_idx[3];

  /** dimension ( 1:1D, 2:2D, 3:3D ) */
  unsigned int dims;

  /** vector length */
  unsigned int vlen;

  /** data type ( 1:scalar data type, 3:vector data type ) */
  unsigned int dtype;

  /** guide cell size */
  size_t gc;

  /** data type ( 4:float, 8:double ) */
  unsigned int rlen;

  /**
   * origin definision
   *
   *  1:Regular
   *  2:Collocate
   *  3:Staggered(1)
   *  4:Staggered(2)
   */
  unsigned int crddef;

  /** extended data */
  unsigned int aux;

  /** voxel size */
  size_t size[3];

  /** origin */
  double origin[3];

  /** pitch */
  double pitch[3];

  /** step */
  size_t step;

  /** time */
  double time;

  /** block size */
  size_t block_size;

  /** data */
  splData data;

  /** gc in sphfile */
  size_t gc_insph;            // add 2010/05/24 for sph format

} splSpx;

extern SplBool splInitSpx(splSpx* obj);
extern SplBool splFinalizeSpx(splSpx* obj, SplBool mem_free_flag);
extern SplBool splSpxSetValues(splSpx* obj, unsigned int dims, unsigned int vlen,
                            unsigned int dtype, size_t gc,
                            unsigned int rlen, unsigned int crddef,
                            size_t* sta_idx, size_t* size,
                            unsigned char* dt);
extern SplBool splSpxSetCoords(splSpx* obj,unsigned int dims,
                            double* org, double* pit);
extern SplBool splSpxSetWholeSize(splSpx* obj, unsigned int dims, size_t* wsize);


/*
 * Getting SPX header infomation(without data-record).
 *
 * mode : SPX read mode
 *  0 : All nodes open the file and it reads data.
 *  1 : Only the node of rank 0 opens the file and data is broadcast
 *      to other nodes. Rank 0 also sets the data read from the file
 *      in the splSpx object.
 *  2 : Only the node of rank 0 opens the file and data is broadcast
 *      to other nodes. However, rank 0 only reads data from the file,
 *      and doesn't set data in the splSpx object.
 */
extern SplBool splGetSpxInfo(const char* fname, splSpx* obj, int mode,
                             int readRank, SPL_Comm grp);

/*
 * Before the call of splReadSpx,
 * splSpx object must be set the following values.
 *  wsize, start_idx, dims, size, dtype,
 *  rlen, data, and vlen(=1:default value).
 *
 * mode : SPX read mode
 *  0 : All nodes open the file and it reads data.
 *  1 : Only the node of rank 0 opens the file and data is broadcast
 *      to other nodes. Rank 0 also sets the data read from the file
 *      in the splSpx object.
 *  2 : Only the node of rank 0 opens the file and data is broadcast
 *      to other nodes. However, rank 0 only reads data from the file,
 *      and doesn't set data in the splSpx object.
 */
extern SplBool splReadSpx(splSvFileList* file_list, int mode,
                          splSpx* obj, int readRank, SPL_Comm grp);

/*
 * mode : SPX write mode
 *  0 : All nodes output data to each file.
 *  1 : The data of each node is gathered in rank 0, and rank 0 writes
 *      the data to one file.
 */
extern SplBool splWriteSpx(const char* fname, int mode,
                           SplBool cmp_flag, SplBool ext_rec_flag,
                           size_t block_size, const splSpx* obj,
                           int writeRank, SPL_Comm grp);

#if 1
extern void splPrintSpx(const splSpx* obj, SPL_Comm grp);
#endif

#if defined(__cplusplus)
}
#endif /* extern "C" */

#endif /* _SPL_FIO_SPX_H_ */

