/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file splfio_sv_file_inline.h
 * Inline function implementation of SvFile I/F
 *
 * DO NOT INCLUDE THIS FILE DIRECTLY
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file splfio_sv_file_inline.h
 * ファイル入出力関数インラインファイル
 *
 * このヘッダーファイルの直接インクルード禁止
 * @author     hira
 * @date       2009/02/20
 */


#ifndef _SPL_FIO_SV_FILE_INLINE_H_
#define _SPL_FIO_SV_FILE_INLINE_H_

#include <string.h>
#include <stdlib.h>
#include "spl_errno.h"

#if defined(__cplusplus)
extern "C" {
#endif

#define SPL_SV_FILE_MEMFUN(rettype) \
        SPL_SV_FILE_INLINE rettype

//extern int spl_err_code;


/**
 * \~english
 * Initialize of splSvFile struct
 * @param      obj          splSvFile struct
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSvFile構造体の初期化を行う。
 * @param      obj          splSvFile構造体オブジェクト
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SPL_SV_FILE_MEMFUN(SplBool)
splInitSvFile(splSvFile* obj) {
  if( !obj ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "splSvFile object is null");
    return SplFalse;
  }
  obj->set_flag = 0;
  obj->start_idx[0] = obj->start_idx[1] = obj->start_idx[2] = (size_t)-1;
  memset((void*)obj->fname, 0, SPL_MAX_STR_SIZE);
  obj->next = NULL;
  return SplTrue;
}

/**
 * \~english
 * Finalize of splSvFile struct
 * @param      obj          splSvFile struct
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSvFile構造体の破棄を行う。
 * @param      obj          splSvFile構造体オブジェクト
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SPL_SV_FILE_MEMFUN(SplBool)
splFinalizeSvFile(splSvFile* obj) {
  if( !obj ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "splSvFile object is null");
    return SplFalse;
  }
  obj->start_idx[0] = obj->start_idx[1] = obj->start_idx[2] = (size_t)-1;
  memset((void*)obj->fname, 0, SPL_MAX_STR_SIZE);
  obj->next = NULL;
  obj->set_flag = 0;
  return SplTrue;
}

/**
 * \~english
 * set File and start index in splSvFile struct
 * @param      obj          splSvFile struct
 * @param     fn          File Name
 * @param      sidx      start index
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSvFile構造体にファイル名、始点インデックスをセットする。
 * @param      obj          splSvFile構造体オブジェクト
 * @param     fn          ファイル名
 * @param      sidx      始点インデックス
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SPL_SV_FILE_MEMFUN(SplBool)
splSetSvFile(splSvFile* obj, const char* fn, const size_t sidx[3]) {
  if( !obj || !fn || !sidx ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "obj, fn or sidx is null");
    return SplFalse;
  }
  if( strlen(fn) >= SPL_MAX_STR_SIZE ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_STR_SIZE_TOO_LONG, "fn too long");
    return SplFalse;
  }

  obj->start_idx[0] = sidx[0];
  obj->start_idx[1] = sidx[1];
  obj->start_idx[2] = sidx[2];

  strcpy((char*)obj->fname, fn);
  obj->next = NULL;
  obj->set_flag = 1;
  return SplTrue;
}


/**
 * \~english
 * set FileName in splSvFile struct
 * @param      obj          splSvFile struct
 * @param     fn          File Name
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSvFile構造体にファイル名をセットする。
 * @param      obj          splSvFile構造体オブジェクト
 * @param     fn          ファイル名
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SPL_SV_FILE_MEMFUN(SplBool)
splSetSvFileName(splSvFile* obj, const char* fn) {
  if( !obj || !fn ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "obj, fn or sidx is null");
    return SplFalse;
  }
  if( strlen(fn) >= SPL_MAX_STR_SIZE ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_STR_SIZE_TOO_LONG, "fn too long");
    return SplFalse;
  }

  obj->start_idx[0] = 0;
  obj->start_idx[1] = 0;
  obj->start_idx[2] = 0;

  strcpy((char*)obj->fname, fn);
  obj->next = NULL;
  obj->set_flag = 0;
  return SplTrue;
}
/**
 * \~english
 * delete File and start index in splSvFile struct
 * @param      fobj          splSvFile struct
 *
 * \~japanese
 * splSvFile構造体オブジェクトを削除する。
 * @param      fobj          splSvFile構造体オブジェクト
 */
SPL_SV_FILE_MEMFUN(void)
splDelSvFile(splSvFile* fobj) {
  if( !fobj ) return;
  if( fobj->next ) splDelSvFile(fobj->next);
  free(fobj);
  return;
}


/**
 * \~english
 * initialize splSvFileList struct object
 * @param      obj          splSvFileList struct
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSvFileList構造体オブジェクトを初期化する。
 * @param      obj          splSvFileList構造体オブジェクト
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SPL_SV_FILE_MEMFUN(SplBool)
splInitSvFileList(splSvFileList* obj) {
  if( !obj ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "splSvFileList object is null");
    return SplFalse;
  }
  obj->list_size = 0;
  obj->top = NULL;
  return SplTrue;
}

/**
 * \~english
 * finalize splSvFileList struct object
 * @param      obj          splSvFile struct
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSvFileList構造体オブジェクトを破棄する。
 * @param      obj          splSvFile構造体オブジェクト
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SPL_SV_FILE_MEMFUN(SplBool)
splFinalizeSvFileList(splSvFileList* obj) {
  if( !obj ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "splSvFileList object is null");
    return SplFalse;
  }
  splDelSvFile(obj->top);
  obj->list_size = 0;
  obj->top = NULL;
  return SplTrue;
}

/**
 * \~english
 * add file information in splSvFileList struct object
 * @param      obj          splSvFileList struct
 * @param     fn          File Name
 * @param      sidx        start index
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSvFileList構造体オブジェクトにファイル情報を追加する。
 * @param      obj          splSvFile構造体オブジェクト
 * @param     fn          ファイル名
 * @param      sidx        始点インデックス
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SPL_SV_FILE_MEMFUN(SplBool)
splAddSvFile(splSvFileList* obj, const char* fn, const size_t sidx[3]) {
  if( !obj || !fn ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "obj or fn is null");
    return SplFalse;
  }
  splSvFile* fobj = (splSvFile*)malloc(sizeof(splSvFile));
  if( !fobj ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "splSvFile malloc size is %d", sizeof(splSvFile));
    return SplFalse;
  }
  splInitSvFile(fobj);
  if( !splSetSvFile(fobj, fn, sidx) ) {
    return SplFalse;
  }
  fobj->next = obj->top;
  obj->top = fobj;
  obj->list_size++;
  return SplTrue;
}


/**
 * \~english
 * add file name in splSvFileList struct object
 * @param      obj          splSvFileList struct
 * @param     fn          File Name
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSvFileList構造体オブジェクトにファイル名を追加する。
 * @param      obj          splSvFile構造体オブジェクト
 * @param     fn          ファイル名
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SPL_SV_FILE_MEMFUN(SplBool)
splAddSvFileName(splSvFileList* obj, const char* fn) {
  if( !obj || !fn ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "obj or fn is null");
    return SplFalse;
  }
  splSvFile* fobj = (splSvFile*)malloc(sizeof(splSvFile));
  if( !fobj ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "splSvFile malloc size is %d", sizeof(splSvFile));
    return SplFalse;
  }
  splInitSvFile(fobj);
  if( !splSetSvFileName(fobj, fn) ) {
    return SplFalse;
  }
  fobj->next = obj->top;
  obj->top = fobj;
  obj->list_size++;
  return SplTrue;
}

#undef SPL_SV_FILE_MEMFUN

#if defined(__cplusplus)
}
#endif /* extern "C" */

#endif /* _SPL_FIO_SV_FILE_INLINE_H_ */

