/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file splmpi_mpi_inline.h
 * Implementation MPI wrapper for Sphere Low Level Library
 *
 * DO NOT INCLUDE THIS FILE DIRECTLY
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file splmpi_mpi_inline.h
 * MPIラッパー関数郡インラインファイル
 *
 * このヘッダーファイルの直接インクルード禁止
 * @author     hira
 * @date       2009/02/20
 */

#ifndef _SPL_MPI_MPI_INLINE_H_
#define _SPL_MPI_MPI_INLINE_H_


/* int MPI_Send(void*, int, MPI_Datatype, int, int, MPI_Comm) */
/**
 * Performs a basic send
 * @param        buf          initial address of send buffer
 * @param        count        number of elements in send buffer
 * @param        datatype    datatype of each send buffer element
 * @param        dest        rank of destination
 * @param        tag          message tag
 * @param        comm        communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Send(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm) {
  return MPI_Send(buf, count, datatype, dest, tag, comm); }


/* int MPI_Recv(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Status *) */
/**
 * Basic receive
 * @param[out]  buf          initial address of receive buffer
 * @param[in]      count        maximum number of elements in receive buffer
 * @param[in]      datatype    datatype of each receive buffer element
 * @param[in]      source      rank of source
 * @param[in]      tag          message tag
 * @param[in]      comm        communicator
 * @param[out]  status      status object
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Recv(void *buf, int count, SPL_Datatype datatype, int source, int tag, SPL_Comm comm, SPL_Status *status) {
  return MPI_Recv(buf, count, datatype, source, tag, comm, status); }

/* int MPI_Get_count(MPI_Status *, MPI_Datatype, int *) */
/**
 * Gets the number of "top level" elements
 * @param[in]      status      return status of receive operation
 * @param[in]      datatype    datatype of each receive buffer element
 * @param[out]  count        number of received elements
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Get_count(SPL_Status *status, SPL_Datatype datatype, int *count) {
  return MPI_Get_count(status, datatype, count); }

/* int MPI_Bsend(void*, int, MPI_Datatype, int, int, MPI_Comm) */
/**
 * Basic send with user-specified buffering
 * @param       buf          initial address of send buffer
 * @param      count        number of elements in send buffer
 * @param       datatype    datatype of each send buffer element
 * @param        dest        rank of destination
 * @param        tag          message tag
 * @param        comm        communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Bsend(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm) {
  return MPI_Bsend(buf, count, datatype, dest, tag, comm); }

/* int MPI_Ssend(void*, int, MPI_Datatype, int, int, MPI_Comm) */
/**
 * Basic synchronous send
 * @param        buf         initial address of send buffer
 * @param        count        number of elements in send buffer
 * @param        datatype    datatype of each send buffer element
 * @param        dest        rank of destination
 * @param        tag          message tag
 * @param        comm        communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Ssend(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm ) {
  return MPI_Ssend(buf, count, datatype, dest, tag, comm ); }

/* int MPI_Rsend(void*, int, MPI_Datatype, int, int, MPI_Comm) */
/**
 * Basic ready send
 * @param        buf         initial address of send buffer
 * @param        count        number of elements in send buffer
 * @param        datatype    datatype of each send buffer element
 * @param        dest        rank of destination
 * @param        tag          message tag
 * @param        comm        communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Rsend(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm) {
  return MPI_Rsend(buf, count, datatype, dest, tag, comm); }

/* int MPI_Buffer_attach( void*, int) */
/**
 * Attaches a user-defined buffer for sending
 * @param        buffer         initial buffer address
 * @param        size        buffer size, in bytes
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Buffer_attach(void *buffer, int size) {
  return MPI_Buffer_attach(buffer, size); }

/* int MPI_Buffer_detach( void*, int*) */
/**
 * Removes an existing buffer (for use in MPI_Bsend etc)
 * @param        bufferptr         initial buffer address
 * @param        size        buffer size, in bytes
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Buffer_detach(void *bufferptr, int *size) {
  return MPI_Buffer_detach(bufferptr, size); }

/* int MPI_Isend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *) */
/**
 * Begins a nonblocking send
 * @param[in]      buf         initial address of send buffer
 * @param[in]      count        number of elements in send buffer
 * @param[in]      datatype    datatype of each send buffer element
 * @param[in]      dest        rank of destination
 * @param[in]      tag          message tag
 * @param[in]      comm        communicator
 * @param[out]  request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Isend(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm, SPL_Request *request) {
  return MPI_Isend(buf, count, datatype, dest, tag, comm, request); }

/* int MPI_Ibsend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *) */
/**
 * Starts a nonblocking buffered send
 * @param[in]      buf         initial address of send buffer
 * @param[in]      count        number of elements in send buffer
 * @param[in]      datatype    datatype of each send buffer element
 * @param[in]      dest        rank of destination
 * @param[in]      tag          message tag
 * @param[in]      comm        communicator
 * @param[out]  request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Ibsend(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm, SPL_Request *request) {
  return MPI_Ibsend(buf, count, datatype, dest, tag, comm, request ); }

/* int MPI_Issend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *) */
/**
 * Starts a nonblocking synchronous send
 * @param[in]      buf         initial address of send buffer
 * @param[in]      count        number of elements in send buffer
 * @param[in]      datatype    datatype of each send buffer element
 * @param[in]      dest        rank of destination
 * @param[in]      tag          message tag
 * @param[in]      comm        communicator
 * @param[out]  request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Issend(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm, SPL_Request *request) {
  return MPI_Issend(buf, count, datatype, dest, tag, comm, request ); }

/* int MPI_Irsend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *) */
/**
 * Starts a nonblocking ready send
 * @param[in]      buf         initial address of send buffer
 * @param[in]      count        number of elements in send buffer
 * @param[in]      datatype    datatype of each send buffer element
 * @param[in]      dest        rank of destination
 * @param[in]      tag          message tag
 * @param[in]      comm        communicator
 * @param[out]  request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Irsend(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm, SPL_Request *request) {
  return MPI_Irsend(buf, count, datatype, dest, tag, comm, request); }

/* int MPI_Irecv(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *) */
/**
 * Begins a nonblocking receive
 * @param[in]      buf          initial address of receive buffer
 * @param[in]      count        maximum number of elements in receive buffer
 * @param[in]      datatype    datatype of each receive buffer element
 * @param[in]      source      rank of source
 * @param[in]      tag          message tag
 * @param[in]      comm        communicator
 * @param[out]  request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Irecv(void *buf, int count, SPL_Datatype datatype, int source, int tag, SPL_Comm comm, SPL_Request *request) {
  return MPI_Irecv(buf, count, datatype, source, tag, comm, request); }

/* int MPI_Wait(MPI_Request *, MPI_Status *) */
/**
 * Waits for an MPI send or receive to complete
 * @param[in]      request     request
 * @param[out]  status      status object
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Wait(SPL_Request *request, SPL_Status *status) {
  return MPI_Wait(request, status); }

/* int MPI_Test(MPI_Request *, int *, MPI_Status *) */
/**
 * Tests for the completion of a send or receive
 * @param[in]      request     communication request
 * @param[out]  flag         true if operation completed
 * @param[out]  status      status object
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Test(SPL_Request *request, int *flag, SPL_Status *status) {
  return MPI_Test(request, flag, status); }

/* int MPI_Request_free(MPI_Request *) */
/**
 * Frees a communication request object
 * @param[in]      request     communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Request_free(SPL_Request *request) {
  return MPI_Request_free(request); }

/* int MPI_Waitany(int, MPI_Request *, int *, MPI_Status *) */
/**
 * Waits for any specified send or receive to complete
 * @param[in]      count        list length
 * @param[in]      array_of_requests     array of requests
 * @param[out]  index      index of handle for operation that completed
 * @param[out]  status      status object
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Waitany(int count, SPL_Request array_of_requests[], int *index, SPL_Status *status) {
  return MPI_Waitany(count, array_of_requests, index, status); }

/* int MPI_Testany(int, MPI_Request *, int *, int *, MPI_Status *) */
/**
 * Tests for completion of any previdously initiated communication
 * @param[in]      count        list length
 * @param[in]      array_of_requests     array of requests
 * @param[out]  index      index of handle for operation that completed
 * @param[out]  flag          true if one of the operations is complete
 * @param[out]  status      status object
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Testany(int count, SPL_Request array_of_requests[], int *index, int *flag, SPL_Status *status) {
  return MPI_Testany(count, array_of_requests, index, flag, status); }

/* int MPI_Waitall(int, MPI_Request *, MPI_Status *) */
/**
 * Waits for all given communications to complete
 * @param[in]      count        list length
 * @param[in]      array_of_requests     array of requests
 * @param[out]  array_of_statuses      array of status objects
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Waitall(int count, SPL_Request array_of_requests[], SPL_Status array_of_statuses[]) {
  return MPI_Waitall(count, array_of_requests, array_of_statuses); }

/* int MPI_Testall(int, MPI_Request *, int *, MPI_Status *) */
/**
 * Tests for the completion of all previously initiated communications
 * @param[in]      count        list length
 * @param[in]      array_of_requests     array of requests
 * @param[in]      flag          flag is true only if all requests have completed.
 * @param[out]  array_of_statuses      array of status objects
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Testall(int count, SPL_Request array_of_requests[], int *flag, SPL_Status array_of_statuses[]) {
  return MPI_Testall(count, array_of_requests, flag, array_of_statuses); }

/* int MPI_Waitsome(int, MPI_Request *, int *, int *, MPI_Status *) */
/**
 * Waits for some given communications to complete
 * @param[in]     incount                length of array_of_requests
 * @param[in]     array_of_requests     array of requests
 * @param[out]  outcount              number of completed requests
 * @param[out]  array_of_indices       array of indices of operations that completed
 * @param[out]  array_of_statuses     array of status objects for operations that completed
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Waitsome(int incount, SPL_Request array_of_requests[], int *outcount, int array_of_indices[], SPL_Status array_of_statuses[]) {
  return MPI_Waitsome(incount, array_of_requests, outcount, array_of_indices, array_of_statuses); }

/* int MPI_Testsome(int, MPI_Request *, int *, int *, MPI_Status *) */
/**
 * Tests for some given communications to complete
 * @param[in]     incount                length of array_of_requests
 * @param[in]     array_of_requests     array of requests
 * @param[out]  outcount              number of completed requests
 * @param[out]  array_of_indices       array of indices of operations that completed
 * @param[out]  array_of_statuses     array of status objects for operations that completed
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Testsome(int incount, SPL_Request array_of_requests[], int *outcount, int array_of_indices[], SPL_Status array_of_statuses[]) {
  return MPI_Testsome(incount, array_of_requests, outcount, array_of_indices, array_of_statuses); }

/* int MPI_Iprobe(int, int, MPI_Comm, int *flag, MPI_Status *) */
/**
 * Nonblocking test for a message
 * @param[in]     source        source rank, or MPI_ANY_SOURCE
 * @param[in]     tag           tag value or MPI_ANY_TAG
 * @param[in]      comm          communicator
 * @param[out]  flag           (logical)
 * @param[out]  status        status object
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Iprobe(int source, int tag, SPL_Comm comm, int *flag, SPL_Status *status) {
  return MPI_Iprobe(source, tag, comm, flag, status); }

/* int MPI_Probe(int, int, MPI_Comm, MPI_Status *) */
/**
 * Blocking test for a message
 * @param[in]     source        source rank, or MPI_ANY_SOURCE
 * @param[in]     tag           tag value or MPI_ANY_TAG
 * @param[in]      comm          communicator
 * @param[out]  status        status object
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Probe(int source, int tag, SPL_Comm comm, SPL_Status *status) {
  return MPI_Probe(source, tag, comm, status); }

/* int MPI_Cancel(MPI_Request *) */
/**
 * Cancels a communication request
 * @param       request        communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Cancel(SPL_Request *request) {
  return MPI_Cancel(request); }

/* int MPI_Test_cancelled(MPI_Status *, int *) */
/**
 * Tests to see if a request was cancelled
 * @param       status        status object
 * @param       flag        (logical)
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Test_cancelled(SPL_Status *status, int *flag) {
  return MPI_Test_cancelled(status, flag); }

/* int MPI_Send_init(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *) */
/**
 * Builds a handle for a standard send
 * @param        buf         initial address of send buffer
 * @param        count        number of elements in send buffer
 * @param        datatype    datatype of each send buffer element
 * @param        dest        rank of destination
 * @param        tag          message tag
 * @param        comm        communicator
 * @param         request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Send_init(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm, SPL_Request *request) {
  return MPI_Send_init(buf, count, datatype, dest, tag, comm, request); }

/* int MPI_Bsend_init(void*, int, MPI_Datatype, int,int, MPI_Comm, MPI_Request *) */
/**
 * Builds a handle for a buffered send
 * @param[in]      buf         initial address of send buffer
 * @param[in]      count        number of elements in send buffer
 * @param[in]      datatype    datatype of each send buffer element
 * @param[in]      dest        rank of destination
 * @param[in]      tag          message tag
 * @param[in]      comm        communicator
 * @param[out]   request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Bsend_init(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm, SPL_Request *request) {
  return MPI_Bsend_init(buf, count, datatype, dest, tag, comm, request); }

/* int MPI_Ssend_init(void*, int, MPI_Datatype, int,int, MPI_Comm, MPI_Request *) */
/**
 * Builds a handle for a synchronous send
 * @param[in]      buf         initial address of send buffer
 * @param[in]      count        number of elements in send buffer
 * @param[in]      datatype    datatype of each send buffer element
 * @param[in]      dest        rank of destination
 * @param[in]      tag          message tag
 * @param[in]      comm        communicator
 * @param[out]   request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Ssend_init(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm, SPL_Request *request) {
  return MPI_Ssend_init(buf, count, datatype, dest, tag, comm, request); }

/* int MPI_Rsend_init(void*, int, MPI_Datatype, int,int, MPI_Comm, MPI_Request *) */
/**
 * Builds a handle for a ready send
 * @param[in]      buf         initial address of send buffer
 * @param[in]      count        number of elements in send buffer
 * @param[in]      datatype    datatype of each send buffer element
 * @param[in]      dest        rank of destination
 * @param[in]      tag          message tag
 * @param[in]      comm        communicator
 * @param[out]   request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Rsend_init(void *buf, int count, SPL_Datatype datatype, int dest, int tag, SPL_Comm comm, SPL_Request *request) {
  return MPI_Rsend_init(buf, count, datatype, dest, tag, comm, request); }

/* int MPI_Recv_init(void*, int, MPI_Datatype, int,int, MPI_Comm, MPI_Request *) */
/**
 * Builds a handle for a receive
 * @param[in]      buf         initial address of receive buffer
 * @param[in]      count       number of elements received
 * @param[in]      datatype    type of each element
 * @param[in]      source      rank of source or MPI_ANY_SOURCE
 * @param[in]      tag          message tag
 * @param[in]      comm        communicator
 * @param[out]   request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Recv_init(void *buf, int count, SPL_Datatype datatype, int source, int tag, SPL_Comm comm, SPL_Request *request) {
  return MPI_Recv_init(buf, count, datatype, source, tag, comm, request); }

/* int MPI_Start(MPI_Request *) */
/**
 * Initiates a communication with a persistent request handle
 * @param         request      communication request
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Start(SPL_Request *request) {
  return MPI_Start(request); }

/* int MPI_Startall(int, MPI_Request *) */
/**
 * Starts a collection of requestsStarts a collection of requests
 * @param         count      list length
 * @param         array_of_requests      array of requests
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Startall(int count, SPL_Request array_of_requests[]) {
  return MPI_Startall(count, array_of_requests); }

/* int MPI_Sendrecv(void *, int, MPI_Datatype,int, int, void *, int, MPI_Datatype, int, int, MPI_Comm, MPI_Status *) */
/**
 * Sends and receives a message
 * @param[in]      sendbuf         initial address of send buffer
 * @param[in]      sendcount        number of elements in send buffer
 * @param[in]      sendtype          type of elements in send buffer
 * @param[in]      dest              rank of destination
 * @param[in]      sendtag          send tag
 * @param[out]   recvbuf          initial address of receive buffer
 * @param[in]      recvcount        number of elements in receive buffer
 * @param[in]      recvtype          type of elements in receive buffer
 * @param[in]      source            rank of source (
 * @param[in]      recvtag           receive tag
 * @param[in]      comm              communicator
 * @param[in]       status            status object
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Sendrecv(void *sendbuf, int sendcount, SPL_Datatype sendtype, int dest, int sendtag, void *recvbuf, int recvcount, SPL_Datatype recvtype, int source, int recvtag, SPL_Comm comm, SPL_Status *status) {
  return MPI_Sendrecv(sendbuf, sendcount, sendtype, dest, sendtag, recvbuf, recvcount, recvtype, source, recvtag, comm, status); }

/* int MPI_Sendrecv_replace(void*, int, MPI_Datatype, int, int, int, int, MPI_Comm, MPI_Status *) */
/**
 * Sends and receives using a single buffer
 * @param[in,out]  buf         initial address of send and receive buffer
 * @param        count        number of elements in send and receive buffer
 * @param        datatype          type of elements in send and receive buffer
 * @param        dest              rank of destination
 * @param        sendtag          send message tag
 * @param        source            rank of source
 * @param        recvtag           receive message tag
 * @param        comm              communicator
 * @param         status            status object
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Sendrecv_replace(void *buf, int count, SPL_Datatype datatype, int dest, int sendtag, int source, int recvtag, SPL_Comm comm, SPL_Status *status) {
  return MPI_Sendrecv_replace(buf, count, datatype, dest, sendtag, source, recvtag, comm, status); }

/* int MPI_Type_contiguous(int, MPI_Datatype, MPI_Datatype *) */
/**
 * Creates a contiguous datatype
 * @param[in]      count           replication count
 * @param[in]      old_type           old datatype
 * @param[out]   newtype             new datatype
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_contiguous(int count, SPL_Datatype old_type, SPL_Datatype *newtype) {
  return MPI_Type_contiguous(count, old_type, newtype); }

/* int MPI_Type_vector(int, int, int, MPI_Datatype, MPI_Datatype *) */
/**
 * Creates a vector (strided) datatype
 * @param[in]      count           number of blocks
 * @param[in]        blocklen        number of elements in each block
 * @param[in]        stride          number of elements between start of each block
 * @param[in]      old_type           old datatype
 * @param[out]   newtype             new datatype
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_vector(int count, int blocklen, int stride, SPL_Datatype old_type, SPL_Datatype *newtype) {
  return MPI_Type_vector(count, blocklen, stride, old_type, newtype); }

/* int MPI_Type_hvector(int, int, MPI_Aint, MPI_Datatype, MPI_Datatype *) */
/**
 * Creates a vector (strided) datatype with offset in bytes
 * @param[in]      count           number of blocks
 * @param[in]        blocklen        number of elements in each block
 * @param[in]        stride          number of elements between start of each block
 * @param[in]      old_type           old datatype
 * @param[out]   newtype             new datatype
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_hvector(int count, int blocklen, SPL_Aint stride, SPL_Datatype old_type, SPL_Datatype *newtype) {
  return MPI_Type_hvector(count, blocklen, stride, old_type, newtype); }

/* int MPI_Type_indexed(int, int *, int *, MPI_Datatype, MPI_Datatype *) */
/**
 * Returns a new data type that represents count blocks.
 * @param[in]      count           number of blocks
 * @param[in]        blocklens        The number of instances of oldtype in each block
 * @param[in]        indices          The displacement of each block in units of extent(oldtype)
 * @param[in]      old_type           old datatype
 * @param[out]   newtype             new datatype
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_indexed(int count, int blocklens[], int indices[], SPL_Datatype old_type, SPL_Datatype *newtype) {
  return MPI_Type_indexed(count, blocklens, indices, old_type, newtype); }

/* int MPI_Type_hindexed(int, int *, MPI_Aint *, MPI_Datatype, MPI_Datatype *) */
/**
 * Returns a new data type that represents count blocks.
 * @param[in]      count           number of blocks
 * @param[in]        blocklens        The number of instances of oldtype in each block
 * @param[in]        indices          A byte displacement for each block
 * @param[in]      old_type           old datatype
 * @param[out]   newtype             new datatype
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_hindexed(int count, int blocklens[], SPL_Aint indices[], SPL_Datatype old_type, SPL_Datatype *newtype) {
  return MPI_Type_hindexed(count, blocklens, indices, old_type, newtype); }

/* int MPI_Type_struct(int, int *, MPI_Aint *, MPI_Datatype *, MPI_Datatype *) */
/**
 * Creates a struct datatype
 * @param[in]      count           number of blocks
 * @param[in]        blocklens        number of elements in each block
 * @param[in]        indices          byte displacement of each block
 * @param[in]      old_types              type of elements in each block
 * @param[out]   newtype             new datatype
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_struct(int count, int blocklens[], SPL_Aint indices[], SPL_Datatype old_types[], SPL_Datatype *newtype) {
  return MPI_Type_struct(count, blocklens, indices, old_types, newtype); }

/* int MPI_Address(void*, MPI_Aint *) */
/**
 * Gets the address of a location in memory
 * @param[in]      location           location in caller memory
 * @param[out]   address        address of location
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Address(void *location, SPL_Aint *address) {
  return MPI_Address(location, address); }

/* int MPI_Type_extent(MPI_Datatype, MPI_Aint *) */
/**
 * Returns the extent of a datatype
 * @param[in]      datatype      datatype
 * @param[out]   extent        datatype extent
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_extent(SPL_Datatype datatype, SPL_Aint *extent) {
  return MPI_Type_extent(datatype, extent); }

/* int MPI_Type_size(MPI_Datatype, int *) */
/**
 * Return the number of bytes occupied by entries in the datatype
 * @param[in]      datatype      datatype
 * @param[out]   size        datatype size
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_size(SPL_Datatype datatype, int *size) {
  return MPI_Type_size(datatype, size); }

/* int MPI_Type_lb(MPI_Datatype, MPI_Aint*) */
/**
 * Returns the lower-bound of a datatype
 * @param[in]      datatype      datatype
 * @param[out]   displacement   displacement of lower bound from origin, in bytes
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_lb(SPL_Datatype datatype, SPL_Aint *displacement) {
  return MPI_Type_lb(datatype, displacement); }

/* int MPI_Type_ub(MPI_Datatype, MPI_Aint*) */
/**
 * Returns the upper bound of a datatype
 * @param[in]      datatype      datatype
 * @param[out]   displacement   displacement of upper bound from origin, in bytes
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_ub(SPL_Datatype datatype, SPL_Aint *displacement) {
  return MPI_Type_ub(datatype, displacement); }

/* int MPI_Type_commit(MPI_Datatype *) */
/**
 * Commits the datatype
 * @param        datatype      datatype
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_commit(SPL_Datatype *datatype) {
  return MPI_Type_commit(datatype); }

/* int MPI_Type_free(MPI_Datatype *) */
/**
 * Frees the datatype
 * @param        datatype      datatype that is freed
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Type_free(SPL_Datatype *datatype) {
  return MPI_Type_free(datatype); }

/* int MPI_Get_elements(MPI_Status *, MPI_Datatype, int *) */
/**
 * Returns the number of basic elements in a datatype
 * @param[in]      status      return status of receive operation
 * @param[in]      datatype      datatype used by receive operation
 * @param[out]        elements      number of received basic elements
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Get_elements(SPL_Status *status, SPL_Datatype datatype, int *elements) {
  return MPI_Get_elements(status, datatype, elements); }

/* int MPI_Pack(void*, int, MPI_Datatype, void *, int, int *,  MPI_Comm) */
/**
 * Packs a datatype into contiguous memory
 * @param[in]      inbuf        input buffer start
 * @param[in]      incount      number of input data items
 * @param[in]      datatype      datatype of each input data item
 * @param[out]        outbuf        output buffer start
 * @param[in]      outcount      output buffer size, in bytes
 * @param[in]      position      current position in buffer, in bytes
 * @param[in]       comm            communicator for packed message
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Pack(void *inbuf, int incount, SPL_Datatype datatype, void *outbuf, int outcount, int *position, SPL_Comm comm) {
  return MPI_Pack(inbuf, incount, datatype, outbuf, outcount, position, comm); }

/* int MPI_Unpack(void*, int, int *, void *, int, MPI_Datatype, MPI_Comm) */
/**
 * Unpack a datatype into contiguous memory
 * @param[in]      inbuf        input buffer start
 * @param[in]      insize         size of input buffer, in bytes
 * @param[in]      position      current position in buffer, in bytes
 * @param[out]        outbuf        output buffer start
 * @param[in]      outcount      number of items to be unpacked
 * @param[in]      datatype      datatype of each input data item
 * @param[in]       comm            communicator for packed message
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Unpack(void *inbuf, int insize, int *position, void *outbuf, int outcount, SPL_Datatype datatype, SPL_Comm comm) {
  return MPI_Unpack(inbuf, insize, position, outbuf, outcount, datatype, comm); }

/* int MPI_Pack_size(int, MPI_Datatype, MPI_Comm, int *) */
/**
 * Returns the upper bound on the amount of space needed to pack a message
 * @param        incount          count argument to packing call
 * @param        datatype      datatype argument to packing call
 * @param         comm            communicator argument to packing call
 * @param         size            upper bound on size of packed message, in bytes
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Pack_size(int incount, SPL_Datatype datatype, SPL_Comm comm, int *size) {
  return MPI_Pack_size(incount, datatype, comm, size); }

/* int MPI_Barrier(MPI_Comm ) */
/**
 * Blocks until all process have reached this routine.
 * @param         comm          communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Barrier(SPL_Comm comm) {
  return MPI_Barrier(comm); }

/* int MPI_Bcast(void*, int, MPI_Datatype, int, MPI_Comm ) */
/**
 * Broadcasts a message from the process with rank "root" to all other processes of the group.
 * @param        buffer          starting address of buffer
 * @param        count          number of entries in buffer
 * @param        datatype      data type of buffer
 * @param        root            rank of broadcast root
 * @param         comm            communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Bcast(void *buffer, int count, SPL_Datatype datatype, int root, SPL_Comm comm) {
  return MPI_Bcast(buffer, count, datatype, root, comm); }

/* int MPI_Gather(void* , int, MPI_Datatype, void*, int, MPI_Datatype, int, MPI_Comm) */
/**
 * Gathers together values from a group of processes
 * @param[in]      sendbuf          starting address of send buffer
 * @param[in]      sendcnt          number of elements in send buffer
 * @param[in]      sendtype      data type of send buffer elements
 * @param[out]        recvbuf          address of receive buffer
 * @param[in]      recvcount        number of elements for any single receive
 * @param[in]      recvtype      data type of recv buffer elements
 * @param[in]      root            rank of receiving process
 * @param[in]       comm            communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Gather(void *sendbuf, int sendcnt, SPL_Datatype sendtype, void *recvbuf, int recvcount, SPL_Datatype recvtype, int root, SPL_Comm comm) {
  return MPI_Gather(sendbuf, sendcnt, sendtype, recvbuf, recvcount, recvtype, root, comm); }

/* int MPI_Gatherv(void* , int, MPI_Datatype, void*, int *, int *, MPI_Datatype, int, MPI_Comm) */
/**
 * Gathers into specified locations from all processes in a group
 * @param[in]      sendbuf          starting address of send buffer
 * @param[in]      sendcnt          number of elements in send buffer
 * @param[in]      sendtype      data type of send buffer elements
 * @param[out]        recvbuf          address of receive buffer
 * @param[in]      recvcnts        integer array (of length group size) containing the number of elements that are received from each process
 * @param[in]      displs        integer array (of length group size). Entry i specifies the displacement relative to recvbuf at which to place the incoming data from process i
 * @param[in]      recvtype      data type of recv buffer elements
 * @param[in]      root            rank of receiving process
 * @param[in]       comm            communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Gatherv(void *sendbuf, int sendcnt, SPL_Datatype sendtype, void *recvbuf, int *recvcnts, int *displs, SPL_Datatype recvtype, int root, SPL_Comm comm) {
  return MPI_Gatherv(sendbuf, sendcnt, sendtype, recvbuf, recvcnts, displs, recvtype, root, comm); }

/* int MPI_Scatter(void* , int, MPI_Datatype, void*, int, MPI_Datatype, int, MPI_Comm) */
/**
 * Sends data from one task to all other tasks in a group
 * @param[in]      sendbuf          starting address of send buffer
 * @param[in]      sendcnt          number of elements in send buffer
 * @param[in]      sendtype      data type of send buffer elements
 * @param[out]        recvbuf          address of receive buffer
 * @param[in]      recvcnt        number of elements for any single receive
 * @param[in]      recvtype      data type of recv buffer elements
 * @param[in]      root            rank of sending process
 * @param[in]       comm            communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Scatter(void *sendbuf, int sendcnt, SPL_Datatype sendtype, void *recvbuf, int recvcnt, SPL_Datatype recvtype, int root, SPL_Comm comm) {
  return MPI_Scatter(sendbuf, sendcnt, sendtype, recvbuf, recvcnt, recvtype, root, comm); }

/* int MPI_Scatterv(void* , int *, int *,  MPI_Datatype, void*, int, MPI_Datatype, int, MPI_Comm) */
/**
 * Scatters a buffer in parts to all tasks in a group
 * @param[in]      sendbuf          starting address of send buffer
 * @param[in]      sendcnts          integer array (of length group size) specifying the number of elements to send to each processor
 * @param[in]      displs          integer array (of length group size). Entry i specifies the displacement (relative to sendbuf from which to take the outgoing data to process i
 * @param[in]      sendtype      data type of send buffer elements
 * @param[out]        recvbuf          address of receive buffer
 * @param[in]      recvcnt        number of elements for any single receive
 * @param[in]      recvtype      data type of recv buffer elements
 * @param[in]      root               rank of sending process
 * @param[in]       comm            communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Scatterv(void *sendbuf, int *sendcnts, int *displs, SPL_Datatype sendtype, void *recvbuf, int recvcnt, SPL_Datatype recvtype, int root, SPL_Comm comm) {
  return MPI_Scatterv(sendbuf, sendcnts, displs, sendtype, recvbuf, recvcnt, recvtype, root, comm); }

/* int MPI_Allgather(void* , int, MPI_Datatype, void*, int, MPI_Datatype, MPI_Comm) */
/**
 * Gathers data from all tasks and distribute it to all
 * @param[in]      sendbuf          starting address of send buffer
 * @param[in]      sendcount         number of elements in send buffer
 * @param[in]      sendtype      data type of send buffer elements
 * @param[out]        recvbuf          address of receive buffer
 * @param[in]      recvcount        number of elements for any single receive
 * @param[in]      recvtype      data type of recv buffer elements
 * @param[in]       comm            communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Allgather(void *sendbuf, int sendcount, SPL_Datatype sendtype, void *recvbuf, int recvcount, SPL_Datatype recvtype, SPL_Comm comm) {
  return MPI_Allgather(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, comm); }

/* int MPI_Allgatherv(void* , int, MPI_Datatype, void*, int *, int *, MPI_Datatype, MPI_Comm) */
/**
 * Gathers data from all tasks and deliver it to all
 * @param[in]      sendbuf          starting address of send buffer
 * @param[in]      sendcount         number of elements in send buffer
 * @param[in]      sendtype      data type of send buffer elements
 * @param[out]        recvbuf          address of receive buffer
 * @param[in]      recvcounts        integer array (of length group size) containing the number of elements that are received from each process
 * @param[in]      displs               integer array (of length group size). Entry i specifies the displacement (relative to recvbuf ) at which to place the incoming data from process i
 * @param[in]      recvtype      data type of receive buffer elements
 * @param[in]       comm            communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Allgatherv(void *sendbuf, int sendcount, SPL_Datatype sendtype, void *recvbuf, int *recvcounts, int *displs, SPL_Datatype recvtype, SPL_Comm comm) {
  return MPI_Allgatherv(sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, comm); }

/* int MPI_Alltoall(void* , int, MPI_Datatype, void*, int, MPI_Datatype, MPI_Comm) */
/**
 * Sends data from all to all processes
 * @param[in]      sendbuf          starting address of send buffer
 * @param[in]      sendcount         number of elements in send buffer
 * @param[in]      sendtype      data type of send buffer elements
 * @param[out]        recvbuf          address of receive buffer
 * @param[in]      recvcnt        number of elements received from any process
 * @param[in]      recvtype      data type of receive buffer elements
 * @param[in]       comm            communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Alltoall(void *sendbuf, int sendcount, SPL_Datatype sendtype, void *recvbuf, int recvcnt, SPL_Datatype recvtype, SPL_Comm comm) {
  return MPI_Alltoall(sendbuf, sendcount, sendtype, recvbuf, recvcnt, recvtype, comm); }

/* int MPI_Alltoallv(void* , int *, int *, MPI_Datatype, void*, int *, int *, MPI_Datatype, MPI_Comm) */
/**
 * Sends data from all to all processes, with a displacement
 * @param[in]      sendbuf          starting address of send buffer
 * @param[in]      sendcnts         integer array equal to the group size specifying the number of elements to send to each processor
 * @param[in]      sdispls            integer array (of length group size). Entry j specifies the displacement (relative to sendbuf from which to take the outgoing data destined for process j
 * @param[in]      sendtype      data type of send buffer elements
 * @param[out]        recvbuf          address of receive buffer
 * @param[in]      recvcnts        integer array equal to the group size specifying the maximum number of elements that can be received from each processor
 * @param[in]      rdispls             integer array (of length group size). Entry i specifies the displacement (relative to recvbuf at which to place the incoming data from process i
 * @param[in]      recvtype      data type of receive buffer elements
 * @param[in]       comm            communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Alltoallv(void *sendbuf, int *sendcnts, int *sdispls, SPL_Datatype sendtype, void *recvbuf, int *recvcnts, int *rdispls, SPL_Datatype recvtype, SPL_Comm comm) {
  return MPI_Alltoallv(sendbuf, sendcnts, sdispls, sendtype, recvbuf, recvcnts, rdispls, recvtype, comm); }

/* int MPI_Reduce(void* , void*, int, MPI_Datatype, MPI_Op, int, MPI_Comm) */
/**
 * Reduces values on all processes to a single value
 * @param[in]      sendbuf          address of send buffer
 * @param[out]   recvbuf          address of receive buffer
 * @param[in]      count              number of elements in send buffer
 * @param[in]      datatype        data type of elements of send buffer
 * @param[in]      op             reduce operation
 * @param[in]      root            rank of root process
 * @param[in]       comm            communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Reduce(void *sendbuf, void *recvbuf, int count, SPL_Datatype datatype, SPL_Op op, int root, SPL_Comm comm) {
  return MPI_Reduce(sendbuf, recvbuf, count, datatype, op, root, comm); }

/* int MPI_Op_create(MPI_User_function *, int, MPI_Op *) */
/**
 * Creates a user-defined combination function handle
 * @param[in]      function          user defined function
 * @param[in]      commute          true if commutative; false otherwise.
 * @param[out]   op               operation
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Op_create(SPL_User_function *function, int commute, SPL_Op *op) {
  return MPI_Op_create(function, commute, op); }

/* int MPI_Op_free( MPI_Op *) */
/**
 * Frees a user-defined combination function handle
 * @param[in]       op               operation
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Op_free(SPL_Op *op) {
  return MPI_Op_free(op); }

/* int MPI_Allreduce(void* , void*, int, MPI_Datatype, MPI_Op, MPI_Comm) */
/**
 * Combines values from all processes and distribute the result back to all processes
 * @param[in]       sendbuf             starting address of send buffer
 * @param[out]         recvbuf            starting address of receive buffer
 * @param[in]       count               number of elements in send buffer
 * @param[in]       datatype            data type of elements of send buffer
 * @param[in]       op               operation
 * @param[in]       comm               communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Allreduce(void *sendbuf, void *recvbuf, int count, SPL_Datatype datatype, SPL_Op op, SPL_Comm comm) {
  return MPI_Allreduce(sendbuf, recvbuf, count, datatype, op, comm); }

/* int MPI_Reduce_scatter(void* , void*, int *, MPI_Datatype, MPI_Op, MPI_Comm) */
/**
 * Combines values and scatters the results
 * @param[in]       sendbuf             starting address of send buffer
 * @param[out]         recvbuf            starting address of receive buffer
 * @param[in]       recvcnts             integer array specifying the number of elements in result distributed to each process. Array must be identical on all calling processes.
 * @param[in]       datatype            data type of elements of input buffer
 * @param[in]       op               operation
 * @param[in]       comm               communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Reduce_scatter(void *sendbuf, void *recvbuf, int *recvcnts, SPL_Datatype datatype, SPL_Op op, SPL_Comm comm) {
  return MPI_Reduce_scatter(sendbuf, recvbuf, recvcnts, datatype, op, comm); }

/* int MPI_Scan(void* , void*, int, MPI_Datatype, MPI_Op, MPI_Comm ) */
/**
 * Computes the scan (partial reductions) of data on a collection of processes
 * @param[in]       sendbuf             starting address of send buffer
 * @param[out]         recvbuf            starting address of receive buffer
 * @param[in]       count             Computes the scan (partial reductions) of data on a collection of processes
 * @param[in]       datatype            data type of elements of input buffer
 * @param[in]       op               operation
 * @param[in]       comm               communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Scan(void *sendbuf, void *recvbuf, int count, SPL_Datatype datatype, SPL_Op op, SPL_Comm comm) {
  return MPI_Scan(sendbuf, recvbuf, count, datatype, op, comm); }

/* int MPI_Group_size(MPI_Group group, int *) */
/**
 * Returns the size of a group
 * @param[in]       group             group
 * @param[out]    size            number of processes in the group
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_size(SPL_Group group, int *size) {
  return MPI_Group_size(group, size); }

/* int MPI_Group_rank(MPI_Group group, int *) */
/**
 * Returns the rank of this process in the given group
 * @param[in]       group             group
 * @param[out]    rank            rank of the calling process in group, or MPI_UNDEFINED if the process is not a member
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_rank(SPL_Group group, int *rank) {
  return MPI_Group_rank(group, rank); }

/* int MPI_Group_translate_ranks (MPI_Group, int, int *, MPI_Group, int *) */
/**
 * Translates the ranks of processes in one group to those in another group
 * @param[in]       group_a             group1
 * @param[in]       n                      number of ranks in ranks1 and ranks2 arrays
 * @param[in]        ranks_a            array of zero or more valid ranks in group1
 * @param[in]        group_b            group2
 * @param[out]    ranks_b            array of corresponding ranks in group2, MPI_UNDEFINED when no correspondence exists.
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_translate_ranks(SPL_Group group_a, int n, int *ranks_a, SPL_Group group_b, int *ranks_b) {
  return MPI_Group_translate_ranks(group_a, n, ranks_a, group_b, ranks_b); }

/* int MPI_Group_compare(MPI_Group, MPI_Group, int *) */
/**
 * Compares two groups
 * @param[in]       group1             group1
 * @param[in]        group2            group2
 * @param[out]    result            integer which is MPI_IDENT if the order and members of the two groups are the same, MPI_SIMILAR if only the members are the same, and MPI_UNEQUAL otherwise
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_compare(SPL_Group group1, SPL_Group group2, int *result) {
  return MPI_Group_compare(group1, group2, result); }

/* int MPI_Comm_group(MPI_Comm, MPI_Group *) */
/**
 * Accesses the group associated with given communicator
 * @param[in]       comm             Communicator
 * @param[out]          group          Group in communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_group(SPL_Comm comm, SPL_Group *group) {
  return MPI_Comm_group(comm, group); }

/* int MPI_Group_union(MPI_Group, MPI_Group, MPI_Group *) */
/**
 * Produces a group by combining two groups
 * @param[in]       group1             group1
 * @param[in]        group2          group2
 * @param[out]    group_out          union group
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_union(SPL_Group group1, SPL_Group group2, SPL_Group *group_out) {
  return MPI_Group_union(group1, group2, group_out); }

/* int MPI_Group_intersection(MPI_Group, MPI_Group, MPI_Group *) */
/**
 * Produces a group as the intersection of two existing groups
 * @param[in]       group1                first group
 * @param[in]        group2          second group
 * @param[out]    group_out          intersection group
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_intersection(SPL_Group group1, SPL_Group group2, SPL_Group *group_out) {
  return MPI_Group_intersection(group1, group2, group_out); }

/* int MPI_Group_difference(MPI_Group, MPI_Group, MPI_Group *) */
/**
 * Makes a group from the difference of two groups
 * @param[in]       group1                first group
 * @param[in]        group2          second group
 * @param[out]    group_out          difference group
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_difference(SPL_Group group1, SPL_Group group2, SPL_Group *group_out) {
  return MPI_Group_difference(group1, group2, group_out); }

/* int MPI_Group_incl(MPI_Group group, int, int *, MPI_Group *) */
/**
 * Produces a group by reordering an existing group and taking only listed members
 * @param[in]       group                group
 * @param[in]        n          number of elements in array ranks (and size of newgroup )
 * @param[in]        ranks             ranks of processes in group to appear in newgroup
 * @param[out]    group_out          new group derived from above, in the order defined by ranks
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_incl(SPL_Group group, int n, int *ranks, SPL_Group *group_out) {
  return MPI_Group_incl(group, n, ranks, group_out); }

/* int MPI_Group_excl(MPI_Group group, int, int *, MPI_Group *) */
/**
 * Produces a group by reordering an existing group and taking only unlisted members
 * @param[in]       group                group
 * @param[in]        n                number of elements in array ranks
 * @param[in]        ranks             array of integer ranks in group not to appear in newgroup
 * @param[out]    newgroup          new group derived from above, preserving the order defined by group
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_excl(SPL_Group group, int n, int *ranks, SPL_Group *newgroup) {
  return MPI_Group_excl(group, n, ranks, newgroup); }

/* int MPI_Group_range_incl(MPI_Group group, int, int [][3], MPI_Group *) */
/**
 * Creates a new group from ranges of ranks in an existing group
 * @param[in]       group                group
 * @param[in]        n                number of triplets in array ranges
 * @param[in]        ranges             a one-dimensional array of integer triplets, of the form (first rank, last rank, stride) indicating ranks in group or processes to be included in newgroup
 * @param[out]    newgroup          new group derived from above, in the order defined by ranges
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_range_incl(SPL_Group group, int n, int ranges[][3], SPL_Group *newgroup) {
  return MPI_Group_range_incl(group, n, ranges, newgroup); }

/* int MPI_Group_range_excl(MPI_Group group, int, int [][3], MPI_Group *) */
/**
 * Produces a group by excluding ranges of processes from an existing group
 * @param[in]       group                group
 * @param[in]        n                number of elements in array ranks
 * @param[in]        ranges             a one-dimensional array of integer triplets of the form (first rank, last rank, stride), indicating the ranks in group of processes to be excluded from the output group newgroup .
 * @param[out]    newgroup          new group derived from above, preserving the order in group
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_range_excl(SPL_Group group, int n, int ranges[][3], SPL_Group *newgroup) {
  return MPI_Group_range_excl(group, n, ranges, newgroup); }

/* int MPI_Group_free(MPI_Group *) */
/**
 * Frees a group
 * @param[in]       group                group
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Group_free(SPL_Group *group) {
  return MPI_Group_free(group); }

/* int MPI_Comm_size(MPI_Comm, int *) */
/**
 * Determines the size of the group associated with a communictor
 * @param[in]       comm              communicator
 * @param[out]    size          number of processes in the group of comm
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_size(SPL_Comm comm, int *size) {
  return MPI_Comm_size(comm, size); }

/* int MPI_Comm_rank(MPI_Comm, int *) */
/**
 * Determines the rank of the calling process in the communicator
 * @param[in]       comm              communicator
 * @param[out]    rank          rank of the calling process in group of comm
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_rank(SPL_Comm comm, int *rank) {
  return MPI_Comm_rank(comm, rank); }

/* int MPI_Comm_compare(MPI_Comm, MPI_Comm, int *) */
/**
 * Compares two communicators
 * @param[in]       comm1              comm1
 * @param[in]       comm2              comm2
 * @param[out]    result          integer which is MPI_IDENT if the contexts and groups are the same
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_compare(SPL_Comm comm1, SPL_Comm comm2, int *result) {
  return MPI_Comm_compare(comm1, comm2, result); }

/* int MPI_Comm_dup(MPI_Comm, MPI_Comm *) */
/**
 * Duplicates an existing communicator with all its cached information
 * @param[in]       comm              communicator
 * @param[out]    comm_out          A new communicator over the same group as comm but with a new context
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_dup(SPL_Comm comm, SPL_Comm *comm_out) {
  return MPI_Comm_dup(comm, comm_out); }

/* int MPI_Comm_create(MPI_Comm, MPI_Group, MPI_Comm *) */
/**
 * Creates a new communicator
 * @param[in]       comm              communicator
 * @param[in]       group             group, which is a subset of the group of comm
 * @param[out]    comm_out        new communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_create(SPL_Comm comm, SPL_Group group, SPL_Comm *comm_out) {
  return MPI_Comm_create(comm, group, comm_out); }

/* int MPI_Comm_split(MPI_Comm, int, int, MPI_Comm *) */
/**
 * Creates a new communicator
 * @param[in]       comm              communicator
 * @param[in]       color           control of subset assignment (nonnegative integer). Processes with the same color are in the same new communicator
 * @param[in]       key             control of rank assigment
 * @param[out]    comm_out        new communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_split(SPL_Comm comm, int color, int key, SPL_Comm *comm_out) {
  return MPI_Comm_split(comm, color, key, comm_out); }

/* int MPI_Comm_free(MPI_Comm *) */
/**
 * Marks the communicator object for deallocation
 * @param         commp              communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_free(SPL_Comm *commp) {
  return MPI_Comm_free(commp); }

/* int MPI_Comm_test_inter(MPI_Comm, int *) */
/**
 * Tests to see if a comm is an inter-communicator
 * @param[in]       comm              communicator
 * @param[out]         flag              (logical)
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_test_inter(SPL_Comm comm, int *flag) {
  return MPI_Comm_test_inter(comm, flag); }

/* int MPI_Comm_remote_size(MPI_Comm, int *) */
/**
 * Determines the size of the remote group associated with an inter-communictor
 * @param[in]       comm              communicator
 * @param[out]         size        number of processes in the group of comm
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_remote_size(SPL_Comm comm, int *size) {
  return MPI_Comm_remote_size(comm, size); }

/* int MPI_Comm_remote_group(MPI_Comm, MPI_Group *) */
/**
 * Accesses the remote group associated with the given inter-communicator
 * @param[in]       comm              communicator
 * @param[out]    group        remote group of communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_remote_group(SPL_Comm comm, SPL_Group *group) {
  return MPI_Comm_remote_group(comm, group); }

/* int MPI_Intercomm_create(MPI_Comm, int, MPI_Comm, int, int, MPI_Comm * ) */
/**
 * Creates an intercommuncator from two intracommunicators
 * @param[in]       local_comm              Local (intra)communicator
 * @param[in]        local_leader        Rank in local_comm of leader (often 0)
 * @param[in]        peer_comm        Remote communicator
 * @param[in]        remote_leader     Rank in peer_comm of remote leader (often 0)
 * @param[in]        tag                Message tag to use in constructing intercommunicator
 * @param[out]    comm_out        Created intercommunicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Intercomm_create (SPL_Comm local_comm, int local_leader, SPL_Comm peer_comm, int remote_leader, int tag, SPL_Comm *comm_out) {
  return MPI_Intercomm_create (local_comm, local_leader, peer_comm, remote_leader, tag, comm_out); }

/* int MPI_Intercomm_merge(MPI_Comm, int, MPI_Comm *) */
/**
 * Creates an intracommuncator from an intercommunicator
 * @param[in]       comm              Intercommunicator
 * @param[in]        high        Used to order the groups of the two intracommunicators within comm when creating the new communicator.
 * @param[out]    comm_out        Created intercommunicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Intercomm_merge(SPL_Comm comm, int high, SPL_Comm *comm_out) {
  return MPI_Intercomm_merge(comm, high, comm_out); }

/* int MPI_Keyval_create(MPI_Copy_function *, MPI_Delete_function *, int *, void*) */
/**
 * Generates a new attribute key
 * @param[in]       copy_fn              Copy callback function for keyval
 * @param[in]        delete_fn        Delete callback function for keyval
 * @param[out]    keyval        key value for future access
 * @param[in]      extra_state       Extra state for callback functions
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Keyval_create(SPL_Copy_function *copy_fn, SPL_Delete_function *delete_fn, int *keyval, void *extra_state) {
  return MPI_Keyval_create(copy_fn, delete_fn, keyval, extra_state); }

/* int MPI_Keyval_free(int *) */
/**
 * Frees attribute key for communicator cache attribute
 * @param    keyval        Frees the integer key value
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Keyval_free(int *keyval) {
  return MPI_Keyval_free(keyval); }

/* int MPI_Attr_put(MPI_Comm, int, void*) */
/**
 * Stores attribute value associated with a key
 * @param    comm           communicator to which attribute will be attached
 * @param    keyval           key value, as returned by MPI_KEYVAL_CREATE
 * @param    attr_value        attribute value
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Attr_put(SPL_Comm comm, int keyval, void *attr_value) {
  return MPI_Attr_put(comm, keyval, attr_value); }

/* int MPI_Attr_get(MPI_Comm, int, void *, int *) */
/**
 * Retrieves attribute value by key
 * @param[in]  comm           communicator to which attribute is attached
 * @param[in]  keyval           key value
 * @param[out]    attr_value          attribute value, unless flag = false
 * @param[out]    flag        true if an attribute value was extracted; false if no attribute is associated with the key
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Attr_get(SPL_Comm comm, int keyval, void *attr_value, int *flag) {
  return MPI_Attr_get(comm, keyval, attr_value, flag); }

/* int MPI_Attr_delete(MPI_Comm, int) */
/**
 * Deletes attribute value associated with a key
 * @param    comm           communicator to which attribute is attached
 * @param    keyval         The key value of the deleted attribute
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Attr_delete(SPL_Comm comm, int keyval) {
  return MPI_Attr_delete(comm, keyval); }

/* int MPI_Topo_test(MPI_Comm, int *) */
/**
 * Determines the type of topology (if any) associated with a communicator
 * @param    comm           communicator
 * @param    top_type         topology type of communicator comm
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Topo_test(SPL_Comm comm, int *top_type) {
  return MPI_Topo_test(comm, top_type); }

/* int MPI_Cart_create(MPI_Comm, int, int *, int *, int, MPI_Comm *) */
/**
 * Makes a new communicator to which topology information has been attached
 * @param[in]  comm_old         input communicator
 * @param[in]  ndims         number of dimensions of cartesian grid
 * @param[in]  dims         integer array of size ndims specifying the number of processes in each dimension
 * @param[in]  periods         logical array of size ndims specifying whether the grid is periodic (true) or not (false) in each dimension
 * @param[in]  reorder            ranking may be reordered (true) or not (false)
 * @param[out]    comm_cart         communicator with new cartesian topology
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Cart_create(SPL_Comm comm_old, int ndims, int *dims, int *periods, int reorder, SPL_Comm *comm_cart) {
  return MPI_Cart_create(comm_old, ndims, dims, periods, reorder, comm_cart); }

/* int MPI_Dims_create(int, int, int *) */
/**
 * Creates a division of processors in a cartesian grid
 * @param[in]  nnodes            number of nodes in a grid
 * @param[in]  ndims            number of cartesian dimensions
 * @param[out]    dims         integer array of size ndims specifying the number of nodes in each dimension
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Dims_create(int nnodes, int ndims, int *dims) {
  return MPI_Dims_create(nnodes, ndims, dims); }

/* int MPI_Graph_create(MPI_Comm, int, int *, int *, int, MPI_Comm *) */
/**
 * Makes a new communicator to which topology information has been attached
 * @param[in]  comm_old            input communicator without topology
 * @param[in]  nnodes            number of nodes in graph
 * @param[in]  index         number of nodes in graph
 * @param[in]  edges         array of integers describing graph edges
 * @param[in]  reorder         ranking may be reordered (true) or not (false)
 * @param[out]    comm_graph         communicator with graph topology added
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Graph_create(SPL_Comm comm_old, int nnodes, int *index, int *edges, int reorder, SPL_Comm *comm_graph) {
  return MPI_Graph_create(comm_old, nnodes, index, edges, reorder, comm_graph); }

/* int MPI_Graphdims_get(MPI_Comm, int *, int *) */
/**
 * Retrieves graph topology information associated with a communicator
 * @param[in]  comm            communicator for group with graph structure
 * @param[out]    nnodes             number of nodes in graph
 * @param[out]    nedges         number of edges in graph
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Graphdims_get(SPL_Comm comm, int *nnodes, int *nedges) {
  return MPI_Graphdims_get(comm, nnodes, nedges); }

/* int MPI_Graph_get(MPI_Comm, int, int, int *, int *) */
/**
 * Retrieves graph topology information associated with a communicator
 * @param[in]        comm            communicator with graph structure
 * @param[in]        maxindex            length of vector index in the calling program
 * @param[in]        maxedges         length of vector edges in the calling program
 * @param[out]    index         array of integers containing the graph structure
 * @param[out]    edges         array of integers containing the graph structure
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Graph_get(SPL_Comm comm, int maxindex, int maxedges, int *index, int *edges) {
  return MPI_Graph_get(comm, maxindex, maxedges, index, edges); }

/* int MPI_Cartdim_get(MPI_Comm, int *) */
/**
 * Retrieves Cartesian topology information associated with a communicator
 * @param[in]        comm            communicator with cartesian structure
 * @param[out]    ndims         number of dimensions of the cartesian structure
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Cartdim_get(SPL_Comm comm, int *ndims) {
  return MPI_Cartdim_get(comm, ndims); }

/* int MPI_Cart_get(MPI_Comm, int, int *, int *, int *) */
/**
 * Retrieves Cartesian topology information associated with a communicator
 * @param[in]        comm            communicator with cartesian structure
 * @param[in]        maxdims         length of vectors dims, periods, and coords in the calling program
 * @param[out]    dims         number of processes for each cartesian dimension
 * @param[out]    periods        periodicity (true/false) for each cartesian dimension
 * @param[out]    coords         coordinates of calling process in cartesian structure
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Cart_get(SPL_Comm comm, int maxdims, int *dims, int *periods, int *coords) {
  return MPI_Cart_get(comm, maxdims, dims, periods, coords); }

/* int MPI_Cart_rank(MPI_Comm, int *, int *) */
/**
 * Determines process rank in communicator given Cartesian location
 * @param[in]        comm            communicator with cartesian structure
 * @param[in]        coords         integer array (of size ndims) specifying the cartesian coordinates of a process
 * @param[out]    rank         rank of specified process
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Cart_rank(SPL_Comm comm, int *coords, int *rank) {
  return MPI_Cart_rank(comm, coords, rank); }

/* int MPI_Cart_coords(MPI_Comm, int, int, int *) */
/**
 * Determines process coords in cartesian topology given rank in group
 * @param[in]        comm            communicator with cartesian structure
 * @param[in]        rank            rank of a process within group of comm
 * @param[in]        maxdims            length of vector coords in the calling program
 * @param[out]    coords         integer array (of size ndims) containing the Cartesian coordinates of specified process
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Cart_coords(SPL_Comm comm, int rank, int maxdims, int *coords) {
  return MPI_Cart_coords(comm, rank, maxdims, coords); }

/* int MPI_Graph_neighbors_count(MPI_Comm, int, int *) */
/**
 * Returns the number of neighbors of a node associated with a graph topology
 * @param[in]        comm            communicator with graph topology
 * @param[in]        rank            rank of process in group of comm
 * @param[out]    nneighbors         number of neighbors of specified process
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Graph_neighbors_count(SPL_Comm comm, int rank, int *nneighbors) {
  return MPI_Graph_neighbors_count(comm, rank, nneighbors); }

/* int MPI_Graph_neighbors(MPI_Comm, int, int, int *) */
/**
 * Returns the neighbors of a node associated with a graph topology
 * @param[in]        comm            communicator with graph topology
 * @param[in]        rank            rank of process in group of comm
 * @param[in]        maxneighbors        size of array neighbors
 * @param[out]    neighbors        ranks of processes that are neighbors to specified process
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Graph_neighbors(SPL_Comm comm, int rank, int maxneighbors, int *neighbors) {
  return MPI_Graph_neighbors(comm, rank, maxneighbors, neighbors); }

/* int MPI_Cart_shift(MPI_Comm, int, int, int *, int *) */
/**
 * Returns the shifted source and destination ranks, given a shift direction and amount
 * @param[in]        comm            communicator with cartesian structure
 * @param[in]        direction            coordinate dimension of shift
 * @param[in]        displ        displacement (> 0: upwards shift, < 0: downwards shift)
 * @param[out]    source        rank of source process
 * @param[out]    dest        rank of destination process
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Cart_shift(SPL_Comm comm, int direction, int displ, int *source, int *dest) {
  return MPI_Cart_shift(comm, direction, displ, source, dest); }

/* int MPI_Cart_sub(MPI_Comm, int *, MPI_Comm *) */
/**
 * Partitions a communicator into subgroups which form lower-dimensional cartesian subgrids
 * @param[in]        comm            communicator with cartesian structure
 * @param[in]        remain_dims        the ith entry of remain_dims specifies whether the ith dimension is kept in the subgrid (true) or is dropped (false)
 * @param[out]          comm_new        communicator containing the subgrid that includes the calling process
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Cart_sub(SPL_Comm comm, int *remain_dims, SPL_Comm *comm_new) {
  return MPI_Cart_sub(comm, remain_dims, comm_new); }

/* int MPI_Cart_map(MPI_Comm, int, int *, int *, int *) */
/**
 * Maps process to Cartesian topology information
 * @param[in]        comm_old            input communicator
 * @param[in]        ndims           number of dimensions of Cartesian structure
 * @param[in]        dims        integer array of size ndims specifying the number of processes in each coordinate direction
 * @param[in]        periods     logical array of size ndims specifying the periodicity specification in each coordinate direction
 * @param[out]          newrank     reordered rank of the calling process; MPI_UNDEFINED if calling process does not belong to grid
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Cart_map(SPL_Comm comm_old, int ndims, int *dims, int *periods, int *newrank) {
  return MPI_Cart_map(comm_old, ndims, dims, periods, newrank); }

/* int MPI_Graph_map(MPI_Comm, int, int *, int *, int *) */
/**
 * Maps process to graph topology information
 * @param[in]        comm_old            input communicator
 * @param[in]        nnodes           number of graph nodes
 * @param[in]        index           integer array specifying the graph structure
 * @param[in]        edges            integer array specifying the graph structure
 * @param[out]    newrank     reordered rank of the calling process; MPI_UNDEFINED if the calling process does not belong to graph
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Graph_map(SPL_Comm comm_old, int nnodes, int *index, int *edges, int *newrank) {
  return MPI_Graph_map(comm_old, nnodes, index, edges, newrank); }

/* int MPI_Get_processor_name(char *, int *) */
/**
 * Gets the name of the processor
 * @param[out]    name            A unique specifier for the actual (as opposed to virtual) node
 * @param[out]    resultlen        Length
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Get_processor_name(char *name, int *resultlen) {
  return MPI_Get_processor_name(name, resultlen); }

/* int MPI_Get_version(int *, int *) */
/**
 * Gets the version of MPI
 * @param[out]    version            Major version of MPI (1 or 2)
 * @param[out]    subversion          Minor version of MPI.
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Get_version(int *version, int *subversion) {
  return MPI_Get_version(version, subversion); }

/* int MPI_Errhandler_create(MPI_Handler_function *, MPI_Errhandler *) */
/**
 * Creates an MPI-style errorhandler
 * @param[in]        function            user defined error handling procedure
 * @param[out]    errhandler          MPI error handler
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Errhandler_create(SPL_Handler_function *function, SPL_Errhandler *errhandler) {
  return MPI_Errhandler_create(function, errhandler); }

/* int MPI_Errhandler_set(MPI_Comm, MPI_Errhandler) */
/**
 * Sets the error handler for a communicator
 * @param          comm            communicator to set the error handler for
 * @param          errhandler      new MPI error handler for communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Errhandler_set(SPL_Comm comm, SPL_Errhandler errhandler) {
  return MPI_Errhandler_set(comm, errhandler); }

/* int MPI_Errhandler_get(MPI_Comm, MPI_Errhandler *) */
/**
 * Gets the error handler for a communicator
 * @param[in]        comm            communicator to get the error handler from
 * @param[out]          errhandler      MPI error handler currently associated with communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Errhandler_get(SPL_Comm comm, SPL_Errhandler *errhandler) {
  return MPI_Errhandler_get(comm, errhandler); }

/* int MPI_Errhandler_free(MPI_Errhandler *) */
/**
 * Frees an MPI-style errorhandler
 * @param          errhandler      MPI error handler
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Errhandler_free(SPL_Errhandler *errhandler) {
  return MPI_Errhandler_free(errhandler); }

/* int MPI_Error_string(int, char *, int *) */
/**
 * Return a string for a given error code
 * @param[in]        errorcode      Error code returned by an MPI routine or an MPI error class
 * @param[out]          string      Text that corresponds to the errorcode
 * @param[out]          resultlen      Length of string
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Error_string(int errorcode, char *string, int *resultlen) {
  return MPI_Error_string(errorcode, string, resultlen); }

/* int MPI_Error_class(int, int *) */
/**
 * Converts an error code into an error class
 * @param[in]        errorcode      Error code returned by an MPI routine
 * @param[out]     errorclass      Error class associated with errorcode
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Error_class(int errorcode, int *errorclass) {
  return MPI_Error_class(errorcode, errorclass); }

/* double MPI_Wtime(void) */
/**
 * Returns an elapsed time on the calling processor
 * @return      Time in seconds since an arbitrary time in the past.
 */
SPL_MPI_INLINE double
SPL_Wtime() {
  return MPI_Wtime(); }

/* double MPI_Wtick(void) */
/**
 * Returns the resolution of MPI_Wtime
 * @return      Time in seconds of resolution of MPI_Wtime
 */
SPL_MPI_INLINE double
SPL_Wtick() {
  return MPI_Wtick(); }

/* int MPI_Init(int *, char ***) */
/**
 * Initialize the MPI execution environment
 * @param          argc      Pointer to the number of arguments
 * @param          argv      Pointer to the argument vector
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Init(int *argc, char ***argv) {
  return MPI_Init(argc, argv); }

/* int MPI_Init_thread( int *, char ***, int, int * ) */
/**
 * Initialize the MPI execution environment
 * @param[in]        argc      Pointer to the number of arguments
 * @param[in]        argv      Pointer to the argument vector
 * @param[in]        required      Level of desired thread support
 * @param[out]          provided      Level of provided thread support
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Init_thread(int *argc, char ***argv, int required, int *provided) {
  return MPI_Init_thread(argc, argv, required, provided); }

/* int MPI_Finalize(void) */
/**
 * Terminates MPI execution environment
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Finalize() {
  return MPI_Finalize(); }

/* int MPI_Initialized(int *) */
/**
 * Indicates whether MPI_Init has been called.
 * @param          flag      Flag is true if MPI_Init has been called and false otherwise.
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Initialized(int *flag) {
  return MPI_Initialized(flag); }

/* int MPI_Abort(MPI_Comm, int) */
/**
 * Terminates MPI execution environment
 * @param          comm              communicator of tasks to abort
 * @param          errorcode            error code to return to invoking environment
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Abort(SPL_Comm comm, int errorcode) {
  return MPI_Abort(comm, errorcode); }

/* MPI-2 communicator naming functions */
/* int MPI_Comm_set_name(MPI_Comm, char *) */
/**
 * give a print name to the communicator
 * @param          comm            Communicator to name
 * @param          name              Name for communicator
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_set_name(SPL_Comm comm, char *name) {
  return MPI_Comm_set_name(comm, name); }

/* int MPI_Comm_get_name(MPI_Comm, char *, int *) */
/**
 * return the print name from the communicator
 * @param[in]        comm            Communicator to get name of
 * @param[out]          namep              One output, contains the name of the communicator. It must be an array of size at least MPI_MAX_NAME_STRING
 * @param[out]          reslen              Number of characters in name
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Comm_get_name(SPL_Comm comm, char *namep, int *reslen) {
  return MPI_Comm_get_name(comm, namep, reslen); }

/* int MPI_Pcontrol(int) */
/**
 * Controls profiling
 * @param          level            Profiling level
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Pcontrol(int level) {
  return MPI_Pcontrol(level); }


/* int MPI_NULL_COPY_FN ( MPI_Comm, int, void *, void *, void *, int * ) */
/* int MPI_NULL_DELETE_FN ( MPI_Comm, int, void *, void * ) */
/* int MPI_DUP_FN ( MPI_Comm, int, void *, void *, void *, int * ) */

/* misc2 (MPI2) */
/* int MPI_Status_f2c( MPI_Fint *, MPI_Status * ) */
/* int MPI_Status_c2f( MPI_Status *, MPI_Fint * ) */
/* int MPI_Finalized( int * ) */
/* int MPI_Type_create_indexed_block(int, int, int *, MPI_Datatype, MPI_Datatype *) */
/* int MPI_Type_get_envelope(MPI_Datatype, int *, int *, int *, int *) */
/* int MPI_Type_get_contents(MPI_Datatype, int, int, int, int *, MPI_Aint *, MPI_Datatype *) */
/* int MPI_Type_create_subarray(int, int *, int *, int *, int, MPI_Datatype, MPI_Datatype *) */
/* int MPI_Type_create_darray(int, int, int, int *, int *, int *, int *, int, MPI_Datatype, MPI_Datatype *) */
/* int MPI_Info_create(MPI_Info *) */
/* int MPI_Info_set(MPI_Info, char *, char *) */
/* int MPI_Info_delete(MPI_Info, char *) */
/* int MPI_Info_get(MPI_Info, char *, int, char *, int *) */
/* int MPI_Info_get_valuelen(MPI_Info, char *, int *, int *) */
/* int MPI_Info_get_nkeys(MPI_Info, int *) */
/* int MPI_Info_get_nthkey(MPI_Info, int, char *) */
/* int MPI_Info_dup(MPI_Info, MPI_Info *) */
/* int MPI_Info_free(MPI_Info *info) */

/* MPI_Fint MPI_Info_c2f(MPI_Info) */
/* MPI_Info MPI_Info_f2c(MPI_Fint) */
/* MPI_Fint MPI_Request_c2f( MPI_Request ) */

/* external */
/* int MPI_Status_set_cancelled( MPI_Status *, int ) */
/**
 * Set the opaque part of an MPI_Status so that MPI_Test_cancelled will return flag
 * @param          status            Status to associate count with (Status)
 * @param          flag               if true indicates that request was cancelled
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Status_set_cancelled(SPL_Status *status, int flag) {
  return MPI_Status_set_cancelled(status, flag); }

/* int MPI_Status_set_elements( MPI_Status *, MPI_Datatype, int ) */
/**
 * Set the opaque part of an MPI_Status so that MPI_Get_elements will return count.
 * @param          status            Status to associate count with (Status)
 * @param          datatype            datatype associated with count
 * @param          count               number of elements to associate with status
 * @return      Error Code (no error=SPL_SUCCESS)
 */
SPL_MPI_INLINE int
SPL_Status_set_elements(SPL_Status *status, SPL_Datatype datatype, int count) {
  return MPI_Status_set_elements(status, datatype, count); }

#endif /* _SPL_MPI_MPI_INLINE_H_ */

