/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
 /**
 * \~english
 * @file sphApp.C
 * Sphere application Class Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphApp.C
 * Sphereアプリケーションクラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */

#include <errno.h>
#include "app/sphApp.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphApp::sphApp():m_sphere(NULL)
{

}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphApp::~sphApp()
{
  // sphSphereオブジェクトの破棄
  sphSphere::destruct();
}


/**
 * usageフラグを設定する。
 * @param       flag    usage flag
 * @return      void
 */
void sphApp::setUsageFlag(bool flag)
{
  m_usageFlag = flag;
}


/**
 * p4フラグを設定する。
 * @param       flag    usage flag
 * @return      void
 */
void sphApp::setP4Flag(bool flag)
{
  m_p4Flag = flag;
}


/**
 * コンフィグレーションファイル名を設定する。
 * @param       config_fname    コンフィグレーションファイル名
 * @return      void
 */
void sphApp::setConfigFile(const char* config_fname)
{
  m_config_fname = config_fname;
}


/**
 * シグナルハンドラー
 * @param       signal_no     シグナル番号
 * @return      void
 */
void sphApp::terminateSignalHandler(int signal_no)
{
  std::cout << "[sphApp] signal handler signal_no=" << signal_no << std::endl;

  // error handler
  sphErrorHandler* errHandler = sphErrorHandler::getInstance();
  errHandler->SPH_ERROR_INFO(SPH_ERR_TERMINATE_SIGNAL, "signal handler signal_no=%d, errno=%d", signal_no, errno);

  int size = 1;
  if (splIsParallel() == SplTrue) {
    SPL_Comm_size(SPL_COMM_WORLD, &size);
    if (size > 1) {
      errHandler->printStackTrace();
      SPL_Abort(SPL_COMM_WORLD, signal_no);
    }
  }

  if (size <= 1) {
    m_sphereExitFlag = true;

    errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, signal_no);
    // プログラムを終了する。
    exit(-1);
  }

}


/**
 * Sphereを実行する。
 * @param       void
 * @return      void
 */
bool sphApp::run(int argc, char** argv)
{

  int retFlg = 0;
  int tmp = 0;
  // 初期化
  if (!initialize()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initialize fault.");
    retFlg = 1;
  }

  // 実行パスを取得する。(rankid=0)
  char abs_path[MAXPATHLEN];
  const char cur_path[MAXPATHLEN] = ".";
  realpath(cur_path, abs_path);
  char buf[MAXPATHLEN];
  // getcwd(buf,sizeof(buf));
  // printf("カレントディレクトリ=%s\n", buf);

  // 並列実行初期化
  if (!initializeParallel(argc, argv)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeParallel fault.");
    retFlg = 1;
  }

//  getcwd(buf,sizeof(buf));
//  printf("カレントディレクトリ=%s\n", buf);

  // 実行パスを取得する。(rankid!=0)
  int myID = 0;
  int commSize = 0;
  SPL_Comm_rank(SPL_COMM_WORLD, &myID);
  SPL_Comm_size(SPL_COMM_WORLD, &commSize);
  if (myID != 0) {
    realpath(cur_path, abs_path);
  }
  else {
    // ランク０はsphereのディレクトリに移動しているので、元に戻す。
    chdir(abs_path);
  }

  /// 実行パスを設定する
  m_runningPath = abs_path;

  // for debug
#ifdef _DEBUGLOG_
  if (myID == 0) {
//    sleep(10);
  }
#endif

  // check initialize in all node
  SPL_Allreduce(&retFlg, &tmp, 1, SPL_INT, SPL_SUM, SPL_COMM_WORLD);
  if( tmp != 0 ){
    // print stack trace
    m_errHandler->printStackTrace();
    // 破棄処理を行う。
    finalize();

    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }

  // コンフィグレーションXMLファイルから読み込みを行う。
  if (PathIsRelative(m_config_fname.c_str())) {
    // 相対パスであるので、実行パスを付加する。
    m_config_fname = m_runningPath + "/" + m_config_fname;
  }

  sphConfigController* config = sphConfigController::getInstance();
  const sphElem* config_elem = NULL;
  const SklCfg::SklSolverConfig* skl_config = NULL;
  if (config->readConfigFile(m_config_fname.c_str())) {
    config_elem = config->getRootElem();
    skl_config = config->getSklSolverConfig();
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "readConfigFile fault. filename=%s", m_config_fname.c_str());
    retFlg = 1;
  }
  if (config_elem == NULL && skl_config == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not read configuration file. filename=%s", m_config_fname.c_str());
    retFlg = 1;
  }
  m_sphere->setConfigElem(config_elem);
  m_sphere->setSklConfig(skl_config);

  /// ソルバー実行
  if( retFlg == 0 ){
    // execute solver
    if (m_sphere->executeSolver() != SPHERE_SUCCESS) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "executeSolver fault.");
      retFlg = 1;
    }
  }

  // エラーメッセージ出力
  if (m_errHandler->existsError()) {
    // print stack trace
    m_errHandler->printStackTrace();
  }

  if (retFlg == 1) {
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
  }

  // 破棄処理を行う。
  finalize();


  return true;
}



/**
 * 初期化を行う。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphApp::initialize()
{

#ifdef SPH_SIGINT_DEFAULT
  signal(SIGINT, SIG_DFL);
#else
  //signal(SIGINT, (sighandler_t)&sphApp::terminateSignalHandler);
  signal(SIGINT, &sphApp::terminateSignalHandler);
#endif // SPH_SIGINT_DEFAULT

  if (m_config_fname.empty()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_config_fname is null.");
    return false;
  }

  if (m_sphere == NULL) {
    m_sphere = sphSphere::getInstance();
  }
  if (m_sphere == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_sphere is null.");
    return false;
  }

  return true;
}


/**
 * 初期化を行う。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphApp::initializeParallel(int argc, char** argv)
{

#ifdef _WIN32
  WSADATA wsadata;
  if (WSAStartup(MAKEWORD(1, 1), &wsadata)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ENVIRONMENT, "WSAStartup failed.");
    return false;
  }
#endif


#ifdef _NON_P4_DEVICE_
  m_p4Flag = true;
#endif

  // Initialize parallel environment
  if (splIsParallel() || m_p4Flag) {
    if (SPL_Init(&argc, &argv) != SPL_SUCCESS ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ENVIRONMENT, "SPL_Init failed.");
      return false;
    }
  }

  return true;

}


/**
 * 破棄処理を行う。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphApp::finalize()
{
  // sphSphereオブジェクトの破棄
  sphSphere::destruct();

  // コンフィグレーションクラス破棄
  sphConfigController::destruct();

  int flag = 0;
  SPL_Initialized(&flag);
  if( flag ) SPL_Finalize();

  return true;
}

