/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphParaExternalFunc.h
 * Parallel External Function Source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphParaExternalFunc.h
 * パラレル外部関数ソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#include "app/sphExternalFunction.h"
#include "app/sphSphere.h"
#include "spl/splmpi.h"

// for Fortran Interface Function
#ifndef _WIN32
// for Linux
#define sph_isParallel_           sph_isparallel_
#define sph_getNumberOfGroup_      sph_getnumberofgroup_
#define sph_getNumberOfNode_      sph_getnumberofnode_
#define sph_getWholeVoxelSize_    sph_getwholevoxelsize_
#define sph_getVoxelDivInfo_      sph_getvoxeldivinfo_
#define sph_getVoxelHeadIndex_    sph_getvoxelheadindex_
#define sph_getVoxelTailIndex_    sph_getvoxeltailindex_
#define sph_getVoxelSize_         sph_getvoxelsize_
#define sph_getCommIDNum_         sph_getcommidnum_
#define sph_getCommIDs_           sph_getcommids_
#define sph_getCommIDEv_          sph_getcommidev_
#define sph_isCommID_             sph_iscommid_
#define sph_getMyID_              sph_getmyid_
#define sph_broadcast_            sph_broadcast_
#define sph_send_                 sph_send_
#define sph_recv_                 sph_recv_
#define sph_iSend_                sph_isend_
#define sph_iRecv_                sph_irecv_
#define sph_allreduce_            sph_allreduce_
#define sph_wait_                 sph_wait_
#define sph_waitAll_              sph_waitall_
#define sph_waitRequestAll_       sph_waitrequestall_
#define sph_abort_                sph_abort_
#define sph_barrier_              sph_barrier_
#define sph_commBndCell_          sph_commbndcell_
#define sph_commPeriodicBndCell_  sph_commperiodicbndcell_
//#define sph_commBndCell2_         sph_commbndcell2_
//#define sph_commCrdBndCell_       sph_commcrdbndcell_
#define sph_printText_            sph_printtext_
#define sph_getTime_              sph_gettime_

#else
// for WIN32
#define sph_isParallel_           SPH_ISPARALLEL
#define sph_getNumberOfGroup_     SPH_GETNUMBEROFGROUP
#define sph_getNumberOfNode_      SPH_GETNUMBEROFNODE
#define sph_getWholeVoxelSize_    SPH_GETWHOLEVOXELSIZE
#define sph_getVoxelDivInfo_      SPH_GETVOXELDIVINFO
#define sph_getVoxelHeadIndex_    SPH_GETVOXELHEADINDEX
#define sph_getVoxelTailIndex_    SPH_GETVOXELTAILINDEX
#define sph_getVoxelSize_         SPH_GETVOXELSIZE
#define sph_getCommIDNum_         SPH_GETCOMMIDNUM
#define sph_getCommIDs_           SPH_GETCOMMIDS
#define sph_getCommIDEv_          SPH_GETCOMMIDEV
#define sph_isCommID_             SPH_ISCOMMID
#define sph_getMyID_              SPH_GETMYID
#define sph_broadcast_            SPH_BROADCAST
#define sph_send_                 SPH_SEND
#define sph_recv_                 SPH_RECV
#define sph_iSend_                SPH_ISEND
#define sph_iRecv_                SPH_IRECV
#define sph_allreduce_            SPH_ALLREDUCE
#define sph_wait_                 SPH_WAIT
#define sph_waitAll_              SPH_WAITALL
#define sph_waitRequestAll_       SPH_WAITREQUESTALL
#define sph_abort_                SPH_ABORT
#define sph_barrier_              SPH_BARRIER
#define sph_commBndCell_          SPH_COMMBNDCELL
#define sph_commPeriodicBndCell_  SPH_COMMPERIODICBNDCELL

//#define sph_commBndCell2_         SPH_COMMBNDCELL2
//#define sph_commCrdBndCell_       SPH_COMMCRDBNDCELL
#define sph_printText_            SPH_PRINTTEXT
#define sph_getTime_              SPH_GETTIME
#endif

extern "C" {


/**
 * ソルバーオブジェクトを取得する。
 * @param[in]       solvID     ソルバーID
 * @return           ソルバーオブジェクト
 */
sphSolverBase* sph_getSolverObject(int solvID)
{
  sphSphere* spherObj = sphSphere::getInstance();
  sphSolverManager* solvMngr = spherObj->getSolverManager();
  return solvMngr->getSolverById(solvID);
}


/**
 * パラレルコントローラを取得する。
 * @param[in]       solvID     ソルバーID
 * @return           パラレルコントローラ
 */
sphParaController* sph_getParaController(int solvID)
{
  sphSolverBase* solvObj = sph_getSolverObject(solvID);
  if (!solvObj) return NULL;
  sphParaController* paraCtlr = solvObj->getParaController();
  return paraCtlr;
}


/**
 * データマネージャを取得する。
 * @param[in]       solvID     ソルバーID
 * @return           データマネージャ
 */
const sphDataManager* sph_getDataManager(int solvID)
{
  sphSolverBase* solvObj = sph_getSolverObject(solvID);
  if (!solvObj) return NULL;
  const sphDataManager* dataMngr = solvObj->getDataManager();
  return dataMngr;
}



/**
 * 逐次実行/並列実行を判断する。
 *
 * 並列実行でも起動プロセス数が１の場合は逐次実行とする。
 * @param[out]       ret     逐次実行/並列実行 (1:並列実行, -1:逐次実行)
 * @return      void
 */
void sph_isParallel_(int* ret)
{
  if( !ret ) return;

  // 起動プロセス数を取得する
  int world_size;
  SPL_Comm_size (SPL_COMM_WORLD, &world_size);
  if( world_size > 1 ) *ret =  1;
  else                        *ret = -1;
  return;
}


/**
 * 並列実行しているソルバーのノード数を取得する。
 * @param[in]       solvID     ソルバーID
 * @param[out]       num        並列実行ソルバーノード数
 * @param[out]       err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_getNumberOfGroup_(int* solvID, int* num, int* err)
{
  if (err) *err = -1;
  if( !num || !solvID || !err) return;
  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  *num = (int)paraCtlr->getNumberOfGroup();
  *err = 1;
  return;
}

/**
 * 並列実行しているソルバーのノード数を取得する。
 * @param[in]       solvID     ソルバーID
 * @param[out]       num        並列実行ソルバーノード数
 * @param[out]       err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_getNumberOfNode_(int* solvID, int* num, int* err)
{
  if (err) *err = -1;
  if( !num || !solvID || !err) return;
  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  *num = (int)paraCtlr->getNumberOfNode();
  *err = 1;
  return;
}

/**
 * ソルバー計算領域全体を取得する。（ガイドセルを含まない）
 * @param[in]       solvID     ソルバーID
 * @param[out]       isz        I方向全体サイズ（ガイドセルを含まない）
 * @param[out]       jsz        I方向全体サイズ（ガイドセルを含まない）
 * @param[out]       ksz        I方向全体サイズ（ガイドセルを含まない）
 * @param[out]       err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_getWholeVoxelSize_(int* solvID, int* isz, int* jsz, int* ksz, int* err)
{
  if (err) *err = -1;
  if( !isz || !jsz || !ksz || !solvID || !err ) return;
  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  const size_t* sz = paraCtlr->getWholeVoxelSize();
  if( !sz ) return;
  *isz = (size_t)sz[0]; *jsz = (size_t)sz[1]; *ksz = (size_t)sz[2];
  *err = 1;
  return;
}

/**
 * ソルバーの分割情報を取得する。
 * @param[in]       solvID     ソルバーID
 * @param[out]       idiv        I方向分割数
 * @param[out]       jdiv        J方向分割数
 * @param[out]       kdiv        K方向分割数
 * @param[out]       err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_getVoxelDivInfo_(int* solvID, int* idiv, int* jdiv, int* kdiv, int* err)
{
  if (err) *err = -1;
  if( !idiv || !jdiv || !kdiv || !solvID || !err ) return;
  const sphSvParaController* paraCtlr = dynamic_cast<const sphSvParaController*>(sph_getParaController(*solvID));
  if (!paraCtlr) return;

  const unsigned int* divInfo = paraCtlr->getVoxelDivInfo();
  if( !divInfo ) return;

  *idiv = (int)divInfo[0]; *jdiv = (int)divInfo[1]; *kdiv = (int)divInfo[2];
  *err = 1;
  return;
}


/**
 * 始点グローバルインデックスを取得する。（ガイドセルを含まない）
 * @param[in]       solvID     ソルバーID
 * @param[in]       id         ノード番号
 * @param[in]       dim        取得方向(0:I方向, 1:J方向, 2:K方向)
 * @param[out]       idx        始点グローバルインデックス(ガイドセルを含まない）
 * @param[out]       err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_getVoxelHeadIndex_(int* solvID, int* id, int* dim, int* idx, int* err)
{
  if (err) *err = -1;
  if( !solvID || !id || !dim || !idx || !err ) return;

  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  *idx = paraCtlr->getVoxelHeadIndex(*id, *dim);
  if( *idx < 0 ) return;

  *err =  1;
  return;
}

/**
 * 終点グローバルインデックスを取得する。（ガイドセルを含まない）
 * @param[in]       solvID     ソルバーID
 * @param[in]       id         ノード番号
 * @param[in]       dim        取得方向(0:I方向, 1:J方向, 2:K方向)
 * @param[out]       idx        終点グローバルインデックス(ガイドセルを含まない）
 * @param[out]       err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_getVoxelTailIndex_(int* solvID,  int* id, int* dim, int* idx, int* err)
{
  if (err) *err = -1;
  if( !solvID || !id || !dim || !idx || !err ) return;

  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  *idx = paraCtlr->getVoxelTailIndex(*id, *dim);
  if( *idx < 0 ) return;

  *err =  1;
  return;
}

/**
 * ボクセルノードサイズを取得する。（ガイドセルを含まない）
 * @param[in]       solvID     ソルバーID
 * @param[out]       isz        I方向ノードサイズ（ガイドセルを含まない）
 * @param[out]       jsz        J方向ノードサイズ（ガイドセルを含まない）
 * @param[out]       ksz        K方向ノードサイズ（ガイドセルを含まない）
 * @param[out]       err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_getVoxelSize_(int* solvID, int* isz, int* jsz, int* ksz, int* err)
{
  if (err) *err = -1;
  if( !solvID || !isz || !jsz || !ksz || !err ) return;

  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  const size_t* sz = paraCtlr->getVoxelSize();
  if( !sz ) return;
  *isz = (int)sz[0]; *jsz = (int)sz[1]; *ksz = (int)sz[2];
  *err = 1;
  return;
}



/**
 * 隣接ノード数を取得する。
 * @param[in]       solvID     ソルバーID
 * @param[in]       i          取得I方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[in]       j          取得J方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[in]       k          取得K方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[out]      num         ノード数
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_getCommIDNum_(int* solvID, int* i, int* j, int* k, int* id, int* err)
{
  if (err) *err = -1;
  if( !solvID || !i || !j || !k || !id || !err ) return;
  const sphSvParaController* paraCtlr = dynamic_cast<const sphSvParaController*>(sph_getParaController(*solvID));
  if (!paraCtlr) return;

  *id = paraCtlr->getCommIDNum(*i, *j, *k);

  *err =  1;
  return;
}


/**
 * 隣接ノード番号リストを取得する。
 * @param[in]       solvID     ソルバーID
 * @param[in]       i          取得I方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[in]       j          取得J方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[in]       k          取得K方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[out]       id        ノード番号リスト
 * @param[out]       err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_getCommIDs_(int* solvID, int* i, int* j, int* k, int* id_list, int* err)
{
  if (err) *err = -1;
  if( !solvID || !i || !j || !k || !id_list || !err ) return;
  const sphSvParaController* paraCtlr = dynamic_cast<const sphSvParaController*>(sph_getParaController(*solvID));
  if (!paraCtlr) return;

  // ノード番号数を取得する。
  int num = paraCtlr->getCommIDNum(*i, *j, *k);
  if (num > 0) {
    // ノード番号リストを取得する。
    const int* list = paraCtlr->getCommIDs(*i, *j, *k);
    register int i;
    for (i=0; i<num; i++) id_list[i] = list[i];
  }

  *err =  1;
  return;
}


/**
 * 隣接ノード番号リストを取得する。（均等分割）
 * @param[in]       solvID     ソルバーID
 * @param[in]       i          取得I方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[in]       j          取得J方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[in]       k          取得K方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[out]       id        隣接ノード番号
 * @param[out]       err        エラー番号（1:success/-1:fault)
 *                              均等分割ではない場合、エラーとする。
 * @return      void
 */
void sph_getCommIDEv_(int* solvID, int* i, int* j, int* k, int* id, int* err)
{
  if (err) *err = -1;
  if( !solvID || !i || !j || !k || !id || !err ) return;
  const sphSvParaController* paraCtlr = dynamic_cast<const sphSvParaController*>(sph_getParaController(*solvID));
  if (!paraCtlr) return;

  if (paraCtlr->isMb()) {
    return;
  }
  // ノード番号数を取得する。
  *id = paraCtlr->getCommIDEv(*i, *j, *k);

  *err =  1;
  return;
}


/**
 * 隣接ノードの有無を判断する。
 * @param[in]       solvID     ソルバーID
 * @param[in]       i          取得I方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[in]       j          取得J方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[in]       k          取得K方向フラグ ( -1:マイナス方向, 1:プラス方向)
 * @param[out]      num         ノード数 (>0:ノード数, 0:隣接ノード無し, <0:最外外郭ノード)
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_isCommID_(int* solvID, int* i, int* j, int* k, int* num, int* err)
{
  if (err) *err = -1;
  if( !solvID || !i || !j || !k || !num || !err ) return;
  const sphSvParaController* paraCtlr = dynamic_cast<const sphSvParaController*>(sph_getParaController(*solvID));
  if (!paraCtlr) return;

  *num = paraCtlr->isCommID(*i, *j, *k);
  *err =  1;
  return;
}


/**
 * 自ノード数を取得する。
 * @param[in]       solvID     ソルバーID
 * @param[out]      id         ノード番号
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_getMyID_(int* solvID, int* id, int* err)
{
  if (err) *err = -1;
  if( !solvID || !id || !err ) return;
  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  *id = paraCtlr->getMyID();

  *err =  1;
  return;
}


/**
 * ブロードキャストを行う。
 * @param[in]       solvID     ソルバーID
 * @param[in/out]       data       送信データ
 * @param[in]       c          送信データ数
 * @param[in]       dtype       データ型番号
 * @param[in]      src         送信元プロセスのランク
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_broadcast_(int* solvID, void* data, int* c, int* dtype, int* src, int* err)
{
  if (err) *err = -1;
  if( !solvID || !data || !c || !dtype || !src || !err ) return;
  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  SPL_Datatype dt_type = spl_get_datatype_byindex(*dtype);
  if (dt_type == SPL_DATATYPE_NULL) return;

  if (! paraCtlr->broadcast(data, *c, dt_type, *src) ) {
    return;
  }
  *err = 1;

  return;
}


/**
 * ブロッキングを行い、データ送信を行う。
 * @param[in]       solvID     ソルバーID
 * @param[in]       data       送信データ
 * @param[in]       c          送信データ数
 * @param[in]       dtype       データ型番号
 * @param[in]      dst         送信先プロセスのランク
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_send_(int* solvID, void* data, int* c, int* dtype, int* dst, int* err)
{
  if (err) *err = -1;
  if( !solvID || !data || !c || !dtype || !dst || !err ) return;
  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  SPL_Datatype dt_type = spl_get_datatype_byindex(*dtype);
  if (dt_type == SPL_DATATYPE_NULL) return;

  if (! paraCtlr->send(data, *c, dt_type, *dst) ) {
    return;
  }
  *err = 1;

  return;
}


/**
 * ブロッキングを行い、データ受信を行う。
 * @param[in]       solvID     ソルバーID
 * @param[out]       data       受信データ
 * @param[in]       c          受信データ数
 * @param[in]       dtype       データ型番号
 * @param[out]      dst         送信先プロセスのランク
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_recv_(int* solvID, void* data, int* c, int* dtype, int* src, int* err)
{
  if (err) *err = -1;
  if( !solvID || !data || !c || !dtype || !src || !err ) return;
  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  SPL_Datatype dt_type = spl_get_datatype_byindex(*dtype);
  if (dt_type == SPL_DATATYPE_NULL) return;

  if (! paraCtlr->recv(data, *c, dt_type, *src) ) {
    return;
  }
  *err = 1;

  return;
}


/**
 * ブロッキングを行い、データ送信を行う。
 * @param[in]       solvID     ソルバーID
 * @param[in]       data       送信データ
 * @param[in]       c          送信データ数
 * @param[in]       dtype       データ型番号
 * @param[in]      dst         送信先プロセスのランク番号
 * @param[out]      key         送信リクエストキー番号
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_iSend_(int* solvID, void* data, int* c, int* dtype, int* dst, int* key, int* err)
{
  if (err) *err = -1;
  if( !solvID || !data || !c || !dtype || !dst || !key || !err ) return;
  sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  SPL_Datatype splDtype = spl_get_datatype_byindex(*dtype);
  if (splDtype == SPL_DATATYPE_NULL) return;

  *key = paraCtlr->iSend(data, *c, splDtype, *dst);
  if( *key <= 0 ) return;

  *err = 1;
  return;
}

/**
 * ブロッキングを行い、データ受信を行う。
 * @param[in]       solvID     ソルバーID
 * @param[out]       data       受信データ
 * @param[in]       c          受信データ数
 * @param[in]       dtype       データ型番号
 * @param[in]       dst         送信先プロセスのランク番号
 * @param[out]      key         受信リクエストキー番号
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_iRecv_(int* solvID, void* data, int* c, int* dtype, int* src, int* key, int* err)
{
  if (err) *err = -1;
  if( !solvID || !data || !c || !dtype || !src || !key || !err ) return;
  sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  SPL_Datatype splDtype = spl_get_datatype_byindex(*dtype);
  if (splDtype == SPL_DATATYPE_NULL) return;

  *key = paraCtlr->iRecv(data, *c, splDtype, *src);
  if( *key <= 0 ) return;

  *err = 1;
  return;
}

/**
 * オールレデュースを行う。
 * @param[in]       solvID     ソルバーID
 * @param[in]       sendbuf       送信データ
 * @param[out]       recvbuf       受信バッファー
 * @param[in]       c           送信データ数
 * @param[in]       dtype       データ型番号
 * @param[in]      optype      オペランド番号
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_allreduce_(int* solvID, void* sendbuf, void* recvbuf, int* c,
                   int* dtype, int* optype, int* err)
{
  if (err) *err = -1;
  if( !solvID || !sendbuf || !recvbuf || !c || !dtype || !optype || !err ) return;
  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  SPL_Datatype splDtype = spl_get_datatype_byindex(*dtype);
  if (splDtype == SPL_DATATYPE_NULL) return;

  SPL_Op spl_op = SPL_OP_NULL;
  if       ( *optype == 100 ) spl_op = SPL_MAX;
  else if  ( *optype == 101 ) spl_op = SPL_MIN;
  else if  ( *optype == 102 ) spl_op = SPL_SUM;
  else if  ( *optype == 103 ) spl_op = SPL_PROD;
  else if  ( *optype == 104 ) spl_op = SPL_LAND;
  else if  ( *optype == 105 ) spl_op = SPL_BAND;
  else if  ( *optype == 106 ) spl_op = SPL_LOR;
  else if  ( *optype == 107 ) spl_op = SPL_BOR;
  else if  ( *optype == 108 ) spl_op = SPL_LXOR;
  else if  ( *optype == 109 ) spl_op = SPL_BXOR;
  else return;

  if( !paraCtlr->allreduce(sendbuf, recvbuf, *c, splDtype, spl_op) ) {
    return;
  }
  *err = 1;
  return;
}


/**
 * 非同期通信の待機を行う。（１つのリクエストのみ）
 * @param[in]       solvID     ソルバーID
 * @param[in]       key        登録リクエストキー
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_wait_(int* solvID, int* key, int* err)
{
  if (err) *err = -1;
  if( !solvID || !key || !err ) return;
  sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  if( !paraCtlr->wait(*key) ) {
    return;
  }
  *err = 1;
  return;
}

/**
 * 非同期通信の待機を行う。（複数のリクエスト）
 * @param[in]       solvID     ソルバーID
 * @param[in]       key_num     リクエストキー数
 * @param[in]       key_list    登録リクエストキーリスト
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_waitAll_(int* solvID, int* key_num, int* key_list, int* err)
{
  if (err) *err = -1;
  if( !solvID || !key_num || !key_list || !err ) return;
  sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  if( !paraCtlr->waitAll((unsigned)*key_num, key_list) ) {
    return;
  }
  *err = 1;
  return;
}


/**
 * 完了待ちとなっている非同期通信の待機を行う。（すべてのリクエスト）
 * @param[in]       solvID     ソルバーID
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_waitRequestAll_(int* solvID, int* err)
{
  if (err) *err = -1;
  if( !solvID || !err ) return;
  sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  if( !paraCtlr->waitRequestAll() ) {
    return;
  }
  *err = 1;
  return;
}


/**
 * 強制終了を行う。
 * @param[in]       errcode          エラーコード
 * @return      void
 */
void sph_abort_(int* errcode)
{
  if( !errcode ) return;
  SPL_Abort(SPL_COMM_WORLD, *errcode);
  return;
}

/**
 * バリア同期を行う。
 * @param[in]       solvID     ソルバーID
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_barrier_(int* solvID, int* err)
{
  if (err) *err = -1;
  if( !solvID || !err ) return;
  const sphParaController* paraCtlr = sph_getParaController(*solvID);
  if (!paraCtlr) return;

  if( !paraCtlr->barrier() ) {
    return;
  }

  *err = 1;
  return;
}


/**
 * 隣接ノードとガイドセルの通信を行う。
 * @param[in]       solvID     ソルバーID
 * @param[in/out]    data      通信データポインタ
 * @param[in]       face       通信面数
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_commBndCell_(int* solvID, void* data, int* face, int* err)
{
  if (err) *err = -1;
  if( !solvID || !data || !face || !err ) return;

  const sphDataManager* dataMngr = sph_getDataManager(*solvID);
  if (!dataMngr) return;

  sphData* dataObj = dataMngr->getDataObj(data);
  if (!dataObj) return;

  if ( !dataObj->comm(*face) ) {
    return;
  }

  *err = 1;
  return;
}



/**
 * 最外殻ノードにて周期境界の通信を指定方向にて行う。
 * @param[in]       solvID     ソルバーID
 * @param[in/out]    data      通信データポインタ
 * @param[in]       face_dir   通信面 (X方向:0, Y方向:1, Z方向:2)
 * @param[in]       axis       通信方向 (プラス方向:+1, マイナス方向:-1)
 * @param[in]       face       通信面数
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_commPeriodicBndCell_(int* solvID, void* data, int* face_dir, int* axis, int* face, int* err)
{

  if (err) *err = -1;
  if( !solvID || !data || !face || !err ) return;

  const sphDataManager* dataMngr = sph_getDataManager(*solvID);
  if (!dataMngr) return;

  sphData* dataObj = dataMngr->getDataObj(data);
  if (!dataObj) return;

  if ( !dataObj->commPeriodic(*face_dir, *axis, *face) ) {
    return;
  }

  *err = 1;
  return;
}



/*****    not support function
void sph_commBndCell2_(int* solvID, void* data, int* face, int* wait_num, int* req, int* err)
{
  return;
}

void sph_commCrdBndCell_(int* solvID, void* data, int* face, int* err)
{
  return;
}
*/


/**
 * テキストファイル出力を行う。
 * @param[in]       file_label     ファイルラベル
 * @param[in]       text              出力テキスト
 * @param[out]      err        エラー番号（1:success/-1:fault)
 * @return      void
 */
void sph_printText_(int* solvID, char* file_label, char* text, int lbl_len, int txt_len)
{
  if( !solvID || !file_label || !text) return;

  // file_labelの空白を取り除く
  char label[64];
  char* pos = strchr(file_label, ' ');
  if (pos == NULL) {
    strcpy(label, file_label);
  }
  else if (pos-file_label < 64) {
    strncpy(label, file_label, pos-file_label);
  }
  else {
    return;
  }

  sphSolverBase* solvObj = sph_getSolverObject(*solvID);
  if (solvObj == NULL) return;
  sphTextFile txt_file;
  if (!solvObj->getTextFile(label, &txt_file)) return;
  txt_file.openFile();

  char* msg = new char[txt_len+1];
  memset(msg, 0x00, txt_len+1);
  strncpy(msg, text, txt_len);
  txt_file.print(msg);
  delete [] msg;

  txt_file.close();

  return;
}


/**
 * 現在時間の取得を行う。
 * @param           void
 * @return          現在時間
 */
double sph_getTime_()
{
  return splGetTime();
}


};    /* extern "C" */

