/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
 /**
 * \~english
 * @file sphSphere.C
 * Sphere Class source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphSphere.C
 * Sphereクラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */
#include "app/sphSphere.h"
#include "solverbase/sphSolverManager.h"
#include "solverbase/sphCouplerManager.h"
#include "parallel/sphParaManager.h"
#include "solverbase/sphFlowBase.h"
#include "solverbase/sphSolverFactory.h"
#include "solverbase/sphCouplerFactory.h"
#include "solverbase/sphFlowFactory.h"
#include "config/sphElem.h"
#include "config/sphElemUtils.h"

/// sphSphere スタティックメンバーの初期化
sphSphere *sphSphere::m_instance = NULL;

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphSphere::sphSphere()
{
  m_solverMngr = new sphSolverManager();
  m_couplerMngr = new sphCouplerManager();
  m_paraMngr = new sphParaManager();
  m_configElem = NULL;
  m_sklconfig = NULL;
  m_flow = NULL;
}

/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphSphere::~sphSphere()
{
  register int i;

  if (m_solverMngr != NULL) delete m_solverMngr;
  if (m_couplerMngr != NULL) delete m_couplerMngr;
  if (m_paraMngr != NULL) delete m_paraMngr;
  // m_configElemの破棄は行わない。
  // sphConfigController::m_rootElemを持っているので、sphConfigControllerの破棄処理にて行う。
//  if (m_configElem != NULL) delete m_configElem;
  if (m_flow != NULL) delete m_flow;

  std::map<int, ProcessGroup* >::iterator it_solv;
  for(it_solv=m_solverGroup.begin();it_solv!=m_solverGroup.end();it_solv++) {
    ProcessGroup* proc_grp = it_solv->second;
    if (proc_grp != NULL) delete proc_grp;
    proc_grp = NULL;
  }
  m_solverGroup.clear();

  std::map<int, ProcessGroup* >::iterator it_coup;
  for(it_coup=m_couplerGroup.begin();it_coup!=m_couplerGroup.end();it_coup++) {
    ProcessGroup* proc_grp = it_coup->second;
    if (proc_grp != NULL) delete proc_grp;
    proc_grp = NULL;
  }
  m_couplerGroup.clear();
}


/**
 * \~english
 * get sphSphere instance for Singleton
 *
 * \~japanese
 * シングルトンクラスであるsphSphereのオブジェクトを取得する。
 */
sphSphere* sphSphere::getInstance()
{
  if (m_instance == NULL) {
    m_instance = new sphSphere();

    // プログラムが終了時に破棄処理を呼ぶ。
    atexit(&destruct);
  }

  return m_instance;
}


/**
 * \~english
 * destruct sphSphere instance for Singleton
 *
 * \~japanese
 * シングルトンクラスであるsphSphereのオブジェクトの破棄を行う。
 */
void sphSphere::destruct()
{
  if (m_instance != NULL) delete m_instance;
  m_instance = NULL;

  return;
}


/**
 * \~english
 * run sphere solver
 * @param       void
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * Sphere ソルバーを実行する。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
int sphSphere::executeSolver()
{
  int ret = SPHERE_SUCCESS;
  if (isSphereConfig()) {
    /// ソルバのノードIDテーブルの作成、プロセスグループの作成を行う。
    if (!createProcessSolverGroup()) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createProcessSolverGroup fault.");
      return SPHERE_ERROR;
    }

    /// カプラのノードIDテーブルの作成、プロセスグループの作成を行う。
    if (!createProcessCouplerGroup()) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createProcessCouplerGroup fault.");
      return SPHERE_ERROR;
    }

    /// ソルバーの生成を行う
    if (!createSolver()) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSolver fault.");
      return SPHERE_ERROR;
    }

    /// カプラーの生成を行う。
    if (!createCoupler()) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createCoupler fault.");
      return SPHERE_ERROR;
    }
    /// フローコントローラの生成を行う。
    if (!createFlow()) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createFlow fault.");
      return SPHERE_ERROR;
    }
  }
  else {
    /// ソルバーの生成を行う
    if (!createSklSolver()) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSklSolver fault.");
      return SPHERE_ERROR;
    }

    /// デフォルトフローコントローラの生成を行う。
    if (!createSklFlow()) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createFlow fault.");
      return SPHERE_ERROR;
    }
  }

  /// ソルバーを実行する。
  ret = m_flow->executeSolvers();

  return ret;
}


/**
 * \~english
 * get solver manager
 * @param       void
 * @return      solver manager
 *
 * \~japanese
 * ソルバーマネージャーを取得する。
 * @param       void
 * @return    ソルバーマネージャー
 */
const sphSolverManager* sphSphere::getSolverManager() const
{
  return m_solverMngr;
}
sphSolverManager* sphSphere::getSolverManager()
{
  return m_solverMngr;
}


/**
 * \~english
 * get coupler manager
 * @param       void
 * @return      coupler manager
 *
 * \~japanese
 * カプラーマネージャーを取得する。
 * @param       void
 * @return    カプラーマネージャー
 */
const sphCouplerManager* sphSphere::getCouplerManager() const
{
  return m_couplerMngr;
}
sphCouplerManager* sphSphere::getCouplerManager()
{
  return m_couplerMngr;
}


/**
 * \~english
 * get parallel manager
 * @param       void
 * @return      parallel manager
 *
 * \~japanese
 * パラレルマネージャーを取得する。
 * @param       void
 * @return      パラレルマネージャー
 */
const sphParaManager* sphSphere::getParaManager() const
{
  return m_paraMngr;
}
sphParaManager* sphSphere::getParaManager()
{

  return m_paraMngr;
}

/**
 * \~english
 * get root configuration element
 * @param       void
 * @return      root configuration element
 *
 * \~japanese
 * コンフィグレーションルート要素を取得する。
 * @param       void
 * @return      コンフィグレーションルート要素
 */
const sphElem* sphSphere::getConfigElem() const
{
  return m_configElem;
}
const sphElem* sphSphere::getConfigElem()
{
  return m_configElem;
}



/**
 * Sklコンフィグレーションクラスを取得する。
 * @param       void
 * @return      Sklコンフィグレーションクラス
 */
const SklCfg::SklSolverConfig* sphSphere::getSklSolverConfig()
{
  return m_sklconfig;
}


/**
 * \~english
 * get flow controller
 * @param       void
 * @return      flow controller
 *
 * \~japanese
 * フローコントローラーを取得する。
 * @param       void
 * @return      フローコントローラー
 */
const sphFlowBase* sphSphere::getFlow() const
{
  return m_flow;
}
sphFlowBase* sphSphere::getFlow()
{
  return m_flow;
}


/**
 * ソルバの追加を行う。
 * @param       solver          ソルバオブジェクト
 * @param       sphObj          sphSphereオブジェクト
 * @param       comm          コミュニケータ
 * @param       group          グループ
 * @return    success(=true)/fault(=false)
 */
bool sphSphere::addSolver(sphSolverBase* solver,
                          sphSphere* sphObj,
                          const sphElem* config_elem,
                          SPL_Comm comm, SPL_Group group)
{
  /// sphSphereオブジェクトの設定を行う。
  solver->setSphere(this);
  /// コンフィグレーションの設定を行う。
  solver->setConfigElem(config_elem);
  /// 作成済みのコミュニケータ、グループを設定する。
  solver->setCommGroup(comm, group);

  /// ファイルコントローラの設定を行う。
  if (!solver->setupFileController()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setupFileController fault");
    return false;
  }

  /// ソルバマネージャに追加を行う。
  return m_solverMngr->addSolver(solver);
}


/**
 * ソルバの追加を行う。
 * @param       solver          ソルバオブジェクト
 * @param       sphObj          sphSphereオブジェクト
 * @param       comm          コミュニケータ
 * @param       group          グループ
 * @return    success(=true)/fault(=false)
 */
bool sphSphere::addSolver(sphSolverBase* solver,
                          sphSphere* sphObj,
                          const SklCfg::SklSolverConfig* sklconfig,
                          SPL_Comm comm, SPL_Group group)
{
  /// sphSphereオブジェクトの設定を行う。
  solver->setSphere(this);
  /// コンフィグレーションの設定を行う。
  solver->setSklConfig(sklconfig);
  /// 作成済みのコミュニケータ、グループを設定する。
  solver->setCommGroup(comm, group);

  /// ファイルコントローラの設定を行う。
  if (!solver->setupFileController()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setupFileController fault");
    return false;
  }

  /// ソルバマネージャに追加を行う。
  return m_solverMngr->addSolver(solver);
}



/**
 * カプラの追加を行う。
 * @param       coupler         カプラオブジェクト
 * @param       sphObj          sphSphereオブジェクト
 * @param       comm          コミュニケータ
 * @param       group          グループ
 * @return    success(=true)/fault(=false)
 */
bool sphSphere::addCoupler(sphCouplerBase* coupler,
                           sphSphere* sphObj,
                           const sphElem* config_elem,
                           SPL_Comm comm, SPL_Group group)
{
  /// sphSphereオブジェクトの設定を行う。
  coupler->setSphere(this);
  /// コンフィグレーションの設定を行う。
  coupler->setConfigElem(config_elem);
  /// 作成済みのコミュニケータ、グループを設定する。
  coupler->setCommGroup(comm, group);

  /// カプラマネージャに追加を行う。
  return m_couplerMngr->addCoupler(coupler);
}


/**
 * フロー制御の設定を行う。
 * @param       フローオブジェクト
 * @return      void
 */
void sphSphere::setFlow(sphFlowBase* flow)
{
  m_flow = flow;
  return;
}

/**
 * コンフィグレーションルート要素を設定する。
 * @param       root_elem       コンフィグレーションルート要素
 * @return      void
 */
void sphSphere::setConfigElem(const sphElem* root_elem)
{
  m_configElem = root_elem;
  return;
}

/**
 *
 * Vsphereコンフィグレーションを設定する。
 * @param       skl_config       Vsphereコンフィグレーション
 * @return      void
 */
void sphSphere::setSklConfig(const SklCfg::SklSolverConfig* skl_config)
{
  m_sklconfig = skl_config;
  return;
}


/**
 * フローコントローラを生成する。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphSphere::createFlow()
{
  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_configElem is null.");
    return false;
  }
  register int i;

  /// コンフィグレーションルート要素からフロー要素の取得
  const sphElem* flow_elem = NULL;
  const char* flow_label = NULL;
  const char* flow_class = NULL;
  int flow_id = 0;
  // <SphFlow>要素取得
  flow_elem = m_configElem->getElem(SPH_CFG_NAME_FLOW);
  if (flow_elem != NULL) {
    // コンフィグレーション記述のフローラベルの取得
    flow_label = flow_elem->getAttrString(SPH_CFG_ATTR_LABEL);
    flow_id = flow_elem->getAttrInt(SPH_CFG_ATTR_ID);
    flow_class = flow_elem->getAttrString(SPH_CFG_ATTR_CLASS);
  }

  /// ソルバ配下のフロー要素を取得する。
  /// ルート直下にフロー要素が無いので、ソルバ要素配下から検索する。
  if (flow_elem == NULL) {
    int solver_cnt = m_solverMngr->getSolverCount();
    for (i=0; i<solver_cnt; i++) {
      const sphSolverBase* solver = m_solverMngr->getSolver(i);
      if (solver == NULL || solver->getConfigElem() == NULL) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "solver object is null.");
        return false;
      }
      flow_elem = solver->getConfigElem()->getElem(SPH_CFG_NAME_FLOW);
      if (flow_elem != NULL) {
        const char* label = flow_elem->getAttrString(SPH_CFG_ATTR_LABEL);
        int id = flow_elem->getAttrInt(SPH_CFG_ATTR_ID);
        const char* class_name = flow_elem->getAttrString(SPH_CFG_ATTR_CLASS);
        /// フローラベルが設定されていない最初のフロー要素であるので、ラベルを設定する。
        if (flow_label == NULL) {
          flow_label = label;
          flow_id = id;
          flow_class = class_name;
        }
        else {
          /// labelとidの一致チェック（フロー制御は、１プロセスに１つしかインスタンスできない。）
          // 同一のフロー要素が複数記述されているのであれば、無視する。
          if (strcasecmp(flow_label, label) != 0 || flow_id != id || strcasecmp(flow_class, class_name) != 0) {
            // error
            m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "duplication SphFlow element.");
            return false;
          }
        }
      }     /* if (flow_elem == NULL) */
    }     /* for (i=0; i<solver_cnt; i++) */
  }     /* if (flow_elem == NULL) */

  /// カプラ配下のフロー要素を取得する。
  /// ルート直下、ソルバ配下にフロー要素が無いので、カプラ要素配下から検索する。
  if (flow_elem == NULL) {
    int coupler_cnt = m_couplerMngr->getCouplerCount();
    for (i=0; i<coupler_cnt; i++) {
      const sphCouplerBase* coupler = m_couplerMngr->getCoupler(i);
      if (coupler == NULL || coupler->getConfigElem() == NULL) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "coupler object is null.");
        return false;
      }
      flow_elem = coupler->getConfigElem()->getElem(SPH_CFG_NAME_FLOW);
      if (flow_elem != NULL) {
        const char* label = flow_elem->getAttrString(SPH_CFG_ATTR_LABEL);
        int id = flow_elem->getAttrInt(SPH_CFG_ATTR_ID);
        const char* class_name = flow_elem->getAttrString(SPH_CFG_ATTR_CLASS);
        /// フローラベルが設定されていない最初のフロー要素であるので、ラベルを設定する。
        if (flow_label == NULL) {
          flow_label = label;
          flow_id = id;
          flow_class = class_name;
        }
        else {
          /// labelとidの一致チェック（フロー制御は、１プロセスに１つしかインスタンスできない。）
          // 同一のフロー要素が複数記述されているのであれば、無視する。
          if (strcasecmp(flow_label, label) != 0 || flow_id != id || strcasecmp(flow_class, class_name) != 0) {
            // error
            m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "duplication SphFlow element.");
            return false;
          }
        }
      }     /* if (flow_elem == NULL) */
    }     /* for (i=0; i<coupler_cnt; i++) */
  }     /* if (flow_elem == NULL) */

  if (flow_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found SphFlow element.");
    return false;
  }

  /// フローコントローラーの作成
  sphFlowFactory flowFactory;
  m_flow = flowFactory.factoryFlow(flow_class, flow_label, flow_id);
  if (m_flow == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create flow object, flowClass=%s, flowLabel=%s, flowId=%d.", flow_class, flow_label, flow_id);
    return false;
  }

  if (flow_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not found flow element, flowClass=%s, flowLabel=%s, flowId=%d.", flow_class, flow_label, flow_id);
    return false;
  }

  m_flow->setSphere(this);

  /// フローコントローラの初期化を行う。
  if (m_flow->initializeFlow(flow_elem) != SPHERE_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeFlow fault.");
    return false;
  }

  return true;
}

/**
 * デフォルトフローコントローラを生成する。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphSphere::createSklFlow()
{
  if (m_sklconfig == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_sklconfig is null.");
    return false;
  }
  register int i;

  /// コンフィグレーションルート要素からフロー要素の取得
  const char* flow_label = m_sklconfig->getSolverType();
  const char* flow_class = m_sklconfig->getSolverType();
  int flow_id = 0;

  /// フローコントローラーの作成
  sphFlowFactory flowFactory;
  m_flow = flowFactory.factoryFlow(flow_class, flow_label, flow_id);
  if (m_flow == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create flow object, flowClass=%s, flowLabel=%s, flowId=%d.", flow_class, flow_label, flow_id);
    return false;
  }

  m_flow->setSphere(this);

  /// フローコントローラの初期化を行う。
  if (m_flow->initializeFlow(m_sklconfig) != SPHERE_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeFlow fault.");
    return false;
  }

  return true;
}


/**
 * ソルバーを生成する。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphSphere::createSolver()
{
  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_configElem is null.");
    return false;
  }

  /// 自プロセス番号の取得を行う。
  int myID = 0;
  SPL_Comm_rank(SPL_COMM_WORLD, &myID);

  sphSolverFactory solvFactory;
  register int i;
  sphElemUtils elemUtils(m_configElem);
  int solv_count = elemUtils.getSolverCount();
  for (i=0; i<solv_count; i++) {
    /// コンフィグレーション要素からソルバ要素の取得
    const sphElem* solv_elem = elemUtils.getSolverElem(i);
    const char* solv_label = solv_elem->getAttrString(SPH_CFG_ATTR_LABEL);
    int solv_id = solv_elem->getAttrInt(SPH_CFG_ATTR_ID);
    const char* solv_class = solv_elem->getAttrString(SPH_CFG_ATTR_CLASS);

    /// プロセスグループ情報を取得する。
    const ProcessGroup* proc_grp = getProcessGroupOfSolver(solv_id);
    if (proc_grp == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "ProcessGroup is null, solverID=%d.", solv_id);
      return false;
    }

    /// 自プロセスは取得ソルバの実行プロセスであるかチェックする。
    if (proc_grp->isProcessGroupId(myID)) {
      /// コミュニケータが生成済みであるかチェックする。
      if (proc_grp->m_comm == SPL_COMM_NULL && splIsParallel()) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "process communication is null, solverID=%d, myID=%d.", solv_id, myID);
        return false;
      }
      /// ソルバーオブジェクト生成
      sphSolverBase* solver = solvFactory.factorySolver(solv_class, solv_label, solv_id);
      if (solver == NULL) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create Solver Object, solverClass=%s, solverLabel=%s, solverID=%d.", solv_class, solv_label, solv_id);
        return false;
      }

      // ソルバー追加
      addSolver(solver, this, solv_elem, proc_grp->m_comm, proc_grp->m_group);
    }
  }

  return true;
}



/**
 * ソルバーを生成する。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphSphere::createSklSolver()
{
  if (m_sklconfig == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_sklconfig is null.");
    return false;
  }

  SPL_Comm comm = SPL_COMM_WORLD;
  SPL_Group group;
  SPL_Comm_group(comm, &group);

  /// 自プロセス番号の取得を行う。
  int myID = 0;
  SPL_Comm_rank(comm, &myID);

  sphSolverFactory solvFactory;
  /// ソルバーオブジェクト生成
  const char* solv_label = m_sklconfig->getSolverType();
  const char* solv_class = m_sklconfig->getSolverType();
  if (solv_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not get SolverType in SklSolverConfig.");
    return false;
  }

  int solv_id = 1;
  sphSolverBase* solver = solvFactory.factorySolver(solv_class, solv_label, solv_id);
  if (solver == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create SklSolver Object, solverClass=%s, solverLabel=%s, solverID=%d.", solv_class, solv_label, solv_id);
    return false;
  }

  // ソルバー追加
  addSolver(solver, this, m_sklconfig, comm, group);

  return true;
}


/**
 * カプラを生成する。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphSphere::createCoupler()
{
  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_configElem is null.");
    return false;
  }

  /// 自プロセス番号の取得を行う。
  int myID = 0;
  SPL_Comm comm;
  SPL_Group group;
  SPL_Comm_rank(SPL_COMM_WORLD, &myID);

  sphCouplerFactory coupFactory;
  register int i;
  sphElemUtils elemUtils(m_configElem);
  int coup_count = elemUtils.getCouplerCount();
  for (i=0; i<coup_count; i++) {
    /// コンフィグレーション要素からカプラ要素の取得
    const sphElem* coup_elem = elemUtils.getCouplerElem(i);
    const char* coupler_label = coup_elem->getAttrString(SPH_CFG_ATTR_LABEL);
    int coupler_id = coup_elem->getAttrInt(SPH_CFG_ATTR_ID);
    const char* coupler_class = coup_elem->getAttrString(SPH_CFG_ATTR_CLASS);

    /// カプラ全体のプロセスグループ情報を取得する。
    const ProcessGroup* proc_grp = getProcessGroupOfCoupler(coupler_id);
    if (proc_grp == NULL) {
      continue;
//      // error
//      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "ProcessGroup is null, couplerID=%d.", coupler_id);
//      return false;
    }
    /// 自プロセスは取得カプラの実行プロセスであるかチェックする。
    if (proc_grp->isProcessGroupId(myID)) {
      /// コミュニケータが生成済みであるかチェックする。
      if (proc_grp->m_comm == SPL_COMM_NULL && splIsParallel() ) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "process communication is null, solverID=%d, myID=%d.", coupler_id, myID);
        return false;
      }

      /// カプラオブジェクト生成
      sphCouplerBase* coupler = coupFactory.factoryCoupler(coupler_class, coupler_label, coupler_id);
      if (coupler == NULL) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create Coupler Object, couplerClass=%s, couplerLabel=%s, couplerID=%d.", coupler_class, coupler_label, coupler_id);
        return false;
      }

      // カプラ追加
      addCoupler(coupler, this, coup_elem, proc_grp->m_comm, proc_grp->m_group);
    }
  }

  return true;
}



/**
 * コンフィグレーション要素からソルバのプロセスグループを作成する。.
 * プロセスグループ情報（プロセス番号リスト）の作成を行う。
 * ソルバのプロセスグループを作成する。
 * @param     void
 * @return    success(=true)/fault(=false)
 */
bool sphSphere::createProcessSolverGroup()
{
  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_configElem is null.");
    return false;
  }

  /// 起動プロセス数を取得する
  int world_size;
  SPL_Comm_size (SPL_COMM_WORLD, &world_size);

  // ソルバIDと割当ノード数
  std::map<int, int> mapSolverSize;

  sphElemUtils elemUtils(m_configElem);

  /// ソルバ毎の割当プロセス数を取得する。
  register int i, j;
  int solver_count = elemUtils.getSolverCount();
  for (i=0; i<solver_count; i++) {
    // <SphSteer>から分割数又はプロセス数を取得する。
    const sphElem* solv_elem = elemUtils.getSolverElem(i);
    if (solv_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "solver index = %d.", i);
      return false;
    }
    int solver_id = solv_elem->getAttrInt(SPH_CFG_ATTR_ID);

    /// ソルバ割当プロセス数を取得する。
    int node_size = elemUtils.getNodeCountFromSteer(solv_elem);

    // 分割数、プロセス数チェック
    if (node_size == 0) {
      node_size = world_size;
    }
    else if (world_size == 1) {
      // 並列実行では無いので、プロセス数＝１を割り当てる
      node_size = 1;
    }
    else if (node_size > world_size) {
      /// 起動プロセス数より多く１つのソルバに割当ることはできない。
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "node size is %d in solver, more than process size(%d)", node_size, world_size);
      return false;
    }

    // サイズリストに追加する。
    mapSolverSize.insert(std::make_pair(solver_id, node_size));
  }

  /// ソルバのノード番号リストを作成, プロセスグループの生成を行う。
  int cur_id = 0;
  std::map<int,int>::iterator it;
  for(it=mapSolverSize.begin();it!=mapSolverSize.end();it++){
    int solver_id = it->first;
    int size = it->second;
    if (size <= 0) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "solver node size is zero. [solverID=%d, size=%d].", solver_id, size);
      return false;
    }
    // ノード番号リスト作成
    int *nodeid_list = new int[size];
    for (j=0; j<size; j++) {
      nodeid_list[j] = cur_id;
      cur_id++;
      if (cur_id >= world_size) cur_id = 0;
    }

    SPL_Comm newComm;
    SPL_Group newGrp;
    /// ソルバのMPIグループ、MPIコミュニケータを作成する（SPL_COMM_WORLDに属するすべてのプロセスで実行する）
    if (splCreateCommGroup(
                  size,
                  nodeid_list,
                  &newComm,
                  &newGrp) != SplTrue) {
      // spl error
      m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
      // sph error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "splCreateCommGroup fault.");
      return false;
    }

    /// プロセスグループ情報の作成を行う。
    ProcessGroup* proc_grp = new ProcessGroup(solver_id, size, nodeid_list);

    /// 作成したグループ、コミュニケータを設定する。
    // グループに属していないプロセスの場合、コミュニケータ(newComm)はSPL_COMM_NULLとなる。
    proc_grp->setCommGroup(newComm, newGrp);

    // ソルバグループ情報リストに追加する。
    if (!m_solverGroup.insert( std::make_pair(solver_id, proc_grp)).second) {
       // error
       m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_solverGoup. [solverID=%d].", solver_id);
       return false;
    }
  }

  return true;
}


/**
 * コンフィグレーション要素からカプラのプロセスグループを作成する。.
 * プロセスグループ情報（プロセス番号リスト）の作成を行う。
 * カプラのプロセスグループを作成する。
 * @param     void
 * @return    success(=true)/fault(=false)
 */
bool sphSphere::createProcessCouplerGroup()
{
  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_configElem is null.");
    return false;
  }

  char msg[1024];
  /// 起動プロセス数を取得する
  int world_size;
  int myID;
  SPL_Comm_size (SPL_COMM_WORLD, &world_size);
  SPL_Comm_rank(SPL_COMM_WORLD, &myID);

  sphElemUtils elemUtils(m_configElem);

  register int i;
  /// カプラ毎の割当プロセスを取得する。
  int coupler_count = elemUtils.getCouplerCount();
  for (i=0; i<coupler_count; i++) {
    // <SphCoupler>要素の取得を行う。
    const sphElem* coup_elem = elemUtils.getCouplerElem(i);
    if (coup_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "coupler index = %d.", i);
      return false;
    }
    // カプラID
    int coupler_id = coup_elem->getAttrInt(SPH_CFG_ATTR_ID);

    // SphCoupler/SphDataCopy/SphRefObj要素の取得を行う。
    const sphElem* ref_elem = coup_elem->getElem(2, SPH_CFG_NAME_DATACOPY, SPH_CFG_NAME_REFOBJ);
    // SphCoupler/SphDataCopy/SphUdtObj要素の取得を行う。
    const sphElem* udt_elem = coup_elem->getElem(2, SPH_CFG_NAME_DATACOPY, SPH_CFG_NAME_UDTOBJ);
    if (ref_elem == NULL || udt_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "%s or %s, coupler index = %d.", SPH_CFG_NAME_REFOBJ, SPH_CFG_NAME_UDTOBJ, i);
      return false;
    }

    /// データ参照元、更新先のソルバIDを取得する。
    int ref_solverId = ref_elem->getAttrInt(SPH_CFG_ATTR_SOLVERID);
    int udt_solverId = udt_elem->getAttrInt(SPH_CFG_ATTR_SOLVERID);

    /// ソルバIDから作成済みのプロセスグループを取得する。
    const ProcessGroup* ref_procgrp = getProcessGroupOfSolver(ref_solverId);
    const ProcessGroup* udt_procgrp = getProcessGroupOfSolver(udt_solverId);
    if (ref_procgrp == NULL || udt_procgrp == NULL) {
      // error
      m_errHandler->SPH_WARNING_INFO(SPH_WARNING_NULL_POINTER, "not found solver group, ref_solverId=%d,upt_solverId=%d.", ref_solverId, udt_solverId);
      continue;
    }

    /// ソルバプロセスグループのIDを結合する
    int size = 0;
    int* nodeid_list = mergeIdList(ref_procgrp->m_nodesize,
                                   ref_procgrp->m_nodelist,
                                   udt_procgrp->m_nodesize,
                                   udt_procgrp->m_nodelist,
                                   size);
    if (size == 0 || nodeid_list == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create merge id list, ref_solverId=%d,upt_solverId=%d.", ref_solverId, udt_solverId);
      return false;
    }

    /// カプラ全体のプロセスグループ = 参照ソルバグループ + 更新ソルバグループ
    SPL_Comm couplerComm;
    SPL_Group couplerGroup;
    /// カプラのMPIグループ、MPIコミュニケータを作成する
    if (splCreateCommGroup(
                      size,
                      nodeid_list,
                      &couplerComm,
                      &couplerGroup) != SplTrue) {
      // spl error
      m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
      // sph error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "splCreateCommGroupUnion fault.");
      return false;
    }

    /// プロセスグループ情報の作成を行う。
    ProcessGroup *coupler_proc_grp = new ProcessGroup(coupler_id, size, nodeid_list);
    if (couplerComm != SPL_COMM_NULL || !splIsParallel()) {
      coupler_proc_grp->setCommGroup(couplerComm, couplerGroup);
    }
    else {
      SPL_Group_free(&couplerGroup);
    }
    // カプラグループ情報リストに追加する。
    if (!m_couplerGroup.insert( std::make_pair(coupler_id, coupler_proc_grp)).second) {
      if (coupler_proc_grp != NULL) delete coupler_proc_grp;
      // error
       m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_couplerGoup. [couplerID=%d].", coupler_id);
       return false;
    }
  }

  return true;
}


/**
 * ソルバIDのプロセスグループ情報を返す。
 * @param     solverID          ソルバID
 * @return    プロセスグループ情報
 */
const sphSphere::ProcessGroup* sphSphere::getProcessGroupOfSolver(int solverID) const
{
  /// ソルバID検索
  std::map<int, ProcessGroup* >::const_iterator it = m_solverGroup.find(solverID);
  if (it == m_solverGroup.end()) {
    return NULL;
  }
  /// プロセスグループ情報を返す
  return it->second;
}

/**
 * カプラIDのプロセスグループ情報を返す。
 * @param     couplerID          カプラID
 * @param     id                リストIndex
 *                              0:カプラ全体グループ
 *                              1:カプラ参照グループ
 *                              2:カプラ更新グループ
 * @return    プロセスグループ情報
 */
const sphSphere::ProcessGroup* sphSphere::getProcessGroupOfCoupler(int couplerID) const
{
  /// カプラID検索
  std::map<int, ProcessGroup* >::const_iterator it = m_couplerGroup.find(couplerID);
  if (it == m_couplerGroup.end()) {
    return NULL;
  }
  /// プロセスグループ情報を返す
  return it->second;
}


/**
 * ２つのIDリストを結合する。.
 * @param[in]     fst_idcnt           第１IDリスト数
 * @param[in]     fst_idlist          第１IDリスト
 * @param[in]     snd_idcnt           第２IDリスト数
 * @param[in]     snd_idlist          第２IDリスト
 * @param[out]    mrg_idcnt           結合IDリスト数
 * @return        結合IDリスト
 */
int* sphSphere::mergeIdList(
              int fst_idcnt, const int* fst_idlist,
              int snd_idcnt, const int* snd_idlist,
              int& mrg_idcnt)
{
  mrg_idcnt = 0;
  int max_count = fst_idcnt + snd_idcnt;
  if (max_count == 0) {
    return NULL;
  }

  register int i, j;
  int* id_list = new int[max_count];
  memset(id_list, 0x00, max_count);

  if (fst_idcnt > 0 && fst_idlist != NULL) {
    memcpy(id_list, fst_idlist, fst_idcnt*sizeof(int));
    mrg_idcnt = fst_idcnt;

    if (snd_idcnt > 0 && snd_idlist != NULL) {
      for (i=0; i<snd_idcnt; i++) {
        bool found = false;
        for (j=0; j<mrg_idcnt; j++) {
          if (snd_idlist[i] == id_list[j]) {
            found = true;
          }
        }
        if (!found) {
          id_list[mrg_idcnt] = snd_idlist[i];
          mrg_idcnt++;
        }
      }
    }
  }
  else if (snd_idcnt > 0 && snd_idlist != NULL) {
    memcpy(id_list, snd_idlist, snd_idcnt*sizeof(int));
    mrg_idcnt = snd_idcnt;
  }

  int *mrg_idlist = NULL;
  if (mrg_idcnt > 0) {
    mrg_idlist = new int[mrg_idcnt];
    memcpy(mrg_idlist, id_list, mrg_idcnt*sizeof(int));
  }

  if (id_list != NULL) delete [] id_list;

  return mrg_idlist;
}


/**
 * Vsphereコンフィグレーションであるかチェックする。
 * @return        true:Vsphereコンフィグレーション
 */
bool sphSphere::isClassicConfig()
{
  return (m_sklconfig != NULL);
}


/**
 * Sphereコンフィグレーションであるかチェックする。
 * @return        true:Sphereコンフィグレーション
 */
bool sphSphere::isSphereConfig()
{
  return (m_configElem != NULL);
}


