/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphErrorHandler.C
 * Error Handler Class Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphErrorHandler.C
 * エラーハンドラークラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */


#include "base/sphErrorHandler.h"

/// sphConfig スタティックメンバーの初期化
sphErrorHandler *sphErrorHandler::m_instance = NULL;


/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ.
 * <メッセージ番号, メッセージ文字列>マップを作成する。
 */
sphErrorHandler::sphErrorHandler() : m_lastErr(NULL)
{
  m_errStack.clear();
  register int i;
  for (i=0; i<SPL_ERR_LASTCODE; i++) {
    const char* msg = spl_get_define_errmsg(i);
    if (msg != NULL) {
      if (!m_msgNumList.insert( std::make_pair(i, std::string(msg))).second) {
        // error
        spl_set_errmsg(SPH_ERR_STL, "can not insert m_msgNumList. [no=%d, msg=%s].", i, msg);
      }
    }
    else {
      break;
    }
  }
  for (i=0; i<SPH_ERR_LASTCODE; i++) {
    const char* msg = shp_error_list[i].message;
    if (msg != NULL) {
      if (!m_msgNumList.insert( std::make_pair(shp_error_list[i].msg_no, std::string(msg))).second ) {
        // error
        spl_set_errmsg(SPH_ERR_STL, "can not insert m_msgNumList. [no=%d, msg=%s].", shp_error_list[i].msg_no, msg);
      }
    }
    else {
      break;
    }
  }
#ifdef SPH_ERRORLEVEL_WARNING
  m_sphBlockWarning = false;
  m_tmpBlockWarning = false;
#else
  m_sphBlockWarning = true;
  m_tmpBlockWarning = true;
#endif

  return;
}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ
 */
sphErrorHandler::~sphErrorHandler()
{
  std::deque<sphMsg*>::const_iterator it = m_errStack.begin();
  while( it != m_errStack.end() ) {
    sphMsg* msg = *it;
    if (msg != NULL) {
      delete msg;
    }
    it++;
  }
  m_errStack.clear();
}


/**
 * \~english
 * get sphErrorHandler instance for Singleton
 *
 * \~japanese
 * シングルトンクラスであるsphErrorHandlerのオブジェクトを取得する。
 */
sphErrorHandler* sphErrorHandler::getInstance()
{
  if (m_instance == NULL) {
    m_instance = new sphErrorHandler();

    // プログラムが終了時に破棄処理を呼ぶ。
    atexit(&destruct);
  }

  return m_instance;
}


/**
 * \~english
 * destruct sphErrorHandler instance for Singleton
 *
 * \~japanese
 * シングルトンクラスであるsphErrorHandlerのオブジェクトの破棄を行う。
 */
void sphErrorHandler::destruct()
{
  if (m_instance != NULL) {
    delete m_instance;
  }
  m_instance = 0;

  return;
}


/**
 * \~english
 * get error message from message list
 * @param      err_no        error number
 * @return    error message
 *
 * \~japanese
 * メッセージリストからメッセージを取得する。
 * @param      err_no        エラー番号
 * @return    エラーメッセージ
 */
std::string sphErrorHandler::getMassage(int err_no) const
{
  std::map<int, std::string>::const_iterator it = m_msgNumList.find(err_no);
  if (it == m_msgNumList.end ()) {
    return std::string();
  }
  return (*it).second;
}


/**
 * \~english
 * insert error message(sphMsg)
 * @param      err_msg        error message object
 * @return     void
 *
 * \~japanese
 * エラーメッセージを追加する。(sphMsg).
 * メッセージオブジェクトをスタックに追加する。/n
 * エラーメッセージがなければ、リストから設定する。
 * @param      err_msg        メッセージオブジェクト
 * @return     void
 */
void sphErrorHandler::error(sphMsg* err_msg)
{
  if (err_msg == NULL) return;
  if (err_msg->m_msgNo == 0) return;

  // メッセージタイプを設定する。
  err_msg->m_type = MSGTYPE_ERROR;

  // メッセージ番号に対応したメッセージを設定する。
  if (err_msg->m_message.empty()) {
    err_msg->m_message = getMassage(err_msg->m_msgNo);
  }

  // エラーメッセージスタックに追加する。
  m_errStack.push_front(err_msg);

  // 最終メッセージ
  m_lastErr = err_msg;

  // 制限数以上のエラーメッセージは削除する
  removeStackEnd();

  return;
};


/**
 * \~english
 * insert error message(msg_no)
 * @param      msg_no        error message number
 * @param      filename      file name
 * @param      lineno          file line number
 * @param        funcname      function name
 * @return     void
 *
 * \~japanese
 * エラーメッセージを追加する。(エラーメッセージ番号).
 * メッセージオブジェクトを生成してスタックに追加する。
 * @param      msg_no        エラーメッセージ番号
 * @param      filename      ファイル名
 * @param      lineno          ファイル行番号
 * @param        funcname      関数名
 * @return     void
 */
void sphErrorHandler::error(int msg_no, const char* filename, int lineno, const char* funcname)
{
  if (msg_no == 0) return;

  // メッセージクラスを生成する。
  sphMsg* err_msg = new sphMsg();
  err_msg->setError(msg_no, filename, lineno, funcname);
  error(err_msg);
  return;
};


/**
 * \~english
 * insert error message(include detail)
 * @param      msg_no        error message number
 * @param      filename      file name
 * @param      lineno          file line number
 * @param        funcname      function name
 * @param        format        detail format
 * @param        ...            detail information parameter
 * @return     void
 *
 * \~japanese
 * エラーメッセージを追加する。(詳細メッセージ付き).
 * メッセージオブジェクトを生成してスタックに追加する。
 * @param      msg_no        エラーメッセージ番号
 * @param      filename      ファイル名
 * @param      lineno          ファイル行番号
 * @param        funcname      関数名
 * @param        format        詳細メッセージフォーマット
 * @param        ...            詳細メッセージパラメータ
 * @return     void
 */
void sphErrorHandler::error(int msg_no, const char* filename, int lineno, const char* funcname, const char* format, ...)
{
  if (msg_no == 0) return;

  // 詳細メッセージを取得する。
  va_list ap;
  va_start(ap, format);
  char buf[1024];
  vsprintf(buf, format, ap);                  // vasprintf to vsprintf at 2010/05/10 by @hira

  // メッセージクラスを生成する。
  sphMsg* error_msg = new sphMsg();
  error_msg->setError(msg_no, filename, lineno, funcname, buf);
  error(error_msg);

  va_end(ap);
  return;
}


/**
 * \~english
 * insert error message(msg_no and detail info)
 * @param      msg_no        error message number
 * @param      info        detail message
 * @return     void
 *
 * \~japanese
 * エラーメッセージを追加する。(エラーメッセージ番号, 詳細メッセージ).
 * メッセージオブジェクトを生成してスタックに追加する。
 * @param      msg_no        エラーメッセージ番号
 * @param      info             詳細メッセージ
 * @return     void
 */
void sphErrorHandler::error_info(int msg_no, const char* info)
{
  if (msg_no == 0) return;

  // メッセージクラスを生成する。
  sphMsg* error_msg = new sphMsg();
  error_msg->m_msgNo = msg_no;
  error_msg->m_info = std::string(info);
  error(error_msg);

  return;
}


/**
 * \~english
 * insert warning message(sphMsg)
 * @param      warning_msg        message object
 * @return     void
 *
 * \~japanese
 * 警告メッセージを追加する。(sphMsg)/n
 * 警告メッセージがなければ、リストから設定する。
 * @param      warning_msg        メッセージオブジェクト
 * @return     void
 */
void sphErrorHandler::warning(sphMsg* warning_msg)
{
  // 警告メッセージのブロックフラグがセットされていれば、警告メッセージは受け取らない。
  if (m_tmpBlockWarning) return;
  if (warning_msg == NULL) return;
  if (warning_msg->m_msgNo == 0) return;

  // メッセージタイプを設定する。
  warning_msg->m_type = MSGTYPE_WARNING;

  // メッセージ番号に対応したメッセージを設定する。
  if (warning_msg->m_message.empty()) {
    warning_msg->m_message = getMassage(warning_msg->m_msgNo);
  }

  // 警告メッセージスタックに追加する。
  m_errStack.push_front(warning_msg);

  // 最終メッセージ
  m_lastErr = warning_msg;

  // 制限数以上のエラーメッセージは削除する
  removeStackEnd();

  return;
}


/**
 * \~english
 * insert waring message(msg_no)
 * @param      msg_no        error message number
 * @param      filename      file name
 * @param      lineno          file line number
 * @param        funcname      function name
 * @return     void
 *
 * \~japanese
 * 警告メッセージを追加する。(警告メッセージ番号).
 * メッセージオブジェクトを生成してスタックに追加する。
 * @param      msg_no        警告メッセージ番号
 * @param      filename      ファイル名
 * @param      lineno          ファイル行番号
 * @param        funcname      関数名
 * @return     void
 */
void sphErrorHandler::warning(int msg_no, const char* filename, int lineno, const char* funcname)
{
  // 警告メッセージのブロックフラグがセットされていれば、警告メッセージは受け取らない。
  if (m_tmpBlockWarning) return;

  if (msg_no == 0) return;

  // メッセージクラスを生成する。
  sphMsg* warning_msg = new sphMsg();
  warning_msg->setWarning(msg_no, filename, lineno, funcname);
  warning(warning_msg);
  return;
}


/**
 * \~english
 * insert waring message(include detail)
 * @param      msg_no        waring message number
 * @param      filename      file name
 * @param      lineno          file line number
 * @param        funcname      function name
 * @param        format        detail format
 * @param        ...            detail information parameter
 * @return     void
 *
 * \~japanese
 * 警告メッセージを追加する。(詳細メッセージ付き).
 * メッセージオブジェクトを生成してスタックに追加する。
 * @param      msg_no        警告メッセージ番号
 * @param      filename      ファイル名
 * @param      lineno          ファイル行番号
 * @param        funcname      関数名
 * @param        format        詳細メッセージフォーマット
 * @param        ...            詳細メッセージパラメータ
 * @return     void
 */
void sphErrorHandler::warning(int msg_no, const char* filename, int lineno, const char* funcname, const char* format, ...)
{
  // 警告メッセージのブロックフラグがセットされていれば、警告メッセージは受け取らない。
  if (m_tmpBlockWarning) return;

  if (msg_no == 0) return;

  // 詳細メッセージを取得する。
  va_list ap;
  va_start(ap, format);
  char buf[1024];
  vsprintf(buf, format, ap);                  // vasprintf to vsprintf at 2010/05/10 by @hira

  // メッセージクラスを生成する。
  sphMsg* warning_msg = new sphMsg();
  warning_msg->setWarning(msg_no, filename, lineno, funcname, buf);
  warning(warning_msg);

  va_end(ap);
  return;
}


/**
 * \~english
 * get last error number
 * @param      void
 * @return     last error number
 *
 * \~japanese
 * 最終メッセージのエラー番号を取得する。
 * @param      void
 * @return     最終エラー番号
 */
int sphErrorHandler::getLastErrorNo() const
{
  if (m_lastErr == NULL) {
    return 0;
  }
  return m_lastErr->m_msgNo;
}


/**
 * \~english
 * check to exist error and warning message
 * @param      void
 * @return     true:exist message/false:not exist message
 *
 * \~japanese
 * エラー・警告メッセージが登録されているかチェックする。
 * @param      void
 * @return     true:エラー・警告メッセージが存在する/false:エラー・警告メッセージが存在しない
 */
bool sphErrorHandler::exists() const
{
  if (m_errStack.size() <= 0) {
    return false;
  }
  return true;
}


/**
 * \~english
 * check to exist error message
 * @param      void
 * @return     true:exist message/false:not exist message
 *
 * \~japanese
 * エラーメッセージが登録されているかチェックする。
 * @param      void
 * @return     true:エラーメッセージが存在する/false:エラーメッセージが存在しない
 */
bool sphErrorHandler::existsError() const
{
  int cnt = 0;
  std::deque<sphMsg*>::const_iterator it = m_errStack.begin();
  while( it != m_errStack.end() ) {
    sphMsg* msg = *it;
    if (msg != NULL) {
      if (msg->m_type == MSGTYPE_ERROR) {
        cnt++;
        return true;
      }
    }
    it++;
  }
  return false;
}


/**
 * \~english
 * check to exist warning message
 * @param      void
 * @return     true:exist message/false:not exist message
 *
 * \~japanese
 * 警告メッセージが登録されているかチェックする。
 * @param      void
 * @return     true:警告メッセージが存在する/false:警告メッセージが存在しない
 */
bool sphErrorHandler::existsWarning() const
{
  int cnt = 0;
  std::deque<sphMsg*>::const_iterator it = m_errStack.begin();
  while( it != m_errStack.end() ) {
    sphMsg* msg = *it;
    if (msg != NULL) {
      if (msg->m_type == MSGTYPE_WARNING) {
        cnt++;
        return true;
      }
    }
    it++;
  }
  return false;
}


/**
 * \~english
 * print stack message to stderr
 * @param      void
 * @return     void
 *
 * \~japanese
 * メッセージスタックを標準エラー出力に出力する。
 * @param      void
 * @return     void
 */
void sphErrorHandler::printStackTrace() const
{

  std::deque<sphMsg*>::const_iterator it = m_errStack.begin();
  while( it != m_errStack.end() ) {
    sphMsg* msg = *it;
    if (msg != NULL) {
      std::cerr << msg->getPrintMessage() << std::endl;
    }
    it++;
  }
  return;
}


/**
 * \~english
 * clear stack message
 * @param      void
 * @return     void
 *
 * \~japanese
 * メッセージスタックをクリアする。
 * @param      void
 * @return     void
 */
void sphErrorHandler::clear()
{
  m_errStack.clear();
}


/**
 * \~english
 * remove limit over message in message stack
 * @param      void
 * @return     void
 *
 * \~japanese
 * メッセージスタックから制限数以上のメッセージを削除する
 * @param      void
 * @return     void
 */
void sphErrorHandler::removeStackEnd()
{
  if (m_errStack.size() < SPH_ERROR_STACK_LIMIT) {
    return;
  }

  std::deque< sphMsg* >::iterator it = m_errStack.begin() + SPH_ERROR_STACK_LIMIT;
  m_errStack.erase(it, m_errStack.end());
}


/**
 * \~english
 * set detail information in last message
 * @param      format        detail format
 * @param      ...            detail information parameter
 * @return     void
 *
 * \~japanese
 * 最終メッセージに詳細情報を設定する。
 * @param        format        詳細メッセージフォーマット
 * @param        ...            詳細メッセージパラメータ
 * @return     void
 */
void sphErrorHandler::setInfo(const char* format, ...)
{
  va_list ap;
  va_start(ap, format);
  char buf[1024];
  vsprintf(buf, format, ap);                  // vasprintf to vsprintf at 2010/05/10 by @hira
  m_lastErr->m_info = buf;
  va_end(ap);
};



/**
 * 警告メッセージは無視して破棄を行う。
 * コンフィグレーションの検索メソッドからの警告メッセージを無視する。
 * @param      void
 * @return     void
 */
void sphErrorHandler::blockWarning()
{
  m_blockCount++;
  m_tmpBlockWarning = true;
  return;
}


/**
 * 警告メッセージのブロックフラグを元に戻す。
 * SPH_ERRORLEVEL_ERRORの設定に戻す。
 * @param      void
 * @return     void
 */
void sphErrorHandler::releaseWarning()
{
  m_blockCount--;
  if (m_blockCount > 0) {
    // まだ警告メッセージブロック中
    return;
  }
  m_tmpBlockWarning = m_sphBlockWarning;
  return;
}


/**
 * 警告メッセージのブロックフラグを取得する。
 * SPH_ERRORLEVEL_ERRORにて設定されたブロックフラグを取得する。
 * @param      void
 * @return     警告メッセージのブロックフラグ（true:警告メッセージは無視する。/false:警告メッセージを受け取る。）
 */
bool sphErrorHandler::getDefaultBlockWarning()
{
  return m_sphBlockWarning;
}



/**
 * エラー発生ファイル情報をセットする。
 * @param      msg_no        エラーメッセージ番号
 * @param      filename      ファイル名
 * @param      lineno          ファイル行番号
 * @param        funcname      関数名
 * @param        format        詳細メッセージフォーマット
 * @param        ...            詳細メッセージパラメータ
 * @return     void
 */
void sphErrorHandler::setFileInfo(const char* filename, int lineno, const char* funcname)
{
  _error_filename = filename;
  _error_lineno = lineno;
  _error_funcname = funcname;

  return;
}


/**
 * エラーメッセージを追加する。(詳細メッセージ付き).
 * メッセージオブジェクトを生成してスタックに追加する。
 * @param      msg_no        エラーメッセージ番号
 * @param        format        詳細メッセージフォーマット
 * @param        ...            詳細メッセージパラメータ
 * @return     void
 */
void sphErrorHandler::error_fmt(int msg_no, const char* format, ...)
{

  if (msg_no == 0) return;

  // 詳細メッセージを取得する。
  va_list ap;
  va_start(ap, format);
  char buf[1024];
  vsprintf(buf, format, ap);                  // vasprintf to vsprintf at 2010/05/10 by @hira

  // メッセージクラスを生成する。
  sphMsg* error_msg = new sphMsg();
  error_msg->setError(msg_no, _error_filename, _error_lineno, _error_funcname, buf);
  error(error_msg);

  va_end(ap);

  return;
}



/**
 * 警告メッセージを追加する。(詳細メッセージ付き).
 * メッセージオブジェクトを生成してスタックに追加する。
 * @param      msg_no        警告メッセージ番号
 * @param        format        詳細メッセージフォーマット
 * @param        ...            詳細メッセージパラメータ
 * @return     void
 */
void sphErrorHandler::warning_fmt(int msg_no, const char* format, ...)
{
  // 警告メッセージのブロックフラグがセットされていれば、警告メッセージは受け取らない。
  if (m_tmpBlockWarning) return;

  if (msg_no == 0) return;

  // 詳細メッセージを取得する。
  va_list ap;
  va_start(ap, format);
  char buf[1024];
  vsprintf(buf, format, ap);                  // vasprintf to vsprintf at 2010/05/10 by @hira

  // メッセージクラスを生成する。
  sphMsg* warning_msg = new sphMsg();
  warning_msg->setWarning(msg_no, _error_filename, _error_lineno, _error_funcname, buf);
  warning(warning_msg);

  va_end(ap);
  return;
}


