/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphConfigController.C
 * Configuration Control Class Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphConfigController.C
 * コンフィギュレーション制御クラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */

#include "config/sphConfigController.h"

/// sphConfigController スタティックメンバーの初期化
sphConfigController *sphConfigController::m_instance = NULL;

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphConfigController::sphConfigController():m_rootElem(NULL), m_sklconfig(NULL)
{
}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ
 */
sphConfigController::~sphConfigController()
{
  if (m_rootElem != NULL) {
    delete m_rootElem;
    m_rootElem = NULL;
  }
  if (m_sklconfig != NULL) {
    delete m_sklconfig;
    m_sklconfig = NULL;
  }
}


/**
 * \~english
 * get sphConfigController instance for Singleton
 *
 * \~japanese
 * シングルトンクラスであるsphConfigのオブジェクトを取得する。
 */
sphConfigController* sphConfigController::getInstance()
{
  if (m_instance == NULL) {
    m_instance = new sphConfigController();

    // プログラムが終了時に破棄処理を呼ぶ。
    atexit(&destruct);
  }

  return m_instance;
}


/**
 * \~english
 * destruct sphConfigController instance for Singleton
 *
 * \~japanese
 * シングルトンクラスであるsphConfigのオブジェクトの破棄を行う。
 */
void sphConfigController::destruct()
{
  if (m_instance != NULL) {
    delete m_instance;
  }
  m_instance = NULL;

  return;
}


/**
 * \~english
 * read Configuration XML file
 * @param      fname        configuration filename
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * コンフィグレーションXMLファイルからパラメータの読み込みを行う。
 * @param      fname        コンフィグレーションXMLファイル名
 * @return    success(=true)/fault(=false)
 */
bool sphConfigController::readConfigFile(const char* fname)
{
  int flg = 0, ret = 0;
  m_xmlFileStack.clear();
  m_filename = fname;
  char real_fname[MAXPATHLEN];

  int myID = 0;
  if (SPL_Comm_rank(SPL_COMM_WORLD, &myID) != SPL_MPI_SUCCESS) {
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Comm_rank error.");
    return false;
  }
  xmlDocPtr xmlDoc = NULL;
  if( myID == 0 ){
    // read Configuration XML file
    xmlDoc = readConfigFile(fname, NULL);
    if (!xmlDoc) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_READ_FILE_XML, "file name is %s", fname);
      flg = 1;
    }
    // 絶対パスの取得
    realpath(fname, real_fname);
    m_filename = real_fname;
  }

  if (!splIsParallel()) {
    if (flg != 0) return false;

    // Vsphereコンフィグレーションであるかチェックする。
    if (isClassicConfig(xmlDocGetRootElement(xmlDoc))) {
      m_sklconfig = new SklCfg::SklSolverConfig();
      m_sklconfig->SklParseConfig(xmlDocGetRootElement(xmlDoc));
      return true;
    }
    else {
      // XMLのパース
      m_rootElem = new sphElem();
      m_rootElem->parseElem(xmlDocGetRootElement(xmlDoc));
      return true;
    }
  }

  if (SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD) != SPL_MPI_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Allreduce error.");
    if (xmlDoc) xmlFreeDoc(xmlDoc);
    return false;
  }
  if( ret != 0 ) {
    if (xmlDoc) xmlFreeDoc(xmlDoc);
    // error
    m_errHandler->SPH_ERROR(SPH_ERR_ALLREDUCE_SUM);
    return false;
  }

  int len = 0;
  xmlChar* bufDoc = NULL;
  if( myID == 0 ){
    xmlDocDumpMemory(xmlDoc, &bufDoc, &len);
  }

  // broadcast data length
  if (SPL_Bcast(&len, 1, SPL_INT, 0, SPL_COMM_WORLD) != SPL_MPI_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Bcast error for data length.");
    if (xmlDoc) xmlFreeDoc(xmlDoc);
    if (bufDoc) xmlFree(bufDoc);
    return false;
  }

  size_t bufSize = len * sizeof(xmlChar) + 1;
  if( myID != 0 ){
    bufDoc = (xmlChar *) xmlMalloc(bufSize);
  }

  if( !bufDoc ){
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MEMORY_ALLOCATE, "bufDoc is null(allocate size=%d)", len+1);
    flg = 1;
  }

  if (SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD) != SPL_MPI_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Allreduce error.");
    if (xmlDoc) xmlFreeDoc(xmlDoc);
    if (bufDoc) xmlFree(bufDoc);
    return false;
  }
  if( ret != 0 ) {
    // error
    m_errHandler->SPH_ERROR(SPH_ERR_ALLREDUCE_SUM);
    if (xmlDoc) xmlFreeDoc(xmlDoc);
    if (bufDoc) xmlFree(bufDoc);
    return false;
  }

  // broadcast stream data
  if (SPL_Bcast(bufDoc, bufSize, SPL_CHAR, 0, SPL_COMM_WORLD) != SPL_MPI_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Bcast error of XML buffer, size=%ld.", bufSize);
    if (xmlDoc) xmlFreeDoc(xmlDoc);
    if (bufDoc) xmlFree(bufDoc);
    return false;
  }

  // broadcast stream data
  if (SPL_Bcast(real_fname, MAXPATHLEN, SPL_CHAR, 0, SPL_COMM_WORLD) != SPL_MPI_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Bcast error of real filename, size=%ld.", MAXPATHLEN);
    if (xmlDoc) xmlFreeDoc(xmlDoc);
    if (bufDoc) xmlFree(bufDoc);
    return false;
  }
  m_filename = real_fname;        // 絶対パスのセット

  // parse from stream data
  if( !parseStream((const char*)bufDoc, len) ){
    flg = 1;
  }

  if (SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD) != SPL_MPI_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Allreduce error.");
    if (xmlDoc) xmlFreeDoc(xmlDoc);
    if (bufDoc) xmlFree(bufDoc);
    return false;
  }
  if( ret != 0 ) {
    // error
    m_errHandler->SPH_ERROR(SPH_ERR_ALLREDUCE_SUM);
    if (xmlDoc) xmlFreeDoc(xmlDoc);
    if (bufDoc) xmlFree(bufDoc);
    return false;
  }

  if (xmlDoc) xmlFreeDoc(xmlDoc);
  if (bufDoc) xmlFree(bufDoc);

  xmlCleanupParser();

  return true;
}


/**
 * \~english
 * read Configuration XML file
 * @param      fname        configuration filename
 * @param      curNode        add node pointer
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * コンフィグレーションXMLファイルから分割ファイルの読み込みを行う。
 * @param      fname        コンフィグレーションXMLファイル名
 * @param      curNode      追加ノードポインタ
 * @return    success(=true)/fault(=false)
 */
xmlDocPtr sphConfigController::readConfigFile(const char* fname, xmlNodePtr curNode)
{
  if (fname == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "read xml file name is null");
    return NULL;
  }

  // XMLファイル名のコピーバッファー
  char xml_fname[MAXPATHLEN];
//  realpath(fname, xml_fname);
  strcpy(xml_fname, fname);

  if (isCircularReadFile(xml_fname)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CIRCULAR_READ_FILE_XML, "circular read xml file name is %s", fname);
    return NULL;
  }

  // stack current read file
  m_xmlFileStack.push_back(std::string(xml_fname));

  // read Configuration XML file
  xmlDocPtr xmlDoc = xmlReadFile(xml_fname, NULL, 0);
  if (xmlDoc == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_PARSE_XML, "unable to create Xml document from file(%s).", fname);
    return NULL;
  }

  // root node
  xmlNodePtr rootNode = xmlDocGetRootElement(xmlDoc);

  // ルートノードの要素名のチェックを行う。
  if (curNode == NULL) {
    if (strcasecmp((const char*)rootNode->name, SPH_CFG_ELEM_SPHERECONFIG) != 0) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_XML, "root element name is not %s.(filename=%s).", SPH_CFG_ELEM_SPHERECONFIG, fname);
      return NULL;
    }
  }

  // create XPath
  // XMLファイル検索 <sph_xml_config_file file_name="xml file"/>
  // sph_xml_config_file要素の大文字・小文字は区別しないXPathの作成
  char xpathExpr[MAXPATHLEN];
//  sprintf(xpathExpr, "//%s[translate(@name,'ABCDEFGHIJKLMNOPQRSTUVWXYZ','abcdefghijklmnopqrstuvwxyz')='%s']", SPH_CFG_ELEM_PARAM, SPH_CFG_NAME_XML_CONFIG);
  sprintf(xpathExpr, "//*[translate(local-name(),'ABCDEFGHIJKLMNOPQRSTUVWXYZ','abcdefghijklmnopqrstuvwxyz')='%s']", SPH_CFG_NAME_XML_CONFIG);

  xmlXPathObjectPtr xpathObj = getXPathObject(rootNode, xpathExpr);
  if(xpathObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_EVALUATE_XPATH, "xpath is %s.", xpathExpr);
    xmlFreeDoc(xmlDoc);
    m_xmlFileStack.pop_back();
    return NULL;
  }

  // NodeSet
  xmlNodeSetPtr nodes = xpathObj->nodesetval;

  // get found size
  int size = nodes->nodeNr;
  for (int i=0; i<size; i++) {
    if(nodes->nodeTab[i]->type != XML_ELEMENT_NODE) continue;

    xmlNodePtr cur = nodes->nodeTab[i];
    // file_name
    xmlChar* attrValue = getXmlAttribute(cur, (const xmlChar*)SPH_CFG_ATTR_FILENAME);
    if (!attrValue) {
      // WARNING
      m_errHandler->SPH_ERROR_INFO(SPH_WARNING_NOT_EXIST_ATTRIBUTE, "attribute is not %s.", SPH_CFG_ATTR_FILENAME);
      xmlXPathFreeObject(xpathObj);
      xmlFreeDoc(xmlDoc);
      // pop xml file
      m_xmlFileStack.pop_back();
      return NULL;
    }

    char rel_fname[MAXPATHLEN];
    // 相対パスの組み立て
    strcpy(xml_fname, fname);
    sprintf(rel_fname, "%s/%s", dirname(xml_fname), (const char*)attrValue);
    xmlFree(attrValue);

    // read Configuration XML file
    xmlDocPtr cfgDoc = readConfigFile(rel_fname, cur);
    if (cfgDoc == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_READ_FILE_XML, "file name is %s.", rel_fname);
      xmlXPathFreeObject(xpathObj);
      xmlFreeDoc(xmlDoc);
      // pop xml file
      m_xmlFileStack.pop_back();
      return NULL;
    }

    // copy root node
    xmlNodePtr copyRoot = xmlDocCopyNode(xmlDocGetRootElement(cfgDoc), xmlDoc, 1);
    if (copyRoot == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_PARSE_XML, "unable to create copy node.");
      xmlXPathFreeObject(xpathObj);
      xmlFreeDoc(cfgDoc);
      // pop xml file
      m_xmlFileStack.pop_back();
      return NULL;
    }

    // add copy root node to current node
    xmlAddChild(cur, copyRoot);

    xmlFreeDoc(cfgDoc);
  }

  xmlXPathFreeObject(xpathObj);

  // pop xml file
  m_xmlFileStack.pop_back();

  return xmlDoc;
}


/**
 * \~english
 * write XML file from xmlDocPtr
 * @param      fname        output file name
 * @param      one_file     true:all elements output 1 file
 *                          false:sph_xml_config_file element output separate file
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * xmlDocPtrをXMLファイルに出力する。
 * @param      fname        出力ファイル名
 * @param      one_file     true:sph_xml_config_file配下の要素もまとめて指定ファイルに出力する。
 *                          false:sph_xml_config_file配下の要素は別ファイルに出力する。
 * @return    success(=true)/fault(=false)
 */
bool sphConfigController::writeConfigFile(const char* fname, bool one_file)
{
  if (m_rootElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_rootElem is null.");
    return false;
  }

  // ルートノードファイル出力
  if (!m_rootElem->writeXmlFile(fname, one_file)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_WRITE_FILE_XML, "file name is %s.", fname);
    return false;
  }

  // sph_xml_config_file配下の要素を別ファイルに出力する。
  if (!one_file) {
    if (!m_rootElem->writeXmlFileInCfgFile(fname)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_WRITE_FILE_XML, "error is writeXmlFileInCfgFile, basefile=%s.", fname);
      return false;
    }
  }

  return true;
}



/**
 * \~english
 * parse XML stream
 * @param      stream    xml stream pointer
 * @param     sz        stream size
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * XML streamをパースする。.
 * 配信されたXML streamをパースして、sphElemを作成する。
 * @param      stream    XMLストリームポインタ
 * @param     sz        XMLストリームサイズ
 * @return    success(=true)/fault(=false)
 */
bool sphConfigController::parseStream(const char* stream, const size_t sz)
{

  if( (sz <= 0) || !stream ){
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "sz or stream is null.");
    return false;
  }

  xmlDocPtr xmlDoc = ::xmlParseMemory(stream, sz);
  if ( !xmlDoc ){
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_PARSE_XML, "unable to create Xml document by xmlParseMemory.");
    return false;
  }

  // ドキュメントのファイル名（絶対パス）を設定する。
  xmlDoc->URL = (const xmlChar*)m_filename.c_str();

  xmlNodePtr cur = xmlDocGetRootElement(xmlDoc);
  if (cur == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_PARSE_XML, "unable to get root element by xmlDocGetRootElement.");
    xmlDoc->URL = NULL;
    xmlFreeDoc(xmlDoc);
    xmlDoc = NULL;
    return false;
  }

  // Vsphereコンフィグレーションであるかチェックする。
  if (isClassicConfig(cur)) {
    m_sklconfig = new SklCfg::SklSolverConfig();
    m_sklconfig->SklParseConfig(cur);
  }
  else {
    // XMLのパース
    m_rootElem = new sphElem();
    m_rootElem->parseElem(cur);
  }
  xmlDoc->URL = NULL;
  xmlFreeDoc(xmlDoc);
  xmlDoc = NULL;

  return true;
}


/**
 * \~english
 * search XPath Node Pointer by Xpath in xml document
 * @param     xpathExpr      XPath string
 * @param     curNode         current node pointer
 * @param     silent           silent mode
 * @return    xmlXPathObjectPtr<br>
 *
 * \~japanese
 * XMLドキュメントからXPath 文字列によって指定されたノードPointerを取得する。
 * @param     xpathExpr      XPath 文字列
 * @param     curNode         検索対象ノードポインタ
 * @param     silent           silent mode<br>
 *                             true:Not foundエラーを出力しない/false:すべてのエラー出力を行う。
 * @return    検索結果ノードポインタ
 */
xmlXPathObjectPtr sphConfigController::getXPathObject(xmlNodePtr curNode, const char* xpathExpr) const
{
  if (curNode == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "xmlNodePtr is null.");
    return NULL;
  }
  if (curNode->doc == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "xmlNodePtr->doc is null.");
    return NULL;
  }
  if (!xpathExpr) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "xpathExpr is null.");
    return NULL;
  }
  xmlDocPtr xmlDoc = curNode->doc;
  xmlXPathContextPtr xpathCtx = xmlXPathNewContext(xmlDoc);
  if (xpathCtx == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_PARSE_XML, "unable to create new XPath context.");
    return NULL;
  }
  xpathCtx->node = curNode;

  // create xpath object
  xmlXPathObjectPtr xpathObj = xmlXPathEvalExpression((const xmlChar*)xpathExpr, xpathCtx);
  if(xpathObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_EVALUATE_XPATH, "xpath=%s", xpathExpr);
    return NULL;
  }

  xmlXPathFreeContext(xpathCtx);

  // NodeSet
  xmlNodeSetPtr nodes = xpathObj->nodesetval;
  if (!nodes) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_XPATH, "xpath=%s", xpathExpr);
    xmlXPathFreeObject(xpathObj);
    xpathObj = NULL;
    return NULL;
  }

  return xpathObj;
}


/**
 * \~english
 * test circular read xml file
 * @param      fname      current read file
 * @return    circular(=true)/not circular(=false)
 *
 * \~japanese
 * 読み込みXMLファイルが読み込み済みであり、循環読み込みとなるかチェックする。.
 * m_xmlFileStackには、子要素方向に読込XMLファイルがスタックされており\n
 * m_xmlFileStackにスタックされているファイルと同一ファイルを読込を行うと循環参照となり\n
 * 無限ループが発生する。\n
 * m_xmlFileStackにスタック済みのファイルであるかチェックを行う。
 * @param      fname      読み込みXMLファイル
 * @return    読み込み済み(=true)/初回読み込み(=false)
 */
bool sphConfigController::isCircularReadFile(const char* fname)
{
  for (std::vector<std::string>::iterator itr=m_xmlFileStack.begin();
        itr!=m_xmlFileStack.end();++itr){
    std::string readFile = *(itr);
    if (strcmp(fname, readFile.c_str()) == 0) {
      return true;
    }
  }
  return false;
}


/**
 * \~english
 * get root element
 * @param      void
 * @return    Root sphElem Object of configuration XML
 *
 * \~japanese
 * コンフィグレーションXMLのルート要素を取得する。
 * @param      void
 * @return    ルート要素sphElemオブジェクト
 */
const sphElem* sphConfigController::getRootElem() const
{
  return m_rootElem;
}


/**
 * Vsphereコンフィグレーションであるかチェックする。
 * @param      ルート要素
 * @return    true:Vsphereコンフィグレーション
 */
bool sphConfigController::isClassicConfig(xmlNodePtr root)
{
  return !isSphereConfig(root);
}

/**
 * Sphereコンフィグレーションであるかチェックする。
 * @param      ルート要素
 * @return    true:Sphereコンフィグレーション
 */
bool sphConfigController::isSphereConfig(xmlNodePtr root)
{
  const char* attr_name = SPH_CFG_ATTR_VER;
  char* attr_value = (char*)(xmlGetProp(root, (const xmlChar*)attr_name));
  if (attr_value == NULL) return false;

  // '0' = 0x30
  int ver_num = attr_value[0] - 0x30;
  if (ver_num != 2) {
    return false;
  }

  return true;

}

/**
 * Vsphereコンフィグレーションであるかチェックする。
 * @return    true:Vsphereコンフィグレーション
 */
bool sphConfigController::isClassicConfig()
{
  if (m_sklconfig != NULL) return true;
  return false;
}

/**
 * Sphereコンフィグレーションであるかチェックする。
 * @param      ルート要素
 * @return    true:Sphereコンフィグレーション
 */
bool sphConfigController::isSphereConfig()
{
  if (m_rootElem != NULL) return true;
  return false;
}


/**
 * Vsphereコンフィグレーションを取得する。
 * @return    Vsphereコンフィグレーション
 */
const SklCfg::SklSolverConfig* sphConfigController::getSklSolverConfig() const
{
  return m_sklconfig;
}


