/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphElem.C
 * XML Element class file
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphConfigController.C
 * XML要素クラスファイル
 * @author     hira
 * @date       2009/02/20
 */
#include "config/sphElem.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphElem::sphElem()
    : m_attrList(NULL), m_childElemList(NULL), m_subTreeList(NULL), m_parentElem(NULL)
{
  /// 初期化を行う。
  initialize();
}


/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphElem::sphElem(const char* elem_name)
    : m_attrList(NULL), m_childElemList(NULL), m_subTreeList(NULL), m_parentElem(NULL)
{
  /// 初期化を行う。
  initialize();
  m_elemName = std::string(elem_name);
}


/**
 * \~english
 * Constructor
 * @param       parentElem        parent element
 * @param       subTreeList     sub tree element list
 *
 * \~japanese
 * コンストラクタ
 * @param       parentElem       親要素
 * @param       subTreeList     外部XMLファイル要素リスト
 */
sphElem::sphElem(const sphElem* parentElem,
                 std::vector<sphElem*>* subTreeList)
    : m_attrList(NULL), m_childElemList(NULL), m_subTreeList(NULL), m_parentElem(NULL)
{
  /// 初期化を行う。
  initialize();

  m_parentElem = parentElem;

  /// 外部XMLファイルリストを設定する。
  if (subTreeList != NULL) {
    m_subTreeList = subTreeList;
  }
}


/**
 * コピーコンストラクタ
 * @param       elem            コピー要素
 */
sphElem::sphElem(const sphElem* elem)
  : m_attrList(NULL), m_childElemList(NULL), m_subTreeList(NULL), m_parentElem(NULL)
{

  /// 初期化を行う。
  initialize();

  m_elemName = elem->m_elemName;

  if (elem->m_attrList != NULL) {
    std::map<std::string, std::string>::const_iterator it_attr = elem->m_attrList->begin();
    for (;  it_attr != elem->m_attrList->end(); it_attr++) {
      std::string attr_name = (*it_attr).first;
      std::string attr_value = (*it_attr).second;
      addAttribute(attr_name, attr_value);
    }
  }

  if (elem->m_childElemList != NULL) {
    std::deque<sphElem*>::const_iterator it_child = elem->m_childElemList->begin();
    for (; it_child != elem->m_childElemList->end(); it_child++) {
      const sphElem* child_elem = *it_child;
      if (child_elem != NULL) {
        // コピーを追加する。
        m_childElemList->push_back(new sphElem(child_elem));
      }
    }
  }

  m_subTreeList = elem->m_subTreeList;
  m_fileName = elem->m_fileName;
  m_parentElem = elem->m_parentElem;

}


/**
 * 初期化を行う。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphElem::initialize()
{
  m_parentElem = NULL;

  /// 外部XMLファイルリストを生成する。
  m_subTreeList = new std::vector<sphElem*>();

  /// 属性リストを生成する。
  m_attrList = new std::map<std::string, std::string, __lesscasecmp>();

  /// 子要素リストを生成する。
  m_childElemList = new std::deque<sphElem*>();

  return true;
}



/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 * 子要素リスト、サブツリーリストに登録しているsphElemオブジェクトをdeleteする。
 */
sphElem::~sphElem()
{
  // 子要素リストからsphElemオブジェクトを解放する。
  if (m_childElemList != NULL) {
    std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
    while( it_child != m_childElemList->end() ) {
      sphElem* child_elem = *it_child;
      if (child_elem != NULL) {
        // sphElemオブジェクトを解放する。
        delete child_elem;
        child_elem = NULL;
      }
      it_child++;
    }
    m_childElemList->clear();
    delete m_childElemList;
    m_childElemList = NULL;
  }

  // サブツリーリストの削除はルート要素のみ
  if (m_parentElem == NULL) {
    if (m_subTreeList != NULL) {
      m_subTreeList->clear();
      // サブツリーリストの解放
      delete m_subTreeList;
      m_subTreeList = NULL;
    }
  }

  return;
}


/**
 * \~english
 * parse XML Node
 * @param      node    xml node pointer
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * parse XML Node
 * @param      node    XMLノード
 * @return    success(=true)/fault(=false)
 */
bool sphElem::parseElem(xmlNodePtr node)
{
  if (node == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "xmlNodePtr is null.");
    return false;
  }
  if (node->name == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "element name is null.");
    return false;
  }
  if (m_subTreeList == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_subTreeList is null.");
    return false;
  }
  if (node->type != XML_ELEMENT_NODE) {
    return false;
  }

  // 要素名の設定
  m_elemName = (const char*)node->name;

  // 属性リストの設定
  xmlAttrPtr prop = node->properties;
  while( prop ) {
    const char* attr_name = (const char*)prop->name;
    char* attr_value = (char*)(xmlGetProp(node, (const xmlChar*)attr_name));
    std::string str_value(attr_value);
    xmlFree(attr_value);

    // 属性リストの追加
    addAttribute(attr_name, str_value);

    prop = prop->next;
  }

  // <sph_xml_config_file>要素、又はルート要素の場合、XMLファイル名を設定する。
  std::string cfg_xml_fname = "";
  if (isSphCfgFileElem() || m_parentElem == NULL) {
    // ルートXML要素から現在のXML要素までの外部ファイルパス（相対パス）を取得する。
    // 外部ファイルが記述されているファイル（パス）の違いを吸収する為に、ルートからの相対パスを取得する。
    cfg_xml_fname = getCanonicalizePath(node);

    // 外部ファイル名を設定する。
    m_fileName = cfg_xml_fname;
  }

  // 子要素を追加する。
  xmlNodePtr child_node = node->children;
  while (child_node) {
    if (child_node->type == XML_ELEMENT_NODE) {
      // 子要素オブジェクトの生成
      sphElem* elem = new sphElem(this, m_subTreeList);

      // 子要素のパースを行う。
      if (elem->parseElem(child_node)) {
        // パース済み要素を子要素リストに追加する。
        m_childElemList->push_back(elem);

        // <sph_xml_config_file>要素の場合は、サブツリーリストに追加する。
        if (!cfg_xml_fname.empty()) {
          // サブツリーリストに追加する。
          if (!addSubTreeList(elem)) {
             // error
             m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_subTreeList. [fname=%s].", cfg_xml_fname.c_str());
             delete elem;
             return false;
          }
        }
      }
      else {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_PARSE_XML, "xml parse error in child element.");
        // パース失敗により解放する。
        delete elem;
      }
    }
    child_node = child_node->next;
  }

  return true;
}


/**
 * \~english
 * write XML file from this element
 * @param      fname        output file name
 * @param      one_file     true:all elements output 1 file
 *                          false:<sph_xml_config_file> element is not output
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * この要素配下をXMLファイルに出力する。
 * @param      fname        出力ファイル名
 * @param      one_file     true:<sph_xml_config_file>配下の要素もまとめて指定ファイルに出力する。
 *                          false:<sph_xml_config_file>配下の要素は出力しない。
 * @return    success(=true)/fault(=false)
 */
bool sphElem::writeXmlFile(const char* fname, bool one_file) const
{

  // XMLドキュメント生成
  xmlDocPtr xmlDoc = xmlNewDoc((const xmlChar*)"1.0");

  // sphElemからxmlNodeを生成して、XMLドキュメントに追加する。
  if (!createElem(xmlDoc, one_file)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CREATE_ELEMENT, "fault createElem.");
    return false;
  }
  if (xmlDoc == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "xmlDoc is null.");
    return false;
  }
  // XMLファイル保存
  int returns = xmlSaveFormatFile(fname, xmlDoc, 1);
  if (returns <= 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_WRITE_FILE_XML, "can not write xmlfile, filename is %s.", fname);
    xmlFreeDoc(xmlDoc);
    return false;
  }

  xmlFreeDoc(xmlDoc);

  return true;
}


/**
 * \~english
 * write XML file from <sph_xml_config_file> element
 * @param      base_fname        output path
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * <sph_xml_config_file>要素をXMLファイルに出力する。.
 * <sph_xml_config_file>要素に記述されているファイル名にて出力する。（ファイル名は変更しない）\n
 * base_fileからの相対パスに出力する。\n
 * 子要素、孫要素からも<sph_xml_config_file>要素を検索して、それぞれ記述されているファイルに出力する。
 * @param      base_fname        出力パス
 * @return    success(=true)/fault(=false)
 */
bool sphElem::writeXmlFileInCfgFile(const char* base_fname) const
{

  if (base_fname == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "base_fname is null.");
    return false;
  }

  // XMLファイル名のコピーバッファー
  char xml_fname[MAXPATHLEN];

  // 子要素から<sph_xml_config_file>要素を検索する。
  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    const sphElem* child_elem = *it_child;
    if (child_elem != NULL) {
      // <sph_xml_config_file>要素であるかチェックする。
      if (child_elem->isSphCfgFileElem()) {
        const char* cfg_fname = child_elem->getAttrString(SPH_CFG_ATTR_FILENAME);

        char rel_fname[MAXPATHLEN];
        // 相対パスの組み立て
        strcpy(xml_fname, base_fname);
        sprintf(rel_fname, "%s/%s", dirname(xml_fname), (const char*)cfg_fname);

        // <sph_xml_config_file>要素の子要素を出力する。
        const sphElem* cfg_elem = child_elem->getElement();
        if (cfg_elem != NULL) {
          // XMLファイル出力
          cfg_elem->writeXmlFile(rel_fname, false);

          // 子要素から<sph_xml_config_file>要素を検索, 出力する。
          if (!cfg_elem->writeXmlFileInCfgFile(rel_fname)) {
            // error
            m_errHandler->SPH_ERROR_INFO(SPH_ERR_WRITE_FILE_XML, "base_fname is %s.", rel_fname);
            return false;
          }
        }
      }
      else {
        if (!child_elem->writeXmlFileInCfgFile(base_fname)) {
          // error
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_WRITE_FILE_XML, "base_fname is %s.", base_fname);
          return false;
        }
      }
    }
    it_child++;
  }

  return true;
}


/**
 * \~english
 * create xml element in xml document
 * @param      xmlDoc    XML Document
 * @param      one_file     true:all elements output 1 file
 *                          false:<sph_xml_config_file> element output separate file
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * XMLドキュメントにノード要素を追加、作成する。
 * @param      xmlDoc    XMLドキュメント
 * @param      one_file    true:<sph_xml_config_file>配下の要素もxmlDocに追加する。
 *                          false:<sph_xml_config_file>配下の要素は追加しない。
 * @return    success(=true)/fault(=false)
 */
bool sphElem::createElem(xmlDocPtr xmlDoc, bool one_file) const
{
  if (xmlDoc == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "xmlDoc is null.");
    return false;
  }
  if (m_elemName.length() <= 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_elemName is empty.");
    return false;
  }

  // 新規XMLノードの生成（XMLドキュメント直下のルート要素）
  xmlDoc->children = xmlNewDocNode(xmlDoc, NULL, (const xmlChar*)(m_elemName.c_str()), NULL);

  // XML属性値の追加
  std::map<std::string, std::string>::const_iterator it_attr = m_attrList->begin();
  while( it_attr != m_attrList->end() ) {
    xmlSetProp(xmlDoc->children, (const xmlChar*)((*it_attr).first.c_str()), (const xmlChar*)((*it_attr).second.c_str()));
    it_attr++;
  }

  // one_fileフラグ=false(追加しない） and <sph_xml_config_file>要素の場合は、終了
  if (!one_file && isSphCfgFileElem()) {
    return true;
  }

  // 子要素を追加する。
  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    const sphElem* child_elem = *it_child;
    if (child_elem != NULL) {
      if (!child_elem->createElem(xmlDoc->children, one_file)) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_CREATE_ELEMENT, "fault createElem.");
        return false;
      }
    }
    it_child++;
  }

  return true;
}


/**
 * \~english
 * create xml element in xml node
 * @param      parent_node    XML node
 * @param      one_file     true:all elements output 1 file
 *                          false:<sph_xml_config_file> element output separate file
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * XMLノードの子要素にノード要素を追加、作成する。
 * @param      parent_node    XMLノード
 * @param      one_file    true:<sph_xml_config_file>配下の要素も追加する。
 *                          false:<sph_xml_config_file>配下の要素は追加しない。
 * @return    success(=true)/fault(=false)
 */
bool sphElem::createElem(xmlNodePtr parent_node, bool one_file) const
{
  if (m_elemName.length() <= 0) {
    return false;
  }

  // XMLノードの子ノードの生成
  xmlNodePtr cur_node = xmlNewChild(parent_node, NULL, (const xmlChar*)(m_elemName.c_str()), NULL);

  // XML属性値の追加
  std::map<std::string, std::string>::const_iterator it_attr = m_attrList->begin();
  while( it_attr != m_attrList->end() ) {
    xmlSetProp(cur_node, (const xmlChar*)((*it_attr).first.c_str()), (const xmlChar*)((*it_attr).second.c_str()));
    it_attr++;
  }

  // one_fileフラグ=false(追加しない） and <sph_xml_config_file>要素の場合は、終了
  if (!one_file && isSphCfgFileElem()) {
    return true;
  }

  // 子要素を追加する。
  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    const sphElem* child_elem = *it_child;
    if (child_elem != NULL) {
      if (!child_elem->createElem(cur_node, one_file)) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_CREATE_ELEMENT, "fault createElem.");
        return false;
      }
    }
    it_child++;
  }

  return true;
}


/**
 * \~english
 * set xml element name
 * @param      name    XML Element name
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * XML要素名を設定する。
 * @param      name    XML要素名
 * @return    success(=true)/fault(=false)
 */
bool sphElem::setElemName(std::string name)
{
  m_elemName = name;
  return true;
}


/**
 * XML属性を追加する。(string).
 * @param      name    XML属性名
 * @param      value    XML属性値(string)
 * @return    success(=true)/fault(=false)
 */
bool sphElem::addAttribute(std::string name, std::string value)
{
  // 属性リストの追加
  if (!m_attrList->insert( std::make_pair(name, value)).second) {
     // error
     m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_attrList. [name=%s, value=%s].", name.c_str(), value.c_str());
     return false;
  }
  return true;
}


/**
 * XML属性を設定する。(string).
 * @param      name    XML属性名
 * @param      value    XML属性値(string)
 * @return    success(=true)/fault(=false)
 */
void sphElem::setAttribute(std::string name, std::string value)
{
  // 属性リストの設定（キーが存在しなければinsertを自動的に行う)
  (*m_attrList)[name] = value;
}


/**
 * \~english
 * set xml attribute(int).
 * @param      name    XML attribute name
 * @param      value    XML attribute value(int)
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * XML属性を設定する。(int).
 * @param      name    XML属性名
 * @param      value    XML属性値(int)
 * @return    success(=true)/fault(=false)
 */
bool sphElem::addAttribute(std::string name, int value)
{
  return addAttribute(name, (long long)value);
}

/**
 * \~english
 * set xml attribute(long).
 * @param      name    XML attribute name
 * @param      value    XML attribute value(long)
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * XML属性を設定する。(long).
 * @param      name    XML属性名
 * @param      value    XML属性値(long)
 * @return    success(=true)/fault(=false)
 */
bool sphElem::addAttribute(std::string name, long value)
{
  return addAttribute(name, (long long)value);
}


/**
 * \~english
 * set xml attribute(long long).
 * @param      name    XML attribute name
 * @param      value    XML attribute value(long long)
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * XML属性を設定する。(long long).
 * @param      name    XML属性名
 * @param      value    XML属性値(long long)
 * @return    success(=true)/fault(=false)
 */
bool sphElem::addAttribute(std::string name, long long value)
{
  std::ostringstream os;
  os << value;
  std::string str_value = os.str();
  addAttribute(name, str_value);

  return true;
}


/**
 * \~english
 * add xml children sphElem Class
 * @param      name    XML Element name
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 子要素のsphElemクラスを追加する。
 * @param      name    XML要素名
 * @return    success(=true)/fault(=false)
 */
bool sphElem::addChildElem(sphElem* child)
{
  m_childElemList->push_back(child);
  return true;
}


/**
 * \~english
 * add external XML file list<external filename, sphElem class>
 * @param      elem    sphElem class
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 外部XMLファイルリスト<ファイル名, sphElemクラス>に追加する。
 * @param      elem        sphElemクラス:値
 * @return    success(=true)/fault(=false)
 */
bool sphElem::addSubTreeList(sphElem* elem)
{

  if (m_subTreeList == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_subTreeList is null.");
    return false;
  }

  // 外部XMLファイルリストの追加
  m_subTreeList->push_back(elem);

  return true;
}


/**
 * \~english
 * get xml element name
 * @param      void
 * @return    XML Element name
 *
 * \~japanese
 * XML要素名を取得する。
 * @param      void
 * @return    XML要素名
 */
const char* sphElem::getElemName() const
{
  return m_elemName.c_str();
}


/**
 * \~english
 * get xml attribute value (string)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(string)
 * @param      name    属性名
 * @return    XML属性値
 */
std::string sphElem::getAttribute(std::string name) const
{
  if (name.empty()) {
    return std::string();
  }
  std::map<std::string, std::string>::const_iterator it = m_attrList->find(name);
  if (it == m_attrList->end ()) {
    m_errHandler->SPH_WARNING_INFO(SPH_WARNING_NOT_EXIST_ATTRIBUTE, "attribute name is %s, <%s>", name.c_str(), m_elemName.c_str());
    return std::string();
  }
  return (*it).second;
}


/**
 * \~english
 * get xml attribute value (string)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(string)
 * @param      name    属性名
 * @return    XML属性値
 */
const char* sphElem::getAttrChar(std::string name) const
{
  if (name.empty()) {
    return NULL;
  }
  std::map<std::string, std::string>::const_iterator it = m_attrList->find(name);
  if (it == m_attrList->end ()) {
    m_errHandler->SPH_WARNING_INFO(SPH_WARNING_NOT_EXIST_ATTRIBUTE, "attribute name is %s, <%s>", name.c_str(), m_elemName.c_str());
    return NULL;
  }
  return (*it).second.c_str();
}

/**
 * \~english
 * get xml attribute value (long long)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(long long)
 * @param      name    属性名
 * @return    XML属性値
 */
long long   sphElem::getAttrInteger (std::string name) const
{
  std::string str_value = getAttribute(name);
  if (str_value.empty()) {
    return 0;
  }
  std::istringstream is(str_value);
  long long value = 0;
  if (str_value.substr(0, 2) == "0x") {
    is >> std::hex >> value;
  }
  else {
    is >> std::dec >> value;
  }

  return value;
}


/**
 * \~english
 * get xml attribute value (double)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(double).
 * オーバーフロー、数値変換エラーチェックは行わない。
 * @param      name    属性名
 * @return    XML属性値
 */
double       sphElem::getAttrReal (std::string name) const
{
  std::string str_value = getAttribute(name);
  if (str_value.empty()) {
    return 0;
  }
  double real = 0.0;
//  real = atof(str_value.c_str());
  std::istringstream is(str_value);
  is >> real;

  return real;
}


/**
 * \~english
 * get xml attribute value (const char*)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(const char*)
 * @param      name    属性名
 * @return    XML属性値
 */
const char* sphElem::getAttrString(const char* name) const
{
  const char* str_value = getAttrChar(std::string(name));
  if (str_value == NULL || strlen(str_value) == 0) return NULL;

  return str_value;
}


/**
 * \~english
 * get xml attribute value (short)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(short).
 * オーバーフロー、数値変換エラーチェックは行わない。(戻り値=0)
 * @param      name    属性名
 * @return    XML属性値
 */
short       sphElem::getAttrShort (const char* name) const
{
  return (short)getAttrInteger(std::string(name));
}


/**
 * \~english
 * get xml attribute value (int)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(int).
 * オーバーフロー、数値変換エラーチェックは行わない。(戻り値=0)
 * @param      name    属性名
 * @return    XML属性値
 */
int         sphElem::getAttrInt   (const char* name) const
{
  return (int)getAttrInteger(std::string(name));
}


/**
 * \~english
 * get xml attribute value (long)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(long).
 * オーバーフロー、数値変換エラーチェックは行わない。(戻り値=0)
 * @param      name    属性名
 * @return    XML属性値
 */
long        sphElem::getAttrLong  (const char* name) const
{
  return (long)getAttrInteger(std::string(name));
}

/**
 * \~english
 * get xml attribute value (long long)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(long long).
 * オーバーフロー、数値変換エラーチェックは行わない。(戻り値=0)
 * @param      name    属性名
 * @return    XML属性値
 */
long long   sphElem::getAttrLLong (const char* name) const
{
  return (long long)getAttrInteger(std::string(name));
}


/**
 * \~english
 * get xml attribute value (float)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(float).
 * オーバーフロー、数値変換エラーチェックは行わない。(戻り値=0.0)
 * @param      name    属性名
 * @return    XML属性値
 */
float       sphElem::getAttrReal4 (const char* name) const
{
  return (float)getAttrReal(std::string(name));
}

/**
 * \~english
 * get xml attribute value (double)
 * @param      name    attribute name
 * @return    XML attribute value
 *
 * \~japanese
 * XML属性値を取得する。(double).
 * オーバーフロー、数値変換エラーチェックは行わない。(戻り値=0.0)
 * @param      name    属性名
 * @return    XML属性値
 */
double      sphElem::getAttrReal8 (const char* name) const
{
  return (double)getAttrReal(std::string(name));
}


/**
 * XML属性値を取得する。(SphDCType).
 * @param      name    属性名
 * @return    XML属性値
 */
SphDCType      sphElem::getAttrSphDcType (const char* name) const
{
  std::string str_value = getAttribute(std::string(name));

  // 小文字変換
//std::transform(str_value.begin(), str_value.end(), str_value.begin(), tolower);  // delete at 20100511 by @hira
  str_value = toLower(str_value);

  if (str_value == "scalar1d") return SPH_DC_ARRAY_1D;       // 1Dデータクラス データ配列{i}
  if (str_value == "scalar2d") return SPH_DC_ARRAY_2D;       // 2Dデータクラス データ配列{i, j}
  if (str_value == "scalar3d") return SPH_DC_ARRAY_3D;       // 3Dデータクラス データ配列{i, j, k}
  if (str_value == "vector1d") return SPH_DC_ARRAY_1DN;      // 1Dデータクラス データ配列{i, l}
  if (str_value == "vector1drv") return SPH_DC_ARRAY_1DNRV;  // 1Dデータクラス データ配列{i, l}
  if (str_value == "vector2d") return SPH_DC_ARRAY_2DN;      // 2Dデータクラス データ配列{i, j, l}
  if (str_value == "vector2drv") return SPH_DC_ARRAY_2DNRV;  // 2Dデータクラス データ配列{l, i, j}
  if (str_value == "vector3d") return SPH_DC_ARRAY_3DN;      // 3Dデータクラス データ配列{i, j, k, l}
  if (str_value == "vector3drv") return SPH_DC_ARRAY_3DNRV;  // 3Dデータクラス データ配列{l, i, j, k}

  if (str_value == "unst_node")     return SPH_DC_UNST_NODE;     // 非構造データクラス（節点、スカラ）
  if (str_value == "unst_node_n")   return SPH_DC_UNST_NODE_N;   // 非構造データクラス（節点、ベクトル）
  if (str_value == "unst_node_nrv") return SPH_DC_UNST_NODE_NRV; // 非構造データクラス（節点、ベクトル、反転）
  if (str_value == "unst_elem")     return SPH_DC_UNST_ELEM;     // 非構造データクラス（要素、スカラ）
  if (str_value == "unst_elem_n")   return SPH_DC_UNST_ELEM_N;   // 非構造データクラス（要素、ベクトル）
  if (str_value == "unst_elem_nrv") return SPH_DC_UNST_ELEM_NRV; // 非構造データクラス（要素、ベクトル、反転）

  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "attribute=%s, value=%s, %s.", name, str_value.c_str(), this->debugPrintElem().c_str());

  return SPH_DC_INVALID;
}


/**
 * XML属性値を取得する。(SPL_Datatype).
 * @param      name    属性名
 * @return    XML属性値
 */
SPL_Datatype   sphElem::getAttrSplDataType (const char* name) const
{
  std::string str_value = getAttribute(std::string(name));

  // 小文字変換
//std::transform(str_value.begin(), str_value.end(), str_value.begin(), tolower);  // delete at 20100511 by @hira
  str_value = toLower(str_value);

  if (str_value == "char") return SPL_CHAR;                     // char型データ
  if (str_value == "unsigned_char") return SPL_UNSIGNED_CHAR;   // unsigned charデータ型
  if (str_value == "short") return SPL_SHORT;                   // short型データ型
  if (str_value == "unsigned_short") return SPL_UNSIGNED_SHORT; // unsigned shortデータ型
  if (str_value == "int") return SPL_INT;                       // int型データ型
  if (str_value == "unsigned_int") return SPL_UNSIGNED;         // unsigned intデータ型
  if (str_value == "long") return SPL_LONG;                     //  long型データ型
  if (str_value == "unsigned_long") return SPL_UNSIGNED_LONG;   // unsigned longデータ型
  if (str_value == "longlong") return SPL_LONG_LONG;            // long long型データ型
  if (str_value == "float") return SPL_FLOAT;                   // float型データ型
  if (str_value == "double") return SPL_DOUBLE;                 // double型データ型

  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "attribute=%s, value=%s, %s.", name, str_value.c_str(), this->debugPrintElem().c_str());

  return SPL_DATATYPE_NULL;
}


/**
 * XML属性値を取得する。(SphCrdDef)
 * @param      name    属性名
 * @return    XML属性値
 */
SphCrdDef sphElem::getAttrSphCrdDef(const char* name) const
{

  std::string str_value = getAttribute(std::string(name));

  // crddefは任意設定項目
  if (str_value.empty()) return CRDDEF_REGULAR;

  // 小文字変換
//std::transform(str_value.begin(), str_value.end(), str_value.begin(), tolower);  // delete at 20100511 by @hira
  str_value = toLower(str_value);

  if (str_value == "regular") return CRDDEF_REGULAR;                     // regular
  if (str_value == "collocate") return CRDDEF_COLLOCATE;   // collocate
  if (str_value == "staggered_1") return CRDDEF_STAGGERED1;                   // staggered_1
  if (str_value == "staggered_2") return CRDDEF_STAGGERED2; // staggered_2

  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "attribute=%s, value=%s, %s.", name, str_value.c_str(), this->debugPrintElem().c_str());

  return CRDDEF_UNKNOWN;
}


/**
 * XML属性値からファイルパスを取得する。(file path)
 * @param[out]      file_path        ファイルパス
 * @param[in]      name    属性名
 * @return    success(=true)/fault(=false)
 */
bool sphElem::getAttrFilePath(char* file_path, const char* name) const
{
  std::string str_value;
  if (name != NULL) {
    str_value = getAttribute(std::string(name));
    if (str_value.empty()) return false;

    // 相対パスか絶対パスか？
    if (!PathIsRelative(str_value.c_str())) {
      // 絶対パス
      strcpy(file_path, str_value.c_str());
      return true;
    }
  }

  // 親要素からXMLファイル名を取得する。
  std::string path = getUpstreamFileName();
  char tmp_path[MAXPATHLEN], *dname;
  strcpy(tmp_path, path.c_str());
  dname = dirname(tmp_path);

  if (name != NULL) {
    // パスの組み立て
    sprintf(file_path, "%s/%s", dname, str_value.c_str());
  }
  else {
    strcpy(file_path, dname);
  }

  return true;
}

/**
 * \~english
 * get xml attribute value (bool)
 * @param      name    attribute name
 * @param      default_value    default value
 * @return      true:XML attribute value="true","on","yes" or not exit attribute name
 *              false:other
 *
 * \~japanese
 * XML属性値を取得する。(bool).
 * @param      name    属性名
 * @param      default_value    属性が存在しなかった時のデフォルト値
 * @return     true:XML属性値="true","on","yes"、又は属性名が存在しない。
 *             false:以外
 */
bool      sphElem::isAttrValue  (const char* name, bool default_value) const
{
  // bool値取得は、警告メッセージを出力しない。
  m_errHandler->blockWarning();
  std::string str_value = getAttribute(std::string(name));
  // 警告メッセージを元に戻す
  m_errHandler->releaseWarning();

  if (str_value.empty()) return default_value;
  if (strcasecmp(str_value.c_str(), "true") == 0) return true;
  if (strcasecmp(str_value.c_str(), "on") == 0) return true;
  if (strcasecmp(str_value.c_str(), "yes") == 0) return true;
  if (strcasecmp(str_value.c_str(), "with") == 0) return true;

  return false;
}


/**
 * \~english
 * get sphElem class by element name(std::string)
 * @param      elem_name    search element name
 * @return     sphElem Class
 *
 * \~japanese
 * 子要素から検索要素名と一致する要素クラスを取得する(std::string).
 * 複数の子要素が検索要素名と一致しても、最初に一致した要素のみを返す。
 * 要素名の大文字・小文字の区別はせず比較する
 * @param      elem_name    検索要素名
 * @return    要素クラス
 */
const sphElem* sphElem::getElement(std::string elem_name) const
{
  if (m_childElemList->size() == 0) {
    return NULL;
  }

  if (elem_name.empty()) {
    return m_childElemList->at(0);
  }
  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    sphElem* child_elem = *it_child;
    if (child_elem->equalsElemName(elem_name)) {
      return child_elem;
    }
    it_child++;
  }

  // エラー
  m_errHandler->SPH_WARNING_INFO(SPH_WARNING_NOT_EXIST_ELEMENT, "element name is %s", elem_name.c_str());

  return NULL;
}


/**
 * \~english
 * get sphElem class by element tree (serial)
 * @param      searchElem    search element tree
 * @return     sphElem Class
 *
 * \~japanese
 * 要素ツリーから検索要素ツリーと一致する要素を取得する (serial).
 * 検索要素ツリーの定義している子要素、孫要素のツリー構成、属性値と一致する要素を検索する。
 * 要素名、属性名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素の検索も行う。
 * @param      searchElem    検索要素ツリー(const sphElem*)
 * @return    検索結果要素(const sphElem*)
 */
const sphElem* sphElem::getSerialElemBySphElem(const sphElem* searchElem) const
{
  if (searchElem == NULL) return NULL;

  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    sphElem* child_elem = *it_child;

    // 子要素と検索トップ要素の検索
    if (child_elem->isSearchMatch(searchElem)) {
      if (searchElem->m_childElemList->size() == 0) {
        return child_elem;
      }

      // 検索要素に子要素を持つので、検索要素の子要素とこの要素の子要素を検索する。
      // 検索要素に複数の子要素が存在した場合を考慮して、検索要素の子要素分回す。
      // 検索要素に複数の子要素が存在した場合は、OR条件と同じ。
      std::deque<sphElem*>::const_iterator it_search_child = searchElem->m_childElemList->begin();
      while( it_search_child != searchElem->m_childElemList->end() ) {
        const sphElem* searchChildElem = *it_search_child;
        const sphElem* result = child_elem->getSerialElemBySphElem(searchChildElem);
        if (result != NULL) {
          return result;
        }
        it_search_child++;
      }
    }
    it_child++;
  }

  return NULL;
}


/**
 * \~english
 * get sphElem class by element list (serial)
 * @param      elem_name    search element name
 * @return     sphElem Class
 *
 * \~japanese
 * 要素ツリーから検索要素名リストと一致する要素を取得する（serial).
 * 子要素から検索要素名リストと一致する子要素、孫要素を検索する。
 * 要素名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素の検索も行う。
 * @param      elem_name     検索要素名
 * @return    検索結果要素(const sphElem*)
 */
const sphElem* sphElem::getElemBySerial(const char* elem_name) const
{
  return getElemBySerial(1, elem_name);
}


/**
 * \~english
 * get sphElem class by element list (serial)
 * @param      num          search element number include first elem_name
 * @param      elem_name00  first search element name
 * @param      ...          search element name list
 * @return     sphElem Class
 *
 * \~japanese
 * 要素ツリーから検索要素名リストと一致する要素を取得する(serial).
 * 子要素から検索要素名リストと一致する子要素、孫要素を検索する。
 * 要素名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素の検索も行う。
 * @param      num           検索要素名数 (elem_name00を含む)
 * @param      elem_name00   最初の検索要素名
 * @param      ...           ２番目以降の検索要素名リスト
 * @return    検索結果要素(const sphElem*)
 */
const sphElem* sphElem::getElemBySerial(unsigned int num, const char* elem_name00, ...) const
{
  if (num == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "num is zero.");
    return NULL;
  }

  // 引数の要素名リストを作成する。
  va_list ap;
  va_start(ap, num);
  std::vector<std::string> arg_list;
  for(int i=0;i<num;i++) {
    const char* arg = NULL;
    if (i==0) arg = elem_name00;
    else arg = va_arg(ap, const char*);
    if (arg == NULL || strlen(arg) <= 0) break;
    arg_list.push_back(std::string(arg));
  }

  // 検索用sphElemを作成する。
  const sphElem* searchElem = createSearchElem(&arg_list);

  // 要素ツリーによる検索実行
  const sphElem* result = getSerialElemBySphElem(searchElem);

  if (searchElem!= NULL) delete searchElem;

  if (result == NULL) {
    // 検索要素名の組み立て（警告出力用）
    std::string elem_list_str;
    for(int i=0;i<arg_list.size();i++) {
      elem_list_str += arg_list[i] + "-";
    }
    if (elem_list_str.size() > 0) {
      elem_list_str.erase(elem_list_str.size()-1);
    }
    // 警告
    m_errHandler->SPH_WARNING_INFO(SPH_WARNING_NOT_EXIST_ELEMENT, "element names is \"%s\"", elem_list_str.c_str());
  }

  va_end(ap);

  return result;
}


/**
 * 検索要素名リストから要素ツリーを作成する。
 * @param      arg_list       検索要素名リスト
 * @return    検索要素ツリー(const sphElem*)
 */
const sphElem* sphElem::createSearchElem(const std::vector<std::string>* arg_list) const
{

  // 検索用sphElemを作成する。
  // 検索ツリーを組み立てる。
  sphElem* searchElem = NULL;
  sphElem* parentElem = NULL;
  sphElem* childElem = NULL;
  parentElem = searchElem;
  std::string arg;
  for(int i=0;i<arg_list->size();i++) {
    arg = arg_list->at(i);
    if (arg.empty()) {
      break;
    }
    if (searchElem == NULL) {
      searchElem = new sphElem();
      searchElem->setElemName(arg);
      parentElem = searchElem;
    }
    else {
      childElem = new sphElem();
      childElem->setElemName(arg);
      parentElem->addChildElem(childElem);
      parentElem = childElem;
    }
  }

  return searchElem;
}


/**
 * \~english
 * get sphElem class by element name (serial)
 * @param      elem_name    search element name
 * @return     sphElem Class
 *
 * \~japanese
 * 子要素から検索要素名と一致する要素クラスを取得する (serial).
 * 複数の子要素が検索要素名と一致しても、最初に一致した要素のみを返す。
 * 要素名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素の検索も行う。
 * @param      elem_name    検索要素名
 * @return    要素クラス
 */
const sphElem* sphElem::getChildElemFirstBySerial(const char* elem_name) const
{
  if (m_childElemList->size() <= 0) {
    return NULL;
  }
  if (elem_name == NULL) {
    return m_childElemList->at(0);
  }

  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    sphElem* child_elem = *it_child;
    if (child_elem->equalsElemName(elem_name)) {
      return child_elem;
    }
    it_child++;
  }

  // 警告
  m_errHandler->SPH_WARNING_INFO(SPH_WARNING_NOT_EXIST_ELEMENT, "element name is \"%s\"", elem_name);

  return NULL;
}


/**
 * \~english
 * get next sphElem class by element name (serial)
 * @param      elem          search previous sphElem
 * @param      elem_name    search element name
 * @return     sphElem Class
 *
 * \~japanese
 * 子要素から検索要素名と一致する次要素を取得する (serial).
 * 複数の子要素が検索要素名と一致しても、最初に一致した要素のみを返す。
 * 要素名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素の検索も行う。
 * @param      elem          事前要素
 * @param      elem_name    検索要素名
 * @return    要素クラス
 */
const sphElem* sphElem::getChildElemNextBySerial(const sphElem* elem, const char* elem_name) const
{
  bool findFlag = false;
  // 子要素から検索する。
  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    sphElem* child_elem = *it_child;
    if (child_elem == elem) {
      if (elem_name == NULL) {
        // 要素名がNULLであるので、次要素を返す。
        return *(++it_child);
      }
      else {
        findFlag = true;
      }
      child_elem = *(++it_child);
      if (it_child == m_childElemList->end()) {
        return NULL;
      }
    }
    if (findFlag) {
      if (child_elem->equalsElemName(elem_name)) {
        return child_elem;
      }
    }
    it_child++;
  }
  return NULL;
}



/**
 * \~english
 * check equals search element
 * @param      searchElem      search element
 * @returm    true:equals/false:not equals
 *
 * \~japanese
 * 検索sphElemと一致する要素をチェックする。
 * @param      searchElem      検索sphElem
 * @returm    true:一致/false:不一致
 *
 */
bool sphElem::isSearchMatch(const sphElem* searchElem) const
{
  if (!equalsElemName(searchElem->m_elemName)) {
    return false;
  }

  std::map<std::string, std::string>::const_iterator it_attr = searchElem->m_attrList->begin();
  while( it_attr != searchElem->m_attrList->end() ) {
    std::string attr_name = (*it_attr).first;
    std::string attr_value = (*it_attr).second;
    std::string this_value = getAttribute(attr_name);
    if (strcasecmp(attr_value.c_str(), this_value.c_str()) != 0) {
      return false;
    }
    it_attr++;
  }

  return true;
}


/**
 * \~english
 * get first element in all tree by element name
 * @param      elem_name      element name
 * @returm     searched element
 *
 * \~japanese
 * 検索要素名と一致する最初の要素を要素ツリーから検索する.
 * 子要素、孫要素から検索を行う。
 * @param      elem_name      検索要素名
 * @returm     検索結果要素
 *
 */
const sphElem* sphElem::getWildElemFirst(const char* elem_name) const
{
  if (elem_name == NULL) return NULL;

  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    sphElem* child_elem = *it_child;
    if (child_elem->equalsElemName(elem_name)) {
      return child_elem;
    }
    else {
      const sphElem* result = getWildElemFirst(elem_name);
      if (result != NULL) {
        return result;
      }
    }
    it_child++;
  }
  return NULL;
}


/**
 * \~english
 * get next element in all tree by element name
 * @param      elem            previous searched element
 * @param      elem_name      element name
 * @returm     searched element
 *
 * \~japanese
 * 検索要素名と一致する次の要素を要素ツリーから検索する.
 * 子要素、孫要素から検索を行う。
 * @param      elem            事前検索要素
 * @param      elem_name      検索要素名
 * @returm     検索結果要素
 *
 */
const sphElem* sphElem::getWildElemNext(sphElem* elem, const char* elem_name) const
{
  if (elem == NULL) return NULL;
  if (elem_name == NULL) return NULL;

  // 要素ツリーから検索を行う。
  bool findFlag = false;
  const sphElem* result = NULL;
  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    sphElem* child_elem = *it_child;
    // 事前検索要素の検索
    if (child_elem == elem) {
      findFlag = true;
      // 事前検索要素が見つかったので、子要素から検索する。
      result = child_elem->getWildElemFirst(elem_name);
      if (result != NULL) {
        return result;
      }
    }
    else {
      if (findFlag) {
        // 事前検索要素は見つかっているので、次に要素名が一致する要素の検索
        if (child_elem->equalsElemName(elem_name)) {
          return child_elem;
        }
        else {
          // 子要素から検索する。
          result = child_elem->getWildElemFirst(elem_name);
          if (result != NULL) {
            return result;
          }
        }
      }
      else {
        // 事前検索要素が見つかっていないので、子要素から事前検索要素の検索
        result = child_elem->getWildElemNext(elem, elem_name);
        if (result == elem) {
          // 事前検索要素のみ見つかった。
          findFlag = true;
        }
        else if (result != NULL) {
          // 事前検索要素が見つかり、次の要素も見つかった。
          return result;
        }
      }
    }
    it_child++;
  }

  // 次の要素は見つからなかったが、事前検索要素は見つかったので、事前検索要素を返す。
  if (findFlag) {
    return elem;
  }

  // 何も見つからなかった。
  return NULL;
}


/**
 * \~english
 * check this element equals <sph_xml_config_file> element
 * @param      void
 * @returm     true:equals/false:not equals
 *
 * \~japanese
 * <sph_xml_config_file>要素であるかチェックする。
 * 要素名が"sph_xml_config_file"であること。
 * 属性に"file_name"を持っていること。
 * 大文字、小文字は区別しない。
 * @param      void
 * @returm     true:equals/false:not equals
 *
 */
bool sphElem::isSphCfgFileElem() const
{
  if (!equalsElemName(SPH_CFG_NAME_XML_CONFIG)) {
    return false;
  }

  if (getAttrString(SPH_CFG_ATTR_FILENAME) == NULL) {
    return false;
  }

  return true;
}



/**
 * \~english
 * debug print this element and child tree to stdout
 * @param      void
 * @returm     void
 *
 * \~japanese
 * この要素配下の要素ツリーを標準出力する.
 * @param      void
 * @returm     void
 *
 */
void sphElem::debugPrint() const
{

  // XMLドキュメント生成
  xmlDocPtr xmlDoc = xmlNewDoc((const xmlChar*)"1.0");

  // sphElemからxmlNodeを生成して、XMLドキュメントに追加する。
  if (!createElem(xmlDoc, true)) {
    return;
  }
  if (xmlDoc == NULL) {
    return;
  }

  // 作成したXMLドキュメントを標準出力に出力する。
  xmlDocFormatDump(stdout, xmlDoc,1) ;

  xmlFreeDoc(xmlDoc);
  return;
}


/**
 * この要素を文字列に組み立てる.
 * @param      void
 * @returm     要素文字列
 *
 */
std::string sphElem::debugPrintElem() const
{
  std::string elem_str = "<" + m_elemName + " ";

  // XML属性値の追加
  std::map<std::string, std::string>::const_iterator it_attr = m_attrList->begin();
  while( it_attr != m_attrList->end() ) {
    elem_str += (*it_attr).first + "='" + (*it_attr).second + "' ";
    it_attr++;
  }
  elem_str += "/>";

  return elem_str;
}


/**
 * \~english
 * set sub tree list
 * @param      subTreeList    sub tree list pointer
 * @returm     void
 *
 * \~japanese
 * サブツリーリストを設定する。
 * @param      subTreeList      サブツリーリスト
 * @returm     void
 *
 */
void sphElem::setSubTreeList(std::vector<sphElem*>* subTreeList)
{
  m_subTreeList = subTreeList;
}


/**
 * \~english
 * remove element equaled sub tree elements from child element list.
 * @param      void
 * @returm     void
 *
 * \~japanese
 * サブツリーリストに登録している要素を子要素リストから削除する.
 * サブツリーリストもNULLにする。
 * sphElemの重複deleteを防ぐ為に、デストラクタから呼ばれる。
 * @param      void
 * @returm     void
 *
 */
void sphElem::removeSubTreeElem()
{
  if (m_subTreeList == NULL) {
    return;
  }

  // サブツリーリストから登録要素の取得
  std::vector<sphElem*>::const_iterator it_tree = m_subTreeList->begin();
  while( it_tree != m_subTreeList->end() ) {
    sphElem* tree_elem = (*it_tree);

    // サブツリーリストに登録要素されている要素と一致する子要素を検索する。
    // 一致した場合、子要素リストから削除する為に、reverse_iteratorを使用し、末尾から削除する。
    std::deque<sphElem*>::reverse_iterator it_child = m_childElemList->rbegin();
    while( it_child != m_childElemList->rend() ) {
      sphElem* child_elem = *it_child;
      if (child_elem == tree_elem) {
        // サブツリーリストに登録要素されている要素と一致する子要素を削除
        m_childElemList->erase((it_child+1).base());
      }
      it_child++;
    }
    it_tree++;
  }

  m_subTreeList = NULL;

}


/**
 * \~english
 * get canonicalize path
 * @param      node           current xml node
 * @return     path name
 *
 * \~japanese
 * 現在ノードの読込ファイル名を取得する.
 * 現在ノードからルートノードまでの<sph_xml_config_file>要素を走査し、ルートノードからの相対パスを組み立てる。
 * パス名の'./', '../'等を正規化を行う。
 * windowsでは'PathCanonicalize'相当、linuxではrealpathの絶対パス部分の削除相当の処理行う。
 * @param      node            現在ノード
 * @returm    読込ファイル名
 *            読込ファイルが存在しない場合は、空文字を返す。
 */
std::string sphElem::getCanonicalizePath(xmlNodePtr node)
{
  char rel_path[MAXPATHLEN] = {0x00};
  char buf_rpath[MAXPATHLEN] = {0x00};
  int cfg_cnt = 0;

  if (node == NULL) {
    return std::string();
  }

  // 現在のノードから親ノード方向へ走査
  xmlNodePtr parent_node = node;
  while (parent_node && parent_node->name) {

    // <sph_xml_config_file>要素の検索
    if (strcasecmp((const char*)(parent_node->name), SPH_CFG_NAME_XML_CONFIG) == 0) {
      // file_name属性からファイル名の取得
      xmlChar* attr_value = getXmlAttribute(parent_node, (const xmlChar*)SPH_CFG_ATTR_FILENAME);
      if (attr_value != NULL) {
        char fname[MAXPATHLEN];
        strcpy(fname, (const char*)attr_value);
        xmlFree(attr_value);

        // 相対パスの組み立て
        if (cfg_cnt == 0) {
          strcpy(rel_path, fname);
        }
        else {
          strcpy(buf_rpath, rel_path);
          sprintf(rel_path, "%s/%s", dirname(fname), (const char*)buf_rpath);
        }
        cfg_cnt++;
      }
    }
    parent_node = parent_node->parent;
  }

  // ドキュメントの読込ファイル名を追加する。
  if (node->doc->URL != NULL) {
    // ドキュメントの読込ファイル名
    char url_path[MAXPATHLEN] = {0x00};
    strcpy(url_path, (const char*)(node->doc->URL));
    if (strlen(rel_path) == 0) {
      // <sph_xml_config_file>要素の検索ができなかった（無かった）ので、ドキュメントの読込ファイル名を返す。
      return std::string(url_path);
    }
    strcpy(buf_rpath, rel_path);
    sprintf(rel_path, "%s/%s", dirname(url_path), (const char*)buf_rpath);
  }

  if (strlen(rel_path) == 0) {
    return std::string();
  }

  // パス名の'./', '../'等を正規化を行う。
  // 実行パス配下の絶対パスに変換後、実行パスを削除し、ルート要素からの相対パスを取得する。
  // 実行パス配下の絶対パスに変換する。
  normalizePath(rel_path, buf_rpath);

  return std::string(buf_rpath);
}


/**
 * \~english
 * get number of element name in child element (serial)
 * @param      elem_name            element name
 * @returm     count or element
 *
 * \~japanese
 * 要素名を持つ子要素数を取得する (serial)。
 * <sph_xml_config_file>要素の検索も行う。
 * @param      elem_name            要素名
 * @returm     検索結果要素数
 */
int sphElem::countNumOfElemBySerial(const char* elem_name) const
{
  if (elem_name == NULL) {
    return m_childElemList->size();
  }
  return countNumOfElemBySerial(1, elem_name);
}


/**
 * \~english
 * get number of element name list in child element (serial)
 * @param      num          search element number include first elem_name
 * @param      elem_name00  first search element name
 * @param      ...          search element name list
 * @returm     count or element
 *
 * \~japanese
 * 要素名リストを持つ子要素数を取得する (serial)。
 * <sph_xml_config_file>要素の検索も行う。
 * @param      num           検索要素名数 (elem_name00を含む)
 * @param      elem_name00   最初の検索要素名
 * @param      ...           ２番目以降の検索要素名リスト
 * @returm     検索結果要素数
 */
int sphElem::countNumOfElemBySerial(unsigned int num, const char* elem_name00, ...) const
{
  if (num == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "num is zero.");
    return -1;
  }

  // 引数の要素名リストを作成する。
  va_list ap;
  va_start(ap, num);
  std::vector<std::string> arg_list;
  for(int i=0;i<num;i++) {
    const char* arg = NULL;
    if (i==0) arg = elem_name00;
    else arg = va_arg(ap, const char*);
    if (arg == NULL || strlen(arg) <= 0) break;
    arg_list.push_back(std::string(arg));
  }

  // 検索用sphElemを作成する。
  const sphElem* searchElem = createSearchElem(&arg_list);

  // 検索要素ツリーによる要素数の取得を行う
  int count = countNumOfSerialElemBySphElem(searchElem);

  if (searchElem!= NULL) delete searchElem;
  va_end(ap);

  return count;
}


/**
 * \~english
 * get number of element in child element by search element tree  (serial)
 * @param      searchElem      search element tree
 * @returm     count or element
 *
 * \~japanese
 * 検索要素ツリーと一致する子要素数を取得する。 (serial)
 * <sph_xml_config_file>要素の検索も行う。
 * @param      searchElem          検索要素ツリー
 * @returm     検索結果要素数
 */
int sphElem::countNumOfSerialElemBySphElem(const sphElem* searchElem) const
{
  if (searchElem == NULL) {
    return m_childElemList->size();
  }

  int num = 0;

  // 子要素から検索を行う。
  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    sphElem* child_elem = *it_child;

    // 子要素と検索トップ要素の検索
    if (child_elem->isSearchMatch(searchElem)) {
      if (searchElem->m_childElemList->size() == 0) {
        // 検索要素に子要素がなければ、インクリメントして、次のこの要素の子要素をチェックする。
        num++;
      }
      else {
        // 検索要素に子要素を持つので、検索要素の子要素とこの要素の子要素を検索する。
        // 検索要素に複数の子要素が存在した場合を考慮して、検索要素の子要素分回す。
        // 検索要素に複数の子要素が存在した場合は、OR条件と同じ。
        std::deque<sphElem*>::const_iterator it_search_child = searchElem->m_childElemList->begin();
        while( it_search_child != searchElem->m_childElemList->end() ) {
          const sphElem* searchChildElem = *it_search_child;
          int child_num = child_elem->countNumOfSerialElemBySphElem(searchChildElem);
          num += child_num;
          it_search_child++;
        }
      }
    }
    it_child++;
  }

  return num;
}


/**
 * \~english
 * equals Element name, compare ignoring case
 * @param      elem_name        destination element name
 * @returm     true:equal/false:not equal
 *
 * \~japanese
 * 要素名が大文字・小文字を区別せずに一致するかチェックする。
 * @param      elem_name          チェック対象要素名
 * @returm     true:equal/false:not equal
 */
bool sphElem::equalsElemName(std::string elem_name) const
{
  if (elem_name.empty() || m_elemName.empty()) {
    return false;
  }

  if (strcasecmp(m_elemName.c_str(), elem_name.c_str()) != 0) {
    return false;
  }

  return true;
}


/**
 * \~english
 * remove relative string like /./, /../, get normalized path
 * @param      path                target path
 * @param      normalized_path       normalized path
 * @returm     normalized path
 *
 * \~japanese
 *  /./, /../ による参照を解決して、正規化されたパス名を生成する。
 * @param      path          正規化を行うパス
 * @param      normalized_path          正規化されたパス
 * @returm     正規化されたパス
 */
char* sphElem::normalizePath(const char *path, char *normalized_path)
{
  char copy_path[MAXPATHLEN];
  char *new_path = normalized_path;
  char *max_path;
  int n;

  strcpy(copy_path, path);
  path = copy_path;
  max_path = copy_path + MAXPATHLEN - 2;

  if (*path != '/') {
    while (*path == '.') {
      if (*path == '.') {
        if (path[1] == '\0' || path[1] == '/') {
          // "./" delete
          path += 2;
          continue;
        }
      }
      if (path > max_path) {
        return NULL;
      }
      *new_path++ = *path++;
    }
  }
  else {
    *new_path++ = '/';
    path++;
  }

  /* Expand each slash-separated pathname component. */
  while (*path != '\0') {
    /* Ignore stray "/". */
    if (*path == '/') {
      path++;
      continue;
    }
    if (*path == '.') {
      /* Ignore ".". */
      if (path[1] == '\0' || path[1] == '/') {
        path++;
        continue;
      }
      if (path[1] == '.') {
        if (path[2] == '\0' || path[2] == '/') {
          path += 2;
          /* Ignore ".." at root. */
          if (new_path == normalized_path + 1)
            continue;
          /* Handle ".." by backing up. */
          while ((--new_path)[-1] != '/')
            ;
          continue;
        }
      }
    }
    /* Safely copy the next pathname component. */
    while (*path != '\0' && *path != '/') {
      if (path > max_path) {
        return NULL;
      }
      *new_path++ = *path++;
    }
    *new_path++ = '/';
  }
  /* Delete trailing slash but don't whomp a lone slash. */
  if (new_path != normalized_path + 1 && new_path[-1] == '/')
    new_path--;
  /* Make sure it's null terminated. */
  *new_path = '\0';
  return normalized_path;
}



/**
 * \~english
 * check to exist xml attribute
 * @param      name    attribute name
 * @return     true:exist XML attribute / false: not exist
 *
 * \~japanese
 * XML属性が存在するかチェックする。
 * @param      name    属性名
 * @return    true:XML属性が存在する。/ false:XML属性が存在しない。
 */
bool sphElem::existAttribute(const char* name) const
{
  if (name == NULL) {
    return false;
  }
  std::map<std::string, std::string>::const_iterator it = m_attrList->find(std::string(name));
  if (it == m_attrList->end ()) {
    return false;
  }
  return true;
}


/**
 * XML子要素が存在するかチェックする。
 * @param      name    要素名
 * @return    true:XML子要素が存在する。/ false:XML子要素が存在しない。
 */
bool sphElem::existChileElement(const char* name) const
{
  if (name == NULL) {
    return false;
  }

  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    const sphElem* child_elem = *it_child;
    if (child_elem->equalsElemName(name)) {
      return true;
    }
    it_child++;
  }
  return false;
}


/**
 * \~english
 * get root element
 * @param      void
 * @return     root element
 *
 * \~japanese
 * ルート要素を取得する。
 * @param     void
 * @return    root element
 */
const sphElem* sphElem::getRootElem() const
{
  if (m_parentElem == NULL) return this;

  const sphElem* rootElem = m_parentElem;
  while (rootElem != NULL) {
    // 親要素がNULLになるまで検索する。
    if (rootElem->m_parentElem == NULL) {
      break;
    }
    rootElem = rootElem->m_parentElem;
  }

  return rootElem;
}


/**
 * 要素名の親要素を取得する。.
 * 要素名がNULLの場合は、直上の親を返す。
 *
 * @param     elem_name         要素名
 * @return    parent element
 */
const sphElem* sphElem::getParentElem(const char* elem_name) const
{
  if (m_parentElem == NULL) return NULL;
  if (elem_name == NULL) return m_parentElem;

  const sphElem* parentElem = m_parentElem;
  while (parentElem != NULL) {
    if (parentElem->equalsElemName(elem_name)) {
      return parentElem;
    }
    // 親要素がNULLになるまで検索する。
    if (parentElem->m_parentElem == NULL) {
      break;
    }

    parentElem = parentElem->m_parentElem;
  }

  return NULL;
}


/**
 * \~english
 * get sphElem class by element tree
 * @param      searchElem    search element tree
 * @return     sphElem Class
 *
 * \~japanese
 * 要素ツリーから検索要素ツリーと一致する要素を取得する.
 * 検索要素ツリーの定義している子要素、孫要素のツリー構成、属性値と一致する要素を検索する。
 * 要素名、属性名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素は検索対象外とする。
 * @param      searchElem    検索要素ツリー(const sphElem*)
 * @return    検索結果要素(const sphElem*)
 */
const sphElem* sphElem::getExceptElemBySphElem(const sphElem* searchElem, const sphElem* prevElem) const
{
  const sphElem* prev_searchedElem = prevElem;

  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    const sphElem* child_elem = *it_child;

    // <sph_xml_config_file>要素であるかチェックする
    if (child_elem->isSphCfgFileElem()) {
      // <sph_xml_config_file>要素であるので、子要素を取得する。
      child_elem = child_elem->getChildElemFirst();
    }

    // 子要素と検索トップ要素の検索
    if (searchElem == NULL || child_elem->isSearchMatch(searchElem)) {
      if (searchElem == NULL) {
        if (prev_searchedElem == NULL) return child_elem;
        else if (child_elem == prev_searchedElem) prev_searchedElem = NULL;
      }
      else {
        if (searchElem->m_childElemList->size() == 0) {
          if (prev_searchedElem == NULL) return child_elem;
          else if (child_elem == prev_searchedElem) prev_searchedElem = NULL;
        }
      }

      if (searchElem != NULL) {
        // 検索要素に子要素を持つので、検索要素の子要素とこの要素の子要素を検索する。
        // 検索要素に複数の子要素が存在した場合を考慮して、検索要素の子要素分回す。
        // 検索要素に複数の子要素が存在した場合は、OR条件と同じ。
        std::deque<sphElem*>::const_iterator it_search_child = searchElem->m_childElemList->begin();
        while( it_search_child != searchElem->m_childElemList->end() ) {
          const sphElem* searchChildElem = *it_search_child;
          const sphElem* result = child_elem->getExceptElemBySphElem(searchChildElem, prev_searchedElem);
          if (result != NULL) {
            return result;
          }
          it_search_child++;
        }
      }
    }
    it_child++;
  }

  return NULL;
}


/**
 * \~english
 * get sphElem class by element list
 * @param      elem_name    search element name
 * @return     sphElem Class
 *
 * \~japanese
 * 要素ツリーから検索要素名リストと一致する要素を取得する.
 * 要素名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素は検索対象外とする。
 * @param      elem_name     検索要素名
 * @return    検索結果要素(const sphElem*)
 */
const sphElem* sphElem::getElem(const char* elem_name) const
{
  if (elem_name == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "elem_name is null.");
    return NULL;
  }
  return getElem(1, elem_name);
}


/**
 * \~english
 * get sphElem class by element list
 * @param      num          search element number include first elem_name
 * @param      elem_name00  first search element name
 * @param      ...          search element name list
 * @return     sphElem Class
 *
 * \~japanese
 * 要素ツリーから検索要素名リストと一致する要素を取得する.
 * 子要素から検索要素名リストと一致する子要素、孫要素を検索する。
 * 要素名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素は検索対象外とする。
 * @param      num           検索要素名数 (elem_name00を含む)
 * @param      elem_name00   最初の検索要素名
 * @param      ...           ２番目以降の検索要素名リスト
 * @return    検索結果要素(const sphElem*)
 */
const sphElem* sphElem::getElem(unsigned int num, const char* elem_name00, ...) const
{
  if (num == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "num is zero.");
    return NULL;
  }

  // 引数の要素名リストを作成する。
  va_list ap;
  va_start(ap, num);
  std::vector<std::string> arg_list;
  for(int i=0;i<num;i++) {
    const char* arg = NULL;
    if (i==0) arg = elem_name00;
    else arg = va_arg(ap, const char*);
    if (arg == NULL || strlen(arg) <= 0) break;
    arg_list.push_back(std::string(arg));
  }

  // 検索用sphElemを作成する。
  const sphElem* searchElem = createSearchElem(&arg_list);

  // 要素ツリーによる検索実行
  const sphElem* result = getExceptElemBySphElem(searchElem);

  if (searchElem!= NULL) delete searchElem;

  if (result == NULL) {
    // 検索要素名の組み立て（警告出力用）
    std::string elem_list_str;
    for(int i=0;i<arg_list.size();i++) {
      elem_list_str += arg_list[i] + "-";
    }
    if (elem_list_str.size() > 0) {
      elem_list_str.erase(elem_list_str.size()-1);
    }
    // 警告
    m_errHandler->SPH_WARNING_INFO(SPH_WARNING_NOT_EXIST_ELEMENT, "element names is \"%s\"", elem_list_str.c_str());
  }

  va_end(ap);

  return result;
}


/**
 * \~english
 * get sphElem class by element name
 * @param      elem_name    search element name
 * @return     sphElem Class
 *
 * \~japanese
 * 子要素から検索要素名と一致する要素クラスを取得する .
 * 複数の子要素が検索要素名と一致しても、最初に一致した要素のみを返す。
 * 要素名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素は検索対象外とする。
 * @param      elem_name    検索要素名
 * @return    要素クラス
 */
const sphElem* sphElem::getChildElemFirst(const char* elem_name) const
{
  return getChildElemFirst(1, elem_name);
}


/**
 * \~english
 * get sphElem class by element list
 * @param      num          search element number include first elem_name
 * @param      elem_name00  first search element name
 * @param      ...          search element name list
 * @return     sphElem Class
 *
 * \~japanese
 * 要素ツリーから検索要素名リストと一致する要素を取得する.
 * 子要素から検索要素名リストと一致する子要素、孫要素を検索する。
 * 要素名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素は検索対象外とする。
 * @param      num           検索要素名数 (elem_name00を含む)
 * @param      elem_name00   最初の検索要素名
 * @param      ...           ２番目以降の検索要素名リスト
 * @return    検索結果要素(const sphElem*)
 */
const sphElem* sphElem::getChildElemFirst(unsigned int num, const char* elem_name00, ...) const
{
  if (num == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "num is zero.");
    return NULL;
  }

  // 引数の要素名リストを作成する。
  va_list ap;
  va_start(ap, num);
  std::vector<std::string> arg_list;
  for(int i=0;i<num;i++) {
    const char* arg = NULL;
    if (i==0) arg = elem_name00;
    else arg = va_arg(ap, const char*);
    if (arg == NULL || strlen(arg) <= 0) break;
    arg_list.push_back(std::string(arg));
  }

  // 検索用sphElemを作成する。
  const sphElem* searchElem = createSearchElem(&arg_list);

  // 要素ツリーによる検索実行
  const sphElem* result = getExceptElemBySphElem(searchElem);

  if (searchElem!= NULL) delete searchElem;

  if (result == NULL) {
    // 検索要素名の組み立て（警告出力用）
    std::string elem_list_str;
    for(int i=0;i<arg_list.size();i++) {
      elem_list_str += arg_list[i] + "-";
    }
    if (elem_list_str.size() > 0) {
      elem_list_str.erase(elem_list_str.size()-1);
    }
    // 警告
    m_errHandler->SPH_WARNING_INFO(SPH_WARNING_NOT_EXIST_ELEMENT, "element names is \"%s\"", elem_list_str.c_str());
  }

  va_end(ap);

  return result;
}

/**
 * \~english
 * get next sphElem class by element name
 * @param      elem          search previous sphElem
 * @param      elem_name    search element name
 * @return     sphElem Class
 *
 * \~japanese
 * 子要素から検索要素名と一致する次要素を取得する.
 * 複数の子要素が検索要素名と一致しても、最初に一致した要素のみを返す。
 * 要素名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素は検索対象外とする。
 * @param      elem          事前要素
 * @param      elem_name    検索要素名
 * @return    要素クラス
 */
const sphElem* sphElem::getChildElemNext(const sphElem* elem, const char* elem_name) const
{
  return getChildElemNext(elem, 1, elem_name);
}


/**
 * \~english
 * get next sphElem class by element name
 * @param      elem          search previous sphElem
 * @param      elem_name    search element name
 * @return     sphElem Class
 *
 * \~japanese
 * 子要素から検索要素名と一致する次要素を取得する.
 * 複数の子要素が検索要素名と一致しても、最初に一致した要素のみを返す。
 * 要素名の大文字・小文字の区別はせず比較する
 * <sph_xml_config_file>要素は検索対象外とする。
 * @param      elem          事前要素
 * @param      elem_name    検索要素名
 * @param      ...           ２番目以降の検索要素名リスト
 * @return    要素クラス
 */
const sphElem* sphElem::getChildElemNext(const sphElem* elem, unsigned int num, const char* elem_name00, ...) const
{
  if (num == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "num is zero.");
    return NULL;
  }

  // 引数の要素名リストを作成する。
  va_list ap;
  va_start(ap, num);
  std::vector<std::string> arg_list;
  for(int i=0;i<num;i++) {
    const char* arg = NULL;
    if (i==0) arg = elem_name00;
    else arg = va_arg(ap, const char*);
    if (arg == NULL || strlen(arg) <= 0) break;
    arg_list.push_back(std::string(arg));
  }

  // 検索用sphElemを作成する。
  const sphElem* searchElem = createSearchElem(&arg_list);

  // 要素ツリーによる検索実行
  const sphElem* result = getExceptElemBySphElem(searchElem, elem);

  if (searchElem!= NULL) delete searchElem;

  // Next検索の場合、最後はNULLになるので、警告メッセージはなし。

  va_end(ap);

  return result;
}



/**
 * \~english
 * get number of element name in child element
 * @param      elem_name            element name
 * @returm     count or element
 *
 * \~japanese
 * 要素名を持つ子要素数を取得する。
 * <sph_xml_config_file>要素は検索対象外とする。
 * @param      elem_name            要素名
 * @returm     検索結果要素数
 */
int sphElem::countNumOfElem(const char* elem_name) const
{
  if (elem_name == NULL) {
    return m_childElemList->size();
  }
  return countNumOfElem(1, elem_name);
}


/**
 * \~english
 * get number of element name list in child element
 * @param      num          search element number include first elem_name
 * @param      elem_name00  first search element name
 * @param      ...          search element name list
 * @returm     count or element
 *
 * \~japanese
 * 要素名リストを持つ子要素数を取得する。
 * <sph_xml_config_file>要素は検索対象外とする。
 * @param      num           検索要素名数 (elem_name00を含む)
 * @param      elem_name00   最初の検索要素名
 * @param      ...           ２番目以降の検索要素名リスト
 * @returm     検索結果要素数
 */
int sphElem::countNumOfElem(unsigned int num, const char* elem_name00, ...) const
{
  if (num == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "num is zero.");
    return -1;
  }

  // 引数の要素名リストを作成する。
  va_list ap;
  va_start(ap, num);
  std::vector<std::string> arg_list;
  for(int i=0;i<num;i++) {
    const char* arg = NULL;
    if (i==0) arg = elem_name00;
    else arg = va_arg(ap, const char*);
    if (arg == NULL || strlen(arg) <= 0) break;
    arg_list.push_back(std::string(arg));
  }

  // 検索用sphElemを作成する。
  const sphElem* searchElem = createSearchElem(&arg_list);

  // 検索要素ツリーによる要素数の取得を行う
  int count = countNumOfExceptElemBySphElem(searchElem);

  if (searchElem!= NULL) delete searchElem;
  va_end(ap);

  return count;
}


/**
 * \~english
 * get number of element in child element by search element tree
 * @param      searchElem      search element tree
 * @returm     count or element
 *
 * \~japanese
 * 検索要素ツリーと一致する子要素数を取得する。
 * <sph_xml_config_file>要素は検索対象外とする。
 * @param      searchElem          検索要素ツリー
 * @returm     検索結果要素数
 */
int sphElem::countNumOfExceptElemBySphElem(const sphElem* searchElem) const
{
  if (searchElem == NULL) {
    return m_childElemList->size();
  }

  int num = 0;

  // 子要素から検索を行う。
  std::deque<sphElem*>::const_iterator it_child = m_childElemList->begin();
  while( it_child != m_childElemList->end() ) {
    const sphElem* child_elem = *it_child;

    // <sph_xml_config_file>要素であるかチェックする
    if (child_elem->isSphCfgFileElem()) {
      // <sph_xml_config_file>要素であるので、子要素を取得する。
      child_elem = child_elem->getChildElemFirst();
    }

    // 子要素と検索トップ要素の検索
    if (child_elem->isSearchMatch(searchElem)) {
      if (searchElem->m_childElemList->size() == 0) {
        // 検索要素に子要素がなければ、インクリメントして、次のこの要素の子要素をチェックする。
        num++;
      }
      else {
        // 検索要素に子要素を持つので、検索要素の子要素とこの要素の子要素を検索する。
        // 検索要素に複数の子要素が存在した場合を考慮して、検索要素の子要素分回す。
        // 検索要素に複数の子要素が存在した場合は、OR条件と同じ。
        std::deque<sphElem*>::const_iterator it_search_child = searchElem->m_childElemList->begin();
        while( it_search_child != searchElem->m_childElemList->end() ) {
          const sphElem* searchChildElem = *it_search_child;
          int child_num = child_elem->countNumOfExceptElemBySphElem(searchChildElem);
          num += child_num;
          it_search_child++;
        }
      }
    }
    it_child++;
  }

  return num;
}


/**
 * リスト要素の子要素を取得する。(@id属性)
 * @param     list_name      リスト要素名 (SphSolverList, SphCouplerList...)
 * @param     sub_name       リスト子要素名 (SphSolver, SphCoupler...)
 * @param     find_id        リスト子要素検索ID (@id属性)
 * @return    リスト子要素
 */
const sphElem* sphElem::getElemListById(
                      const char* list_name,
                      const char* sub_name,
                      int find_id) const
{

  // 検索用sphElemの作成
  // <SphSolver/CouplerList>要素
  sphElem* search_listelem = new sphElem();
  search_listelem->setElemName(list_name);
  // <SphSolver/Coupler>要素
  sphElem* search_subelem = new sphElem();
  search_subelem->setElemName(sub_name);
  // @id=solverId
  search_subelem->addAttribute(SPH_CFG_ATTR_ID, find_id);
  search_listelem->addChildElem(search_subelem);

  // 検索実行
  const sphElem* solv_elem = getExceptElemBySphElem(search_listelem);

  delete search_listelem;

  return solv_elem;
}


/**
 * リスト要素の子要素を取得する。(@label属性)
 * @param     list_name      リスト要素名 (SphSolverList, SphCouplerList...)
 * @param     sub_name       リスト子要素名 (SphSolver, SphCoupler...)
 * @param     find_id        リスト子要素検索ラベル (@label属性)
 * @return    リスト子要素
 */
const sphElem* sphElem::getElemListByLabel(
                      const char* list_name,
                      const char* sub_name,
                      const char* find_label) const
{

  // 検索用sphElemの作成
  // <SphSolver/CouplerList>要素
  sphElem* search_listelem = new sphElem();
  search_listelem->setElemName(list_name);
  // <SphSolver/Coupler>要素
  sphElem* search_subelem = new sphElem();
  search_subelem->setElemName(sub_name);
  // @id=find_label
  search_subelem->addAttribute(SPH_CFG_ATTR_LABEL, std::string(find_label));
  search_listelem->addChildElem(search_subelem);

  // 検索実行
  const sphElem* solv_elem = getExceptElemBySphElem(search_listelem);

  delete search_listelem;

  return solv_elem;
}


/**
 * 要素のXMLファイル名を取得する。.
 * 親要素からファイル名を検索する。
 * @param     void
 * @return    XMLファイル名
 */
std::string sphElem::getUpstreamFileName() const
{
  // 自要素から検索を行う。
  std::string filename = m_fileName;
  if (!filename.empty()) return filename;

  // 親要素から検索を行う。
  const sphElem* parentElem = m_parentElem;
  if (parentElem == NULL) return "";

  filename = parentElem->m_fileName;
  while (filename.empty()) {
    // 親の親へ移動する。
    parentElem = parentElem->m_parentElem;
    if (parentElem == NULL) {
      break;
    }
    filename = parentElem->m_fileName;
  }

  return filename;
}



/**
 * 子要素から検索要素名と属性値の一致する要素を取得する.
 * @param      elem_name     検索要素名
 * @param      attr_name     検索属性名
 * @param      attr_value     検索属性値
 * @return    検索結果要素(const sphElem*)
 */
const sphElem* sphElem::getChildElemByAttr(
                      const char* elem_name,
                      const char* attr_name,
                      const char* attr_value) const
{

  // 検索用sphElemを作成する。
  sphElem searchElem;
  searchElem.setElemName(elem_name);
  searchElem.addAttribute(attr_name, attr_value);

  // 要素ツリーによる検索実行
  return getExceptElemBySphElem(&searchElem);
}


/**
 * stringを小文字文字列に変換する。
 * @param       value   変換対象文字列
 * @return      小文字変換文字列
 */
std::string sphElem::toLower(std::string value)
{

  int len = value.length();
  std::string lower = std::string(value);
  for (int i=0; i<len; i++) {
    //lower[i] = std::tolower(value[i]);
    int chr = value[i];
    if ( chr >= 'A' && chr <= 'Z' ) {
      chr += 'a' - 'A';    /* 小文字に変換 */
    }
    lower[i] = chr;
  }

  return lower;
}


