/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphElemUtils.C
 * Sphere Configuration Element Utilities Class source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphElemUtils.C
 * Sphereコンフィギュレーション要素ユーティリティクラスソースファイル
 * @author     hira
 * @date       2009/02/20
 */


#include "config/sphElemUtils.h"


/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphElemUtils::sphElemUtils()
{
}


/**
 * コンストラクタ
 * @param       elem            コンフィギュレーション要素
 */
sphElemUtils::sphElemUtils(const sphElem* elem)
{
  m_referElem = elem;
}

/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphElemUtils::~sphElemUtils()
{

}


/**
 * 有効なソルバ数を取得する。
 * @return    有効なソルバ数
 */
int sphElemUtils::getSolverCount() const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return 0;
  }

  const sphElem* config_elem = m_referElem->getRootElem();

  if (config_elem == NULL) {
    return 0;
  }

  // <SphSolverList>要素取得
  const sphElem*solv_list = config_elem->getElem(SPH_CFG_NAME_SOLVER_LIST);
  if (solv_list == NULL) {
    return 0;
  }

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  int count = 0;
  // <SphSolver>要素検索
  const sphElem* solv_elem = solv_list->getChildElemFirst(SPH_CFG_NAME_SOLVER);
  while (solv_elem != NULL) {
    const char* solv_label = solv_elem->getAttrString(SPH_CFG_ATTR_LABEL);
    int solv_id = solv_elem->getAttrInt(SPH_CFG_ATTR_ID);
    // @mode属性取得 (デフォルト：true)
    bool solv_enable = solv_elem->isAttrValue(SPH_CFG_ATTR_ENABLE, true);
    if (solv_enable) {
      count++;
    }

    // 次要素検索
    solv_elem = solv_list->getChildElemNext(solv_elem, SPH_CFG_NAME_SOLVER);
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  return count;
}


/**
 * 有効なカプラ数を取得する。
 * @return    有効なカプラ数
 */
int sphElemUtils::getCouplerCount() const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return 0;
  }

  // ルート要素を取得する。
  const sphElem* config_elem = m_referElem->getRootElem();

  if (config_elem == NULL) {
    return 0;
  }

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  // <SphCouplerList>要素取得
  const sphElem* coup_list = config_elem->getElem(SPH_CFG_NAME_COUPLER_LIST);
  if (coup_list == NULL) {
    // 警告メッセージを元に戻す。
    m_errHandler->releaseWarning();
    return 0;
  }

  int count = 0;
  // <SphCoupler>要素検索
  const sphElem* coup_elem = coup_list->getChildElemFirst(SPH_CFG_NAME_COUPLER);
  while (coup_elem != NULL) {
    const char* coup_label = coup_elem->getAttrString(SPH_CFG_ATTR_LABEL);
    int coup_id = coup_elem->getAttrInt(SPH_CFG_ATTR_ID);
    // @mode属性取得 (デフォルト：true)
    bool coup_enable = coup_elem->isAttrValue(SPH_CFG_ATTR_ENABLE, true);
    if (coup_enable) {
      count++;
    }

    // 次要素検索
    coup_elem = coup_list->getChildElemNext(coup_elem, SPH_CFG_NAME_COUPLER);
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  return count;
}



/**
 * 有効なソルバ要素を取得する。(index)
 * @param     index       ソルバーリストのindex (<SphSolver>要素の@id属性ではない)
 * @return    ソルバ要素
 */
const sphElem* sphElemUtils::getSolverElem(int index) const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return NULL;
  }

  // ルート要素を取得する。
  const sphElem* config_elem = m_referElem->getRootElem();

  if (config_elem == NULL) {
    return NULL;
  }

  // <SphSolverList>要素取得
  const sphElem*solv_list = config_elem->getElem(SPH_CFG_NAME_SOLVER_LIST);
  if (solv_list == NULL) {
    return NULL;
  }

  int count = 0;
  // <SphSolver>要素検索
  const sphElem* result_elem = NULL;
  const sphElem* solv_elem = solv_list->getChildElemFirst(SPH_CFG_NAME_SOLVER);
  while (solv_elem != NULL) {
    const char* solv_label = solv_elem->getAttrString(SPH_CFG_ATTR_LABEL);
    int solv_id = solv_elem->getAttrInt(SPH_CFG_ATTR_ID);
    // @mode属性取得 (デフォルト：true)
    bool solv_enable = solv_elem->isAttrValue(SPH_CFG_ATTR_ENABLE, true);
    if (solv_enable) {
      if (count == index) {
        result_elem = solv_elem;
      }
      count++;
    }

    // 次要素検索
    solv_elem = solv_list->getChildElemNext(solv_elem, SPH_CFG_NAME_SOLVER);
  }

  return result_elem;
}


/**
 * 有効なカプラ要素を取得する。(index)
 * @param     index       カプラリストのindex (<SphCoupler>要素の@id属性ではない)
 * @return    有効なカプラ要素
 */
const sphElem* sphElemUtils::getCouplerElem( int index) const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return NULL;
  }

  // ルート要素を取得する。
  const sphElem* config_elem = m_referElem->getRootElem();

  if (config_elem == NULL) {
    return NULL;
  }

  // <SphCouplerList>要素取得
  const sphElem* coup_list = config_elem->getElem(SPH_CFG_NAME_COUPLER_LIST);
  if (coup_list == NULL) {
    return NULL;
  }

  int count = 0;
  // <SphCoupler>要素検索
  const sphElem* result_elem = NULL;
  const sphElem* coup_elem = coup_list->getChildElemFirst(SPH_CFG_NAME_COUPLER);
  while (coup_elem != NULL) {
    const char* coup_label = coup_elem->getAttrString(SPH_CFG_ATTR_LABEL);
    int coup_id = coup_elem->getAttrInt(SPH_CFG_ATTR_ID);
    // @mode属性取得 (デフォルト：true)
    bool coup_enable = coup_elem->isAttrValue(SPH_CFG_ATTR_ENABLE, true);
    if (coup_enable) {
      if (count == index) {
        result_elem = coup_elem;
        break;
      }
      count++;
    }

    // 次要素検索
    coup_elem = coup_list->getChildElemNext(coup_elem, SPH_CFG_NAME_COUPLER);
  }

  return result_elem;
}


/**
 * 有効なソルバ要素を取得する。(@id属性)
 * @param     solverId       ソルバID(<SphSolver>要素の@id属性)
 * @return    ソルバ要素
 */
const sphElem* sphElemUtils::getElemBySolverId(int solverId) const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return NULL;
  }

  // ルート要素を取得する。
  const sphElem* config_elem = m_referElem->getRootElem();
  if (config_elem == NULL) return NULL;

  // ソルバ要素を@id属性で検索する。
  const sphElem* solv_elem = config_elem->getElemListById(
                                SPH_CFG_NAME_SOLVER_LIST,
                                SPH_CFG_NAME_SOLVER,
                                solverId);

  // ソルバ要素が有効であるかチェックする。
  if (!solv_elem->isAttrValue(SPH_CFG_ATTR_ENABLE, true)) {
    return NULL;
  }

  return solv_elem;
}


/**
 * 有効なカプラ要素を取得する。(@id属性)
 * @param     couplerId       カプラID(<SphCoupler>要素の@id属性)
 * @return    有効なカプラ要素
 */
const sphElem* sphElemUtils::getElemByCouplerId(int couplerId) const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return NULL;
  }

  // ルート要素を取得する。
  const sphElem* config_elem = m_referElem->getRootElem();
  if (config_elem == NULL) return NULL;


  // ソルバ要素を@id属性で検索する。
  const sphElem* coup_elem = config_elem->getElemListById(
                                SPH_CFG_NAME_COUPLER_LIST,
                                SPH_CFG_NAME_COUPLER,
                                couplerId);

  // カプラ要素が有効であるかチェックする。
  if (!coup_elem->isAttrValue(SPH_CFG_ATTR_ENABLE, true)) {
    return NULL;
  }

  return coup_elem;
}


/**
 * 有効なソルバ要素を取得する。(@label属性)
 * @param     solverLabel       ソルバラベル(<SphSolver>要素の@label属性)
 * @return    ソルバ要素
 */
const sphElem* sphElemUtils::getElemBySolverLabel(const char* solverLabel) const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return NULL;
  }

  // ルート要素を取得する。
  const sphElem* config_elem = m_referElem->getRootElem();
  if (config_elem == NULL) return NULL;

  // ソルバ要素を@label属性で検索する。
  const sphElem* solv_elem = config_elem->getElemListByLabel(
                                SPH_CFG_NAME_SOLVER_LIST,
                                SPH_CFG_NAME_SOLVER,
                                solverLabel);

  // ソルバ要素が有効であるかチェックする。
  if (!solv_elem->isAttrValue(SPH_CFG_ATTR_ENABLE, true)) {
    return NULL;
  }

  return solv_elem;
}


/**
 * 有効なカプラ要素を取得する。(@label属性)
 * @param     couplerLabel       カプラID(<SphCoupler>要素の@label属性)
 * @return    有効なカプラ要素
 */
const sphElem* sphElemUtils::getElemByCouplerLabel(const char* couplerLabel) const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return NULL;
  }

  // ルート要素を取得する。
  const sphElem* config_elem = m_referElem->getRootElem();
  if (config_elem == NULL) return NULL;

  // カプラ要素を@label属性で検索する。
  const sphElem* coup_elem = config_elem->getElemListByLabel(
                                SPH_CFG_NAME_COUPLER_LIST,
                                SPH_CFG_NAME_COUPLER,
                                couplerLabel);

  // カプラ要素が有効であるかチェックする。
  if (!coup_elem->isAttrValue(SPH_CFG_ATTR_ENABLE, true)) {
    return NULL;
  }

  return coup_elem;
}


/**
 * ソルバ要素のノード数を取得する。(@id属性)
 * @param     solverId       ソルバID(<SphSolver>要素の@id属性)
 * @return    ソルバ要素のノード数
 */
int sphElemUtils::getNodeCountBySolverId(int solverId) const
{
  // ソルバ要素を@id属性で検索する。
  const sphElem* solv_elem = getElemBySolverId(solverId);
  if (solv_elem == NULL) {
    return 0;
  }

  return getNodeCountFromSteer(solv_elem);
}


/**
 * カプラ要素のノード数を取得する。(@id属性)
 * @param     couplerId       ソルバID(<SphCoupler>要素の@id属性)
 * @return    カプラ要素のノード数
 */
int sphElemUtils::getNodeCountByCouplerId(int couplerId) const
{
  // カプラ要素を@id属性で検索する。
  const sphElem* coup_elem = getElemByCouplerId(couplerId);
  if (coup_elem == NULL) {
    return 0;
  }

  return getNodeCountFromSteer(coup_elem);
}


/**
 * SphSteer要素からノード数を取得する。
 * @param     elem           ソルバ要素／カプラ要素
 * @return    ソルバ要素／カプラ要素のノード数
 */
int sphElemUtils::getNodeCountFromSteer(const sphElem* elem) const
{
  if (elem == NULL) return 0;

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  register int i;
  int node_count = 0;
  // dims
  int dims = elem->getAttrInt(SPH_CFG_ATTR_DIMS);
  if (dims == 0) dims = 3;

  // <SphNodeList>要素の検索
  const sphElem* node_list = elem->getElem(
                                2,
                                SPH_CFG_NAME_STEER,
                                SPH_CFG_NAME_NODELIST);
  if (node_list != NULL) {
    // <SphNode>のカウント
    node_count = node_list->countNumOfElem(SPH_CFG_NAME_NODE);
  }

  if (node_count == 0) {
    // <SphVoxelDivision>要素の検索
    const sphElem* div_elem = elem->getElem(
                                  2,
                                  SPH_CFG_NAME_STEER,
                                  SPH_CFG_NAME_DIVISION);
    if (div_elem != NULL) {
      sphElemUtils searcher(div_elem);
      /// <SphVoxelDivision>要素のi, j, kの取得を行う。
      const char *attr_idx[] = {SPH_CFG_ATTR_I, SPH_CFG_ATTR_J, SPH_CFG_ATTR_K};
      unsigned int voxelDivision[3] = {1, 1, 1};
      if (searcher.getDimsAttributeValues(dims, SPH_CFG_NAME_DIVISION, attr_idx, voxelDivision)) {
        // ３次元分割数にする
        node_count = 1;
        for (i=0; i<dims; i++) node_count *= voxelDivision[i];
      }
      else {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "invalid <SphVoxelDivision> or attributes, %s", div_elem->debugPrintElem().c_str());
        node_count = 0;
      }
    }
  }

  if (node_count == 0) {
    // <SphNodeProcess>要素の検索
    const sphElem* proc_elem = elem->getElem(
                                  2,
                                  SPH_CFG_NAME_STEER,
                                  SPH_CFG_NAME_NODEPROCESS);
    if (proc_elem != NULL) {
      node_count = proc_elem->getAttrInt(SPH_CFG_ATTR_VALUE);
    }
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  return node_count;
}


/**
 * SphFileList/SphMultiFile要素からSphFile要素を識別ラベルから検索を行う。
 * @param[in]     fileLabel      ファイル識別ラベル
 * @return    SphFile要素
 */
const sphElem* sphElemUtils::getFileElemByFileLabel(const char* fileLabel) const
{
  if (fileLabel == NULL || m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getElemByFileLabel arguments");
    return NULL;
  }

  const sphElem* config_elem = m_referElem->getRootElem();

  /// SphFileList/SphFile[@label] or SphMultiFile/SphFile[@label]の検索を行う。
  // <SphFile>要素からlabel検索を行う。
  const sphElem* file_elem = config_elem->getElemListByLabel(SPH_CFG_NAME_FILELIST, SPH_CFG_NAME_FILE, fileLabel);
  if (file_elem == NULL) {
    // <SphFileList>要素の取得を行う。
    const sphElem* list_elem = config_elem->getElem(SPH_CFG_NAME_FILELIST);
    if (list_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "element name is %s.", SPH_CFG_NAME_FILELIST);
      return NULL;
    }
    // <SphMultiFile>要素からlabelの検索を行う。
    file_elem = config_elem->getElemListByLabel(SPH_CFG_NAME_MULTIFILE, SPH_CFG_NAME_FILE, fileLabel);
    if (file_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "label is not found.[label=%s]", fileLabel);
      return NULL;
    }
  }
  return file_elem;
}



/**
 * SphFileList/SphMultiFile要素のSphFile要素を識別IDから検索を行う。
 * @param[in]     fileId      ファイル識別ID
 * @return    SphFile要素
 */
const sphElem* sphElemUtils::getFileElemByFileId(int fileId) const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getFileElemByFileId arguments");
    return NULL;
  }

  const sphElem* config_elem = m_referElem->getRootElem();

  /// SphFileList/SphFile[@id] or SphMultiFile/SphFile[@id]の検索を行う。
  // <SphFile>要素からlabel検索を行う。
  const sphElem* file_elem = config_elem->getElemListById(SPH_CFG_NAME_FILELIST, SPH_CFG_NAME_FILE, fileId);
  if (file_elem == NULL) {
    // <SphFileList>要素の取得を行う。
    const sphElem* list_elem = config_elem->getElem(SPH_CFG_NAME_FILELIST);
    if (list_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "element name is %s.", SPH_CFG_NAME_FILELIST);
      return NULL;
    }
    // <SphMultiFile>要素からlabelの検索を行う。
    file_elem = list_elem->getElemListById(SPH_CFG_NAME_MULTIFILE, SPH_CFG_NAME_FILE, fileId);
    if (file_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "label is not found.[id=%s]", fileId);
      return NULL;
    }
  }
  return file_elem;
}


/**
 * SphFile/SphMultiFile要素から参照元のSphFile/SphMultiFile要素を取得する。
 * @param     void
 * @return    参照元SphFile/SphMultiFile要素
 */
const sphElem* sphElemUtils::getFileElem() const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getFileElemByFileId arguments");
    return NULL;
  }

  if ( !m_referElem->equalsElemName(SPH_CFG_NAME_FILE)
       && !m_referElem->equalsElemName(SPH_CFG_NAME_MULTIFILE) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphFile> or <SphMultiFile>.");
    return NULL;
  }

  // 親要素がない場合(SklSolverConfig)は親要素が<SphFileList>であるかチェックはしない。
  if (m_referElem->getParentElem(NULL) == NULL) {
    // 親要素がないので、この要素<SphFile>とする
    return m_referElem;
  }
  /// 親要素が<SphFileList>であるかチェックする。
  const sphElem* parentElem = m_referElem->getParentElem(SPH_CFG_NAME_FILELIST);
  if (parentElem != NULL) {
    // <SphFileList>配下の<SphFile>であるので、検索の必要はない。
    return m_referElem;
  }


  // ルート要素
  const sphElem* config_elem = m_referElem->getRootElem();
  const sphElem* src_elem = NULL;
  if (m_referElem->existAttribute(SPH_CFG_ATTR_ID)) {
    // idから、FileListを検索する。
    int fileId = m_referElem->getAttrInt(SPH_CFG_ATTR_ID);
    src_elem = config_elem->getElemListById(SPH_CFG_NAME_FILELIST, m_referElem->getElemName(), fileId);
  }
  else if (m_referElem->existAttribute(SPH_CFG_ATTR_LABEL)) {
    // labelから、FileListを検索する。
    const char* fileLabel = m_referElem->getAttrString(SPH_CFG_ATTR_LABEL);
    src_elem = config_elem->getElemListByLabel(SPH_CFG_NAME_FILELIST, m_referElem->getElemName(), fileLabel);
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "invalid attribute, not exist 'id' or  'label'[%s]", m_referElem->debugPrintElem().c_str());
  }

  if (src_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "src element is not found, %s", m_referElem->debugPrintElem().c_str());
  }

  return src_elem;
}


/**
 * SphFileListからSphFile/SphMultiFile要素を識別ラベルから検索を行う。
 * @param[in]     fileLabel      ファイル識別ラベル
 * @return    SphFile/SphMultiFile要素
 */
const sphElem* sphElemUtils::getElemByFileLabel(const char* fileLabel) const
{
  if (fileLabel == NULL || m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getElemByFileLabel arguments");
    return NULL;
  }

  const sphElem* config_elem = m_referElem->getRootElem();

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  /// SphFileList/SphFile[@label] or SphMultiFile/SphFile[@label]の検索を行う。
  // <SphFile>要素からlabel検索を行う。
  const sphElem* file_elem = config_elem->getElemListByLabel(SPH_CFG_NAME_FILELIST, SPH_CFG_NAME_FILE, fileLabel);

  if (file_elem == NULL) {
    // <SphMultiFile>要素からlabel検索を行う。
    file_elem = config_elem->getElemListByLabel(SPH_CFG_NAME_FILELIST, SPH_CFG_NAME_MULTIFILE, fileLabel);
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  return file_elem;
}


/**
 * <SphDomainInfo>要素から情報の取得を行う.
 * @param[in]     dims      次元数
 * @param[out]     size      ボクセルサイズ
 * @param[out]     origin    ボクセル原点座標
 * @param[out]    pitch     ボクセルピッチ
 * @param[out]    start     始点インデックス
 * @return    success(=true)/fault(=false)
 */
bool sphElemUtils::getDomainInfo(
                          int dims,
                          size_t size[3],
                          double origin[3],
                          double pitch[3],
                          size_t start[3]) const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return false;
  }

  register int i;

  /// <SphDomainInfo>要素の取得を行う。
  const sphElem* domain_elem = NULL;
  if (m_referElem->equalsElemName(SPH_CFG_NAME_SOLVER)) {
    domain_elem = m_referElem->getElem(SPH_CFG_NAME_DOMAININFO);
  }
  else if (m_referElem->equalsElemName(SPH_CFG_NAME_DOMAININFO)) {
    domain_elem = m_referElem;
  }
  if (domain_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found <SphDomainInfo>");
    return false;
  }

  sphElemUtils searcher(domain_elem);
  /// <SphVoxelSize>要素のix, jx, kxの取得を行う。
  const sphElem* size_elem = domain_elem->getElem(SPH_CFG_NAME_VOXELSIZE);
  if (size_elem != NULL) {
    const char *attr_size[] = {SPH_CFG_ATTR_IX, SPH_CFG_ATTR_JX, SPH_CFG_ATTR_KX};
    if (!searcher.getDimsAttributeValues(dims, SPH_CFG_NAME_VOXELSIZE, attr_size, size)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "invalid <SphVoxelSize> or attributes");
      return false;
    }
  }

  /// <SphVoxelOrigin>要素のox, oy, ozの取得を行う。
  const sphElem* org_elem = domain_elem->getElem(SPH_CFG_NAME_VOXELORIGIN);
  if (org_elem != NULL) {                   // <SphVoxelOrigin>要素存在チェック（任意要素）
    const char *attr_org[] = {SPH_CFG_ATTR_OX, SPH_CFG_ATTR_OY, SPH_CFG_ATTR_OZ};
    if (!searcher.getDimsAttributeValues(dims, SPH_CFG_NAME_VOXELORIGIN, attr_org, origin)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "invalid <SphVoxelOrigin> or attributes");
      return false;
    }
  }

  /// <SphVoxelPitch>要素のdx,dy,dzの取得を行う。
  const char *attr_dbl[] = {SPH_CFG_ATTR_DX, SPH_CFG_ATTR_DY, SPH_CFG_ATTR_DZ};
  const sphElem* pitch_elem = domain_elem->getElem(SPH_CFG_NAME_VOXELPITCH);
  if (pitch_elem != NULL) {                   // <SphVoxelPitch>要素存在チェック（任意要素）
    if (!searcher.getDimsAttributeValues(dims, SPH_CFG_NAME_VOXELPITCH, attr_dbl, pitch)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "invalid <SphVoxelPitch> or attributes");
      return false;
    }
  }

  /// <SphVoxelWidth>要素のdx,dy,dzの取得を行う。
  double width[3] = {0, 0, 0};
  const sphElem* width_elem = domain_elem->getElem(SPH_CFG_NAME_VOXELWIDTH);
  if (width_elem != NULL) {                   // <SphVoxelWidth>要素存在チェック（任意要素）
    if (!searcher.getDimsAttributeValues(dims, SPH_CFG_NAME_VOXELWIDTH, attr_dbl, width)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "invalid <SphVoxelWidth> or attributes");
      return false;
    }
  }

  if (pitch_elem == NULL && width_elem != NULL) {
    /// ボクセルピッチを設定する。
    for (i=0; i<3; i++) pitch[i] = 0.0;
    for (i=0; i<dims; i++) {
      if (size[i] > 0) {
        pitch[i] = width[i]/size[i];
      }
    }
  }

  /// <SphStartIndex>要素のix, jx, kxの取得を行う。
  if (start != NULL) {
    const sphElem* start_elem = domain_elem->getElem(SPH_CFG_NAME_STARTINDEX);
    if (start_elem != NULL) {
      const char *attr_sidx[] = {SPH_CFG_ATTR_I, SPH_CFG_ATTR_J, SPH_CFG_ATTR_K};
      if (!searcher.getDimsAttributeValues(dims, SPH_CFG_NAME_STARTINDEX, attr_sidx, start)) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "invalid <SphVoxelSize> or attributes");
        return false;
      }
    }
  }

  return true;
}



/**
 * 自要素が属するSphSolver/SphSteer/SphDir[@input]を取得する。
 * @param        void
 * @return    success(=true)/fault(=false)
 */
const char* sphElemUtils::getInputBaseDir() const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return false;
  }

  const sphElem* solverElem = m_referElem;
  if (!m_referElem->equalsElemName(SPH_CFG_NAME_SOLVER)) {
    // 自身が属するSphSolver要素を取得する。
    solverElem = m_referElem->getParentElem(SPH_CFG_NAME_SOLVER);
  }

  if (solverElem != NULL) {
    // SphSolver/SphSteer/SphDir[@input]の取得
    const sphElem* dir_elem = solverElem->getElem(2, SPH_CFG_NAME_STEER, SPH_CFG_NAME_BASEDIR);
    if (dir_elem != NULL) {
      const char* basedir = dir_elem->getAttrString(SPH_CFG_ATTR_INPUTDIR);
      return basedir;
    }
  }

  return NULL;

}

/**
 * 自要素が属するSphSolver/SphSteer/SphDir[@output]を取得する。
 * @param        void
 * @return    success(=true)/fault(=false)
 */
const char* sphElemUtils::getOutputBaseDir() const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return false;
  }

  const sphElem* solverElem = m_referElem;
  if (!m_referElem->equalsElemName(SPH_CFG_NAME_SOLVER)) {
    // 自身が属するSphSolver要素を取得する。
    solverElem = m_referElem->getParentElem(SPH_CFG_NAME_SOLVER);
  }

  if (solverElem != NULL) {
    // SphSolver/SphSteer/SphDir[@output]の取得
    const sphElem* dir_elem = solverElem->getElem(2, SPH_CFG_NAME_STEER, SPH_CFG_NAME_BASEDIR);
    if (dir_elem != NULL) {
      const char* basedir = dir_elem->getAttrString(SPH_CFG_ATTR_OUTPUTDIR);
      return basedir;
    }
  }

  return NULL;

}


/**
 * ファイルフォーマットによって書式化されたファイル名を取得する。
 * @param[out]      file_path       ファイルパス名
 * @param[in]      basedir         基準フォルダ名
 * @param[in]      steps          ステップ番号
 * @param[in]      rankId          ランク番号
 * @param[in]      sub_flag        サブブロックフラグ(true:サブブロック)
 * @return    success(=true)/fault(=false)
 */
bool sphElemUtils::getFileFormatPath(
                      char* file_path,
                      const char* basedir,
                      int steps, int rankId,
                      bool sub_flag)
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_referElem is null");
    return false;
  }

  if ( !m_referElem->equalsElemName(SPH_CFG_NAME_FILE) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphFile> or <SphMultiFile>.");
    return false;
  }

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  // 分散出力であるか取得する？
  bool multi_flag = true;

  // ファイル入出力モードの取得
  SphCfgFileIOMode cfg_mode = sphElemUtils(m_referElem).getFileIoMode();
  if      (cfg_mode == CFG_FILE_DEFAULT) multi_flag = true;     // mode属性が存在しない場合、multiモードとする。
  else if (cfg_mode == CFG_FILE_MULTI) multi_flag = true;
  else if (cfg_mode == CFG_FILE_PROCESS) multi_flag = true;
  else if (cfg_mode == CFG_FILE_WHOLE) multi_flag = false;
  else {
    // 警告メッセージを元に戻す。
    m_errHandler->releaseWarning();
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "invalid mode attribute.");
    return false;
  }

  char file_name[MAXPATHLEN] = {0};
  /// SphFileNameFormat要素の取得
  const sphElem* fmt_elem = m_referElem->getElem(SPH_CFG_NAME_FILEFORMAT);
  if (fmt_elem != NULL) {
    /// ファイル名書式によってファイル名を作成する。
    const char* fmt = fmt_elem->getAttrString(SPH_CFG_ATTR_FORMAT);
    const char* params = fmt_elem->getAttrString(SPH_CFG_ATTR_VALUE);
    if (strcasecmp(params, "rank") == 0) {
      sprintf(file_name, fmt, rankId);
    }
    else if (strcasecmp(params, "step") == 0) {
      sprintf(file_name, fmt, steps);
    }
    else if (strcasecmp(params, "step_rank") == 0) {
      sprintf(file_name, fmt, steps, rankId);
    }
    else if (strcasecmp(params, "rank_step") == 0) {
      sprintf(file_name, fmt, rankId, steps);
    }
    else {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "@value attribute is '%s', %s.", params, fmt_elem->debugPrintElem().c_str());
    }
  }
  else {
    /// prefix    ファイル名接頭文字
    const char* prefix = m_referElem->getAttrString(SPH_CFG_ATTR_PREFIX);
    if (prefix != NULL) {
      /// ファイル名接頭文字によってファイル名を作成する。
      // ファイル拡張子＝ファイルフォーマット
      const char* attr_ext = m_referElem->getAttrString(SPH_CFG_ATTR_FORMAT);
      std::string ext = sphElem::toLower(std::string(attr_ext));

      int commSize = 0;
      SPL_Comm_size(SPL_COMM_WORLD, &commSize);

      // 出力ファイル名形式
      bool id_format = false;
      if (sub_flag && multi_flag) id_format = true;   // サブブロック実行 + 分散出力 = MYID付き
      else if (multi_flag && commSize > 1)  id_format = true;   // 分散出力 + 並列実行 = MYID付き
                                                      // 分散出力 + 逐次実行 = MYIDなし

      if (!id_format) {
        // 単独実行
        /// ["prefix"属性]_[ステップ数(８桁)].["format"属性]
        sprintf(file_name, "%s_%08d.%s", prefix, steps, ext.c_str());
      }
      else {
        /// ["prefix"属性]_[ステップ数(８桁)]_id[ランク番号(6桁)].["format"属性]
        sprintf(file_name, "%s_%08d_id%06d.%s", prefix, steps, rankId, ext.c_str());
      }
    }
    else {
      /// file_name属性の取得
      const char* attr_fname = m_referElem->getAttrString(SPH_CFG_ATTR_FILENAME);
      if (attr_fname == NULL) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "can not get formated file name, %s.", m_referElem->debugPrintElem().c_str());
      }
      else {
        strcpy(file_name, attr_fname);
      }
    }
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  if (strlen(file_name) == 0) {
    return false;
  }

  // 相対パスか絶対パスか？
  if (!PathIsRelative(file_name)) {
    strcpy(file_path, file_name);
    return true;
  }

  if (basedir != NULL) {
    // ソルバのbasedirからの相対パス
    // 相対パスであるので、basedirを付加する。
    sprintf(file_path, "%s/%s", basedir, file_name);
  }
  else {
    // 記述XMLファイルパスからの相対パス
    char attr_path[MAXPATHLEN];
    /// XMLファイルの記述パスの取得
    m_referElem->getAttrFilePath(attr_path);
    sprintf(file_path, "%s/%s", attr_path, file_name);
  }

  return true;

}


/**
 * ファイルパス名を取得する。
 * @param[out]      file_path       ファイルパス名
 * @param[in]      basedir         基準フォルダ名
 * @return    success(=true)/fault(=false)
 */
bool sphElemUtils::getFilePath(
                      char* file_path,
                      const char* basedir) const
{
  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getFileElemByFileId arguments");
    return false;
  }

  if ( !m_referElem->equalsElemName(SPH_CFG_NAME_FILE) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphFile> or <SphMultiFile>.");
    return false;
  }

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  char file_name[MAXPATHLEN] = {0};
  /// file_name属性の取得
  const char* attr_fname = m_referElem->getAttrString(SPH_CFG_ATTR_FILENAME);
  if (attr_fname == NULL) {
    // 警告メッセージを元に戻す。
    m_errHandler->releaseWarning();
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "can not get @file_name, %s.", m_referElem->debugPrintElem().c_str());
    return false;
  }
  else {
    strcpy(file_name, attr_fname);
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  if (strlen(file_name) == 0) {
    return false;
  }

  // 相対パスか絶対パスか？
  if (!PathIsRelative(file_name)) {
    strcpy(file_path, file_name);
    return true;
  }

  if (basedir != NULL) {
    // ソルバのbasedirからの相対パス
    // 相対パスであるので、basedirを付加する。
    sprintf(file_path, "%s/%s", basedir, file_name);
  }
  else {
    // 記述XMLファイルパスからの相対パス
    char attr_path[MAXPATHLEN];
    /// XMLファイルの記述パスの取得
    m_referElem->getAttrFilePath(attr_path);
    sprintf(file_path, "%s/%s", attr_path, file_name);
  }

  return true;

}



/**
 * ファイル入出力モードを取得する。
 * @return    ファイル入出力モード
 */
SphCfgFileIOMode sphElemUtils::getFileIoMode() const
{

  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getFileElemByFileId arguments");
    return CFG_FILE_UNKNOWN;
  }

  if ( !m_referElem->equalsElemName(SPH_CFG_NAME_FILE) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphFile>.");
    return CFG_FILE_UNKNOWN;
  }

  // mode属性が存在しない
  if (!m_referElem->existAttribute(SPH_CFG_ATTR_MODE)) {
    return CFG_FILE_DEFAULT;
  }

  const char* mode_attr = m_referElem->getAttrString(SPH_CFG_ATTR_MODE);

  // ファイル入出力モードの取得
  SphCfgFileIOMode mode = CFG_FILE_UNKNOWN;
  if (strcasecmp(mode_attr, SPH_CFG_FILEIO_MULTI) == 0) mode = CFG_FILE_MULTI;
  else if (strcasecmp(mode_attr, SPH_CFG_FILEIO_WHOLE) == 0) mode = CFG_FILE_WHOLE;
  else if (strcasecmp(mode_attr, SPH_CFG_FILEIO_PROCESS) == 0) mode = CFG_FILE_PROCESS;

  return mode;
}


/**
 * ファイル出力ステップであるかチェックする。
 *
 * SphFile/SphFileOption[@interval]の属性値をチェックする。
 * @param     current_step        現在ステップ
 * @return    true=出力ステップ
 */
bool sphElemUtils::isFileInterval(size_t current_step) const
{

  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getFileElemByFileId arguments");
    return false;
  }

  const sphElem* option_elem = NULL;
  if ( m_referElem->equalsElemName(SPH_CFG_NAME_FILE) ) {
    // <SphFileOption>要素：任意要素を取得する。
    if (m_referElem->existChileElement(SPH_CFG_NAME_FILEOPTION)) {
      option_elem = m_referElem->getElem(SPH_CFG_NAME_FILEOPTION);
    }
  }
  else if ( m_referElem->equalsElemName(SPH_CFG_NAME_FILEOPTION) ) {
    option_elem = m_referElem;
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphFile>.");
    return false;
  }

  size_t interval = 0;
  if (option_elem != NULL) {
    interval = option_elem->getAttrInt(SPH_CFG_ATTR_INTERVAL);
  }

  /// 出力ステップであるか、intervalをチェックする。
  if( interval != 0 && (current_step%interval !=0) ) {
    return false;
  }

  return true;
}


/**
 * ファイル始点インデックスタイプを取得する。
 * @return    ファイル始点インデックスタイプ
 */
SphCfgAssignType sphElemUtils::getStartIndexAssignType() const
{

  if ( m_referElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid getFileElemByFileId arguments");
    return CFG_ASSIGN_UNKNOWN;
  }

  const sphElem *startidx_elem = NULL;
  if ( m_referElem->equalsElemName(SPH_CFG_NAME_FILE) ) {
    if (m_referElem->existChileElement(SPH_CFG_NAME_STARTINDEX)) {
      startidx_elem = m_referElem->getElem(SPH_CFG_NAME_STARTINDEX);
    }
  }
  else if ( m_referElem->equalsElemName(SPH_CFG_NAME_STARTINDEX) ) {
    startidx_elem = m_referElem;
  }

  if (startidx_elem == NULL) {
    return CFG_ASSIGN_UNKNOWN;
  }
  // assign属性が存在しない
  if (!startidx_elem->existAttribute(SPH_CFG_ATTR_ASSIGN)) {
    return CFG_ASSIGN_UNKNOWN;
  }

  const char* assign_attr = startidx_elem->getAttrString(SPH_CFG_ATTR_ASSIGN);

  // ファイル始点インデックスタイプの取得
  SphCfgAssignType assign = CFG_ASSIGN_UNKNOWN;
  if (strcasecmp(assign_attr, SPH_CFG_ASSIGN_BLOCK) == 0) assign = CFG_ASSIGN_BLOCK;
  else if (strcasecmp(assign_attr, SPH_CFG_ASSIGN_WHOLE) == 0) assign = CFG_ASSIGN_WHOLE;
  else if (strcasecmp(assign_attr, SPH_CFG_ASSIGN_PROCESS) == 0) assign = CFG_ASSIGN_PROCESS;

  return assign;
}



