/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphDataManager.C
 * Data Manager Class Source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphDataManager.C
 * データマネージャクラスソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#include "data/sphDataManager.h"
#include "data/sphDataUtils.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphDataManager::sphDataManager()
{
  m_mapDataList.clear();
}


/**
 * \~english
 * Destoructor
 *
 * \~japanese
 * デストラクタ
 */
sphDataManager::~sphDataManager()
{
  /// データオブジェクトの破棄
  std::map<std::string, sphData*>::iterator itr = m_mapDataList.begin();
  for( ; itr != m_mapDataList.end(); itr++){
    sphData* dataObj = itr->second;
    if( dataObj != NULL ) delete dataObj;
    dataObj = NULL;
  }
  m_mapDataList.clear();
}


/**
 * ボクセルデータを作成、登録する。
 *
 * @param      data_elem        データオブジェクト要素
 * @param      paraCtrlr      パラレルコントローラー
 * @return ボクセルデータオブジェクト
 */
sphData* sphDataManager::createVoxelArray(
                      const sphElem* data_elem,
                      const sphParaController *paraCtrlr)
{
  if (data_elem == NULL || paraCtrlr == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createVoxelArray arguments is null.");
    return NULL;
  }

  if (!data_elem->equalsElemName(SPH_CFG_NAME_DATAOBJ)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphDataObj>.");
    return NULL;
  }

  m_errHandler->blockWarning();            // 警告メッセージを無効にする。

  /// データオブジェクト要素の属性値の取得を行う。
  const char* data_label = data_elem->getAttrString(SPH_CFG_ATTR_LABEL);
  int data_id = data_elem->getAttrInt(SPH_CFG_ATTR_ID);
  SphDCType dctype = data_elem->getAttrSphDcType(SPH_CFG_ATTR_FORMAT);
  SPL_Datatype datatype = data_elem->getAttrSplDataType(SPH_CFG_ATTR_DATATYPE);
  const char* domain = data_elem->getAttrString(SPH_CFG_ATTR_DOMAIN);
  SphCrdDef crddef = data_elem->getAttrSphCrdDef(SPH_CFG_ATTR_CRDDEF);
  size_t data_num = data_elem->getAttrLLong(SPH_CFG_ATTR_DATANUM);
  /// データ数の記述がないベクトルデータクラスの場合は、次元数とする。
  if (data_num == 0) {
    switch(dctype) {
    case SPH_DC_ARRAY_2DN:  /* (i, j, n) */
    case SPH_DC_ARRAY_2DNRV:  /* (n, i, j) */
      data_num = 2;
      break;
    case SPH_DC_ARRAY_3DN:  /* (i, j, k, n) */
    case SPH_DC_ARRAY_3DNRV:  /* (n, i, j, k) */
      data_num = 3;
      break;
    default:
      data_num = 1;
      break;
    }
  }
  size_t gc = data_elem->getAttrLLong(SPH_CFG_ATTR_GC);

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  /// チェックパラメータ
  if (data_label == NULL || dctype == SPH_DC_INVALID || datatype == SPL_DATATYPE_NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "%s", data_elem->debugPrintElem().c_str());
    return NULL;
  }

  sphData* dataObj = NULL;
  if (domain == NULL || strcasecmp(domain, "node") == 0) {
    // 自ノード領域のデータ生成を行う。
    dataObj = createVoxelArray(
                          dctype, datatype,
                          data_label,
                          paraCtrlr,
                          gc,
                          data_num,
                          crddef);
    if (dataObj == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
      return NULL;
    }
  }
  else if (strcasecmp(domain, "whole") == 0) {
    // 計算領域全体のボクセルサイズを取得する。
    const sphSvParaController* svCtrlr = dynamic_cast<const sphSvParaController*>(paraCtrlr);
    if( !svCtrlr ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController dynamic_cast error.");
      return NULL;
    }
    const size_t* wsize = svCtrlr->getWholeVoxelSize();
    // 計算領域全体のデータ生成を行う。
    dataObj = createVoxelArray(
                          dctype, datatype,
                          data_label,
                          wsize,
                          gc,
                          data_num,
                          crddef);
    if (dataObj == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
      return NULL;
    }
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "can not support sphData[@domain]=%s, %s.", domain, data_elem->debugPrintElem().c_str());
    return NULL;
  }

  return dataObj;
}


/**
 * ボクセルデータを作成、登録する。
 *
 * サイズパラメータは、データクラスタイプに関係無く、i,j,k,lの順番とする。<br>
 * （SPH_DC_ARRAY_3DNRVデータクラスタイプでも、lを先頭にする必要はない。）<br>
 * 作成したボクセルデータは、データ識別ラベルと対にしてマップに登録する。
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      label          データ識別ラベル
 * @param      paraCtrlr      パラレルコントローラー
 * @param      gc            ガイドセルサイズ
 * @param      data_num      セルのデータ数
 * @param      crddef      データ定義位置
 * @return ボクセルデータオブジェクト
 */
sphData* sphDataManager::createVoxelArray(
                      SphDCType dcType,
                      SPL_Datatype dataType,
                      const char* label,
                      const sphParaController *paraCtrlr,
                      const size_t gc,
                      const size_t data_num,
                      SphCrdDef crddef)
{

  // データクラス作成
  sphData* voxel = createVoxelArray(
                            dcType, dataType,
                            paraCtrlr,
                            gc, data_num,
                            crddef);
  if (voxel == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object, label=%s.", label);
    return NULL;
  }

  /// データオブジェクトの登録
  if (!registDataObj(label, voxel)) {
    delete voxel;
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "registData fault, label=%s.", label);
    return NULL;
  }

  return voxel;
}



/**
 * ボクセルデータを作成する。(登録無し）
 *
 * サイズパラメータは、データクラスタイプに関係無く、i,j,k,lの順番とする。<br>
 * （SPH_DC_ARRAY_3DNRVデータクラスタイプでも、lを先頭にする必要はない。）<br>
 * 作成したボクセルデータは、登録しない。
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      paraCtrlr      パラレルコントローラー
 * @param      gc            ガイドセルサイズ
 * @param      data_num      セルのデータ数
 * @param      crddef      データ定義位置
 * @return ボクセルデータオブジェクト
 */
sphData* sphDataManager::createVoxelArray(
                              SphDCType dcType,
                              SPL_Datatype dataType,
                              const sphParaController *paraCtrlr,
                              const size_t gc,
                              const size_t data_num,
                              SphCrdDef crddef)
{

  std::string label = "";
  // データクラス作成
  sphData* voxel = NULL;

  // char型
  if (dataType == SPL_CHAR) {
    sphVoxArray<char>* obj = new sphVoxArray<char>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // unsigned char型
  else if (dataType == SPL_UNSIGNED_CHAR) {
    sphVoxArray<unsigned char>* obj = new sphVoxArray<unsigned char>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // short型
  else if (dataType == SPL_SHORT) {
    sphVoxArray<short>* obj = new sphVoxArray<short>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // unsigned short型
  else if (dataType == SPL_UNSIGNED_SHORT) {
    sphVoxArray<unsigned short>* obj = new sphVoxArray<unsigned short>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // int型
  else if (dataType == SPL_INT) {
    sphVoxArray<int>* obj = new sphVoxArray<int>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // unsigned int型
  else if (dataType == SPL_UNSIGNED) {
    sphVoxArray<unsigned int>* obj = new sphVoxArray<unsigned int>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // long型
  else if (dataType == SPL_LONG) {
    sphVoxArray<long>* obj = new sphVoxArray<long>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // unsigned long型
  else if (dataType == SPL_UNSIGNED_LONG) {
    sphVoxArray<unsigned long>* obj = new sphVoxArray<unsigned long>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // long long型(SPL_LONG_LONG=SPL_LONG_LONG_INT)
  else if (dataType == SPL_LONG_LONG) {
    sphVoxArray<long long>* obj = new sphVoxArray<long long>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // float型
  else if (dataType == SPL_FLOAT) {
    sphVoxArray<float>* obj = new sphVoxArray<float>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // double型
  else if (dataType == SPL_DOUBLE) {
    sphVoxArray<double>* obj = new sphVoxArray<double>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support datatype[=%d].", dataType);
  }

  if (voxel == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  else {
    if (voxel->getDataPointer() == NULL) {
      delete voxel;
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data array.");
      return NULL;
    }
  }

  return voxel;
}


/**
 * ボクセルデータを作成、登録する。
 *
 * サイズパラメータは、データクラスタイプに関係無く、i,j,k,data_numの順番とする。<br>
 * data_numはセルのデータ数となる。<br/>
 * （SPH_DC_ARRAY_3DNRVデータクラスタイプでも、セルデータ数を先頭にする必要はない。）<br>
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      label          データ識別ラベル
 * @param      array_size    データサイズ
 * @param      gc            ガイドセルサイズ
 * @param      data_num      l サイズ (セルのデータ数)
 * @param      crddef      データ定義位置
 * @return ボクセルデータオブジェクト
 */
sphData* sphDataManager::createVoxelArray(
                              SphDCType dcType,
                              SPL_Datatype dataType,
                              const char* label,
                              const size_t array_size[3],
                              const size_t gc,
                              const size_t data_num,
                              SphCrdDef crddef)
{

  // データクラス作成
  sphData* voxel = createVoxelArray(
                            dcType, dataType,
                            array_size,
                            gc, data_num, crddef);
  if (voxel == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object, label=%s.", label);
    return NULL;
  }

  /// データオブジェクトの登録
  if (!registDataObj(label, voxel)) {
    delete voxel;
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "registData fault, label=%s.", label);
    return NULL;
  }

  return voxel;
}


/**
 * ボクセルデータを作成する。(登録無し）
 *
 * サイズパラメータは、データクラスタイプに関係無く、i,j,k,data_numの順番とする。<br>
 * data_numはセルのデータ数となる。<br/>
 * （SPH_DC_ARRAY_3DNRVデータクラスタイプでも、セルデータ数を先頭にする必要はない。）<br>
 * 作成したボクセルデータは、登録しない。
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      array_size    データサイズ
 * @param      gc            ガイドセルサイズ
 * @param      data_num      l サイズ (セルのデータ数)
 * @param      crddef      データ定義位置
 * @return ボクセルデータオブジェクト
 */
sphData* sphDataManager::createVoxelArray(
                              SphDCType dcType,
                              SPL_Datatype dataType,
                              const size_t array_size[3],
                              const size_t gc,
                              const size_t data_num,
                              SphCrdDef crddef)
{

  // データクラス作成
  sphData* voxel = NULL;

  // char型
  if (dataType == SPL_CHAR) {
    sphVoxArray<char>* obj = new sphVoxArray<char>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  // unsigned char型
  else if (dataType == SPL_UNSIGNED_CHAR) {
    sphVoxArray<unsigned char>* obj = new sphVoxArray<unsigned char>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  // short型
  else if (dataType == SPL_SHORT) {
    sphVoxArray<short>* obj = new sphVoxArray<short>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  // unsigned short型
  else if (dataType == SPL_UNSIGNED_SHORT) {
    sphVoxArray<unsigned short>* obj = new sphVoxArray<unsigned short>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  // int型
  else if (dataType == SPL_INT) {
    sphVoxArray<int>* obj = new sphVoxArray<int>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  // unsigned int型
  else if (dataType == SPL_UNSIGNED) {
    sphVoxArray<unsigned int>* obj = new sphVoxArray<unsigned int>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  // long型
  else if (dataType == SPL_LONG) {
    sphVoxArray<long>* obj = new sphVoxArray<long>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  // unsigned long型
  else if (dataType == SPL_UNSIGNED_LONG) {
    sphVoxArray<unsigned long>* obj = new sphVoxArray<unsigned long>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  // long long型(SPL_LONG_LONG=SPL_LONG_LONG_INT)
  else if (dataType == SPL_LONG_LONG) {
    sphVoxArray<long long>* obj = new sphVoxArray<long long>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  // float型
  else if (dataType == SPL_FLOAT) {
    sphVoxArray<float>* obj = new sphVoxArray<float>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  // double型
  else if (dataType == SPL_DOUBLE) {
    sphVoxArray<double>* obj = new sphVoxArray<double>(dcType, dataType, array_size, gc, data_num);
    obj->setCrdDef(crddef);
    voxel = obj;
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support datatype[=%d].", dataType);
  }

  if (voxel == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  else {
    if (voxel->getDataPointer() == NULL) {
      delete voxel;
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data array.");
      return NULL;
    }
  }

  return voxel;
}


sphData*
sphDataManager::createUnstData(const sphElem* data_elem,
                               const sphParaController *paraCtrlr) {
  if (data_elem == NULL || paraCtrlr == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createVoxelArray arguments is null.");
    return NULL;
  }

  if (!data_elem->equalsElemName(SPH_CFG_NAME_DATAOBJ)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphDataObj>.");
    return NULL;
  }

  m_errHandler->blockWarning();            // 警告メッセージを無効にする。

  /// データオブジェクト要素の属性値の取得を行う。
  const char* data_label = data_elem->getAttrString(SPH_CFG_ATTR_LABEL);
  SphDCType dctype = data_elem->getAttrSphDcType(SPH_CFG_ATTR_FORMAT);
  SPL_Datatype datatype = data_elem->getAttrSplDataType(SPH_CFG_ATTR_DATATYPE);
  size_t data_num = data_elem->getAttrLLong(SPH_CFG_ATTR_DATANUM);
  /// データ数の記述がないベクトルデータクラスの場合は、次元数とする。
  if (data_num == 0) {
    switch(dctype) {
    case SPH_DC_UNST_NODE:
    case SPH_DC_UNST_ELEM:
      data_num = 1;
      break;
    case SPH_DC_UNST_NODE_N:
    case SPH_DC_UNST_NODE_NRV:
    case SPH_DC_UNST_ELEM_N:
    case SPH_DC_UNST_ELEM_NRV:
      data_num = 3;
/* 次元数が取得できないため、ベクトルデータは3次元のみとする
      if( dims == 1 )      data_num = 1;
      else if( dims == 2 ) data_num = 2;
      else                 data_num = 3;
*/
      break;
    default:
      data_num = 1;
      break;
    }
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  /// チェックパラメータ
  if (data_label == NULL || dctype == SPH_DC_INVALID || datatype == SPL_DATATYPE_NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "%s", data_elem->debugPrintElem().c_str());
    return NULL;
  }

  return createUnstData(dctype, datatype, data_label, paraCtrlr, data_num);
}


sphData*
sphDataManager::createUnstData(SphDCType dcType,
                               SPL_Datatype dtType,
                               const char* label,
                               const sphParaController *paraCtrlr,
                               const size_t vlen) {
  // データクラス作成
  sphData* unst = NULL;

  // char型
  if (dtType == SPL_CHAR) {
    sphUnstData<char>* obj = new sphUnstData<char>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // unsigned char型
  else if (dtType == SPL_UNSIGNED_CHAR) {
    sphUnstData<unsigned char>* obj = new sphUnstData<unsigned char>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // short型
  else if (dtType == SPL_SHORT) {
    sphUnstData<short>* obj = new sphUnstData<short>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // unsigned short型
  else if (dtType == SPL_UNSIGNED_SHORT) {
    sphUnstData<unsigned short>* obj = new sphUnstData<unsigned short>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // int型
  else if (dtType == SPL_INT) {
    sphUnstData<int>* obj = new sphUnstData<int>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // unsigned int型
  else if (dtType == SPL_UNSIGNED) {
    sphUnstData<unsigned int>* obj = new sphUnstData<unsigned int>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // long型
  else if (dtType == SPL_LONG) {
    sphUnstData<long>* obj = new sphUnstData<long>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // unsigned long型
  else if (dtType == SPL_UNSIGNED_LONG) {
    sphUnstData<unsigned long>* obj = new sphUnstData<unsigned long>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // long long型(SPL_LONG_LONG=SPL_LONG_LONG_INT)
  else if (dtType == SPL_LONG_LONG) {
    sphUnstData<long long>* obj = new sphUnstData<long long>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // float型
  else if (dtType == SPL_FLOAT) {
    sphUnstData<float>* obj = new sphUnstData<float>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // double型
  else if (dtType == SPL_DOUBLE) {
    sphUnstData<double>* obj = new sphUnstData<double>(label, dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support datatype[=%d].", dtType);
  }

  if (unst == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  else {
    if (unst->getDataPointer() == NULL) {
      delete unst;
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create unstruct data object.");
      return NULL;
    }
  }

  /// データオブジェクトの登録
  if (!registDataObj(label, unst)) {
    delete unst;
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "registData fault, label=%s.", label);
    return NULL;
  }

  return unst;
}


sphData*
sphDataManager::createUnstData(SphDCType dcType,
                               SPL_Datatype dtType,
                               const sphParaController *paraCtrlr,
                               const size_t vlen) {
  // データクラス作成
  sphData* unst = NULL;

  // char型
  if (dtType == SPL_CHAR) {
    sphUnstData<char>* obj = new sphUnstData<char>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // unsigned char型
  else if (dtType == SPL_UNSIGNED_CHAR) {
    sphUnstData<unsigned char>* obj = new sphUnstData<unsigned char>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // short型
  else if (dtType == SPL_SHORT) {
    sphUnstData<short>* obj = new sphUnstData<short>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // unsigned short型
  else if (dtType == SPL_UNSIGNED_SHORT) {
    sphUnstData<unsigned short>* obj = new sphUnstData<unsigned short>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // int型
  else if (dtType == SPL_INT) {
    sphUnstData<int>* obj = new sphUnstData<int>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // unsigned int型
  else if (dtType == SPL_UNSIGNED) {
    sphUnstData<unsigned int>* obj = new sphUnstData<unsigned int>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // long型
  else if (dtType == SPL_LONG) {
    sphUnstData<long>* obj = new sphUnstData<long>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // unsigned long型
  else if (dtType == SPL_UNSIGNED_LONG) {
    sphUnstData<unsigned long>* obj = new sphUnstData<unsigned long>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // long long型(SPL_LONG_LONG=SPL_LONG_LONG_INT)
  else if (dtType == SPL_LONG_LONG) {
    sphUnstData<long long>* obj = new sphUnstData<long long>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // float型
  else if (dtType == SPL_FLOAT) {
    sphUnstData<float>* obj = new sphUnstData<float>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  // double型
  else if (dtType == SPL_DOUBLE) {
    sphUnstData<double>* obj = new sphUnstData<double>(dcType, dtType, paraCtrlr, vlen);
    unst = obj;
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support datatype[=%d].", dtType);
  }

  if (unst == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  else {
    if (unst->getDataPointer() == NULL) {
      delete unst;
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create unstruct data object.");
      return NULL;
    }
  }

  return unst;
}


/**
 * \~english
 * Get Voxel Data Array Pointer From registered map by data label
 * @param      label          Data Label
 * @return    sphData*      Data Class Pointer
 *
 * \~japanese
 * データ識別ラベルからボクセルデータを登録済みのマップから取得する。
 * @param      label          データ識別ラベル
 * @return    sphData*      データクラスポインタ
 */
sphData* sphDataManager::getDataObj(const char* label) const
{
  std::map<std::string, sphData*>::const_iterator itr = m_mapDataList.find(std::string(label));
  if( itr == m_mapDataList.end() ){
    // error
    m_errHandler->SPH_WARNING_INFO(SPH_WARNING_NOT_EXIST_LABEL, "data label is %s.", label);
    return NULL;
  }

  return itr->second;
}



/**
 * \~english
 * Get Voxel Data Array Pointer From registered map by data pointer
 * @param      ptr          Data pointer
 * @return    sphData*      Data Class Pointer
 *
 * \~japanese
 * データポインタからボクセルデータを登録済みのマップから取得する。
 * @param      ptr          データポインタ
 * @return    sphData*      データクラスポインタ
 */
sphData* sphDataManager::getDataObj(void* ptr) const
{

  std::map<std::string, sphData*>::const_iterator itr = m_mapDataList.begin();
  for( ; itr != m_mapDataList.end(); itr++){
    sphData* dataObj = itr->second;
    if (!dataObj) continue;

    if( ptr == dataObj->getDataPointer() ) {
      return dataObj;
    }
  }

  // error
  m_errHandler->SPH_WARNING_INFO(SPH_WARNING_NOT_EXIST_DATA, "not found data pointer[=%p].", ptr);

  return NULL;
}

/**
 * \~english
 * Delete Data From registered map
 * @param      label          Data Label
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 登録済みのマップからデータを削除する。
 * @param      label          削除データ識別ラベル
 * @return    success(=true)/fault(=false)
 */
bool sphDataManager::deleteDataObj(const char* label)
{
  sphData* data = getDataObj(label);
  if (data == NULL) return false;

  delete data;
  data = NULL;

  m_mapDataList.erase(std::string(label));

  return true;
}


/**
 * 登録済みのマップからデータを削除する。
 * @param      data          削除データオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphDataManager::deleteDataObj(sphData* data)
{
  std::string label = data->getLabel();

  return deleteDataObj(label.c_str());
}


/**
 * データオブジェクトを追加する。
 * @param       label        識別ラベル
 * @param       data         データオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphDataManager::registDataObj(const char* label, sphData* data)
{
  if( !label || !data ) return false;

  std::map<std::string, sphData*>::const_iterator itr = m_mapDataList.find(label);
  if( itr != m_mapDataList.end() ){
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL_DUPLICATEKEY, "can not insert m_mapDataList for duplicate label[=%s].", label);
    return false;
  }

  /// m_mapDataListにデータオブジェクトを追加する。
  if( !((m_mapDataList.insert(std::make_pair(std::string(label), data))).second) ){
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_mapDataList [label=%s].", label);
    return false;
  }

  /// データオブジェクトにラベルを設定する。
  data->setLabel(label);

  return true;
}


/**
 * 部分領域のボクセルデータクラスを作成、登録する。
 *
 * 部分領域ボクセルデータクラスの作成においては、"domain"属性(=node|whole)は無視する。(@domain="node"のみ)
 * @param      data_elem        データオブジェクト要素
 * @param      paraCtrlr      パラレルコントローラー
 * @return ボクセルデータオブジェクト
 */
sphData* sphDataManager::createVoxelContainer(
                      const sphElem* data_elem,
                      const sphParaController *paraCtrlr)
{
  if (data_elem == NULL || paraCtrlr == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createVoxelContainer arguments is null.");
    return NULL;
  }

  if (!data_elem->equalsElemName(SPH_CFG_NAME_DATAOBJ)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphDataObj>.");
    return NULL;
  }

  m_errHandler->blockWarning();            // 警告メッセージを無効にする。

  /// データオブジェクト要素の属性値の取得を行う。
  const char* data_label = data_elem->getAttrString(SPH_CFG_ATTR_LABEL);
  int data_id = data_elem->getAttrInt(SPH_CFG_ATTR_ID);
  SphDCType dctype = data_elem->getAttrSphDcType(SPH_CFG_ATTR_FORMAT);
  SPL_Datatype datatype = data_elem->getAttrSplDataType(SPH_CFG_ATTR_DATATYPE);
  const char* domain = data_elem->getAttrString(SPH_CFG_ATTR_DOMAIN);
  SphCrdDef crddef = data_elem->getAttrSphCrdDef(SPH_CFG_ATTR_CRDDEF);
  size_t data_num = data_elem->getAttrLLong(SPH_CFG_ATTR_DATANUM);
  /// データ数の記述がないベクトルデータクラスの場合は、次元数とする。
  if (data_num == 0) {
    switch(dctype) {
    case SPH_DC_ARRAY_2DN:  /* (i, j, n) */
    case SPH_DC_ARRAY_2DNRV:  /* (n, i, j) */
      data_num = 2;
      break;
    case SPH_DC_ARRAY_3DN:  /* (i, j, k, n) */
    case SPH_DC_ARRAY_3DNRV:  /* (n, i, j, k) */
      data_num = 3;
      break;
    default:
      data_num = 1;
      break;
    }
  }
  size_t gc = data_elem->getAttrLLong(SPH_CFG_ATTR_GC);

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  /// チェックパラメータ
  if (data_label == NULL || dctype == SPH_DC_INVALID || datatype == SPL_DATATYPE_NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "%s", data_elem->debugPrintElem().c_str());
    return NULL;
  }

  sphData* dataObj = NULL;
  // 自ノード領域のデータ生成を行う。
  dataObj = createVoxelArray(
                        dctype, datatype,
                        data_label,
                        paraCtrlr,
                        gc,
                        data_num,
                        crddef);
  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }

  return dataObj;
}


/**
 * 部分領域のボクセルデータクラスを作成、登録する。
 *
 * サイズパラメータは、データクラスタイプに関係無く、i,j,k,lの順番とする。<br>
 * （SPH_DC_ARRAY_3DNRVデータクラスタイプでも、lを先頭にする必要はない。）<br>
 * 作成したボクセルデータは、データ識別ラベルと対にしてマップに登録する。
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      label          データ識別ラベル
 * @param      paraCtrlr      パラレルコントローラー
 * @param      gc            ガイドセルサイズ
 * @param      data_num      セルのデータ数
 * @param      crddef      データ定義位置
 * @return ボクセルデータオブジェクト
 */
sphData* sphDataManager::createVoxelContainer(
                      SphDCType dcType,
                      SPL_Datatype dataType,
                      const char* label,
                      const sphParaController *paraCtrlr,
                      const size_t gc,
                      const size_t data_num,
                      SphCrdDef crddef)
{

  // データクラス作成
  sphData* voxel = createVoxelContainer(
                            dcType, dataType,
                            paraCtrlr,
                            gc, data_num,
                            crddef);
  if (voxel == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object, label=%s.", label);
    return NULL;
  }

  /// データオブジェクトの登録
  if (!registDataObj(label, voxel)) {
    delete voxel;
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "registData fault, label=%s.", label);
    return NULL;
  }

  return voxel;
}



/**
 * 部分領域のボクセルデータクラスを作成する。(登録無し）
 *
 * サイズパラメータは、データクラスタイプに関係無く、i,j,k,lの順番とする。<br>
 * （SPH_DC_ARRAY_3DNRVデータクラスタイプでも、lを先頭にする必要はない。）<br>
 * 作成したボクセルデータは、登録しない。
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      paraCtrlr      パラレルコントローラー
 * @param      gc            ガイドセルサイズ
 * @param      data_num      セルのデータ数
 * @param      crddef      データ定義位置
 * @return ボクセルデータオブジェクト
 */
sphData* sphDataManager::createVoxelContainer(
                              SphDCType dcType,
                              SPL_Datatype dataType,
                              const sphParaController *paraCtrlr,
                              const size_t gc,
                              const size_t data_num,
                              SphCrdDef crddef)
{

  std::string label = "";
  // データクラス作成
  sphData* voxel = NULL;

  // char型
  if (dataType == SPL_CHAR) {
    sphVoxContainer<char>* obj = new sphVoxContainer<char>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // unsigned char型
  else if (dataType == SPL_UNSIGNED_CHAR) {
    sphVoxContainer<unsigned char>* obj = new sphVoxContainer<unsigned char>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // short型
  else if (dataType == SPL_SHORT) {
    sphVoxContainer<short>* obj = new sphVoxContainer<short>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // unsigned short型
  else if (dataType == SPL_UNSIGNED_SHORT) {
    sphVoxContainer<unsigned short>* obj = new sphVoxContainer<unsigned short>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // int型
  else if (dataType == SPL_INT) {
    sphVoxContainer<int>* obj = new sphVoxContainer<int>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // unsigned int型
  else if (dataType == SPL_UNSIGNED) {
    sphVoxContainer<unsigned int>* obj = new sphVoxContainer<unsigned int>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // long型
  else if (dataType == SPL_LONG) {
    sphVoxContainer<long>* obj = new sphVoxContainer<long>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // unsigned long型
  else if (dataType == SPL_UNSIGNED_LONG) {
    sphVoxContainer<unsigned long>* obj = new sphVoxContainer<unsigned long>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // long long型(SPL_LONG_LONG=SPL_LONG_LONG_INT)
  else if (dataType == SPL_LONG_LONG) {
    sphVoxContainer<long long>* obj = new sphVoxContainer<long long>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // float型
  else if (dataType == SPL_FLOAT) {
    sphVoxContainer<float>* obj = new sphVoxContainer<float>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  // double型
  else if (dataType == SPL_DOUBLE) {
    sphVoxContainer<double>* obj = new sphVoxContainer<double>(label, dcType, dataType, paraCtrlr, gc, data_num, crddef);
    voxel = obj;
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support datatype[=%d].", dataType);
  }

  if (voxel == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }

  return voxel;
}

