/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphFileController.C
 * Sphere File Controller Class source file
 * @author     hira
 * @date       2009/09/07
 *
 * \~japanese
 * @file sphFileController.C
 * Sphere ファイルコントローラークラスソースファイル
 * @author     hira
 * @date       2009/09/07
 */

#include "fileio/sphFileController.h"
#include "fileio/sphSbxFile.h"
#include "fileio/sphSpxFile.h"
#include "fileio/sphSphFile.h"
#include "fileio/sphSbxHeader.h"
#include "fileio/sphSpxHeader.h"
#include "fileio/sphSphHeader.h"
#include "config/sphElemUtils.h"


/**
 * コンストラクタ
 */
sphFileController::sphFileController()
                      : m_svFileObj(NULL), m_fileType(NULL), m_fileElem(NULL),
                        m_fileHeader(NULL)
{
  m_formatRankId = 0;
  m_formatSteps = 0;
  m_assignedStartIdx[0] = m_assignedStartIdx[1] = m_assignedStartIdx[2] = 0;
}



/**
 * デストラクタ
 */
sphFileController::~sphFileController()
{
  if (m_svFileObj != NULL) delete m_svFileObj;
  if (m_fileHeader != NULL) delete m_fileHeader;
}


/**
 * ファイル要素を設定する。
 * @param       fileElem       ファイル要素
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::initFileCtrl(const sphElem* file_elem)
{

  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "file_elem is null.");
    return false;
  }

  if ( !file_elem->equalsElemName(SPH_CFG_NAME_FILE)
       && !file_elem->equalsElemName(SPH_CFG_NAME_MULTIFILE) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphFile> or <SphMultiFile>.");
    return false;
  }

  // 参照元のSphFile要素を取得する。
  const sphElem* src_elem = sphElemUtils(file_elem).getFileElem();

  if (src_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "src element is not found, %s.", file_elem->debugPrintElem().c_str());
    return false;
  }

  // 読込パラメータの取得
  const char* format = src_elem->getAttrString(SPH_CFG_ATTR_FORMAT);
  m_fileType = format;

  // ファイルオブジェクトの生成
  if (m_svFileObj != NULL) delete m_svFileObj;
  m_svFileObj = createFileObj(format);
  if (m_svFileObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_svFileObj is null");
    return false;
  }

  // ファイルのヘッダー情報クラスを生成する。
  if (m_fileHeader != NULL) delete m_fileHeader;
  m_fileHeader = dynamic_cast<sphSvHeader*>(createFileHeader(format));
  if (m_fileHeader == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileHeader is null");
    return false;
  }

  // ファイル要素
  m_fileElem = file_elem;


  return true;
}



/**
 * ファイルフォーマットを設定する。
 *
 * writeFile(filename,...)用の初期化メソッド
 * @param       format       ファイルフォーマット('sbx', 'spx', 'sph')
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::initFileCtrl(const char* format)
{

  // ファイルフォーマット
  m_fileType = format;

  // ファイルオブジェクトの生成
  if (m_svFileObj != NULL) delete m_svFileObj;
  m_svFileObj = createFileObj(format);
  if (m_svFileObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_svFileObj is null");
    return false;
  }

  // ファイルのヘッダー情報クラスを生成する。
  if (m_fileHeader != NULL) delete m_fileHeader;
  m_fileHeader = dynamic_cast<sphSvHeader*>(createFileHeader(format));
  if (m_fileHeader == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileHeader is null");
    return false;
  }

  return true;
}



/**
 * ファイルフォーマット文字列からファイルオブジェクトを生成する。
 * @param     file_type         ファイルフォーマット文字列
 * @return    ファイルオブジェクト
 */
sphSvFile* sphFileController::createFileObj(const char* file_type)
{

  if (strcasecmp(file_type, "sbx") == 0) {
    return new sphSbxFile();
  }
  else if (strcasecmp(file_type, "spx") == 0) {
    return new sphSpxFile();
  }
  else if (strcasecmp(file_type, "sph") == 0) {
    return new sphSphFile();
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support file type[=%s].", file_type);
    return NULL;
  }
}


/**
 * サポート対象のファイルフォーマットであるかチェックする。。
 * @param     void
 * @return    true:サポート対象/false:サポート対象外
 */
bool sphFileController::isSupprotedFormat()
{
  if (m_svFileObj != NULL) return true;
  return false;
}


/**
 * 読み込みモードを設定する。
 * @param        mode    読み込みモード
 *                       - ０：全てのノードがファイルをオープンする
 *                       - １：識別子readRankのノードがファイルをオープンし、読み込んだ値を他のノードにブロードキャストする。
 *                            このとき、識別子readRankのノードも情報を保持する。
 * @return    void
 */
void sphFileController::setReadMode(int mode)
{
  if (m_svFileObj == NULL) return;
  m_svFileObj->setReadMode(mode);
}



/**
 * 書き込みモードを設定する。
 * @param        mode    書き込みモード
 *                       - ０：全てのノードがファイルをオープンする
 *                       - １：識別子readRankのノードがファイルをオープンし、他のノードから情報を集めてデータを書き出す。
 *                            このとき、識別子readRankのノードも出力情報を持っている。
 * @return    void
 */
void sphFileController::setWriteMode(int mode)
{
  if (m_svFileObj == NULL) return;
  m_svFileObj->setWriteMode(mode);
}


/**
 * 圧縮書込みを行うかのフラグを設定する。
 * @param     flag    圧縮書込フラグ（true:圧縮書込みを行う。, false:圧縮書込みを行わない。）
 * @return    なし
 */
void sphFileController::setCompressionFlag(bool flag)
{
  if (m_svFileObj == NULL) return;
  m_svFileObj->setCompressionFlag(flag);
}


/**
 * 圧縮ブロックサイズを設定する。
 * @param     block_size        圧縮ブロックサイズ
 * @return    なし
 */
void sphFileController::setBlockSize(size_t block_size)
{
  m_fileHeader->setBlockSize(block_size);
}


/**
 * 拡張レコードを書込みかのフラグを設定する。
 * @param     flag    拡張レコードフラグ（true:拡張レコードを書き込む。, false:拡張レコードを書き込まない。）
 * @return    なし
 */
void sphFileController::setExtRecodeFlag(bool flag)
{
  if (m_svFileObj == NULL) return;
  m_svFileObj->setExtRecodeFlag(flag);
}



/**
 * 読み込みファイルリストにファイル、始点インデックスを追加する。
 * @param      filename      追加ファイル名
 * @param      sidx        読込始点インデックス
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::addSvFileInfo(const char* filename, const size_t sidx[3])
{
  if (m_svFileObj == NULL) return false;
  return m_svFileObj->addSvFileInfo(filename, sidx);
}


/**
 * 読み込みファイルリストにファイルを追加する。
 * @param      filename      追加ファイル名
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::addSvFileName(const char* filename)
{
  if (m_svFileObj == NULL) return false;
  return m_svFileObj->addSvFileName(filename);
}

/**
 * 読み込みファイルリストをクリアする。
 * @param     なし
 * @return    なし
 */
void sphFileController::clearSvFileList()
{
  if (m_svFileObj == NULL) return;
  m_svFileObj->clearSvFileList();
}


/**
 * 読み込みファイルリストを設定する。
 * @param     filelist          ファイルリスト
 * @return    なし
 */
void sphFileController::setSvFileList(splSvFileList* filelist)
{
  if (m_svFileObj == NULL) return;
  return m_svFileObj->setSvFileList(filelist);
}


/**
 * 単一ファイルからファイルのヘッダー情報のみの取得を行う.
 * @param[in]       filename    読込ファイル名
 * @param[out]      header      ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::getFileInfo(sphFileHeader* header)
{
  if (m_svFileObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_svFileObj is null");
    return false;
  }
  if (m_fileElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileElem is null");
    return false;
  }

  m_errHandler->blockWarning();          // 警告メッセージを無効にする。

  char file_name[MAXPATHLEN] = {0x00};
  if (!sphElemUtils(m_fileElem).getFileFormatPath(file_name, m_inputDir, m_formatSteps, m_formatRankId)) {
    // error:@file_name is not found
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "can not get file_name");
    return false;
  }
  const char* type = m_fileElem->getAttrString(SPH_CFG_ATTR_TYPE);
  int io_nodeid = m_fileElem->getAttrInt(SPH_CFG_ATTR_IO_NODEID);

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  if (type == NULL || strcasecmp(type, "in") != 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "type is '%s'.", type);
    return false;
  }
  int mode_num = 0;

  // モード番号の設定
  SphCfgFileIOMode cfg_mode = sphElemUtils(m_fileElem).getFileIoMode();
  if      (cfg_mode == CFG_FILE_DEFAULT) mode_num = 0;     // mode属性が存在しない場合、multiモードとする。
  else if (cfg_mode == CFG_FILE_MULTI) mode_num = 0;
  else if (cfg_mode == CFG_FILE_PROCESS) mode_num = 0;
  else if (cfg_mode == CFG_FILE_WHOLE) mode_num = 1;
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "invalid mode.");
    return false;
  }

  setReadMode(mode_num);                // 読込モード
  setRankId(io_nodeid);             // 読込ランク番号

  return m_svFileObj->getFileInfo(file_name, header);
}


/**
 * 読込済みのファイルのヘッダー情報のみの取得を行う.
 * @param[out]      header     ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
const sphFileHeader* sphFileController::getFileInfo() const
{
  if (m_svFileObj == NULL) return NULL;
  return m_svFileObj->getFileInfo();
}



/**
 * 読込済みのファイルのヘッダー情報のみの取得を行う.
 * @param[out]      header     ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
sphFileHeader* sphFileController::getFileInfo()
{
  if (m_svFileObj == NULL) return NULL;
  return m_svFileObj->getFileInfo();
}


/**
 * ファイルのヘッダー情報の設定を行う.
 * @param[in]      header     ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::setFileInfo(const sphFileHeader* header)
{
  if (m_svFileObj == NULL) return false;
  return m_svFileObj->setFileInfo(header);
}


/**
 *  ファイルの読み込みを行う。.
 *  ファイルリストの作成は、addFileNameメソッドにて追加する。\n
 *  読込パラメータ(mode, readRank, procGrp)は、他のメソッド(setter)にて設定する。\n
 *  読込と同時にデータクラスを作成する。
 *  読込後のファイルヘッダー情報は、getFileInfoにより取得する。\n
 * @param       void
 * @return    success(=true)/fault(=false)
 *         戻り値は、読み込みの成否のみで、エラー番号は、getErrorCodeにて渡す。
 */
sphData* sphFileController::readFile()
{
  if (m_svFileObj == NULL) return NULL;
  if (m_fileElem != NULL) {
    return readFile(m_fileElem);
  }
  else {
    return m_svFileObj->readFile();
  }
}



/**
 *  ファイルの読み込みを行う。.
 *  ファイルリストの作成は、addFileNameメソッドにて追加する。\n
 *  読込パラメータ(mode, readRank, procGrp)は、他のメソッド(setter)にて設定する。\n
 *  読込と同時にデータクラスを作成する。
 *  読込後のファイルヘッダー情報は、getFileInfoにより取得する。\n
 * @param       void
 * @return    success(=true)/fault(=false)
 *         戻り値は、読み込みの成否のみで、エラー番号は、getErrorCodeにて渡す。
 */
sphData* sphFileController::readFile(const sphElem* file_elem)
{
  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "readFile arguments is null");
    return NULL;
  }
  if (m_svFileObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_svFileObj is null");
    return NULL;
  }

  // 参照元のSphFile要素を取得する。
  const sphElem* src_elem = sphElemUtils(file_elem).getFileElem();
  if (src_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "src_elem is null");
    return NULL;
  }

  m_errHandler->blockWarning();          // 警告メッセージを無効にする。

  /// 読込パラメータの取得
  // <SphFile/SphMultiFile>
  const char* type = src_elem->getAttrString(SPH_CFG_ATTR_TYPE);
  const char* mode = src_elem->getAttrString(SPH_CFG_ATTR_MODE);
  int io_nodeid = src_elem->getAttrInt(SPH_CFG_ATTR_IO_NODEID);
  // <SphFileOption>
  const sphElem* opt_elem = file_elem->getElem(SPH_CFG_NAME_FILEOPTION);
  const char* handling = NULL;
  int gc_insph = 0;
  if (opt_elem != NULL) {
    handling = opt_elem->getAttrString(SPH_CFG_ATTR_BITSHANDLING);
    gc_insph = opt_elem->getAttrInt(SPH_CFG_ATTR_GCINSPH);
  }
  /// ファイル名、始点インデックスを設定する。
  const char* top_fname = addFileList(file_elem);

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  if (top_fname == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "not exist file name");
    return NULL;
  }


  if ( src_elem->equalsElemName(SPH_CFG_NAME_FILE) ) {
    if (type == NULL || strcasecmp(type, "in") != 0) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "type is '%s'.", type);
      return NULL;
    }
  }

  int mode_num = 0;        // "multi"=0をデフォルトとする。

  // モード番号の設定
  SphCfgFileIOMode cfg_mode = sphElemUtils(src_elem).getFileIoMode();
  if      (cfg_mode == CFG_FILE_DEFAULT) mode_num = 0;     // mode属性が存在しない場合、multiモードとする。
  else if (cfg_mode == CFG_FILE_MULTI) mode_num = 0;
  else if (cfg_mode == CFG_FILE_PROCESS) mode_num = 0;
  else if (cfg_mode == CFG_FILE_WHOLE) mode_num = 1;
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "invalid mode.");
    return NULL;
  }
  setReadMode(mode_num);                // 読込モード
  setRankId(io_nodeid);             // 読込ランク番号

  register int i;
  int dims = m_fileHeader->getDims();
  sphData* dataObj = m_svFileObj->getData();
  if (dataObj == NULL) {
    // ファイルのヘッダー情報を取得する。
    sphSvHeader* header = dynamic_cast<sphSvHeader*>(createFileHeader(m_fileType));
    if (!m_svFileObj->getFileInfo(top_fname, header)) {
      if (header != NULL) delete header;
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "getFileInfo fault, filename=%s", top_fname);
      return NULL;
    }

    // 読込データ型を取得、設定する。
    m_fileHeader->setDtype(header->getDtype());
    m_fileHeader->setVlen(header->getVlen());
    m_fileHeader->setRlen(header->getRlen());
    if (header != NULL) delete header;
  }
  else {
    // データオブジェクトから読込データ型を取得、設定する。
    if (!setDataType(dataObj->getDataType())) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setDataType fault, filename=%s", top_fname);
      return NULL;
    }
    if (!setVectorType(dataObj->getDcType(), dataObj->getNumOfDtElem())) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setVectorType fault, filename=%s", top_fname);
      return NULL;
    }
  }

  // SPHファイルのGCを設定する。
  setGcInSph(gc_insph);

  /// 読込ヘッダー情報を設定する。
  setFileInfo(m_fileHeader);

  /// 読込実行
  sphData* data = m_svFileObj->readFile();

  // SBXファイルの場合、ビット操作を行うかチェックする。
  if (strcasecmp(m_fileType, "sbx") == 0) {
    if (handling != NULL) {
      if (strcasecmp(handling, SPH_CFG_SBX_MASKUPPER)) {
        // 上位ビットをマスクする。
        data = m_svFileObj->maskUpperBits();
      }
      else if (strcasecmp(handling, SPH_CFG_SBX_SHIFTLOW)) {
        // 下位ビットをシフトする。
        data = m_svFileObj->shiftLowBits();
      }
      else {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "invalid attribute[@%s], %s", SPH_CFG_ATTR_BITSHANDLING, opt_elem->debugPrintElem().c_str());
      }
    }
  }

  return data;
}



/**
 *  ファイルに書込みを行う。
 *  書込みパラメータ、データは、他のメソッド(setter)にて設定する。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::writeFile()
{
  if (m_svFileObj == NULL) return false;
  if (m_fileElem != NULL) {
    return writeFile(m_fileElem);
  }
  else {
    return m_svFileObj->writeFile();
  }
}


/**
 *  ファイルに書込みを行う。
 * @param       file_elem       書込ファイル要素
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::writeFile(const sphElem* file_elem)
{
  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "writeFile arguments is null");
    return false;
  }
  if (m_svFileObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_svFileObj is null");
    return false;
  }

  sphData* dataObj = m_svFileObj->getData();
  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_DATA_IS_NULL, "sphData is null to write.");
    return false;
  }

  // 参照元のSphFile要素を取得する。
  const sphElem* src_elem = sphElemUtils(file_elem).getFileElem();

  m_errHandler->blockWarning();          // 警告メッセージを無効にする。
  /// 書込パラメータの取得
  // <SphFile>
  const char* type = src_elem->getAttrString(SPH_CFG_ATTR_TYPE);
  int io_nodeid = src_elem->getAttrInt(SPH_CFG_ATTR_IO_NODEID);

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  if (type == NULL || strcasecmp(type, "out") != 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "type is '%s'.", type);
    return false;
  }
  int mode_num = 0;
  // モード番号の設定
  SphCfgFileIOMode cfg_mode = sphElemUtils(src_elem).getFileIoMode();
  if      (cfg_mode == CFG_FILE_DEFAULT) mode_num = 0;     // mode属性が存在しない場合、multiモードとする。
  else if (cfg_mode == CFG_FILE_MULTI) mode_num = 0;
  else if (cfg_mode == CFG_FILE_PROCESS) mode_num = 0;
  else if (cfg_mode == CFG_FILE_WHOLE) mode_num = 1;
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "invalid mode.");
    return false;
  }

  // 部分領域管理実行であるかチェックする。
  bool sub_flag = false;
  const sphParaController* para_ctrl = getData()->getParaController();
  if (para_ctrl == NULL) {
    // パラレルコントローラを持っていないので、WHOLE(mode=1)出力は不可
    mode_num = 0;
  }
  else {
    sub_flag = (para_ctrl->getNumberOfSubBlock() > 1);
  }

  setWriteMode(mode_num);                // 書込モード
  setRankId(io_nodeid);             // 書込ランク番号

  m_errHandler->blockWarning();          // 警告メッセージを無効にする。

  const sphElem* header_elem = src_elem->getElem(SPH_CFG_NAME_FILEOPTION);
  if (header_elem != NULL) {
    // 拡張レコードの出力有無
    bool extrec = header_elem->isAttrValue(SPH_CFG_ATTR_EXTRECORD, false);
    setExtRecodeFlag(extrec);

    // 圧縮ブロックサイズ
    if (header_elem->existAttribute(SPH_CFG_ATTR_BLOCKSIZE)) {
      size_t block_size = header_elem->getAttrLong(SPH_CFG_ATTR_BLOCKSIZE);
      setBlockSize(block_size);
      if (block_size == 0) {
        // 圧縮ブロックサイズが０の場合、圧縮を行わない。
        setCompressionFlag(false);
      }
    }

    // 実数タイプ
    int rlen = header_elem->getAttrInt(SPH_CFG_ATTR_RLEN);
    if (rlen != 0) {
      setRlen(rlen);
    }

    // AUX
    // SBXファイルデータの占有ビット数 = 下位４ビット
    unsigned int aux_vaildnum = (unsigned int)header_elem->getAttrInt(SPH_CFG_ATTR_SBXVAILDNUM);
    // SBXファイルデータのデータ存在情報 = 上位４ビット
    unsigned int aux_vaildflag = (unsigned int)header_elem->getAttrInt(SPH_CFG_ATTR_SBXVAILDFLAG);
    unsigned int aux = (aux_vaildflag << 4) + aux_vaildnum;
    setAux(aux);
  }
  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  /// ファイル名を取得する。
  char fname[MAXPATHLEN] = {0};
  if (!sphElemUtils(src_elem).getFileFormatPath(
                                      fname,
                                      m_outputDir,
                                      m_formatSteps, m_formatRankId,
                                      sub_flag) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "can not get write file name");
    return false;
  }
  addSvFileName(fname);

  /// データ種別
  if (!setDataType(dataObj->getDataType())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setDataType fault");
    return false;
  }
  /// ベクトル数
  if (!setVectorType(dataObj->getDcType(), dataObj->getNumOfDtElem())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setVectorType fault");
    return false;
  }

  /// 読込ヘッダー情報を設定する。
  setFileInfo(m_fileHeader);

  /// 書込実行
  return m_svFileObj->writeFile();

}


/**
 * ファイルに書込みを行う。
 * @param filename         出力ファイル名
 * @param mode              出力モード(0=分散出力, 1=一括出力)
 * @param write_rankid      出力ノード番号
 * @param compressFlag      圧縮フラグ(true=データ圧縮を行う)
 * @param extRecodeFlag     拡張レコード出力フラグ(true=拡張レコードの出力を行う)
 * @param aux               AUXレコードの値
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::writeFile(
                          const char* filename,
                          int mode,
                          int write_rankid,
                          bool compressFlag,
                          bool extRecodeFlag,
                          unsigned int aux)
{
  if (m_svFileObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_svFileObj is null");
    return false;
  }

  sphData* dataObj = m_svFileObj->getData();
  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_DATA_IS_NULL, "sphData is null to write.");
    return false;
  }

  /// 書込パラメータ
  setWriteMode(mode);                // 書込モード
  setRankId(write_rankid);             // 書込ランク番号
  setExtRecodeFlag(extRecodeFlag);      // 拡張レコードの出力有無
  setCompressionFlag(compressFlag);        // 圧縮フラグ
  setAux(aux);                        // auxレコード
  addSvFileName(filename);         // 出力ファイル名

  /// データ種別
  if (!setDataType(dataObj->getDataType())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setDataType fault");
    return false;
  }
  /// ベクトル数
  if (!setVectorType(dataObj->getDcType(), dataObj->getNumOfDtElem())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setVectorType fault");
    return false;
  }

  /// 読込ヘッダー情報を設定する。
  setFileInfo(m_fileHeader);

  /// 書込実行
  return m_svFileObj->writeFile();

}


/**
 * \~english
 * sphDataデータクラスを取得する.
 * @param     void
 * @return    sphDataクラス
 */
sphData* sphFileController::getData()
{
  if (m_svFileObj == NULL) return NULL;
  return m_svFileObj->getData();
}


/**
 * sphDataデータクラスを設定する.
 * @param     data        sphDataクラス
 * @return    void
 */
void sphFileController::setData(sphData* data)
{
  if (m_svFileObj == NULL) return;
  m_svFileObj->setData(data);
}


/**
 * 上位ビットデータをマスクし(0とする）、下位ビットのみのデータとする.
 * @param      void
 * @return    sphData pointer
 */
sphData* sphFileController::maskUpperBits()
{
  if (m_svFileObj == NULL) return NULL;
  return m_svFileObj->maskUpperBits();
}

/**
 *  下位ビットをシフトし、上位のみのデータとする。.
 * @param      void
 * @return    sphData pointer
 */
sphData* sphFileController::shiftLowBits()
{
  if (m_svFileObj == NULL) return NULL;
  return m_svFileObj->shiftLowBits();
}



/**
 *  読込/書込プロセスグループを設定する
 * @param      comm                  読込/書込プロセスグループ
 * @return    void
 */
void sphFileController::setProcGrp(SPL_Comm comm)
{
  m_ioComm = comm;
  if (m_svFileObj == NULL) return;
  return m_svFileObj->setProcGrp(comm);
}

/**
 *  読込/書込ランク番号を設定する
 * @param      rankId                   読込/書込ランク番号
 * @return    void
 */
void sphFileController::setRankId(int rankId)
{
  if (m_svFileObj == NULL) return;
  return m_svFileObj->setRankId(rankId);
}


/**
 *  ファイルヘッダーオブジェクトを生成する.
 *  破棄は呼出側で行うこと。
 * @param      void
 * @return     ファイルヘッダーオブジェクト
 */
sphFileHeader* sphFileController::createFileHeader(const char* file_type)
{
  sphFileHeader* header = NULL;
  if (strcasecmp(file_type, "sbx") == 0) {
    header = new sphSbxHeader();
  }
  else if (strcasecmp(file_type, "spx") == 0) {
    header = new sphSpxHeader();
  }
  else if (strcasecmp(file_type, "sph") == 0) {
    header = new sphSphHeader();
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support file type[=%s].", file_type);
    return NULL;
  }

  return header;
}



/**
 *  ファイル要素からファイルリストを作成する
 * @param      file_elem                   ファイル要素
 * @return     先頭ファイル名
 */
const char* sphFileController::addFileList(const sphElem* file_elem)
{

  // 参照元のSphFile要素を取得する。
  sphElemUtils elemUtils(file_elem);
  const sphElem* src_elem = elemUtils.getFileElem();
  // ソルバのbasedirを取得する。
  const char* basedir = NULL;
  if (m_inputDir != NULL) {
    basedir = m_inputDir;
  }
  else {
    basedir = elemUtils.getInputBaseDir();
  }
  int dims = m_fileHeader->getDims();

  /// ファイル名、始点インデックスを設定する。
  const char* top_fname = NULL;
  const char *attr_idx[] = {SPH_CFG_ATTR_I, SPH_CFG_ATTR_J, SPH_CFG_ATTR_K};
  const sphElem* elem = NULL;
  if (src_elem->equalsElemName(SPH_CFG_NAME_FILE)) {
    elem = src_elem;
  }
  else if (src_elem->equalsElemName(SPH_CFG_NAME_MULTIFILE)) {
    elem = src_elem->getChildElemFirst(SPH_CFG_NAME_FILE);
  }

  while (elem != NULL) {
    char file_name[MAXPATHLEN] = {0x00};
    if (!sphElemUtils(elem).getFileFormatPath(file_name, basedir, m_formatSteps, m_formatRankId)) {
      // error:@file_name is not found
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "can not get file_name");
      return false;
    }

    // 始点インデックスを取得する。
    const sphElem* start_elem = elem->getElem(SPH_CFG_NAME_STARTINDEX);

    if (top_fname == NULL) {
      top_fname = file_name;
    }
    if (start_elem != NULL) {
      // 始点インデックスを取得する。
      size_t start_idx[3] = {0, 0, 0};
      // 始点インデックスに@assign属性が設定されている場合は、m_assignedStartIdxから始点インデックスを設定する。
      SphCfgAssignType assign = sphElemUtils(start_elem).getStartIndexAssignType();
      if (assign == CFG_ASSIGN_UNKNOWN) {
        if (!sphElemUtils(start_elem).getDimsAttributeValues(dims, SPH_CFG_NAME_STARTINDEX, attr_idx, start_idx)) {
          // error
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "invalid attributes in SphStartIndex");
          return NULL;
        }
      }
      else {
        start_idx[0] = m_assignedStartIdx[0];
        start_idx[1] = m_assignedStartIdx[1];
        start_idx[2] = m_assignedStartIdx[2];
      }
      // ファイル名、始点インデックスを設定する。
      addSvFileInfo(file_name, start_idx);
    }
    else {
      // ファイル名を設定する。
      addSvFileName(file_name);
    }

    // 次要素検索
    if (file_elem->equalsElemName(SPH_CFG_NAME_MULTIFILE)) {
      elem = src_elem->getChildElemNext(elem, SPH_CFG_NAME_FILE);
    }
    else {
      elem = NULL;
    }
  }

  return top_fname;
}



/**
 * 次元数、ボクセルサイズ、始点インデックス、ガイドセル、全体サイズを設定する
 * @param      dims            次元数
 * @param      size            ボクセルサイズ（ガイドセルを含まない）
 * @param      sta_idx         始点インデックス
 * @param       gc             ガイドセル
 * @param      wsize           全体ボクセルサイズ（ガイドセルを含まない）
 * @return     void
 */
void sphFileController::setVoxelInfo(
                  int dims,
                  const size_t size[3],
                  const size_t sta_idx[3],
                  size_t gc,
                  const size_t wsize[3])
{
  register int i;
  size_t header_size[3] = {1, 1, 1};
  size_t header_wsize[3] = {1, 1, 1};
  for (i=0; i<dims; i++) {
    header_size[i] = size[i] + 2*gc;
    if (wsize != NULL) {
      header_wsize[i] = wsize[i] + 2*gc;
    }
  }
  m_fileHeader->setDims(dims);
  m_fileHeader->setSize(header_size);
  m_fileHeader->setStartIdx(sta_idx);
  m_fileHeader->setGc(gc);
  if (wsize != NULL) {
    m_fileHeader->setWholeSize(header_wsize);
  }

}


/**
 * 入力フォルダを設定する
 * @param      basedir         入力フォルダ
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::setInputBaseDir(const char* basedir)
{
  if (basedir == NULL) return false;
  strcpy(m_inputDir, basedir);
  return true;
}

/**
 * 出力フォルダを設定する
 * @param      basedir         出力フォルダ
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::setOutputBaseDir(const char* basedir)
{
  if (basedir == NULL) return false;
  strcpy(m_outputDir, basedir);
  return true;
}

/**
 * ファイルフォーマットのランク番号を設定する
 * @param      rankid         ランク番号
 * @return     void
 */
void sphFileController::setFileFormatRankId(int rankid)
{
  m_formatRankId = rankid;
  return;
}


/**
 * ファイルフォーマットのステップ番号を設定する
 * @param      steps         ステップ番号
 * @return     void
 */
void sphFileController::setFileFormatSteps(size_t steps)
{
  m_formatSteps = steps;
  return;
}


/**
 * データ種別を設定する。
 * @param     dtype             データ型タイプ
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::setDataType(SPL_Datatype dtype)
{
  int ndlen = spl_get_datasize(dtype);
  /// SBXファイルの場合は、dtypeにセットする
  if (strcasecmp(m_fileType, "sbx") == 0) {
    if (dtype != SPL_UNSIGNED_CHAR && dtype != SPL_UNSIGNED_SHORT && dtype != SPL_UNSIGNED) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support dtype[=%d] of %s.", dtype, m_fileType);
      return false;
    }
    m_fileHeader->setDtype(ndlen);
  }
  /// SPX, SPHファイルの場合は、rlenにセットする。
  else if ( (strcasecmp(m_fileType, "spx") == 0)
        ||  (strcasecmp(m_fileType, "sph") == 0) ) {
    if (dtype != SPL_FLOAT && dtype != SPL_DOUBLE) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support dtype[=%d] of %s.", dtype, m_fileType);
      return false;
    }
    m_fileHeader->setRlen(ndlen);
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support file type[=%s].", m_fileType);
    return false;
  }

  return true;
}


/**
 * ベクトル長を設定する。
 * @param     dctype             データクラス種別
 * @param     data_num            ベクトル長
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::setVectorType(SphDCType dctype, size_t data_num)
{
  /// SBXファイルの場合は、データクラス、ベクトル長チェックを行う。
  if (strcasecmp(m_fileType, "sbx") == 0) {
    switch(dctype) {
    case SPH_DC_ARRAY_1D:    /* (i) */
    case SPH_DC_ARRAY_2D:    /* (i, j) */
    case SPH_DC_ARRAY_3D:    /* (i, j, k) */
      break;
    case SPH_DC_ARRAY_2DNRV:  /* (n, i, j) */
    case SPH_DC_ARRAY_1DNRV:  /* (n, i) */
    case SPH_DC_ARRAY_3DNRV:  /* (n, i, j, k) */
    case SPH_DC_ARRAY_1DN:  /* (i, n) */
    case SPH_DC_ARRAY_2DN:  /* (i, j, n) */
    case SPH_DC_ARRAY_3DN:  /* (i, j, k, n) */
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support dctype[SPH_DC_ARRAY_1DN/2DN/3DN/1DNRV/2DNRV/3DNRV].");
      return false;
      break;
    }
    if (data_num != 1) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support data_num[=%d].", data_num);
      return false;
    }
  }
  /// SPX, SPHファイルの場合は、vlenにセットする。
  else if  ( (strcasecmp(m_fileType, "spx") == 0)
          || (strcasecmp(m_fileType, "sph") == 0) ) {
    int dtype = 1;
    switch(dctype) {
    case SPH_DC_ARRAY_1D:    /* (i) */
    case SPH_DC_ARRAY_2D:    /* (i, j) */
    case SPH_DC_ARRAY_3D:    /* (i, j, k) */
      dtype = 1;
      break;
    case SPH_DC_ARRAY_2DNRV:  /* (n, i, j) */
    case SPH_DC_ARRAY_1DNRV:  /* (n, i) */
    case SPH_DC_ARRAY_3DNRV:  /* (n, i, j, k) */
      dtype = 3;
      break;
    case SPH_DC_ARRAY_1DN:  /* (i, n) */
    case SPH_DC_ARRAY_2DN:  /* (i, j, n) */
    case SPH_DC_ARRAY_3DN:  /* (i, j, k, n) */
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support dctype[SPH_DC_ARRAY_1DN/2DN/3DN].");
      return false;
      break;
    }
    // データ種別（1:スカラ、３:ベクトル）
    m_fileHeader->setDtype(dtype);
    // ベクトル長
    m_fileHeader->setVlen(data_num);
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support file type[=%s].", m_fileType);
    return false;
  }
  return true;
}


/**
 * 実数タイプを設定する。
 * @param     rlen             実数タイプ（4:float, 8:double)
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::setRlen(unsigned int rlen)
{
  /// SBXファイルの場合は、rlenにセットする
  if (strcasecmp(m_fileType, "sbx") == 0) {
    // 実数タイプ
    m_fileHeader->setRlen(rlen);
  }
  /// SPXファイルの場合は、何もしない (setDataTypeで設定する）
  else if (strcasecmp(m_fileType, "spx") == 0) { }
  /// SPHファイルの場合は、何もしない (setDataTypeで設定する）
  else if (strcasecmp(m_fileType, "sph") == 0) { }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support file type[=%s].", m_fileType);
    return false;
  }
  return true;
}



/**
 * AUXを設定する。
 * @param     aux             AUX
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::setAux(unsigned int aux)
{
  if (m_fileHeader == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileHeader is null");
    return false;
  }
  m_fileHeader->setAux(aux);
  return true;
}


/**
 * ステップ、時間を設定する。
 * @param     step             ステップ
 * @param     time             時刻
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::setStepTime(size_t step, double time)
{
  if (m_fileHeader == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileHeader is null");
    return false;
  }

  /// SBXファイルの場合は、何もしない
  if (strcasecmp(m_fileType, "sbx") == 0) {
  }
  /// SPX, SPHファイルの場合, ステップ、時間を設定する。
  else if (  (strcasecmp(m_fileType, "spx") == 0)
          || (strcasecmp(m_fileType, "sph") == 0) ) {
    // 実数タイプ
    m_fileHeader->setStepTime(step, time);
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not support file type[=%s].", m_fileType);
    return false;
  }

  /// ファイル名書式用のステップ数を設定する。
  m_formatSteps = step;

  return true;
}


/**
 * 原点座標、ピッチ、定義位置を設定する。
 * @param     org[3]           原点座標
 * @param     pitch[3]         ピッチ
 * @param     crddef           定義位置
 * @return    success(=true)/fault(=false)
 */
bool sphFileController::setCoordsInfo(double org[3], double pitch[3], SphCrdDef crddef)
{
  if (m_fileHeader == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileHeader is null");
    return false;
  }
  m_fileHeader->setOrigin(org);
  m_fileHeader->setPitch(pitch);
  m_fileHeader->setCrddef((unsigned int)crddef);

  return true;
}



/**
 * SPHファイル内のGCを設定する。
 * @param       gc_insph            SPHファイルに内包するGC
 */
void sphFileController::setGcInSph(int gc_insph)
{

  sphSphHeader* header = dynamic_cast<sphSphHeader*>(m_fileHeader);
  if (header == NULL) return;

  header->m_gc_insph = gc_insph;

}


/**
 * StartIndex要素の始点インデックスを設定する。
 *
 * ファイル読込時のSPX,SBXファイルの始点インデックスを設定する。
 * StartIndex[@assign]の記述に従って始点インデックスを設定する。
 * SPX,SBXファイルの拡張レコードの始点インデックスに相当する。
 * @param sta_idx     始点インデックス
 */
void sphFileController::setAssignedStartIndex(const size_t sta_idx[3])
{
  m_assignedStartIdx[0] = sta_idx[0];
  m_assignedStartIdx[1] = sta_idx[1];
  m_assignedStartIdx[2] = sta_idx[2];
}



