/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSbxFile.C
 * SBX File Input/Output Class
 * @author     hira
 * @date       2009/09/07
 *
 * \~japanese
 * @file sphSbxFile.C
 * SBXファイル入出力クラス
 * @author     hira
 * @date       2009/09/07
 */

#include "fileio/sphSbxFile.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphSbxFile::sphSbxFile():m_header(NULL)
{
  // 初期化する。
  initialize();
}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 * データ配列の破棄を行う。
 */
sphSbxFile::~sphSbxFile()
{
  // 破棄処理を行う。(データ配列の破棄を行う。）
  finalize();
}


/**
 * \~english
 *
 * \~japanese
 *  初期化を行う.
 *  メンバーの初期化（デフォルト値の代入）を行う.
 * @param      なし
 * @return    success(=true)/fault(=false)
 */
bool sphSbxFile::initialize()
{
  // SBXヘッダー情報
  m_header = NULL;

  // 入出力データ
  m_sphArray = NULL;

  /// メンバーの初期化
  m_readMode = 0;    // 読み込みモード
  m_writeMode = 0;   // 書込みモード
  m_compression = true;    // 圧縮書込みを行うかのフラグ
  m_extRecode = false;    // 拡張レコードを書込みかのフラグ

  return true;
}


/**
 * \~english
 *
 * \~japanese
 *  破棄処理を行う.
 *  splSbx, splSvFileList構造体の破棄処理を行う。
 * @param     mem_free_flag    splSbx->data(splData)の開放フラグ
 *                  true:開放を行う。, false:開放を行わない。
 * @return    success(=true)/fault(=false)
 */
bool sphSbxFile::finalize()
{
  // Sbxヘッダー情報の破棄処理
  if (m_header != NULL) {
    m_header->finalize();
    m_header = NULL;
  }

  return true;
}


/**
 * \~english
 *
 *
 * \~japanese
 *  複数のSBXファイルの読み込みを行う。.
 *  ファイルリストの作成は、addFileNameメソッドにて追加する。\n
 *  読込パラメータ(mode, readRank, procGrp)は、他のメソッド(setter)にて設定する。\n
 *  読込と同時にデータクラスを作成する。
 *  読込後のファイルヘッダー情報は、getFileInfoにより取得する。\n
 * @return    ファイルデータ
 */
sphData* sphSbxFile::readFile()
{
  register int i, j;
  int flg = 0, tmp = 0;

  if (flg != 1) {
    // 読込情報が設定されていないので、読込ファイルから作成する。
    if (m_header == NULL) {
      m_header = new sphSbxHeader();
      if (!createFileHeader(m_header)) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createFileHeader error.");
        flg = 1;
      }
    }
  }

  if (flg != 1) {
    // 読込情報のチェック
    if (!m_header->checkHeader()) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "sphSbxHeader->checkHeader error.");
      flg = 1;
    }
  }

  // splSbx構造体オブジェクトの生成
  splSbx* obj = new splSbx;
  if (!splInitSbx(obj)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault splInitSbx.");
    flg = 1;
  }

  if (flg != 1) {
    if (m_header != NULL) {
      // 読込ヘッダー情報をsplSbx構造体、データにセットする。
      m_header->getSplSbxHeader(obj);
    }

    // 読込データ配列の作成
    size_t buf_size = m_header->getVoxelDataLength();
    if (buf_size == 0) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "malloc allocate size is zero.");
      flg = 1;
    }
    if (m_sphArray == NULL) {
      if (buf_size > 0) {
        // sphDataクラスを生成する。(作成データポインタをファイルデータ読込データ(obj->data.c)にセットする。)
        m_sphArray = createSphArray(obj);
        if (m_sphArray == NULL) {
          // error
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_MEMORY_ALLOCATE, "allocate size is %ld.", buf_size);
          flg = 1;
        }
      }
    }
    else {
      // 生成済みsphDataクラスのポインタを読込データポインタにセットする。
      // サイズチェック
      size_t array_size = m_sphArray->getArrayLength();         // GC込みデータ長
      SPL_Datatype array_dtype = m_sphArray->getDataType();
      if (array_size*spl_get_datasize(array_dtype) >= buf_size) {
        obj->data.c = (unsigned char*)m_sphArray->getDataPointer();
      }
      else {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not equal data length to readFile, header=%ld, dataObj=%ld.", buf_size, array_size*spl_get_datasize(array_dtype));
        flg = 1;
      }
    }
  }

  // error check
  if( flg != 0 ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_ALLREDUCE_SUM, "pre readFile error");
    splFinalizeSbx(obj, true);
    delete obj;
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return NULL;
  }

  // SBXファイルの読込を行う。
  if (!splReadSbx(
        m_fileList,
        m_readMode,
        obj,
        m_rankId,
        m_procGrp)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_READ_FILE, "fault splReadSbx.");
    free(obj->data.c);
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return NULL;
  }

  // ヘッダー情報を設定する。
  m_header->setSplSbxHeader(obj);

  /// データオブジェクトにデータ定義位置を設定する。
  if (m_sphArray->getCrdDef() == CRDDEF_UNKNOWN) {
    if (obj->crddef <= 4) {
      m_sphArray->setCrdDef((SphCrdDef)obj->crddef);
    }
  }

  // splSbx構造体オブジェクトの破棄（読込データは破棄しない）
  splFinalizeSbx(obj, false);

  delete obj;

  return m_sphArray;
}


/**
 * \~english
 *
 *
 * \~japanese
 * 読込済みのSBXファイルのヘッダー情報のみの取得を行う.
 * @param[out]      header     ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
const sphFileHeader* sphSbxFile::getFileInfo() const
{
  return m_header;
}


/**
 * \~english
 *
 *
 * \~japanese
 * 読込済みのSBXファイルのヘッダー情報のみの取得を行う.
 * @param[out]      header     ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
sphFileHeader* sphSbxFile::getFileInfo()
{
  return m_header;
}

/**
 * \~english
 *
 *
 * \~japanese
 * SBXファイルのヘッダー情報の設定を行う.
 * 引数ヘッダー情報(header)からコピーを作成して、m_headerにヘッダー情報を設定する。
 * @param[in]      header     ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
bool sphSbxFile::setFileInfo(const sphFileHeader* header)
{

  if (header == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "header is null.");
    return false;
  }

  const sphSbxHeader* src_header = dynamic_cast<const sphSbxHeader*>(header);
  if (src_header == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "cast error from sphFileHeader* to sphSbxHeader*.");
    return false;
  }

  if (m_header == NULL) {
    m_header = new sphSbxHeader();
  }
  // ヘッダー情報をコピーする。
  *m_header = *src_header;

  return true;
}


/**
 * \~english
 *
 *
 * \~japanese
 * 単一SBXファイルからSBXファイルのヘッダー情報のみの取得を行う.
 * 読込ヘッダー情報は、メンバーのヘッダー情報変数(m_header)にはセットされない。
 * @param[in]       filename    読込ファイル名
 * @param[out]      header      ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
bool sphSbxFile::getFileInfo(const char* filename, sphFileHeader* header)
{
  int flg = 0, tmp = 0;
  sphSbxHeader* sbxHeader = NULL;

  if (filename == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "filename is null.");
    flg = 1;
  }

  if (header == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "header is null.");
    flg = 1;
  }
  else {
    sbxHeader = dynamic_cast<sphSbxHeader*>(header);
    if (sbxHeader == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "cast error from sphFileHeader* to sphSbxHeader*.");
      flg = 1;
    }
  }

  // splSbx構造体オブジェクトの生成
  splSbx* obj = new splSbx;
  if (!splInitSbx(obj)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault splInitSbx.");
    flg = 1;
  }

  // error check
  if( flg != 0 ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_ALLREDUCE_SUM, "pre getFileInfo error");
    splFinalizeSbx(obj, true);
    delete obj;
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }

  // SBXファイル情報を取得する。
  if (!splGetSbxInfo(filename,
          obj,
          m_readMode,
          m_rankId,
          m_procGrp)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault splGetSbxInfo, filename is %s.", filename);
    splFinalizeSbx(obj, false);
    delete obj;
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }

  // ヘッダー情報(splSbx構造体）からヘッダー情報(sphSbxHeader) を設定する。
  sbxHeader->setSplSbxHeader(obj);

  // splSbx構造体オブジェクトの破棄（読込データは破棄しない）
  splFinalizeSbx(obj, false);
  delete obj;

  return true;
}


/**
 * \~english
 *
 * \~japanese
 *  単一SBXファイルに書込みを行う。
 *  書込みパラメータ、データは、他のメソッド(setter)にて設定する。
 * @param      filename    書込ファイル名
 * @param      procGrp     プロセスグループ
 * @return    success(=true)/fault(=false)
 */
bool sphSbxFile::writeFile()
{

  int flg = 0, tmp = 0;

  // 書き込み前チェック、準備を行う。
  void* pData = preWriteFile();
  if (pData == NULL) {
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault preWriteFile.");
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }

  // 書込ファイルの取得
  char filename[MAXPATHLEN];
  getFileName(0, filename);

  // splSbx構造体オブジェクトの生成
  splSbx* obj = new splSbx;
  if (!splInitSbx(obj)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault splInitSbx.");
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }
  // データポインタをセットする。
  obj->data.c = (unsigned char*)pData;

  // ヘッダー情報をsplSbx構造体に設定する。
  m_header->getSplSbxHeader(obj);

  // SBXファイルへの書込を行う。
  if (!splWriteSbx(filename,
          m_writeMode,
          (SplBool)m_compression,
          (SplBool)m_extRecode,
          m_header->m_block_size,
          obj,
          m_rankId,
          m_procGrp)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_WRITE_FILE, "fault splWriteSbx.");
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }

  return true;
}


/**
 * \~english
 * set mask on Upper bits
 * \~japanese
 * 上位ビットデータをマスクし(0とする）、下位ビットのみのデータとする.
 * 占有ビット数 : 上位データビットのビット数
 * データ存在情報 :   0 = 上位、下位ビット共に有効な情報が記述されている場合
 *           1 = 下位のみに有効な情報が記述されている場合
 *           2 = 上位のみに有効な情報が記述されている場合
 * - unsigned charデータタイプのみマスクを行う。(return false)
 * - 占有ビット数=0の場合、何もしない。(return true)
 * - データ存在情報=0の場合、マスクを行う。(return true)
 * - データ存在情報=1の場合、マスクを行う。(return true)
 * - データ存在情報=2の場合、すべて0x00とする。(return true)
 * - データクラスが設定済みの場合は、エラーとする。（何もしない）
 * @param      なし
 * @return    success(=true)/fault(=false)
 */
sphData* sphSbxFile::maskUpperBits()
{
  if (m_header == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_header is null.");
    return NULL;
  }

  // unsigned charデータタイプのみマスクを行う。
  if (m_header->m_dtype != 1) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_DTYPE, "only 'unsigned char' dtype, this dtype is %d.\n", m_header->m_dtype);
    return NULL;
  }
  // 次元数チェック
  if (m_header->m_dims == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_DIMS, "dims is zero.");
    return NULL;
  }
  // 読込データ
  if (m_sphArray == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_DATA_IS_NULL, "m_sphArray is null.");
    return NULL;
  }

  register int i;

  // AUXデータの取得
  unsigned int bits, dataflag;
  m_header->getDataExistFlags(bits, dataflag);
  // 占有ビット数=0の場合、何もしない。
  if (bits == 0) {
    // error
    m_errHandler->SPH_WARNING_INFO(SPH_WARNING_DO_NOTHING, "bits in aux is %d.", bits);
    return m_sphArray;
  }

  // データ取得
  sphArray<unsigned char>* array = dynamic_cast<sphArray<unsigned char>*>(m_sphArray);
  if (array == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "cast error from sphData* to sphArray*.");
    return false;
  }
  unsigned char* data = array->getData();

  // データサイズ取得
  if (data == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_DATA_IS_NULL, "char pointer in data is null.");
    return NULL;
  }
  size_t size = array->getArrayLength();
  if (size == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ARRAY_SIZE, "size is zero.");
    return NULL;
  }

  // データ存在情報=2の場合、すべて0x00とする。
  if (dataflag == 2) {
    memset(data, 0x00, size);
  }
  // データ存在情報=0,1の場合、マスクを行う。
  else if (dataflag == 0 || dataflag == 1) {
    unsigned char mask = 0xFF >> bits;
    for (i=0; i<size; i++) {
      *(data+i) = *(data+i) & mask;
    }
  }

  return m_sphArray;
}


/**
 * \~english
 *
 * \~japanese
 *  下位ビットをシフトし、上位のみのデータとする。.
 * 占有ビット数 : 上位データビットのビット数
 * データ存在情報 :   0 = 上位、下位ビット共に有効な情報が記述されている場合
 *           1 = 下位のみに有効な情報が記述されている場合
 *           2 = 上位のみに有効な情報が記述されている場合
 * - unsigned charデータタイプのみマスクを行う。
 * - 占有ビット数=8の場合、何もしない。
 * - データ存在情報=0の場合、シフトを行う。
 * - データ存在情報=1の場合、すべて0x00とする。
 * - データ存在情報=2の場合、シフトを行う。
 * - データクラスが設定済みの場合は、エラーとする。（何もしない）
 * @param      なし
 * @return    success(=true)/fault(=false)
 */
sphData* sphSbxFile::shiftLowBits()
{

  if (m_header == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_header is null.");
    return NULL;
  }

  // unsigned charデータタイプのみシフトを行う。
  if (m_header->m_dtype != 1) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_DTYPE, "only 'unsigned char' dtype, this dtype is %d.\n", m_header->m_dtype);
    return NULL;
  }
  // 次元数チェック
  if (m_header->m_dims == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_DIMS, "dims is zero.");
    return NULL;
  }
  // 読込データ
  if (m_sphArray == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_DATA_IS_NULL, "m_sphArray is null.");
    return NULL;
  }

  register int i;
  // AUXデータの取得
  unsigned int bits, dataflag;
  m_header->getDataExistFlags(bits, dataflag);
  // 占有ビット数=8の場合、何もしない。
  if (bits >= 8) {
    // error
    m_errHandler->SPH_WARNING_INFO(SPH_WARNING_DO_NOTHING, "bits in aux is %d.", bits);
    return m_sphArray;
  }

  // データ取得
  sphArray<unsigned char>* array = dynamic_cast<sphArray<unsigned char>*>(m_sphArray);
  if (array == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "cast error from sphData* to sphArray*.");
    return NULL;
  }
  unsigned char* data = array->getData();

  // データサイズ取得
  if (data == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_DATA_IS_NULL, "char pointer in data is null.");
    return NULL;
  }
  size_t size = array->getArrayLength();
  if (size == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ARRAY_SIZE, "size is zero.");
    return NULL;
  }

  // 占有ビット数=0の場合、すべて0x00とする。
  unsigned int shift = 8 - bits;
  // データ存在情報=1の場合、すべて0x00とする。
  if (dataflag == 1 || shift == 8) {
    memset(data, 0x00, size);
  }
  // データ存在情報=0,2の場合、シフトを行う。
  else if (dataflag == 0 || dataflag == 1) {
    for (i=0; i<size; i++) {
      *(data+i) = *(data+i) >> shift;
    }
  }

  return m_sphArray;
}


/**
 * \~english
 * set read mode
 * @param        mode    read mode
 *                       - 0:all nodes read
 *                       - 1:only readRank read data
 *                       - 2:only readRank read data, and readRank not expand data.
 * @return      void
 *
 * \~japanese
 * 読み込みモードを設定する。
 * @param        mode    読み込みモード
 *                       - ０：全てのノードがファイルをオープンする
 *                       - １：識別子readRankのノードがファイルをオープンし、読み込んだ値を他のノードにブロードキャストする。
 *                            このとき、識別子readRankのノードも情報を保持する。
 *                       - ２：識別子readRankのノードがファイルをオープンし、読み込んだ値を他のノードにブロードキャストする。
 *                            このとき、識別子readRankのノードは情報を保持しない。
 * @return    void
 */
void sphSbxFile::setReadMode(int mode)
{
  m_readMode = mode;
}



/**
 * \~english
 * set write mode
 * @param        mode    write mode
 *                       - 0:all nodes write
 *                       - 1:only writeRank write data. Output gather data.
 *                       - 2:only writeRank write data. Output gather data.
 *                           and writeRank has not my file object.
 * @return      void
 *
 * \~japanese
 * 書き込みモードを設定する。
 * @param        mode    書き込みモード
 *                       - ０：全てのノードがファイルをオープンする
 *                       - １：識別子readRankのノードがファイルをオープンし、他のノードから情報を集めてデータを書き出す。
 *                            このとき、識別子readRankのノードも出力情報を持っている。
 *                       - ２：識別子readRankのノードがファイルをオープンし、他のノードから情報を集めてデータを書き出す。
 *                            このとき、識別子readRankのノードは出力情報を持たない。
 * @return    void
 */
void sphSbxFile::setWriteMode(int mode)
{
  m_writeMode = mode;
}

/**
 * \~english
 *
 * \~japanese
 * 圧縮書込みを行うかのフラグを設定する。
 * @param     flag    圧縮書込フラグ（true:圧縮書込みを行う。, false:圧縮書込みを行わない。）
 * @return    なし
 */
void sphSbxFile::setCompressionFlag(bool flag)
{
  m_compression = flag;
}


/**
 * \~english
 *
 * \~japanese
 * 圧縮書込みを行うかのフラグを取得する。
 * @param     void
 * @return    圧縮書込フラグ（true:圧縮書込みを行う。, false:圧縮書込みを行わない。）
 */
bool sphSbxFile::isCompressionFlag()
{
  return m_compression;
}


/**
 * 拡張レコードを書込みかのフラグを設定する。
 * @param     flag    拡張レコードフラグ（true:拡張レコードを書き込む。, false:拡張レコードを書き込まない。）
 * @return    なし
 */
void sphSbxFile::setExtRecodeFlag(bool flag)
{
  m_extRecode = flag;
}


/**
 * \~english
 *
 * \~japanese
 * sphDataクラスを生成する.
 * - データクラスタイプ : SPH_DC_ARRAY_1D, SPH_DC_ARRAY_2D, SPH_DC_ARRAY_1D
 * - データ型タイプ : SPL_UNSIGNED_CHAR, SPL_UNSIGNED_SHORT, SPL_UNSIGNED
 * @param     data    sphDataクラス
 * @return    success(=true)/fault(=false)
 */
sphData* sphSbxFile::createSphArray(splSbx* sbxObj)
{

  if (sbxObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "splSbx object is null.");
    return NULL;
  }

  // 次元数チェック
  if (sbxObj->dims == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_DIMS, "dims is zero.");
    return NULL;
  }
  // データサイズ取得
  register int i;
  size_t size[3] = {0, 0, 0};
  for (i=0; i<sbxObj->dims; i++) {
    // サイズチェック
    if (sbxObj->size[i] <= sbxObj->gc*2) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ARRAY_SIZE, "size have to be more than gc*2, size[%d]:%ld <= gc[%d]:%ld * 2", i, sbxObj->size[i], i, sbxObj->gc);
      return NULL;
    }
    // ヘッダー情報のサイズはGCを含んだサイズであるので、データクラスのサイズのGCを含まないサイズにする。
    size[i] = sbxObj->size[i] - sbxObj->gc*2;
  }

  // データクラス作成
  sphData* voxel = NULL;

  // データクラスタイプ
  SphDCType dctype = SPH_DC_INVALID;
  if (sbxObj->dims == 1) {
    if (sbxObj->vlen == 1) dctype = SPH_DC_ARRAY_1D;
    else  dctype = SPH_DC_ARRAY_1DN;
  }
  else if (sbxObj->dims == 2) {
    if (sbxObj->vlen == 1) dctype = SPH_DC_ARRAY_2D;
    else  dctype = SPH_DC_ARRAY_2DN;
  }
  else if (sbxObj->dims == 3) {
    if (sbxObj->vlen == 1) dctype = SPH_DC_ARRAY_3D;
    else  dctype = SPH_DC_ARRAY_3DN;
  }

  // データ型タイプ
  SPL_Datatype datatype = SPL_DATATYPE_NULL;
  if (sbxObj->dtype == 1) {
    datatype = SPL_UNSIGNED_CHAR;
    sphVoxArray<unsigned char>* array =
                new sphVoxArray<unsigned char>(
                          dctype,
                          datatype,
                          size,
                          sbxObj->gc,
                          sbxObj->vlen);
    if (array != NULL) sbxObj->data.c = array->getData();
    voxel = array;
  }
  else if (sbxObj->dtype == 2) {
    datatype = SPL_UNSIGNED_SHORT;
    sphVoxArray<unsigned short>* array =
                new sphVoxArray<unsigned short>(
                          dctype,
                          datatype,
                          size,
                          sbxObj->gc,
                          sbxObj->vlen);
    if (array != NULL) sbxObj->data.s = array->getData();
    voxel = array;

  }
  else if (sbxObj->dtype == 4) {
    datatype = SPL_UNSIGNED;
    sphVoxArray<unsigned int>* array =
                new sphVoxArray<unsigned int>(
                          dctype,
                          datatype,
                          size,
                          sbxObj->gc,
                          sbxObj->vlen);
    if (array != NULL) sbxObj->data.i = array->getData();
    voxel = array;
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_DTYPE, "invalid dtype=%d.", sbxObj->dtype);
    return NULL;
  }

  if (voxel == NULL || sbxObj->data.c == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_DATA_IS_NULL, "can not create sphData.");
    return NULL;
  }

  // ファイルクラスにデータクラスを設定する。
  m_sphArray = voxel;

  return voxel;
}


/**
 * \~english
 * get sphData pointer
 * @param     void
 * @return    sphData object pointer
 *
 * \~japanese
 * sphDataデータクラスを取得する.
 * @param     void
 * @return    sphDataクラス
 */
sphData* sphSbxFile::getData()
{
  return m_sphArray;
}


/**
 * \~english
 * set sphData pointer
 * @param     data        sphData object pointer
 * @return    void
 *
 * \~japanese
 * sphDataデータクラスを設定する.
 * @param     data        sphDataクラス
 * @return    void
 */
void sphSbxFile::setData(sphData* data)
{
  m_sphArray = data;
}


/**
 * \~english
 * get read mode
 * @param     void
 * @return    read mode
 *
 * \~japanese
 * 読み込みモードを取得する。
 * @param     void
 * @return    読込モード
 */
int sphSbxFile::getReadMode() const
{
  return m_readMode;
}

/**
 * \~english
 * get write mode
 * @param     void
 * @return    write mode
 *
 * \~japanese
 * 書き込みモードを取得する。
 * @param     void
 * @return    書込モード
 */
int sphSbxFile::getWriteMode() const
{
  return m_writeMode;
}


