/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSpxFile.C
 * SPX File Input/Output Class
 * @author     hira
 * @date       2009/09/07
 *
 * \~japanese
 * @file sphSpxFile.C
 * SPXファイル入出力クラス
 * @author     hira
 * @date       2009/09/07
 */

#include "fileio/sphSpxFile.h"


/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphSpxFile::sphSpxFile():m_header(NULL)
{
  // 初期化する。
  initialize();
}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 * データ配列の破棄を行う。
 */
sphSpxFile::~sphSpxFile()
{
  // 破棄処理を行う。(データ配列の破棄を行う。）
  finalize();
}


/**
 * \~english
 *
 * \~japanese
 *  初期化を行う.
 *  メンバーの初期化（デフォルト値の代入）を行う.
 * @param      なし
 * @return    success(=true)/fault(=false)
 */
bool sphSpxFile::initialize()
{
  // SPXヘッダー情報
  m_header = NULL;

  // 入出力データ
  m_sphArray = NULL;

  /// メンバーの初期化
  m_readMode = 0;    // 読み込みモード
  m_writeMode = 0;   // 書込みモード
  m_compression = true;    // 圧縮書込みを行うかのフラグ
  m_extRecode = false;    // 拡張レコードを書込みかのフラグ

  return true;
}


/**
 * \~english
 *
 * \~japanese
 *  破棄処理を行う.
 *  splSpx, splSvFileList構造体の破棄処理を行う。
 * @param     mem_free_flag    splSpx->data(splData)の開放フラグ
 *                  true:開放を行う。, false:開放を行わない。
 * @return    success(=true)/fault(=false)
 */
bool sphSpxFile::finalize()
{
  // SPXヘッダー情報の破棄処理
  if (m_header != NULL) {
    m_header->finalize();
    m_header = NULL;
  }

  return true;
}



/**
 * \~english
 *
 *
 * \~japanese
 *  複数のSPXファイルの読み込みを行う。.
 *  ファイルリストの作成は、addFileNameメソッドにて追加する。\n
 *  読込パラメータ(mode, readRank, procGrp)は、他のメソッド(setter)にて設定する。\n
 *  読込と同時にデータクラスを作成する。
 *  読込後のファイルヘッダー情報は、getFileInfoにより取得する。\n
 * @return    ファイルデータ
 */
sphData* sphSpxFile::readFile()
{
  register int i, j;
  int flg = 0, tmp = 0;

  if (flg != 1) {
    // 読込情報が設定されていないので、読込ファイルから作成する。
    if (m_header == NULL) {
      m_header = new sphSpxHeader();
      if (!createFileHeader(m_header)) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createFileHeader error.");
        flg = 1;
      }
    }
  }

  if (flg != 1) {
    // 読込情報のチェック
    if (!m_header->checkHeader()) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "sphSpxHeader->checkHeader error.");
      flg = 1;
    }
  }

  // splSpx構造体オブジェクトの生成
  splSpx* obj = new splSpx;
  if (!splInitSpx(obj)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault splInitSpx.");
    flg = 1;
  }

  if (flg != 1) {
    if (m_header != NULL) {
      // 読込ヘッダー情報をsplSpx構造体、データにセットする。
      m_header->getSplSpxHeader(obj);
    }

    // 読込データ配列の作成
    size_t buf_size = m_header->getVoxelDataLength();
    if (buf_size == 0) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "malloc allocate size is zero.");
      flg = 1;
    }
    if (m_sphArray == NULL) {
      if (buf_size > 0) {
        // sphDataクラスを生成する。(作成データポインタをファイルデータ読込データ(obj->data.c)にセットする。)
        m_sphArray = createSphArray(obj);
        if (m_sphArray == NULL) {
          // error
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_MEMORY_ALLOCATE, "allocate size is %ld.", buf_size);
          flg = 1;
        }
      }
    }
    else {
      // 生成済みsphDataクラスのポインタを読込データポインタにセットする。
      // サイズチェック
      size_t array_size = m_sphArray->getArrayLength();         // GC込みデータ長
      SPL_Datatype array_dtype = m_sphArray->getDataType();
      if (array_size*spl_get_datasize(array_dtype) == buf_size) {
        obj->data.c = (unsigned char*)m_sphArray->getDataPointer();
      }
      else {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not equal data length to readFile, header=%ld, dataObj=%ld.", buf_size, array_size*spl_get_datasize(array_dtype));
        flg = 1;
      }
    }
  }

  // error check
  if( flg != 0 ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_ALLREDUCE_SUM, "pre readFile error");
    splFinalizeSpx(obj, true);
    delete obj;
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return NULL;
  }

  // SPXファイルの読込を行う。
  if (!splReadSpx(
        m_fileList,
        m_readMode,
        obj,
        m_rankId,
        m_procGrp)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_READ_FILE, "fault splReadSpx.");
    free(obj->data.c);
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return NULL;
  }

  // ヘッダー情報を設定する。
  m_header->setSplSpxHeader(obj);

  /// データオブジェクトにデータ定義位置を設定する。
  if (m_sphArray->getCrdDef() == CRDDEF_UNKNOWN) {
    if (obj->crddef <= 4) {
      m_sphArray->setCrdDef((SphCrdDef)obj->crddef);
    }
  }

  // splSpx構造体オブジェクトの破棄（読込データは破棄しない）
  splFinalizeSpx(obj, false);

  delete obj;

  return m_sphArray;
}


/**
 * \~english
 * set read mode
 * @param        mode    read mode
 *                       - 0:all nodes read
 *                       - 1:only readRank read data
 *                       - 2:only readRank read data, and readRank not expand data.
 * @return      void
 *
 * \~japanese
 * 読み込みモードを設定する。
 * @param        mode    読み込みモード
 *                       - ０：全てのノードがファイルをオープンする
 *                       - １：識別子readRankのノードがファイルをオープンし、読み込んだ値を他のノードにブロードキャストする。
 *                            このとき、識別子readRankのノードも情報を保持する。
 *                       - ２：識別子readRankのノードがファイルをオープンし、読み込んだ値を他のノードにブロードキャストする。
 *                            このとき、識別子readRankのノードは情報を保持しない。
 * @return    void
 */
void sphSpxFile::setReadMode(int mode)
{
  m_readMode = mode;
}


/**
 * \~english
 * set write mode
 * @param        mode    write mode
 *                       - 0:all nodes write
 *                       - 1:only writeRank write data. Output gather data.
 *                       - 2:only writeRank write data. Output gather data.
 *                           and writeRank has not my file object.
 * @return      void
 *
 * \~japanese
 * 書き込みモードを設定する。
 * @param        mode    書き込みモード
 *                       - ０：全てのノードがファイルをオープンする
 *                       - １：識別子writeRankのノードがファイルをオープンし、他のノードから情報を集めてデータを書き出す。
 *                            このとき、識別子writeRankのノードも出力情報を持っている。
 *                       - ２：識別子writeRankのノードがファイルをオープンし、他のノードから情報を集めてデータを書き出す。
 *                            このとき、識別子writeRankのノードは出力情報を持たない。
 * @return    void
 */
void sphSpxFile::setWriteMode(int mode)
{
  m_writeMode = mode;
}


/**
 * \~english
 *
 *
 * \~japanese
 * 読込済みのSPXファイルのヘッダー情報のみの取得を行う.
 * @param[out]      header     ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
const sphFileHeader* sphSpxFile::getFileInfo() const
{
  return m_header;
}


/**
 * \~english
 *
 *
 * \~japanese
 * 読込済みのSPXファイルのヘッダー情報のみの取得を行う.
 * @param[out]      header     ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
sphFileHeader* sphSpxFile::getFileInfo()
{
  return m_header;
}


/**
 * \~english
 *
 *
 * \~japanese
 * SPXファイルのヘッダー情報の設定を行う.
 * 引数ヘッダー情報(header)からコピーを作成して、m_headerにヘッダー情報を設定する。
 * @param[in]      header     ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
bool sphSpxFile::setFileInfo(const sphFileHeader* header)
{

  if (header == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "header is null.");
    return false;
  }

  const sphSpxHeader* src_header = dynamic_cast<const sphSpxHeader*>(header);
  if (src_header == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "cast error from sphFileHeader* to sphSpxHeader*.");
    return false;
  }

  if (m_header == NULL) {
    m_header = new sphSpxHeader();
  }
  // ヘッダー情報のみをコピーする。
  *m_header = *src_header;

  return true;
}


/**
 * \~english
 *
 *
 * \~japanese
 * 単一SPXファイルからSPXファイルのヘッダー情報のみの取得を行う.
 * 読込ヘッダー情報は、メンバーのヘッダー情報変数(m_header)にはセットされない。
 * @param[in]       filename    読込ファイル名
 * @param[out]      header      ファイルヘッダー情報返値
 * @return    success(=true)/fault(=false)
 */
bool sphSpxFile::getFileInfo(const char* filename, sphFileHeader* header)
{
  int flg = 0, tmp = 0;
  sphSpxHeader* spxHeader = NULL;

  if (filename == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "filename is null.");
    flg = 1;
  }

  if (header == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "header is null.");
    flg = 1;
  }
  else {
    spxHeader = dynamic_cast<sphSpxHeader*>(header);
    if (spxHeader == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "cast error from sphFileHeader* to sphSpxHeader*.");
      flg = 1;
    }
  }

  // splSpx構造体オブジェクトの生成
  splSpx* obj = new splSpx;
  if (!splInitSpx(obj)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault splInitSpx.");
    flg = 1;
  }

  // error check
  if( flg != 0 ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_ALLREDUCE_SUM, "pre getFileInfo error");
    splFinalizeSpx(obj, true);
    delete obj;
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }

  // SPXファイル情報を取得する。
  if (!splGetSpxInfo(filename,
          obj,
          m_readMode,
          m_rankId,
          m_procGrp)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault splGetSpxInfo, filename is %s.", filename);
    splFinalizeSpx(obj, false);
    delete obj;
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }

  // ヘッダー情報(splSpx構造体）からヘッダー情報(sphSpxHeader) を設定する。
  spxHeader->setSplSpxHeader(obj);

  // splSpx構造体オブジェクトの破棄（読込データは破棄しない）
  splFinalizeSpx(obj, false);
  delete obj;

  return true;
}


/**
 * \~english
 *
 * \~japanese
 *  単一SPXファイルに書込みを行う。
 *  書込みパラメータ、データは、他のメソッド(setter)にて設定する。
 * @param      filename    書込ファイル名
 * @param      procGrp     プロセスグループ
 * @return    success(=true)/fault(=false)
 */
bool sphSpxFile::writeFile()
{
  int flg = 0, tmp = 0;

  // 書き込み前チェック、準備を行う。
  void* pData = preWriteFile();
  if (pData == NULL) {
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault preWriteFile.");
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }

  // 書込ファイルの取得
  char filename[MAXPATHLEN];
  getFileName(0, filename);

  // splSpx構造体オブジェクトの生成
  splSpx* obj = new splSpx;
  if (!splInitSpx(obj)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault splInitSpx.");
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }

  // データポインタをセットする。
  obj->data.c = (unsigned char*)pData;

  // ヘッダー情報をsplSpx構造体に設定する。
  m_header->getSplSpxHeader(obj);

  // SPXファイルへの書込を行う。
  if (!splWriteSpx(filename,
          m_writeMode,
          (SplBool)m_compression,
          (SplBool)m_extRecode,
          m_header->m_block_size,
          obj,
          m_rankId,
          m_procGrp)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_WRITE_FILE, "fault splWriteSpx.");
    // print error
    if (splIsParallel()) m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
  }

  return true;
}


/**
 * \~english
 *
 * \~japanese
 * 圧縮書込みを行うかのフラグを設定する。
 * @param     flag    圧縮書込フラグ（true:圧縮書込みを行う。, false:圧縮書込みを行わない。）
 * @return    なし
 */
void sphSpxFile::setCompressionFlag(bool flag)
{
  m_compression = flag;
}


/**
 * \~english
 *
 * \~japanese
 * 圧縮書込みを行うかのフラグを取得する。
 * @param     void
 * @return    圧縮書込フラグ（true:圧縮書込みを行う。, false:圧縮書込みを行わない。）
 */
bool sphSpxFile::isCompressionFlag()
{
  return m_compression;
}

/**
 * \~english
 *
 * \~japanese
 * 拡張レコードを書込みかのフラグを設定する。
 * @param     flag    拡張レコードフラグ（true:拡張レコードを書き込む。, false:拡張レコードを書き込まない。）
 * @return    なし
 */
void sphSpxFile::setExtRecodeFlag(bool flag)
{
  m_extRecode = flag;
}


/**
 * \~english
 *
 * \~japanese
 * sphDataクラスを生成する.
 * - データクラスタイプ : SPH_DC_ARRAY_1D, SPH_DC_ARRAY_2D, SPH_DC_ARRAY_1D
 * - データ型タイプ : SPL_UNSIGNED_CHAR, SPL_UNSIGNED_SHORT, SPL_UNSIGNED
 * @param     data    sphDataクラス
 * @return    success(=true)/fault(=false)
 */
sphData* sphSpxFile::createSphArray(splSpx* spxObj)
{

  if (spxObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "splSpx object is null.");
    return NULL;
  }

  // 次元数チェック
  if (spxObj->dims == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_DIMS, "dims is zero.");
    return NULL;
  }
  // データサイズ取得
  register int i;
  size_t size[3] = {0, 0, 0};
  for (i=0; i<spxObj->dims; i++) {
    // サイズチェック
    if (spxObj->size[i] <= spxObj->gc*2) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ARRAY_SIZE, "size have to be more than gc*2, size[%d]:%ld <= gc[%d]:%ld * 2", i, spxObj->size[i], i, spxObj->gc);
      return NULL;
    }
    // ヘッダー情報のサイズはGCを含んだサイズであるので、データクラスのサイズのGCを含まないサイズにする。
    size[i] = spxObj->size[i] - spxObj->gc*2;
  }

  // データクラス作成
  sphData* voxel = NULL;

  // データクラスタイプ
  SphDCType dctype = SPH_DC_INVALID;
  if (spxObj->dims == 1) {
    if (spxObj->vlen == 1) dctype = SPH_DC_ARRAY_1D;
    else  dctype = SPH_DC_ARRAY_1DNRV;
  }
  else if (spxObj->dims == 2) {
    if (spxObj->vlen == 1) dctype = SPH_DC_ARRAY_2D;
    else  dctype = SPH_DC_ARRAY_2DNRV;
  }
  else if (spxObj->dims == 3) {
    if (spxObj->vlen == 1) dctype = SPH_DC_ARRAY_3D;
    else  dctype = SPH_DC_ARRAY_3DNRV;
  }

  // 実数サイズ
  SPL_Datatype datatype = SPL_DATATYPE_NULL;
  if (spxObj->rlen == 4) {
    datatype = SPL_FLOAT;
    sphVoxArray<float>* array =
                new sphVoxArray<float>(
                        dctype,
                        datatype,
                        size,
                        spxObj->gc,
                        spxObj->vlen);
    if (array != NULL) spxObj->data.f = array->getData();
    voxel = array;
  }
  else if (spxObj->rlen == 8) {
    datatype = SPL_DOUBLE;
    sphVoxArray<double>* array =
                new sphVoxArray<double>(
                        dctype,
                        datatype,
                        size,
                        spxObj->gc,
                        spxObj->vlen);
    if (array != NULL) spxObj->data.d = array->getData();
    voxel = array;
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_DTYPE, "invalid rlen=%d.", spxObj->rlen);
    return NULL;
  }

  if (voxel == NULL || spxObj->data.c == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_DATA_IS_NULL, "can not create sphData.");
    return NULL;
  }

  // ファイルクラスにデータクラスを設定する。
  m_sphArray = voxel;

  return voxel;
}


/**
 * \~english
 * get sphData pointer
 * @param     void
 * @return    sphData object pointer
 *
 * \~japanese
 * sphDataデータクラスを取得する.
 * @param     void
 * @return    sphDataクラス
 */
sphData* sphSpxFile::getData()
{
  return m_sphArray;
}


/**
 * \~english
 * set sphData pointer
 * @param     data        sphData object pointer
 * @return    void
 *
 * \~japanese
 * sphDataデータクラスを設定する.
 * @param     data        sphDataクラス
 * @return    void
 */
void sphSpxFile::setData(sphData* data)
{
  m_sphArray = data;
}



/**
 * \~english
 * get read mode
 * @param     void
 * @return    read mode
 *
 * \~japanese
 * 読み込みモードを取得する。
 * @param     void
 * @return    読込モード
 */
int sphSpxFile::getReadMode() const
{
  return m_readMode;
}

/**
 * \~english
 * get write mode
 * @param     void
 * @return    write mode
 *
 * \~japanese
 * 書き込みモードを取得する。
 * @param     void
 * @return    書込モード
 */
int sphSpxFile::getWriteMode() const
{
  return m_writeMode;
}




