/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSvFile.C
 * Sphere Simple Voxel File Input/Output Class
 * @author     hira
 * @date       2009/09/07
 *
 * \~japanese
 * @file sphSvFile.C
 * Sphere Simple Voxel ファイル入出力クラス
 * @author     hira
 * @date       2009/09/07
 */

#include "fileio/sphSvFile.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphSvFile::sphSvFile() : m_fileList(NULL)
{
  // 初期化を行う。
  initialize();
}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ
 */
sphSvFile::~sphSvFile()
{
  // 破棄処理
  finalize();
}


/**
 * \~english
 *
 * \~japanese
 *  初期化を行う.
 *  メンバーの初期化（デフォルト値の代入）を行う.
 * @param      なし
 * @return    success(=true)/fault(=false)
 */
bool sphSvFile::initialize()
{
  // ファイルリスト構造体
  if (m_fileList != NULL) {
    splFinalizeSvFileList(m_fileList);
    delete m_fileList;
    m_fileList = NULL;
  }
  /// splSvFileList構造体生成
  m_fileList = new splSvFileList;

  // splSvFileList構造体を初期化する。
  if (!splInitSvFileList(m_fileList)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "splInitSvFileList fault");
    return false;
  }

  return true;
}


/**
 * \~english
 *
 * \~japanese
 *  破棄処理を行う.
 *  splSbx, splSvFileList構造体の破棄処理を行う。
 * @param     mem_free_flag    splSbx->data(splData)の開放フラグ
 *                  true:開放を行う。, false:開放を行わない。
 * @return    success(=true)/fault(=false)
 */
bool sphSvFile::finalize()
{
  // ファイルリストの破棄処理
  if (m_fileList != NULL) {
    splFinalizeSvFileList(m_fileList);
    delete m_fileList;
    m_fileList = NULL;
  }

  return true;
}


/**
 * \~english
 * set mask on Upper bits, get lower bits data
 * @param      void
 * @return    sphData pointer
 *
 * \~japanese
 * 上位ビットデータをマスクし(0とする）、下位ビットのみのデータとする.
 * @param      void
 * @return    sphData pointer
 */
sphData* sphSvFile::maskUpperBits()
{
  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "not support maskUpperBits");
  return NULL;
}

/**
 * \~english
 * shift lower bits, get upper bits data
 * @param      void
 * @return    sphData pointer
 *
 * \~japanese
 *  下位２ビットをシフトし、上位のみのデータとする。.
 * @param      void
 * @return    sphData pointer
 */
sphData* sphSvFile::shiftLowBits()
{
  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "not support shiftLowBits");
  return NULL;
}


/**
 * \~english
 * create sphFileHeader infomation
 * @param      header           file header
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ファイルヘッダー情報を作成する.
 * @param      header           ファイルヘッダー情報
 * @return    success(=true)/fault(=false)
 */
bool sphSvFile::createFileHeader(sphSvHeader* header)
{
  register int i, j;
  int flg = 0, tmp = 0;

  if (header == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "header is nullt.");
    flg = 1;
  }

  if (getReadMode() != 0) {
    if (splBcastSvFileList(m_fileList, m_rankId, m_procGrp) != SplTrue) {
      // spl error
      m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
      // sph error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "fault bcastSvFileList.");
      flg = 1;
    }
  }

  size_t list_cnt = getFileListCount();
  if (list_cnt <= 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not set filelist.");
    flg = 1;
  }

  // error check
  tmp = 0;
  SPL_Allreduce(&flg, &tmp, 1, SPL_INT, SPL_SUM, m_procGrp);
  if( tmp != 0 ) {
    // error
    m_errHandler->SPH_ERROR(SPH_ERR_ALLREDUCE_SUM);
    return false;
  }

  // 読込情報が設定されていないので、読込ファイルから作成する。
  size_t obj_size[3] = {1, 1, 1};
  size_t obj_sidx[3] = {(size_t)-1, (size_t)-1, (size_t)-1};
  size_t obj_eidx[3] = {0, 0, 0};
  size_t gc;
  bool set_flag = false;
  for (i=0; i<list_cnt; i++) {
    splSvFile* spl_file = getSplFile(i);
    if (spl_file == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not get splSvFile object.");
      flg = 1;
    }
    else {
      // 読込ヘッダー情報が設定されていないので、ファイルヘッダーから取得する。
      if (!getFileInfo(spl_file->fname, header)) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_READ_FILE, "fault getFileInfo, filename is %s.", spl_file->fname);
        flg = 1;
      }
    }
    if (flg == 1) continue;

    unsigned int dims  = header->getDims();
    const size_t* size = header->getSize();
    const size_t* wsize = header->getWholeSize();
    const size_t* start_idx = header->getStartIdx();
    gc = header->getGc();
    size_t wlen = 1;
    if (wsize != NULL) {
      for (j=0; j<dims; j++) wlen *= wsize[j];
    }
    else {
      wlen = 0;
    }

    for (j=0; j<dims; j++) {
      size_t sidx = 0, eidx = 0;
      if (spl_file->set_flag == 1) sidx = spl_file->start_idx[j];
      else if (wlen > 0 && start_idx != NULL) sidx = start_idx[j];
      else {
        // set_flag == 0 and wsize == 0
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not set start_idx, file_name is %s.", spl_file->fname);
        flg = 1;
      }
      if (size[j] <= (2 * gc)) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "invalid size or gc in file. [filename=%s, size[%d]=%ld, gc=%ld", spl_file->fname, j, size[j], gc);
        flg = 1;
      }

      // 始点、終点インデックスの最小、最大値を取得する。
      eidx = sidx + size[j] - 2 * gc;
      obj_sidx[j] = (obj_sidx[j] < sidx) ? obj_sidx[j] : sidx;
      obj_eidx[j] = (obj_eidx[j] > eidx) ? obj_eidx[j] : eidx;

      if (flg == 1)  break;
    }
    if (flg == 1)  break;
  }

  if (flg != 1) {
    for (i=0;i<header->getDims();i++) {
      if (obj_sidx[i] > obj_eidx[i]) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "invalid start and end index in file. [start_idx[%d]=%ld, end_idx[%d]=%ld", i, obj_sidx[i], i, obj_eidx[i]);
        flg = 1;
      }
      obj_size[i] = obj_eidx[i] - obj_sidx[i] + 2 * gc;
    }
  }
  else {
    return false;
  }

  // サイズを設定する。
  header->setSize(obj_size);

  // 始点インデックスを設定する。
  for (i=header->getDims();i<3;i++) obj_sidx[i] = 0;
  header->setStartIdx(obj_sidx);

  return true;
}


/**
 * \~english
 *
 * \~japanese
 * 読み込みファイル名を取得する。
 * 読み込みファイルリストからid番目を取得する。
 * @param[in]      id    ファイルリストインデックス（０〜"ファイルリストサイズ-1"）
 * @param[out]      filename    ファイル名
 * @param[out]      sidx_i      I方向始点グローバルインデックス
 * @param[out]      sidx_j      J方向始点グローバルインデックス
 * @param[out]      sidx_k      K方向始点グローバルインデックス
 * @return    success(=true)/fault(=false)
 */
bool sphSvFile::getFileName(
            int id,
            char* filename,
            size_t& sidx_i, size_t& sidx_j, size_t& sidx_k)
{
  if (filename == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "filename is null.");
    return false;
  }

  splSvFile* fobj = getSplFile(id);
  if (fobj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "splSvFile object is null.");
    return false;
  }

  // 読込ファイル情報のコピー
  strcpy(filename, fobj->fname);
  sidx_i = fobj->start_idx[0];
  sidx_j = fobj->start_idx[1];
  sidx_k = fobj->start_idx[1];

  return true;
}


/**
 * \~english
 *
 * \~japanese
 * 読み込みファイル名を取得する。
 * 読み込みファイルリストからid番目を取得する。
 * @param[in]      id    ファイルリストインデックス（０〜"ファイルリストサイズ-1"）
 * @param[out]      filename    ファイル名
 * @return    success(=true)/fault(=false)
 */
bool sphSvFile::getFileName(
            int id,
            char* filename)
{
  size_t sidx_x, sidx_y, sidx_z;
  return getFileName(id, filename, sidx_x, sidx_y, sidx_z);
}


/**
 * \~english
 *
 * \~japanese
 * 読み込みファイル情報を取得する。
 * 読み込みファイルリストからid番目を取得する。
 * @param[in]      id    ファイルリストインデックス（０〜"ファイルリストサイズ-1"）
 * @return    読み込みファイル情報
 */
/// 読み込みファイル情報を取得する。
splSvFile* sphSvFile::getSplFile(int id)
{
  splSvFile* fobj = m_fileList->top;
  if (fobj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "File List is null.");
    return false;
  }

  if (m_fileList->list_size <= id) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "invalid id. [list_size=%d, get_id=%d]\n", m_fileList->list_size, id);
    return false;
  }

  register size_t i;
  for (i=0; i<m_fileList->list_size-1-id; i++) {
    fobj = fobj->next;
    if (fobj == NULL) break;
  }
  if (fobj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "file name is null in file list. [list_size=%d, get id=%d]\n", m_fileList->list_size, id);
    return false;
  }
  return fobj;
}


/**
 * \~english
 *
 * \~japanese
 * 読み込みファイルリストのファイル数を取得する。
 * @param      なし
 * @return    ファイルリストのファイル数
 */
size_t sphSvFile::getFileListCount()
{
  return m_fileList->list_size;
}



/**
 * 読み込みファイルリストにファイル、始点インデックスを追加する。
 * @param      filename      追加ファイル名
 * @param      sidx        読込始点インデックス
 * @return    success(=true)/fault(=false)
 */
bool sphSvFile::addSvFileInfo(const char* filename, const size_t sidx[3])
{
  if (filename == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "filename is null");
    return false;
  }

  // ファイルリストにファイルを追加する。
  if (!splAddSvFile(m_fileList, filename, sidx)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault splAddSvFile.");
    return false;
  }

  return true;
}


/**
 * \~english
 *
 * \~japanese
 * 読み込みファイルリストにファイルを追加する。
 * @param      filename      追加ファイル名
 * @return    success(=true)/fault(=false)
 */
bool sphSvFile::addSvFileName(const char* filename)
{
  if (filename == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "filename is null");
    return false;
  }

  // ファイルリストにファイルを追加する。
  if (!splAddSvFileName(m_fileList, filename)) {
    // spl error
    m_errHandler->error_info(spl_get_errcode(), spl_get_errinfo());
    // sph error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "fault splAddSvFileName.");
    return false;
  }

  return true;
}


/**
 * \~english
 *
 * \~japanese
 * 読み込みファイルリストをクリアする。
 * @param     なし
 * @return    なし
 */
void sphSvFile::clearSvFileList()
{
  // ファイルリストの破棄処理
  if (m_fileList != NULL) {
    splFinalizeSvFileList(m_fileList);
  }
}

/**
 * \~english
 *
 * \~japanese
 * 読み込みファイルリストを設定する。
 * @param     filelist          ファイルリスト
 * @return    なし
 */
void sphSvFile::setSvFileList(splSvFileList* filelist)
{
  // 読み込みファイルリストを設定する。
  m_fileList = filelist;
}


/**
 * \~english
 *
 * \~japanese
 * 書込前準備を行う。
 * @param     header          ファイルヘッダー
 * @return    書込データポインタ
 */
void* sphSvFile::preWriteFile()
{
  // 書込ヘッダー情報が設定されていること。
  sphSvHeader* header = dynamic_cast<sphSvHeader*>(getFileInfo());
  if (header == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_header is null.");
    return NULL;
  }

  // 書込情報のチェック
  if (!header->checkHeader()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "sphSvHeader->checkHeader error.");
    return NULL;
  }

  // 書込ファイルが設定されていること。
  if (getFileListCount() <= 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "write file is empty.");
    return NULL;
  }

  // 圧縮フラグが設定されている場合、ブロックサイズを自動設定する。
  if (this->isCompressionFlag() && header->getBlockSize() <= 0) {
    header->setBlockSizeBySize();
  }

  sphData* data = this->getData();
  void* pData = NULL;
  if (data == NULL) {
    // 書込データが設定されいない。
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_sphArray is null.");
    return NULL;
  }
  // データクラスからデータを取得する。
  unsigned int dtype = header->getDtype();
  unsigned int rlen = header->getRlen();
  const size_t* dc_size;
  size_t gc;
  size_t vlen;
  unsigned int dims;
  if (header->getFormat() == "sbx") {
    if (dtype == 1) {
      sphArray<unsigned char>* array = dynamic_cast<sphArray<unsigned char>*>(data);
      if (array == NULL) {
        // error
        m_errHandler->SPH_ERROR(SPH_ERR_CAST_NULL);
        return NULL;
      }
      pData = array->getData();
      dc_size = array->getArraySize();
      gc = array->getGc();
      dims = array->getDims();
      vlen = array->getNumOfDtElem();
    }
    else if (dtype == 2) {
      sphArray<unsigned short>* array = dynamic_cast<sphArray<unsigned short>*>(data);
      if (array == NULL) {
        // error
        m_errHandler->SPH_ERROR(SPH_ERR_CAST_NULL);
        return NULL;
      }
      pData = array->getData();
      dc_size = array->getArraySize();
      gc = array->getGc();
      dims = array->getDims();
      vlen = array->getNumOfDtElem();
    }
    else if (dtype == 4) {
      sphArray<unsigned int>* array = dynamic_cast<sphArray<unsigned int>*>(data);
      if (array == NULL) {
        // error
        m_errHandler->SPH_ERROR(SPH_ERR_CAST_NULL);
        return NULL;
      }
      pData = array->getData();
      dc_size = array->getArraySize();
      gc = array->getGc();
      dims = array->getDims();
      vlen = array->getNumOfDtElem();
    }
    else {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_DTYPE, "dtype have to be 1, 2 or 4.");
      return NULL;
    }
  }
  else if ( (header->getFormat() == "spx")
         || (header->getFormat() == "sph") ) {
    // データクラスからデータを取得する。
    if (rlen == 4) {
      sphArray<float>* array = dynamic_cast<sphArray<float>*>(data);
      if (array == NULL) {
        // error
        m_errHandler->SPH_ERROR(SPH_ERR_CAST_NULL);
        return NULL;
      }
      pData = array->getData();
      dc_size = array->getArraySize();
      gc = array->getGc();
      dims = array->getDims();
      vlen = array->getNumOfDtElem();
    }
    else if (rlen == 8) {
      sphArray<double>* array = dynamic_cast<sphArray<double>*>(data);
      if (array == NULL) {
        // error
        m_errHandler->SPH_ERROR(SPH_ERR_CAST_NULL);
        return NULL;
      }
      pData = array->getData();
      dc_size = array->getArraySize();
      gc = array->getGc();
      dims = array->getDims();
      vlen = array->getNumOfDtElem();
    }
    else {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_RLEN, "rlen have to be 4 or 8.");
      return NULL;
    }
  }

  // check data and write header
  if (dims != header->getDims()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_DIMS, "dims is difference in data.[data=%d, header=%d]", dims,  header->getDims());
    return NULL;
  }

  if (vlen != header->getVlen()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VLEN, "vlen is difference in data.[data=%ld, header=%ld]", vlen,  header->getVlen());
    return NULL;
  }

  register int i;
  int mode = getWriteMode();
  const size_t* wr_size = header->getSize();
  for (i=0; i<dims; i++) {
    if (dc_size[i]+gc*2 != wr_size[i]) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ARRAY_SIZE, "size is difference in data.[size[%d],gc=%ld,%ld, header[%d]=%ld", i, dc_size[i],gc, i, wr_size[i]);
      return NULL;
    }
  }
/****
  // set block size for gather
  if (mode == 1) {
    const size_t* whole_size = header->getWholeSize();
    if (header->getBlockSize() <= 0) {
      size_t gathersize = whole_size[0];
      if (dims == 1) gathersize = whole_size[0];
      else if (dims == 2) gathersize = whole_size[0];
      else if (dims == 3) gathersize = whole_size[0]*whole_size[1];
      if (gathersize > 0) {
        header->setBlockSize(gathersize);
      }
      else {
        header->setBlockSizeBySize();
      }
    }
  }
****/

  return pData;
}


