/*
 * sphTextFile.C
 *
 *  Created on: 2010/01/04
 *      Author: hira
 */

#include "fileio/sphTextFile.h"
#include "config/sphElemUtils.h"
#include "spl/util_endian.h"


/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphTextFile::sphTextFile()
{
  m_configElem = NULL;
  m_fp = NULL;
  m_steptime = NULL;
  /// デフォルト出力モード=1 : ノード0のみが出力を行う。
  m_writeMode = 1;
  m_interval = 0;
  m_filename.erase();
}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphTextFile::~sphTextFile()
{
  if (m_fp != NULL) fclose(m_fp);
  m_fp = NULL;
}


/**
 * コンフィグレーションファイル要素を設定する。
 * @param       config_elem       コンフィグレーションファイル要素
 * @return      void
 */
bool sphTextFile::setConfigElem(const sphElem* config_elem)
{
  if (config_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "setConfigElem arguments is null.");
    return false;
  }
  m_configElem = config_elem;

  m_errHandler->blockWarning();          // 警告メッセージを無効にする
  /// ファイル名を取得する。
  if (m_configElem->existAttribute(SPH_CFG_ATTR_FILENAME)) {
    if (!m_configElem->existAttribute(SPH_CFG_ATTR_PREFIX)) {
      if (m_configElem->getElem(SPH_CFG_NAME_FILEFORMAT) == NULL) {
        // 固定ファイル名であるので、取得しておく。
        char file_name[MAXPATHLEN];
        if (!sphElemUtils(m_configElem).getFileFormatPath(file_name, m_outputPath.c_str(), 0, 0)) {
          // error
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "sphElemUtils::getFileFormatPath fault.");
          return false;
        }
        m_filename = file_name;
      }
    }
  }

  /// インターバルステップ
  const sphElem* opt_elem = m_configElem->getElem(SPH_CFG_NAME_FILEOPTION);
  if (opt_elem != NULL) {
    m_interval = m_configElem->getAttrLLong(SPH_CFG_ATTR_INTERVAL);
  }

  m_rankId = m_configElem->getAttrInt(SPH_CFG_ATTR_IO_NODEID);

  // ファイル入出力モード
  SphCfgFileIOMode cfg_mode = sphElemUtils(m_configElem).getFileIoMode();

  // モード番号の設定
  m_writeMode = 1;
  if (cfg_mode == CFG_FILE_DEFAULT || cfg_mode == CFG_FILE_WHOLE)  m_writeMode = 1;
  else if (cfg_mode == CFG_FILE_MULTI)  m_writeMode = 0;
  else if (cfg_mode == CFG_FILE_PROCESS)  m_writeMode = 0;
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "invalid mode.");
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  return true;
}


/**
 * 現在ステップ・時刻クラスを設定する。
 * @param       steptime       現在ステップ・時刻
 * @return      void
 */
bool sphTextFile::setStepTime(const sphStepTime* steptime)
{
  if (steptime == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "setStepTime arguments is null.");
    return false;
  }
  m_steptime = steptime;
  return true;
}



/**
 * ファイル名を設定する。
 * @param[in]       file_name       ファイル名
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::setFileName(const char* file_name)
{
  m_filename = file_name;
  return true;
}


/**
 * ファイル名を取得する。
 * @param[out]       file_name       ファイル名
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::getFileName(char* file_name) const
{
  if (file_name == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "getFileName arguments is null.");
    return false;
  }

  file_name[0] = 0x00;
  if (!m_filename.empty()) {
    strcpy(file_name, m_filename.c_str());
    return true;
  }

  if (m_configElem == NULL) {
    return false;
  }

  /// コンフィグレーションファイル要素から取得する。
  int myID = 0;
  if (m_procGrp != SPL_COMM_NULL) SPL_Comm_rank(m_procGrp, &myID);

  size_t step = (m_steptime != NULL) ? m_steptime->getTotalStep() : 0;
  sphElemUtils elem_utils(m_configElem);
  if (!elem_utils.getFileFormatPath(file_name, m_outputPath.c_str(), step, myID)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "sphElemUtils::getFileFormatPath fault.");
    return false;
  }

  return true;
}


/**
 * インターバルステップを設定する。
 * @param[out]       interval       インターバルステップ
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::setInterval(size_t interval)
{
  m_interval = interval;
  return true;
}


/**
 * インターバルステップを取得する。
 * @param[out]       void
 * @return    インターバルステップ
 */
size_t sphTextFile::getInterval() const
{
  return m_interval;
}




/**
 * 出力ディレクトリを設定する。
 * @param[in]       dir       出力ディレクトリ
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::setBaseDir(const char* dir)
{
  m_outputPath = dir;
  return true;
}


/**
 * 出力ディレクトリを取得する。
 * @param[in]       dir       出力ディレクトリ
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::getBaseDir(char* dir) const
{
  if (dir == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "getBaseDir arguments is null.");
    return false;
  }
  strcpy(dir, m_outputPath.c_str());
  return true;
}


/**
 *  書込プロセスグループを設定する
 * @param      procGrp                  書込プロセスグループ
 * @return    void
 */
void sphTextFile::setProcGrp(SPL_Comm procGrp)
{
  m_procGrp = procGrp;
}

/**
 *  書込ランク番号を設定する
 * @param      rankId                   書込ランク番号
 * @return    void
 */
void sphTextFile::setRankId(int rankId)
{
  m_rankId = rankId;
}


/**
 *  書込プロセスグループを取得する
 * @param      void
 * @return    書込プロセスグループ
 */
SPL_Comm sphTextFile::getProcGrp() const
{
  return m_procGrp;
}


/**
 *  書込ランク番号を取得する
 * @param      void
 * @return    書込ランク番号
 */
int sphTextFile::getRankId() const
{
  return m_rankId;
}


/**
 * ファイルをオープンする。
 * @param      init     初期化フラグ(true:ファイルを上書きモードでオープンする。/ false:ファイルを追加モードでオープンする。)
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::openFile(bool init)
{
  // 出力ランク番号であるか
  if (m_writeMode == 1) {
    int myID = 0;
//    (m_procGrp != SPL_COMM_NULL) ? SPL_Comm_rank(m_procGrp, &myID) : 0;
    if (m_procGrp != SPL_COMM_NULL) {
      SPL_Comm_rank(m_procGrp, &myID);
    }
    if (myID != m_rankId) {
      // 出力ランク番号ではない。
      return true;
    }
  }

  char file_name[MAXPATHLEN];

  if (!getFileName(file_name)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not get file name.");
    return false;
  }

  if (init) {
    m_fp = fopen(file_name, "w");
  }
  else {
    m_fp = fopen(file_name, "a");
  }
  if (m_fp == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_WRITE_FILE, "can not open file, [filename=%s].", file_name);
    return false;
  }

  return true;
}


/**
 * ファイルに文字列を出力する。(最大出力サイズ=4096バイト)
 * @param      format           出力書式
 * @param      ...              出力パラメータ
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::print(const char* format, ...) const
{
  if (m_fp == NULL) return true;

  // 出力ステップであるか？
  if (m_steptime != NULL) {
    if (!m_steptime->chkInterval(m_interval)) return true;
  }

  // メッセージを取得する。
  va_list ap;
  va_start(ap, format);
  char buf[4096];
  vsprintf(buf, format, ap);                  // vasprintf to vsprintf at 2010/05/10 by @hira

  // ファイル出力する。
  fprintf(m_fp, buf);

  va_end(ap);

  return true;
}


/**
 * ファイルに文字列を出力する。(std::string)
 * @param      msg           出力文字列
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::print(std::string msg) const
{
  if (m_fp == NULL) return true;

  // 出力ステップであるか？
  if (m_steptime != NULL) {
    if (!m_steptime->chkInterval(m_interval)) return true;
  }

  // ファイル出力する。
  fprintf(m_fp, msg.c_str());

  return true;
}



/**
 * ファイルに文字列を出力し、コンソールにも出力する。(最大出力サイズ=4096バイト)
 * @param      format           出力書式
 * @param      ...              出力パラメータ
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::printWithConsole(const char* format, ...) const
{
  if (m_fp == NULL) return true;

  // 出力ステップであるか？
  if (m_steptime != NULL) {
    if (!m_steptime->chkInterval(m_interval)) return true;
  }

  // メッセージを取得する。
  va_list ap;
  va_start(ap, format);
  char buf[4096];
  vsprintf(buf, format, ap);                  // vasprintf to vsprintf at 2010/05/10 by @hira

  // ファイル出力する。
  fprintf(m_fp, buf);

  // コンソール出力を行う。
  printf(buf);

  va_end(ap);

  return true;
}


/**
 * ファイルに文字列を出力し、コンソールにも出力する。(std::string)
 * @param      msg           出力文字列
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::printWithConsole(std::string msg) const
{
  if (m_fp == NULL) return true;

  // 出力ステップであるか？
  if (m_steptime != NULL) {
    if (!m_steptime->chkInterval(m_interval)) return true;
  }

  // ファイル出力する。
  fprintf(m_fp, msg.c_str());

  // コンソール出力を行う。
  printf(msg.c_str());

  return true;
}


/**
 * ファイルに文字列を出力する。(最大出力サイズ=4096バイト)
 * @param      format           出力書式
 * @param      ...              出力パラメータ
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::printLine(const char* format, ...) const
{
  if (m_fp == NULL) return true;

  // 出力ステップであるか？
  if (m_steptime != NULL) {
    if (!m_steptime->chkInterval(m_interval)) return true;
  }

  // メッセージを取得する。
  va_list ap;
  va_start(ap, format);
  char buf[4096];
  vsprintf(buf, format, ap);                  // vasprintf to vsprintf at 2010/05/10 by @hira

  // ファイル出力する。
  fprintf(m_fp, buf);
  fprintf(m_fp, "\n");

  va_end(ap);

  return true;
}


/**
 * ファイルに文字列を出力し、コンソールにも出力する。(最大出力サイズ=4096バイト)
 * @param      format           出力書式
 * @param      ...              出力パラメータ
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::printLineWithConsole(const char* format, ...) const
{
  if (m_fp == NULL) return true;

  // 出力ステップであるか？
  if (m_steptime != NULL) {
    if (!m_steptime->chkInterval(m_interval)) return true;
  }

  // メッセージを取得する。
  va_list ap;
  va_start(ap, format);
  char buf[4096];
  vsprintf(buf, format, ap);                  // vasprintf to vsprintf at 2010/05/10 by @hira

  // ファイル出力する。
  fprintf(m_fp, buf);
  fprintf(m_fp, "\n");

  // コンソール出力を行う。
  printf(buf);
  printf("\n");

  va_end(ap);

  return true;
}



/**
 * データを16進数文字列として出力する。
 * @param      data           出力データ
 * @param      dtype           データ型
 * @param      size           出力サイズ
 * @param      data_num        1格子のデータ数
 * @param      rev_flag        true:SPH_DC_ARRAY_3DNRV, false:SPH_DC_ARRAY_3DN
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::printDumpHex(
                      const void* data, SPL_Datatype dtype, const size_t size[3],
                      size_t data_num, bool rev_flag) const
{
  if (m_fp == NULL) return true;

  // 出力ステップであるか？
  if (m_steptime != NULL) {
    if (!m_steptime->chkInterval(m_interval)) return true;
  }

  register size_t i, j, k, l, n;
  int dataofsize = spl_get_datasize(dtype);

  size_t vlen = data_num;
  size_t dlen = 1;
  if (rev_flag) {
    vlen = 1;
    dlen = data_num;
  }

  for (l=0; l<vlen; l++) {
    if (vlen > 1) {
      fprintf(m_fp,"vlen=%ld\n", l);
    }

    for (k=0; k<size[2]; k++) {
      fprintf(m_fp,"k=%ld\n", k);
      if (dlen > 1) {
        fprintf(m_fp,"j/i   0 - %ld [vlen=%ld]\n", size[0]-1, dlen);
      }
      else {
        fprintf(m_fp,"j/i   0 - %ld\n", size[0]-1);
      }

      for (j=0; j<size[1]; j++) {
        fprintf(m_fp, "%05ld:", j);
        for (i=0; i<size[0]; i++) {
          for (n=0; n<dlen; n++) {

            char hex[64] = {0};
            size_t idx;
            if (!rev_flag) {
                idx = size[0]*size[1]*size[2]*l + size[0]*size[1]*k + size[0]*j + i;
            }
            else {
              idx = size[0]*size[1]*dlen*k + size[0]*dlen*j + dlen*i + n;
            }

            if (dataofsize == 1) {
              sprintf(hex, "%02X ", ((unsigned char*)data)[idx]);
            }
            else if (dataofsize == 2) {
              unsigned short value = ((unsigned short*)data)[idx];
              BSwap16(&value);
              sprintf(hex, "%04X ", value);
            }
            else if (dataofsize == 4) {
              unsigned int value = ((unsigned int*)data)[idx];
              BSwap32(&value);
              sprintf(hex, "%08X ", value);
            }
            else if (dataofsize == 8) {
              unsigned long long value = ((unsigned long long*)data)[idx];
              BSwap64(&value);
              unsigned int hex_value;
              hex_value = (unsigned int)(value >> 32);
              char hexU[64] = "";
              sprintf(hexU, "%08X", hex_value);
              hex_value = (unsigned int)(value & 0x00000000FFFFFFFF);
              sprintf(hex, "%s%08X ", hexU, hex_value);
            }
            fprintf(m_fp, hex);
          }
        }
        fprintf(m_fp,"\n");
      }
      fprintf(m_fp,"\n");
    }
    fprintf(m_fp,"\n");
  }
  fprintf(m_fp,"\n");

  return true;
}


/**
 * データを数値文字列として出力する。
 * @param      data           出力データ
 * @param      dtype           データ型
 * @param      size           出力サイズ
 * @param      data_num        1格子のデータ数
 * @param      rev_flag        true:SPH_DC_ARRAY_3DNRV, false:SPH_DC_ARRAY_3DN
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::printDump(
                    const void* data, SPL_Datatype dtype, const size_t size[3],
                    size_t data_num, bool rev_flag) const
{

  if (m_fp == NULL) return true;

  // 出力ステップであるか？
  if (m_steptime != NULL) {
    if (!m_steptime->chkInterval(m_interval)) return true;
  }

  register size_t i, j, k, l, n;
  int dataofsize = spl_get_datasize(dtype);

  size_t vlen = data_num;
  size_t dlen = 1;
  if (rev_flag) {
    vlen = 1;
    dlen = data_num;
  }

  for (l=0; l<vlen; l++) {
    if (vlen > 1) {
      fprintf(m_fp,"vlen=%ld\n", l);
    }

    for (k=0; k<size[2]; k++) {
      fprintf(m_fp,"k=%ld\n", k);

      if (dlen > 1) {
        fprintf(m_fp,"j/i   0 - %ld [vlen=%ld]\n", size[0]-1, dlen);
      }
      else {
        fprintf(m_fp,"j/i   0 - %ld\n", size[0]-1);
      }
      for (j=0; j<size[1]; j++) {
        fprintf(m_fp, "%05ld:", j);
        for (i=0; i<size[0]; i++) {
          for (n=0; n<dlen; n++) {

            char hex[64] = {0};
            size_t idx;
            if (!rev_flag) {
                idx = size[0]*size[1]*size[2]*l + size[0]*size[1]*k + size[0]*j + i;
            }
            else {
              idx = size[0]*size[1]*dlen*k + size[0]*dlen*j + dlen*i + n;
            }

            if (dtype == SPL_CHAR) {
              char value = ((char*)data)[idx];
              sprintf(hex, "%3d ", value);
            }
            else if (dtype == SPL_UNSIGNED_CHAR) {
              unsigned char value = ((unsigned char*)data)[idx];
              sprintf(hex, "%3d ", value);
            }
            else if (dtype == SPL_SHORT) {
              short value = ((short*)data)[idx];
              sprintf(hex, "%5d ", value);
            }
            else if (dtype == SPL_UNSIGNED_SHORT) {
              unsigned short value = ((unsigned short*)data)[idx];
              sprintf(hex, "%5d ", value);
            }
            else if (dtype == SPL_INT) {
              int value = ((int*)data)[idx];
              sprintf(hex, "%8d ", value);
            }
            else if (dtype == SPL_UNSIGNED) {
              unsigned int value = ((unsigned int*)data)[idx];
              sprintf(hex, "%8d ", value);
            }
            else if (dtype == SPL_LONG) {
              long value = ((long*)data)[idx];
              sprintf(hex, "%8ld ", value);
            }
            else if (dtype == SPL_UNSIGNED_LONG) {
              unsigned long value = ((unsigned long*)data)[idx];
              sprintf(hex, "%8ld ", value);
            }
            else if (dtype == SPL_LONG_LONG_INT) {
              long long value = ((long long*)data)[idx];
              sprintf(hex, "%8ld ", value);
            }
            else if (dtype == SPL_FLOAT) {
              float value = ((float*)data)[idx];
              sprintf(hex, "%e ", value);
            }
            else if (dtype == SPL_DOUBLE) {
              double value = ((double*)data)[idx];
              sprintf(hex, "%e ", value);
            }
            fprintf(m_fp, hex);
          }
        }
        fprintf(m_fp,"\n");
      }
      fprintf(m_fp,"\n");
    }
    fprintf(m_fp,"\n");
  }
  fprintf(m_fp,"\n");


  return true;
}


/**
 * データを16進数文字列として出力する。
 * @param      data           出力データオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::printDumpHex(const sphData* data) const
{
  if (data == NULL) return false;

  const void* pdata = data->getDataPointer();
  SPL_Datatype dtype = data->getDataType();
  const size_t* size = data->_getArraySize();
  size_t data_num = data->getNumOfDtElem();
  SphDCType dctype = data->getDcType();

  bool rev_flag = false;
  switch (dctype) {
  case SPH_DC_ARRAY_1DNRV:
  case SPH_DC_ARRAY_2DNRV:
  case SPH_DC_ARRAY_3DNRV:
    rev_flag = true;
  }

  return printDumpHex(pdata, dtype, size, data_num, rev_flag);
}


/*
 * データを数値文字列として出力する。
 * @param      data           出力データオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::printDump(const sphData* data) const
{
  if (data == NULL) return false;

  const void* pdata = data->getDataPointer();
  SPL_Datatype dtype = data->getDataType();
  const size_t* size = data->_getArraySize();
  size_t data_num = data->getNumOfDtElem();
  SphDCType dctype = data->getDcType();

  bool rev_flag = false;
  switch (dctype) {
  case SPH_DC_ARRAY_1DNRV:
  case SPH_DC_ARRAY_2DNRV:
  case SPH_DC_ARRAY_3DNRV:
    rev_flag = true;
  }

  return printDump(pdata, dtype, size, data_num, rev_flag);
}



/**
 * ファイルをクローズする。
 * @param      void
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::close()
{
  if (m_fp != NULL) fclose(m_fp);
  m_fp = NULL;
  return true;
}


/**
 * バッファからファイルへ出力する。
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphTextFile::flush()
{
  if (m_fp != NULL) {
    if (fflush(m_fp) != 0) {
      return false;
    }
  }

  return true;
}




