/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * @file sphMultiParaController.C
 * 部分領域管理パラレルコントローラソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#include "parallel/sphMultiParaController.h"


/**
 * コンストラクタ
 * @param      paraKey        パラレルコントローラー識別キー
 */
sphMultiParaController::sphMultiParaController(unsigned int paraKey) : sphSvParaController(paraKey)
{
  m_mapChildPara.clear();
  m_processGroupId = NULL;
  m_nodeCount = 0;
}


/**
 * デストラクタ
 */
sphMultiParaController::~sphMultiParaController()
{
  // 子パラレルコントローラの削除
  std::map<int, sphSvParaController*>::iterator itr = m_mapChildPara.begin();
  for( ; itr != m_mapChildPara.end(); itr++){
    // パラレルコントローラ
    sphSvParaController* paraCtrl = itr->second;
    if (paraCtrl != NULL) delete paraCtrl;
    paraCtrl = NULL;
  }
  m_mapChildPara.clear();
  if (!m_processGroupId) delete []m_processGroupId;
}

/**
 * 計算領域サイズ、始点インデックス、分割情報から部分領域の初期化を行う。
 *
 * EV分割用のパラレルコントローラを初期化する。
 * @param[in]       wsize   計算領域サイズ
 * @param[in]       wsta    グローバル始点インデックス
 * @param[in]       div     プロセス分割情報(I,j,K方向の分割数)
 * @param[in]       localblock_count     ローカルブロック数
 * @return    success(=true)/fault(=false)
 */
bool sphMultiParaController::initializeVoxel(
                      const size_t wsize[3],
                      const size_t wsta[3],
                      unsigned int proc_div[3],
                      unsigned int localblock_count)
{
  register size_t i, j, l, n;
  bool ret;

  if (!wsize || !wsta || !proc_div) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "wsize or wsta or proc_div is null.");
    return false;
  }

  // get node(rank) count from mpi
  int nodeCnt = this->getNumberOfGroup();
  unsigned int dist_div[3] = {proc_div[0], proc_div[1], proc_div[2]};

  if( dist_div[0]*dist_div[1]*dist_div[2] == 0 ) {
    /// プロセス数からボクセルサイズを自動分割する
    if( !decideDivPattern(nodeCnt, wsize, dist_div) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "decideDivPattern fault");
      return false;
    }
  }
  if( dist_div[0]*dist_div[1]*dist_div[2] != nodeCnt ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "Invalid division values, [%d * %d * %d != %d]", dist_div[0], dist_div[1], dist_div[2], nodeCnt);
    return false;
  }

  /// 均等分割したすべてのサブブロックのサイズ、始点インデックスを取得する。
  // １プロセスノードのボクセルサイズ
  const size_t tmp_size[3] = {wsize[0]/dist_div[0], wsize[1]/dist_div[1], wsize[2]/dist_div[2]};

  // サブブロックのI,J,K方向分割数を取得する。
  unsigned int local_div[3] = {0, 0, 0};
  if (!decideDivPattern(localblock_count, tmp_size, local_div)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "decideDivPattern fault");
    return false;
  }

  // 計算領域全体をサブブロックで分割した場合のI,J,K方向分割数=プロセス分割情報(div)*サブブロック分割数(local_div)
  unsigned int sub_div[3] = {dist_div[0]*local_div[0], dist_div[1]*local_div[1], dist_div[2]*local_div[2]};
  unsigned int node_count = sub_div[0]*sub_div[1]*sub_div[2];

  // すべてのサブブロックのサイズ、始点インデックス、グループ番号(=プロセス番号)リストの作成
  size_t (*node_size)[3] = new size_t[node_count][3];
  size_t (*node_stratidx)[3] = new size_t[node_count][3];
  unsigned int *groupid_table = new unsigned int[node_count];
  memset(node_size, 0x00, node_count*3*sizeof(size_t));
  memset(node_stratidx, 0x00, node_count*3*sizeof(size_t));
  memset(groupid_table, 0x00, node_count*sizeof(unsigned int));

  /// プロセスのボクセルサイズを分割数に従って分割し、サブブロックのノードサイズ、始点インデックスを取得する。
  if (!createVoxelDistribution(
                      wsize,
                      sub_div,
                      node_size,
                      node_stratidx)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "decideDivPattern fault");
    return false;
  }

  // グループ番号(=プロセス番号)の割当を行う。
  for (n=0; n<node_count; n++) {
    int ks = n / (sub_div[0]*sub_div[1]);
    int kx = n / (sub_div[0]*sub_div[1]*local_div[2]);
    int js = (n - ks*sub_div[0]*sub_div[1]) / sub_div[0];
    int jx = (n - ks*sub_div[0]*sub_div[1]) / (sub_div[0]*local_div[1]);
    int is = (n - ks*sub_div[0]*sub_div[1] - js*sub_div[0]);
    int ix = (n - ks*sub_div[0]*sub_div[1] - js*sub_div[0]) / local_div[0];
    int proc_no = ix + (jx*dist_div[0]) + (kx*dist_div[0]*dist_div[1]);
    groupid_table[n] = proc_no;
  }

  // ボクセル情報の設定を行う。
  ret = this->initializeVoxel(node_count, wsize, wsta, node_size, node_stratidx, groupid_table);
  delete [] node_size; delete [] node_stratidx; delete [] groupid_table;

  if (!ret) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeLocalBlock is fault.");
    return false;
  }

  // 分割数を設定する
  // プロセスの分割数
  this->setVoxelDivInfo(dist_div);

  // ノードマッピングテーブルを作成する。
  if (!this->createNodeMap()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createNodeMap fault");
    return false;
  }

  // サブブロックの分割数
  std::map<int, sphSvParaController*>::iterator itr = m_mapChildPara.begin();
  for( ; itr != m_mapChildPara.end(); itr++){
    // node info
    sphSvParaController* sub_ctrl = itr->second;
    sub_ctrl->setVoxelDivInfo(sub_div);
    sub_ctrl->setVoxelDivLocalInfo(local_div);

    // サブブロックのノードマッピングテーブルを作成する。
    if (!sub_ctrl->createNodeMap()) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createNodeMap fault");
      return false;
    }
  }

  return true;
}


/**
 * 計算領域サイズ、始点インデックス、分割数から部分領域の初期化を行う。
 * @param[in]       wsize   計算領域サイズ
 * @param[in]       wsta    グローバル始点インデックス
 * @param[in]       proc_num    分割数
 * @param[in]       localblock_count     ローカルブロック数
 * @return    success(=true)/fault(=false)
 */
bool sphMultiParaController::initializeVoxel(
                      const size_t wsize[3],
                      const size_t wsta[3],
                      unsigned int proc_num,
                      unsigned int localblock_count)
{
  /// 分割数からI,J,Kの分割情報を算出する。
  unsigned int proc_div[3] = {0, 0, 0};
  if (!decideDivPattern(proc_num, wsize, proc_div)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "decideDivPattern fault");
    return false;
  }


  /// sphMultiParaController::initializeVoxel(wsize, wsta, div, localblock_count)を呼ぶ。
  return this->initializeVoxel(wsize, wsta, proc_div, localblock_count);

}


/**
 * ノード位置、サイズ情報リストから部分領域の初期化を行う。
 *
 * 引数に設定されたすべてのノード情報リストから、初期化を行う。<br>
 * 他のノード情報は収集を行わず、引数のノード情報リストから隣接面情報を作成する。
 * @param      node_count      ノード情報リスト数
 * @param      wsize        ソルバ全体領域のサイズ
 * @param      wstart       ソルバ全体領域の始点インデックス
 * @param      node_size        サブブロックノードサイズリスト   (node_count x [ilen, jlen, klen])
 * @param      node_sta        サブブロック始点インデックスリスト(node_count x [iwsta, jwsta, kwsta])
 * @param groupid_table       グループID(=プロセスID)テーブル=[node_count]
 * @return    success(=true)/fault(=false)
 */
bool sphMultiParaController::initializeVoxel(
                      unsigned int node_count,
                      const size_t wsize[3],
                      const size_t wstart[3],
                      const size_t (*node_size)[3],
                      const size_t (*node_sta)[3],
                      const unsigned int *groupid_table)
{
  register size_t i, n;
  bool ret;

  if (!node_count || !node_sta || !node_size || !groupid_table) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "node_size or node_sta or groupid_table is null.");
    return false;
  }

  // グループID(=プロセスID)テーブルの作成
  m_processGroupId = new int[node_count];
  m_nodeCount = node_count;

  /// 引数のノード情報リストからプロセス毎の領域をノード情報リストを作成する。
  /// - 引数のノード情報リストはサブブロックの領域情報である。
  // プロセス数を取得する。(=最大プロセス番号+1)
  int proc_count = 0;
  for (n=0; n<node_count; n++) {
    int proc_id = groupid_table[n];
    m_processGroupId[n] = groupid_table[n];
    if (proc_id > proc_count) proc_count = proc_id;
  }
  proc_count += 1;

  // 自ランク番号
  int myRank = this->getMyRank();
  size_t my_size[3] = {1, 1, 1};
  size_t my_start[3] = {0, 0, 0};

  size_t (*proc_size)[3] = new size_t[proc_count][3];
  size_t (*proc_startidx)[3] = new size_t[proc_count][3];
  size_t (*proc_endidx)[3] = new size_t[proc_count][3];
  memset(proc_size, 0x00, proc_count*3*sizeof(size_t));
  memset(proc_startidx, 0xFF, proc_count*3*sizeof(size_t));
  memset(proc_endidx, 0x00, proc_count*3*sizeof(size_t));

  for (n=0; n<node_count; n++) {
    size_t proc_id = groupid_table[n];
    if (proc_id >= proc_count) {
      delete []proc_size; delete []proc_startidx;
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_BUFFER_OVERFLOW, "invalid process table.");
      return false;
    }

    // プロセス内の最小始点インデックス、最大終点インデックスを取得する。
    for (i=0; i<3; i++) {
      // 最大終点インデックスからプロセスサイズの取得
      if (proc_endidx[proc_id][i] < node_sta[n][i] + node_size[n][i]) {
        proc_endidx[proc_id][i] = node_sta[n][i] + node_size[n][i] - 1;
      }
      // 最小始点インデックスrの取得
      if (proc_startidx[proc_id][i] > node_sta[n][i]) {
        proc_startidx[proc_id][i] = node_sta[n][i];
      }
    }
  }
  for (n=0; n<proc_count; n++) {
    for (i=0; i<3; i++) {
      proc_size[n][i] = proc_endidx[n][i] - proc_startidx[n][i] + 1;
    }

    // 自プロセス領域サイズ
    if (n == myRank) {
      for (i=0; i<3; i++) {
        my_size[i] = proc_size[n][i];
        my_start[i] = proc_startidx[n][i];
      }
    }
  }

  /// 自プロセス計算領域のパラレルコントローラの作成を行う。
  ret = sphSvParaController::initializeVoxel(
                                proc_count,
                                wsize, wstart,
                                proc_size, proc_startidx);
  delete []proc_size; delete []proc_startidx; delete [] proc_endidx;

  if (!ret) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeVoxel is fault.");
    return false;
  }

  /// createLocalBlockにより子パラレルコントローラを生成、登録を行う。
  if (!initializeLocalBlock(
                        node_count,
                        my_size, my_start,
                        node_size, node_sta,
                        groupid_table)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeLocalBlock is fault.");
    return false;
  }

  return true;
}

/**
 * サブブロックのパラレルコントローラを生成、登録を行う。
 *
 * @param node_count         サブブロック総数
 * @param psize              プロセス領域のサイズ
 * @param pstart             プロセスの始点インデックス
 * @param node_size          サブブロックノードサイズ
 * @param node_startidx      サブブロックノード始点インデックス
 * @param groupid_table      サブブロックグループID(=プロセスID)
 * @return    success(=true)/fault(=false)
 */
bool sphMultiParaController::initializeLocalBlock(
              unsigned int node_count,
              const size_t psize[3],
              const size_t pstart[3],
              const size_t (*node_size)[3],
              const size_t (*node_startidx)[3],
              const unsigned int *groupid_table)
{
  register size_t i, n;

  if (!node_startidx || !node_size || !groupid_table) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "node_size or node_sta or groupid_table is null.");
    return false;
  }
  if (node_count <= 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "node_count is zero.");
    return false;
  }

  // 自ノード番号
  size_t myId = this->getMyID();

  // 自ノードのボクセルサイズ
  const size_t* mySize = getVoxelSize();
  // 自ノードの始点インデックス
  size_t myStartIdx[3];
  this->getVoxelStartIndex(myStartIdx);
  double myOrigin[3], myPitch[3];
  // 全体計算領域原点座標
  this->getVoxelOrigin(myOrigin);
  // 全体計算領域ピッチ
  this->getVoxelPitch(myPitch);

  // サブブロックのパラレルコントローラを作成する。(ノード情報マップ作成用)
  sphSvParaController* tmpl_ctrl = new sphSvParaController(this, myId);

  // ソルバのコミュニケータ、グループを設定する。
  tmpl_ctrl->initialize(this->getSplComm());

  // サブブロック間の隣接情報の作成を行う。
  if (!tmpl_ctrl->initializeVoxel(
                              node_count,
                              psize, pstart,
                              node_size, node_startidx,
                              groupid_table)) {
    delete tmpl_ctrl;
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "subblock initializeVoxel.");
    return false;
  }

  // ノード情報マップを取得する。
  MapNodeInfo* nodeInfo = tmpl_ctrl->getMapNodeInfo();

  SPL_Comm spl_comm = this->getSplComm();
  // ノード０〜
  int localid = 0;
  for (n=0; n<node_count; n++) {
    if (groupid_table[n] != myId) {
      continue;
    }

    // サブブロックのパラレルコントローラのインスタンス
    sphSvParaController* sub_ctrl = new sphSvParaController(this, n);

    // ソルバのコミュニケータ、グループを設定する。
    sub_ctrl->setCommGroup(spl_comm);

    // プロセス領域のボクセルサイズ、始点インデックス、原点座標、ピッチを設定する。
    sub_ctrl->setWholeVoxelSize(mySize);
    sub_ctrl->setReferIndex(myStartIdx);
    sub_ctrl->setVoxelOrigin(myOrigin);
    sub_ctrl->setVoxelPitch(myPitch);

    // サブブロック間の隣接情報を設定する。
    sub_ctrl->setMapNodeInfo(nodeInfo);

    // 生成サブブロックパラレルコントローラをテーブルに追加する。
    if (!addChildParaController(localid, sub_ctrl)) {
      delete tmpl_ctrl;
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "subblock addChildParaController.");
      return false;
    }
    localid++;
  }

  // ノード情報マップ作成用パラレルコントローラの破棄
  delete tmpl_ctrl;

  return true;
}


/**
 * ローカル番号からサブブロックのパラレルコントローラを取得する。
 *
 * ローカル番号は、プロセス内のサブブロックのみに０〜採番したIDである。
 * @param localid   ローカル番号（０〜）
 * @return  サブブロックパラレルコントローラ
 */
const sphSvParaController* sphMultiParaController::getChildParaController(int localid) const
{
  // キー=localidのサブブロックのパラレルコントローラを取得する。
  std::map<int, sphSvParaController*>::const_iterator it = m_mapChildPara.find(localid);
  if(it == m_mapChildPara.end()) return NULL;
  return it->second;
}


/**
 * サブブロックのパラレルコントローラを追加する。
 * @param localid   ローカル番号（０〜）=key
 * @param subctrl   サブブロックパラレルコントローラ = value
 * @return  サブブロックパラレルコントローラ
 */
bool sphMultiParaController::addChildParaController(
                        int localid,
                        sphSvParaController* subctrl)
{
  if (subctrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSvParaController is null.");
    return false;
  }

  // サブブロックのパラレルコントローラを追加する。
  if( !((m_mapChildPara.insert(std::make_pair(localid, subctrl))).second) ){
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_mapChildPara [label=%d].", localid);
    return false;
  }

  return true;
}

/**
 * プロセス内のサブブロックパラレルコントローラ数を取得する。
 * @return    サブブロックパラレルコントローラ数
 */
int sphMultiParaController::getNumberOfChildCtrl() const
{
  return m_mapChildPara.size();
}


/**
 * オブジェクトの文字列表現を返します。(デバッグ用)
 * @return    オブジェクトの文字列表現
 */
std::string sphMultiParaController::toString() const
{

  std::string msg = "sphMultiParaController class\n";
  char buf[1024] = {0};
  // 自身のボクセル情報
  msg += std::string("管理パラレルコントローラ(親クラス)情報\n");
  msg += sphSvParaController::toString() + '\n';

  // 部分領域数
  msg += std::string("部分領域パラレルコントローラ(子クラス)情報\n");
  int num = getNumberOfChildCtrl();
  sprintf(buf, "Child ParaContoller Size=%d", num);
  msg += std::string(buf) + '\n';

  for (int i=0; i<num; i++) {
    const sphSvParaController* para = getChildParaController(i);
    sprintf(buf, "Child ParaController[%d]", i);
    msg += std::string(buf) + '\n';
    msg += para->toString() + '\n';
  }
  return msg;
}


/**
 * ノードIDからサブブロックのパラレルコントローラを取得する。
 *
 * ノードIDは、すべてのプロセスのすべてのサブブロックにユニークに採番されたIDである。
 * @param nodeid   ノードID（すべてのサブブロックのユニークID）
 * @return  サブブロックパラレルコントローラ
 */
const sphSvParaController*
sphMultiParaController::getChildParaControllerByNodeId(int nodeid) const
{
  std::map<int, sphSvParaController*>::const_iterator itr = m_mapChildPara.begin();
  for ( ;itr != m_mapChildPara.end(); itr++) {
    sphSvParaController* paraCtrl = (*itr).second;
    int myid = paraCtrl->getMyID();
    if (myid == nodeid) {
      return paraCtrl;
    }
  }
  return NULL;
}



/**
 * サブブロックパラレルコントローラテーブルにノードIDのパラレルコントローラが存在するかチェックする。
 *
 * 自プロセスのサブブロックにノードIDのサブブロックが存在するかチェックする。
 * ノードIDは、すべてのプロセスのすべてのサブブロックにユニークに採番されたIDである。
 * @param nodeid  ノードID（すべてのサブブロックのユニークID）
 * @return    true=同一プロセス内に存在する。
 */
bool sphMultiParaController::isContainNodeId(int nodeid) const
{
  if (nodeid < 0) return false;
  return (getChildParaControllerByNodeId(nodeid) != NULL);
}

/**
 * サブブロックが属するプロセス番号を取得する。
 *
 * プロセスに属しない場合、SPL_PROC_NULLを返す。
 * @param nodeid      サブブロックID
 * @return      プロセス番号
 */
int sphMultiParaController::getProcessId(int nodeid) const
{
  if (nodeid < 0) return SPL_PROC_NULL;
  if (nodeid >= m_nodeCount) return SPL_PROC_NULL;
  return m_processGroupId[nodeid];

}


/**
 * すべてのプロセスのサブブロック総数を取得する。
 * @return    すべてのプロセスのサブブロック総数
 */
int sphMultiParaController::getNumberOfSubBlock() const
{
  return m_nodeCount;
}


/**
 * 部分領域管理をサポートしているか判断する。
 * @return    true=部分領域管理サポート
 */
bool sphMultiParaController::isMultiParaController() const
{
  return true;
}


/**
 * 並列実行であるかチェックする。
 *
 * 並列実行であっても、ボクセル分割数が１の場合はfalse(=逐次)を返す。
 * 逐次実行であっても、サブブロック総数が１より大きい場合はtrue(=並列)を返す。
 * @return    true=並列実行, false=逐次実行
 */
bool sphMultiParaController::isParallel() const
{
  // ノード数が１より多い場合は、true
  if (getNumberOfNode() > 1) return true;
  if (getNumberOfSubBlock() > 1) return true;
  return false;
}

#if 0     // not use delete at 2011/01/09 by @hira

/**
 * ノード間演算を行う。
 *
 * サブブロック間でバッファにローカルの演算を行う。（MPI_Allreduceは行わない。)
 * プロセス間でMPI_Allreduceを行うには、execAllreduceを実行する。
 * @param      sendbuf        送信データ
 * @param      recvbuf        受信データ
 * @param      c              送信データ数
 * @param      datatype      送信データタイプ
 * @param      operation      演算の種類
 * @param      mpi_label      MPI識別文字列
 * @return    success(=true)/fault(=false)
 */
bool sphMultiParaController::allreduce(void* sendbuf, void* recvbuf,
                               unsigned int c, SPL_Datatype datatype,
                               SPL_Op operation,
                               const char*mpi_label)
{

  register int i;
  if( !splIsParallel() ){
    int dsz = getMPIDatatypeSize(datatype);
    if (dsz <= 0) return false;
    memcpy(recvbuf, sendbuf, dsz*c);
    return true;
  }

  if( m_splComm == SPL_COMM_NULL ) return false;
  if (mpi_label == NULL) {
    if ( (spl_mpi_return_info.m_mpiErr = SPL_Allreduce(sendbuf,recvbuf,c,datatype,operation,m_splComm)) == SPL_MPI_SUCCESS ) {
      return true;
    }
    return false;
  }

  // MPI_Allreduce通信バッファに演算結果をセットするのみ。
  // MPI_Allreduceは行わない。
  std::string label = std::string(mpi_label);
  std::map<std::string, MPI_Allreduce_Buffer*>::iterator itr = m_mapAllReduceBuffer.find(label);

  // MPIデータ型サイズ
  int dsz = getMPIDatatypeSize(datatype);

  if (itr == m_mapAllReduceBuffer.end()) {
    // 新規作成
    MPI_Allreduce_Buffer *mpi_buf = new MPI_Allreduce_Buffer;
    size_t len = c*dsz;
    mpi_buf->sendbuf = malloc(len);
    mpi_buf->recvbuf = malloc(len);
    memset(mpi_buf->sendbuf, 0x00, len);
    memset(mpi_buf->recvbuf, 0x00, len);

    // 送信バッファーにコピーする。
    memcpy(mpi_buf->sendbuf, sendbuf, len);
    mpi_buf->c = c;
    mpi_buf->datatype = datatype;
    mpi_buf->operation = operation;

    // 通信バッファテーブルに追加する。
    if( !((m_mapAllReduceBuffer.insert(std::make_pair(label, mpi_buf))).second) ){
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_mapAllReduceBuffer [label=%s].", label.c_str());
      return false;
    }
    return true;
  }
  else {
    int err_no = 0;
    // 演算を行い、通信バッファ（送信バッファ）に演算結果をセットする
    MPI_Allreduce_Buffer* mpi_buf = itr->second;

    // 最大値
    if (operation == SPL_MAX) {
      if (datatype == SPL_CHAR) {
        max( (char*)sendbuf, (char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_CHAR || datatype == SPL_BYTE) {
        max( (unsigned char*)sendbuf, (unsigned char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_SHORT) {
        max( (short*)sendbuf, (short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_SHORT) {
        max( (unsigned short*)sendbuf, (unsigned short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_INT) {
        max( (int*)sendbuf, (int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED) {
        max( (unsigned int*)sendbuf, (unsigned int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_LONG) {
        max( (long*)sendbuf, (long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_LONG) {
        max( (unsigned long*)sendbuf, (unsigned long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == MPI_LONG_LONG_INT || datatype == MPI_LONG_LONG) {
        max( (long long*)sendbuf, (long long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_FLOAT || datatype == SPL_DOUBLE ||  datatype == SPL_REAL) {
        if (dsz == 4) {
          max( (float*)sendbuf, (float*)mpi_buf->sendbuf, c );
        }
        else if (dsz == 8) {
          max( (double*)sendbuf, (double*)mpi_buf->sendbuf, c );
        }
        else {
          err_no = -2;
        }
      }
      else {
        err_no = -1;
      }
    }
    // 最小値
    else if (operation == SPL_MIN) {
      if (datatype == SPL_CHAR) {
        min( (char*)sendbuf, (char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_CHAR || datatype == SPL_BYTE) {
        min( (unsigned char*)sendbuf, (unsigned char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_SHORT) {
        min( (short*)sendbuf, (short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_SHORT) {
        min( (unsigned short*)sendbuf, (unsigned short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_INT) {
        min( (int*)sendbuf, (int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED) {
        min( (unsigned int*)sendbuf, (unsigned int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_LONG) {
        min( (long*)sendbuf, (long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_LONG) {
        min( (unsigned long*)sendbuf, (unsigned long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == MPI_LONG_LONG_INT || datatype == MPI_LONG_LONG) {
        min( (long long*)sendbuf, (long long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_FLOAT || datatype == SPL_DOUBLE ||  datatype == SPL_REAL) {
        if (dsz == 4) {
          min( (float*)sendbuf, (float*)mpi_buf->sendbuf, c );
        }
        else if (dsz == 8) {
          min( (double*)sendbuf, (double*)mpi_buf->sendbuf, c );
        }
        else {
          err_no = -2;
        }
      }
      else {
        err_no = -1;
      }
    }
    // 合計
    else if (operation == SPL_SUM) {
      if (datatype == SPL_CHAR) {
        sum( (char*)sendbuf, (char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_CHAR || datatype == SPL_BYTE) {
        sum( (unsigned char*)sendbuf, (unsigned char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_SHORT) {
        sum( (short*)sendbuf, (short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_SHORT) {
        sum( (unsigned short*)sendbuf, (unsigned short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_INT) {
        sum( (int*)sendbuf, (int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED) {
        sum( (unsigned int*)sendbuf, (unsigned int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_LONG) {
        sum( (long*)sendbuf, (long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_LONG) {
        sum( (unsigned long*)sendbuf, (unsigned long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == MPI_LONG_LONG_INT || datatype == MPI_LONG_LONG) {
        sum( (long long*)sendbuf, (long long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_FLOAT || datatype == SPL_DOUBLE ||  datatype == SPL_REAL) {
        if (dsz == 4) {
          sum( (float*)sendbuf, (float*)mpi_buf->sendbuf, c );
        }
        else if (dsz == 8) {
          sum( (double*)sendbuf, (double*)mpi_buf->sendbuf, c );
        }
        else {
          err_no = -2;
        }
      }
      else {
        err_no = -1;
      }
    }
    // 積
    else if (operation == SPL_PROD) {
      if (datatype == SPL_CHAR) {
        prod( (char*)sendbuf, (char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_CHAR || datatype == SPL_BYTE) {
        prod( (unsigned char*)sendbuf, (unsigned char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_SHORT) {
        prod( (short*)sendbuf, (short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_SHORT) {
        prod( (unsigned short*)sendbuf, (unsigned short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_INT) {
        prod( (int*)sendbuf, (int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED) {
        prod( (unsigned int*)sendbuf, (unsigned int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_LONG) {
        prod( (long*)sendbuf, (long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_LONG) {
        prod( (unsigned long*)sendbuf, (unsigned long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == MPI_LONG_LONG_INT || datatype == MPI_LONG_LONG) {
        prod( (long long*)sendbuf, (long long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_FLOAT || datatype == SPL_DOUBLE ||  datatype == SPL_REAL) {
        if (dsz == 4) {
          prod( (float*)sendbuf, (float*)mpi_buf->sendbuf, c );
        }
        else if (dsz == 8) {
          prod( (double*)sendbuf, (double*)mpi_buf->sendbuf, c );
        }
        else {
          err_no = -2;
        }
      }
      else {
        err_no = -1;
      }
    }
    // 論理積
    else if (operation == SPL_LAND) {
      if (datatype == SPL_CHAR) {
        land( (char*)sendbuf, (char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_CHAR || datatype == SPL_BYTE) {
        land( (unsigned char*)sendbuf, (unsigned char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_SHORT) {
        land( (short*)sendbuf, (short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_SHORT) {
        land( (unsigned short*)sendbuf, (unsigned short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_INT) {
        land( (int*)sendbuf, (int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED) {
        land( (unsigned int*)sendbuf, (unsigned int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_LONG) {
        land( (long*)sendbuf, (long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_LONG) {
        land( (unsigned long*)sendbuf, (unsigned long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == MPI_LONG_LONG_INT || datatype == MPI_LONG_LONG) {
        land( (long long*)sendbuf, (long long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_FLOAT || datatype == SPL_DOUBLE ||  datatype == SPL_REAL) {
        if (dsz == 4) {
          land( (float*)sendbuf, (float*)mpi_buf->sendbuf, c );
        }
        else if (dsz == 8) {
          land( (double*)sendbuf, (double*)mpi_buf->sendbuf, c );
        }
        else {
          err_no = -2;
        }
      }
      else {
        err_no = -1;
      }
    }
    // ビット演算の積
    else if (operation == SPL_BAND) {
      if (datatype == SPL_CHAR) {
        band( (char*)sendbuf, (char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_CHAR || datatype == SPL_BYTE) {
        band( (unsigned char*)sendbuf, (unsigned char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_SHORT) {
        band( (short*)sendbuf, (short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_SHORT) {
        band( (unsigned short*)sendbuf, (unsigned short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_INT) {
        band( (int*)sendbuf, (int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED) {
        band( (unsigned int*)sendbuf, (unsigned int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_LONG) {
        band( (long*)sendbuf, (long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_LONG) {
        band( (unsigned long*)sendbuf, (unsigned long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == MPI_LONG_LONG_INT || datatype == MPI_LONG_LONG) {
        band( (long long*)sendbuf, (long long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_FLOAT || datatype == SPL_DOUBLE ||  datatype == SPL_REAL) {
        // 実数はビット演算は行えない。
      }
      else {
        err_no = -1;
      }
    }
    // 論理和
    else if (operation == SPL_LOR) {
      if (datatype == SPL_CHAR) {
        lor( (char*)sendbuf, (char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_CHAR || datatype == SPL_BYTE) {
        lor( (unsigned char*)sendbuf, (unsigned char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_SHORT) {
        lor( (short*)sendbuf, (short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_SHORT) {
        lor( (unsigned short*)sendbuf, (unsigned short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_INT) {
        lor( (int*)sendbuf, (int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED) {
        lor( (unsigned int*)sendbuf, (unsigned int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_LONG) {
        lor( (long*)sendbuf, (long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_LONG) {
        lor( (unsigned long*)sendbuf, (unsigned long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == MPI_LONG_LONG_INT || datatype == MPI_LONG_LONG) {
        lor( (long long*)sendbuf, (long long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_FLOAT || datatype == SPL_DOUBLE ||  datatype == SPL_REAL) {
        if (dsz == 4) {
          lor( (float*)sendbuf, (float*)mpi_buf->sendbuf, c );
        }
        else if (dsz == 8) {
          lor( (double*)sendbuf, (double*)mpi_buf->sendbuf, c );
        }
        else {
          err_no = -2;
        }
      }
      else {
        err_no = -1;
      }
    }
    // ビット演算の和
    else if (operation == SPL_BOR) {
      if (datatype == SPL_CHAR) {
        bor( (char*)sendbuf, (char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_CHAR || datatype == SPL_BYTE) {
        bor( (unsigned char*)sendbuf, (unsigned char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_SHORT) {
        bor( (short*)sendbuf, (short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_SHORT) {
        bor( (unsigned short*)sendbuf, (unsigned short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_INT) {
        bor( (int*)sendbuf, (int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED) {
        bor( (unsigned int*)sendbuf, (unsigned int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_LONG) {
        bor( (long*)sendbuf, (long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_LONG) {
        bor( (unsigned long*)sendbuf, (unsigned long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == MPI_LONG_LONG_INT || datatype == MPI_LONG_LONG) {
        bor( (long long*)sendbuf, (long long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_FLOAT || datatype == SPL_DOUBLE ||  datatype == SPL_REAL) {
        // 実数はビット演算は行えない。
      }
      else {
        err_no = -1;
      }
    }
    // 排他的論理和
    else if (operation == SPL_LXOR) {
      if (datatype == SPL_CHAR) {
        lxor( (char*)sendbuf, (char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_CHAR || datatype == SPL_BYTE) {
        lxor( (unsigned char*)sendbuf, (unsigned char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_SHORT) {
        lxor( (short*)sendbuf, (short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_SHORT) {
        lxor( (unsigned short*)sendbuf, (unsigned short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_INT) {
        lxor( (int*)sendbuf, (int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED) {
        lxor( (unsigned int*)sendbuf, (unsigned int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_LONG) {
        lxor( (long*)sendbuf, (long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_LONG) {
        lxor( (unsigned long*)sendbuf, (unsigned long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == MPI_LONG_LONG_INT || datatype == MPI_LONG_LONG) {
        lxor( (long long*)sendbuf, (long long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_FLOAT || datatype == SPL_DOUBLE ||  datatype == SPL_REAL) {
        // 実数は排他的論理和は行えない。
      }
      else {
        err_no = -1;
      }
    }
    // ビット演算の排他的論理和
    else if (operation == SPL_BXOR) {
      if (datatype == SPL_CHAR) {
        bxor( (char*)sendbuf, (char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_CHAR || datatype == SPL_BYTE) {
        bxor( (unsigned char*)sendbuf, (unsigned char*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_SHORT) {
        bxor( (short*)sendbuf, (short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_SHORT) {
        bxor( (unsigned short*)sendbuf, (unsigned short*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_INT) {
        bxor( (int*)sendbuf, (int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED) {
        bxor( (unsigned int*)sendbuf, (unsigned int*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_LONG) {
        bxor( (long*)sendbuf, (long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_UNSIGNED_LONG) {
        bxor( (unsigned long*)sendbuf, (unsigned long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == MPI_LONG_LONG_INT || datatype == MPI_LONG_LONG) {
        bxor( (long long*)sendbuf, (long long*)mpi_buf->sendbuf, c );
      }
      else if (datatype == SPL_FLOAT || datatype == SPL_DOUBLE ||  datatype == SPL_REAL) {
        // 実数は排他的論理和は行えない。
      }
      else {
        err_no = -1;
      }
    }
    if (err_no == -1) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "not support SPL_Datatype.");
      return false;
    }
    else if (err_no == -2) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "not support SPL_Op.");
      return false;
    }
    return true;
  }
}


/**
 * ノード間演算(MPI_Allreduce)の通信を行う。
 *
 * MPI_Allreduce通信バッファテーブルからMPI_Allreduceをプロセス間にて実行する。
 * 実行結果はrecvbufにセットする。
 * mpi_labelがNULLの場合はすべてのテーブルに対してMPI_Allreduceを実行する。
 * @param      mpi_label      通信MPI識別文字列
 * @return    success(=true)/fault(=false)
 */
bool sphMultiParaController::execAllreduce(const char*mpi_label)
{

  if( m_splComm == SPL_COMM_NULL ) return false;

  if (mpi_label != NULL) {
    // 指定MPI識別文字列のバッファーを実行する。
    std::string label = std::string(mpi_label);
    std::map<std::string, MPI_Allreduce_Buffer*>::iterator itr = m_mapAllReduceBuffer.find(label);

    if (itr == m_mapAllReduceBuffer.end()) return false;

    // 通信バッファ
    MPI_Allreduce_Buffer* mpi_buf = itr->second;

    // MPI_Allreduiceの実行
    if ( (spl_mpi_return_info.m_mpiErr
                   = SPL_Allreduce(
                          mpi_buf->sendbuf, mpi_buf->recvbuf,
                          mpi_buf->c, mpi_buf->datatype,
                          mpi_buf->operation, m_splComm)) != SPL_MPI_SUCCESS ) {
      return false;
    }
  }
  else {
    // 通信バッファテーブルのすべてのMPI通信を行う。
    std::map<std::string, MPI_Allreduce_Buffer*>::iterator itr = m_mapAllReduceBuffer.begin();
    std::map<std::string, MPI_Allreduce_Buffer*>::iterator itr_end = m_mapAllReduceBuffer.end();
    for (; itr != itr_end; itr++) {
      // 通信バッファ
      MPI_Allreduce_Buffer* mpi_buf = itr->second;

      // MPI_Allreduiceの実行
      if ( (spl_mpi_return_info.m_mpiErr
                     = SPL_Allreduce(
                            mpi_buf->sendbuf, mpi_buf->recvbuf,
                            mpi_buf->c, mpi_buf->datatype,
                            mpi_buf->operation, m_splComm)) != SPL_MPI_SUCCESS ) {
        return false;
      }
    }
  }
  return true;
}


/**
 * MPI_Allreduce通信バッファテーブルをクリアする。
 *
 * mpi_labelがNULLの場合はすべてのテーブルをクリアする。
 * @param      mpi_label      削除MPI識別文字列
 * @return    success(=true)/fault(=false)
 */
bool sphMultiParaController::clearAllreduceBuffer(const char*mpi_label)
{

  if (mpi_label != NULL) {
    // 指定MPI識別文字列のバッファーをクリアする。
    std::string label = std::string(mpi_label);
    std::map<std::string, MPI_Allreduce_Buffer*>::iterator itr = m_mapAllReduceBuffer.find(label);

    if (itr == m_mapAllReduceBuffer.end()) return false;

    // 通信バッファ
    MPI_Allreduce_Buffer* mpi_buf = itr->second;
    free(mpi_buf->sendbuf);
    free(mpi_buf->recvbuf);
    delete mpi_buf;
    mpi_buf = NULL;
    m_mapAllReduceBuffer.erase(itr);
  }
  else {
    // 通信バッファテーブルのすべてのMPI通信を行う。
    std::map<std::string, MPI_Allreduce_Buffer*>::iterator itr = m_mapAllReduceBuffer.begin();
    std::map<std::string, MPI_Allreduce_Buffer*>::iterator itr_end = m_mapAllReduceBuffer.end();
    for (; itr != itr_end; itr++) {
      // 通信バッファ
      MPI_Allreduce_Buffer* mpi_buf = itr->second;

      free(mpi_buf->sendbuf);
      free(mpi_buf->recvbuf);
      delete mpi_buf;
      mpi_buf = NULL;
    }
    m_mapAllReduceBuffer.clear();
  }

  return true;
}

/**
 * MPI_Allreduce通信バッファを取得する。
 *
 * @param      mpi_label      MPI識別文字列
 * @return    MPI_Allreduce通信バッファ
 */
MPI_Allreduce_Buffer* sphMultiParaController::getAllreduceBuffer(const char*mpi_label)
{

  if (mpi_label == NULL) return NULL;

  // 指定MPI識別文字列のバッファーを取得する。
  std::string label = std::string(mpi_label);
  std::map<std::string, MPI_Allreduce_Buffer*>::iterator itr = m_mapAllReduceBuffer.find(label);

  if (itr == m_mapAllReduceBuffer.end()) return NULL;

  // 通信バッファ
  MPI_Allreduce_Buffer* mpi_buf = itr->second;

  return mpi_buf;
}

#endif
