/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphParaController.C
 * Parallel Controller Class Source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphParaController.C
 * パラレルコントローラークラスソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#include "parallel/sphParaController.h"


/**
 * \~english
 * Constructor
 * @param      paraKey      parallel Controller Key
 *
 * \~japanese
 * コンストラクタ
 * @param      paraKey      パラレルマネージャ識別キー
 */
sphParaController::sphParaController(unsigned int paraKey)
{
  m_paraKey = paraKey;
  m_hostNameList.clear();
  m_idTable = NULL;
  m_splComm = SPL_COMM_NULL;
  m_splGroup = SPL_GROUP_NULL;
  m_myID = 0;
  m_immediateRequest = new KeyMapping<SPL_Request>();
}


/**
 * \~english
 *
 * \~japanese
 * デストラクタ
 */
sphParaController::~sphParaController()
{
  m_hostNameList.clear();
  if (m_idTable != NULL) {
    delete[] m_idTable;
  }
  if (m_immediateRequest != NULL) delete m_immediateRequest;
}



/**
 * コミュニケータ、グループを設定する。
 * @param       comm    コミュニケータ
 * @param       group   グループ
 * @return      void
 */
void sphParaController::setCommGroup(SPL_Comm comm, SPL_Group group)
{
  m_splComm = comm;
  if (group != SPL_GROUP_NULL) {
    m_splGroup = group;
  }
  else {
    SPL_Comm_group(m_splComm, &m_splGroup);
  }

  return;
}


/**
 * \~english
 * Create Host Name Table
 *
 * \~japanese
 * ホスト名テーブルを作成する。
 */
bool sphParaController::createHostNameTable()
{
  // get comm size
  unsigned int nnum = getNumberOfGroup();
  if( nnum <= 0 ) return false;
  // get comm id
  int myid = getMyRank();
  if( myid == SPL_PROC_NULL ) return false;

  m_hostNameList.clear();
  int flg = 0;
  char hostName[256];
  if( gethostname(hostName, sizeof(hostName)) != 0 ) flg = 1;
  if( flg == 1 ) printf("error gethostname.\n");
  int tmp = 0;
  allreduce(&flg, &tmp, 1, SPL_INT, SPL_SUM);
  if( tmp != 0 ) return false;

  for(register int n=0; n<nnum; n++) {
    char recvHost[256] = {0x00};
    if (n==myid) {
      broadcast(hostName, sizeof(hostName), SPL_CHAR, n);
      strcpy(recvHost, hostName);
    }
    else {
      broadcast(recvHost, sizeof(recvHost), SPL_CHAR, n);
    }
    std::string strName(recvHost);
    m_hostNameList.push_back(strName);
  }
  return true;
}


/**
 * \~english
 * get Host Name
 * @param      id      rank number in process group
 * @return    const char*      Host Name
 *
 * \~japanese
 * ホスト名を取得する。
 * @param      id      プロセスグループ内のランク番号
 * @return    const char*      ホスト名
 */
const char* sphParaController::getHostName(int id)
{
  if( id < 0 ) return NULL;
  if (m_idTable == NULL) return NULL;

  int num = getNumberOfGroup();
  if( id >= num ) return NULL;
  int wid = m_idTable[id];
  if( wid >= m_hostNameList.size() ) return NULL;
  return m_hostNameList.at(wid).c_str();
}


/**
 * \~english
 * create Rank Table in process group
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * プロセスグループ内のランク番号テーブルを作成する。
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::createGrpIDTable()
{
  // get my node id in process group
  int id  = getMyRank();
  m_myID = id;
  if( id == SPL_PROC_NULL ) return true;
  // get my global node id
  int wid = getMyWorldRank();
  if( wid  == SPL_PROC_NULL  ) return false;
  if (m_idTable) {
    delete[] m_idTable;
    m_idTable = NULL;
  }

  int num = getNumberOfGroup();
  m_idTable = new int[num];
  int* tmp  = new int[num];
  memset(m_idTable, 0, sizeof(int)*num);
  memset(tmp, 0, sizeof(int)*num);
  m_idTable[id] = wid;

  allreduce(m_idTable, tmp, num, SPL_INT, SPL_SUM);
  memcpy(m_idTable, tmp, sizeof(int)*num);
  delete [] tmp;

  return true;
}

/**
 * \~english
 * initialize
 *
 * Initialize MPI and create Host Name Table.<br>
 * This function must call one time, when Program start.
 * @param      argc    the number of arguments
 * @param      argv    the argument vector
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 初期化処理を行う。
 *
 * MPIの実行環境の初期化、ホスト名テーブルの作成を行う。
 * - プログラムの起動時に１回だけ呼び出さなければならない。
 * @param      argc    プログラム起動引数数
 * @param      argv    プログラム起動引数
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::initialize(SPL_Comm comm, SPL_Group group)
{
  /// コミュニケータを設定する。
  setCommGroup(comm, group);

  // create Host Name table
  if( !createHostNameTable() ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "createHostNameTable error.");
    return false;
  }

  // create IDTabel
  if( !createGrpIDTable() ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "createGrpIDTable error.");
    return false;
  }

  return true;
}


/**
 * \~english
 * finalize
 *
 * \~japanese
 * 終了処理を行う。
 */
void sphParaController::finalize()
{
  if( !splIsParallel() ) return;

  // Terminates MPI execution environment
  finalizeMPI();

  return;
}


/**
 * \~english
 * get process group rank table
 * @return      const int*        process group rank table
 *
 * \~japanese
 * プロセスグループランク番号テーブルを取得する。
 * @return      const int*        プロセスグループランク番号テーブル
 */
const int* sphParaController::getGrpIDTable() const
{
  return m_idTable;
}


/**
 * \~english
 * get mpi communicator
 * @return SPL_Comm      mpi communicator
 *
 * \~japanese
 * MPI コミュニケータを取得する。
 * @return SPL_Comm      MPI コミュニケータ
 */
SPL_Comm sphParaController::getSplComm() const
{
  return m_splComm;
}


/**
 * \~english
 * set Parallel Controller Identifier Key
 * @param      paraKey        Parallel Controller Identifier Key
 *
 * \~japanese
 * パラレルコントローラー識別キーを設定する。
 * @param      paraKey        パラレルコントローラー識別キー
 */
void sphParaController::setParaKey(unsigned int paraKey)
{
  m_paraKey = paraKey;
}

/**
 * \~english
 * get parallel Controller Identifier Key
 * @return      paraKey        Parallel Controller Identifier Key
 *
 * \~japanese
 * パラレルコントローラー識別キーを取得する。
 * @return      paraKey        パラレルコントローラー識別キー
 */
unsigned int sphParaController::getParaKey() const
{
  return m_paraKey;
}


/**
 * \~english
 * set Parallel Controller Identifier Key
 * @param      paraKey        Parallel Controller Identifier Key
 *
 * \~japanese
 * パラレルコントローラー識別キーを設定する。
 * @param      paraKey        パラレルコントローラー識別キー
 */
void sphParaController::setSubKey(unsigned int subKey)
{
  m_subKey = subKey;
}

/**
 * \~english
 * get parallel Controller Identifier Key
 * @return      paraKey        Parallel Controller Identifier Key
 *
 * \~japanese
 * パラレルコントローラー識別キーを取得する。
 * @return      paraKey        パラレルコントローラー識別キー
 */
unsigned int sphParaController::getSubKey() const
{
  return m_subKey;
}


/**
 * プロセスグループ内の自ノード番号を取得する。
 *
 * 部分領域管理の場合は、サブブロックIDを取得する。
 * @return    プロセスグループ内ノード番号
 */
int sphParaController::getMyID() const
{
  return m_myID;
}

/**
 * プロセスグループ内の自ノード番号を設定する。
 *
 * 部分領域管理の場合は、サブブロックIDを設定する。
 * @param   myid    プロセスグループ内ノード番号、サブブロック番号
 */
void sphParaController::setMyID(int myid)
{
  m_myID = myid;
}


/**
 * \~english
 * create process group
 * @param      idNum          Node(Rank) List count
 * @param      idList          Node(Rank) List of process group
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * プロセスグループを作成する。
 * @param      idNum          プロセスグループノードリスト数
 * @param      idList          プロセスグループノードリスト
 * @return    success(=true)/fault(=false)
 */
SPL_Comm sphParaController::createProcessGroup(int idNum, int* idList, SPL_Comm world_comm)
{

  // create Host Name table in mpi_comm_world
  if( !createHostNameTable() ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ENVIRONMENT, "createHostNameTable error.");
    return SPL_COMM_NULL;
  }

  // create comm group in mpi
  if (!createCommGroupByComm(idNum, idList, world_comm, &m_splComm, &m_splGroup )) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "createCommGroupMPI error.");
    return SPL_COMM_NULL;
  }
  if (m_splComm == SPL_COMM_NULL) {
    return SPL_COMM_NULL;
  }

  // create IDTabel
  if( !createGrpIDTable() ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ENVIRONMENT, "createGrpIDTable error.");
    return SPL_COMM_NULL;
  }

  return m_splComm;
}


/**
 * \~english
 * Get Whole Voxel Size in process group
 * @return    Whole Voxel Size {i, j, k}
 *
 * \~japanese
 * プロセスグループ内のすべてのノードを結合したサイズを取得する。
 * @return    ボクセルサイズ {i, j, k}
 */
const size_t* sphParaController::getWholeVoxelSize() const
{
  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "sphParaController::getWholeVoxelSize().");
  return NULL;
}


/**
 * \~english
 * Get Voxel Head Index in node
 * @param      nodeId        nodeid
 * @param      dirId          direction(I=0, J=1, K=2)
 * @return    Voxel Head Index
 *
 * \~japanese
 * 始点インデックスを取得する。
 * @param      nodeId        ノード番号
 * @param      dirId          取得方向(I=0, J=1, K=2)
 * @return    始点インデックス
 */
size_t sphParaController::getVoxelHeadIndex(int nodeId,
                                      unsigned int dirId) const
{
  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "sphParaController::getVoxelHeadIndex().");
  return 0;
}


/**
 * \~english
 * Get Voxel Tail Index in node
 * @param      nodeId        nodeid
 * @param      dirId          direction(I=0, J=1, K=2)
 * @return    Voxel Tail Index
 *
 * \~japanese
 * 終点インデックスを取得する。
 * @param      nodeId        ノード番号
 * @param      dirId          取得方向(I=0, J=1, K=2)
 * @return    終点インデックス
 */
size_t sphParaController::getVoxelTailIndex(int nodeId,
                                          unsigned int dirId) const
{
  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "sphParaController::getVoxelTailIndex().");
  return 0;
}


/**
 * \~english
 * Get Voxel Size of node
 * @param      nodeId      node id (not parameter=my node)
 * @return    node size {i, j, k}<br>
 *             NULL = not exist node. or node size is zero.
 *
 * \~japanese
 * ノードサイズを取得する。
 * @param      nodeId      ノード番号（指定しない場合：自ノード）
 * @return    ノードサイズ {i, j, k}
 *             NULL：ノードが存在しない。又はノードサイズが０。
 */
const size_t* sphParaController::getVoxelSize(int nodeId) const
{
  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "sphParaController::getVoxelSize().");
  return NULL;
}


/**
 * ピッチを取得する。
 * @param      pitch          ピッチ
 * @return    void
 */
bool sphParaController::getVoxelPitch(double pitch[3]) const
{
  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "sphParaController::getVoxelPitch().");
  return false;
}


/**
 * 始点インデックスを取得する。
 * @param      sta_idx        始点インデックス（ガイドセルを含まない）
 * @return    void
 */
void sphParaController::getVoxelStartIndex(size_t sta_idx[3]) const
{
  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "sphParaController::getVoxelStartIndex().");
  return;

}


/**
 * \~english
 * get node count in node information class list
 * @return    node count
 *
 * \~japanese
 * ノード情報クラスリストのノード数を取得する。
 * @return    ノード数
 */
int  sphParaController::getNumberOfNode() const
{
  // error
  m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "sphParaController::getNumberOfNode().");
  return 0;
}


/**
 * プロセス内のサブブロックパラレルコントローラ数を取得する。
 *
 * sphMultiParaControllerクラスのみ有効である。他派生クラスは0を返す。
 * @return    0:固定
 */
int sphParaController::getNumberOfChildCtrl() const
{
  return 0;
}


/**
 * 部分領域管理をサポートしているか判断する。
 * @return    true=部分領域管理サポート
 */
bool sphParaController::isMultiParaController() const
{
  return false;
}


/**
 * 全プロセスのサブブロック総数を取得する。
 *
 * sphMultiParaControllerクラス以外の場合はgetNumberOfNodeと等価である。
 * @return    全プロセスのサブブロック総数
 */
int sphParaController::getNumberOfSubBlock() const
{
  return this->getNumberOfNode();
}


/**
 * オブジェクトの文字列表現を返します。(デバッグ用)
 * @return    オブジェクトの文字列表現
 */
std::string sphParaController::toString() const
{

  std::string msg = "sphParaController class\n";
  char buf[1024] = {0};

  if( splIsParallel() ) {
    msg += "splIsParallel = Running MPI\n";
  }
  else {
    msg += "splIsParallel = Not MPI\n";
  }

  int num = getNumberOfGroup();
  sprintf(buf, "NumberOfGroup=%d", num);
  msg += std::string(buf) + '\n';

  msg += "GroupIdTable:\n";
  const int* idTable = getGrpIDTable();
  sprintf(buf, "    ");
  for (int i=0; i<num; i++) {
    sprintf(buf, "%s, %d", buf, idTable[i]);
    if ((i+1)%32 == 0) {
      msg += std::string(buf) + '\n';
      sprintf(buf, "    ");
    }
  }
  msg += std::string(buf) + '\n';

  msg += "HostNameTable:\n";
  for (int i=0; i<num; i++) {
    std::string name = m_hostNameList.at(i);
    sprintf(buf, "    hostname[%d] = %s", i, name.c_str());
    msg += std::string(buf) + '\n';
  }

  unsigned int paraKey = getParaKey();
  sprintf(buf, "paraKey=%d", paraKey);
  msg += std::string(buf) + '\n';

  SPL_Comm splComm = getSplComm();
  sprintf(buf, "splComm=%d", splComm);
  msg += std::string(buf) + '\n';

  return msg;
}

