/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphParaMPI.C
 * MPI Wrapper Function Source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphParaMPI.C
 * MPIラッパー関数ソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#include "parallel/sphParaController.h"


/**
 * \~english
 * MPI Routine Return information
 * \~japanese
 * MPI実行結果情報
 */
Spl_Mpi_ReturnInfo spl_mpi_return_info;

/**
 * \~english
 * Initialize the MPI execution environment
 * @param      argc    the number of arguments
 * @param      argv    the argument vector
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * MPI実行環境の初期化を行う。
 * @param      argc    プログラム起動引数数
 * @param      argv    プログラム起動引数
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::initMPI(int* argc, char*** argv)
{
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if (!argc) return false;
  if (!argv) return false;

  // Initialize MPI
  if( (spl_mpi_return_info.m_mpiErr = SPL_Init(argc, argv)) != SPL_MPI_SUCCESS) return false;

  return true;
}


/**
 * \~english
 * Terminates MPI execution environment
 *
 * \~japanese
 * MPI実行環境の終了処理を行う。
 */
void sphParaController::finalizeMPI()
{
  if( !splIsParallel() ) return;

  if (m_splComm == SPL_COMM_WORLD) {
    int flag;
    SPL_Initialized(&flag);
    if( flag ) SPL_Finalize();
  }
  else {
    if( m_splComm != SPL_COMM_NULL ) {
      SPL_Comm_free(&m_splComm);
      SPL_Group_free(&m_splGroup);
    }
  }
  return;
}


/**
 * \~english
 * get process count in process group
 * @return    process count in process group<br>
 *            if not parallel, return value is 1.<br>
 *            if error occur,  return value is 0.<br>
 *
 * \~japanese
 * プロセスグループ内の実行プロセス数を取得する。
 * @return    プロセスグループ内の実行プロセス数<br>
 *             単独実行時：1<br>
 *            エラー：0<br>
 */
unsigned int sphParaController::getNumberOfGroup() const
{
  if( !splIsParallel() ) return 1;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if( m_splComm == SPL_COMM_NULL ) return 0;
  int size;
  if ( (spl_mpi_return_info.m_mpiErr = SPL_Comm_size(m_splComm, &size)) == SPL_MPI_SUCCESS ) {
    if (size >= 0) return size;
  }
  return 0;
}

/**
 * \~english
 * get all process count
 * @return    all process count<br>
 *            if not parallel, return value is 1.<br>
 *            if error occur,  return value is 0.<br>
 *
 * \~japanese
 * すべての実行プロセス数を取得する。
 * @return    すべての実行プロセス数<br>
 *             単独実行時：1<br>
 *            エラー：0<br>
 */
unsigned int sphParaController::getNumberOfWorld() const
{
  if( !splIsParallel() ) return 1;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  int size;
  if( (spl_mpi_return_info.m_mpiErr = SPL_Comm_size(SPL_COMM_WORLD, &size)) == SPL_MPI_SUCCESS ) {
    if (size >= 0) return size;
  }
  return 0;
}

/**
 * \~english
 * get my rank index in process group
 * @return    my rank index<br>
 *            if error occur,  return value is SPL_PROC_NULL.<br>
 *
 * \~japanese
 * プロセスグループ内の自ランク番号（ノード番号）を取得する。<br>
 * @return    自ランク番号（ノード番号）<br>
 *            エラー：SPL_PROC_NULL<br>
 */
int sphParaController::getMyRank() const
{

  if( !splIsParallel() ) return 0;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if( m_splComm == SPL_COMM_NULL ) return SPL_PROC_NULL;
  int id;
  if( (spl_mpi_return_info.m_mpiErr = SPL_Comm_rank(m_splComm, &id)) == SPL_MPI_SUCCESS )
    return id;
  else return SPL_PROC_NULL;
}

/**
 * \~english
 * get my rank index in all process
 * @return    my rank index<br>
 *            if error occur,  return value is SPL_PROC_NULL.
 *
 * \~japanese
 * すべてのプロセス内の自ランク番号（ノード番号）を取得する。
 * @return    自ランク番号（ノード番号）<br>
 *            エラー：SPL_PROC_NULL
 */
int sphParaController::getMyWorldRank() const
{

  if( !splIsParallel() ) return 0;
  int id;
  if(  (spl_mpi_return_info.m_mpiErr = SPL_Comm_rank(SPL_COMM_WORLD, &id)) == SPL_MPI_SUCCESS )
    return id;
  else return SPL_PROC_NULL;
}


/**
 * \~english
 * Create Process Group
 * @param      idNum      node list count
 * @param      idList    node list
 * @param      newComm    new MPI_Comm
 * @param      newGrp    new MPI_Group
 * @return    SPL_Comm  Create new communicator<br>
 *                       if error occur, return value is SPL_COMM_NULL
 *
 * \~japanese
 * 新規にプロセスグループを作成する。
 * @param      idNum      作成ノード数
 * @param      idList    作成ノードリスト
 * @param      newComm    新規作成コミュニケータ
 * @param      newGrp    新規作成グループ
 * @return    SPL_Comm    作成コミュニケータ<br>
 *                         作成エラー：SPL_COMM_NULL
 */
bool sphParaController::createCommGroupByWorld(int idNum, int* idList, SPL_Comm *newComm, SPL_Group *newGrp ) const
{
  if( !splIsParallel() ) return true;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;

  if( (idNum < 1) || !idList ) return false;

  *newComm == SPL_COMM_NULL;

  // create MPI_COMM
  SPL_Group oldGrp;

  if ( (spl_mpi_return_info.m_mpiErr = SPL_Comm_group(SPL_COMM_WORLD, &oldGrp)) != SPL_MPI_SUCCESS) {
    return false;
  }
  if ( (spl_mpi_return_info.m_mpiErr = SPL_Group_incl(oldGrp, idNum, idList, newGrp)) != SPL_MPI_SUCCESS) {
    return false;
  }
  if ( (spl_mpi_return_info.m_mpiErr = SPL_Comm_create(SPL_COMM_WORLD, *newGrp, newComm)) != SPL_MPI_SUCCESS) {
    return false;
  }

  return true;

}


/**
 * \~english
 * Create Process Group
 * @param      idNum      node list count
 * @param      idList    node list
 * @param      newComm    new MPI_Comm
 * @param      newGrp    new MPI_Group
 * @return    SPL_Comm  Create new communicator<br>
 *                       if error occur, return value is SPL_COMM_NULL
 *
 * \~japanese
 * 既存プロセスグループからプロセスグループを作成する。
 * @param      idNum      作成ノード数
 * @param      idList    作成ノードリスト
 * @param      oldComm    作成済みMPIコミュニケータ
 * @param      newComm    新規作成コミュニケータ
 * @param      newGrp    新規作成グループ
 * @return    SPL_Comm    作成コミュニケータ<br>
 *                         作成エラー：SPL_COMM_NULL
 */
bool sphParaController::createCommGroupByComm(int idNum, int* idList, SPL_Comm oldComm, SPL_Comm *newComm, SPL_Group *newGrp ) const
{
  if( !splIsParallel() ) return true;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;

  if( (idNum < 1) || !idList ) return false;

  *newComm == SPL_COMM_NULL;

  // create MPI_COMM
  SPL_Group oldGrp;

  if ( (spl_mpi_return_info.m_mpiErr = SPL_Comm_group(oldComm, &oldGrp)) != SPL_MPI_SUCCESS) {
    return false;
  }
  if ( (spl_mpi_return_info.m_mpiErr = SPL_Group_incl(oldGrp, idNum, idList, newGrp)) != SPL_MPI_SUCCESS) {
    return false;
  }
  if ( (spl_mpi_return_info.m_mpiErr = SPL_Comm_create(oldComm, *newGrp, newComm)) != SPL_MPI_SUCCESS) {
    return false;
  }

  return true;

}

/**
 * \~english
 * allreduce
 *
 * Combines values from all processes and distribute the result back to all processes<br>
 * if not mpi executable environment, sendbuf is copied to recvbuf.
 * @param      sendbuf        send buffer
 * @param      recvbuf        receive buffer
 * @param      c              send data count
 * @param      datatype      send data type
 * @param      operation      Combine operation
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ノード間演算を行う。
 *
 * MPI環境ではない場合、送信データを受信データにコピーする。
 * @param      sendbuf        送信データ
 * @param      recvbuf        受信データ
 * @param      c              送信データ数
 * @param      datatype      送信データタイプ
 * @param      operation      演算の種類
 * @param      mpi_label      MPI識別文字列
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::allreduce(void* sendbuf, void* recvbuf,
                               unsigned int c, SPL_Datatype datatype,
                               SPL_Op operation,
                               const char*mpi_label) const
{

  if( !splIsParallel() ){
    int dsz = getMPIDatatypeSize(datatype);
    if (dsz <= 0) return false;
    memcpy(recvbuf, sendbuf, dsz*c);
    return true;
  }
  else {
    if( m_splComm == SPL_COMM_NULL ) return false;
    if ( (spl_mpi_return_info.m_mpiErr = SPL_Allreduce(sendbuf,recvbuf,c,datatype,operation,m_splComm)) == SPL_MPI_SUCCESS ) {
      return true;
    }

    return false;
  }
}


/**
 * \~english
 * broadcast
 *
 * Broadcasts a message from the process with rank "root" to all other processes of the group.
 * @param      data          send/receive buffer
 * @param      c              send/receive data count
 * @param      datatype      send/receive data type
 * @param      id      rank of broadcast root
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * プロセスグループ内の他のプロセスにデータを送信する。
 * @param      data          送信/受信データ
 * @param      c              送信/受信データ数
 * @param      datatype      送信/受信データタイプ
 * @param      id            送信元プロセスのランク
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::broadcast(void* data, int c, SPL_Datatype datatype, int id) const
{
  if( !splIsParallel() ) return true;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if( !data ) return false;
  if( m_splComm == SPL_COMM_NULL ) return false;

  if( (spl_mpi_return_info.m_mpiErr = SPL_Bcast(data, c, datatype, id, m_splComm)) == SPL_MPI_SUCCESS ) {
    return true;
  }
  else return false;
}



/**
 * \~english
 * barrier
 *
 * Blocks until all process have reached this routine.
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * バリア同期を行う。
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::barrier() const
{
  if( !splIsParallel() ) return true;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if( m_splComm == SPL_COMM_NULL ) return false;

  if( (spl_mpi_return_info.m_mpiErr = SPL_Barrier(m_splComm)) == SPL_SUCCESS ) {
    return true;
  }
  else {
    return false;
  }
}


/**
 * \~english
 * Performs a basic blocking send
 * @param      data      send buffer
 * @param      c          number of elements in send buffer
 * @param      datatype  data type in send buffer
 * @param      dst        rank of destination
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ブロッキングを行い、データ送信を行う。
 * @param      data      送信データ
 * @param      c          送信データ数
 * @param      datatype  送信データタイプ
 * @param      dst        送信先ランク番号
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::send(void* data, int c,
                      SPL_Datatype datatype, int dst) const
{
  if( !splIsParallel() ) return true;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if( m_splComm == SPL_COMM_NULL ) return false;
  if( !data ) return false;

  int tag = 1;
  if( (spl_mpi_return_info.m_mpiErr = SPL_Send(data, c, datatype, dst, tag, m_splComm)) == SPL_SUCCESS ) {
    return true;
  }else return false;
}



/**
 * \~english
 * Basic blocking receive
 * @param      data      receive buffer
 * @param      c          number of elements in receive buffer
 * @param      datatype  data type in receive buffer
 * @param      src        rank of source
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ブロッキングを行い、データ受信を行う。
 * @param      data      受信データ
 * @param      c          受信データ数
 * @param      datatype  受信データタイプ
 * @param      src        送信元ランク番号
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::recv(void* data, int c, SPL_Datatype datatype, int src) const
{
  if( !splIsParallel() ) return true;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if( m_splComm == SPL_COMM_NULL ) return false;
  if( !data ) return false;

  int tag = 1;
  if( (spl_mpi_return_info.m_mpiErr = SPL_Recv(data, c, datatype, src, tag, m_splComm, &(spl_mpi_return_info.m_mpiStatus))) == SPL_SUCCESS ) {
    return true;
  }else return false;
}


/**
 * ブロッキングを行わず、データ送信を行う。.
 * 通信リクエスト番号は、リクエスト登録リストに登録し、登録キーを返す。
 * @param      data      送信データ
 * @param      c          送信データ数
 * @param      datatype  送信データタイプ
 * @param      dst        送信先ランク番号
 * @return     登録キー(>0:登録キー, <=0:送信/登録失敗)
 */
int sphParaController::iSend(void* data, int c, SPL_Datatype datatype, int dst)
{
  int key = -1;
  if( !splIsParallel() ) return 1;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if( m_splComm == SPL_COMM_NULL ) return key;
  if( !data ) return key;

  int tag = 1;
  SPL_Request *req = m_immediateRequest->create();
  if( !req ) return key;
  if( (spl_mpi_return_info.m_mpiErr = SPL_Isend(data, c, datatype, dst, tag, m_splComm, req)) != SPL_SUCCESS ){
    return key;
  }

  /// リクエスト番号を登録する。
  key = m_immediateRequest->add(req);

  return key;
}


/**
 * ブロッキングを行わず、データ受信を行う。
 * @param      data      受信データ
 * @param      c          受信データ数
 * @param      datatype  受信データタイプ
 * @param      src        送信元ランク番号
 * @return     登録キー(>0:登録キー, <=0:送信/登録失敗)
 */
int sphParaController::iRecv(void* data, int c, SPL_Datatype datatype, int src)
{
  int key = -1;
  if( !splIsParallel() ) return 1;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if( m_splComm == SPL_COMM_NULL ) return key;
  if( !data ) return key;

  int tag = 1;
  SPL_Request *req = m_immediateRequest->create();
  if( !req ) return key;
  if( (spl_mpi_return_info.m_mpiErr = SPL_Irecv(data, c, datatype, src, tag, m_splComm, req)) != SPL_SUCCESS ) {
    return key;
  }

  /// リクエスト番号を登録する。
  key = m_immediateRequest->add(req);

  return key;
}



/**
 * 単一ノンブロッキング通信の完了待ちを行う。
 * @param      key          リクエスト登録キー
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::wait(int key)
{
  if( !splIsParallel() ) return true;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;

  SPL_Request* req = m_immediateRequest->get(key);
  if( !req ) return false;

  if( (spl_mpi_return_info.m_mpiErr = SPL_Wait(req, &(spl_mpi_return_info.m_mpiStatus))) != SPL_SUCCESS ) {
    return false;
  }

  /// リクエスト番号を破棄する。
  m_immediateRequest->del(key);

  return true;
}


/**
 * リクエスト登録キーリストのノンブロッキング通信の完了待ちを行う。
 * @param      key_num      キーリスト数
 * @param      key_list          リクエスト登録キーリスト
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::waitAll(unsigned key_num, int* key_list)
{
  if( !splIsParallel() ) return true;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;

  SPL_Status* stat_list = new SPL_Status[key_num];
  SPL_Request* req_list = new SPL_Request[key_num];
  register unsigned i;
  for(i=0; i<key_num; i++) {
    SPL_Request* req = m_immediateRequest->get(key_list[i]);
    if( req ) {
      req_list[i] = *req;
    }
    else {
      req_list[i] = SPL_REQUEST_NULL; // mod 20100423 @tamaki - MPI_ -> SPL_
    }
  }

  bool ret = false;
  if( (spl_mpi_return_info.m_mpiErr = SPL_Waitall(key_num, req_list, stat_list)) == SPL_SUCCESS ) {
    ret = true;
    // 送受信完了により登録リストから破棄する。
    for(i=0; i<key_num; i++) {
      m_immediateRequest->del(key_list[i]);
    }
  }

  delete [] stat_list;
  delete [] req_list;

  return ret;
}


/**
 * 完了待ちとなっているすべてのリクエストの完了待ちを行う。
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::waitRequestAll()
{
  if( !splIsParallel() ) return true;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;

  int key_num = m_immediateRequest->count();
  if (key_num <= 0) {
    return true;
  }
  int* key_list = new int[key_num];

  // リクエストキーリストを取得する。
  if (m_immediateRequest->getKeyList(key_list) != key_num) {
    delete [] key_list;
    return false;
  }

  // ノンブロッキング通信の完了待ちを行う。
  bool ret = waitAll(key_num, key_list);
  delete [] key_list;

  return ret;
}


/**
 * \~english
 * Terminates MPI execution environment
 * @param      errcode          error code to return
 *
 * \~japanese
 * MPI実行環境を終了する。
 * @param      errcode          エラーコード
 */
void sphParaController::abort(int errcode) const
{
  SPL_Abort(SPL_COMM_WORLD, errcode);
}


/**
 * \~english
 * Gathers together values from a group of processes
 * @param      send_data      send buffer
 * @param      send_cnt          number of elements in send buffer
 * @param      send_datatype  data type in send buffer
 * @param      recv_data      receive buffer
 * @param      recv_cnt          number of elements in receive buffer
 * @param      recv_datatype  data type in receive buffer
 * @param      root        rank of receiving process
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * １つのプロセスが全プロセスからデータ収集を行う。
 * @param      send_data      送信データ
 * @param      send_cnt       送信データ数
 * @param      send_datatype  送信データタイプ
 * @param      recv_data      受信データ
 * @param      recv_cnt       受信データ数
 * @param      recv_datatype  受信データタイプ
 * @param      root            受信ランク番号
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::gather(void* send_data, int send_cnt, SPL_Datatype send_datatype,
                       void* recv_data, int recv_cnt, SPL_Datatype recv_datatype,
                       int root) const
{
  if( !splIsParallel() ) return true;
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if( m_splComm == SPL_COMM_NULL ) return false;
  if( !send_data ) return false;
  if ( getMyRank() == root) {
    if (!recv_data) return false;
  }

  // MPI_Gather
  if( (spl_mpi_return_info.m_mpiErr = SPL_Gather(send_data, send_cnt, send_datatype,
                                recv_data, recv_cnt, recv_datatype, root,
                                m_splComm)) == SPL_SUCCESS ) {
    return true;
  }else return false;

}


/**
 * \~english
 * allgather
 *
 * Gathers data from all tasks and distribute it to all<br>
 * if not mpi executable environment, send_data is copied to recv_data.
 * @param      send_data      send buffer
 * @param      send_cnt          number of elements in send buffer
 * @param      send_datatype  data type in send buffer
 * @param      recv_data      receive buffer
 * @param      recv_cnt          number of elements in receive buffer
 * @param      recv_datatype  data type in receive buffer
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * データ収集
 *
 * プロセスグループ内のすべてのプロセスに対しデータ収集を行う。<br>
 * MPI実行環境ではない場合、送信データを受信データにコピーする。
 * @param      send_data      送信データ
 * @param      send_cnt       送信データ数
 * @param      send_datatype  送信データタイプ
 * @param      recv_data      受信データ
 * @param      recv_cnt       受信データ数
 * @param      recv_datatype  受信データタイプ
 * @return    success(=true)/fault(=false)
 */
bool sphParaController::allgather(void* send_data, int send_cnt, SPL_Datatype send_datatype,
                       void* recv_data, int recv_cnt, SPL_Datatype recv_datatype) const
{
  if( !splIsParallel() ) {
    int dsz = getMPIDatatypeSize(send_datatype);
    if (dsz <= 0) return false;
    memcpy(recv_data, send_data, dsz*send_cnt);
    return true;
  }
  spl_mpi_return_info.m_mpiErr = SPL_ERR_NULL_POINTER;
  if( m_splComm == SPL_COMM_NULL ) return false;
  if( !send_data ) return false;
  if (!recv_data) return false;

  // MPI_Allgather
  if( (spl_mpi_return_info.m_mpiErr = SPL_Allgather(send_data, send_cnt, send_datatype,
                                recv_data, recv_cnt, recv_datatype,
                                m_splComm)) == SPL_SUCCESS ) {
    return true;
  }else return false;
}


/**
 * \~english
 * get datatype size
 *
 * Return the number of bytes occupied by entries in the datatype<br>
 * if not install MPI, You can get only base datatype(SPL_CHAR, SPL_INT, SPL_LONG...).
 * @param      datatype        datatype
 * @return    size in datatype<br>
 *             Error:0    can not get datatype
 *
 * \~japanese
 * データタイプのサイズを取得する。
 *
 * MPI環境が設定されていない場合は、基本データタイプ(=SPL_CHAR, SPL_INT, SPL_LONG...)のみ取得可能である。（派生データタイプは取得不可）
 * @param      datatype      データタイプ
 * @return    データタイプサイズ<br>
 *             ０：データタイプ取得不可（エラー）
 */
int sphParaController::getMPIDatatypeSize(SPL_Datatype datatype) const
{

  int size = 0;
  if( !splIsParallel() ) {
    size = spl_get_datasize(datatype);
  }
  else {
    // MPI_Type_size
    if( (spl_mpi_return_info.m_mpiErr = SPL_Type_size(datatype, &size)) != SPL_SUCCESS ) {
      return 0;
    }
  }
  return size;
}


