/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphParaManager.C
 * Parallel Manager Class Source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphParaManager.C
 * パラレルマネージャークラスソースファイル
 * @author     hira
 * @date       2009/02/20
 */
#include <math.h>
#include "parallel/sphParaManager.h"
#include "parallel/sphSvParaController.h"
#include "parallel/sphMultiParaController.h"
#include "parallel/sphUnstParaController.h"
#include "config/sphElemUtils.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphParaManager::sphParaManager()
{
  m_mapKey = 0;
  m_mapParaCtrl.clear();

}

/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 * パラレルコントローラの破棄を行う。
 */
sphParaManager::~sphParaManager()
{
  std::map<int, sphParaController*>::const_iterator it = m_mapParaCtrl.begin();
  for ( ;it != m_mapParaCtrl.end(); it++) {
    sphParaController* paraCtrl = (*it).second;
    if (paraCtrl != NULL) delete paraCtrl;
    paraCtrl = NULL;
  }
  m_mapParaCtrl.clear();
}


/**
 * 構造格子パラレルコントローラ作成、追加する。（ソルバコンフィグレーション要素）.
 * ソルバコンフィグレーション要素からパラレルコントローラを生成する。
 * @param       solverElem    ソルバコンフィグレーション要素
 * @param       comm    コミュニケータ
 * @return      生成パラレルコントローラ
 */
sphParaController*
sphParaManager::createSvParaController(
                            const sphElem* solverElem,
                            SPL_Comm comm)
{

  register int i;
  int ret = SPHERE_SUCCESS;
  sphParaController* paraCtrl = NULL;

  if (solverElem == NULL || (comm == SPL_COMM_NULL && splIsParallel())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createSvParaController arguments is null.");
    return NULL;
  }

  /// 起動プロセス数を取得する
  int world_size;
  SPL_Comm_size (SPL_COMM_WORLD, &world_size);

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  /// 次元数の取得を行う。
  int dims = solverElem->getAttrInt(SPH_CFG_ATTR_DIMS);
  if (dims == 0) dims = 3;      // デフォルト＝３
  if (dims < 0 || dims > 3) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid dims(=%d)", dims);
    ret = SPHERE_TERMINATE;
  }

  sphElemUtils elemUtils(solverElem);
  /// <SphDomainInfo>要素からサイズ、原点座標、ピッチを取得する。
  size_t voxel_size[3] = {1, 1, 1};
  size_t voxel_start[3] = {0, 0, 0};
  double voxel_origin[3], voxel_pitch[3];
  if (!elemUtils.getDomainInfo(dims, voxel_size, voxel_origin, voxel_pitch, voxel_start)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid <SphDomainInfo>.");
    ret = SPHERE_TERMINATE;
  }

  /// <SphDomainInfo>要素の取得を行う。
  const sphElem* domain_elem = solverElem->getElem(SPH_CFG_NAME_DOMAININFO);
  if (domain_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found <SphDomainInfo>");
    ret = SPHERE_TERMINATE;
  }

  /// <SphSteer>要素の取得を行う。
  const sphElem* steer_elem = solverElem->getElem(SPH_CFG_NAME_STEER);
  if (steer_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found <SphSteer>");
    ret = SPHERE_TERMINATE;
  }

  /// SphParallelTemplate要素が有効であるかチェックする。
  if (!validParallelTemplate(solverElem)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "invalid <SphParallelTemplate>");
    return NULL;
  }
  // SphEqualPartitioning要素
  const sphElem* ep_elem = solverElem->getElem(3,
                              SPH_CFG_NAME_STEER,
                              SPH_CFG_NAME_PARALLELTEMPLATE,
                              SPH_CFG_NAME_EQUALPARTITIONING);

  bool mbx_grouping = isMultiplyConnectedPartitioning(solverElem);
  int subblock_count = countEqualPartitioning(solverElem);

  /// <SphNodeProcess>要素からソルバプロセス数の取得を行う。
  const sphElem* proc_elem = steer_elem->getElem(SPH_CFG_NAME_NODEPROCESS);
  int proc_num = 0;
  if (proc_elem != NULL) {
    proc_num = proc_elem->getAttrInt(SPH_CFG_ATTR_VALUE);
  }
  else {
    proc_num = world_size;
  }
  if (world_size == 1) {
    // 逐次実行であるので分割数=1
    proc_num = 1;
  }

  if ( ret != SPHERE_SUCCESS ) {
    // 警告メッセージを元に戻す。
    m_errHandler->releaseWarning();
    return NULL;
  }

  /// <SphNodeList>要素からパラレルコントローラの作成を行う。
  /// 但し、SphEqualPartitioning要素が記述されている場合は無視する。
  const sphElem* nodelist_elem = NULL;
  if (ep_elem == NULL) {
    nodelist_elem = domain_elem->getElem(SPH_CFG_NAME_NODELIST);
    if (nodelist_elem != NULL) {
      // パラレルコントローラの生成を行う。
      paraCtrl = createSvParaController(
                                  dims,
                                  voxel_size,
                                  voxel_start,
                                  nodelist_elem,
                                  voxel_origin,
                                  voxel_pitch,
                                  comm,
                                  mbx_grouping);
    }
    else {
      if (mbx_grouping) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "not found <SphNodeList>");
        paraCtrl = NULL;
      }
    }
  }

  /// <SphVoxelDivision>要素からパラレルコントローラの作成を行う。
  const sphElem* div_elem = NULL;
  if (nodelist_elem == NULL) {
    div_elem = steer_elem->getElem(SPH_CFG_NAME_DIVISION);
    if (div_elem != NULL) {
      /// <SphVoxelDivision>要素のi, j, kの取得を行う。
      unsigned int div[3] = {1, 1, 1};
      if (world_size == 1) {
        // 単独実行であるので分割数=1
        div[0] = div[1] = div[2] = 1;
      }
      else {
        const char *attr_idx[] = {SPH_CFG_ATTR_I, SPH_CFG_ATTR_J, SPH_CFG_ATTR_K};
        if (!sphElemUtils(div_elem).getDimsAttributeValues(dims, SPH_CFG_NAME_DIVISION, attr_idx, div)) {
          // error
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "can not get SphDivision[@i, @j, @k], %s.", div_elem->debugPrintElem().c_str());
          div[0] = div[1] = div[2] = 0;
        }
      }
      int proc_num = 1;
      for (i=0; i<dims; i++) proc_num *= div[i];
      if (proc_num >= 1) {
        /// sphSvParaControllerを生成する。
        paraCtrl = createSvParaController(
                                voxel_size,
                                voxel_start,
                                voxel_origin,
                                voxel_pitch,
                                div,
                                comm,
                                subblock_count);
      }
    }
  }

  /// <SphNodeProcess>要素、起動プロセス数からパラレルコントローラの作成を行う。
  if (nodelist_elem == NULL && div_elem == NULL) {
    /// sphSvParaControllerを生成する。
    paraCtrl = createSvParaController(
                              voxel_size,
                              voxel_start,
                              voxel_origin,
                              voxel_pitch,
                              proc_num,
                              comm,
                              subblock_count);
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  if (paraCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not create sphParaController.");
    return NULL;
  }

  return paraCtrl;
}


/**
 * 部分領域データ管理機能の構造格子パラレルコントローラ作成、追加する。（ノードリスト要素）.
 *
 * ノードリスト要素からパラレルコントローラを生成する。
 * SphNode[@group=プロセス番号]に従ってソルバプロセスに割り当てる。
 * SphNode[@group]が存在しない場合は、ソルバプロセスに均等に割り当てる。
 * @param       dims         次元数
 * @param      wsize        ソルバ全体領域のサイズ
 * @param      wstart       ソルバ全体領域の始点インデックス
 * @param       node_list    ノードリスト要素
 * @param       voxelorigin    原点座標
 * @param       voxelpitch     ピッチ
 * @param       comm    コミュニケータ
 * @param       mbx_grouping    部分領域フラグ(true=部分領域データ管理機能)
 * @return      生成パラレルコントローラ
 */
sphParaController*
sphParaManager::createSvParaController(
                            int dims,
                            const size_t wsize[3],
                            const size_t wstart[3],
                            const sphElem* nodelist_elem,
                            const double voxelorigin[3],
                            const double voxelpitch[3],
                            SPL_Comm comm,
                            bool mbx_grouping)
{
  /// パラメータチェック
  if (nodelist_elem == NULL
      || voxelorigin == NULL || voxelpitch == NULL
      || (comm == SPL_COMM_NULL && splIsParallel())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createSvParaController arguments is null.");
    return NULL;
  }

  register int i;
  int ret = SPHERE_SUCCESS;

  /// <SphNode>要素数を取得する。
  int elem_num = nodelist_elem->countNumOfElem(SPH_CFG_NAME_NODE);

  /// 設定済みのソルバグループのプロセス数を取得する。
  int commSize = 0;
  SPL_Comm_size(comm, &commSize);
  if (elem_num < commSize) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "solver process is less than <SphNode> size , solver process=%d, <SphNode> size=%d", commSize, elem_num);
    return NULL;
  }

  /// ノードリストからノードサイズ、始点インデックスを取得する。
  size_t (*node_size)[3] = new size_t[elem_num][3];
  size_t (*node_start)[3] = new size_t[elem_num][3];
  unsigned int *groupid_table = new unsigned int[elem_num];
  memset(node_size, 0x00, elem_num*3*sizeof(size_t));
  memset(node_start, 0x00, elem_num*3*sizeof(size_t));
  memset(groupid_table, 0x00, elem_num*sizeof(unsigned int));
  const sphElem* node_elem = nodelist_elem->getChildElemFirst(SPH_CFG_NAME_NODE);
  bool grouping = false;
  int sub_num = ceil(elem_num/commSize);
  int except_num = sub_num*commSize - elem_num;
  if (node_elem->existAttribute(SPH_CFG_ATTR_GROUPID)) {
    grouping = true;
  }

  int node_id = 0;
  while (node_elem != NULL) {
    int node_id = node_elem->getAttrInt(SPH_CFG_ATTR_NODEID);
    if (node_id >= elem_num) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "nodeID is big in SphNode, SphNodeList count=%d, SphNode[@nodeID]=%d", elem_num, node_id);
      ret = SPHERE_ERROR;
      break;
    }

    // group属性出現の排他的論理和(XOR):group属性があったり、なかったりしてはならない。
    if ( node_elem->existAttribute(SPH_CFG_ATTR_GROUPID) ^ grouping ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "invalid SphNode[@group]");
      ret = SPHERE_ERROR;
      break;
    }

    // SphNode[@group=プロセス番号]はソルバプロセス数以上の場合はエラーとする。
    unsigned int group_id = node_elem->getAttrInt(SPH_CFG_ATTR_GROUPID);
    if (commSize > 1) {
      if (group_id >= commSize) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "group is large in SphNode, comm_size=%d, SphNode[@group]=%d", commSize, group_id);
        ret = SPHERE_ERROR;
        break;
      }
    }
    else {
      // 逐次実行の場合は、グループID(=プロセス番号)は０とする。
      group_id = 0;
    }
    // group_tableに割り当てる。
    if (grouping && mbx_grouping) {
      // @group=プロセス番号による割当
      groupid_table[node_id] = group_id;
    }
    else {
      // 均等割当
      int proc_id = node_id/sub_num;
      if (proc_id < commSize-except_num) {
        groupid_table[node_id] = proc_id;
      }
      else {
        groupid_table[node_id] = proc_id - 1;
      }
    }

    sphElemUtils elemUtils(node_elem);
    // <SphSize>の取得
    const char *attr_size[] = {SPH_CFG_ATTR_IX, SPH_CFG_ATTR_JX, SPH_CFG_ATTR_KX};
    if (elemUtils.getDimsAttributeValues(dims, SPH_CFG_NAME_SIZE, attr_size, node_size[node_id])) {
      // ３次元サイズにする
      for (i=dims; i<3; i++) node_size[node_id][i] = 1;
    }
    else {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "invalid <SphSize> or attributes");
      ret = SPHERE_ERROR;
    }

    // <SphStartIndex>の取得
    const char *attr_idx[] = {SPH_CFG_ATTR_I, SPH_CFG_ATTR_J, SPH_CFG_ATTR_K};
    if (!elemUtils.getDimsAttributeValues(dims, SPH_CFG_NAME_STARTINDEX, attr_idx, node_start[node_id])) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "invalid <SphStartIndex> or attributes");
      ret = SPHERE_ERROR;
    }

    node_elem = nodelist_elem->getChildElemNext(node_elem, SPH_CFG_NAME_NODE);
  }

  sphParaController* paraCtrl = NULL;
  if (ret == SPHERE_SUCCESS) {
    /// パラレルマネージャを介してパラレルコントローラの生成、初期化を行う。
    // 部分領域データ管理機能を利用しない場合は、groupid_tableはNULLを設定する。
    paraCtrl = createSvParaController(
                                elem_num,
                                wsize,
                                wstart,
                                node_size,
                                node_start,
                                voxelorigin,
                                voxelpitch,
                                comm,
                                mbx_grouping ? groupid_table : NULL);
    if (paraCtrl == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSvParaController fault");
      ret = SPHERE_ERROR;
    }
  }

  delete [] node_size; delete [] node_start; delete [] groupid_table;

  if (ret == SPHERE_ERROR) {
    return NULL;
  }

  return paraCtrl;
}


/**
 * 構造格子パラレルコントローラ作成、追加する。（ボクセルサイズ分割）.
 *
 * ボクセルサイズを分割数に応じてノード分割してパラレルコントローラを生成する。
 * ノードをサブブロック分割数に再分割してサブブロックを作成する。
 * サブブロック分割数が１より大きい場合は、部分領域データ管理機能パラレルマネージャを生成する。
 * @param       voxelsize    ボクセルサイズ
 * @param       startindex    始点インデックス
 * @param       voxelorigin    原点座標
 * @param       voxelpitch     ピッチ
 * @param       division      分割数
 * @param       comm    コミュニケータ
 * @param       block_count      サブブロック分割数
 * @return      生成パラレルコントローラ
 */
sphParaController* sphParaManager::createSvParaController(
                          const size_t voxelsize[3],
                          const size_t startindex[3],
                          const double voxelorigin[3],
                          const double voxelpitch[3],
                          unsigned int division[3],
                          SPL_Comm comm,
                          unsigned int block_count)
{

  /// パラメータチェック
  if (voxelsize == NULL || startindex == NULL
      || voxelorigin == NULL || voxelpitch == NULL
      || division == NULL || (comm == SPL_COMM_NULL  && splIsParallel())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createSvParaController arguments is null.");
    return NULL;
  }

  if (voxelsize[0]*voxelsize[1]*voxelsize[2] == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "voxelsize is zero, VoxelSize={%ld,%ld,%ld}.", voxelsize[0], voxelsize[1], voxelsize[2]);
    return NULL;
  }

  // コミュニケーショングループサイズと分割数をチェックする。
  int commSize = 0;
  SPL_Comm_size(comm, &commSize);
  if (division[0]*division[1]*division[2] != 0
      && division[0]*division[1]*division[2] != commSize) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "division is not equal process size, Division={%d,%d,%d}, Process Size=%d.", division[0], division[1], division[2], commSize);
    return NULL;
  }

  /// パラレルコントローラの生成を行う。
  sphSvParaController* paraCtrl = NULL;
  if (block_count <= 1) {
    paraCtrl = new sphSvParaController();
  }
  else {
    // 部分領域データ管理機能パラレルマネージャ生成
    paraCtrl = new sphMultiParaController();
  }


  /// ソルバグループのプロセス数と同じであるので、ソルバのコミュニケータ、グループを設定する。
  paraCtrl->initialize(comm);

  /// 原点座標、ピッチを設定する。
  paraCtrl->setVoxelOrigin(voxelorigin);
  paraCtrl->setVoxelPitch(voxelpitch);

  /// パラレルコントローラにボクセルサイズ、始点インデックス、分割数を設定する。
  if (!paraCtrl->initializeVoxel(voxelsize, startindex, division, block_count)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeVoxel fault");
    delete paraCtrl;
    paraCtrl = NULL;
    return NULL;
  }

  /// パラレルマネージャに追加する。
  int mapKey;
  if ( (mapKey = addParaController(paraCtrl)) == 0 ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "addParaController fault");
    delete paraCtrl;
    paraCtrl = NULL;
    return NULL;
  }

  return paraCtrl;
}


/**
 * 構造格子パラレルコントローラ作成、追加する。（プロセス数）.
 *
 * ボクセルサイズをプロセス数に応じてノード分割してパラレルコントローラを生成する。
 * サブブロック分割数が１より大きい場合は、部分領域データ管理機能パラレルマネージャを生成する。
 * @param       voxelsize    ボクセルサイズ
 * @param       startindex    始点インデックス
 * @param       voxelorigin    原点座標
 * @param       voxelpitch     ピッチ
 * @param       proc_num      プロセス数
 * @param       comm    コミュニケータ
 * @param       block_count      サブブロック分割数
 * @return      生成パラレルコントローラ
 */
sphParaController* sphParaManager::createSvParaController(
                          const size_t voxelsize[3],
                          const size_t startindex[3],
                          const double voxelorigin[3],
                          const double voxelpitch[3],
                          unsigned int proc_num,
                          SPL_Comm comm,
                          unsigned int block_count)
{
  /// パラメータチェック
  if (voxelsize == NULL || startindex == NULL
      || voxelorigin == NULL || voxelpitch == NULL
      || (comm == SPL_COMM_NULL && splIsParallel())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createSvParaController arguments is null.");
    return NULL;
  }

  if (voxelsize[0]*voxelsize[1]*voxelsize[2] == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "voxelsize is zero, VoxelSize={%ld,%ld,%ld}.", voxelsize[0], voxelsize[1], voxelsize[2]);
    return NULL;
  }

  // コミュニケーショングループサイズと分割数をチェックする。
  int commSize = 0;
  SPL_Comm_size(comm, &commSize);
  if (proc_num > commSize) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "a lot of process count[%d] than community count[%d]", proc_num, commSize);
    return NULL;
  }

  /// パラレルコントローラの生成を行う。
  sphSvParaController* paraCtrl = NULL;
  if (block_count <= 1) {
    paraCtrl = new sphSvParaController();
  }
  else {
    // 部分領域データ管理機能パラレルマネージャ生成
    paraCtrl = new sphMultiParaController();
  }

  /// ソルバグループのプロセス数と同じであるので、ソルバのコミュニケータ、グループを設定する。
  paraCtrl->initialize(comm);

  /// 原点座標、ピッチを設定する。
  paraCtrl->setVoxelOrigin(voxelorigin);
  paraCtrl->setVoxelPitch(voxelpitch);

  /// パラレルコントローラにボクセルサイズ、始点インデックス、分割数を設定する。
  if (!paraCtrl->initializeVoxel(voxelsize, startindex, proc_num, block_count)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeVoxel fault");
    delete paraCtrl;
    paraCtrl = NULL;
    return NULL;
  }

  /// パラレルマネージャに追加する。
  int mapKey;
  if ( (mapKey = addParaController(paraCtrl)) == 0 ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "addParaController fault");
    delete paraCtrl;
    paraCtrl = NULL;
    return NULL;
  }

  return paraCtrl;
}


/**
 * 構造格子パラレルコントローラ作成、追加する。（ボクセルサイズリスト）.
 *
 * ボクセルサイズリストからパラレルコントローラを生成する。
 * グループID(=ランク番号)テーブルが設定されている場合は、部分領域データ管理機能パラレルマネージャを生成する。
 * @param       proc_num      プロセス数
 * @param      wsize        ソルバ全体領域のサイズ
 * @param      wstart       ソルバ全体領域の始点インデックス
 * @param       voxelsize    ボクセルサイズリスト
 * @param       voxelorigin    原点座標
 * @param       voxelpitch     ピッチ
 * @param       comm    コミュニケータ
 * @param       groupid_table    グループID(=ランク番号)テーブル
 * @return      生成パラレルコントローラ
 */
sphParaController* sphParaManager::createSvParaController(
                          unsigned int node_num,
                          const size_t wsize[3],
                          const size_t wstart[3],
                          const size_t (*voxelsize)[3],
                          const size_t (*startindex)[3],
                          const double voxelorigin[3],
                          const double voxelpitch[3],
                          SPL_Comm comm,
                          const unsigned int *groupid_table)
{
  /// パラメータチェック
  if (voxelsize == NULL || startindex == NULL || node_num == 0
      || voxelorigin == NULL || voxelpitch == NULL
      || (comm == SPL_COMM_NULL && splIsParallel()) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createSvParaController arguments is null.");
    return NULL;
  }

  /// パラレルコントローラの生成を行う。
  sphSvParaController* paraCtrl = NULL;
  if (groupid_table == NULL) {
    // コミュニケーショングループサイズとノードをチェックする。
    int commSize = 0;
    SPL_Comm_size(comm, &commSize);
    if (node_num != commSize) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "solver process is not equal Node size , solver process=%d, Node size=%d", commSize, node_num);
      return NULL;
    }

    paraCtrl = new sphSvParaController();
  }
  else {
    // 部分領域データ管理機能パラレルマネージャ生成
    paraCtrl = new sphMultiParaController();
  }

  /// ソルバのコミュニケータ、グループを設定する。
  paraCtrl->initialize(comm);

  /// 原点座標、ピッチを設定する。
  paraCtrl->setVoxelOrigin(voxelorigin);
  paraCtrl->setVoxelPitch(voxelpitch);

  /// パラレルコントローラにボクセルサイズ、始点インデックスを設定する。
  if (!paraCtrl->initializeVoxel(node_num, wsize, wstart, voxelsize, startindex, groupid_table)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeVoxel fault");
    if (paraCtrl != NULL) delete paraCtrl;
    paraCtrl = NULL;
    return NULL;
  }

  /// パラレルマネージャに追加する。
  int mapKey;
  if ( (mapKey = addParaController(paraCtrl)) == 0 ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "addParaController fault");
    delete paraCtrl;
    paraCtrl = NULL;
    return NULL;
  }

  return paraCtrl;
}


/**
 * 構造格子パラレルコントローラ作成、追加する。（グループ作成）.
 *
 * ノード番号リストから新しいグループを生成してパラレルコントローラを生成する。
 * old_commに属するすべてのノードにて、この関数を呼び出し新しいグループを生成しなければならない。
 * ノード番号リストに属するノードにて戻り値がNULLの場合は、パラレルコントローラの生成失敗である。
 * ノード番号リストに属しないノードにては、パラレルコントローラの生成は行わないので、戻り値がNULL（正常）である。
 * グループID(=ランク番号)テーブルが設定されている場合は、部分領域データ管理機能パラレルマネージャを生成する。
 * @param       proc_num      プロセス数
 * @param       voxelsize    ボクセルサイズリスト
 * @param       startindex    始点インデックスリスト
 * @param       voxelorigin    原点座標
 * @param       voxelpitch     ピッチ
 * @param       old_comm          旧コミュニケータ
 * @param       procid_list       プロセス番号リスト
 * @param       groupid_table    グループID(=ランク番号)テーブル
 * @return      生成パラレルコントローラ
 */
sphParaController* sphParaManager::createSvParaControllerForNewGroup(
                          unsigned int proc_num,
                          const size_t wsize[3],
                          const size_t wstart[3],
                          const size_t (*voxelsize)[3],
                          const size_t (*startindex)[3],
                          const double voxelorigin[3],
                          const double voxelpitch[3],
                          SPL_Comm old_comm,
                          int* procid_list,
                          const unsigned int *groupid_table)
{
  /// パラメータチェック
  if (voxelsize == NULL || startindex == NULL || proc_num == 0
      || voxelorigin == NULL || voxelpitch == NULL
      || (old_comm == SPL_COMM_NULL  && splIsParallel())
      || procid_list == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createSvParaController arguments is null.");
    return NULL;
  }

  /// パラレルコントローラの生成を行う。
  sphSvParaController* paraCtrl = NULL;
  if (groupid_table == NULL) {
    paraCtrl = new sphSvParaController();
  }
  else {
    // 部分領域データ管理機能パラレルマネージャ生成
    paraCtrl = new sphMultiParaController();
  }

  /// コミュニケータ、グループを作成する。
  SPL_Comm comm = paraCtrl->createProcessGroup(proc_num, procid_list, old_comm);
  if (comm == SPL_COMM_NULL) {
    /// 計算領域ノードリストに含まれないノードでは、COMM_NULLとなるので、エラーではない。
    delete paraCtrl;
    paraCtrl = NULL;
    return NULL;
  }

  /// ソルバのコミュニケータ、グループを設定する。
  paraCtrl->initialize(comm);

  /// 原点座標、ピッチを設定する。
  paraCtrl->setVoxelOrigin(voxelorigin);
  paraCtrl->setVoxelPitch(voxelpitch);

  /// パラレルコントローラにボクセルサイズ、始点インデックスを設定する。
  if (!paraCtrl->initializeVoxel(proc_num, wsize, wstart, voxelsize, startindex, groupid_table)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeVoxel fault");
    if (paraCtrl != NULL) delete paraCtrl;
    paraCtrl = NULL;
    return NULL;
  }

  /// パラレルマネージャに追加する。
  int mapKey;
  if ( (mapKey = addParaController(paraCtrl)) == 0 ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "addParaController fault");
    delete paraCtrl;
    paraCtrl = NULL;
    return NULL;
  }

  return paraCtrl;
}


/**
 * パラレルコントローラを追加する。
 * @param       paraCtrl        パラレルコントローラ
 * @return    success(=true)/fault(=false)
 */
int sphParaManager::addParaController(sphParaController* paraCtrl)
{
  if (paraCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "addParaController arguments is null.");
    return 0;
  }

  /// KEYのインクリメント
  m_mapKey++;
  if (m_mapKey == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "m_mapKey is zero, (limit add map).");
    return 0;
  }

  // パラレルコントローラの追加
  if (!m_mapParaCtrl.insert( std::make_pair(m_mapKey, paraCtrl)).second) {
    m_mapKey--;
     // error
     m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_mapParaCtrl. [m_mapKey=%d].", m_mapKey);
     return 0;
  }

  /// 識別キーを設定する。
  paraCtrl->setParaKey(m_mapKey);

  return m_mapKey;
}


/**
 * パラレルコントローラの登録数を取得する。
 * @param       comm    コミュニケータ
 * @param       paraCtrl        パラレルコントローラ
 * @return    success(=true)/fault(=false)
 */
int sphParaManager::getNumberOfParaController() const
{
  return m_mapParaCtrl.size();
}


/**
 * インデックス番号のパラレルコントローラを取得する。
 * @param       id    インデックス番号
 * @return    パラレルコントローラ
 */
const sphParaController* sphParaManager::getParaControllerByIndex(int id) const
{
  int count = 0;
  std::map<int, sphParaController*>::const_iterator it = m_mapParaCtrl.begin();
  for ( ;it != m_mapParaCtrl.end(); it++) {
    if (count == id) {
      return (*it).second;
    }
    count++;
  }
  return NULL;
}


/**
 * マップキー番号のパラレルコントローラを取得する。
 * @param       map_key    マップキー番号
 * @return    パラレルコントローラ
 */
const sphParaController* sphParaManager::getParaController(int map_key) const
{
  std::map<int, sphParaController*>::const_iterator it = m_mapParaCtrl.find(map_key);
  if (it == m_mapParaCtrl.end ()) {
    return NULL;
  }
  return (*it).second;
}


/**
 * コミュニケータ番号のパラレルコントローラを取得する。
 * @param       comm    コミュニケータ
 * @return    パラレルコントローラ
 */
const sphParaController* sphParaManager::getParaControllerByComm(SPL_Comm comm) const
{
  std::map<int, sphParaController*>::const_iterator it = m_mapParaCtrl.begin();
  for ( ;it != m_mapParaCtrl.end(); it++) {
    const sphParaController* paraCtrl = (*it).second;
    if (paraCtrl->getSplComm() == comm) {
      return paraCtrl;
    }
  }
  return NULL;
}


/**
 * コミュニケータ番号のパラレルコントローラのマップキーを取得する。
 * @param       comm    コミュニケータ
 * @return    マップキー
 */
int sphParaManager::getKeyNumber(SPL_Comm comm) const
{
  std::map<int, sphParaController*>::const_iterator it = m_mapParaCtrl.begin();
  for ( ;it != m_mapParaCtrl.end(); it++) {
    int map_key = (*it).first;
    const sphParaController* paraCtrl = (*it).second;
    if (paraCtrl->getSplComm() == comm) {
      return map_key;
    }
  }
  return 0;
}


/**
 * コミュニケータ番号のパラレルコントローラを削除する。
 * @param       comm    コミュニケータ
 * @return     void
 */
void sphParaManager::removeParaControllerByComm(SPL_Comm comm)
{
  std::map<int, sphParaController*>::iterator it = m_mapParaCtrl.begin();
  for ( ;it != m_mapParaCtrl.end(); it++) {
    sphParaController* paraCtrl = (*it).second;
    if (paraCtrl->getSplComm() == comm) {
      if (paraCtrl != NULL) delete paraCtrl;
      paraCtrl = NULL;
      m_mapParaCtrl.erase(it);
      return;
    }
  }
  return;
}



/**
 * インデックス番号のパラレルコントローラを削除する。
 * @param       index    インデックス番号
 * @return     void
 */
void sphParaManager::removeParaController(int map_key)
{
  std::map<int, sphParaController*>::iterator it = m_mapParaCtrl.begin();
  for ( ;it != m_mapParaCtrl.end(); it++) {
    sphParaController* paraCtrl = (*it).second;
    if ((*it).first == map_key) {
      if (paraCtrl != NULL) delete paraCtrl;
      paraCtrl = NULL;
      m_mapParaCtrl.erase(it);
      return;
    }
  }
  return;
}



/**
 * 構造格子パラレルコントローラ作成、追加する。（ソルバコンフィグレーション要素）.
 * ソルバコンフィグレーション要素からパラレルコントローラを生成する。
 * @param       configElem    ソルバコンフィグレーション要素
 * @param       comm    コミュニケータ
 * @param       mesh    非構造格子データ
 * @return      生成パラレルコントローラ
 */
sphParaController*
sphParaManager::createUnstParaController(
                            const sphElem* solverElem,
                            SPL_Comm comm )
{

  register int i;
  int ret = SPHERE_SUCCESS;
  sphParaController* paraCtrl = NULL;

  if (solverElem == NULL || (comm == SPL_COMM_NULL && splIsParallel())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createSvParaController arguments is null.");
    return NULL;
  }

  /// 起動プロセス数を取得する
  int world_size;
  SPL_Comm_size (SPL_COMM_WORLD, &world_size);

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  /// 次元数の取得を行う。
  int dims = solverElem->getAttrInt(SPH_CFG_ATTR_DIMS);
  if (dims == 0) dims = 3;      // デフォルト＝３
  if (dims < 0 || dims > 3) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid dims(=%d)", dims);
    ret = SPHERE_TERMINATE;
  }

  sphElemUtils elemUtils(solverElem);

  /// <SphSteer>要素の取得を行う。
  const sphElem* steer_elem = solverElem->getElem(SPH_CFG_NAME_STEER);
  if (steer_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found <SphSteer>");
    ret = SPHERE_TERMINATE;
  }

  if ( ret != SPHERE_SUCCESS ) {
    // 警告メッセージを元に戻す。
    m_errHandler->releaseWarning();
    return NULL;
  }

  /// <SphNodeProcess>要素からパラレルコントローラの作成を行う。
  const sphElem* proc_elem = NULL;
  proc_elem = steer_elem->getElem(SPH_CFG_NAME_NODEPROCESS);
  if (proc_elem != NULL) {
    int proc_num =proc_elem->getAttrInt(SPH_CFG_ATTR_VALUE);
    if (world_size == 1) {
      // 単独実行であるので分割数=1
      proc_num = 1;
    }
    /// sphUnstParaControllerを生成する。
    paraCtrl = createUnstParaController(proc_num, comm);
  }

  /// <SphNodeProcess>要素が存在しない場合は、ソルバグループのプロセス数にてパラレルコントローラの作成を行う。
  if (proc_elem == NULL) {
    /// sphUnstParaControllerを生成する。
    int proc_num = 0;
    paraCtrl = createUnstParaController(proc_num, comm);
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  if (paraCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not create sphParaController.");
    return NULL;
  }

  // 非構造格子データのセットアップを行う
  sphUnstParaController* unstParaCtrl = dynamic_cast<sphUnstParaController*>(paraCtrl);
  if( !unstParaCtrl ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not create sphUnstParaController.");
    return NULL;
  }
  int commSize = 0;
  SPL_Comm_size(comm, &commSize);
  if( commSize == 1 ) {          // for serial
    sphUnstMesh::MeshDataType dtype = sphUnstMesh::MESH_DOUBLE;
    char fn_node[MAXPATHLEN]={0}, fn_elem[MAXPATHLEN]={0},
         fn_grp[MAXPATHLEN]={0}, fn_bnd[MAXPATHLEN]={0};

    // parser
    const sphElem* domain_elem = NULL;
    if( !(domain_elem = solverElem->getElem(SPH_CFG_NAME_DOMAININFO)) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found <SphDomainInfo>");
      return NULL;
    }
    const sphElem* fileElem = NULL;
    const sphElem* srcFileElem = NULL;
    const char* flabel = NULL;
    // node.dat
    if( !(fileElem = domain_elem->getElem(2, "SphUnstNode", "SphFile")) ) {
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can't get SphUnstNode::SphFile");
      return NULL;
    }
    sphElemUtils nodeElem(fileElem);
    if( !( srcFileElem = nodeElem.getFileElem() ) ) {
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can't get SphFile");
      return NULL;
    }
    sphElemUtils nodeSrcPath(srcFileElem);
    if( !nodeSrcPath.getFilePath(fn_node, elemUtils.getInputBaseDir()) ) {
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "can't create file name");
      return NULL;
    }

    // elem.dat
    if( !(fileElem = domain_elem->getElem(2, "SphUnstElem", "SphFile")) ) {
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can't get SphUnstElem::SphFile");
      return NULL;
    }
    sphElemUtils elemElem(fileElem);
    if( !( srcFileElem = elemElem.getFileElem() ) ) {
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can't get SphFile");
      return NULL;
    }
    sphElemUtils elemSrcPath(srcFileElem);
    if( !elemSrcPath.getFilePath(fn_elem, elemUtils.getInputBaseDir()) ) {
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "can't create file name");
      return NULL;
    }

    // group.dat
    if( (fileElem = domain_elem->getElem(2, "SphUnstGroup", "SphFile")) ) {
      sphElemUtils grpElem(fileElem);
      if( !( srcFileElem = grpElem.getFileElem() ) ) {
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can't get SphFile");
        return NULL;
      }
      sphElemUtils grpSrcPath(srcFileElem);
      if( !grpSrcPath.getFilePath(fn_grp, elemUtils.getInputBaseDir()) ) {
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "can't create file name");
        return NULL;
      }
    }

    // boundary.dat
    if( (fileElem = domain_elem->getElem(2, "SphUnstBoundary", "SphFile")) ) {
      sphElemUtils bndElem(fileElem);
      if( !( srcFileElem = bndElem.getFileElem() ) ) {
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can't get SphFile");
        return NULL;
      }
      sphElemUtils bndSrcPath(srcFileElem);
      if( !bndSrcPath.getFilePath(fn_bnd, elemUtils.getInputBaseDir()) ) {
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "can't create file name");
        return NULL;
      }
    }

    if( !unstParaCtrl->createMesh(dtype, fn_node, fn_elem, fn_grp, fn_bnd) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT,
                                   "can not create Unstructured-Mesh.");
      return NULL;
    }
  }
  else if( commSize > 1 ) {      // for parallel
    sphUnstMesh::MeshDataType dtype = sphUnstMesh::MESH_DOUBLE;
    char filename[MAXPATHLEN] = {0};

    // parser
    const sphElem* domain_elem = NULL;
    if( !(domain_elem = solverElem->getElem(SPH_CFG_NAME_DOMAININFO)) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found <SphDomainInfo>");
      return NULL;
    }
    const sphElem* fileElem = NULL;
    const sphElem* srcFileElem = NULL;
    const char* flabel = NULL;
    // dist_mesh
    if( !(fileElem = domain_elem->getElem(2, "SphUnstDistMesh", "SphFile")) ) {
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can't get SphUnstDistMesh::SphFile");
      return NULL;
    }
    sphElemUtils distMeshElem(fileElem);
    if( !( srcFileElem = distMeshElem.getFileElem() ) ) {
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can't get SphFile");
      return NULL;
    }
    const char* baseDir = elemUtils.getInputBaseDir();
    if( !sphElemUtils(srcFileElem).getFileFormatPath(filename, baseDir, 0, unstParaCtrl->getMyRank()) ) {
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "can't create file name");
      return NULL;
    }

    if( !unstParaCtrl->createMesh(dtype, filename) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT,
                                   "can not create Unstructured-Mesh.");
      return NULL;
    }
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT,
                                 "can not create Unstructured-Mesh.");
    return NULL;
  }

  return paraCtrl;
}


/**
 * 構造格子パラレルコントローラ作成、追加する。
 * ソルバコンフィグレーション要素からパラレルコントローラを生成する。
 * 非構造格子データはファイルを指定して読み込む（逐次版）
 * @param       configElem    ソルバコンフィグレーション要素
 * @param       comm          コミュニケータ
 * @param       fn_node       節点ファイル
 * @param       fn_elem       要素ファイル
 * @param       fn_grp        グループファイル
 * @param       fn_bnd        境界条件ファイル
 * @param       dtype         精度
 * @return      生成パラレルコントローラ
 */
sphParaController*
sphParaManager::createUnstParaController(
                            const sphElem* solverElem,
                            SPL_Comm comm,
                            const char* fn_node,
                            const char* fn_elem,
                            const char* fn_grp,
                            const char* fn_bnd,
                            sphUnstMesh::MeshDataType dtype )
{

  register int i;
  int ret = SPHERE_SUCCESS;
  sphParaController* paraCtrl = NULL;

  if (solverElem == NULL || (comm == SPL_COMM_NULL && splIsParallel())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createSvParaController arguments is null.");
    return NULL;
  }

  /// 起動プロセス数を取得する
  int world_size;
  SPL_Comm_size (SPL_COMM_WORLD, &world_size);

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  /// 次元数の取得を行う。
  int dims = solverElem->getAttrInt(SPH_CFG_ATTR_DIMS);
  if (dims == 0) dims = 3;      // デフォルト＝３
  if (dims < 0 || dims > 3) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid dims(=%d)", dims);
    ret = SPHERE_TERMINATE;
  }

  sphElemUtils elemUtils(solverElem);

  /// <SphSteer>要素の取得を行う。
  const sphElem* steer_elem = solverElem->getElem(SPH_CFG_NAME_STEER);
  if (steer_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found <SphSteer>");
    ret = SPHERE_TERMINATE;
  }

  if ( ret != SPHERE_SUCCESS ) {
    // 警告メッセージを元に戻す。
    m_errHandler->releaseWarning();
    return NULL;
  }

  /// <SphNodeProcess>要素からパラレルコントローラの作成を行う。
  const sphElem* proc_elem = NULL;
  proc_elem = steer_elem->getElem(SPH_CFG_NAME_NODEPROCESS);
  if (proc_elem != NULL) {
    int proc_num =proc_elem->getAttrInt(SPH_CFG_ATTR_VALUE);
    if (world_size == 1) {
      // 単独実行であるので分割数=1
      proc_num = 1;
    }
    /// sphUnstParaControllerを生成する。
    paraCtrl = createUnstParaController(proc_num, comm);
  }

  /// <SphNodeProcess>要素が存在しない場合は、ソルバグループのプロセス数にてパラレルコントローラの作成を行う。
  if (proc_elem == NULL) {
    /// sphUnstParaControllerを生成する。
    int proc_num = 0;
    paraCtrl = createUnstParaController(proc_num, comm);
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  if (paraCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not create sphParaController.");
    return NULL;
  }

  // 非構造格子データのセットアップを行う
  sphUnstParaController* unstParaCtrl = dynamic_cast<sphUnstParaController*>(paraCtrl);
  if( !unstParaCtrl ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not create sphUnstParaController.");
    return NULL;
  }
  int commSize = 0;
  SPL_Comm_size(comm, &commSize);
  if( commSize == 1 ) {          // for serial
    if( !unstParaCtrl->createMesh(dtype, fn_node, fn_elem, fn_grp, fn_bnd) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT,
                                   "can not create Unstructured-Mesh.");
      return NULL;
    }
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT,
                                 "can not create Unstructured-Mesh.");
    return NULL;
  }

  return paraCtrl;
}


/**
 * 構造格子パラレルコントローラ作成、追加する。（ソルバコンフィグレーション要素）.
 * ソルバコンフィグレーション要素からパラレルコントローラを生成する。
 * 非構造格子データはファイルを指定して読み込む（並列版）
 * @param       configElem    ソルバコンフィグレーション要素
 * @param       comm          コミュニケータ
 * @param       dist_fname    分散非構造格子ファイル
 * @param       dtype         精度
 * @return      生成パラレルコントローラ
 */
sphParaController*
sphParaManager::createUnstParaController(
                            const sphElem* solverElem,
                            SPL_Comm comm,
                            const char* dist_fname,
                            sphUnstMesh::MeshDataType dtype )

{

  register int i;
  int ret = SPHERE_SUCCESS;
  sphParaController* paraCtrl = NULL;

  if (solverElem == NULL || (comm == SPL_COMM_NULL && splIsParallel())) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createUnstParaController arguments is null.");
    return NULL;
  }

  /// 起動プロセス数を取得する
  int world_size;
  SPL_Comm_size (SPL_COMM_WORLD, &world_size);

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  /// 次元数の取得を行う。
  int dims = solverElem->getAttrInt(SPH_CFG_ATTR_DIMS);
  if (dims == 0) dims = 3;      // デフォルト＝３
  if (dims < 0 || dims > 3) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid dims(=%d)", dims);
    ret = SPHERE_TERMINATE;
  }

  sphElemUtils elemUtils(solverElem);

  /// <SphSteer>要素の取得を行う。
  const sphElem* steer_elem = solverElem->getElem(SPH_CFG_NAME_STEER);
  if (steer_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found <SphSteer>");
    ret = SPHERE_TERMINATE;
  }

  if ( ret != SPHERE_SUCCESS ) {
    // 警告メッセージを元に戻す。
    m_errHandler->releaseWarning();
    return NULL;
  }

  /// <SphNodeProcess>要素からパラレルコントローラの作成を行う。
  const sphElem* proc_elem = NULL;
  proc_elem = steer_elem->getElem(SPH_CFG_NAME_NODEPROCESS);
  if (proc_elem != NULL) {
    int proc_num =proc_elem->getAttrInt(SPH_CFG_ATTR_VALUE);
    if (world_size == 1) {
      // 単独実行であるので分割数=1
      proc_num = 1;
    }
    /// sphUnstParaControllerを生成する。
    paraCtrl = createUnstParaController(proc_num, comm);
  }

  /// <SphNodeProcess>要素が存在しない場合は、ソルバグループのプロセス数にてパラレルコントローラの作成を行う。
  if (proc_elem == NULL) {
    /// sphUnstParaControllerを生成する。
    int proc_num = 0;
    paraCtrl = createUnstParaController(proc_num, comm);
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  if (paraCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not create sphParaController.");
    return NULL;
  }

  // 非構造格子データのセットアップを行う
  sphUnstParaController* unstParaCtrl = dynamic_cast<sphUnstParaController*>(paraCtrl);
  if( !unstParaCtrl ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not create sphUnstParaController.");
    return NULL;
  }
  int commSize = 0;
  SPL_Comm_size(comm, &commSize);
  if( commSize > 1 ) {      // for parallel
    if( !unstParaCtrl->createMesh(dtype, dist_fname) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT,
                                   "can not create Unstructured-Mesh.");
      return NULL;
    }
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT,
                                 "can not create Unstructured-Mesh.");
    return NULL;
  }

  return paraCtrl;
}


/**
 * 構造格子パラレルコントローラ作成、追加する。（プロセス数）.
 * ボクセルサイズをプロセス数に応じてノード分割してパラレルコントローラを生成する。
 * @param       proc_num      プロセス数
 * @param       comm    コミュニケータ
 * @param       mesh    非構造格子データ
 * @return      生成パラレルコントローラ
 */
sphParaController* sphParaManager::createUnstParaController(
                          unsigned int proc_num,
                          SPL_Comm comm )
{
  /// パラメータチェック
  if (comm == SPL_COMM_NULL && splIsParallel()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createUnstParaController arguments is null.");
    return NULL;
  }

  // コミュニケーショングループサイズと分割数をチェックする。
  int commSize = 0;
  SPL_Comm_size(comm, &commSize);
  if (proc_num > commSize) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "a lot of process count[%d] than community count[%d]", proc_num, commSize);
    return NULL;
  }

  /// パラレルコントローラの生成を行う。
  sphUnstParaController* paraCtrl = new sphUnstParaController();

  /// ソルバグループのプロセス数と同じであるので、ソルバのコミュニケータ、グループを設定する。
  paraCtrl->initialize(comm);

  /// パラレルマネージャに追加する。
  int mapKey;
  if ( (mapKey = addParaController(paraCtrl)) == 0 ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "addParaController fault");
    delete paraCtrl;
    paraCtrl = NULL;
    return NULL;
  }

  return paraCtrl;
}




/**
 * ParallelTemplate要素が有効であるかチェックする。
 * @param       solverElem    ソルバコンフィグレーション要素
 * @return      true=正常
 */
bool
sphParaManager::validParallelTemplate(const sphElem* solverElem)
{
  /// SphParallelTemplate要素は任意要素であるので、存在しない場合は正常とする。
  if (!existsParallelTemplate(solverElem)) {
    return true;
  }

  // SphMultiplyConnectedPartitioning要素
  const sphElem* mcp_elem = solverElem->getElem(3,
                              SPH_CFG_NAME_STEER,
                              SPH_CFG_NAME_PARALLELTEMPLATE,
                              SPH_CFG_NAME_MULTIPLYPARTITIONING);

  // SphEqualPartitioning要素
  const sphElem* ep_elem = solverElem->getElem(3,
                              SPH_CFG_NAME_STEER,
                              SPH_CFG_NAME_PARALLELTEMPLATE,
                              SPH_CFG_NAME_EQUALPARTITIONING);

  /// SphMultiplyConnectedPartitioning要素とSphEqualPartitioning要素が定義されている場合は、エラーとする。
  if (mcp_elem != NULL && ep_elem != NULL) {
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "duplicate element SphMultiplyConnectedPartitioning and SphEqualPartitioning");
    return false;
  }

  /// SphMultiplyConnectedPartitioning要素もSphEqualPartitioning要素も定義されていない場合は、エラーとする。
  if (mcp_elem == NULL && ep_elem == NULL) {
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "not found element SphMultiplyConnectedPartitioning and SphEqualPartitioning");
    return false;
  }

  return true;
}

/**
 * SphParallelTemplate要素が存在しているかチェックする。
 * @param solverElem    ソルバコンフィグレーション要素
 * @return    true=SphParallelTemplate要素が存在している
 */
bool
sphParaManager::existsParallelTemplate(const sphElem* solverElem)
{
  // SphMultiplyConnectedPartitioning要素
  const sphElem* mcp_elem = solverElem->getElem(2,
                              SPH_CFG_NAME_STEER,
                              SPH_CFG_NAME_PARALLELTEMPLATE);

  return (mcp_elem != NULL);
}

/**
 * ブロックグループ(mbx_grouping)の属性値を取得する。
 *
 * @param solverElem    ソルバコンフィグレーション要素
 * @return    true=ブロックグループ有効
 */
bool
sphParaManager::isMultiplyConnectedPartitioning(const sphElem* solverElem)
{
  // SphMultiplyConnectedPartitioning要素
  const sphElem* mcp_elem = solverElem->getElem(3,
                              SPH_CFG_NAME_STEER,
                              SPH_CFG_NAME_PARALLELTEMPLATE,
                              SPH_CFG_NAME_MULTIPLYPARTITIONING);
  if (mcp_elem == NULL) return false;

  return mcp_elem->isAttrValue(SPH_CFG_ATTR_MBXGROUPING, false);
}

/**
 * number_of_subblockの値を取得する。
 * @param solverElem    ソルバコンフィグレーション要素
 * @return    number_of_subblock値
 */
int
sphParaManager::countEqualPartitioning(const sphElem* solverElem)
{
  // SphEqualPartitioning要素
  const sphElem* ep_elem = solverElem->getElem(3,
                              SPH_CFG_NAME_STEER,
                              SPH_CFG_NAME_PARALLELTEMPLATE,
                              SPH_CFG_NAME_EQUALPARTITIONING);
  if (ep_elem == NULL) return 0;
  int count = ep_elem->getAttrInt(SPH_CFG_ATTR_NUMBEROFSUBBLOCK);
  return count;
}


