/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSvNodeInfo.C
 * Node Information Class Source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphSvNodeInfo.C
 * ノード情報クラスソースファイル
 * @author     hira
 * @date       2009/02/20
 */


#include <math.h>
#include "parallel/sphSvNodeInfo.h"


/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphSvNodeInfo::sphSvNodeInfo()
{
  m_nodeID = SPL_PROC_NULL;
  m_voxMin[0] = m_voxMin[1] = m_voxMin[2] = m_voxMax[0] = m_voxMax[1] = m_voxMax[2] = 0;
  m_sbxNames.clear();
  m_offsetInfo.clear();
  m_voxSize[0] = m_voxSize[1] = m_voxSize[2] = 0;
  register unsigned int i, j, k;
  for(i=0; i<3; i++){
    for(j=0; j<3; j++){
      for(k=0; k<3; k++){
        m_idMtrxNum[i][j][k] = 0;
        m_idMtrx[i][j][k] = NULL;
      }
    }
  }
}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ
 */
sphSvNodeInfo::~sphSvNodeInfo()
{
  m_offsetInfo.clear();
  register unsigned int i, j, k;
  for(i=0; i<3; i++){
    for(j=0; j<3; j++){
      for(k=0; k<3; k++){
        if (m_idMtrx[i][j][k] != NULL) {
          delete m_idMtrx[i][j][k];
        }
      }
    }
  }
}

/**
 * ノードIDを取得する。
 * @return
 */
int sphSvNodeInfo::getNodeID() const
{
  return m_nodeID;
}


/**
 * \~english
 * Get Voxel Size
 * @return        const size_t*        Voxel Size
 *
 * \~japanese
 * ボクセルサイズを取得する。
 * @return        const size_t*        ボクセルサイズ
 */
const size_t* sphSvNodeInfo::getVoxelSize() const
{
  return m_voxSize;
}


/**
 * \~english
 * Get Voxel Start Index
 * @return        const size_t*        Voxel Start Index
 *
 * \~japanese
 * ボクセル始点インデックスを取得する。(ガイドセルは含まない領域）
 * @return        const size_t*        ボクセル始点インデックス
 */
const size_t* sphSvNodeInfo::getVoxelMin() const
{
  return m_voxMin;
}


/**
 * \~english
 * Get Voxel End Index
 * @return        const size_t*        Voxel End Index
 *
 * \~japanese
 * ボクセル終点インデックスを取得する。(ガイドセルは含まない領域）
 * @return        const size_t*        ボクセル終点インデックス
 */
const size_t* sphSvNodeInfo::getVoxelMax() const
{
  return m_voxMax;
}


/**
 * \~english
 * Get Contacted Node Count
 * @return    int      Contacted Node Count
 *
 * \~japanese
 * 接合ノード数を取得する。
 * @return    int      接合ノード数
 */
int sphSvNodeInfo::getOffsetDivNum() const
{
  return m_offsetInfo.size();
}


/**
 * \~english
 * Get Contacted Node List
 * @return    int      Contacted Node List
 *
 * \~japanese
 * 接合ノードリストを取得する。
 * @return    int      接合ノードリスト
 */
int sphSvNodeInfo::getOffsetDivIds(int* nodeids) const
{
  int size = m_offsetInfo.size();
  if (size <= 0) {
    return 0;
  }
  if (nodeids == NULL) {
    nodeids = new int[size];
  }

  OffsetDiv::const_iterator itr = m_offsetInfo.begin();
  OffsetDiv::const_iterator itrEnd = m_offsetInfo.end();
  int i = 0;
  for( ; itr != itrEnd; itr++, i++){
    nodeids[i] = (*itr)->getRankId();
  }

  return size;
}


/**
 * \~english
 * Get Contacted Node Information
 * @param      nodeid      NodeId
 * @return          Contacted Node Information
 *
 * \~japanese
 * 接合ノードリストを取得する。
 * @return        接合ノード情報
 */
const sphOffsetDiv* sphSvNodeInfo::getOffsetDivInfo(const int nodeid) const
{
  if (nodeid < 0) return NULL;
  int size = m_offsetInfo.size();
  if (size <= 0) {
    return NULL;
  }

  OffsetDiv::const_iterator itr = m_offsetInfo.begin();
  OffsetDiv::const_iterator itrEnd = m_offsetInfo.end();
  int i = 0;
  for( ; itr != itrEnd; itr++, i++){
    if ((*itr)->getRankId() == nodeid) {
      return (const sphOffsetDiv*)(*itr);
    }
  }

  return NULL;
}


/**
 * \~english
 * Set Node Indexes
 * @param      nodeid      node id
 * @param      min_i        I start index
 * @param      min_j        J start index
 * @param      min_k        K start index
 * @param      max_i        I end index
 * @param      max_j        J end index
 * @param      max_k        K end index
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ノードインデックスを設定する。
 * @param      nodeid       ノード番号
 * @param      min_i        I方向始点インデックス
 * @param      min_j        J方向始点インデックス
 * @param      min_k        K方向始点インデックス
 * @param      max_i        I方向終点インデックス
 * @param      max_j        J方向終点インデックス
 * @param      max_k        K方向終点インデックス
 * @return    success(=true)/fault(=false)
 */
bool sphSvNodeInfo::setNodeInfo(int nodeid, size_t min_i, size_t min_j, size_t min_k, size_t max_i, size_t max_j, size_t max_k)
{
  // node id
  m_nodeID = nodeid;

  if ( (min_i == 0) && (min_j == 0) && (min_k == 0)
      && (max_i == 0) && (max_j == 0) && (max_k == 0) ) {
    m_voxMin[0] = m_voxMin[1] = m_voxMin[2] = 0;
    m_voxMax[0] = m_voxMax[1] = m_voxMax[2] = 0;
    m_voxSize[0] = m_voxSize[1] = m_voxSize[2] = 0;
    return true;
  }

  // min coordinate
  m_voxMin[0] = (min_i <= max_i) ? min_i : max_i;
  m_voxMin[1] = (min_j <= max_j) ? min_j : max_j;
  m_voxMin[2] = (min_k <= max_k) ? min_k : max_k;

  // max coordinate
  m_voxMax[0] = (min_i > max_i) ? min_i : max_i;
  m_voxMax[1] = (min_j > max_j) ? min_j : max_j;
  m_voxMax[2] = (min_k > max_k) ? min_k : max_k;

  m_voxSize[0] = m_voxMax[0] - m_voxMin[0] + 1;
  m_voxSize[1] = m_voxMax[1] - m_voxMin[1] + 1;
  m_voxSize[2] = m_voxMax[2] - m_voxMin[2] + 1;

  return true;
}


/**
 * \~english
 * add Contacted Node Information
 * @param      offset_node      Contacted Node Information
 * @return      >0    offset div count<br>
 *           =0    not find offset div<br>
 *           <0    error code<br>
 *           =-1    initialize error<br>
 *           =-2    offset error (duplicate node)<br>
 *
 * \~japanese
 * 接合ノード情報を追加する。
 * @param      offset_node      追加接合ノード情報
 * @return      >0    接合ノード数<br>
 *           =0        接合ノードは存在しない。<br>
 *           <0        エラーコード<br>
 *           =-1      初期設定エラー<br>
 *           =-2      重複ノードエラー<br>
 */
int sphSvNodeInfo::addOffsetDiv(sphSvNodeInfo* offset_node)
{
  // check node info
  // nodeid
  if (m_nodeID == SPL_PROC_NULL) {
    return -1;
  }
  // size check
  if (m_voxSize[0]*m_voxSize[1]*m_voxSize[2] == 0) return 0;

  // nodeid
  if (offset_node->m_nodeID == SPL_PROC_NULL) {
    return -1;
  }

  // size check
  if (offset_node->m_voxSize[0]*offset_node->m_voxSize[1]*offset_node->m_voxSize[2] == 0) return 0;


/******************************************************
  // check duplicate Voxel
  // check inside node in src box
  long coord_cnt = 0;
  int chk_box[6];
  for (int i=0; i<3; i++) chk_box[i] = m_voxMin[i];
  for (int i=0; i<3; i++) chk_box[3+i] = m_voxMax[i];
  for (unsigned int k=offset_node->m_voxMin[2]; k<=offset_node->m_voxMax[2]; k++) {
    for (unsigned int j=offset_node->m_voxMin[1]; j<=offset_node->m_voxMax[1]; j++) {
      for (unsigned int i=offset_node->m_voxMin[0]; i<=offset_node->m_voxMax[0]; i++) {
        // only face coords
        if ((i>offset_node->m_voxMin[0] && i<offset_node->m_voxMax[0]) && (j>offset_node->m_voxMin[1] && j<offset_node->m_voxMax[1]) && (k>offset_node->m_voxMin[2] && k<offset_node->m_voxMax[2])) {
          continue;
        }
        if ((i>offset_node->m_voxMin[0] && i<offset_node->m_voxMax[0]) && (j>offset_node->m_voxMin[1] && j<offset_node->m_voxMax[1]) ) {
          continue;
        }
        if ((i>offset_node->m_voxMin[0] && i<offset_node->m_voxMax[0]) && (k>offset_node->m_voxMin[2] && k<offset_node->m_voxMax[2])) {
          continue;
        }
        if ( (j>offset_node->m_voxMin[1] && j<offset_node->m_voxMax[1]) && (k>offset_node->m_voxMin[2] && k<offset_node->m_voxMax[2])) {
          continue;
        }

        // check inside node
        int coord[3] = {i, j, k};
        if (IsInsideCoord(chk_box, coord) ) {
          fprintf(stdout, "Inside coord in box:%d, coord[id:%d]=%d, %d, %d.\n", m_nodeID, offset_node->m_nodeID, coord[0], coord[1], coord[2]);
          return -2;
        }
        coord_cnt++;
      }
    }
  }

  // check inside node in dest box
  coord_cnt = 0;
  for (int i=0; i<3; i++) chk_box[i] = offset_node->m_voxMin[i];
  for (int i=0; i<3; i++) chk_box[3+i] = offset_node->m_voxMax[i];
  for (unsigned int k=m_voxMin[2]; k<=m_voxMax[2]; k++) {
    for (unsigned int j=m_voxMin[1]; j<=m_voxMax[1]; j++) {
      for (unsigned int i=m_voxMin[0]; i<=m_voxMax[0]; i++) {
        // only face coords
        if ((i>m_voxMin[0] && i<m_voxMax[0]) && (j>m_voxMin[1] && j<m_voxMax[1]) && (k>m_voxMin[2] && k<m_voxMax[2])) {
          continue;
        }
        if ((i>m_voxMin[0] && i<m_voxMax[0]) && (j>m_voxMin[1] && j<m_voxMax[1]) ) {
          continue;
        }
        if ((i>m_voxMin[0] && i<m_voxMax[0]) && (k>m_voxMin[2] && k<m_voxMax[2])) {
          continue;
        }
        if ( (j>m_voxMin[1] && j<m_voxMax[1]) && (k>m_voxMin[2] && k<m_voxMax[2])) {
          continue;
        }
        // check inside node
        int coord[3] = {i, j, k};
        if (IsInsideCoord(chk_box, coord) ) {
          fprintf(stdout, "Inside coord in box:%d, coord[id:%d]=%d, %d, %d.\n", offset_node->m_nodeID, m_nodeID, coord[0], coord[1], coord[2]);
          return -2;
        }
        coord_cnt++;
      }
    }
  }
  *******************************************/

  int cnt = 0;
  // search -I, +I, -J, +J, -K, +K face
  for (int kdir=-1; kdir<=1; kdir++) {
    for (int jdir=-1; jdir<=1; jdir++) {
      for (int idir=-1; idir<=1; idir++) {
        if (abs(idir) + abs(jdir) + abs(kdir) != 1) {
          continue;
        }

        // -I face : -1, 0, 0
        // +I face : +1, 0, 0
        // -J face : 0, -1, 0
        // +J face : 0, +1, 0
        // -K face : 0, 0, -1
        // +K face : 0, 0, +1

        size_t face_src[6] = {0, 0, 0, 0, 0, 0};
        size_t face_dest[6] = {0, 0, 0, 0, 0, 0};
        size_t offset_face[6] = {0, 0, 0, 0, 0, 0};

        // get source face :face[6] = {min_i, min_j, min_k, max_i, max_j, max_k}
        if ( !getNodeFace(idir, jdir, kdir, face_src) ) continue;
        // get destination face :face[6] = {min_i, min_j, min_k, max_i, max_j, max_k}
        if ( !offset_node->getNodeFace((-1)*idir, (-1)*jdir, (-1)*kdir, face_dest) ) continue;

        // first check : face_src +1 (-1) = face_dest
        if (idir) {
          if (face_src[0] + idir != face_dest[0]) {
            continue;
          }
        }
        else if (jdir) {
          if (face_src[1] + jdir != face_dest[1]) {
            continue;
          }
        }
        else if (kdir) {
          if (face_src[2] + kdir != face_dest[2]) {
            continue;
          }
        }

        // get offset face
        if ( !getOffsetFace(face_src, face_dest, offset_face) ) continue;

        // Insert vector of offset info
        sphOffsetDiv* offset = new sphOffsetDiv();
        offset->setOffsetSize(&offset_face[0], &offset_face[3]);
        offset->setRankId(offset_node->m_nodeID);

        if (idir == 1) {
          offset->setFaceType(XFACE_PLUS);
        }
        else if (idir == -1) {
          offset->setFaceType(XFACE_MINUS);
        }
        else if (jdir == 1) {
          offset->setFaceType(YFACE_PLUS);
        }
        else if (jdir == -1) {
          offset->setFaceType(YFACE_MINUS);
        }
        else if (kdir == 1) {
          offset->setFaceType(ZFACE_PLUS);
        }
        else if (kdir == -1) {
          offset->setFaceType(ZFACE_MINUS);
        }

        // check inserted offset info
        OffsetDiv::const_iterator itr = m_offsetInfo.begin();
        OffsetDiv::const_iterator itrEnd = m_offsetInfo.end();
        bool find = false;
        for( ; itr != itrEnd; itr++){
          if ((*itr)->getRankId() == offset->getRankId()) {
            find = true;
            break;
          }
        }
        if (!find) {
          m_offsetInfo.push_back( offset);
        }
        cnt++;
      }
    }
  }

  return cnt;

}


/**
 * \~english
 * Get Node start/end index for face
 * @param[in]    idir            I direction
 * @param[in]    jdir            J direction
 * @param[in]    kdir            K direction
 * @param[out]      face            face start/end index[6]
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 指定面の始点／終点インデックスを取得する。
 *
 * ６面のいずれかの指定面の始点／終点インデックスを取得する。
 * @param[in]    idir            I方向 (+1:プラス方向面、-1:マイナス方向面)
 * @param[in]    jdir            J方向 (+1:プラス方向面、-1:マイナス方向面)
 * @param[in]    kdir            K方向 (+1:プラス方向面、-1:マイナス方向面)
 * @param[out]      face      指定面の始点／終点インデックス={min_i,min_j,min_k,max_i,max_j,max_k}
 * @return    success(=true)/fault(=false)
 */
bool sphSvNodeInfo::getNodeFace(int idir, int jdir, int kdir, size_t* face)
{
  if (face == NULL) {
    face = new size_t[6];
  }

  face[0] = m_voxMin[0];
  face[1] = m_voxMin[1];
  face[2] = m_voxMin[2];
  face[3] = m_voxMax[0];
  face[4] = m_voxMax[1];
  face[5] = m_voxMax[2];

  if (idir < 0) {
    face[3] = m_voxMin[0];
    return true;
  }
  else if (idir > 0) {
    face[0] = m_voxMax[0];
    return true;
  }
  else if (jdir < 0) {
    face[4] = m_voxMin[1];
    return true;
  }
  else if (jdir > 0) {
    face[1] = m_voxMax[1];
    return true;
  }
  else if (kdir < 0) {
    face[5] = m_voxMin[2];
    return true;
  }
  else if (kdir > 0) {
    face[2] = m_voxMax[2];
    return true;
  }
  return false;
}


/**
 * \~english
 * Get Offset start/end index for src face and dest face
 * @param[in]    face_src              src face index={min_i,min_j,min_k,max_i,max_j,max_k}
 * @param[in]    face_dest            dest face index={min_i,min_j,min_k,max_i,max_j,max_k}
 * @param[out]      face_offset      offset index={min_i,min_j,min_k,max_i,max_j,max_k}
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ２つの面の接合面を取得する。
 * @param[in]    face_src              元面インデックス={min_i,min_j,min_k,max_i,max_j,max_k}
 * @param[in]    face_dest            相手面インデックス={min_i,min_j,min_k,max_i,max_j,max_k}
 * @param[out]      face_offset          接合面インデックス={min_i,min_j,min_k,max_i,max_j,max_k}
 * @return    success(=true)/fault(=false)
 */
bool sphSvNodeInfo::getOffsetFace(const size_t* face_src, const size_t* face_dest, size_t* face_offset)
{
  if (face_src == NULL) return false;
  if (face_dest == NULL) return false;
  if (face_offset == NULL) {
    face_offset = new size_t[6];
  }

  size_t* offset_min = NULL;
  size_t* offset_max = NULL;
  size_t chk_box_src[6];
  size_t chk_box_dest[6];

  // copy to check box from src
  for (int i=0; i<6; i++) chk_box_src[i] = face_src[i];

  // copy to check box from dest
  for (int i=0; i<6; i++) chk_box_dest[i] = face_dest[i];

  // cross coord(max 64Points)
  size_t cross_coords[64][3];
  memset(cross_coords, 0x00, 64*3*sizeof(size_t));

  int pointcnt = 0;
  size_t diff = 0;
  // check I face
  if ( (face_src[0] == face_src[3]) && (face_dest[0] == face_dest[3]) ) {
    // gap of 2 faces
    diff = face_dest[0] - face_src[0];

    // diff=1 is border on 2 faces
    if (abs(diff) == 1) {

      // move box to same src face
      chk_box_dest[0] = chk_box_dest[0] - diff;
      chk_box_dest[3] = chk_box_dest[3] - diff;
    }
  }

  // check J face
  else if ( (face_src[1] == face_src[4]) && (face_dest[1] == face_dest[4]) ) {
    // gap of 2 faces
    diff = face_dest[1] - face_src[1];

    // diff=1 is border on 2 faces
    if (abs(diff) == 1) {

      // move box to same src face
      chk_box_dest[1] = chk_box_dest[1] - diff;
      chk_box_dest[4] = chk_box_dest[4] - diff;
    }
  }

  // check K face
  else if ( (face_src[2] == face_src[5]) && (face_dest[2] == face_dest[5]) ) {
    // gap of 2 faces
    diff = face_dest[2] - face_src[2];

    // diff=1 is border on 2 faces
    if (abs(diff) == 1) {

      // move box to same src face
      chk_box_dest[2] = chk_box_dest[2] - diff;
      chk_box_dest[5] = chk_box_dest[5] - diff;
    }
  }
  else {
    return false;
  }

  if (abs(diff) != 1) {
    return false;
  }

  // get cross coord
  pointcnt = getCrossCoords(cross_coords, &chk_box_src[0], &chk_box_src[3], &chk_box_dest[0], &chk_box_dest[3]);
  if (pointcnt > 0) {
    for (int i=0; i<pointcnt;i++) {
      if (!isInsideCoord(chk_box_src, cross_coords[i])) {
        continue;
      }

      if (!isInsideCoord(chk_box_dest, cross_coords[i])) {
        continue;
      }

      if (!offset_min) {
        offset_min = new size_t[3];
        for (int ic=0; ic<3; ic++) offset_min[ic] = cross_coords[i][ic];
      }
      if (!offset_max) {
        offset_max = new size_t[3];
        for (int ic=0; ic<3; ic++) offset_max[ic] = cross_coords[i][ic];
      }
      if (offset_min[0]>cross_coords[i][0] || offset_min[1]>cross_coords[i][1] || offset_min[2]>cross_coords[i][2]) {
        for (int ic=0; ic<3; ic++) offset_min[ic] = cross_coords[i][ic];
      }
      else if (offset_max[0]<cross_coords[i][0] || offset_max[1]<cross_coords[i][1] || offset_max[2]<cross_coords[i][2]) {
        for (int ic=0; ic<3; ic++) offset_max[ic] = cross_coords[i][ic];
      }
    }
  }

  if (!offset_min && !offset_max) {
    // get contact coord
    pointcnt = getContactCoords(cross_coords, &chk_box_src[0], &chk_box_src[3], &chk_box_dest[0], &chk_box_dest[3]);
    if (pointcnt == 2) {
      offset_min = new size_t[3];
      offset_max = new size_t[3];
      for (int ic=0; ic<3; ic++) offset_min[ic] = cross_coords[0][ic];
      for (int ic=0; ic<3; ic++) offset_max[ic] = cross_coords[1][ic];
    }
  }

  // get min point and max point
  if (offset_min && offset_max) {
    for (int i=0; i<3; i++) face_offset[i] = offset_min[i];
    for (int i=0; i<3; i++) face_offset[3+i] = offset_max[i];

    if (offset_min) delete[] offset_min;
    if (offset_max) delete[] offset_max;
    return true;
  }

  if (offset_min) delete[] offset_min;
  if (offset_max) delete[] offset_max;

  return false;
}


/**
 * \~english
 * Check Inside point in box.
 * @param      box                  box index={min_i,min_j,min_k,max_i,max_j,max_k}
 * @param      face_dest            point index={i,j,k}
 * @return    inside(=true)/outside(=false)
 *
 * \~japanese
 * ボックスの中に点が存在するかチェックする。
 * @param      box              ボックスインデックス={min_i,min_j,min_k,max_i,max_j,max_k}
 * @param      coord            点インデックス={i,j,k}
 * @return    内側(=true)/外側(=false)
 */
bool sphSvNodeInfo::isInsideCoord(const size_t* box, const size_t* coord)
{
  if (box == NULL) return false;
  if (coord == NULL) return false;

  if ( (coord[0] >= box[0] && coord[1] >= box[1] && coord[2] >= box[2])
    && (coord[0] <= box[3] && coord[1] <= box[4] && coord[2] <= box[5]) ) {
    return true;
  }
  return false;
}


/**
 * \~english
 * Get File Name
 * @param      file_attr        File Attribute Value
 * @return     file name
 *
 * \~japanese
 * SBXファイル名を取得する。
 * @param      file_attr        ファイル属性値
 * @return       ファイル名
 */
const char* sphSvNodeInfo::getSbxFilename(const char* file_attr) const
{

  if (m_sbxNames.size() <= 0) return NULL;

  std::map<std::string, std::string>::const_iterator itr;
  std::map<std::string, std::string>::const_iterator itrEnd;
  if (file_attr == NULL) {
    itr = m_sbxNames.begin();
    return &itr->second[0];
  }
  itr = m_sbxNames.begin();
  itrEnd = m_sbxNames.end();
  for( ; itr != itrEnd; itr++){
    const char* attr = &itr->first[0];
    const char* file = &itr->second[0];
    if( !strcasecmp(attr, file_attr) ) {
      return file;
    }
  }

  return NULL;
}


/**
 * \~english
 * Set File Name
 * @param      sbx_attr        File Attribute Value
 * @param      sbx_file        File Name
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * SBXファイル名を設定する。
 * @param      sbx_attr        ファイル属性値
 * @param      sbx_file        ファイル名
 * @return    success(=true)/fault(=false)
 */
bool sphSvNodeInfo::setSbxFilename(const char* sbx_attr, const char* sbx_file)
{
  m_sbxNames.insert(std::make_pair(sbx_attr, sbx_file));
  return true;
}

/**
 * オブジェクトの文字列表現を返します。(デバッグ用)
 * @return    オブジェクトの文字列表現
 */
std::string sphSvNodeInfo::toString() const
{
  std::string msg = "sphSvNodeInfo class\n";
  char buf[1024] = {0};
  sprintf(buf, "m_nodeID=%d", m_nodeID);
  msg += std::string(buf) + '\n';
  sprintf(buf, "m_groupID=%d", m_groupID);
  msg += std::string(buf) + '\n';
  sprintf(buf, "m_voxMin=%ld, %ld, %ld", m_voxMin[0], m_voxMin[1], m_voxMin[2]);
  msg += std::string(buf) + '\n';
  sprintf(buf, "m_voxMax=%ld, %ld, %ld", m_voxMax[0], m_voxMax[1], m_voxMax[2]);
  msg += std::string(buf) + '\n';
  sprintf(buf, "m_voxSize=%ld, %ld, %ld", m_voxSize[0], m_voxSize[1], m_voxSize[2]);
  msg += std::string(buf) + '\n';

  int cnt = 0;
  OffsetDiv::const_iterator itr = m_offsetInfo.begin();
  OffsetDiv::const_iterator itrEnd = m_offsetInfo.end();
  for( ; itr != itrEnd; itr++){
    if ((*itr)->getRankId() == SPL_PROC_NULL) continue;
    const size_t* offsetMin = (*itr)->getOffsetMin();
    const size_t* offsetMax = (*itr)->getOffsetMax();
    if (offsetMin[0] == 0 && offsetMin[1] == 0 && offsetMin[2] == 0
      && offsetMax[0] == 0 && offsetMax[1] == 0 && offsetMax[2] == 0) continue;

    sprintf(buf, "接合面情報[%d]", cnt);
    msg += std::string(buf) + '\n';
    sprintf(buf, "    隣接相手ノードID=%d", (*itr)->getRankId());
    msg += std::string(buf) + '\n';

    sprintf(buf, "    接合面インデックス 始点[I,J,K]/終点[I,J,K]=%ld, %ld, %ld/%ld, %ld, %ld", offsetMin[0], offsetMin[1], offsetMin[2], offsetMax[0], offsetMax[1], offsetMax[2]);
    msg += std::string(buf) + '\n';
    cnt++;
  }

  return msg;
}


/**
 * \~english
 * get cross point index in 2 box
 * @param[out] cross_coords          cross point indexes list={i,j,k}*n
 * @param[in]    coord_src_min        src start index={i,j,k}
 * @param[in]    coord_src_max        src end index={i,j,k}
 * @param[in]    coord_dest_min        dest start index={i,j,k}
 * @param[in]    coord_dest_max        dest end index={i,j,k}
 * @return    cross point count
 *
 * \~japanese
 * ２つのボクセルの境界線が交わる位置のインデックスを取得する。
 * @param[out] cross_coords          接点インデックスリスト={i,j,k}*n
 * @param[in]    coord_src_min        src始点インデックス={i,j,k}
 * @param[in]    coord_src_max        src終点インデックス={i,j,k}
 * @param[in]    coord_dest_min        dest始点インデックス={i,j,k}
 * @param[in]    coord_dest_max        dest終点インデックス={i,j,k}
 * @return    接点数
 */
int sphSvNodeInfo::getCrossCoords(size_t (*cross_coords)[3], size_t* coord_src_min, size_t* coord_src_max, size_t* coord_dest_min, size_t* coord_dest_max) const
{

  int cnt = 0;
  size_t* x_coords = NULL;
  size_t* y_coords = NULL;
  size_t* z_coords = NULL;
  for (int ix=0; ix<4; ix++) {
    if (ix != 0) {
      if ( coord_src_min[0] == coord_src_max[0]
          && coord_src_min[0] == coord_dest_min[0]
          && coord_src_min[0] == coord_dest_max[0] ) {
        continue;
      }
    }

    x_coords = NULL;
    y_coords = NULL;
    z_coords = NULL;

    if (ix==0) x_coords = coord_src_min;
    if (ix==1) x_coords = coord_src_max;
    if (ix==2) x_coords = coord_dest_min;
    if (ix==3) x_coords = coord_dest_max;

    for (int iy=0; iy<4; iy++) {
      if (iy != 0) {
        if ( coord_src_min[1] == coord_src_max[1]
            && coord_src_min[1] == coord_dest_min[1]
            && coord_src_min[1] == coord_dest_max[1] ) {
          continue;
        }
      }
      if (iy==0) y_coords = coord_src_min;
      if (iy==1) y_coords = coord_src_max;
      if (iy==2) y_coords = coord_dest_min;
      if (iy==3) y_coords = coord_dest_max;

      for (int iz=0; iz<4; iz++) {
        if (iz != 0) {
          if ( coord_src_min[2] == coord_src_max[2]
              && coord_src_min[2] == coord_dest_min[2]
              && coord_src_min[2] == coord_dest_max[2] ) {
            continue;
          }
        }

        if (iz==0) z_coords = coord_src_min;
        if (iz==1) z_coords = coord_src_max;
        if (iz==2) z_coords = coord_dest_min;
        if (iz==3) z_coords = coord_dest_max;

        if (x_coords && y_coords && z_coords) {
          cross_coords[cnt][0] = x_coords[0];
          cross_coords[cnt][1] = y_coords[1];
          cross_coords[cnt][2] = z_coords[2];

          cnt++;
        }
      }
    }
  }

  return cnt;
}


/**
 * \~english
 * get contact point index in 2 face
 * @param[out] cross_coords          cross point indexes list={i,j,k}*2
 * @param[in]    coord_src_min        src start index={i,j,k}
 * @param[in]    coord_src_max        src end index={i,j,k}
 * @param[in]    coord_dest_min        dest start index={i,j,k}
 * @param[in]    coord_dest_max        dest end index={i,j,k}
 * @return    cross point count
 *
 * \~japanese
 * ２つの面が接する面の始点・終点インデックスを取得する。
 * @param[out] cross_coords          接点インデックスリスト={i,j,k}*2
 * @param[in]    coord_src_min        src始点インデックス={i,j,k}
 * @param[in]    coord_src_max        src終点インデックス={i,j,k}
 * @param[in]    coord_dest_min        dest始点インデックス={i,j,k}
 * @param[in]    coord_dest_max        dest終点インデックス={i,j,k}
 * @return    接点数
 */
int sphSvNodeInfo::getContactCoords(size_t (*cross_coords)[3], size_t* coord_src_min, size_t* coord_src_max, size_t* coord_dest_min, size_t* coord_dest_max) const
{

  int cnt = 0;

  int cont_cnt[2][3] = {{0, 0, 0}, {0, 0, 0}};
  for (int i=0; i<3; i++) {
    if ((coord_src_min[i] == coord_src_max[i]) && (coord_dest_min[i] == coord_dest_max[i])) {
      cross_coords[0][i] = cross_coords[1][i] = coord_src_min[i];
      cont_cnt[0][i] = cont_cnt[1][i] = 1;
      continue;
    }
    if (abs(coord_src_min[i] - coord_dest_max[i]) == 1) {
      cross_coords[0][i] = cross_coords[1][i] = coord_src_min[i];
      cont_cnt[0][i] = cont_cnt[1][i] = 1;
      continue;
    }
    if (abs(coord_src_max[i] - coord_dest_min[i]) == 1) {
      cross_coords[0][i] = cross_coords[1][i] = coord_src_max[i];
      cont_cnt[0][i] = cont_cnt[1][i] = 1;
      continue;
    }
    if (abs(coord_src_min[i] - coord_dest_max[i]) == 1) {
      cross_coords[0][i] = cross_coords[1][i] = coord_src_min[i];
      cont_cnt[0][i] = cont_cnt[1][i] = 1;
      continue;
    }
    if (abs(coord_src_max[i] - coord_dest_min[i]) == 1) {
      cross_coords[0][i] = cross_coords[1][i] = coord_src_max[i];
      cont_cnt[0][i] = cont_cnt[1][i] = 1;
      continue;
    }

    if (coord_src_min[i] >= coord_dest_min[i] && coord_src_min[i] <= coord_dest_max[i]) {
      cross_coords[0][i] = coord_src_min[i];
      cont_cnt[0][i] = 1;
    }
    if (coord_src_max[i] >= coord_dest_min[i] && coord_src_max[i] <= coord_dest_max[i]) {
      cross_coords[1][i] = coord_src_max[i];
      cont_cnt[1][i] = 1;
    }
    if (coord_dest_min[i] >= coord_src_min[i] && coord_dest_min[i] <= coord_src_max[i]) {
      cross_coords[0][i] = coord_dest_min[i];
      cont_cnt[0][i] = 1;
    }
    if (coord_dest_max[i] >= coord_dest_min[i] && coord_dest_max[i] <= coord_src_max[i]) {
      cross_coords[1][i] = coord_dest_max[i];
      cont_cnt[1][i] = 1;
    }
  }

  int set_check = 0;
  for (int i=0; i<2; i++) {
    for (int j=0; j<3; j++) {
      if (cont_cnt[i][j] == 1) {
        set_check++;
      }
    }
  }

  if (set_check == 6) {
    cnt = 2;
  }

  return cnt;
}



/**
 * \~english
 * Get contact node count
 * @param      i            I direction
 * @param      j            J direction
 * @param      k            K direction
 * @return    contact node count
 *
 * \~japanese
 * 指定面の接合ノード数を取得する。
 * @param      i            I方向 (+1:プラス方向面、-1:マイナス方向面)
 * @param      j            J方向 (+1:プラス方向面、-1:マイナス方向面)
 * @param      k            K方向 (+1:プラス方向面、-1:マイナス方向面)
 * @return    接合ノード数
 */
int sphSvNodeInfo::getCommIDNum(int i, int j, int k) const
{
  if( ((i<-1)||(i>1)) || ((j<-1)||(j>1)) || ((k<-1)||(k>1)) ) return -1;
  return m_idMtrxNum[i+1][j+1][k+1];
}



/**
 * \~english
 * Get contact nodeId list
 * @param      i            I direction
 * @param      j            J direction
 * @param      k            K direction
 * @return    contact nodeId List
 *
 * \~japanese
 * 指定面の接合ノード番号リストを取得する。
 * @param      i            I方向 (+1:プラス方向面、-1:マイナス方向面)
 * @param      j            J方向 (+1:プラス方向面、-1:マイナス方向面)
 * @param      k            K方向 (+1:プラス方向面、-1:マイナス方向面)
 * @return    接合ノードリスト
 */
const int* sphSvNodeInfo::getCommIDs(int i, int j, int k) const
{
  if( ((i<-1)||(i>1)) || ((j<-1)||(j>1)) || ((k<-1)||(k>1)) ) return NULL;
  return m_idMtrx[1+i][1+j][1+k];
}



/**
 * \~english
 * create contact node matrix
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 接合ノードリストを作成する。
 *
 * ボクセルの面別に接合ノード数、接合ノードリストを作成する。
 * @return    success(=true)/fault(=false)
 */
bool sphSvNodeInfo::createIdMatrix()
{
  int dirMtrx[6][3] = { {-1, 0, 0}, {1, 0, 0},
                        {0, -1, 0}, {0, 1, 0},
                        {0, 0, -1}, {0, 0, 1} };

  register int dir;
  for(int dir=0; dir<6; dir++) {
    int i = dirMtrx[dir][0];
    int j = dirMtrx[dir][1];
    int k = dirMtrx[dir][2];

    std::vector<int> offsetids;
    offsetids.clear();

    OffsetDiv::const_iterator itr = m_offsetInfo.begin();
    OffsetDiv::const_iterator itrEnd = m_offsetInfo.end();

    for( ; itr != itrEnd; itr++){
      sphOffsetDiv* offsetDiv = *itr;

      if (offsetDiv->getRankId() < 0) continue;
      const size_t* offsetMin = offsetDiv->getOffsetMin();
      const size_t* offsetMax = offsetDiv->getOffsetMax();
      if (offsetMin[0] == 0 && offsetMin[1] == 0 && offsetMin[2] == 0
        && offsetMax[0] == 0 && offsetMax[1] == 0 && offsetMax[2] == 0) {
        continue;
      }

      int face = -1;
      int dir = -1;
      if (i == -1 || i == 1) {
        face = 0;
        dir = i;
      }
      else if (j == -1 || j == 1) {
        face = 1;
        dir = j;
      }
      else if (k == -1 || k == 1) {
        face = 2;
        dir = k;
      }

      size_t min_face = offsetMin[face];
      size_t max_face = offsetMax[face];
      if (min_face != max_face) {
        continue;
      }

      if (dir == -1) {
        if (m_voxMin[face] != min_face) {
          continue;
        }
      }
      else if (dir == 1) {
        if (m_voxMax[face] != min_face) {
          continue;
        }
      }
      else {
        continue;
      }

      // add list
      offsetids.push_back(offsetDiv->getRankId());
    }

    // set matrix
    if (offsetids.size() > 0) {
      int* list = new int[offsetids.size()];
      for (register int id=0; id<offsetids.size();id++) {
        list[id] = offsetids.at(id);
      }
      m_idMtrx[i+1][j+1][k+1] = list;
    }
    else {
      m_idMtrx[i+1][j+1][k+1] = NULL;
    }
    m_idMtrxNum[i+1][j+1][k+1] = offsetids.size();
  }

  return true;
}


/**
 * グループID(=ランクID)を取得する。
 *
 * 部分領域データ管理機能用
 * @return      グループID(=ランクID)
 */
int sphSvNodeInfo::getGroupID() const
{
  return m_groupID;
}

/**
 * グループID(=ランクID)を設定する。
 *
 * 部分領域データ管理機能用
 * @param groupid     グループID(=ランクID)
 */
void sphSvNodeInfo::setGroupID(int groupid)
{
  m_groupID = groupid;
}

