/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSvParaController.C
 * Simple Voxel Type Parallel Controller Source File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file sphSvParaController.C
 * 構造格子パラレルコントローラーソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#include <set>
#include "parallel/sphSvParaController.h"


/**
 * コンストラクタ
 * @param      paraKey        パラレルコントローラー識別キー
 */
sphSvParaController::sphSvParaController(unsigned int paraKey) : sphParaController(paraKey)
{
  m_nodeInfo = new MapNodeInfo;
  m_voxSize[0] = m_voxSize[1] = m_voxSize[2] = 0;
  m_nodeInfo->clear();
  m_voxMinIdx[0] = m_voxMinIdx[1] = m_voxMinIdx[2] = 0;
  m_voxelOrigin[0] = m_voxelOrigin[1] = m_voxelOrigin[2] = 0.0;
  m_voxelPitch[0] = m_voxelPitch[1] = m_voxelPitch[2] = 0.0;
  m_divInfo[0] = m_divInfo[1] = m_divInfo[2] = 0;
  m_divLocalInfo[0] = m_divLocalInfo[1] = m_divLocalInfo[2] = 0;
  m_parentCtrl = NULL;
}


/**
 * コンストラクタ(サブボックス)
 *
 * 部分データ管理機能(sphMultiParaController)の子パラレルコントローラの生成用のコンストラクタである。
 * @param      parent        親パラレルコントローラ
 * @param      myId          サブボックスID(0〜サブボックス総数-1)
 */
sphSvParaController::sphSvParaController(const sphParaController* parent, int myId) : sphParaController()
{
  m_nodeInfo = new MapNodeInfo;
  m_voxSize[0] = m_voxSize[1] = m_voxSize[2] = 0;
  m_nodeInfo->clear();
  m_voxMinIdx[0] = m_voxMinIdx[1] = m_voxMinIdx[2] = 0;
  m_voxelOrigin[0] = m_voxelOrigin[1] = m_voxelOrigin[2] = 0.0;
  m_voxelPitch[0] = m_voxelPitch[1] = m_voxelPitch[2] = 0.0;
  m_divInfo[0] = m_divInfo[1] = m_divInfo[2] = 0;
  m_parentCtrl = parent;
  m_myID = myId;

}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ
 */
sphSvParaController::~sphSvParaController()
{
/****************
  // node info map
  MapNodeInfo::iterator itr = m_nodeInfo->begin();
  for( ; itr != m_nodeInfo->end(); itr++){
    // node info
    sphSvNodeInfo* node_info = itr->second;
    if (node_info != NULL) delete node_info;
    node_info = NULL;
  }
  m_nodeInfo->clear();
*****************/
  if (!m_nodeEvMap) {
    for(int i=0; i < m_divInfo[0]; i++) {
        for(int j=0; j < m_divInfo[1]; j++) {
            delete[] m_nodeEvMap[i][j];
        }
        delete[] m_nodeEvMap[i];
    }
    delete[] m_nodeEvMap;
  }
}

/**
 * 計算領域サイズ、始点インデックス、分割情報からボクセル情報の初期化を行う。
 * @param[in]       wsize   計算領域サイズ
 * @param[in]       wsta    グローバル始点インデックス
 * @param[in]       proc_num    分割数
 * @param[in]       localblock_count     ローカルブロック数(部分領域データ管理用)
 * @return    success(=true)/fault(=false)
 */
bool sphSvParaController::initializeVoxel(
                              const size_t wsize[3],
                              const size_t wsta[3],
                              unsigned int proc_num,
                              unsigned int localblock_count)
{
  if ( (m_splComm == SPL_COMM_NULL || m_splGroup == SPL_GROUP_NULL) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "communicator or group is null.");
    return false;
  }

  register int i, n;
  // get my node(rank) number from mpi
  int myID = this->getMyRank();

  // get node(rank) count from mpi
  if (proc_num == 0) {
    proc_num = this->getNumberOfGroup();
  }

  unsigned int dist_div[3] = {0, 0, 0};
  /// プロセス数からボクセルサイズを自動分割する
  if( !decideDivPattern(proc_num, wsize, dist_div) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "decideDivPattern fault");
    return false;
  }

  return initializeVoxel(wsize, wsta, dist_div);
}


/**
 * 計算領域サイズ、始点インデックス、分割情報からボクセル情報の初期化を行う。
 * @param[in]       wsize   計算領域サイズ
 * @param[in]       wsta    グローバル始点インデックス
 * @param[in/out]       divNode    分割情報
 * @param[in]       localblock_count     ローカルブロック数(部分領域データ管理用)
 * @return    success(=true)/fault(=false)
 */
bool sphSvParaController::initializeVoxel(
                              const size_t wsize[3],
                              const size_t wsta[3],
                              unsigned int divNode[3],
                              unsigned int localblock_count)
{
  if ( (m_splComm == SPL_COMM_NULL || m_splGroup == SPL_GROUP_NULL) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "communicator or group is null.");
    return false;
  }

  register int i, n;
  // get my node(rank) number from mpi
  int myID = this->getMyRank();

  // get node(rank) count from mpi
  int nodeCnt = this->getNumberOfGroup();

  unsigned int dist_div[3] = {divNode[0], divNode[1], divNode[2]};

  if( dist_div[0]*dist_div[1]*dist_div[2] == 0 ) {
    /// プロセス数からボクセルサイズを自動分割する
    if( !decideDivPattern(nodeCnt, wsize, dist_div) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "decideDivPattern fault");
      return false;
    }
  }
  if( dist_div[0]*dist_div[1]*dist_div[2] != nodeCnt ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "Invalid division values, [%d * %d * %d != %d]", dist_div[0], dist_div[1], dist_div[2], nodeCnt);
    return false;
  }

  /// 計算領域ボクセルサイズを分割数に従ってノード別に分割し、ノードサイズ、始点インデックスを取得する。
  size_t (*distSize)[3] = new size_t[nodeCnt][3];
  size_t (*distStart)[3] = new size_t[nodeCnt][3];
  if (!createVoxelDistribution(wsize, dist_div, distSize, distStart)) {
    delete [] distSize; delete [] distStart;
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "decideDivPattern fault");
    return false;
  }

  /// 始点インデックスに引数で渡されたグローバル始点インデックスを加算する。
  for (n=0; n<nodeCnt; n++) {
    for (i=0; i<3; i++) distStart[n][i] += wsta[i];
  }

  // ノード位置、サイズ情報リストからボクセル情報の初期化を行う。
  if (!this->initializeVoxel(nodeCnt, wsize, wsta, distSize, distStart)) {
    delete [] distSize; delete [] distStart;
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeVoxel fault");
    return false;
  }

  // 分割情報を格納する。
  for (i=0; i<3; i++) m_divInfo[i] = dist_div[i];

  // ノードマッピングテーブルを作成する。
  if (!this->createNodeMap()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createNodeMap fault");
    return false;
  }

  delete [] distSize; delete [] distStart;
  return true;
}


/**
 * 自ノード位置、サイズ情報からボクセル情報の初期化を行う。
 *
 * 引数に設定された自ノード情報から、初期化を行う。<br>
 * 他のノード情報は収集を行い、隣接面情報を作成する。
 * @param      wsize        ソルバ全体領域のサイズ
 * @param      wstart       ソルバ全体領域の始点インデックス
 * @param      mysize       自ノードの[I,J,K]I方向サイズ
 * @param      mysta        自ノードの[I,J,K]方向グローバル始点インデックス
 * @return    success(=true)/fault(=false)
 */
bool sphSvParaController::initializeVoxel(
                            const size_t wsize[3],
                            const size_t wstart[3],
                            const size_t mysize[3],
                            const size_t mysta[3])
{

  // get my node(rank) number from mpi
  int myID = this->getMyRank();

  // get node(rank) count from mpi
  int nodeCnt = this->getNumberOfGroup();

  // create node list
  int* nodes = new int[nodeCnt];

  // node number gather
  int node_data = myID;
  if (!this->allgather(&node_data, 1, SPL_INT, nodes, 1, SPL_INT)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "can not get nodes id list by allgather.");
    delete [] nodes;
    this->abort(-1);
    return false;
  }

  // node indexes gather
  long long* idx_data = new long long[6];
  idx_data[0] = mysta[0];
  idx_data[1] = mysta[1];
  idx_data[2] = mysta[2];
  idx_data[3] = mysize[0];
  idx_data[4] = mysize[1];
  idx_data[5] = mysize[2];
  long long* idx_recv = new long long[6*nodeCnt];
  if (!this->allgather(idx_data, 6, SPL_LONG_LONG, idx_recv, 6, SPL_LONG)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "can not get nodes id list by allgather.");
    delete [] nodes;
    delete [] idx_data;
    delete [] idx_recv;
    this->abort(-1);
    return false;
  }


  size_t (*pNode_sta)[3] = new size_t[nodeCnt][3];
  size_t (*pNode_len)[3] = new size_t[nodeCnt][3];
  for (int i=0; i<nodeCnt; i++) {
    for (int j=0; j<3; j++) {
      pNode_sta[i][j] = idx_recv[i*6+j];
      pNode_len[i][j] = idx_recv[i*6+j+3];
    }
  }

  bool ret = initializeVoxel(nodeCnt, wsize, wstart, pNode_len, pNode_sta);

  delete [] nodes;
  delete [] idx_data;
  delete [] idx_recv;
  delete [] pNode_sta;
  delete [] pNode_len;

  if (!ret) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "initializeVoxel error.");
    return false;
  }

  return true;
}


/**
 * ノード位置、サイズ情報リストからボクセル情報の初期化を行う。
 *
 * 引数に設定されたすべてのノード情報リストから、初期化を行う。<br>
 * 他のノード情報は収集を行わず、引数のノード情報リストから隣接面情報を作成する。
 * @param      nodeCnt      ノード情報リスト数
 * @param      wsize        ソルバ全体領域のサイズ
 * @param      wstart       ソルバ全体領域の始点インデックス
 * @param      node_size        ノードサイズリスト   (nodeCnt x [ilen, jlen, klen])
 * @param      node_sta        グローバル始点インデックスリスト(nodeCnt x [iwsta, jwsta, kwsta])
 * @param      groupid_table    グループID(=ランクID)
 * @return    success(=true)/fault(=false)
 */
bool sphSvParaController::initializeVoxel(
                            unsigned int nodeCnt,
                            const size_t wsize[3],
                            const size_t wstart[3],
                            const size_t (*node_size)[3],
                            const size_t (*node_sta)[3],
                            const unsigned int *groupid_table)
{
  if (!wsize || !wstart || !node_sta || !node_size) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "node_size or node_sta is null.");
    return false;
  }

  int myRankId = this->getMyRank();

  // add node info
  register int i, dim;
  size_t voxMin[3] = {(size_t)-1, (size_t)-1, (size_t)-1};
  size_t voxMax[3] = {0, 0, 0};
  for (i=0; i<nodeCnt; i++) {
    if (node_size[i][0]*node_size[i][1]*node_size[i][2] == 0) continue;

    // グループID(=ランクID)を設定する。
    unsigned int groupid = 0;
    if (groupid_table == NULL) groupid = i;
    else groupid = groupid_table[i];

    // ノード情報を追加する。
    if (!addNodeInfo(i, node_sta[i][0], node_sta[i][1], node_sta[i][2],
                        node_sta[i][0]+node_size[i][0]-1,
                        node_sta[i][1]+node_size[i][1]-1,
                        node_sta[i][2]+node_size[i][2]-1,
                        groupid)) {

      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "addNodeInfo error, nodeid=%d.", i);
      return false;
    }

    // 始点、終点インデックスを取得する。
    if (groupid_table == NULL || groupid == myRankId) {
      for (dim=0; dim<3; dim++) {
        if (voxMin[dim] > node_sta[i][dim]) voxMin[dim] = node_sta[i][dim];
        if (voxMax[dim] < node_sta[i][dim]+node_size[i][dim]-1) voxMax[dim] = node_sta[i][dim]+node_size[i][dim]-1;
      }
    }
  }

  // 始点、終点インデックスをソルバ全体領域の始点、サイズのチェックを行う。
  for (dim=0; dim<3; dim++) {
    if (voxMin[dim] < wstart[dim]) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "node start index[=%ld,%ld,%ld] error. [<whole start index=%ld,%ld,%ld]", voxMin[0],voxMin[1],voxMin[2], wstart[0],wstart[1],wstart[2]);
      return false;
    }
    if ( voxMax[dim] > wstart[dim] + wsize[dim] - 1 ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "node max index[=%ld,%ld,%ld] error. [>whole max index=%ld,%ld,%ld]", voxMax[0],voxMax[1],voxMax[2],wsize[0]+wstart[0]-1,wsize[1]+wstart[1]-1,wsize[2]+wstart[2]-1);
      return false;
    }
    // set whole size and start index
    m_voxMinIdx[dim] = wstart[dim];
    m_voxSize[dim] = wsize[dim];
  }

  // チェックノード数
  if (getNumberOfNode() <= 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "node count is zero ");
    return false;
  }

  // create adjacent node information
  if (!makeNodeOffset()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INITIALIZE, "makeNodeOffset error.");
    return false;
  }

  return true;
}


/**
 * \~english
 * Get Whole Voxel Size in process group
 * @return    Whole Voxel Size {i, j, k}
 *
 * \~japanese
 * プロセスグループ内のすべてのノードを結合したサイズを取得する。
 * @return    ボクセルサイズ {i, j, k}
 */
const size_t* sphSvParaController::getWholeVoxelSize() const
{
  return m_voxSize;
}

/**
 * プロセスグループ内のすべてのノードを結合したサイズを設定する。
 * @param     size    ボクセルサイズ {i, j, k}
 */
void sphSvParaController::setWholeVoxelSize(const size_t size[3])
{
  register int i;
  for (i=0; i< 3; i++) m_voxSize[i] = size[i];
}


/**
 * \~english
 * Get Voxel Head Index in node
 * @param      nodeId        nodeid
 * @param      dirId          direction(I=0, J=1, K=2)
 * @return    Voxel Head Index
 *
 * \~japanese
 * 始点インデックスを取得する。
 * @param      nodeId        ノード番号
 * @param      dirId          取得方向(I=0, J=1, K=2)
 * @return    始点インデックス
 */
size_t sphSvParaController::getVoxelHeadIndex(int nodeId,
                                      unsigned int dirId) const
{
  if (nodeId < 0 || dirId > 2) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "invalid nodeId or dirId, nodeId=%d, dirId=%d.", nodeId, dirId);
    return 0;
  }

  const sphSvNodeInfo* nodeinfo = getNodeInfo(nodeId);
  if (!nodeinfo) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
    return 0;
  }
  const size_t* voxel_min = nodeinfo->getVoxelMin();
  if (!voxel_min) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "voxel_min is null");
    return 0;
  }

  return voxel_min[dirId];

}


/**
 * \~english
 * Get Voxel Tail Index in node
 * @param      nodeId        nodeid
 * @param      dirId          direction(I=0, J=1, K=2)
 * @return    Voxel Tail Index
 *
 * \~japanese
 * 終点インデックスを取得する。
 * @param      nodeId        ノード番号
 * @param      dirId          取得方向(I=0, J=1, K=2)
 * @return    終点インデックス
 */
size_t sphSvParaController::getVoxelTailIndex(int nodeId,
                                          unsigned int dirId) const
{
  if (nodeId < 0 || dirId > 2) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "nodeId=%d, dirId=%d", nodeId, dirId);
    return 0;
  }

  const sphSvNodeInfo* nodeinfo = getNodeInfo(nodeId);
  if (!nodeinfo) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
    return 0;
  }
  const size_t* voxel_max = nodeinfo->getVoxelMax();
  if (!voxel_max) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "voxel_max is null");
    return 0;
  }

  return voxel_max[dirId];

}


/**
 * \~english
 * Get Voxel Size of node
 * @param      nodeId      node id (not parameter=my node)
 * @return    node size {i, j, k}<br>
 *             NULL = not exist node. or node size is zero.
 *
 * \~japanese
 * ノードサイズを取得する。
 * @param      nodeId      ノード番号（指定しない場合：自ノード）
 * @return    ノードサイズ {i, j, k}
 *             NULL：ノードが存在しない。又はノードサイズが０。
 */
const size_t* sphSvParaController::getVoxelSize(int nodeId) const
{
  const sphSvNodeInfo* nodeinfo = getNodeInfo(nodeId);
  if (!nodeinfo) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
    return NULL;
  }
  const size_t* size = nodeinfo->getVoxelSize();
  if (!size) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "size is null");
    return NULL;
  }
  if (size[0]*size[1]*size[2] == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "size is zero");
    return NULL;
  }

  return size;
}


/**
 * \~english
 * Get adjacent node count in one direction
 *
 * Parameter<br>
 * - i, j, k = -1,  0,  0 : -I Direction Face
 * - i, j, k =  1,  0,  0 : +I Direction Face
 * - i, j, k =  0, -1,  0 : -J Direction Face
 * - i, j, k =  0,  1,  0 : +J Direction Face
 * - i, j, k =  0,  0, -1 : -K Direction Face
 * - i, j, k =  0,  0,  1 : +K Direction Face
 *
 * @param      i      I Direction (-1, 0, +1)
 * @param      j      J Direction (-1, 0, +1)
 * @param      k      K Direction (-1, 0, +1)
 * @return    adjacent node count
 *
 * \~japanese
 * 指定方向のノード面に隣接するノード数を取得する。
 *
 * パラメータ<br>
 * - i, j, k = -1,  0,  0 : -I 方向面
 * - i, j, k =  1,  0,  0 : +I 方向面
 * - i, j, k =  0, -1,  0 : -J 方向面
 * - i, j, k =  0,  1,  0 : +J 方向面
 * - i, j, k =  0,  0, -1 : -K 方向面
 * - i, j, k =  0,  0,  1 : +K 方向面
 *
 * @param      i      I 方向 (-1, 0, +1)
 * @param      j      J 方向 (-1, 0, +1)
 * @param      k      K 方向 (-1, 0, +1)
 * @return    隣接するノード数
 */
int sphSvParaController::getCommIDNum (int i, int j, int k) const
{
  if( ((i<-1)||(i>1)) || ((j<-1)||(j>1)) || ((k<-1)||(k>1)) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "-1>[i,j,k=%d,%d,%d]>+1", i, j, k);
    return -1;
  }

  if (i+j+k != -1 && i+j+k != 1) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "[i+j+k=%d,%d,%d] != -1, +1", i, j, k);
    return -1;
  }

  const sphSvNodeInfo* nodeinfo = getNodeInfo();
  if (!nodeinfo) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
    return -1;
  }

  return nodeinfo->getCommIDNum(i, j, k);
}


/**
 * \~english
 * Get adjacent nodes list in one direction
 *
 * Parameter<br>
 * - i, j, k = -1,  0,  0 : -I Direction Face
 * - i, j, k =  1,  0,  0 : +I Direction Face
 * - i, j, k =  0, -1,  0 : -J Direction Face
 * - i, j, k =  0,  1,  0 : +J Direction Face
 * - i, j, k =  0,  0, -1 : -K Direction Face
 * - i, j, k =  0,  0,  1 : +K Direction Face
 *
 * @param      i      I Direction (-1, 0, +1)
 * @param      j      J Direction (-1, 0, +1)
 * @param      k      K Direction (-1, 0, +1)
 * @return    adjacent node list
 *
 * \~japanese
 * 指定方向のノード面に隣接するノードリストを取得する。
 *
 * パラメータ<br>
 * - i, j, k = -1,  0,  0 : -I 方向面
 * - i, j, k =  1,  0,  0 : +I 方向面
 * - i, j, k =  0, -1,  0 : -J 方向面
 * - i, j, k =  0,  1,  0 : +J 方向面
 * - i, j, k =  0,  0, -1 : -K 方向面
 * - i, j, k =  0,  0,  1 : +K 方向面
 *
 * @param      i      I 方向 (-1, 0, +1)
 * @param      j      J 方向 (-1, 0, +1)
 * @param      k      K 方向 (-1, 0, +1)
 * @return    隣接するノードリスト
 */
const int* sphSvParaController::getCommIDs (int i, int j, int k) const
{
  if( ((i<-1)||(i>1)) || ((j<-1)||(j>1)) || ((k<-1)||(k>1)) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "-1>[i,j,k=%d,%d,%d]>+1", i, j, k);
    return NULL;
  }

  if (i+j+k != -1 && i+j+k != 1) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "[i+j+k=%d,%d,%d] != -1, +1", i, j, k);
    return NULL;
  }

  const sphSvNodeInfo* nodeinfo = getNodeInfo();
  if (!nodeinfo) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
    return NULL;
  }

  return nodeinfo->getCommIDs(i, j, k);
}


/**
 * \~english
 * Get node face status in one direction
 *
 * Parameter<br>
 * - i, j, k = -1,  0,  0 : -I Direction Face
 * - i, j, k =  1,  0,  0 : +I Direction Face
 * - i, j, k =  0, -1,  0 : -J Direction Face
 * - i, j, k =  0,  1,  0 : +J Direction Face
 * - i, j, k =  0,  0, -1 : -K Direction Face
 * - i, j, k =  0,  0,  1 : +K Direction Face
 *
 * @param      i      I Direction (-1, 0, +1)
 * @param      j      J Direction (-1, 0, +1)
 * @param      k      K Direction (-1, 0, +1)
 * @return    node face status<br>
 *               > 0 : adjacent node count<br>
 *               = 0 : not exist adjacent node<br>
 *               < 0 : voxel border face
 *
 * \~japanese
 * 指定方向のノード面に隣接するノードの有無・最外郭を取得する。
 *
 * パラメータ<br>
 * - i, j, k = -1,  0,  0 : -I 方向面
 * - i, j, k =  1,  0,  0 : +I 方向面
 * - i, j, k =  0, -1,  0 : -J 方向面
 * - i, j, k =  0,  1,  0 : +J 方向面
 * - i, j, k =  0,  0, -1 : -K 方向面
 * - i, j, k =  0,  0,  1 : +K 方向面
 *
 * @param      i      I 方向 (-1, 0, +1)
 * @param      j      J 方向 (-1, 0, +1)
 * @param      k      K 方向 (-1, 0, +1)
 * @return    ノード面状態<br>
 *               > 0 : 隣接ノード数<br>
 *               = 0 : 隣接ノードなし（ボクセル内側）<br>
 *               < 0 : 最外殻面
 */
int  sphSvParaController::isCommID(int i, int j, int k) const
{
  if( ((i<-1)||(i>1)) || ((j<-1)||(j>1)) || ((k<-1)||(k>1)) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "-1>[i,j,k=%d,%d,%d]>+1", i, j, k);
    return -1;
  }

  if (i+j+k != -1 && i+j+k != 1) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "[i+j+k=%d,%d,%d] != -1, +1", i, j, k);
    return -1;
  }

  const sphSvNodeInfo* nodeinfo = getNodeInfo();
  if (!nodeinfo) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
    return -1;
  }

  int num = nodeinfo->getCommIDNum(i, j, k);
  if (num > 0) {
    return num;
  }

  // check my node size
  const size_t* node_sz = nodeinfo->getVoxelSize();
  if (!node_sz) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "node_sz is null");
    return -1;
  }

  // check outside face
  const size_t* voxMinIdx = this->getReferIndex();
  const size_t* voxel_sz = getWholeVoxelSize();
  const size_t* node_min = nodeinfo->getVoxelMin();
  const size_t* node_max = nodeinfo->getVoxelMax();

  if ( (i==-1) && (node_min[0] == voxMinIdx[0]) ) return -1;
  if ( (i==1) && (node_max[0] == voxMinIdx[0] + voxel_sz[0]-1) ) return -1;
  if ( (j==-1) && (node_min[1] == voxMinIdx[1]) ) return -1;
  if ( (j==1) && (node_max[1] == voxMinIdx[1] + voxel_sz[1]-1) ) return -1;
  if ( (k==-1) && (node_min[2] == voxMinIdx[2]) ) return -1;
  if ( (k==1) && (node_max[2] == voxMinIdx[2] + voxel_sz[2]-1) ) return -1;

  return 0;

}



/**
 * 指定方向のノード面に隣接するノード番号を取得する。
 *
 * パラメータ<br>
 * - i, j, k = -1,  0,  0 : -I 方向面<br/>
 * - i, j, k =  1,  0,  0 : +I 方向面<br/>
 * - i, j, k =  0, -1,  0 : -J 方向面<br/>
 * - i, j, k =  0,  1,  0 : +J 方向面<br/>
 * - i, j, k =  0,  0, -1 : -K 方向面<br/>
 * - i, j, k =  0,  0,  1 : +K 方向面<br/>
 *
 * @param      i      I 方向 (-1, 0, +1)
 * @param      j      J 方向 (-1, 0, +1)
 * @param      k      K 方向 (-1, 0, +1)
 * @return    隣接するノード番号<br/>
 *               最外殻の場合、-1を返す。<br/>
 *               エラーの場合、-2を返す。<br/>
 */
int sphSvParaController::getCommIDEv (int i, int j, int k) const
{
  // 均等分割チェック
  int len = 1;
  register int n, dm;
  for (n=0; n<3; n++) {
    len *= this->m_divInfo[n];
  }
  if (len <= 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "this solver not is EV division.");
    return -2;
  }

  if( ((i<-1)||(i>1)) || ((j<-1)||(j>1)) || ((k<-1)||(k>1)) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "-1>[i,j,k=%d,%d,%d]>+1", i, j, k);
    return -2;
  }

  if (i+j+k != -1 && i+j+k != 1) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "[i+j+k=%d,%d,%d] != -1, +1", i, j, k);
    return -2;
  }

  const sphSvNodeInfo* nodeinfo = getNodeInfo();
  if (!nodeinfo) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
    return -2;
  }

  // 隣接ノード数
  int num = nodeinfo->getCommIDNum(i, j, k);
  const size_t* voxel_min = nodeinfo->getVoxelMin();
  const size_t* voxel_max = nodeinfo->getVoxelMax();

  if (num <= 0) {
    return -1;
  }

  int nodeid = -1;
  const int* nodeIds = nodeinfo->getCommIDs(i, j, k);
  for (n=0; n<num; n++) {

    /// 通信面情報を取得する。
    const sphOffsetDiv* divinfo = nodeinfo->getOffsetDivInfo(nodeIds[n]);
    const size_t* div_min = divinfo->getOffsetMin();
    const size_t* div_max = divinfo->getOffsetMax();
    FaceType ftype = divinfo->getFaceType();

    // 斜め方向の通信チェック
    bool isslant = false;
    int ext_dm = 0;
    if (ftype == XFACE_MINUS || ftype == XFACE_PLUS) ext_dm = 0;
    else if (ftype == YFACE_MINUS || ftype == YFACE_PLUS) ext_dm = 1;
    else if (ftype == ZFACE_MINUS || ftype == ZFACE_PLUS) ext_dm = 2;
    for (dm = 0; dm <3; dm++) {
      if (dm == ext_dm) continue;
      if (div_min[dm] != voxel_min[dm]) isslant = true;
      if (div_max[dm] != voxel_max[dm]) isslant = true;
    }
    if (isslant) {
      // 斜め方向の通信であるので除外する。
      continue;
    }
    // 最初に見つかったノード番号を返す。
    nodeid = nodeIds[n];
    break;
  }

  return nodeid;
}



/**
 * Periodic通信の最外殻の通信相手ノード番号を取得する。
 *
 * @param      face     通信面 (0:X面、1:Y面、2:Z面)
 * @param      dir      通信方向 (-1:マイナス方向、+1:プラス方向)
 * @return    Periodic通信するノード番号<br/>
 *               最外殻ではない場合、-1を返す。<br/>
 *               エラーの場合、-2を返す。<br/>
 */
int sphSvParaController::getCommIDPeriodic (int face, int dir) const
{
  // 均等分割チェック
  int len = 1;
  register int i;
  for (i=0; i<3; i++) {
    len *= this->m_divInfo[i];
  }
  if (len <= 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOT_SUPPORT_METHOD, "this solver not is EV division.");
    return -2;
  }

  // 自身が指定方向が最外殻であるかチェックする。
  int idir = 0, jdir=0, kdir=0;
  switch (face) {
    case 0:
      idir = dir;
      break;
    case 1:
      jdir = dir;
      break;
    case 2:
      kdir = dir;
      break;
  }
  int num = isCommID(idir, jdir, kdir);
  if (num != -1) {
    // 最外殻ではない
    return -1;
  }

  const sphSvNodeInfo* nodeinfo = getNodeInfo();
  if (!nodeinfo) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
    return -2;
  }
  // ノード始点、終点インデックス
  const size_t* voxel_min = nodeinfo->getVoxelMin();
  const size_t* voxel_max = nodeinfo->getVoxelMax();

  int nodeid = -1;
  // node info map
  MapNodeInfo::const_iterator itr = m_nodeInfo->begin();
  for( ; itr != m_nodeInfo->end(); itr++){
    // node info
    int other_nodeid = itr->first;
    sphSvNodeInfo* other_info = itr->second;
    int other_num = other_info->getCommIDNum((-1)*idir, (-1)*jdir, (-1)*kdir);
    if (other_num > 0) {
      continue;
    }

    // ノード始点、終点インデックス
    const size_t* other_min = other_info->getVoxelMin();
    const size_t* other_max = other_info->getVoxelMax();

    // 同一方向のノードであるか？
    bool isaxis = true;
    for (i=0; i<3; i++) {
      if (i == face) continue;
      if (voxel_min[i] != other_min[i]) isaxis = false;
      if (voxel_max[i] != other_max[i]) isaxis = false;
    }
    if (!isaxis) {
      // 別軸のノード
      continue;
    }

    // EV型であるので隣接面がなければ最外殻である。
    nodeid = other_nodeid;
    break;
  }

  return nodeid;
}


/**
 * \~english
 * get start Index in total voxel
 * @return    start Index in total voxel
 *
 * \~japanese
 * 全ノードの始点インデックスを取得する。
 * @return    全ノードの始点インデックス
 */
const size_t* sphSvParaController::getReferIndex() const
{
  return m_voxMinIdx;
}


/**
 * 全ノードの始点インデックスを設定する。
 * @param     sidx     全ノードの始点インデックス
 */
void sphSvParaController::setReferIndex(const size_t sidx[3])
{
  register int i;
  for (i=0; i<3; i++) m_voxMinIdx[i] = sidx[i];
}


/**
 * \~english
 * get node Information
 * @param      nodeId      node id (not parameter=my node)
 * @return    Node Information Class
 *
 * \~japanese
 * ノード情報クラスを取得する。
 * @param      nodeId      ノード番号（指定しない場合：自ノード）
 * @return    ノード情報クラス
 */
const sphSvNodeInfo* sphSvParaController::getNodeInfo(int nodeId) const
{
  if (nodeId == -1) {
    nodeId = m_myID;
  }
  MapNodeInfo::const_iterator itr = m_nodeInfo->find(nodeId);
  MapNodeInfo::const_iterator itrEnd = m_nodeInfo->end();
  if( itr == itrEnd ) return NULL;
  return itr->second;
}


/**
 * \~english
 * get node count in node information class list
 * @return    node count
 *
 * \~japanese
 * ノード情報クラスリストのノード数を取得する。
 * @return    ノード数
 */
int  sphSvParaController::getNumberOfNode() const
{
  return m_nodeInfo->size();
}


/**
 * \~english
 * get adjacent total node count of all direction
 * @return    adjacent total node count
 *
 * \~japanese
 * すべての隣接ノード数を取得する。
 * @return    全隣接ノード数
 */
int  sphSvParaController::getOffsetNodeNum()
{
  const sphSvNodeInfo* nodeinfo = getNodeInfo();
  if (!nodeinfo) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
    return 0;
  }
  int num = nodeinfo->getOffsetDivNum();

  return num;
}


/**
 * \~english
 * add Node information class
 * @param      nodeid        node id
 * @param      min_i          start I index
 * @param      min_j          start J index
 * @param      min_k          start K index
 * @param      max_i          end I index
 * @param      max_j          end J index
 * @param      max_k          end K index
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ノード情報クラスリストへの追加を行う。
 * @param      nodeid        ノード番号
 * @param      min_i          I 始点インデックス
 * @param      min_j          J 始点インデックス
 * @param      min_k          K 始点インデックス
 * @param      max_i          I 終点インデックス
 * @param      max_j          J 終点インデックス
 * @param      max_k          K 終点インデックス
 * @param      groupid          グループID(=ランクID)
 * @return    success(=true)/fault(=false)
 */
bool sphSvParaController::addNodeInfo(
                        int nodeid,
                        size_t min_i, size_t min_j, size_t min_k,
                        size_t max_i, size_t max_j, size_t max_k,
                        unsigned int groupid)
{
  sphSvNodeInfo* nodeinfo = new sphSvNodeInfo();
  nodeinfo->setNodeInfo(nodeid, min_i, min_j, min_k, max_i, max_j, max_k);
  nodeinfo->setGroupID(groupid);
  if ( !addNodeInfo(nodeid, nodeinfo) ){
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "addNodeInfo is error");
    delete nodeinfo; return false;
  }
  return true;
}


/**
 * \~english
 * add Node information class
 * @param      nodeid        node id
 * @param      nodeinfo      Added Node information class
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ノード情報クラスリストへの追加を行う。
 * @param      nodeid        ノード番号
 * @param      nodeinfo      追加ノード情報クラス
 * @return    success(=true)/fault(=false)
 */
bool sphSvParaController::addNodeInfo(int nodeid, sphSvNodeInfo* nodeinfo)
{

  if( !((m_nodeInfo->insert(std::make_pair(nodeid, nodeinfo))).second) ){
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_nodeInfo. [nodeid=%ld].", nodeid);
    return false;
  }

  return true;
}


/**
 * \~english
 * create adjacent data
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 隣接面情報を作成する。
 * @return    success(=true)/fault(=false)
 */
bool sphSvParaController::makeNodeOffset()
{

  // node info map
  MapNodeInfo::iterator itr = m_nodeInfo->begin();
  MapNodeInfo::iterator itrEnd = m_nodeInfo->end();
  for( ; itr != itrEnd; itr++){
    // node info
    sphSvNodeInfo* node_src = itr->second;

    MapNodeInfo::iterator itr_dest = m_nodeInfo->begin();
    MapNodeInfo::iterator itrEnd_dest = m_nodeInfo->end();
    for( ; itr_dest != itrEnd_dest; itr_dest++){

      if (itr == itr_dest) {
        continue;
      }

      sphSvNodeInfo* node_dest = itr_dest->second;

      int divcnt = node_src->addOffsetDiv(node_dest);
      if (divcnt < 0) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "addOffsetDiv is error");
        return false;
      }
    }

    // create ID Matrix Table
    node_src->createIdMatrix();

  }
  return true;
}



/**
 * 分割数を算出する。
 * @param[in]       divNum      全ノード数
 * @param[in]       voxSize     ボクセルサイズ
 * @param[out]      divPttn     算出結果分割数
 * @return    success(=true)/fault(=false)
 */
bool sphSvParaController::decideDivPattern(unsigned int divNum,
                                    const size_t voxSize[3],
                                    unsigned int divPttn[3]) const
{
  if( (divNum==0) || !voxSize || !divPttn ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "decideDivPattern arguments is null.");
    return false;
  }
  if( (voxSize[0]==0) || (voxSize[1]==0) || (voxSize[2]==0) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "voxel size is zero.");
    return false;
  }
  if( divNum == 1 ){
    divPttn[0] = divPttn[1] = divPttn[2] = 1;
    return true;
  }

  divPttn[0] = divPttn[1] = divPttn[2] = 0;
  long long minCommSize = voxSize[0]*voxSize[1]
                            +voxSize[1]*voxSize[2]
                            +voxSize[2]*voxSize[0];

  size_t n;
  for(n=1; n<=divNum; n++){
    size_t i, j, k;
    for(i=1; i<=n; i++){
      if( (voxSize[0] / i) < 1 ) break;
      size_t jmax = n/i;
      for(j=1; j<=jmax; j++){
        if( (voxSize[1] / j) < 1 ) break;
        size_t kmax = jmax/j;
        for(k=1; k<=kmax; k++){
          if( (voxSize[2] / k) < 1 ) break;
          if( i*j*k == divNum){
            long long commSize;
            size_t tmp_div[3] = {i, j, k};
            /// 通信セル数を算出して、最小通信サイズとなる分割数を取得する。
            if( (commSize=calcCommSize(tmp_div, voxSize)) == 0 ) break;
            if( commSize < minCommSize ){
              divPttn[0] = i; divPttn[1] = j; divPttn[2] = k;
              minCommSize = commSize;
            }
          }
        }
      }
    }
  }

  if( (divPttn[0]==0) || (divPttn[1]==0) || (divPttn[2]==0) ){
    divPttn[0] = divPttn[1] = divPttn[2] = 1;
    int maxLenElem = 2;
    if( voxSize[maxLenElem] < voxSize[1] ) maxLenElem = 1;
    if( voxSize[maxLenElem] < voxSize[0] ) maxLenElem = 0;
    divPttn[maxLenElem] = divNum;
  }

  return true;
}


/**
 * 通信面のセル数を算出する。
 * @param       div         分割数
 * @param       voxSize     ボクセルサイズ
 * @return      通信面のセル数
 */
long long sphSvParaController::calcCommSize(const size_t div[3],
                                        const size_t voxSize[3]) const
{
  if(div == NULL) return 0;
  if( voxSize == NULL) return 0;

  register int i;
  size_t Len[3];
  for (i=0; i<3; i++) {
    Len[i] = voxSize[i] / div[i];
    if( Len[i] == 0 ) return 0;
  }

  long long commFace[3];
  if( div[0] != 1) commFace[0] = Len[1]*Len[2]*div[0];
  else commFace[0] = 0;
  if( div[1] != 1) commFace[1] = Len[2]*Len[0]*div[1];
  else commFace[1] = 0;
  if( div[2] != 1) commFace[2] = Len[0]*Len[1]*div[2];
  else commFace[2] = 0;

  return (commFace[0] + commFace[1] + commFace[2]);
}


/**
 * 分割ノードのサイズ、始点インデックスを算出する。
 * @param[in]       voxSize           計算領域サイズ
 * @param[in]       divNode           分割数
 * @param[out]       distEnd          分割ノードサイズ (配列=分割ノード数*[3:i,j,k]）
 * @param[out]       distStart        分割ノード始点インデックス (配列=分割ノード数*[3:i,j,k]）
 * @return      通信面のセル数
 */
bool sphSvParaController::createVoxelDistribution(
                              const size_t voxSize[3],
                              const unsigned int divNode[3],
                              size_t (*distSize)[3],
                              size_t (*distStart)[3] ) const
{
  if(!distSize||!distStart||!voxSize||!divNode) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "CreateVoxelDistribution arguments is null.");
    return false;
  }

  // 全プロセス数
  unsigned int totalNode = divNode[0]*divNode[1]*divNode[2];

  register int dim, n;
  memset(distSize, 0x00, totalNode*3*sizeof(size_t));
  memset(distStart, 0x00, totalNode*3*sizeof(size_t));

  size_t *start_p[3] = {NULL, NULL, NULL};
  for (n=0; n<3; n++) start_p[n] = new size_t[divNode[n]];
  if (start_p[0] == NULL || start_p[1] == NULL || start_p[2] == NULL) {
    for (n=0; n<3; n++) if (start_p[n] != NULL) delete [] start_p[n];
    return false;
  }

  size_t base_len[3];
  for (n=0; n<3; n++) base_len[n] = voxSize[n] / divNode[n];

  size_t hasuu[3];
  for (n=0; n<3; n++) hasuu[n] = voxSize[n] % divNode[n];

  /// 分割を行った時の端数を先頭ノードから均等に割り振る。
  for(dim=0; dim<3; dim++){
    start_p[dim][0] = 0;
    for(n=1; n<divNode[dim]; n++){
      unsigned int tmp;
      if( hasuu[dim] == 0 ) tmp = 0;
      else { tmp = 1; hasuu[dim]--; };
      start_p[dim][n] = start_p[dim][n-1] + base_len[dim] + tmp;
    }
  }

  register unsigned int i, j, k;
  size_t (*distEnd)[3] = new size_t[totalNode][3];
  if (distEnd == NULL) {
    for (n=0; n<3; n++) if (start_p[n] != NULL) delete [] start_p[n];
    return false;
  }

  for(k=0; k<divNode[2]; k++){
    for(j=0; j<divNode[1]; j++){
      for(i=0; i<divNode[0]; i++){
        unsigned int id = divNode[0]*divNode[1]*k + divNode[0]*j + i;
        distStart[id][0] = start_p[0][i];
        distStart[id][1] = start_p[1][j];
        distStart[id][2] = start_p[2][k];
        if( i == (divNode[0]-1) ) distEnd[id][0] = voxSize[0] - 1;
        else                      distEnd[id][0] = start_p[0][i+1] - 1;
        if( j == (divNode[1]-1) ) distEnd[id][1] = voxSize[1] - 1;
        else                      distEnd[id][1] = start_p[1][j+1] - 1;
        if( k == (divNode[2]-1) ) distEnd[id][2] = voxSize[2] - 1;
        else                      distEnd[id][2] = start_p[2][k+1] - 1;
      }
    }
  }

  bool ret = true;
  for (n=0; n<totalNode; n++) {
    for (dim=0; dim<3; dim++) {
      // 始点インデックスと終点インデックスをチェックする。
      if (distEnd[n][dim] < distStart[n][dim]) {
        for (n=0; n<3; n++) if (start_p[n] != NULL) delete [] start_p[n];
        if (distEnd != NULL) delete [] distEnd;
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Invalid start and end index, nodeno=%d,dim=%d,startIndex=%ld, endIndex=%ld", n, dim, distStart[n][dim], distEnd[n][dim]);
        return false;
      }
      // ノードサイズを算出する。
      distSize[n][dim] = distEnd[n][dim] - distStart[n][dim] + 1;
    }
  }

  for (n=0; n<3; n++) if (start_p[n] != NULL) delete [] start_p[n];
  if (distEnd != NULL) delete [] distEnd;

  return true;
}


/**
 * 計算領域全体の原点座標を取得する。
 * @param      org          計算領域全体の原点座標
 * @return    void
 */
bool sphSvParaController::getVoxelOrigin(double org[3]) const
{
  register int i;
  for (i=0; i<3; i++) org[i] = m_voxelOrigin[i];
  return true;
}


/**
 * 計算領域全体の原点座標を設定する。
 * @param      org          計算領域全体の原点座標
 * @return    void
 */
void sphSvParaController::setVoxelOrigin(const double org[3])
{
  register int i;
  for (i=0; i<3; i++) {
    m_voxelOrigin[i] = org[i];
  }
  return;
}


/**
 * 自ノードの原点座標を取得する。
 * @param      org          自ノードの原点座標
 * @return    void
 */
bool sphSvParaController::getNodeOrigin(double org[3]) const
{
  register int i;
  const sphSvNodeInfo* my_nodeInfo = getNodeInfo();
  if (my_nodeInfo == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
    return false;
  }
  const size_t* sta_idx = my_nodeInfo->getVoxelMin();
  for (i=0; i<3; i++) {
    org[i] = m_voxelOrigin[i]+sta_idx[i]*m_voxelPitch[i];
  }

  return true;
}


/**
 * ピッチを取得する。
 * @param      pitch          ピッチ
 * @return    void
 */
bool sphSvParaController::getVoxelPitch(double pitch[3]) const
{
  register int i;
  for (i=0; i<3; i++) pitch[i] = m_voxelPitch[i];
  return true;
}


/**
 * ピッチを設定する。
 * @param      pitch        ピッチ
 * @return    void
 */
void sphSvParaController::setVoxelPitch(const double pitch[3])
{
  register int i;
  for (i=0; i<3; i++) {
    m_voxelPitch[i] = pitch[i];
  }
  return;
}


/**
 * 始点インデックスを取得する。
 * @param      sta_idx        始点インデックス（ガイドセルを含まない）
 * @return    void
 */
void sphSvParaController::getVoxelStartIndex(size_t sta_idx[3]) const
{
  register int i;
  for (i=0; i<3; i++) {
    sta_idx[i] = getVoxelHeadIndex(m_myID, i);
  }
  return;

}

/**
 * 分割情報を取得する。
 * @param      void
 * @return    分割情報
 */
const unsigned int* sphSvParaController::getVoxelDivInfo() const
{
  return m_divInfo;
}


/**
 * 分割情報を設定する。
 * @param      div    分割情報
 */
void sphSvParaController::setVoxelDivInfo(unsigned int div[3])
{
  m_divInfo[0] = div[0];
  m_divInfo[1] = div[1];
  m_divInfo[2] = div[2];
}


/**
 * サブブロックの分割情報を設定する。
 * @param      div    分割情報
 */
void sphSvParaController::setVoxelDivLocalInfo(unsigned int div[3])
{
  m_divLocalInfo[0] = div[0];
  m_divLocalInfo[1] = div[1];
  m_divLocalInfo[2] = div[2];
}


/**
 * EV分割であるかチェックする。
 * @return     true:EV分割
 */
bool sphSvParaController::isEv() const
{
  register int i;
  const unsigned int *div = getVoxelDivInfo();
  int div_cnt = 1;
  for (i=0; i<3; i++) div_cnt *= div[i];
  if( div_cnt <= 0 ) {
    // 均等分割ではない。
    return false;
  }

  return true;
}


/**
 * MB分割であるかチェックする。
 * @return     true:MB分割
 */
bool sphSvParaController::isMb() const
{
  return !isEv();
}


/**
 * 非格子構造であるかチェックする。
 * @return     true:非格子構造
 */
bool sphSvParaController::isUnst() const
{
  return false;
}


/**
 * ノード情報マップを取得する。
 * @return  ノード情報マップ
 */
MapNodeInfo* sphSvParaController::getMapNodeInfo()
{
  return m_nodeInfo;
}

/**
 * ノード情報マップを取得する。
 * @param   info  ノード情報マップ
 */
void sphSvParaController::setMapNodeInfo(MapNodeInfo* info)
{
  m_nodeInfo = info;
}


/**
 * オブジェクトの文字列表現を返します。(デバッグ用)
 * @return    オブジェクトの文字列表現
 */
std::string sphSvParaController::toString() const
{

  std::string msg = "sphSvParaController class\n";
  char buf[1024] = {0};

  // MyId
  sprintf(buf, "m_myID=%d", m_myID);
  msg += std::string(buf) + '\n';

  // Voxel Size
  sprintf(buf, "計算領域ボクセルサイズ = %ld, %ld, %ld", m_voxSize[0], m_voxSize[1], m_voxSize[2]);
  msg += std::string(buf) + '\n';
  // Voxel Index
  sprintf(buf, "計算領域始点インデックス = %ld, %ld, %ld", m_voxMinIdx[0], m_voxMinIdx[1], m_voxMinIdx[2]);
  msg += std::string(buf) + '\n';
  // 原点座標
  sprintf(buf, "計算領域原点座標 = %f, %f, %f", m_voxelOrigin[0], m_voxelOrigin[1], m_voxelOrigin[2]);
  msg += std::string(buf) + '\n';
  // ピッチ
  sprintf(buf, "計算領域ピッチ = %f, %f, %f", m_voxelPitch[0], m_voxelPitch[1], m_voxelPitch[2]);
  msg += std::string(buf) + '\n';

  if (isEv()) {
    msg += "Parallel Mode = EqualPartitioning\n";
    sprintf(buf, "    I,J,K方向分割数 = %d, %d, %d", m_divInfo[0], m_divInfo[1], m_divInfo[2]);
    msg += std::string(buf) + '\n';
  }
  else {
    msg += "Parallel Mode = MultiplyConnectedPartitioning\n";
  }

  msg += '\n';
  // ノード情報
  sprintf(buf, "ノード情報:size=%d", (int)m_nodeInfo->size());
  msg += std::string(buf) + '\n';
  MapNodeInfo::const_iterator itr = m_nodeInfo->begin();
  MapNodeInfo::const_iterator itrEnd = m_nodeInfo->end();
  for( ; itr != itrEnd; itr++){
    sphSvNodeInfo* nodeinfo = itr->second;
    msg += nodeinfo->toString();
    msg += '\n';
  }

  return msg;
}


/**
 * 親パラレルコントローラ(=sphMultiParaController)を取得する。
 *
 * 部分領域管理のサブブロック領域における親のプロセス管理パラレルコントローラを取得する。
 * @return      親パラレルコントローラ(=sphMultiParaController)
 */
const sphParaController* sphSvParaController::getParentController() const
{
  return m_parentCtrl;
}



/**
 * 並列実行であるかチェックする。
 *
 * 並列実行であっても、ボクセル分割数が１の場合はfalse(=逐次)を返す。
 * @return    true=並列実行, false=逐次実行
 */
bool sphSvParaController::isParallel() const
{
  // ノード数が１より多い場合は、true
  if (getNumberOfNode() <= 1) return false;
  return true;
}


/**
 * 隣接ノードIDテーブルを取得する。
 *
 * nIDTableは以下の面番号により面方向のノードIDをセットして戻す。<br/>
 *     nIDTable[6] = {X_MINUS=0, X_PLUS=1, Y_MINUS=2, Y_PLUS=3, Z_MINUS=4, Z_PLUS=5}<br/>
 * ノードIDテーブルは分割方法によりセットする値が異なる。<br/>
 *   EV分割の場合は、隣接ノード番号をセットする。<br/>
 *     nIDTable = 隣接するノード番号<br/>
 *                最外殻の場合、-1を返す。<br/>
 *                エラーの場合、-2を返す。<br/>
 *   MB分割の場合は、隣接ノード数をセットする。<br/>
 *     nIDTable > 0 : 隣接ノード数<br>
 *              = 0 : 隣接ノードなし（ボクセル内側）<br>
 *              < 0 : 最外殻面
 * @param[out] nIDTable      隣接ノードIDテーブル
 * @return        true=success
 */
bool sphSvParaController::getCommIDTable(int nIDTable[6]) const
{
  if (nIDTable == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nIDTable is null.");
    return false;
  }

  if (this->isEv()) {
    nIDTable[0] = this->getCommIDEv(-1, 0, 0);  // XMINUS
    nIDTable[2] = this->getCommIDEv( 0,-1, 0);  // YMINUS
    nIDTable[4] = this->getCommIDEv( 0, 0,-1);  // ZMINUS
    nIDTable[1]  = this->getCommIDEv( 1, 0, 0);  // XPLUS
    nIDTable[3]  = this->getCommIDEv( 0, 1, 0);  // YPLUS
    nIDTable[5]  = this->getCommIDEv( 0, 0, 1);  // ZPLUS
  }
  else if (this->isMb()) {
    nIDTable[0] = this->isCommID(-1, 0, 0);  // XMINUS
    nIDTable[2] = this->isCommID( 0,-1, 0);  // YMINUS
    nIDTable[4] = this->isCommID( 0, 0,-1);  // ZMINUS
    nIDTable[1]  = this->isCommID( 1, 0, 0);  // XPLUS
    nIDTable[3]  = this->isCommID( 0, 1, 0);  // YPLUS
    nIDTable[5]  = this->isCommID( 0, 0, 1);  // ZPLUS
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid parallel.");
    return false;
  }

  return true;
}


/**
 * ノードIDの分割位置マッピングテーブルを作成する
 *
 * m_nodeEvMap[0:I方向分割数-1][0:J方向分割数-1][0:K方向分割数-1] = ノード番号
 * ノード番号が存在しない位置にはSPL_PROC_NULLをセットする。
 * @return     true=success
 */
bool sphSvParaController::createNodeMap()
{
  if (!isEv()) {
    // EV分割ではない。
    return true;
  }

  int node_count = this->getNumberOfNode();
  register int i, j, k, n;

  if (node_count != m_divInfo[0]*m_divInfo[1]*m_divInfo[2]) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INTERNAL_IMPOSSIBLE, "node count is not match division.[node count=%d, div=%d,%d,%d]", node_count, m_divInfo[0], m_divInfo[1], m_divInfo[2]);
    return false;
  }

  // 全ノードの始点インデックスを取得する
  std::set<size_t> start_i;
  std::set<size_t> start_j;
  std::set<size_t> start_k;

  for (n=0; n<node_count; n++) {

    const sphSvNodeInfo* nodeinfo = getNodeInfo(n);
    if (!nodeinfo) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "nodeinfo is null");
      return false;
    }
    const size_t* start_idx = nodeinfo->getVoxelMin();
    if (!start_idx) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "voxel_min is null");
      return false;
    }
    start_i.insert(start_idx[0]);
    start_j.insert(start_idx[1]);
    start_k.insert(start_idx[2]);
  }

  // ノード番号テーブルの生成を行う。
  m_nodeEvMap = new int**[m_divInfo[0]];
  for(i=0; i < m_divInfo[0]; i++) {
    m_nodeEvMap[i] = new int*[m_divInfo[1]];
    for(j=0; j < m_divInfo[1]; j++) {
      m_nodeEvMap[i][j] = new int[m_divInfo[2]];
    }
  }

  for (i=0; i < m_divInfo[0]; i++) {
    for (j=0; j < m_divInfo[1]; j++) {
      for (k=0; k < m_divInfo[2]; k++) {
        m_nodeEvMap[i][j][k] = SPL_PROC_NULL;
      }
    }
  }

  // 各ノードの始点インデックスからノード分割位置を設定する。
  for (n=0; n<node_count; n++) {
    const sphSvNodeInfo* nodeinfo = getNodeInfo(n);
    const size_t* start_idx = nodeinfo->getVoxelMin();
    int nodeID = nodeinfo->getNodeID();

    // 始点インデックスのインデックスを取得する。
    std::set<size_t>::iterator it_i = start_i.find(start_idx[0]);
    int i_idx = std::distance( start_i.begin(), it_i );
    std::set<size_t>::iterator it_j = start_j.find(start_idx[1]);
    int j_idx = std::distance( start_j.begin(), it_j );
    std::set<size_t>::iterator it_k = start_k.find(start_idx[2]);
    int k_idx = std::distance( start_k.begin(), it_k );
    m_nodeEvMap[i_idx][j_idx][k_idx] = nodeID;
  }

  return true;
}

/**
 * ノードIDの分割位置マッピングテーブルを取得する。
 *
 * m_nodeEvMap[0:I方向分割数-1][0:J方向分割数-1][0:K方向分割数-1] = ノード番号
 * ノード番号が存在しない位置にはSPL_PROC_NULLがセットされている。
 * @param[out] i       I方向分割数
 * @param[out] j       J方向分割数
 * @param[out] k       K方向分割数
 * @return      ノードIDの分割位置マッピングテーブル
 */
const int*** sphSvParaController::getNodeMap(unsigned int& i, unsigned int& j, unsigned int& k) const
{
  i = m_divInfo[0];
  j = m_divInfo[1];
  k = m_divInfo[2];

  return (const int***)m_nodeEvMap;
}



/**
 * 自ノードの分割位置を取得する。
 *
 * ノード番号が存在しない場合は、falseを返す。
 * @param[out] i       I方向分割位置 (0 〜 I方向分割数-1)
 * @param[out] j       J方向分割位置 (0 〜 J方向分割数-1)
 * @param[out] k       K方向分割位置 (0 〜 K方向分割数-1)
 * @return      true=ノード番号位置の取得成功
 */
bool sphSvParaController::getNodeDivPosition(unsigned int& i_idx, unsigned int& j_idx, unsigned int& k_idx) const
{
  register size_t i, j, k;

  if (!m_nodeEvMap) return false;

  int myID = this->getMyID();
  for (i=0; i < m_divInfo[0]; i++) {
    for (j=0; j < m_divInfo[1]; j++) {
      for (k=0; k < m_divInfo[2]; k++) {
        if (myID = m_nodeEvMap[i][j][k]) {
          i_idx = i;
          j_idx = j;
          k_idx = k;
          return true;
        }
      }
    }
  }
  return false;
}


