/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphUnstMeshDouble.C
 * Double Unstructured Mesh Information Class Source File
 * @author     tamaki
 * @date       2010/04/22
 *
 * \~japanese
 * @file sphUnstMeshDouble.C
 * 倍精度非構造格子情報クラスソースファイル
 * @author     tamaki
 * @date       2010/04/22
 */

#include "parallel/sphUnstMeshDouble.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphUnstMeshDouble::sphUnstMeshDouble() {
  sphUnstMesh::m_meshDtType = MESH_DOUBLE;
  m_sData = NULL;
  m_rData = NULL;
  xyz = NULL;
}

/**
 * \~english
 * Destoructor
 *
 * \~japanese
 * デストラクタ
 */
sphUnstMeshDouble::~sphUnstMeshDouble() {
  if( m_sData ) {
    for(int i=0; i<m_nPE; i++) if( m_sData[i] ) delete [] m_sData[i];
    delete [] m_sData;
  }
  if( m_rData ) {
    for(int i=0; i<m_nPE; i++) if( m_rData[i] ) delete [] m_rData[i];
    delete [] m_rData;
  }
  if( xyz ) delete [] xyz;
}

bool
sphUnstMeshDouble::readNode(const char* fn) {
  FILE *fp;
  int i;
  /* ******** Open node file ******** */
  if( !fn ) return false;
  if( (fp = fopen(fn, "r")) == NULL ) return false;
  /* ******** Input variables ******** */
  fscanf(fp,"%d",&nNode);
  nnNode = nNode;
 
  /* ------------ Memory Allocation ------------ */
  if( !(xyz = new double[nNode*3]) ) return false;
  memset(xyz, 0, sizeof(double)*nNode*3);
  for ( i = 0; i < nNode; i++ ){
    fscanf(fp,"%lf %lf %lf",&xyz[3*i],&xyz[3*i+1],&xyz[3*i+2]);
  }
  fclose(fp);
  return true;
}

bool
sphUnstMeshDouble::comm(double  *x) {
  int i,j,index;

  if ( m_nPE > 1 ){
    /* ******** Memory allocation ******** */
    m_sData = new double*[m_nPE];
    if ( !m_sData ) {
      fprintf(stderr,"Error: Fail to allocate temp array for message passing.\n");
      exit(0);
    }

    m_rData = new double*[m_nPE];
    if ( !m_rData ) {
      fprintf(stderr,"Error: Fail to allocate temp array for message passing.\n");
      exit(0);
    }

    m_sRequest = new SPL_Request[m_nPE];
    if ( !m_sRequest ) {
      fprintf(stderr,"Error: Fail to allocate temp array for message passing.\n");
      exit(0);
    }

    m_rRequest = new SPL_Request[m_nPE];
    if ( !m_rRequest ) {
      fprintf(stderr,"Error: Fail to allocate temp array for message passing.\n");
      exit(0);
    }

    m_sStatus = new SPL_Status[m_nPE];
    if ( !m_sStatus ) {
      fprintf(stderr,"Error: Fail to allocate temp array for message passing.\n");
      exit(0);
    }

    m_rStatus = new SPL_Status[m_nPE];
    if ( !m_rStatus ) {
      fprintf(stderr,"Error: Fail to allocate temp array for message passing.\n");
      exit(0);
    }

    for ( i = 0; i < m_nPE; i++ ){
      m_sData[i] = new double[m_commTable[i]];
      memset(m_sData[i], 0, sizeof(double)*m_commTable[i]);
      m_rData[i] = new double[m_commTable[i]];
      memset(m_rData[i], 0, sizeof(double)*m_commTable[i]);
    }

    /* ******** Pickup data for sending ******** */
    for   ( i = 0; i < m_nPE;          i++ ){
      for ( j = 0; j < m_commTable[i]; j++ ){
        index = m_commList[i][j];
        m_sData[i][j] = x[index];
      }
    }

    /* ******** Send data ******** */
    for   ( i = 0; i < m_nPE; i++ ){
      if ( m_PElist[i] != m_myRank ){
        SPL_Isend( m_sData[i], m_commTable[i], SPL_DOUBLE, m_PElist[i],
                   0, SPL_COMM_WORLD, &m_sRequest[i]);
      }
    }

    /* ******** Receive data ******** */
    for   ( i = 0; i < m_nPE; i++ ){
      if ( m_PElist[i] != m_myRank ){
        SPL_Irecv( m_rData[i], m_commTable[i], SPL_DOUBLE, m_PElist[i],
                   0, SPL_COMM_WORLD, &m_rRequest[i]);
      }
    }
  }
  else{
    return true;
  }
  return true;
}

bool
sphUnstMeshDouble::wait(double  *x) {
  int      i,j,index;

  if ( m_nPE > 1 ){
    /* ******** Finalize communication ******** */
    for   ( i = 0; i < m_nPE; i++ ){
      if ( m_PElist[i] != m_myRank ){
        SPL_Wait( &m_sRequest[i], &m_sStatus[i]);
      }
    }
    for   ( i = 0; i < m_nPE; i++ ){
      if ( m_PElist[i] != m_myRank ){
        SPL_Wait( &m_rRequest[i], &m_rStatus[i]);
      }
    }

    /* ******** Adding data ******** */
    for   ( i = 0; i < m_nPE;          i++ ){
      for ( j = 0; j < m_commTable[i]; j++ ){
        index = m_commList[i][j];
        x[index] += m_rData[i][j];
      }
    }

    /* ------------------------ */
    for ( i = 0; i < m_nPE; i++ ){
      delete [] m_sData[i]; m_sData[i] = NULL;
      delete [] m_rData[i]; m_rData[i] = NULL;
    }
    delete [] m_sData; m_sData = NULL;
    delete [] m_rData; m_rData = NULL;
    delete [] m_sRequest; m_sRequest = NULL;
    delete [] m_rRequest; m_rRequest = NULL;
    delete [] m_sStatus; m_sStatus = NULL;
    delete [] m_rStatus; m_rStatus = NULL;
  }
  else{
    return true;
  }
  return true;
}

#define NAME_LEN 256

bool
sphUnstMeshDouble::readDistMesh(const char* filename) {
  int   i;
  FILE *fp;

  if( !filename ) return false;
  /* ------------ File open ------------ */
  if ( (fp=fopen(filename,"rb")) == NULL ){
    fprintf(stderr,"\nError: Fail to open mesh file!");
    fprintf(stderr,"\nError: Filename: %s!\n",filename);
    return false;
  }

  /* ------------ MESH ------------ */
  fread(&nNode,sizeof(int),1,fp);          /* nNode     */
  fread(&nnNode,sizeof(int),1,fp);         /* nnNode    */

  node = new int[nnNode];
  memset(node, 0, sizeof(int)*nnNode);
  fread( node,sizeof(int),nnNode,fp); /* node list */

  fread(&nInner,sizeof(int),1,fp);
  fread(&nOverlapped,sizeof(int),1,fp);

  elem = new int[nInner+nOverlapped];
  memset(elem, 0, sizeof(int)*(nInner+nOverlapped));
  fread( elem,sizeof(int),nInner+nOverlapped,fp);

  nop = new int[(nInner+nOverlapped)*8];
  memset(nop, 0, sizeof(int)*(nInner+nOverlapped)*8);
  fread( nop,sizeof(int),(nInner+nOverlapped)*8,fp);

  /* ------------ GRP ------------ */
  /* ------------ Node group ------------ */
  fread(&nNgrp,sizeof(int),1,fp);
  if ( nNgrp > 0 ){
    NgrpName = new char*[nNgrp];
    for ( i = 0; i < nNgrp; i++ ){
      NgrpName[i] = new char[NAME_LEN];
      if( fread(NgrpName[i],sizeof(char),128,fp) != 128 ) {
        printf("DEBUG@tama ERROR1\n"); fflush(stdout);
        return false;
      }
    }
    NgrpIndex = new int[nNgrp+1];
    memset(NgrpIndex, 0, sizeof(int)*(nNgrp+1));
    if( fread( NgrpIndex,sizeof(int),nNgrp+1,fp) != nNgrp+1 ) {
      printf("DEBUG@tama ERROR2\n"); fflush(stdout);
      return false;
    }
    if ( NgrpIndex[nNgrp] ){
      Ngrp = new int[NgrpIndex[nNgrp]];
      memset(Ngrp, 0, sizeof(int)*NgrpIndex[nNgrp]);
      if( fread( Ngrp,sizeof(int),NgrpIndex[nNgrp],fp) != NgrpIndex[nNgrp] ) {
        printf("DEBUG@tama ERROR3\n"); fflush(stdout);
        return false;
      }
    }
  }
  /* ------------ Element group ------------ */
  fread(&nEgrp,sizeof(int),1,fp);
  if ( nEgrp > 0 ){
    EgrpName = new char*[nEgrp];
    for ( i = 0; i < nEgrp; i++ ){
      EgrpName[i] = new char[NAME_LEN];
      if( fread(EgrpName[i],sizeof(char),128,fp) != 128 ) {
        printf("DEBUG@tama ERROR4\n"); fflush(stdout);
        return false;
      }
    }
    EgrpIndex = new int[nEgrp+1];
    memset(EgrpIndex, 0, sizeof(int)*(nEgrp+1));
    fread( EgrpIndex,sizeof(int),nEgrp+1,fp);
    if ( EgrpIndex[nEgrp] ){
      Egrp = new int[EgrpIndex[nEgrp]];
      memset(Egrp, 0, sizeof(int)*EgrpIndex[nEgrp]);
      if( fread( Egrp,sizeof(int),EgrpIndex[nEgrp],fp) != EgrpIndex[nEgrp] ) {
        printf("DEBUG@tama ERROR5\n"); fflush(stdout);
        return false;
      }
    }
  }
  /* ------------ Surface group ------------ */
  fread(&nSgrp,sizeof(int),1,fp);
  if ( nSgrp > 0 ){
    SgrpName = new char*[nSgrp];
    //SgrpName = (char**)new char[nSgrp][NAME_LEN];
    //tamaki memset(SgrpName, 0, sizeof(char)*nSgrp*NAME_LEN);
    for ( i = 0; i < nSgrp; i++ ){
      SgrpName[i] = new char[NAME_LEN];
      if( fread(SgrpName[i],sizeof(char),128,fp) != 128 ) {
        printf("DEBUG@tama ERROR6\n"); fflush(stdout);
        return false;
      }
    }
    SgrpIndex = new int[nSgrp+1];
    memset(SgrpIndex, 0, sizeof(int)*(nSgrp+1));
    fread( SgrpIndex,sizeof(int),nSgrp+1,fp);
    if ( SgrpIndex[nSgrp] ){
      Sgrp = new int[SgrpIndex[nSgrp]*2];
      memset(Sgrp, 0, sizeof(int)*SgrpIndex[nSgrp]*2);
      if( fread( Sgrp,sizeof(int),SgrpIndex[nSgrp]*2,fp) != SgrpIndex[nSgrp] ) {
        printf("DEBUG@tama ERROR7\n"); fflush(stdout);
        return false;
      }
    }
  }
  /* ------------ COMM ------------ */
  int tmp_nPE;
  if( fread(&tmp_nPE,sizeof(int),1,fp) != 1 ) {
    printf("DEBUG@tama ERROR8\n"); fflush(stdout);
    return false;
  }
  if( tmp_nPE != m_nPE ) {
    printf("DEBUG@tama ERROR9\n"); fflush(stdout);
    // mod 20100601@tamaki
    m_nPE = tmp_nPE;
    //return false;
  }

  if ( m_nPE > 0 ){
    m_PElist = new int[m_nPE];
    memset(m_PElist, 0, sizeof(int)*m_nPE);
    fread( m_PElist,sizeof(int),m_nPE,fp);
    m_commTable = new int[m_nPE];
    memset(m_commTable, 0, sizeof(int)*m_nPE);
    fread(m_commTable, sizeof(int), m_nPE,fp);
    m_commList = new int*[m_nPE];
    if ( !m_commList ){
      fprintf(stderr,"Error: In ReadMesh(Renumber.c)\n");
      fprintf(stderr,"Error: Fail to allocate commList memory!\n");
      exit(0);
    }
    for ( i = 0; i < m_nPE; i++ ){
      m_commList[i] = new int[m_commTable[i]];
      memset(m_commList[i], 0, sizeof(int)*m_commTable[i]);
      fread( m_commList[i],sizeof(int),m_commTable[i],fp);
    }
  }
  /* ------------ Coordinate ------------ */
  int tmp_nnNode;
  fread(&tmp_nnNode,sizeof(int),1,fp);
  if( nnNode != tmp_nnNode ) {
    return false;
  }
  xyz = new double[nnNode*3];
  memset(xyz, 0, sizeof(double)*nnNode*3);
  fread( xyz,sizeof(double),nnNode*3,fp);

  fclose(fp);
  return true;
}

#undef NAME_LEN

