/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphUnstParaController.C
 * Unstructured Mesh Type Parallel Controller Source File
 * @author     hira
 * @date       2010/04/01
 *
 * \~japanese
 * @file sphUnstParaController.C
 * 非構造格子パラレルコントローラーソースファイル
 * @author     hira
 * @date       2010/04/01
 */

#include "parallel/sphUnstParaController.h"
#include "parallel/sphUnstMeshDouble.h"
#include "parallel/sphUnstMeshFloat.h"


/**
 * \~english
 * Constructor
 * @param      paraKey        parallel controller key
 *
 * \~japanese
 * コンストラクタ
 * @param      paraKey        パラレルコントローラー識別キー
 */
sphUnstParaController::sphUnstParaController(unsigned int paraKey) : sphParaController(paraKey)
{
  m_mesh = NULL;
}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ
 */
sphUnstParaController::~sphUnstParaController()
{
  if( m_mesh ) delete m_mesh;
}


/**
 * 非構造格子データをセットする。
 * @param      mesh          非構造格子データ
 * @return     success(=true)/fault(=false)
 */
bool
sphUnstParaController::setMesh(sphUnstMesh* mesh) {
  if( !mesh ) return false;
  m_mesh = mesh;
  return true;
}


/**
 * 非構造格子データを取得する。
 * @return     非構造格子データオブジェクト
 */
const sphUnstMesh*
sphUnstParaController::getMesh() const {
  return (const sphUnstMesh*)m_mesh;
}


/**
 * 非構造格子データを取得する。
 * @return     非構造格子データオブジェクト
 */
sphUnstMesh*
sphUnstParaController::getMesh() {
  return m_mesh;
}

/**
 * ファイルから逐次版非構造格子データを読み込む。
 * @param     dtype     浮動小数点数精度（現在はDOUBLEのみ）
 * @param     fn_node   節点情報ファイル
 * @param     fn_elem   要素情報ファイル
 * @param     fn_grp    グループ情報ファイル
 * @param     fn_bnd    境界情報ファイル
 * @return    success(=true)/fault(=false)
 */
bool
sphUnstParaController::createMesh(sphUnstMesh::MeshDataType dtype,
                                  const char* fn_node,
                                  const char* fn_elem,
                                  const char* fn_grp,
                                  const char* fn_bnd) {
  if( !fn_node || !fn_elem ) return false;

  if( m_mesh ) delete m_mesh;
  m_mesh = NULL;

  if( dtype == sphUnstMesh::MESH_DOUBLE ) {
    if( !(m_mesh = dynamic_cast<sphUnstMesh*>( new sphUnstMeshDouble() )) ) {
      // error
      return false;
    }
  }
  else if ( dtype == sphUnstMesh::MESH_FLOAT ) {
    if( !(m_mesh = dynamic_cast<sphUnstMesh*>( new sphUnstMeshFloat() )) ) {
      // error
      return false;
    }
  }
  else {
    // error
    return false;
  }

  SPL_Comm_size(m_splComm, &m_mesh->m_nPE);
  m_mesh->m_nPE -= 1; // "m_mesh->m_nPE" is "MPI_Comm_size() - 1" -> ("num of proc" - "myproc")
  m_mesh->m_myRank = m_myID;
  m_mesh->m_comm = m_splComm;

  if( !m_mesh->readNode(fn_node) ) {
    // error
    return false;
  }
  if( !m_mesh->readElement(fn_elem) ) {
    // error
    return false;
  }
  if( fn_grp ) {
    if( !m_mesh->readGroup(fn_grp) ) {
      // error
      return false;
    }
  }
  if( fn_bnd ) {
    if( !m_mesh->readBoundary(fn_bnd) ) {
      // error
      return false;
    }
  }

  return true;
}

/**
 * ファイルから並列版非構造格子データを読み込む。
 * @param     file_prefix    分散非構造格子ファイル
 * @return    success(=true)/fault(=false)
 */
bool
sphUnstParaController::createMesh(sphUnstMesh::MeshDataType dtype,
                                  const char* filename ) {
  if( !filename ) return false;

  if( m_mesh ) delete m_mesh;
  m_mesh = NULL;

  if( dtype == sphUnstMesh::MESH_DOUBLE ) {
    if( !(m_mesh = dynamic_cast<sphUnstMesh*>( new sphUnstMeshDouble() )) ) {
      // error
      return false;
    }
  }
  else if ( dtype == sphUnstMesh::MESH_FLOAT ) {
    if( !(m_mesh = dynamic_cast<sphUnstMesh*>( new sphUnstMeshFloat() )) ) {
      // error
      return false;
    }
  }
  else {
    // error
    return false;
  }

  SPL_Comm_size(m_splComm, &m_mesh->m_nPE);
  m_mesh->m_nPE -= 1; // "m_mesh->m_nPE" is "MPI_Comm_size() - 1" -> ("num of proc" - "myproc")
  m_mesh->m_myRank = m_myID;
  m_mesh->m_comm = m_splComm;

  if( !m_mesh->readDistMesh(filename) ) {
    // error
    return false;
  }

  return true;
}


/**
 * 計算領域全体の原点座標を取得する。
 * @param      org          計算領域全体の原点座標
 * @return    void
 */
bool sphUnstParaController::getVoxelOrigin(double org[3]) const
{
  return false;
}


/**
 * 自ノードの原点座標を取得する。
 * @param      org          自ノードの原点座標
 * @return    void
 */
bool sphUnstParaController::getNodeOrigin(double org[3]) const
{
  return false;
}



/**
 * EV分割であるかチェックする。
 * @return     true:EV分割
 */
bool sphUnstParaController::isEv() const
{
  return false;
}


/**
 * MB分割であるかチェックする。
 * @return     true:MB分割
 */
bool sphUnstParaController::isMb() const
{
  return false;
}


/**
 * 非格子構造であるかチェックする。
 * @return     true:非格子構造
 */
bool sphUnstParaController::isUnst() const
{
  return true;
}


/**
 * 並列実行であるかチェックする。
 *
 * 並列実行であっても、ノード数が１の場合はfalse(=逐次)を返す。
 * @return    true=並列実行, false=逐次実行
 */
bool sphUnstParaController::isParallel() const
{
  if (m_mesh == NULL) return false;

  // ノード数が１より多い場合は、true
  if (m_mesh->nNode <= 1) return false;
  return true;
}


