/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */
/**
 * @file SklCfgDomainInfo.C
 * コンフィグレーションDomainInfo要素クラスソースファイル
 * @author     riken
 * @date       2010/10/01
 */

#include <string.h> // add 20090629
#include "skl_config/SklCfgDefName.h"
#include "skl_config/SklCfgDomainInfo.h"

namespace SklCfg {

/**
 * コンストラクタ
 */
SklCfgDomainInfo::SklCfgDomainInfo()
{
  m_voxOrg = NULL;
  m_voxSize = NULL;
  m_voxWidth = NULL;
  m_voxPitch = NULL;
}

/**
 * デストラクタ
 */
SklCfgDomainInfo::~SklCfgDomainInfo()
{
  if( m_voxOrg ) delete [] m_voxOrg;
  if( m_voxSize ) delete [] m_voxSize;
  if( m_voxWidth ) delete [] m_voxWidth;
  if( m_voxPitch ) delete [] m_voxPitch;
}

bool SklCfgDomainInfo::CheckNodeName(const char* nname)
{
  if( !nname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0001_);
    return false;
  }

  if( StrCaseCompare(nname,
                     _SKL_CFG_NNAME_DOMAININFO_,
                     strlen(_SKL_CFG_NNAME_DOMAININFO_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, nname);
    return false;
  }

  return true;
}

bool SklCfgDomainInfo::ParseVoxelOrigin(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_DOMAININFO_VOXEL_ORIGIN_,
                     strlen(_SKL_CFG_NNAME_DOMAININFO_VOXEL_ORIGIN_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlChar* tmpStr = NULL;
  float origin[3] = {0.0f, 0.0f, 0.0f};
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_XORG_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_ORIGIN_,
              _SKL_CFG_ANAME_DOMAININFO_XORG_);
    return false;
  }
  origin[0] = (float)atof((const char*)tmpStr);
  xmlFree(tmpStr);
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_YORG_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_ORIGIN_,
              _SKL_CFG_ANAME_DOMAININFO_YORG_);
    return false;
  }
  origin[1] = (float)atof((const char*)tmpStr);
  xmlFree(tmpStr);
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_ZORG_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_ORIGIN_,
              _SKL_CFG_ANAME_DOMAININFO_ZORG_);
    return false;
  }
  origin[2] = (float)atof((const char*)tmpStr);
  xmlFree(tmpStr);

  if( m_voxOrg ){
    CfgWarnMsg("element \"%s\" is redefined. This \"%s=(%f,%f,%f)\" is disregarded.",
               _SKL_CFG_NNAME_DOMAININFO_VOXEL_ORIGIN_,
               _SKL_CFG_NNAME_DOMAININFO_VOXEL_ORIGIN_,
               origin[0], origin[1], origin[2]);
    return true;
  }

  if( !(m_voxOrg = new float[3]) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  m_voxOrg[0] = origin[0];
  m_voxOrg[1] = origin[1];
  m_voxOrg[2] = origin[2];

  return true;
}

/**
 * XML要素からボクセルサイズをパースする。
 * @param node      XML要素
 * @return      true=success, false=fault
 */
bool SklCfgDomainInfo::ParseVoxelSize(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_DOMAININFO_VOXEL_SIZE_,
                     strlen(_SKL_CFG_NNAME_DOMAININFO_VOXEL_SIZE_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlChar* tmpStr = NULL;
  unsigned int  size[3] = {0, 0, 0};
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_IMAX_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_SIZE_,
              _SKL_CFG_ANAME_DOMAININFO_IMAX_);
    return false;
  }
  size[0] = (unsigned int)atoi((const char*)tmpStr);
  xmlFree(tmpStr);
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_JMAX_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_SIZE_,
              _SKL_CFG_ANAME_DOMAININFO_JMAX_);
    return false;
  }
  size[1] = (unsigned int)atoi((const char*)tmpStr);
  xmlFree(tmpStr);
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_KMAX_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_SIZE_,
              _SKL_CFG_ANAME_DOMAININFO_KMAX_);
    return false;
  }
  size[2] = (unsigned int)atoi((const char*)tmpStr);
  xmlFree(tmpStr);

  if( m_voxSize ){
    CfgWarnMsg("element \"%s\" is redefined. This \"%s=(%d,%d,%d)\" is disregarded.",
               _SKL_CFG_NNAME_DOMAININFO_VOXEL_SIZE_,
               _SKL_CFG_NNAME_DOMAININFO_VOXEL_SIZE_,
               size[0], size[1], size[2]);
    return true;
  }

  if( !(m_voxSize = new unsigned int[3]) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  m_voxSize[0] = size[0];
  m_voxSize[1] = size[1];
  m_voxSize[2] = size[2];

  return true;
}

/**
 * XML要素からボクセル幅をパースする。
 * @param node      XML要素
 * @return      true=success, false=fault
 */
bool SklCfgDomainInfo::ParseVoxelWidth(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_DOMAININFO_VOXEL_WIDTH_,
                     strlen(_SKL_CFG_NNAME_DOMAININFO_VOXEL_WIDTH_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlChar* tmpStr = NULL;
  float width[3] = {0.0f, 0.0f, 0.0f};
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_XWID_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_WIDTH_,
              _SKL_CFG_ANAME_DOMAININFO_XWID_);
    return false;
  }
  width[0] = (float)atof((const char*)tmpStr);
  xmlFree(tmpStr);
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_YWID_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_WIDTH_,
              _SKL_CFG_ANAME_DOMAININFO_YWID_);
    return false;
  }
  width[1] = (float)atof((const char*)tmpStr);
  xmlFree(tmpStr);
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_ZWID_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_WIDTH_,
              _SKL_CFG_ANAME_DOMAININFO_ZWID_);
    return false;
  }
  width[2] = (float)atof((const char*)tmpStr);
  xmlFree(tmpStr);

  if( m_voxWidth ){
    CfgWarnMsg("element \"%s\" is redefined. This \"%s=(%f,%f,%f)\" is disregarded.",
               _SKL_CFG_NNAME_DOMAININFO_VOXEL_WIDTH_,
               _SKL_CFG_NNAME_DOMAININFO_VOXEL_WIDTH_,
               width[0], width[1], width[2]);
    return true;
  }

  if( !(m_voxWidth = new float[3]) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  m_voxWidth[0] = width[0];
  m_voxWidth[1] = width[1];
  m_voxWidth[2] = width[2];

  return true;
}

/**
 * XML要素からピッチをパースする。
 * @param node      XML要素
 * @return      true=success, false=fault
 */
bool SklCfgDomainInfo::ParseVoxelPitch(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_DOMAININFO_VOXEL_PITCH_,
                     strlen(_SKL_CFG_NNAME_DOMAININFO_VOXEL_PITCH_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlChar* tmpStr = NULL;
  float pitch[3] = {0.0f, 0.0f, 0.0f};
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_XPIT_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_PITCH_,
              _SKL_CFG_ANAME_DOMAININFO_XPIT_);
    return false;
  }
  pitch[0] = (float)atof((const char*)tmpStr);
  xmlFree(tmpStr);
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_YPIT_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_PITCH_,
              _SKL_CFG_ANAME_DOMAININFO_YPIT_);
    return false;
  }
  pitch[1] = (float)atof((const char*)tmpStr);
  xmlFree(tmpStr);
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_DOMAININFO_ZPIT_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_DOMAININFO_VOXEL_PITCH_,
              _SKL_CFG_ANAME_DOMAININFO_ZPIT_);
    return false;
  }
  pitch[2] = (float)atof((const char*)tmpStr);
  xmlFree(tmpStr);

  if( m_voxPitch ){
    CfgWarnMsg("element \"%s\" is redefined. This \"%s=(%f,%f,%f)\" is disregarded.",
               _SKL_CFG_NNAME_DOMAININFO_VOXEL_PITCH_,
               _SKL_CFG_NNAME_DOMAININFO_VOXEL_PITCH_,
               pitch[0], pitch[1], pitch[2]);
    return true;
  }

  if( !(m_voxPitch = new float[3]) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  m_voxPitch[0] = pitch[0];
  m_voxPitch[1] = pitch[1];
  m_voxPitch[2] = pitch[2];

  return true;
}

bool SklCfgDomainInfo::ParseSklDefNode(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !StrCaseCompare((const char*)node->name,
                      _SKL_CFG_NNAME_DOMAININFO_VOXEL_ORIGIN_,
                      strlen(_SKL_CFG_NNAME_DOMAININFO_VOXEL_ORIGIN_)) ){
    if( !ParseVoxelOrigin(node) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
  }
  else if( !StrCaseCompare((const char*)node->name,
                           _SKL_CFG_NNAME_DOMAININFO_VOXEL_SIZE_,
                           strlen(_SKL_CFG_NNAME_DOMAININFO_VOXEL_SIZE_)) ){
    if( !ParseVoxelSize(node) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
  }
  else if( !StrCaseCompare((const char*)node->name,
                           _SKL_CFG_NNAME_DOMAININFO_VOXEL_WIDTH_,
                           strlen(_SKL_CFG_NNAME_DOMAININFO_VOXEL_WIDTH_)) ){
    if( !ParseVoxelWidth(node) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
  }
  else if( !StrCaseCompare((const char*)node->name,
                           _SKL_CFG_NNAME_DOMAININFO_VOXEL_PITCH_,
                           strlen(_SKL_CFG_NNAME_DOMAININFO_VOXEL_PITCH_)) ){
    if( !ParseVoxelPitch(node) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
  }

  return true;
}

bool SklCfgDomainInfo::Verification()
{
  if( !m_voxSize ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgWarnMsg("VoxelSize is not set.");
  }
  if( !m_voxOrg ){
    CfgWarnMsg("VoxelOrigin is not set.");
  }
  if( m_voxWidth && m_voxPitch ){
    CfgWarnMsg("Both the VoxelPitch and VoxelWidth are set.");
  }
  else if( !m_voxWidth && !m_voxPitch ) {
    CfgWarnMsg("Both the VoxelPitch and VoxelWidth are not set.");
  }
  return true;
}
/* 20070115
bool SklCfgDomainInfo::Verification()
{
  if( !m_voxSize ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0006_, "VoxelSize is not set.");
    return false;
  }

  if( !m_voxOrg ){
    CfgWarnMsg("VoxelOrigin is not set.");
  }

  if( !m_voxWidth && m_voxPitch ) {
    m_voxWidth = new float[3];
    m_voxWidth[0] = m_voxPitch[0] * (m_voxSize[0] - 1);
    m_voxWidth[1] = m_voxPitch[1] * (m_voxSize[1] - 1);
    m_voxWidth[2] = m_voxPitch[2] * (m_voxSize[2] - 1);
  }
  else if( m_voxWidth && !m_voxPitch ) {
    m_voxPitch = new float[3];
    m_voxPitch[0] = m_voxWidth[0] / (m_voxSize[0] - 1);
    m_voxPitch[1] = m_voxWidth[1] / (m_voxSize[1] - 1);
    m_voxPitch[2] = m_voxWidth[2] / (m_voxSize[2] - 1);
  }
  else if( m_voxWidth && m_voxPitch ){
    CfgWarnMsg("Both the VoxelPitch and VoxelWidth are set.");
    CfgWarnMsg("VoxelPitch information is given to priority.");
    CfgWarnMsg("VoxelWidth is generated from VoxelPitch and VoxelSize.");
    m_voxWidth[0] = m_voxPitch[0] * (m_voxSize[0] - 1);
    m_voxWidth[1] = m_voxPitch[1] * (m_voxSize[1] - 1);
    m_voxWidth[2] = m_voxPitch[2] * (m_voxSize[2] - 1);
  }
  else if( !m_voxWidth && !m_voxPitch ) {
    CfgWarnMsg("Both the VoxelPitch and VoxelWidth are not set.");
  }

  return true;
}
*/


/**
 * 原点座標を取得する。
 * @param[out] x    X方向原点座標
 * @param[out] y    Y方向原点座標
 * @param[out] z    Z方向原点座標
 * @return      true=success, false=fault
 */
bool SklCfgDomainInfo::GetVoxelOrigin(float& x, float& y, float& z) const
{
  if( !m_voxOrg ) {
    // delete 2008.01.07
    // CfgErrMsg(_SKL_CFG_ERR_0050_);
    return false;
  }
  x = m_voxOrg[0];
  y = m_voxOrg[1];
  z = m_voxOrg[2];
  return true;
}

/**
 * ボクセルサイズを取得する。
 * @param[out] i    I方向ボクセルサイズ
 * @param[out] j    J方向ボクセルサイズ
 * @param[out] k    K方向ボクセルサイズ
 * @return      true=success, false=fault
 */
bool SklCfgDomainInfo::GetVoxelSize(unsigned int& i,
                                    unsigned int& j,
                                    unsigned int& k) const
{
  if( !m_voxSize ) {
    CfgErrMsg(_SKL_CFG_ERR_0051_);
    return false;
  }
  i = m_voxSize[0];
  j = m_voxSize[1];
  k = m_voxSize[2];
  return true;
}


/**
 * ボクセル幅を取得する。
 * @param[out] x    X方向ボクセル幅
 * @param[out] y    Y方向ボクセル幅
 * @param[out] z    Z方向ボクセル幅
 * @return      true=success, false=fault
 */
bool SklCfgDomainInfo::GetVoxelWidth(float& x, float& y, float& z) const
{
  if( !m_voxWidth ) {
    // delete 2008.01.07
    // CfgErrMsg(_SKL_CFG_ERR_0052_);
    return false;
  }
  x = m_voxWidth[0];
  y = m_voxWidth[1];
  z = m_voxWidth[2];
  return true;
}

/**
 * ピッチを取得する。
 * @param[out] x       X方向ピッチ
 * @param[out] y       Y方向ピッチ
 * @param[out] z       Z方向ピッチ
 * @return      true=success, false=fault
 */
bool SklCfgDomainInfo::GetVoxelPitch(float& x, float& y, float& z) const
{
  if( !m_voxPitch ) {
    // delete 2008.01.07
    // CfgErrMsg(_SKL_CFG_ERR_0053_);
    return false;
  }
  x = m_voxPitch[0];
  y = m_voxPitch[1];
  z = m_voxPitch[2];
  return true;
}

#ifdef DEBUG
void SklCfgDomainInfo::Print()
{
  cout << " <-- DomainInfo --" << endl << flush;

  if( m_voxOrg ){
    cout << "  voxel origin : "
         << m_voxOrg[0] << ", " << m_voxOrg[1] << ", " << m_voxOrg[2]
         << endl << flush;
  }else{
    cout << "  voxel origin not set." << endl << flush;
  }

  if( m_voxSize ){
    cout << "  voxel size   : "
         << m_voxSize[0] << ", " << m_voxSize[1] << ", " << m_voxSize[2]
         << endl << flush;
  }else{
    cout << "  voxel size not set." << endl << flush;
  }

  if( m_voxWidth ){
    cout << "  voxel width  : "
         << m_voxWidth[0] << ", " << m_voxWidth[1] << ", " << m_voxWidth[2]
         << endl << flush;
  }else{
    cout << "  voxel width not set." << endl << flush;
  }

  if( m_voxPitch ){
    cout << "  voxel pitch  : "
         << m_voxPitch[0] << ", " << m_voxPitch[1] << ", " << m_voxPitch[2]
         << endl << flush;
  }else{
    cout << "  voxel pitch not set." << endl << flush;
  }

  cout << "  -- DomainInfo -->" << endl << flush;

}
#endif //DEBUG

} // end of namespace SklCfg

