/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */
/**
 * @file SklCfgHistory.C
 * コンフィグレーションHistory要素クラスソースファイル
 * @author     riken
 * @date       2010/10/01
 */

#include "skl_config/SklCfgDefName.h"
#include "skl_config/SklCfgHistory.h"
#include "skl_config/vfvPathUtil.h"

namespace SklCfg {

/**
 * コンストラクタ
 */
SklCfgHistory::SklCfgHistory()
{
  m_interval = 0;
  m_fname = NULL;
  m_attr = NULL;
}

/**
 * デストラクタ
 */
SklCfgHistory::~SklCfgHistory()
{
  if( m_fname ) delete [] m_fname;
  if( m_attr ) delete [] m_attr;
}

/**
 * 要素名が一致しているかチェックする。
 * @param nname     要素名
 * @return          true=一致
 */
bool SklCfgHistory::CheckNodeName(const char* nname)
{
  if( !nname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0001_);
    return false;
  }

  if( StrCaseCompare(nname,
                     _SKL_CFG_NNAME_STEER_HISTORY_,
                     strlen(_SKL_CFG_NNAME_STEER_HISTORY_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, nname);
    return false;
  }

  return true;
}

bool SklCfgHistory::ParseNode(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !CheckNodeName((const char*)node->name) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  if( !ParseSklDefNode(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  if( !Verification() ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  return true;
}

bool SklCfgHistory::ParseSklDefNode(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !CheckNodeName((const char*)node->name) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  if( !ParseInterval(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }
  if( !ParseFileName(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }
  if( !ParseAttr(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }
  return true;
}

bool SklCfgHistory::ParseInterval(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  xmlChar* tmpStr;
  tmpStr = xmlGetProp(node, (const xmlChar*)_SKL_CFG_ANAME_STEER_INTERVAL_);
  if( !tmpStr ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_STEER_HISTORY_,
              _SKL_CFG_ANAME_STEER_INTERVAL_);
    return false;
  }
  m_interval = (unsigned int)atoi((const char*)tmpStr);
  xmlFree(tmpStr);

  return true;
}

bool SklCfgHistory::ParseFileName(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  xmlChar* tmpStr;
  tmpStr = xmlGetProp(node, (const xmlChar*)_SKL_CFG_ANAME_STEER_FILE_NAME_);
  if( !tmpStr ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_STEER_HISTORY_,
              _SKL_CFG_ANAME_STEER_FILE_NAME_);
    return false;
  }
  if( m_fname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0008_);
    return false;
  }
  size_t len = strlen((const char*)tmpStr);
  if( !(m_fname = new char[len + 1]) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  memset(m_fname, 0, sizeof(char)*(len + 1));
  strncpy(m_fname, (const char*)tmpStr, len+1);
  xmlFree(tmpStr);

  return true;
}

bool SklCfgHistory::ParseAttr(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  xmlChar* tmpStr;
  tmpStr = xmlGetProp(node, (const xmlChar*)_SKL_CFG_ANAME_STEER_ATTR_);
  if( !tmpStr ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_STEER_HISTORY_,
              _SKL_CFG_ANAME_STEER_ATTR_);
    return false;
  }
  if( m_attr ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0034_);
    return false;
  }
  size_t len = strlen((const char*)tmpStr);
  if( !(m_attr = new char[len + 1]) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  memset(m_attr, 0, sizeof(char)*(len + 1));
  strncpy(m_attr, (const char*)tmpStr, len+1);
  xmlFree(tmpStr);

  return true;
}

bool SklCfgHistory::CreateFileName(const char* basedir)
{
  if( !m_fname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0033_);
    return false;
  }
  if( !basedir ){
    string strfname(m_fname);
    strfname = VFV::vfvPath_normalize(strfname);
    char* tmpfname = new char[strfname.size() + 1];
    if( !tmpfname ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0005_);
      return false;
    }
    strcpy(tmpfname, strfname.c_str());
    delete [] m_fname;
    m_fname = tmpfname;
    return true;
  }

  string strfname(m_fname);
  string strbasedir(basedir);
  string path;
#ifndef WIN32
  path.clear();
#endif //WIN32

  if( !VFV::vfvPath_isAbsolute(strfname) ){
    path = strbasedir;
    path += VFV::vfvPath_getDelimChar();
    path += strfname;
  }
  else path = strfname;
  path = VFV::vfvPath_normalize(path);

  char* tmpfname = new char[path.size() + 1];
  if( !tmpfname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  strcpy(tmpfname, path.c_str());
  delete [] m_fname;
  m_fname = tmpfname;

  return true;
}

bool SklCfgHistory::Verification()
{
  if( !m_fname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0006_, "History File Name");
    return false;
  }
  if( m_interval == 0 ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0006_, "History Interval is zero");
    return false;
  }
  return true;
}

unsigned int SklCfgHistory::GetInterval() const
{
  return m_interval;
}

const char* SklCfgHistory::GetFileName() const
{
  return (const char*)m_fname;
}

const char* SklCfgHistory::GetAttr() const
{
  return (const char*)m_attr;
}

#ifdef DEBUG
void SklCfgHistory::Print()
{
  cout << "   <-- History -- : " << flush;

  if( m_interval != 0 ){
    cout << "(interval)" << m_interval << ", " << flush;
  }else{
    cout << "(interval) not set, " << flush;
  }
  if( m_fname ){
    cout << "(basename)" << m_fname << ", " << flush;
  }else{
    cout << "(basename) not set, " << flush;
  }
  cout << "-->" << endl << flush;
}
#endif //DEBUG

} // end of namespace SklCfg

