/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */
/**
 * @file SklCfgInFile.C
 * コンフィグレーションInFile要素クラスソースファイル
 * @author     riken
 * @date       2010/10/01
 */

#include "skl_config/SklCfgDefName.h"
#include "skl_config/SklCfgInFile.h"
#include "skl_config/vfvPathUtil.h"

namespace SklCfg {
/**
 * コンストラクタ
 */
SklCfgInFile::SklCfgInFile()
{
  m_attr = NULL;
  m_format = NULL;
  m_fname = NULL;
  m_multiIn = false;
  m_coordID = -1;
}

/**
 * デストラクタ
 */
SklCfgInFile::~SklCfgInFile()
{
  if( m_attr ) delete [] m_attr;
  if( m_format ) delete [] m_format;
  if( m_fname ) delete [] m_fname;
}

bool SklCfgInFile::CheckNodeName(const char* nname)
{
  if( !nname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0001_);
    return false;
  }

  if( StrCaseCompare(nname,
                     _SKL_CFG_NNAME_STEER_IN_FILE_,
                     strlen(_SKL_CFG_NNAME_STEER_IN_FILE_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, nname);
    return false;
  }

  return true;
}

bool SklCfgInFile::ParseNode(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !CheckNodeName((const char*)node->name) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  if( !ParseSklDefNode(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  if( !Verification() ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  return true;
}

bool SklCfgInFile::ParseSklDefNode(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !CheckNodeName((const char*)node->name) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  if( !ParseAttr(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }
  if( !ParseFormat(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }
  if( !ParseFileName(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }
  if( !ParseMultiInput(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }
  if( !ParseCoordID(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  return true;
}

/**
 * XML要素からファイル属性をパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgInFile::ParseAttr(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  xmlChar* tmpStr;
  tmpStr = xmlGetProp(node, (const xmlChar*)_SKL_CFG_ANAME_STEER_ATTR_);
  if( !tmpStr ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_STEER_IN_FILE_,
              _SKL_CFG_ANAME_STEER_ATTR_);
    return false;
  }
  if( m_attr ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0009_);
    return false;
  }
  size_t len = strlen((const char*)tmpStr);
  if( !(m_attr = new char[len + 1]) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  memset(m_attr, 0, sizeof(char)*(len + 1));
  strncpy(m_attr, (const char*)tmpStr, len+1);
  xmlFree(tmpStr);

  return true;
}

/**
 * XML要素からファイルフォーマットをパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgInFile::ParseFormat(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  xmlChar* tmpStr;
  tmpStr = xmlGetProp(node, (const xmlChar*)_SKL_CFG_ANAME_STEER_FORMAT_);
  if( !tmpStr ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_STEER_IN_FILE_,
              _SKL_CFG_ANAME_STEER_FORMAT_);
    return false;
  }
  if( m_format ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0010_);
    return false;
  }
  size_t len = strlen((const char*)tmpStr);
  if( !(m_format = new char[len + 1]) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  memset(m_format, 0, sizeof(char)*(len + 1));
  strncpy(m_format, (const char*)tmpStr, len+1);
  xmlFree(tmpStr);

  return true;
}

/**
 * XML要素からファイル名をパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgInFile::ParseFileName(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  xmlChar* tmpStr;
  tmpStr = xmlGetProp(node, (const xmlChar*)_SKL_CFG_ANAME_STEER_FILE_NAME_);
  if( !tmpStr ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_STEER_IN_FILE_,
              _SKL_CFG_ANAME_STEER_FILE_NAME_);
    return false;
  }
  if( m_fname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0011_);
    return false;
  }
  size_t len = strlen((const char*)tmpStr);
  if( !(m_fname = new char[len + 1]) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  memset(m_fname, 0, sizeof(char)*(len + 1));
  strncpy(m_fname, (const char*)tmpStr, len+1);
  xmlFree(tmpStr);

  return true;
}

/**
 * XML要素から分散入力フラグをパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgInFile::ParseMultiInput(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  xmlChar* tmpStr;
  tmpStr = xmlGetProp(node, (const xmlChar*)_SKL_CFG_ANAME_STEER_MULTI_IO_);
  if( tmpStr ){
    if( !strcasecmp((char*)tmpStr, "yes") ) m_multiIn = true;
  }
  xmlFree(tmpStr);

  return true;
}

/**
 * XML要素からcoordIDをパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgInFile::ParseCoordID(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  xmlChar* tmpStr;
  tmpStr = xmlGetProp(node, (const xmlChar*)_SKL_CFG_ANAME_STEER_COORD_ID_);
  if( tmpStr ){
    m_coordID = atoi((const char*)tmpStr);
  }
  xmlFree(tmpStr);

  return true;
}

/**
 * ファイル名にディレクトリパスを付加する。
 * @param basedir    ディレクトリパス
 * @return      true=success, false=fault
 */
bool SklCfgInFile::CreateFileName(const char* basedir)
{
  if( !m_fname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0033_);
    return false;
  }
  if( !basedir ){
    string strfname(m_fname);
    strfname = VFV::vfvPath_normalize(strfname);
    char* tmpfname = new char[strfname.size() + 1];
    if( !tmpfname ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0005_);
      return false;
    }
    strcpy(tmpfname, strfname.c_str());
    delete [] m_fname;
    m_fname = tmpfname;
    return true;
  }

  string strfname(m_fname);
  string strbasedir(basedir);
  string path;
#ifndef WIN32
  path.clear();
#endif //WIN32

  if( !VFV::vfvPath_isAbsolute(strfname) ){
    path = strbasedir;
    path += VFV::vfvPath_getDelimChar();
    path += strfname;
  }
  else path = strfname;
  path = VFV::vfvPath_normalize(path);

  char* tmpfname = new char[path.size() + 1];
  if( !tmpfname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  strcpy(tmpfname, path.c_str());
  delete [] m_fname;
  m_fname = tmpfname;

  return true;
}

bool SklCfgInFile::Verification()
{
  if( !m_fname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0006_, "InFile File Name");
    return false;
  }
  if( !m_attr ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0006_, "InFile Attribute");
    return false;
  }
  if( !m_format ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0006_, "InFile Format");
    return false;
  }

  return true;
}

/**
 * 分散入力フラグを"true"に設定する。
 */
void SklCfgInFile::SetMultiInput()
{
  m_multiIn = true;
}


/**
 * 分散入力フラグを"false"に設定する。
 */
void SklCfgInFile::UnsetMultiInput()
{
  m_multiIn = false;
}

/**
 * ファイル属性を取得する。
 * @return    ファイル属性
 */
const char* SklCfgInFile::GetAttr() const
{
  return (const char*)m_attr;
}

/**
 * ファイルフォーマットを取得する。
 * @return      ファイルフォーマット
 */
const char* SklCfgInFile::GetFormat() const
{
  return (const char*)m_format;
}

/**
 * ファイルタイプを取得する。
 * @return    ファイルタイプ
 */
SklFileType SklCfgInFile::GetFileType() const
{
  if( !m_format ) return SKL_FILE_TYPE_UNKOWN;

  if( !strcasecmp(m_format, _SKL_CFG_ANAME_STEER_FILE_TYPE_SPH) ) return SKL_SPH;
  else if( !strcasecmp(m_format, _SKL_CFG_ANAME_STEER_FILE_TYPE_SVX) ) return SKL_SVX;
  else if( !strcasecmp(m_format, _SKL_CFG_ANAME_STEER_FILE_TYPE_OTV) ) return SKL_OTV;
  else if( !strcasecmp(m_format, _SKL_CFG_ANAME_STEER_FILE_TYPE_OVX) ) return SKL_OVX;
  else if( !strcasecmp(m_format, _SKL_CFG_ANAME_STEER_FILE_TYPE_FDV) ) return SKL_FDV;
  else if( !strcasecmp(m_format, _SKL_CFG_ANAME_STEER_FILE_TYPE_NBF) ) return SKL_NBF;
  else if( !strcasecmp(m_format, _SKL_CFG_ANAME_STEER_FILE_TYPE_NVX) ) return SKL_NVX;
  else if( !strcasecmp(m_format, _SKL_CFG_ANAME_STEER_FILE_TYPE_VCR) ) return SKL_VCAR;
  else return SKL_FILE_TYPE_UNKOWN;
}

/**
 * ファイル名を取得する。
 * @return    ファイル名
 */
const char* SklCfgInFile::GetFileName() const
{
  return (const char*)m_fname;
}

/**
 * 分散入力フラグを取得する。
 * @return    分散入力フラグ
 */
bool SklCfgInFile::IsMultiInput() const
{
  return m_multiIn;
}

/**
 * coordIDを取得する。
 * @return    coordID
 */
int SklCfgInFile::GetCoordID() const
{
  return m_coordID;
}

/**
 * InFile要素の属性を取得する。（ファイルフォーマット）
 * @param[out] attr      ファイル属性
 * @param[out] format    ファイルフォーマット
 * @param[out] fname     ファイル名
 * @return      true=success, false=fault
 */
bool SklCfgInFile::GetData(const char** attr,
                           const char** format,
                           const char** fname) const
{
  if( !attr || !format || !fname ) {
    CfgErrMsg(_SKL_CFG_ERR_0054_,
              "SklCfgInFile::GetData(attr, format, fname) const");
    if( !attr )   CfgErrMsg(_SKL_CFG_ERR_0055_, "attr is NULL");
    if( !format ) CfgErrMsg(_SKL_CFG_ERR_0055_, "format is NULL");
    if( !fname )  CfgErrMsg(_SKL_CFG_ERR_0055_, "fname is NULL");
    return false;
  }
  *attr = m_attr;
  *format = m_format;
  *fname = m_fname;
  return true;
}

/**
 * InFile要素の属性を取得する。（ファイルタイプ）
 * @param[out] attr      ファイル属性
 * @param[out] format    ファイルタイプ
 * @param[out] fname     ファイル名
 * @return      true=success, false=fault
 */
bool SklCfgInFile::GetData(const char** attr,
                           SklFileType* format,
                           const char** fname) const
{
  if( !attr || !format || !fname ) {
    CfgErrMsg(_SKL_CFG_ERR_0054_,
              "SklCfgInFile::GetData(attr, format, fname) const");
    if( !attr )   CfgErrMsg(_SKL_CFG_ERR_0055_, "attr is NULL");
    if( !format ) CfgErrMsg(_SKL_CFG_ERR_0055_, "format is NULL");
    if( !fname )  CfgErrMsg(_SKL_CFG_ERR_0055_, "fname is NULL");
    return false;
  }
  const char* fmt;
  if( !GetData(attr, &fmt, fname) ) return false;
  *format = GetFileType();
  return true;
}

#ifdef DEBUG
void SklCfgInFile::Print()
{
  cout << "   <-- InFile -- : " << flush;

  if( m_attr ){
    cout << "(attr)" << m_attr << ", " << flush;
  }else{
    cout << "(attr) not set, " << flush;
  }
  if( m_format ){
    cout << "(format)" << m_format << ", " << flush;
  }else{
    cout << "(format) not set, " << flush;
  }
  if( m_fname ){
    cout << "(fname)" << m_fname << ", " << flush;
  }else{
    cout << "(fname) not set, " << flush;
  }
  cout << "-->" << endl << flush;
}
#endif //DEBUG

} // end of namespace SklCfg

