/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */
/**
 * @file SklCfgNodeBase.C
 * コンフィグレーション要素基底クラスソースファイル
 * @author     riken
 * @date       2010/10/01
 */

#include <string.h> // add 20090629
#include "skl_config/SklCfgDefName.h"
#include "skl_config/SklCfgNodeBase.h"

namespace SklCfg {

/**
 * コンストラクタ
 */
SklCfgNodeBase::SklCfgNodeBase()
{
}

/**
 * デストラクタ
 */
SklCfgNodeBase::~SklCfgNodeBase()
{

}

/**
 * XML要素をコンフィグレーションクラスにパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgNodeBase::ParseNode(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !Parse(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  if( !Verification() ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  return true;
}

/**
 * XML要素をコンフィグレーションクラスにパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgNodeBase::Parse(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !CheckNodeName((const char*)node->name) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  xmlNodePtr cur = node->xmlChildrenNode;
  while ( cur ) {
    if ( cur->type == XML_COMMENT_NODE ) {
      cur = cur->next;
      continue;
    }

    if( !StrCaseCompare((const char*)cur->name,
                        _SKL_CFG_NNAME_USRDEF_ELEM_,
                        strlen(_SKL_CFG_NNAME_USRDEF_ELEM_)) ){
      // Parse user defined element(CfgElem)
      CfgElem* elem = new CfgElem;
      if( !elem ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !elem->SetAttr(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        delete elem;
        return false;
      }
      if( !m_root.AddElem(elem) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        delete elem;
        return false;
      }
    }
    else if( !StrCaseCompare((const char*)cur->name,
                             _SKL_CFG_NNAME_USRDEF_PARAM_,
                             strlen(_SKL_CFG_NNAME_USRDEF_PARAM_)) ){
      // Parse user defined element(CfgParam)
      CfgParam* param = new CfgParam;
      if( !param ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !param->SetAttr(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        delete param;
        return false;
      }
      if( !m_root.AddParam(param) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        delete param;
        return false;
      }
    }
    else {
      //Parse SPHERE Defined element
      if( !ParseSklDefNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }

    cur = cur->next;
  }

  return true;
}


/**
 * Param要素を追加する。
 * @param param    Param要素
 * @return      true=success, false=fault
 */
bool SklCfgNodeBase::AddParam(CfgParam* param)
{
  if( !param ) return false;
  m_root.AddParam(param);
  return true;
}

/**
 * Elem要素を追加する。
 * @param elem    elem要素
 * @return      true=success, false=fault
 */
bool SklCfgNodeBase::AddElem (CfgElem* elem )
{
  if( !elem ) return false;
  m_root.AddElem(elem);
  return true;
}


/**
 * 子要素をクリアする。
 */
void SklCfgNodeBase::ClearRoot() {
  m_root.ClearList();
}

/**
 * 子要素クラスを取得する。
 * @return      子要素クラス
 */
const CfgElem* SklCfgNodeBase::GetRootElem() const
{
  return &m_root;
}

/**
 * 子要素クラスの要素数を取得する。
 * @return      子要素数
 */
unsigned int SklCfgNodeBase::GetSize() const
{
  return m_root.GetSize();
}

/**
 * 子要素クラスのElem要素数を取得する。
 * @return      Elem要素数
 */
unsigned int SklCfgNodeBase::GetElemSize() const
{
  return m_root.GetElemSize();
}

/**
 * 子要素クラスのParam要素数を取得する。
 * @return      Param要素数
 */
unsigned int SklCfgNodeBase::GetParamSize() const
{
  return m_root.GetParamSize();
}

/**
 * 子要素の最初のElem要素を取得する。(name検索)
 * @param name    検索name値
 * @return      Elem要素
 */
const CfgElem* SklCfgNodeBase::GetElemFirst(const char* name) const
{
  return m_root.GetElemFirst(name);
}

/**
 * 子要素の最初のElem要素を取得する。(name, id, comment検索)
 * @param name    検索name値
 * @param id      検索id値
 * @param cmt     検索comment値
 * @return        Elem要素
 */
const CfgElem* SklCfgNodeBase::GetElemFirst(const char* name,
                                            int id, const char* cmt) const
{
  return m_root.GetElemFirst(name, id, cmt);
}

/**
 * 子要素の次のElem要素を取得する。(name検索)
 * @param elem      前のElem要素
 * @param name      検索name値
 * @return          Elem要素
 */
const CfgElem* SklCfgNodeBase::GetElemNext(const CfgElem* elem,
                                           const char* name) const
{
  if( !elem ) return NULL;
  return m_root.GetElemNext(elem, name);
}


/**
 * 子要素の次のElem要素を取得する。(name, id, comment検索)
 * @param elem      前のElem要素
 * @param name    検索name値
 * @param id      検索id値
 * @param cmt     検索comment値
 * @return        Elem要素
 */
const CfgElem* SklCfgNodeBase::GetElemNext(const CfgElem* elem,
                                           const char* name,
                                           int id,
                                           const char* cmt) const
{
  if( !elem ) return NULL;
  return m_root.GetElemNext(elem, name, id, cmt);
}

/**
 * 子要素の最初のParam要素を取得する。(name検索)
 * @param name    検索name値
 * @return      Param要素
 */
const CfgParam* SklCfgNodeBase::GetParamFirst(const char* name) const
{
  return m_root.GetParamFirst(name);
}

/**
 * 子要素の最初のParam要素を取得する。(name, id, comment検索)
 * @param name    検索name値
 * @param id      検索id値
 * @param cmt     検索comment値
 * @return        Param要素
 */
const CfgParam* SklCfgNodeBase::GetParamFirst(const char* name,
                                            int id, const char* cmt) const
{
  return m_root.GetParamFirst(name, id, cmt);
}

/**
 * 子要素の次のParam要素を取得する。(name検索)
 * @param param      前のParam要素
 * @param name      検索name値
 * @return          Param要素
 */
const CfgParam* SklCfgNodeBase::GetParamNext(const CfgParam* param,
                                             const char* name) const
{
  if( !param ) return NULL;
  return m_root.GetParamNext(param, name);
}

/**
 * 子要素の次のParam要素を取得する。(name, id, comment検索)
 * @param param      前のParam要素
 * @param name    検索name値
 * @param id      検索id値
 * @param cmt     検索comment値
 * @return        Param要素
 */
const CfgParam* SklCfgNodeBase::GetParamNext(const CfgParam* param,
                                             const char* name,
                                             int id, const char* cmt) const
{
  if( !param ) return NULL;
  return m_root.GetParamNext(param, name, id, cmt);
}

} // end of namespace SklCfg

