/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */
/**
 * @file SklCfgSteer.C
 * コンフィグレーションSteer要素クラスソースファイル
 * @author     riken
 * @date       2010/10/01
 */

#include <string.h> // add 20090629
#include "skl_config/SklCfgDefName.h"
#include "skl_config/SklCfgSteer.h"

namespace SklCfg {
/**
 * コンストラクタ
 */
SklCfgSteer::SklCfgSteer()
{
  m_calcSteps = 1000;
  m_calcStepsFlg = false;
  m_strType = Initial;
  m_strTypeFlg = false;
  m_chkModeFlg = false;
  m_chkModeFile = NULL;
  m_inputData = NULL;
  m_outputData = NULL;
}

/**
 * デストラクタ
 */
SklCfgSteer::~SklCfgSteer()
{
  if( m_chkModeFile ) delete [] m_chkModeFile;
  if( m_inputData ) delete m_inputData;
  if( m_outputData ) delete m_outputData;
}

bool SklCfgSteer::CheckNodeName(const char* nname)
{
  if( !nname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0001_);
    return false;
  }

  if( StrCaseCompare(nname,
                     _SKL_CFG_NNAME_STEER_,
                     strlen(_SKL_CFG_NNAME_STEER_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, nname);
    return false;
  }

  return true;
}

bool SklCfgSteer::ParseSklDefNode(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !StrCaseCompare((const char*)node->name,
                      _SKL_CFG_NNAME_STEER_CALC_STEPS_,
                      strlen(_SKL_CFG_NNAME_STEER_CALC_STEPS_)) ){
    if( !ParseCalcSteps(node) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
  }
  else if( !StrCaseCompare((const char*)node->name,
                           _SKL_CFG_NNAME_STEER_START_COND_,
                           strlen(_SKL_CFG_NNAME_STEER_START_COND_)) ){
    if( !ParseStartCondition(node) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
  }
  else if( !StrCaseCompare((const char*)node->name,
                           _SKL_CFG_NNAME_STEER_INPUT_DATA_,
                           strlen(_SKL_CFG_NNAME_STEER_INPUT_DATA_)) ){
    if( !ParseInputData(node) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
  }
  else if( !StrCaseCompare((const char*)node->name,
                           _SKL_CFG_NNAME_STEER_OUTPUT_DATA_,
                           strlen(_SKL_CFG_NNAME_STEER_OUTPUT_DATA_)) ){
    if( !ParseOutputData(node) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
  }
  else if( !StrCaseCompare((const char*)node->name,
                           _SKL_CFG_NNAME_STEER_CHECK_MODE_,
                           strlen(_SKL_CFG_NNAME_STEER_CHECK_MODE_)) ){
    if( !ParseCheckMode(node) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
  }

  return true;
}

/**
 * XML要素から計算ステップ数をパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgSteer::ParseCalcSteps(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_STEER_CALC_STEPS_,
                     strlen(_SKL_CFG_NNAME_STEER_CALC_STEPS_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlChar* tmpStr = NULL;
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_STEER_STEP_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_STEER_CALC_STEPS_,
              _SKL_CFG_ANAME_STEER_STEP_);
    return false;
  }

  if( m_calcStepsFlg ){
    CfgWarnMsg("element \"%s\" is redefined. \"%s=%d\" is disregarded.",
               _SKL_CFG_NNAME_STEER_CALC_STEPS_,
               _SKL_CFG_NNAME_STEER_CALC_STEPS_,
               atoi((const char*)tmpStr));
    xmlFree(tmpStr);
    return true;
  }

  m_calcSteps = (unsigned int)atoi((const char*)tmpStr);
  xmlFree(tmpStr);
  m_calcStepsFlg = true;

  return true;
}

/**
 * XML要素から開始条件をパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgSteer::ParseStartCondition(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_STEER_START_COND_,
                     strlen(_SKL_CFG_NNAME_STEER_START_COND_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlChar* tmpStr = NULL;
  if( !(tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_STEER_TYPE_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_STEER_START_COND_,
              _SKL_CFG_ANAME_STEER_TYPE_);
    return false;
  }

  if( m_strTypeFlg ){
    CfgWarnMsg("element \"%s\" is redefined. \"%s=%s\" is disregarded.",
               _SKL_CFG_NNAME_STEER_START_COND_,
               _SKL_CFG_NNAME_STEER_START_COND_,
               (const char*)tmpStr);
    xmlFree(tmpStr);
    return true;
  }

  if( !StrCaseCompare((const char*)tmpStr,
                      _SKL_CFG_ANAME_STEER_TYPE_INITIAL_,
                      strlen(_SKL_CFG_ANAME_STEER_TYPE_INITIAL_)) ){
    m_strType = Initial;
  }
  else if( !StrCaseCompare((const char*)tmpStr,
                           _SKL_CFG_ANAME_STEER_TYPE_RESTART_,
                           strlen(_SKL_CFG_ANAME_STEER_TYPE_RESTART_)) ){
    m_strType = Restart;
  }
  else {
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0007_,
              _SKL_CFG_NNAME_STEER_START_COND_,
              _SKL_CFG_ANAME_STEER_TYPE_,
              (const char*)tmpStr);
    xmlFree(tmpStr);
    return false;
  }
  xmlFree(tmpStr);
  m_strTypeFlg = true;

  return true;
}

/**
 * XML要素からInputData要素をパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgSteer::ParseInputData(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_STEER_INPUT_DATA_,
                     strlen(_SKL_CFG_NNAME_STEER_INPUT_DATA_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  if( m_inputData ){
    CfgWarnMsg("element \"%s\" is redefined. This \"%s\" is disregarded.",
               _SKL_CFG_NNAME_STEER_INPUT_DATA_,
               _SKL_CFG_NNAME_STEER_INPUT_DATA_);
    return true;
  }

  if( !(m_inputData = new SklCfgSteerInputData) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }

  if( !m_inputData->ParseNode(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  return true;
}

/**
 * XML要素からOutputData要素をパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgSteer::ParseOutputData(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_STEER_OUTPUT_DATA_,
                     strlen(_SKL_CFG_NNAME_STEER_OUTPUT_DATA_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  if( m_outputData ){
    CfgWarnMsg("element \"%s\" is redefined. This \"%s\" is disregarded.",
               _SKL_CFG_NNAME_STEER_OUTPUT_DATA_,
               _SKL_CFG_NNAME_STEER_OUTPUT_DATA_);
    return true;
  }

  if( !(m_outputData = new SklCfgSteerOutputData) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }

  if( !m_outputData->ParseNode(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  return true;
}

/**
 * XML要素からソルバー起動条件をパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklCfgSteer::ParseCheckMode(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_STEER_CHECK_MODE_,
                     strlen(_SKL_CFG_NNAME_STEER_CHECK_MODE_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  m_chkModeFlg = true;

  xmlChar* tmpStr = NULL;
  if( (tmpStr = xmlGetProp(node,
                 (const xmlChar*)_SKL_CFG_ANAME_STEER_FILE_NAME_)) ){
    if( m_chkModeFile ) delete [] m_chkModeFile;
    unsigned int len = strlen((char*)tmpStr);
    if( !(m_chkModeFile = new char[len+1]) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0005_);
      return false;
    }
    memset(m_chkModeFile, 0, sizeof(char)*(len+1));
    strcpy(m_chkModeFile, (char*)tmpStr);
  }

  return true;
}

bool SklCfgSteer::Verification()
{
  return true;
}

/**
 * 計算ステップ数を取得する。
 * @return    計算ステップ数
 */
unsigned int SklCfgSteer::GetCalcSteps() const
{
  return m_calcSteps;
}

/**
 * 開始条件を取得する。
 * @return    開始条件
 */
StartType SklCfgSteer::GetStartType() const
{
  return m_strType;
}

/**
 * InputData/InFile要素数を取得する。
 * @return    InFile要素数
 */
unsigned int SklCfgSteer::GetInFileSize() const
{
  if( !m_inputData ) return 0;
  return m_inputData->GetInFileSize();
}

/**
 * 最初のInputData/InFile要素を取得する。
 * @return    InFile要素クラス
 */
const SklCfgInFile* SklCfgSteer::GetInFileFirst() const
{
  if( !m_inputData ) return NULL;
  return m_inputData->GetInFileFirst();
}

/**
 * 次のInputData/InFile要素を取得する。
 * @param file    前のInFile要素クラス
 * @return    InFile要素クラス
 */
const SklCfgInFile*
      SklCfgSteer::GetInFileNext(const SklCfgInFile* file) const
{
  if( !file ) return NULL;
  if( !m_inputData ) return NULL;
  return m_inputData->GetInFileNext(file);
}

/**
 * 出力インターバルを取得する。
 * @return    出力インターバル
 */
unsigned int SklCfgSteer::GetMonitorInterval() const
{
  if( !m_outputData ) return 0;
  return m_outputData->GetMonitorInterval();
}

/**
 * OutputData/OutFile要素数を取得する。
 * @return    OutFile要素数
 */
unsigned int SklCfgSteer::GetOutFileSize() const
{
  if( !m_outputData ) return 0;
  return m_outputData->GetOutFileSize();
}

/**
 * 最初のOutputData/OutFile要素を取得する。
 * @return    OutFile要素クラス
 */
const SklCfgOutFile* SklCfgSteer::GetOutFileFirst() const
{
  if( !m_outputData ) return NULL;
  return m_outputData->GetOutFileFirst();
}

/**
 * 次のOutputData/OutFile要素を取得する。
 * @return    OutFile要素クラス
 */
const SklCfgOutFile*
      SklCfgSteer::GetOutFileNext(const SklCfgOutFile* file) const
{
  if( !file ) return NULL;
  if( !m_outputData ) return NULL;
  return m_outputData->GetOutFileNext(file);
}


/**
 * OutputData/History要素数を取得する。
 * @return    History要素数
 */
unsigned int SklCfgSteer::GetHistorySize() const
{
  if( !m_outputData ) return 0;
  return m_outputData->GetHistorySize();
}

/**
 * 最初のOutputData/History要素を取得する。
 * @return    History要素クラス
 */
const SklCfgHistory* SklCfgSteer::GetHistoryFirst() const
{
  if( !m_outputData ) return NULL;
  return m_outputData->GetHistoryFirst();
}

/**
 * 次のOutputData/History要素を取得する。
 * @return    History要素クラス
 */
const SklCfgHistory*
      SklCfgSteer::GetHistoryNext(const SklCfgHistory* history) const
{
  if( !history ) return NULL;
  if( !m_outputData ) return NULL;
  return m_outputData->GetHistoryNext(history);
}

/**
 * ソルバー起動条件を取得する。
 * @return    ソルバー起動条件
 */
bool SklCfgSteer::IsCheckMode() const
{
  return m_chkModeFlg;
}

const char* SklCfgSteer::GetCheckModeFile() const
{
  return m_chkModeFile;
}

/**
 * 入力フォルダを取得する。
 * @return     入力フォルダ
 */
const char* SklCfgSteer::getInputBaseDir() const
{
  if (m_inputData == NULL) return NULL;
  return m_inputData->GetBaseDir();
}

/**
 * 出力フォルダを取得する。
 * @return     出力フォルダ
 */
const char* SklCfgSteer::getOutputBaseDir() const
{
  if (m_outputData == NULL) return NULL;
  return m_outputData->GetBaseDir();
}


#ifdef DEBUG
void SklCfgSteer::Print()
{
  cout << " <-- Steer --" << endl << flush;

  cout << "  calculation steps : " << m_calcSteps << endl << flush;
  cout << "  start condition   : ";
  if( m_strType == Initial ) cout << "Initial";
  else if( m_strType == Restart ) cout << "Restart";
  else cout << "Unknown";
  cout << endl << flush;

  if( m_inputData ) m_inputData->Print();
  if( m_outputData ) m_outputData->Print();

  cout << "  -- Steer -->" << endl << flush;
}
#endif //DEBUG

} // end of namespace SklCfg

