/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */
/**
 * @file SklCfgSteerInputData.C
 * コンフィグレーションInputData要素クラスソースファイル
 * @author     riken
 * @date       2010/10/01
 */

#include <string.h> // add 20090629
#include "skl_config/SklCfgDefName.h"
#include "skl_config/SklCfgSteerInputData.h"

namespace SklCfg {
/**
 * コンストラクタ
 */
SklCfgSteerInputData::SklCfgSteerInputData()
{
  m_baseDir = NULL;
  m_inFileList.clear();
}

/**
 * デストラクタ
 */
SklCfgSteerInputData::~SklCfgSteerInputData()
{
  if( m_baseDir ) delete [] m_baseDir;
  if( !m_inFileList.empty() ){
    list<SklCfgInFile*>::iterator itr = m_inFileList.begin();
    list<SklCfgInFile*>::iterator itrEnd = m_inFileList.end();
    for( ; itr != itrEnd ; itr++ ) delete *itr;
  }
  m_inFileList.clear();
}

bool SklCfgSteerInputData::CheckNodeName(const char* nname)
{
  if( !nname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0001_);
    return false;
  }

  if( StrCaseCompare(nname,
                     _SKL_CFG_NNAME_STEER_INPUT_DATA_,
                     strlen(_SKL_CFG_NNAME_STEER_INPUT_DATA_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_);
    return false;
  }

  return true;
}

bool SklCfgSteerInputData::ParseNode(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !CheckNodeName((const char*)node->name) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  if( !ParseSklDefNode(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  if( !Verification() ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  return true;
}

bool SklCfgSteerInputData::ParseSklDefNode(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !CheckNodeName((const char*)node->name) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  xmlChar* tmpStr;
  tmpStr = xmlGetProp(node, (const xmlChar*)_SKL_CFG_ANAME_STEER_BASE_DIR_);
  if( tmpStr ){
    if( m_baseDir ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0012_);
      return false;
    }
    size_t len = strlen((const char*)tmpStr);
    if( !(m_baseDir = new char[len + 1]) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0005_);
      return false;
    }
    memset(m_baseDir, 0, sizeof(char)*(len + 1));
    strncpy(m_baseDir, (const char*)tmpStr, len+1);
    xmlFree(tmpStr);
  }
/*
  else {
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_STEER_INPUT_DATA_,
              _SKL_CFG_ANAME_STEER_BASE_DIR_);
    return false;
  }
*/

  xmlNodePtr cur = node->xmlChildrenNode;
  while ( cur ) {
    if ( cur->type == XML_COMMENT_NODE ) {
      cur = cur->next;
      continue;
    }

    if( !StrCaseCompare((const char*)cur->name,
                        _SKL_CFG_NNAME_STEER_IN_FILE_,
                        strlen(_SKL_CFG_NNAME_STEER_IN_FILE_)) ){
      SklCfgInFile* infile = new SklCfgInFile();
      if( !infile ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !infile->ParseNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        delete infile;
        return false;
      }
      if( !infile->CreateFileName(m_baseDir) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
      m_inFileList.push_back(infile);
    }

    cur = cur->next;
  }

  return true;
}

bool SklCfgSteerInputData::Verification()
{
  return true;
}

/**
 * ディレクトリパスを取得する。
 * @return    ディレクトリパス
 */
const char* SklCfgSteerInputData::GetBaseDir() const
{
  return (const char*)m_baseDir;
}

/**
 * InFile要素数を取得する。
 * @return    InFile要素数
 */
unsigned int SklCfgSteerInputData::GetInFileSize() const
{
  return (unsigned int)(m_inFileList.size());
}

/**
 * 最初のInFile要素を取得する。
 * @return    InFile要素クラス
 */
const SklCfgInFile* SklCfgSteerInputData::GetInFileFirst() const
{
  if( m_inFileList.size() == 0 ) return NULL;
  return m_inFileList.front();
}

/**
 * 次のInFile要素を取得する。
 * @param file    前のInFile要素
 * @return    InFile要素クラス
 */
const SklCfgInFile*
      SklCfgSteerInputData::GetInFileNext(const SklCfgInFile* file) const
{
  if( !file ) return NULL;
  list<SklCfgInFile*>::const_iterator itr = m_inFileList.begin();
  list<SklCfgInFile*>::const_iterator itrEnd = m_inFileList.end();
  for( ; itr != itrEnd; itr++ ){
    if( file == (*itr) ){
      list<SklCfgInFile*>::const_iterator itr2 = itr;
      itr2++;
      if( itr2 == itrEnd ) return NULL;
      return *itr2;
    }
  }
  return NULL;
}

#ifdef DEBUG
void SklCfgSteerInputData::Print()
{
  cout << "  <-- SteerInputData --" << endl << flush;

  if( m_baseDir ){
    cout << "   base directory   : " << m_baseDir << endl << flush;
  }else{
    cout << "   base directory not set" << endl << flush;
  }
  if( !m_inFileList.empty() ){
    list<SklCfgInFile*>::iterator itr = m_inFileList.begin();
    list<SklCfgInFile*>::iterator itrEnd = m_inFileList.end();
    for( ; itr != itrEnd ; itr++ ) (*itr)->Print();
  }

  cout << "   -- SteerInputData -->" << endl << flush;
}
#endif //DEBUG

} // end of namespace SklCfg

