/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */
/**
 * @file SklCfgUsrDefNode.C
 * コンフィグレーション要素クラスソースファイル
 * @author     riken
 * @date       2010/10/01
 */

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <iostream>
#include <string>
#include "skl_config/SklCfgDefName.h"
#include "skl_config/SklCfgUsrDefNode.h"

namespace SklCfg {
/**
 * コンストラクタ
 */
/* Add XML Parser Interface 2007.09 begin */
CfgIdt::CfgIdt() {
  Init();
}

/**
 * デストラクタ
 */
CfgIdt::CfgIdt(const char* name) {
  Init();
  if( !SetIdent(name) ) {
    fprintf(stderr, "Can't create CfgIdt object.\n");
    return;
  }
}

/**
 * コンストラクタ
 * @param name    name属性
 * @param id      id属性
 */
CfgIdt::CfgIdt(const char* name, int id) {
  Init();
  if( !SetIdent(name, id) ) {
    fprintf(stderr, "Can't create CfgIdt object.\n");
    return;
  }
}


/**
 * コンストラクタ
 * @param name    name属性
 * @param comment      comment属性
 */
CfgIdt::CfgIdt(const char* name,         const char* comment) {
  Init();
  if( !SetIdent(name, comment) ) {
    fprintf(stderr, "Can't create CfgIdt object.\n");
    return;
  }
}

/**
 * コンストラクタ
 * @param name    name属性
 * @param id      id属性
 * @param comment      comment属性
 */
CfgIdt::CfgIdt(const char* name, int id, const char* comment) {
  Init();
  if( !SetIdent(name, id, comment) ) {
    fprintf(stderr, "Can't create CfgIdt object.\n");
    return;
  }
}

/**
 * デストラクタ
 */
CfgIdt::~CfgIdt() {
  m_idtStr.clear();
  if( m_name ) delete [] m_name;
  if( m_comment ) delete [] m_comment;
}

/**
 * 初期化を行う。
 */
void CfgIdt::Init() {
  m_idtStr.clear();
  m_name = NULL;
  m_idFlg = false;
  m_id = -1;
  m_comment = NULL;
}

/**
 * 要素識別子文字列を更新する。
 * @return      true=success, false=fault
 */
bool CfgIdt::UpdateIdtStr() {
  if( !m_name ) return false;
  m_idtStr.clear();
  m_idtStr = string(m_name);
  if( IsSetID() ) {
    m_idtStr += string(":");
    char tmp[256];
    sprintf(tmp, "%d", m_id);
    m_idtStr += string(tmp);
    if( m_comment ) {
      m_idtStr += string(":");
      m_idtStr += string(m_comment);
    }
  }
  else {
    if( m_comment ) {
      m_idtStr += string(":*:");
      m_idtStr += string(m_comment);
    }
  }
  return true;
}

/**
 * name属性を設定する
 * @param name    name属性
 * @return      true=success, false=fault
 */
bool CfgIdt::SetName(const char* name) {
  if( m_name ) delete [] m_name;
  m_name = NULL;
  if( !name ) return true;
  if( !(m_name = new char[strlen(name)+1]) ) {
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  strcpy(m_name, name);
  return true;
}

/**
 * id属性を設定する。
 * @param id    id属性
 * @return      true=success, false=fault
 */
bool CfgIdt::SetID(int id) {
  m_id = id;
  m_idFlg = true;
  return true;
}

/**
 * id属性を設定をクリアする。
 * @return      true=success, false=fault
 */
bool CfgIdt::UnsetID() {
  m_id = -1;
  m_idFlg = false;
  return true;
}

/**
 * comment属性を設定する。
 * @param comment   comment属性
 * @return      true=success, false=fault
 */
bool CfgIdt::SetComment(const char* comment) {
  if( m_comment ) delete [] m_comment;
  m_comment = NULL;
  if( !comment ) return true;
  if( !(m_comment = new char[strlen(comment)+1]) ) {
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  strcpy(m_comment, comment);
  return true;
}

/**
 * XML要素からパースを行う。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool CfgIdt::Parse(xmlNodePtr node) {
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }
  return true;
}

/**
 * 要素識別子文字列から属性を設定する。
 * @param idt_str   要素識別子文字列
 * @return      true=success, false=fault
 */
bool CfgIdt::SetIdtFromStr(const char* idt_str) {
  if( !idt_str ) return false;

  if( m_name ) delete [] m_name;
  m_name = NULL;
  m_idFlg = false;
  m_id = -1;
  if( m_comment ) delete [] m_comment;
  m_comment = NULL;

  size_t str_len = strlen(idt_str);
  char* buf = new char[str_len + 1];
  if( !buf ) return false;
  strcpy(buf, idt_str);
  size_t idt_num = 1;
  register size_t m;
  for(m=0; m<=str_len; m++) {
    if( buf[m] == ':' ) idt_num++;
  }
  if( idt_num > 3 ) {
    CfgErrMsg("CfgIdt::SetIdtFromStr(const char* idt_str)");
    CfgErrMsg(" Invalid parameter.(idt_str : %s)", idt_str);
    return false;
  }

  const char *str[3];
  str[0] = str[1] = str[2] = NULL;
  register size_t n = 0;
  register size_t i = 0;
  str[i++] = &buf[0];
  while( n <= str_len ) {
    if( buf[n] == ':' ) {
      buf[n] = '\0';
      str[i++] = &buf[n+1];
    }
    n++;
    if( i >= idt_num ) break;
  }

  switch( i ) {
  case 3:
    if( !str[2] ) { delete [] buf; return false; }
    if( !(m_comment = new char[strlen(str[2])+1]) ) {
      delete [] buf;
      CfgErrMsg("CfgIdt::SetIdtFromStr(const char* idt_str)");
      CfgErrMsg(_SKL_CFG_ERR_0005_);
      return false;
    }
    strcpy(m_comment, str[2]);
  case 2:
    if( !str[1] ) {
      delete [] buf;
      CfgErrMsg("CfgIdt::SetIdtFromStr(const char* idt_str)");
      CfgErrMsg(" Invalid parameter.(idt_str : %s)", idt_str);
      return false;
    }
    if( (str[1][0] == '*') || (str[1][0] == '\0') ) {
      CfgErrMsg("CfgIdt::SetIdtFromStr(const char* idt_str)");
      CfgErrMsg(" Invalid parameter.(idt_str : %s)", idt_str);
      m_idFlg = false;
    }
    else {
      m_idFlg = true;
      m_id = atoi(str[1]);
    }
  case 1:
    if( !str[0] ) {
      delete [] buf;
      CfgErrMsg("CfgIdt::SetIdtFromStr(const char* idt_str)");
      CfgErrMsg(" Invalid parameter.(idt_str : %s)", idt_str);
      return false;
    }
    if( !(m_name = new char[strlen(str[0])+1]) ) {
      delete [] buf;
      CfgErrMsg("CfgIdt::SetIdtFromStr(const char* idt_str)");
      CfgErrMsg(_SKL_CFG_ERR_0005_);
      return false;
    }
    strcpy(m_name, str[0]);
    break;
  default:
    delete [] buf;
    CfgErrMsg("CfgIdt::SetIdtFromStr(const char* idt_str)");
    CfgErrMsg(" Invalid parameter.(idt_str : %s)", idt_str);
    return false;
  }

  delete [] buf;
  if( !UpdateIdtStr() ) {
    fprintf(stderr, "Can't create CfgIdt object.\n");
    return false;
  }
  return true;
}

/**
 * name属性を設定する。
 * @param name    name属性
 * @return      true=success, false=fault
 */
bool CfgIdt::SetIdent(const char* name) {
  if( !SetName(name) ) return false;
  if( !UnsetID() ) return false;
  if( m_comment ) delete [] m_comment;
  m_comment = NULL;
  if( !UpdateIdtStr() ) {
    fprintf(stderr, "Can't create CfgIdt object.\n");
    return false;
  }
  return true;
}

/**
 * name, id属性を設定する。
 * @param name    name属性
 * @param id    id属性
 * @return      true=success, false=fault
 */
bool CfgIdt::SetIdent(const char* name, int id) {
  if( !SetName(name) ) return false;
  if( !SetID(id) ) return false;
  if( m_comment ) delete [] m_comment;
  m_comment = NULL;
  if( !UpdateIdtStr() ) {
    fprintf(stderr, "Can't create CfgIdt object.\n");
    return false;
  }
  return true;
}

/**
 * name, comment属性を設定する。
 * @param name    name属性
 * @param comment   comment属性
 * @return      true=success, false=fault
 */
bool CfgIdt::SetIdent(const char* name,         const char* comment) {
  if( !SetName(name) ) return false;
  if( !UnsetID() ) return false;
  if( !SetComment(comment) ) return false;
  if( !UpdateIdtStr() ) {
    fprintf(stderr, "Can't create CfgIdt object.\n");
    return false;
  }
  return true;
}

/**
 * name, id, comment属性を設定する。
 * @param name    name属性
 * @param id    id属性
 * @param comment   comment属性
 * @return      true=success, false=fault
 */
bool CfgIdt::SetIdent(const char* name, int id, const char* comment) {
  if( !SetName(name) ) return false;
  if( !SetID(id) ) return false;
  if( !SetComment(comment) ) return false;
  if( !UpdateIdtStr() ) {
    fprintf(stderr, "Can't create CfgIdt object.\n");
    return false;
  }
  return true;
}

/**
 * 要素識別子文字列を取得する。
 * @param[out] str   要素識別子文字列
 * @param[in] slen    要素識別子文字列長
 * @return    要素識別子文字列
 */
char* CfgIdt::GetIdtStr(char* str, size_t slen) const {
  size_t idt_str_len = 0;
  if( !m_name ) return NULL;

  std::string tmp_str = std::string(m_name);
  if( m_idFlg ) {
    tmp_str += std::string(":");
    char id_str[128];
    sprintf(id_str, "%d", m_id);
    tmp_str += std::string(id_str);
    if( m_comment ) {
      tmp_str += std::string(":");
      tmp_str += std::string(m_comment);
    }
  }
  else {
    if( m_comment ) {
      tmp_str += std::string(":");
      tmp_str += std::string("*");
      tmp_str += std::string(":");
      tmp_str += std::string(m_comment);
    }
  }
  if( slen <= tmp_str.size() ) {
    CfgErrMsg("_SKL_CFG_ERR_0056_");
    return NULL;
  }
  strcpy(str, tmp_str.c_str());

  return str;
}

/**
 * 要素識別子文字列を取得する。
 * @return    要素識別子文字列
 */
const char* CfgIdt::GetIdtStr() const {
  if( m_idtStr.size() == 0 ) return NULL;
  return m_idtStr.c_str();
}

/**
 * '='オペレータ
 *
 * 要素識別子クラスをコピーする。
 * @param src   元要素識別子クラス
 * @return      コピー要素識別子クラス
 */
CfgIdt& CfgIdt::operator=(const CfgIdt& src) {
  if( m_name ) delete [] m_name;
  if( m_comment ) delete [] m_comment;
  Init();
  size_t len = strlen(src.m_name);
  m_name = new char[len+1];
  strcpy(m_name, src.m_name);
  m_idFlg = src.m_idFlg;
  m_id = src.m_id;
  len = strlen(src.m_comment);
  m_comment = new char[len+1];
  strcpy(m_comment, src.m_comment);
  return *this;
}

/**
 * 要素識別子文字列を作成する。
 * @param[out] str   要素識別子文字列
 * @param[in] slen    要素識別子文字列長
 * @param[in] name    name属性
 * @return    要素識別子文字列
 */
char* CfgIdt::genIdt(char* str, size_t slen, const char* name) {
  if( !name ) {
    fprintf(stderr, _SKL_CFG_ERR_0055_, "name = NULL");
    return NULL;
  }
  if( slen <= strlen(name) ) {
    fprintf(stderr, _SKL_CFG_ERR_0056_);
    return NULL;
  }
  strcpy(str, name);
  return str;
}

/**
 * 要素識別子文字列を作成する。
 * @param[out] str   要素識別子文字列
 * @param[in] slen    要素識別子文字列長
 * @param[in] name    name属性
 * @param[in] id      id属性
 * @return    要素識別子文字列
 */
char* CfgIdt::genIdt(char* str, size_t slen, const char* name, int id) {
  if( !name ) {
    fprintf(stderr, _SKL_CFG_ERR_0055_, "name = NULL");
    return NULL;
  }
  std::string tmp_str = std::string(name);
  tmp_str += std::string(":");
  char id_str[128];
  sprintf(id_str, "%d", id);
  tmp_str += std::string(id_str);
  if( slen <= tmp_str.size() ) {
    fprintf(stderr, _SKL_CFG_ERR_0056_);
    return NULL;
  }
  strcpy(str, tmp_str.c_str());
  return str;
}

/**
 * 要素識別子文字列を作成する。
 * @param[out] str   要素識別子文字列
 * @param[in] slen    要素識別子文字列長
 * @param[in] name    name属性
 * @param[in] comment   comment属性
 * @return    要素識別子文字列
 */
char* CfgIdt::genIdt(char* str, size_t slen,
                     const char* name, const char* comment) {
  if( !name || !comment ) {
    if( !name ) fprintf(stderr, _SKL_CFG_ERR_0055_, "name = NULL");
    if( !comment ) fprintf(stderr, _SKL_CFG_ERR_0055_, "comment = NULL");
    return NULL;
  }
  std::string tmp_str = std::string(name);
  tmp_str += std::string(":");
  tmp_str += std::string("*");
  tmp_str += std::string(":");
  tmp_str += std::string(comment);
  if( slen <= tmp_str.size() ) {
    fprintf(stderr, _SKL_CFG_ERR_0056_);
    return NULL;
  }
  strcpy(str, tmp_str.c_str());
  return str;
}

/**
 * 要素識別子文字列を作成する。
 * @param[out] str   要素識別子文字列
 * @param[in] slen    要素識別子文字列長
 * @param[in] name    name属性
 * @param[in] id      id属性
 * @param[in] comment   comment属性
 * @return    要素識別子文字列
 */
char* CfgIdt::genIdt(char* str, size_t slen,
                     const char* name, int id, const char* comment) {
  if( !name || !comment ) {
    if( !name ) fprintf(stderr, _SKL_CFG_ERR_0055_, "name = NULL");
    if( !comment ) fprintf(stderr, _SKL_CFG_ERR_0055_, "comment = NULL");
    return NULL;
  }
  std::string tmp_str = std::string(name);
  tmp_str += std::string(":");
  char id_str[128];
  sprintf(id_str, "%d", id);
  tmp_str += std::string(id_str);
  tmp_str += std::string(":");
  tmp_str += std::string(comment);
  if( slen <= tmp_str.size() ) {
    fprintf(stderr, _SKL_CFG_ERR_0056_);
    return NULL;
  }
  strcpy(str, tmp_str.c_str());
  return str;
}


/* Add XML Parser Interface 2007.09 end */
/**
 * コンストラクタ
 */
CfgBase::CfgBase()
{
  m_name = NULL;
  m_id = -1;
  m_idFlg = false;
  m_comment = NULL;
}

/**
 * デストラクタ
 */
CfgBase::~CfgBase()
{
  if( m_name ) delete [] m_name;
  if( m_comment ) delete [] m_comment;
}

/**
 * name属性を設定する。
 * @param name    name属性
 * @return      true=success, false=fault
 */
bool CfgBase::SetName(const char* name)
{
  if( !name ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0029_);
    return false;
  }
  unsigned int len = strlen(name)+1;
  if( m_name ) delete [] m_name;
  m_name = NULL;
  m_name = new char[len];
  if( !m_name ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  strncpy(m_name, name, len);
  return true;
}

/**
 * id属性を設定する。
 * @param id    id属性
 * @return      true=success, false=fault
 */
bool CfgBase::SetID(const int id)
{
  m_id = id;
  m_idFlg = true;
  return true;
}

/**
 * comment属性を設定する
 * @param cmt   comment属性
 * @return      true=success, false=fault
 */
bool CfgBase::SetComment(const char* cmt)
{
  if( !cmt ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0030_);
    return false;
  }
  unsigned int len = strlen(cmt)+1;
  if( m_comment ) delete [] m_comment;
  m_comment = NULL;
  m_comment = new char[len];
  if( !m_comment ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  strncpy(m_comment, cmt, len);
  return true;
}

/**
 * name属性を取得する。
 * @return    name属性
 */
const char* CfgBase::GetName() const
{
  return (const char*)m_name;
}

/**
 * id属性が設定されているかチェックする。
 * @return    true=id属性設定
 */
bool CfgBase::isSetID() const
{
  return m_idFlg;
}

/**
 * id属性を取得する。
 * @return    id属性
 */
int CfgBase::GetID() const
{
  return m_id;
}

/**
 * comment属性を取得する
 * @return    comment属性
 */
const char* CfgBase::GetComment() const
{
  return (const char*)m_comment;
}

/**
 * コンストラクタ
 */
CfgElem::CfgElem()
{
  m_prmList.clear();
  m_elmList.clear();
}

/**
 * デストラクタ
 */
CfgElem::~CfgElem()
{
  if( !m_prmList.empty() ){
    list<CfgParam*>::const_iterator itr = m_prmList.begin();
    list<CfgParam*>::const_iterator itrEnd = m_prmList.end();
    for( ; itr != itrEnd ; itr++ ) delete *itr;
  }
  m_prmList.clear();

  if( !m_elmList.empty() ){
    list<CfgElem*>::const_iterator itr = m_elmList.begin();
    list<CfgElem*>::const_iterator itrEnd = m_elmList.end();
    for( ; itr != itrEnd ; itr++ ) delete *itr;
  }
  m_elmList.clear();
}

/**
 * Param要素を追加する。
 * @param param   Param要素
 * @return      true=success, false=fault
 */
bool CfgElem::AddParam(CfgParam* param)
{
  if( !param ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0031_);
    return false;
  }
  m_prmList.push_back(param);
  return true;
}

/**
 * Elem要素を追加する。
 * @param elem   Elem要素
 * @return      true=success, false=fault
 */
bool CfgElem::AddElem(CfgElem* elem)
{
  if( !elem ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0032_);
    return false;
  }
  m_elmList.push_back(elem);
  return true;
}

/**
 * XML要素からパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool CfgElem::Parse(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_USRDEF_ELEM_,
                     strlen(_SKL_CFG_NNAME_USRDEF_ELEM_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlNodePtr cur = node->xmlChildrenNode;
  while ( cur ) {
    if ( cur->type == XML_COMMENT_NODE ) {
      cur = cur->next;
      continue;
    }

    if( !StrCaseCompare((const char*)cur->name,
                        _SKL_CFG_NNAME_USRDEF_ELEM_,
                        strlen(_SKL_CFG_NNAME_USRDEF_ELEM_)) ){
      // parse Elem tag
      CfgElem* elem = new CfgElem;
      if( !elem ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !elem->SetAttr(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        delete elem;
        return false;
      }
      if( !AddElem(elem) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        delete elem;
        return false;
      }
    }
    else if( !StrCaseCompare((const char*)cur->name,
                             _SKL_CFG_NNAME_USRDEF_PARAM_,
                             strlen(_SKL_CFG_NNAME_USRDEF_PARAM_)) ){
      //parse Param tag
      CfgParam* param = new CfgParam;
      if( !param ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !param->SetAttr(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        delete param;
        return false;
      }
      if( !AddParam(param) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        delete param;
        return false;
      }
    }
    cur = cur->next;
  }

  return true;
}

/**
 * XML要素からパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool CfgElem::ParseNode(xmlNodePtr node)
{
  return Parse(node);
}

/**
 * XML要素から属性を設定する。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool CfgElem::SetAttr(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_USRDEF_ELEM_,
                     strlen(_SKL_CFG_NNAME_USRDEF_ELEM_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlChar* tmpStr = NULL;

  // get name attribute
  if( !(tmpStr = xmlGetProp(node,
                    (const xmlChar*)_SKL_CFG_ANAME_USRDEF_ELEM_NAME_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_USRDEF_ELEM_,
              _SKL_CFG_ANAME_USRDEF_ELEM_NAME_);
    return false;
  }
  if( !SetName((const char*)tmpStr) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    xmlFree(tmpStr);
    return false;
  }
  xmlFree(tmpStr);

  // get id attribute
  if( (tmpStr = xmlGetProp(node,
                   (const xmlChar*)_SKL_CFG_ANAME_USRDEF_ELEM_ID_)) ){
    int id = atoi((const char*)tmpStr);
    if( !SetID(id) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      xmlFree(tmpStr);
      return false;
    }
    xmlFree(tmpStr);
  }

  // get comment attribute
  if( (tmpStr = xmlGetProp(node,
                   (const xmlChar*)_SKL_CFG_ANAME_USRDEF_ELEM_COMMENT_)) ){
    if( !SetComment((const char*)tmpStr) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      xmlFree(tmpStr);
      return false;
    }
    xmlFree(tmpStr);
  }

  if( !Parse(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  return true;
}

/**
 * 子要素をクリアする。
 */
void CfgElem::ClearList() {
  if( !m_prmList.empty() ){
    list<CfgParam*>::const_iterator itr = m_prmList.begin();
    list<CfgParam*>::const_iterator itrEnd = m_prmList.end();
    for( ; itr != itrEnd ; itr++ ) delete *itr;
  }
  m_prmList.clear();

  if( !m_elmList.empty() ){
    list<CfgElem*>::const_iterator itr = m_elmList.begin();
    list<CfgElem*>::const_iterator itrEnd = m_elmList.end();
    for( ; itr != itrEnd ; itr++ ) delete *itr;
  }
  m_elmList.clear();
}

/**
 * 子要素数を取得する。
 * @return    子要素数
 */
unsigned int CfgElem::GetSize() const
{
  unsigned int prmListNum = (unsigned int)m_prmList.size();
  unsigned int elmListNum = (unsigned int)m_elmList.size();
  return (prmListNum + elmListNum);
}

/**
 * 子Elem要素数を取得する
 * @return    子Elem要素数
 */
unsigned int CfgElem::GetElemSize() const
{
  return (unsigned int)(m_elmList.size());
}


/**
 * 子Param要素数を取得する
 * @return    子Param要素数
 */
unsigned int CfgElem::GetParamSize() const
{
  return (unsigned int)(m_prmList.size());
}

/**
 * 指定された属性の最初のElem要素を取得する
 * @param name    name属性
 * @return    Elem要素
 */
const CfgElem* CfgElem::GetElemFirst(const char* name) const
{
  if( m_elmList.empty() ) return NULL;
  if( !name ) return m_elmList.front();

  list<CfgElem*>::const_iterator itr = m_elmList.begin();
  list<CfgElem*>::const_iterator itrEnd = m_elmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( !StrCaseCompare(name, (*itr)->GetName()) ) return *itr;
  }

  return NULL;
}

/**
 * 指定された属性の最初のElem要素を取得する
 * @param name    name属性
 * @param id      id属性
 * @param cmt     comment属性
 * @return    Elem要素
 */
const CfgElem* CfgElem::GetElemFirst(const char* name,
                                     int id, const char* cmt) const
{
  if( m_elmList.empty() ) return NULL;

  list<CfgElem*>::const_iterator itr = m_elmList.begin();
  list<CfgElem*>::const_iterator itrEnd = m_elmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( id == (*itr)->GetID() ){
      if( !name || !StrCaseCompare(name, (*itr)->GetName()) ){
        if( !cmt || !StrCaseCompare(cmt, (*itr)->GetComment()) ){
          // find!
          return *itr;
        }
      }
    }
  }
  return NULL;
}

/**
 * 指定された属性の最初のElem要素を取得する
 * @param name    name属性
 * @param cmt     comment属性
 * @return    Elem要素
 */
const CfgElem* CfgElem::GetElemFirst(const char* name, const char* cmt) const
{
  if( m_elmList.empty() ) return NULL;
  if( !name || !cmt ) return NULL;

  list<CfgElem*>::const_iterator itr = m_elmList.begin();
  list<CfgElem*>::const_iterator itrEnd = m_elmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( !StrCaseCompare(name, (*itr)->GetName()) 
        && !StrCaseCompare(cmt, (*itr)->GetComment()) ) {
      // find!
      return *itr;
    }
  }
  return NULL;
}

/**
 * 指定された属性の次のElem要素を取得する
 * @param elem      前の要素
 * @param name    name属性
 * @return    Elem要素
 */
const CfgElem* CfgElem::GetElemNext(const CfgElem* elem,
                                    const char* name) const
{
  if( !elem ) return NULL;

  list<CfgElem*>::const_iterator itr = m_elmList.begin();
  list<CfgElem*>::const_iterator itrEnd = m_elmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( elem == (*itr) ){
      list<CfgElem*>::const_iterator itr2 = itr;
      itr2++;
      if( itr2 == itrEnd ) return NULL;
      if( !name ) return *itr2;
      else{
        for( ; itr2 != itrEnd; itr2++){
          if( !StrCaseCompare(name, (*itr2)->GetName()) ) return *itr2;
        }
      }
    }
  }

  return NULL;
}

/**
 * 指定された属性の次のElem要素を取得する
 * @param elem      前の要素
 * @param name    name属性
 * @param id      id属性
 * @param cmt     comment属性
 * @return    Elem要素
 */
const CfgElem* CfgElem::GetElemNext(const CfgElem* elem,
                                    const char* name,
                                    int id,
                                    const char* cmt) const
{
  if( !elem ) return NULL;

  list<CfgElem*>::const_iterator itr = m_elmList.begin();
  list<CfgElem*>::const_iterator itrEnd = m_elmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( elem == (*itr) ){
      list<CfgElem*>::const_iterator itr2 = itr;
      itr2++;
      if( itr2 == itrEnd ) return NULL;
      for( ; itr2 != itrEnd; itr2++){
        if( id == (*itr2)->GetID() ){
          if( !name || !StrCaseCompare(name, (*itr2)->GetName()) ){
            if( !cmt || !StrCaseCompare(cmt, (*itr2)->GetComment()) ){
              // find!
              return *itr2;
            }
          }
        }
      }
    }
  }
  return NULL;
}

/**
 * 指定された属性の次のElem要素を取得する
 * @param elem      前の要素
 * @param name    name属性
 * @param cmt     comment属性
 * @return    Elem要素
 */
const CfgElem* CfgElem::GetElemNext(const CfgElem* elem,
                                    const char* name, const char* cmt) const
{
  if( !elem || !name || !cmt ) return NULL;

  list<CfgElem*>::const_iterator itr = m_elmList.begin();
  list<CfgElem*>::const_iterator itrEnd = m_elmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( elem == (*itr) ){
      list<CfgElem*>::const_iterator itr2 = itr;
      itr2++;
      if( itr2 == itrEnd ) return NULL;
      for( ; itr2 != itrEnd; itr2++){
        if( !StrCaseCompare(name, (*itr2)->GetName())
            && !StrCaseCompare(cmt, (*itr2)->GetComment()) ) {
          // find!
          return *itr2;
        }
      }
    }
  }
  return NULL;
}

/**
 * 指定された属性の最初のParam要素を取得する
 * @param name    name属性
 * @return    Param要素
 */
const CfgParam* CfgElem::GetParamFirst(const char* name) const
{
  if( m_prmList.empty() ) return NULL;
  if( !name ) return m_prmList.front();

  list<CfgParam*>::const_iterator itr = m_prmList.begin();
  list<CfgParam*>::const_iterator itrEnd = m_prmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( !StrCaseCompare(name, (*itr)->GetName()) ) return *itr;
  }
  return NULL;
}

/**
 * 指定された属性の最初のParam要素を取得する
 * @param name    name属性
 * @param id      id属性
 * @param cmt     comment属性
 * @return    Param要素
 */
const CfgParam* CfgElem::GetParamFirst(const char* name,
                                       int id, const char* cmt) const
{
  if( m_prmList.empty() ) return NULL;

  list<CfgParam*>::const_iterator itr = m_prmList.begin();
  list<CfgParam*>::const_iterator itrEnd = m_prmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( id == (*itr)->GetID() ){
      if( !name || !StrCaseCompare(name, (*itr)->GetName()) ){
        if( !cmt || !StrCaseCompare(cmt, (*itr)->GetComment()) ){
          // find!
          return *itr;
        }
      }
    }
  }
  return NULL;
}

/**
 * 指定された属性の最初のParam要素を取得する
 * @param name    name属性
 * @param cmt     comment属性
 * @return    Param要素
 */
const CfgParam* CfgElem::GetParamFirst(const char* name, const char* cmt) const
{
  if( m_prmList.empty() ) return NULL;
  if( !name || !cmt ) return NULL;

  list<CfgParam*>::const_iterator itr = m_prmList.begin();
  list<CfgParam*>::const_iterator itrEnd = m_prmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( !StrCaseCompare(name, (*itr)->GetName())
        && !StrCaseCompare(cmt, (*itr)->GetComment()) ) {
      // find!
      return *itr;
    }
  }
  return NULL;
}

/**
 * 指定された属性の次のParam要素を取得する
 * @param param    前のParam要素
 * @param name    name属性
 * @return    Param要素
 */
const CfgParam* CfgElem::GetParamNext(const CfgParam* param,
                                      const char* name) const
{
  if( !param ) return NULL;

  list<CfgParam*>::const_iterator itr = m_prmList.begin();
  list<CfgParam*>::const_iterator itrEnd = m_prmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( param == (*itr) ){
      list<CfgParam*>::const_iterator itr2 = itr;
      itr2++;
      if( itr2 == itrEnd ) return NULL;
      if( !name ) return *itr2;
      else{
        for( ; itr2 != itrEnd; itr2++){
          if( !StrCaseCompare(name, (*itr2)->GetName()) ) return *itr2;
        }
      }
    }
  }
  return NULL;
}

/**
 * 指定された属性の次のParam要素を取得する
 * @param param    前のParam要素
 * @param name    name属性
 * @param id      id属性
 * @param cmt     comment属性
 * @return    Param要素
 */
const CfgParam* CfgElem::GetParamNext(const CfgParam* param,
                                      const char* name,
                                      int id,
                                      const char* cmt) const
{
  if( !param ) return NULL;

  list<CfgParam*>::const_iterator itr = m_prmList.begin();
  list<CfgParam*>::const_iterator itrEnd = m_prmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( param == (*itr) ){
      list<CfgParam*>::const_iterator itr2 = itr;
      itr2++;
      if( itr2 == itrEnd ) return NULL;
      for( ; itr2 != itrEnd; itr2++){
        if( id == (*itr2)->GetID() ){
          if( !name || !StrCaseCompare(name, (*itr2)->GetName()) ){
            if( !cmt || !StrCaseCompare(cmt, (*itr2)->GetComment()) ){
              // find!
              return *itr2;
            }
          }
        }
      }
    }
  }
  return NULL;
}

/**
 * 指定された属性の次のParam要素を取得する
 * @param param    前のParam要素
 * @param name    name属性
 * @param cmt     comment属性
 * @return    Param要素
 */
const CfgParam* CfgElem::GetParamNext(const CfgParam* param,
                                      const char* name, const char* cmt) const
{
  if( !param || !name || !cmt ) return NULL;

  list<CfgParam*>::const_iterator itr = m_prmList.begin();
  list<CfgParam*>::const_iterator itrEnd = m_prmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( param == (*itr) ){
      list<CfgParam*>::const_iterator itr2 = itr;
      itr2++;
      if( itr2 == itrEnd ) return NULL;
      for( ; itr2 != itrEnd; itr2++){
        if( !StrCaseCompare(name, (*itr2)->GetName())
            && !StrCaseCompare(cmt, (*itr2)->GetComment()) ) {
          // find!
          return *itr2;
        }
      }
    }
  }
  return NULL;
}

/*
 *  Add XML Parser Interface 2007.09 begin
 */
/**
 * 要素識別子から最初のElem要素を取得する。
 * @param idt   要素識別子
 * @return      Elem要素
 */
const CfgElem* CfgElem::GetElemFirst(CfgIdt& idt) const {
  if( m_elmList.empty() ) return NULL;
  const char* name = idt.GetName();
  if( !name ) return m_elmList.front();

  list<CfgElem*>::const_iterator itr = m_elmList.begin();
  list<CfgElem*>::const_iterator itrEnd = m_elmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( !StrCaseCompare(name, (*itr)->GetName()) ) {
      if( idt.IsSetID() ) {
        int id = idt.GetID();
        if( id == (*itr)->GetID() ) {
          const char* comment = idt.GetComment();
          if( !comment ) {
            // find!
            return *itr;
          }
          if( !StrCaseCompare(comment, (*itr)->GetComment()) ) {
            // find!
            return *itr;
          }
        }
      }
      else {
        const char* comment = idt.GetComment();
        if( !comment ) {
          // find!
          return *itr;
        }
        if( !StrCaseCompare(comment, (*itr)->GetComment()) ) {
          // find!
          return *itr;
        }
      }
    }
  }

  return NULL;
}

/**
 * 要素識別子から次のElem要素を取得する。
 * @param elem   前のElem要素
 * @param idt   要素識別子
 * @return      Elem要素
 */
const CfgElem* CfgElem::GetElemNext(const CfgElem* elem, CfgIdt& idt) const {
  if( !elem ) return NULL;

  list<CfgElem*>::const_iterator itr = m_elmList.begin();
  list<CfgElem*>::const_iterator itrEnd = m_elmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( elem == (*itr) ){
      list<CfgElem*>::const_iterator itr2 = itr;
      itr2++;
      if( itr2 == itrEnd ) return NULL;
      const char* name = idt.GetName();
      if( !name ) return *itr2;
      else{
        for( ; itr2 != itrEnd; itr2++){
          if( !StrCaseCompare(name, (*itr2)->GetName()) ) {
            if( idt.IsSetID() ) {
              int id = idt.GetID();
              if( id == (*itr2)->GetID() ) {
                const char* comment = idt.GetComment();
                if( !comment ) {
                  // find!
                  return *itr2;
                }
                if( !StrCaseCompare(comment, (*itr2)->GetComment()) ) {
                  // find!
                  return *itr2;
                }
              }
            }
            else {
              const char* comment = idt.GetComment();
              if( !comment ) {
                // find!
                return *itr2;
              }
              if( !StrCaseCompare(comment, (*itr2)->GetComment()) ) {
                // find!
                return *itr2;
              }
            }
          }
        }
      }
    }
  }

  return NULL;
}

/**
 * 要素識別子から最初のParam要素を取得する。
 * @param idt   要素識別子
 * @return      Param要素
 */
const CfgParam* CfgElem::GetParamFirst(CfgIdt& idt) const {
  if( m_prmList.empty() ) return NULL;
  const char* name = idt.GetName();
  if( !name ) return m_prmList.front();

  list<CfgParam*>::const_iterator itr = m_prmList.begin();
  list<CfgParam*>::const_iterator itrEnd = m_prmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( !StrCaseCompare(name, (*itr)->GetName()) ) {
      if( idt.IsSetID() ) {
        int id = idt.GetID();
        if( id == (*itr)->GetID() ) {
          const char* comment = idt.GetComment();
          if( !comment ) {
            // find!
            return *itr;
          }
          if( !StrCaseCompare(comment, (*itr)->GetComment()) ) {
            // find!
            return *itr;
          }
        }
      }
      else {
        const char* comment = idt.GetComment();
        if( !comment ) {
          // find!
          return *itr;
        }
        if( !StrCaseCompare(comment, (*itr)->GetComment()) ) {
          // find!
          return *itr;
        }
      }
    }
  }

  return NULL;
}

/**
 * 要素識別子から次のParam要素を取得する。
 * @param param   前のPram要素
 * @param idt   要素識別子
 * @return      Param要素
 */
const CfgParam* CfgElem::GetParamNext(const CfgParam* param, CfgIdt& idt) const {
  if( !param ) return NULL;

  list<CfgParam*>::const_iterator itr = m_prmList.begin();
  list<CfgParam*>::const_iterator itrEnd = m_prmList.end();
  for( ; itr != itrEnd; itr++ ){
    if( param == (*itr) ){
      list<CfgParam*>::const_iterator itr2 = itr;
      itr2++;
      if( itr2 == itrEnd ) return NULL;
      const char* name = idt.GetName();
      if( !name ) return *itr2;
      else{
        for( ; itr2 != itrEnd; itr2++){
          if( !StrCaseCompare(name, (*itr2)->GetName()) ) {
            if( idt.IsSetID() ) {
              int id = idt.GetID();
              if( id == (*itr2)->GetID() ) {
                const char* comment = idt.GetComment();
                if( !comment ) {
                  // find!
                  return *itr2;
                }
                if( !StrCaseCompare(comment, (*itr2)->GetComment()) ) {
                  // find!
                  return *itr2;
                }
              }
            }
            else {
              const char* comment = idt.GetComment();
              if( !comment ) {
                // find!
                return *itr2;
              }
              if( !StrCaseCompare(comment, (*itr2)->GetComment()) ) {
                // find!
                return *itr2;
              }
            }
          }
        }
      }
    }
  }

  return NULL;
}

/**
 * 子要素のParam要素からvalue属性を取得する。
 * @param format      取得データ型
 * @param ...         要素識別子, 変数参照
 * @return      true=success, false=fault
 */
bool CfgElem::GetValues(const char* format, ...) const {
  if( !format ) return false;
  const char* delim = ":";
//----------------------------------------------------------------
  if( strlen(format) > 255 ) {
    CfgErrMsg(_SKL_CFG_ERR_0057_);
    return false;
  }
  char fmt[256];
  memset(fmt, 0, 256);
  strcpy(fmt, format);
  char* token = strtok(fmt, delim);
//----------------------------------------------------------------
//  char* token = strtok(format, delim);
//----------------------------------------------------------------
  va_list vl;
  va_start(vl, format);
  while( token ) {
    if( !strcmp(token, "i") ) {      // integer
      const char* ident_str = va_arg(vl, const char*);
      CfgIdt ident;
      if( !ident.SetIdtFromStr(ident_str) ) {
        va_end(vl);
        return false;
      }
      int* value = va_arg(vl, int*);
      if( !value ) {
        va_end(vl);
        CfgErrMsg(_SKL_CFG_ERR_0058_);
        return false;
      }
      const CfgParam* param = GetParamFirst(ident);
      if( !param ) {
        va_end(vl);
        return false;
      }
      if( !param->GetData(value) ) {
        va_end(vl);
        return false;
      }
    }
    else if( !strcmp(token, "f") ) { // float
      const char* ident_str = va_arg(vl, const char*);
      CfgIdt ident;
      if( !ident.SetIdtFromStr(ident_str) ) {
        va_end(vl);
        return false;
      }
      float* value = va_arg(vl, float*);
      if( !value ) {
        va_end(vl);
        CfgErrMsg(_SKL_CFG_ERR_0058_);
        return false;
      }
      const CfgParam* param = GetParamFirst(ident);
      if( !param ) {
        va_end(vl);
        return false;
      }
      if( !param->GetData(value) ) {
        va_end(vl);
        return false;
      }
    }
    else if( !strcmp(token, "d") ) { // double
      const char* ident_str = va_arg(vl, const char*);
      CfgIdt ident;
      if( !ident.SetIdtFromStr(ident_str) ) {
        va_end(vl);
        return false;
      }
      double* value = va_arg(vl, double*);
      if( !value ) {
        va_end(vl);
        CfgErrMsg(_SKL_CFG_ERR_0058_);
        return false;
      }
      const CfgParam* param = GetParamFirst(ident);
      if( !param ) {
        va_end(vl);
        return false;
      }
      if( !param->GetData(value) ) {
        va_end(vl);
        return false;
      }
    }
    else if( !strcmp(token, "s") ) { // string
      const char* ident_str = va_arg(vl, const char*);
      CfgIdt ident;
      if( !ident.SetIdtFromStr(ident_str) ) {
        va_end(vl);
        return false;
      }
// ----- modify 20080724 begin forT2K ----------------------------------
//      char** value = va_arg(vl, char**);
// ---------------------------------------------------------------------
      char** value = (char**)va_arg(vl, char*);
// ----- modify 20080724 end -------------------------------------------
      if( !value ) {
        va_end(vl);
        CfgErrMsg(_SKL_CFG_ERR_0058_);
        return false;
      }
      const CfgParam* param = GetParamFirst(ident);
      if( !param ) {
        va_end(vl);
        return false;
      }
      const char* tmp_str;
      if( !param->GetData(&tmp_str) ) {
        va_end(vl);
        return false;
      }
      *value = (char*)tmp_str;
    }
    else if( !strcmp(token, "e") ) { // CfgElem
      const char* ident_str = va_arg(vl, const char*);
      CfgIdt ident;
      if( !ident.SetIdtFromStr(ident_str) ) {
        va_end(vl);
        return false;
      }
// ----- modify 20080724 begin for T2K ---------------------------------
//      const CfgElem** value = va_arg(vl, const CfgElem**);
// ---------------------------------------------------------------------
      const CfgElem** value = (const CfgElem**)va_arg(vl, const CfgElem*);
// ----- modify 20080724 end -------------------------------------------
      if( !value ) {
        va_end(vl);
        CfgErrMsg(_SKL_CFG_ERR_0058_);
        return false;
      }
      if( !(*value = GetElemFirst(ident)) ) {
        va_end(vl);
        return false;
      }
    }
    else {
      va_end(vl);
      return false;
    }
    token = strtok(NULL, delim);
  }
  va_end(vl);
  return true;
}

/**
 * 子要素のParam要素からvalue属性を取得する。(int型)
 * @param num     取得数
 * @param ...         要素識別子, 変数参照
 * @return      true=success, false=fault
 */
bool CfgElem::GetIntValues(unsigned int num, ...) const {
  va_list vl;
  va_start(vl, num);
  register unsigned int n;
  for(n=0; n<num; n++) {
    const char* ident_str = va_arg(vl, const char*);
    CfgIdt ident;
    if( !ident.SetIdtFromStr(ident_str) ) {
      va_end(vl);
      return false;
    }
    int* value = va_arg(vl, int*);
    if( !value ) {
      va_end(vl);
      CfgErrMsg(_SKL_CFG_ERR_0058_);
      return false;
    }
    const CfgParam* param = GetParamFirst(ident);
    if( !param ) {
      va_end(vl);
      return false;
    }
    if( !param->GetData(value) ) {
      va_end(vl);
      return false;
    }
  }
  va_end(vl);
  return true;
}

/**
 * 子要素のParam要素からvalue属性を取得する。(float型)
 * @param num     取得数
 * @param ...         要素識別子, 変数参照
 * @return      true=success, false=fault
 */
bool CfgElem::GetFloatValues(unsigned int num, ...) const {
  va_list vl;
  va_start(vl, num);
  register unsigned int n;
  for(n=0; n<num; n++) {
    const char* ident_str = va_arg(vl, const char*);
    CfgIdt ident;
    if( !ident.SetIdtFromStr(ident_str) ) {
      va_end(vl);
      return false;
    }
    float* value = va_arg(vl, float*);
    if( !value ) {
      va_end(vl);
      CfgErrMsg(_SKL_CFG_ERR_0058_);
      return false;
    }
    const CfgParam* param = GetParamFirst(ident);
    if( !param ) {
      va_end(vl);
      return false;
    }
    if( !param->GetData(value) ) {
      va_end(vl);
      return false;
    }
  }
  va_end(vl);
  return true;
}

/**
 * 子要素のParam要素からvalue属性を取得する。(double型)
 * @param num     取得数
 * @param ...         要素識別子, 変数参照
 * @return      true=success, false=fault
 */
bool CfgElem::GetDoubleValues(unsigned int num, ...) const {
  va_list vl;
  va_start(vl, num);
  register unsigned int n;
  for(n=0; n<num; n++) {
    const char* ident_str = va_arg(vl, const char*);
    CfgIdt ident;
    if( !ident.SetIdtFromStr(ident_str) ) {
      va_end(vl);
      return false;
    }
    double* value = va_arg(vl, double*);
    if( !value ) {
      va_end(vl);
      CfgErrMsg(_SKL_CFG_ERR_0058_);
      return false;
    }
    const CfgParam* param = GetParamFirst(ident);
    if( !param ) {
      va_end(vl);
      return false;
    }
    if( !param->GetData(value) ) {
      va_end(vl);
      return false;
    }
  }
  va_end(vl);
  return true;
}

/**
 * 子要素のParam要素からvalue属性を取得する。(string型)
 * @param num     取得数
 * @param ...         要素識別子, 変数参照
 * @return      true=success, false=fault
 */
bool CfgElem::GetStrValues(unsigned int num, ...) const {
  va_list vl;
  va_start(vl, num);
  register unsigned int n;
  for(n=0; n<num; n++) {
    const char* ident_str = va_arg(vl, const char*);
    CfgIdt ident;
    if( !ident.SetIdtFromStr(ident_str) ) {
      va_end(vl);
      return false;
    }
// ----- modify 20080724 begin for T2K ---------------------------------
//    char** value = va_arg(vl, char**);
// ---------------------------------------------------------------------
    char** value = (char**)va_arg(vl, char*);
// ----- modify 20080724 end -------------------------------------------
    if( !value ) {
      va_end(vl);
      CfgErrMsg(_SKL_CFG_ERR_0058_);
      return false;
    }
    const CfgParam* param = GetParamFirst(ident);
    if( !param ) {
      va_end(vl);
      return false;
    }
    const char* tmp_str;
    if( !param->GetData(&tmp_str) ) {
      va_end(vl);
      return false;
    }
    *value = (char*)tmp_str;
  }
  va_end(vl);
  return true;
}

/**
 * 子要素のParam要素からvalue属性を取得する。(CfgElem型)
 * @param num     取得数
 * @param ...         要素識別子, 変数参照
 * @return      true=success, false=fault
 */
bool CfgElem::GetElemValues(unsigned int num, ...) const {
  va_list vl;
  va_start(vl, num);
  register unsigned int n;
  for(n=0; n<num; n++) {
    const char* ident_str = va_arg(vl, const char*);
    CfgIdt ident;
    if( !ident.SetIdtFromStr(ident_str) ) {
      va_end(vl);
      return false;
    }
// ----- modify 20080724 begin for T2K ---------------------------------
//    const CfgElem** value = va_arg(vl, const CfgElem**);
// ---------------------------------------------------------------------
    const CfgElem** value = (const CfgElem**)va_arg(vl, const CfgElem*);
// ----- modify 20080724 end -------------------------------------------
    if( !value ) {
      va_end(vl);
      CfgErrMsg(_SKL_CFG_ERR_0058_);
      return false;
    }
    if( !(*value = GetElemFirst(ident)) ) {
      va_end(vl);
      return false;
    }
  }
  va_end(vl);
  return true;
}

/**
 * value属性を取得する。(int)
 * @param[in] ident     要素識別子
 * @param[out] value     value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(CfgIdt ident, int* value) const {
  if( !value ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "value = NULL");
    return false;
  }
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  if( !param->GetData(value) ) return false;
  return true;
}


/**
 * value属性を取得する。(float)
 * @param[in] ident     要素識別子
 * @param[out] value     value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(CfgIdt ident, float* value) const {
  if( !value ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "value = NULL");
    return false;
  }
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  if( !param->GetData(value) ) return false;
  return true;
}

/**
 * value属性を取得する。(double)
 * @param[in] ident     要素識別子
 * @param[out] value     value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(CfgIdt ident, double* value) const {
  if( !value ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "value = NULL");
    return false;
  }
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  if( !param->GetData(value) ) return false;
  return true;
}

/**
 * value属性を取得する。(char*)
 * @param[in] ident     要素識別子
 * @param[out] value     value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(CfgIdt ident, const char** value) const {
  if( !value ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "value = NULL");
    return false;
  }
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  if( !param->GetData(value) ) return false;
  return true;
}

/**
 * 子Elem要素を取得する。
 * @param[in] ident     要素識別子
 * @param[out] value     子Elem要素
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(CfgIdt ident, const CfgElem** value) const {
  if( !value ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "value = NULL");
    return false;
  }
  const CfgElem* elem = GetElemFirst(ident);
  if( !elem ) return false;
  *value = elem;
  return true;
}

/**
 * value属性を取得する。(int)
 * @param[in] ident_str     要素識別子文字列
 * @param[out] value     value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(const char* ident_str, int* value) const {
  if( !ident_str || !value ) {
    if( !ident_str ) CfgErrMsg(_SKL_CFG_ERR_0055_, "ident_str = NULL");
    if( !value ) CfgErrMsg(_SKL_CFG_ERR_0055_, "value = NULL");
    return false;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) return false;
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  if( !param->GetData(value) ) return false;
  return true;
}

/**
 * value属性を取得する。(float)
 * @param[in] ident_str     要素識別子文字列
 * @param[out] value     value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(const char* ident_str, float* value) const {
  if( !ident_str || !value ) {
    if( !ident_str ) CfgErrMsg(_SKL_CFG_ERR_0055_, "ident_str = NULL");
    if( !value ) CfgErrMsg(_SKL_CFG_ERR_0055_, "value = NULL");
    return false;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) return false;
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  if( !param->GetData(value) ) return false;
  return true;
}

/**
 * value属性を取得する。(double)
 * @param[in] ident_str     要素識別子文字列
 * @param[out] value     value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(const char* ident_str, double* value) const {
  if( !ident_str || !value ) {
    if( !ident_str ) CfgErrMsg(_SKL_CFG_ERR_0055_, "ident_str = NULL");
    if( !value ) CfgErrMsg(_SKL_CFG_ERR_0055_, "value = NULL");
    return false;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) return false;
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  if( !param->GetData(value) ) return false;
  return true;
}

/**
 * value属性を取得する。(char*)
 * @param[in] ident_str     要素識別子文字列
 * @param[out] value     value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(const char* ident_str, const char** value) const {
  if( !ident_str || !value ) {
    if( !ident_str ) CfgErrMsg(_SKL_CFG_ERR_0055_, "ident_str = NULL");
    if( !value ) CfgErrMsg(_SKL_CFG_ERR_0055_, "value = NULL");
    return false;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) return false;
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  if( !param->GetData(value) ) return false;
  return true;
}

/**
 * 子Elem要素を取得する。
 * @param[in] ident_str     要素識別子文字列
 * @param[out] value     子Elem要素
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(const char* ident_str, const CfgElem** value) const {
  if( !ident_str || !value ) {
    if( !ident_str ) CfgErrMsg(_SKL_CFG_ERR_0055_, "ident_str = NULL");
    if( !value ) CfgErrMsg(_SKL_CFG_ERR_0055_, "value = NULL");
    return false;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) return false;
  const CfgElem* elem = GetElemFirst(ident);
  if( !elem ) return false;
  *value = elem;
  return true;
}


/**
 * value属性を取得する。(int)
 * @param ident     要素識別子
 * @return      value属性
 */
int CfgElem::GetIntValue(CfgIdt ident) const {
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  int value;
  if( !param->GetData(&value) ) return false;
  return value;
}

/**
 * value属性を取得する。(float)
 * @param ident     要素識別子
 * @return      value属性
 */
float CfgElem::GetFloatValue(CfgIdt ident) const {
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  float value;
  if( !param->GetData(&value) ) return false;
  return value;
}

/**
 * value属性を取得する。(double)
 * @param ident     要素識別子
 * @return      value属性
 */
double CfgElem::GetDoubleValue(CfgIdt ident) const {
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  double value;
  if( !param->GetData(&value) ) return false;
  return value;
}

/**
 * value属性を取得する。(char*)
 * @param ident     要素識別子
 * @return      value属性
 */
const char* CfgElem::GetStrValue(CfgIdt ident) const {
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return false;
  const char* value;
  if( !param->GetData(&value) ) return false;
  return value;
}

/**
 * 子Elem要素を取得する。
 * @param ident     要素識別子
 * @return      子Elem要素
 */
const CfgElem* CfgElem::GetElemValue(CfgIdt ident) const {
  return GetElemFirst(ident);
}

/**
 * value属性を取得する。(int)
 * @param ident_str     要素識別子文字列
 * @return    value属性
 */
int CfgElem::GetIntValue(const char* ident_str) const {
  if( !ident_str ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "ident_str = NULL");
    return 0;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) return 0;
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return 0;
  int value;
  if( !param->GetData(&value) ) return 0;
  return value;
}

/**
 * value属性を取得する。(float)
 * @param ident_str     要素識別子文字列
 * @return    value属性
 */
float CfgElem::GetFloatValue(const char* ident_str) const {
  if( !ident_str ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "ident_str = NULL");
    return 0.0f;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) return 0.0f;
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return 0.0f;
  float value;
  if( !param->GetData(&value) ) return 0.0f;
  return value;
}

/**
 * value属性を取得する。(double)
 * @param ident_str     要素識別子文字列
 * @return    value属性
 */
double CfgElem::GetDoubleValue(const char* ident_str) const {
  if( !ident_str ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "ident_str = NULL");
    return 0.0f;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) return 0.0f;
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return 0.0f;
  double value;
  if( !param->GetData(&value) ) return 0.0f;
  return value;
}

/**
 * value属性を取得する。(char*)
 * @param ident_str     要素識別子文字列
 * @return    value属性
 */
const char* CfgElem::GetStrValue(const char* ident_str) const {
  if( !ident_str ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "ident_str = NULL");
    return NULL;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) return NULL;
  const CfgParam* param = GetParamFirst(ident);
  if( !param ) return NULL;
  const char* value;
  if( !param->GetData(&value) ) return NULL;
  return value;
}

/**
 * 子Elem要素を取得する。
 * @param ident_str     要素識別子文字列
 * @return      子Elem要素
 */
const CfgElem* CfgElem::GetElemValue(const char* ident_str) const {
  if( !ident_str ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "ident_str = NULL");
    return NULL;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) return NULL;
  return GetElemFirst(ident);
}

/**
 * ベクトルvalue属性を取得する。(int)
 * @param[in] u     u成分要素識別子
 * @param[in] v     v成分要素識別子
 * @param[in] w     w成分要素識別子
 * @param[out] uv    u成分value属性
 * @param[out] vv    v成分value属性
 * @param[out] wv    w成分value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetVctValue(CfgIdt u, CfgIdt v, CfgIdt w,
                          int* uv, int* vv, int* wv) const {
  if( !uv || !vv || !wv ) {
    if( !uv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "uv = NULL");
    if( !vv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "vv = NULL");
    if( !wv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "wv = NULL");
    return false;
  }
  const CfgParam* uparam = NULL;
  const CfgParam* vparam = NULL;
  const CfgParam* wparam = NULL;
  if( !(uparam = GetParamFirst(u)) ) return false;
  if( !(vparam = GetParamFirst(v)) ) return false;
  if( !(wparam = GetParamFirst(w)) ) return false;
  int tmp_uv, tmp_vv, tmp_wv;
  if( !uparam->GetData(&tmp_uv) ) return false;
  if( !vparam->GetData(&tmp_vv) ) return false;
  if( !wparam->GetData(&tmp_wv) ) return false;
  *uv = tmp_uv;
  *vv = tmp_vv;
  *wv = tmp_wv;
  return true;
}

/**
 * ベクトルvalue属性を取得する。(float)
 * @param[in] u     u成分要素識別子
 * @param[in] v     v成分要素識別子
 * @param[in] w     w成分要素識別子
 * @param[out] uv    u成分value属性
 * @param[out] vv    v成分value属性
 * @param[out] wv    w成分value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetVctValue(CfgIdt u, CfgIdt v, CfgIdt w,
                          float* uv, float* vv, float* wv) const {
  if( !uv || !vv || !wv ) {
    if( !uv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "uv = NULL");
    if( !vv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "vv = NULL");
    if( !wv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "wv = NULL");
    return false;
  }
  const CfgParam* uparam = NULL;
  const CfgParam* vparam = NULL;
  const CfgParam* wparam = NULL;
  if( !(uparam = GetParamFirst(u)) ) return false;
  if( !(vparam = GetParamFirst(v)) ) return false;
  if( !(wparam = GetParamFirst(w)) ) return false;
  float tmp_uv, tmp_vv, tmp_wv;
  if( !uparam->GetData(&tmp_uv) ) return false;
  if( !vparam->GetData(&tmp_vv) ) return false;
  if( !wparam->GetData(&tmp_wv) ) return false;
  *uv = tmp_uv;
  *vv = tmp_vv;
  *wv = tmp_wv;
  return true;
}

/**
 * ベクトルvalue属性を取得する。(double)
 * @param[in] u     u成分要素識別子
 * @param[in] v     v成分要素識別子
 * @param[in] w     w成分要素識別子
 * @param[out] uv    u成分value属性
 * @param[out] vv    v成分value属性
 * @param[out] wv    w成分value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetVctValue(CfgIdt u, CfgIdt v, CfgIdt w,
                          double* uv, double* vv, double* wv) const {
  if( !uv || !vv || !wv ) {
    if( !uv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "uv = NULL");
    if( !vv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "vv = NULL");
    if( !wv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "wv = NULL");
    return false;
  }
  const CfgParam* uparam = NULL;
  const CfgParam* vparam = NULL;
  const CfgParam* wparam = NULL;
  if( !(uparam = GetParamFirst(u)) ) return false;
  if( !(vparam = GetParamFirst(v)) ) return false;
  if( !(wparam = GetParamFirst(w)) ) return false;
  double tmp_uv, tmp_vv, tmp_wv;
  if( !uparam->GetData(&tmp_uv) ) return false;
  if( !vparam->GetData(&tmp_vv) ) return false;
  if( !wparam->GetData(&tmp_wv) ) return false;
  *uv = tmp_uv;
  *vv = tmp_vv;
  *wv = tmp_wv;
  return true;
}

/**
 * ベクトルvalue属性を取得する。(char*)
 * @param[in] u     u成分要素識別子
 * @param[in] v     v成分要素識別子
 * @param[in] w     w成分要素識別子
 * @param[out] uv    u成分value属性
 * @param[out] vv    v成分value属性
 * @param[out] wv    w成分value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetVctValue(CfgIdt u, CfgIdt v, CfgIdt w,
                          const char** uv,
                          const char** vv,
                          const char** wv) const {
  if( !uv || !vv || !wv ) {
    if( !uv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "uv = NULL");
    if( !vv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "vv = NULL");
    if( !wv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "wv = NULL");
    return false;
  }
  const CfgParam* uparam = NULL;
  const CfgParam* vparam = NULL;
  const CfgParam* wparam = NULL;
  if( !(uparam = GetParamFirst(u)) ) return false;
  if( !(vparam = GetParamFirst(v)) ) return false;
  if( !(wparam = GetParamFirst(w)) ) return false;
  const char *tmp_uv, *tmp_vv, *tmp_wv;
  if( !uparam->GetData(&tmp_uv) ) return false;
  if( !vparam->GetData(&tmp_vv) ) return false;
  if( !wparam->GetData(&tmp_wv) ) return false;
  *uv = tmp_uv;
  *vv = tmp_vv;
  *wv = tmp_wv;
  return true;
}

/**
 * ベクトル子Elem要素を取得する。
 * @param[in] u     u成分要素識別子
 * @param[in] v     v成分要素識別子
 * @param[in] w     w成分要素識別子
 * @param[out] uv    u成分value属性
 * @param[out] vv    v成分value属性
 * @param[out] wv    w成分value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetVctValue(CfgIdt u, CfgIdt v, CfgIdt w,
                          const CfgElem** uv,
                          const CfgElem** vv,
                          const CfgElem** wv) const {
  if( !uv || !vv || !wv ) {
    if( !uv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "uv = NULL");
    if( !vv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "vv = NULL");
    if( !wv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "wv = NULL");
    return false;
  }
  const CfgElem* uelem = NULL;
  const CfgElem* velem = NULL;
  const CfgElem* welem = NULL;
  if( !(uelem = GetElemFirst(u)) ) return false;
  if( !(velem = GetElemFirst(v)) ) return false;
  if( !(welem = GetElemFirst(w)) ) return false;
  *uv = uelem;
  *vv = velem;
  *wv = welem;
  return true;
}

/**
 * ベクトルvalue属性を取得する。(int)
 * @param[in] u     u成分要素識別子文字列
 * @param[in] v     v成分要素識別子文字列
 * @param[in] w     w成分要素識別子文字列
 * @param[out] uv    u成分value属性
 * @param[out] vv    v成分value属性
 * @param[out] wv    w成分value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetVctValue(const char* u, const char* v, const char* w,
                          int* uv, int* vv, int* wv) const {
  if( !u || !v || !w || !uv || !vv || !wv ) {
    if( !u ) CfgErrMsg(_SKL_CFG_ERR_0055_, "u = NULL");
    if( !v ) CfgErrMsg(_SKL_CFG_ERR_0055_, "v = NULL");
    if( !w ) CfgErrMsg(_SKL_CFG_ERR_0055_, "w = NULL");
    if( !uv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "uv = NULL");
    if( !vv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "vv = NULL");
    if( !wv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "wv = NULL");
    return false;
  }
  CfgIdt uid, vid, wid;
  if( !uid.SetIdtFromStr(u) ) return false;
  if( !vid.SetIdtFromStr(v) ) return false;
  if( !wid.SetIdtFromStr(w) ) return false;
  const CfgParam* uparam = NULL;
  const CfgParam* vparam = NULL;
  const CfgParam* wparam = NULL;
  if( !(uparam = GetParamFirst(uid)) ) return false;
  if( !(vparam = GetParamFirst(vid)) ) return false;
  if( !(wparam = GetParamFirst(wid)) ) return false;
  int tmp_uv, tmp_vv, tmp_wv;
  if( !uparam->GetData(&tmp_uv) ) return false;
  if( !vparam->GetData(&tmp_vv) ) return false;
  if( !wparam->GetData(&tmp_wv) ) return false;
  *uv = tmp_uv;
  *vv = tmp_vv;
  *wv = tmp_wv;
  return true;
}

/**
 * ベクトルvalue属性を取得する。(float)
 * @param[in] u     u成分要素識別子文字列
 * @param[in] v     v成分要素識別子文字列
 * @param[in] w     w成分要素識別子文字列
 * @param[out] uv    u成分value属性
 * @param[out] vv    v成分value属性
 * @param[out] wv    w成分value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetVctValue(const char* u, const char* v, const char* w,
                          float* uv, float* vv, float* wv) const {
  if( !u || !v || !w || !uv || !vv || !wv ) {
    if( !u ) CfgErrMsg(_SKL_CFG_ERR_0055_, "u = NULL");
    if( !v ) CfgErrMsg(_SKL_CFG_ERR_0055_, "v = NULL");
    if( !w ) CfgErrMsg(_SKL_CFG_ERR_0055_, "w = NULL");
    if( !uv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "uv = NULL");
    if( !vv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "vv = NULL");
    if( !wv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "wv = NULL");
    return false;
  }
  CfgIdt uid, vid, wid;
  if( !uid.SetIdtFromStr(u) ) return false;
  if( !vid.SetIdtFromStr(v) ) return false;
  if( !wid.SetIdtFromStr(w) ) return false;
  const CfgParam* uparam = NULL;
  const CfgParam* vparam = NULL;
  const CfgParam* wparam = NULL;
  if( !(uparam = GetParamFirst(uid)) ) return false;
  if( !(vparam = GetParamFirst(vid)) ) return false;
  if( !(wparam = GetParamFirst(wid)) ) return false;
  float tmp_uv, tmp_vv, tmp_wv;
  if( !uparam->GetData(&tmp_uv) ) return false;
  if( !vparam->GetData(&tmp_vv) ) return false;
  if( !wparam->GetData(&tmp_wv) ) return false;
  *uv = tmp_uv;
  *vv = tmp_vv;
  *wv = tmp_wv;
  return true;
}

/**
 * ベクトルvalue属性を取得する。(double)
 * @param[in] u     u成分要素識別子文字列
 * @param[in] v     v成分要素識別子文字列
 * @param[in] w     w成分要素識別子文字列
 * @param[out] uv    u成分value属性
 * @param[out] vv    v成分value属性
 * @param[out] wv    w成分value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetVctValue(const char* u, const char* v, const char* w,
                          double* uv, double* vv, double* wv) const {
  if( !u || !v || !w || !uv || !vv || !wv ) {
    if( !u ) CfgErrMsg(_SKL_CFG_ERR_0055_, "u = NULL");
    if( !v ) CfgErrMsg(_SKL_CFG_ERR_0055_, "v = NULL");
    if( !w ) CfgErrMsg(_SKL_CFG_ERR_0055_, "w = NULL");
    if( !uv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "uv = NULL");
    if( !vv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "vv = NULL");
    if( !wv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "wv = NULL");
    return false;
  }
  CfgIdt uid, vid, wid;
  if( !uid.SetIdtFromStr(u) ) return false;
  if( !vid.SetIdtFromStr(v) ) return false;
  if( !wid.SetIdtFromStr(w) ) return false;
  const CfgParam* uparam = NULL;
  const CfgParam* vparam = NULL;
  const CfgParam* wparam = NULL;
  if( !(uparam = GetParamFirst(uid)) ) return false;
  if( !(vparam = GetParamFirst(vid)) ) return false;
  if( !(wparam = GetParamFirst(wid)) ) return false;
  double tmp_uv, tmp_vv, tmp_wv;
  if( !uparam->GetData(&tmp_uv) ) return false;
  if( !vparam->GetData(&tmp_vv) ) return false;
  if( !wparam->GetData(&tmp_wv) ) return false;
  *uv = tmp_uv;
  *vv = tmp_vv;
  *wv = tmp_wv;
  return true;
}

/**
 * ベクトルvalue属性を取得する。(char*)
 * @param[in] u     u成分要素識別子文字列
 * @param[in] v     v成分要素識別子文字列
 * @param[in] w     w成分要素識別子文字列
 * @param[out] uv    u成分value属性
 * @param[out] vv    v成分value属性
 * @param[out] wv    w成分value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetVctValue(const char* u, const char* v, const char* w,
                          const char** uv,
                          const char** vv,
                          const char** wv) const {
  if( !u || !v || !w || !uv || !vv || !wv ) {
    if( !u ) CfgErrMsg(_SKL_CFG_ERR_0055_, "u = NULL");
    if( !v ) CfgErrMsg(_SKL_CFG_ERR_0055_, "v = NULL");
    if( !w ) CfgErrMsg(_SKL_CFG_ERR_0055_, "w = NULL");
    if( !uv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "uv = NULL");
    if( !vv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "vv = NULL");
    if( !wv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "wv = NULL");
    return false;
  }
  CfgIdt uid, vid, wid;
  if( !uid.SetIdtFromStr(u) ) return false;
  if( !vid.SetIdtFromStr(v) ) return false;
  if( !wid.SetIdtFromStr(w) ) return false;
  const CfgParam* uparam = NULL;
  const CfgParam* vparam = NULL;
  const CfgParam* wparam = NULL;
  if( !(uparam = GetParamFirst(uid)) ) return false;
  if( !(vparam = GetParamFirst(vid)) ) return false;
  if( !(wparam = GetParamFirst(wid)) ) return false;
  const char *tmp_uv, *tmp_vv, *tmp_wv;
  if( !uparam->GetData(&tmp_uv) ) return false;
  if( !vparam->GetData(&tmp_vv) ) return false;
  if( !wparam->GetData(&tmp_wv) ) return false;
  *uv = tmp_uv;
  *vv = tmp_vv;
  *wv = tmp_wv;
  return true;
}

/**
 * ベクトル子Elem要素を取得する。
 * @param[in] u     u成分要素識別子文字列
 * @param[in] v     v成分要素識別子文字列
 * @param[in] w     w成分要素識別子文字列
 * @param[out] uv    u成分value属性
 * @param[out] vv    v成分value属性
 * @param[out] wv    w成分value属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetVctValue(const char* u, const char* v, const char* w,
                          const CfgElem** uv,
                          const CfgElem** vv,
                          const CfgElem** wv) const {
  if( !u || !v || !w || !uv || !vv || !wv ) {
    if( !u ) CfgErrMsg(_SKL_CFG_ERR_0055_, "u = NULL");
    if( !v ) CfgErrMsg(_SKL_CFG_ERR_0055_, "v = NULL");
    if( !w ) CfgErrMsg(_SKL_CFG_ERR_0055_, "w = NULL");
    if( !uv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "uv = NULL");
    if( !vv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "vv = NULL");
    if( !wv ) CfgErrMsg(_SKL_CFG_ERR_0055_, "wv = NULL");
    return false;
  }
  CfgIdt uid, vid, wid;
  if( !uid.SetIdtFromStr(u) ) return false;
  if( !vid.SetIdtFromStr(v) ) return false;
  if( !wid.SetIdtFromStr(w) ) return false;
  const CfgElem* uelem = NULL;
  const CfgElem* velem = NULL;
  const CfgElem* welem = NULL;
  if( !(uelem = GetElemFirst(uid)) ) return false;
  if( !(velem = GetElemFirst(vid)) ) return false;
  if( !(welem = GetElemFirst(wid)) ) return false;
  *uv = uelem;
  *vv = velem;
  *wv = welem;
  return true;
}

/**
 * num階層下にあるvalue属性を取得する。(int)
 * @param[out] val   value属性
 * @param[in] num     探索階層数
 * @param[in] ...     探索name属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(int* val, unsigned int num, ...) const {
  if( !val ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "val = NULL");
    return false;
  }
  va_list vl;
  va_start(vl, num);
  const char* ident_str = va_arg(vl, const char*);
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) {
    va_end(vl);
    return false;
  }
  const CfgElem* elem = GetElemFirst(ident);
  if( !elem ) {
    va_end(vl);
    return false;
  }

  register unsigned int n;
  for(n=1; n<num-1; n++) {
    if( !(ident_str = va_arg(vl, const char*)) ) {
      va_end(vl);
      CfgErrMsg(_SKL_CFG_ERR_0058_);
      return false;
    }
    if( !ident.SetIdtFromStr(ident_str) ) {
      va_end(vl);
      return false;
    }
    if( !(elem = elem->GetElemFirst(ident)) ) {
      va_end(vl);
      return false;
    }
  }

  ident_str = va_arg(vl, const char*);
  if( !ident_str ) {
    va_end(vl);
    CfgErrMsg(_SKL_CFG_ERR_0058_);
    return false;
  }
  if( !ident.SetIdtFromStr(ident_str) ) {
    va_end(vl);
    return false;
  }
  const CfgParam* param = elem->GetParamFirst(ident);
  if( !param ) {
    va_end(vl);
    return false;
  }
  if( !param->GetData(val) ) {
    va_end(vl);
    return false;
  }

  va_end(vl);
  return true;
}

/**
 * num階層下にあるvalue属性を取得する。(float)
 * @param[out] val   value属性
 * @param[in] num     探索階層数
 * @param[in] ...     探索name属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(float* val, unsigned int num, ...) const {
  if( !val ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "val = NULL");
    return false;
  }
  va_list vl;
  va_start(vl, num);
  const char* ident_str = va_arg(vl, const char*);
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) {
    va_end(vl);
    return false;
  }
  const CfgElem* elem = GetElemFirst(ident);
  if( !elem ) {
    va_end(vl);
    return false;
  }

  register unsigned int n;
  for(n=1; n<num-1; n++) {
    if( !(ident_str = va_arg(vl, const char*)) ) {
      va_end(vl);
      CfgErrMsg(_SKL_CFG_ERR_0058_);
      return false;
    }
    if( !ident.SetIdtFromStr(ident_str) ) {
      va_end(vl);
      return false;
    }
    if( !(elem = elem->GetElemFirst(ident)) ) {
      va_end(vl);
      return false;
    }
  }

  ident_str = va_arg(vl, const char*);
  if( !ident_str ) {
    CfgErrMsg(_SKL_CFG_ERR_0058_);
    va_end(vl);
    return false;
  }
  if( !ident.SetIdtFromStr(ident_str) ) {
    va_end(vl);
    return false;
  }
  const CfgParam* param = elem->GetParamFirst(ident);
  if( !param ) {
    va_end(vl);
    return false;
  }
  if( !param->GetData(val) ) {
    va_end(vl);
    return false;
  }

  va_end(vl);
  return true;
}

/**
 * num階層下にあるvalue属性を取得する。(double)
 * @param[out] val   value属性
 * @param[in] num     探索階層数
 * @param[in] ...     探索name属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(double* val, unsigned int num, ...) const {
  if( !val ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "val = NULL");
    return false;
  }
  va_list vl;
  va_start(vl, num);
  const char* ident_str = va_arg(vl, const char*);
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) {
    va_end(vl);
    return false;
  }
  const CfgElem* elem = GetElemFirst(ident);
  if( !elem ) {
    va_end(vl);
    return false;
  }

  register unsigned int n;
  for(n=1; n<num-1; n++) {
    if( !(ident_str = va_arg(vl, const char*)) ) {
      va_end(vl);
      CfgErrMsg(_SKL_CFG_ERR_0058_);
      return false;
    }
    if( !ident.SetIdtFromStr(ident_str) ) {
      va_end(vl);
      return false;
    }
    if( !(elem = elem->GetElemFirst(ident)) ) {
      va_end(vl);
      return false;
    }
  }

  ident_str = va_arg(vl, const char*);
  if( !ident_str ) {
    CfgErrMsg(_SKL_CFG_ERR_0058_);
    va_end(vl);
    return false;
  }
  if( !ident.SetIdtFromStr(ident_str) ) {
    va_end(vl);
    return false;
  }
  const CfgParam* param = elem->GetParamFirst(ident);
  if( !param ) {
    va_end(vl);
    return false;
  }
  if( !param->GetData(val) ) {
    va_end(vl);
    return false;
  }

  va_end(vl);
  return true;
}

/**
 * num階層下にあるvalue属性を取得する。(char*)
 * @param[out] val   value属性
 * @param[in] num     探索階層数
 * @param[in] ...     探索name属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(const char** val, unsigned int num, ...) const {
  if( !val ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "val = NULL");
    return false;
  }
  va_list vl;
  va_start(vl, num);
  const char* ident_str = va_arg(vl, const char*);
  if( !ident_str ) {
    CfgErrMsg(_SKL_CFG_ERR_0058_);
    va_end(vl);
    return false;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) {
    va_end(vl);
    return false;
  }
  const CfgElem* elem = GetElemFirst(ident);
  if( !elem ) {
    va_end(vl);
    return false;
  }

  register unsigned int n;
  for(n=1; n<num-1; n++) {
    if( !(ident_str = va_arg(vl, const char*)) ) {
      CfgErrMsg(_SKL_CFG_ERR_0058_);
      va_end(vl);
      return false;
    }
    if( !ident.SetIdtFromStr(ident_str) ) {
      va_end(vl);
      return false;
    }
    if( !(elem = elem->GetElemFirst(ident)) ) {
      va_end(vl);
      return false;
    }
  }

  ident_str = va_arg(vl, const char*);
  if( !ident_str ) {
    CfgErrMsg(_SKL_CFG_ERR_0058_);
    va_end(vl);
    return false;
  }
  if( !ident.SetIdtFromStr(ident_str) ) {
    va_end(vl);
    return false;
  }
  const CfgParam* param = elem->GetParamFirst(ident);
  if( !param ) {
    va_end(vl);
    return false;
  }
  if( !param->GetData(val) ) {
    va_end(vl);
    return false;
  }

  va_end(vl);
  return true;
}

/**
 * num階層下にあるElem要素を取得する。(char*)
 * @param[out] val   Elem要素
 * @param[in] num     探索階層数
 * @param[in] ...     探索name属性
 * @return      true=success, false=fault
 */
bool CfgElem::GetValue(const CfgElem** val, unsigned int num, ...) const {
  if( !val ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "val = NULL");
    return false;
  }
  va_list vl;
  va_start(vl, num);
  const char* ident_str = va_arg(vl, const char*);
  if( !ident_str ) {
    CfgErrMsg(_SKL_CFG_ERR_0058_);
    va_end(vl);
    return false;
  }
  CfgIdt ident;
  if( !ident.SetIdtFromStr(ident_str) ) {
    va_end(vl);
    return false;
  }
  const CfgElem* elem = GetElemFirst(ident);
  if( !elem ) {
    va_end(vl);
    return false;
  }

  register unsigned int n;
  for(n=1; n<num; n++) {
    if( !(ident_str = va_arg(vl, const char*)) ) {
      CfgErrMsg(_SKL_CFG_ERR_0058_);
      va_end(vl);
      return false;
    }
    if( !ident.SetIdtFromStr(ident_str) ) {
      va_end(vl);
      return false;
    }
    if( !(elem = elem->GetElemFirst(ident)) ) {
      va_end(vl);
      return false;
    }
  }
  *val = elem;

  va_end(vl);
  return true;
}

/*
 *  Add XML Parser Interface 2007.09 end
 */
/**
 * コンストラクタ
 */
CfgParam::CfgParam()
{
  m_dType = DTYPE_UNKNOWN;
  m_valFlag = false;
}

/**
 * デストラクタ
 */
CfgParam::~CfgParam()
{
  ClearStr();
}

/**
 * dType属性を取得する。
 * @return    データ型タイプ
 */
DataType CfgParam::GetDataType() const
{
  return m_dType;
}

/**
 * value属性を取得する。(PrmVal)
 * @param[out] val   value属性
 * @return      データ型タイプ
 */
DataType CfgParam::GetData(PrmVal& val) const
{
  switch(m_dType){
    case _INT:
      val.ival = m_value.ival;
      break;
    case _REAL:
      val.fval = m_value.fval;
      break;
    case _STR:
      {
        if( !m_value.sval ) return DTYPE_UNKNOWN;
        unsigned int len = strlen(m_value.sval)+1;
        val.sval = new char[len];
        if( !val.sval ) return DTYPE_UNKNOWN;
        strncpy(val.sval, m_value.sval, len);
      }
      break;
    default:
      break;
  }
  return m_dType;
}


/**
 * value属性を取得する。(int)
 * @param[out] val   value属性
 * @return      true=success, false=fault
 */
bool CfgParam::GetData(int* val) const
{
  if( !val ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "val = NULL");
    return false;
  }
  if( m_dType != _INT ) {
    CfgErrMsg(_SKL_CFG_ERR_0059_, "int");
    return false;
  }
  *val = m_value.ival;
  return true;
}

/**
 * value属性を取得する。(float)
 * @param[out] val   value属性
 * @return      true=success, false=fault
 */
bool CfgParam::GetData(float* val) const
{
  if( !val ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "val = NULL");
    return false;
  }
  if( m_dType != _REAL ) {
    CfgErrMsg(_SKL_CFG_ERR_0059_, "float");
    return false;
  }
  *val = (float)m_value.fval;
  return true;
}

/**
 * value属性を取得する。(double)
 * @param[out] val   value属性
 * @return      true=success, false=fault
 */
bool CfgParam::GetData(double* val) const
{
  if( !val ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "val = NULL");
    return false;
  }
  if( m_dType != _REAL ) {
    CfgErrMsg(_SKL_CFG_ERR_0059_, "double");
    return false;
  }
  *val = m_value.fval;
  return true;
}

/**
 * value属性を取得する。(char*)
 * @param[out] val   value属性
 * @return      true=success, false=fault
 */
bool CfgParam::GetData(const char** val) const
{
  if( !val ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "val = NULL");
    return false;
  }
  if( m_dType != _STR ) {
    CfgErrMsg(_SKL_CFG_ERR_0059_, "char*");
    return false;
  }
  *val = m_value.sval;
  return true;
}

/**
 * value属性が設定されているかチェックする。
 * @return    true=value属性設定済み
 */
bool CfgParam::IsSetValue() const
{
  return m_valFlag;
}

/**
 * value属性を設定する。(double)
 * @param val   value属性
 * @return      true=success, false=fault
 */
bool CfgParam::SetData(double val)
{
  if( !ClearStr() ) return false;
  m_value.fval = val;
  m_dType = _REAL;
  return true;
}

/**
 * value属性を設定する。(int)
 * @param val   value属性
 * @return      true=success, false=fault
 */
bool CfgParam::SetData(int val)
{
  if( !ClearStr() ) return false;
  m_value.ival = val;
  m_dType = _INT;
  return false;
}

/**
 * value属性を設定する。(char*)
 * @param val   value属性
 * @return      true=success, false=fault
 */
bool CfgParam::SetData(const char* val)
{
  if( !val ) return false;
  if( !ClearStr() ) return false;
  unsigned int len = strlen(val)+1;
  m_value.sval = new char[len];
  if( !m_value.sval ){
    m_dType = DTYPE_UNKNOWN;
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  strncpy(m_value.sval, val, len);
  m_dType = _STR;
  return true;
}

/**
 * value属性を設定する。(DataType)
 * @param type   データ型タイプ
 * @param val   value属性
 * @return      true=success, false=fault
 */
bool CfgParam::SetData(DataType type, char* val)
{
  if( type == DTYPE_UNKNOWN ) {
    CfgErrMsg("Invalid data type.");
    return false;
  }
  if( !val ) {
    CfgErrMsg(_SKL_CFG_ERR_0055_, "val = NULL");
    return false;
  }

  if( !ClearStr() ) return false;
  switch(type){
    case _INT:
      m_value.ival = atoi(val);
      m_dType = _INT;
      break;
    case _REAL:
      m_value.fval = atof(val);
      m_dType = _REAL;
      break;
    case _STR:
      {
        unsigned int len = strlen(val)+1;
        m_value.sval = new char[len];
        if( !m_value.sval ){
          CfgErrMsg(_SKL_CFG_ERR_0005_);
          m_dType = DTYPE_UNKNOWN;
          return false;
        }
        strncpy(m_value.sval, val, len);
        m_dType = _STR;
      }
      break;
    default:
      break;
  }

  return true;
}

/**
 * 設定属性をクリアする。
 */
bool CfgParam::ClearStr()
{
  if( (m_dType==_STR)&& m_value.sval ) delete [] m_value.sval;
  m_value.sval = NULL;
  m_dType = DTYPE_UNKNOWN;
  return true;
}

/**
 * XML要素からパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool CfgParam::Parse(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_USRDEF_PARAM_,
                     strlen(_SKL_CFG_NNAME_USRDEF_PARAM_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  return true;
}

/**
 * XML要素から属性を設定する。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool CfgParam::SetAttr(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_USRDEF_PARAM_,
                     strlen(_SKL_CFG_NNAME_USRDEF_PARAM_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlChar* tmpStr = NULL;

  // get name attribute
  if( !(tmpStr = xmlGetProp(node,
                    (const xmlChar*)_SKL_CFG_ANAME_USRDEF_PARAM_NAME_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_USRDEF_PARAM_,
              _SKL_CFG_ANAME_USRDEF_PARAM_NAME_);
    return false;
  }
  if( !SetName((const char*)tmpStr) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    xmlFree(tmpStr);
    return false;
  }
  xmlFree(tmpStr);

#if 0 // ----- modify 20100510 begin -----
  // get dtype attribute
  DataType dtype = DTYPE_UNKNOWN;
  if( !(tmpStr = xmlGetProp(node,
                    (const xmlChar*)_SKL_CFG_ANAME_USRDEF_PARAM_DTYPE_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_USRDEF_PARAM_,
              _SKL_CFG_ANAME_USRDEF_PARAM_DTYPE_);
    return false;
  }
  if( !StrCaseCompare((const char*)tmpStr,
                      _SKL_CFG_AVAL_USRDEF_PARAM_INT_) ) dtype = _INT;
  else if( !StrCaseCompare((const char*)tmpStr,
                           _SKL_CFG_AVAL_USRDEF_PARAM_REAL_) ) dtype = _REAL;
  else if( !StrCaseCompare((const char*)tmpStr,
                           _SKL_CFG_AVAL_USRDEF_PARAM_STRING_) ) dtype = _STR;
  else dtype = DTYPE_UNKNOWN;
  xmlFree(tmpStr);

  // get value attribute
  if( !(tmpStr = xmlGetProp(node,
                    (const xmlChar*)_SKL_CFG_ANAME_USRDEF_PARAM_VALUE_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_USRDEF_PARAM_,
              _SKL_CFG_ANAME_USRDEF_PARAM_VALUE_);
    return false;
  }
  if( !SetData(dtype, (char*)tmpStr) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    xmlFree(tmpStr);
    return false;
  }
  xmlFree(tmpStr);
#else
  // get value and dtype attribute
  DataType dtype = DTYPE_UNKNOWN;
  if( tmpStr = xmlGetProp(node,
                    (const xmlChar*)_SKL_CFG_ANAME_USRDEF_PARAM_VALUE_) ){
    xmlChar* tmpStr2 = NULL;
    if( tmpStr2 = xmlGetProp(node,
                      (const xmlChar*)_SKL_CFG_ANAME_USRDEF_PARAM_DTYPE_) ){
      if( !StrCaseCompare((const char*)tmpStr2,
                          _SKL_CFG_AVAL_USRDEF_PARAM_INT_) ) dtype = _INT;
      else if( !StrCaseCompare((const char*)tmpStr2,
                               _SKL_CFG_AVAL_USRDEF_PARAM_REAL_) ) dtype = _REAL;
      else if( !StrCaseCompare((const char*)tmpStr2,
                               _SKL_CFG_AVAL_USRDEF_PARAM_STRING_) ) dtype = _STR;
      else dtype = DTYPE_UNKNOWN;
      xmlFree(tmpStr2);
    }

    if( !SetData(dtype, (char*)tmpStr) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      xmlFree(tmpStr);
      return false;
    }
    xmlFree(tmpStr);
    m_valFlag = true;
  }
#endif  // ----- modify 20100511 end -----

  // get id attribute
  if( (tmpStr = xmlGetProp(node,
                   (const xmlChar*)_SKL_CFG_ANAME_USRDEF_PARAM_ID_)) ){
    int id = atoi((const char*)tmpStr);
    if( !SetID(id) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      xmlFree(tmpStr);
      return false;
    }
    xmlFree(tmpStr);
  }

  // get comment attribute
  if( (tmpStr = xmlGetProp(node,
                   (const xmlChar*)_SKL_CFG_ANAME_USRDEF_PARAM_COMMENT_)) ){
    if( !SetComment((const char*)tmpStr) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      xmlFree(tmpStr);
      return false;
    }
    xmlFree(tmpStr);
  }

  return true;
}

/**
 * 属性値を出力する。
 */
void CfgBase::Print() const {
  if( m_name ) printf("m_name=\"%s\",", m_name);
  else         printf("m_name=NULL,");
  printf("id=%d,", m_id);
  if( m_idFlg ) printf("m_idFlg=TRUE,");
  else          printf("m_idFlg=false,");
  if( m_comment ) printf("m_comment=\"%s\",", m_comment);
  else            printf("m_comment=NULL,");
}

/**
 * 属性値を出力する。
 */
void CfgParam::Print() const {
  printf("<Param ");
  CfgBase::Print();
  switch(m_dType) {
    case _INT:
      printf("m_dType=INT,value=%d,", m_value.ival);
      break;
    case _REAL:
      printf("m_dType=REAL,value=%e,", m_value.fval);
      break;
    case _STR:
      printf("m_dType=STRING,value=%s,", m_value.sval);
      break;
    default:
      break;
  }
  printf("/>\n");
}

/**
 * 属性値を出力する。
 */
void CfgElem::Print() const {
  printf("<Elem ");
  CfgBase::Print();
  printf(">\n");

  const CfgParam *param = GetParamFirst();
  while( param ) {
    param->Print();
    param = GetParamNext(param);
  }

  const CfgElem* elem = GetElemFirst();
  while( elem ) {
    elem->Print();
    elem = GetElemNext(elem);
  }

/*
  list<CfgParam*>::const_iterator prmitr = m_prmList.begin();
  list<CfgParam*>::const_iterator prmitrEnd = m_prmList.end();
  for( ; prmitr != prmitrEnd; prmitr++ ) (*prmitr)->Print();
  list<CfgElem*>::const_iterator elmitr = m_elmList.begin();
  list<CfgElem*>::const_iterator elmitrEnd = m_elmList.end();
  for( ; elmitr != elmitrEnd; elmitr++ ) (*elmitr)->Print();
*/
  printf("</Elem>\n");
}

} // end of namespace SklCfg

