/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */
/**
 * @file SklIBCConfig.C
 * InnerBoundaryコンフィグレーション情報格納クラスソースファイル
 * @author     riken
 * @date       2010/10/01
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "skl_config/SklCfgDefName.h"
#include "skl_config/SklIBCConfig.h"

namespace SklCfg {

/**
 * コンストラクタ
 */
SklIBCConfig::SklIBCConfig()
{
  Initialize();
}

/**
 * デストラクタ
 */
SklIBCConfig::~SklIBCConfig()
{
  Initialize();
}

/**
 * 初期化を行う。
 */
void SklIBCConfig::Initialize()
{
  m_bnd = NULL;
}

/**
 * BoundaryXMLファイルからパースを行う。
 *
 * 逐次、並列実行によりパース方法を切り替える。
 * @param fname   XMLファイル名
 * @return      true=success, false=fault
 */
// modify for vsphere_cmp at 2008/08/15 by @hira
//     from ParaMngr to ParaCmpo
bool SklIBCConfig::SklParseConfig(const char* fname)
{
  if( !fname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0017_);
    return false;
  }

  if (splIsParallel()) {
    if( !SklParseConfigFileForParallel(fname) ) return false;
  } else {
    if( !SklParseConfigFile(fname) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
  }

  return true;
}

/**
 * BoundaryXMLファイルからパースを行う。(逐次実行)
 *
 * 自ノードが直接ファイルをオープンする。
 * @param fname   XMLファイル名
 * @return      true=success, false=fault
 */
bool SklIBCConfig::SklParseConfigFile(const char* fname)
{
  if( !fname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0017_);
    return false;
  }

  int fd;
  struct stat st_buf;

  if( (fd = open(fname, O_RDONLY)) < 0 ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0018_, fname);
    return false;
  }
  if( fstat(fd, &st_buf) < 0 ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0019_);
    return false;
  }
  if( st_buf.st_size < 1 ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0020_);
    return false;
  }

  char* strs = new char[st_buf.st_size + 1];
  if( !strs ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  size_t str_size = st_buf.st_size;
  size_t sz = str_size;
  do {
    sz -= read(fd, &strs[st_buf.st_size-sz], sz);
  } while ( sz );
  close(fd);

  if( !ParseConfigStream(str_size, strs) ){
    delete [] strs;
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  delete [] strs;

  return true;
}

/**
 * BoundaryXMLファイルからパースを行う。(並列実行)
 *
 * マスタノードがファイルをオープンし、配信する。
 * @param fname   XMLファイル名
 * @return      true=success, false=fault
 */
// modify for vsphere_cmp at 2008/08/15 by @hira
//     from ParaMngr to ParaCmpo
bool SklIBCConfig::SklParseConfigFileForParallel(const char* fname)
{
  int flg, ret;
  flg = 0;

  int myID = 0;
  if (SPL_Comm_rank(SPL_COMM_WORLD, &myID) != SPL_MPI_SUCCESS) {
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Comm_rank error.");
    return false;
  }
  if( myID == 0 ){
    if( !fname ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0017_);
      flg = 1;
    }
  }

  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) return false;

  int fd = -1;
  struct stat st_buf;
  if( myID == 0 ){
    if( (fd = open(fname, O_RDONLY)) < 0 ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0018_, fname);
      flg = 1;
    }
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) return false;

  if( myID == 0 ){
    if( fstat(fd, &st_buf) < 0 ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0019_);
      flg = 1;
    }
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) { if( myID == 0 ) close(fd); return false; }

  if( myID == 0 ){
    if( st_buf.st_size < 1 ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0020_);
      flg = 1;
    }
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) { if( myID == 0 ) close(fd); return false; }

  off_t len = 0;
  if( myID == 0 ){
    len = st_buf.st_size;
  }
  SPL_Datatype dt_type = SPL_DATATYPE_NULL;
  if( sizeof(off_t) == sizeof(unsigned) ) {
    dt_type = SPL_UNSIGNED;
  }
  else if( sizeof(off_t) == sizeof(long long) ) {
    dt_type = SPL_LONG_LONG;
  }
  if( SPL_Bcast(&len, 1, dt_type, 0, SPL_COMM_WORLD) != SPL_MPI_SUCCESS ){
    if( myID == 0 ) close(fd); return false;
  }

  char* strs = new char[len + 1];
  if( !strs ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    flg = 1;
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) { if( myID == 0 ) close(fd); return false; }

  if( myID == 0 ){
    size_t str_size = st_buf.st_size;
    size_t sz = str_size;
    do {
      sz -= read(fd, &strs[st_buf.st_size-sz], sz);
    } while ( sz );
    close(fd);
  }
  if( SPL_Bcast(strs, len+1, SPL_CHAR, 0, SPL_COMM_WORLD) != SPL_MPI_SUCCESS ){
    delete [] strs;
    return false;
  }

  if( !ParseConfigStream(len, strs) ){
    delete [] strs;
    flg = 1;
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) return false;

  delete [] strs;
  return true;
}

/**
 * XML文字列からパースを行う。
 * @param sz      XML文字列サイズ
 * @param stream    XML文字列
 * @return      true=success, false=fault
 */
bool SklIBCConfig::ParseConfigStream(const size_t sz, const char* stream)
{
  if( (sz <= 0) || !stream ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0021_);
    return false;
  }
  xmlDocPtr doc = ::xmlParseMemory(stream, sz);
  if ( !doc ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0022_);
    return false;
  }

  xmlNodePtr cur = xmlDocGetRootElement(doc);
  while ( cur ) {
    //---- skip Comments
    if ( cur->type == XML_COMMENT_NODE ) {
      cur = cur->next;
      continue;
    }
    break;
  }
  if( ! cur ){
    xmlFreeDoc(doc);
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0023_);
    return false;
  }

  //---- parse tree
  if ( !SklParse(cur) ){
    xmlFreeDoc(doc);
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  xmlFreeDoc(doc);
  return true;
}

/**
 * XML要素からパースを行う。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklIBCConfig::SklParse(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !Parse(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  return true;
}

/**
 * XML要素からパースを行う。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklIBCConfig::Parse(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_BC_TABLE_,
                     strlen(_SKL_CFG_NNAME_BC_TABLE_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlNodePtr cur = node->xmlChildrenNode;
  while ( cur ) {
    if ( cur->type == XML_COMMENT_NODE ) {
      cur = cur->next;
      continue;
    }

    if( !StrCaseCompare((const char*)cur->name,
                        _SKL_CFG_NNAME_INNER_BND_,
                        strlen(_SKL_CFG_NNAME_INNER_BND_)) ){
      if( !m_bnd ) {
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
      m_bnd->ClearRoot();
      if( !m_bnd->ParseNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }

    cur = cur->next;
  }

  return true;
}

/**
 * InnerBoundary要素を設定する。
 * @param bnd   InnerBoundary要素
 */
void SklIBCConfig::SetBnd(SklCfgInnerBnd* bnd) {
  m_bnd = bnd;
}

/**
 * InnerBoundary要素を取得する。
 * @return   InnerBoundary要素
 */
SklCfgInnerBnd* SklIBCConfig::GetBnd() const {
  return m_bnd;
}

#ifdef DEBUG
void SklIBCConfig::Print() const
{
}
#endif //DEBUG

} // end of namespace SklCfg

