/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2009
 *
 */
/**
 * @file SklSolverConfig.C
 * V-Sphereソルバーコンフィグレーション情報格納クラスソースファイル
 * @author     riken
 * @date       2010/10/01
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "skl_config/SklCfgDefName.h"
#include "skl_config/SklSolverConfig.h"

namespace SklCfg {
/**
 * コンストラクタ
 */
SklSolverConfig::SklSolverConfig()
{
  Initialize();
}

/**
 * デストラクタ
 */
SklSolverConfig::~SklSolverConfig()
{
  if( m_domainInfo ) delete m_domainInfo;
  if( m_steer ) delete m_steer;
  if( m_parameter ) delete m_parameter;
  if( m_innerBnd ) delete m_innerBnd;
  if( m_outerBnd ) delete m_outerBnd;
  if( m_mdmTable ) delete m_mdmTable;
  if( m_mbxTable ) delete m_mbxTable;
  if( m_usrDefNode ) delete m_usrDefNode;

  Initialize();
}

/**
 * 初期化を行う。
 */
void SklSolverConfig::Initialize()
{
  m_domainInfo = NULL;
  m_steer = NULL;
  m_parameter = NULL;
  m_innerBnd = NULL;
  m_outerBnd = NULL;
  m_mdmTable = NULL;
  m_mbxTable = NULL;
  m_usrDefNode = NULL;
  m_solverType = NULL;
}

/**
 * コンフィグレーションXMLファイルからパースを行う。
 *
 * 逐次、並列実行によりパース方法を切り替える。
 * @param fname   XMLファイル名
 * @return      true=success, false=fault
 */
// modify for vsphere_cmp at 2008/08/15 by @hira
//     from ParaMngr to ParaCmpo
bool SklSolverConfig::SklParseConfig(const char* fname)
{
  if( !fname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0017_);
    return false;
  }

  // read Sphere-Configuration file
  if (splIsParallel()) {
    if( !SklParseConfigFileForParallel(fname) ) return false;
  } else {
    if( !SklParseConfigFile(fname) ) return false;
  }

  // read Boundary-Condition table file
  if( !m_innerBnd ) {
    if( !ReadBCFileForInner() ) return false;
  }
  if( !m_outerBnd ) {
    if( !ReadBCFileForOuter() ) return false;
  }

  // read Medium table file
  if( !m_mdmTable ) {
    if( !ReadMDMTableFile() ) return false;
  }

  // read MultiBox table file
  if( !m_mbxTable ) {
    if( !ReadMBoxTableFile() ) return false;
  }

  return true;
}

/**
 * コンフィグレーションXMLファイルからパースを行う。(逐次実行)
 *
 * 自ノードが直接ファイルをオープンする。
 * @param fname   XMLファイル名
 * @return      true=success, false=fault
 */
bool SklSolverConfig::SklParseConfigFile(const char* fname)
{
  if( !fname ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0017_);
    return false;
  }

  int fd;
  struct stat st_buf;

  if( (fd = open(fname, O_RDONLY)) < 0 ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0018_, fname);
    return false;
  }
  if( fstat(fd, &st_buf) < 0 ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0019_);
    return false;
  }
  if( st_buf.st_size < 1 ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0020_);
    return false;
  }

  char* strs = new char[st_buf.st_size + 1];
  if( !strs ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    return false;
  }
  size_t str_size = st_buf.st_size;
  size_t sz = str_size;
  do {
    sz -= read(fd, &strs[st_buf.st_size-sz], sz);
  } while ( sz );
  close(fd);

  if( !ParseConfigStream(str_size, strs) ){
    delete [] strs;
    return false;
  }

  delete [] strs;
  return true;
}

/**
 * コンフィグレーションXMLファイルからパースを行う。(並列実行)
 *
 * マスタノードがファイルをオープンし、配信する。
 * @param fname   XMLファイル名
 * @return      true=success, false=fault
 */
// modify for vsphere_cmp at 2008/08/15 by @hira
//     from ParaMngr to ParaCmpo
bool SklSolverConfig::SklParseConfigFileForParallel(const char* fname)
{
  int flg, ret;
  flg = 0;

  int myID = 0;
  if (SPL_Comm_rank(SPL_COMM_WORLD, &myID) != SPL_MPI_SUCCESS) {
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Comm_rank error.");
    return false;
  }
  if( myID == 0 ){
    if( !fname ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0017_);
      flg = 1;
    }
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) return false;

  int fd = -1;
  struct stat st_buf;
  if( myID == 0 ){
    if( (fd = open(fname, O_RDONLY)) < 0 ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0018_, fname);
      flg = 1;
    }
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) return false;

  if( myID == 0 ){
    if( fstat(fd, &st_buf) < 0 ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0019_);
      flg = 1;
    }
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) { if( myID == 0 ) close(fd); return false; }

  if( myID == 0 ){
    if( st_buf.st_size < 1 ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0020_);
      flg = 1;
    }
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) { if( myID == 0 ) close(fd); return false; }

  off_t len = 0;
  if( myID == 0 ){
    len = st_buf.st_size;
  }
  SPL_Datatype dt_type = SPL_DATATYPE_NULL;
  if( sizeof(off_t) == sizeof(unsigned) ) {
    dt_type = SPL_UNSIGNED;
  }
  else if( sizeof(off_t) == sizeof(long long) ) {
    dt_type = SPL_LONG_LONG;
  }
  if( SPL_Bcast(&len, 1, dt_type, 0, SPL_COMM_WORLD) != SPL_MPI_SUCCESS ){
    if( myID == 0 ) close(fd); return false;
  }

  char* strs = new char[len + 1];
  if( !strs ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0005_);
    flg = 1;
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) { if( myID == 0 ) close(fd); return false; }

  if( myID == 0 ){
    size_t str_size = st_buf.st_size;
    size_t sz = str_size;
    do {
      sz -= read(fd, &strs[st_buf.st_size-sz], sz);
    } while ( sz );
    close(fd);
  }
  if( SPL_Bcast(strs, len+1, SPL_CHAR, 0, SPL_COMM_WORLD) != SPL_MPI_SUCCESS ){
    delete [] strs;
    return false;
  }

  if( !ParseConfigStream(len, strs) ){
    delete [] strs;
    flg = 1;
  }
  SPL_Allreduce(&flg, &ret,1,SPL_INT,SPL_SUM,SPL_COMM_WORLD);
  if( ret != 0 ) return false;

  delete [] strs;
  return true;
}

/**
 * XML文字列からパースを行う。
 * @param sz      XML文字列サイズ
 * @param stream    XML文字列
 * @return      true=success, false=fault
 */
bool SklSolverConfig::ParseConfigStream(const size_t sz, const char* stream)
{
  if( (sz <= 0) || !stream ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0021_);
    return false;
  }
  xmlDocPtr doc = ::xmlParseMemory(stream, sz);
  if ( !doc ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0022_);
    return false;
  }

  xmlNodePtr cur = xmlDocGetRootElement(doc);
  while ( cur ) {
    //---- skip Comments
    if ( cur->type == XML_COMMENT_NODE ) {
      cur = cur->next;
      continue;
    }
    break;
  }
  if( ! cur ){
    xmlFreeDoc(doc);
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0023_);
    return false;
  }

  //---- parse tree
  if ( !SklParse(cur) ){
    xmlFreeDoc(doc);
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  xmlFreeDoc(doc);
  return true;
}

/**
 * XML要素からパースを行う。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklSolverConfig::SklParse(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( !Parse(node) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  return true;
}

/**
 * XML要素からパースを行う。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklSolverConfig::Parse(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_SPHERECONFIG_,
                     strlen(_SKL_CFG_NNAME_SPHERECONFIG_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlChar* tmpStr;
  tmpStr = xmlGetProp(node, (const xmlChar*)_SKL_CFG_ANAME_SOLVERTYPE_);
  if( !tmpStr ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0004_,
              _SKL_CFG_NNAME_SPHERECONFIG_,
              _SKL_CFG_ANAME_SOLVERTYPE_);
    return false;
  }

  // SolverTyepを設定する
  int len = strlen((const char*)tmpStr);
  m_solverType = new char[len + 1];
  memset(m_solverType, 0x00, sizeof(char)*(len + 1));
  strncpy(m_solverType, (const char*)tmpStr, len + 1);

  xmlFree(tmpStr);

  xmlNodePtr cur = node->xmlChildrenNode;
  while ( cur ) {
    if ( cur->type == XML_COMMENT_NODE ) {
      cur = cur->next;
      continue;
    }

    if( !StrCaseCompare((const char*)cur->name,
                        _SKL_CFG_NNAME_DOMAININFO_,
                        strlen(_SKL_CFG_NNAME_DOMAININFO_)) ){
      if( m_domainInfo ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0024_);
        return false;
      }
      if( !(m_domainInfo = new SklCfgDomainInfo) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !m_domainInfo->ParseNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }
    else if( !StrCaseCompare((const char*)cur->name,
                             _SKL_CFG_NNAME_STEER_,
                             strlen(_SKL_CFG_NNAME_STEER_)) ){
      if( m_steer ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0025_);
        return false;
      }
      if( !(m_steer = new SklCfgSteer) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !m_steer->ParseNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }
    else if( !StrCaseCompare((const char*)cur->name,
                             _SKL_CFG_NNAME_PARAMETER_,
                             strlen(_SKL_CFG_NNAME_PARAMETER_)) ){
      if( m_parameter ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0026_);
        return false;
      }
      if( !(m_parameter = new SklCfgParameter) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !m_parameter->ParseNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }
    else if( !StrCaseCompare((const char*)cur->name,
                             _SKL_CFG_NNAME_BC_TABLE_,
                             strlen(_SKL_CFG_NNAME_BC_TABLE_)) ){
      if( !ParseBCTable(cur) ) {
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }
    else if( !StrCaseCompare((const char*)cur->name,
                             _SKL_CFG_NNAME_MDM_TABLE_,
                             strlen(_SKL_CFG_NNAME_MDM_TABLE_)) ){
      if( m_mdmTable ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0028_);
        return false;
      }
      if( !(m_mdmTable = new SklCfgMdmTable) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !m_mdmTable->ParseNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }
    else if( !StrCaseCompare((const char*)cur->name,
                             _SKL_CFG_NNAME_MBX_TABLE_,
                             strlen(_SKL_CFG_NNAME_MBX_TABLE_)) ){
      if( m_mbxTable ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0028_);
        return false;
      }
      if( !(m_mbxTable = new SklCfgMBoxTable) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !m_mbxTable->ParseNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }
    else if( !StrCaseCompare((const char*)cur->name,
                             _SKL_CFG_NNAME_USRDEF_ELEM_,
                             strlen(_SKL_CFG_NNAME_USRDEF_ELEM_)) ){
      if( !m_usrDefNode ){
        if( !(m_usrDefNode = new SklCfgUsrDefTopNode) ){
          CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
          CfgErrMsg(_SKL_CFG_ERR_0005_);
          return false;
        }
      }
      if( !m_usrDefNode->ParseNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }

    cur = cur->next;
  }

  return true;
}

/**
 * 要素タイプのコンフィグレーション要素クラスを取得する。
 * @param node      要素タイプ
 * @return      コンフィグレーション要素クラス
 */
SklCfgNodeBase* SklSolverConfig::GetNode(SKLNODE node) const
{
  switch(node){
    case NODE_UNKNOWN:
      return NULL;
    case DOMAININFO:
      return m_domainInfo;
    case STEER:
      return m_steer;
    case PARAMETER:
      return m_parameter;
    case INNERBND:
      return m_innerBnd;
    case OUTERBND:
      return m_outerBnd;
    case MDMTBL:
      return m_mdmTable;
    case MBOXTBL:
      return m_mbxTable;
    case USERDEF:
      return m_usrDefNode;
    default:
      return NULL;
  }
}

/**
 * XML要素からBC_Table要素をパースする。
 * @param node    XML要素
 * @return      true=success, false=fault
 */
bool SklSolverConfig::ParseBCTable(xmlNodePtr node)
{
  if( !node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0003_);
    return false;
  }

  if( StrCaseCompare((const char*)node->name,
                     _SKL_CFG_NNAME_BC_TABLE_,
                     strlen(_SKL_CFG_NNAME_BC_TABLE_)) ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0002_, (const char*)node->name);
    return false;
  }

  xmlNodePtr cur = node->xmlChildrenNode;
  while ( cur ) {
    if ( cur->type == XML_COMMENT_NODE ) {
      cur = cur->next;
      continue;
    }

    if( !StrCaseCompare((const char*)cur->name,
                        _SKL_CFG_NNAME_INNER_BND_,
                        strlen(_SKL_CFG_NNAME_INNER_BND_)) ){
      if( m_innerBnd ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0027_);
        return false;
      }
      if( !(m_innerBnd = new SklCfgInnerBnd) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !m_innerBnd->ParseNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }
    else if( !StrCaseCompare((const char*)cur->name,
                             _SKL_CFG_NNAME_OUTER_BND_,
                             strlen(_SKL_CFG_NNAME_OUTER_BND_)) ){
      if( m_outerBnd ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0028_);
        return false;
      }
      if( !(m_outerBnd = new SklCfgOuterBnd) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        CfgErrMsg(_SKL_CFG_ERR_0005_);
        return false;
      }
      if( !m_outerBnd->ParseNode(cur) ){
        CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
        return false;
      }
    }

    cur = cur->next;
  }

  return true;
}

/**
 * InnerBoundary要素をパースする。
 * @return      true=success, false=fault
 */
bool SklSolverConfig::ReadBCFileForInner()
{
  const CfgElem* top = GetTop(DOMAININFO);
  if( !top ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  const CfgParam* param = top->GetParamFirst(_SKL_CFG_ANAME_DOMAININFO_BCTBL_);
  if( !param ) return true;

  SklIBCConfig bcCfg;
  SklCfgInnerBnd* bnd = NULL;
  SklCfgInnerBnd* bnd_bkup = NULL;
  while( param ){
    const char* fname = NULL;
    if( !param->GetData(&fname) || !fname ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0035_);
      return false;
    }

    if( !(bnd = new SklCfgInnerBnd) ) {
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0005_);
      return false;
    }
    bcCfg.SetBnd(bnd);

    if( !bcCfg.SklParseConfig(fname) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }

    if( bnd ) {
      if( bnd->GetSize() > 0 ) {
        if( bnd_bkup ) delete bnd_bkup;
        bnd_bkup = bnd;
      }
      else delete bnd;
    }

    param = top->GetParamNext(param, _SKL_CFG_ANAME_DOMAININFO_BCTBL_);
  }

  if( !m_innerBnd ) {
    m_innerBnd = bnd_bkup;
  }
  else {
    if( bnd_bkup ) delete bnd_bkup;
  }

  return true;
}

/**
 * OuterBoundary要素をパースする。
 * @return      true=success, false=fault
 */
bool SklSolverConfig::ReadBCFileForOuter()
{
  const CfgElem* top = GetTop(DOMAININFO);
  if( !top ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  const CfgParam* param = top->GetParamFirst(_SKL_CFG_ANAME_DOMAININFO_BCTBL_);
  if( !param ) return true;

  SklOBCConfig bcCfg;
  SklCfgOuterBnd* bnd = NULL;
  SklCfgOuterBnd* bnd_bkup = NULL;
  while( param ){
    const char* fname = NULL;
    if( !param->GetData(&fname) || !fname ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0035_);
      return false;
    }

    if( !(bnd = new SklCfgOuterBnd) ) {
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0005_);
      return false;
    }
    bcCfg.SetBnd(bnd);

    if( !bcCfg.SklParseConfig(fname) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }

    if( bnd ) {
      if( bnd->GetSize() > 0 ) {
        if( bnd_bkup ) delete bnd_bkup;
        bnd_bkup = bnd;
      }
      else delete bnd;
    }

    param = top->GetParamNext(param, _SKL_CFG_ANAME_DOMAININFO_BCTBL_);
  }

  if( !m_outerBnd ) {
    m_outerBnd = bnd_bkup;
  }
  else {
    if( bnd_bkup ) delete bnd_bkup;
  }

  return true;
}

/**
 * Medium_Table要素をパースする。
 * @return      true=success, false=fault
 */
bool SklSolverConfig::ReadMDMTableFile()
{
  if( m_mdmTable ) return true;

  const CfgElem* top = GetTop(DOMAININFO);
  if( !top ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  const CfgParam* param = top->GetParamFirst(_SKL_CFG_ANAME_DOMAININFO_MDMTBL_);
  if( !param ) return true;

  SklMDMConfig mdmCfg;
  SklCfgMdmTable* mdmTbl = NULL;
  while( param ){
    const char* fname = NULL;
    if( !param->GetData(&fname) || !fname ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0035_);
      return false;
    }

    if( mdmTbl ) delete mdmTbl;
    if( !(mdmTbl = new SklCfgMdmTable) ) {
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0035_);
      return false;
    }
    mdmCfg.SetMdmTable(mdmTbl);

    if( !mdmCfg.SklParseConfig(fname) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
    param = top->GetParamNext(param, _SKL_CFG_ANAME_DOMAININFO_MDMTBL_);
  }

  if( !m_mdmTable ) {
    m_mdmTable = mdmTbl;
  }
  else {
    if( mdmTbl ) delete mdmTbl;
  }

  return true;
}

/**
 * MultiBoxDivTable要素をパースする。
 * @return      true=success, false=fault
 */
bool SklSolverConfig::ReadMBoxTableFile()
{
  if( m_mbxTable ) return true;

  const CfgElem* top = GetTop(DOMAININFO);
  if( !top ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  const CfgParam* param = top->GetParamFirst(_SKL_CFG_ANAME_DOMAININFO_MBXTBL_);
  if( !param ) return true;

  SklMBoxConfig mbxCfg;
  SklCfgMBoxTable* mbxTbl = NULL;
  while( param ){
    const char* fname = NULL;
    if( !param->GetData(&fname) || !fname ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0035_);
      return false;
    }

    if( mbxTbl ) delete mbxTbl;
    if( !(mbxTbl = new SklCfgMBoxTable) ) {
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      CfgErrMsg(_SKL_CFG_ERR_0035_);
      return false;
    }
    mbxCfg.SetMBoxTable(mbxTbl);

    if( !mbxCfg.SklParseConfig(fname) ){
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
    param = top->GetParamNext(param, _SKL_CFG_ANAME_DOMAININFO_MBXTBL_);
  }

  if( !m_mbxTable ) {
    m_mbxTable = mbxTbl;
  }
  else {
    if( mbxTbl ) delete mbxTbl;
  }

  return true;
}

/**
 * コンフィグレーションのトップ要素の取得を行う。
 * @param node    要素種別
 * @param name        name属性値
 * @return      トップ要素
 */
const CfgElem* SklSolverConfig::GetTop(SKLNODE node, const char* name) const
{
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return NULL;
  if( (node == USERDEF) && name ) return n->GetElemFirst(name);
  return n->GetRootElem();
}

/**
 * コンフィグレーションの要素数の取得を行う。
 * @param node    要素種別
 * @param name        name属性値
 * @return            要素数
 */
unsigned int SklSolverConfig::GetSize(SKLNODE node, const char* name) const
{
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return 0;
  if( (node == USERDEF) && name ){
    const CfgElem* elem = n->GetElemFirst(name);
    if( !elem ) return 0;
    return elem->GetSize();
  }
  return n->GetSize();
}

/**
 * コンフィグレーションのElem要素数の取得を行う。
 * @param node    要素種別
 * @param name        name属性値
 * @return            要素数
 */
unsigned int SklSolverConfig::GetElemSize(SKLNODE node, const char* name) const
{
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return 0;
  if( (node == USERDEF) && name ){
    const CfgElem* elem = n->GetElemFirst(name);
    if( !elem ) return 0;
    return elem->GetElemSize();
  }
  return n->GetElemSize();
}

/**
 * コンフィグレーションのParam要素数の取得を行う。
 * @param node    要素種別
 * @param name        name属性値
 * @return            要素数
 */
unsigned int SklSolverConfig::GetParamSize(SKLNODE node, const char* name) const
{
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return 0;
  if( (node == USERDEF) && name ){
    const CfgElem* elem = n->GetElemFirst(name);
    if( !elem ) return 0;
    return elem->GetParamSize();
  }
  return n->GetParamSize();
}

/**
 * コンフィグレーションの最初のElem要素の取得を行う。
 * @param node    要素種別
 * @param name        name属性値
 * @return            最初のElem要素
 */
const CfgElem* SklSolverConfig::GetElemFirst(SKLNODE node,
                                             const char* name) const
{
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return NULL;
  return n->GetElemFirst(name);
}

/**
 * コンフィグレーションの最初のElem要素の取得を行う。
 * @param node    要素種別
 * @param name        name属性値
 * @param id          id属性値
 * @param cmt         comment属性値
 * @return            最初のElem要素
 */
const CfgElem* SklSolverConfig::GetElemFirst(SKLNODE node, const char* name,
                                             int id, const char* cmt) const
{
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return NULL;
  return n->GetElemFirst(name, id, cmt);
}

/**
 * コンフィグレーションの次のElem要素の取得を行う。
 * @param node    要素種別
 * @param elem        前のElem要素
 * @param name        name属性値
 * @return            次のElem要素
 */
const CfgElem* SklSolverConfig::GetElemNext(SKLNODE node,
                                            const CfgElem* elem,
                                            const char* name) const
{
  if( !elem ) return NULL;
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return NULL;
  return n->GetElemNext(elem, name);
}

/**
 * コンフィグレーションの次のElem要素の取得を行う。
 * @param node    要素種別
 * @param elem        前のElem要素
 * @param name        name属性値
 * @param id          id属性値
 * @param cmt         comment属性値
 * @return            次のElem要素
 */
const CfgElem* SklSolverConfig::GetElemNext(SKLNODE node,
                                            const CfgElem* elem,
                                            const char* name,
                                            int id,
                                            const char* cmt) const
{
  if( !elem ) return NULL;
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return NULL;
  return n->GetElemNext(elem, name, id, cmt);
}

/**
 * コンフィグレーションの最初のParam要素の取得を行う。
 * @param node    要素種別
 * @param name        name属性値
 * @return            最初のParam要素
 */
const CfgParam* SklSolverConfig::GetParamFirst(SKLNODE node,
                                               const char* name) const
{
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return NULL;
  return n->GetParamFirst(name);
}

/**
 * コンフィグレーションの最初のParam要素の取得を行う。
 * @param node    要素種別
 * @param name        name属性値
 * @param id          id属性値
 * @param cmt         comment属性値
 * @return            最初のParam要素
 */
const CfgParam* SklSolverConfig::GetParamFirst(SKLNODE node, const char* name,
                                               int id, const char* cmt) const
{
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return NULL;
  return n->GetParamFirst(name, id, cmt);
}

/**
 * コンフィグレーションの次のParam要素の取得を行う。
 * @param node    要素種別
 * @param param        前のParam要素
 * @param name        name属性値
 * @return            次のParam要素
 */
const CfgParam* SklSolverConfig::GetParamNext(SKLNODE node,
                                              const CfgParam* param,
                                              const char* name) const
{
  if( !param ) return NULL;
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return NULL;
  return n->GetParamNext(param, name);
}

/**
 * コンフィグレーションの次のParam要素の取得を行う。
 * @param node    要素種別
 * @param param        前のParam要素
 * @param name        name属性値
 * @param id          id属性値
 * @param cmt         comment属性値
 * @return            次のParam要素
 */
const CfgParam* SklSolverConfig::GetParamNext(SKLNODE node,
                                              const CfgParam* param,
                                              const char* name,
                                              int id,
                                              const char* cmt) const
{
  if( !param ) return NULL;
  SklCfgNodeBase* n = GetNode(node);
  if( !n ) return NULL;
  return n->GetParamNext(param, name, id, cmt);
}

/**
 * コンフィグレーションから原点座標を取得する。(float)
 * @param[out] x       X方向原点座標
 * @param[out] y       X方向原点座標
 * @param[out] z       X方向原点座標
 * @return      true=success, false=fault
 */
bool SklSolverConfig::GetVoxelOrigin(float& x, float& y, float& z) const
{
  if( !m_domainInfo ) return false;
  return m_domainInfo->GetVoxelOrigin(x, y, z);
}

/**
 * コンフィグレーションからボクセルサイズを取得する。
 * @param[out] i     I方向ボクセルサイズ
 * @param[out] j     J方向ボクセルサイズ
 * @param[out] k     K方向ボクセルサイズ
 * @return      true=success, false=fault
 */
bool SklSolverConfig::GetVoxelSize(unsigned int& i,
                                   unsigned int& j,
                                   unsigned int& k) const
{
  if( !m_domainInfo ) return false;
  return m_domainInfo->GetVoxelSize(i, j, k);
}

/**
 * コンフィグレーションからボクセル幅を取得する。
 * @param[out] x       X方向幅
 * @param[out] y       Y方向幅
 * @param[out] z       Z方向幅
 * @return      true=success, false=fault
 */
bool SklSolverConfig::GetVoxelWidth(float& x, float& y, float& z) const
{
  if( !m_domainInfo ) return false;
  return m_domainInfo->GetVoxelWidth(x, y, z);
}

/**
 * コンフィグレーションからピッチを取得する。
 * @param[out] x       X方向ピッチ
 * @param[out] y       Y方向ピッチ
 * @param[out] z       Z方向ピッチ
 * @return      true=success, false=fault
 */
bool SklSolverConfig::GetVoxelPitch(float& x, float& y, float& z) const
{
  if( !m_domainInfo ) return false;
  return m_domainInfo->GetVoxelPitch(x, y, z);
}

/**
 * コンフィグレーションから計算ステップ数を取得する。
 *
 * ”CalculationSteps”要素（計算ステップ数の記述）
 * @return      計算ステップ数
 */
unsigned int SklSolverConfig::GetCalcSteps() const
{
  if( !m_steer ) return 0;
  return m_steer->GetCalcSteps();
}

/**
 * コンフィグレーションからソルバーの開始条件を取得する。
 *
 * ”StartCondition”要素（ソルバーの開始条件の記述）
 * @return      ソルバーの開始条件
 */
StartType SklSolverConfig::GetStartType() const
{
  if( !m_steer ) return Initial;
  return m_steer->GetStartType();
}

/**
 * コンフィグレーションからソルバーの起動条件を取得する。
 *
 * ”CheckMode”要素（ソルバーの起動条件の指定）
 * @return      ソルバーの起動条件
 */
bool SklSolverConfig::IsCheckMode() const
{
  if( !m_steer ) return false;
  return m_steer->IsCheckMode();
}

/**
 * コンフィグレーションからソルバ起動条件ファイルを取得する。
 * @return    ソルバ起動条件ファイル
 */
const char* SklSolverConfig::GetCheckModeFile() const
{
  if( !m_steer ) return NULL;
  return m_steer->GetCheckModeFile();
}

/**
 * コンフィグレーションから入力ファイル数を取得する。
 * @return      入力ファイル数
 */
unsigned int SklSolverConfig::GetInFileSize() const
{
  if( !m_steer ) return 0;
  return m_steer->GetInFileSize();
}

/**
 * コンフィグレーションから入力ファイルの最初の要素を取得する。
 * @return      最初の入力ファイル要素
 */
const SklCfgInFile* SklSolverConfig::GetInFileFirst() const
{
  if( !m_steer ) return NULL;
  return m_steer->GetInFileFirst();
}

/**
 * コンフィグレーションから入力ファイルの次のの要素を取得する。
 * @param file    前の入力ファイル要素
 * @return        次の入力ファイル要素
 */
const SklCfgInFile*
      SklSolverConfig::GetInFileNext(const SklCfgInFile* file) const
{
  if( !file ) return NULL;
  if( !m_steer ) return NULL;
  return m_steer->GetInFileNext(file);
}

/**
 * コンフィグレーションから出力インターバルを取得する。
 * @return      出力インターバル
 */
unsigned int SklSolverConfig::GetMonitorInterval() const
{
  if( !m_steer ) return 0;
  return m_steer->GetMonitorInterval();
}

/**
 * コンフィグレーションから出力ファイル数を取得する。
 * @return      出力ファイル数
 */
unsigned int SklSolverConfig::GetOutFileSize() const
{
  if( !m_steer ) return 0;
  return m_steer->GetOutFileSize();
}

/**
 * コンフィグレーションから出力ファイルの最初の要素を取得する。
 * @return      最初の出力ファイル要素
 */
const SklCfgOutFile* SklSolverConfig::GetOutFileFirst() const
{
  if( !m_steer ) return NULL;
  return m_steer->GetOutFileFirst();
}

/**
 * コンフィグレーションから出力ファイルの次のの要素を取得する。
 * @param file    前の出力ファイル要素
 * @return        次の出力ファイル要素
 */
const SklCfgOutFile*
      SklSolverConfig::GetOutFileNext(const SklCfgOutFile* file) const
{
  if( !file ) return NULL;
  if( !m_steer ) return NULL;
  return m_steer->GetOutFileNext(file);
}

/**
 * コンフィグレーションからヒストリファイル数を取得する。
 * @return        ヒストリファイル数
 */
unsigned int SklSolverConfig::GetHistorySize() const
{
  if( !m_steer ) return 0;
  return m_steer->GetHistorySize();
}

/**
 * コンフィグレーションからヒストリファイルの最初の要素を取得する。
 * @return      最初のヒストリファイル要素
 */
const SklCfgHistory* SklSolverConfig::GetHistoryFirst() const
{
  if( !m_steer ) return NULL;
  return m_steer->GetHistoryFirst();
}

/**
 * コンフィグレーションからヒストリファイルの次のの要素を取得する。
 * @param history    前のヒストリファイル要素
 * @return        次のヒストリファイル要素
 */
const SklCfgHistory*
      SklSolverConfig::GetHistoryNext(const SklCfgHistory* history) const
{
  if( !history ) return NULL;
  if( !m_steer ) return NULL;
  return m_steer->GetHistoryNext(history);
}

/**
 * MPI通信性能の最適化要素の取得を行う。
 * @param vc        ガイドセル
 * @param loop_max  最大通信回数
 * @return      true=success, false=fault
 */
bool
SklSolverConfig::IsChkCommPerform(unsigned int& vc, size_t& loop_max) const
{
  const CfgElem* top = GetTop(STEER);
  if( !top ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  const CfgElem* elem = top->GetElemFirst("chk_comm_perform");
  if( !elem ) return false;

  int val;
  const CfgParam* param = elem->GetParamFirst("vc");
  if( param ) {
    if( !param->GetData(&val) ) {
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
    vc = (unsigned int)val;
  }

  param = elem->GetParamFirst("loop_max");
  if( param ) {
    if( !param->GetData(&val) ) {
      CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
      return false;
    }
    loop_max = (size_t)val;
  }

  return true;
}

/**
 * MPI通信テスト終了フラグを取得する。
 * @return    true=MPI通信テスト後終了
 */
bool
SklSolverConfig::IsExitAfterChkCommPerform() const {
  const CfgElem* top = GetTop(STEER);
  if( !top ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return true; // exit
  }
  const CfgElem* elem1 = top->GetElemFirst("chk_comm_perform");
  if( !elem1 ) return false;
  const CfgElem* elem2 = elem1->GetElemFirst("exit");
  if( !elem2 ) return false;
  return true;
}

#ifdef DEBUG
void SklSolverConfig::Print()
{
  if( m_domainInfo ) m_domainInfo->Print();
  else cout << " <-- DomainInfo not set -->" << endl << flush;

  if( m_steer ) m_steer->Print();
  else cout << " <-- Steer not set -->" << endl << flush;

  if( m_parameter ) m_parameter->Print();
  else cout << " <-- Parameter not set -->" << endl << flush;

  if( m_innerBnd ) m_innerBnd->Print();
  else cout << " <-- InnerBoundary not set -->" << endl << flush;

  if( m_outerBnd ) m_outerBnd->Print();
  else cout << " <-- OuterBoundary not set -->" << endl << flush;
}

#endif //DEBUG

/**
 * SolverTypeを取得する。
 * @return    SolverType
 */
const char* SklSolverConfig::getSolverType() const
{
  return m_solverType;
}

/**
 * 入力フォルダを取得する。
 * @return     入力フォルダ
 */
const char* SklSolverConfig::getInputBaseDir() const
{
  if (m_steer == NULL) return NULL;
  return m_steer->getInputBaseDir();
}

/**
 * 出力フォルダを取得する。
 * @return     出力フォルダ
 */
const char* SklSolverConfig::getOutputBaseDir() const
{
  if (m_steer == NULL) return NULL;
  return m_steer->getOutputBaseDir();
}


/**
 * コンフィグレーションXMLルート要素からパースを行う。
 * @param root_node     XMLルート要素
 * @return      true=success, false=fault
 */
bool SklSolverConfig::SklParseConfig(xmlNodePtr root_node)
{
  if( !root_node ){
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    CfgErrMsg(_SKL_CFG_ERR_0017_);
    return false;
  }

  //---- parse tree
  if ( !SklParse(root_node) ) {
    CfgDbgMsg("%s (%d): ",__FILE__, __LINE__);
    return false;
  }

  // read Boundary-Condition table file
  if( !m_innerBnd ) {
    if( !ReadBCFileForInner() ) return false;
  }
  if( !m_outerBnd ) {
    if( !ReadBCFileForOuter() ) return false;
  }

  // read Medium table file
  if( !m_mdmTable ) {
    if( !ReadMDMTableFile() ) return false;
  }

  // read MultiBox table file
  if( !m_mbxTable ) {
    if( !ReadMBoxTableFile() ) return false;
  }

  return true;
}

} // end of namespace SklCfg



