/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphCouplerBase.C
 * Sphere Coupler super Class Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphCouplerBase.C
 * Sphereカプラー基底クラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */
#include "solverbase/sphCouplerBase.h"
#include "data/sphDataManager.h"
#include "solverbase/sphSolverBase.h"
#include "config/sphElemUtils.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphCouplerBase::sphCouplerBase()
{
  /// 初期化を行う。
  initialize();

}


/**
 * \~english
 * Constructor
 * @param     label     coupler label
 *
 * \~japanese
 * コンストラクタ
 * @param     label     カプラーラベル
 */
sphCouplerBase::sphCouplerBase(const char* label) : sphSphereObject(label, 0)
{
  /// 初期化を行う。
  initialize();
}


/**
 * \~english
 * Constructor
 * @param     label     coupler label
 * @param     id     coupler id
 *
 * \~japanese
 * コンストラクタ
 * @param     label     カプラーラベル
 * @param     id     カプラーID
 */
sphCouplerBase::sphCouplerBase(const char* label, int id) : sphSphereObject(label, id)
{
  /// 初期化を行う。
  initialize();
}

/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphCouplerBase::~sphCouplerBase()
{
  if (m_couplerComm != SPL_COMM_WORLD) {
    if (m_couplerComm != SPL_COMM_NULL) SPL_Comm_free(&m_couplerComm);
    if (m_couplerGroup != SPL_GROUP_NULL) SPL_Group_free(&m_couplerGroup);
  }

  std::map<int, NodeBoundaryInfo*>::iterator it_send = m_sendNodeTable.begin();
  for( ; it_send!=m_sendNodeTable.end(); it_send++) {
    NodeBoundaryInfo* info = it_send->second;
    delete info;
    info = NULL;
  }
  m_sendNodeTable.clear();

  std::map<int, NodeBoundaryInfo*>::iterator it_recv = m_recvNodeTable.begin();
  for( ; it_recv!=m_recvNodeTable.end(); it_recv++) {
    NodeBoundaryInfo* info = it_recv->second;
    if (info->m_recvBuf != NULL) delete info->m_recvBuf;
    info->m_recvBuf = NULL;
    delete info;
    info = NULL;
  }
  m_recvNodeTable.clear();

}




/**
 * 初期化を行う。
 * @param     void
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerBase::initialize()
{

  m_couplerComm = SPL_COMM_NULL;
  m_couplerGroup = SPL_GROUP_NULL;
  m_refSolverID = 0;
  m_udtSolverID = 0;

  return true;
}


/**
 * カプラ全体のコミュニケータ、グループを設定する.
 * @param               comm    コミュニケータ
 * @param               group   グループ
 * @return              void
 */
void sphCouplerBase::setCommGroup(SPL_Comm comm, SPL_Group group)
{
  m_couplerComm = comm;
  m_couplerGroup = group;
  return;
}


/**
 * カプラ全体のコミュニケータを取得する.
 * @param               void
 * @return              コミュニケータ(SPL_Comm)
 */
SPL_Comm sphCouplerBase::getSplComm() const
{
  return m_couplerComm;
}


/**
 * カプラ全体のグループを取得する.
 * @param               void
 * @return              グループ(SPL_Group)
 */
SPL_Group sphCouplerBase::getSplGroup() const
{
  return m_couplerGroup;
}


/**
 * 参照ソルバIDを取得する.
 * @param               void
 * @return              参照ソルバID
 */
int sphCouplerBase::getReferSolverId() const
{
  return m_refSolverID;
}

/**
 * 更新ソルバIDを取得する.
 * @param               void
 * @return              更新ソルバID
 */
int sphCouplerBase::getUpdateSolverId() const
{
  return m_udtSolverID;
}


/**
 * 参照データオブジェクトをデータマネージャから取得する。
 * @param       void
 * @return      参照データオブジェクト
 */
sphData* sphCouplerBase::getReferDataInDataManager(const char* ref_reg_label)
{
  sphData* dataObj = NULL;
  if (ref_reg_label == NULL) {
    dataObj = getDataObjectInDataManager(SPH_CFG_NAME_REFOBJ);
  }
  else {
    /// インスタンスされているソルバオブジェクトの取得を行う。
    sphSolverManager* solvMngr = getSolverManager();
    const sphSolverBase* solverObj = solvMngr->getSolverById(m_refSolverID);
    if (solverObj == NULL) {
      /// 自身にはソルバはインスタンスされていない。
      return NULL;
    }
    /// データマネージャからデータオブジェクトを取得する。
    const sphDataManager* dataMngr = solverObj->getDataManager();
    dataObj = dataMngr->getDataObj(ref_reg_label);
  }

  return dataObj;
}

/**
 * 更新データオブジェクトをデータマネージャから取得する。
 * @param       void
 * @return      更新データオブジェクト
 */
sphData* sphCouplerBase::getUpdateDataInDataManager(const char* upd_reg_label)
{
  sphData* dataObj = NULL;
  if (upd_reg_label == NULL) {
    dataObj = getDataObjectInDataManager(SPH_CFG_NAME_UDTOBJ);
  }
  else {
    /// インスタンスされているソルバオブジェクトの取得を行う。
    sphSolverManager* solvMngr = getSolverManager();
    sphSolverBase* solverObj = solvMngr->getSolverById(m_udtSolverID);
    if (solverObj == NULL) {
      /// 自身にはソルバはインスタンスされていない。
      return NULL;
    }
    /// データマネージャからデータオブジェクトを取得する。
    sphDataManager* dataMngr = solverObj->getDataManager();
    dataObj = dataMngr->getDataObj(upd_reg_label);
  }

  return dataObj;
}


/**
 * 参照/更新データオブジェクトを取得する。
 * @param       copy_name       取得要素名（SphRefObj又はSphUdtObj)
 * @return      参照/更新データオブジェクト
 */
sphData* sphCouplerBase::getDataObjectInDataManager(const char* elem_name )
{

  /// ソルバIDを取得する。
  const sphElem* obj_elem = m_configElem->getElem(2, SPH_CFG_NAME_DATACOPY, elem_name);
  if (obj_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found element[%s/%s/%s].", m_configElem->getElemName(), SPH_CFG_NAME_DATACOPY, elem_name);
    return NULL;
  }
  int solverId = obj_elem->getAttrInt(SPH_CFG_ATTR_SOLVERID);

  /// インスタンスされているソルバオブジェクトの取得を行う。
  sphSolverManager* solvMngr = getSolverManager();
  sphSolverBase* solverObj = solvMngr->getSolverById(solverId);
  if (solverObj == NULL) {
    /// 自身にはソルバはインスタンスされていない。
    return NULL;
  }

  /// データオブジェクトを取得する。
  const sphElem* data_elem = obj_elem->getElem(SPH_CFG_NAME_DATAOBJ);
  if (data_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found element[%s/%s].", obj_elem->getElemName(), SPH_CFG_NAME_DATAOBJ);
    return NULL;
  }
  /// データ識別文字列を取得する。
  const char* data_reg_label = NULL;
  if (data_elem->existAttribute(SPH_CFG_ATTR_ID)) {
    /// データ識別番号指定であるので、ソルバのデータリストからデータ識別文字列を検索する。
    int dataID = data_elem->getAttrInt(SPH_CFG_ATTR_ID);
    // ソルバ要素を取得する。
    const sphElem* solver_elem = sphElemUtils(m_configElem).getElemBySolverId(solverId);
    if (solver_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found element[SphSolver[@id=%d].", solverId);
      return NULL;
    }
    // データリストからデータ要素を取得する。
    const sphElem* dataobj_elem = solver_elem->getElemListById(SPH_CFG_NAME_DATALIST, SPH_CFG_NAME_DATAOBJ, dataID);
    if (dataobj_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found element[SphSolver[@id=%d]/%s/%s[@id=%d].", solverId, SPH_CFG_NAME_DATALIST, SPH_CFG_NAME_DATAOBJ, dataID);
      return NULL;
    }
    data_reg_label = dataobj_elem->getAttrString(SPH_CFG_ATTR_LABEL);
  }
  else {
    data_reg_label = data_elem->getAttrString(SPH_CFG_ATTR_LABEL);
  }
  if (data_reg_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "can not get data label");
    return NULL;
  }

  /// データマネージャからデータオブジェクトを取得する。
  const sphDataManager* dataMngr = solverObj->getDataManager();
  sphData* dataObj = dataMngr->getDataObj(data_reg_label);

  return dataObj;

}


/**
 * 参照データオブジェクトを取得する。
 * @param       void
 * @return      参照データオブジェクト
 */
sphData* sphCouplerBase::getReferData()
{
  return m_refDataObj;
}

/**
 * 更新データオブジェクトを取得する。
 * @param       void
 * @return      更新データオブジェクト
 */
sphData* sphCouplerBase::getUpdateData()
{
  return m_uptDataObj;
}


