/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphCouplerManager.C
 * Sphere Coupler Manager Class Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphCouplerManager.C
 * Sphereカプラーマネージャークラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */

#include "solverbase/sphCouplerManager.h"
#include "solverbase/sphCouplerBase.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphCouplerManager::sphCouplerManager()
{
  initialize();

}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphCouplerManager::~sphCouplerManager()
{
  finalize();
}


/**
 * 初期化処理
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerManager::initialize()
{
  m_mapCoupler.clear();
  return true;
}


/**
 * 破棄処理
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerManager::finalize()
{
  std::map<int, sphCouplerBase*>::const_iterator it = m_mapCoupler.begin();
  if(it == m_mapCoupler.end()) return true;
  for( ; it!=m_mapCoupler.end(); it++) {
    sphCouplerBase* coupler = it->second;
    delete coupler;
    coupler = NULL;
  }
  m_mapCoupler.clear();

  return true;
}


/**
 * カプラ数の取得を行う。
 * @param       void
 * @return      カプラ数
 */
int sphCouplerManager::getCouplerCount() const
{
  return m_mapCoupler.size();
}


/**
 * カプラの取得を行う。（インデックス）
 * @param       インデックス
 * @return      カプラオブジェクト
 */
sphCouplerBase* sphCouplerManager::getCoupler(int id) const
{
  if (id < 0 || id >= m_mapCoupler.size()) return NULL;

  std::map<int, sphCouplerBase*>::const_iterator it = m_mapCoupler.begin();
  register int i;
  for (i=0; i<id; i++, it++);
  return it->second;
}


/**
 * カプラの取得を行う。（カプラ識別ID）
 * @param       カプラ識別ID
 * @return      カプラオブジェクト
 */
sphCouplerBase* sphCouplerManager::getCouplerById(int couplerId) const
{
  std::map<int, sphCouplerBase*>::const_iterator it = m_mapCoupler.find(couplerId);
  if(it == m_mapCoupler.end()) return NULL;
  return it->second;
}


/**
 * カプラの取得を行う。（カプラ識別文字列）
 * @param       カプラ識別文字列
 * @return      カプラオブジェクト
 */
sphCouplerBase* sphCouplerManager::getCouplerByLabel(const char* label) const
{
  std::map<int, sphCouplerBase*>::const_iterator it = m_mapCoupler.begin();
  if(it == m_mapCoupler.end()) return NULL;

  for( ; it!=m_mapCoupler.end(); it++) {
    sphCouplerBase* coupler = it->second;
    std::string coupler_label = coupler->getSphereLabel();
    if (coupler_label == label) {
      return it->second;
    }
  }

  return NULL;
}


/**
 * カプラの追加を行う。
 * @param       カプラオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerManager::addCoupler(sphCouplerBase* coupler)
{
  if (!m_mapCoupler.insert( std::make_pair(coupler->getSphereId(), coupler)).second) {
     // error
     m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_mapCoupler.");
     delete coupler;
     coupler = NULL;
     return false;
  }

  return true;
}


