/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphCouplerSv2Sv.C
 * Sphere Simple Voxel to Simple Voxel Coupler super Class Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphCouplerSv2Sv.C
 * Sphere構造格子To構造格子カプラー基底クラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */

#include <float.h>
#include <math.h>
#include "solverbase/sphCouplerSv2Sv.h"
#include "solverbase/sphSolverBase.h"
#include "config/sphElemUtils.h"
#include "data/sphDataUtils.h"
#include "data/sphDataManager.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphCouplerSv2Sv::sphCouplerSv2Sv()
{
  // 初期化を行う。
  initialize();
}

/**
 * \~english
 * Constructor
 * @param     label     coupler label
 *
 * \~japanese
 * コンストラクタ
 * @param     label     カプラーラベル
 */
sphCouplerSv2Sv::sphCouplerSv2Sv(const char* label) : sphCouplerBase(label, 0)
{
  // 初期化を行う。
  initialize();

}


/**
 * \~english
 * Constructor
 * @param     label     coupler label
 * @param     id     coupler id
 *
 * \~japanese
 * コンストラクタ
 * @param     label     カプラーラベル
 * @param     id     カプラーID
 */
sphCouplerSv2Sv::sphCouplerSv2Sv(const char* label, int id) : sphCouplerBase(label, id)
{
  // 初期化を行う。
  initialize();

}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphCouplerSv2Sv::~sphCouplerSv2Sv()
{

}


/**
 * 初期化処理
 * @param        void
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerSv2Sv::initialize()
{

  m_refDataObj = NULL;
  m_uptDataObj = NULL;
  m_extGc = 0;

  return true;
}


/**
 * カプラの初期化を行う（データラベル指定）.
 * 通信ノード、ボクセル情報のテーブルを作成する。
 * @param               ref_reg_label         参照データ登録ラベル
 * @param               upd_reg_label         更新データ登録ラベル
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int
sphCouplerSv2Sv::initializeCoupler(const char* ref_reg_label, const char* upd_reg_label)
{
  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_configElem is null.");
    return SPHERE_TERMINATE;
  }

  /// 拡張ガイドセル値を取得する。
  const sphElem* copy_elem = NULL;
  copy_elem = m_configElem->getElem(SPH_CFG_NAME_DATACOPY);
  if (copy_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found element[%s/%s].", m_configElem->getElemName(), SPH_CFG_NAME_DATACOPY);
    return SPHERE_TERMINATE;
  }
  m_extGc = copy_elem->getAttrInt(SPH_CFG_ATTR_EXTGC);

  /// 参照ソルバIDと更新ソルバIDを取得する。
  const sphElem* elem = NULL;
  elem = m_configElem->getElem(2, SPH_CFG_NAME_DATACOPY, SPH_CFG_NAME_REFOBJ);
  if (elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found element[%s/%s/%s].", m_configElem->getElemName(), SPH_CFG_NAME_DATACOPY, SPH_CFG_NAME_REFOBJ);
    return SPHERE_TERMINATE;
  }
  m_refSolverID = elem->getAttrInt(SPH_CFG_ATTR_SOLVERID);

  elem = m_configElem->getElem(2, SPH_CFG_NAME_DATACOPY, SPH_CFG_NAME_UDTOBJ);
  if (elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found element[%s/%s/%s].", m_configElem->getElemName(), SPH_CFG_NAME_DATACOPY, SPH_CFG_NAME_UDTOBJ);
    return SPHERE_TERMINATE;
  }
  m_udtSolverID = elem->getAttrInt(SPH_CFG_ATTR_SOLVERID);

  /// 更新データオブジェクトと参照オブジェクトを取得する。
  m_refDataObj = getReferDataInDataManager(ref_reg_label);
  m_uptDataObj = getUpdateDataInDataManager(upd_reg_label);

  /// 参照・更新データから通信ノードテーブルを作成する。
  if (!createCommNodeTable(m_refDataObj, m_uptDataObj, m_extGc)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createCommNodeTable fault");
    return SPHERE_TERMINATE;
  }

  return SPHERE_SUCCESS;
}


/**
 * \~english
 * loop Coupler.
 * @param               void
 * @return              -1:error, 0:forced terminated, 1:normality
 *
 * \~japanese
 * カプラーのステップ実行を行う.
 * @param               void
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int
sphCouplerSv2Sv::executeCoupler()
{

  // 参照データのガイドセルの通信を行う。
  if (!commBndCell()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "commBndCell fault");
    return SPHERE_ERROR;
  }

  /// 参照データオブジェクトから受信バッファーにデータを受信する。
  if (!communicateData()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "receiveData fault");
    return SPHERE_ERROR;
  }

  /// 受信バッファーから更新データオブジェクトに補間を行い、データをセットする。
  if (!interpolate()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "interpolate fault");
    return SPHERE_ERROR;
  }


  return SPHERE_SUCCESS;
}


/**
 * \~english
 * post Coupler.
 * @param               void
 * @return              -1:error, 0:forced terminated, 1:normality
 *
 * \~japanese
 * カプラーの破棄処理を行う.
 * @param               void
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int
sphCouplerSv2Sv::postCoupler()
{
  return SPHERE_SUCCESS;
}


/**
 * 参照データオブジェクトの取得を行う.
 * 参照データのノードからデータを受信する。
 * @param               void
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerSv2Sv::communicateData()
{

  int myID = 0;
  int commSize = 0;
  SPL_Comm_rank(m_couplerComm, &myID);
  SPL_Comm_size(m_couplerComm, &commSize);

  register int i, n;
  bool ret = true;

  /// 非同期通信のリクエストリストの作成を行う。
  int comm_size = m_sendNodeTable.size()+m_recvNodeTable.size();
  SPL_Request* req = new SPL_Request[comm_size];
  SPL_Status* stat = new SPL_Status[comm_size];
  int req_id = 0;

  /// 送信テーブルから送信先ノードに参照データを非同期送信する。
  std::map<int, NodeBoundaryInfo*>::iterator it_send = m_sendNodeTable.begin();
  for( ; it_send!=m_sendNodeTable.end(); it_send++) {
    NodeBoundaryInfo* sendInfo = it_send->second;
    /// 送信先が自ノードの場合は、送信しない。
    if (sendInfo->m_nodeID != myID) {
      if (SPL_Isend(m_refDataObj->getDataPointer(),
                    1, sendInfo->m_comType,
                    sendInfo->m_nodeID,
                    0,
                    m_couplerComm,
                    &req[req_id]) != SPL_SUCCESS) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Isend error.");
        ret = false;
        break;
      }
    }
    else {
      req[req_id] = (SPL_Request)SPL_REQUEST_NULL;
    }
    req_id++;
  }

  if (ret) {
    /// 受信テーブルから受信元ノードから参照データを非同期受信する。
    std::map<int, NodeBoundaryInfo*>::iterator it_recv = m_recvNodeTable.begin();
    for( ; it_recv!=m_recvNodeTable.end(); it_recv++) {
      int id = it_recv->first;
      NodeBoundaryInfo* recvInfo = it_recv->second;
      /// 受信元ノードが自ノードの場合は、受信を行わない。
      if (recvInfo->m_nodeID != myID) {
        if(SPL_Irecv(recvInfo->m_recvBuf->getDataPointer(),
                     1, recvInfo->m_comType,
                     recvInfo->m_nodeID,
                     0,
                     m_couplerComm,
                     &req[req_id]) != SPL_SUCCESS) {
          // error
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Irecv error.");
          ret = false;
          break;
        }
      }
      else {
        req[req_id] = (SPL_Request)SPL_REQUEST_NULL;
      }
      req_id++;
    }
  }

  if (ret) {
    /// 非同期送受信の終了待ちを行う。
    if (SPL_Waitall(comm_size, req, stat) != SPL_MPI_SUCCESS) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "wailall error.");
      ret = false;
    }
  }

  if (req) delete[] req;
  if (stat) delete[] stat;

  if (!ret) {
    m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return ret;
  }

  /// 自身の参照データから受信バッファーにコピーを行う。
  it_send = m_sendNodeTable.begin();
  for( ; it_send!=m_sendNodeTable.end(); it_send++) {
    NodeBoundaryInfo* sendInfo = it_send->second;

    // debug
    std::map<int, NodeBoundaryInfo*>::iterator it_recv = m_recvNodeTable.begin();
    for( ; it_recv!=m_recvNodeTable.end(); it_recv++) {
      int id = it_recv->first;
      NodeBoundaryInfo* recvInfo = it_recv->second;
      if (recvInfo != NULL)
        int nodeid = recvInfo->m_nodeID;
    }
    // debug


    if (sendInfo->m_nodeID == myID) {
      std::map<int, NodeBoundaryInfo*>::iterator it_recv = m_recvNodeTable.find(myID);
      if (it_recv == m_recvNodeTable.end()) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INTERNAL_IMPOSSIBLE, "not exist myID[=%d] in m_recvNodeTable.", myID);
        ret = false;
        break;
      }
      NodeBoundaryInfo* recvInfo = it_recv->second;
      size_t cp_start_idx[3] = {0, 0, 0};
      if (!sphDataUtils::copyArray(
                  recvInfo->m_recvBuf,
                  m_refDataObj,
                  sendInfo->m_size,
                  cp_start_idx,
                  sendInfo->m_sta_idx) ) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "copyArray fault.");
        ret = false;
        break;
      }
    }
  }

  return ret;
}


/**
 * 受信バッファーから更新データに補間を行い、データをセットする
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerSv2Sv::interpolate()
{
  // 更新データオブジェクトを持っていないノードも存在する。
  if (m_uptDataObj == NULL) return true;

  register size_t i, j, k, ix, jx, kx, d, l;

  /// 更新データの次元、サイズ、原点座標、ピッチ、ベクトル数、ガイドセル数を取得する。
  int dims = m_uptDataObj->getDims();
  const size_t* udt_size = m_uptDataObj->_getArraySize();
  double udt_org[3] = {0.0, 0.0, 0.0};
  double udt_pitch[3] = {0.0, 0.0, 0.0};
  m_uptDataObj->getDataOrigin(udt_org);
  m_uptDataObj->getDataPitch(udt_pitch);
  size_t data_num = m_uptDataObj->getNumOfDtElem();
  size_t gc = m_uptDataObj->getGc();
  /// 更新データの定義位置
  SphCrdDef udt_crddef = m_uptDataObj->getCrdDef();

  /// 受信ノード情報テーブルから受信バッファーを取得する。
  std::map<int, NodeBoundaryInfo*>::iterator it_recv = m_recvNodeTable.begin();
  for( ; it_recv!=m_recvNodeTable.end(); it_recv++) {
    NodeBoundaryInfo* recvInfo = it_recv->second;
    /// 受信データの定義位置
    SphCrdDef ref_crddef = (SphCrdDef)recvInfo->m_crddef;
    int ref_faceFlag[3][2] = {0};
    memcpy(ref_faceFlag, recvInfo->m_faceFlag, sizeof(ref_faceFlag));

    /// 受信バッファーからデータを取得する。
    sphData* recvBuf = recvInfo->m_recvBuf;
    const void* recv_pdata = recvBuf->getDataPointer();

    /// 受信バッファーのサイズを取得する。
    const size_t* buf_size = recvBuf->_getArraySize();

    /// 受信データの近傍８点から更新データの線形補間を行う。
    double udt_coors[3] = {0.0, 0.0, 0.0};              // 更新データの座標 {x, y, z}
    double buf_coors_sta[3] = {0.0, 0.0, 0.0};          // 近傍８点の始点座標 {x, y, z}
    size_t buf_idx[3] = {0, 0, 0};                      // 近傍８点のインデックス {i, j, k}
    double buf_value[2][2][2] = {0};                    // 近傍８点のデータ [i,i+1][j,j+1][k,k+1]
    double buf_coors[2][2][2][3] = {0};                 // 近傍８点の座標 [i,i+1][j,j+1][k,k+1][x,y,z]

    /// 更新データの始点インデックス、終点インデックス間で受信データからセットする
    for (l=0; l<data_num; l++) {

      for (k=recvInfo->m_sta_idx[2]; k<=recvInfo->m_end_idx[2]; k++) {
        // 更新データの座標を取得する。
        udt_coors[2] = (udt_org[2]-udt_pitch[2]*gc) + udt_pitch[2]*k;
        if (udt_crddef == CRDDEF_COLLOCATE) udt_coors[2] += udt_pitch[2]/2;
        else if (udt_crddef == CRDDEF_STAGGERED1 && l != 2) udt_coors[2] += udt_pitch[2]/2;
        else if (udt_crddef == CRDDEF_STAGGERED2 && l != 2) udt_coors[2] += udt_pitch[2]/2;

        // 更新データの近傍受信データのインデックスを取得する
        double diff = udt_coors[2] - recvInfo->m_originNode[2];
        if (ref_crddef == CRDDEF_COLLOCATE) diff -= recvInfo->m_pitchNode[2]/2;
        else if (ref_crddef == CRDDEF_STAGGERED1 && l != 2) diff -= recvInfo->m_pitchNode[2]/2;
        else if (ref_crddef == CRDDEF_STAGGERED2 && l != 2) diff -= recvInfo->m_pitchNode[2]/2;
        // 参照データが更新データを内包しないので更新しない。
        if (diff < (-1)*DBL_EPSILON) continue;

        // staggerdの場合、最外殻は無効データ(u=j,k, v=i,k, w=i,j)
        if ((ref_crddef == CRDDEF_STAGGERED1 || ref_crddef == CRDDEF_STAGGERED2)
            && l != 2 && ref_faceFlag[2][1] == -1
            && k==recvInfo->m_end_idx[2]) {
          continue;
        }

        buf_idx[2] = (size_t)((diff) / recvInfo->m_pitchNode[2]);

        for (j=recvInfo->m_sta_idx[1]; j<=recvInfo->m_end_idx[1]; j++) {
          // 更新データの座標を取得する。
          udt_coors[1] = (udt_org[1]-udt_pitch[1]*gc) + udt_pitch[1]*j;
          if (udt_crddef == CRDDEF_COLLOCATE) udt_coors[1] += udt_pitch[1]/2;
          else if (udt_crddef == CRDDEF_STAGGERED1 && l != 1) udt_coors[1] += udt_pitch[1]/2;
          else if (udt_crddef == CRDDEF_STAGGERED2 && l != 1) udt_coors[1] += udt_pitch[1]/2;

          // 更新データの近傍受信データのインデックスを取得する
          diff = udt_coors[1] - recvInfo->m_originNode[1];
          if (ref_crddef == CRDDEF_COLLOCATE) diff -= recvInfo->m_pitchNode[1]/2;
          else if (ref_crddef == CRDDEF_STAGGERED1 && l != 1) diff -= recvInfo->m_pitchNode[1]/2;
          else if (ref_crddef == CRDDEF_STAGGERED2 && l != 1) diff -= recvInfo->m_pitchNode[1]/2;
          // 参照データが更新データを内包しないので更新しない。
          if (diff < (-1)*DBL_EPSILON) continue;

          // staggerdの場合、最外殻は無効データ(u=j,k, v=i,k, w=i,j)
          if ((ref_crddef == CRDDEF_STAGGERED1 || ref_crddef == CRDDEF_STAGGERED2)
              && l != 1 && ref_faceFlag[1][1] == -1
              && j==recvInfo->m_end_idx[1]) {
            continue;
          }

          buf_idx[1] = (size_t)((diff) / recvInfo->m_pitchNode[1]);

          for (i=recvInfo->m_sta_idx[0]; i<=recvInfo->m_end_idx[0]; i++) {
            // 更新データの座標を取得する。
            udt_coors[0] = (udt_org[0]-udt_pitch[0]*gc) + udt_pitch[0]*i;
            if (udt_crddef == CRDDEF_COLLOCATE) udt_coors[0] += udt_pitch[0]/2;
            else if (udt_crddef == CRDDEF_STAGGERED1 && l != 0) udt_coors[0] += udt_pitch[0]/2;
            else if (udt_crddef == CRDDEF_STAGGERED2 && l != 0) udt_coors[0] += udt_pitch[0]/2;

            // 更新データの近傍受信データのインデックスを取得する
            diff = udt_coors[0] - recvInfo->m_originNode[0];
            if (ref_crddef == CRDDEF_COLLOCATE) diff -= recvInfo->m_pitchNode[0]/2;
            else if (ref_crddef == CRDDEF_STAGGERED1 && l != 0) diff -= recvInfo->m_pitchNode[0]/2;
            else if (ref_crddef == CRDDEF_STAGGERED2 && l != 0) diff -= recvInfo->m_pitchNode[0]/2;
            // 参照データが更新データを内包しないので更新しない。
            if (diff < (-1)*DBL_EPSILON) continue;

            // staggerdの場合、最外殻は無効データ(u=j,k, v=i,k, w=i,j)
            if ((ref_crddef == CRDDEF_STAGGERED1 || ref_crddef == CRDDEF_STAGGERED2)
                && l != 0 && ref_faceFlag[0][1] == -1
                && i==recvInfo->m_end_idx[0]) {
              continue;
            }

            buf_idx[0] = (size_t)((diff) / recvInfo->m_pitchNode[0]);

            for (d=0; d<dims; d++) {
              // 近傍受信データが最外郭の場合は、近傍インデックス位置を最外郭-1にする。
              // 線形補間には更新データ位置の双端のデータが必要であるので、最外郭-1 - 最外郭とする。
              if (udt_crddef == CRDDEF_COLLOCATE && ref_faceFlag[d][1] == -1) {
                // collocateの場合、プラス側最外郭は無効データであるので、最外郭-2 - 最外郭-1とする。
                if (buf_idx[d]+2 >= buf_size[d]) {
                  buf_idx[d] = buf_size[d] - 3;
                }
              }
              else if (udt_crddef == CRDDEF_STAGGERED1 && ref_faceFlag[d][1] == -1) {
                if (l != d) {
                  // STAGGERED1の場合、プラス側最外郭(u,v,wの時, jk,ik,ij)は無効データであるので、最外郭-2 - 最外郭-1とする。
                  if (buf_idx[d]+2 >= buf_size[d]) {
                    buf_idx[d] = buf_size[d] - 3;
                  }
                }
              }
              else if (udt_crddef == CRDDEF_STAGGERED2 && ref_faceFlag[d][1] == -1) {
                // STAGGERED2の場合、プラス側最外郭は無効データであるので、最外郭-2 - 最外郭-1とする。
                if (buf_idx[d]+2 >= buf_size[d]) {
                  buf_idx[d] = buf_size[d] - 3;
                }
              }
              else if (udt_crddef == CRDDEF_STAGGERED2 && ref_faceFlag[d][0] == -1) {
                if (l == d) {
                  // STAGGERED2の場合、マイナス側最外郭(u,v,wの時, i,j,k)は無効データであるので、1-2とする。
                  if (buf_idx[d] == 0) {
                    buf_idx[d] = 1;
                  }
                }
              }
              if (buf_idx[d]+1 >= buf_size[d]) {
                buf_idx[d] = buf_size[d] - 2;
              }

              // 受信データの近傍８点の始点座標を取得する
              buf_coors_sta[d] = recvInfo->m_originNode[d] + recvInfo->m_pitchNode[d]*buf_idx[d];
              if (ref_crddef == CRDDEF_COLLOCATE) buf_coors_sta[d] += recvInfo->m_pitchNode[d]/2;
              else if (ref_crddef == CRDDEF_STAGGERED1 && l != d) buf_coors_sta[d] += recvInfo->m_pitchNode[d]/2;
              else if (ref_crddef == CRDDEF_STAGGERED2 && l != d) buf_coors_sta[d] += recvInfo->m_pitchNode[d]/2;
            }

            // 近傍受信データインデックスと+1のインデックス位置の座標、値を取得して周囲８点の座標、値を取得する。
            bool buf_min = false;
            bool buf_max = false;
            memset(buf_value, 0x00, sizeof(double)*2*2*2);
            for (kx=0; kx<2; kx++) {
              for (jx=0; jx<2; jx++) {
                for (ix=0; ix<2; ix++) {
                  bool coord_min = true;
                  bool coord_max = true;
                  for (d=0; d<dims; d++) {
                    double delta = 0.0;
                    if (d == 0) delta = recvInfo->m_pitchNode[d]*ix;
                    else if (d == 1) delta = recvInfo->m_pitchNode[d]*jx;
                    else if (d == 2) delta = recvInfo->m_pitchNode[d]*kx;
                    // 近傍8点の座標を取得する。
                    buf_coors[ix][jx][kx][d] = buf_coors_sta[d] + delta;

                    // 内包する座標であるかチェックする。
                    coord_min &= (udt_coors[d] >= buf_coors[ix][jx][kx][d]-DBL_EPSILON);
                    coord_max &= (udt_coors[d] <= buf_coors[ix][jx][kx][d]+DBL_EPSILON);
                  }
                  // 内包する座標であるかチェックする。
                  if (coord_min) buf_min = true;
                  if (coord_max) buf_max = true;

                  buf_value[ix][jx][kx] = getRealValue(recvBuf, buf_idx[0]+ix, buf_idx[1]+jx, buf_idx[2]+kx, l);
                }
              }
            }
            // 内包する座標であるかチェックする。
            if (!buf_min || !buf_max) {
              // 更新データの座標を内包していないので、更新除外する。
              continue;
            }

            // 近傍8点から線形補間を行う
            double udt_value = interpolateLinear(dims, udt_coors, buf_coors, buf_value);

            // 線形補間結果から更新データをセットする。
            if (!setData(udt_value, i, j, k, l)) {
              // error
              m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setData fault.");
              return false;
            }

          }     // for (i=recvInfo->m_sta_idx[0];...)
        }     // for (j=recvInfo->m_sta_idx[1];...)
      }     // for (k=recvInfo->m_sta_idx[2];...)
    }     // for (l=0;...)
  }     // for( ; it_recv!=m_recvNodeTable.end();

  return true;
}



/**
 * 参照・更新データオブジェクトから通信ノードテーブルを作成する。
 * @param       refDataObj       参照データオブジェクト
 * @param       udtDataObj       更新データオブジェクト
 * @param       ext_gc           拡張ガイドセル
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerSv2Sv::createCommNodeTable(
                        const sphData* refDataObj,
                        const sphData* udtDataObj,
                        size_t ext_gc)
{

  int myID = 0;
  int commSize = 0;
  SPL_Comm_rank(m_couplerComm, &myID);
  SPL_Comm_size(m_couplerComm, &commSize);

  SPL_Datatype dt_size_t = SPL_DATATYPE_NULL;
  if( sizeof(size_t) == 4 )      dt_size_t = SPL_INT;
  else if( sizeof(size_t) == 8 ) dt_size_t = SPL_LONG_LONG;
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INTERNAL_IMPOSSIBLE, "can not get sizeof size_t.");
    return false;
  }

  /// データを受信する（コピー先）の始点座標、幅
  register int i, n, m;
  double start_coords[3] = {0.0, 0.0, 0.0};
  double width[3] = {0.0, 0.0, 0.0};
  unsigned int dims = 0;
  unsigned int crddef = 0;
  int dtype_id = 0;
  if (udtDataObj != NULL) {
    // 次元数を取得する。
    dims = udtDataObj->getDims();
    // サイズを取得する。
    const size_t* vox_size = udtDataObj->getArraySize();
    // ガイドセルを取得する。
    size_t gc = udtDataObj->getGc();
    // １格子のデータ数を取得する。
    size_t data_num = udtDataObj->getNumOfDtElem();
    // データ定義位置を取得する。
    crddef = (unsigned int)udtDataObj->getCrdDef();
    // 原点座標を取得する。
    double org[3];
    udtDataObj->getDataOrigin(org);
    // ピッチを取得する。
    double pitch[3];
    udtDataObj->getDataPitch(pitch);
    // データクラスを取得する。
    SphDCType dctype = udtDataObj->getDcType();
    // データ型を取得する
    SPL_Datatype dtype = udtDataObj->getDataType();
    dtype_id = spl_get_dataindex_bydatatype(dtype);

    for (i=0; i<dims; i++) {
      start_coords[i] = org[i] - gc*pitch[i] - ext_gc*pitch[i];
      width[i] = (vox_size[i] + 2*gc + 2*ext_gc - 1)*pitch[i];
      // CRDDEF_COLLOCATE,CRDDEF_STAGGERED2の場合、vox_sizeは領域サイズ-1であるので、１加算する。
      if (crddef == CRDDEF_COLLOCATE || crddef == CRDDEF_STAGGERED2) {
        width[i] += pitch[i];
      }
    }
  }

  // 始点座標と幅を送信バッファーに組み立てる。
  double send_buf[6] = {0};
  for (i=0; i<3; i++) send_buf[i] = start_coords[i];
  for (i=0; i<3; i++) send_buf[i+3] = width[i];

  // 始点座標と幅を全ノードに送信する。（定義位置に関係無く全領域幅を送信する）
  double* recv_buf = new double[commSize*6];
  memset(recv_buf, 0x00, sizeof(double)*commSize*6);
  if( SPL_Allgather(send_buf, 6, SPL_DOUBLE,
                    recv_buf, 6, SPL_DOUBLE,
                    m_couplerComm) != SPL_SUCCESS ) {
    if (!recv_buf) delete [] recv_buf;
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Allgather fault.");
    m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
    return false;
  }

  // 受信した始点座標と幅から送信ノードテーブルを作成する。
  double* recv_sta_crds = NULL;
  double* recv_width = NULL;
  for (n=0; n<commSize; n++) {
    // データを送信する始点座標、幅
    recv_sta_crds = recv_buf + n * 6;
    recv_width = recv_buf + n * 6 + 3;

    // 受信した始点座標と幅から送信ノードテーブルを作成する。
    if (!createSendTable(refDataObj, n, recv_sta_crds, recv_width)) {
      if (!recv_buf) delete [] recv_buf;
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSendTable fault.");
      return false;
    }
  }
  if (!recv_buf) delete [] recv_buf;

  // 通信情報構造体
  typedef struct  {
    int node_id;                // 送信元ノードID
    int dims;                   // 次元数
    int dtype_id;               // データタイプ番号
    int crddef;               // データ定義位置
    size_t size[3];             // 送信サイズ
    double origin[3];           // 送信始点原点座標
    double pitch[3];            // 送信ピッチ
    int faceFlag[3][2];            // 最外郭送信フラグ
  } CommInfo;
  CommInfo send_strc_buf;

  // 通信情報構造体の通信タイプを作成する。
  SPL_Datatype commType = SPL_DATATYPE_NULL;
  int          block[8] = {1, 1, 1, 1, 3, 3, 3, 6};
  SPL_Datatype old_types[8] = {SPL_INT, SPL_INT, SPL_INT, SPL_INT, dt_size_t, SPL_DOUBLE, SPL_DOUBLE, SPL_INT};
//  SPL_Aint     disp[6] = {0, 4, 8, 12, 36, 60};
  SPL_Aint     disp[8];
  SPL_Address( &send_strc_buf, disp);
  SPL_Address( &(send_strc_buf.dims), disp+1);
  SPL_Address( &(send_strc_buf.dtype_id), disp+2);
  SPL_Address( &(send_strc_buf.crddef), disp+3);
  SPL_Address( send_strc_buf.size, disp+4);
  SPL_Address( send_strc_buf.origin, disp+5);
  SPL_Address( send_strc_buf.pitch, disp+6);
  SPL_Address( send_strc_buf.faceFlag, disp+7);
  // disp[0]を0とした値に換算する
  for (i=7; i>=0; i--) disp[i] -= disp[0];

  SPL_Type_struct( 8, block, disp, old_types, &commType);
  SPL_Type_commit( &commType );
  if (commType == SPL_DATATYPE_NULL && splIsParallel()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "created data type is null.");
    return false;
  }

  // 受信データ
  CommInfo* recv_strc_buf = new CommInfo[commSize];
  bool result = true;
  /// 送信ノードテーブルから送信先に送信情報を送信する。
  for (n=0; n<commSize; n++) {
    int recv_nodeid = n;                // 送信先ノード番号
    memset(&send_strc_buf, 0x00, sizeof(CommInfo));
    memset(recv_strc_buf, 0x00, sizeof(CommInfo)*commSize);

    // 送信ノード情報をセットする。
    send_strc_buf.node_id = myID;             // 自ノード番号
    std::map<int, NodeBoundaryInfo*>::const_iterator it = m_sendNodeTable.find(recv_nodeid);
    if(it != m_sendNodeTable.end()) {
      NodeBoundaryInfo* sendInfo = it->second;
      send_strc_buf.dims = sendInfo->m_dims;                // 次元数
      send_strc_buf.dtype_id = spl_get_dataindex_bydatatype(sendInfo->m_dtType);  // データタイプ番号
      send_strc_buf.crddef = sendInfo->m_crddef;        // データ定義位置
      for (i=0; i<send_strc_buf.dims; i++) {
        send_strc_buf.size[i] = sendInfo->m_size[i];
        send_strc_buf.origin[i] = sendInfo->m_originNode[i];
        send_strc_buf.pitch[i] = sendInfo->m_pitchNode[i];
      }
      memcpy(send_strc_buf.faceFlag, sendInfo->m_faceFlag, sizeof(send_strc_buf.faceFlag));
    }

    if (splIsParallel()) {
      // 送信先ノード番号にGatherを行う。
      if (SPL_Gather( &send_strc_buf, 1, commType,
                      recv_strc_buf, 1, commType,
                      recv_nodeid, m_couplerComm) != SPL_SUCCESS ) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Gather fault.");
        result = false;
      }
    }
    else {
      memcpy(&recv_strc_buf[0], &send_strc_buf, sizeof(send_strc_buf));
    }

    if (!result) break;

    // 受信側が自ノード、且つ受信データを持つ場合
    if (myID == recv_nodeid && udtDataObj != NULL) {
      for (m=0; m<commSize; m++) {
        // 送受信IDチェック
        if (recv_strc_buf[m].node_id != m) {
          // error : ノードIDが異なる。
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "invalid node_id[=%d] and receive id=%d.", m, recv_strc_buf[m].node_id);
          result = false;
        }
        size_t size = 1;
        int recv_dims = recv_strc_buf[m].dims;
        if (recv_dims == 0) recv_dims = dims;
        for (i=0; i<recv_dims; i++) size *= recv_strc_buf[m].size[i];
        if (size == 0) {
          // 受信データがない
          continue;
        }

        // 自データ情報と送信情報のチェックを行う
        if (recv_strc_buf[m].dims != dims) {
          // error : 次元数が異なる。
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not equals my dims[=%d] and receive[id=%d] dims[=%d].", dims, m, recv_strc_buf[m].dims);
          result = false;
        }
        if (recv_strc_buf[m].dtype_id != dtype_id) {
          // error : データタイプが異なる
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "not equals my dtype[=%d] and receive[id=%d] dtype[=%d].", dtype_id, m, recv_strc_buf[m].dtype_id);
          result = false;
        }

        if (!result) break;
        // 受信サイズ、始点座標、ピッチから受信ノードテーブルを作成する。
        if (!createReceiveTable(
                          udtDataObj,
                          m,
                          recv_strc_buf[m].crddef,
                          recv_strc_buf[m].size,
                          recv_strc_buf[m].origin,
                          recv_strc_buf[m].pitch,
                          recv_strc_buf[m].faceFlag)) {
          // error
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createReceiveTable fault.");
          result = false;
        }
        if (!result) break;
      }
    }
    if (!result) break;
  }

  if (!recv_strc_buf) delete [] recv_strc_buf;
  SPL_Type_free( &commType );

  if (!result) {
    m_errHandler->printStackTrace();
    SPL_Abort(SPL_COMM_WORLD, m_errHandler->getLastErrorNo());
  }

  return result;
}



/**
 * 参照データオブジェクトから送信ノードテーブルを作成する。.
 * 送信ノードテーブル
 *      m_nodeID          送信先ノードID
 *      m_sta_idx        送信する参照データの始点インデックス（ガイドセルを含む）
 *      m_end_idx        終点インデックス（ガイドセルを含む）：送信ノードでは未使用
 *      m_size           送信する参照データのサイズ（ガイドセルを含む）
 *      m_originNode     送信する参照データの原点座標
 *      m_pitchNode      送信する参照データのピッチ
 *      splDtype         送信通信タイプ
 *      m_recvBuf        NULL（送信ノードでは未使用）
 * @param       refDataObj       参照データオブジェクト
 * @param       src_nodeid       送信先ノード番号
 * @param       src_start_coords       送信始点座標
 * @param       src_width       送信幅
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerSv2Sv::createSendTable(
                      const sphData* refDataObj,
                      int src_nodeid,
                      const double src_start_coords[3],
                      const double src_width[3])
{
  if (src_start_coords == NULL || src_width == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createSendTable arguments is null.");
    return false;
  }

  SPL_Datatype dt_size_t = SPL_DATATYPE_NULL;
  if( sizeof(size_t) == 4 )      dt_size_t = SPL_INT;
  else if( sizeof(size_t) == 8 ) dt_size_t = SPL_LONG_LONG;
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INTERNAL_IMPOSSIBLE, "can not get sizeof size_t.");
    return false;
  }

  int myID = 0;
  int commSize = 0;
  SPL_Comm_rank(m_couplerComm, &myID);
  SPL_Comm_size(m_couplerComm, &commSize);

  register int i;
  unsigned int dims = 0;
  size_t send_sta_idx[3] = {0, 0, 0};
  size_t send_end_idx[3] = {0, 0, 0};
  size_t send_size[3] = {0, 0, 0};
  SPL_Datatype splDtype = SPL_DATATYPE_NULL;
  double send_sta_coords[3] = {0.0, 0.0, 0.0};
  double send_pitch[3] = {0.0, 0.0, 0.0};
  bool not_send = false;

  // 参照ソルバを持っていないプロセスも存在する。。
  if (refDataObj == NULL) {
    return true;
  }

  // 次元数を取得する。
  dims = refDataObj->getDims();
  // サイズを取得する。
  const size_t* vox_size = refDataObj->getArraySize();
  // ガイドセルを取得する。
  size_t gc = refDataObj->getGc();
  // １格子のデータ数を取得する。
  size_t data_num = refDataObj->getNumOfDtElem();
  // 原点座標を取得する。
  double org[3];
  refDataObj->getDataOrigin(org);
  // ピッチを取得する。
  double pitch[3];
  refDataObj->getDataPitch(pitch);
  // データクラスを取得する。
  SphDCType dctype = refDataObj->getDcType();
  // データ型を取得する
  SPL_Datatype dtype = refDataObj->getDataType();
  // データ定義位置を取得する。
  SphCrdDef crddef = refDataObj->getCrdDef();

  // CRDDEF_COLLOCATE,CRDDEF_STAGGERED2の場合、原点位置をpitch/2加算する。
//  if (crddef == CRDDEF_COLLOCATE) {
//    for (i=0;i<dims;i++) org[i] += pitch[i]/2;
//  }

  // 送信始点座標、幅から送信する始点インデックス、サイズ、始点座標を取得する
  double diff[3] = {0.0, 0.0, 0.0};
  for (i=0; i<dims; i++) {
    /// 送信始点インデックスを取得する。
    diff[i] = src_start_coords[i] - org[i];
    if (diff[i] >= 0.0) {
      // 差分長さをピッチで割った値がセル数
      // 送信始点がプラス側であるので、切り下げる。（送信領域の外側が更新データへのコピー領域）
      size_t diff_idx = (size_t)floor(diff[i]/pitch[i]);
      send_sta_idx[i] = diff_idx + gc;
      // 送信始点インデックスが領域外であるかチェックする。
      if (send_sta_idx[i] >= vox_size[i] + gc*2) {
        // 送信対象ノードではない。
        not_send = true;
      }
    }
    else if (diff[i] < 0.0) {
      // 差分長さをピッチで割った値がセル数
      // 送信始点がマイナス側であるので、切り上げる。（送信領域の外側が更新データへのコピー領域）
      size_t diff_idx = (size_t)ceil( (-1.0*diff[i])/pitch[i]);
      if (diff_idx >= gc) {
        send_sta_idx[i] = 0;
      }
      else {
        send_sta_idx[i] = gc - diff_idx;
      }
    }

    /// 送信終点インデックスを取得する。
    diff[i] = src_start_coords[i]+src_width[i] - org[i];
    if (diff[i] >= 0.0) {
      // 送信終点がプラス側であるので、切り上げる。（送信領域の外側が更新データへのコピー領域）
      size_t diff_idx = (size_t)ceil(diff[i]/pitch[i]);
      if (diff_idx + gc > 0) {
        send_end_idx[i] = diff_idx + gc;
      }
      else {
        send_end_idx[i] = 0;
      }
      // 送信終点インデックスが領域外であるかチェックする。
      if (send_end_idx[i] >= vox_size[i] + gc*2) {
        send_end_idx[i] = vox_size[i] + gc*2 - 1;
      }
    }
    else if (diff[i] < 0.0) {
      // 送信終点がマイナス側であるので、切り下げる。（送信領域の外側が更新データへのコピー領域）
      size_t diff_idx = (size_t)floor( (-1.0*diff[i])/pitch[i]);
      // 送信終点が領域外であるかチェックする。
      if (diff_idx + 1 > gc) {
        // 送信対象ノードではない。
        not_send = true;
      }
      else {
        // 送信終点インデックスを取得する。
        send_end_idx[i] = gc - diff_idx - 1;
      }
    }
    if (not_send) break;

    // 送信サイズを取得する。
    if (send_sta_idx[i] <= send_end_idx[i]) {
      send_size[i] = send_end_idx[i] - send_sta_idx[i] + 1;
    }
    if (send_sta_idx[i] + send_size[i] > vox_size[i] + gc*2) {
      send_size[i] = vox_size[i] + gc*2 - send_sta_idx[i];
    }

    // 送信始点座標、ピッチを取得する。
    send_sta_coords[i] = ((double)send_sta_idx[i]-(double)gc)*pitch[i]+org[i];
    send_pitch[i] = pitch[i];
  }

  // 送信サイズが０の場合は送信対象外である。
  if (!not_send) {
    // check send size
    size_t len = 1;
    for (i=0; i<dims; i++) len *= send_size[i];
    if (len == 0) {
      // 送信対象ノードではない。
      not_send = true;
    }
  }

  if (not_send) {
    // 送信対象ノードではない。
    return true;
  }

  /// 無効データを送信するかチェックする。
  int faceFlag[3][2] = {0};
  // 最外郭面
  const sphSvParaController* paraCtrl = dynamic_cast<const sphSvParaController*>(refDataObj->getParaController());
  const size_t *whole_size =  paraCtrl->getWholeVoxelSize();
  size_t vox_startidx[3] = {0, 0, 0};
  paraCtrl->getVoxelStartIndex(vox_startidx);

  if (paraCtrl != NULL) {
    for (i=0; i<dims; i++) {
      // マイナス方向が最外郭であるか？最外郭面を送信するか？
      if (vox_startidx[i] + send_sta_idx[i] == 0) {
        faceFlag[i][0] = paraCtrl->isCommID( ((i==0)?-1:0), ((i==1)?-1:0), ((i==2)?-1:0) );
      }
      // プラス方向が最外郭であるか？最外郭面を送信するか？
      if (vox_startidx[i] + send_sta_idx[i]+send_size[i] >= whole_size[i] + gc*2) {
        faceFlag[i][1] = paraCtrl->isCommID( ((i==0)?1:0), ((i==1)?1:0), ((i==2)?1:0) );
      }
    }
  }

  /// 送信ノードテーブルを作成する。
  size_t max[3] = {0, 0, 0};
  for (i=0; i<dims; i++) max[i] = vox_size[i] + gc*2 - 1;

  // 送信データタイプを作成する。
  splDtype = createCommType(dctype, dtype, max, send_sta_idx, send_end_idx, data_num);
  if (splDtype != SPL_DATATYPE_NULL || !splIsParallel()) {
    /// 送信情報を作成する。
    NodeBoundaryInfo* sendInfo = new NodeBoundaryInfo();
    sendInfo->m_nodeID = src_nodeid;             // 送信先ノードID
    for (i=0; i<3; i++) {
      sendInfo->m_sta_idx[i] = send_sta_idx[i];  // 送信始点インデックス（ガイドセルを含む）
      sendInfo->m_size[i] = send_size[i];        // 送信サイズ（ガイドセルを含む）
      sendInfo->m_originNode[i] = send_sta_coords[i];        // 送信原点座標
      sendInfo->m_pitchNode[i] = send_pitch[i];        // 送信ピッチ
    }
    sendInfo->m_crddef = crddef;                // 参照データの定義位置
    sendInfo->m_comType = splDtype;
    memcpy(sendInfo->m_faceFlag, faceFlag, sizeof(faceFlag));
    sendInfo->m_dims = dims;
    sendInfo->m_dtType = dtype;

    // 送信ノードテーブルに追加する。
    if( !((m_sendNodeTable.insert(std::make_pair(src_nodeid, sendInfo))).second) ){
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not add NodeBoundaryInfo for send. [send nodeID=%d].", src_nodeid);
      return false;
    }
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not create comm data type.");
    return false;
  }

  return true;
}



/**
 * 更新データオブジェクトから受信ノードテーブルを作成する。.
 * 受信ノードテーブル
 *      m_nodeID         受信元ノードID
 *      m_sta_idx        受信バッファーをコピーする更新データの始点インデックス（ガイドセルを含む）
 *      m_end_idx        受信バッファーをコピーする更新データの終点インデックス（ガイドセルを含む）
 *      m_size           受信する参照データのサイズ（ガイドセルを含む）
 *      m_originNode     受信する参照データの原点座標
 *      m_pitchNode      受信する参照データのピッチ
 *      m_crddef         受信する参照データの定義位置
 *      splDtype         受信通信タイプ
 *      m_recvBuf        受信バッファー
 * @param       udtDataObj       更新データオブジェクト
 * @param       recv_nodeid      受信ノードID
 * @param       crddef          データ定義位置
 * @param       recv_size        受信サイズ
 * @param       recv_start_coords        受信始点座標
 * @param       recv_pitch        受信ピッチ
 * @param       faceFlag          最外郭送信フラグ
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerSv2Sv::createReceiveTable(
                          const sphData* udtDataObj,
                          int recv_nodeid,
                          int recv_crddef,
                          const size_t recv_size[3],
                          const double recv_start_coords[3],
                          const double recv_pitch[3],
                          const int faceFlag[3][2])
{

  if (recv_size == NULL || recv_start_coords == NULL || recv_pitch == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createReceiveTable arguments is null.");
    return false;
  }

  register int i;
  // 参照ソルバを持っていないプロセスも存在する。。
  if (udtDataObj == NULL) {
    return true;
  }

  // 次元数を取得する。
  unsigned int dims = udtDataObj->getDims();
  // サイズを取得する。
  const size_t* vox_size = udtDataObj->getArraySize();
  // ガイドセルを取得する。
  size_t gc = udtDataObj->getGc();
  // １格子のデータ数を取得する。
  size_t data_num = udtDataObj->getNumOfDtElem();
  // 原点座標を取得する。
  double org[3];
  udtDataObj->getDataOrigin(org);
  // ピッチを取得する。
  double pitch[3];
  udtDataObj->getDataPitch(pitch);
  // データクラスを取得する。
  SphDCType dctype = udtDataObj->getDcType();
  // データ型を取得する
  SPL_Datatype dtype = udtDataObj->getDataType();
  // 定義位置
  SphCrdDef crddef = udtDataObj->getCrdDef();

  // CRDDEF_COLLOCATE,CRDDEF_STAGGERED2の場合、原点位置をpitch/2加算する。
//  if (crddef == CRDDEF_COLLOCATE) {
//    for (i=0;i<dims;i++) org[i] += pitch[i]/2;
//  }

  // データ長チェック
  size_t len = 1;
  for (i=0; i<dims; i++) len *= recv_size[i];
  if (len == 0) {
    return true;
  }

  // 受信バッファーから更新データにコピーする始点インデックスを算出する。
  size_t upd_sta_idx[3] = {0, 0, 0};
  size_t upd_end_idx[3] = {0, 0, 0};
  size_t upd_size[3] = {0, 0, 0};
  double diff[3] = {0.0, 0.0, 0.0};
  bool not_recv = false;
  for (i=0; i<dims; i++) {
    // 受信始点座標と更新データの始点座標との差分長さ
    diff[i] = recv_start_coords[i] - org[i];
    if (diff[i] >= 0.0) {
      // 差分長さをピッチで割った値がセル数
      // 受信始点がプラス側であるので、切り上げる。（受信領域の内側が更新データへのコピー領域）
      size_t diff_idx = (size_t)ceil(diff[i]/pitch[i]);
      upd_sta_idx[i] = diff_idx + gc;
      if (upd_sta_idx[i] >= vox_size[i] + gc*2) {
        // 更新データと参照データとの重複領域のあるノードのサイズを送信してきているので、領域外になることはありえない。
        not_recv = true;
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid receive start coords over than size, [recv_start_coords[%d]=%f,org[%d]=%f,pitch[%d]=%f,size[%d]=%ld,gc=%ld].", i, recv_start_coords[i], i, org[i], i, pitch[i], i, vox_size[i], gc);
        return false;
      }
    }
    else if (diff[i] < 0.0) {
      // 差分長さをピッチで割った値がセル数
      // 受信始点がマイナス側であるので、切り下げる。（受信領域の内側が更新データへのコピー領域）
      size_t diff_idx = (size_t)floor( (-1.0*diff[i])/pitch[i]);
      if (diff_idx > gc) {
        upd_sta_idx[i] = 0;
      }
      else {
        upd_sta_idx[i] = gc - diff_idx;
      }
    }

    // 受信終点座標と更新データの終点座標との差分長さ
    //diff[i] = recv_start_coords[i]+recv_pitch[i]*(recv_size[i]-1) - org[i];
    diff[i] = recv_start_coords[i]+recv_pitch[i]*(recv_size[i]) - org[i];
    if (diff[i] >= 0.0) {
      // 差分長さをピッチで割った値がセル数
      // 受信終点がプラス側であるので、切り下げる。（受信領域の内側が更新データへのコピー領域）
      size_t diff_idx = (size_t)floor(diff[i]/pitch[i]);
      upd_end_idx[i] = diff_idx + gc;
      if (upd_end_idx[i] >= vox_size[i] + gc*2) {
        upd_end_idx[i] = vox_size[i] + gc*2 - 1;
      }
    }
    else if (diff[i] < 0.0) {
      // 差分長さをピッチで割った値がセル数
      // 受信終点がマイナス側であるので、切り上げる。（受信領域の内側が更新データへのコピー領域）
      size_t diff_idx = (size_t)ceil( (-1.0*diff[i])/pitch[i]);
      if (diff_idx > gc) {
        not_recv = true;
        // 更新データと参照データとの重複領域のあるノードのサイズを送信してきているので、領域外になることはありえない。
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid receive end coords less than origin, [recv_start_coords[%d]=%f,recv_pitch[%d]=%f,recv_size[%d]=%ld,org[%d]=%f,pitch[%d]=%f,size[%d]=%ld,gc=%ld].", i, recv_start_coords[i], i, recv_pitch[i], i, recv_size[i], i, org[i], i, pitch[i], i, vox_size[i], gc);
        return false;
      }
      else {
        upd_end_idx[i] = gc - diff_idx;
      }
    }
    // 更新データにコピーするサイズの算出, 及びチェックを行う。
    if (upd_sta_idx[i] <= upd_end_idx[i]) {
      upd_size[i] = upd_end_idx[i] - upd_sta_idx[i] + 1;
    }
    else {
      // 更新データと参照データとの重複領域のあるノードのサイズを送信してきているので、領域外になることはありえない。
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid receive end index less than start index, [upd_sta_idx[%d]=%ls,upd_end_idx[%d]=%ld,recv_start_coords[%d]=%f,recv_pitch[%d]=%f,recv_size[%d]=%ld,org[%d]=%f,pitch[%d]=%f,size[%d]=%ld,gc=%ld].", i, upd_sta_idx[i], i, upd_end_idx[i], i, recv_start_coords[i], i, recv_pitch[i], i, recv_size[i], i, org[i], i, pitch[i], i, vox_size[i], gc);
      return false;
    }
    if (upd_sta_idx[i] + upd_size[i] > vox_size[i] + gc*2) {
      // 更新データと参照データとの重複領域のあるノードのサイズを送信してきているので、領域外になることはありえない。
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid receive end index less than start index, [udt_sta_idx[%d]=%ls,udt_size[%d]=%ld,recv_start_coords[%d]=%f,recv_pitch[%d]=%f,recv_size[%d]=%ld,org[%d]=%f,pitch[%d]=%f,size[%d]=%ld,gc=%ld].", i, upd_sta_idx[i], i, upd_size[i], i, recv_start_coords[i], i, recv_pitch[i], i, recv_size[i], i, org[i], i, pitch[i], i, vox_size[i], gc);
      return false;
    }
  }

  /// 受信バッファーを作成する
  /// インスタンスされているソルバオブジェクトの取得を行う。
  const sphSolverManager* solvMngr = getSolverManager();
  sphSolverBase* solverObj = solvMngr->getSolverById(m_udtSolverID);
  if (solverObj == NULL) {
    // 更新データオブジェクトが存在するのに更新ソルバが存在しないことはない。
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INTERNAL_IMPOSSIBLE, "can not get solver object.");
    return false;
  }

  // データ型別にインスタンスを生成しなければいけないので、データマネージャのcreateVoxelArrayを使用する。
  sphDataManager* dataMnger = solverObj->getDataManager();
  sphData* recv_buf = dataMnger->createVoxelArray(dctype, dtype, recv_size, 0, data_num);
  if (recv_buf == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create receive buffer.");
    return false;
  }

  // 受信通信タイプを作成する。
  size_t max[3] = {0, 0, 0};
  size_t buf_sta_idx[3] = {0, 0, 0};
  size_t buf_end_idx[3] = {0, 0, 0};
  for (i=0; i<dims; i++) max[i] = recv_size[i] - 1;
  for (i=0; i<dims; i++) buf_end_idx[i] = recv_size[i] - 1;
  SPL_Datatype splDtype = createCommType(dctype, dtype, max, buf_sta_idx, buf_end_idx, data_num);

  if (splDtype != SPL_DATATYPE_NULL || !splIsParallel()) {
    /// 受信情報を作成する。
    NodeBoundaryInfo* recvInfo = new NodeBoundaryInfo();
    recvInfo->m_nodeID = recv_nodeid;             // 受信先ノードID
    for (i=0; i<3; i++) {
      recvInfo->m_sta_idx[i] = upd_sta_idx[i];  // 更新始点インデックス
      recvInfo->m_end_idx[i] = upd_end_idx[i];  // 更新終点インデックス
      recvInfo->m_size[i] = recv_size[i];        // 受信サイズ
      recvInfo->m_originNode[i] = recv_start_coords[i];        // 受信原点座標
      recvInfo->m_pitchNode[i] = recv_pitch[i];        // 受信ピッチ
    }
    recvInfo->m_comType = splDtype;
    recvInfo->m_recvBuf = recv_buf;
    recvInfo->m_crddef = recv_crddef;
    memcpy(recvInfo->m_faceFlag, faceFlag, sizeof(recvInfo->m_faceFlag));

    if( !((m_recvNodeTable.insert(std::make_pair(recv_nodeid, recvInfo))).second) ){
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not add NodeBoundaryInfo for receive. [receive nodeID=%d].", recv_nodeid);
      return false;
    }
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not create comm data type.");
    return false;
  }

  return true;

}


/**
 * MPI通信タイプの作成を行う。
 * @param    dcType    データクラスタイプ
 * @param    dtType    データ型
 * @param    max    最大通信領域インデックス
 * @param    sta    開始通信領域インデックス
 * @param    end    終了通信領域インデックス
 * @param    data_num    １格子データ数
 * @return    通信タイプ
 */
SPL_Datatype
sphCouplerSv2Sv::createCommType(
                    SphDCType dcType,
                    SPL_Datatype dtType,
                    const size_t max[3],
                    const size_t sta[3],
                    const size_t end[3],
                    const size_t data_num)
{

  if (max == NULL || sta == NULL || end == NULL || data_num == 0) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "createCommType arguments is null.");
    return SPL_DATATYPE_NULL;
  }

  SPL_Datatype newType = SPL_DATATYPE_NULL;

  size_t l_end = data_num - 1;

  switch (dcType) {
    case (SPH_DC_ARRAY_1D):
      splCreateCommType1D(0, max[0], sta[0], end[0], dtType, &newType);
      break;
    case (SPH_DC_ARRAY_1DN):
      splCreateCommType2D(0, max[0], 0,
                          sta[0], end[0], 0, l_end, dtType, &newType);
      break;
    case (SPH_DC_ARRAY_1DNRV):
      splCreateCommType2D(0, l_end, 0,
                          0, l_end, sta[0], end[0],
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_2D):
      splCreateCommType2D(0, max[0], 0,
                          sta[0], end[0], sta[1], end[1],
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_2DN):
      splCreateCommType3D(0, max[0], 0, max[1], 0,
                          sta[0], end[0], sta[1], end[1], 0, l_end,
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_2DNRV):
      splCreateCommType3D(0, l_end, 0, max[0], 0,
                          0, l_end, sta[0], end[0], sta[1], end[1],
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_3D):
      splCreateCommType3D(0, max[0], 0, max[1], 0,
                          sta[0], end[0], sta[1], end[1], sta[2], end[2],
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_3DN):
      splCreateCommType4D(0, max[0], 0, max[1], 0, max[2], 0,
                          sta[0], end[0], sta[1], end[1], sta[2], end[2], 0, l_end,
                          dtType, &newType);
      break;
    case (SPH_DC_ARRAY_3DNRV):
      splCreateCommType4D(0, l_end, 0, max[0], 0, max[1], 0,
                          0, l_end, sta[0], end[0], sta[1], end[1], sta[2], end[2],
                          dtType, &newType);
      break;
    default:
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support dctype=%d.", (int)dcType);
      break;
  }

  return newType;
}



/**
 * データの指定インデックス位置の値をdouble型で取得する。.
 * sphVoxArray型へのdynamic_castは行わずに値を取得する。<br/>
 * i_idx,j_idx,k_idx,l_idxは、データクラスタイプによっても不変とする。<br/>
 * ２次元データの場合、k_idx値は不定値とする。<br/>
 * スカラデータの場合、l_idx値は不定値とする。<br/>
 * @param    dataObj    データオブジェクト
 * @param    i_idx      I方向インデックス（ガイドセルは含まない）
 * @param    j_idx      J方向インデックス（ガイドセルは含まない）
 * @param    k_idx      K方向インデックス（ガイドセルは含まない）
 * @param    l_idx      data_numインデックス
 * @return    データ値(double)
 */
double
sphCouplerSv2Sv::getRealValue(
                    sphData* dataObj,
                    size_t i_idx, size_t j_idx, size_t k_idx, size_t l_idx)
{

  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "getRealValue arguments is null.");
    return 0.0;
  }

  int dims = dataObj->getDims();
  const size_t* size = dataObj->_getArraySize();
  SphDCType dcType = dataObj->getDcType();
  SPL_Datatype datatype = dataObj->getDataType();
  int sizeofdata = spl_get_datasize(datatype);
  const void* dataPtr = dataObj->getDataPointer();
  size_t length = dataObj->getArrayLength();

  register int i;
  size_t idx = 0;

  switch (dcType) {
    case (SPH_DC_ARRAY_1D):
    case (SPH_DC_ARRAY_2D):
    case (SPH_DC_ARRAY_3D):
      if (dims >= 3) idx += size[0]*size[1]*k_idx;
      if (dims >= 2) idx += size[0]*j_idx;
      idx += i_idx;
      break;
    case (SPH_DC_ARRAY_1DN):
    case (SPH_DC_ARRAY_2DN):
    case (SPH_DC_ARRAY_3DN):
      idx += size[0]*size[1]*size[2]*l_idx;
      if (dims >= 3) idx += size[0]*size[1]*k_idx;
      if (dims >= 2) idx += size[0]*j_idx;
      idx += i_idx;
      break;
    case (SPH_DC_ARRAY_1DNRV):
    case (SPH_DC_ARRAY_2DNRV):
    case (SPH_DC_ARRAY_3DNRV):
      if (dims >= 3) idx += size[0]*size[1]*k_idx;
      if (dims >= 2) idx += size[0]*j_idx;
      idx += i_idx;
      idx += l_idx;
      break;
    default:
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support dctype=%d.", (int)dcType);
      return 0.0;
      break;
  }
  if (idx >= length) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "index is over than length,i,j,k,l=%ld,%ld,%ld,%ld,idx=%ld,array_length=%ld.", i_idx, j_idx, k_idx, l_idx, idx, length);
    return 0.0;
  }

  void* valuePtr = (char*)dataPtr + idx*sizeofdata;
  if( datatype == SPL_CHAR ) {
    char* ptr = (char*)valuePtr;
    char value = (*ptr);
    return (double)value;
  }
  else if( datatype == SPL_UNSIGNED_CHAR ) {
    unsigned char* ptr = (unsigned char*)valuePtr;
    unsigned char value = (*ptr);
    return (double)value;
  }
  else if( datatype == SPL_SHORT ) {
    short* ptr = (short*)valuePtr;
    short value = (*ptr);
    return (double)value;
  }
  else if( datatype == SPL_UNSIGNED_SHORT ) {
    unsigned short* ptr = (unsigned short*)valuePtr;
    unsigned short value = (*ptr);
    return (double)value;
  }
  else if( datatype == SPL_INT ) {
    int* ptr = (int*)valuePtr;
    int value = (*ptr);
    return (double)value;
  }
  else if( datatype == SPL_UNSIGNED ) {
    unsigned int* ptr = (unsigned int*)valuePtr;
    unsigned int value = (*ptr);
    return (double)value;
  }
  else if( datatype == SPL_LONG ) {
    long* ptr = (long*)valuePtr;
    long value = (*ptr);
    return (double)value;
  }
  else if( datatype == SPL_UNSIGNED_LONG ) {
    unsigned long* ptr = (unsigned long*)valuePtr;
    unsigned long value = (*ptr);
    return (double)value;
  }
  else if( datatype == SPL_LONG_LONG || datatype == SPL_LONG_LONG_INT) {
    long long* ptr = (long long*)valuePtr;
    long long value = (*ptr);
    return (double)value;
  }
  else if( datatype == SPL_FLOAT ) {
    float* ptr = (float*)valuePtr;
    float value = (*ptr);
    return (double)value;
  }
  else if( datatype == SPL_DOUBLE ) {
    double* ptr = (double*)valuePtr;
    double value = (*ptr);
    return (double)value;
  }

  return 0.0;
}



/**
 * データの指定インデックス位置の値を設定する。.
 * sphVoxArray型へのdynamic_castは行わずに値を設定する。<br/>
 * i_idx,j_idx,k_idx,l_idxは、データクラスタイプによっても不変とする。<br/>
 * ２次元データの場合、k_idx値は不定値とする。
 * スカラデータの場合、l_idx値は不定値とする。
 * @param    value      設定データ値
 * @param    i_idx      I方向インデックス（ガイドセルは含まない）
 * @param    j_idx      J方向インデックス（ガイドセルは含まない）
 * @param    k_idx      K方向インデックス（ガイドセルは含まない）
 * @param    l_idx      data_numインデックス
 * @return    success(=true)/fault(=false)
 */
bool
sphCouplerSv2Sv::setData(
                    double& value,
                    size_t i_idx, size_t j_idx, size_t k_idx, size_t l_idx)
{
  if (m_uptDataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_uptDataObj is null.");
    return SPL_DATATYPE_NULL;
  }

  int dims = m_uptDataObj->getDims();
  const size_t* size = m_uptDataObj->_getArraySize();
  SphDCType dcType = m_uptDataObj->getDcType();
  SPL_Datatype datatype = m_uptDataObj->getDataType();
  int sizeofdata = spl_get_datasize(datatype);
  void* dataPtr = m_uptDataObj->getDataPointer();

  register int i;
  size_t idx = 0;

  switch (dcType) {
    case (SPH_DC_ARRAY_1D):
    case (SPH_DC_ARRAY_2D):
    case (SPH_DC_ARRAY_3D):
      if (dims >= 3) idx += size[0]*size[1]*k_idx;
      if (dims >= 2) idx += size[0]*j_idx;
      idx += i_idx;
      break;
    case (SPH_DC_ARRAY_1DN):
    case (SPH_DC_ARRAY_2DN):
    case (SPH_DC_ARRAY_3DN):
      idx += size[0]*size[1]*size[2]*l_idx;
      if (dims >= 3) idx += size[0]*size[1]*k_idx;
      if (dims >= 2) idx += size[0]*j_idx;
      idx += i_idx;
      break;
    case (SPH_DC_ARRAY_1DNRV):
    case (SPH_DC_ARRAY_2DNRV):
    case (SPH_DC_ARRAY_3DNRV):
      if (dims >= 3) idx += size[0]*size[1]*k_idx;
      if (dims >= 2) idx += size[0]*j_idx;
      idx += i_idx;
      idx += l_idx;
      break;
    default:
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "can not support dctype=%d.", (int)dcType);
      return false;
      break;
  }

  void* valuePtr = (char*)dataPtr + idx*sizeofdata;
  if( datatype == SPL_CHAR ) {
    char* ptr = (char*)valuePtr;
    (*ptr) = (char)value;
  }
  else if( datatype == SPL_UNSIGNED_CHAR ) {
    unsigned char* ptr = (unsigned char*)valuePtr;
    (*ptr) = (unsigned char)value;
  }
  else if( datatype == SPL_SHORT ) {
    short* ptr = (short*)valuePtr;
    (*ptr) = (short)value;
  }
  else if( datatype == SPL_UNSIGNED_SHORT ) {
    unsigned short* ptr = (unsigned short*)valuePtr;
    (*ptr) = (unsigned short)value;
  }
  else if( datatype == SPL_INT ) {
    int* ptr = (int*)valuePtr;
    (*ptr) = (int)value;
  }
  else if( datatype == SPL_UNSIGNED ) {
    unsigned int* ptr = (unsigned int*)valuePtr;
    (*ptr) = (unsigned int)value;
  }
  else if( datatype == SPL_LONG ) {
    long* ptr = (long*)valuePtr;
    (*ptr) = (long)value;
  }
  else if( datatype == SPL_UNSIGNED_LONG ) {
    unsigned long* ptr = (unsigned long*)valuePtr;
    (*ptr) = (unsigned long)value;
  }
  else if( datatype == SPL_LONG_LONG || datatype == SPL_LONG_LONG_INT) {
    long long* ptr = (long long*)valuePtr;
    (*ptr) = (long long)value;
  }
  else if( datatype == SPL_FLOAT ) {
    float* ptr = (float*)valuePtr;
    (*ptr) = (float)value;
  }
  else if( datatype == SPL_DOUBLE ) {
    double* ptr = (double*)valuePtr;
    (*ptr) = (double)value;
  }
  else {
    return false;
  }

  return true;
}


/**
 * 線形補間を行う。.
 * @param    dims       次元数
 * @param    dst_coors[dims]      補間を算出する座標 (dims=xyz)
 * @param    src_coors[i][j][k][dims]      近傍８点の座標 (i=2,j=2,k=2,dims=xyz)
 * @param    src_value[i][j][k]      近傍８点の値 (i=2,j=2,k=2)
 * @return    補間結果値
 */
double
sphCouplerSv2Sv::interpolateLinear(
                        int dims,
                        const double dst_coors[3],
                        const double src_coors[2][2][2][3],
                        const double src_value[2][2][2])
{

  if (dst_coors == NULL || src_coors == NULL || src_value == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "interpolateLinear arguments is null.");
    return 0.0;
  }

  register int i, j, k;
  double interp_value0[3] = {0.0, 0.0, 0.0};
  double interp_value1[3] = {0.0, 0.0, 0.0};
  double interp_value = 0.0;

  // 1次元(j=0.k=0):(i[0],j[0])-(i[1],j[0]),k=0
  // {0,0,0}-{1,0,0}のライン上のx位置の値を算出する。
  i=0, j=0, k=0;
  interp_value0[0] = linearValue(
                        src_coors[i][j][k][0], src_value[i][j][k],
                        src_coors[i+1][j][k][0], src_value[i+1][j][k],
                        dst_coors[0]);
  if (dims == 1) return interp_value0[0];

  // 1次元(j=1.k=0):(i[0],j[1])-(i[1],j[1]),k=0
  // {0,1,0}-{1,1,0}のライン上のx位置の値を算出する。
  i=0, j=1, k=0;
  interp_value0[1] = linearValue(
                        src_coors[i][j][k][0], src_value[i][j][k],
                        src_coors[i+1][j][k][0], src_value[i+1][j][k],
                        dst_coors[0]);

  // 2次元(k=0):(i[0],j[0])-(i[1],j[1]),k=0
  // {0,0,0}-{1,1,0}の面上のy位置の値を算出する。
  i=0, j=0, k=0;
  interp_value0[2] = linearValue(
                        src_coors[i][j][k][1], interp_value0[0],
                        src_coors[i+1][j+1][k][1], interp_value0[1],
                        dst_coors[1]);
  if (dims == 2) return interp_value0[2];

  // 1次元(j=0.k=1):(i[0],j[0])-(i[1],j[0]),k=1
  // {0,0,1}-{1,0,1}のライン上のx位置の値を算出する。
  i=0, j=0, k=1;
  interp_value1[0] = linearValue(
                        src_coors[i][j][k][0], src_value[i][j][k],
                        src_coors[i+1][j][k][0], src_value[i+1][j][k],
                        dst_coors[0]);
  // 1次元(j=1.k=1):(i[0],j[1])-(i[1],j[1]),k=1
  // {0,1,1}-{1,1,1}のライン上のx位置の値を算出する。
  i=0, j=1, k=1;
  interp_value1[1] = linearValue(
                        src_coors[i][j][k][0], src_value[i][j][k],
                        src_coors[i+1][j][k][0], src_value[i+1][j][k],
                        dst_coors[0]);

  // 2次元(k=1):(i[0],j[0])-(i[1],j[1]),k=1
  // {0,0,1}-{1,1,1}の面上のy位置の値を算出する。
  i=0, j=0, k=1;
  interp_value1[2] = linearValue(
                        src_coors[i][j][k][1], interp_value1[0],
                        src_coors[i+1][j+1][k][1], interp_value1[1],
                        dst_coors[1]);

  // 3次元
  // {0,0,0}-{1,1,1}の面上のz位置の値を算出する。
  i=0, j=0, k=0;
  interp_value = linearValue(
                        src_coors[i][j][k][2], interp_value0[2],
                        src_coors[i+1][j+1][k+1][2], interp_value1[2],
                        dst_coors[2]);

  return interp_value;
}


/**
 * ２点間の線形補間を行う。.
 * @param    ix0            1点目の座標
 * @param    dx0            1点目の値
 * @param    ix1            2点目の座標
 * @param    dx1            2点目の値
 * @param    ix             補間を行う座標
 * @return    補間結果値
 */
double
sphCouplerSv2Sv::linearValue(double ix0, double dx0,
                            double ix1, double dx1,
                            double ix)
{
  double param = (ix-ix0)/(ix1-ix0);
  return (1-param)*dx0+param*dx1;
}



/**
 * 参照データのガイドセル領域の通信を行い、データの同期を取る。
 * @return    success(=true)/fault(=false)
 */
bool sphCouplerSv2Sv::commBndCell()
{
  // 参照データ
  sphData* ref_data = getReferData();
  if (ref_data == NULL) {
    // 参照データを持っていないノードも存在する。
    return true;
  }

  // ガイドセル領域の通信を行う。
  return ref_data->comm();

}



