/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphFlowBase.C
 * Sphere Flow super Class Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphFlowBase.C
 * Sphereフロー基底クラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */
#include "solverbase/sphFlowBase.h"
#include "solverbase/sphSolverBase.h"
#include "solverbase/sphCouplerBase.h"
#include "config/sphElemUtils.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphFlowBase::sphFlowBase() : sphSphereObject()
{
  // 初期化を行う。
  initialize();
}


/**
 * \~english
 * Constructor
 * @param     label     flow label
 *
 * \~japanese
 * コンストラクタ
 * @param     label     フローラベル
 */
sphFlowBase::sphFlowBase(const char* label) : sphSphereObject(label, 0)
{
  // 初期化を行う。
  initialize();

}


/**
 * \~english
 * Constructor
 * @param     label     flow label
 * @param     id     flow id
 *
 * \~japanese
 * コンストラクタ
 * @param     label     フローラベル
 * @param     id     フローID
 */
sphFlowBase::sphFlowBase(const char* label, int id) : sphSphereObject(label, id)
{

  // 初期化を行う。
  initialize();

}

/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphFlowBase::~sphFlowBase()
{

}

/**
 * 初期処理を行う.
 * @param     void
 * @return    success(=true)/fault(=false)
 */
bool sphFlowBase::initialize()
{
  m_steptime = new sphStepTime();
  return true;
}


/**
 * フローを初期設定する.
 * コンフィグレーションからステップ、時刻を設定する。
 * @param       config_elem       コンフィグレーションフロー要素
 * @return    success(=true)/fault(=false)
 */
int sphFlowBase::initializeFlow(const sphElem* config_elem)
{
  // コンフィグレーションフロー要素の設定
  setConfigElem(config_elem);

  /// ステップ、時刻を取得する。
  // <SphStepTime>
  const sphElem* elem = config_elem->getElem(SPH_CFG_NAME_STEPTIME);
  if (elem != NULL) {
    m_steptime->setMaxStep(elem->getAttrLLong(SPH_CFG_ATTR_MAXSTEP));
    m_steptime->setBaseStep(elem->getAttrLLong(SPH_CFG_ATTR_BASESTEP));
    m_steptime->setBaseTime(elem->getAttrReal8(SPH_CFG_ATTR_BASETIME));
    m_steptime->setDeltaTime(elem->getAttrReal8(SPH_CFG_ATTR_DELTATIME));
  }

  /// 最大ステップ数のデフォルト値を設定する。
  if (m_steptime->getMaxStep() == 0) {
    m_steptime->setMaxStep(SPH_CFG_DEF_MAX_STEPS);
  }

  /// １ステップ時間のデフォルト値を設定する。
  if (m_steptime->getDeltaTime() == 0) {
    m_steptime->setDeltaTime(SPH_CFG_DEF_DELTA_TIME);
  }

  return SPHERE_SUCCESS;
}


/**
 * フローを初期設定する.
 * コンフィグレーションからステップ、時刻を設定する。
 * @param       config_elem       コンフィグレーションフロー要素
 * @return    success(=true)/fault(=false)
 */
int sphFlowBase::initializeFlow(const SklCfg::SklSolverConfig* sklconfig)
{

  /// ステップ、時刻を取得する。
  if (sklconfig != NULL) {
    m_steptime->setMaxStep(sklconfig->GetCalcSteps());
    m_steptime->setBaseStep(0);
    m_steptime->setBaseTime(0.0);
    /// １ステップ時間のデフォルト値を設定する。
    m_steptime->setDeltaTime(SPH_CFG_DEF_DELTA_TIME);
  }

  /// 最大ステップ数のデフォルト値を設定する。
  if (m_steptime->getMaxStep() == 0) {
    m_steptime->setMaxStep(SPH_CFG_DEF_MAX_STEPS);
  }

  return SPHERE_SUCCESS;
}


/**
 * ステップ・時刻管理クラスを取得する
 * @param       void
 * @return    ステップ・時刻管理クラス
 */
sphStepTime* sphFlowBase::getStepTime() const
{
  return m_steptime;
}



