/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSimpleFlow.C
 * Sphere Simple Voxel Flow super Class Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphSimpleFlow.C
 * Sphere構造格子フロー基底クラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */
#include "solverbase/sphSimpleFlow.h"
#include "solverbase/sphSolverBase.h"
#include "solverbase/sphCouplerBase.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphSimpleFlow::sphSimpleFlow()
{

}

/**
 * \~english
 * Constructor
 * @param     label     flow label
 *
 * \~japanese
 * コンストラクタ
 * @param     label     フローラベル
 */
sphSimpleFlow::sphSimpleFlow(const char* label) : sphFlowBase(label)
{
}


/**
 * \~english
 * Constructor
 * @param     label     flow label
 * @param     id     flow id
 *
 * \~japanese
 * コンストラクタ
 * @param     label     フローラベル
 * @param     id     フローID
 */
sphSimpleFlow::sphSimpleFlow(const char* label, int id) : sphFlowBase(label, id)
{

}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphSimpleFlow::~sphSimpleFlow()
{

}


/**
 * フローを初期設定する.
 * コンフィグレーションフロー要素からノードIDリストを作成する。
 * @param       config_elem       コンフィグレーションフロー要素
 * @return      -1:error, 0:forced terminated, 1:normality
 */
int sphSimpleFlow::initializeFlow(const sphElem* config_elem)
{
  int ret = SPHERE_SUCCESS;
  // コンフィグレーションフロー要素の設定
  if ((ret = sphFlowBase::initializeFlow(config_elem)) != SPHERE_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "sphFlowBase::initializeFlow fault.");
    return ret;
  }

  return ret;
}

/**
 * フローを初期設定する.
 * コンフィグレーションフロー要素からノードIDリストを作成する。
 * @param       config_elem       コンフィグレーションフロー要素
 * @return      -1:error, 0:forced terminated, 1:normality
 */
int sphSimpleFlow::initializeFlow(const SklCfg::SklSolverConfig* sklconfig)
{
  int ret = SPHERE_SUCCESS;
  // コンフィグレーションフロー要素の設定
  if ((ret = sphFlowBase::initializeFlow(sklconfig)) != SPHERE_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "sphFlowBase::initializeFlow fault.");
    return ret;
  }

  return ret;
}


/**
 * ソルバを実行する。
 * @param       void
 * @return      -1:error, 0:forced terminated, 1:normality
 */
int sphSimpleFlow::executeSolvers()
{
  register int i;
  int result = 0;
  int ret = 0, tmp = 0;

  // initialize
  result = initializeSolvers();
  if (result != SPHERE_SUCCESS) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeSolver fault, return code=%d.", result);
    return result;
  }

  // check initializeSolver in all node
  SPL_Barrier(SPL_COMM_WORLD);

  // main loop
  result = mainSolvers();
  if (result != SPHERE_SUCCESS) {
    return result;
  }

  // finalize
  result = postSolvers();
  if (result != SPHERE_SUCCESS) {
    return result;
  }

  return SPHERE_SUCCESS;
}



/**
 * ソルバを初期化する。
 * @param       void
 * @return      -1:error, 0:forced terminated, 1:normality
 */
int sphSimpleFlow::initializeSolvers()
{

  register int i;
  int result = SPHERE_SUCCESS;
  const sphSolverManager* solvMngr = this->getSolverManager();
  const sphCouplerManager* coupMngr = this->getCouplerManager();

  // initialize
  int solvCnt = solvMngr->getSolverCount();
  for (i=0; i<solvCnt; i++) {
    sphSolverBase* solver = solvMngr->getSolver(i);
    result = solver->initializeSolver(m_steptime);
    if (result != SPHERE_SUCCESS) {
      return result;
    }
  }

  int coupCnt = coupMngr->getCouplerCount();
  for (i=0; i<coupCnt; i++) {
    sphCouplerBase* coupler = coupMngr->getCoupler(i);
    result = coupler->initializeCoupler();
    if (result != SPHERE_SUCCESS) {
      return result;
    }
  }
  return result;
}



/**
 * ソルバメインループを実行する。
 * @param       void
 * @return      -1:error, 0:forced terminated, 1:normality
 */
int sphSimpleFlow::mainSolvers()
{

  register int i, j;
  register size_t step;
  int result = SPHERE_SUCCESS;

  const sphSolverManager* solvMngr = this->getSolverManager();
  const sphCouplerManager* coupMngr = this->getCouplerManager();

  // main loop
  int solvCnt = solvMngr->getSolverCount();
  int coupCnt = coupMngr->getCouplerCount();

  /// 最大実行ステップを取得する。
  size_t maxStep = m_steptime->getMaxStep();

  /// ソルバをmaxStepループを行う。
  for(step=1; step<=maxStep; step++){
    // 現在の実行ステップ数を設定する。
    m_steptime->setCurrentStep(step);
    // 時間をインクリメントする。
    m_steptime->incrementTime();

    /// シグナル受信による終了をチェックする。
    if( isSphereExit() ) {
      return SPHERE_TERMINATE;
    }

    // ソルバ数ループ
    for (i=0; i<solvCnt; i++) {
      sphSolverBase* solver = solvMngr->getSolver(i);

      // チェックモードによる実行をチェックする。
      if (solver->isCheckMode()) {
        continue;
      }

      int solvId = solver->getSphereId();

      /// ソルバの"loopSolver"を呼び出す。
      result = solver->loopSolver(m_steptime);
      if (result != SPHERE_SUCCESS) {
        return result;
      }
    }    // for (i=0; i<solvCnt; i++)

    /// カプラの実行を行う。
    for (j=0; j<coupCnt; j++) {
      sphCouplerBase* coupler = coupMngr->getCoupler(j);
      result = coupler->executeCoupler();

      if (result != SPHERE_SUCCESS) {
        return result;
      }
    }    // for (j=0; j<coupCnt; j++)

  }  // for (step=1; step<=maxStep; step++)

  return result;

}


/**
 * ソルバ破棄処理を実行する。
 * @param       void
 * @return      -1:error, 0:forced terminated, 1:normality
 */
int sphSimpleFlow::postSolvers()
{
  register int i;
  int result = 1;

  const sphSolverManager* solvMngr = this->getSolverManager();
  const sphCouplerManager* coupMngr = this->getCouplerManager();

  // finalize
  int solvCnt = solvMngr->getSolverCount();
  for (i=0; i<solvCnt; i++) {
    sphSolverBase* solver = solvMngr->getSolver(i);
    result = solver->postSolver(m_steptime);
    if (result != 1) {
      return result;
    }
  }

  int coupCnt = coupMngr->getCouplerCount();
  for (i=0; i<coupCnt; i++) {
    sphCouplerBase* coupler = coupMngr->getCoupler(i);
    result = coupler->postCoupler();
    if (result != 1) {
      return result;
    }
  }

  return result;
}



