/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSolverAllocate.C
 * Sphere Solver super Class for Data Allocate Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphSolverAllocate.C
 * Sphereソルバー基底クラス（データ生成）ソースファイル
 * @author     hira
 * @date       2009/10/01
 */
#include "solverbase/sphSolverBase.h"
#include "config/sphElemUtils.h"


/**
 * データクラス生成を行う。(ラベル)
 * @param               cfg_label           データ要素識別ラベル
 * @return              データオブジェクト
 */
sphData* sphSolverBase::allocateDataObj(const char* cfg_label)
{
  if (cfg_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "allocateDataObj arguments is null.");
    return NULL;
  }
  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
    return NULL;
  }

  /// SphDataList/SphDataObj[@label]の検索を行う。
  const sphElem* data_elem = m_configElem->getElemListByLabel(SPH_CFG_NAME_DATALIST, SPH_CFG_NAME_DATAOBJ, cfg_label);
  if (data_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "label is not found.[label=%s]", cfg_label);
    return NULL;
  }

  /// <SphDataObj>要素からデータ生成を行う。
  sphData* dataObj = allocateDataObj(data_elem);
  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }

  return dataObj;

}



/**
 * データクラス生成を行う。(データオブジェクト要素)
 * @param               data_elem           データオブジェクト要素
 * @return              データオブジェクト
 */
sphData* sphSolverBase::allocateDataObj(const sphElem* data_elem)
{
  if (data_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "allocateDataObj arguments is null.");
    return NULL;
  }

  if (!data_elem->equalsElemName(SPH_CFG_NAME_DATAOBJ)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphDataObj>.");
    return NULL;
  }

  if (m_paraCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_paraCtrl is null.");
    return NULL;
  }

  /// データクラスと次元数をチェックする。
  // 自ノード領域のデータ生成を行う。
  SphDCType dctype = data_elem->getAttrSphDcType(SPH_CFG_ATTR_FORMAT);
  const char* data_label = data_elem->getAttrString(SPH_CFG_ATTR_LABEL);
  sphData* dataObj = NULL;
  switch(dctype) {
  case SPH_DC_ARRAY_1D:
  case SPH_DC_ARRAY_1DN:
  case SPH_DC_ARRAY_1DNRV:
  case SPH_DC_ARRAY_2D:
  case SPH_DC_ARRAY_2DN:
  case SPH_DC_ARRAY_2DNRV:
  case SPH_DC_ARRAY_3D:
  case SPH_DC_ARRAY_3DN:
  case SPH_DC_ARRAY_3DNRV: {
      sphDataUtils dataUtils;
      if (dataUtils.getDimsOfSphDCType(dctype) != m_solverDims) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "invalid DataClass Type, not match dims in Solver.[label=%s]", data_label);
        return NULL;
      }
      // 子ボクセルパラレルコントローラ数によって生成クラスを振り分ける。
      if (!m_paraCtrl->isMultiParaController()) {
        // シンプルボクセルデータクラス
        dataObj = m_dataMngr->createVoxelArray(data_elem, m_paraCtrl);
      }
      else {
        // 部分領域ボクセルデータクラス
        dataObj = m_dataMngr->createVoxelContainer(data_elem, m_paraCtrl);
      }
    }
    break;
  case SPH_DC_UNST_NODE:
  case SPH_DC_UNST_NODE_N:
  case SPH_DC_UNST_NODE_NRV:
  case SPH_DC_UNST_ELEM:
  case SPH_DC_UNST_ELEM_N:
  case SPH_DC_UNST_ELEM_NRV:
    dataObj = m_dataMngr->createUnstData(data_elem, m_paraCtrl);
    break;
  default:
    break;
  }
  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.[label=%s]", data_label);
    return NULL;
  }

  m_errHandler->blockWarning();          // 警告メッセージを無効にする。

  /// 子要素に<SphFile>, <SphMultiFile>要素が存在しているか？
  const sphElem* file_elem = getChildFileElem(data_elem, "in");
  if (file_elem == NULL) {
    // 子要素に<SphFile>要素が存在しないので、<SphMultiFile>要素を検索する。
    file_elem = data_elem->getElem(SPH_CFG_NAME_MULTIFILE);
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  if (file_elem != NULL) {
    /// ファイルからデータを読み込む。
    if ( loadFile(file_elem, dataObj) == NULL) {
      /// 生成済みdataObjの破棄
      m_dataMngr->deleteDataObj(dataObj->getLabel().c_str());
      dataObj = NULL;
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "loadFile fault.");
    }
  }

  return dataObj;
}



/**
 * データクラス生成を行う。(パラレルコントローラ指定).
 * 作成したデータオブジェクトは、データマネージャに登録する。
 * @param      reg_label          データ登録ラベル
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      paraCtrlr      パラレルコントローラー
 * @param      gc            ガイドセルサイズ
 * @param      data_num      セルのデータ数
 * @param      crddef        データ定義位置
 * @return     データオブジェクト
 */
sphData* sphSolverBase::allocateDataObj(
                      const char* reg_label,
                      SphDCType dcType,
                      SPL_Datatype dataType,
                      const size_t gc,
                      const size_t data_num,
                      SphCrdDef crddef)
{
  /// データクラスと次元数をチェックする。
  sphData* dataObj = NULL;
  switch(dcType) {
  case SPH_DC_ARRAY_1D:
  case SPH_DC_ARRAY_1DN:
  case SPH_DC_ARRAY_1DNRV:
  case SPH_DC_ARRAY_2D:
  case SPH_DC_ARRAY_2DN:
  case SPH_DC_ARRAY_2DNRV:
  case SPH_DC_ARRAY_3D:
  case SPH_DC_ARRAY_3DN:
  case SPH_DC_ARRAY_3DNRV: {
      sphDataUtils dataUtils;
      if (dataUtils.getDimsOfSphDCType(dcType) != m_solverDims) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "invalid DataClass Type, not match dims in Solver.");
        return NULL;
      }
      // 子ボクセルパラレルコントローラ数によって生成クラスを振り分ける。
      if (!m_paraCtrl->isMultiParaController()) {
        // シンプルボクセルデータクラス
        dataObj = m_dataMngr->createVoxelArray(dcType, dataType, reg_label, m_paraCtrl, gc, data_num, crddef);
      }
      else {
        // 部分領域ボクセルデータクラス
        dataObj = m_dataMngr->createVoxelContainer(dcType, dataType, reg_label, m_paraCtrl, gc, data_num, crddef);
      }
    }
    break;
  case SPH_DC_UNST_NODE:
  case SPH_DC_UNST_NODE_N:
  case SPH_DC_UNST_NODE_NRV:
  case SPH_DC_UNST_ELEM:
  case SPH_DC_UNST_ELEM_N:
  case SPH_DC_UNST_ELEM_NRV:
    dataObj = m_dataMngr->createUnstData(dcType, dataType, reg_label, m_paraCtrl, data_num);
    break;
  default:
    break;
  }

  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  return dataObj;
}


/**
 * データクラス生成を行う。(パラレルコントローラ指定).
 * 作成したデータオブジェクトは、データマネージャに登録しない。
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      gc            ガイドセルサイズ
 * @param      data_num      セルのデータ数
 * @param      crddef        データ定義位置
 * @return     データオブジェクト
 */
sphData* sphSolverBase::allocateDataObj(
                      SphDCType dcType,
                      SPL_Datatype dataType,
                      const size_t gc,
                      const size_t data_num,
                      SphCrdDef crddef)
{
  /// データクラスと次元数をチェックする。
  sphData* dataObj = NULL;
  switch(dcType) {
  case SPH_DC_ARRAY_1D:
  case SPH_DC_ARRAY_1DN:
  case SPH_DC_ARRAY_1DNRV:
  case SPH_DC_ARRAY_2D:
  case SPH_DC_ARRAY_2DN:
  case SPH_DC_ARRAY_2DNRV:
  case SPH_DC_ARRAY_3D:
  case SPH_DC_ARRAY_3DN:
  case SPH_DC_ARRAY_3DNRV: {
      sphDataUtils dataUtils;
      if (dataUtils.getDimsOfSphDCType(dcType) != m_solverDims) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "invalid DataClass Type, not match dims in Solver.");
        return NULL;
      }
      // 子ボクセルパラレルコントローラ数によって生成クラスを振り分ける。
      if (!m_paraCtrl->isMultiParaController()) {
        // シンプルボクセルデータクラス
        dataObj = m_dataMngr->createVoxelArray(dcType, dataType, m_paraCtrl, gc, data_num, crddef);
      }
      else {
        // 部分領域ボクセルデータクラス
        dataObj = m_dataMngr->createVoxelContainer(dcType, dataType, m_paraCtrl, gc, data_num, crddef);
      }
    }
    break;
  case SPH_DC_UNST_NODE:
  case SPH_DC_UNST_NODE_N:
  case SPH_DC_UNST_NODE_NRV:
  case SPH_DC_UNST_ELEM:
  case SPH_DC_UNST_ELEM_N:
  case SPH_DC_UNST_ELEM_NRV:
    dataObj = m_dataMngr->createUnstData(dcType, dataType, m_paraCtrl, data_num);
    break;
  default:
    break;
  }

  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  return dataObj;
}



/**
 * データクラス生成を行う。(サイズ指定).
 *
 * 作成したデータオブジェクトは、データマネージャに登録する。<br/>
 * データクラスの破棄はSPHEREで行う。<br/>
 * 他のノードとの通信は行わない。
 * 部分領域ボクセルデータクラスの生成は行わない。
 * @param      reg_label          データ登録ラベル
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      size[3]        データ領域サイズ
 * @param      gc            ガイドセルサイズ
 * @param      data_num      セルのデータ数
 * @return     データオブジェクト
 */
sphData* sphSolverBase::allocateDataObjExcept(
                      const char* reg_label,
                      SphDCType dcType,
                      SPL_Datatype dataType,
                      const size_t size[3],
                      const size_t gc,
                      const size_t data_num)
{

  sphData* dataObj = NULL;
  switch(dcType) {
  case SPH_DC_ARRAY_1D:
  case SPH_DC_ARRAY_1DN:
  case SPH_DC_ARRAY_1DNRV:
  case SPH_DC_ARRAY_2D:
  case SPH_DC_ARRAY_2DN:
  case SPH_DC_ARRAY_2DNRV:
  case SPH_DC_ARRAY_3D:
  case SPH_DC_ARRAY_3DN:
  case SPH_DC_ARRAY_3DNRV:
    dataObj = m_dataMngr->createVoxelArray(
                            dcType, dataType, reg_label,
                            size,
                            gc, data_num);
    break;
  case SPH_DC_UNST_NODE:
  case SPH_DC_UNST_NODE_N:
  case SPH_DC_UNST_NODE_NRV:
  case SPH_DC_UNST_ELEM:
  case SPH_DC_UNST_ELEM_N:
  case SPH_DC_UNST_ELEM_NRV:
    dataObj = m_dataMngr->createUnstData(dcType, dataType, reg_label, m_paraCtrl, data_num);
    break;
  default:
    break;
  }

  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  return dataObj;
}


/**
 * データクラス生成を行う。(サイズ指定).
 * 作成したデータオブジェクトは、データマネージャに登録しない。<br/>
 * データクラスの破棄は呼出し側で行う必要がある。<br/>
 * 他のノードとの通信は行わない。
 * 部分領域ボクセルデータクラスの生成は行わない。
 * @param      dcType        データクラスタイプ
 * @param      dtType        データタイプ
 * @param      size[3]        データ領域サイズ
 * @param      gc            ガイドセルサイズ
 * @param      data_num      セルのデータ数
 * @return     データオブジェクト
 */
sphData* sphSolverBase::allocateDataObjExcept(
                      SphDCType dcType,
                      SPL_Datatype dataType,
                      const size_t size[3],
                      const size_t gc,
                      const size_t data_num)
{

  sphData* dataObj = NULL;
  switch(dcType) {
  case SPH_DC_ARRAY_1D:
  case SPH_DC_ARRAY_1DN:
  case SPH_DC_ARRAY_1DNRV:
  case SPH_DC_ARRAY_2D:
  case SPH_DC_ARRAY_2DN:
  case SPH_DC_ARRAY_2DNRV:
  case SPH_DC_ARRAY_3D:
  case SPH_DC_ARRAY_3DN:
  case SPH_DC_ARRAY_3DNRV:
    dataObj = m_dataMngr->createVoxelArray(
                            dcType, dataType,
                            size,
                            gc, data_num);
    break;
  case SPH_DC_UNST_NODE:
  case SPH_DC_UNST_NODE_N:
  case SPH_DC_UNST_NODE_NRV:
  case SPH_DC_UNST_ELEM:
  case SPH_DC_UNST_ELEM_N:
  case SPH_DC_UNST_ELEM_NRV:
    dataObj = m_dataMngr->createUnstData(dcType, dataType, m_paraCtrl, data_num);
    break;
  default:
    break;
  }

  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
    return NULL;
  }
  return dataObj;
}

