/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSolverBase.C
 * Sphere Solver super Class Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphSolverBase.C
 * Sphereソルバー基底クラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */
#include "solverbase/sphSolverBase.h"
#include "solverbase/sphFlowBase.h"
#include "config/sphElemUtils.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphSolverBase::sphSolverBase() : sphSphereObject()
{
  /// 初期化を行う。
  initialize();
}


/**
 * \~english
 * Constructor
 * @param     label     solver label
 *
 * \~japanese
 * コンストラクタ
 * @param     label     ソルバーラベル
 */
sphSolverBase::sphSolverBase(const char* label) : sphSphereObject(label, 0)
{
  /// 初期化を行う。
  initialize();
}


/**
 * \~english
 * Constructor
 * @param     label     solver label
 * @param     id     solver id
 *
 * \~japanese
 * コンストラクタ
 * @param     label     ソルバーラベル
 * @param     id     ソルバーID
 */
sphSolverBase::sphSolverBase(const char* label, int id) : sphSphereObject(label, id)
{

  /// 初期化を行う。
  initialize();
}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphSolverBase::~sphSolverBase()
{
  if (m_solverComm != SPL_COMM_WORLD) {
    if (m_solverComm != SPL_COMM_NULL) SPL_Comm_free(&m_solverComm);
    if (m_solverGroup != SPL_GROUP_NULL) SPL_Group_free(&m_solverGroup);
  }
  if (m_dataMngr != NULL) delete m_dataMngr;
  if (m_fileCtrl != NULL) delete m_fileCtrl;
}



/**
 * 初期化を行う。
 * @param     void
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::initialize()
{

  m_solverComm = SPL_COMM_NULL;
  m_solverGroup = SPL_GROUP_NULL;
  m_dataMngr = new sphDataManager();
  m_paraCtrl = NULL;
  m_fileCtrl = new sphFileController();
  m_sklconfig = NULL;

  return true;
}


/**
 * ソルバーのコミュニケータ、グループを設定する.
 * @param               comm    コミュニケータ
 * @param               group   グループ
 * @return              void
 */
void sphSolverBase::setCommGroup(SPL_Comm comm, SPL_Group group)
{
  m_solverComm = comm;
  m_solverGroup = group;
  return;
}



/**
 * カプラ全体のコミュニケータを取得する.
 * @param               void
 * @return              コミュニケータ(SPL_Comm)
 */
SPL_Comm sphSolverBase::getSplComm() const
{
  return m_solverComm;
}


/**
 * カプラ全体のグループを取得する.
 * @param               void
 * @return              グループ(SPL_Group)
 */
SPL_Group sphSolverBase::getSplGroup() const
{
  return m_solverGroup;
}


/**
 * パラレルコントローラを設定する.
 * @param      paraCtrl       パラレルコントローラ
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::setParaController(sphParaController* paraCtrl)
{
  if (paraCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "paraCtrl is null");
    return false;
  }
  m_paraCtrl = paraCtrl;
  return true;
}

/**
 * パラレルコントローラを取得する (const).
 * @param      void
 * @return    パラレルコントローラ
 */
const sphParaController* sphSolverBase::getParaController() const
{
  return m_paraCtrl;
}


/**
 * パラレルコントローラを取得する.
 * @param      void
 * @return    パラレルコントローラ
 */
sphParaController* sphSolverBase::getParaController()
{
  return m_paraCtrl;
}



/**
 * ソルバモードを取得する.
 * @param      void
 * @return    ソルバモード
 */
const std::string sphSolverBase::getSolverMode() const
{
  return m_solverMode;
}


/**
 * データマネージャを取得する(const).
 * @param      void
 * @return    データマネージャ
 */
const sphDataManager* sphSolverBase::getDataManager() const
{
  return m_dataMngr;
}


/**
 * データマネージャを取得する.
 * @param      void
 * @return    データマネージャ
 */
sphDataManager* sphSolverBase::getDataManager()
{
  return m_dataMngr;
}


/**
 * 生成済みのデータオブジェクトを取得する。
 * @param               label           データラベル
 * @return    データオブジェクト
 */
sphData* sphSolverBase::getDataObj(const char* label)
{
  return m_dataMngr->getDataObj(label);
}


/**
 * データオブジェクトを登録する。
 * @param               label           データラベル
 * @param               dataObj           データオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::registerDataObj(const char* label, sphData* dataObj)
{

  return m_dataMngr->registDataObj(label, dataObj);
}

/**
 * データオブジェクトを破棄する。（データラベル）
 * @param               label           データラベル
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::deleteDataObj(const char* label)
{
  return m_dataMngr->deleteDataObj(label);
}

/**
 * データオブジェクトを破棄する。（データオブジェクト）
 * @param               dataObj           データオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::deleteDataObj(sphData* dataObj)
{
  if (m_dataMngr->deleteDataObj(dataObj)) {
    return true;
  }


  delete dataObj;
  dataObj = NULL;

  return true;
}



/**
 * 自ボクセルサイズを取得する。
 * @param               size        ボクセルサイズ（ガイドセルを含まない）
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::getVoxelSize(size_t size[3])
{
  if (m_paraCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_paraCtrl is null");
    return false;
  }

  const size_t* vox_sz = m_paraCtrl->getVoxelSize();
  register int i;
  for (i=0; i<3; i++) size[i] = vox_sz[i];

  return true;
}


/**
 * 自ボクセルの始点インデックスを取得する。
 * @param               sidx        始点インデックス
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::getVoxelStartIndex(size_t sidx[3])
{
  if (m_paraCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_paraCtrl is null");
    return false;
  }

  m_paraCtrl->getVoxelStartIndex(sidx);

  return true;
}



/**
 * 計算領域全体のボクセルサイズを取得する。
 * @param               wsize        計算領域全体ボクセルサイズ（ガイドセルを含まない）
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::getWholeVoxelSize(size_t wsize[3])
{
  if (m_paraCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_paraCtrl is null");
    return false;
  }

  const size_t* vox_wsz = m_paraCtrl->getWholeVoxelSize();
  register int i;
  for (i=0; i<3; i++) wsize[i] = vox_wsz[i];

  return true;
}


/**
 * ソルバ次元数を取得する。
 * @param     void
 * @return    ソルバ次元数
 */
int sphSolverBase::getDims() const
{
  return m_solverDims;
}

/**
 * ソルバ次元数を設定する。
 * @param     ソルバ次元数
 */
void sphSolverBase::setDims(int dims)
{
  m_solverDims = dims;
}


/**
 * ステップ・時刻管理クラスを取得する。(const)
 * @param     void
 * @return    ステップ・時刻管理クラス
 */
const sphStepTime* sphSolverBase::getStepTime() const
{
  const sphFlowBase* flow = getFlow();
  return flow->getStepTime();
}


/**
 * ステップ・時刻管理クラスを取得する。
 * @param     void
 * @return    ステップ・時刻管理クラス
 */
sphStepTime* sphSolverBase::getStepTime()
{
  sphFlowBase* flow = getFlow();
  return flow->getStepTime();
}



/**
 * ファイルコントローラに必要な情報を設定する。
 * @param               void
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::setupFileController()
{
  if (m_fileCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileCtrl is null.");
    return false;
  }
  const char* inputDir = ".";
  const char* outputDir = ".";
  if (!isClassicConfig()) {
    // Sphereコンフィグレーションの場合
    sphElemUtils elem_utils(m_configElem);

    /// 入出力フォルダを設定する。
    inputDir = elem_utils.getInputBaseDir();
    outputDir = elem_utils.getOutputBaseDir();
  }
  // SklSolverConfigの場合は、"./"とする。

  /// 入出力フォルダを設定する。
  m_fileCtrl->setInputBaseDir(inputDir);
  m_fileCtrl->setOutputBaseDir(outputDir);

  m_fileCtrl->setProcGrp(m_solverComm);             // 読込コミュニケータ

  /// 自ランク番号を設定する。
  int myID = 0;
  SPL_Comm_rank(m_solverComm, &myID);
  m_fileCtrl->setFileFormatRankId(myID);

  return true;

}


/**
 * Vsphereコンフグレーションを設定する。
 * @param      skl_config      Vsphereコンフグレーション
 * @return     void
 */
void sphSolverBase::setSklConfig(const SklCfg::SklSolverConfig* skl_config)
{
  m_sklconfig = skl_config;
  return;
}



/**
 * Vsphereコンフグレーションを取得する。
 * @param      void
 * @return     Vsphereコンフグレーション
 */
const SklCfg::SklSolverConfig* sphSolverBase::getSklConfig() const
{
  return m_sklconfig;
}


/**
 * Vsphereコンフグレーションであるかチェックする。
 * @param      void
 * @return     true:Vsphereコンフグレーション
 */
bool sphSolverBase::isClassicConfig()
{
  return (m_sklconfig != NULL);
}


/**
 * ソルバーの実行モードがチェックモードであるかチェックする。
 * @param      void
 * @return     true:チェックモード
 */
bool sphSolverBase::isCheckMode() const
{
  if (m_solverMode.c_str() == NULL) return false;
  return (strcasecmp(m_solverMode.c_str(), SPH_CFG_SOLVER_CHECKMODE) == 0);
}


