/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSolverFileIO.C
 * Sphere Solver super Class for fileIO Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphSolverFileIO.C
 * Sphereソルバー基底クラス（ファイル入出力）ソースファイル
 * @author     hira
 * @date       2009/10/01
 */
#include "solverbase/sphSolverBase.h"
#include "config/sphElemUtils.h"
#include "fileio/sphFileController.h"
#include "sph_mem_util.h"
#include "debugLog.h"


/**
 * ファイル識別ラベルによってファイル出力を行う。.
 * データラベルはコンフグレーションのデータ識別文字列ではなく、データマネージャ登録済みのラベルである。
 * @param[in]      file_label          ファイル識別ラベル
 * @param[in]      data_reg_label          データラベル（データマネージャ登録済みラベル）
 * @return     true=success
 */
bool sphSolverBase::writeFile(
                        const char* file_label,
                        const char* data_reg_label)
{
  if (file_label == NULL || data_reg_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "writeFile arguments is null");
    return false;
  }

  const sphElem* file_elem = NULL;
  if (isClassicConfig()) {
    file_elem = createSklOutFileElem(file_label);
  }
  else {
    if (m_configElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
      return NULL;
    }
    /// ファイル要素の取得を行う。
    file_elem = sphElemUtils(m_configElem).getElemByFileLabel(file_label);
    if (file_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "SphFile[@label=%s]", file_label);
      return false;
    }
  }
  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "file_elem is null");
    return NULL;
  }

  // 出力ガイドセルフラグ
  bool gc_flag = getGcOutputFlag(file_elem);

  /// データオブジェクトの取得を行う。
  sphData* dataObj = getDataObj(data_reg_label);
  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "not register label=%s in DataManager.", data_reg_label);
    return false;
  }

  /// 実行ステップ
  sphStepTime* steptime = getStepTime();
  size_t step = steptime->getTotalStep();
  /// 実行時間
  double time = steptime->getTotalTime();

  /// 書き込みを実行する。
  if (!writeFile(file_elem, dataObj, step, time, gc_flag)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "writeFile fault.");
    return false;
  }

  return true;
}


/**
 * データ識別ラベルによってファイル出力を行う。.
 * データ識別ラベルはコンフグレーションのデータ識別文字列であり、データマネージャにデータ識別ラベルでデータが登録されていなければなりません。
 * データ識別ラベルのデータ要素には、ファイル出力用のSphFile[@type='out]要素が記述されていなければなりません。
 * @param[in]      data_cfg_label          データラベル（コンフグレーションのデータ識別文字列）
 * @return     true=success
 */
bool sphSolverBase::writeFile(const char* data_cfg_label)
{
  if (data_cfg_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "writeFile arguments is null");
    return false;
  }

  // Sphereコンフィグレーション
  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
    return NULL;
  }

  /// データ要素の取得を行う。
  const sphElem* data_elem = m_configElem->getElemListByLabel(SPH_CFG_NAME_DATALIST, SPH_CFG_NAME_DATAOBJ, data_cfg_label);
  if (data_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "SphDataObj[@label=%s]", data_cfg_label);
    return false;
  }

  /// 子要素の<SphFile[@type='out'>要素を検索する。
  const sphElem* file_elem = getChildFileElem(data_elem, "out");
  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "not exist SphFile in SphDataObj[@label=%s], %s", data_cfg_label);
    return false;
  }

  // 出力ガイドセルフラグ
  bool gc_flag = getGcOutputFlag(file_elem);

  /// データオブジェクトの取得を行う。
  sphData* data = getDataObj(data_cfg_label);
  if (data == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "not register label=%s in DataManager.", data_cfg_label);
    return false;
  }

  /// 実行ステップ
  sphStepTime* steptime = getStepTime();
  size_t step = steptime->getTotalStep();
  /// 実行時間
  double time = steptime->getTotalTime();

  /// 書き込みを実行する。
  if (!writeFile(file_elem, data, step, time, gc_flag)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "writeFile fault.");
    return false;
  }

  return true;
}


/**
 * データオブジェクトをファイル出力する。.
 * @param[in]      file_label          ファイル識別ラベル
 * @param[in]      dataObj               書込データ
 * @return     true=success
 */
bool sphSolverBase::writeFile(
                        const char* file_label,
                        sphData* dataObj)
{
  if (file_label == NULL || dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "writeFile arguments is null");
    return false;
  }

  const sphElem* file_elem = NULL;
  if (isClassicConfig()) {
    // Vsphereコンフィグレーション
    file_elem = createSklOutFileElem(file_label);
  }
  else {
    // Sphereコンフィグレーション
    if (m_configElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
      return NULL;
    }
    /// ファイル要素の取得を行う。
    file_elem = sphElemUtils(m_configElem).getElemByFileLabel(file_label);
    if (file_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "SphFile[@label=%s]", file_label);
      return false;
    }
  }
  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "file_elem is null");
    return NULL;
  }

  // 出力ガイドセルフラグ
  bool gc_flag = getGcOutputFlag(file_elem);

  /// 実行ステップ
  sphStepTime* steptime = getStepTime();
  size_t step = steptime->getTotalStep();
  /// 実行時間
  double time = steptime->getTotalTime();

  /// 書き込みを実行する。
  if (!writeFile(file_elem, dataObj, step, time, gc_flag)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "writeFile fault.");
    return false;
  }

  return true;
}


/**
 * データ識別ラベルによってファイル出力を行う。.
 * データ識別ラベルはコンフグレーションのデータ識別文字列であり、データマネージャにデータ識別ラベルでデータが登録されていなければなりません。
 * データ識別ラベルのデータ要素には、ファイル出力用のSphFile[@type='out]要素が記述されていなければなりません。
 * @param[in]      data_cfg_label          データラベル（コンフグレーションのデータ識別文字列）
 * @param[in]      step               書込ステップ数
 * @param[in]      time               書込時間
 * @param[in]      gc_flag            ガイドセル出力フラグ（true:データオブジェクトのガイドセルを出力する/false:ガイドセル＝０で出力する。)
 * @param[in]      force              強制出力フラグ（true:SphFileOption[@interval]に従わない/false:SphFileOption[@interval]に従う）
 * @return     true=success
 */
bool sphSolverBase::writeFile(const char* data_cfg_label,
                          size_t step, double time,
                          bool gc_flag,
                          bool force)
{
  if (data_cfg_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "writeFile arguments is null");
    return false;
  }

  // Sphereコンフィグレーション
  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
    return NULL;
  }

  /// データ要素の取得を行う。
  const sphElem* data_elem = m_configElem->getElemListByLabel(SPH_CFG_NAME_DATALIST, SPH_CFG_NAME_DATAOBJ, data_cfg_label);
  if (data_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "SphDataObj[@label=%s]", data_cfg_label);
    return false;
  }

  /// 子要素の<SphFile[@type='out'>要素を検索する。
  const sphElem* file_elem = getChildFileElem(data_elem, "out");
  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "not exist SphFile in SphDataObj[@label=%s], %s", data_cfg_label);
    return false;
  }

  /// データオブジェクトの取得を行う。
  sphData* dataObj = getDataObj(data_cfg_label);
  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "not register label=%s in DataManager.", data_cfg_label);
    return false;
  }

  /// 書き込みを実行する。
  if (!writeFile(file_elem, dataObj, step, time, gc_flag, force)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "writeFile fault.");
    return false;
  }

  return true;
}


/**
 * データオブジェクトをファイル出力する。.
 * @param[in]      file_label          ファイル識別ラベル
 * @param[in]      dataObj               書込データ
 * @param[in]      step               書込ステップ数
 * @param[in]      time               書込時間
 * @param[in]      gc_flag            ガイドセル出力フラグ（true:データオブジェクトのガイドセルを出力する/false:ガイドセル＝０で出力する。)
 * @param[in]      force              強制出力フラグ（true:SphFileOption[@interval]に従わない/false:SphFileOption[@interval]に従う）
 * @return     true=success
 */
bool sphSolverBase::writeFile(
                        const char* file_label,
                        sphData* dataObj,
                        size_t step, double time,
                        bool gc_flag,
                        bool force)
{
  if (file_label == NULL || dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "writeFile arguments is null");
    return false;
  }

  const sphElem* file_elem = NULL;
  if (isClassicConfig()) {
    // Vsphereコンフィグレーション
    file_elem = createSklOutFileElem(file_label);
  }
  else {
    // Sphereコンフィグレーション
    if (m_configElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
      return NULL;
    }
    /// ファイル要素の取得を行う。
    file_elem = sphElemUtils(m_configElem).getElemByFileLabel(file_label);
    if (file_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "SphFile[@label=%s]", file_label);
      return false;
    }
  }
  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "file_elem is null");
    return NULL;
  }

  if (!force) {
    /// 出力ステップであるか、intervalをチェックする。
    if (!sphElemUtils(file_elem).isFileInterval(step)) {
      return true;
    }
  }

  /// 書き込みを実行する。
  if (!writeFile(file_elem, dataObj, step, time, gc_flag, force)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "writeFile fault.");
    return false;
  }

  return true;
}


/**
 * データオブジェクトをファイル出力する。
 *
 * 部分管理データクラスとボクセルデータクラスにより出力メソッドを分岐する。
 * @param[in]      file_elem          ファイル要素
 * @param[in]      dataObj               書込データオブジェクト
 * @param[in]      step               書込ステップ数
 * @param[in]      time               書込時間
 * @param[in]      gc_flag            ガイドセル出力フラグ（true:データオブジェクトのガイドセルを出力する/false:ガイドセル＝０で出力する。)
 * @param[in]      force              強制出力フラグ（true:SphFileOption[@interval]に従わない/false:SphFileOption[@interval]に従う）
 * @return     データオブジェクト
 */
bool sphSolverBase::writeFile(const sphElem* file_elem,
                            sphData* dataObj,
                            size_t step, double time,
                            bool gc_flag,
                            bool force)
{

  if (file_elem == NULL || dataObj == NULL || m_fileCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "writeFile arguments is null");
    return false;
  }

  bool ret = false;
  // 部分領域管理データクラスであるかチェックする。
  if (!dataObj->isVoxContainer()) {
    // ボクセルデータクラス用のファイル出力メソッド
    ret = writeFileVoxArray(file_elem, dataObj, step, time, gc_flag, force);
  }
  else {
    // 部分領域管理データクラス用のファイル出力メソッド
    ret = writeFileVoxContainer(file_elem, dataObj, step, time, gc_flag, force);
  }

  return ret;
}


/**
 * ボクセルデータオブジェクトをファイル出力する。.
 * @param[in]      file_elem          ファイル要素
 * @param[in]      dataObj               書込データオブジェクト
 * @param[in]      step               書込ステップ数
 * @param[in]      time               書込時間
 * @param[in]      gc_flag            ガイドセル出力フラグ（true:データオブジェクトのガイドセルを出力する/false:ガイドセル＝０で出力する。)
 * @param[in]      force              強制出力フラグ（true:SphFileOption[@interval]に従わない/false:SphFileOption[@interval]に従う）
 * @return      true=success
 */
bool sphSolverBase::writeFileVoxArray(const sphElem* file_elem,
                            sphData* dataObj,
                            size_t step, double time,
                            bool gc_flag,
                            bool force)
{
  if (file_elem == NULL || dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "writeFile arguments is null");
    return false;
  }
  if (m_fileCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileCtrl is null");
    return false;
  }

  if (!force) {
    /// 出力ステップであるか、intervalをチェックする。
    if (!sphElemUtils(file_elem).isFileInterval(step)) {
      return true;
    }
  }

  /// データオブジェクトから原点座標、ピッチを設定する。
  const sphParaController* para_ctrl = dataObj->getParaController();
  if (para_ctrl == NULL) {
    para_ctrl = this->getParaController();
  }

  double data_org[3] = {0.0, 0.0, 0.0};
  double data_pitch[3] = {0.0, 0.0, 0.0};
  // 原点座標の取得
  if (!para_ctrl->getNodeOrigin(data_org)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "getDataOrigin fault.");
    return false;
  }

  if (!para_ctrl->getVoxelPitch(data_pitch)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "getDataPitch fault.");
    return false;
  }

  register int i;
  /// データオブジェクトから読込サイズ、始点インデックスを設定する。
  const size_t* voxel_size = dataObj->getArraySize();
  size_t dims = dataObj->getDims();
  size_t gc = dataObj->getGc();
  SphCrdDef crddef = dataObj->getCrdDef();

  sphData* writeDataObj = dataObj;
  if (gc > 0 && gc_flag == false) {
    // ガイドセルを出力しない。
    gc = 0;
    // ガイドセルなしのデータオブジェクトの生成
    writeDataObj = dataObj->createVoxelArray(gc, true);    // gc=0, データコピー
    if (writeDataObj == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create write data object.");
      return false;
    }
  }

  /// パラレルコントローラから始点インデックスを取得する。
  size_t voxel_sidx[3] = {0, 0, 0};
  para_ctrl->getVoxelStartIndex(voxel_sidx);

  /// 計算領域全体サイズを取得する。
  const size_t* whole_size = para_ctrl->getWholeVoxelSize();

  // 定義位置がCollocateの場合、全体サイズを-1する。（デフォルト定義位置が変更された場合、変更する必要がある。 at 2010/06/22 by @hira)
  size_t data_whole_size[3] = {whole_size[0], whole_size[1], whole_size[2]};
  if (crddef == CRDDEF_COLLOCATE) {
    for (i=0; i<3; i++) data_whole_size[i] -= 1;
  }

  /// ファイル要素を設定する。
  if (!m_fileCtrl->initFileCtrl(file_elem)) {
    if (writeDataObj != dataObj) delete writeDataObj;
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setFileElem fault.");
    return false;
  }

  /// 原点座標、ピッチを設定する。
  m_fileCtrl->setCoordsInfo(data_org, data_pitch, dataObj->getCrdDef());

  /// 出力領域のサイズ、始点インデックス、ガイドセル, 計算領域全体サイズを設定する。
  m_fileCtrl->setVoxelInfo(dims, voxel_size, voxel_sidx, gc, data_whole_size);

  /// 実行ステップ、時間を設定する。
  m_fileCtrl->setStepTime(step, time);

  /// 書込データオブジェクトを設定する。
  m_fileCtrl->setData(writeDataObj);

  // 出力自ノードIDの設定
  int myID = para_ctrl->getMyID();
  m_fileCtrl->setFileFormatRankId(myID);

  /// ファイルへ書き込みを行う
  bool ret = m_fileCtrl->writeFile();
  if (!ret) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "writeFile fault.");
  }

  if (writeDataObj != dataObj) delete writeDataObj;

  return ret;
}



/**
 * 部分管理データオブジェクトをファイル出力する。.
 * @param[in]      file_elem          ファイル要素
 * @param[in]      dataObj               書込データオブジェクト
 * @param[in]      step               書込ステップ数
 * @param[in]      time               書込時間
 * @param[in]      gc_flag            ガイドセル出力フラグ（true:データオブジェクトのガイドセルを出力する/false:ガイドセル＝０で出力する。)
 * @param[in]      force              強制出力フラグ（true:SphFileOption[@interval]に従わない/false:SphFileOption[@interval]に従う）
 * @return     true=success
 */
bool sphSolverBase::writeFileVoxContainer(
                            const sphElem* file_elem,
                            sphData* dataObj,
                            size_t step,
                            double time,
                            bool gc_flag,
                            bool force)
{
  if (file_elem == NULL || dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "writeFile arguments is null");
    return false;
  }
  if (m_fileCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileCtrl is null");
    return false;
  }

  if (!force) {
    /// 出力ステップであるか、intervalをチェックする。
    if (!sphElemUtils(file_elem).isFileInterval(step)) {
      return true;
    }
  }

  register int i, j, k, l, lr, n, c;
  bool ret = false;
  int vox_num = dataObj->getNumberOfVoxArray();

  // ファイル入出力モード
  SphCfgFileIOMode cfg_mode = sphElemUtils(file_elem).getFileIoMode();
  if (cfg_mode == CFG_FILE_DEFAULT || cfg_mode == CFG_FILE_MULTI) {
    // 部分領域データのサブブロック単位のマルチ出力の場合、サブブロック毎に出力する。
    for (i=0; i<vox_num; i++) {
      sphData* writeDataObj = dataObj->getVoxArray(i);
      if (writeDataObj == NULL) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_DATA_IS_NULL, "getVoxArray is null [id=%d]", i);
        return false;
      }

      /// 書き込みを実行する。
      if (!writeFile(file_elem, writeDataObj, step, time, gc_flag, force)) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "writeFile fault.");
        return false;
      }
    }
    ret = true;
  }
  else if (cfg_mode == CFG_FILE_PROCESS) {
    int gc = 0;
    if (gc_flag) gc = dataObj->getGc();

    // 部分領域データオブジェクトのプロセス領域のデータオブジェクトのコピーを生成する。
    sphData* writeDataObj = dataObj->createVoxelArray(gc, true);

    /// 書き込みを実行する。
    ret = writeFile(file_elem, writeDataObj, step, time, gc_flag, force);

    // 生成コピーデータオブジェクトの破棄
    delete writeDataObj;

    if (!ret) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "writeFile fault.");
      return false;
    }
  }
  else if (cfg_mode == CFG_FILE_WHOLE) {
    // 部分領域管理データクラスの場合、プロセス間にて領域の重なりがあるかもしれないので
    // SPLファイルIO関数でGATHERしない。
    // ここでGATHERを行う。
    int writeRank = file_elem->getAttrInt(SPH_CFG_ATTR_IO_NODEID);
    SphDCType dcType = dataObj->getDcType();
    SPL_Datatype dataType = dataObj->getDataType();
    const size_t gc  = dataObj->getGc();
    const size_t data_num  = dataObj->getNumOfDtElem();
    SphCrdDef crddef = dataObj->getCrdDef();
    int dtype_sz = spl_get_datasize(dataType);

    sphData* writeDataObj = NULL;
    const sphSvParaController *solv_para = dynamic_cast<const sphSvParaController*>(this->getParaController());
    int myId = solv_para->getMyID();
    size_t *procs_size = NULL;
    size_t *procs_startid = NULL;
    SPL_Datatype dt_size_t = (sizeof(size_t) == 4) ? SPL_INT : SPL_LONG_LONG;

    // 計算領域全体サイズを取得する。
    const size_t* whole_size = solv_para->getWholeVoxelSize();
    const size_t* whole_start = solv_para->getReferIndex();
    int proc_num = solv_para->getNumberOfNode();

    // 出力ガイドセル
    int write_gc = (gc_flag) ? gc : 0;

    ret = true;
    if (writeRank == myId) {
      // ソルバ全体の書込バッファーを用意する。
      writeDataObj = allocateDataObjExcept(
                                        dcType,
                                        dataType,
                                        whole_size,
                                        write_gc,
                                        data_num);

      int proc_num = solv_para->getNumberOfNode();
      procs_size = new size_t[proc_num*3];
      procs_startid = new size_t[proc_num*3];
      memset(procs_size, 0x00, sizeof(size_t)*proc_num*3);
      memset(procs_startid, 0x00, sizeof(size_t)*proc_num*3);
    }

    // 部分領域データオブジェクトのプロセス領域のデータオブジェクトのコピーを生成する。
    sphData* myDataObj = dataObj->createVoxelArray(write_gc, true);

    size_t my_size[3] = {0};
    memcpy(my_size, myDataObj->_getArraySize(), sizeof(size_t)*3);   // ガイドセルを含むサイズ
    size_t my_startid[3] = {0, 0, 0};
    myDataObj->getDataStartIdx(my_startid);      // ガイドセルを含まない始点インデックス

    ret &= (SPL_Gather(my_size, 3, dt_size_t,
                  procs_size, 3, dt_size_t,
                  writeRank, solv_para->getSplComm()) == SPL_SUCCESS );

    ret &= (SPL_Gather(my_startid, 3, dt_size_t,
                  procs_startid, 3, dt_size_t,
                  writeRank, solv_para->getSplComm()) == SPL_SUCCESS );

    if (ret) {

      if (writeRank == myId) {

        const size_t *writeData_size = writeDataObj->_getArraySize();
        char* write_p =(char*)writeDataObj->getDataPointer();

        for (n=0; n<proc_num; n++) {
          unsigned char* recv_buf = NULL;
          size_t data_len = procs_size[n*3]*procs_size[n*3+1]*procs_size[n*3+2]*data_num*dtype_sz;
          recv_buf = new unsigned char[sizeof(unsigned char)*data_len];
          memset(recv_buf, 0x00, sizeof(char)*data_len);

          if (n == myId) {
            memcpy(recv_buf, myDataObj->getDataPointer(), data_len);
          }
          else {
            SPL_Status status;
            if ( SPL_Recv(recv_buf,
                          data_len,
                          SPL_UNSIGNED_CHAR,
                          n, 1, solv_para->getSplComm(), &status) != SPL_SUCCESS ) {
              ret = false;
              delete [] recv_buf;
              // error
              m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Recv fault.");
              break;
            }

          }

          // 受信データをソルバ全体の書込バッファーにセットする。
          size_t recv_startid[3] = {procs_startid[n*3], procs_startid[n*3+1], procs_startid[n*3+2]};
          size_t recv_size[3] = {procs_size[n*3], procs_size[n*3+1], procs_size[n*3+2]};

          int dt_num = data_num, dt_num_rv = 1;
          if (dcType == SPH_DC_ARRAY_1DNRV || dcType == SPH_DC_ARRAY_2DNRV || dcType == SPH_DC_ARRAY_3DNRV) {
            dt_num = 1; dt_num_rv = data_num;
          }

          for (l=0; l<dt_num; l++) {
            for (k=0; k<recv_size[2]; k++) {
              for (j=0; j<recv_size[1]; j++) {

                size_t write_idx = writeData_size[0]*writeData_size[1]*writeData_size[2]*l
                          + writeData_size[0]*writeData_size[1]*(k+recv_startid[2])
                          + writeData_size[0]*(j+recv_startid[1])
                          + recv_startid[0];
                size_t recv_idx = recv_size[0]*recv_size[1]*recv_size[2]*l
                          + recv_size[0]*recv_size[1]*k
                          + recv_size[0]*j;
                write_idx *= dtype_sz;
                recv_idx *= dtype_sz;
                write_idx *= dt_num_rv;
                recv_idx *= dt_num_rv;
                size_t cp_len = recv_size[0]*dt_num_rv*dtype_sz;
                // 0x00以外のデータのコピーを行う。
                memcpyex(write_p+write_idx, recv_buf+recv_idx, cp_len, dtype_sz);

/**************
                for (i=0; i<recv_size[0]; i++) {
                  for (lr=0; lr<dt_num_rv; lr++) {
                    size_t write_idx = writeData_size[0]*writeData_size[1]*writeData_size[2]*l
                              + writeData_size[0]*writeData_size[1]*(k+recv_startid[2])
                              + writeData_size[0]*(j+recv_startid[1])
                              + (i+recv_startid[0])
                              + lr;
                    size_t recv_idx = recv_size[0]*recv_size[1]*recv_size[2]*l
                              + recv_size[0]*recv_size[1]*k
                              + recv_size[0]*j
                              + i + lr;
                    write_idx *= dtype_sz;
                    recv_idx *= dtype_sz;
                    for (c=0; c<dtype_sz; c++) {
                      *(write_p+write_idx+c) |= *(recv_buf+recv_idx+c);
                    }     // for (c=0; c<dtype_sz; c++)
                  }     // for (lr=0; lr=dt_num_rv; lr++)
                }     // for (i=0; i<recv_size[0]; i++)
*******************/

              }     // for (j=0; j<recv_size[1]; j++)
            }     // for (k=0; k<recv_size[2]; k++)
          }     // for (l=0; l<dt_num; l++)
          delete [] recv_buf;

        }
      }
      else {
        size_t data_len = my_size[0]*my_size[1]*my_size[2]*data_num*dtype_sz;
        // send my data
        if ( SPL_Send((void*)myDataObj->getDataPointer(),
                              data_len,
                              SPL_UNSIGNED_CHAR,
                              writeRank, 1, solv_para->getSplComm()) != SPL_SUCCESS ) {
          // error
          m_errHandler->SPH_ERROR_INFO(SPH_ERR_MPI_ERROR, "SPL_Send fault.");
          ret = false;
        }
      }
    }

    if (ret) {
      // 書込ノードのみファイル出力する。
      if (writeRank == myId) {
        // 原点座標、ピッチ
        double vox_org[3], vox_pit[3];
        solv_para->getVoxelOrigin(vox_org);
        solv_para->getVoxelPitch(vox_pit);

        // 始点インデックス
        size_t vox_sidx[3];
        solv_para->getVoxelStartIndex(vox_sidx);

        /// 書き込みを実行する。
        ret = writeFileExcept(file_elem, writeDataObj, vox_org, vox_pit, step, time, vox_sidx, myId);
      }
    }

    if (writeDataObj) delete writeDataObj;
    if (myDataObj) delete myDataObj;
    if (procs_size) delete [] procs_size;
    if (procs_startid) delete [] procs_startid;
  }
  else {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "invalid mode attribute.");
    return false;
  }

  return ret;

}



/**
 * ボクセルデータオブジェクトをファイル出力する。.
 * @param[in]      file_elem          ファイル要素
 * @param[in]      dataObj               書込データオブジェクト
 * @param[in]      vox_org            書込データ原点座標
 * @param[in]      vox_pit            書込データピッチ
 * @param[in]      time               書込時間
 * @param[in]      step               書込ステップ数
 * @param[in]      time               書込時間
 * @param[in]      vox_startidx       書込データ始点インデックス
 * @param[in]      gc_flag            ガイドセル出力フラグ（true:データオブジェクトのガイドセルを出力する/false:ガイドセル＝０で出力する。)
 * @return      true=success
 */
bool sphSolverBase::writeFileExcept(const sphElem* file_elem,
                            sphData* dataObj,
                            double vox_org[3],
                            double vox_pit[3],
                            size_t step,
                            double time,
                            size_t vox_sidx[3],
                            int blockid)
{
  if (file_elem == NULL || dataObj == NULL || vox_org == NULL || vox_pit == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "writeFile arguments is null");
    return false;
  }
  if (m_fileCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileCtrl is null");
    return false;
  }

  register int i;
  /// データオブジェクトから読込サイズ、始点インデックスを設定する。
  const size_t* voxel_size = dataObj->getArraySize();
  size_t dims = dataObj->getDims();
  size_t gc = dataObj->getGc();
  SphCrdDef crddef = dataObj->getCrdDef();

  /// パラレルコントローラから始点インデックスを取得する。
  const sphParaController *para_ctrl = this->getParaController();

  /// 計算領域全体サイズを取得する。
  const size_t* whole_size = para_ctrl->getWholeVoxelSize();

  // 定義位置がCollocateの場合、全体サイズを-1する。（デフォルト定義位置が変更された場合、変更する必要がある。 at 2010/06/22 by @hira)
  size_t data_whole_size[3] = {whole_size[0], whole_size[1], whole_size[2]};
  if (crddef == CRDDEF_COLLOCATE) {
    for (i=0; i<3; i++) data_whole_size[i] -= 1;
  }

  /// ファイル要素を設定する。
  if (!m_fileCtrl->initFileCtrl(file_elem)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setFileElem fault.");
    return false;
  }

  /// 原点座標、ピッチを設定する。
  m_fileCtrl->setCoordsInfo(vox_org, vox_pit, dataObj->getCrdDef());

  /// 出力領域のサイズ、始点インデックス、ガイドセル, 計算領域全体サイズを設定する。
  m_fileCtrl->setVoxelInfo(dims, voxel_size, vox_sidx, gc, data_whole_size);

  /// 実行ステップ、時間を設定する。
  m_fileCtrl->setStepTime(step, time);

  /// 書込データオブジェクトを設定する。
  m_fileCtrl->setData(dataObj);

  // 出力自ノードIDの設定
  int myID = 0;
  if (blockid < 0) {
    myID = para_ctrl->getMyID();
  }
  m_fileCtrl->setFileFormatRankId(myID);

  /// ファイルへ書き込みを行う
  bool ret = m_fileCtrl->writeFile();
  if (!ret) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "writeFile fault.");
  }

  return ret;
}


/**
 * データ要素配下の入出力タイプと一致するファイル要素を取得する。
 * @param[in]      file_elem          ファイル要素
 * @param[in]      type               入出力タイプ ("in"/"out")
 * @return         ファイル要素
 */
const sphElem* sphSolverBase::getChildFileElem(const sphElem* data_elem, const char* type)
{
  /// 子要素の<SphFile[@type='in'/'out']>要素を検索する。
  const sphElem* file_elem = NULL;
  const sphElem* child_elem = NULL;
  const sphElem* found_elem = NULL;
  // 子要素からSphFile[@type='in'/'out']を検索する。
  child_elem = data_elem->getChildElemFirst(SPH_CFG_NAME_FILE);
  while (child_elem != NULL) {
    found_elem = sphElemUtils(child_elem).getFileElem();
    if (found_elem != NULL) {
      // @type='in'/'out'であるかチェックする。
      const char* attr_type = found_elem->getAttrString(SPH_CFG_ATTR_TYPE);
      if (strcasecmp(type, attr_type) == 0) {
        file_elem = found_elem;
        break;
      }
    }
    child_elem = data_elem->getChildElemNext(child_elem, SPH_CFG_NAME_FILE);
  }
  return file_elem;
}



/**
 * ファイルヘッダー情報を取得する。（ファイル要素識別ラベル）
 * @param[in]      file_label          ファイル要素識別ラベル
 * @param[out]     header         ヘッダー情報クラスオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::getFileHeader(const char* file_label, sphFileHeader* header)
{
  if (file_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "getFileHeader arguments is null.");
    return false;
  }

  if (m_fileCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileCtrl is null");
    return false;
  }

  const sphElem* file_elem = NULL;
  if (isClassicConfig()) {
    file_elem = createSklInFileElem(file_label);
  }
  else {
    if (m_configElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
      return NULL;
    }

    /// SphFileList/SphFile[@label] or SphMultiFile/SphFile[@label]の検索を行う。
    sphElemUtils elemUtils(m_configElem);
    file_elem = elemUtils.getFileElemByFileLabel(file_label);
    if (file_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "label is not found.[label=%s]", file_label);
      return false;
    }
  }

  /// <SphFile>要素からファイルヘッダー情報の取得を行う。
  if (!getFileHeader(file_elem, header)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "getFileHeader fault.");
    return false;
  }

  return true;
}


/**
 * ファイルヘッダー情報を取得する。（SphFile要素）
 * @param[in]      file_elem       SphFile要素
 * @param[out]     header         ヘッダー情報クラスオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::getFileHeader(const sphElem* file_elem, sphFileHeader* header)
{

  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "getFileHeader arguments is null.");
    return false;
  }

  if (m_fileCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileCtrl is null");
    return false;
  }
  if (!file_elem->equalsElemName(SPH_CFG_NAME_FILE)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphFile>.");
    return false;
  }

  /// ファイル要素を設定する。
  if (!m_fileCtrl->initFileCtrl(file_elem)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setFileElem fault.");
    return false;
  }

  /// ヘッダー情報を取得する。
  if (!m_fileCtrl->getFileInfo(header)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "getFileHeader fault.");
    return false;
  }

  return true;
}



/**
 * ファイル識別ラベルからファイルを読込み、データクラス生成を行う。
 * @param[in]      file_label          ファイル識別ラベル
 * @param[in]      size                読込サイズ
 * @param[in]      start_idx           読込始点インデックス
 * @param[in]      gc                  ガイドセル値
 * @return     データオブジェクト
 */
sphData* sphSolverBase::loadFile(
                      const char* file_label,
                      const size_t size[3],
                      const size_t start_idx[3],
                      size_t gc)
{
  if (file_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "loadFile arguments is null.");
    return NULL;
  }

  if (m_fileCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileCtrl is null");
    return NULL;
  }
  const sphElem* file_elem = NULL;
  if (isClassicConfig()) {
    file_elem = createSklInFileElem(file_label);
  }
  else {
    if (m_configElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
      return NULL;
    }

    /// SphFileList/SphFile[@label] or SphFileList/SphMultiFile[@label]の検索を行う。
    sphElemUtils elemUtils(m_configElem);
    file_elem = elemUtils.getElemByFileLabel(file_label);
    if (file_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "label is not found.[label=%s]", file_label);
      return NULL;
    }
  }
  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "file_elem is null");
    return NULL;
  }

  /// 読込ファイル要素を設定する。
  if (!m_fileCtrl->initFileCtrl(file_elem)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setFileElem fault.");
    return NULL;
  }

  /// サイズ、始点インデックス、ガイドセルを設定する。
  int dims = getDims();
  m_fileCtrl->setVoxelInfo(dims, size, start_idx, gc);

  /// ファイルから読込みを行う
  sphData* dataObj = m_fileCtrl->readFile();
  if (dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "can not create Data Object by readFile.");
  }

  return dataObj;
}


/**
 * ファイル識別ラベルからファイルを読込み、データオブジェクトにデータを設定する。.
 * 引数のデータオブジェクトは生成済みのデータオブジェクトを渡す。
 * @param[in]      file_label          ファイル識別ラベル
 * @param[in]      dataObj             生成済みデータオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::loadFile(const char* file_label, sphData* dataObj)
{
  if (file_label == NULL || dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "loadFile arguments is null.");
    return false;
  }
  if (m_fileCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_fileCtrl is null");
    return false;
  }

  const sphElem* file_elem = NULL;
  if (isClassicConfig()) {
    file_elem = createSklInFileElem(file_label);
  }
  else {
    if (m_configElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
      return NULL;
    }
    /// SphFileList/SphFile[@label] or SphFileList/SphMultiFile[@label]の検索を行う。
    sphElemUtils elemUtils(m_configElem);
    file_elem = elemUtils.getElemByFileLabel(file_label);
    if (file_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "label is not found.[label=%s]", file_label);
      return false;
    }
  }

  register int n;
  // ファイル入出力モードの取得
  SphCfgFileIOMode mode = sphElemUtils(file_elem).getFileIoMode();
  if (mode == CFG_FILE_UNKNOWN) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CONFIG_INVALID_ATTR_VALUE, "invalid mode attribute.");
    return false;
  }
  else if (mode == CFG_FILE_DEFAULT) {
    // 分割入力をデフォルトとする。
    mode = CFG_FILE_MULTI;
  }
  bool ret = false;

  // データクラスにパラレルコントローラを持っていないかチェックする。
  if (dataObj->getParaController() == NULL) {
    // ソルバクラスのパラレルコントローラから始点インデックスを取得する。
    sphParaController *para_ctrl = this->getParaController();
    size_t start_idx[3] = {0, 0, 0};
    para_ctrl->getVoxelStartIndex(start_idx);
    const size_t* voxel_size = dataObj->getArraySize();
    sphData* read_data = loadFile(file_label, voxel_size, start_idx, dataObj->getGc());
    if (read_data == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "loadFile fault");
      return false;
    }
    if (!dataObj->copyArray(read_data)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "copyArray fault.");
      return false;
    }
    delete read_data;
  }
  // 部分領域管理データクラスであるかチェックする。
  else if (dataObj->isVoxContainer() && mode == CFG_FILE_MULTI) {
    // 部分領域管理データクラスで分割入力の場合、サブブロック単位で読込を行う。
    int vox_num = dataObj->getNumberOfVoxArray();
    for (n=0; n<vox_num; n++) {
      sphData *sub_data = dataObj->getVoxArray(n);
      // loadFile
      if (!loadFile(file_elem, sub_data)) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "loadFile fault [subblock id = %d].", n);
        return false;
      }
    }
  }
  else {
    // ファイルから読込を行う。
    if (!loadFile(file_elem, dataObj)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "loadFile fault.");
      return false;
    }
  }

  return true;
}

/**
 * ファイル要素、データオブジェクト要素からデータクラス生成を行う。
 * @param[in]      file_elem       SphFile/SphMultiFile要素
 * @param[in]      dataObj         データオブジェクト
 * @return     データオブジェクト
 */
bool sphSolverBase::loadFile(const sphElem* file_elem, sphData* dataObj)
{
  if (file_elem == NULL || dataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "loadFile arguments is null.");
    return false;
  }

  /// ファイル要素名チェック
  if (!file_elem->equalsElemName(SPH_CFG_NAME_FILE)
      && !file_elem->equalsElemName(SPH_CFG_NAME_MULTIFILE)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphFile> or <SphMultiFile>.");
    return false;
  }

  // データクラスからパラレルコントローラの取得を行う。
  const sphParaController *para_ctrl = dataObj->getParaController();
  if (para_ctrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "element is not <SphFile> or <SphMultiFile>.");
    return false;
  }

  register int i;
  /// 読込ファイル要素を設定する。
  if (!m_fileCtrl->initFileCtrl(file_elem)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initFileCtrl fault.");
    return false;
  }

  /// データオブジェクトから読込サイズを取得する。
  const size_t* voxel_size = dataObj->getArraySize();
  size_t dims = dataObj->getDims();
  size_t gc = dataObj->getGc();

  /// パラレルコントローラから始点インデックスを取得する。
  size_t voxel_sidx[3] = {0, 0, 0};
  para_ctrl->getVoxelStartIndex(voxel_sidx);

  /// サイズ、始点インデックス、ガイドセルを設定する。
  m_fileCtrl->setVoxelInfo(dims, voxel_size, voxel_sidx, gc);

  /// 部分領域管理データクラスの場合、プロセス領域のデータクラスを生成する。
  sphData *procDataObj = NULL;
  // 部分領域管理データクラスであるかチェックする。
  if (dataObj->isVoxContainer()) {
    // プロセス領域のデータクラスを生成する。
    procDataObj = dataObj->createVoxelArray();      // gc = dataObjと同じ, データコピーなし

    /// 読込結果を格納するデータオブジェクトを設定する。
    m_fileCtrl->setData(procDataObj);
  }
  else {
    /// 読込結果を格納するデータオブジェクトを設定する。
    m_fileCtrl->setData(dataObj);
  }

  // 入力ファイル名フォーマット用のノード番号、ステップを設定する。
  m_fileCtrl->setFileFormatSteps(this->getStepTime()->getCurrentStep());
  m_fileCtrl->setFileFormatRankId(para_ctrl->getMyID());

  // ファイルの始点インデックスを設定する。
  // 始点インデックスに@assign属性が設定されている場合は、m_assignedStartIdxを設定する。
  SphCfgAssignType assign = sphElemUtils(file_elem).getStartIndexAssignType();
  if (assign == CFG_ASSIGN_BLOCK) {
    m_fileCtrl->setAssignedStartIndex(voxel_sidx);
  }
  else if (assign == CFG_ASSIGN_PROCESS) {
    size_t proc_sidx[3] = {0, 0, 0};
    this->getParaController()->getVoxelStartIndex(proc_sidx);
    m_fileCtrl->setAssignedStartIndex(proc_sidx);
  }
  else if (assign == CFG_ASSIGN_WHOLE) {
    const sphSvParaController *solv_para = dynamic_cast<const sphSvParaController*>(this->getParaController());
    m_fileCtrl->setAssignedStartIndex(solv_para->getReferIndex());
  }

  /// ファイルから読込みを行う
  sphData* fileDataObj = m_fileCtrl->readFile();
  if (fileDataObj == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "Data Object of readFile is null.");
    if (procDataObj != NULL) delete procDataObj;
    return false;
  }

  // ボクセルデータクラスの場合は終了する。
  if (procDataObj == NULL) {
    return true;
  }

  // 部分領域管理データクラスの場合、読込プロセス領域のデータをサブブロック領域にコピーする。
  dataObj->copyVoxelArray(fileDataObj, fileDataObj->getGc(), dataObj->getGc());
  delete procDataObj;

  return true;

}



/**
 * ガイドセル出力フラグを取得する。。
 * @param[in]      file_elem       SphFile/SphMultiFile要素
 * @return       ガイドセル出力フラグ（true:データオブジェクトのガイドセルを出力する/false:ガイドセル＝０で出力する。)
 *
 */
bool sphSolverBase::getGcOutputFlag(const sphElem* file_elem)
{
  bool gc_flag = false;
  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "getGcOutputFlag arguments is null");
    return gc_flag;
  }

  /// <SphFileOption>要素を取得する。
  const sphElem* option_elem = file_elem->getElem(SPH_CFG_NAME_FILEOPTION);
  if (option_elem != NULL) {
    gc_flag = option_elem->isAttrValue(SPH_CFG_ATTR_GCFLAG, false);
  }

  return gc_flag;
}



/**
 * テキストファイルオブジェクトを取得する。
 * @param[in]      file_label          ファイル要素識別ラベル
 * @param[out]     text_file           テキストファイルオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphSolverBase::getTextFile(const char* file_label, sphTextFile* text_file, bool initopen)
{
  if (file_label == NULL || text_file == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "getTextFile arguments is null.");
    return false;
  }

  const sphElem* file_elem = NULL;
  if (isClassicConfig()) {
    // Vsphereコンフィグレーション
    file_elem = createSklTextFileElem(file_label);

    /// 出力フォルダを設定する。
    const char* outputDir = ".";
    text_file->setBaseDir(outputDir);
  }
  else {
    // Sphereコンフィグレーション
    if (m_configElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
      return NULL;
    }
    /// ファイル要素の取得を行う。
    sphElemUtils elem_utils(m_configElem);
    file_elem = elem_utils.getElemByFileLabel(file_label);
    if (file_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "SphFile[@label=%s]", file_label);
      return false;
    }

    /// 出力フォルダを設定する。
    const char* outputDir = elem_utils.getOutputBaseDir();
    text_file->setBaseDir(outputDir);
  }
  if (file_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "file_elem is null.");
    return false;
  }

  /// 読込ファイル名書式のステップ数を設定する。
  const sphStepTime* steptime = getStepTime();
  text_file->setStepTime(steptime);

  /// プロセスコミュニケータ
  text_file->setProcGrp(m_solverComm);

  /// ファイル要素を設定する。
  text_file->setConfigElem(file_elem);

  // ファイルをオープンする。
  if (initopen) {
    return text_file->openFile(false);
  }

  return true;
}



/**
 * SklSolverConfigからファイル入力要素(sphElem)を作成する。
 * @param[in]      file_label          ファイル要素識別ラベル
 * @return    ファイル入力要素(sphElem)
 */
const sphElem* sphSolverBase::createSklInFileElem(const char* file_label)
{
  if (file_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "file_label is null.");
    return NULL;
  }
  if (m_sklconfig == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_sklconfig is null.");
    return NULL;
  }

  // InputData要素からファイル識別ラベル(attr)を検索する。
  const SklCfg::SklCfgInFile* cfg = m_sklconfig->GetInFileFirst();
  while(cfg){
    const char *tmp_attr = cfg->GetAttr();
    const char *format = cfg->GetFormat();
    const char *fname = NULL;
    if ( cfg->GetData(&tmp_attr, &format, &fname) ){
      if( !strcasecmp(tmp_attr, file_label) ) {
        // sphElemを作成する。
        sphElem* elem = new sphElem(SPH_CFG_NAME_FILE);
        elem->addAttribute(SPH_CFG_ATTR_LABEL, tmp_attr);
        elem->addAttribute(SPH_CFG_ATTR_FORMAT, format);
        elem->addAttribute(SPH_CFG_ATTR_FILENAME, fname);
        elem->addAttribute(SPH_CFG_ATTR_TYPE, "in");
        bool mio = cfg->IsMultiInput();
        if (mio) {
          elem->addAttribute(SPH_CFG_ATTR_MODE, SPH_CFG_FILEIO_MULTI);
        }
        else {
          elem->addAttribute(SPH_CFG_ATTR_MODE, SPH_CFG_FILEIO_WHOLE);
        }
        return elem;
      }
    }
    cfg = m_sklconfig->GetInFileNext(cfg);
  }

  return NULL;

}


/**
 * SklSolverConfigからファイル出力要素(sphElem)を作成する。
 * @param[in]      file_label          ファイル要素識別ラベル
 * @return    ファイル入力要素(sphElem)
 */
const sphElem* sphSolverBase::createSklOutFileElem(const char* file_label)
{
  if (file_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "file_label is null.");
    return NULL;
  }
  if (m_sklconfig == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_sklconfig is null.");
    return NULL;
  }

  // OutputData要素からファイル識別ラベル(attr)を検索する。
  const SklCfg::SklCfgOutFile* cfg = m_sklconfig->GetOutFileFirst();
  while(cfg){
    const char *tmp_attr = cfg->GetAttr();
    const char *format = cfg->GetFormat();
    const char *prefix = cfg->GetBaseName();
    if( !strcasecmp(tmp_attr, file_label) ) {
      // sphElemを作成する。
      sphElem* elem = new sphElem(SPH_CFG_NAME_FILE);
      elem->addAttribute(SPH_CFG_ATTR_LABEL, tmp_attr);
      elem->addAttribute(SPH_CFG_ATTR_FORMAT, format);
      elem->addAttribute(SPH_CFG_ATTR_PREFIX, prefix);
      elem->addAttribute(SPH_CFG_ATTR_TYPE, "out");
      // multi属性
      SphCfgFileIOMode mio = cfg->GetMultiMode();
      if (mio == CFG_FILE_DEFAULT) {
        elem->addAttribute(SPH_CFG_ATTR_MODE, SPH_CFG_FILEIO_MULTI);
      }
      else if ( mio == CFG_FILE_MULTI ) {
        elem->addAttribute(SPH_CFG_ATTR_MODE, SPH_CFG_FILEIO_MULTI);
      }
      else if ( mio == CFG_FILE_WHOLE ) {
        elem->addAttribute(SPH_CFG_ATTR_MODE, SPH_CFG_FILEIO_WHOLE);
      }
      else if ( mio == CFG_FILE_PROCESS ) {
        elem->addAttribute(SPH_CFG_ATTR_MODE, SPH_CFG_FILEIO_PROCESS);
      }
      else {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "'multi' attribute invalid.");
        return NULL;
      }

      // interval属性
      int interval = cfg->GetInterval();
      if (interval != 0) {
        // SphFileOption要素の作成
        sphElem* opt_elem = new sphElem(SPH_CFG_NAME_FILEOPTION);
        std::ostringstream stream;
        stream << interval;
        opt_elem->addAttribute(SPH_CFG_ATTR_INTERVAL, stream.str());
        elem->addChildElem(opt_elem);
      }
      return elem;
    }
    cfg = m_sklconfig->GetOutFileNext(cfg);
  }

  return NULL;

}



/**
 * SklSolverConfigのHistory要素からファイル出力要素(sphElem)を作成する。
 * @param[in]      file_label          ファイル要素識別ラベル
 * @return    ファイル入力要素(sphElem)
 */
const sphElem* sphSolverBase::createSklTextFileElem(const char* file_label)
{
  if (file_label == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "file_label is null.");
    return NULL;
  }
  if (m_sklconfig == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_sklconfig is null.");
    return NULL;
  }

  char file_name[MAXPATHLEN] = {0};
  int interval = 1;
  SphCfgFileIOMode mio = CFG_FILE_DEFAULT;
  {
    // OutputData要素からファイル識別ラベル(attr)+format="ascii"を検索する。
    const SklCfg::SklCfgOutFile* cfg = m_sklconfig->GetOutFileFirst();
    while(cfg){
      const char *tmp_attr = cfg->GetAttr();
      const char *format = cfg->GetFormat();
      const char *prefix = cfg->GetBaseName();
      if( !strcasecmp(tmp_attr, file_label) && !strcasecmp(format, "ascii")) {
        // 出力ファイル名
        sprintf(file_name, "%s.txt", prefix);
        // interval属性
        interval = cfg->GetInterval();
        // multi属性
        mio = cfg->GetMultiMode();
        break;
      }
      cfg = m_sklconfig->GetOutFileNext(cfg);
    }
  }

  // History要素から検索
  if (strlen(file_name) <= 0) {
    const SklCfg::SklCfgHistory* history = m_sklconfig->GetHistoryFirst();
    while( history ){
      const char *tmp_attr = history->GetAttr();
      const char *name = history->GetFileName();
      if( !strcasecmp(tmp_attr, file_label)) {
        // 出力ファイル名
        strcpy(file_name, name);
        // interval属性
        interval = history->GetInterval();
        break;
      }
      history = m_sklconfig->GetHistoryNext(history);
    }
  }
  if (strlen(file_name) > 0) {
    // sphElemを作成する。
    sphElem* elem = new sphElem(SPH_CFG_NAME_FILE);
    elem->addAttribute(SPH_CFG_ATTR_LABEL, file_label);
    elem->addAttribute(SPH_CFG_ATTR_FORMAT, "text");
    elem->addAttribute(SPH_CFG_ATTR_FILENAME, file_name);
    elem->addAttribute(SPH_CFG_ATTR_TYPE, "out");
    // multi属性
    // テキストファイルの場合は、一括出力をデフォルトとする。
    if (mio == CFG_FILE_DEFAULT) {
      elem->addAttribute(SPH_CFG_ATTR_MODE, SPH_CFG_FILEIO_WHOLE);
    }
    else if ( mio == CFG_FILE_MULTI ) {
      elem->addAttribute(SPH_CFG_ATTR_MODE, SPH_CFG_FILEIO_MULTI);
    }
    else if ( mio == CFG_FILE_WHOLE ) {
      elem->addAttribute(SPH_CFG_ATTR_MODE, SPH_CFG_FILEIO_WHOLE);
    }
    else if ( mio == CFG_FILE_PROCESS ) {
      elem->addAttribute(SPH_CFG_ATTR_MODE, SPH_CFG_FILEIO_PROCESS);
    }
    else {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "'multi' attribute invalid.");
      return NULL;
    }

    // interval属性
    if (interval != 0) {
      // SphFileOption要素の作成
      sphElem* opt_elem = new sphElem(SPH_CFG_NAME_FILEOPTION);
      std::ostringstream stream;
      stream << interval;
      opt_elem->addAttribute(SPH_CFG_ATTR_INTERVAL, stream.str());
      elem->addChildElem(opt_elem);
    }
    return elem;
  }

  return NULL;

}
