/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSolverManager.C
 * Sphere Solver Manager Class Source File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphSolverManager.h
 * Sphereソルバーマネージャークラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */
#include "solverbase/sphSolverManager.h"
#include "solverbase/sphSolverBase.h"

/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphSolverManager::sphSolverManager()
{
  initialize();

}


/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphSolverManager::~sphSolverManager()
{
  finalize();
}


/**
 * 初期化処理
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphSolverManager::initialize()
{
  m_mapSolver.clear();
  return true;
}


/**
 * 破棄処理
 * @param       void
 * @return    success(=true)/fault(=false)
 */
bool sphSolverManager::finalize()
{
  std::map<int, sphSolverBase*>::const_iterator it = m_mapSolver.begin();
  if(it == m_mapSolver.end()) return true;
  for( ; it!=m_mapSolver.end(); it++) {
    sphSolverBase* solver = it->second;
    delete solver;
    solver = NULL;
  }
  m_mapSolver.clear();

  return true;
}


/**
 * ソルバ数の取得を行う。
 * @param       void
 * @return      ソルバ数
 */
int sphSolverManager::getSolverCount() const
{
  return m_mapSolver.size();
}


/**
 * ソルバの取得を行う。（インデックス）
 * @param       インデックス
 * @return      ソルバオブジェクト
 */
sphSolverBase* sphSolverManager::getSolver(int id) const
{
  if (id < 0 || id >= m_mapSolver.size()) return NULL;

  std::map<int, sphSolverBase*>::const_iterator it = m_mapSolver.begin();
  register int i;
  for (i=0; i<id; i++, it++);
  return it->second;
}


/**
 * ソルバの取得を行う。（ソルバ識別ID）
 * @param       ソルバ識別ID
 * @return      ソルバオブジェクト
 */
sphSolverBase* sphSolverManager::getSolverById(int solverId) const
{
  std::map<int, sphSolverBase*>::const_iterator it = m_mapSolver.find(solverId);
  if(it == m_mapSolver.end()) return NULL;
  return it->second;
}


/**
 * ソルバの取得を行う。（ソルバ識別文字列）
 * @param       ソルバ識別文字列
 * @return      ソルバオブジェクト
 */
sphSolverBase* sphSolverManager::getSolverByLabel(const char* label) const
{
  std::map<int, sphSolverBase*>::const_iterator it = m_mapSolver.begin();
  if(it == m_mapSolver.end()) return NULL;

  for( ; it!=m_mapSolver.end(); it++) {
    sphSolverBase* solver = it->second;
    std::string solver_label = solver->getSphereLabel();
    if (solver_label == label) {
      return it->second;
    }
  }

  return NULL;
}


/**
 * ソルバの追加を行う。
 * @param       ソルバオブジェクト
 * @return    success(=true)/fault(=false)
 */
bool sphSolverManager::addSolver(sphSolverBase* solver)
{
  if (!m_mapSolver.insert( std::make_pair(solver->getSphereId(), solver)).second) {
     // error
     m_errHandler->SPH_ERROR_INFO(SPH_ERR_STL, "can not insert m_mapSolver.");
     delete solver;
     solver = NULL;
     return false;
  }

  return true;
}


