/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSolverSv.C
 * Sphere Simple Voxel Solver super Class Spurce File
 * @author     hira
 * @date       2009/10/01
 *
 * \~japanese
 * @file sphSolverSv.C
 * Sphere構造格子ソルバー基底クラスソースファイル
 * @author     hira
 * @date       2009/10/01
 */
#include "solverbase/sphSolverSv.h"
#include "config/sphElemUtils.h"
#include "data/sphDataUtils.h"


/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphSolverSv::sphSolverSv()
{
  /// 初期化を行う。
  initialize();

}


/**
 * \~english
 * Constructor
 * @param     label     solver label
 *
 * \~japanese
 * コンストラクタ
 * @param     label     ソルバーラベル
 */
sphSolverSv::sphSolverSv(const char* solv_label) : sphSolverBase(solv_label)
{
  /// 初期化を行う。
  initialize();
}


/**
 * \~english
 * Constructor
 * @param     label     solver label
 * @param     id     solver id
 *
 * \~japanese
 * コンストラクタ
 * @param     label     ソルバーラベル
 * @param     id     ソルバーID
 */
sphSolverSv::sphSolverSv(const char* solv_label, int solv_id) : sphSolverBase(solv_label, solv_id)
{
  /// 初期化を行う。
  initialize();

}

/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphSolverSv::~sphSolverSv()
{


}


/**
 * 初期化を行う。
 * @param     void
 * @return    success(=true)/fault(=false)
 */
bool sphSolverSv::initialize()
{
  m_solverDims = 3;

  return true;
}


/**
 * \~english
 * initialize Solver.
   * @param               steptime         step, time class object
 * @return              -1:error, 0:forced terminated, 1:normality
 *
 * \~japanese
 * ソルバーの初期化を行う.
   * @param               steptime          実行ステップ・時刻クラス
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int
sphSolverSv::initializeSolver(sphStepTime* steptime)
{
  int ret = SPHERE_SUCCESS;

  // コンフィグレーションがSklSolverConfigであるので、終了する。
  if (!isClassicConfig()) {
    if (m_configElem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_configElem is null");
      return SPHERE_TERMINATE;
    }
  }

  /// <SphSolver>要素から基本動作に必要な情報を取得する。
  if ( (ret = setupSolver()) != SPHERE_SUCCESS ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setupSolver fault");
    return ret;
  }

  /// sphParaControllerを作成する。
  if ( (ret = initializeVoxel()) != SPHERE_SUCCESS ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeVoxel fault");
    return ret;
  }

  /// <SphDataList>要素から情報収集、データ生成を行う。
  if ( (ret = setupDataList()) != SPHERE_SUCCESS ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setupDataList fault");
    return ret;
  }

  return ret;
}


/**
 * <SphDataList>要素から情報収集、データ生成を行う。
 * @param               void
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverSv::setupDataList()
{
  // コンフィグレーションがSklSolverConfigであるので、終了する。
  if (isClassicConfig()) {
    return SPHERE_SUCCESS;
  }

  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
    return SPHERE_TERMINATE;
  }

  // 警告メッセージを無効にする。
  m_errHandler->blockWarning();

  /// SphDataList/SphDataObj要素の取得を行う。
  const sphElem *data_elem = m_configElem->getChildElemFirst(2, SPH_CFG_NAME_DATALIST, SPH_CFG_NAME_DATAOBJ);
  while (data_elem != NULL) {
    /// instance属性が"instanceであれば、データオブジェクトを生成する。
    const char* instance = data_elem->getAttrString(SPH_CFG_ATTR_INSTANCE);
    if (instance == NULL || strcasecmp(instance, "initialize") == 0) {
      if (allocateDataObj(data_elem) == NULL) {
        // 警告メッセージを元に戻す。
        m_errHandler->releaseWarning();
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
        return SPHERE_ERROR;
      }
    }
    data_elem = m_configElem->getChildElemNext(data_elem, 2, SPH_CFG_NAME_DATALIST, SPH_CFG_NAME_DATAOBJ);
  }

  // 警告メッセージを元に戻す。
  m_errHandler->releaseWarning();

  return SPHERE_SUCCESS;

}


/**
 * <SphSolver>要素から基本情報を取得する。
 * @param               void
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverSv::setupSolver()
{
  register int i;
  int ret = SPHERE_SUCCESS;

  if (this->isClassicConfig()) {
    // Vsphereコンフィグレーション
    bool check_mode = m_sklconfig->IsCheckMode();
    if (check_mode) {
      m_solverMode = std::string(SPH_CFG_SOLVER_CHECKMODE);
    }
  }
  else {
    // Sphereコンフィグレーション
    /// ソルバのモードの取得を行う。
    if (m_configElem->existAttribute(SPH_CFG_ATTR_MODE)) {
      const char* mode = m_configElem->getAttrString(SPH_CFG_ATTR_MODE);
      if ( mode != NULL ) m_solverMode = std::string(mode);
    }

    /// 次元数の取得を行う。
    m_solverDims = m_configElem->getAttrInt(SPH_CFG_ATTR_DIMS);
    if (m_solverDims == 0) m_solverDims = 3;      // デフォルト＝３
    if (m_solverDims < 0 || m_solverDims > 3) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid dims(=%d)", m_solverDims);
      return SPHERE_TERMINATE;
    }
  }

  return ret;

}


/**
 * sphParaControllerの生成を行う。
 * <SphSolver>要素からパラレルコントローラの生成を行う。
 * @param               void
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverSv::initializeVoxel()
{

  register int i;
  int ret = SPHERE_SUCCESS;

  if ( m_solverComm == SPL_COMM_NULL && splIsParallel() )  {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSolverSv parameters is null");
    return SPHERE_TERMINATE;
  }

  // パラレルコントローラの生成
  sphParaManager* paraMngr = getParaManager();
  if (isClassicConfig()) {

    // パラレルコントローラの生成, 追加
    if ( !initializeVoxelSklConfig() ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeVoxelFromSklConfig fault");
      ret = SPHERE_TERMINATE;
    }

  }
  else {
    // パラレルコントローラの生成, 追加
    if ( !setParaController(
                  paraMngr->createSvParaController(
                            m_configElem, m_solverComm)) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSvParaController fault");
      ret = SPHERE_TERMINATE;
    }
  }

  return ret;

}



/**
 * sphParaControllerの生成を行う。
 * <SphSolver>要素からパラレルコントローラの生成を行う。
 * @param               void
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverSv::initializeVoxelSklConfig()
{

  register int i;
  int ret = SPHERE_SUCCESS;

  if (m_sklconfig == NULL || (m_solverComm == SPL_COMM_NULL && splIsParallel()) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSolverSv parameters(member) is null");
    return SPHERE_TERMINATE;
  }

  // ソルバグループのプロセス数を取得する。
  SPL_Comm comm = m_solverComm;
  int commSize = 0;
  SPL_Comm_size(m_solverComm, &commSize);

  // パラレルコントローラの生成
  sphParaManager* paraMngr = getParaManager();

  // ボクセルサイズ、原点座標、ピッチを取得する。
  size_t voxel_size[3] = {0, 0, 0};
  size_t voxel_start[3] = {0, 0, 0};
  double voxel_org[3] = {0, 0, 0};
  double voxel_pit[3] = {0, 0, 0};
  if (!getConfigDomainInfo(voxel_size, voxel_org, voxel_pit)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "can not get DomainInfo");
    return SPHERE_TERMINATE;
  }

  // ParallelTemplate要素の取得
  bool isTemp = existsConfigParallelTemplate();
  bool isEp = existsConfigEqualPartitioning();
  int sub_num = getConfigNumberOfSubblock();
  bool isMcp = existsConfigMultiplyPartitioning();
  bool mbx_grouping = isConfigMbxGrouping();

  // EV分割、MV分割のコンフィグレーション定義が、true^true, false^falseの場合は定義エラーとする。
  if (!(isEp ^ isMcp)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "duplicate configuration EV^MB");
    return SPHERE_TERMINATE;
  }

  if (isMcp && m_sklconfig->GetTop(SklCfg::MBOXTBL) == NULL) {   // not found MBXTBL
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found MBXTBL in SklConfig");
    ret = SPHERE_TERMINATE;
  }
  else if (isMcp && m_sklconfig->GetTop(SklCfg::MBOXTBL) != NULL) {   // MB
    // MB分割パラレルコントローラの生成を行う。
    if (!initializeVoxel(voxel_size, voxel_org, voxel_pit, voxel_start)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid initializeVoxelMb()");
      ret = SPHERE_TERMINATE;
    }
  }
  else {    // EV
    // コンフィグレーションからI,J,K分割数を取得する。
    unsigned int voxel_div[3] = {0, 0, 0};
    if (!getConfigVoxelDivision(voxel_div)) {
      voxel_div[0]=voxel_div[1]=voxel_div[2]=0;
    }

    // パラレルコントローラの生成, 追加
    // voxel_divが{0,0,0}ならば、プロセス数から自動分割を行う。
    if ( !setParaController(
                  paraMngr->createSvParaController(
                      voxel_size, voxel_start,
                      voxel_org, voxel_pit,
                      voxel_div,
                      m_solverComm,
                      sub_num)) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSvParaController fault");
      ret = SPHERE_TERMINATE;
    }
  }

  return ret;

}


/**
 * sphParaControllerの生成を行う。(MB分割)
 * 分割情報からパラレルコントローラの生成を行う。
 * @param               voxelsize       ボクセルサイズ
 * @param               voxel_origin    ボクセル原点座標（NULL:SphDoaminInfoから取得する）
 * @param               voxel_pitch      ボクセルピッチ（NULL:SphDoaminInfoから取得する）
 * @param               voxel_start     ボクセル始点インデックス
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverSv::initializeVoxel(
                    const size_t voxel_size[3],
                    const double voxel_origin[3],
                    const double voxel_pitch[3],
                    const size_t voxel_start[3])
{

  register int i;
  int ret = SPHERE_SUCCESS;

  if (m_solverComm == SPL_COMM_NULL && splIsParallel()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSolverSv parameters(member) is null");
    return SPHERE_TERMINATE;
  }

  // ソルバグループのプロセス数を取得する。
  SPL_Comm comm = m_solverComm;
  int commSize = 0;
  SPL_Comm_size(m_solverComm, &commSize);

  // パラレルコントローラの生成
  sphParaManager* paraMngr = getParaManager();

  size_t startidx[3] = {0, 0, 0};
  if (voxel_start != NULL) {
    startidx[0]=voxel_start[0];startidx[1]=voxel_start[1];startidx[2]=voxel_start[2];
  }
  // ParallelTemplate要素の取得
  bool isTemp = existsConfigParallelTemplate();
  bool isEp = existsConfigEqualPartitioning();
  int sub_num = getConfigNumberOfSubblock();
  bool isMcp = existsConfigMultiplyPartitioning();
  bool mbx_grouping = isConfigMbxGrouping();

  // EV分割、MV分割のコンフィグレーション定義が、true^true, false^falseの場合は定義エラーとする。
  if (!(isEp ^ isMcp)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "duplicate configuration EV^MB");
    return SPHERE_TERMINATE;
  }

  // コンフィグレーションからI,J,K分割数の取得を行う。
  unsigned int voxel_div[3] = {0, 0, 0};
  if (!getConfigVoxelDivision(voxel_div)) {
    voxel_div[0]=voxel_div[1]=voxel_div[2]=0;
  }

  if (isEp) {   // EV分割
    // パラレルコントローラの生成, 追加
    // voxel_divが{0,0,0}ならば、プロセス数から自動分割を行う。
    if ( !setParaController(
                  paraMngr->createSvParaController(
                      voxel_size, startidx,
                      voxel_origin, voxel_pitch,
                      voxel_div,
                      m_solverComm,
                      sub_num)) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSvParaController fault");
      ret = SPHERE_TERMINATE;
    }
  }
  else if (isClassicConfig()) {      // V-Sphere Configuration
    if (m_sklconfig->GetTop(SklCfg::MBOXTBL) == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "SklCfg::MBOXTBL is null");
      return SPHERE_TERMINATE;
    }

    // MultiBoxDivTable/Elem name="node"の数の取得
    const SklCfg::CfgElem* elem = m_sklconfig->GetElemFirst(SklCfg::MBOXTBL, "node");
    unsigned int elem_num = 0;
    while (elem) {
      elem_num++;
      elem = m_sklconfig->GetElemNext(SklCfg::MBOXTBL, elem, "node");
    }
    size_t (*node_size)[3] = new size_t[elem_num][3];
    size_t (*node_start)[3] = new size_t[elem_num][3];
    unsigned int *node_group = new unsigned int[elem_num];
    memset(node_size, 0x00, sizeof(size_t)*elem_num*3);
    memset(node_start, 0x00, sizeof(size_t)*elem_num*3);
    memset(node_group, 0x00, sizeof(unsigned int)*elem_num);

    // サブブロック数
    int sub_count = (elem_num+(commSize-1))/commSize;

    // MBXTBLからノードリストの取得を行う。
    elem = m_sklconfig->GetElemFirst(SklCfg::MBOXTBL, "node");
    const SklCfg::CfgParam* param = NULL;
    while (elem) {
      int min[3], max[3], group;
      min[0] = min[1] = min[2] = max[0] = max[1] = max[2] = 0;

      // ID
      int id = elem->GetID();
      if (id >= elem_num) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ATTRIBUTE, "invalid id value in node element.");
        ret = SPHERE_TERMINATE;
        break;
      }
      param = elem->GetParamFirst("xmin");
      if (param) param->GetData(&min[0]);
      param = elem->GetParamFirst("ymin");
      if (param) param->GetData(&min[1]);
      param = elem->GetParamFirst("zmin");
      if (param) param->GetData(&min[2]);
      param = elem->GetParamFirst("xmax");
      if (param) param->GetData(&max[0]);
      param = elem->GetParamFirst("ymax");
      if (param) param->GetData(&max[1]);
      param = elem->GetParamFirst("zmax");
      if (param) param->GetData(&max[2]);
      param = elem->GetParamFirst("group");
      if (commSize > 1) {
        if (param) {
          param->GetData(&group);
          node_group[id] = group;
        }
        else {
          group = id / sub_count;
          node_group[id] = group;
        }
      }

      for (i=0; i<3; i++) {
        node_start[id][i] = min[i];
        node_size[id][i] = max[i] - min[i] + 1;
      }

      elem = m_sklconfig->GetElemNext(SklCfg::MBOXTBL, elem, "node");
    }
    if (ret == SPHERE_SUCCESS) {
      // パラレルコントローラの生成, 追加
      if ( !setParaController(
                    paraMngr->createSvParaController(
                        elem_num,
                        voxel_size, startidx,
                        node_size, node_start,
                        voxel_origin, voxel_pitch,
                        m_solverComm,
                        (mbx_grouping)?node_group:NULL)) ) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSvParaController fault");
        ret = SPHERE_TERMINATE;
      }
    }
    delete [] node_size; delete [] node_start; delete [] node_group;
  }
  else {      // Sphere Configuration
    const sphElem* nodelist_elem = m_configElem->getElem(2, SPH_CFG_NAME_DOMAININFO, SPH_CFG_NAME_NODELIST);
    if (nodelist_elem == NULL) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "not found <SphNodeList>");
      return SPHERE_TERMINATE;
    }
      // パラレルコントローラの生成を行う。
    // パラレルコントローラの生成, 追加
    if ( !setParaController(
                  paraMngr->createSvParaController(
                                  m_solverDims,
                                  voxel_size,
                                  startidx,
                                  nodelist_elem,
                                  voxel_origin,
                                  voxel_pitch,
                                  m_solverComm,
                                  mbx_grouping)) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSvParaController fault");
      ret = SPHERE_TERMINATE;
    }
  }
  return ret;

}


/**
 * sphParaControllerの生成を行う。
 * @param               nodes_elem       <SphNodeList>要素
 * @param               voxel_size      ボクセルサイズ（NULL:SphDoaminInfoから取得する）
 * @param               voxel_start     ボクセル始点インデックス（NULL:SphDoaminInfoから取得する）
 * @param               voxel_origin    ボクセル原点座標（NULL:SphDoaminInfoから取得する）
 * @param               voxel_pitch      ボクセルピッチ（NULL:SphDoaminInfoから取得する）
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverSv::initializeVoxel(
                        const sphElem* nodelist_elem,
                        const size_t voxel_size[3],
                        const size_t voxel_start[3],
                        const double voxel_origin[3],
                        const double voxel_pitch[3])
{
  register int i;
  int ret = SPHERE_SUCCESS;

  if (nodelist_elem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "initializeVoxel arguments is null.");
    return SPHERE_TERMINATE;
  }

  if (m_configElem == NULL || (m_solverComm == SPL_COMM_NULL  && splIsParallel()) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSolverSv parameters(member) is null");
    return SPHERE_TERMINATE;
  }

  sphElemUtils elemUtils(m_configElem);
  /// <SphDomainInfo>要素からサイズ、原点座標、ピッチを取得する。
  size_t tmp_size[3] = {1, 1, 1};
  size_t tmp_start[3] = {1, 1, 1};
  double tmp_origin[3] = {0.0, 0.0, 0.0};
  double tmp_pitch[3] = {0.0, 0.0, 0.0};
  if (voxel_origin == NULL || voxel_pitch == NULL) {
    if (!elemUtils.getDomainInfo(m_solverDims, tmp_size, tmp_origin, tmp_pitch, tmp_start)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid <SphDomainInfo>.");
      ret = SPHERE_TERMINATE;
    }
  }
  else {
    for (i=0; i<m_solverDims; i++) {
      tmp_origin[i] = voxel_origin[i];
      tmp_pitch[i] = voxel_pitch[i];
    }
  }

  /// <SphNode>要素数を取得する。
  int proc_num = nodelist_elem->countNumOfElem(SPH_CFG_NAME_NODE);

  /// 設定済みのソルバグループのプロセス数を取得する。
  SPL_Comm comm = m_solverComm;
  int commSize = 0;
  SPL_Comm_size(m_solverComm, &commSize);
  if (proc_num != commSize) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "solver process is not equals group process, solver node=%d, group process=%d", proc_num, commSize);
    return SPHERE_TERMINATE;
  }

  if (ret == SPHERE_SUCCESS) {
    /// パラレルマネージャを介してパラレルコントローラの生成、初期化を行う。
    sphParaManager* paraMngr = getParaManager();
    if ( !setParaController(
                  paraMngr->createSvParaController(
                                m_solverDims,
                                tmp_size,
                                tmp_start,
                                nodelist_elem,
                                tmp_origin,
                                tmp_pitch,
                                m_solverComm)) ) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSvParaController fault");
      ret = SPHERE_TERMINATE;
    }
  }

  return ret;
}


/**
 * sphParaControllerの生成を行う。(分割情報)
 * 分割情報からパラレルコントローラの生成を行う。
 * @param               voxelsize       ボクセルサイズ
 * @param               division        分割数
 * @param               voxel_origin    ボクセル原点座標（NULL:SphDoaminInfoから取得する）
 * @param               voxel_pitch      ボクセルピッチ（NULL:SphDoaminInfoから取得する）
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverSv::initializeVoxel(
                    const size_t voxelsize[3],
                    const unsigned int division[3],
                    const double voxel_origin[3],
                    const double voxel_pitch[3])
{

  register int i;
  int ret = SPHERE_SUCCESS;

  if (voxelsize == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "initializeVoxel arguments is null.");
    return SPHERE_TERMINATE;
  }

  if (m_solverComm == SPL_COMM_NULL  && splIsParallel()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSolverSv parameters(member) is null");
    return SPHERE_TERMINATE;
  }

  // ParallelTemplate要素の取得
  bool isTemp = existsConfigParallelTemplate();
  bool isEp = existsConfigEqualPartitioning();
  int sub_num = getConfigNumberOfSubblock();
  bool isMcp = existsConfigMultiplyPartitioning();
  bool mbx_grouping = isConfigMbxGrouping();

  // EV分割、MV分割のコンフィグレーション定義が、true^true, false^falseの場合は定義エラーとする。
  if (!(isEp ^ isMcp)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "duplicate configuration EV^MB");
    return SPHERE_TERMINATE;
  }

  if (isMcp) {
    size_t voxel_start[3] = {0, 0, 0};
    // MB分割パラレルコントローラの生成を行う。
    if (!initializeVoxel(voxelsize, voxel_origin, voxel_pitch, voxel_start)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid initializeVoxelMb()");
      ret = SPHERE_TERMINATE;
    }
    return SPHERE_SUCCESS;
  }
  else if (!isEp) {
    // EV分割ではないのでエラーとする。
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "Configuration is not EqualPartitioning.");
    return SPHERE_TERMINATE;
  }

  /// 分割数を取得する。
  int node_count = 1;
  if (division != NULL) {
    for (i=0; i<3; i++) node_count *= division[i];
  }
  else {
    node_count = 0;
  }
  if (node_count == 0) {
    // 実行プロセス数からパラレルコントローラの生成を行う。(proc_num=0によりプロセス数自動取得）
    if (initializeVoxel(voxelsize, 0, voxel_origin, voxel_pitch) != SPHERE_SUCCESS) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "Division is zero[%ld,%ld,%ld].", division[0], division[1], division[2]);
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeVoxel(proc_num) failed.");
      return SPHERE_TERMINATE;
    }
    return SPHERE_SUCCESS;
  }

  /// <SphDomainInfo>要素からサイズ、原点座標、ピッチを取得する。
  size_t tmp_size[3] = {1, 1, 1};
  double tmp_origin[3] = {0.0, 0.0, 0.0};
  double tmp_pitch[3] = {0.0, 0.0, 0.0};
  if (voxel_origin == NULL || voxel_pitch == NULL) {
    // コンフィグレーションから原点座標、ピッチを取得する。
    if (!getConfigDomainInfo(tmp_size, tmp_origin, tmp_pitch)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "can not get DomainInfo");
      return SPHERE_TERMINATE;
    }
  }
  else {
    for (i=0; i<m_solverDims; i++) {
      tmp_origin[i] = voxel_origin[i];
      tmp_pitch[i] = voxel_pitch[i];
    }
  }

  /// パラレルマネージャを介してパラレルコントローラの生成、初期化を行う。
  sphParaManager* paraMngr = getParaManager();
  size_t startindex[3] = {0, 0, 0};
  unsigned int tmp_div[3] = {division[0], division[1], division[2]};
  if ( !setParaController(
                  paraMngr->createSvParaController(
                              voxelsize,
                              startindex,
                              tmp_origin,
                              tmp_pitch,
                              tmp_div,
                              m_solverComm,
                              sub_num)) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSvParaController fault");
    ret = SPHERE_TERMINATE;
  }

  return ret;

}


/**
 * sphParaControllerの生成を行う。(分割プロセス数)
 *
 * 分割プロセス数からパラレルコントローラの生成を行う。
 * @param               voxelsize       ボクセルサイズ
 * @param               proc_num        分割プロセス数（0：ソルバプロセス数を設定する）
 * @param               voxel_origin    ボクセル原点座標（NULL:SphDoaminInfoから取得する）
 * @param               voxel_pitch      ボクセルピッチ（NULL:SphDoaminInfoから取得する）
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverSv::initializeVoxel(
                    const size_t voxelsize[3],
                    int proc_num,
                    const double voxel_origin[3],
                    const double voxel_pitch[3])
{

  register int i;
  int ret = SPHERE_SUCCESS;

  if (voxelsize == NULL ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "initializeVoxel arguments is null.");
    return SPHERE_TERMINATE;
  }

  if (m_solverComm == SPL_COMM_NULL && splIsParallel()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSolverSv parameters(member) is null");
    return SPHERE_TERMINATE;
  }

  // ParallelTemplate要素の取得
  bool isTemp = existsConfigParallelTemplate();
  bool isEp = existsConfigEqualPartitioning();
  int sub_num = getConfigNumberOfSubblock();
  bool isMcp = existsConfigMultiplyPartitioning();
  bool mbx_grouping = isConfigMbxGrouping();

  // EV分割、MV分割のコンフィグレーション定義が、true^true, false^falseの場合は定義エラーとする。
  if (!(isEp ^ isMcp)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "duplicate configuration EV^MB");
    return SPHERE_TERMINATE;
  }

  // EV分割用のメソッドであるので、EqualPartitioningの記述が必要である。
  if (!isEp) {
    // EV分割ではないのでエラーとする。
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "Configuration is not EqualPartitioning.");
    return SPHERE_TERMINATE;
  }

  if (proc_num <= 0) {
    SPL_Comm_size(m_solverComm, &proc_num);
  }

  /// <SphDomainInfo>要素からサイズ、原点座標、ピッチを取得する。
  size_t tmp_size[3] = {1, 1, 1};
  double tmp_origin[3] = {0.0, 0.0, 0.0};
  double tmp_pitch[3] = {0.0, 0.0, 0.0};
  if (voxel_origin == NULL || voxel_pitch == NULL) {
    // コンフィグレーションから原点座標、ピッチを取得する。
    if (!getConfigDomainInfo(tmp_size, tmp_origin, tmp_pitch)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "can not get DomainInfo");
      return SPHERE_TERMINATE;
    }
  }
  else {
    for (i=0; i<m_solverDims; i++) {
      tmp_origin[i] = voxel_origin[i];
      tmp_pitch[i] = voxel_pitch[i];
    }
  }

  /// パラレルマネージャを介してパラレルコントローラの生成、初期化を行う。
  sphParaManager* paraMngr = getParaManager();
  size_t startindex[3] = {0, 0, 0};
  if ( !setParaController(
            paraMngr->createSvParaController(
                              voxelsize,
                              startindex,
                              tmp_origin,
                              tmp_pitch,
                              proc_num,
                              m_solverComm,
                              sub_num)) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSvParaController fault");
    ret = SPHERE_TERMINATE;
  }


  return ret;

}

/**
 * phParaControllerの生成を行う。（ボクセルサイズリスト）.
 *
 * ボクセルサイズリストからパラレルコントローラを生成する。
 * グループID(=ランク番号)テーブルが設定されている場合は、部分領域データ管理機能パラレルマネージャを生成する。
 * @param       node_num      ノード数
 * @param      wsize        ソルバ全体領域のサイズ
 * @param      wstart       ソルバ全体領域の始点インデックス
 * @param       voxelsize    ボクセルサイズリスト
 * @param       voxelstart    ボクセル始点インデックスリスト
 * @param       voxel_origin    原点座標
 * @param       voxel_pitch     ピッチ
 * @param       groupid_table    グループID(=ランク番号)テーブル
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverSv::initializeVoxel(
                          unsigned int node_num,
                          const size_t wsize[3],
                          const size_t wstart[3],
                          const size_t (*voxelsize)[3],
                          const size_t (*voxelstart)[3],
                          const double voxel_origin[3],
                          const double voxel_pitch[3],
                          const unsigned int *groupid_table)
{

  register int i;
  int ret = SPHERE_SUCCESS;

  if (node_num == 0 || wsize == NULL || wstart == NULL || voxelsize == NULL || voxelstart == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "initializeVoxel arguments is null.");
    return SPHERE_TERMINATE;
  }

  if (m_solverComm == SPL_COMM_NULL && splIsParallel()) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSolverSv parameters(member) is null");
    return SPHERE_TERMINATE;
  }

  // MB分割用のメソッドであるので、MultiplyConnectedPartitioningの記述が必要である。
  if (!existsConfigMultiplyPartitioning()) {
    // MB分割ではないのでエラーとする。
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "Configuration is not MultiplyConnectedPartitioning.");
    return SPHERE_TERMINATE;
  }

  /// <SphDomainInfo>要素からサイズ、原点座標、ピッチを取得する。
  size_t tmp_size[3] = {1, 1, 1};
  double tmp_origin[3] = {0.0, 0.0, 0.0};
  double tmp_pitch[3] = {0.0, 0.0, 0.0};
  if (voxel_origin == NULL || voxel_pitch == NULL) {
    // コンフィグレーションから原点座標、ピッチを取得する。
    if (!getConfigDomainInfo(tmp_size, tmp_origin, tmp_pitch)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_ELEMENT, "can not get DomainInfo");
      return SPHERE_TERMINATE;
    }
  }
  else {
    for (i=0; i<m_solverDims; i++) {
      tmp_origin[i] = voxel_origin[i];
      tmp_pitch[i] = voxel_pitch[i];
    }
  }

  /// パラレルマネージャを介してパラレルコントローラの生成、初期化を行う。
  sphParaManager* paraMngr = getParaManager();
  if ( !setParaController(
            paraMngr->createSvParaController(
                                node_num,
                                wsize,
                                wstart,
                                voxelsize,
                                voxelstart,
                                tmp_origin,
                                tmp_pitch,
                                m_solverComm,
                                groupid_table)) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createSvParaController fault");
    ret = SPHERE_TERMINATE;
  }

  return ret;

}


/**
 * パラレルコントローラーを設定する.
 * sphSvParaControllerへのアクセスを容易にする為にm_svParaCtrlへ設定を行う。
 * 主はsphParaManagerのmapに存在する。m_svParaCtrlはテンポラリ的ポインタ
 * @param               paraCtrl
 * @return    success(=true)/fault(=false)
 */
bool sphSolverSv::setParaController(sphParaController* paraCtrl)
{
  if (!sphSolverBase::setParaController(paraCtrl)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "sphSolverBase::setParaController fault");
    return false;
  }

  m_svParaCtrl = dynamic_cast<sphSvParaController*>(paraCtrl);
  if (m_svParaCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphSvParaController cast is null");
    return false;
  }

  return true;
}


/**
 * パラレルコントローラにボクセル原点座標, ピッチの設定を行う。
 * @param               org       ボクセル原点座標
 * @param               pitch       ボクセルピッチ
 * @return    success(=true)/fault(=false)
 */
bool sphSolverSv::setVoxelCoords(const double org[3], const double pitch[3])
{
  if (m_svParaCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSvParaController is null.");
    return false;
  }
  /// パラレルコントローラにボクセル原点座標、ピッチの設定を行う。
  m_svParaCtrl->setVoxelOrigin(org);
  m_svParaCtrl->setVoxelPitch(pitch);

  // TODO:カプラの再初期化を行う。


  return true;
}


/**
 * コンフィグレーションにParallelTemplate要素が存在するかチェックする。
 *
 * @return  true=ParallelTemplate要素が存在する
 */
bool sphSolverSv::existsConfigParallelTemplate()
{
  if (isClassicConfig()) {      // V-Sphere Configuration

    const SklCfg::CfgElem* elem = m_sklconfig->GetElemFirst(SklCfg::STEER, "ParallelTemplate");
    if (elem == NULL) return false;
    return true;
  }
  else {      // Sphere Configuration
    // SphEqualPartitioning要素
    const sphElem* temp_elem = m_configElem->getElem(2,
                                SPH_CFG_NAME_STEER,
                                SPH_CFG_NAME_PARALLELTEMPLATE);
    if (temp_elem == NULL) return false;
    return true;

  }
}


/**
 * コンフィグレーションにEqualPartitioning要素が存在するかチェックする。
 *
 *
 * @return  true=EqualPartitioning要素が存在する。
 */
bool sphSolverSv::existsConfigEqualPartitioning()
{
  bool existsMbxElem = false;
  if (isClassicConfig()) {      // V-Sphere Configuration
    existsMbxElem = (m_sklconfig->GetTop(SklCfg::MBOXTBL) != NULL);
  }
  else {
    const sphElem* nodelist_elem = m_configElem->getElem(2, SPH_CFG_NAME_DOMAININFO, SPH_CFG_NAME_NODELIST);
    existsMbxElem = (nodelist_elem != NULL);
  }

  // ParallelTemplate要素が存在しない場合は、MBOXTBLが存在していればMB分割とする
  if (!existsConfigParallelTemplate()) {
    return !existsMbxElem;
  }

  if (isClassicConfig()) {      // V-Sphere Configuration
    const SklCfg::CfgElem* elem = m_sklconfig
                  ->GetElemFirst(SklCfg::STEER, "ParallelTemplate")
                    ->GetElemFirst("EqualPartitioning");
    if (elem == NULL) return false;
    return true;
  }
  else {      // Sphere Configuration
    // SphEqualPartitioning要素
    const sphElem* ep_elem = m_configElem->getElem(3,
                                SPH_CFG_NAME_STEER,
                                SPH_CFG_NAME_PARALLELTEMPLATE,
                                SPH_CFG_NAME_EQUALPARTITIONING);
    if (ep_elem == NULL) return false;
    return true;
  }
}


/**
 * コンフィグレーションからEqualPartitioning[@number_of_subblock]の設定値を取得する。
 *
 * @return  "number_of_subblock"の設定値
 */
int sphSolverSv::getConfigNumberOfSubblock()
{

  if (isClassicConfig()) {      // V-Sphere Configuration
    const SklCfg::CfgElem* elem = m_sklconfig->GetElemFirst(SklCfg::STEER, "ParallelTemplate");
    if (elem == NULL) return 0;
    elem = elem->GetElemFirst("EqualPartitioning");
    if (elem == NULL) return 0;
    const SklCfg::CfgParam* param = elem->GetParamFirst(SPH_CFG_ATTR_NUMBEROFSUBBLOCK);
    if (param == NULL) return 0;
    int value = 0;
    param->GetData(&value);

    return value;
  }
  else {      // Sphere Configuration
    // SphEqualPartitioning要素
    const sphElem* ep_elem = m_configElem->getElem(3,
                                SPH_CFG_NAME_STEER,
                                SPH_CFG_NAME_PARALLELTEMPLATE,
                                SPH_CFG_NAME_EQUALPARTITIONING);
    if (ep_elem == NULL) return 0;
    int count = ep_elem->getAttrInt(SPH_CFG_ATTR_NUMBEROFSUBBLOCK);
    return count;
  }
}


/**
 * コンフィグレーションにMultiplyConnectedPartitioning要素が存在するかチェックする。
 *
 * MBOTBL, SphNodeList要素が存在していなければならない。
 * @return  true=MultiplyConnectedPartitioning要素が存在する。
 */
bool sphSolverSv::existsConfigMultiplyPartitioning()
{

  bool existsMbxElem = false;
  if (isClassicConfig()) {      // V-Sphere Configuration
    existsMbxElem = (m_sklconfig->GetTop(SklCfg::MBOXTBL) != NULL);
  }
  else {
    const sphElem* nodelist_elem = m_configElem->getElem(2, SPH_CFG_NAME_DOMAININFO, SPH_CFG_NAME_NODELIST);
    existsMbxElem = (nodelist_elem != NULL);
  }

  // ParallelTemplate要素が存在しない場合は、MBOXTBLが存在していればMB分割とする
  if (!existsConfigParallelTemplate()) {
    return existsMbxElem;
  }

  if (isClassicConfig()) {      // V-Sphere Configuration
    const SklCfg::CfgElem* elem = m_sklconfig
                          ->GetElemFirst(SklCfg::STEER, "ParallelTemplate")
                            ->GetElemFirst("MultiplyConnectedPartitioning");
    if (elem == NULL) return false;

    return existsMbxElem;
  }
  else {      // Sphere Configuration
    // SphMultiplyConnectedPartitioning要素
    const sphElem* mcp_elem = m_configElem->getElem(3,
                                SPH_CFG_NAME_STEER,
                                SPH_CFG_NAME_PARALLELTEMPLATE,
                                SPH_CFG_NAME_MULTIPLYPARTITIONING);
    if (mcp_elem == NULL) return false;
    return existsMbxElem;
  }
}


/**
 * コンフィグレーションからMultiplyConnectedPartitioning[@mbx_grouping]の設定値を取得する。
 *
 * @return  true = "mbx_grouping=yes"
 */
bool sphSolverSv::isConfigMbxGrouping()
{
  if (isClassicConfig()) {      // V-Sphere Configuration

    const SklCfg::CfgElem* elem = m_sklconfig->GetElemFirst(SklCfg::STEER, "ParallelTemplate");
    if (elem == NULL) return false;
    elem = elem->GetElemFirst("MultiplyConnectedPartitioning");
    if (elem == NULL) return false;
    const SklCfg::CfgParam* param = elem->GetParamFirst(SPH_CFG_ATTR_MBXGROUPING);
    if (param == NULL) return 0;
    const char *value;
    param->GetData(&value);
    if (strcasecmp(value, "yes") == 0) {
      return true;
    }
    return false;
  }
  else {
    // Sphere Configuration
    // SphMultiplyConnectedPartitioning要素
    const sphElem* mcp_elem = m_configElem->getElem(3,
                                SPH_CFG_NAME_STEER,
                                SPH_CFG_NAME_PARALLELTEMPLATE,
                                SPH_CFG_NAME_MULTIPLYPARTITIONING);
    if (mcp_elem == NULL) return false;
    return mcp_elem->isAttrValue(SPH_CFG_ATTR_MBXGROUPING, false);
  }

}


/**
 * コンフィグレーションからサイズ、原点座標、ピッチを取得する。
 * @param   sz      サイズ
 * @param   org     原点座標
 * @param   pitch   ピッチ
 * @return  true=success
 */
bool sphSolverSv::getConfigDomainInfo(size_t sz[3], double org[3], double pitch[3])
{
  register int i;
  if (isClassicConfig()) {      // V-Sphere Configuration
    /// <DomainInfo>要素からサイズ、原点座標、ピッチを取得する。
    unsigned int elem_size[3] = {0, 0, 0};
    float elem_org[3] = {0.0, 0.0, 0.0};
    float elem_pit[3] = {0.0, 0.0, 0.0};
    if (!m_sklconfig->GetVoxelSize(elem_size[0], elem_size[1], elem_size[2])) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can not get VoxelSize in SklConfig");
      return false;
    }
    if (!m_sklconfig->GetVoxelOrigin(elem_org[0], elem_org[1], elem_org[2])) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can not get VoxelOrigin in SklConfig");
      return false;
    }
    if (!m_sklconfig->GetVoxelPitch(elem_pit[0], elem_pit[1], elem_pit[2])) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_NOTFOUND_ELEMENT, "can not get VoxelPitch in SklConfig");
      return false;
    }
    for (i=0; i<3; i++) {
      sz[i] = elem_size[i];
      org[i] = elem_org[i];
      pitch[i] = elem_pit[i];
    }
  }
  else {
    /// <SphDomainInfo>要素からサイズ、原点座標、ピッチを取得する。
    sphElemUtils elemUtils(m_configElem);
    if (!elemUtils.getDomainInfo(m_solverDims, sz, org, pitch)) {
      // error
      m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "invalid <SphDomainInfo>.");
      return false;
    }
  }

  return true;
}


/**
 * コンフィグレーションからI,J,K分割数を取得する。
 * @param[out]   voxel_div      I,J,K分割数
 * @return  true=success
 */
bool sphSolverSv::getConfigVoxelDivision(unsigned int voxel_div[3])
{
  voxel_div[0]=voxel_div[1]=voxel_div[2]=0;

  // ソルバグループのプロセス数を取得する。
  int commSize = 0;
  SPL_Comm_size(m_solverComm, &commSize);
  if (commSize == 1) {
    voxel_div[0]=voxel_div[1]=voxel_div[2]=1;
    return true;
  }

  register int i;
  if (isClassicConfig()) {      // V-Sphere Configuration
    // VoxelDivisionMethodからI,J,K方向分割数を取得する。
    const SklCfg::CfgElem* div_elem = m_sklconfig->GetElemFirst(SklCfg::STEER, "VoxelDivisionMethod");
    int elem_div[3] = {0, 0, 0};
    if (div_elem == NULL) {
      return false;
    }

    const SklCfg::CfgParam* param;
    if ( (param = div_elem->GetParamFirst("I")) ) param->GetData(&elem_div[0]);
    if ( (param = div_elem->GetParamFirst("J")) ) param->GetData(&elem_div[1]);
    if ( (param = div_elem->GetParamFirst("K")) ) param->GetData(&elem_div[2]);
    voxel_div[0]=elem_div[0]; voxel_div[1]=elem_div[1]; voxel_div[2]=elem_div[2];
  }
  else {

    /// <SphSteer>要素の取得を行う。
    const sphElem* steer_elem = m_configElem->getElem(SPH_CFG_NAME_STEER);
    if (steer_elem == NULL) {
      return false;
    }
    if (!steer_elem->existChileElement(SPH_CFG_NAME_DIVISION)) {
      return false;
    }
    const sphElem* div_elem = steer_elem->getElem(SPH_CFG_NAME_DIVISION);
    if (div_elem == NULL) {
      return false;
    }
    /// <SphVoxelDivision>要素のi, j, kの取得を行う。
    const char *attr_idx[] = {SPH_CFG_ATTR_I, SPH_CFG_ATTR_J, SPH_CFG_ATTR_K};
    if (!sphElemUtils(div_elem).getDimsAttributeValues(m_solverDims, SPH_CFG_NAME_DIVISION, attr_idx, voxel_div)) {
      voxel_div[0]=voxel_div[1]=voxel_div[2]=0;
      return false;
    }
  }

  return true;
}


