/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file sphSolverUnst.C
 * Sphere Unstructured Mesh Solver super Class Spurce File
 * @author     hira
 * @date       2010/04/01
 *
 * \~japanese
 * @file sphSolverUnst.C
 * Sphere非構造格子ソルバー基底クラスソースファイル
 * @author     hira
 * @date       2010/04/01
 */
#include "solverbase/sphSolverUnst.h"
#include "config/sphElemUtils.h"
#include "data/sphDataUtils.h"


/**
 * \~english
 * Constructor
 *
 * \~japanese
 * コンストラクタ
 */
sphSolverUnst::sphSolverUnst()
{
  /// 初期化を行う。
  initialize();

}


/**
 * \~english
 * Constructor
 * @param     label     solver label
 *
 * \~japanese
 * コンストラクタ
 * @param     label     ソルバーラベル
 */
sphSolverUnst::sphSolverUnst(const char* solv_label) : sphSolverBase(solv_label)
{
  /// 初期化を行う。
  initialize();
}


/**
 * \~english
 * Constructor
 * @param     label     solver label
 * @param     id     solver id
 *
 * \~japanese
 * コンストラクタ
 * @param     label     ソルバーラベル
 * @param     id     ソルバーID
 */
sphSolverUnst::sphSolverUnst(const char* solv_label, int solv_id) : sphSolverBase(solv_label, solv_id)
{
  /// 初期化を行う。
  initialize();

}

/**
 * \~english
 * Destructor
 *
 * \~japanese
 * デストラクタ.
 */
sphSolverUnst::~sphSolverUnst()
{


}


/**
 * 初期化を行う。
 * @param     void
 * @return    success(=true)/fault(=false)
 */
bool sphSolverUnst::initialize()
{
  m_solverDims = 3;

  return true;
}


/**
 * \~english
 * initialize Solver.
   * @param               steptime         step, time class object
 * @return              -1:error, 0:forced terminated, 1:normality
 *
 * \~japanese
 * ソルバーの初期化を行う.
   * @param               steptime          実行ステップ・時刻クラス
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int
sphSolverUnst::initializeSolver(sphStepTime* steptime)
{
  int ret = SPHERE_SUCCESS;

  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "m_configElem is null");
    return SPHERE_TERMINATE;
  }

  /// <SphSolver>要素から基本動作に必要な情報を取得する。
  if ( (ret = setupSolver()) != SPHERE_SUCCESS ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setupSolver fault");
    return ret;
  }

  /// sphParaControllerを作成する。
  if ( (ret = initializeUnstMesh()) != SPHERE_SUCCESS ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "initializeUnstMesh fault");
    return ret;
  }

  /// <SphDataList>要素から情報収集、データ生成を行う。
  if ( (ret = setupDataList()) != SPHERE_SUCCESS ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "setupDataList fault");
    return ret;
  }

  return ret;
}


/**
 * <SphDataList>要素から情報収集、データ生成を行う。
 * @param               void
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverUnst::setupDataList()
{
  if (m_configElem == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_PARAMETER, "m_configElem is null.");
    return SPHERE_TERMINATE;
  }

  /// SphDataList/SphDataObj要素の取得を行う。
  const sphElem *data_elem = m_configElem->getChildElemFirst(2, SPH_CFG_NAME_DATALIST, SPH_CFG_NAME_DATAOBJ);
  while (data_elem != NULL) {
    /// instance属性が"instanceであれば、データオブジェクトを生成する。
    const char* instance = data_elem->getAttrString(SPH_CFG_ATTR_INSTANCE);
    if (instance == NULL || strcasecmp(instance, "initialize") == 0) {
      if (allocateDataObj(data_elem) == NULL) {
        // error
        m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "can not create data object.");
        return SPHERE_ERROR;
      }
    }
    data_elem = m_configElem->getChildElemNext(data_elem, 2, SPH_CFG_NAME_DATALIST, SPH_CFG_NAME_DATAOBJ);
  }

  return SPHERE_SUCCESS;

}


/**
 * <SphSolver>要素から基本情報を取得する。
 * @param               void
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverUnst::setupSolver()
{
  register int i;
  int ret = SPHERE_SUCCESS;

  if (m_configElem == NULL) return SPHERE_TERMINATE;

  /// ソルバのモードの取得を行う。
  if (m_configElem->existAttribute(SPH_CFG_ATTR_MODE)) {
    const char* mode = m_configElem->getAttrString(SPH_CFG_ATTR_MODE);
    if ( mode != NULL ) m_solverMode = mode;
  }

  /// 次元数の取得を行う。
  m_solverDims = m_configElem->getAttrInt(SPH_CFG_ATTR_DIMS);
  if (m_solverDims == 0) m_solverDims = 3;      // デフォルト＝３
  if (m_solverDims < 0 || m_solverDims > 3) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_INVALID_VALUE, "invalid dims(=%d)", m_solverDims);
    return SPHERE_TERMINATE;
  }

  return ret;

}


/**
 * sphParaControllerの生成を行う。
 * <SphSolver>要素からパラレルコントローラの生成を行う。
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverUnst::initializeUnstMesh()
{

  register int i;
  int ret = SPHERE_SUCCESS;

  if (m_configElem == NULL || (m_solverComm == SPL_COMM_NULL && splIsParallel()) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSolverUnst parameters(member) is null");
    return SPHERE_TERMINATE;
  }

  sphParaManager* paraMngr = getParaManager();
  if ( !setParaController(
                paraMngr->createUnstParaController(
                          m_configElem, m_solverComm)) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createUnstParaController fault");
    ret = SPHERE_TERMINATE;
  }

  return ret;

}


/**
 * sphParaControllerの生成を行う。
 * <SphSolver>要素からパラレルコントローラの生成を行う。
 * 非構造格子データはファイルを指定して読み込む（逐次実行用）
 * @param    fn_node   節点ファイル
 * @param    fn_elem   要素ファイル
 * @param    fn_grp    グループファイル
 * @param    fn_bnd    境界条件ファイル
 * @param    dtype     精度（sphUnstMesh::MESH_DOUBLE or sphUnstMesh::MESH_FLOAT）
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverUnst::initializeUnstMesh(const char* fn_node,
                                      const char* fn_elem,
                                      const char* fn_grp,
                                      const char* fn_bnd,
                                      sphUnstMesh::MeshDataType dtype )
{

  register int i;
  int ret = SPHERE_SUCCESS;

  if (m_configElem == NULL || (m_solverComm == SPL_COMM_NULL && splIsParallel()) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSolverUnst parameters(member) is null");
    return SPHERE_TERMINATE;
  }

  sphParaManager* paraMngr = getParaManager();
  if ( !setParaController(
                paraMngr->createUnstParaController(
                          m_configElem, m_solverComm, fn_node, fn_elem, fn_grp, fn_bnd, dtype)) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createUnstParaController fault");
    ret = SPHERE_TERMINATE;
  }

  return ret;

}


/**
 * sphParaControllerの生成を行う。
 * <SphSolver>要素からパラレルコントローラの生成を行う。
 * 非構造格子データはdist_fnameから読み込む（並列実行に対応した分散格子ファイル）
 * @param       dist_fname    分散非構造格子ファイル
 * @param       dtype         精度（sphUnstMesh::MESH_DOUBLE or sphUnstMesh::MESH_FLOAT）
 * @return              -1:error, 0:forced terminated, 1:normality
 */
int sphSolverUnst::initializeUnstMesh(const char* dist_fname,
                                      sphUnstMesh::MeshDataType dtype)
{

  register int i;
  int ret = SPHERE_SUCCESS;

  if (m_configElem == NULL || (m_solverComm == SPL_COMM_NULL && splIsParallel()) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_NULL_POINTER, "sphSolverUnst parameters(member) is null");
    return SPHERE_TERMINATE;
  }

  sphParaManager* paraMngr = getParaManager();
  if ( !setParaController(
                paraMngr->createUnstParaController(
                          m_configElem, m_solverComm, dist_fname, dtype)) ) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "createUnstParaController fault");
    ret = SPHERE_TERMINATE;
  }

  return ret;

}


/**
 * パラレルコントローラーを設定する.
 * sphUnstParaControllerへのアクセスを容易にする為にm_svParaCtrlへ設定を行う。
 * 主はsphParaManagerのmapに存在する。m_svParaCtrlはテンポラリ的ポインタ
 * @param               paraCtrl
 * @return    success(=true)/fault(=false)
 */
bool sphSolverUnst::setParaController(sphParaController* paraCtrl)
{
  if (!sphSolverBase::setParaController(paraCtrl)) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_RETURN_FAULT, "sphSolverBase::setParaController fault");
    return false;
  }

  m_unstParaCtrl = dynamic_cast<sphUnstParaController*>(paraCtrl);
  if (m_unstParaCtrl == NULL) {
    // error
    m_errHandler->SPH_ERROR_INFO(SPH_ERR_CAST_NULL, "sphUnstParaController cast is null");
    return false;
  }

  return true;
}


