/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file spl.c
 * Declaration of Message for Sphere Low Level Library
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file spl.c
 * Sphereメッセージ定義ファイル
 * @author     hira
 * @date       2009/02/20
 */


#include <stdio.h>
#include <sys/time.h>
#include <stdarg.h>
#include "spl/spl.h"

#if defined(__cplusplus)
extern "C" {
#endif

static const char* _error_file = NULL;
static int _error_line = -1;
static const char* _error_func = NULL;

/**
 * \~english
 * Error Code in sphere
 * \~japanese
 * エラーコードグローバル変数
 */
int spl_err_code = SPL_SUCCESS;
char spl_err_info[SPL_MAX_STR_SIZE];

/**
 * \~english
 * Error Message List
 * \~japanese
 * エラーメッセージ文字列リスト
 */
const char* spl_err_str[SPL_ERR_LASTCODE+1] = {
  "Success"        /*   0:SPL_SUCCESS */
 ,"MPI error"        /*   1:SPL_ERR_MPI_ERROR */
 ,"Invalid parameter"      /*   2:SPL_ERR_INVALID_PARAMETER */
 ,"Memory allocate error"    /*   3:SPL_ERR_MEMORY_ALLOCATE */
 ,"File open error"      /*   4:SPL_ERR_OPEN_FILE */
 ,"File read error"      /*   5:SPL_ERR_READ_FILE */
 ,"File write error"      /*   6:SPL_ERR_WRITE_FILE */
 ,"File seek error"      /*   7:SPL_ERR_SEEK_FILE */
 ,"Illegal EOF was detected"    /*   8:SPL_ERR_DETECT_ILLEGAL_EOF */
 ,"Pointer is NULL"      /*   9:SPL_ERR_NULL_POINTER */
 ,"String length too long"    /*  10:SPL_ERR_STR_SIZE_TOO_LONG */
 ,"Invalid data type"  /*  11:SPL_ERR_INVALID_SPLSV_DTYPE */
 ,"Data pointer is NULL"    /*  12:SPL_ERR_DATA_IS_NULL */
 ,"Invalid(or different) Array size"  /*  13:SPL_ERR_INVALID_ARRAY_SIZE */
 ,"Invalid file magick No"    /*  14:SPL_ERR_INVALID_MAGICK_NO */
 ,"Invalid file endian type"    /*  15:SPL_ERR_INVALID_ENDIAN_TYPE */
 ,"Compress data failed"    /*  16:SPL_ERR_COMPRESS_DATA */
 ,"Uncompress data failed"    /*  17:SPL_ERR_UNCOMPRESS_DATA */
 ,"Buffer over flow"      /*  18:SPL_ERR_BUFFER_OVERFLOW */
 ,"Constractor Error"/*  19:SPL_ERR_CONSTRACTOR */
 ,"STL error"/*  20:SPL_ERR_STL */
 ,"Initialize error"/*  21:SPL_ERR_INITIALIZE */
 ,"Invalid Environment"/*  22:SPL_ERR_INVALID_ENVIRONMENT */
 ,"Cast result is null"/*  23:SPL_ERR_CAST_NULL */
 ,"Not suport method"/*  24:SPL_ERR_NOT_SUPPORT_METHOD */
 ,"Not found in list"/*  25:SPL_ERR_NOT_FOUND_IN_LIST */
 ,"Allreduce sum value failed"/*  26: SPL_ERR_ALLREDUCE_SUM*/
 ,"File fstat error"/*  27:SPL_ERR_FSTAT_FILE */
 ,"Internal impossible error."/*  28: SPL_ERR_INTERNAL_IMPOSSIBLE*/
 ,"Invalid global start index"/*  29: SPL_ERR_START_IDX */
 ,"Recode size error." /*  30: SPL_ERR_RECODE_SIZE */
 ,""/*  31: */
 ,""/*  32: */
 ,""/*  33: */
 ,""/*  34: */
 ,""/*  35: */
 ,""/*  36: */
 ,""/*  37: */
 ,""/*  38: */
 ,""/*  39: */
 ,""/*  40: */
 ,""/*  41: */
 ,""/*  42: */
 ,""/*  43: */
 ,""/*  44: */
 ,""/*  45: */
 ,""/*  46: */
 ,""/*  47: */
 ,""/*  48: */
 ,""/*  49: */
 ,""/*  50: SPL_ERR_LASTCODE*/
};


/**
 * \~english
 * Print Error Message
 * \~japanese
 * エラーメッセージを表示する。
 * @param        str      エラーメッセージ
 */
void
spl_err(const char* str) {
  if( str ) {
    fprintf(stderr, " (error) %s : %s\n", spl_err_str[spl_err_code], str);
  }
  else {
    fprintf(stderr, " (error) %s\n", spl_err_str[spl_err_code]);
  }
}

/**
 * \~english
 * Print Message
 * @param        format      message format
 * @param        ...          message value
 * \~japanese
 * メッセージを標準出力に表示する。
 * @param        format      出力メッセージフォーマット
 * @param        ...          出力メッセージ値
 */
void
spl_msg(const char* format, ...) {
  if( !format ) return;
  char buf[1024];
  va_list vl;
  va_start(vl, format);
  vsprintf(buf, format, vl);
  va_end(vl);
  fprintf(stdout, " SPHERE MSG : %s\n", buf);
  return;
}

/**
 * \~english
 * Print Error Message
 * @param        format      message format
 * @param        ...          message value
 * \~japanese
 * エラーメッセージを表示する。
 * @param        format      出力メッセージフォーマット
 * @param        ...          出力メッセージ値
 */
void
spl_errmsg(const char* format, ...) {
  if( !format ) return;
  char buf[1024];
  va_list vl;
  va_start(vl, format);
  vsprintf(buf, format, vl);
  va_end(vl);
  fprintf(stderr, " SPHERE ERROR : %s\n", buf);
  return;
}


/**
 * \~english
 * Print Warning Message
 * @param        format      message format
 * @param        ...          message value
 * \~japanese
 * 警告メッセージを表示する。
 * @param        format      出力メッセージフォーマット
 * @param        ...          出力メッセージ値
 */
void
spl_wrnmsg(const char* format, ...) {
  if( !format ) return;
  char buf[1024];
  va_list vl;
  va_start(vl, format);
  vsprintf(buf, format, vl);
  va_end(vl);
  fprintf(stderr, " SPHERE WARNNING : %s\n", buf);
  return;
}

/**
 * \~english
 * get current time
 * @return        second value
 * \~japanese
 * 現在時刻を取得する。
 * @return        時刻（秒）
 */
double
splGetTime() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  double t = 0.0;
  t += (double)tv.tv_sec;
  t += (double)tv.tv_usec * 1.0e-6;
  return t;
}


/**
 * \~english
 * set Error Message
 * @param        err_code        error code
 * @param        format          additional message format
 * @param        ...              additional message value
 *
 * \~japanese
 * エラーメッセージを設定する。
 *
 * DBG_MSGを定義したときのみエラーメッセージを出力する。
 * @param        err_code        エラーコード
 * @param        format          付加メッセージフォーマット
 * @param        ...              付加メッセージ値
 */
void
spl_set_errmsg(int err_code, const char* format, ...)
{
  char *err_str = NULL;
  if (err_code > SPL_SUCCESS && err_code <= SPL_ERR_LASTCODE) {
    err_str = (char*)spl_err_str[err_code];
  }

  char buf[SPL_MAX_STR_SIZE] = "";
  if( format ) {
    va_list vl;
    va_start(vl, format);
    vsprintf(buf, format, vl);
    va_end(vl);
  }

  spl_err_code = err_code;
  strcpy(spl_err_info, buf);
  return;
}

/**
 * エラー発生箇所のファイル名、行番号、関数名を設定する。
 * spl_set_error_file -> spl_set_errinfo_fmtと対で使用する。
 * @param        file_name          エラー発生ファイル名
 * @param        lineno             エラー発生行番号
 * @param        func               エラー発生関数名
 */
void spl_set_error_file(const char* file_name, int lineno, const char* func)
{
  _error_file = file_name;
  _error_line = lineno;
  _error_func = func;
}


/**
 * \~english
 * set Error Message
 * @param        err_code        error code
 * @param        filename         error file name
 * @param        lineno           error line number
 * @param        funcname         error function name
 * @param        format          additional message format
 * @param        ...              additional message value
 *
 * \~japanese
 * エラーメッセージを設定する。
 *
 * DBG_MSGを定義したときのみエラーメッセージを出力する。
 * @param        err_code        エラーコード
 * @param        filename          エラー発生ファイル名
 * @param        lineno             エラー発生行番号
 * @param        funcname               エラー発生関数名
 * @param        format          付加メッセージフォーマット
 * @param        ...              付加メッセージ値
 */
void
spl_set_errinfo(int err_code, const char* filename, int lineno, const char* funcname, const char* format, ...)
{
  char *err_str = NULL;
  if (err_code > SPL_SUCCESS && err_code <= SPL_ERR_LASTCODE) {
    err_str = (char*)spl_err_str[err_code];
  }

  char buf[SPL_MAX_STR_SIZE] = "";
  if( format ) {
    va_list vl;
    va_start(vl, format);
    vsprintf(buf, format, vl);
    va_end(vl);
  }

  spl_err_code = err_code;
  sprintf(spl_err_info, "%s (%s[%d]:%s)", buf, filename, lineno, funcname);

  return;
}


/**
 * \~english
 * set Error Message
 * @param        err_code        error code
 * @param        format          additional message format
 * @param        ...              additional message value
 *
 * \~japanese
 * エラーメッセージを設定する。
 *
 * DBG_MSGを定義したときのみエラーメッセージを出力する。
 * spl_set_error_file -> spl_set_errinfo_fmtと対で使用する。
 * @param        err_code        エラーコード
 * @param        format          付加メッセージフォーマット
 * @param        ...              付加メッセージ値
 */
void
spl_set_errinfo_fmt(int err_code, const char* format, ...)
{
  char *err_str = NULL;
  if (err_code > SPL_SUCCESS && err_code <= SPL_ERR_LASTCODE) {
    err_str = (char*)spl_err_str[err_code];
  }

  char buf[SPL_MAX_STR_SIZE] = "";
  if( format ) {
    va_list vl;
    va_start(vl, format);
    vsprintf(buf, format, vl);
    va_end(vl);
  }

  spl_err_code = err_code;
  sprintf(spl_err_info, "%s (%s[%d]:%s)", buf, _error_file, _error_line, _error_func);

  return;
}


/**
 * \~english
 * clear error code
 *
 * \~japanese
 * エラーコードをクリアする。
 */
void
spl_clear_errcode() {
  spl_err_code = SPL_SUCCESS;
  spl_err_info[0] = '\0';
}

/**
 * \~english
 * set error code
 * @param        err_code      error code
 * \~japanese
 * エラーコードを設定する。
 * @param        err_code      エラーコード
 */
void
spl_set_errcode(int err_code) {
  spl_err_code = err_code;
  spl_err_info[0] = '\0';
}

/**
 * \~english
 * get error code
 * @return          error code
 * \~japanese
 * エラーコードを取得する。
 * @return            エラーコード
 */
int
spl_get_errcode() {
  return spl_err_code;
}

/**
 * \~english
 * get defined error code
 * @param        err_code      error code
 * @return          error message
 * \~japanese
 * 定義済みのエラーコードを取得する。
 * @param        err_code      エラー番号
 * @return            エラーメッセージ
 */
const char*
spl_get_errinfo() {
  return spl_err_info;
}

/**
 * \~english
 * get defined error code
 * @param        err_code      error code
 * @return          error message
 * \~japanese
 * 定義済みのエラーコードを取得する。
 * @param        err_code      エラー番号
 * @return            エラーメッセージ
 */
const char*
spl_get_define_errmsg(int err_no) {
  return spl_err_str[err_no];
}

/**
 * \~english
 * print error message to stderr
 * @param        　void
 * @return          void
 * \~japanese
 * 標準エラー出力にエラーメッセージを出力する。
 * @param        　void
 * @return          void
 */
void
spl_print_error_message() {
  fprintf(stderr, "Spl_Error[%d]:%s info=%s \n", spl_err_code, spl_err_str[spl_err_code], spl_err_info);
}

#if defined(__cplusplus)
}
#endif /* extern "C" */


