/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file public_splfio_sph.c
 * Implementation of public SPH subroutine.
 * @author     yoh
 * @date       2009/11/27
 *
 * \~japanese
 * @file public_splfio_sph.c
 * SPHファイル入出力公開関数郡ソースファイル
 * @author     yoh
 * @date       2009/11/27
 */

#include <stdio.h>
#include <limits.h>

#include "spl/splfio_spx.h"
#include "spl/util_endian.h"
#include "splfio_spx_io.h"
#include "splfio_sph_io.h"

#if defined(__cplusplus)
extern "C" {
#endif

/**
 * \~english
 * read SPH
 * @param[in]      file_list     file list
 * @param[in]      mode          read mode
 *                               - 0:all nodes read
 *                               - 1:only readRank read data
 *                               - 2:only readRank read data, and readRank
 *                                   not expand data.
 * @param[in,out]  obj           splSpx Object
 * @param[in]      readRank      Rank that reads a file
 *                               (It is effective at mode 1 or 2.)
 * @param[in]      grp           Communication group
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * SPHファイルを読み込む。
 * @param[in]      file_list     ファイルリスト
 * @param[in]      mode          読込モード
 *                               - 0:すべてのノードにて読み込みを行う。
 *                               - 1:readRankノードが読み込みを行う。
 *                               - 2:only readRank read data, and readRank
                                     not expand data.
 * @param[in,out]  obj           splSpxオブジェクト
 * @param[in]      readRank      ファイル読込ノード番号(モード1, 2で有効)
 * @param[in]      grp           プロセスグループ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splReadSph(splSvFileList* file_list,  /* input  */
           int mode,                  /* input  */
           splSpx* obj,               /* input/output */
           int readRank,              /* input */
           SPL_Comm grp               /* input  */ ) {
  int myrank;
  splSvFile* pp;
  int flg = 0, tmp = 0;

  if( !file_list || !obj ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "file_list or obj is null.");
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  if( !obj->size ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "obj->size is null.");
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  register unsigned int n;
  size_t dest_len;
  dest_len = (obj->dims >= 1) ? 1 : 0;
  for(n=0; n<obj->dims; n++) dest_len *= obj->size[n];

  if (dest_len > 0) {
    // check splSpx object
    if (!checkSpx(obj, SplFalse)) {
      flg = 1;
    }
  }

  if (flg != 0) {
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  // none parallel is mode = 0
  if (!splIsParallel()) mode = 0;

  if( mode == 0 ) myrank = 0;
  else            SPL_Comm_rank(grp, &myrank);

  // copy FileList
  if (mode != 0) {
    if (splBcastSvFileList(file_list, readRank, grp) != SplTrue) {
      spl_print_error_message();
      SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
      return SplFalse;
    }
  }

  pp = file_list->top;
  while( pp ) {

    register unsigned int m;
    const char* fname;
    FILE* fp;
    EMatchType etype;
    splSpx fobj;

    fname = (const char*)pp->fname;

    // 初期化
    splInitSpx(&fobj);

    /* getSphInfo */
    if( getSphInfo(mode, readRank, grp, fname, &fobj, &etype) == SplFalse ) {
      flg = 1;
    }

    if (!flg) {
      // check dims in file
      if (fobj.dims != obj->dims) {
        SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER,
		       "not equals dims in file [dims in file=%d, "
		       "dims in parameter=%d].", fobj.dims, obj->dims);
        flg = 1;
      }

      // check rlen in file
      if (fobj.rlen != obj->rlen) {
        SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER,
		       "not equals rlen in file [rlen in file=%d, "
		       "rlen in parameter=%d].", fobj.rlen, obj->rlen);
        flg = 1;
      }

      /* set paramter of obj */
      if( pp->set_flag == 1 ) {
        for(m=0; m<fobj.dims; m++) {
          fobj.start_idx[m] = pp->start_idx[m];
        }
      }
      else {
        if (file_list->list_size > 1) {
          // error
          SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER,
                         "not set start_idx in file_list");
          flg = 1;
        }
      }

      // set header from file
      for(m=0; m<fobj.dims; m++) {
        obj->pitch[m] = fobj.pitch[m];
        obj->origin[m] = fobj.origin[m]
                         + ((long long)obj->start_idx[m]
			    - (long long)fobj.start_idx[m]) * fobj.pitch[m];
        obj->wsize[m] = fobj.wsize[m];
      }
      obj->rlen = fobj.rlen;
      obj->crddef = fobj.crddef;
      obj->aux = fobj.aux;
      obj->block_size = fobj.block_size;
      obj->step = fobj.step;
      obj->time = fobj.time;
    }

    // check error between all nodes
    if (flg != 0) {
      spl_print_error_message();
      SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
      return SplFalse;
    }

    fp = NULL;
    /* open file */
    if( (mode == 0) || (myrank == readRank) ) {
      if( (fp = fopen(fname, "rb")) == NULL ) {
        SPL_ERROR_INFO(SPL_ERR_OPEN_FILE, "file name is %s.", fname);
        flg = 1;
      }
    }
    else {
      fp = NULL;
    }

    // gc in sph file from 'configuration xml file'
    fobj.gc = obj->gc_insph;

    /* read data record */
    if( readSphDataRecord(mode, readRank, grp, fp, etype,
                          obj->dims, fobj.vlen, obj->vlen,
                          fobj.gc, obj->gc, fobj.rlen,
                          fobj.size, fobj.start_idx,
                          obj->size, obj->start_idx,
                          obj->data.c) == SplFalse ) {
      if( (mode == 0) || (myrank == readRank) ) fclose(fp);
      flg = 1;
    }

    /* file close */
    if (fp) fclose(fp);

    // check error between all nodes
    if (flg != 0) {
      spl_print_error_message();
      SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
      return SplFalse;
    }

    /* increment pp */
    pp = pp->next;
  }

  return SplTrue;
}

/**
 * \~english
 * write SPH
 * @param        fname        write file name
 * @param        mode         write mode
 *                            - 0:all nodes write
 *                            - 1:only writeRank write data. Output gather data.
 *                            - 2:only writeRank write data. Output gather data.
 *                                and writeRank has not my splSpx object.
 * @param        block_size   compress block size, or gather send/receive
 *                            buffer size
 * @param         obj         splSpx Object
 * @param        writeRank    Rank that write a file
 *                            (It is effective at mode 1 or 2.)
 * @param        grp          Communication group
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * SPHファイルを書き込む。
 * @param        fname        書込ファイル名
 * @param        mode         書込モード
 *                            - 0:すべてのノードがファイルの書き込みを行う。
 *                            - 1:writeRankノードのみがデータ収集を行い、
 *                                ファイルに書き込む。
 *                            - 2:splSpxオブジェクトを持たないwriteRankノードが
 *                                データ収集を行い、ファイルに書き込む。
 * @param        block_size   圧縮ブロックサイズ、又はモード1,2における
 *                            データ収集バッファーサイズ
 * @param         obj         splSpxオブジェクト
 * @param        writeRank    ファイル書込ノード番号(モード1, 2で有効)
 * @param        grp          プロセスグループ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splWriteSph(const char* fname,    /* input */
            int mode,             /* input */
            size_t block_size,    /* input */
            const splSpx* obj,    /* input */
            int writeRank,        /* input */
            SPL_Comm grp          /* input */ ) {

  int myrank = 0;
  FILE* fp = NULL;
  register unsigned int n;
  int flg = 0, tmp = 0;

  SPL_Datatype dt_size_t;
  if( sizeof(size_t) == 4 )      dt_size_t = SPL_INT;
  else if( sizeof(size_t) == 8 ) dt_size_t = SPL_LONG_LONG;
  else                           return SplFalse;

  /* parameter check */
  if( !obj ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "splSpx object is null.");
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  // check splSpx object
  if (!checkSpx(obj, SplTrue)) {
    flg = 1;
  }

  // none parallel is mode = 0
  if (!splIsParallel()) mode = 0;

  if( mode == 0 )  myrank = 0;
  else             SPL_Comm_rank(grp, &myrank);

  size_t dest_len;
  dest_len = (obj->dims >= 1) ? 1 : 0;
  for(n=0; n<obj->dims; n++) dest_len *= obj->size[n];

  if ((mode == 0 &&  dest_len > 0) || (mode != 0 && myrank == writeRank)) {
    if( !fname ) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "fname is null.");
      flg = 1;
    }
    if( (fp = fopen(fname, "wb")) == NULL ) {
      SPL_ERROR_INFO(SPL_ERR_OPEN_FILE, "file name is %s", fname);
      flg = 1;
    }
  }
  else {
    fp = NULL;
  }

  // check error between all nodes
  if (flg != 0) {
    if( fp ) fclose(fp);
    fp = NULL;
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  size_t wsidx[3] = {0,0,0};
  size_t write_size[3] = {0,0,0};
  size_t write_sidx[3] = {0,0,0};
  if (!flg) {
    size_t wlen = 1;
    for(n=0; n<obj->dims; n++) wlen *= obj->wsize[n];

    // bcast wlen
    SPL_Bcast(&wlen, 1, dt_size_t, writeRank, grp);
    if (wlen == 0) {
      // not set wsize
      // get whole size and whole start index
      if (getSvWholeSize(mode, obj->dims, obj->size, obj->start_idx, obj->wsize,
			 wsidx, write_size, write_sidx, grp) == SplFalse) {
        flg = 1;
      }
    }
    else {
      if (mode != 0) {
        for(n=0; n<obj->dims; n++) {
          wsidx[n] = 0;
          write_size[n] = obj->wsize[n];
          write_sidx[n] = obj->start_idx[n];
        }
      }
      else {
        for(n=0; n<obj->dims; n++) {
          wsidx[n] = obj->start_idx[n];
          write_size[n] = obj->size[n];
          write_sidx[n] = obj->start_idx[n];
        }
      }
    }

    if (mode != 0) {
      // calculate origin/pitch coords
      if (getVoxelCoords(obj->dims, obj->size, obj->start_idx, wsidx,
			 obj->origin, obj->pitch, grp) == SplFalse) {
        flg = 1;
      }
    }
  }

  // set block_size for gather
  if (mode != 0 && block_size == 0) {
    block_size = 1;
    for(n=0; n<obj->dims; n++) block_size *= write_size[n];
  }

  // check error between all nodes
  if (flg != 0) {
    if( fp ) fclose(fp);
    fp = NULL;
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  if (!flg) {
    if( writeSphHeaderRecord(fp, mode, writeRank, grp,
                             obj->vlen, obj->rlen,
                             write_size, obj->origin, obj->pitch,
                             obj->step, obj->time) == SplFalse ) {
      if( fp ) fclose(fp);
      fp = NULL;
      flg = 1;
    }
  }

  if (!flg) {
    if( writeSphDataRecord(mode, writeRank, grp, fp, block_size,
                           obj->vlen, obj->rlen,
                           obj->size, obj->data.c,
                           write_size, write_sidx) == SplFalse) {
      if( fp ) fclose(fp);
      fp = NULL;
      flg = 1;
    }
  }

  if( fp ) fclose(fp);
  fp = NULL;

  // check error between all nodes
  if (flg != 0) {
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  return SplTrue;
}

/**
 * \~english
 * get SPH File Information
 * @param[in]      fname       file name
 * @param[out]  obj            splSpx Object
 * @param[in]      mode        read mode
 *                             - 0:all nodes read
 *                             - 1:only readRank read data
 *                             - 2:only readRank read data, and readRank
 *                                 not expand data.
 * @param[in]      readRank    Rank that reads a file (It is effective at mode 1 or 2.)
 * @param[in]      grp         Communication group
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * SPHファイル情報を取得する。
 * @param[in]      fname       ファイル名
 * @param[out]  obj            splSpxオブジェクト
 * @param[in]      mode        読込モード
 *                             - 0:すべてのノードにて読み込みを行う。
 *                             - 1:readRankノードが読み込みを行う。
 *                             - 2:only readRank read data, and readRank
 *                                 not expand data.
 * @param[in]      readRank    ファイル読込ノード番号(モード1, 2で有効)
 * @param[in]      grp         プロセスグループ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splGetSphInfo(const char* fname,            /* input  */
              splSpx* obj,                  /* output */
              int mode,                     /* input  */
              int readRank,                 /* input  */
              SPL_Comm grp                  /* input  */ ) {
  EMatchType etype;

  if( getSphInfo(mode, readRank, grp, fname, obj, &etype) == SplFalse ) {
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  return SplTrue;
}

#if defined(__cplusplus)
}
#endif /* extern "C" */
