/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file public_splfio_spx.c
 * Implementation of public SPX subroutine.
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file public_splfio_spx.c
 * SPXファイル入出力公開関数郡ソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#include <stdio.h>
#include <limits.h>

#include "spl/splfio_spx.h"
#include "spl/util_endian.h"
#include "splfio_spx_io.h"

#if defined(__cplusplus)
extern "C" {
#endif

//extern int spl_err_code;

/**
 * \~english
 * Initialize splSpx object
 * @param        obj        splSpx Object
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSpxオブジェクトを初期化する。
 * @param        obj        splSpxオブジェクト
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splInitSpx(splSpx* obj) {
  register int i;
  if( !obj ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "splSpx object is null");
    return SplFalse;
  }
  for(i=0; i<3; i++) {
    obj->wsize[i] = 0;
    obj->start_idx[i] = 0;
    obj->size[i] = 0;
    obj->origin[i] = 0.0;
    obj->pitch[i] = 0.0;
  }
  obj->dims = 0;
  obj->vlen = 1;
  obj->dtype = 0;
  obj->gc = 0;
  obj->rlen = 0;
  obj->crddef = 1;
  obj->aux = 0;
  obj->step = 0;
  obj->time = 0.0;
  obj->block_size = 0.0;
  obj->data.c = NULL;
  obj->gc_insph = 0;
  return SplTrue;
}


/**
 * \~english
 * Finalize splSpx object
 * @param        obj                splSpx Object
 * @param        mem_free_flag      data memory free flag<br>
 *                                 SplTrue:memory free/SplFalse:not memory free
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSpxオブジェクトを破棄する。
 * @param        obj                splSpxオブジェクト
 * @param        mem_free_flag      データ領域のメモリ解放フラグ<br>
 *                                 SplTrue:メモリ解放する。/SplFalse:メモリ解放しない。
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splFinalizeSpx(splSpx* obj, SplBool mem_free_flag) {
  register int i;
  if( !obj ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "splSpx object is null");
    return SplFalse;
  }
  if( mem_free_flag ) {
    switch( obj->rlen ) {
      case 4:
        free(obj->data.f);
        obj->data.f = NULL;
        break;
      case 8:
        free(obj->data.d);
        obj->data.d = NULL;
        break;
      default:
        // error
        SPL_ERROR_INFO(SPL_ERR_INVALID_SPLSV_DTYPE, "rlen is %d", obj->rlen);
        return SplFalse;
        break;
    }
  }
  for(i=0; i<3; i++) {
    obj->wsize[i] = 0;
    obj->start_idx[i] = 0;
    obj->size[i] = 0;
    obj->origin[i] = 0.0;
    obj->pitch[i] = 0.0;
  }
  obj->dims = 0;
  obj->vlen = 1;
  obj->dtype = 0;
  obj->gc = 0;
  obj->rlen = 0;
  obj->crddef = 1;
  obj->aux = 0;
  obj->step = 0;
  obj->time = 0.0;
  obj->block_size = 0;
  obj->gc_insph = 0;
  return SplTrue;
}


/**
 * \~english
 * set value in splSpx object
 * @param        obj                splSpx Object
 * @param        dims              dimension ( 1:1D, 2:2D, 3:3D )
 * @param        vlen              vector length
 * @param        dtype              data type ( 1:unsigned char, 2:unsigned short, 4:unsigned int )
 * @param        gc                guide cell size
 * @param        rlen              real data size (4:float, 8:double)
 * @param        crfdef            coords define value (1:Regular, 2:Collocate, 3:Staggered(1), 4:Staggered(2))
 * @param        sta_idx            global start index
 * @param        size              size
 * @param        dt                data pointer
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSpxオブジェクトにSPX情報を設定する。。
 * @param        obj                splSpxオブジェクト
 * @param        dims              次元数 ( 1:1D, 2:2D, 3:3D )
 * @param        vlen              ベクトルデータ長(1 or 3)
 * @param        dtype              データ型タイプ ( 1:unsigned char, 2:unsigned short, 4:unsigned int )
 * @param        gc                ガイドセルサイズ (4:float, 8:double)
 * @param        rlen              実数データのサイズ (1:Regular, 2:Collocate, 3:Staggered(1), 4:Staggered(2))
 * @param        crfdef            原点座標の定義位置
 * @param        sta_idx            始点インデックス
 * @param        size              サイズ
 * @param        dt                データポインタ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splSpxSetValues(splSpx* obj,
                unsigned int dims,
                unsigned int vlen,
                unsigned int dtype,
                size_t gc,
                unsigned int rlen,
                unsigned int crddef,
                size_t* sta_idx,
                size_t* size,
                unsigned char* dt) {
  if( !obj || !sta_idx || !size ) {        // error
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, " obj, sta_idx or size is null.");
    return SplFalse;
  }

  obj->dims = dims;
  obj->vlen = vlen;
  obj->dtype = dtype;
  obj->gc = gc;
  obj->rlen = rlen;
  obj->crddef = crddef;
  memcpy(obj->start_idx, sta_idx, sizeof(size_t)*dims);
  memcpy(obj->size, size, sizeof(size_t)*dims);
  obj->data.c = dt;

  return SplTrue;
}


/**
 * \~english
 * set origin coords and pitch in splSpx object
 * @param        obj                splSpx Object
 * @param        dims              dimension ( 1:1D, 2:2D, 3:3D )
 * @param        org               origin coords
 * @param        pit               pitch
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSpxオブジェクトの原点座標, ピッチを設定する。
 *
 * 設定された原点座標, ピッチは、書込(splWriteSpx)にて使用する。<br>
 * 読込(splReadSpx)では使用しない。
 * @param        obj                splSpxオブジェクト
 * @param        dims              次元数 ( 1:1D, 2:2D, 3:3D )
 * @param        org              原点座標
 * @param        pit              ピッチ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splSpxSetCoords(splSpx* obj,
                    unsigned int dims,
                    double* org,
                    double* pit)
{
    if( !obj || !org || !pit ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, " obj, org or pit is null.");
      return SplFalse;
    }

    memcpy(obj->origin, org, sizeof(double)*dims);
    memcpy(obj->pitch, pit, sizeof(double)*dims);

    return SplTrue;
}

/**
 * \~english
 * set extended value in splSpx object
 * @param        obj                splSpx Object
 * @param        dims              dimension ( 1:1D, 2:2D, 3:3D )
 * @param        wsize              global voxel size
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSpxオブジェクトの拡張領域レコード情報を設定する。
 *
 * 設定された拡張領域情報は、書込(splWriteSpx)にて使用する。
 * @param        obj                splSpxオブジェクト
 * @param        dims              次元数 ( 1:1D, 2:2D, 3:3D )
 * @param        wsize              全体ボクセル数
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splSpxSetWholeSize(splSpx* obj,
                    unsigned int dims,
                    size_t* wsize)
{
    if( !obj || !wsize  ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, " obj or wsize is null.");
      return SplFalse;
    }

    memcpy(obj->wsize, wsize, sizeof(size_t)*dims);

    return SplTrue;
}


/**
 * \~english
 * read SPX
 * @param[in]      file_list          file list
 * @param[in]      mode              read mode
 *                                 - 0:all nodes read
 *                                 - 1:only readRank read data
 *                                 - 2:only readRank read data, and readRank not expand data.
 * @param[in,out]  obj             splSpx Object
 * @param[in]      readRank          Rank that reads a file (It is effective at mode 1 or 2.)
 * @param[in]      grp                Communication group
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * SPXファイルを読み込む。
 * @param[in]      file_list          ファイルリスト
 * @param[in]      mode              読込モード
 *                                 - 0:すべてのノードにて読み込みを行う。
 *                                 - 1:readRankノードが読み込みを行う。
 *                                 - 2:only readRank read data, and readRank not expand data.
 * @param[in,out]  obj             splSpxオブジェクト
 * @param[in]      readRank          ファイル読込ノード番号(モード1, 2で有効)
 * @param[in]      grp                プロセスグループ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splReadSpx(splSvFileList* file_list,  /* input  */
           int mode,                  /* input  */
           splSpx* obj,               /* input/output */
           int readRank,              /* input */
           SPL_Comm grp               /* input  */ ) {
  int myrank;
  splSvFile* pp;
  int flg = 0, tmp = 0;

  if( !file_list || !obj ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "file_list or obj is null.");
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  if( !obj->size ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "obj->size is null.");
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  register unsigned int n;
  size_t dest_len;
  dest_len = (obj->dims >= 1) ? 1 : 0;
  for(n=0; n<obj->dims; n++) dest_len *= obj->size[n];

  if (dest_len > 0) {
    // check splSpx object
    if (!checkSpx(obj, SplFalse)) {
      flg = 1;
    }
  }

  if (flg != 0) {
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  // none parallel is mode = 0
  if (!splIsParallel()) mode = 0;

  if( mode == 0 )  myrank = 0;
  else   SPL_Comm_rank(grp, &myrank);

  // copy FileList
  if (mode != 0) {
    if (splBcastSvFileList(file_list, readRank, grp) != SplTrue) {
      spl_print_error_message();
      SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
      return SplFalse;
    }
  }

  pp = file_list->top;
  while( pp ) {

    register unsigned int m;
    const char* fname;
    FILE* fp;
    EMatchType etype;
    unsigned long long blk_sz, cmp_flg;
    SplBool has_ext_record;
    splSpx fobj;

    fname = (const char*)pp->fname;

    /* getSpxInfo */
    if( getSpxInfo(mode, readRank, grp, fname, &fobj,
                   &etype, &blk_sz, &cmp_flg, &has_ext_record) == SplFalse ) {
      flg = 1;
    }

    if (!flg) {
      // check dims in file
      if (fobj.dims != obj->dims) {
        SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "not equals dims in file [dims in file=%d, dims in parameter=%d].", fobj.dims, obj->dims);
        flg = 1;
      }

      // check rlen in file
      if (fobj.rlen != obj->rlen) {
        SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "not equals rlen in file [rlen in file=%d, rlen in parameter=%d].", fobj.rlen, obj->rlen);
        flg = 1;
      }

      // check vlen in file
      if (fobj.vlen != obj->vlen) {
        SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "not equals vlen in file [vlen in file=%d, vlen in parameter=%d].", fobj.vlen, obj->vlen);
        flg = 1;
      }

      /* set paramter of obj */
      if( pp->set_flag == 1 ) {
        if( has_ext_record == SplTrue ) {
          int chk_flag = 0;
          for(m=0; m<fobj.dims; m++) {
            if( pp->start_idx[m] != fobj.start_idx[m] ) {
              // error
              SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "invalid start_idx, filename=%s,  pp->start_idx[%d]=%ld, fobj.start_idx[%d]=%ld", fname, m, pp->start_idx[m], m, fobj.start_idx[m]);
              chk_flag = 1;
              break;
            }
          }
          if( chk_flag != 0 ) {
            flg = 1;
          }
        }
        for(m=0; m<fobj.dims; m++) {
          fobj.start_idx[m] = pp->start_idx[m];
/**** 2009/10/08 delete @hira
          if( fobj.start_idx[m] == 0 ) {
            obj->pitch[m] = fobj.pitch[m];
            obj->origin[m] = fobj.origin[m] + obj->start_idx[m] * obj->pitch[m];
          }
****/
        }
      }
      else {
        if( has_ext_record == SplFalse ) {
          if (file_list->list_size > 1) {
            // error
            SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "not set start_idx in file_list");
            flg = 1;
          }
        }
      }

      // check overlap area in read file
      if( mode == 0 ) {
        size_t fobj_sta_idx[3] = {0, 0, 0};
        size_t fobj_end_idx[3] = {0, 0, 0};
        size_t obj_sta_idx[3] = {0, 0, 0};
        size_t obj_end_idx[3] = {0, 0, 0};
        size_t ov_start_idx[3] = {0, 0, 0};
        size_t ov_size[3] = {0, 0, 0};
        // GC領域分を加算して領域の重なりを求める。
        for(m=0; m<fobj.dims; m++) {
          fobj_sta_idx[m] = (fobj.start_idx[m] <= fobj.gc) ? 0 : fobj.start_idx[m] - fobj.gc;
          obj_sta_idx[m] = (obj->start_idx[m] <= obj->gc) ? 0 : obj->start_idx[m] - obj->gc;
          // sizeはGC込みサイズ(eidx = sidx + (size-2*gc)-1   +  GC領域分を加算)
          fobj_end_idx[m] = fobj.start_idx[m] + fobj.size[m] + fobj.gc - 1;
          obj_end_idx[m] = obj->start_idx[m] + obj->size[m] + obj->gc - 1;
        }
        if (getOverlapAreas(
                      fobj.dims,
                      fobj_sta_idx, fobj_end_idx,
                      obj_sta_idx, obj_end_idx,
                      ov_start_idx, ov_size) != SplTrue) {
          /* next */
          pp = pp->next;
          continue;
        }
        size_t ov_len = 1;
        for(m=0; m<fobj.dims; m++) ov_len *= ov_size[m];
        if (ov_len == 0) {
          pp = pp->next;
          continue;
        }
      }

      // set header from file
      for(m=0; m<fobj.dims; m++) {
        obj->pitch[m] = fobj.pitch[m];
        obj->origin[m] = fobj.origin[m]
                         + ((long long)obj->start_idx[m]-(long long)fobj.start_idx[m]) * fobj.pitch[m];
        obj->wsize[m] = fobj.wsize[m];
      }
      obj->rlen = fobj.rlen;
      obj->crddef = fobj.crddef;
      obj->aux = fobj.aux;
      obj->block_size = fobj.block_size;
      obj->step = fobj.step;
      obj->time = fobj.time;
    }

/**** delete 2009/10/10 by @hira
      // check skip read file. (only mode=0)
      if( mode == 0 ) {
        int chk_flag = 0;
        for(m=0; m<fobj.dims; m++) {
          if(   ( obj->start_idx[m] > (fobj.start_idx[m]+fobj.size[m]) )
             || ( (obj->start_idx[m]+obj->size[m]) < fobj.start_idx[m] ) ) {
            chk_flag = 1;
          }
        }
        if( chk_flag == 1 ) {
          pp = pp->next;
          continue;
        }
      }
****/

    // check error between all nodes
//    SPL_Allreduce(&flg, &tmp, 1, SPL_INT, SPL_SUM, grp);
//    if( tmp != 0 ) return SplFalse;
    if (flg != 0) {
      spl_print_error_message();
      SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
      return SplFalse;
    }

    fp = NULL;
    /* open file */
    if( (mode == 0) || (myrank == readRank) ) {
      if( (fp = fopen(fname, "rb")) == NULL ) {
        SPL_ERROR_INFO(SPL_ERR_OPEN_FILE, "file name is %s.", fname);
        flg = 1;
      }
    }
    else {
      fp = NULL;
    }

    /* read data record */
    if( readSpxDataRecord(mode, readRank, grp, fp, etype, cmp_flg,
                          obj->dims, fobj.vlen, obj->vlen,
                          fobj.gc, obj->gc, fobj.rlen,
                          fobj.size, fobj.start_idx,
                          obj->size, obj->start_idx,
                          blk_sz, obj->data.c) == SplFalse ) {
      if( (mode == 0) || (myrank == readRank) ) fclose(fp);
      flg = 1;
    }

    /* file close */
    if (fp) fclose(fp);

    // check error between all nodes
//    SPL_Allreduce(&flg, &tmp, 1, SPL_INT, SPL_SUM, grp);
//    if( tmp != 0 ) return SplFalse;
    if (flg != 0) {
      spl_print_error_message();
      SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
      return SplFalse;
    }

    /* increment pp */
    pp = pp->next;
  }

  return SplTrue;
}


/**
 * \~english
 * write SPX
 * @param        fname              write file name
 * @param        mode              write mode
 *                                 - 0:all nodes write
 *                                 - 1:only writeRank write data. Output gather data.
 *                                 - 2:only writeRank write data. Output gather data.
 *                                   and writeRank has not my splSpx object.
 * @param        cmp_flag          compress flag (SplTrue:compress, SplFalse:not compress)
 * @param        ext_rec_flag      write extension recode flag (SplTrue:write, SplFalse:not write)
 * @param        block_size        compress block size, or gather send/receive buffer size
 * @param         obj             splSpx Object
 * @param        writeRank          Rank that write a file (It is effective at mode 1 or 2.)
 * @param        grp                Communication group
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * SPXファイルを書き込む。
 * @param        fname              書込ファイル名
 * @param        mode                書込モード
 *                                 - 0:すべてのノードがファイルの書き込みを行う。
 *                                 - 1:writeRankノードのみがデータ収集を行い、ファイルに書き込む。
 *                                 - 2:splSpxオブジェクトを持たないwriteRankノードがデータ収集を行い、ファイルに書き込む。
 * @param        cmp_flag            データ圧縮フラグ (SplTrue:圧縮を行う, SplFalse:圧縮を行わない)
 * @param        ext_rec_flag        拡張レコードの書込フラグ (SplTrue:書込を行う, SplFalse:書込を行わない)
 * @param        block_size          圧縮ブロックサイズ、又はモード1,2におけるデータ収集バッファーサイズ
 * @param         obj             splSpxオブジェクト
 * @param        writeRank          ファイル書込ノード番号(モード1, 2で有効)
 * @param        grp                プロセスグループ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splWriteSpx(const char* fname,    /* input */
            int mode,             /* input */
            SplBool cmp_flag,     /* input */
            SplBool ext_rec_flag, /* input */
            size_t block_size,    /* input */
            const splSpx* obj,    /* input */
            int writeRank,        /* input */
            SPL_Comm grp          /* input */ ) {

  int myrank = 0;
  FILE* fp = NULL;
  register unsigned int n;
  int flg = 0, tmp = 0;

  SPL_Datatype dt_size_t;
  if( sizeof(size_t) == 4 )      dt_size_t = SPL_INT;
  else if( sizeof(size_t) == 8 ) dt_size_t = SPL_LONG_LONG;
  else                           return SplFalse;

  /* parameter check */
  if( !obj ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "splSpx object is null.");
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

/**** delete at 2009/10/15
  if (mode != 0 && block_size == 0) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "block_size is zero, mode1 need block_size for gather.");
    flg = 1;
  }
****/

  // check splSpx object
  if (!checkSpx(obj, SplTrue)) {
    flg = 1;
  }

  // none parallel is mode = 0
  if (!splIsParallel()) mode = 0;

  if( mode == 0 )  myrank = 0;
  else             SPL_Comm_rank(grp, &myrank);

  size_t dest_len;
  dest_len = (obj->dims >= 1) ? 1 : 0;
  for(n=0; n<obj->dims; n++) dest_len *= obj->size[n];

  if ((mode == 0 &&  dest_len > 0) || (mode != 0 && myrank == writeRank)) {
    if( !fname ) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "fname is null.");
      flg = 1;
    }
    if( (fp = fopen(fname, "wb")) == NULL ) {
      SPL_ERROR_INFO(SPL_ERR_OPEN_FILE, "file name is %s", fname);
      flg = 1;
    }
  }
  else {
    fp = NULL;
  }

  // check error between all nodes
//  SPL_Allreduce(&flg, &tmp, 1, SPL_INT, SPL_SUM, grp);
//  if( tmp != 0 ) {
  if (flg != 0) {
    if( fp ) fclose(fp);
    fp = NULL;
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  size_t wsidx[3] = {0,0,0};
  size_t write_size[3] = {0,0,0};
  size_t write_sidx[3] = {0,0,0};
  if (!flg) {
    size_t wlen = 1;
    for(n=0; n<obj->dims; n++) wlen *= obj->wsize[n];

    // bcast wlen
    if (mode != 0) {
      SPL_Bcast(&wlen, 1, dt_size_t, writeRank, grp);
    }
    if (wlen == 0) {
      // not set wsize
      // get whole size and whole start index
      if (getSvWholeSize(mode, obj->dims, obj->size, obj->start_idx, obj->wsize, wsidx, write_size, write_sidx, grp) == SplFalse) {
        flg = 1;
      }
    }
    else {
      if (mode != 0) {
        for(n=0; n<obj->dims; n++) {
          wsidx[n] = 0;
          write_size[n] = obj->wsize[n];
          write_sidx[n] = obj->start_idx[n];
        }
      }
      else {
        for(n=0; n<obj->dims; n++) {
          wsidx[n] = obj->start_idx[n];
          write_size[n] = obj->size[n];
          write_sidx[n] = obj->start_idx[n];
        }
      }
    }

    if (mode != 0) {
      // calculate origin/pitch coords
      if (getVoxelCoords(obj->dims, obj->size, obj->start_idx, wsidx, obj->origin, obj->pitch, grp) == SplFalse) {
        flg = 1;
      }
    }
  }

  // set block_size for gather
  if (mode != 0 && block_size == 0) {
    block_size = 1;
    for(n=0; n<obj->dims; n++) block_size *= write_size[n];
  }

  // check error between all nodes
//  SPL_Allreduce(&flg, &tmp, 1, SPL_INT, SPL_SUM, grp);
//  if( tmp != 0 ) {
  if (flg != 0) {
    if( fp ) fclose(fp);
    fp = NULL;
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  if (!flg) {
    size_t cmp_size = (cmp_flag) ? block_size : 0;
    if( writeSpxHeaderRecord(fp, mode, writeRank, grp,
                             obj->dims, obj->vlen, obj->dtype, obj->gc,
                             obj->rlen, obj->crddef, obj->aux, cmp_size,
                             write_size, obj->origin, obj->pitch,
                             obj->step, obj->time) == SplFalse ) {
//      SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "writeSpxHeaderRecord error.");
      if( fp ) fclose(fp);
      fp = NULL;
      flg = 1;
    }
  }

  if (!flg) {
    if( writeSpxDataRecord(mode, writeRank, grp, fp, cmp_flag, block_size,
                           obj->dims, obj->vlen, obj->rlen,
                           obj->size, obj->data.c,
                           write_size, write_sidx) == SplFalse) {
//      SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "writeSpxDataRecord error.");
      if( fp ) fclose(fp);
      fp = NULL;
      flg = 1;
    }
  }

  if (!flg) {
    if( ext_rec_flag ) {
      if( writeSpxExtRecord(fp, mode, writeRank, grp, obj->dims,
                            obj->wsize, wsidx) == SplFalse ) {
        if( fp ) fclose(fp);
        fp = NULL;
        flg = 1;
      }
    }
  }

  if( fp ) fclose(fp);
  fp = NULL;

  // check error between all nodes
//  SPL_Allreduce(&flg, &tmp, 1, SPL_INT, SPL_SUM, grp);
//  if( tmp != 0 ) return SplFalse;
  if (flg != 0) {
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  return SplTrue;
}


/**
 * \~english
 * get SPX File Information
 * @param[in]      fname            file name
 * @param[out]  obj             splSpx Object
 * @param[in]      mode              read mode
 *                                 - 0:all nodes read
 *                                 - 1:only readRank read data
 *                                 - 2:only readRank read data, and readRank not expand data.
 * @param[in]      readRank          Rank that reads a file (It is effective at mode 1 or 2.)
 * @param[in]      grp                Communication group
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * SPXファイル情報を取得する。
 * @param[in]      fname            ファイル名
 * @param[out]  obj             splSpxオブジェクト
 * @param[in]      mode              読込モード
 *                                 - 0:すべてのノードにて読み込みを行う。
 *                                 - 1:readRankノードが読み込みを行う。
 *                                 - 2:only readRank read data, and readRank not expand data.
 * @param[in]      readRank          ファイル読込ノード番号(モード1, 2で有効)
 * @param[in]      grp                プロセスグループ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splGetSpxInfo(const char* fname,            /* input  */
              splSpx* obj,                  /* output */
              int mode,                     /* input  */
              int readRank,                 /* input  */
              SPL_Comm grp                  /* input  */ ) {
  EMatchType etype;
  unsigned long long blk_sz, cmp_flg;
  SplBool has_ext_record;

  if( getSpxInfo(mode, readRank, grp, fname, obj,
                 &etype, &blk_sz, &cmp_flg, &has_ext_record) == SplFalse ) {
    spl_print_error_message();
    SPL_Abort(SPL_COMM_WORLD, spl_get_errcode());
    return SplFalse;
  }

  return SplTrue;
}

#if 1

/**
 * \~english
 * print SPX Information
 * @param        obj             splSpx Object
 * @param        grp                Communication group
 *
 * \~japanese
 * SPXファイル情報を表示する。
 * @param        obj             splSpxオブジェクト
 * @param        grp                プロセスグループ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
void
splPrintSpx(const splSpx* obj, SPL_Comm grp) {
  int rank;
  if( !obj ) return;
  SPL_Comm_rank(grp, &rank);
  fprintf(stdout, "[%d] splSpx Object {\n", rank);
  fprintf(stdout, " dims      = %d\n", (int)obj->dims);
  fprintf(stdout, " wsize     = (%d, %d, %d)\n",
          (int)obj->wsize[0], (int)obj->wsize[1], (int)obj->wsize[2]);
  fprintf(stdout, " start_idx = (%d, %d, %d)\n",
          (int)obj->start_idx[0],(int)obj->start_idx[1],(int)obj->start_idx[2]);  fprintf(stdout, " size      = (%d, %d, %d)\n",
          (int)obj->size[0], (int)obj->size[1], (int)obj->size[2]);
  fprintf(stdout, " origin    = (%e, %e, %e)\n",
          obj->origin[0], obj->origin[1], obj->origin[2]);
  fprintf(stdout, " pitch     = (%e, %e, %e)\n",
          obj->pitch[0], obj->pitch[1], obj->pitch[2]);
  fprintf(stdout, " vlen      = %d\n", (int)obj->vlen);
  fprintf(stdout, " dtype     = %d\n", (int)obj->dtype);
  fprintf(stdout, " gc        = %d\n", (int)obj->gc);
  fprintf(stdout, " rlen      = %d\n", (int)obj->rlen);
  fprintf(stdout, " crddef    = %d\n", (int)obj->crddef);
  fprintf(stdout, " aux       = %d\n", (int)obj->aux);
  fprintf(stdout, " data      = %p\n", obj->data.c);
  fprintf(stdout, " step      = %d\n", (int)obj->step);
  fprintf(stdout, " time      = %e\n", obj->time);
  fprintf(stdout, "}\n");
  return;
}
#endif

#if defined(__cplusplus)
}
#endif /* extern "C" */

