/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file splfio_filetype_check.c
 * Implementation of endian utility procedure.
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file splfio_filetype_check.c
 * SBX/SPXファイルのエンディアンユーティリティソースファイル
 * @author     hira
 * @date       2009/02/20
 */


#include <strings.h>
#include "splfio_filetype_check.h"
#include "spl/spl.h"
#include "spl/spl_errno.h"

#if defined(__cplusplus)
extern "C" {
#endif

//extern int spl_err_code;

/**
 * \~english
 * SBX Bytes Data endian check procedure
 * @param        magick        SBX Byte data
 * @return    EMatchType
 *                 - Match:little endian
 *                 - UnMatch:big endian
 *                 - UnKnown:error
 *
 * \~japanese
 * SBXバイトデータチェック及びエンディアンチェックを行う
 * @param        magick        SBX バイトデータ
 * @return    EMatchType
 *                 - Match:little endian
 *                 - UnMatch:big endian
 *                 - UnKnown:error
 */
EMatchType
isMatchEndianSbxMagick(int magick) {

  if( strncasecmp((char*)&magick, SPL_SBX_MAGICK_STR, 4) == 0 ) {
    return Match;
  }
/**** modify at 2009/10/15 by @hira
  else if( strncasecmp((char*)&magick, SPL_SBX_MAGICK_STR_REV, 4) == 0 ) {
      return UnMatch;
  }
****/
  else {
    char* magick_rev = (char*)&magick;
    BSwap32(magick_rev);
    if( strncasecmp(magick_rev, SPL_SBX_MAGICK_STR, 4) == 0 ) {
      return UnMatch;
    }
  }
  return UnKnown;
}

/**
 * \~english
 * SBX File endian check procedure
 * @param        fname      file name
 * @param        grp        Communication group
 * @return    EMatchType
 *                 - Match:little endian
 *                 - UnMatch:big endian
 *                 - UnKnown:error
 *
 * \~japanese
 * SBXファイルチェック及びエンディアンチェックを行う
 * @param        fname      ファイル名
 * @param        grp        MPIプロセスグループ
 * @return    EMatchType
 *                 - Match:little endian
 *                 - UnMatch:big endian
 *                 - UnKnown:error
 */
EMatchType
isMatchEndianSbxFile(const char* fname, SPL_Comm grp) {
  int magick;
  int rank;
  SPL_Comm_rank(grp, &rank);
  if( rank == 0 ) {
    FILE* fp;
    if( !fname ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "fname is null.");
      return UnKnown;
    }
    if( (fp = fopen(fname, "rb")) == NULL ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_OPEN_FILE, "fname is %s.", fname);
      return UnKnown;
    }
    if( fread(&magick, sizeof(int), 1, fp) != 1 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "fread error, fname is %s.", fname);
      return UnKnown;
    }
    fclose(fp);
  }
  SPL_Bcast(&magick, 1, SPL_INT, 0, grp);
/**** modify at 2009/10/15 by @hira
  if( strncasecmp((char*)&magick, SPL_SBX_MAGICK_STR, 4) == 0 ) {
    return Match;
  }
  else if( strncasecmp((char*)&magick, SPL_SBX_MAGICK_STR_REV, 4) == 0 ) {
    return UnMatch;
  }
  else return UnKnown;
****/
  return isMatchEndianSbxMagick(magick);
}


/**
 * \~english
 * SPX Bytes Data endian check procedure
 * @param        magick        SPX Byte data
 * @return    EMatchType
 *                 - Match:little endian
 *                 - UnMatch:big endian
 *                 - UnKnown:error
 *
 * \~japanese
 * SPXバイトデータチェック及びエンディアンチェックを行う
 * @param        magick        SPX バイトデータ
 * @return    EMatchType
 *                 - Match:little endian
 *                 - UnMatch:big endian
 *                 - UnKnown:error
 */
EMatchType
isMatchEndianSpxMagick(int magick) {

  if( strncasecmp((char*)&magick, SPL_SPX_MAGICK_STR, 4) == 0 ) {
    return Match;
  }

/**** modify at 2009/10/15 by @hira
  else if( strncasecmp((char*)&magick, SPL_SPX_MAGICK_STR_REV, 4) == 0 ) {
      return UnMatch;
  }
****/
  else {
    char* magick_rev = (char*)&magick;
    BSwap32(magick_rev);
    if( strncasecmp(magick_rev, SPL_SPX_MAGICK_STR, 4) == 0 ) {
      return UnMatch;
    }
  }

  return UnKnown;
}


/**
 * \~english
 * SPX File endian check procedure
 * @param        fname      file name
 * @param        grp        Communication group
 * @return    EMatchType
 *                 - Match:little endian
 *                 - UnMatch:big endian
 *                 - UnKnown:error
 *
 * \~japanese
 * SPXファイルチェック及びエンディアンチェックを行う
 * @param        fname      ファイル名
 * @param        grp        MPIプロセスグループ
 * @return    EMatchType
 *                 - Match:little endian
 *                 - UnMatch:big endian
 *                 - UnKnown:error
 */
EMatchType
isMatchEndianSpxFile(const char* fname, SPL_Comm grp) {
  int magick;
  int rank;
  SPL_Comm_rank(grp, &rank);
  if( rank == 0 ) {
    FILE* fp;
    if( !fname ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "fname is null.");
      return UnKnown;
    }
    if( (fp = fopen(fname, "rb")) == NULL ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_OPEN_FILE, "fname is %s.", fname);
      return UnKnown;
    }
    if( fread(&magick, sizeof(int), 1, fp) != 1 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "fread error, fname is %s.", fname);
      return UnKnown;
    }
    fclose(fp);
  }
  SPL_Bcast(&magick, 1, SPL_INT, 0, grp);
/**** modify at 2009/10/15 by @hira
  if( strncasecmp((char*)&magick, SPL_SPX_MAGICK_STR, 4) == 0 ) {
    return Match;
  }
  else if( strncasecmp((char*)&magick, SPL_SPX_MAGICK_STR_REV, 4) == 0 ) {
    return UnMatch;
  }
  else return UnKnown;
****/
  return isMatchEndianSpxMagick(magick);
}


/**
 * \~english
 * Bytes Data endian check procedure
 * @param        magick        SPX Byte data
 * @param        ftype        FileType(SBX or SPX)
 * @return    EMatchType
 *                 - Match:little endian
 *                 - UnMatch:big endian
 *                 - UnKnown:error
 *
 * \~japanese
 * バイトデータチェック及びエンディアンチェックを行う
 * @param        magick        SPX バイトデータ
 * @param        ftype        ファイルタイプ(SBX or SPX)
 * @return    EMatchType
 *                 - Match:little endian
 *                 - UnMatch:big endian
 *                 - UnKnown:error
 */
EMatchType
isMatchEndianSvMagick(int magick, SplFileType ftype) {
  if( ftype == SPH_FMT_SBX ) {
    return isMatchEndianSbxMagick(magick);
  }
  else if( ftype == SPH_FMT_SPX ) {
    return isMatchEndianSpxMagick(magick);
  }
  return UnKnown;
}


/**
 * \~english
 * File format check procedure
 * @param        fname      file name
 * @param        grp        Communication group
 * @return    SBX File(=SplTrue)/not SBX File(=SplFalse)
 *
 * \~japanese
 * SPXファイルチェックを行う。
 * @param        fname      ファイル名
 * @param        grp        MPIプロセスグループ
 * @return    SBX File(=SplTrue)/not SBX File(=SplFalse)
 */
SplBool
isSbxFile(const char* fname, SPL_Comm grp) {
  int magick;
  int rank;
  SPL_Comm_rank(grp, &rank);
  if( rank == 0 ) {
    FILE* fp;
    if( !fname ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "fname is null.");
      return SplFalse;
    }
    if( (fp = fopen(fname, "rb")) == NULL ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_OPEN_FILE, "fname is %s.", fname);
      return SplFalse;
    }
    if( fread(&magick, sizeof(int), 1, fp) != 1 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "fread error, fname is %s.", fname);
      return SplFalse;
    }
    fclose(fp);
  }
  SPL_Bcast(&magick, 1, SPL_INT, 0, grp);

/**** modify at 2009/10/15 by @hira
  if( strncasecmp((char*)&magick, SPL_SBX_MAGICK_STR, 4) != 0 ) {
    if( strncasecmp((char*)&magick, SPL_SBX_MAGICK_STR_REV, 4) != 0 ) {
      return SplFalse;
    }
  }
****/
  if (isMatchEndianSbxMagick(magick) == UnKnown) {
    return SplFalse;
  }
  return SplTrue;
}

#if defined(__cplusplus)
}
#endif /* extern "C" */

