/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file splfio_sbx_io.
 * Implementation of SBX utility subroutine.
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file splfio_sbx_io.c
 * SBXファイルの低レベル入出力関数ソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#include <strings.h>
#include <stdlib.h>
#include <zlib.h>
#include "spl/spl.h"
#include "spl/util_endian.h"
#include "spl/splmpi.h"
#include "spl/splfio_sbx.h"
#include "splfio_sv_read.h"
#include "splfio_sv_write.h"
#include "splfio_sv_util.h"

#ifdef _DEBUGLOG_
#include "debugLog.h"
#endif

#define SPL_SBX_ATTR_RECORD_SIZE       40 /* bytes */
#define SPL_SBX_SIZE_RECORD_SIZE       24 /* bytes */
#define SPL_SBX_ORIGIN_RECORD_ELEM_NUM  3
#define SPL_SBX_PITCH_RECORD_ELEM_NUM   3

#if defined(__cplusplus)
extern "C" {
#endif


/**
 * \~english
 * copy splSbx Object
 * @param[out]   dst        destination splSbx Object
 * @param[in]      src        source splSbx Object
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSbxオブジェクトをコピーする。
 * @param[out]   dst        コピー先splSbxオブジェクト
 * @param[in]      src        コピー元splSbxオブジェクト
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
copySbx(splSbx* dst, const splSbx* src) {
  if( !dst || !src ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "dst or src is null.");
    return SplFalse;
  }

  register int i;
  for(i=0; i<3; i++) {
    dst->wsize[i] = src->wsize[i];
    dst->start_idx[i] = src->start_idx[i];
    dst->size[i] = src->size[i];
    dst->origin[i] = src->origin[i];
    dst->pitch[i] = src->pitch[i];
  }
  dst->dims = src->dims;
  dst->vlen = src->vlen;
  dst->dtype = src->dtype;
  dst->gc = src->gc;
  dst->rlen = src->rlen;
  dst->crddef = src->crddef;
  dst->aux = src->aux;
#if 0 /* SBX has no time-record */
  dst->step = src->step;
  dst->time = src->time;
#endif
  if( src->data.c && dst->data.c ) {
    if(    (src->size[0] <= dst->size[0])
        && (src->size[1] <= dst->size[1])
        && (src->size[2] <= dst->size[2]) ) {
      size_t len = src->dtype;
      for (i=0; i<dst->dims;i++) len *= dst->size[i];
      memcpy(dst->data.c, src->data.c, len);
    }
    else {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_ARRAY_SIZE, "src->size[%ld,%ld,%ld]<=dst->size[%ld,%ld,%ld]", src->size[0], src->size[1], src->size[2], dst->size[0], dst->size[1], dst->size[2]);
      return SplFalse;
    }
  }
  else {
    // error
    SPL_ERROR_INFO(SPL_ERR_DATA_IS_NULL, "src->data or dst->data is null");
    return SplFalse;
  }

  return SplTrue;
}


/**
 * \~english
 * get Sbx Information
 * @param[in]             mode              read mode
 * @param[in]            readRank           Rank that reads a file (It is effective at mode 1 or 2.)
 * @param[in]            grp                Communication group
 * @param[in]            fname              file name
 * @param[out]        obj                splSbx Object
 * @param[out]        etype              Endian match type
 * @param[out]        blk_sz            block size
 * @param[out]        cmp_flg            compress flag
 * @param[out]        has_ext_record    has extend record
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * SBXファイルから情報を取得する。
 * @param[in]             mode              読込モード
 * @param[in]            readRank           ファイル読込ノード番号(モード1, 2で有効)
 * @param[in]            grp                プロセスグループ
 * @param[in]            fname              ファイル名
 * @param[out]        obj                splSbxオブジェクト
 * @param[out]        etype              エンディアンタイプ
 * @param[out]        blk_sz            ブロックサイズ
 * @param[out]        cmp_flg            圧縮フラグ
 * @param[out]        has_ext_record    拡張レコードの有無
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
getSbxInfo(int mode,                    /* input  */
           int readRank,                /* input  */
           SPL_Comm grp,                /* input  */
           const char* fname,           /* input  */
           splSbx* obj,                 /* output */
           EMatchType* etype,           /* output */
           unsigned long long* blk_sz,  /* output */
           unsigned long long* cmp_flg, /* output */
           SplBool* has_ext_record      /* output */ ) {
  FILE* fp = NULL;
  int myrank;
  int flg = 0;
  int tmp = 0;
  register int n;
  long offset;

  if( !obj || !etype || !blk_sz || !cmp_flg || !has_ext_record ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "obj,etype,blk_sz,cmp_flg or has_ext_record is null");
    return SplFalse;
  }

  if( mode == 0 )  myrank = 0;
  else             SPL_Comm_rank(grp, &myrank);

  if( (mode == 0) || (myrank == readRank) ) {
    /* read header */
    if( !fname ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "fname is null");
      flg = 1;
    }
    else {
      if ( (fp = fopen(fname, "rb")) == NULL ) {
        SPL_ERROR_INFO(SPL_ERR_OPEN_FILE, "[node=%d] file name = %s.", myrank, fname);
        flg = 1;
      }
    }
  }

//  tmp = 0;
//  SPL_Allreduce(&flg, &tmp, 1, SPL_INT, SPL_SUM, grp);
//  if( tmp != 0 ) {
//    return SplFalse;
//  }
  if (flg != 0) {
    return SplFalse;
  }

  if( (mode == 0) || (myrank == readRank) ) {
    if (!flg) {
      if( readSvAttrRecord(fp,
                          SPH_FMT_SBX,
                          etype,
                          &obj->dims,
                          &obj->vlen,
                          &obj->dtype,
                          &obj->gc,
                          &obj->rlen,
                          &obj->crddef,
                          &obj->aux,
                          blk_sz) == SplFalse ) {
        flg = 1;
      }
    }
    if (!flg) {
      offset = SPL_SBX_ATTR_RECORD_SIZE;
      if( readSvSizeRecord(fp, offset, *etype, obj->size) == SplFalse ) {
        flg = 1;
      }
    }

    if (!flg) {
      offset += SPL_SBX_SIZE_RECORD_SIZE;
      if( readSvOriginRecord(fp, offset, *etype,
                            obj->rlen, obj->origin) == SplFalse ) {
        flg = 1;
      }
    }
    if (!flg) {
      offset += SPL_SBX_ORIGIN_RECORD_ELEM_NUM*obj->rlen;
      if( readSvPitchRecord(fp, offset, *etype,
                            obj->rlen, obj->pitch) == SplFalse ) {
        flg = 1;
      }
    }

    if (!flg) {
      /* set compress flag (blk_sz -> 0:uncompress, other:compress) */
      *cmp_flg = *blk_sz;

      offset += SPL_SBX_PITCH_RECORD_ELEM_NUM*obj->rlen;
      if( readSvExtRecord(fp, offset, *etype,
                          obj->rlen, obj->wsize,
                          obj->start_idx, has_ext_record) == SplFalse ) {
        flg = 1;
      }
    }
    if (fp) fclose(fp);
  }

//  tmp = 0;
//  SPL_Allreduce(&flg, &tmp, 1, SPL_INT, SPL_SUM, grp);
//  if( tmp != 0 ) {
//    return SplFalse;
//  }
  if (flg != 0) {
    return SplFalse;
  }

  /* broadcast header */
  if( mode != 0 ) {
    size_t size_t_buf[10];
    unsigned int int_buf[8];
    unsigned long long ll_buf[2];
    double double_buf[6];

    SPL_Datatype dt_size_t;
    if( sizeof(size_t) == 4 )      dt_size_t = SPL_INT;
    else if( sizeof(size_t) == 8 ) dt_size_t = SPL_LONG_LONG;
    else                           return SplFalse;

    if( myrank == readRank ) {
      size_t_buf[0] = obj->wsize[0];
      size_t_buf[1] = obj->wsize[1];
      size_t_buf[2] = obj->wsize[2];
      size_t_buf[3] = obj->start_idx[0];
      size_t_buf[4] = obj->start_idx[1];
      size_t_buf[5] = obj->start_idx[2];
      size_t_buf[6] = obj->gc;
      size_t_buf[7] = obj->size[0];
      size_t_buf[8] = obj->size[1];
      size_t_buf[9] = obj->size[2];
    }
    SPL_Bcast(size_t_buf, 10, dt_size_t, readRank, grp);
    if( myrank != readRank ) {
      obj->wsize[0]     = size_t_buf[0];
      obj->wsize[1]     = size_t_buf[1];
      obj->wsize[2]     = size_t_buf[2];
      obj->start_idx[0] = size_t_buf[3];
      obj->start_idx[1] = size_t_buf[4];
      obj->start_idx[2] = size_t_buf[5];
      obj->gc           = size_t_buf[6];
      obj->size[0]      = size_t_buf[7];
      obj->size[1]      = size_t_buf[8];
      obj->size[2]      = size_t_buf[9];
    }

    if( myrank == readRank ) {
      int_buf[0] = obj->dims;
      int_buf[1] = obj->vlen;
      int_buf[2] = obj->dtype;
      int_buf[3] = obj->rlen;
      int_buf[4] = obj->crddef;
      int_buf[5] = obj->aux;
      int_buf[6] = *etype;
      int_buf[7] = *has_ext_record;
    }
    SPL_Bcast(int_buf, 8, SPL_INT, readRank, grp);
    if( myrank != readRank ) {
      obj->dims       = int_buf[0];
      obj->vlen       = int_buf[1];
      obj->dtype      = int_buf[2];
      obj->rlen       = int_buf[3];
      obj->crddef     = int_buf[4];
      obj->aux        = int_buf[5];
      *etype          = (EMatchType)int_buf[6];
      *has_ext_record = int_buf[7];
    }

    if( myrank == readRank ) {
      double_buf[0] = obj->origin[0];
      double_buf[1] = obj->origin[1];
      double_buf[2] = obj->origin[2];
      double_buf[3] = obj->pitch[0];
      double_buf[4] = obj->pitch[1];
      double_buf[5] = obj->pitch[2];
    }
    SPL_Bcast(double_buf, 6, SPL_DOUBLE, readRank, grp);
    if( myrank != readRank ) {
      obj->origin[0] = double_buf[0];
      obj->origin[1] = double_buf[1];
      obj->origin[2] = double_buf[2];
      obj->pitch[0]  = double_buf[3];
      obj->pitch[1]  = double_buf[4];
      obj->pitch[2]  = double_buf[5];
    }

    if( myrank == readRank ) {
      ll_buf[0] = *blk_sz;
      ll_buf[1] = *cmp_flg;
    }
    SPL_Bcast(ll_buf, 2, SPL_LONG_LONG, readRank, grp);
    if( myrank != readRank ) {
      *blk_sz = ll_buf[0];
      *cmp_flg = ll_buf[1];
    }
  }
  obj->block_size = *blk_sz;

  return SplTrue;
}


/**
 * read SBX data record
 *
 * Attension!
 *   The buffer(comp_buf and exp_buf) is not allocated in this subroutine.
 * @param[in]     mode              read mode
 * @param[in]      readRank         Rank that reads a file (It is effective at mode 1 or 2.)
 * @param[in]      grp              Communication group
 * @param[in]      fp                file pointer
 * @param[in]      eType            Endian match type
 * @param[in]      cmp_flg          compress flag
                                    - 0     : data block is uncompressed
                                    - other : data block is compressed
 * @param[in]      dims              Dimension
 * @param[in]      file_vlen        Vector length in file
 * @param[in]      dest_vlen        Vector length in output array(dst)
 * @param[in]      dtype            Data type(1or2or4)
 * @param[in]      file_gc          Size of guide cell of data described in file
 * @param[in]      dest_gc          Size of guide cell of "dst"
 * @param[in]      rlen              Real size(4or8)
 * @param[in]      file_vsz          Size of Voxel described in file
 * @param[in]      file_vsta        Start index of data described in file
 * @param[in]      dest_vsz          Size of Voxel of "dst"
 * @param[in]      dest_vsta        Start index of "dst"
 * @param[in]      vox_num_per_blk  Size of number of voxel  in a compressed data record.
 * @param[out]   dst              Output data array
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSbxDataRecord(int mode,
                  int readRank,
                  SPL_Comm grp,
                  FILE* fp,
                  EMatchType eType,
                  unsigned long long cmp_flg,
                  unsigned int dims,
                  unsigned int file_vlen,
                  unsigned int dest_vlen,
                  unsigned int dtype,
                  size_t file_gc,
                  size_t dest_gc,
                  unsigned int rlen,
                  size_t file_vsz[3],
                  size_t file_vsta[3],
                  size_t dest_vsz[3],
                  size_t dest_vsta[3],
                  size_t vox_num_per_blk,
                  unsigned char* dst) {
  long file_offset;

  /* parameter check */
  if( !file_vsz || !file_vsta || !dest_vsz || !dest_vsta ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "file_vsz, file_vsta, dest_vsz or dest_vsta is null");
    return SplFalse;
  }

  if( fp ) {
    /* seek file */
    file_offset =  SPL_SBX_ATTR_RECORD_SIZE + SPL_SBX_SIZE_RECORD_SIZE
                   + (  SPL_SBX_ORIGIN_RECORD_ELEM_NUM
                      + SPL_SBX_PITCH_RECORD_ELEM_NUM  ) * rlen
#if 0 /* SBX has no time-record */
            + (8 + (SPL_SBX_TIME_RECORD_ELEM_NUM - 1) * rlen )
#endif
            ;

    if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek error");
      return SplFalse;
    }
  }

  // read data record
  if (!readSvDataRecord(
                        mode,
                        readRank,
                        grp,
                        fp,
                        eType,
                        cmp_flg,
                        dims,
                        file_vlen,
                        dest_vlen,
                        dtype,
                        file_gc,
                        dest_gc,
                        file_vsz,
                        file_vsta,
                        dest_vsz,
                        dest_vsta,
                        vox_num_per_blk,
                        dst)) {
    return SplFalse;
  }
  return SplTrue;
}


/**
 * write header record in SBX File
 *
 * @param[in]      fp                file pointer
 * @param[in]       mode              write mode
 * @param[in]      writeRank         Rank that write a file (It is effective at mode 1 or 2.)
 * @param[in]      grp              Communication group
 * @param[in]      dims              Dimension
 * @param[in]      vlen              Vector length
 * @param[in]      dtype            Data type(1or2or4)
 * @param[in]      gc                Size of guide cell
 * @param[in]      rlen              Real size(4or8)
 * @param[in]      crddef            coords define
 * @param[in]      aux              extension data
 * @param[in]      block_size       Size of number of voxel  in a compressed data record.
 * @param[in]      size              size of voxel
 * @param[in]      org              origin of voxel
 * @param[in]      pit              pitch of voxel
 * @param[out]         wsize            total size of nodes
 * @param[in]       sta_idx          gloval start index of nodes
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSbxHeaderRecord(FILE* fp,
                     int mode,
                     int writeRank,
                     SPL_Comm grp,
                     unsigned int dims,
                     unsigned int vlen,
                     unsigned int dtype,
                     size_t gc,
                     unsigned int rlen,
                     unsigned int crddef,
                     unsigned int aux,
                     unsigned long long block_size,
                     const size_t* size,
                     const double* org,
                     const double* pit
                     ) {
  long file_offset;
  int myrank = 0;
  register unsigned int n;

  if( !size || !org || !pit ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "size, org or pit is null");
    return SplFalse;
  }

  if( mode == 0 )  myrank = 0;
  else             SPL_Comm_rank(grp, &myrank);

  if (fp) {
    if( writeSvAttrRecord(fp, SPH_FMT_SBX, dims, vlen, dtype, gc,
                          rlen, crddef, aux, block_size) == SplFalse ) {
      return SplFalse;
    }
  }

  if( fp ) {
    file_offset = SPL_SBX_ATTR_RECORD_SIZE;
    if( writeSvSizeRecord(fp, file_offset, dims, size) == SplFalse ) {
      return SplFalse;
    }
  }

  if( fp ) {
    file_offset += SPL_SBX_SIZE_RECORD_SIZE;
    if( writeSvOriginRecord(fp, file_offset, dims, rlen, org) == SplFalse ) {
      return SplFalse;
    }
  }

  if( fp ) {
    file_offset += SPL_SBX_ORIGIN_RECORD_ELEM_NUM*rlen;
    if( writeSvPitchRecord(fp, file_offset, dims, rlen, pit) == SplFalse ) {
      return SplFalse;
    }
  }

  return SplTrue;
}



/**
 * write extension record in SBX File
 *
 * @param        fp                file pointer
 * @param         mode              write mode
 * @param        writeRank         Rank that write a file (It is effective at mode 1 or 2.)
 * @param        grp              Communication group
 * @param        dims              Dimension
 * @param         wsize            total size of nodes
 * @param         sta_idx          gloval start index of nodes
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSbxExtRecord(FILE* fp,
                  int mode,
                  int writeRank,
                  SPL_Comm grp,
                  unsigned int dims,
                  const size_t* wsize,
                  const size_t* sta_idx ) {
  int myrank;

  if( !wsize || !sta_idx ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "wsize or sta_idx is null");
    return SplFalse;
  }

  if( mode == 0 )  myrank = 0;
  else             SPL_Comm_rank(grp, &myrank);

  if( fp) {
    if( writeSvExtRecord(fp, dims, wsize, sta_idx) == SplFalse ) {
      return SplFalse;
    }
  }

  return SplTrue;
}


/**
 * write header record in SBX File
 *
 * @param         mode              write mode
 * @param        writeRank         Rank that write a file (It is effective at mode 1 or 2.)
 * @param        grp              Communication group
 * @param        fp                file pointer
 * @param        cmp_flg          compress flag
 *                                   - 0     : data block is uncompressed
 *                                   - other : data block is compressed
 * @param        vox_num_per_blk   Size of number of voxel  in a compressed data record.
 * @param        dims              Dimension
 * @param        vlen              Vector length
 * @param        dtype            Data type(1or2or4)
 * @param        vsz              size of voxel
 * @param        data              data of voxel
 * @param         wsize            total size of nodes
 * @param         sta_idx          gloval start index of nodes
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSbxDataRecord(int mode,
                   int writeRank,
                   SPL_Comm grp,
                   FILE* fp,
                   unsigned long long cmp_flg,
                   size_t vox_num_per_blk,
                   unsigned int dims,
                   unsigned int vlen,
                   unsigned int dtype,
                   const size_t* vsz,
                   const unsigned char* data,
                   const size_t* wsize,
                   const size_t* sta_idx) {

  if( !vsz || !wsize || !sta_idx) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "vsz, wsize or sta_idx is null");
    return SplFalse;
  }

  // write data record
  if (!writeSvDataRecord(
                        mode,
                        writeRank,
                        grp,
                        fp,
                        cmp_flg,
                        vox_num_per_blk,
                        dims,
                        vlen,
                        dtype,
                        vsz,
                        data,
                        wsize,
                        sta_idx)) {
    return SplFalse;
  }

  return SplTrue;
}


/**
 * \~english
 * check splSbx Object
 * @param[in]     uobj        splSbx object from user
 * @param[in]      writeFlag    check flag
 *                               - SplTrue  : for write
 *                               - SplFalse : for read
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * splSbx Objectのパラメータ値をチェックする。
 * @param[in]     uobj        ユーザ設定のsplSbxオブジェクト
 * @param[in]      writeFlag    チェック対象フラグ
 *                               - SplTrue  : 書込時チェック
 *                               - SplFalse : 読込時チェック
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
checkSbx(const splSbx* uobj, SplBool writeFlag)
{
  // size and data check
  register unsigned int n;
  size_t dest_len;

  // dims
  if (uobj->dims == 0 || uobj->dims > 3) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "dims is invalid [dims=%d]", uobj->dims);
    return SplFalse;
  }

  // vlen
  if (uobj->vlen == 0) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "vlen is invalid [vlen=%d]", uobj->vlen);
    return SplFalse;
  }
  if (uobj->vlen != 1) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "WARNNING : vlen is not 1 [vlen=%d].", uobj->vlen);
  }

  // dtype
  if (uobj->dtype != 1 && uobj->dtype != 2 && uobj->dtype != 4) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "dtype is invalid [dtype=%d].", uobj->dtype);
    return SplFalse;
  }

  if (writeFlag) {
    // rlen
    if (uobj->rlen != 4 && uobj->rlen != 8) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "rlen is invalid [rlen=%d].", uobj->rlen);
      return SplFalse;
    }

    // crddef
    if (uobj->crddef == 0 || uobj->crddef >= 5) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "crddef is invalid [crddef=%d]", uobj->crddef);
      return SplFalse;
    }
  }

  // data
  dest_len = 1;
  for(n=0; n<uobj->dims; n++) dest_len *= uobj->size[n];
  if (dest_len > 0 && uobj->data.c == NULL) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "data is null.");
    return SplFalse;
  }

  return SplTrue;

}


#if defined(__cplusplus)
}
#endif /* extern "C" */

#undef SPL_SBX_ATTR_RECORD_SIZE
#undef SPL_SBX_SIZE_RECORD_SIZE
#undef SPL_SBX_ORIGIN_RECORD_ELEM_NUM
#undef SPL_SBX_PITCH_RECORD_ELEM_NUM

