/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file splfio_sph_io.c
 * Implementation of SPH utility subroutine.
 * @author     yoh
 * @date       2009/11/27
 *
 * \~japanese
 * @file splfio_sph_io.c
 * SPHファイルの低レベル入出力関数ソースファイル
 * @author     yoh
 * @date       2009/02/27
 */

#include <strings.h>
#include <stdlib.h>
#include <zlib.h>
#include "spl/spl.h"
#include "spl/util_endian.h"
#include "spl/splmpi.h"
#include "spl/splfio_sph.h"
#include "splfio_sv_read.h"
#include "splfio_sv_write.h"
#include "splfio_sv_util.h"

#define SPL_SPH_ATTR_RECORD_SIZE       16 /* bytes */
#define SPL_SPH_SIZE_RECORD_SIZE1      20
#define SPL_SPH_SIZE_RECORD_SIZE2      32
#define SPL_SPH_TIME_RECORD_SIZE1      16
#define SPL_SPH_TIME_RECORD_SIZE2      24

#if defined(__cplusplus)
extern "C" {
#endif


/**
 * \~english
 * get SPH Information
 * @param[in]         mode         read mode
 * @param[in]         readRank     Rank that reads a file
 *                                 (It is effective at mode 1 or 2.)
 * @param[in]         grp          Communication group
 * @param[in]         fname        file name
 * @param[out]        obj          splSpx Object
 * @param[out]        etype        Endian match type
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * SPHファイルから情報を取得する。
 * @param[in]         mode         読込モード
 * @param[in]         readRank     ファイル読込ノード番号(モード1, 2で有効)
 * @param[in]         grp          プロセスグループ
 * @param[in]         fname        ファイル名
 * @param[out]        obj          splSpxオブジェクト
 * @param[out]        etype        エンディアンタイプ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
getSphInfo( int mode,                    /* input  */
            int readRank,                /* input  */
            SPL_Comm grp,                /* input  */
            const char* fname,           /* input  */
            splSpx* obj,                 /* output */
            EMatchType* etype            /* output */ ) {

  FILE* fp = NULL;
  int myrank;
  int flg = 0;
  int tmp = 0;
  long long step;
  double dtime;
  register int n;
  long offset;

  if( !obj || !etype ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "obj or etype is null");
    return SplFalse;
  }

  if( mode == 0 )  myrank = 0;
  else             SPL_Comm_rank(grp, &myrank);

  if( (mode == 0) || (myrank == readRank) ) {

    /* read header */
    if( !fname ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "fname is null");
      flg = 1;
    }
    else {
      if ( (fp = fopen(fname, "rb")) == NULL ) {
        SPL_ERROR_INFO(SPL_ERR_OPEN_FILE, "[node=%d] file name = %s.",
            myrank, fname);
        flg = 1;
      }
    }
  }

  if (flg != 0) {
    return SplFalse;
  }

  if( (mode == 0) || (myrank == readRank) ) {
    if (!flg) {
      int ibuff[4];
      if ( fread(ibuff, sizeof(int), 4, fp) != 4 ) {
        flg = 1;
      }
      if (!flg) {
        *etype = UnKnown;
        if ( ibuff[0] == 8 )
          *etype = Match;
        else {
          BSwapVec(ibuff, 4);
          if ( ibuff[0] == 8 )
            *etype = UnMatch;
          else {
            flg = 1;
            // error
            SPL_ERROR_INFO(SPL_ERR_RECODE_SIZE, "recode size error");
          }
        }
      }
      if (!flg) {
        obj->dims = 3;
        if ( ibuff[1] == 1 ) { // scalar
          obj->vlen = 1;
          obj->dtype = 1;
        } else if ( ibuff[1] == 2 ) { // vector
          obj->vlen = 3;
          obj->dtype = 3;
        } else {
          flg = 1;
          // error
          SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "invalid sph:dtype");
        }
        obj->gc = 0;
        if ( ibuff[2] == 1 ) { // single precision
          obj->rlen = 4;
        } else if ( ibuff[2] == 2 ) { // double precision
          obj->rlen = 8;
        } else {
          flg = 1;
          // error
          SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "invalid sph:rlen");
        }
        obj->crddef = 1;
        obj->aux = 0;
      }
    }

    int idmy = 0;       // レコード長

    // size
    if (!flg) {
      if ( fread(&idmy, sizeof(int), 1, fp) != 1 ) flg = 1;
      if ( obj->rlen == 4 ) {
        int ibuff[3] = {0};
        if ( fread(ibuff, sizeof(int), 3, fp) != 3 ) flg = 1;
        if ( *etype == UnMatch ) BSwapVec(ibuff, 3);
        obj->size[0] = ibuff[0];
        obj->size[1] = ibuff[1];
        obj->size[2] = ibuff[2];
      } else {
        long long lbuff[3] = {0};
        if ( fread(lbuff, sizeof(long long), 3, fp) != 3 ) flg = 1;
        if ( *etype == UnMatch ) DBSwapVec(lbuff, 3);
        obj->size[0] = lbuff[0];
        obj->size[1] = lbuff[1];
        obj->size[2] = lbuff[2];
      }
      if ( fread(&idmy, sizeof(int), 1, fp) != 1 ) flg = 1;
    }

    // origin
    if (!flg) {
      if ( fread(&idmy, sizeof(int), 1, fp) != 1 ) flg = 1;
      if ( obj->rlen == 4 ) {
        float fbuff[3] = {0};
        if ( fread(fbuff, sizeof(float), 3, fp) != 3 ) flg = 1;
        if ( *etype == UnMatch ) BSwapVec(fbuff, 3);
        obj->origin[0] = fbuff[0];
        obj->origin[1] = fbuff[1];
        obj->origin[2] = fbuff[2];
      } else {
        double dbuff[3] = {0};
        if ( fread(dbuff, sizeof(double), 3, fp) != 3 ) flg = 1;
        if ( *etype == UnMatch ) DBSwapVec(dbuff, 3);
        obj->origin[0] = dbuff[0];
        obj->origin[1] = dbuff[1];
        obj->origin[2] = dbuff[2];
      }
      if ( fread(&idmy, sizeof(int), 1, fp) != 1 ) flg = 1;
    }

    // pitch
    if (!flg) {
      if ( fread(&idmy, sizeof(int), 1, fp) != 1 ) flg = 1;
      if ( obj->rlen == 4 ) {
        float fbuff[3] = {0};
        if ( fread(fbuff, sizeof(float), 3, fp) != 3 ) flg = 1;
        if ( *etype == UnMatch ) BSwapVec(fbuff, 3);
        obj->pitch[0] = fbuff[0];
        obj->pitch[1] = fbuff[1];
        obj->pitch[2] = fbuff[2];
      } else {
        double dbuff[3] = {0};
        if ( fread(dbuff, sizeof(double), 3, fp) != 3 ) flg = 1;
        if ( *etype == UnMatch ) DBSwapVec(dbuff, 3);
        obj->pitch[0] = dbuff[0];
        obj->pitch[1] = dbuff[1];
        obj->pitch[2] = dbuff[2];
      }
      if ( fread(&idmy, sizeof(int), 1, fp) != 1 ) flg = 1;
    }

    // step, time
    if (!flg) {
      if ( fread(&idmy, sizeof(int), 1, fp) != 1 ) flg = 1;
      if ( obj->rlen == 4 ) {
        int ibuff = 0;
        float fbuff = 0;
        if ( fread(&ibuff, sizeof(int), 1, fp) != 1 ) flg = 1;
        if ( fread(&fbuff, sizeof(float), 1, fp) != 1 ) flg = 1;
        if ( *etype == UnMatch ) BSwapVec(&ibuff, 1);
        if ( *etype == UnMatch ) BSwapVec(&fbuff, 1);
        step = ibuff;
        dtime = fbuff;
      } else {
        long long lbuff = 0;
        double dbuff = 0;
        if ( fread(&lbuff, sizeof(long long), 1, fp) != 1 ) flg = 1;
        if ( fread(&dbuff, sizeof(double), 1, fp) != 1 ) flg = 1;
        if ( *etype == UnMatch ) DBSwapVec(&lbuff, 1);
        if ( *etype == UnMatch ) DBSwapVec(&dbuff, 1);
        step = lbuff;
        dtime = dbuff;
      }
      if ( fread(&idmy, sizeof(int), 1, fp) != 1 ) flg = 1;

      obj->step = (size_t)step;
      obj->time = dtime;
    }

    if (fp) fclose(fp);
  }

  if (flg != 0) {
    return SplFalse;
  }

  /* broadcast header */
  if( mode != 0 ) {
    size_t size_t_buf[11];
    unsigned int int_buf[8];
    unsigned long long ll_buf[2];
    double double_buf[7];

    SPL_Datatype dt_size_t;
    if( sizeof(size_t) == 4 )      dt_size_t = SPL_INT;
    else if( sizeof(size_t) == 8 ) dt_size_t = SPL_LONG_LONG;
    else                           return SplFalse;

    if( myrank == readRank ) {
      size_t_buf[0]  = obj->wsize[0];
      size_t_buf[1]  = obj->wsize[1];
      size_t_buf[2]  = obj->wsize[2];
      size_t_buf[3]  = obj->start_idx[0];
      size_t_buf[4]  = obj->start_idx[1];
      size_t_buf[5]  = obj->start_idx[2];
      size_t_buf[6]  = obj->gc;
      size_t_buf[7]  = obj->size[0];
      size_t_buf[8]  = obj->size[1];
      size_t_buf[9]  = obj->size[2];
      size_t_buf[10] = obj->step;
    }
    SPL_Bcast(size_t_buf, 11, dt_size_t, readRank, grp);
    if( myrank != readRank ) {
      obj->wsize[0]     = size_t_buf[0];
      obj->wsize[1]     = size_t_buf[1];
      obj->wsize[2]     = size_t_buf[2];
      obj->start_idx[0] = size_t_buf[3];
      obj->start_idx[1] = size_t_buf[4];
      obj->start_idx[2] = size_t_buf[5];
      obj->gc           = size_t_buf[6];
      obj->size[0]      = size_t_buf[7];
      obj->size[1]      = size_t_buf[8];
      obj->size[2]      = size_t_buf[9];
      obj->step         = size_t_buf[10];
    }

    if( myrank == readRank ) {
      int_buf[0] = obj->dims;
      int_buf[1] = obj->vlen;
      int_buf[2] = obj->dtype;
      int_buf[3] = obj->rlen;
      int_buf[4] = obj->crddef;
      int_buf[5] = obj->aux;
      int_buf[6] = *etype;
    }
    SPL_Bcast(int_buf, 7, SPL_INT, readRank, grp);
    if( myrank != readRank ) {
      obj->dims       = int_buf[0];
      obj->vlen       = int_buf[1];
      obj->dtype      = int_buf[2];
      obj->rlen       = int_buf[3];
      obj->crddef     = int_buf[4];
      obj->aux        = int_buf[5];
      *etype          = (EMatchType)int_buf[6];
    }

    if( myrank == readRank ) {
      double_buf[0] = obj->origin[0];
      double_buf[1] = obj->origin[1];
      double_buf[2] = obj->origin[2];
      double_buf[3] = obj->pitch[0];
      double_buf[4] = obj->pitch[1];
      double_buf[5] = obj->pitch[2];
      double_buf[6] = obj->time;
    }
    SPL_Bcast(double_buf, 7, SPL_DOUBLE, readRank, grp);
    if( myrank != readRank ) {
      obj->origin[0] = double_buf[0];
      obj->origin[1] = double_buf[1];
      obj->origin[2] = double_buf[2];
      obj->pitch[0]  = double_buf[3];
      obj->pitch[1]  = double_buf[4];
      obj->pitch[2]  = double_buf[5];
      obj->time      = double_buf[6];
    }

  }

  obj->block_size = 0;

  return SplTrue;
}

/**
 * read SPH data record
 *
 * Attension!
 *   The buffer(comp_buf and exp_buf) is not allocated in this subroutine.
 * @param[in]      mode         read mode
 * @param[in]      readRank     Rank that reads a file
 *                              (It is effective at mode 1 or 2.)
 * @param[in]      grp          Communication group
 * @param[in]      fp           file pointer
 * @param[in]      eType        Endian match type
 * @param[in]      dims         Dimension
 * @param[in]      file_vlen    Vector length in file
 * @param[in]      dest_vlen    Vector length in output array(dst)
 * @param[in]      file_gc      Size of guide cell of data described in file
 * @param[in]      dest_gc      Size of guide cell of "dst"
 * @param[in]      rlen         Real size(4or8)
 * @param[in]      file_vsz     Size of Voxel described in file
 * @param[in]      file_vsta    Start index of data described in file
 * @param[in]      dest_vsz     Size of Voxel of "dst"
 * @param[in]      dest_vsta    Start index of "dst"
 * @param[out]     dst          Output data array
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSphDataRecord(int mode,
                  int readRank,
                  SPL_Comm grp,
                  FILE* fp,
                  EMatchType eType,
                  unsigned int dims,
                  unsigned int file_vlen,
                  unsigned int dest_vlen,
                  size_t file_gc,
                  size_t dest_gc,
                  unsigned int rlen,
                  size_t file_vsz[3],
                  size_t file_vsta[3],
                  size_t dest_vsz[3],
                  size_t dest_vsta[3],
                  unsigned char* dst) {
  long file_offset;
  size_t dest_len;
  register unsigned int n;
  dest_len = 1;
  for(n=0; n<dims; n++) dest_len *= dest_vsz[n];

  /* parameter check */
  if( !file_vsz || !file_vsta || !dest_vsz || !dest_vsta ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER,
        "file_vsz, file_vsta, dest_vsz or dest_vsta is null");
    return SplFalse;
  }

  if( fp ) {
    /* seek file */
    if ( rlen == 4 )
      file_offset = SPL_SPH_ATTR_RECORD_SIZE + SPL_SPH_SIZE_RECORD_SIZE1 * 3
      + SPL_SPH_TIME_RECORD_SIZE1 + 4;
    else
      file_offset = SPL_SPH_ATTR_RECORD_SIZE + SPL_SPH_SIZE_RECORD_SIZE2 * 3
      + SPL_SPH_TIME_RECORD_SIZE2 + 4;

    if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek error");
      return SplFalse;
    }
  }

  // read data record
  {
    size_t blk_sz;
    unsigned char *blk_buf;
    int gc_offset;
    size_t file_vend[3], dest_vend[3];
    size_t ov_gloval_sidx[3], ov_gloval_size[3];
    size_t src_sz_exgc[3], dest_sz_exgc[3];
    size_t data_len = rlen;

    for(n=0; n<dims; n++) {
      src_sz_exgc[n] = file_vsz[n] - 2*file_gc;
      dest_sz_exgc[n] = dest_vsz[n] - 2*dest_gc;
    }

    /* allocate data-block buffer */
    blk_sz = sizeof(char)*data_len*file_vlen;
    for(n=0; n<dims; n++) blk_sz *= file_vsz[n];
    if( (blk_buf = (unsigned char*)malloc(sizeof(char)*blk_sz)) == NULL ) {
      SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE,
          "malloc length = %ld", sizeof(char)*blk_sz);
      return SplFalse;
    }

    /* read from file */
    if( readSvDataBlock(mode, readRank, grp, fp, eType,
        blk_sz, blk_sz, blk_buf) == SplFalse ) {
      free(blk_buf);
      return SplFalse;
    }

    if (dst != NULL && dest_len > 0) {
      // copy destination buffer
      size_t src_gloval_sidx[3], src_gloval_eidx[3];
      size_t dst_gloval_sidx[3], dst_gloval_eidx[3];

      /* corrects dst_index and set gc_offset */
      gc_offset = dest_gc - file_gc;
      for(n=0; n<dims; n++){
        src_gloval_sidx[n] = (file_vsta[n] > file_gc)?file_vsta[n] - file_gc:0;
        dst_gloval_sidx[n] = (dest_vsta[n] > dest_gc)?dest_vsta[n] - dest_gc:0;
        src_gloval_eidx[n] = src_gloval_sidx[n] + file_vsz[n] - 1;
        dst_gloval_eidx[n] = dst_gloval_sidx[n] + dest_vsz[n] - 1;
      }

      /* calc overlap area */
      if( getOverlapAreas(dims, src_gloval_sidx, src_gloval_eidx,
          dst_gloval_sidx, dst_gloval_eidx,
          ov_gloval_sidx, ov_gloval_size) == SplFalse ) {
        free(blk_buf);
        return SplFalse;
      }
      /* copy buffer */
      if( copySvBuffer(dims,
          data_len*file_vlen, data_len*dest_vlen,
          file_vsta, src_sz_exgc, file_gc,
          dest_vsta, dest_sz_exgc, dest_gc,
          ov_gloval_sidx, ov_gloval_size, 0,
          blk_buf, dst) != SplTrue ) {
        free(blk_buf);
        return SplFalse;
      }
    }     // if (dst != NULL && dest_len > 0)

    free(blk_buf);
  }

  return SplTrue;
}

/**
 * write header record in SPH File
 *
 * @param[in]      fp           file pointer
 * @param[in]      mode         write mode
 * @param[in]      writeRank    Rank that write a file
 *                              (It is effective at mode 1 or 2.)
 * @param[in]      grp          Communication group
 * @param[in]      vlen         Vector length
 * @param[in]      rlen         Real size(4or8)
 * @param[in]      size         size of voxel
 * @param[in]      org          origin of voxel
 * @param[in]      pit          pitch of voxel
 * @param[in]      step         step count
 * @param[in]      time         time
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSphHeaderRecord( FILE* fp,
                      int mode,
                      int writeRank,
                      SPL_Comm grp,
                      unsigned int vlen,
                      unsigned int rlen,
                      const size_t* size,
                      const double* org,
                      const double* pit,
                      size_t step,
                      double time) {
  long file_offset;
  int myrank = 0;
  register unsigned int n;
  int idmy = 0;       // レコード長
  int ibuff[8] = {0x00};
  float fbuff[3] = {0.0, 0.0, 0.0};
  long long lbuff[3] = {0x00};
  double dbuff[3] = {0.0, 0.0, 0.0};

  if( !size || !org || !pit) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "size, org or pit is null");
    return SplFalse;
  }

  if( mode == 0 )  myrank = 0;
  else             SPL_Comm_rank(grp, &myrank);

  if (fp) {
    if( fseek(fp, 0, SEEK_SET) != 0 ) {
      SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is 0.");
      return SplFalse;
    }

    ibuff[0] = ibuff[3] = 8;
    ibuff[1] = (vlen == 1) ? 1 : 2;
    ibuff[2] = (rlen == 4) ? 1 : 2;
    if( fwrite(ibuff, sizeof(int), 4, fp) != 4 ) {
      SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "attribute write error");
      return SplFalse;
    }
  }

  // size
  if( fp ) {
    if ( rlen == 4 ) {
      ibuff[0] = ibuff[4] = 12;
      ibuff[1] = size[0];
      ibuff[2] = size[1];
      ibuff[3] = size[2];
      if( fwrite(ibuff, sizeof(int), 5, fp) != 5 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "size write error");
        return SplFalse;
      }
    } else {
      idmy = 24;
      lbuff[0] = size[0];
      lbuff[1] = size[1];
      lbuff[2] = size[2];
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "size write error");
        return SplFalse;
      }
      if( fwrite(lbuff, sizeof(long long), 3, fp) != 3 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "size write error");
        return SplFalse;
      }
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "size write error");
        return SplFalse;
      }
    }
  }

  // origin
  if( fp ) {
    if ( rlen == 4 ) {
      idmy = 12;
      fbuff[0] = org[0];
      fbuff[1] = org[1];
      fbuff[2] = org[2];
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "org write error");
        return SplFalse;
      }
      if( fwrite(fbuff, sizeof(float), 3, fp) != 3 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "org write error");
        return SplFalse;
      }
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "org write error");
        return SplFalse;
      }
    } else {
      idmy = 24;
      dbuff[0] = org[0];
      dbuff[1] = org[1];
      dbuff[2] = org[2];
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "org write error");
        return SplFalse;
      }
      if( fwrite(dbuff, sizeof(double), 3, fp) != 3 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "org write error");
        return SplFalse;
      }
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "org write error");
        return SplFalse;
      }
    }
  }

  if( fp ) {
    if ( rlen == 4 ) {
      idmy = 12;
      fbuff[0] = pit[0];
      fbuff[1] = pit[1];
      fbuff[2] = pit[2];
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "pitch write error");
        return SplFalse;
      }
      if( fwrite(fbuff, sizeof(float), 3, fp) != 3 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "pitch write error");
        return SplFalse;
      }
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "pitch write error");
        return SplFalse;
      }
    } else {
      idmy = 24;
      dbuff[0] = pit[0];
      dbuff[1] = pit[1];
      dbuff[2] = pit[2];
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "pitch write error");
        return SplFalse;
      }
      if( fwrite(dbuff, sizeof(double), 3, fp) != 3 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "pitch write error");
        return SplFalse;
      }
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "pitch write error");
        return SplFalse;
      }
    }
  }

  if( fp ) {
    if ( rlen == 4 ) {
      idmy = 8;
      ibuff[0] = step;
      fbuff[0] = time;
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "time write error");
        return SplFalse;
      }
      if( fwrite(ibuff, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "time write error");
        return SplFalse;
      }
      if( fwrite(fbuff, sizeof(float), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "time write error");
        return SplFalse;
      }
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "time write error");
        return SplFalse;
      }
    } else {
      idmy = 16;
      lbuff[0] = step;
      dbuff[0] = time;
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "time write error");
        return SplFalse;
      }
      if( fwrite(lbuff, sizeof(long long), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "time write error");
        return SplFalse;
      }
      if( fwrite(dbuff, sizeof(double), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "time write error");
        return SplFalse;
      }
      if( fwrite(&idmy, sizeof(int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "time write error");
        return SplFalse;
      }
    }
  }

  return SplTrue;
}


/**
 * write data record in SPH File
 *
 * @param        mode            write mode
 * @param        writeRank       Rank that write a file
 *                               (It is effective at mode 1 or 2.)
 * @param        grp             Communication group
 * @param        fp              file pointer
 * @param        vox_num_per_blk Size of number of voxel in a write data record.
 * @param        vlen            Vector length
 * @param        rlen            Real Data size(4 or 8)
 * @param        vsz             size of voxel
 * @param        data            data of voxel
 * @param        wsize           total size of nodes
 * @param        sta_idx         gloval start index of nodes
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSphDataRecord( int mode,
                    int writeRank,
                    SPL_Comm grp,
                    FILE* fp,
                    size_t vox_num_per_blk,
                    unsigned int vlen,
                    unsigned int rlen,
                    const size_t* vsz,
                    const unsigned char* data,
                    const size_t* wsize,
                    const size_t* sta_idx) {
  register unsigned int n, i;
  unsigned int dims = 3;
  unsigned int data_len = rlen;

  if( !vsz || !wsize || !sta_idx) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "vsz, wsize or sta_idx is null");
    return SplFalse;
  }

  if( mode == 0 ) {
    unsigned int data_size;
    data_size = data_len * vlen;
    for(n=0; n<dims; n++) data_size *= vsz[n];
    if (data_size == 0) {
      return SplTrue;
    }

    if ( fp ) {
      if( fwrite(&data_size, sizeof(unsigned int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "data sz write error");
        return SplFalse;
      }

      if( fwrite(data, sizeof(char), data_size, fp) !=data_size ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "data write error");
        return SplFalse;
      }

      if( fwrite(&data_size, sizeof(unsigned int), 1, fp) != 1 ) {
        SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "data sz2 write error");
        return SplFalse;
      }
    }
  }
  else { // mode 1: Gather Data
    register unsigned long long i, j, k;

    int myrank, proc_size;
    SPL_Comm_rank(grp, &myrank);
    SPL_Comm_size(grp, &proc_size);

    unsigned char* write_buf = NULL;
    unsigned char* rec_buf = NULL;
    unsigned char* send_buf = NULL;
    long long *send_lens = NULL;
    long long *nodes_size = NULL;
    long long  *nodes_sta_idx = NULL;
    long long  *sends_sta_idx = NULL;
    long long my_size[3] = {0, 0, 0};
    long long  my_sta[3] = {0, 0, 0};
    long long my_send_sta[3] = {0, 0, 0};
    long long tmp_write_sta_idx[3] = {0, 0, 0};
    long long tmp_write_size[3] = {0, 0, 0};
    size_t write_size[3] = {0, 0, 0};
    size_t write_sta_idx[3] = {0, 0, 0};
    size_t write_buf_sz;
    size_t write_pos_idx = 0;
    long long write_len = 0;
    long long dtype_len = 0;
    long long whole_byte_size;
    long long whole_len;
    int dims_value;

    whole_byte_size = 0;
    if (myrank == writeRank) {
      write_len = vox_num_per_blk;
      dtype_len = vlen * data_len;

      // set write start_index and size
      for(n=0; n<dims; n++) {
//        tmp_write_sta_idx[n] = sta_idx[n];
        tmp_write_size[n] = wsize[n];
      }
    }

    // BCast
    SPL_Bcast(&write_len, 1, SPL_LONG_LONG, writeRank, grp);
    SPL_Bcast(&dtype_len, 1, SPL_LONG_LONG, writeRank, grp);
//    SPL_Bcast(&write_sta_idx, 3, SPL_LONG_LONG, writeRank, grp);
    SPL_Bcast(&tmp_write_size, 3, SPL_LONG_LONG, writeRank, grp);

    // get whole length
    whole_len = 1;
    for(n=0; n<dims; n++) {
//      write_sta_idx[n] = tmp_write_sta_idx[n];
      write_size[n] = tmp_write_size[n];
      whole_len *= write_size[n];
    }
    whole_byte_size = whole_len*dtype_len;
    write_buf_sz = write_len * dtype_len;
    if (write_buf_sz <= 0) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "write buffer size is zero.");
      return SplFalse;
    }
    if (whole_byte_size <= 0) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "whole size is zero.");
      return SplFalse;
    }

    for (n=0; n<dims; n++) {
      my_sta[n] = (long long)sta_idx[n];
    }

    // get my send area
    size_t my_ov_sidx[3] = {0,0,0};
    size_t my_ov_eidx[3] = {0,0,0};
    size_t my_ov_size[3] = {0,0,0};
    size_t my_ov_len = 0;
    size_t end_idx[3] = {0,0,0};
    size_t end_widx[3] = {0,0,0};
    for (i=0; i<dims; i++) {
      end_idx[i] = sta_idx[i] + vsz[i] - 1;
      end_widx[i] = write_sta_idx[i] + write_size[i] - 1;
    }
    if (getOverlapAreas(dims, sta_idx, end_idx, write_sta_idx, end_widx,
        my_ov_sidx, my_ov_size) == SplTrue) {
      my_ov_len = 1;
      for (n=0; n<dims; n++) {
        my_ov_eidx[n] = my_ov_sidx[n] + my_ov_size[n] - 1;
        my_size[n] = (long long)my_ov_size[n];
      }
    }

    // allocate buffer memory
    // send buffer from nodes
    send_buf = (unsigned char*)malloc(sizeof(char)*write_buf_sz);
    if (!send_buf) {
      SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld",
          sizeof(char)*write_buf_sz);
      return SplFalse;
    }
    memset(send_buf, 0, sizeof(char)*write_buf_sz);

    if (myrank == writeRank) {
      // write buffer to file
      write_buf = (unsigned char*)malloc(sizeof(char)*write_buf_sz);
      // reveive buffer from nodes
      rec_buf = (unsigned char*)malloc(sizeof(char)*write_buf_sz);
      // send data length list from nodes
      send_lens = (long long*)malloc(sizeof(long long)*proc_size);
      // nodes size list
      nodes_size = (long long*)malloc(sizeof(long long)*proc_size*dims);
      // nodes start index list
      nodes_sta_idx = (long long*)malloc(sizeof(long long)*proc_size*dims);
      // send data start index list
      sends_sta_idx = (long long*)malloc(sizeof(long long)*proc_size*dims);
      if (!write_buf || !rec_buf || !send_buf || !send_lens || !nodes_size
          || !nodes_sta_idx || !sends_sta_idx ) {
        SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld",
            sizeof(char)*write_buf_sz);
        if (write_buf) free(write_buf);
        if (rec_buf) free(rec_buf);
        if (send_buf) free(send_buf);
        if (send_lens) free(send_lens);
        if (nodes_size) free(nodes_size);
        if (nodes_sta_idx) free(nodes_sta_idx);
        if (sends_sta_idx) free(sends_sta_idx);
        return SplFalse;
      }
      memset(write_buf, 0, sizeof(char)*write_buf_sz);
      memset(rec_buf, 0, sizeof(char)*write_buf_sz);
      memset(send_lens, 0, sizeof(long long)*proc_size);
      memset(nodes_size, 0, sizeof(long long)*proc_size*dims);
      memset(nodes_sta_idx, 0, sizeof(long long)*proc_size*dims);
      memset(sends_sta_idx, 0, sizeof(long long)*proc_size*dims);
    }

    // gather nodes size
    if (SPL_Gather(my_size, dims, SPL_LONG_LONG,
        nodes_size, dims, SPL_LONG_LONG,
        writeRank, grp) != SPL_SUCCESS ) {
      SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Gather Error.");
      if (write_buf) free(write_buf);
      if (rec_buf) free(rec_buf);
      if (send_buf) free(send_buf);
      if (send_lens) free(send_lens);
      if (nodes_size) free(nodes_size);
      if (nodes_sta_idx) free(nodes_sta_idx);
      if (sends_sta_idx) free(sends_sta_idx);
      return SplFalse;
    }

    // gather nodes start index
    if (SPL_Gather(my_sta, dims, SPL_LONG_LONG,
        nodes_sta_idx, dims, SPL_LONG_LONG,
        writeRank, grp) != SPL_SUCCESS ) {
      SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Gather Error.");
      if (write_buf) free(write_buf);
      if (rec_buf) free(rec_buf);
      if (send_buf) free(send_buf);
      if (send_lens) free(send_lens);
      if (nodes_size) free(nodes_size);
      if (nodes_sta_idx) free(nodes_sta_idx);
      if (sends_sta_idx) free(sends_sta_idx);
      return SplFalse;
    }

    long long remain_len;
    size_t sta_widx[3] = {0, 0, 0};
    size_t node_size[3] = {1, 1, 1};
    size_t node_sta_idx[3] = {0, 0, 0};
    size_t send_sta_idx[3] = {0, 0, 0};
    long long send_len;
    size_t copy_len;
    long long loc_len;
    remain_len = whole_byte_size;
    write_pos_idx = 0;
    unsigned int data_size = whole_byte_size;

    // write data size in data recode in file
    if (myrank == writeRank) {
      if ( fwrite(&data_size, sizeof(unsigned int), 1, fp) != 1 ) {
        if (write_buf) free(write_buf);
        if (rec_buf) free(rec_buf);
        if (send_buf) free(send_buf);
        if (send_lens) free(send_lens);
        if (nodes_size) free(nodes_size);
        if (nodes_sta_idx) free(nodes_sta_idx);
        if (sends_sta_idx) free(sends_sta_idx);
        return SplFalse;
      }
    }

    // loop data block
    while( remain_len > 0 ) {

      // clear buffer
      if (write_buf) memset(write_buf, 0, sizeof(char)*write_buf_sz);
      if (rec_buf) memset(rec_buf, 0, sizeof(char)*write_buf_sz);

      my_send_sta[0] = my_send_sta[1] = my_send_sta[2] =0;
      sta_widx[0] = sta_widx[1] = sta_widx[2] =0;

      size_t sidx[3] = {0,0,0};
      size_t len;
      if (my_ov_len > 0) {
        // get write start index
        getIndexFromPos(dims, write_pos_idx, write_size, sta_widx);

        // get overlap start index and lenght in my node
        getOverlapLength(dims, write_size, sta_widx, write_len,
            my_ov_sidx, my_ov_eidx, sidx, &len);
      }

      if (len > 0) {
        for (n=0; n<dims; n++) {
          my_send_sta[n] = (long long)sidx[n];
        }

        if (len > write_len) {
          SPL_ERROR_INFO(SPL_ERR_INTERNAL_IMPOSSIBLE,
              "overlap length overfolow [buf_size=%d,ov_len=%d].",
              write_len, len);
          len = write_len;
        }

        // get send data array index in my node
        size_t local_sidx[3] = {0,0,0}, local_eidx[3] = {0,0,0},
            local_cp_sidx[3] = {0,0,0};
        for (n=0; n<dims; n++) {
          local_sidx[n] = my_ov_sidx[n]-sta_idx[n];
          local_cp_sidx[n] = sidx[n]-sta_idx[n];
          local_eidx[n] = my_ov_eidx[n]-sta_idx[n];
        }

        memset(send_buf, 0, sizeof(char)*write_buf_sz);
        // create send buffer
        if (copyLengthBuffer(dims, data_len * vlen, vsz, local_sidx,
            local_eidx, local_cp_sidx, len, write_buf_sz,
            data, send_buf) != SplTrue) {
          if (write_buf) free(write_buf);
          if (rec_buf) free(rec_buf);
          if (send_buf) free(send_buf);
          if (send_lens) free(send_lens);
          if (nodes_size) free(nodes_size);
          if (nodes_sta_idx) free(nodes_sta_idx);
          if (sends_sta_idx) free(sends_sta_idx);
          return SplFalse;
        }
      }

      // gather send length from node
      send_len = len;
      if (SPL_Gather(&send_len, 1, SPL_LONG_LONG,
          send_lens, 1, SPL_LONG_LONG,
          writeRank, grp) != SPL_SUCCESS ) {
        SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Gather Error.");
        if (write_buf) free(write_buf);
        if (rec_buf) free(rec_buf);
        if (send_buf) free(send_buf);
        if (send_lens) free(send_lens);
        if (nodes_size) free(nodes_size);
        if (nodes_sta_idx) free(nodes_sta_idx);
        if (sends_sta_idx) free(sends_sta_idx);
        return SplFalse;
      }
      // gather send start index from node
      if (SPL_Gather(my_send_sta, dims, SPL_LONG_LONG,
          sends_sta_idx, dims, SPL_LONG_LONG,
          writeRank, grp) != SPL_SUCCESS ) {
        SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Gather Error.");
        if (write_buf) free(write_buf);
        if (rec_buf) free(rec_buf);
        if (send_buf) free(send_buf);
        if (send_lens) free(send_lens);
        if (nodes_size) free(nodes_size);
        if (nodes_sta_idx) free(nodes_sta_idx);
        if (sends_sta_idx) free(sends_sta_idx);
        return SplFalse;
      }

      SPL_Request request = SPL_REQUEST_NULL;
      if (send_len > 0) {
        // send my data
        if ( SPL_Isend((void*)send_buf,
            send_len*dtype_len,
            SPL_UNSIGNED_CHAR,
            writeRank, 1, grp,&request) != SPL_SUCCESS ) {
          SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Isend Error.");
          if (write_buf) free(write_buf);
          if (rec_buf) free(rec_buf);
          if (send_buf) free(send_buf);
          if (send_lens) free(send_lens);
          if (nodes_size) free(nodes_size);
          if (nodes_sta_idx) free(nodes_sta_idx);
          if (sends_sta_idx) free(sends_sta_idx);
          return SplFalse;
        }
      }

      // receive data from nodes, and copy write buffer
      if (myrank == writeRank) {
        for (n=0; n<proc_size; n++) {
          if (send_lens[n] == 0) continue;

          if (rec_buf) memset(rec_buf, 0, sizeof(char)*write_buf_sz);
          // receive data from nodes
          SPL_Status status;
          if ( SPL_Recv(rec_buf,
              send_lens[n]*dtype_len,
              SPL_UNSIGNED_CHAR,
              n, 1, grp, &status) != SPL_SUCCESS ) {
            SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Recv Error.");
            if (write_buf) free(write_buf);
            if (rec_buf) free(rec_buf);
            if (send_buf) free(send_buf);
            if (send_lens) free(send_lens);
            if (nodes_size) free(nodes_size);
            if (nodes_sta_idx) free(nodes_sta_idx);
            if (sends_sta_idx) free(sends_sta_idx);
            return SplFalse;
          }

          // receive node size, start index in node, start index in receive data
          node_size[0] = node_size[1] = node_size[2] = 1;
          node_sta_idx[0] = node_sta_idx[1] = node_sta_idx[2] = 0;
          for (i=0; i<dims; i++) {
            node_size[i] = nodes_size[n*dims+i];
            node_sta_idx[i] = nodes_sta_idx[n*dims+i];
            send_sta_idx[i] = sends_sta_idx[n*dims+i];
          }

          // get start and end position in receive data
          size_t lidx = 0;
          size_t my_idx[3];
          size_t widx;
          // copy reveive buffer to write buffer
          while (lidx < send_lens[n]) {
            my_idx[0]=my_idx[1]=my_idx[2]=0;
            // get copy start node local index
            getLocalIndexFromPos(dims, node_size, node_sta_idx,
                send_sta_idx, lidx, my_idx);

            // get global position
            getPosFromIndex(dims, write_size, my_idx, &widx);

            // get write position
            widx = widx - write_pos_idx;

            // get copy length
            if (lidx == 0) {
              loc_len = node_size[0] - (send_sta_idx[0] - node_sta_idx[0]);
            }
            else {
              loc_len = node_size[0];
            }
            if (lidx + loc_len >= send_lens[n]) {
              loc_len = send_lens[n] - lidx;
            }

            // copy buffer
            if (loc_len > 0) {
              // get overlap len
              size_t loc_ov_str;
              size_t idx = (send_sta_idx[0]+lidx)%node_size[0];
              copy_len = 0;
              getOverlapArea(write_sta_idx[0], write_size[0]-1, idx,
                  idx + loc_len -1, &loc_ov_str, &copy_len);
              if (copy_len > 0) {
                if (sizeof(char)*write_buf_sz
                    >= widx*dtype_len + copy_len*dtype_len &&
                    sizeof(char)*write_buf_sz
                    >= lidx*dtype_len + copy_len*dtype_len) {
                  memcpy(write_buf+widx*dtype_len,
                      rec_buf+lidx*dtype_len,
                      copy_len*dtype_len);
                }
                else {
                  fprintf(stderr, "invalid over allocate memory\n");
                }
              }
            }
            lidx += loc_len;
          }      // lidx < send_lens[n]
        }        // for (n=0; n<proc_size; n++)

        // write data recode
        if( remain_len < write_buf_sz ) {
          write_buf_sz = remain_len;
        }
        // write uncompress data
        if( fwrite(write_buf, sizeof(char), write_buf_sz, fp)
            != write_buf_sz ) {
          if (write_buf) free(write_buf);
          if (rec_buf) free(rec_buf);
          if (send_buf) free(send_buf);
          if (send_lens) free(send_lens);
          if (nodes_size) free(nodes_size);
          if (nodes_sta_idx) free(nodes_sta_idx);
          if (sends_sta_idx) free(sends_sta_idx);
          return SplFalse;
        }
      } // myrank == writeRank

      // wait for isend
      SPL_Status wait_status;
      if (send_len > 0) {
        if (SPL_Wait(&request, &wait_status) != SPL_SUCCESS ) {
          SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Wait Error.");
          if (write_buf) free(write_buf);
          if (rec_buf) free(rec_buf);
          if (send_buf) free(send_buf);
          if (send_lens) free(send_lens);
          if (nodes_size) free(nodes_size);
          if (nodes_sta_idx) free(nodes_sta_idx);
          if (sends_sta_idx) free(sends_sta_idx);
          return SplFalse;
        }
      }

      // get remain length
      remain_len -= write_buf_sz;
      write_pos_idx += write_len;
      if( remain_len < write_buf_sz ) {
        write_buf_sz = remain_len;
        write_len = remain_len/dtype_len;
      }

    } /* while( remain_len > 0 ) */

    // write eof in data recode
    if (myrank == writeRank) {
      if ( fwrite(&data_size, sizeof(unsigned int), 1, fp) != 1 ) {
        if (write_buf) free(write_buf);
        if (rec_buf) free(rec_buf);
        if (send_buf) free(send_buf);
        if (send_lens) free(send_lens);
        if (nodes_size) free(nodes_size);
        if (nodes_sta_idx) free(nodes_sta_idx);
        if (sends_sta_idx) free(sends_sta_idx);
        return SplFalse;
      }
    }

    if (write_buf) free(write_buf);
    if (rec_buf) free(rec_buf);
    if (send_buf) free(send_buf);
    if (send_lens) free(send_lens);
    if (nodes_size) free(nodes_size);
    if (nodes_sta_idx) free(nodes_sta_idx);
    if (sends_sta_idx) free(sends_sta_idx);
  }

  return SplTrue;
}

#if defined(__cplusplus)
}
#endif /* extern "C" */
