/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file splfio_sv_read.c
 * Prototype declaration of file read subroutine related to Sv(Simple Voxel).
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file splfio_sv_read.c
 * Sv(Simple Voxel)の低レベルファイル読込関数群ソースファイル
 * @author     hira
 * @date       2009/02/20
 */

#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <zlib.h>
#include "splfio_sv_read.h"
#include "splfio_sv_util.h"

#ifdef _DEBUGLOG_
#include "debugLog.h"
#endif

#if defined(__cplusplus)
extern "C" {
#endif

//int spl_err_code;

/* -----------------------------------------------------------------------*/
/*  Simple-Voxel File read procedure                                      */
/* -----------------------------------------------------------------------*/

/**
 * read Attribute record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param[in]           fp                file pointer
 * @param[in]            fmt               File Type
 * @param[out]        endian_type       Endian match type
 * @param[out]        dims              Dimension
 * @param[out]        vlen              Vector length
 * @param[out]        dtype             Data type(1or2or4)
 * @param[out]        gc                Size of guide cell
 * @param[out]        rlen              Real size(4or8)
 * @param[out]        crddef            coords define
 * @param[out]        aux                block_size
 * @param[out]        block_size        Size of number of voxel  in a compressed data record.
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvAttrRecord(FILE* fp,                      /* input  */
                 SplFileType fmt,               /* input  */
                 EMatchType* endian_type,       /* output */
                 unsigned int *dims,            /* output */
                 unsigned int *vlen,            /* output */
                 unsigned int *dtype,           /* output */
                 size_t *gc,                    /* output */
                 unsigned int *rlen,            /* output */
                 unsigned int *crddef,          /* output */
                 unsigned int *aux,             /* output */
                 unsigned long long* block_size /* output */ ) {
  int magick;
  int tmp_gc;
  EMatchType eType;
  if( !fp || !endian_type || !dims || !vlen
     || !dtype || !gc || !rlen || !crddef || !aux || !block_size ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvAttrRecord arguments is null");
    return SplFalse;
  }

  /* read magick */
  if( fseek(fp, 0, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is 0");
    return SplFalse;
  }
  if( fread(&magick, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_MAGICK_NO, "magick read error");
    return SplFalse;
  }

  eType = UnKnown;
  if( (eType = isMatchEndianSvMagick(magick, fmt)) == UnKnown ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_ENDIAN_TYPE, "isMatchEndianSvMagick error");
    return SplFalse;
  }
  *endian_type = eType;

  /* read dims */
  if( fread(dims, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "dims read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap32(dims);

  /* read vlen */
  if( fread(vlen, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "vlen read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap32(vlen);

  /* read dtype */
  if( fread(dtype, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "dtype read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap32(dtype);

  /* read gc */
  if( fread(&tmp_gc, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "gc read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap32(&tmp_gc);
  *gc = (size_t)tmp_gc;

  /* read rlen */
  if( fread(rlen, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "rlen read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap32(rlen);

  /* read crddef */
  if( fread(crddef, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "crddef read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap32(crddef);

  /* read aux */
  if( fread(aux, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "aux read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap32(aux);

  /* read block size */
  if( fread(block_size, sizeof(long long), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "block_size read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap64(block_size);

  return SplTrue;
}


/**
 * read Size record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param[in]           fp                file pointer
 * @param[in]            file_offset      file seek length
 * @param[in]            eType             Endian match type
 * @param[out]        size              Voxel size
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvSizeRecord(FILE* fp,               /* input */
                 long file_offset,       /* input */
                 EMatchType eType,       /* input */
                 size_t* size            /* output */ ) {
  unsigned long long sz[3];

  if( !fp || !size || (eType == UnKnown) ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvSizeRecord arguments is null.");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  /* read size */
  if( fread(sz, sizeof(long long), 3, fp) != 3 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "sz read error");
    return SplFalse;
  }
  if( eType == UnMatch ) DBSwapVec(sz, 3);
  size[0] = (size_t)sz[0];
  size[1] = (size_t)sz[1];
  size[2] = (size_t)sz[2];

  return SplTrue;
}


/**
 * read Origin record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param[in]           fp                file pointer
 * @param[in]            file_offset      file seek length
 * @param[in]            eType             Endian match type
 * @param[in]            rlen               Real Data size
 * @param[out]        org              Voxel Origin
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvOriginRecord(FILE* fp,               /* input */
                   long file_offset,       /* input */
                   EMatchType eType,       /* input */
                   int rlen,               /* input */
                   double* org             /* output */ ) {
  if( !fp || !org || (eType == UnKnown) ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvOriginRecord arguments is null.");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  /* read origin */
  if( rlen == 4 ) {       /* single */
    float o[3];
    if( fread(o, sizeof(float), 3, fp) != 3 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "origin read error");
      return SplFalse;
    }
    if( eType == UnMatch ) BSwapVec(o, 3);
    org[0] = (double)o[0];
    org[1] = (double)o[1];
    org[2] = (double)o[2];
  }
  else if( rlen == 8 ) {  /* double */
    if( fread(org, sizeof(double), 3, fp) != 3 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "org read error");
      return SplFalse;
    }
    if( eType == UnMatch ) DBSwapVec(org, 3);
  }
  else {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_SPLSV_DTYPE, "invalid rlen=%d", rlen);
    return SplFalse;
  }

  return SplTrue;
}


/**
 * read Pitch record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param[in]           fp                file pointer
 * @param[in]            file_offset      file seek length
 * @param[in]            eType             Endian match type
 * @param[in]            rlen               Real Data size
 * @param[out]        pit              Voxel Pitch
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvPitchRecord(FILE* fp,               /* input */
                  long file_offset,       /* input */
                  EMatchType eType,       /* input */
                  int rlen,               /* input */
                  double* pit             /* output */ ) {
  if( !fp || !pit || (eType == UnKnown) ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvPitchRecord arguments is null.");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  /* read pitch */
  if( rlen == 4 ) {       /* single */
    float p[3];
    if( fread(p, sizeof(float), 3, fp) != 3 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "pitch read error");
      return SplFalse;
    }
    if( eType == UnMatch ) BSwapVec(p, 3);
    pit[0] = (double)p[0];
    pit[1] = (double)p[1];
    pit[2] = (double)p[2];
  }
  else if( rlen == 8 ) {  /* double */
    if( fread(pit, sizeof(double), 3, fp) != 3 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "pitch read error");
      return SplFalse;
    }
    if( eType == UnMatch ) DBSwapVec(pit, 3);
  }
  else {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_SPLSV_DTYPE, "invalid rlen=%d", rlen);
    return SplFalse;
  }

  return SplTrue;
}


/**
 * read Time record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param[in]           fp                file pointer
 * @param[in]            file_offset      file seek length
 * @param[in]            eType             Endian match type
 * @param[in]            rlen               Real Data size
 * @param[out]        step              step count
 * @param[out]        time              time
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvTimeRecord(FILE* fp,         /* input */
                 long file_offset, /* input */
                 EMatchType eType, /* input */
                 int rlen,         /* input */
                 long long* step,  /* output */
                 double* time      /* output */ ) {
  if( !fp || !step || !time || (eType == UnKnown) ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvTimeRecord arguments is null.");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  /* read step */
  if( fread(step, sizeof(long long), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "step read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap64(step);

  /* read time */
  if( rlen == 4 ) {       /* single */
    float ftime;
    if( fread(&ftime, sizeof(float), 1, fp) != 1 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "time read error");
      return SplFalse;
    }
    if( eType == UnMatch ) BSwap32(&ftime);
    *time = (double)ftime;
  }
  else if( rlen == 8 ) {  /* double */
    if( fread(time, sizeof(double), 1, fp) != 1 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "time read error");
      return SplFalse;
    }
    if( eType == UnMatch ) BSwap64(time);
  }
  else {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_SPLSV_DTYPE, "invalid rlen=%d", rlen);
    return SplFalse;
  }

  return SplTrue;
}
#if 0
SplBool
readSvTimeRecord(FILE* fp,               /* input */
                 long file_offset,       /* input */
                 EMatchType eType,       /* input */
                 int rlen,               /* input */
                 long long* step,        /* output */
                 float* ftime,           /* output */
                 double* dtime           /* output */ ) {
  if( !fp ||
      !step ||
      ((rlen==4)&&(!ftime)) ||
      ((rlen==8)&&(!dtime)) ||
      (eType == UnKnown) ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvTimeRecord arguments is null.");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  /* read step */
  if( fread(step, sizeof(long long), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "step read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap64(step);

  /* read time */
  if( rlen == 4 ) {       /* single */
    if( !ftime ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "ftime is null.");
      return SplFalse;
    }
    if( fread(ftime, sizeof(float), 1, fp) != 1 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "time read error");
      return SplFalse;
    }
    if( eType == UnMatch ) BSwap32(ftime);
  }
  else if( rlen == 8 ) {  /* double */
    if( !dtime ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "dtime is null.");
      return SplFalse;
    }
    if( fread(dtime, sizeof(double), 1, fp) != 1 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "time read error");
      return SplFalse;
    }
    if( eType == UnMatch ) BSwap64(dtime);
  }
  else {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_SPLSV_DTYPE, "invalid rlen=%d", rlen);
    return SplFalse;
  }

  return SplTrue;
}
#endif


/**
 * read compress flag for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param[in]           fp                file pointer
 * @param[in]            file_offset      file seek length
 * @param[in]            eType             Endian match type
 * @param[out]        cmp_flg           compress size
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
checkCompDtBlk(FILE* fp,                   /* input */
               long file_offset,           /* input */
               EMatchType eType,           /* input */
               unsigned long long* cmp_flg /* output */ ) {
  if( !fp || !cmp_flg ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "checkCompDtBlk arguments is null.");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  /* read cmp_flg */
  if( fread(cmp_flg, sizeof(long long), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "cmp_flg read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap64(cmp_flg);

  return SplTrue;
}


/**
 * read compress block size for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.<br>
 * caution:<br>
 *      This procedure doesn't seek to the fp.<br>
 * @param[in]           fp                file pointer
 * @param[in]            eType             Endian match type
 * @param[out]        comp_blk_sz       compress block size
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvBlkSz(FILE* fp,            /* input */
            EMatchType eType,    /* input */
            size_t* comp_blk_sz  /* output */ ) {
  if( !fp || !comp_blk_sz ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvBlkSz arguments is null.");
    return SplFalse;
  }

  unsigned long long tmp_comp_blk_sz;
  if( fread(&tmp_comp_blk_sz, sizeof(long long), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "tmp_comp_blk_sz read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap64(&tmp_comp_blk_sz);
  *comp_blk_sz = (size_t)tmp_comp_blk_sz;

  return SplTrue;
}


/**
 * read block data size of data record. for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.<br>
 * caution:<br>
 *      This procedure doesn't seek to the fp.<br>
 * @param[in]           mode              Read mode ( 0:all nodes read, 1:only readRank node read)
 * @param[in]           readRank          Rank that reads a file
 * @param[in]           grp               Communication group
 * @param[in]           fp                file pointer
 * @param[in]            eType             Endian match type
 * @param[out]        comp_blk_sz       block data size
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvDataBlkSz(int mode,
                int readRank,
                SPL_Comm grp,
                FILE* fp,
                EMatchType eType,
                size_t* comp_blk_sz) {
  /* parameter check */
  if( !comp_blk_sz ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvDataBlkSz arguments is null.");
    return SplFalse;
  }

  if( mode != 0 ) {
    int myrank;
    unsigned long long tmp_comp_blk_sz;
    SPL_Comm_rank(grp, &myrank);
    if( myrank == readRank ) {
      if( readSvBlkSz(fp, eType, comp_blk_sz) != SplTrue ) {
        return SplFalse;
      }
      tmp_comp_blk_sz = *comp_blk_sz;
    }
    SPL_Bcast(&tmp_comp_blk_sz, 1, SPL_LONG_LONG, readRank, grp);
    if( myrank != readRank ) {
      *comp_blk_sz = (size_t)tmp_comp_blk_sz;
    }
  }
  else {
    if( readSvBlkSz(fp, eType, comp_blk_sz) != SplTrue ) {
      return SplFalse;
    }
  }

  return SplTrue;
}


/**
 * read data of data record. for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.<br>
 * caution:<br>
 *      This procedure doesn't seek to the fp.<br>
 * @param[in]           fp                file pointer
 * @param[in]            dt_blk_sz         data block size
 * @param[in]            buf_sz             buffer size
 * @param[out]        buf               read budder
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvDtBlk(FILE* fp,
            size_t dt_blk_sz,
            size_t buf_sz,
            unsigned char* buf ) {
  if( !fp || !buf || (dt_blk_sz > buf_sz) ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvDtBlk arguments is null.");
    return SplFalse;
  }

  if( fread(buf, sizeof(char), dt_blk_sz, fp) != dt_blk_sz ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "data block read error");
    return SplFalse;
  }

  return SplTrue;
}


/**
 * read data of data record. for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.<br>
 * caution:<br>
 *      This procedure doesn't seek to the fp.<br>
 * @param[in]           mode              Read mode ( 0:all nodes read, 1:only readRank node read)
 * @param[in]           readRank          Rank that reads a file
 * @param[in]           grp               Communication group
 * @param[in]           fp                file pointer
 * @param[in]            eType              Endian match type
 * @param[in]            dt_blk_sz         data block size
 * @param[in]            buf_sz             buffer size
 * @param[out]        buf               read budder
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvDataBlock(int mode,
                int readRank,
                SPL_Comm grp,
                FILE* fp,
                EMatchType eType,
                size_t dt_blk_sz,
                size_t buf_sz,
                unsigned char* buf ) {
  /* parameter check */
  if( !buf ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvDataBlock arguments is null.");
    return SplFalse;
  }

  /* read compress data from file */
  if( mode == 0 ) {
    if( readSvDtBlk(fp, dt_blk_sz, buf_sz, buf) != SplTrue ) {
      return SplFalse;
    }
  }
  else {
    int myrank;
    SPL_Comm_rank(grp, &myrank);
    if( myrank == readRank ) {
      if( readSvDtBlk(fp, dt_blk_sz, buf_sz, buf) != SplTrue ) {
        return SplFalse;
      }
    }
    SPL_Bcast(buf, dt_blk_sz, SPL_CHAR, readRank, grp);
  }

  return SplTrue;
}


/**
 * read extend record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.<br>
 * caution:<br>
 *       This procedure seeks fp in front of Data-Record.<br>
 *      file_offset specifies the position of DataRecord.
 * @param[in]           fp                file pointer
 * @param[in]           file_offset       file seek length
 * @param[in]            eType              Endian match type
 * @param[in]            rlen               Real data size
 * @param[out]        wsz               gloval size
 * @param[out]        widx               gloval start index
 * @param[out]        has_ext_record    has extend record
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvExtRecord(FILE* fp,                /* input */
                long file_offset,        /* input */
                EMatchType eType,        /* input */
                int rlen,                /* input */
                size_t* wsz,             /* output */
                size_t* widx,            /* output */
                SplBool* has_ext_record  /* output */ ) {
  SplBool write_flag;
  unsigned long long tmp_wsz[3], tmp_widx[3];
  if( !has_ext_record ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvExtRecord arguments is null.");
    return SplFalse;
  }
  *has_ext_record = SplFalse;

  if( !fp || !wsz || !widx || (eType == UnKnown) ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvExtRecord arguments is null.");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  unsigned long long cbs;
  if( fread(&cbs, sizeof(long long), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "cbs read error");
    return SplFalse;
  }
  if( eType == UnMatch ) BSwap64(&cbs);
  while(cbs != 0) {
    if( fseek(fp, cbs, SEEK_CUR) != 0 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", cbs);
      return SplFalse;
    }
    if( fread(&cbs, sizeof(long long), 1, fp) != 1 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_READ_FILE, "cbs read error");
      return SplFalse;
    }
    if( eType == UnMatch ) BSwap64(&cbs);
  }

  /* read world voxel size */
  write_flag = SplTrue;
  if( fread(tmp_wsz, sizeof(long long), 3, fp) != 3 ) {
    if( feof(fp) != 0 ) {
      write_flag = SplFalse;
    }
    else {
      // error
      SPL_ERROR_INFO(SPL_ERR_DETECT_ILLEGAL_EOF, "eof error");
      return SplFalse;
    }
  }
  if( eType == UnMatch ) DBSwapVec(wsz, 3);
  if( write_flag == SplTrue ) {
    wsz[0] = (size_t)tmp_wsz[0];
    wsz[1] = (size_t)tmp_wsz[1];
    wsz[2] = (size_t)tmp_wsz[2];
  }
  else {
    memset(wsz, 0, sizeof(size_t)*3);
    memset(widx, 0, sizeof(size_t)*3);
    return SplTrue;
  }

  /* read world index */
  if( fread(tmp_widx, sizeof(long long), 3, fp) != 3 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_READ_FILE, "tmp_widx read error");
    return SplFalse;
  }
  if( eType == UnMatch ) DBSwapVec(widx, 3);
  widx[0] = (size_t)tmp_widx[0];
  widx[1] = (size_t)tmp_widx[1];
  widx[2] = (size_t)tmp_widx[2];

  *has_ext_record = SplTrue;
  return SplTrue;
}


/**
 * Two or more data block is read.
 *
 * This function is used only for the compressed data block.
 *
 * @param[in]           mode                Read mode ( 0:all nodes read, 1:only readRank node read)
 * @param[in]           readRank             Rank that reads a file
 * @param[in]           grp                  Communication group
 * @param[in]           fp                  file pointer
 * @param[in]           eType                Endian match type
 * @param[in]           decomp_flag          decompress flag
 * @param[in]           dtype                Data type(1or2or4)
 * @param[in,out]     cmp_blk_buf_sz       compress buffer size
 * @param[in]           ext_blk_buf_sz      extend buffer size
 * @param[in]           dest_buf_sz          destination buffer size
 * @param[in]           surplus_buf_sz      surplus buffer size
 * @param[in,out]     cmp_blk_buf          compress buffer
 * @param[out]        ext_blk_buf          extend buffer
 * @param[out]        dest_buf            destination buffer
 * @param[out]        surplus_buf          surplus buffer
 * @param[out]        dest_effective_len  destination effective length
 * @param[out]        surplus_effective_len  surplus effective length
 * @return          error code
 *                     - -1 : Error<br>
 *                     -  0 : Success<br>
 *                     -  1 : Success and Reaches to last data block
 */
int
readSvCmpDtBlks(int mode,                      /* input */
                int readRank,                  /* input */
                SPL_Comm grp,                  /* input */
                FILE* fp,                      /* input */
                EMatchType eType,              /* input */
                SplBool decomp_flag,           /* input */
                unsigned int dtype,            /* input */
                size_t* cmp_blk_buf_sz,        /* input/output */
                size_t ext_blk_buf_sz,         /* input */
                size_t dest_buf_sz,            /* input */
                size_t surplus_buf_sz,         /* input */
                unsigned char** cmp_blk_buf,   /* tmp buffer/output */
                unsigned char* ext_blk_buf,    /* tmp buffer */
                unsigned char* dest_buf,       /* output */
                unsigned char* surplus_buf,    /* output */
                size_t* dest_effective_len,    /* output */
                size_t* surplus_effective_len  /* output */) {
  size_t cmp_blk_sz;

  /* parameter check */
  if(    !cmp_blk_buf || !ext_blk_buf || !dest_buf || !surplus_buf
      || !dest_effective_len || !surplus_effective_len ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvCmpDtBlks arguments is null.");
    return -1;
  }

  *dest_effective_len = *surplus_effective_len = 0;
  while( *surplus_effective_len == 0 ) {
    /* read block size from fp */
    if( readSvDataBlkSz(mode, readRank, grp, fp, eType, &cmp_blk_sz) != SplTrue ) {
      return -1;
    }
    if( cmp_blk_sz == 0 ) { return 1; } // It reaches to last data block.

    /* Reallocate cmp_blk_buf, if needed. */
    if( !(*cmp_blk_buf) || (cmp_blk_sz > *cmp_blk_buf_sz) ) {
      if( *cmp_blk_buf ) free(*cmp_blk_buf);
      *cmp_blk_buf = (unsigned char*)malloc(sizeof(char)*cmp_blk_sz);
      if( !(*cmp_blk_buf) ) {
        // error
        SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld", sizeof(char)*cmp_blk_sz);
        return -1;
      }
      *cmp_blk_buf_sz = cmp_blk_sz;
    }
    /* read data block from fp */
    if( readSvDataBlock(mode, readRank, grp, fp, eType,
                        cmp_blk_sz, *cmp_blk_buf_sz, *cmp_blk_buf) != SplTrue ) {
      return -1;
    }

    /* decompress data */
    if( decomp_flag == SplTrue ) {
      size_t extended_data_sz;
      if( decompSvDataBlock(eType, dtype,
                            *cmp_blk_buf_sz, ext_blk_buf_sz,
                            *cmp_blk_buf, ext_blk_buf,
                            &extended_data_sz) != SplTrue ) {
        return -1;
      }
      if( copyBuffer(extended_data_sz,
                     dest_buf_sz - (*dest_effective_len),
                     surplus_buf_sz - (*surplus_effective_len),
                     ext_blk_buf,
                     dest_buf + (*dest_effective_len),
                     surplus_buf + (*surplus_effective_len),
                     surplus_effective_len) != SplTrue ) {
        return -1;
      }
      (*dest_effective_len) += (extended_data_sz - (*surplus_effective_len));
    }
  }

  return 0;
}



/**
 * read SBX data record
 *
 * Attension!
 *   The buffer(comp_buf and exp_buf) is not allocated in this subroutine.
 * @param[in]     mode              read mode
 * @param[in]      readRank         Rank that reads a file (It is effective at mode 1 or 2.)
 * @param[in]      grp              Communication group
 * @param[in]      fp                file pointer
 * @param[in]      eType            Endian match type
 * @param[in]      cmp_flg          compress flag
                                    - 0     : data block is uncompressed
                                    - other : data block is compressed
 * @param[in]      dims              Dimension
 * @param[in]      file_vlen        Vector length in file
 * @param[in]      dest_vlen        Vector length in output array(dst)
 * @param[in]      data_len         Data type length (1,2,4,8)
 * @param[in]      file_gc          Size of guide cell of data described in file
 * @param[in]      dest_gc          Size of guide cell of "dst"
 * @param[in]      file_vsz          Size of Voxel described in file
 * @param[in]      file_vsta        Start index of data described in file
 * @param[in]      dest_vsz          Size of Voxel of "dst"
 * @param[in]      dest_vsta        Start index of "dst"
 * @param[in]      vox_num_per_blk  Size of number of voxel  in a compressed data record.
 * @param[out]   dst              Output data array
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
readSvDataRecord(int mode,
                  int readRank,
                  SPL_Comm grp,
                  FILE* fp,
                  EMatchType eType,
                  unsigned long long cmp_flg,
                  unsigned int dims,
                  unsigned int file_vlen,
                  unsigned int dest_vlen,
                  unsigned int data_len,
                  size_t file_gc,
                  size_t dest_gc,
                  size_t file_vsz[3],
                  size_t file_vsta[3],
                  size_t dest_vsz[3],
                  size_t dest_vsta[3],
                  size_t vox_num_per_blk,
                  unsigned char* dst) {

  register unsigned int n;
  size_t dest_len;

  /* parameter check */
  if( !file_vsz || !file_vsta || !dest_vsz || !dest_vsta ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "readSvDataRecord arguments is null.");
    return SplFalse;
  }

  dest_len = 1;
  for(n=0; n<dims; n++) dest_len *= dest_vsz[n];

  /* read data block and copy for dst buffer */
  if( cmp_flg == 0 ) { /* read uncompressed data block */
    size_t blk_sz, tmp_blk_sz;
    unsigned char *blk_buf;
    int gc_offset;
    size_t file_vend[3], dest_vend[3];
    size_t ov_gloval_sidx[3], ov_gloval_size[3];
    size_t src_sz_exgc[3], dest_sz_exgc[3];

    for(n=0; n<dims; n++) {
      src_sz_exgc[n] = file_vsz[n] - 2*file_gc;
      dest_sz_exgc[n] = dest_vsz[n] - 2*dest_gc;
    }

    /* read compress block data size */
    if( readSvDataBlkSz(mode, readRank, grp, fp, eType, &tmp_blk_sz) == SplFalse ) {
      return SplFalse;
    }

    /* allocate data-block buffer */
    blk_sz = sizeof(char)*data_len*file_vlen;
    for(n=0; n<dims; n++) blk_sz *= file_vsz[n];
    if( blk_sz != tmp_blk_sz ) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "invaild block size [size=%ld,%ld,%ld,block size=%ld].", file_vsz[0], file_vsz[1], file_vsz[2], tmp_blk_sz);
      return SplFalse;
    }
    if( (blk_buf = (unsigned char*)malloc(sizeof(char)*blk_sz)) == NULL ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld", sizeof(char)*blk_sz);
      return SplFalse;
    }

    /* read from file */
    if( readSvDataBlock(mode, readRank, grp, fp, eType,
                        blk_sz, blk_sz, blk_buf) == SplFalse ) {
      free(blk_buf);
      return SplFalse;
    }

    if (dst != NULL && dest_len > 0) {
      // copy destination buffer
      size_t src_gloval_sidx[3], src_gloval_eidx[3];
      size_t dst_gloval_sidx[3], dst_gloval_eidx[3];

      /* corrects dst_index and set gc_offset */
      gc_offset = dest_gc - file_gc;
      for(n=0; n<dims; n++){
//        if( (dest_vsta[n] == 0) && (dest_gc > file_gc) ) {
//          gc_offset = dest_gc - file_gc;
//        }
//        else {
//          gc_offset = 0;
//        }
//        file_vend[n] = file_vsta[n] + file_vsz[n] - 1;
//        dest_vend[n] = (dest_vsta[n] + dest_vsz[n] - 1) - gc_offset;
//        if (gc_offset > 0) {
//          dest_ov_sta[n] = (dest_vsta[n] > gc_offset) ? dest_vsta[n] - gc_offset : 0;
//        }
//        else {
//          dest_ov_sta[n] = dest_vsta[n];
//        }
//        dest_vend[n] = dest_ov_sta[n] + dest_vsz[n] - 1;

/**** 2009/10/08 modify by gc by @hira
        file_vend[n] = file_vsta[n] + (file_vsz[n]-2*file_gc) - 1;
        dest_vend[n] = dest_vsta[n] + (dest_vsz[n]-2*dest_gc) - 1;
****/
        src_gloval_sidx[n] = (file_vsta[n] > file_gc)?file_vsta[n] - file_gc:0;
        dst_gloval_sidx[n] = (dest_vsta[n] > dest_gc)?dest_vsta[n] - dest_gc:0;
        src_gloval_eidx[n] = src_gloval_sidx[n] + file_vsz[n] - 1;
        dst_gloval_eidx[n] = dst_gloval_sidx[n] + dest_vsz[n] - 1;
      }

      /* calc overlap area */
/**** 2009/10/08 modify by gc by @hira
      if( getOverlapAreas(dims, file_vsta, file_vend, dest_vsta, dest_vend,
                          ov_gloval_sidx, ov_gloval_size) == SplFalse ) {
        free(blk_buf);
        return SplFalse;
      }
****/
      if( getOverlapAreas(dims, src_gloval_sidx, src_gloval_eidx, dst_gloval_sidx, dst_gloval_eidx,
                          ov_gloval_sidx, ov_gloval_size) == SplFalse ) {
        free(blk_buf);
        return SplFalse;
      }
      /* copy buffer */
/**** 2009/10/09 modify by @hira
      if( copySvBuffer(dims,
                      data_len*file_vlen, data_len*dest_vlen,
                      file_vsta, src_sz_exgc, file_gc,
                      dest_vsta, dest_sz_exgc, dest_gc,
                      ov_gloval_sidx, ov_gloval_size, dest_gc,
                      blk_buf, dst) != SplTrue ) {
****/
      if( copySvBuffer(dims,
                      data_len*file_vlen, data_len*dest_vlen,
                      file_vsta, src_sz_exgc, file_gc,
                      dest_vsta, dest_sz_exgc, dest_gc,
                      ov_gloval_sidx, ov_gloval_size, 0,
                      blk_buf, dst) != SplTrue ) {
      free(blk_buf);
        return SplFalse;
      }
    }     // if (dst != NULL && dest_len > 0)

    free(blk_buf);
  }
  else { /* read compressed data blocks */
    int ret;
    unsigned char *cmp_blk_buf, *ext_blk_buf, *src_buf, *surplus_buf;
    size_t cmp_blk_buf_sz, ext_blk_buf_sz, src_buf_sz, surplus_buf_sz;
    size_t gc_offset;
    size_t src_effective_len, surplus_effective_len;
    unsigned char *tmp_src_buf;
    size_t tmp_src_buf_sz, src_tmp_surface, readed_vox_num;
    size_t dest_vend[3], src_vsz[3];
    size_t total_file_vox_num;
    size_t src_sz_exgc[3], dest_sz_exgc[3];
    size_t tmp_src_sidx[3], tmp_src_eidx[3];
    size_t tmp_dst_sidx[3], tmp_dst_eidx[3];

/*
    size_t readed_vox_num;
    register unsigned int n;
    size_t surplus_effect_len, src_buf_pointer;
    size_t src_vsz[3], src_tmp_surface;
*/

    /* allocate buffer */
    cmp_blk_buf = ext_blk_buf = src_buf = surplus_buf = NULL;
    cmp_blk_buf_sz = ext_blk_buf_sz = src_buf_sz = surplus_buf_sz = 0;

    /* allocate ext_blk_buf (Fixed array length) */
    ext_blk_buf_sz = vox_num_per_blk*data_len*file_vlen;
    if( (ext_blk_buf = (unsigned char*)malloc(sizeof(char)*ext_blk_buf_sz)) == NULL ) {
       // error
       SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld", sizeof(char)*ext_blk_buf_sz);
      if( cmp_blk_buf ) free(cmp_blk_buf);
      return SplFalse;
    }

    /* allocate cmp_blk_buf */
    cmp_blk_buf_sz = ext_blk_buf_sz;
    if( (cmp_blk_buf = (unsigned char*)malloc(sizeof(char)*cmp_blk_buf_sz)) == NULL ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld", sizeof(char)*cmp_blk_buf_sz);
      return SplFalse;
    }

    /* allocate src_buf (Fixed array length) */
    if( getReadBufLength(dims, file_vsz, data_len, file_vlen, &src_buf_sz) == SplFalse ) {
      if( cmp_blk_buf ) free(cmp_blk_buf);
      if( ext_blk_buf ) free(ext_blk_buf);
      return SplFalse;
    }
    if( (src_buf = (unsigned char*)malloc(sizeof(char)*src_buf_sz)) == NULL ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld", sizeof(char)*src_buf_sz);
      if( cmp_blk_buf ) free(cmp_blk_buf);
      if( ext_blk_buf ) free(ext_blk_buf);
      return SplFalse;
    }

    /* allocate surplus_buf (Fixed array length) */
    surplus_buf_sz = ext_blk_buf_sz;
    if( (surplus_buf = (unsigned char*)malloc(sizeof(char)*surplus_buf_sz)) == NULL ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld", sizeof(char)*surplus_buf_sz);
      if( cmp_blk_buf ) free(cmp_blk_buf);
      if( ext_blk_buf ) free(ext_blk_buf);
      if( src_buf ) free(src_buf);
      return SplFalse;
    }

    for(n=0; n<dims; n++) src_vsz[n] = file_vsz[n];
    src_tmp_surface = 1;
    for(n=0; n<dims-1; n++) src_tmp_surface *= file_vsz[n];
    src_vsz[dims-1] = src_buf_sz / (data_len * file_vlen * src_tmp_surface);

    // get size excepted gc
    for(n=0; n<dims; n++) {
      src_sz_exgc[n] = src_vsz[n] - 2*file_gc;
      dest_sz_exgc[n] = dest_vsz[n] - 2*dest_gc;
    }

    /* corrects dst_index and set gc_offset */
    for(n=0; n<dims; n++){
//      if( (dest_vsta[n] == 0) && (dest_gc > file_gc) ) {
//        gc_offset = dest_gc - file_gc;
//      }
//      else {
//        gc_offset = 0;
//      }
//      if (dest_vsz[n] > 0) {
//        dest_vend[n] = (dest_vsta[n] + dest_vsz[n] - 1) - gc_offset;
//      }
//      else {
//        dest_vend[n] = dest_vsta[n];
//      }
      dest_vend[n] = dest_vsta[n] + (dest_vsz[n]-2*dest_gc) - 1;
    }

    /*  */
    total_file_vox_num = 1;
    for(n=0; n<dims; n++) { total_file_vox_num *= file_vsz[n]; }

    readed_vox_num = 0;
    src_effective_len = surplus_effective_len = 0;
    tmp_src_buf = src_buf;
    tmp_src_buf_sz = src_buf_sz;
    size_t remain_len = 0;
    ret = 0;
    while( (ret = readSvCmpDtBlks(mode, readRank, grp, fp, eType, SplTrue, data_len,
                                  &cmp_blk_buf_sz, ext_blk_buf_sz,
                                  tmp_src_buf_sz, surplus_buf_sz,
                                  &cmp_blk_buf, ext_blk_buf, tmp_src_buf, surplus_buf,
                                  &src_effective_len, &surplus_effective_len)) != -1 ) {

      if (dst != NULL && dest_len > 0) {
        // copy destination buffer
        size_t ov_gloval_sidx[3], ov_gloval_sz[3];
        size_t src_gloval_sidx[3], src_gloval_eidx[3];
        size_t src_tmp_idx[3];

        /* calc start index of ext_blk_buf */
        if( getIndexFromPos(dims, readed_vox_num, file_vsz, src_tmp_idx) == SplFalse ) {
            if( cmp_blk_buf ) free(cmp_blk_buf);
            if( ext_blk_buf ) free(ext_blk_buf);
            if( src_buf ) free(src_buf);
            if( surplus_buf ) free(surplus_buf);
            return SplFalse;
        }

        /* calc overlap area */
/**** 2009/10/09 modify by @hira
        for(n=0; n<dims; n++) {
          src_gloval_sidx[n] = src_tmp_idx[n] + file_vsta[n];
          src_gloval_eidx[n] = src_gloval_sidx[n] + (src_vsz[n]-2*file_gc) - 1;
        }
        if( getOverlapAreas(dims, src_gloval_sidx, src_gloval_eidx,
                            dest_vsta, dest_vend,
                            ov_gloval_sidx, ov_gloval_sz) == SplFalse ) {
            if( cmp_blk_buf ) free(cmp_blk_buf);
            if( ext_blk_buf ) free(ext_blk_buf);
            if( src_buf ) free(src_buf);
            if( surplus_buf ) free(surplus_buf);
            return SplFalse;
        }

        // copySvBuffer
        if( copySvBuffer(dims,
                        file_vlen*data_len, dest_vlen*data_len,
                        src_gloval_sidx, src_sz_exgc, file_gc,
                        dest_vsta, dest_sz_exgc, dest_gc,
                        ov_gloval_sidx, ov_gloval_sz, dest_gc,
                        src_buf, dst) == SplFalse ) {
          if( cmp_blk_buf ) free(cmp_blk_buf);
          if( ext_blk_buf ) free(ext_blk_buf);
          if( src_buf ) free(src_buf);
          if( surplus_buf ) free(surplus_buf);
          return SplFalse;
        }
****/
        for(n=0; n<dims; n++) {
          src_gloval_sidx[n] = src_tmp_idx[n] + file_vsta[n];
          src_gloval_eidx[n] = src_gloval_sidx[n] + (src_vsz[n]-2*file_gc) - 1;

          tmp_src_sidx[n] = (src_gloval_sidx[n] > file_gc)?src_gloval_sidx[n] - file_gc:0;
          tmp_dst_sidx[n] = (dest_vsta[n] > dest_gc)?dest_vsta[n] - dest_gc:0;
          tmp_src_eidx[n] = tmp_src_sidx[n] + file_vsz[n] - 1;
          tmp_dst_eidx[n] = tmp_dst_sidx[n] + dest_vsz[n] - 1;

        }
        if( getOverlapAreas(dims, tmp_src_sidx, tmp_src_eidx,
                            tmp_dst_sidx, tmp_dst_eidx,
                            ov_gloval_sidx, ov_gloval_sz) == SplFalse ) {
            if( cmp_blk_buf ) free(cmp_blk_buf);
            if( ext_blk_buf ) free(ext_blk_buf);
            if( src_buf ) free(src_buf);
            if( surplus_buf ) free(surplus_buf);
            return SplFalse;
        }

        // copySvBuffer
        if( copySvBuffer(dims,
                        file_vlen*data_len, dest_vlen*data_len,
                        src_gloval_sidx, src_sz_exgc, file_gc,
                        dest_vsta, dest_sz_exgc, dest_gc,
                        ov_gloval_sidx, ov_gloval_sz, 0,
                        src_buf, dst) == SplFalse ) {
          if( cmp_blk_buf ) free(cmp_blk_buf);
          if( ext_blk_buf ) free(ext_blk_buf);
          if( src_buf ) free(src_buf);
          if( surplus_buf ) free(surplus_buf);
          return SplFalse;
        }
      }      // if (dst != NULL && dest_len > 0)

      if (src_buf_sz < surplus_effective_len) {
        break;
      }

// mod tamaki 20090908 begin -----------------------
//      memset(src_buf, NULL, src_buf_sz);
//--------------------------------------------------
      memset(src_buf, 0, src_buf_sz);
// mod tamaki 20090908 end   -----------------------

      /* Correction */
      memcpy(src_buf, surplus_buf, surplus_effective_len);
      tmp_src_buf = src_buf + surplus_effective_len;
      tmp_src_buf_sz = src_buf_sz - surplus_effective_len;

      /* initialize */
      readed_vox_num += ((src_effective_len+remain_len) / data_len / file_vlen);
      remain_len = surplus_effective_len;

      if( readed_vox_num >= total_file_vox_num ) break;
      if( ret == 1 ) break;

      src_effective_len = surplus_effective_len = 0;
    } /* while */

    if( cmp_blk_buf ) free(cmp_blk_buf);
    if( ext_blk_buf ) free(ext_blk_buf);
    if( src_buf ) free(src_buf);
    if( surplus_buf ) free(surplus_buf);

    if( ret == -1 ) return SplFalse;
  } /* if( cmp_flg == 0 ) */

  return SplTrue;
}



#if defined(__cplusplus)
}
#endif /* extern "C" */

