/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file splfio_sv_util.c
 * Prototype declaration of utility subroutine related to Sv(Simple Voxel).
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file splfio_sv_util.c
 * Sv(Simple Voxel)のユーティリティソースファイル
 * @author     hira
 * @date       2009/02/20
 */


#include "spl/util_endian.h"
#include "splfio_sv_util.h"
#include <sys/types.h>
#include <string.h>
#include <zlib.h>

#ifdef _DEBUGLOG_
#include "debugLog.h"
#endif

#define SPL_SV_MAX_BUF_SIZE 16000000 /* 16 MByte */

#if defined(__cplusplus)
extern "C" {
#endif

//extern int spl_err_code;


/**
 * \~english
 * get index from position
 * @param[in]       dims        Dimension ( 1 or 2 or 3 )
 * @param[in]      position    Length of 1D
 * @param[in]      vsz         Voxel size
 * @param[out]   idx        Index
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * 始点(0,0,0)からの位置のボクセルインデックスを取得する。
 * @param[in]       dims        次元数 ( 1 or 2 or 3 )
 * @param[in]      position    始点(0,0,0)からの長さ
 * @param[in]      vsz         ボクセルサイズ
 * @param[out]   idx         ボクセルインデックス
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
getIndexFromPos(unsigned int dims,
                size_t position,
                const size_t* vsz,
                size_t* idx ) {
  register int n;
  register int i;
  size_t surplus;
  size_t Coef[3];

  if( !vsz || !idx || (dims > 3) ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "getIndexFromPos arguments is null");
    return SplFalse;
  }

  memset(Coef, 0, sizeof(size_t)*3);
  Coef[0] = 1;
  for(n=1; n<dims; n++) Coef[n] = Coef[n-1]*vsz[n-1];

  surplus = position;
  for(n=(int)dims-1; n>=0; n--) {
    if (Coef[n] > 0) {
      idx[n] = surplus / Coef[n];
      surplus = surplus % Coef[n];
    }
    else {
      idx[n] = 0;
    }
  }

  if( surplus != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "surplus is zero");
    return SplFalse;
  }

  for(n=(int)dims-1; n>=0; n--) {
    if (idx[n] >= vsz[n]) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "invalid idx, idx[%d]=%ld >= vsz[%d]=%ld", n, idx[n], n, vsz[n]);
      for(i=0; n<(int)dims; n++) {
        idx[n] = vsz[n]-1;
      }
      return SplFalse;
    }
  }

  return SplTrue;
}


/**
 * \~english
 * calculate default size of read buffer
 *
 * MAX buffer size is SPL_SV_MAX_BUF_SIZE.
 * @param[in]       dims        Dimension ( 1 or 2 or 3 )
 * @param[in]      vsize        Voxel size of src file
 * @param[in]      dtype         Data type<br>
 *                               - 1 -> unsigned char<br>
 *                               - 2 -> unsigned short<br>
 *                               - 4 -> unsigned int or float<br>
 *                               - 8 -> double<br>
 * @param[in]       vec_len       Vector length
 * @param[out]   buf_len       Buffer length
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * 読み込みバッファーサイズの既定値を計算する。
 *
 * 最大バッファーサイズは、SPL_SV_MAX_BUF_SIZEとする。
 * @param[in]       dims        次元数 ( 1 or 2 or 3 )
 * @param[in]      vsize        ボクセルサイズ
 * @param[in]      dsize        データサイズ<br>
 *                               - 1 -> unsigned char<br>
 *                               - 2 -> unsigned short<br>
 *                               - 4 -> unsigned int or float<br>
 *                               - 8 -> double<br>
 * @param[in]       vec_len       ベクトル長さ
 * @param[out]   buf_len       バッファーサイズ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
getReadBufLength(unsigned int dims,
                 const size_t* vsize,
                 unsigned int dsize,
                 unsigned int vec_len,
                 size_t* buf_len) {
  size_t surface_size, surface_num, outside_index_size;

  /* parameter check */
  if( !vsize || !buf_len || (dims > 3) ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "getReadBufLength arguments is null");
    return SplFalse;
  }

  switch( dims ) {
  case 1:
    surface_size = dsize*vec_len;
    outside_index_size = vsize[0];
    break;
  case 2:
    surface_size = vsize[0]*dsize*vec_len;
    outside_index_size = vsize[1];
    break;
  case 3:
    surface_size = vsize[0]*vsize[1]*dsize*vec_len;
    outside_index_size = vsize[2];
    break;
  default:
    return 0;
  }

  if( SPL_SV_MAX_BUF_SIZE > surface_size ) {
    surface_num = SPL_SV_MAX_BUF_SIZE / surface_size;
  }else{
    surface_num = 1;
  }
  if( surface_num > outside_index_size ) surface_num = outside_index_size;
  *buf_len = surface_size * surface_num;

  return SplTrue;
}

/**
 * get overlap index for 1D
 *
 * <pre>
 *           src_gloval_sidx       src_gloval_eidx
 *                  +-------------------+
 *   dgs     dge
 * #1 +-------+
 *             dgs    dge
 * #2           +-------+
 *                       dgs     dge
 * #3                     +-------+
 *                                 dgs     dge
 * #4                               +-------+
 *                                           dgs     dge
 * #5                                         +-------+
 *          dgs                                   dge
 * #6        +-------------------------------------+
 * </pre>
 *
 * @param[in]        src_gloval_sidx        source start index
 * @param[in]        src_gloval_eidx        source end index
 * @param[in]        dst_gloval_sidx        destination start index
 * @param[in]        dst_gloval_eidx        destination end index
 * @param[out]    ov_gloval_sidx        overlap start index
 * @param[out]    ov_gloval_size        overlap size
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
getOverlapArea(const size_t  src_gloval_sidx,
               const size_t  src_gloval_eidx,
               const size_t  dst_gloval_sidx,
               const size_t  dst_gloval_eidx,
               size_t* ov_gloval_sidx,
               size_t* ov_gloval_size ) {
  /* parameter check */
  if( !ov_gloval_sidx || !ov_gloval_size ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "getOverlapArea arguments is null");
    return SplFalse;
  }

  if(   (dst_gloval_eidx < src_gloval_sidx)
     || (src_gloval_eidx < dst_gloval_sidx) ) {       /* # 1,5 */
    // modify 2009/10/08 by @hira
//    *ov_gloval_sidx = *ov_gloval_size = 0;
    *ov_gloval_sidx = (dst_gloval_eidx<src_gloval_eidx)?dst_gloval_eidx:src_gloval_eidx;
    *ov_gloval_size = 0;
    return SplTrue;
  }
  else if(   (dst_gloval_sidx <  src_gloval_sidx)
          && (src_gloval_sidx <= dst_gloval_eidx)
          && (dst_gloval_eidx <= src_gloval_eidx) ) { /* # 2 */
    *ov_gloval_sidx = src_gloval_sidx;
    *ov_gloval_size = dst_gloval_eidx - src_gloval_sidx + 1;

  }
  else if(   (src_gloval_sidx <= dst_gloval_sidx)
          && (dst_gloval_eidx <= src_gloval_eidx) ) { /* # 3 */
    *ov_gloval_sidx = dst_gloval_sidx;
    *ov_gloval_size = dst_gloval_eidx - dst_gloval_sidx + 1;
  }
  else if(   (src_gloval_sidx <= dst_gloval_sidx)
          && (dst_gloval_sidx <= src_gloval_eidx)
          && (src_gloval_eidx <  dst_gloval_eidx) ) { /* # 4 */
    *ov_gloval_sidx = dst_gloval_sidx;
    *ov_gloval_size = src_gloval_eidx - dst_gloval_sidx + 1;
  }
  else if(   (dst_gloval_sidx <  src_gloval_sidx)
          && (src_gloval_eidx <  dst_gloval_eidx) ) { /* # 6 */
    *ov_gloval_sidx = src_gloval_sidx;
    *ov_gloval_size = src_gloval_eidx - src_gloval_sidx + 1;
  }
  else {
    return SplFalse;
  }

  return SplTrue;
}


/**
 * get overlap indexes for 3D
 *
 * @param[in]        dims                    Dimension ( 1 or 2 or 3 )
 * @param[in]        src_gloval_sidx        source start index
 * @param[in]        src_gloval_eidx        source end index
 * @param[in]        dst_gloval_sidx        destination start index
 * @param[in]        dst_gloval_eidx        destination end index
 * @param[out]    ov_gloval_sidx        overlap start index
 * @param[out]    ov_gloval_size        overlap size
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
getOverlapAreas(unsigned int dims,
                const size_t* src_gloval_sidx,
                const size_t* src_gloval_eidx,
                const size_t* dst_gloval_sidx,
                const size_t* dst_gloval_eidx,
                size_t* ov_gloval_sidx,
                size_t* ov_gloval_size ) {
  register unsigned int n;

  /* parameter check */
  if(    !src_gloval_sidx || !src_gloval_eidx
      || !dst_gloval_sidx || !dst_gloval_eidx
      || !ov_gloval_sidx || !ov_gloval_size ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "getOverlapAreas arguments is null");
    return SplFalse;
  }

  for(n=0; n<dims; n++) {
    if( getOverlapArea(src_gloval_sidx[n], src_gloval_eidx[n],
                       dst_gloval_sidx[n], dst_gloval_eidx[n],
                       &ov_gloval_sidx[n], &ov_gloval_size[n]) != SplTrue ) {
      return SplFalse;
    }
  }

  return SplTrue;
}

/**
 * copy buffer
 *
 * <pre>
 *              dst_buf        surplus_buf
 *                 |<-dst_buf_size->|<-       surplus_buf_size       ->|
 *  +--------------+----------------+----------------------------------+
 *  |              ||||||||||||||||||//////////////////////////        |
 *  +--------------+----------------+----------------------------------+
 *                 |                |<-surplus_effective_len->|
 *                 +------------------------------------------+
 *          src_buf|==========================================|
 *                 +------------------------------------------+
 *                 |<-             src_buf_size             ->|
 *
 *  surplus_buf_size >= -surplus_effective_len
 * </pre>
 *
 * @param[in]          src_buf_size              Buffer size of src_buf (byte)
 * @param[in]          dst_buf_size              Buffer size of dst_buf (byte)
 * @param[in]          surplus_buf_size          Buffer size of surplus_buf_size (byte)
 * @param[in]          src_buf                   Buffer of src
 * @param[out]      dst_buf                   Buffer of dst
 * @param[out]      surplus_buf               Buffer of surplus
 * @param[out]      surplus_effective_len     Effective data length of surplus_buf (byte)
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
copyBuffer(size_t src_buf_size,
           size_t dst_buf_size,
           size_t surplus_buf_size,
           const unsigned char* src_buf,
           unsigned char* dst_buf,
           unsigned char* surplus_buf,
           size_t* surplus_effective_len) {
  size_t dst_copy_buf_size, tmp_surplus_effective_len;

  /* parameter check */
  if( !src_buf || !dst_buf || !surplus_buf || !surplus_effective_len ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "copyBuffer arguments is null");
    return SplFalse;
  }

  dst_copy_buf_size = 0;
  tmp_surplus_effective_len = 0;
  if( src_buf_size <= dst_buf_size ) {
    dst_copy_buf_size = src_buf_size;
  }
  else {
    dst_copy_buf_size = dst_buf_size;
    tmp_surplus_effective_len = src_buf_size - dst_buf_size;
    if( tmp_surplus_effective_len > surplus_buf_size ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "tmp_surplus_effective_len(=%ld) > surplus_buf_size(=%ld)", tmp_surplus_effective_len, surplus_buf_size);
      return SplFalse;
    }
  }

  memcpy(dst_buf, src_buf, sizeof(char)*dst_copy_buf_size);
  if( tmp_surplus_effective_len > 0 ) {
    memcpy(surplus_buf, src_buf+dst_copy_buf_size, sizeof(char)*tmp_surplus_effective_len);
  }
  *surplus_effective_len = tmp_surplus_effective_len;

  return SplTrue;
}

/**
 * Simple voxel data copy to dst from src.
 *
 * <pre>
 * src and dst array is index(vlen, ix, jx, kx)
 *                                               (ix, jx, kx)
 *   +-----------------------------------------------X
 *   |         +------------+                        |
 *   |         |src         |                        |
 *   |         |            |                        |
 *   |         |            |                        |
 *   |     +---+------------D----------------+       |
 *   |     |   |            |            dst |       |
 *   |     |   |            |                |       |
 *   |     |   |            |                |       |
 *   |     B---C------------+----------------+       |
 *   |         |            |                        |
 *   |         |            |                        |
 *   |         A------------+                        |
 *   X-----------------------------------------------+
 * (0,0,0)
 *  A : src_gloval_sidx (input)
 *  B : dst_gloval_sidx (input)
 *  C : cp_gloval_sidx (output)
 *  D : cp_gloval_eidx (output)
 *  cp_gloval_size = cp_gloval_eidx - cp_gloval_sidx + 1
 * </pre>
 *
 * @param[in]          dims              Dimension(1 -> 1D, 2 -> 2D, 3 -> 3D)
 * @param[in]          src_dlen          src data length in cell (sbx:vlen*dtype, spx:vlen*rlen)
 * @param[in]          dst_vlen          dst data length in cell (sbx:vlen*dtype, spx:vlen*rlen)
 * @param[in]          src_gloval_sidx    src gloval start index (except guide cell)
 * @param[in]          src_size          src array size (except guide cell)
 * @param[in]          src_gc             src guild cell
 * @param[in]          dst_gloval_sidx   dst gloval start index (except guide cell)
 * @param[in]          dst_size           dst array size (except guide cell)
 * @param[in]          dst_gc             dst guild cell
 * @param[in]          cp_gloval_sidx     copy gloval start index (except guide cell)
 * @param[in]          cp_gloval_size     copy array size (except guide cell)
 * @param[in]          cp_gc               copy guild cell
 * @param[in]          src               src array
 * @param[out]      dst                  dst array
 */
SplBool
copySvBuffer(
             unsigned int dims,
             unsigned int src_dlen,
             unsigned int dst_dlen,
             const size_t* src_gloval_sidx,
             const size_t* src_size,
             unsigned int src_gc,
             const size_t* dst_gloval_sidx,
             const size_t* dst_size,
             unsigned int dst_gc,
             const size_t* cp_gloval_sidx,
             const size_t* cp_gloval_size,
             unsigned int cp_gc,
             const unsigned char* src,
             unsigned char* dst ) {
  register long long i, j, k;
  long long tmp_src_gloval_sidx[3], tmp_dst_gloval_sidx[3];
  long long tmp_src_gloval_eidx[3], tmp_dst_gloval_eidx[3];
  long long tmp_cp_gloval_sidx[3], tmp_cp_gloval_eidx[3];
  long long src_sz_gc[3], dst_sz_gc[3];
  long long src_len, dst_len;
  int chk_flag;

  // parameter check
  if( !src_size || !src_gloval_sidx || !dst_size || !dst_gloval_sidx
     || !cp_gloval_sidx || !cp_gloval_size || !src || !dst ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "copySvBuffer arguments is null");
    return SplFalse;
  }

  // check copy size
  chk_flag = 0;
  for(i=0; i<dims; i++) {
    if( cp_gloval_size[i] == 0 ) chk_flag = 1;
  }
  // There is no copy when the cp_gloval_size[n] is 0
  if( chk_flag == 1 ) {
    return SplTrue;
  }

  memset(tmp_src_gloval_sidx, 0, sizeof(long long)*3);
  memset(tmp_dst_gloval_sidx, 0, sizeof(long long)*3);
  memset(tmp_cp_gloval_sidx, 0, sizeof(long long)*3);
  memset(tmp_cp_gloval_eidx, 0, sizeof(long long)*3);
  memset(tmp_src_gloval_eidx, 0, sizeof(long long)*3);
  memset(tmp_dst_gloval_eidx, 0, sizeof(long long)*3);

  src_len = src_dlen;
  dst_len = dst_dlen;
  for(i=0; i<dims; i++) {
    tmp_src_gloval_sidx[i] = src_gloval_sidx[i] - src_gc;
    tmp_dst_gloval_sidx[i] = dst_gloval_sidx[i] - dst_gc;
    tmp_src_gloval_eidx[i] = src_gloval_sidx[i] + (src_size[i]+src_gc) - 1;
    tmp_dst_gloval_eidx[i] = dst_gloval_sidx[i] + (dst_size[i]+dst_gc) - 1;

    tmp_cp_gloval_sidx[i] = cp_gloval_sidx[i] - dst_gc;
    if (tmp_cp_gloval_sidx[i] < tmp_src_gloval_sidx[i]) {
      tmp_cp_gloval_sidx[i] = tmp_src_gloval_sidx[i];
    }
    if (tmp_cp_gloval_sidx[i] < tmp_dst_gloval_sidx[i]) {
      tmp_cp_gloval_sidx[i] = tmp_dst_gloval_sidx[i];
    }

    tmp_cp_gloval_eidx[i] = cp_gloval_sidx[i] + (cp_gloval_size[i]+cp_gc) - 1;
    if (tmp_cp_gloval_eidx[i] > tmp_src_gloval_eidx[i]) {
      tmp_cp_gloval_eidx[i] = tmp_src_gloval_eidx[i];
    }
    if (tmp_cp_gloval_eidx[i] > tmp_dst_gloval_eidx[i]) {
      tmp_cp_gloval_eidx[i] = tmp_dst_gloval_eidx[i];
    }
    src_len *= src_size[i]+2*src_gc;
    dst_len *= dst_size[i]+2*dst_gc;
    src_sz_gc[i] = src_size[i]+2*src_gc;
    dst_sz_gc[i] = dst_size[i]+2*dst_gc;
  }

  long long src_idx;
  long long dst_idx;
  int cp_len;

  if (src_dlen <= dst_dlen) {
    cp_len = src_dlen;
  }
  else {
    cp_len = dst_dlen;
  }
  for(k=tmp_cp_gloval_sidx[2]; k<=tmp_cp_gloval_eidx[2]; k++) {
    for(j=tmp_cp_gloval_sidx[1]; j<=tmp_cp_gloval_eidx[1]; j++) {
      for(i=tmp_cp_gloval_sidx[0]; i<=tmp_cp_gloval_eidx[0]; i++) {
        src_idx =    src_sz_gc[0]*src_sz_gc[1]*(k-tmp_src_gloval_sidx[2])
                   + src_sz_gc[0]*(j-tmp_src_gloval_sidx[1])
                   + (i-tmp_src_gloval_sidx[0]);
        src_idx *= src_dlen;

        dst_idx =    dst_sz_gc[0]*dst_sz_gc[1]*(k-tmp_dst_gloval_sidx[2])
                   + dst_sz_gc[0]*(j-tmp_dst_gloval_sidx[1])
                   + (i-tmp_dst_gloval_sidx[0]);
        dst_idx *= dst_dlen;

        if (src_idx < 0) continue;
        if (src_idx > src_len) continue;
        if (dst_idx < 0) continue;
        if (dst_idx > dst_len) continue;

        memcpy(dst+dst_idx, src+src_idx, cp_len);
      }
    }
  }

  return SplTrue;
}


/**
 * \~english
 * Compress data.
 * @param[in]       ext_buf_sz      original data buffer size
 * @param[in]      cmp_buf_sz       compress buffer size
 * @param[in]       ext_buf          original data buffer
 * @param[in]       cmp_buf          compress buffer
 * @param[out]    compressed_data_sz    compressed data size
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * データを圧縮する。
 * @param[in]       ext_buf_sz      圧縮前データバッファーサイズ
 * @param[in]      cmp_buf_sz       圧縮バッファーサイズ
 * @param[in]       ext_buf          圧縮前データバッファー
 * @param[out]    cmp_buf          圧縮バッファー
 * @param[out]    compressed_data_sz    圧縮後データサイズ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
compSvDataBlock(size_t ext_buf_sz,
                size_t cmp_buf_sz,
                unsigned char* ext_buf,
                unsigned char* cmp_buf,
                size_t* compressed_data_sz) {
  size_t tmp_cmp_buf_sz;

  /* parameter check */
  if( !ext_buf || !cmp_buf || !compressed_data_sz ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "compSvDataBlock arguments is null");
    return SplFalse;
  }

  tmp_cmp_buf_sz = cmp_buf_sz;
  if( compress(cmp_buf, &tmp_cmp_buf_sz, ext_buf, ext_buf_sz) != Z_OK) {
    // error
    SPL_ERROR_INFO(SPL_ERR_UNCOMPRESS_DATA, "tmp_cmp_buf_sz=%ld, ext_buf_sz=%ld", tmp_cmp_buf_sz, ext_buf_sz);
    return SplFalse;
  }
  *compressed_data_sz = tmp_cmp_buf_sz;

  return SplTrue;
}

/*
 *  decompSvDataBlock :
 *    Extended compress data.
 *
 */

/**
 * \~english
 * Extended compress data.
 * @param[in]        eType          Endian match type
 * @param[in]        dtype          Data type<br>
 *                               - 1 -> unsigned char<br>
 *                               - 2 -> unsigned short<br>
 *                               - 4 -> unsigned int or float<br>
 *                               - 8 -> double<br>
 * @param[in]      cmp_buf_sz       compress buffer size
 * @param[in]       ext_buf_sz      decompressed data buffer size
 * @param[in]       cmp_buf          compress buffer
 * @param[in]       ext_buf          decompressed data buffer
 * @param[out]    extended_data_sz    decompressed data size
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * データを解凍する。
 * @param[in]        eType          エンディアンタイプ
 * @param[in]        dtype          データタイプ<br>
 *                               - 1 -> unsigned char<br>
 *                               - 2 -> unsigned short<br>
 *                               - 4 -> unsigned int or float<br>
 *                               - 8 -> double<br>
 * @param[in]      cmp_buf_sz       圧縮データバッファーサイズ
 * @param[in]       ext_buf_sz      解凍バッファーサイズ
 * @param[in]       cmp_buf         圧縮データバッファー
 * @param[out]    ext_buf         解凍バッファー
 * @param[out]    extended_data_sz    解凍後データサイズ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
decompSvDataBlock(EMatchType eType,
                  unsigned int dtype,
                  size_t cmp_buf_sz,
                  size_t ext_buf_sz,
                  unsigned char* cmp_buf,
                  unsigned char* ext_buf,
                  size_t* extended_data_sz) {
  size_t tmp_ext_buf_sz;

  /* parameter check */
  if( !cmp_buf || !ext_buf || !extended_data_sz ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "decompSvDataBlock arguments is null");
    return SplFalse;
  }

  tmp_ext_buf_sz = ext_buf_sz;
  if( uncompress(ext_buf, &tmp_ext_buf_sz, cmp_buf, cmp_buf_sz) != Z_OK) {
    // error
    SPL_ERROR_INFO(SPL_ERR_UNCOMPRESS_DATA, "tmp_ext_buf_sz=%ld, cmp_buf_sz=%ld", tmp_ext_buf_sz, cmp_buf_sz);
    return SplFalse;
  }
  *extended_data_sz = tmp_ext_buf_sz;

  if( eType == UnMatch ) {
    switch( dtype ) {
      case 2:
        SBSwapVec(ext_buf, tmp_ext_buf_sz/dtype);
        break;
      case 4:
        BSwapVec(ext_buf, tmp_ext_buf_sz/dtype);
        break;
      default:
        break;
    }
  }

  return SplTrue;
}



/**
 * \~english
 * get overlap index and length from gloval start index and length
 * @param[in]      dims              dimension
 * @param[in]      gloval_size       gloval size
 * @param[in]      gloval_sidx       gloval start index
 * @param[in]      gloval_len        gloval data length
 * @param[in]      bnd_sidx          boundary start index
 * @param[in]      bnd_eidx          boundary end index
 * @param[out]      ov_sidx          overlaped start index
 * @param[out]      ov_len          overlaped data length
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 始点インデックスからデータ長までの範囲と重なる領域の始点インデックス、データ長を取得する。
 * @param[in]      dims              次元数
 * @param[in]      gloval_size       全体領域のサイズ
 * @param[in]      gloval_sidx       全体領域の始点インデックス
 * @param[in]      gloval_len        全体領域の始点インデックスからのデータ長さ
 * @param[in]      bnd_sidx          重なる境界領域の始点インデックス
 * @param[in]      bnd_eidx          重なる境界領域の終点インデックス
 * @param[out]      ov_sidx          重なった領域の始点インデックス
 * @param[out]      ov_len          重なった領域のデータ長さ
 * @return    success(=true)/fault(=false)
 */
SplBool
getOverlapLength(unsigned int dims,
                const size_t* gloval_size,
                const size_t* gloval_sidx,
                const size_t gloval_len,
                const size_t* bnd_sidx,
                const size_t* bnd_eidx,
                size_t* ov_sidx,
                size_t* ov_len ) {
  register unsigned int n;
  register size_t i;
  size_t gloval_eidx[3] = {0, 0, 0};
  size_t bnd_sta[3] = {0, 0, 0};
  size_t bnd_end[3] = {0, 0, 0};
  size_t sta_pos;
  size_t end_pos;
  size_t plus;
  size_t sta_glv_pos;
  size_t end_glv_pos;
  size_t sta_bnd_pos;
  size_t end_bnd_pos;
  size_t ov_gloval_sidx;
  size_t ov_gloval_size;

  /* parameter check */
  if(    !gloval_size || !gloval_sidx
      || !bnd_sidx || !bnd_eidx
      || !ov_sidx || !ov_len ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "getOverlapLength arguments is null");
    return SplFalse;
  }

  *ov_len = 0;
  for(n=0; n<dims; n++) {
    if (bnd_sidx[n] > bnd_eidx[n]) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "Boundry index invalid.(bnd_sidx[%d]:%d > bnd_eidx[%d]:%d)",
                                  n, bnd_sidx[n], n, bnd_eidx[n]);
      return SplFalse;
    }
  }

  // get array index
  getPosFromIndex(dims, gloval_size, gloval_sidx, &sta_glv_pos);
  end_glv_pos = sta_glv_pos + gloval_len - 1;
  getIndexFromPos(dims, end_glv_pos, gloval_size, gloval_eidx);

  // get boundary array inde
  getPosFromIndex(dims, gloval_size, bnd_sidx, &sta_bnd_pos);
  getPosFromIndex(dims, gloval_size, bnd_eidx, &end_bnd_pos);

  if (sta_glv_pos > end_bnd_pos) {
    // not overlap
    *ov_len = 0;
    return SplTrue;
  }

  if (end_glv_pos < sta_bnd_pos) {
    // not overlap
    *ov_len = 0;
    return SplTrue;
  }

  sta_pos = (sta_glv_pos < sta_bnd_pos) ? sta_bnd_pos : sta_glv_pos;
  end_pos = (end_glv_pos > end_bnd_pos) ? end_bnd_pos : end_glv_pos;

  plus = 0;
  for (i=sta_pos;i<=end_pos;i+= plus) {
    getIndexFromPos(dims, i, gloval_size, bnd_sta);
    if (i==sta_pos) {
      if (sta_pos == sta_glv_pos) {
        plus = gloval_size[0]-gloval_sidx[0];
      }
      else {    // sta_pos == sta_bnd_pos
        plus = gloval_size[0]-bnd_sidx[0];
      }
    }
    else {
      plus = gloval_size[0];
    }
    if (i+plus-1 <= end_pos) {
      getIndexFromPos(dims, i+plus-1, gloval_size, bnd_end);
    }
    else {
      getIndexFromPos(dims, end_pos, gloval_size, bnd_end);
    }

    int isov = 1;
    for (n=1; n<dims; n++) {
      if (bnd_sta[n] >= bnd_sidx[n] && bnd_sta[n] <= bnd_eidx[n]) {
        isov = 1;
      }
      else {
        isov = 0;
        break;
      }
    }

    if (isov) {
      ov_gloval_sidx = 0;
      ov_gloval_size = 0;
      getOverlapArea(bnd_sta[0], bnd_end[0],
                    bnd_sidx[0], bnd_eidx[0],
                    &ov_gloval_sidx, &ov_gloval_size);
      if (ov_gloval_size > 0) {
        if (*ov_len == 0) {
          ov_sidx[0]=ov_gloval_sidx;
          for(n=1; n<dims; n++) ov_sidx[n] = bnd_sta[n];
        }
        *ov_len += ov_gloval_size;
      }
    }
    if (plus <= 0) break;
  }
  if (*ov_len <= 0) {
    return SplFalse;
  }

  return SplTrue;
}


/**
 * \~english
 * get position from index
 * @param[in]       dims        Dimension ( 1 or 2 or 3 )
 * @param[in]      vsz         Voxel size
 * @param[in]       idx        Index
 * @param[out]    position    Length of 1D
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * ボクセルインデックスから  始点(0,0,0)からの位置を取得する。
 * @param[in]       dims        次元数 ( 1 or 2 or 3 )
 * @param[in]      vsz         ボクセルサイズ
 * @param[out]   idx         ボクセルインデックス
 * @param[in]      position    始点(0,0,0)からの長さ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
getPosFromIndex(unsigned int dims,
                const size_t* vsz,
                const size_t* idx,
                size_t* position) {
  register int n;
  size_t surplus;
  size_t Coef[3];

  if( !vsz || !idx || (dims > 3) ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "getPosFromIndex arguments is null");
    return SplFalse;
  }

  memset(Coef, 0, sizeof(size_t)*3);
  Coef[0] = 1;
  for(n=1; n<dims; n++) Coef[n] = Coef[n-1]*vsz[n-1];

  *position = 0;
  for(n=0; n<dims; n++) {
    *position += Coef[n]*idx[n];
  }

  return SplTrue;
}

/**
 * \~english
 * get index by add length
 * @param      dims             dimension
 * @param      wsize            all voxel size
 * @param      bnd_size         node voxel size
 * @param      bnd_sta          node start index
 * @param      add_sta          add start index in node voxel
 * @param      add_len           add length in node voxel
 * @param      idx               added index
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * ノード領域の始点からの加算長さの位置のインデックスを取得する。
 *
 * ノード領域内の開始位置からノード領域内で指定データ長さ分を加算したノード位置を求める。<br>
 * ノード領域のバッファー内のデータ範囲を全体領域にコピーする際に開始、終了インデックスを算出する時に使用する。
 * @param      dims             次元
 * @param      wsize            全計算領域サイズ
 * @param      bnd_size         ノード領域サイズ
 * @param      bnd_sta          ノード領域の始点インデックス
 * @param      add_sta          加算開始インデックス
 * @param      add_len           加算長さ
 * @param      idx               加算位置のインデックス
 * @return    success(=true)/fault(=false)
 */
SplBool
getLocalIndexFromPos(unsigned int dims,
                const size_t* bnd_size,
                const size_t* bnd_sta,
                const size_t* add_sta,
                const size_t add_len,
                size_t* idx) {
  register int n;

  if( (dims > 3) || !bnd_size || !bnd_sta || !idx ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "getLocalIndexFromPos arguments is null");
    return SplFalse;
  }

  // get position in local voxel
  size_t local_sta[3];
  for (n=0; n< dims; n++) {
    if (add_sta[n] < bnd_sta[n]) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "add start index is less than local voxel start index.(add_sta[%d]:%d > bnd_sta[%d]:%d).",
                                  n, add_sta[n], n, bnd_sta[n]);
      return SplFalse;
    }
    local_sta[n] = add_sta[n] - bnd_sta[n];
  }

  // get local position in local voxel
  size_t loc_sta_pos;
  getPosFromIndex(dims, bnd_size, local_sta, &loc_sta_pos);

  size_t loc_total = 1;
  for (n=0; n< dims; n++) {
    loc_total *= bnd_size[n];
  }

  // check over max index
  if (loc_total <= loc_sta_pos+add_len) {
    // over max index
    for (n=0; n< dims; n++) {
      idx[n] = bnd_sta[n]+bnd_size[n]-1;
    }
    return SplTrue;
  }

  // get added possition index in local voxel
  size_t loc_idx[3];
  getIndexFromPos(dims, loc_sta_pos+add_len, bnd_size, loc_idx);

  // convert local index to global index
  for (n=0; n< dims; n++) {
    idx[n] = loc_idx[n] + bnd_sta[n];
  }

  return SplTrue;
}


SplBool
copyLengthBuffer(
             unsigned int dims,
             const size_t src_dtype,
             const size_t* src_size,
             const size_t* src_sidx,
             const size_t* src_eidx,
             const size_t* cp_sidx,
             const size_t cp_len,
             const size_t dst_len,
             const unsigned char* src_buf,
             unsigned char* dst_buf )
{
  register long long i, j, k;

  // parameter check
  if( !src_size || !src_sidx || !src_eidx || !src_buf || !dst_buf
      || src_dtype <= 0 || dst_len <= 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "copySvBuffer arguments is null");
    return SplFalse;
  }

  // check copy indexes
  for(i=0; i<dims; i++) {
    if( src_sidx[i] > src_eidx[i] ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "Invalid src indexes, i=%d, sidx=%ld, eidx=%ld", src_sidx[i], src_eidx[i]);
      return SplFalse;
    }
  }

  size_t sidx[3] = {0,0,0};
  size_t eidx[3] = {0,0,0};
  size_t size[3] = {1,1,1};
  size_t csidx[3] = {0,0,0};
  size_t src_len = src_dtype;
  for (i=0; i<dims; i++) {
    sidx[i] = src_sidx[i];
    eidx[i] = src_eidx[i];
    size[i] = src_size[i];
    csidx[i] = cp_sidx[i];
    src_len *= src_size[i];
  }
  size_t len = 0;
  size_t cp_bytes = src_eidx[0] - src_sidx[0] + 1;
  cp_bytes *= src_dtype;
  size_t idx = 0, cp_eidx = 0, dst_idx = 0;
  size_t rest_len = 0;
  size_t start_index = size[0]*size[1]*csidx[2]+size[0]*csidx[1]+csidx[0];
  for(k=csidx[2]; k<=eidx[2]; k++) {
    if (rest_len == 0) j=csidx[1];
    else j=sidx[1];
    for(; j<=eidx[1]; j++) {
      if (rest_len == 0) {
        idx =    size[0]*size[1]*k
                     + size[0]*j
                     + csidx[0];
        len = (eidx[0]-csidx[0]+1)*src_dtype;
      }
      else {
        idx =    size[0]*size[1]*k
                     + size[0]*j
                     + sidx[0];
        len = cp_bytes;
      }
      idx *= src_dtype;
      if (rest_len + len > cp_len*src_dtype) {
        len = cp_len*src_dtype - rest_len;
      }

      if (idx+len > src_len) {
        // error
        SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "Invalid memcpy, cp_sidx=%ld, cp_len=%ld, src_len=%ld", cp_sidx, cp_len, src_len);
        return SplFalse;
      }
      if (dst_idx+len > dst_len) {
        // error
        SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "Invalid memcpy, dst_idx=%ld, cp_len=%ld, dst_len=%ld", dst_idx, cp_len, dst_len);
        return SplFalse;
      }
      memcpy(dst_buf+dst_idx, src_buf+idx, len);
      dst_idx += len;
      rest_len += len;
      if (rest_len >= cp_len*src_dtype) {
        return SplTrue;
      }
    }
  }

  return SplTrue;
}



#if defined(__cplusplus)
}
#endif /* extern "C" */

#undef SPL_SV_MAX_BUF_SIZE

