/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file splfio_sv_write.c
 * Prototype declaration of file write subroutine related to Sv(Simple Voxel).
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file splfio_sv_write.c
 * Sv(Simple Voxel)の低レベルファイル書込関数群ソースファイル
 * @author     hira
 * @date       2009/02/20
 */


#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <zlib.h>
#include "splfio_sv_write.h"
#include "splfio_sv_util.h"

#if defined(__cplusplus)
extern "C" {
#endif

//int spl_err_code;

/* -----------------------------------------------------------------------*/
/*  Simple-Voxel File write procedure                                      */
/* -----------------------------------------------------------------------*/

/**
 * write Attribute record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param             fp                file pointer
 * @param              fmt               File Type
 * @param             endian_type       Endian match type
 * @param              dims              Dimension
 * @param              vlen              Vector length
 * @param              dtype             Data type(1or2or4)
 * @param              gc                Size of guide cell
 * @param              rlen              Real size(4or8)
 * @param              crddef            coords define
 * @param              aux                block_size
 * @param              block_size        Size of number of voxel  in a compressed data record.
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvAttrRecord(FILE* fp,                     /* input */
                  SplFileType fmt,              /* input */
                  unsigned int dims,            /* input */
                  unsigned int vlen,            /* input */
                  unsigned int dtype,           /* input */
                  size_t gc,                    /* input */
                  unsigned int rlen,            /* input */
                  unsigned int crddef,          /* input */
                  unsigned int aux,             /* input */
                  unsigned long long block_size /* input */ ) {
  int magick;
  char magick_str[4];
  int tmp_gc;

  /* parameter check */
  if( !fp ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "writeSvAttrRecord arguments is null");
    return SplFalse;
  }

  /* seek file */
  if( fseek(fp, 0, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is 0.");
    return SplFalse;
  }

  /* write magick */
  switch ( fmt ) {
  case SPH_FMT_SBX:
    strcpy(magick_str, SPL_SBX_MAGICK_STR);
    magick = *(int*)magick_str;
    break;
  case SPH_FMT_SPX:
    strcpy(magick_str, SPL_SPX_MAGICK_STR);
    magick = *(int*)magick_str;
    break;
  default:
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "invalid fmt=%d", (int)fmt);
    return SplFalse;
  }
  if( fwrite(&magick, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "magick write error");
    return SplFalse;
  }

  /* write dims */
  if( fwrite(&dims, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "dims write error");
    return SplFalse;
  }

  /* write vlen */
  if( fwrite(&vlen, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "vlen write error");
    return SplFalse;
  }

  /* write dtype */
  if( fwrite(&dtype, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "dtype write error");
    return SplFalse;
  }

  /* write gc */
  tmp_gc = (int)gc;
  if( fwrite(&tmp_gc, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "gc write error");
    return SplFalse;
  }

  /* write rlen */
  if( fwrite(&rlen, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "rlen write error");
    return SplFalse;
  }

  /* write crddef */
  if( fwrite(&crddef, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "crddef write error");
    return SplFalse;
  }

  /* write aux */
  if( fwrite(&aux, sizeof(int), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "aux write error");
    return SplFalse;
  }

  /* write block size */
  if( fwrite(&block_size, sizeof(long long), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "block_size write error");
    return SplFalse;
  }

  return SplTrue;
}


/**
 * write Size record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param             fp                file pointer
 * @param              file_offset      file seek length
 * @param              dims             Dimension
 * @param             size              Voxel size
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvSizeRecord(FILE* fp,          /* input */
                  long file_offset,  /* input */
                  unsigned int dims, /* input */
                  const size_t* size /* input */ ) {
  register unsigned int n;
  unsigned long long sz[3];

  if( !fp || !size ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "writeSvSizeRecord arguments is null");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  /* write size */
  sz[0] = sz[1] = sz[2] = 1;
  for(n=0; n<dims; n++) {
    sz[n] = (unsigned long long)size[n];
  }
  if( fwrite(sz, sizeof(long long), 3, fp) != 3 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "sz write error");
    return SplFalse;
  }

  return SplTrue;
}


/**
 * write Origin record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param             fp                file pointer
 * @param              file_offset      file seek length
 * @param              dims             Dimension
 * @param              rlen             Real Data size
 * @param              org              Voxel Origin
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvOriginRecord(FILE* fp,          /* input */
                    long file_offset,  /* input */
                    unsigned int dims, /* input */
                    int rlen,          /* input */
                    const double* org  /* input */ ) {
  register unsigned int n;
  if( !fp || !org ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "writeSvOriginRecord arguments is null");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  /* write origin */
  if( rlen == 4 ) {       /* single */
    float o[3];
    memset(o, 0, sizeof(float)*3);
    for(n=0; n<dims; n++) { o[n] = (float)org[n]; }
    if( fwrite(o, sizeof(float), 3, fp) != 3 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "origin write error");
      return SplFalse;
    }
  }
  else if( rlen == 8 ) {  /* double */
    double o[3];
    memset(o, 0, sizeof(double)*3);
    for(n=0; n<dims; n++) { o[n] = (double)org[n]; }
    if( fwrite(o, sizeof(double), 3, fp) != 3 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "origin write error");
      return SplFalse;
    }
  }
  else {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_SPLSV_DTYPE, "invalid rlen=%d", rlen);
    return SplFalse;
  }

  return SplTrue;
}


/**
 * write Pitch record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param           fp                file pointer
 * @param            file_offset      file seek length
 * @param            dims             Dimension
 * @param              rlen               Real Data size
 * @param        pit              Voxel Pitch
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvPitchRecord(FILE* fp,          /* input */
                   long file_offset,  /* input */
                   unsigned int dims, /* input */
                   int rlen,          /* input */
                   const double* pit  /* input */ ) {
  register unsigned int n;

  if( !fp || !pit ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "writeSvPitchRecord arguments is null");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  /* write pitch */
  if( rlen == 4 ) {       /* single */
    float p[3];
    memset(p, 0, sizeof(float)*3);
    for(n=0; n<dims; n++) {
      p[n] = (float)pit[n];
    }
    if( fwrite(p, sizeof(float), 3, fp) != 3 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "pitch write error");
      return SplFalse;
    }
  }
  else if( rlen == 8 ) {  /* double */
    double p[3];
    memset(p, 0, sizeof(double)*3);
    for(n=0; n<dims; n++) {
      p[n] = (double)pit[n];
    }
    if( fwrite(p, sizeof(double), 3, fp) != 3 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "pitch write error");
      return SplFalse;
    }
  }
  else {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_SPLSV_DTYPE, "invalid rlen=%d", rlen);
    return SplFalse;
  }

  return SplTrue;
}


/**
 * write Time record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.
 *
 * @param             fp                file pointer
 * @param              file_offset      file seek length
 * @param              rlen               Real Data size
 * @param              step              step count
 * @param              time              time
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvTimeRecord(FILE* fp,          /* input */
                  long file_offset,  /* input */
                  int rlen,          /* input */
                  long long step,    /* input */
                  double time        /* input */ ) {
  if( !fp ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "writeSvTimeRecord arguments is null");
    return SplFalse;
  }

  if( fseek(fp, file_offset, SEEK_SET) != 0 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_SEEK_FILE, "fseek length is %ld.", file_offset);
    return SplFalse;
  }

  /* write step */
  if( fwrite(&step, sizeof(long long), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "step write error");
    return SplFalse;
  }

  /* write time */
  if( rlen == 4 ) {       /* single */
    float ftime;
    ftime = (float)time;
    if( fwrite(&ftime, sizeof(float), 1, fp) != 1 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "time write error");
      return SplFalse;
    }
  }
  else if( rlen == 8 ) {  /* double */
    if( fwrite(&time, sizeof(double), 1, fp) != 1 ) {
      // error
      SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "time write error");
      return SplFalse;
    }
  }
  else {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_SPLSV_DTYPE, "invalid rlen=%d", rlen);
    return SplFalse;
  }

  return SplTrue;
}


/**
 * write compress block size for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.<br>
 * caution:<br>
 *      This procedure doesn't seek to the fp.<br>
 * @param             fp                file pointer
 * @param              comp_blk_sz       compress block size
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvBlkSz(FILE* fp,          /* input */
             size_t comp_blk_sz /* input */ ) {
  if( !fp ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "writeSvBlkSz arguments is null");
    return SplFalse;
  }

  unsigned long long tmp_comp_blk_sz;
  tmp_comp_blk_sz = (unsigned long long)comp_blk_sz;
  if( fwrite(&tmp_comp_blk_sz, sizeof(long long), 1, fp) != 1 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "tmp_comp_blk_sz write error");
    return SplFalse;
  }

  return SplTrue;
}


/**
 * write block data size of data record. for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.<br>
 * caution:<br>
 *      This procedure doesn't seek to the fp.<br>
 * @param             mode              write mode
 * @param             writeRank          Rank that write a file
 * @param             grp               Communication group
 * @param             fp                file pointer
 * @param              comp_blk_sz       block data size
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvDataBlkSz(int mode,
                 int writeRank,
                 SPL_Comm grp,
                 FILE* fp,
                 size_t comp_blk_sz) {
  if( mode != 0 ) {
    int myrank;
    SPL_Comm_rank(grp, &myrank);
    if( myrank == writeRank ) {
      if( writeSvBlkSz(fp, comp_blk_sz) != SplTrue ) {
        return SplFalse;
      }
    }
  }
  else {
    if( writeSvBlkSz(fp, comp_blk_sz) != SplTrue ) {
      return SplFalse;
    }
  }

  return SplTrue;
}


/**
 * write data of data record. for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.<br>
 * caution:<br>
 *      This procedure doesn't seek to the fp.<br>
 * @param             fp                file pointer
 * @param              buf_sz             buffer size
 * @param              buf               write budder
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvDtBlk(FILE* fp,
             size_t buf_sz,
             const unsigned char* buf ) {
  if( !fp || !buf ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "writeSvDtBlk arguments is null");
    return SplFalse;
  }

  if( fwrite(buf, sizeof(char), buf_sz, fp) != buf_sz ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "data block write error");
    return SplFalse;
  }

  return SplTrue;
}


/**
 * write data of data record. for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.<br>
 * caution:<br>
 *      This procedure doesn't seek to the fp.<br>
 * @param             mode              write mode
 * @param             writeRank          Rank that write a file
 * @param             grp               Communication group
 * @param             fp                file pointer
 * @param              buf_sz             buffer size
 * @param              buf               write budder
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvDataBlock(int mode,
                 int writeRank,
                 SPL_Comm grp,
                 FILE* fp,
                 size_t buf_sz,
                 const unsigned char* buf ) {
  /* parameter check */
  if( !buf ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "writeSvDataBlock arguments is null");
    return SplFalse;
  }

  /* write compress data from file */
  if( mode == 0 ) {
    if( writeSvDtBlk(fp, buf_sz, buf) != SplTrue ) {
      return SplFalse;
    }
  }
  else {
    int myrank;
    SPL_Comm_rank(grp, &myrank);
    if( myrank == writeRank ) {
      if( writeSvDtBlk(fp, buf_sz, buf) != SplTrue ) {
        return SplFalse;
      }
    }
  }

  return SplTrue;
}


/**
 * write extend record for Simple-Voxel File
 *
 * This subroutine is not possible to use it directly in a parallel mode.<br>
 * caution:<br>
 *       This procedure seeks fp in front of Data-Record.<br>
 *      file_offset specifies the position of DataRecord.
 * @param             fp                file pointer
 * @param             dims              Dimension
 * @param              wsz               gloval size
 * @param              widx               gloval start index
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvExtRecord(FILE* fp,          /* input */
                 unsigned int dims, /* input */
                 const size_t* wsz, /* input */
                 const size_t* widx /* input */ ) {
  register unsigned int n;
  unsigned long long tmp_wsz[3], tmp_widx[3];

  if( !fp || !wsz || !widx ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "writeSvExtRecord arguments is null");
    return SplFalse;
  }

  tmp_wsz[0] = tmp_wsz[1] = tmp_wsz[2] = 1;
  tmp_widx[0] = tmp_widx[1] = tmp_widx[2] = 0;
  for(n=0; n<dims; n++) {
    tmp_wsz[n] = (unsigned long long)wsz[n];
    tmp_widx[n] = (unsigned long long)widx[n];
  }

  /* write world voxel size */
  if( fwrite(tmp_wsz, sizeof(long long), 3, fp) != 3 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "tmp_wsz write error");
    return SplFalse;
  }

  /* write world index */
  if( fwrite(tmp_widx, sizeof(long long), 3, fp) != 3 ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_WRITE_FILE, "tmp_widx write error");
    return SplFalse;
  }

  return SplTrue;
}



/**
 * write header record in SBX/SPX File
 *
 * @param         mode              write mode
 * @param        writeRank         Rank that write a file (It is effective at mode 1 or 2.)
 * @param        grp              Communication group
 * @param        fp                file pointer
 * @param        cmp_flg          compress flag
 *                                   - 0     : data block is uncompressed
 *                                   - other : data block is compressed
 * @param        vox_num_per_blk   Size of number of voxel  in a compressed data record.
 * @param        dims              Dimension
 * @param        vlen              Vector length
 * @param        data_len            Data type(1or2or4)
 * @param        vsz              size of voxel in my node
 * @param        data              data of voxel
 * @param         wsize            total size of nodes (significant at mode 1 or 2)
 * @param         sta_idx          gloval start index of nodes (significant at mode 1 or 2)
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
writeSvDataRecord(int mode,
                   int writeRank,
                   SPL_Comm grp,
                   FILE* fp,
                   unsigned long long cmp_flg,
                   size_t vox_num_per_blk,
                   unsigned int dims,
                   unsigned int vlen,
                   unsigned int data_len,
                   const size_t* vsz,
                   const unsigned char* data,
                   const size_t* wsize,
                   const size_t* sta_idx)
{
//  if( !vsz || !data || !wsize || !sta_idx) {
  if( !vsz || !wsize || !sta_idx) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "writeSvDataRecord arguments is null");
    return SplFalse;
  }

  register unsigned int n, i;

  if( mode == 0 ) {
    size_t data_size;
    data_size = data_len * vlen;
    for(n=0; n<dims; n++) data_size *= vsz[n];
    if (data_size == 0) {
      return SplTrue;
    }
    if( cmp_flg ) { /* write compress data */
      size_t size_per_blk;
      size_per_blk = vox_num_per_blk * vlen * data_len;
      long long remain_len;
      size_t pp, cmp_buf_sz;
      unsigned char* cmp_buf;
      cmp_buf_sz = size_per_blk;
      int plus = (int)((float)(sizeof(char)*cmp_buf_sz)*0.001) + 12;
      if( (cmp_buf = (unsigned char*)malloc(sizeof(char)*cmp_buf_sz+plus)) == NULL ) {
        // error
        SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld", sizeof(char)*cmp_buf_sz+plus);
        return SplFalse;
      }
      remain_len = data_size;
      pp = 0;
      while( remain_len > 0 ) {

        size_t tmp_cmp_buf_sz;
        tmp_cmp_buf_sz = cmp_buf_sz+plus;

        int ret = Z_OK;
        if( (ret = compress(cmp_buf, &tmp_cmp_buf_sz, data+pp, cmp_buf_sz)) != Z_OK ) {
          // error
          SPL_ERROR_INFO(SPL_ERR_COMPRESS_DATA, "tmp_cmp_buf_sz=%ld, cmp_buf_sz=%ld", tmp_cmp_buf_sz, cmp_buf_sz);
          free(cmp_buf);
          return SplFalse;
        }

        if( writeSvDataBlkSz(mode, writeRank, grp, fp, tmp_cmp_buf_sz) == SplFalse ) {
          free(cmp_buf);
          return SplFalse;
        }
        if( writeSvDataBlock(mode, writeRank, grp, fp,
                            tmp_cmp_buf_sz, cmp_buf) == SplFalse ) {
          free(cmp_buf);
          return SplFalse;
        }

        pp += cmp_buf_sz;
        remain_len -= cmp_buf_sz;
        if( remain_len < cmp_buf_sz ) cmp_buf_sz = remain_len;
      } /* while( remain_len > 0 ) */
      free(cmp_buf);
    }
    else { /* write uncompress data */
      if( writeSvDataBlkSz(mode, writeRank, grp, fp, data_size) == SplFalse ) {
        return SplFalse;
      }
      if( writeSvDataBlock(mode, writeRank, grp, fp, data_size, data) == SplFalse ) {
        return SplFalse;
      }
    }
    if( writeSvDataBlkSz(mode, writeRank, grp, fp, 0) == SplFalse ) {
      return SplFalse;
    }
  }
  //
  // mode 1 Gather Data
  //
  else if( mode != 0 ) {
    register unsigned long long i, j, k;

    int myrank, proc_size;
    SPL_Comm_rank(grp, &myrank);
    SPL_Comm_size(grp, &proc_size);

    unsigned char* write_buf = NULL;
    unsigned char* rec_buf = NULL;
    unsigned char* send_buf = NULL;
    unsigned char* cmp_buf = NULL;
    long long *send_lens = NULL;
    long long *nodes_size = NULL;
    long long  *nodes_sta_idx = NULL;
    long long  *sends_sta_idx = NULL;
    long long my_size[3] = {0, 0, 0};
    long long  my_sta[3] = {0, 0, 0};
    long long my_send_sta[3] = {0, 0, 0};
    long long tmp_write_sta_idx[3] = {0, 0, 0};
    long long tmp_write_size[3] = {0, 0, 0};
    size_t write_size[3] = {0, 0, 0};
    size_t write_sta_idx[3] = {0, 0, 0};
    size_t write_buf_sz;
    size_t write_pos_idx = 0;
    long long write_len = 0;
    long long dtype_len = 0;
    long long whole_byte_size;
    long long whole_len;
    int dims_value;
    int cmp_plus = 0;

    whole_byte_size = 0;
    if (myrank == writeRank) {
      write_len = vox_num_per_blk;
      dtype_len = vlen * data_len;

      // set write start_index and size
      for(n=0; n<dims; n++) {
//        tmp_write_sta_idx[n] = sta_idx[n];
        tmp_write_size[n] = wsize[n];
      }
    }

    // BCast
    dims_value = dims;
    SPL_Bcast(&dims_value, 1, SPL_INT, writeRank, grp);
    SPL_Bcast(&write_len, 1, SPL_LONG_LONG, writeRank, grp);
    SPL_Bcast(&dtype_len, 1, SPL_LONG_LONG, writeRank, grp);
//    SPL_Bcast(&write_sta_idx, 3, SPL_LONG_LONG, writeRank, grp);
    SPL_Bcast(&tmp_write_size, 3, SPL_LONG_LONG, writeRank, grp);
    dims = dims_value;
    if (dims <= 0) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "dims size is zero.");
      return SplFalse;
    }

    // get whole length
    whole_len = 1;
    for(n=0; n<dims; n++) {
//      write_sta_idx[n] = tmp_write_sta_idx[n];
      write_size[n] = tmp_write_size[n];
      whole_len *= write_size[n];
    }
    whole_byte_size = whole_len*dtype_len;
    write_buf_sz = write_len * dtype_len;
    if (write_buf_sz <= 0) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "write buffer size is zero.");
      return SplFalse;
    }
    if (whole_byte_size <= 0) {
      SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "whole size is zero.");
      return SplFalse;
    }

    for (n=0; n<dims; n++) {
      my_sta[n] = (long long)sta_idx[n];
    }

    // get my send area
    size_t my_ov_sidx[3] = {0,0,0};
    size_t my_ov_eidx[3] = {0,0,0};
    size_t my_ov_size[3] = {0,0,0};
    size_t my_ov_len = 0;
    size_t end_idx[3] = {0,0,0};
    size_t end_widx[3] = {0,0,0};
    for (i=0; i<dims; i++) {
      end_idx[i] = sta_idx[i] + vsz[i] - 1;
      end_widx[i] = write_sta_idx[i] + write_size[i] - 1;
    }
    if (getOverlapAreas(dims, sta_idx, end_idx, write_sta_idx, end_widx, my_ov_sidx, my_ov_size) == SplTrue) {
      my_ov_len = 1;
      for (n=0; n<dims; n++) {
        my_ov_eidx[n] = my_ov_sidx[n] + my_ov_size[n] - 1;
        my_size[n] = (long long)my_ov_size[n];
      }
    }

    // allocate buffer memory
    // send buffer from nodes
    send_buf = (unsigned char*)malloc(sizeof(char)*write_buf_sz);
    if (!send_buf) {
      // error
      SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld", sizeof(char)*write_buf_sz);
      return SplFalse;
    }
    memset(send_buf, 0, sizeof(char)*write_buf_sz);

    if (myrank == writeRank) {
      // write buffer to file
      write_buf = (unsigned char*)malloc(sizeof(char)*write_buf_sz);
      // reveive buffer from nodes
      rec_buf = (unsigned char*)malloc(sizeof(char)*write_buf_sz);
      // send data length list from nodes
      send_lens = (long long*)malloc(sizeof(long long)*proc_size);
      // nodes size list
      nodes_size = (long long*)malloc(sizeof(long long)*proc_size*dims);
      // nodes start index list
      nodes_sta_idx = (long long*)malloc(sizeof(long long)*proc_size*dims);
      // send data start index list
      sends_sta_idx = (long long*)malloc(sizeof(long long)*proc_size*dims);
      // compress buffer
      cmp_plus = (int)((float)(sizeof(char)*write_buf_sz)*0.001) + 12;
      cmp_buf = (unsigned char*)malloc(sizeof(char)*write_buf_sz + cmp_plus);
      if (!write_buf || !rec_buf || !send_buf || !send_lens || !nodes_size
          || !nodes_sta_idx || !sends_sta_idx ) {
        // error
        SPL_ERROR_INFO(SPL_ERR_MEMORY_ALLOCATE, "malloc length = %ld", sizeof(char)*write_buf_sz);
        if (write_buf) free(write_buf);
        if (rec_buf) free(rec_buf);
        if (send_buf) free(send_buf);
        if (send_lens) free(send_lens);
        if (nodes_size) free(nodes_size);
        if (nodes_sta_idx) free(nodes_sta_idx);
        if (sends_sta_idx) free(sends_sta_idx);
        if (cmp_buf) free(cmp_buf);
        return SplFalse;
      }
      memset(write_buf, 0, sizeof(char)*write_buf_sz);
      memset(rec_buf, 0, sizeof(char)*write_buf_sz);
      memset(send_lens, 0, sizeof(long long)*proc_size);
      memset(nodes_size, 0, sizeof(long long)*proc_size*dims);
      memset(nodes_sta_idx, 0, sizeof(long long)*proc_size*dims);
      memset(sends_sta_idx, 0, sizeof(long long)*proc_size*dims);
      memset(cmp_buf, 0, sizeof(char)*write_buf_sz + cmp_plus);
    }

    // gather nodes size
    if (SPL_Gather(my_size, dims, SPL_LONG_LONG,
                  nodes_size, dims, SPL_LONG_LONG,
                  writeRank, grp) != SPL_SUCCESS ) {
      SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Gather Error.");
      if (write_buf) free(write_buf);
      if (rec_buf) free(rec_buf);
      if (send_buf) free(send_buf);
      if (send_lens) free(send_lens);
      if (nodes_size) free(nodes_size);
      if (nodes_sta_idx) free(nodes_sta_idx);
      if (sends_sta_idx) free(sends_sta_idx);
      if (cmp_buf) free(cmp_buf);
      return SplFalse;
    }

    // gather nodes start index
    if (SPL_Gather(my_sta, dims, SPL_LONG_LONG,
                  nodes_sta_idx, dims, SPL_LONG_LONG,
                  writeRank, grp) != SPL_SUCCESS ) {
      SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Gather Error.");
      if (write_buf) free(write_buf);
      if (rec_buf) free(rec_buf);
      if (send_buf) free(send_buf);
      if (send_lens) free(send_lens);
      if (nodes_size) free(nodes_size);
      if (nodes_sta_idx) free(nodes_sta_idx);
      if (sends_sta_idx) free(sends_sta_idx);
      if (cmp_buf) free(cmp_buf);
      return SplFalse;
    }

    long long remain_len;
    size_t sta_widx[3] = {0, 0, 0};
    size_t node_size[3] = {1, 1, 1};
    size_t node_sta_idx[3] = {0, 0, 0};
    size_t send_sta_idx[3] = {0, 0, 0};
    long long send_len;
    size_t copy_len;
    long long loc_len;
    remain_len = whole_byte_size;
    write_pos_idx = 0;

    // write data size in data recode in file
    if (myrank == writeRank) {
      if (!cmp_flg) {
        // uncompress
        if( writeSvDataBlkSz(mode, writeRank, grp, fp, whole_byte_size) == SplFalse ) {
          if (write_buf) free(write_buf);
          if (rec_buf) free(rec_buf);
          if (send_buf) free(send_buf);
          if (send_lens) free(send_lens);
          if (nodes_size) free(nodes_size);
          if (nodes_sta_idx) free(nodes_sta_idx);
          if (sends_sta_idx) free(sends_sta_idx);
          if (cmp_buf) free(cmp_buf);
          return SplFalse;
        }
      }
    }

    // loop data block
    while( remain_len > 0 ) {

      // clear buffer
      if (write_buf) memset(write_buf, 0, sizeof(char)*write_buf_sz);
      if (rec_buf) memset(rec_buf, 0, sizeof(char)*write_buf_sz);

      my_send_sta[0] = my_send_sta[1] = my_send_sta[2] =0;
      sta_widx[0] = sta_widx[1] = sta_widx[2] =0;

      size_t sidx[3] = {0,0,0};
      size_t len;
      if (my_ov_len > 0) {
        // get write start index
        getIndexFromPos(dims, write_pos_idx, write_size, sta_widx);

        // get overlap start index and lenght in my node
        getOverlapLength(dims, write_size, sta_widx, write_len, my_ov_sidx, my_ov_eidx, sidx, &len);       // modify 2009/10/13 by @hira
      }

      if (len > 0) {
        for (n=0; n<dims; n++) {
          my_send_sta[n] = (long long)sidx[n];
        }

        if (len > write_len) {
          SPL_ERROR_INFO(SPL_ERR_INTERNAL_IMPOSSIBLE, "overlap length overfolow [buf_size=%d,ov_len=%d].", write_len, len);
          len = write_len;
        }

        // get send data array index in my node
        size_t local_sidx[3] = {0,0,0}, local_eidx[3] = {0,0,0}, local_cp_sidx[3] = {0,0,0};
        for (n=0; n<dims; n++) {
          local_sidx[n] = my_ov_sidx[n]-sta_idx[n];
          local_cp_sidx[n] = sidx[n]-sta_idx[n];
          local_eidx[n] = my_ov_eidx[n]-sta_idx[n];
        }

        memset(send_buf, 0, sizeof(char)*write_buf_sz);
        // create send buffer
        if (copyLengthBuffer(dims, data_len * vlen, vsz, local_sidx, local_eidx, local_cp_sidx, len, write_buf_sz , data, send_buf) != SplTrue) {
          if (write_buf) free(write_buf);
          if (rec_buf) free(rec_buf);
          if (send_buf) free(send_buf);
          if (send_lens) free(send_lens);
          if (nodes_size) free(nodes_size);
          if (nodes_sta_idx) free(nodes_sta_idx);
          if (sends_sta_idx) free(sends_sta_idx);
          if (cmp_buf) free(cmp_buf);
          return SplFalse;
        }
      }

      // gather send length from node
      send_len = len;
      if (SPL_Gather(&send_len, 1, SPL_LONG_LONG,
                    send_lens, 1, SPL_LONG_LONG,
                    writeRank, grp) != SPL_SUCCESS ) {
        SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Gather Error.");
        if (write_buf) free(write_buf);
        if (rec_buf) free(rec_buf);
        if (send_buf) free(send_buf);
        if (send_lens) free(send_lens);
        if (nodes_size) free(nodes_size);
        if (nodes_sta_idx) free(nodes_sta_idx);
        if (sends_sta_idx) free(sends_sta_idx);
        if (cmp_buf) free(cmp_buf);
        return SplFalse;
      }

      // gather send start index from node
      if (SPL_Gather(my_send_sta, dims, SPL_LONG_LONG,
                    sends_sta_idx, dims, SPL_LONG_LONG,
                    writeRank, grp) != SPL_SUCCESS ) {
        SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Gather Error.");
        if (write_buf) free(write_buf);
        if (rec_buf) free(rec_buf);
        if (send_buf) free(send_buf);
        if (send_lens) free(send_lens);
        if (nodes_size) free(nodes_size);
        if (nodes_sta_idx) free(nodes_sta_idx);
        if (sends_sta_idx) free(sends_sta_idx);
        if (cmp_buf) free(cmp_buf);
        return SplFalse;
      }

      SPL_Request request = SPL_REQUEST_NULL;
      if (send_len > 0) {
        // send my data
        if ( SPL_Isend((void*)send_buf,
                      send_len*dtype_len,
                      SPL_UNSIGNED_CHAR,
                      writeRank, 1, grp,&request) != SPL_SUCCESS ) {
          SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Isend Error.");
          if (write_buf) free(write_buf);
          if (rec_buf) free(rec_buf);
          if (send_buf) free(send_buf);
          if (send_lens) free(send_lens);
          if (nodes_size) free(nodes_size);
          if (nodes_sta_idx) free(nodes_sta_idx);
          if (sends_sta_idx) free(sends_sta_idx);
          if (cmp_buf) free(cmp_buf);
          return SplFalse;
        }
      }

      // receive data from nodes, and copy write buffer
      if (myrank == writeRank) {
        for (n=0; n<proc_size; n++) {
          if (send_lens[n] == 0) continue;

          if (rec_buf) memset(rec_buf, 0, sizeof(char)*write_buf_sz);
          // receive data from nodes
          SPL_Status status;
          if ( SPL_Recv(rec_buf,
                        send_lens[n]*dtype_len,
                        SPL_UNSIGNED_CHAR,
                        n, 1, grp, &status) != SPL_SUCCESS ) {
            SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Recv Error.");
            if (write_buf) free(write_buf);
            if (rec_buf) free(rec_buf);
            if (send_buf) free(send_buf);
            if (send_lens) free(send_lens);
            if (nodes_size) free(nodes_size);
            if (nodes_sta_idx) free(nodes_sta_idx);
            if (sends_sta_idx) free(sends_sta_idx);
            if (cmp_buf) free(cmp_buf);
            return SplFalse;
          }

          // receive node size, start index in node, start index in receive data
          node_size[0] = node_size[1] = node_size[2] = 1;
          node_sta_idx[0] = node_sta_idx[1] = node_sta_idx[2] = 0;
          for (i=0; i<dims; i++) {
            node_size[i] = nodes_size[n*dims+i];
            node_sta_idx[i] = nodes_sta_idx[n*dims+i];
            send_sta_idx[i] = sends_sta_idx[n*dims+i];
          }

          // get start and end position in receive data
          size_t lidx = 0;
          size_t my_idx[3];
          size_t widx;
          // copy reveive buffer to write buffer
          while (lidx < send_lens[n]) {
            my_idx[0]=my_idx[1]=my_idx[2]=0;
            // get copy start node local index
            getLocalIndexFromPos(dims, node_size, node_sta_idx, send_sta_idx, lidx, my_idx);

            // get global position
            getPosFromIndex(dims, write_size, my_idx, &widx);

            // get write position
            widx = widx - write_pos_idx;

            // get copy length
            if (lidx == 0) {
              loc_len = node_size[0] - (send_sta_idx[0] - node_sta_idx[0]);
            }
            else {
              loc_len = node_size[0];
            }
            if (lidx + loc_len >= send_lens[n]) {
              loc_len = send_lens[n] - lidx;
            }

            // copy buffer
            if (loc_len > 0) {
              // get overlap len
              size_t loc_ov_str;
              size_t idx = (send_sta_idx[0]+lidx)%node_size[0];
              copy_len = 0;
              getOverlapArea(write_sta_idx[0], write_size[0]-1, idx, idx + loc_len -1, &loc_ov_str, &copy_len);
              if (copy_len > 0) {
                if (sizeof(char)*write_buf_sz >= widx*dtype_len + copy_len*dtype_len
                    && sizeof(char)*write_buf_sz >= lidx*dtype_len + copy_len*dtype_len) {
                  memcpy(write_buf+widx*dtype_len,
                        rec_buf+lidx*dtype_len,
                        copy_len*dtype_len);
                }
                else {
                  fprintf(stderr, "invalid over allocate memory\n");
                }
              }
            }
            lidx += loc_len;
          }       // lidx < send_lens[n]
        }        // for (n=0; n<proc_size; n++)

        // write data recode
        if (cmp_flg) {
          // write compress recode
          size_t tmp_cmp_buf_sz;
          tmp_cmp_buf_sz = write_buf_sz + cmp_plus;

          memset(cmp_buf, 0, sizeof(char)*write_buf_sz + cmp_plus);
          // compress data
          if( compress(cmp_buf, &tmp_cmp_buf_sz, write_buf, write_buf_sz) != Z_OK ) {
            // error
            SPL_ERROR_INFO(SPL_ERR_COMPRESS_DATA, "tmp_cmp_buf_sz=%ld, write_buf_sz=%ld", tmp_cmp_buf_sz, write_buf_sz);
            if (write_buf) free(write_buf);
            if (rec_buf) free(rec_buf);
            if (send_buf) free(send_buf);
            if (send_lens) free(send_lens);
            if (nodes_size) free(nodes_size);
            if (nodes_sta_idx) free(nodes_sta_idx);
            if (sends_sta_idx) free(sends_sta_idx);
            if (cmp_buf) free(cmp_buf);
            return SplFalse;
          }

          // write compress data size
          if( writeSvDataBlkSz(mode, writeRank, grp, fp, tmp_cmp_buf_sz) == SplFalse ) {
            if (write_buf) free(write_buf);
            if (rec_buf) free(rec_buf);
            if (send_buf) free(send_buf);
            if (send_lens) free(send_lens);
            if (nodes_size) free(nodes_size);
            if (nodes_sta_idx) free(nodes_sta_idx);
            if (sends_sta_idx) free(sends_sta_idx);
            if (cmp_buf) free(cmp_buf);
            return SplFalse;
          }
          // write compress data
          if( writeSvDataBlock(mode, writeRank, grp, fp,
                               tmp_cmp_buf_sz, cmp_buf) == SplFalse ) {
            if (write_buf) free(write_buf);
            if (rec_buf) free(rec_buf);
            if (send_buf) free(send_buf);
            if (send_lens) free(send_lens);
            if (nodes_size) free(nodes_size);
            if (nodes_sta_idx) free(nodes_sta_idx);
            if (sends_sta_idx) free(sends_sta_idx);
            if (cmp_buf) free(cmp_buf);
            return SplFalse;
          }
        }
        else {
          if( remain_len < write_buf_sz ) {
            write_buf_sz = remain_len;
          }
          // write uncompress data
          if( writeSvDataBlock(mode, writeRank, grp, fp, write_buf_sz, write_buf) == SplFalse ) {
            if (write_buf) free(write_buf);
            if (rec_buf) free(rec_buf);
            if (send_buf) free(send_buf);
            if (send_lens) free(send_lens);
            if (nodes_size) free(nodes_size);
            if (nodes_sta_idx) free(nodes_sta_idx);
            if (sends_sta_idx) free(sends_sta_idx);
            if (cmp_buf) free(cmp_buf);
            return SplFalse;
          }
        }

      } // myrank == writeRank

      // wait for isend
      SPL_Status wait_status;
      if (send_len > 0) {
        if (SPL_Wait(&request, &wait_status) != SPL_SUCCESS ) {
          SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "MPI Wait Error.");
          if (write_buf) free(write_buf);
          if (rec_buf) free(rec_buf);
          if (send_buf) free(send_buf);
          if (send_lens) free(send_lens);
          if (nodes_size) free(nodes_size);
          if (nodes_sta_idx) free(nodes_sta_idx);
          if (sends_sta_idx) free(sends_sta_idx);
          if (cmp_buf) free(cmp_buf);
          return SplFalse;
        }
      }

      // get remain length
      remain_len -= write_buf_sz;
      write_pos_idx += write_len;
      if( remain_len < write_buf_sz ) {
        write_buf_sz = remain_len;
        write_len = remain_len/dtype_len;
      }

    } /* while( remain_len > 0 ) */

    // write eof in data recode
    if (myrank == writeRank) {
      if( writeSvDataBlkSz(mode, writeRank, grp, fp, 0) == SplFalse ) {
        if (write_buf) free(write_buf);
        if (rec_buf) free(rec_buf);
        if (send_buf) free(send_buf);
        if (send_lens) free(send_lens);
        if (nodes_size) free(nodes_size);
        if (nodes_sta_idx) free(nodes_sta_idx);
        if (sends_sta_idx) free(sends_sta_idx);
        if (cmp_buf) free(cmp_buf);
        return SplFalse;
      }
    }

    if (write_buf) free(write_buf);
    if (rec_buf) free(rec_buf);
    if (send_buf) free(send_buf);
    if (send_lens) free(send_lens);
    if (nodes_size) free(nodes_size);
    if (nodes_sta_idx) free(nodes_sta_idx);
    if (sends_sta_idx) free(sends_sta_idx);
    if (cmp_buf) free(cmp_buf);

  }

  return SplTrue;
}


/**
 * \~english
 * get Whole voxel size and start index
 * @param[in]      mode             mode
 * @param[in]      dims             dimension
 * @param[in]      my_size          my node size
 * @param[in]      my_sidx          my node start index
 * @param[out]      wsize           Whole nodes size
 * @param[out]      wsidx           Whole nodes start index
 * @param[out]      write_size      boundry box size
 * @param[out]      write_sidx      boundry box start index
 * @param        grp                Communication group
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 全体のボクセルサイズ、始点インデックスを取得する。
 *
 * 全ノードから最小始点インデックス、最大終点インデックスを取得して算出する。
 * @param[in]      mode             モード
 * @param[in]      dims             次元
 * @param[in]      my_size          自ノードサイズ
 * @param[in]      my_sidx          自ノード始点インデックス
 * @param[out]      wsize            全体ボクセルサイズ
 * @param[out]      wsidx            全体ボクセル始点インデックス
 * @param[out]      write_size       全体境界領域サイズ
 * @param[out]      write_sidx       全体境界領域始点インデックス
 * @param[in]        grp                プロセスグループ
 * @return    success(=true)/fault(=false)
 */
SplBool
getSvWholeSize(int mode,
              unsigned int dims,
              const size_t* my_size,
              const size_t* my_sidx,
              size_t* wsize,
              size_t* wsidx,
              size_t* write_size,
              size_t* write_sidx,
              SPL_Comm grp)
{

  register unsigned int n;
  int myrank = 0;

  if (!my_size || !my_sidx || !wsize || !wsidx || !write_size || !write_sidx) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "size or sidx is null.");
    return SplFalse;
  }

  if (!splIsParallel()) {
    for (n=0;n<dims;n++) {
      wsize[n] = my_size[n];
      wsidx[n] = my_sidx[n];
      write_size[n] = my_size[n];
      write_sidx[n] = my_sidx[n];
    }
    return SplTrue;
  }

  // get my length
  size_t my_len = 1;
  for(n=0; n<dims; n++) my_len *= my_size[n];

  if (mode == 0) {
    for(n=0; n<dims; n++) {
      write_size[n] = my_size[n];
      if (my_len > 0) {
        write_sidx[n] = my_sidx[n];
      }
      else {
        write_sidx[n] = 0;
      }
    }
    return SplTrue;
  }

  SPL_Comm_rank(grp, &myrank);

  // get start index
  unsigned long long tmp_min_idx[3] = {0, 0, 0};
  unsigned long long min_idx[3] = {0, 0, 0};
  for(n=0; n<dims; n++) {
    if (my_len > 0) {
      tmp_min_idx[n] = (unsigned long long)(my_sidx[n]);
    }
    else {
      tmp_min_idx[n] = __LONG_LONG_MAX__;
    }
  }
  SPL_Allreduce(tmp_min_idx, min_idx, dims, SPL_LONG_LONG, SPL_MIN, grp);

  // get max index
  unsigned long long tmp_max_idx[3] = {0, 0, 0};
  unsigned long long max_idx[3] = {0, 0, 0};
  for(n=0; n<dims; n++) {
    if (my_size[n] > 0) {
      tmp_max_idx[n] = (unsigned long long)(my_sidx[n] + my_size[n] - 1);
    }
    else {
      tmp_max_idx[n] = 0;
    }
  }
  SPL_Allreduce(tmp_max_idx, max_idx, dims, SPL_LONG_LONG, SPL_MAX, grp);

  // get whole size and start index
  for (n=0; n<dims; n++) {
    wsidx[n] = min_idx[n];
    wsize[n] = max_idx[n] + 1;
  }

  for(n=0; n<dims; n++) {
    if (max_idx[n] < min_idx[n]) {
      SPL_ERROR_INFO(SPL_ERR_INTERNAL_IMPOSSIBLE, "not consistency [max_idx=%ld<min_idx=%ld]", max_idx[n], min_idx[n]);
      return SplFalse;
    }
    write_size[n] = max_idx[n] - min_idx[n] + 1;
    if (my_sidx[n] < wsidx[n]) {
      SPL_ERROR_INFO(SPL_ERR_INTERNAL_IMPOSSIBLE, "not consistency [start_idx=%ld<wsidx=%ld]", my_sidx[n], wsidx[n]);
      return SplFalse;
    }
    if (my_len > 0) {
      write_sidx[n] = my_sidx[n] - wsidx[n];
    }
    else {
      write_sidx[n] = 0;
    }
  }

  return SplTrue;
}

/**
 * \~english
 * get Whole voxel size and start index
 * @param[in]      dims             dimension
 * @param[in]      my_size          my node size
 * @param[in]      my_sidx          my node start index
 * @param[out]      wsize           Whole nodes size
 * @param[out]      wsidx           Whole nodes start index
 * @param        grp                Communication group
 * @return    success(=true)/fault(=false)
 *
 * \~japanese
 * 全体のボクセルサイズ、始点インデックスを取得する。
 *
 * 全ノードから最小始点インデックス、最大終点インデックスを取得して算出する。
 * @param[in]      dims             次元
 * @param[in]      my_size          自ノードサイズ
 * @param[in]      my_sidx          自ノード始点インデックス
 * @param[out]      wsize            全体ボクセルサイズ
 * @param[out]      wsidx            全体ボクセル始点インデックス
 * @param[in]        grp                プロセスグループ
 * @return    success(=true)/fault(=false)
 */
SplBool
getVoxelCoords(
              unsigned int dims,
              const size_t* my_size,
              const size_t* my_sidx,
              const size_t* wsidx,
              double* worg,
              double* pit,
              SPL_Comm grp)
{

  register unsigned int n;
  int myrank = 0;

  if (!wsidx || !worg || !pit) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "sidx, org or pit is null.");
    return SplFalse;
  }

  if (!splIsParallel()) {
    return SplTrue;
  }

  SPL_Comm_rank(grp, &myrank);

  size_t my_len = 1;
  for(n=0; n<dims; n++) my_len *= my_size[n];

  // get min start index
  unsigned long long tmp_min_idx[3] = {0, 0, 0};
  unsigned long long min_idx[3] = {0, 0, 0};
  for(n=0; n<dims; n++) {
    if (my_len > 0) {
      tmp_min_idx[n] = (unsigned long long)(my_sidx[n]);
    }
    else {
      tmp_min_idx[n] = __LONG_LONG_MAX__;
    }
  }
  SPL_Allreduce(tmp_min_idx, min_idx, dims, SPL_LONG_LONG, SPL_MIN, grp);

  // get min origin coords
  double tmp_min_org[3];
  double min_org[3] = {DBL_MAX, DBL_MAX, DBL_MAX};
  memset(tmp_min_org, 0, sizeof(double)*3);
  for (n=0;n<dims;n++) {
    tmp_min_org[n] = (my_len > 0) ? worg[n] : DBL_MAX;
  }
  SPL_Allreduce((void*)tmp_min_org, min_org, dims, SPL_DOUBLE, SPL_MIN, grp);

  // get pitch
  double pit_len = 0.0;
  double wpit[3];
  for (n=0;n<dims;n++) {
    if (pit[n] == 0.0)  pit_len = 0.0;
    wpit[n] = pit[n];
  }
  if (pit_len == 0.0) {
    memset(wpit, 0, sizeof(double)*3);
    double mypit[3] = {DBL_MAX, DBL_MAX, DBL_MAX};
    for (n=0;n<dims;n++) {
      mypit[n] = (my_len > 0) ? pit[n] : DBL_MAX;
    }
    SPL_Allreduce((void*)mypit, wpit, 3, SPL_DOUBLE, SPL_MIN, grp);
    for (n=0;n<dims;n++) pit[n] = wpit[n];
  }
  for (n=0; n<dims; n++) {
    worg[n] = min_org[n] - (min_idx[n] - wsidx[n]) * pit[n];
  }

  return SplTrue;
}

#if defined(__cplusplus)
}
#endif /* extern "C" */

