/*************************************************************
   SPHERE - Skeleton for PHysical and Engineering REsearch
   Copyright (c) RIKEN, Japan. All right reserved. 2004-2010
 *************************************************************/
/**
 * \~english
 * @file spl_mpi.c
 * MPI Support Functions File
 * @author     hira
 * @date       2009/02/20
 *
 * \~japanese
 * @file spl_mpi.c
 * MPIサポート関数群ファイル
 * @author     hira
 * @date       2009/02/20
 */

#include <stdio.h>
#include <stdarg.h>
#include "spl/splmpi.h"

#if defined(__cplusplus)
extern "C" {
#endif

//extern int spl_err_code;

/**
 * \~english
 * Print Message with my rank number
 * @param        grp          MPI communicator
 * @param        format      message format
 * @param        ...          message value
 *
 * \~japanese
 * 自ノード番号とメッセージを表示する。
 * @param        grp           MPIコミュニケータ
 * @param        format      出力メッセージフォーマット
 * @param        ...          出力メッセージ値
 */
void
spl_pmsg(SPL_Comm grp, const char* format, ...) {
  if( !format ) return;
  char buf[1024];
  va_list vl;
  va_start(vl, format);
  vsprintf(buf, format, vl);
  va_end(vl);
  int rank;
  SPL_Comm_rank(grp, &rank);
  fprintf(stdout, " SPHERE MSG[%d] : %s\n", rank, buf);
  return;
}

/**
 * \~english
 * Print Error Message with my rank number
 * @param        grp          MPI communicator
 * @param        format      message format
 * @param        ...          message value
 *
 * \~japanese
 * 自ノード番号とエラーメッセージを表示する。
 * @param        grp           MPIコミュニケータ
 * @param        format      出力メッセージフォーマット
 * @param        ...          出力メッセージ値
 */
void
spl_perrmsg(SPL_Comm grp, const char* format, ...) {
  if( !format ) return;
  char buf[1024];
  va_list vl;
  va_start(vl, format);
  vsprintf(buf, format, vl);
  va_end(vl);
  int rank;
  SPL_Comm_rank(grp, &rank);
  fprintf(stderr, " SPHERE ERROR[%d] : %s\n", rank, buf);
  return;
}

/**
 * \~english
 * Print Warning Message with my rank number
 * @param        grp          MPI communicator
 * @param        format      message format
 * @param        ...          message value
 *
 * \~japanese
 * 自ノード番号と警告メッセージを表示する。
 * @param        grp           MPIコミュニケータ
 * @param        format      出力メッセージフォーマット
 * @param        ...          出力メッセージ値
 */
void
spl_pwrnmsg(SPL_Comm grp, const char* format, ...) {
  if( !format ) return;
  char buf[1024];
  va_list vl;
  va_start(vl, format);
  vsprintf(buf, format, vl);
  va_end(vl);
  int rank;
  SPL_Comm_rank(grp, &rank);
  fprintf(stderr, " SPHERE WARNNING[%d] : %s\n", rank, buf);
  return;
}


/**
 * \~english
 * create datatype of array type to send/receive for 1D
 * @param[in]      ilen          I size
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * 通信データタイプを作成する。(1D-Arrayタイプ)
 *
 * 通信データサイズを指定して作成する派生型データタイプ。
 * @param[in]      ilen          I方向通信サイズ
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
inline SplBool
splCreateCommType1DA(
                     size_t ilen,
                     SPL_Datatype oldType, SPL_Datatype* newType)
{
  if( !newType ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "newType is null.");
    return SplFalse;
  }
  if( SPL_Type_contiguous(ilen, oldType, newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_contiguous error.");
    return SplFalse;
  }
  if( SPL_Type_commit(newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_commit error.");
    return SplFalse;
  }
  return SplTrue;

}

/**
 * \~english
 * create datatype of index type to send/receive for 1D
 * @param[in]      imin          I min index
 * @param[in]      imax          I max index
 * @param[in]      ista          I start index
 * @param[in]      iend          I end index
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * 通信データタイプを作成する。(1D-Indexタイプ)
 *
 * 通信データの始点、終点インデックスを指定して作成する派生型データタイプ。
 * @param[in]      imin          I方向領域始点インデックス
 * @param[in]      imax          I方向領域終点インデックス
 * @param[in]      ista          I方向通信始点インデックス
 * @param[in]      iend          I方向通信終点インデックス
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
inline SplBool
splCreateCommType1D(size_t imin, size_t imax,
                    size_t ista, size_t iend,
                    SPL_Datatype oldType, SPL_Datatype* newType)
{
  if( !newType ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "newType is null.");
    return SplFalse;
  }

  SPL_Datatype tmpType;
  int ext;
  int block[2];
  SPL_Aint disp[2];
  SPL_Datatype type[2];

  if ( SPL_Type_size(oldType, &ext)!= SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_size error.");
    return SplFalse;
  }

  size_t ilen = iend - ista + 1;

  if( SPL_Type_vector(1, ilen, imax-imin+1,
                      oldType, &tmpType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_vector error.");
    return SplFalse;
  }

  block[0] = 1;
  block[1] = 1;
  disp[0] = (SPL_Aint)0;
  disp[1] = (SPL_Aint)((ista -imin) * ext);
  type[0] = SPL_LB;
  type[1] = tmpType;

  if( SPL_Type_struct(2, block, disp, type, newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_struct error.");
    return SplFalse;
  }

  if( SPL_Type_commit(newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_commit error.");
    return SplFalse;
  }
  return SplTrue;
}

/**
 * \~english
 * create datatype of array type to send/receive for 2D
 * @param[in]      imin          I min index
 * @param[in]      imax          I max index
 * @param[in]      ilen          I size
 * @param[in]      jlen          J size
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * 通信データタイプを作成する。(2D-Arrayタイプ)
 *
 * 通信データサイズを指定して作成する派生型データタイプ。<br>
 * SPH_DC_ARRAY_1DN, SPH_DC_ARRAY_1DNV, SPH_DC_ARRAY_2Dのデータクラスタイプにて使用可能である。<br>
 * パラメータの設定順番(i,j)は、データの配列順に従って入れ替えること。<br>
 * SPH_DC_ARRAY_1DNVの場合は(l,i)である。
 * @param[in]      imin          I方向領域始点インデックス
 * @param[in]      imax          I方向領域終点インデックス
 * @param[in]      ilen          I方向サイズ
 * @param[in]      jlen          J方向サイズ
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splCreateCommType2DA(size_t imin, size_t imax,
                     size_t ilen, size_t jlen,
                     SPL_Datatype oldType, SPL_Datatype* newType) {
  if( !newType ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "newType is null.");
    return SplFalse;
  }
  if( SPL_Type_vector(jlen, ilen, imax-imin+1,
                      oldType, newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_vector error.");
    return SplFalse;
  }
  if( SPL_Type_commit(newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_commit error.");
    return SplFalse;
  }
  return SplTrue;
}

/**
 * \~english
 * create datatype of index type to send/receive for 2D
 * @param[in]      imin          I min index
 * @param[in]      imax          I max index
 * @param[in]      jmin          J min index
 * @param[in]      ista          I start index
 * @param[in]      iend          I end index
 * @param[in]      jsta          J start index
 * @param[in]      jend          J end index
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * 通信データタイプを作成する。(2D-Indexタイプ)
 *
 * 通信データの始点、終点インデックスを指定して作成する派生型データタイプ。<br>
 * SPH_DC_ARRAY_1DN, SPH_DC_ARRAY_1DNV, SPH_DC_ARRAY_2Dのデータクラスタイプにて使用可能である。<br>
 * パラメータの設定順番(i,j)は、データの配列順に従って入れ替えること。<br>
 * SPH_DC_ARRAY_1DNVの場合は(l,i)である。
 * @param[in]      imin          I方向領域始点インデックス
 * @param[in]      imax          I方向領域終点インデックス
 * @param[in]      jmin          J方向領域始点インデックス
 * @param[in]      ista          I方向通信始点インデックス
 * @param[in]      iend          I方向通信終点インデックス
 * @param[in]      jsta          J方向通信始点インデックス
 * @param[in]      jend          J方向通信終点インデックス
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splCreateCommType2D(size_t imin, size_t imax, size_t jmin,
                    size_t ista, size_t iend, size_t jsta, size_t jend,
                    SPL_Datatype oldType, SPL_Datatype* newType) {
  int size;
  size_t ilen, jlen;
  SPL_Datatype tmpType, type[2];
  int block[2];
  SPL_Aint disp[2];
  if( !newType ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "newType is null.");
    return SplFalse;
  }
  if( SPL_Type_size(oldType, &size) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_size error.");
    return SplFalse;
  }
  ilen = iend - ista + 1;
  jlen = jend - jsta + 1;
  if( SPL_Type_vector(jlen, ilen, imax-imin+1,
                      oldType, &tmpType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_vector error.");
    return SplFalse;
  }
  block[0] = 1;
  block[1] = 1;
  disp[0] = 0;
  disp[1] = ((imax-imin+1)*(jsta-jmin)+(ista-imin))*size;
  type[0] = SPL_LB;
  type[1] = tmpType;
  if( SPL_Type_struct(2, block, disp, type, newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_struct error.");
    return SplFalse;
  }
  if( SPL_Type_commit(newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_commit error.");
    return SplFalse;
  }
  return SplTrue;
}

/**
 * \~english
 * create datatype of array type to send/receive for 3D
 * @param[in]      imin          I min index
 * @param[in]      imax          I max index
 * @param[in]      jmin          J min index
 * @param[in]      jmax          J max index
 * @param[in]      ilen          I size
 * @param[in]      jlen          J size
 * @param[in]      klen          K size
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * 通信データタイプを作成する。(3D-Arrayタイプ)
 *
 * 通信データサイズを指定して作成する派生型データタイプ。<br>
 * SPH_DC_ARRAY_2DN, SPH_DC_ARRAY_2DNV, SPH_DC_ARRAY_3Dのデータクラスタイプにて使用可能である。<br>
 * パラメータの設定順番(i,j,k)は、データの配列順に従って入れ替えること。<br>
 * SPH_DC_ARRAY_2DNVの場合は(l,i,j)である。
 * @param[in]      imin          I方向領域始点インデックス
 * @param[in]      imax          I方向領域終点インデックス
 * @param[in]      jmin          J方向領域始点インデックス
 * @param[in]      jmax          J方向領域終点インデックス
 * @param[in]      ilen          I方向サイズ
 * @param[in]      jlen          J方向サイズ
 * @param[in]      klen          K方向サイズ
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splCreateCommType3DA(size_t imin, size_t imax,
                     size_t jmin, size_t jmax,
                     size_t ilen, size_t jlen, size_t klen,
                     SPL_Datatype oldType, SPL_Datatype* newType) {
  int size;
  SPL_Datatype tmpType;
  SPL_Aint dist;
  if( !newType ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "newType is null.");
    return SplFalse;
  }
  if( SPL_Type_size(oldType, &size) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_size error.");
    return SplFalse;
  }
  if( SPL_Type_vector(jlen, ilen, imax-imin+1,
                      oldType, &tmpType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_vector error.");
    return SplFalse;
  }
  dist = (imax-imin+1)*(jmax-jmin+1)*size;
  if( SPL_Type_hvector(klen, 1, dist, tmpType, newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_hvector error.");
    return SplFalse;
  }
  if( SPL_Type_commit(newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_commit error.");
    return SplFalse;
  }
  return SplTrue;
}

/**
 * \~english
 * create datatype of index type to send/receive for 2D
 * @param[in]      imin          I min index
 * @param[in]      imax          I max index
 * @param[in]      jmin          J min index
 * @param[in]      jmax          J max index
 * @param[in]      kmin          K min index
 * @param[in]      ista          I start index
 * @param[in]      iend          I end index
 * @param[in]      jsta          J start index
 * @param[in]      jend          J end index
 * @param[in]      ksta          K start index
 * @param[in]      kend          K end index
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * 通信データタイプを作成する。(2D-Indexタイプ)
 *
 * 通信データの始点、終点インデックスを指定して作成する派生型データタイプ。<br>
 * SPH_DC_ARRAY_2DN, SPH_DC_ARRAY_2DNV, SPH_DC_ARRAY_3Dのデータクラスタイプにて使用可能である。<br>
 * パラメータの設定順番(i,j,k)は、データの配列順に従って入れ替えること。<br>
 * SPH_DC_ARRAY_2DNVの場合は(l,i,j)である。
 * @param[in]      imin          I方向領域始点インデックス
 * @param[in]      imax          I方向領域終点インデックス
 * @param[in]      jmin          J方向領域始点インデックス
 * @param[in]      jmax          J方向領域終点インデックス
 * @param[in]      kmin          K方向領域始点インデックス
 * @param[in]      ista          I方向通信始点インデックス
 * @param[in]      iend          I方向通信終点インデックス
 * @param[in]      jsta          J方向通信始点インデックス
 * @param[in]      jend          J方向通信終点インデックス
 * @param[in]      ksta          K方向通信始点インデックス
 * @param[in]      kend          K方向通信終点インデックス
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splCreateCommType3D(size_t imin, size_t imax,
                    size_t jmin, size_t jmax,
                    size_t kmin,
                    size_t ista, size_t iend,
                    size_t jsta, size_t jend,
                    size_t ksta, size_t kend,
                    SPL_Datatype oldType, SPL_Datatype* newType) {
  int size;
  size_t ilen, jlen, klen;
  SPL_Datatype tmpType, tmpType2, type[2];
  int block[2];
  SPL_Aint disp[2];
  SPL_Aint dist;
  if( !newType ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "newType is null.");
    return SplFalse;
  }
  if( SPL_Type_size(oldType, &size) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_size error.");
    return SplFalse;
  }
  ilen = iend - ista + 1;
  jlen = jend - jsta + 1;
  klen = kend - ksta + 1;
  if( SPL_Type_vector(jlen, ilen, imax-imin+1,
                      oldType, &tmpType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_vector error.");
    return SplFalse;
  }
  dist = (imax-imin+1)*(jmax-jmin+1)*size;
  if( SPL_Type_hvector(klen, 1, dist, tmpType, &tmpType2) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_hvector error.");
    return SplFalse;
  }
  block[0] = 1;
  block[1] = 1;
  disp[0] = 0;
  disp[1] = (  (imax-imin+1)*(jmax-jmin+1)*(ksta-kmin)
             + (imax-imin+1)*(jsta-jmin) + (ista-imin) ) * size;
  type[0] = SPL_LB;
  type[1] = tmpType2;
  if( SPL_Type_struct(2, block, disp, type, newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_struct error.");
    return SplFalse;
  }
  if( SPL_Type_commit(newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_commit error.");
    return SplFalse;
  }
  return SplTrue;
}

/**
 * \~english
 * create datatype of array type to send/receive for 4D
 * @param[in]      imin          I min index
 * @param[in]      imax          I max index
 * @param[in]      jmin          J min index
 * @param[in]      jmax          J max index
 * @param[in]      kmin          K min index
 * @param[in]      kmax          K max index
 * @param[in]      ilen          I size
 * @param[in]      jlen          J size
 * @param[in]      klen          K size
 * @param[in]      llen          l size
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * 通信データタイプを作成する。(4D-Arrayタイプ)
 *
 * 通信データサイズを指定して作成する派生型データタイプ。<br>
 * SPH_DC_ARRAY_3DN, SPH_DC_ARRAY_3DNVのデータクラスタイプにて使用可能である。<br>
 * パラメータの設定順番(i,j,k,l)は、データの配列順に従って入れ替えること。<br>
 * SPH_DC_ARRAY_3DNVの場合は(l,i,j,k)である。
 * @param[in]      imin          I方向領域始点インデックス
 * @param[in]      imax          I方向領域終点インデックス
 * @param[in]      jmin          J方向領域始点インデックス
 * @param[in]      jmax          J方向領域終点インデックス
 * @param[in]      kmin          K方向領域始点インデックス
 * @param[in]      kmax          K方向領域終点インデックス
 * @param[in]      ilen          I方向サイズ
 * @param[in]      jlen          J方向サイズ
 * @param[in]      klen          K方向サイズ
 * @param[in]      llen          セルのデータ数
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
inline SplBool
splCreateCommType4DA(size_t imin, size_t imax,
                     size_t jmin, size_t jmax,
                     size_t kmin, size_t kmax,
                     size_t ilen, size_t jlen, size_t klen,
                     size_t llen,
                     SPL_Datatype oldType, SPL_Datatype* newType)
{
  int ext;
  SPL_Datatype tempType1, tempType2;
  SPL_Aint dist;

  if( !newType ) {
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "newType is null.");
    return SplFalse;
  }

  size_t isize = imax-imin+1;
  size_t jsize = jmax-jmin+1;
  size_t ksize = kmax-kmin+1;

  if( SPL_Type_size(oldType, &ext) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_size error.");
    return SplFalse;
  }
  if ( SPL_Type_vector(jlen, ilen, isize, oldType, &tempType1) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_vector error.");
    return SplFalse;
  }

  dist = (SPL_Aint)(isize*jsize*ext);

  if ( SPL_Type_hvector(klen, 1, dist, tempType1, &tempType2) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_hvector error.");
    return SplFalse;
  }
  dist = (SPL_Aint)(isize*jsize*ksize*ext);
  if ( SPL_Type_hvector(llen, 1, dist, tempType2, newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_hvector error.");
    return SplFalse;
  }

  if ( SPL_Type_commit(newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_commit error.");
    return SplFalse;
  }

  return SplTrue;
}

/**
 * \~english
 * create datatype of index type to send/receive for 4D
 * @param[in]      imin          I min index
 * @param[in]      imax          I max index
 * @param[in]      jmin          J min index
 * @param[in]      jmax          J max index
 * @param[in]      kmin          K min index
 * @param[in]      kmax          K max index
 * @param[in]      lmax          l min index
 * @param[in]      ista          I start index
 * @param[in]      iend          I end index
 * @param[in]      jsta          J start index
 * @param[in]      jend          J end index
 * @param[in]      ksta          K start index
 * @param[in]      kend          K end index
 * @param[in]      lsta          l start index
 * @param[in]      lend          l end index
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * 通信データタイプを作成する。(4D-Indexタイプ)
 *
 * 通信データの始点、終点インデックスを指定して作成する派生型データタイプ。<br>
 * SPH_DC_ARRAY_3DN, SPH_DC_ARRAY_3DNVのデータクラスタイプにて使用可能である。<br>
 * パラメータの設定順番(i,j,k,l)は、データの配列順に従って入れ替えること。<br>
 * SPH_DC_ARRAY_3DNVの場合は(l,i,j,k)である。
 * @param[in]      imin          I方向領域始点インデックス
 * @param[in]      imax          I方向領域終点インデックス
 * @param[in]      jmin          J方向領域始点インデックス
 * @param[in]      jmax          J方向領域終点インデックス
 * @param[in]      kmin          K方向領域始点インデックス
 * @param[in]      kmax          K方向領域終点インデックス
 * @param[in]      lmin          セルのデータ数始点インデックス
 * @param[in]      ista          I方向通信始点インデックス
 * @param[in]      iend          I方向通信終点インデックス
 * @param[in]      jsta          J方向通信始点インデックス
 * @param[in]      jend          J方向通信終点インデックス
 * @param[in]      ksta          K方向通信始点インデックス
 * @param[in]      kend          K方向通信終点インデックス
 * @param[in]      lsta          セルのデータ数始点インデックス
 * @param[in]      lend          セルのデータ数終点インデックス
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
inline SplBool
splCreateCommType4D(size_t imin, size_t imax,
                    size_t jmin, size_t jmax,
                    size_t kmin, size_t kmax,
                    size_t lmin,
                    size_t ista, size_t iend,
                    size_t jsta, size_t jend,
                    size_t ksta, size_t kend,
                    size_t lsta, size_t lend,
                    SPL_Datatype oldType, SPL_Datatype* newType)
{
  SPL_Datatype tempType, tempType2, tempType3;
  int ext;
  int block[2];
  SPL_Aint disp[2];
  SPL_Datatype type[2];

  if ( SPL_Type_size(oldType, &ext) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_size error.");
    return SplFalse;
  }

  size_t ilen = iend - ista + 1;
  size_t jlen = jend - jsta + 1;
  size_t klen = kend - ksta + 1;
  size_t llen = lend - lsta + 1;
  size_t isize = imax - imin +1;
  size_t jsize = jmax - jmin +1;
  size_t ksize = kmax - kmin +1;

  if ( SPL_Type_vector(jlen, ilen, isize, oldType, &tempType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_vector error.");
    return SplFalse;
  }

  SPL_Aint dist = (SPL_Aint)(isize*jsize*ext);
  if ( SPL_Type_hvector(klen, 1, dist, tempType, &tempType2) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_hvector error.");
    return SplFalse;
  }

  dist = (SPL_Aint)(isize*jsize*ksize*ext);
  if ( SPL_Type_hvector(llen, 1, dist, tempType2, &tempType3) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_hvector error.");
    return SplFalse;
  }

  block[0] = 1;
  block[1] = 1;
  disp[0] = (SPL_Aint)0;
  disp[1] = (SPL_Aint)(
              ( isize * jsize * ksize * (lsta-lmin)
              + isize * jsize * (ksta-kmin)
               + isize * (jsta-jmin)
               + (ista-imin) ) * ext);
  type[0] = SPL_LB;
  type[1] = tempType3;

  if ( SPL_Type_struct(2, block, disp, type, newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_struct error.");
    return SplFalse;
  }

  if (SPL_Type_commit(newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_commit error.");
    return SplFalse;
  }

  return SplTrue;
}


/**
 * 通信データタイプを作成する。(Vectorタイプ)
 *
 * 通信データの始点、終点インデックスを指定して作成する派生型データタイプ。<br>
 * ベクトルデータの個々の要素に対して領域、始点、終点インデックスを指定を行う。<br>
 * 2Dベクトル（データ数＝２）、3Dベクトル（データ数＝３）のみ使用できる。
 * @param[in]      dims          次元数
 * @param[in]      min           領域終点インデックス
 * @param[in]      max           領域終点インデックス
 * @param[in]      sta           通信始点インデックス
 * @param[in]      end           通信終点インデックス
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
inline SplBool
splCreateCommTypeVector(
                    int dims,
                    size_t min[3], size_t max[3],
                    size_t sta[3][3], size_t end[3][3],
                    SPL_Datatype oldType, SPL_Datatype* newType)
{
  SPL_Datatype tempType[3], tempType2, tempType3;
  int ext;
  int block[4];
  SPL_Aint disp[4];
  SPL_Datatype type[4];
  SPL_Datatype vectType[3];
  int data_num = dims;

  if ( SPL_Type_size(oldType, &ext) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_size error.");
    return SplFalse;
  }

  register size_t i, j;
  size_t len[3][3];
  size_t size[3][3];
  for (j=0; j<3; j++) {
    for (i=0; i<3; i++) {
      len[j][i] = 1;
      size[j][i] = 1;
    }
  }
  for (j=0; j<data_num; j++) {
    for (i=0; i<dims; i++) {
      len[j][i] = end[j][i] - sta[j][i] + 1;
      size[j][i] = max[i] - min[i] + 1;
    }
  }

  for (j=0; j<data_num; j++) {
    if ( SPL_Type_vector(len[j][1], len[j][0], size[j][0], oldType, &tempType[j]) != SPL_SUCCESS ) {
      SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_vector error.");
      return SplFalse;
    }
    if (data_num > 2) {
      SPL_Aint dist = (SPL_Aint)(size[j][0]*size[j][1]*ext);
      if ( SPL_Type_hvector(len[j][2], 1, dist, tempType[j], &vectType[j]) != SPL_SUCCESS ) {
        SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_hvector error.");
        return SplFalse;
      }
    }
    else {
      vectType[j] = tempType[j];
    }
  }

  block[0] = 1;
  block[1] = 1;
  block[2] = 1;
  block[3] = 1;
  disp[0] = (SPL_Aint)0;
  disp[1] = (SPL_Aint)(
              ( size[0][0] * size[0][1] * (sta[0][2]-min[2])
               + size[0][0] * (sta[0][1]-min[1])
               + (sta[0][0]-min[0]) ) * ext);

  disp[2] = (SPL_Aint)(
              ( size[0][0] * size[0][1] * size[0][2]
               + size[1][0] * size[1][1] * (sta[1][2]-min[2])
               + size[1][0] * (sta[1][1]-min[1])
               + (sta[1][0]-min[0]) ) * ext);

  disp[3] = (SPL_Aint)(
              ( size[0][0] * size[0][1] * size[0][2]
               + size[1][0] * size[1][1] * size[1][2]
               + size[2][0] * size[2][1] * (sta[2][2]-min[2])
               + size[2][0] * (sta[2][1]-min[1])
               + (sta[2][0]-min[0]) ) * ext);

  type[0] = SPL_LB;
  type[1] = vectType[0];
  type[2] = vectType[1];
  type[3] = vectType[2];


  if ( SPL_Type_struct(data_num+1, block, disp, type, newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_struct error.");
    return SplFalse;
  }

  if (SPL_Type_commit(newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_commit error.");
    return SplFalse;
  }

  return SplTrue;
}



/**
 * 通信データタイプを作成する。(VectorRVタイプ)
 *
 * 通信データの始点、終点インデックスを指定して作成する派生型データタイプ。<br>
 * ベクトルデータの個々の要素に対して領域、始点、終点インデックスを指定を行う。<br>
 * 2DベクトルRV（データ数＝２）、3DベクトルRV（データ数＝３）のみ使用できる。
 * @param[in]      dims          次元数
 * @param[in]      min           領域終点インデックス
 * @param[in]      max           領域終点インデックス
 * @param[in]      sta           通信始点インデックス
 * @param[in]      end           通信終点インデックス
 * @param[in]      oldType      old datatype
 * @param[out]        newType      new datatype
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
inline SplBool
splCreateCommTypeVectorRv(
                    int dims,
                    size_t min[3], size_t max[3],
                    size_t sta[3][3], size_t end[3][3],
                    SPL_Datatype oldType, SPL_Datatype* newType)
{
  SPL_Datatype tempType[3], tempType2[3], tempType3;
  int ext;
  int block[4];
  SPL_Aint disp[4];
  SPL_Datatype type[4];
  SPL_Datatype vectType[3];
  int data_num = dims;
  SPL_Aint dist;

  if ( SPL_Type_size(oldType, &ext) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_size error.");
    return SplFalse;
  }

  register size_t i, j;
  size_t len[3][3];
  size_t size[3][3];
  for (j=0; j<3; j++) {
    for (i=0; i<3; i++) {
      len[j][i] = 1;
      size[j][i] = 1;
    }
  }
  for (j=0; j<data_num; j++) {
    for (i=0; i<dims; i++) {
      len[j][i] = end[j][i] - sta[j][i] + 1;
      size[j][i] = max[i] - min[i] + 1;
    }
  }

  for (j=0; j<data_num; j++) {
    if ( SPL_Type_vector(len[j][0], 1, data_num, oldType, &tempType[j]) != SPL_SUCCESS ) {
      SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_vector error.");
      return SplFalse;
    }
    dist = (SPL_Aint)(size[j][0]*ext*data_num);
    if ( SPL_Type_hvector(len[j][1], 1, dist, tempType[j], &tempType2[j]) != SPL_SUCCESS ) {
      SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_vector error.");
      return SplFalse;
    }
    if (data_num > 2) {
      dist = (SPL_Aint)(size[j][0]*size[j][1]*ext*data_num);
      if ( SPL_Type_hvector(len[j][2], 1, dist, tempType2[j], &vectType[j]) != SPL_SUCCESS ) {
        SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_hvector error.");
        return SplFalse;
      }
    }
    else {
      vectType[j] = tempType2[j];
    }
  }

  block[0] = 1;
  block[1] = 1;
  block[2] = 1;
  block[3] = 1;
  disp[0] = (SPL_Aint)0;
  disp[1] = (SPL_Aint)( (
              (   size[0][0] * size[0][1] * (sta[1][2]-min[2])
                + size[0][0] * (sta[0][1]-min[1])
                + (sta[0][0]-min[0]) ) * data_num) * ext );

  disp[2] = (SPL_Aint)( (
              (  size[1][0] * size[1][1] * (sta[1][2]-min[2])
               + size[1][0] * (sta[1][1]-min[1])
               + (sta[1][0]-min[0]) ) * data_num + 1 ) * ext );

  disp[3] = (SPL_Aint)( (
              (  size[2][0] * size[2][1] * (sta[2][2]-min[2])
               + size[2][0] * (sta[2][1]-min[1])
               + (sta[2][0]-min[0]) ) * data_num + 2 ) * ext );

  type[0] = SPL_LB;
  type[1] = vectType[0];
  type[2] = vectType[1];
  type[3] = vectType[2];


  if ( SPL_Type_struct(data_num+1, block, disp, type, newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_struct error.");
    return SplFalse;
  }

  if (SPL_Type_commit(newType) != SPL_SUCCESS ) {
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Type_commit error.");
    return SplFalse;
  }

  return SplTrue;
}


/**
 * \~english
 * get data size
 * @param        datatype          datatype
 * @return    data size
 *
 * \~japanese
 * データタイプのサイズを取得する。.
 * 基本データ型のみサイズを取得する。<br/>
 * 派生データ型の場合は、取得しない。（SPL_Type_sizeを使用すること。)
 * @param        datatype          データタイプ
 * @return    データサイズ
 */
int spl_get_datasize(SPL_Datatype datatype)
{

  int size = 0;
       if( datatype == SPL_CHAR )           { size = sizeof(char);           }
  else if( datatype == SPL_UNSIGNED_CHAR )  { size = sizeof(unsigned char);  }
  else if( datatype == SPL_SHORT )          { size = sizeof(short);          }
  else if( datatype == SPL_UNSIGNED_SHORT ) { size = sizeof(unsigned short); }
  else if( datatype == SPL_INT )            { size = sizeof(int);            }
  else if( datatype == SPL_UNSIGNED )       { size = sizeof(unsigned);       }
  else if( datatype == SPL_LONG )           { size = sizeof(long);           }
  else if( datatype == SPL_UNSIGNED_LONG )  { size = sizeof(unsigned long);  }
  else if( datatype == SPL_LONG_LONG )      { size = sizeof(long long);      }
  else if( datatype == SPL_LONG_LONG_INT )  { size = sizeof(long long);      }
  else if( datatype == SPL_FLOAT )          { size = sizeof(float);          }
  else if( datatype == SPL_DOUBLE )         { size = sizeof(double);         }
  else { size = 0; }
  return size;
}


/**
 * \~english
 * get datatype from data size
 * @param        dtsize          data size
 * @return    datatype
 *
 * \~japanese
 * データサイズからMPIデータタイプを取得する。
 * @param        dtsize          データサイズ
 * @return    MPIデータタイプ
 */
SPL_Datatype spl_get_datatype(int dtsize)
{

  int size = 0;
  if (dtsize == sizeof(unsigned char)) {
    return SPL_UNSIGNED_CHAR;
  }
  else if (dtsize == sizeof(unsigned short)) {
    return SPL_UNSIGNED_SHORT;
  }
  else if (dtsize == sizeof(unsigned)) {
    return SPL_UNSIGNED;
  }
  else if (dtsize == sizeof(unsigned long)) {
    return SPL_UNSIGNED_LONG;
  }
  else if (dtsize == sizeof(long long)) {
    return SPL_LONG_LONG_INT;
  }
  return SPL_DATATYPE_NULL;
}


inline SplBool
splIsAllEquals(void *sendbuf, int count, SPL_Datatype datatype, SPL_Comm comm)
{
  if( !splIsParallel() ) {
    return SplTrue;
  }

  int dsize = spl_get_datasize(datatype)*count;
  void* result = malloc(dsize);
  int flg = 0, tmp = 0;

  SPL_Allreduce(sendbuf, result, count, datatype, SPL_BAND, comm);
  if( memcmp(sendbuf, result, dsize) != 0) {
    flg = 1;
  }
  free(result);

  // error check
  tmp = 0;
  SPL_Allreduce(&flg, &tmp, 1, SPL_INT, SPL_SUM, comm);
  if( tmp != 0 ) {
    return SplFalse;
  }

  return SplTrue;
}


/**
 * Gather SVFileList
 *
 * @param[in,out]   file_list       master filelist and create filelist
 * @param[in]      readRank         Rank that reads a file (It is effective at mode 1 or 2.)
 * @param[in]      grp              Communication group
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splBcastSvFileList( splSvFileList* file_list,
                  int readRank,
                  SPL_Comm grp )
{
  register long i;
  register int j;

  if( !splIsParallel() ) {
    return SplTrue;
  }

  int myrank;
  SPL_Comm_rank(grp, &myrank);

  /* parameter check */
  if( !file_list ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "bcastSvFileList arguments is null.");
    return SplFalse;
  }

  if (myrank != readRank) {
    splFinalizeSvFileList(file_list);
  }

  SPL_Datatype dt_size_t = SPL_DATATYPE_NULL;
  if( sizeof(size_t) == 4 )      dt_size_t = SPL_INT;
  else if( sizeof(size_t) == 8 ) dt_size_t = SPL_LONG_LONG;
  else {
    // error
    SPL_ERROR_INFO(SPL_ERR_INTERNAL_IMPOSSIBLE, "can not get sizeof size_t.");
    return SplFalse;
  }
  size_t list_size = (file_list != NULL)?file_list->list_size:0;
  SPL_Bcast(&list_size, 1, dt_size_t, readRank, grp);
  if (list_size == 0) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "invalid list size, list_size is zero.");
    return SplFalse;
  }

  int* set_flag_list = (int*)malloc(sizeof(int)*list_size);
  size_t* start_idx_list= (size_t*)malloc(sizeof(size_t)*list_size*3);
  int fname_size = sizeof(file_list->top->fname);
  char* fname_list= (char*)malloc(sizeof(char)*list_size*fname_size);

  if (myrank == readRank) {
    splSvFile* next = file_list->top;
    for (i=0; i<list_size; i++) {
      *(set_flag_list + i) = next->set_flag;
      memcpy(start_idx_list + i*3, next->start_idx, sizeof(size_t)*3);
      memcpy(fname_list + i*fname_size, next->fname, fname_size);

      next = next->next;
    }
  }

  // Bcast
  SPL_Bcast(set_flag_list, list_size, SPL_INT, readRank, grp);
  SPL_Bcast(start_idx_list, list_size*3, dt_size_t, readRank, grp);
  SPL_Bcast(fname_list, list_size*fname_size, SPL_CHAR, readRank, grp);

  int set_flag;
  size_t start_idx[3];
  char* fname;
  if (myrank != readRank) {
    for (i=list_size-1; i>=0; i--) {
      set_flag = set_flag_list[i];
      for (j=0; j<3; j++) start_idx[j] = start_idx_list[i*3+j];
      fname = fname_list+fname_size*i;

      if (set_flag == 1) {
        splAddSvFile(file_list, fname, start_idx);
      }
      else {
        splAddSvFileName(file_list, fname);
      }
    }
  }

  free(set_flag_list);
  free(start_idx_list);
  free(fname_list);

  return SplTrue;

}


/**
 * \~english
 * Create Process Group
 * @param      idNum      node list count
 * @param      idList    node list
 * @param      newComm    new MPI_Comm
 * @param      newGrp    new MPI_Group
 * @return    success(=SplTrue)/fault(=SplFalse)
 *
 * \~japanese
 * ノード番号リストからプロセスグループを作成する。
 * @param      idNum      作成ノード数
 * @param      idList    作成ノードリスト
 * @param      newComm    新規作成コミュニケータ
 * @param      newGrp    新規作成グループ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splCreateCommGroup(int idNum, int* idList, SPL_Comm *newComm, SPL_Group *newGrp )
{
  if( !splIsParallel() ) {
    *newComm = 0;
    *newGrp = 0;
    return SplTrue;
  }
  int err_no = 0;

  if( (idNum < 1) || !idList ) {
    // error
    SPL_ERROR_INFO(SPL_ERR_INVALID_PARAMETER, "id list is empty");
    return SplFalse;
  }

  SPL_Comm oldComm = SPL_COMM_WORLD;
  *newComm == SPL_COMM_NULL;

  // create MPI_COMM
  SPL_Group oldGrp;

  if ( (err_no = SPL_Comm_group(SPL_COMM_WORLD, &oldGrp)) != SPL_MPI_SUCCESS) {
    // error
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Comm_group fault.[error_no=%d]", err_no);
    return SplFalse;
  }
  if ( (err_no = SPL_Group_incl(oldGrp, idNum, idList, newGrp)) != SPL_MPI_SUCCESS) {
    // error
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Group_incl fault.[error_no=%d]", err_no);
    return SplFalse;
  }
  if ( (err_no = SPL_Comm_create(SPL_COMM_WORLD, *newGrp, newComm)) != SPL_MPI_SUCCESS) {
    // error
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Comm_create fault.[error_no=%d]", err_no);
    return SplFalse;
  }

  SPL_Group_free(&oldGrp);

  return SplTrue;

}

/**
 * ２つのグループを結合してプロセスグループを作成する。
 * @param      oldGroup1      グループ１
 * @param      oldGroup2      グループ２
 * @param      newComm       新規作成コミュニケータ
 * @param      newGrp        新規作成グループ
 * @return    success(=SplTrue)/fault(=SplFalse)
 */
SplBool
splCreateCommGroupUnion(SPL_Group oldGroup1, SPL_Group oldGroup2, SPL_Comm *newComm, SPL_Group *newGrp )
{
  if( !splIsParallel() ) {
    *newComm = 0;
    *newGrp = 0;
    return SplTrue;
  }
  int err_no = 0;

  SPL_Comm oldComm = SPL_COMM_WORLD;
  *newComm == SPL_COMM_NULL;

  // create MPI_COMM
  if ( (err_no = SPL_Group_union(oldGroup1, oldGroup2, newGrp)) != SPL_MPI_SUCCESS) {
    // error
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Group_union fault.[error_no=%d]", err_no);
    return SplFalse;
  }

  if ( (err_no = SPL_Comm_create(oldComm, *newGrp, newComm)) != SPL_MPI_SUCCESS) {
    // error
    SPL_ERROR_INFO(SPL_ERR_MPI_ERROR, "SPL_Comm_create fault.[error_no=%d]", err_no);
    return SplFalse;
  }

  return SplTrue;

}



/**
 * \~english
 * get data index
 * @param        datatype          datatype
 * @return    data index
 *
 * \~japanese
 * データタイプのMPICHの型番号を取得する。.
 * データタイプを送受信する場合にOPENMPIは構造体の為に送受信できないので、データタイプの一意の番号を取得する。
 * @param        datatype          データタイプ
 * @return       データタイプ番号
 */
int spl_get_dataindex_bydatatype(SPL_Datatype datatype)
{
  int type = 0;
  if( datatype == SPL_CHAR )                { type = 1; } // SPH_CHAR
  else if( datatype == SPL_UNSIGNED_CHAR )  { type = 2; } // SPH_UNSIGNED_CHAR
  else if( datatype == SPL_SHORT )          { type = 4; } // SPH_SHORT
  else if( datatype == SPL_UNSIGNED_SHORT ) { type = 5; } // SPH_UNSIGNED_SHORT
  else if( datatype == SPL_INT )            { type = 6; } // SPH_INTEGER
  else if( datatype == SPL_UNSIGNED )       { type = 7; } // SPH_UNSIGNED
  else if( datatype == SPL_LONG )           { type = 8; } // SPH_LONG
  else if( datatype == SPL_UNSIGNED_LONG )  { type = 9; } // SPH_UNSIGNED_LONG
  else if( datatype == SPL_LONG_LONG )      { type = 13;} // SPH_LONG_LONG
  else if( datatype == SPL_LONG_LONG_INT )  { type = 13;} // SPH_LONG_LONG
  else if( datatype == SPL_FLOAT )          { type = 10;} // SPH_REAL4 or SPH_FLOAT
  else if( datatype == SPL_DOUBLE )         { type = 11;} // SPH_REAL8 or SPH_DOUBLE
  else if( datatype == SPL_REAL )         { type = 26;} // Fortran types real
  else { type = 0; }

  return type;
}


/**
 * \~english
 * get data index
 * @param        datatype          datatype
 * @return    data index
 *
 * \~japanese
 * データ番号からデータタイプを取得する。.
 * @param        dataindex          データ番号
 * @return       データタイプ番号
 */
SPL_Datatype spl_get_datatype_byindex(int dataindex)
{
  SPL_Datatype datatype = SPL_DATATYPE_NULL;
  if      ( dataindex == 1  /* SPH_CHAR                */ ) { datatype = SPL_CHAR; }
  else if ( dataindex == 2  /* SPH_UNSIGNED_CHAR       */ ) { datatype = SPL_UNSIGNED_CHAR; }
  else if ( dataindex == 4  /* SPH_SHORT               */ ) { datatype = SPL_SHORT; }
  else if ( dataindex == 5  /* SPH_UNSIGNED_SHORT      */ ) { datatype = SPL_UNSIGNED_SHORT; }
  else if ( dataindex == 6  /* SPH_INTEGER             */ ) { datatype = SPL_INT; }
  else if ( dataindex == 7  /* SPH_UNSIGNED            */ ) { datatype = SPL_UNSIGNED; }
  else if ( dataindex == 8  /* SPH_LONG                */ ) { datatype = SPL_LONG; }
  else if ( dataindex == 9  /* SPH_UNSIGNED_LONG       */ ) { datatype = SPL_UNSIGNED_LONG; }
  else if ( dataindex == 13 /* SPH_LONG_LONG           */ ) { datatype = SPL_LONG_LONG_INT; }
  else if ( dataindex == 10 /* SPH_REAL4 or SPH_FLOAT  */ ) { datatype = SPL_FLOAT;}
  else if ( dataindex == 11 /* SPH_REAL8 or SPH_DOUBLE */ ) { datatype = SPL_DOUBLE;}
  else if ( dataindex == 26 /* Fortran types real */      ) { datatype = SPL_REAL;}
  else { datatype = SPL_DATATYPE_NULL; }

  return datatype;
}


#if defined(__cplusplus)
}
#endif /* extern "C" */

