/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2008
 *
 */

#ifndef _SKL_VOXEL_DATA_SET_H_
#define _SKL_VOXEL_DATA_SET_H_

#include <stdio.h> 
#include <string.h>
#include <stdlib.h>
#include <iostream>
#include <vector>
// modify(Vsphere 1.6.0) start for vsphere_win at 2008/06/20 by @hira
#ifndef _WIN32
#include <unistd.h> // for debug         // for linux
#else
#include "sph_win32_util.h"   // for win32
#endif
// modify(Vsphere 1.6.0) end for vsphere_win at 2008/06/20 by @hira
#include <stdarg.h> // for debug

class SklVoxDataSet {
friend class SklVoxDataCntl;
protected:
  // for debug
  void SklErrMsg(char* format, ...){
    if( !format ) return;
    char buf[1024];
    va_list vl;
    va_start(vl, format);
    vsprintf(buf, format, vl);
    va_end(vl);
    fprintf(stderr, "%s", buf);
    fprintf(stderr, "\n");
    fflush(stderr);
    sleep(1);
    return;
  }
public:
  typedef enum { UNKNOWN_FMT=0,
                 SVX_FMT     = 1,
                 FDV_FMT     = 2,
                 SPH_FMT     = 3,
                 NBF_FMT     = 4,
                 P3DXYZ_FMT  = 5,
                 P3DFUNC_FMT = 6,
                 P3DQ_FMT    = 7,
                 SBX_FMT     = 8} FmtType;
  typedef enum { UNKNOWN_SVTYPE=0, TYPE_SCALAR=1, TYPE_VECTOR=2} SVType;
  typedef enum { UNKNOWN_DATATYPE=0, TYPE_SINGLE=1, TYPE_DOUBLE=2} DataType;
  typedef enum { SVX_VOLUME = 0,
                 SVX_AREA_X = 1,
                 SVX_AREA_Y = 2,
                 SVX_AREA_Z = 3,
                 SVX_MID    = 4,
                 SVX_BCC    = 5,
                 SVX_BCF_X  = 6,
                 SVX_BCF_Y  = 7,
                 SVX_BCF_Z  = 8,
                 FDV_GRID   = 0,
                 FDV_SCALAR = 1,
                 FDV_VECTOR = 2,
                 SPH_DATA   = 0,
                 NBF_DATA   = 0,
                 SBX_DATA   = 0} DataKind;

  typedef enum { UNKNOWN_CRDDEFTYPE=0,
          TYPE_REGULAR=1,
          TYPE_COLLOCATE=2,
          TYPE_STAGGERED1 = 3,
          TYPE_STAGGERED2 = 4 } CoordDefType;
  typedef enum { UNKNOWN_VSZTYPE=0, TYPE_CHAR = 1, TYPE_SHORT = 2, TYPE_INT = 4 } VszType;

protected:

  std::string m_fname;

  FmtType m_fmt;
  SVType m_svType;
  DataType m_dType;

  // for Single precision
  unsigned int m_sz[3];
  float m_org[3];
  float m_pit[3];
  int m_step;
  float m_time;
  float *m_float_data;

  // for Double precision
  unsigned long long m_lsz[3];
  double m_dorg[3];
  double m_dpit[3];
  long long m_lstep;
  double m_dtime;
  double *m_double_data;

  unsigned m_DATANUM;

  bool SetSVType(SVType type);
  bool SetDataType(DataType type);
  bool SetSize(unsigned ix, unsigned jx, unsigned kx);
  bool SetSize(unsigned long long ix,
               unsigned long long jx,
               unsigned long long kx);
  bool SetOrigin(float ox, float oy, float oz);
  bool SetOrigin(double ox, double oy, double oz);
  bool SetPitch(float dx, float dy, float dz);
  bool SetPitch(double dx, double dy, double dz);
  bool SetStep(int step);
  bool SetStep(long long step);
  bool SetTime(float time);
  bool SetTime(double time);

  bool SetFileName(const char* fname);

public:
  SklVoxDataSet();
  virtual ~SklVoxDataSet();

  virtual int GetFormat();
  virtual int GetDataType();
  virtual int GetSVType();

  virtual bool GetSize(unsigned& ix, unsigned& jx, unsigned& kx) const;
  virtual bool GetSize(unsigned long long& ix,
                       unsigned long long& jx,
                       unsigned long long& kx) const;
  virtual bool GetOrigin(float& ox, float& oy, float& oz) const;
  virtual bool GetOrigin(double& ox, double& oy, double& oz) const;
  virtual bool GetPitch(float& dx, float& dy, float& dz) const;
  virtual bool GetPitch(double& dx, double& dy, double& dz) const;
  virtual bool GetStep(int& step) const;
  virtual bool GetStep(long long& step) const;
  virtual bool GetTime(float& time) const;
  virtual bool GetTime(double& time) const;

  virtual bool Read(const char* fname,
                    const unsigned* size,
                    unsigned vc = 0,
                    bool multi_io_flag = false,
                    int procGrp = 0) = 0;
  virtual bool Write(const char* fname,
                     unsigned gc = 0,
                     bool multi_io_flag = false,
                     int procGrp = 0) = 0;

  const char* GetFileName() const;
  float* GetSingleData() const;
  double* GetDoubleData() const;
};

inline SklVoxDataSet::SklVoxDataSet()
{
  m_fname.clear();

  m_fmt = UNKNOWN_FMT;
  m_svType = UNKNOWN_SVTYPE;
  m_dType = UNKNOWN_DATATYPE;

  memset(m_sz, 0, sizeof(unsigned)*3);
  memset(m_org, 0, sizeof(float)*3);
  memset(m_pit, 0, sizeof(float)*3);
  m_step = 0;
  m_time = 0.0f;
  m_float_data = NULL;

  memset(m_lsz, 0, sizeof(long long)*3);
  memset(m_dorg, 0, sizeof(double)*3);
  memset(m_dpit, 0, sizeof(double)*3);
  m_lstep = 0;
  m_dtime = 0.0;
  m_double_data = NULL;

  m_DATANUM = 0;

}

inline SklVoxDataSet::~SklVoxDataSet()
{
  if (m_float_data) delete[] m_float_data;
  if (m_double_data) delete[] m_double_data;
  m_fname.clear();
}

inline bool SklVoxDataSet::SetSVType(SVType type)
{
  m_svType = type;
  return true;
}

inline bool SklVoxDataSet::SetDataType(DataType type)
{
  m_dType = type;
  return true;
}

inline bool SklVoxDataSet::SetSize(unsigned ix, unsigned jx, unsigned kx)
{
  m_sz[0] = ix; m_sz[1] = jx; m_sz[2] = kx;
  m_lsz[0] = (unsigned long long)ix;
  m_lsz[1] = (unsigned long long)jx;
  m_lsz[2] = (unsigned long long)kx;
  return true;
}

inline bool SklVoxDataSet::SetSize(unsigned long long ix,
                                   unsigned long long jx,
                                   unsigned long long kx)
{
  m_lsz[0] = ix; m_lsz[1] = jx; m_lsz[2] = kx;
  m_sz[0] = (unsigned)ix; m_sz[1] = (unsigned)jx; m_sz[2] = (unsigned)kx;
  return true;
}

inline bool SklVoxDataSet::SetOrigin(float ox, float oy, float oz)
{
  m_org[0] = ox; m_org[1] = oy; m_org[2] = oz;
  m_dorg[0] = (double)ox; m_dorg[1] = (double)oy; m_dorg[2] = (double)oz;
  return true;
}

inline bool SklVoxDataSet::SetOrigin(double ox, double oy, double oz)
{
  m_dorg[0] = ox; m_dorg[1] = oy; m_dorg[2] = oz;
  m_org[0] = (float)ox; m_org[1] = (float)oy; m_org[2] = (float)oz;
  return true;
}

inline bool SklVoxDataSet::SetPitch(float dx, float dy, float dz)
{
  m_pit[0] = dx; m_pit[1] = dy; m_pit[2] = dz;
  m_dpit[0] = (double)dx; m_dpit[1] = (double)dy; m_dpit[2] = (double)dz;
  return true;
}

inline bool SklVoxDataSet::SetPitch(double dx, double dy, double dz)
{
  m_dpit[0] = dx; m_dpit[1] = dy; m_dpit[2] = dz;
  m_pit[0] = (float)dx; m_pit[1] = (float)dy; m_pit[2] = (float)dz;
  return true;
}

inline bool SklVoxDataSet::SetStep(int step)
{
  m_step = step;
  m_lstep = (long long)step;
  return true;
}

inline bool SklVoxDataSet::SetStep(long long step)
{
  m_lstep = step;
  m_step = (int)step;
  return true;
}

inline bool SklVoxDataSet::SetTime(float time)
{
  m_time = time;
  m_dtime = (double)time;
  return true;
}

inline bool SklVoxDataSet::SetTime(double time)
{
  m_dtime = time;
  m_time = (float)time;
  return true;
}

inline int SklVoxDataSet::GetFormat()
{
  return m_fmt;
}

inline int SklVoxDataSet::GetDataType()
{
  return m_dType;
}

inline int SklVoxDataSet::GetSVType()
{
  return m_svType;
}

inline bool SklVoxDataSet::GetSize(unsigned& ix,
                                   unsigned& jx,
                                   unsigned& kx) const
{
  if( m_dType != TYPE_SINGLE ) {
    ix = (unsigned)m_lsz[0];
    jx = (unsigned)m_lsz[1];
    kx = (unsigned)m_lsz[2];
  }
  else {
    ix = m_sz[0]; jx = m_sz[1]; kx = m_sz[2];
  }
  return true;
}

inline bool SklVoxDataSet::GetSize(unsigned long long& ix,
                                   unsigned long long& jx,
                                   unsigned long long& kx) const
{
  if( m_dType != TYPE_DOUBLE ) return false;
  ix = m_lsz[0]; jx = m_lsz[1]; kx = m_lsz[2];
  return true;
}

inline bool SklVoxDataSet::GetOrigin(float& ox, float& oy, float& oz) const
{
  if( m_dType != TYPE_SINGLE ) return false;
  ox = m_org[0]; oy = m_org[1]; oz = m_org[2];
  return true;
}

inline bool SklVoxDataSet::GetOrigin(double& ox, double& oy, double& oz) const
{
  if( m_dType != TYPE_DOUBLE ) return false;
  ox = m_dorg[0]; oy = m_dorg[1]; oz = m_dorg[2];
  return true;
}

inline bool SklVoxDataSet::GetPitch(float& dx, float& dy, float& dz) const
{
  if( m_dType != TYPE_SINGLE ) return false;
  dx = m_pit[0]; dy = m_pit[1]; dz = m_pit[2];
  return true;
}

inline bool SklVoxDataSet::GetPitch(double& dx, double& dy, double& dz) const
{
  if( m_dType != TYPE_DOUBLE ) return false;
  dx = m_dpit[0]; dy = m_dpit[1]; dz = m_dpit[2];
  return true;
}

inline bool SklVoxDataSet::GetStep(int& step) const
{
  if( m_dType != TYPE_SINGLE ) {
    step = (int)m_lstep;
  }
  else {
    step = m_step;
  }
  return true;
}

inline bool SklVoxDataSet::GetStep(long long& step) const
{
  if( m_dType != TYPE_DOUBLE ) return false;
  step = m_lstep;
  return true;
}

inline bool SklVoxDataSet::GetTime(float& time) const
{
  if( m_dType != TYPE_SINGLE ) return false;
  time = m_time;
  return true;
}

inline bool SklVoxDataSet::GetTime(double& time) const
{
  if( m_dType != TYPE_DOUBLE ) return false;
  time = m_dtime;
  return true;
}


inline bool SklVoxDataSet::SetFileName(const char* fname)
{
  if( !fname ) return false;
  m_fname = fname;
  return true;
}

inline const char* SklVoxDataSet::GetFileName() const
{
  if( m_fname.size() == 0 ) return NULL;
  return m_fname.c_str();
}


inline float* SklVoxDataSet::GetSingleData() const
{
  return m_float_data;
}

inline double* SklVoxDataSet::GetDoubleData() const
{
  return m_double_data;
}

#endif // _SKL_VOXEL_DATA_SET_H_

