//============================================================================
// Name        : spxTools.C
// Author      :
// Version     :
// Copyright   :
// Description :
//============================================================================

#include <string.h>
#include <stdlib.h>
#include <iostream>

#include "SklSphDataSet.h"
#include "spl/splmpi.h"
#include "spl/splfio.h"

using namespace std;

/// sph2spx version
const char version[] = "Version:1.0.0 ";


/**
 * Usage
 *
 */
void usage() {
  fprintf(stdout, "\n");
  fprintf(stdout, " Usage : sph2spx [-gc gc-value -crd crddef -cmp block-size -o spx-file] sph-file \n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  Convert SPH file to SPX file.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  [Parameter]\n");
  fprintf(stdout, "    -o,--output spx-file       output SPX file name. (defualt:output.spx)\n");
  fprintf(stdout, "    sph-file                   input SPH file name.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  [Options]\n");
  fprintf(stdout, "    -gc gc-value               gc value. (defualt=0)\n");
  fprintf(stdout, "    -crd crddef                Origin Coords Define value. (defualt=1:Regular)\n");
  fprintf(stdout, "                                  1:Regular\n");
  fprintf(stdout, "                                  2:Collocate\n");
  fprintf(stdout, "                                  3:Staggered(1)\n");
  fprintf(stdout, "                                  4:Staggered(2)\n");
  fprintf(stdout, "    -cmp block-size            compress block size. (defualt=0:uncompress)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  [Others]\n");
  fprintf(stdout, "    -i,--info file-name        print SPX/SPH information.\n");
  fprintf(stdout, "    -v,--version               print version.\n");
  fprintf(stdout, "    -h,-help                   print this.\n");
  fprintf(stdout, "\n");
}


/**
 * convert SPH to SPX
 * @param     fname        output SPX File Name
 * @param     sphData      input SPH Data
 * @param     gc          guide cell value
 * @param     crddef      coords define value
 * @param     block_size    compress block size
 * @return    bool        success(=true)/failed(=false)
 */
bool convertSpx(const char* fname, SklSphDataSet* sphData, int gc, int crddef, int block_size)
{

  splSbx obj;
  if( splInitSbx(&obj) == SplFalse ) {
    return false;
  }

  // write spx file
  splSpx spxObj;
  if( splInitSpx(&spxObj) == SplFalse ) {
    fprintf(stderr, "splInitSpx failed.\n");
    return false;
  }


  // 次元数
  spxObj.dims = 3;
  unsigned int sz[3] = {0, 0, 0};
  sphData->GetSize(sz[0], sz[1], sz[2]);
  if (sz[2] == 0 || sz[2] == 1) {
    spxObj.dims = 2;
  }
  if (sz[1] == 0 || sz[1] == 1) {
    spxObj.dims = 1;
  }

  // ベクトル長、データ種別
  spxObj.vlen = 1;
  spxObj.dtype = 1;
  if( sphData->GetSVType() == SklVoxDataSet::TYPE_SCALAR ) {
    spxObj.vlen = 1;
    spxObj.dtype = 1;
  }
  else if( sphData->GetSVType() == SklVoxDataSet::TYPE_VECTOR ) {
    spxObj.vlen = 3;
    spxObj.dtype = 3;
  }

  // ガイドセルの数
  spxObj.gc = gc;

  // 実数データのサイズ
  spxObj.rlen = 4;
  if( sphData->GetDataType() == SklVoxDataSet::TYPE_SINGLE ) {
    spxObj.rlen = 4;
  }
  else if( sphData->GetDataType() == SklVoxDataSet::TYPE_DOUBLE ) {
    spxObj.rlen = 8;
  }

  // 原点座標の定義位置
  spxObj.crddef = crddef;

  // ボクセルサイズ
  spxObj.size[0] = sz[0];
  spxObj.size[1] = sz[1];
  spxObj.size[2] = sz[2];

  // 原点座標レコード
  if( sphData->GetDataType() == SklVoxDataSet::TYPE_SINGLE ) {
    float origin[3];
    sphData->GetOrigin(origin[0], origin[1], origin[2]);
    spxObj.origin[0] = origin[0];
    spxObj.origin[1] = origin[1];
    spxObj.origin[2] = origin[2];
  }
  else if( sphData->GetDataType() == SklVoxDataSet::TYPE_DOUBLE ) {
    double origin[3];
    sphData->GetOrigin(origin[0], origin[1], origin[2]);
    spxObj.origin[0] = origin[0];
    spxObj.origin[1] = origin[1];
    spxObj.origin[2] = origin[2];
  }

  // ボクセルピッチレコード
  if( sphData->GetDataType() == SklVoxDataSet::TYPE_SINGLE ) {
    float pitch[3];
    sphData->GetPitch(pitch[0], pitch[1], pitch[2]);
    spxObj.pitch[0] = pitch[0];
    spxObj.pitch[1] = pitch[1];
    spxObj.pitch[2] = pitch[2];
  }
  else if( sphData->GetDataType() == SklVoxDataSet::TYPE_DOUBLE ) {
    double pitch[3];
    sphData->GetPitch(pitch[0], pitch[1], pitch[2]);
    spxObj.pitch[0] = pitch[0];
    spxObj.pitch[1] = pitch[1];
    spxObj.pitch[2] = pitch[2];
  }

  // 時刻レコード
  if( sphData->GetDataType() == SklVoxDataSet::TYPE_SINGLE ) {
    int step;
    float time;
    sphData->GetStep(step);
    sphData->GetTime(time);
    spxObj.step = step;
    spxObj.time = time;
  }
  else if ( sphData->GetDataType() == SklVoxDataSet::TYPE_DOUBLE ) {
    long long step;
    double time;
    sphData->GetStep(step);
    sphData->GetTime(time);
    spxObj.step = step;
    spxObj.time = time;
  }

  // データ
  if( sphData->GetDataType() == SklVoxDataSet::TYPE_SINGLE ) {
    spxObj.data.f = sphData->GetSingleData();
  }
  else if ( sphData->GetDataType() == SklVoxDataSet::TYPE_DOUBLE ) {
    spxObj.data.d = sphData->GetDoubleData();
  }

  int mode = 0;
  int readRank = 0;
  int grp = SPL_COMM_WORLD;
  SplBool compFlg = SplFalse;
  if (block_size != 0) {
    compFlg = SplTrue;
  }
  else {
    block_size = spxObj.size[0]*spxObj.size[1];
  }
  if( splWriteSpx(fname, 0, compFlg, SplFalse, block_size,
                  &spxObj, readRank, grp) == SplFalse ) {
    fprintf(stderr, "splWriteSbx failed.\n");
    return false;
  }

  if( splFinalizeSpx(&spxObj, SplFalse) == SplFalse ) {
    fprintf(stderr, "splFinalizeSpx failed.\n");
  }

  return true;
}


/**
 * print SPX File Information
 * @param     fname      SPX File Name
 * @return    std::string      Information
 */
std::string printSpx(const char *fname)
{

  splSpx spxObj;
  if( splInitSpx(&spxObj) == SplFalse ) {
    fprintf(stderr, "splInitSbx failed.\n");
    return "";
  }

  int mode = 0;
  int readRank = 0;
  int grp = SPL_COMM_WORLD;
  if( splGetSpxInfo(fname, &spxObj, mode, readRank, grp) == SplFalse ) {
    fprintf(stderr, "splGetSpxInfo failed.\n");
    return "";
  }

  std::string msg;
  char value[1024];

  sprintf(value, "File Name            : %s", fname);
  msg += value;
  msg += "\n";

  // 次元数
  sprintf(value, "Dimension            : %d", spxObj.dims);
  msg += value;
  msg += "\n";

  // ベクトル長
  sprintf(value, "Vector Length        : %d", spxObj.vlen);
  msg += value;
  msg += "\n";

  // データ種別
  if( spxObj.dtype == 1 ) {
    msg += "Data Type            : Scalar Type";
  }
  else if( spxObj.dtype == 3 ) {
    msg += "Data Type            : Vector Type";
  }
  else {
    msg += "Data Type            : Unknown Type";
  }
  msg += "\n";

  // ガイドセルの数
  sprintf(value, "Guide Cell           : %ld", spxObj.gc);
  msg += value;
  msg += "\n";

  // 実数データのサイズ
  if( spxObj.rlen == 4 ) {
    msg += "Real Data Size       : float";
  }
  else if( spxObj.rlen == 8 ) {
    msg += "Real Data Size       : double";
  }
  else {
    msg += "Real Data Size       : Unknown";
  }
  msg += "\n";

  // 原点座標の定義位置
  if( spxObj.crddef == 1 ) {
    msg += "Origin Coords Define : Regular (=1)";
  }
  else if( spxObj.crddef == 2 ) {
    msg += "Origin Coords Define : Collocate (=2)";
  }
  else if( spxObj.crddef == 3 ) {
    msg += "Origin Coords Define : Staggered(1) (=3)";
  }
  else if( spxObj.crddef == 4 ) {
    msg += "Origin Coords Define : Staggered(2) (=4)";
  }
  else {
    msg += "Origin Coords Define : Unknown";
  }
  msg += "\n";

  // ボクセルサイズ
  sprintf(value, "Voxel Size           : %ld, %ld, %ld", spxObj.size[0], spxObj.size[1], spxObj.size[2]);
  msg += value;
  msg += "\n";

  // 原点座標レコード
  if( spxObj.rlen == 4 ) {
    sprintf(value, "Origin Coords        : %f, %f, %f", (float)spxObj.origin[0], (float)spxObj.origin[1], (float)spxObj.origin[2]);
    msg += value;
  }
  else if( spxObj.rlen == 8 ) {
    sprintf(value, "Origin Coords        : %lf, %lf, %lf", spxObj.origin[0], spxObj.origin[1], spxObj.origin[2]);
    msg += value;
  }
  msg += "\n";

  // ボクセルピッチレコード
  if( spxObj.rlen == 4 ) {
    sprintf(value, "Voxel Pitch          : %f, %f, %f", (float)spxObj.pitch[0], (float)spxObj.pitch[1], (float)spxObj.pitch[2]);
    msg += value;
  }
  else if( spxObj.rlen == 8 ) {
    sprintf(value, "Voxel Pitch          : %lf, %lf, %lf", spxObj.pitch[0], spxObj.pitch[1], spxObj.pitch[2]);
    msg += value;
  }
  msg += "\n";

  // 時刻レコード
  sprintf(value, "Time Step Count      : %ld", spxObj.step);
  msg += value;
  msg += "\n";
  if( spxObj.rlen == 4 ) {
    sprintf(value, "Time                 : %f", spxObj.time);
    msg += value;
  }
  else if( spxObj.rlen == 8 ) {
    sprintf(value, "Time                 : %lf", spxObj.time);
    msg += value;
  }
  msg += "\n";

  if( splFinalizeSpx(&spxObj, SplFalse) == SplFalse ) {
    fprintf(stderr, "splFinalizeSbx failed.\n");
  }
  return msg;
}


/**
 * main
 *
 */
int main(int argc, char** argv) {

  splSbx obj;
  size_t blksize = 0;
  size_t i;

  if( argc <= 1 ) {
    usage();
    return -1;
  }

  if (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "--help") == 0) {
    usage();
    return EXIT_SUCCESS;
  }
  else if (strcmp(argv[1], "-v") == 0 || strcmp(argv[1], "--version") == 0) {
    cout << version << endl;
    return EXIT_SUCCESS;
  }
  else if (strcmp(argv[1], "-i") == 0 || strcmp(argv[1], "--info") == 0) {
    if( argc < 3 ) {
      usage();
      return -1;
    }
    if (strcmp(argv[2]+strlen(argv[2])-3, "sph") == 0) {
      SklSphDataSet* sph = new SklSphDataSet();
      sph->Read(argv[2]);
      std::string msg = sph->printInformation();
      cout << msg << endl;

      delete sph;
    }
    else if (strcmp(argv[2]+strlen(argv[2])-3, "spx") == 0) {
      std::string msg = printSpx(argv[2]);
      cout << msg << endl;
    }
    else {
      usage();
      return -1;
    }
  }
  else {
    char *sphName = NULL;
    char *spxName = "output.spx";
    int gc = 0;
    int blk_size = 0;
    int crddef = 1;
    for (int i=1; i<argc;i++) {
      if (strcmp(argv[i], "-gc") == 0) {
        gc = atoi(argv[i+1]);
      }
      else if (strcmp(argv[i], "-crd") == 0) {
        crddef = atoi(argv[i+1]);
      }
      else if (strcmp(argv[i], "-cmp") == 0) {
        blk_size = atoi(argv[i+1]);
      }
      else if (strcmp(argv[i], "-o") == 0 || strcmp(argv[i], "--output") == 0) {
        spxName = argv[i+1];
      }
    }
    sphName = argv[argc-1];

    cout << "convert : " << sphName << "  >>>  " << spxName << endl;
    // read sph file
    SklSphDataSet* sph = new SklSphDataSet();
    if (!sph->Read(sphName)) {
      fprintf(stderr, "convert failed [SPH read error].\n");
      return -1;
    }

    if (!convertSpx(spxName, sph, gc, crddef, blk_size)) {
      fprintf(stderr, "convert failed [SPX write error].\n");
      return -1;
    }
    cout << "convert success!" << endl;
  }


  return EXIT_SUCCESS;
}
